--[[--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author tdonahue 2013-07-28

------------------------------------------------------------------------------]]

local KSClass = require 'KSClass'

local KSJsonEncoder = require 'KSJsonEncoder'

local KSJsonDecoder = require 'KSJsonDecoder'

local WFCloseSupport = require 'WFCloseSupport'

local WFArguments = require "WFArguments"

local REAppLuaInterface = require 'REAppLuaInterface'

local GroupPixServerUrls = require 'GroupPixServerUrls'

--------------------------------------------------------------------------------

local GroupPixPin = 'GroupPixPin'

--------------------------------------------------------------------------------

local methods = { }

--------------------------------------------------------------------------------

function methods:__init( )
	
	debug.print( "GroupPixPin init" )
	
end

--------------------------------------------------------------------------------

function methods:getPinFormat( --[[OzUser]] ozUser )

	WFArguments.check( GroupPixPin, 'OzUser' )
	
	local options = {
	    requestHeaders = {
			{ header = 'Connection', value = 'close' },
	    },
	}

	-- GET
	-- /v1/pins/format
	local pinUrl = string.format( '%s/format', GroupPixServerUrls.pinsUrl )

	local resultP = ozUser:getP( pinUrl, options )
		:mapDataP( function( status, data )
			--debug.print( data )
			local pinFormat = KSJsonDecoder.decodeWebSafeJson( data )
			
			local result = {}
			result[ "pinLength" ] = pinFormat.length
			result[ "pinRegex" ] = pinFormat.regex
			result[ "pinInitialState" ] = pinFormat.initialState
			
	        REAppLuaInterface.performResultCallback( "updatePinFormat", result )

			return 'data', pinFormat
		end )

	resultP:runUntilDone()

end

--------------------------------------------------------------------------------

function methods:createPinForCatalog( --[[OzUser]] ozUser, --[[OzCatalogService]] catalog )
	
	WFArguments.check( GroupPixPin, 'OzUser', 'OzCatalogService' )	

	-- POST
	-- /v1/pins/
	local pinUrl = string.format( '%s', GroupPixServerUrls.pinsUrl )

	local content = KSJsonEncoder.encodeOzUploadJson { catalogId = catalog._catalogId, userId = ozUser:id() } 

	local postArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Connection', value = 'close' },
		},
		requestMethod = 'POST',
	}
	
	local resultP = ozUser:putP( pinUrl, content, postArgs )
		:mapDataP( function( status, data, headers )
			
			debug.print( data )
			local pin = KSJsonDecoder.decodeWebSafeJson( data )
			
			-- setting pin was successful, set the pin value on the catalog payload as well
			return self:setCatalogPinPayload( ozUser, catalog, pin )
			
		end )

	resultP:runUntilDone()
end

--------------------------------------------------------------------------------

function methods:getPin( --[[OzUser]] ozUser, --[[string]] pinCode )

	WFArguments.check( GroupPixPin, 'OzUser', 'string' )	

	local options = {
	    requestHeaders = {
			{ header = 'Connection', value = 'close' },
	    },
	}

	-- GET
	-- /v1/pins/{pinCode}
	local pinUrl = string.format( '%s/%s', GroupPixServerUrls.pinsUrl, pinCode )

	local resultP = ozUser:getP( pinUrl, options )
		:mapDataP( function( status, data )
			--debug.print( data )
			local pin = KSJsonDecoder.decodeWebSafeJson( data )
			
			local result = {}
			result[ "eventId" ] = pin.catalogId
			result[ "pinCode" ] = pin.pinCode
			result[ "pinState" ] = pin.pinState
			result[ "pinExpires" ] = pin.pinExpires
			result[ "inviteShortUrl" ] = pin.inviteShortUrl
			
	        REAppLuaInterface.performResultCallback( "getPinCode", result )

			return 'data', pin
		end )

	resultP:runUntilDone()

end

--------------------------------------------------------------------------------

function methods:addUserByPin( --[[OzUser]] ozUser, --[[string]] pinCode, --[[string]] userId )

	WFArguments.check( GroupPixPin, 'OzUser', 'string', 'string' )	

	-- Update
	-- POST
	-- /v1/pins/{pinCode}
	local pinUrl = string.format( '%s/%s', GroupPixServerUrls.pinsUrl, pinCode )

	local content = KSJsonEncoder.encodeOzUploadJson { users = { userId } } 

	local postArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Connection', value = 'close' },
		},
		requestMethod = 'POST',
	}
	
	debug.print( "add user by pin" )
	
	local resultP = ozUser:putP( pinUrl, content, postArgs )
		:whenResolvedDo( function( status, data, headers )
			debug.print( string.format( "add user by pin status:%s", status ) )
			
			local result = {}

			if status == 'data' then
				debug.print( "request shared catalog reload()" )
				ozUser:requestSharedCatalogReload()
				
				-- match the catalog id at the end of the string, '$' tells the pattern to start
				-- at the end of the string, then match a-f or 0-9 0 or more times
				-- given a url such as: https://api.oznext.com/v1.0/catalogs/2cae36a77b634458858663121e029ec5
				local catalogId = string.match( headers.Location, "[a-f0-9]*$" )
				debug.print( string.format( "catalogId = %s", catalogId ) )
				
				-- add the catalogId that the user was added to
				result[ "catalogId" ] = catalogId
				
			end
			
			result[ "result" ] = (status == 'data')
			result[ "pinCode" ] = pinCode
			
	        REAppLuaInterface.performResultCallback( "addUserByPinCode", result )

		end )
		
	resultP:runUntilDone()
end


--------------------------------------------------------------------------------

function methods:changePinState( --[[OzUser]] ozUser, --[[OzCatalogService]] catalog, --[[string]] pinCode, --[[string]] pinState )

	WFArguments.check( GroupPixPin, 'OzUser', 'OzCatalogService', 'string', 'string' )	

	-- Update
	-- POST
	-- /v1/pins/{pinCode}
	local pinUrl = string.format( '%s/%s', GroupPixServerUrls.pinsUrl, pinCode )

	local content = KSJsonEncoder.encodeOzUploadJson { pinState = pinState } 

	local postArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Connection', value = 'close' },
		},
		requestMethod = 'POST',
	}

	local resultP = ozUser:putP( pinUrl, content, postArgs )
		:mapDataP( function( status, data )
			debug.print( data )
			
			local pin = KSJsonDecoder.decodeWebSafeJson( data )

			-- changing pin state was successful, set the pinstate value on the catalog payload as well
			return self:setCatalogPinPayload( ozUser, catalog, pin )

		end )
		
	resultP:runUntilDone()
end

--------------------------------------------------------------------------------

function methods:deletePinForCatalog( --[[OzUser]] ozUser, --[[OzCatalogService]] catalog, --[[string]] pinCode )
	
	WFArguments.check( GroupPixPin, 'OzUser', 'OzCatalogService', 'string' )	

	-- DELETE
	-- /v1/pins/{pincode}
	local pinUrl = string.format( '%s/%s', GroupPixServerUrls.pinsUrl, pinCode )

	local options = {
	    requestMethod = 'DELETE',
	    requestHeaders = {
			{ header = 'Connection', value = 'close' },
	    },
	}
	
	local resultP = ozUser:getP( pinUrl, options )
		:mapDataP( function( status, data )
			debug.print( string.format( "pin code %s, deleted", pinCode ) )
			
			local pin = {
				pinCode = "",
				pinState = "",
				pinExpires = "",
				inviteShortUrl = "",
			}
			
			-- deleting pin was successful, set an empty value for the pinCode on the catalog payload as well
			return self:setCatalogPinPayload( ozUser, catalog, pin )
		end )

	resultP:runUntilDone()
end

--------------------------------------------------------------------------------

function methods:setCatalogPinPayload( --[[OzUser]] ozUser, --[[OzCatalogService]] catalog, --[[table]] pin )
	
	WFArguments.check( GroupPixPin, 'OzUser', 'OzCatalogService', 'table' )	
		
	
	local resultP = catalog._docStore:applyOverlayToDocumentP {
		docType = 'catalog',
		docId = catalog._catalogId,
		overlay = {
			payload = {
				[ 'X-Event-Pin' ] = pin.pinCode,
				[ 'X-Event-PinState' ] = pin.pinState,
				[ 'X-Event-PinExpires' ] = pin.pinExpires,
				[ 'X-Event-PinInviteShortUrl' ] = pin.inviteShortUrl,
			}
		},
	}

--[[
	resultP:mapP( function( status, data ) 
	
		debug.print( string.format( "status=%s, data=%s", status, data ) )
		
	end ):runUntilDone()
--]]

	resultP:runUntilDone()
			
	local result = {}
	result[ "eventId" ] = catalog._catalogId
	result[ "pinCode" ] = pin.pinCode
	result[ "pinState" ] = pin.pinState
	result[ "pinExpires" ] = pin.pinExpires
	result[ "inviteShortUrl" ] = pin.inviteShortUrl

	REAppLuaInterface.performResultCallback( "updatePinCode", result )
	
	-- request a reload to update the catalogs
	ozUser:requestCatalogReload()

end

--------------------------------------------------------------------------------

WFCloseSupport.addCloseSupportMethods( methods )

return KSClass 'GroupPixPin'( methods )
