--[[----------------------------------------------------------------------------

KSClass.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-01-18

@reviewed mhamburg 2011-03-08

------------------------------------------------------------------------------]]

--[[!


\class KSClass

\brief This namespace allows you to define Lua classes.

\c KSClass is a function that takes a class name and a prototype-like description of that class.
It returns a factory object that, when called, constructs new instances of the class.

This typical example, adapted from the unit tests, creates a class named "myClass",
describing it with an in-line table. It then creates an object of this type using the
returned factory function, and calls one of the methods

\code
	local gotHere = nil

	local myClass = KSClass 'myClass' {
	
		__init = function( self, ghValue )
			gotHere = ghValue
		end,

		x = 42,
		
		xPlusPlus = function( self )
			local oldX = self.x
			self.x = oldX + 1
			return oldX
			return self.x
		end,
		
		modifier1,
		modifier2
		
	}
	
	local obj1 = myClass()

	assert( obj1:xPlusPlus() == 42, "myObject:x++ = 42" )
	assert( obj1:xPlusPlus() == 43, "myObject:x++ = 43" )
\endcode

 
The table that defines this class contains a mixture of methods,
metamethods, and initial values. These are simply copied through to all
new instances of this class.

The table is interpreted as follows:
<ul>
	<li>Entries in the array index positions (\c modifer1 and \c modifier2) are
	  expected to be functions that will be given an opportunity to
	  modify the prototype before any other construction takes place.
	  A typical example is WFCloseSupport:addCloseSupportMethods(). </li>

	<li>\c __init is interpreted as a constructor method. It takes a \c self
	  parameter, followed by any other parameters that are passed in
	  to the factory function.
	
	<li>Any items whose name matches a Lua metamethod (such as \c __add, but with the
	    exception of \c __newindex) are
	  pulled out and placed in the metatable for an object of this class.</li>
	
	<li>Objects of this class are protected against read/write to
	  non-existent keys (such as \c __index and \c __newindex) unless you specify
	 \c __index = false or nil, or \c __newindex = false. You may not specify
	  any other value for \c __newindex. You may provide an alternative
	  implementation for \c __index, but you lose the read protection in that
	  case.</li>
	  
	<li>A \c __tostring metamethod is provided unless you override that with
	  \c __tostring = false or a function of your own. The default method
	  simply returns the name of the class.</li>
		
	<li>Any string key whose value is a function is removed from the prototype
	  and added to the \c __index table, thus making it a method on the class.
	  It is an error to have a function in the table with a non-string key.	</li>
	
	<li>Any remaining keys are copied as-is into new instances of the class.
	  WARNING: This is a shallow copy. Do not use a table key or value unless
	  you want to share the table across all instances of the class in a
	  given universe.</li>
</ul>

This alternate syntax can be useful if you are declaring a larger class and want to
intersperse member function declarations with other declarations.

\code
local myClass = KSClass( 'myClass', { ... } )
\endcode
	 
You can use the \c KSClass object itself to verify that an object is
of that type:
\code
	local myObj1 = myClass()
	assert( myClass:isClassOf( myObj1 ), "myClass:isClassOf( myObj1 )" )
\endcode
This mechanism is intentionally minimal. It does not provide support
for some features commonly found in class systems, such as inheritance
and mix-in classes. It does not currently support properties, although
we are still thinking about that.
 
]]

--------------------------------------------------------------------------------

local assert = assert
local error = error
local pairs = pairs
local ipairs = ipairs
local setmetatable = setmetatable
local tostring = tostring
local type = type

local string_format = string.format

local KSLeaks = require 'KSLeaks'
local KSLeaks_registerObject = assert( KSLeaks.registerObject )

local KSPrivateKeys = require 'KSPrivateKeys'
local KSPrivateKeys_isPrivateKey = assert( KSPrivateKeys.isPrivateKey )

local WFArguments = require 'WFArguments'
local WFArguments_check = assert( WFArguments.check )
local WFDebugTraceback_traceback = require 'WFDebugTraceback'.traceback

local DEBUG = _G.DEBUG


--------------------------------------------------------------------------------

local function constantFn( value )
	
	-- Minimal version of KSFunctionUtils.constantFn so we don't have to drag
	-- in that whole namespace.
	
	return function()
		return value
	end

end

--------------------------------------------------------------------------------

--[[ We can use the type metafield to ease checking at the expense of allowing
class forging. Since that's abusive and we don't have plug-ins, let's go for it
for now. ]]

local USE_TYPE_METAFIELD = true

local CHECK_METHOD_TARGETS = _G.DEBUG

--------------------------------------------------------------------------------

local metamethodNames = {

	"__init",
		-- not a metamethod as defined by Lua, but we use it here

	--"__index",	-- __index gets special handling (see createIndexTable...)
	--"__newindex", -- Don't allow overrides of __newindex.

	"__call",
	"__eq",
	"__lt",
	"__add",
	"__sub",
	"__mul",
	"__div",
	"__pow",
	"__unm",
	"__concat",
	"__tostring",

}

--------------------------------------------------------------------------------

-- The set of keys that will not generate errors when reading even if not
-- present. (Instead, they will just produce nil.)

local kAllowedKeys = {
	close = true -- make it safe to test for close support
}

--------------------------------------------------------------------------------

local function wrapMethod( className, method )

	if CHECK_METHOD_TARGETS and type( className ) == 'string' then
	
		return function( obj, ... )
			WFArguments_check( className )
			return method( obj, ... )
		end
		
	else
	
		return method
		
	end

end

--------------------------------------------------------------------------------

local function createIndexTableForClass( className, prototype, mt, classObject )

	-- [scouten 2011-01-18]: If we add properties to this class system,
	-- then this approach may no longer work.

	local indexTable = {}
	
	if type( prototype.__index ) == 'function' then
	
		local indexFunc = prototype.__index
		prototype.__index = false
		
		local baseTable = createIndexTableForClass( className, prototype, mt, classObject )
		
		return function( t, k )
			if baseTable[ k ] ~= nil then
				return baseTable[ k ]
			else
				return indexFunc( t, k )
			end
		end
	
	end
	
	-- Disallow read from unknown fields?

	if prototype.__index == nil then
	
		local indexTableMT = {
		
			__index = function( t, k )
				if kAllowedKeys[ k ] or KSPrivateKeys_isPrivateKey( k )  then
					return nil -- we ran out on the index chain but we don't want to error
				else
					local message = string_format( "%s: illegal attempt to read key %s",
										className, tostring( k ) )
					if DEBUG then
						debug.print( message )
						debug.print( WFDebugTraceback_traceback { "KSClass" } )
					end
					error( message, 2 )
				end
			end,
		
		}
		
		setmetatable( indexTable, indexTableMT )

	elseif prototype.__index ~= false then
	
		error( "KSClass: __index must be nil, false, or function", 3 )
	
	end
	
	-- Scan prototype for method definitions.
	
	for k, v in pairs( prototype ) do
	
		if type( v ) == 'function' then
			assert( type( k ) == 'string' )
			indexTable[ k ] = wrapMethod( className, v )
			prototype[ k ] = nil
		end
	
	end
	
	return indexTable

end

--------------------------------------------------------------------------------

local function createMetatableForClass( className, prototype, classObject )

	local mt = {}
	
	if prototype.__tostring == nil then
		mt.__tostring = constantFn( string_format( "%s instance", className ) )
	end
	
	-- Pull metamethods out from prototype and place in metatable. Note that this skips
	-- false entries.
	
	-- Always clear the metamethods. This means that all of the default cases have
	-- to be handled ahead of this copy step.

	for _, mmName in ipairs( metamethodNames ) do
	
		if prototype[ mmName ] then
			assert( type( prototype[ mmName ] ) == 'function' )
			mt[ mmName ] = prototype[ mmName ]
		end
		
		-- Always clear since false just means "no default"
	
		prototype[ mmName ] = nil

	end
	
	-- Always create __index metamethod.
	
	mt.__index = createIndexTableForClass( className, prototype, mt, classObject )

	prototype.__index = nil
	
	-- Create __newindex metamethods unless specifically asked not to do so.
	-- Also check that the entry for __newindex is nil or false.
	
	assert( not prototype.__newindex, "KSClass: prototype.__newindex must be nil or false" )
	
	if prototype.__newindex == nil then
	
		function mt.__newindex( t, k, v )
			if KSPrivateKeys_isPrivateKey( k ) then
				rawset( t, k, v )
			else
				local message = string_format( "%s: illegal attempt to write key %s",
														className, tostring( k ) )
				if DEBUG then
					debug.print( message )
					debug.print( WFDebugTraceback_traceback{ "KSClass" } )
				end
				error( message, 2 )
			end
		end
		
	end
	
	-- We've processed the __newindex entry if any.
	
	prototype.__newindex = nil

	-- Protect the metatable from access. Metatable is here only to support
	-- isClassOf and to provide a __tostring method for classes.
	
	mt.__metatable = setmetatable( {}, {
		__tostring = constantFn(
				string_format( "KSClass instance metatable: %s", className )
		) } )
		
	if USE_TYPE_METAFIELD then
		mt.__type = className
	else
		error( "We haven't implemented type checking without the __type field" )
	end

	return mt

end

--------------------------------------------------------------------------------

-- Apply the modifiers in prototype to the target.

local function applyPrototypeToTarget( prototype, target )
	
	local modifiers
	
	-- Copy any methods into the target
	
	if prototype ~= target then

		-- Extract the modifiers

		modifiers = { }

		for i = 1, #prototype do
			modifiers[ i ] = prototype[ i ]
		end

		-- Copy any methods into the target
	
		for k, v in pairs( prototype ) do
			if not modifiers[ k ] then
				assert( target[ k ] == nil, "Method defined twice: " .. k )
				target[ k ] = v
			end
		end
		
	else
	
		modifiers = prototype
		
	end
	
	-- Walk forward. We use a while loop in case any of the modifiers add further
	-- modifiers.
	
	local i = 1
	
	while true do

		local modifier = modifiers[ i ]
		
		if modifier == nil then break end

		if type( modifier ) == 'function' then
		
			modifier( target )

		elseif type( modifier ) == 'table' then
			
			applyPrototypeToTarget( modifier, target )
			
		else
		
			error( "Expected a function or table. Got " .. type( modifier ) )

		end
		
		if prototype == target then
			prototype[ i ] = nil
		end
		
		i = i + 1

	end
	
end

--------------------------------------------------------------------------------

local function adjustPrototypeForModifiers( prototype )

	applyPrototypeToTarget( prototype, prototype )
	
end

--------------------------------------------------------------------------------

local kStandardModifiers = {

	bindMethod = function( self, mname )
		WFArguments_check( '_', 'string' )
		local method = self[ mname ]
		return function( ... ) return method( self, ... ) end
	end,

	bindDynamicMethod = function( self, mname )
		WFArguments_check( '_', 'string' )
		return function( ... ) return self[ mname ]( self, ... ) end
	end,

}

--------------------------------------------------------------------------------

local function adjustPrototypeForStandardModifiers( prototype )

	applyPrototypeToTarget( kStandardModifiers, prototype )
	
end

--------------------------------------------------------------------------------

local function newObject( className, prototype, mt, ... )

	local newInstance = {}
	
	for k, v in pairs( prototype ) do
		newInstance[ k ] = v
	end
	
	setmetatable( newInstance, mt )
	
	KSLeaks_registerObject( className, newInstance )
	
	local initFunc = mt.__init
	
	if initFunc then
		initFunc( newInstance, ... )
	end
	
	return newInstance

end

--------------------------------------------------------------------------------

local function createClassObject( className, prototype )

	WFArguments_check( 'string', 'table' )

	assert( className ~= '', "className must be a non-empty string" )

    -- mt is the metatable for objects of the class; mt.__metatable is used by isClassOf;
    -- classObjectMT is the metatable of the class object.

	local classNameExpanded = string_format( "KSClass: %s", className )
	
	-- The prototype can include a list of standard modifier functions (like
	-- addCloseSupport) that we apply now before doing any further construction.
	
	adjustPrototypeForModifiers( prototype )
	
	adjustPrototypeForStandardModifiers( prototype )

	local classObject = {}

	local mt = createMetatableForClass( className, prototype, classObject )
	
	local mt_metatable = mt.__metatable
	
	local classObjectMT = {
	
		__call = function( _, ... )
			-- The _ parameter is the class object.
			return newObject( className, prototype, mt, ... )
		end,
		
		__tostring = constantFn( classNameExpanded ),
		
		__index = {
		
			isClassOf = function( self, object )
			
				-- NOTE: This isn't robust against class forging since anyone can
				-- use getmetatable to get this key and stick it on their class
				-- as the __metatable entry.
				
				-- GRUNGE [mhamburg 2011-03-08]: Not necessarily worth fixing since
				-- it's abusive, but the correct fix is probably to maintain a weak
				-- set of members of the class.
				
				-- [mhamburg 2011-03-08]: At the expense of losing some security
				-- around the function but at a gain in simplicity, isClassOf could just
				-- be a field on the class rather than having to reside in an __index
				-- table.

				return getmetatable( object ) == mt_metatable
			
			end,
		
		},
		
		__newindex = function( t, k )
			error( "KSClass object is read-only", 2 )
		end,
		
		__metatable = classNameExpanded,
	
	}
	
	return setmetatable( classObject, classObjectMT )

end

--------------------------------------------------------------------------------

local function KSClass( className, prototype )

	-- We accept either of the following syntaxes:

		-- local class = KSClass( className, prototype )

	-- or:
	
		-- local class = KSClass( className )( prototype )
		-- which looks prettier when you replace prototype with
		-- an inline table construction.

	if not prototype then
	
		return function( prototype )
			return createClassObject( className, prototype )
		end
	
	else
	
		return createClassObject( className, prototype )

	end

end

--------------------------------------------------------------------------------

return KSClass
