--[[----------------------------------------------------------------------------

KSClassTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-01-17

@reviewed epoley 2011-03-08

------------------------------------------------------------------------------]]

local KSClass = require 'KSClass'
local KSPrivateKeys = require 'KSPrivateKeys'
local KSTestUtils = require 'KSTestUtils'

local KSClassTest = {}

--------------------------------------------------------------------------------

function KSClassTest.testBasicClass()

	local gotHere

	local fooClass = KSClass 'fooClass' {
	
		__init = function( self, ghValue )
			gotHere = ghValue
		end,
		
	}
	
	local foo1 = fooClass( 1 )
	assert( type( foo1 ) == 'table', "didn't create a table" )
	assert( gotHere == 1, "didn't get to init method" )

	local foo2 = fooClass( 2 )
	assert( type( foo2 ) == 'table', "didn't create a table (2)" )
	assert( gotHere == 2, "didn't get to second init method" )
	
	assert( foo1 ~= foo2, "didn't create distinct objects" )

end

--------------------------------------------------------------------------------

function KSClassTest.testDefaultMetamethods()

	local fooClass = KSClass 'fooClass' {}
	
	local foo1 = fooClass( 1 )
	
	local success = pcall( function() return foo1.x end )
	assert( not success, "__index didn't throw" )
	
	success = pcall( function() foo1.y = 42 end )
	assert( not success, "__newindex didn't throw" )
	
	local tostr = tostring( foo1 )
	assert( tostr == 'fooClass instance', "__tostring didn't return class name" )

end

--------------------------------------------------------------------------------

function KSClassTest.testDisableDefaultMetamethods()

	local fooClass = KSClass 'fooClass' {
	
		__index = false,
		__newindex = false,
		__tostring = false,

	}
	
	local foo1 = fooClass( 1 )
	
	local success = pcall( function() return foo1.x end )
	assert( success, "__index did throw" )
	
	success = pcall( function() foo1.y = 42 end )
	assert( success, "__newindex did throw" )
	
	local tostr = tostring( foo1 )
	assert( tostr ~= 'fooClass instance', "__tostring returned class name" )

end

--------------------------------------------------------------------------------

function KSClassTest.testDefaultPropertiesCopied()

	local fooClass = KSClass 'fooClass' { x = 42 }
	
	local foo1 = fooClass()
	local foo2 = fooClass()
	
	assert( foo1.x == 42, "foo1.x == 42" )
	foo1.x = 43
	
	assert( foo1.x == 43, "foo1.x == 43 after assignment" )
	assert( foo2.x == 42, "foo2.x == 42, unaffected by assignment to foo1" )

end

--------------------------------------------------------------------------------

function KSClassTest.testOnlyDefaultPropertiesCopied()

	local fooClass = KSClass 'fooClass' { x = 42, __tostring = false }
	
	local foo1 = fooClass()
	
	for k, v in pairs( foo1 ) do
		assert( k == 'x', 'non-property copied: ' .. tostring( k ) )
	end

end

--------------------------------------------------------------------------------

function KSClassTest.testDefaultTablePropertiesCopied()

	local fooClass = KSClass 'fooClass' {
		__init = function( self )
			self.t = { }
		end,
		t = false,
		
	}
	
	local foo1 = fooClass()
	local foo2 = fooClass()
	
	local t1 = foo1.t
	local t2 = foo2.t
	
	assert( #t1 == 0, "#t1 == 0" )
	assert( #t2 == 0, "#t2 == 0" )
	
	foo1.t[ 1 ] = true
	
	assert( #t1 == 1, "#t1 == 1"  )
	assert( #t2 == 0, "#t2 == 0" )
	
end

--------------------------------------------------------------------------------

function KSClassTest.testIsClassOf()

	local fooClass = KSClass 'fooClass' {}
	local barClass = KSClass 'barClass' {}
	
	local foo1 = fooClass()
	local bar1 = barClass()

	assert( fooClass:isClassOf( foo1 ), "fooClass:isClassOf( foo1 )" )
	assert( not barClass:isClassOf( foo1 ), "barClass:isClassOf( foo1 )" )
	
	assert( barClass:isClassOf( bar1 ), "barClass:isClassOf( bar1 )" )
	assert( not fooClass:isClassOf( bar1 ), "fooClass:isClassOf( bar1 )" )
	
	assert( not fooClass:isClassOf( _G ), "fooClass:isClassOf( _G )" )

	assert( not fooClass:isClassOf( 1 ), "fooClass:isClassOf( _G )" )

	assert( not fooClass:isClassOf( nil ), "fooClass:isClassOf( _G )" )

end

--------------------------------------------------------------------------------

function KSClassTest.testMethods()

	local fooClass = KSClass 'fooClass' {
	
		x = 42,
		
		xPlusPlus = function( self )
			local oldX = self.x
			self.x = oldX + 1
			return oldX
		end,
	
	}

	local foo1 = fooClass()
	
	assert( foo1:xPlusPlus() == 42, "foo:x++ = 42" )
	assert( foo1:xPlusPlus() == 43, "foo:x++ = 43" )
	assert( foo1:xPlusPlus() == 44, "foo:x++ = 44" )

end

--------------------------------------------------------------------------------

function KSClassTest.testCustomToString()

	local fooClass = KSClass 'fooClass' {
	
		x = 42,
		
		__tostring = function( self )
			return string.format( "Foo with the value of %d", self.x )
		end,
	
	}
	
	local foo = fooClass()
	
	local foostring = tostring( foo )
	assert( foostring == "Foo with the value of 42", "tostring returned an incorrect value: " .. foostring )
	
end

--------------------------------------------------------------------------------

function KSClassTest.testSetMetatableProtection()

	local fooClass = KSClass 'fooClass' {}

	local foo1 = fooClass()
	
	assert( KSTestUtils.testFails( function() setmetatable( foo1, { a = "b" } ) end ) )

end

--------------------------------------------------------------------------------

function KSClassTest.testClassModifiers()

	local function modifier1( proto )
		proto.sayHello = function( self ) return "Hello!" end
	end

	local function modifier2( proto )
		proto.sayGoodbye = function( self ) return "Goodbye!" end
	end

	local fooClass = KSClass 'fooClass' { modifier1, modifier2 }
	
	local foo = fooClass()
	
	assert( foo:sayHello() == "Hello!", "Should have added sayHello method." )

	assert( foo:sayGoodbye() == "Goodbye!", "Should have added sayGoodbye method." )

end

--------------------------------------------------------------------------------

function KSClassTest.testFieldProtection()

	local fooClass = KSClass 'fooClass' { }
	
	local foo = fooClass()
	
	assert( KSTestUtils.testFails( function() return foo.x end ), "Should not be able to read a missing field" )
	
	assert( KSTestUtils.testFails( function() foo.x = 3 end ), "Should not be able to assign to missing fields" )
	
	local xk = KSPrivateKeys.makePrivateKey( "x" )
	
	assert( foo[ xk ] == nil, "Should be able to read from private key" )
	
	foo[ xk ] = 3 -- Should be able to assign
	
	assert( foo[ xk ] == 3, "Assignment should set private key" )
	
	assert( foo.close == nil, "Should be able to read the missing close field" )
	
	assert( KSTestUtils.testFails( function() foo.close = 3 end ), "Should not be able to assign to the close field" )
	
end

--------------------------------------------------------------------------------

function KSClassTest.testClassArgumentChecking()

	local bazClass = KSClass 'bazClass' { }
	
	local baz = bazClass()
	
	local WFArguments = require 'WFArguments'
	
	local function takeBaz( b ) WFArguments.check( 'bazClass' ) end
	
	takeBaz( baz )
	
	assert( not pcall( takeBaz, { } ), "Should fail with non class object" )
	
end

--------------------------------------------------------------------------------

function KSClassTest.testBindMethod()

	local helloClass = KSClass 'helloClass' {
		_me = "unspecified",
		__init = function( self, me )
			self._me = me
		end,
		hello = function( self, x )
			return "Hello " .. x .. " from " .. self._me
		end
	}
	
	local fred = helloClass( "Fred" )
	
	local bhello = fred:bindMethod 'hello'
	
	assert( bhello( "Ginger" ) == "Hello Ginger from Fred",
		"Bound function returned the wrong value." )
	
end

--------------------------------------------------------------------------------

function KSClassTest.testBindDynamicMethod()

	local helloClass = KSClass 'helloClass' {
		_me = "unspecified",
		__init = function( self, me )
			self._me = me
		end,
		hello = function( self, x )
			return "Hello " .. x .. " from " .. self._me
		end
	}
	
	local fred = helloClass( "Fred" )
	
	local bhello = fred:bindDynamicMethod( 'hello' )
	
	assert( bhello( "Ginger" ) == "Hello Ginger from Fred",
		"Bound function returned the wrong value." )
		
	rawset( fred, 'hello', function( self, x )
		return "Greetings " .. x .. " from " .. self._me
	end )
	
	assert( bhello( "Ginger" ) == "Greetings Ginger from Fred",
		"Dynamically bound function returned the wrong value." )
	
end

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-10-10]: Add a test for the new __index = function support.

--------------------------------------------------------------------------------

return KSClassTest
