--[[----------------------------------------------------------------------------

KSCommandManager.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author erwright 2011-05-05

------------------------------------------------------------------------------]]

	-- core.logging
local KSLogging = require 'KSLogging'

local KSReactive = require 'KSReactive'

local WFCommandManagerImp = require 'WFCommandManagerImp'

local KSCommandManager = { }

local commands = { }

local canExecuteCommandBs = { }

local commandManagerLog = KSLogging.log 'KSCommandManager'
--	:setLevel( KSLogging.DEBUG )
	
--------------------------------------------------------------------------------

-- Registers a command.  Will overwrite any existing command with the same name.
--
-- Arguments:
--   command:        a string name (required)
--   fn:             a function that can take any nubmer of parameters and return any number of values (required)
--   optCanExecuteB: behavior to disable the command, returning status and optional reason code (optional)

function KSCommandManager.registerCommand( command, fn, optCanExecuteB )

	assert( type( command ) == 'string' )
	assert( type( fn ) == 'function' )
	if type( optCanExecuteB ) ~= 'nil' then
		local _, isBehavior = KSReactive.isEventNode( optCanExecuteB )
		assert( isBehavior, "Didn't get a behavior for optCanExecuteB for command: " .. command )
	end

	if commands[ command ] ~= nil then
		commandManagerLog:info( "KSCommandManager: overwriting registration for command: %s", command )
	end
	
	commandManagerLog:debug( "KSCommandManager: Registering command: %s", command )
		
	commands[ command ] = fn
	canExecuteCommandBs[ command ] = optCanExecuteB or KSReactive.constantB( true )
	
	-- Ensure that the command/enabled state gets sent over to the native WFCommandManager proxy.
	-- then ensure future changes get sent over, too.
	
	KSCommandManager.canExecuteCommandB( command ):onDataB( function( canExecute )
		WFCommandManagerImp.updateCommandState( command, not not canExecute )
	end )
	
end

--------------------------------------------------------------------------------

function KSCommandManager.isCommandRegistered( command )

	assert( type( command ) == 'string' )

	return commands[ command ] ~= nil
	
end

--------------------------------------------------------------------------------

function KSCommandManager.canExecuteCommandB( command )

	assert( type( command ) == 'string' )

	local canExecuteB = canExecuteCommandBs[ command ]
	
	if not canExecuteB then
		commandManagerLog:error( "KSCommandManager: canExecuteCommandB for unregistered command: %s", command )
		return KSReactive.constantB( false )
	end

	return canExecuteB:relayB()
	
end

--------------------------------------------------------------------------------

-- Returns ( boolean[, string ] ):
--   boolean = canExecute
--   string  = optional reason code for why the command is disabled

function KSCommandManager.canExecuteCommand( command )
	
	local canExecute, reason = KSCommandManager.canExecuteCommandB( command ):valueNow()
	assert( type( canExecute ) == 'boolean', "canExecuteCommandB() got: " .. tostring( canExecute ) .. " for command: " .. command )
	
	commandManagerLog:debug( "KSCommandManager: command %s, can execute: %s", command, canExecute )
	
	return canExecute, reason

end

--------------------------------------------------------------------------------

function KSCommandManager.executeCommand( command, ... )

	assert( type( command ) == 'string' )

	commandManagerLog:debug( "KSCommandManager: Executing command: %s", command )

	local fn = commands[ command ]
	if not fn then
		commandManagerLog:error( "KSCommandManager: executeCommand called for unregistered command: %s", command )
		return
	end
	
	if not KSCommandManager.canExecuteCommand( command ) then
		-- FIXME [erwright 2011-12-06]:  Bump warning level to error for this particular
		-- log so we can tell if we are actually getting the store receipts. Return
		-- this to lower debug level after we confirm bug is fixed (presumably after
		-- 1.1.) This instrumentation is to help us understand bug 3061577.
		-- PR #3062887: Reminder to restore this to :debug.
	
		--commandManagerLog:debug( "KSCommandManager: attempted to execute disabled command: %s", command )
		commandManagerLog:error( "KSCommandManager: attempted to execute disabled command: %s", command )
		
		return
	end
	
	return fn( ... )
end

--------------------------------------------------------------------------------

return KSCommandManager
