--[[----------------------------------------------------------------------------

KSDataCache.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This is the public API to KSDataCache.

-- FIXME [scouten 2011-04-05]: Document it.

--------------------------------------------------------------------------------

@author mhamburg 2011-03-xx (extracted from BabluImp.lua by scouten 2011-04-05)

@reviewed scouten 2011-03-02

General review comment: This file will probably undergo significant revision
in M1.

@reviewed epoley 2011-03-08
@reviewed mhamburg 2011-03-11

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSPromises = require 'KSPromises'
local WFCloseSupport = require 'WFCloseSupport'


--============================================================================--

-- @mark KSDataCache class
--[[!
	\class KSDataCache
	\brief This namepace allows manipulation of a database.
	A database is represented as a table of simple keys whose
	values are KSData objects.

]]--

local dataCacheMethods = {

	_readEP = false,
		-- (WFWorkerThread) event processor that implements reading from the cache

	_writeEP = false,
		-- (WFWorkerThread) event processor that implements writing to the cache
	
	_pathP = false,
		-- (P< string >) path to data cache file (should be used only for disk space
		-- management)
		
	_localCache = false,
		-- (fully weak table) map from key to data
		
	_readPromiseCache = false,
		-- (fully weak table) map from key to promise for data
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function dataCacheMethods:__init( optName )

	self._writeEP = WFWorkerThread( 'KSDataCacheImp.lua',
										 'write:' .. ( optName or '' ) )
										
	-- Wait for the write thread before attempting the read thread. We use the
	-- fact that the worker thread API can take a promise for the argument
	-- as a way to delay construction of the result.
	
	local readArgP = self._writeEP:threadIsReadyP():mapDataP( function()
		return 'data', 'read:' .. ( optName or '' )
	end )

	self._readEP = WFWorkerThread( 'KSDataCacheImp.lua', readArgP )

	self:addCloseTarget( self._writeEP )
	self:addCloseTarget( self._readEP )
	
	local fullyWeakMT = { __mode = 'kv' }
	
	self._localCache = setmetatable( { }, fullyWeakMT )

	self._readPromiseCache = setmetatable( { }, fullyWeakMT )

end

--------------------------------------------------------------------------------
-- Create a request against the write processor but do so only after the
-- read processor is ready. This balances with the fact that we won't start
-- the read processor until write processor is ready.

local function writeRequestP( self, ... )
	local n = select( '#', ... )
	local args = { ... }
	return self._readEP:threadIsReadyP():mapDataP( function()
		return self._writeEP:requestP( unpack( args, 1, n ) )
	end )
end

--------------------------------------------------------------------------------
-- (P< string >) path to database file (should be used only for disk space
-- management)

--[[!
      Retrieves the physical path to the database file, for the purpose of disk
      space management. Do not manipulate the database file directly.

      \return The file path.
]]--

function dataCacheMethods:pathP()

	if not self._pathP then
		self._pathP = self._readEP:requestP( 'path' )
	end
	
	return self._pathP

end

--------------------------------------------------------------------------------

--[[!
       \brief Creates a promise that provides a data value.

       \param path (string) The key of the data value.
       \param debugMe (boolean) When \c true, tracks retain/release
	    issues with the data object that is created.

       \return The promise object.
]]--

function dataCacheMethods:getP( --[[string]] path, --[[boolean]] debugMe )

	-- debugMe helps you to debug retain/release issues with the NSData object
	-- that gets created.

	-- FIXME [mhamburg 2011-04-13]: Is it worth tracking these so that they
	-- could be closed if need be?
	
	local cachedP = self._readPromiseCache[ path ]
	
	if cachedP and not cachedP:isCanceled() then
		return cachedP
	end
	
	local cachedData = WFCloseSupport.fetchCached( self._localCache, path )
	
	if cachedData then
		local p = KSPromises.resolvedPromiseP( 'data', cachedData )
		self._readPromiseCache[ path ] = p
		return p
	end
	
	local readP = self._readEP:requestP( 'fetchCachedData', path, debugMe )
	
	-- Since we're multithreading, a failure to read might actually be caused
	-- by the data just not having been written yet. So, we try again on the
	-- write EP. This has the downside of resulting in two attempts to read, but
	-- if we're routinely missing in the data cache then it probably isn't
	-- doing its job.
	
	-- Note that we terminate on an actual error from the data cache. A nil
	-- result just means that the data wasn't present.
	
	readP = readP:mapP( function( status, data )
	
		if status == 'data' then
	
			if data then

				self._localCache[ path ] = data
				return 'data', data

			elseif WFCloseSupport.hasBeenOrWillBeClosed( self._writeEP ) then
			
				return 'error', "Read request outlived the data cache"

			else
			
				local wp = writeRequestP( self, 'fetchCachedData', path, debugMe )
				
				local p = wp:observeP( function( status, data )
					if status == 'data' and data then
						self._localCache[ path ] = data
					end
				end )
					
				return p
					
			end
		
		else
		
			debug.print( "Data cache read error:", status, data )
			return status, data

		end

	end )
	
	-- Cache it
	
	self._readPromiseCache[ path ] = readP
	
	-- And return it.
	
	return readP
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Creates a promise that provides keys in the database.

	\param optStartAfterPath  (string) The key of a data value at which to start.
		If not supplied, starts at the beginning.
	\param optLimit	(number) The number of keys to retrieve. If not supplied,
		retrieves all keys after the start key.

	\return The promise object.

]]--

function dataCacheMethods:enumerateKeysP( --[[string]] optStartAfterPath, --[[number]] optLimit )

	-- This is a read operation but we do it on the write thread because we
	-- don't want to block the read thread waiting for the enumeration.

	return writeRequestP( self, 'enumerateKeys', optStartAfterPath, optLimit )

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a promise that stores a data value in the database.

	\param path (string) The key of the data value.
	\param data (KSData) The new data value.

	\return The promise object.
]]--

function dataCacheMethods:putP( --[[string]] path, --[[KSData]] data )

	self._localCache[ path ] = data

	local resultP = writeRequestP( self, 'storeCachedData', path, data )
	
	resultP:runUntilDone()
		-- No cancel support since we just want to put.
	
	return resultP

end

--------------------------------------------------------------------------------

--[[!
 	\brief Creates a promise that removes a data value from the database.

       \param path (string) The key of the data value.

	\return The promise object.
]]--

function dataCacheMethods:removeP( --[[string]] path )
	
	local resultP = writeRequestP( self, 'removeData', path )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

--[[!
	\brief Garbage collects the database.
]]--

function dataCacheMethods:vacuum()

	return writeRequestP( self, 'vacuum' )

end

--------------------------------------------------------------------------------

--[[!
	\brief Shuts down the connection to the database and deletes it from disk.

]]--

function dataCacheMethods:haltAndDeleteP()

	assert( self._writeEP and self._readEP )

	-- Shutdown the write event processor and wait for confirmation.

	local stopWriteP = writeRequestP( self, 'haltAndDelete', false )

	-- Then stop the read EP, ask it to delete the files, and wait for confirmation.
	
	local stopReadP = stopWriteP:mapDataP( function()
	
		return self._readEP:requestP( 'haltAndDelete', true )
		
	end )
	
	stopReadP:whenResolvedDo( function()

		self:close()
	
	end ):runUntilDone()
	
	return stopReadP

end

--------------------------------------------------------------------------------

return KSClass 'KSDataCache'( dataCacheMethods )
