--[[----------------------------------------------------------------------------

KSDatabase.lua

@author scouten 2011-02-18

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This namespace contains two classes that do most of the work for a database-backed
event processor.


SCHEMA DEFINITION
`````````````````

See KSDatabaseSchema.lua for information on how database schemas are defined.


DATABASE SERVICE
````````````````

Once the schema is defined, the event processor init script will then convert
that to a running database by calling

	KSDatabase.startDatabaseWithSchema( schema, ... )
		-- Using ... passes desired database path through.

This call:

	(1) Opens the SQLite database connection.
	(2) Ensures that the database schema matches what was specified in the schema
		declaration above.
	(3) Registers standard message handlers so that a KSDatabaseService can communicate
		with it.

In most cases, this should mark the end of the init script. However, it is possible
that the event processor may need more direct access to the database later, so a
reference to the database object is returned by this call.

Once this call completes, the database schema is frozen for the lifetime of the
event processor. It is permitted to change the database schema object after this
point, but any such changes will be ignored for this database instance.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'

	-- wichitafoundation.core.string
local KSStringUtils = require 'KSStringUtils'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.sqlite.database
local WFSqliteDatabaseImp = require 'WFSqliteDatabaseImp'


--------------------------------------------------------------------------------

local COLLECT_DIAGNOSTIC_LOG_DATA = true


--============================================================================--

-- @mark SQL statement cache

--------------------------------------------------------------------------------

local function getStatement( self, sql )

	local cache = self._stmtCache
	
	local stmt = cache[ sql ]
	
	if stmt then

		-- WARNING [scouten 2011-04-12]: If we ever start using APIs from
		-- WFSqliteDatabaseImp that don't reset the statement upon completion,
		-- this needs to come back.

		--stmt:reset()
	
	else

		stmt = self._db:prepare( sql )
		cache[ sql ] = stmt

		-- If we haven't cached bindByPosition and gatherArrayOfArray yet,
		-- do it now. This saves us a bunch of overhead in callmethod of
		-- AgDirectObject.
		
		if not self._bindByPositionMethod then
			self._bindByPositionMethod = stmt:getMethod 'bindByPosition'
			self._runToCompletionMethod = stmt:getMethod 'runToCompletion'
			self._gatherArrayOfArrayMethod = stmt:getMethod 'gatherArrayOfArray'
		end
	
	end
	
	return stmt

end

--------------------------------------------------------------------------------

local function killStatement( self, sql )

	local cache = self._stmtCache
	
	local stmt = cache[ sql ]
	
	if stmt then
	
		stmt:finalize()
		cache[ sql ] = nil
	
	end
	
end


--============================================================================--

-- @mark schema migration

--------------------------------------------------------------------------------

local function configureDatabaseForSchema( db, schema, path )

	-- Add required version info schema.
	
	schema:addTable [[CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )]]

	-- Turn on WAL support if so requested. (But don't when using in-memory
	-- databases, since that is kind of pointless and seems to hurt performance.)

	if ( not path:match( '^:memory:' ) ) and schema._useWAL then
		db:execOnce( "PRAGMA journal_mode = WAL;" )
	end

	-- Make sure all desired tables exist.

	local tableIntrospectionSql = [[SELECT tbl_name, sql FROM sqlite_master WHERE type == "table"]]
	
	local tableSchemaRows = db:gatherArrayOfArray( tableIntrospectionSql )
	killStatement( db, tableIntrospectionSql )
	
	local existingTables = {}
	
	for i = 1, #tableSchemaRows do
		
		local schemaRow = tableSchemaRows[ i ]
		existingTables[ schemaRow[ 1 ] ] = schemaRow[ 2 ]
			-- existingTables[ tblName ] = sql
	
	end
	
	for tableName, tableDescription in pairs( schema._tables ) do
	
		local existingTableSql = existingTables[ tableName ]
		
		if existingTableSql then
			
			-- Table already exists in DB.
			
			if existingTableSql ~= tableDescription._sql then

				-- Table exists and doesn't match desired schema.
				
				if tableDescription._onConflict == 'replace' then
				
					-- Delete existing table and start over.
					
					db:execOnce( "DROP TABLE " .. tableName )
					db:execOnce( tableDescription._sql )
				
				else
				
					error( string.format(
						"KSDatabase: conflicting definition for table %s\n\n%s",
						tableName, existingTableSql ), 2 )
				
				end
			
			end
		
		else
		
			-- No such table in DB: Create it now.
			
			db:execOnce( tableDescription._sql )
		
		end
	
	end
	
	-- Schema version check. (This might have to move ahead of the table
	-- creation phase above in future versions when we have a known incompatibility,
	-- but this will suffice to lock v1 clients out of future DBs and provide
	-- enough info for future clients to know they have to upgrade.)
	
	local currentVersion, backwardCompatibleToVersion =
						schema:currentAndBackwardCompatibleVersions()

	local schemaVersionCheckSql = [[SELECT currentVersion, backwardCompatibleToVersion
						FROM schemaVersion]]
	
	local schemaVersionRows = db:gatherArrayOfArray( schemaVersionCheckSql )
	killStatement( db, schemaVersionCheckSql )
	
	if not schemaVersionRows then
		error( "KSDatabase: no response from schema query" )
	end
	
	if #schemaVersionRows > 1 then
		error( "KSDatabase: multiple schema version rows" )
	end

	local schemaVersionRow = schemaVersionRows[ 1 ]
	if schemaVersionRow then
	
		local db_currentVersion = schemaVersionRow[ 1 ]
		local db_backwardCompatibleVersion = schemaVersionRow[ 2 ]
		
		if type( db_backwardCompatibleVersion ) ~= 'string'
			or db_backwardCompatibleVersion > currentVersion
		then
			error( "KSDatabase: database format is newer than current software" )
		end
		
		if type( currentVersion ) ~= 'string' or currentVersion < db_currentVersion then
			error( "KSDatabase: database version is older than current, no upgrade path" )
		end
	
		-- FIXME [scouten 2011-08-13]: In a future version, we'll need to
		-- upgrade the current/backward values to match the new values once
		-- we've upgraded the schema.
	
	else
		
		-- No pre-existing schema statement: Probably a new DB..
	
		db:execOnce( [[INSERT INTO schemaVersion VALUES( ?, ? )]], currentVersion,
							backwardCompatibleToVersion )

	end
	
	-- FIXME [scouten 2011-02-18]: Remove unwanted tables? NNTD
	
	-- Make sure all desired indexes exist.
	
	local indexIntrospectionSql = [[SELECT name, sql FROM sqlite_master WHERE type == "index"]]
	
	local indexSchemaRows = db:gatherArrayOfArray( indexIntrospectionSql )
	killStatement( db, indexIntrospectionSql )
	
	local existingIndexes = {}
	
	for i = 1, #indexSchemaRows do
		
		local schemaRow = indexSchemaRows[ i ]
		existingIndexes[ schemaRow[ 1 ] ] = schemaRow[ 2 ]
			-- existingIndexes[ tblName ] = sql
	
	end
	
	for indexName, indexDescription in pairs( schema._indexes ) do
	
		local existingIndexSql = existingIndexes[ indexName ]
		
		if existingIndexSql then
			
			-- Index already exists in DB.
			
			if existingIndexSql ~= indexDescription._sql then

				-- Index exists and doesn't match desired schema.
				-- Unlike tables, we don't bother with a conflict strategy;
				-- we always blow away any unwanted index and start over.
				
				db:execOnce( "DROP INDEX " .. indexName )
				db:execOnce( indexDescription._sql )
				
			end
		
		else
		
			-- No such index in DB: Create it now.
			
			db:execOnce( indexDescription._sql )
		
		end
	
	end
	
	-- FIXME [scouten 2011-03-18]: Remove unwanted indexes.

end


--============================================================================--

-- @mark database wrapper class

local databaseMethods = {

	_db = false,
		-- (WFSqliteDatabaseImp)
	
	_stmtCache = false,
		-- (table) cache of prepared statements
			-- key: (string) SQL to compile
			-- value: (WFSqliteStatement)
	
	_path = false,
		-- (string) path to the database
	
	_hasActiveTransaction = false,
		-- (boolean) true if we're in the midst of a :withTransactionDo block

	_bindByPositionMethod = false,
		-- (function) optimized bypass for bindByPosition method of WFSqliteStatement

	_runToCompletionMethod = false,
		-- (function) optimized bypass for runToCompletion method of WFSqliteStatement
	
	_gatherArrayOfArrayMethod = false,
		-- (function) optimized bypass for gatherArrayOfArray method of WFSqliteStatement

	_PERF_gatherStats = false,
		-- (boolean) true to gather performance data about queries run against
		-- this database
	
	_PERF_stmtStats = false,
		-- (table) stats for each statement
			-- key: (string) SQL from statement
			-- value: (table) info on statement (count, min, max, total)

	_PERF_queriesInTransaction = false,
		-- (table) while in a transaction, this accumulates the queries that were executed
			-- in this transaction

	_log = false,
		-- (KSLogging.log)

}

--------------------------------------------------------------------------------

local dbInstanceCounter = 1

--------------------------------------------------------------------------------

-- @mark KSDatabase class
--[[!

	\class KSDatabase
	\brief This namespace provides direct access to an SQLite database.

	\sa KSDatabaseSchema

	A worker thread's init script defines a schema, then converts
	it to a running database by calling
	
	<pre>
	   KSDatabase.startDatabaseWithSchema( schema, ... )
	</pre>
	
	Use the additional arguments to pass through the desired database path.
	
	This call opens the SQLite database connection, ensures that the database schema
	matches what was specified in the schema declaration, and
	registers standard message handlers so that a KSDatabaseService can communicate
	with it.
	
	In most cases, this should mark the end of the init script.
	The call returns a reference to the database object,which represent this
	database connection; the thread can use it for direct access to the database if needed.
	 
	Once this call completes, the database schema is frozen for the lifetime of the
	thread. Any change you make to the database schema object are ignored for this
	database instance.

]]--

function databaseMethods:__init( path, schema )

	if type( path ) ~= 'string' then
		error( "KSDatabase.startDatabaseWithSchema: first argument must be path", 2 )
	end

	self._log = KSLogging.log( string.format( "KSDatabase %s %d in thread %s", path,
											  dbInstanceCounter,
											  WFCurrentThread.DEBUG_getThreadName() ) )
--		:setLevel( KSLogging.DEBUG )

	dbInstanceCounter = dbInstanceCounter + 1

	if path:sub( 1, 8 ) == ':memory:' then

		path = ':memory:'

	elseif path:sub( 1, 6 ) == ':temp:' then

		-- Per http://www.sqlite.org/inmemorydb.html, this will cause an
		-- on-disk DB to be created and then destroyed when the connection
		-- is destroyed.

		path = ''

	end

	self._path = path
	
	self._log:info( "Make database" )

	self._db = assert( WFSqliteDatabaseImp.makeDatabase( path ) )
	
	self._stmtCache = {}
	
	-- Collect diagnostic log data?
	
	if COLLECT_DIAGNOSTIC_LOG_DATA then
	
		self:PERF_collectQueryPerformanceStats()
	
		WFDiagnosticLog.logDataRequestE()
			:onDataE( function()
				self:sendDiagnosticLogData()
			end )
	
	end
	
	-- Ensure that we have desired schema in place.
	
	local success, message = pcall( function()
		configureDatabaseForSchema( self, schema, path )
	end )
	
	if not success then
	
		-- If we fail here, DB is probably corrupt. Close it out now.
	
		self._db:close()
		self._db = nil
		
		error( message )

	end
	
	WFCloseSupport.mustCloseAtThreadTermination( self )

end

--------------------------------------------------------------------------------

local function assertDatabaseNotClosed( self )

	local db = self._db
	
	if not db then
		self._log:error( "Database used after it was closed via %s", ':traceback:' )
		error( "KSDatabase: database was closed", 3 )
	end
	
	return db

end

--------------------------------------------------------------------------------
--[[!
	Closes this database connection.

]]--

function databaseMethods:close()

	if self._db then
	
		self._log:info( "Finalizing statements in preparation for closing database" )

		for _, stmt in pairs( self._stmtCache ) do
			stmt:finalize()
		end
		self._stmtCache = false
	
		self._log:info( "Closing database" )
	
		self._db:close()
		self._db = false

	end

end

--------------------------------------------------------------------------------

--[[!
	Retrieves the disk path to this database file.

	\return (string) The path.

]]--

function databaseMethods:path()

	return self._path

end

--------------------------------------------------------------------------------

local sqlContinuationIndent = "\n" .. string.rep( " ", 48 )

--------------------------------------------------------------------------------

local queryStartTime

local function finish_profileQuery( self, sql, ... )

	local elapsedTime = debug.timeNow() - queryStartTime
	
	if sql == 'COMMIT' and self._PERF_queriesInTransaction then

		sql = 'COMMIT :: queries were:' .. sqlContinuationIndent .. table.concat( self._PERF_queriesInTransaction, sqlContinuationIndent )
		self._PERF_queriesInTransaction = false

	elseif self._hasActiveTransaction and self._PERF_queriesInTransaction and sql ~= 'BEGIN TRANSACTION' then
	
		self._PERF_queriesInTransaction[ #self._PERF_queriesInTransaction + 1 ] = sql
	
	end
	
	local stat = self._PERF_stmtStats[ sql ]
	
	if stat then
		stat.count = stat.count + 1
		stat.min = math.min( stat.min, elapsedTime )
		stat.max = math.max( stat.max, elapsedTime )
		stat.total = stat.total + elapsedTime
	else
		stat = {
			sql = sql,
			count = 1,
			min = elapsedTime,
			max = elapsedTime,
			total = elapsedTime,
		}
		self._PERF_stmtStats[ sql ] = stat
	end
	
	self._PERF_gatherStats = true

	return ...

end

--------------------------------------------------------------------------------

local function profileQuery( self, sql, methodName, ... )

	self._PERF_gatherStats = false

	queryStartTime = debug.timeNow()

	return finish_profileQuery( self, sql, self[ methodName ]( self, sql, ... ) )

end

--------------------------------------------------------------------------------

local function finish_execOnce( self, sql, ... )

	killStatement( self, sql )
	
	return ...

end

--------------------------------------------------------------------------------

--[[!
	Executes an SQL query exactly once.

	\param sql (string) The query.
	\param ... Any additional parameters needed for the query.

	\return Multiple results, with each result representing one row as an array.
		Because this is limited by the size of the Lua stack, you should
		use  \c :gatherArrayOfArray for a query that can return more than a few rows.
]]--

function databaseMethods:execOnce( --[[string]] sql, ... )

	self._log:info( "execOnce %s", sql )

	local stmt = getStatement( self, sql )
	stmt:bindByPosition( ... )
	
	return finish_execOnce( self, sql, stmt:runToCompletion() )

end

--------------------------------------------------------------------------------

--[[!
	Executes an SQL query, and caches the query so that it executes
	faster on subsequent calls.

	\param sql (string) The query.
	\param ... Any additional parameters needed for the query.

	\return Multiple results, with each result representing one row as an array.
		Because this is limited by the size of the Lua stack, you should
		use  \c :gatherArrayOfArray for a query that can return more than a few rows.

]]--

function databaseMethods:exec( --[[string]] sql, ... )

	self._log:info( "exec %s", sql )

	if self._PERF_gatherStats then
		return profileQuery( self, sql, 'exec', ... )
	end

	local stmt = getStatement( self, sql )
	self._bindByPositionMethod( stmt, ... )
	
	return self._runToCompletionMethod( stmt )
	
end

--------------------------------------------------------------------------------

--[[!
	Executes an SQL query and returns the result as an array of arrays.

	\param sql (string) The query.
	\param ... Any additional parameters needed for the query.

	\return The result as an array of arrays, where each contained array
		represents one row.


]]--

function databaseMethods:gatherArrayOfArray( --[[string]] sql, ... )

	self._log:info( "gatherArrayOfArray %s", sql )

	if self._PERF_gatherStats then
		return profileQuery( self, sql, 'gatherArrayOfArray', ... )
	end

	local stmt = getStatement( self, sql )
	self._bindByPositionMethod( stmt, ... )

	return self._gatherArrayOfArrayMethod( stmt )

end

--------------------------------------------------------------------------------

--[[!
       Retrieves the most recently inserted row.

       \return The row ID.
]]--

function databaseMethods:getLastInsertRowId()

	self._log:info( "getLastInsertRowId" )

	local db = assertDatabaseNotClosed( self )
	return db:getLastInsertRowID()

end

--------------------------------------------------------------------------------

--[[!
	Sets the change counter for this connection.

	\param changeCounter (number) The new value.

]]--

function databaseMethods:setChangeCounter( --[[number]] changeCounter )

	self._log:info( "setChangeCounter %s", changeCounter )

	local db = assertDatabaseNotClosed( self )
	db:setChangeCounter( changeCounter )

end

--------------------------------------------------------------------------------

--[[!
	Retrieves the change counter for this connection.

	\return (number) The current counter value.


]]--

function databaseMethods:getChangeCounter()

	self._log:info( "getChangeCounter" )

	local db = assertDatabaseNotClosed( self )
	return db:getChangeCounter()

end

--------------------------------------------------------------------------------

local function finish_withTransactionDo( self, success, ... )

	self._log:info( "finish_withTransactionDo success %s", success )

	self._hasActiveTransaction = false

	if success then
		self:exec "COMMIT"
		return ...
	else
		self:exec "ROLLBACK"
		error( ... )
	end

end

--------------------------------------------------------------------------------

--[[!
       Applies a function within a single database transaction.
       
       <ul><li>Do not call this from within a function that is already being called by
       \c withTransactionDo  (transactions cannot be nested). </li>
       
       <li>You cannot use \c coroutine.yield within this call.</li></ul>

       \param func The function, which calls functions such as \c :exec to execute
		SQL queries within the transaction, and returns normally to commit.
		If the function returns an error, the transaction is rolled back.

       \param ... Any additional parameters for the function.

       \return The result.
]]--

function databaseMethods:withTransactionDo( --[[fn]] func, ... )

	self._log:info( "withTransactionDo" )

	if self._hasActiveTransaction then
		error( "KSDatabase:withTransactionDo: nested transaction disallowed", 2 )
	end

	self._hasActiveTransaction = true
	self._PERF_queriesInTransaction = {}
	
	self:exec "BEGIN TRANSACTION"
	
	return finish_withTransactionDo( self, pcall( func, ... ) )
	
end

--------------------------------------------------------------------------------

--[[!
       Collects performance data for the connection.
]]--

function databaseMethods:PERF_collectQueryPerformanceStats()

	self._PERF_gatherStats = true
	
	--debug.print( "Database for path", self._path, "configured to gather perf data" )
	
	self._PERF_stmtStats = {}

end

--------------------------------------------------------------------------------

local function appendExplainResults( self, eqpPrefix, sql, output )

	local eqpSql = string.format( "%s %s", eqpPrefix, sql )
	local eqpRows = self:gatherArrayOfArray( eqpSql )

	killStatement( self, eqpSql )
	
	if #eqpRows > 0 then

		local eqpRowsFormatted = {}
		for q = 1, #eqpRows do
			eqpRowsFormatted[ q ] = table.concat( eqpRows[ q ], " | " )
		end
		
		local eqpSummary = table.concat( eqpRowsFormatted, sqlContinuationIndent )
		
		output[ #output + 1 ] = string.format( "%s%s\n\n", sqlContinuationIndent, eqpSummary )
	
	else
	
		output[ #output + 1 ] = "\n"
	
	end

end

--------------------------------------------------------------------------------

--[[!
      Generates a pretty-print report on performance data.

      \return (table) The report.
]]--

function databaseMethods:PERF_generateQueryStatsReport()

	if not self._PERF_gatherStats then return end
	
	local output = {
		"\nQUERY PROFILE SUMMARY FOR DATABASE ", self._path,
		"\n",
		"\n   Count       Min       Max       Total  SQL",
		"\n--------  --------  --------  ----------  -------------",
		"\n\n",
	}

	local sortedStats = {}
	local n = 0
	
	for k, v in pairs( self._PERF_stmtStats ) do
		if not v.sql:match( "^EXPLAIN " ) then
			n = n + 1
			sortedStats[ n ] = v
		end
	end
	
	table.sort( sortedStats, function( a, b )
	
		if a.total ~= b.total then
			return a.total > b.total
		else
			return a.sql < b.sql
		end
	
	end )

	for i = 1, #sortedStats do

		local stat = sortedStats[ i ]
		
		local sql = stat.sql:gsub( "\n%s+", sqlContinuationIndent )
		
		output[ #output + 1 ] =
			string.format(
				"%8.0f  %8.3f  %8.3f  %10.3f  %s\n",
					stat.count,
					stat.min,
					stat.max,
					stat.total,
					sql )
		
		if i <= 10 and not stat.sql:match( "^COMMIT" ) then
			appendExplainResults( self, "EXPLAIN", stat.sql, output )
			appendExplainResults( self, "EXPLAIN QUERY PLAN", stat.sql, output )
		else
			output[ #output + 1 ] = '\n'
		end

	end
	
	return table.concat( output )

end

--------------------------------------------------------------------------------

--[[!
	Reports  performance data to the debugger.

 ]]--

function databaseMethods:PERF_dumpQueryPerformanceStats()

	if not self._PERF_gatherStats then return end
	
	debug.print( self:PERF_generateQueryStatsReport() )

	self._PERF_gatherStats = false

end

--------------------------------------------------------------------------------

--[[!

	Reports performance data to the diagnostic log.

]]--

function databaseMethods:sendDiagnosticLogData()

	if not self._PERF_gatherStats then return end

	-- FIXME [scouten 2012-05-31]: Very hasty translation of console log format
	-- to HTML. Maybe change this to "real" HTML someday?

	local rawReport = self:PERF_generateQueryStatsReport()
	
	local htmlReport = string.format( "<pre>%s</pre>",
									  KSStringUtils.escapeHtml( rawReport ) )

	local idSafeThreadName = WFCurrentThread.DEBUG_getThreadName()
	idSafeThreadName = idSafeThreadName:gsub( ' :: ', '+' )

	WFDiagnosticLog.addSectionToLog {

		id = string.format( 'KSDatabase-%s-%s',
							self._path,
							idSafeThreadName ),
		
		name = string.format( "Database Performance: %s",
							  WFCurrentThread.DEBUG_getThreadName() ),
		
		weight = 2000,
		
		body = htmlReport,

	}

end

--------------------------------------------------------------------------------

--[[!
	Retrieves the path of this database as a display string.

	\return The display string.
]]--

function databaseMethods:__tostring()

	return string.format( "KSDatabase( %s )", self._path )

end

--------------------------------------------------------------------------------

WFCloseSupport.addCloseSupportMethods( databaseMethods )

--------------------------------------------------------------------------------

return KSClass 'KSDatabase'( databaseMethods )
