--[[----------------------------------------------------------------------------

KSDatabaseSchema.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

SCHEMA DEFINITION
`````````````````

The typical usage pattern is that the event processor init script will create
a database schema description by calling

	local schema = KSDatabaseSchema()
	
and adding schema elements (tables and indices) to it by calling

	schema:addTable()
	
and similar methods.

GRUNGE [scouten 2011-02-18]: Finish writing this documentation once the API stabilizes.

--------------------------------------------------------------------------------

@author scouten 2011-02-18

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'

	-- wichitafoundation.sqlite.database
local WFSqliteUtils = require 'WFSqliteUtils'


--============================================================================--

-- @mark schema table class

local schemaTableMethods = {

	_tableName = false,
		-- (string) table name

	_sql = false,
		-- (string) SQL needed to create this table
	
	_onConflict = false,
		-- (optional, string) what to do if this table exists and schema doesn't match
			-- "replace": delete existing table and start over
			-- (nil): raise an error

}

--------------------------------------------------------------------------------

function schemaTableMethods:__init( sql )

	local tableName = WFSqliteUtils.tableNameFromCreateTableStatement( sql )

	if not tableName or #tableName < 1 then
		error( "KSDatabaseSchema:addTable: could not find table name in " .. sql, 5 )
	end
	
	self._tableName = tableName
	self._sql = sql

end

--------------------------------------------------------------------------------

function schemaTableMethods:onConflict( conflictPolicy )

	if conflictPolicy == 'replace' then

		self._onConflict = conflictPolicy

	else
	
		error( "KSDatabaseSchema:addTable:onConflict: invalid policy " .. tostring( conflictPolicy ), 2 )
	
	end

	return self

end

--------------------------------------------------------------------------------

local schemaTableClass = KSClass 'KSDatabaseSchemaTable'( schemaTableMethods )


--============================================================================--

-- @mark schema index class

local schemaIndexMethods = {

	_indexName = false,
		-- (string) index name

	_sql = false,
		-- (string) SQL needed to create this index
	
}

--------------------------------------------------------------------------------

function schemaIndexMethods:__init( sql )

	local _, _, indexName = sql:find( "([%a_][%w_]*)%s+[Oo][Nn]%s+" )
	
	if not indexName or #indexName < 1 then
		error( "KSDatabaseSchema:addIndex: could not find index name in " .. sql, 5 )
	end
	
	self._indexName = indexName
	self._sql = sql

end

--------------------------------------------------------------------------------

local schemaIndexClass = KSClass 'KSDatabaseSchemaIndex'( schemaIndexMethods )


--============================================================================--

-- @mark KSDatabaseSchema class
--[[!

	\class KSDatabaseSchema
	\brief Represents a database schema.
	
		A worker thread can create an object of this type and use its
		methods to add schema elements (tables and indices).
		Use the schema to create an SQLite database instance.

		\sa KSDatabase

]]--

local schemaMethods = {

	_tables = false,
		-- (table)
			-- key: (string) table name
			-- value: (KSSchemaTable) info about this table definition
	
	_indexes = false,
		-- (table)
			-- key: (string) index name
			-- value: (KSSchemaIndex) info about this index definition
	
	_useWAL = false,
		-- (boolean) true if WAL mode was requested
	
	_currentVersion = false,
		-- (string) current schema version (typically app version)
	
	_backwardCompatibleToVersion = false,
		-- (string) set to < _currentVersion to assert compatibility with
			-- older software
			-- older software will fail to launch if it's version is < this
	
}

--------------------------------------------------------------------------------

function schemaMethods:__init( currentVersion, backwardCompatibleToVersion )

	self._tables = {}
	self._indexes = {}

	self._currentVersion = currentVersion or '0'
	self._backwardCompatibleToVersion = backwardCompatibleToVersion or currentVersion or '0'
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves the version information with which this schema was created.

	\return Multiple values: the current version, and the earliest version
		with which this one is backward compatible.

]]--

function schemaMethods:currentAndBackwardCompatibleVersions()

	return self._currentVersion, self._backwardCompatibleToVersion

end

--------------------------------------------------------------------------------

--[[!
	\brief Enables the write-ahead feature of SQLite.

]]--

function schemaMethods:useWAL()

	self._useWAL = true

	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a table to this schema.

	\param sql (string) An SQL statement that defines the new table.

	\return (table) The new table object.
]]--

function schemaMethods:addTable( --[[string]] sql )

	local tableEntry = schemaTableClass( sql )
	self._tables[ tableEntry._tableName ] = tableEntry
	
	return tableEntry

end

--------------------------------------------------------------------------------

--[[!
	\brief Adds an index to this schema.

	\param sql (string) An SQL statement that defines the new index.

	\return The new index entry.

]]--

function schemaMethods:addIndex( --[[string]] sql )

	local indexEntry = schemaIndexClass( sql )
	self._indexes[ indexEntry._indexName ] = indexEntry
	
	return indexEntry

end

--------------------------------------------------------------------------------

return KSClass 'KSDatabaseSchema'( schemaMethods )
