--[[----------------------------------------------------------------------------

KSDate.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@reviewed epoley 2011-03-07
@reviewed aksherry 2011-03-09

------------------------------------------------------------------------------]]

	-- core.os
local KSDateImp = require 'KSDateImp' -- intended only for use by this namespace


--============================================================================--
--[[!

This namespace provides cross-platform utilities for manipulating date-time values in Lua.

]]--

--@mark class KSDate

local KSDate = {}

--------------------------------------------------------------------------------
--[[!
	Retrieves the current date-time.

	\return A number of seconds.
]]--

KSDate.timeNow = KSDateImp.timeNow

--[[!
	Retrieves the current time in terms of clock ticks.

	\return A number of clock ticks.
]]--

KSDate.timeNowForAnimation = KSDateImp.timeNowForAnimation

--[[!
	Retrieves the time since the parent application was launched.

	\return A number of seconds.
]]--

KSDate.timeIntervalSinceAppLaunch = KSDateImp.timeIntervalSinceAppLaunch

--[[!
	Retrieves the current date-time and reports it as an ISO 8601
	formatted string with components normalized to the time zone
	of device's local calendar setting.

	\return  An ISO 8601 Date string.
]]--

KSDate.currentDateTimeInLocalizedComponents = KSDateImp.currentDateTimeInLocalizedComponents

local decomposeTime = assert( KSDateImp.decomposeTime )

local KSDateImp_politeDate = assert( KSDateImp.politeDate )

--------------------------------------------------------------------------------

--[[!
     Retrieves the current or a specified date-time as a string with formatting options.

     	\param optTime If supplied, the date-time value to format. If nil, retrieves
			and formats the current time.
	\param optUseLocalTimeZone When true, uses the time zone of device's local calendar setting.
			If nil or false, uses GMT.
	\param formatString A Lua format string.

	\return The formatted date-time string.
]]--

local function getTime( optTime, optUseLocalTimeZone, formatString )

	-- Break time down into its components. If optTime is nil, we'll use current time.
	-- if optUseLocalTimeZone is true, we'll use the timezone set by the system, otherwise GMT.
	
	local year, month, day, hour, minute, second = decomposeTime( optTime, optUseLocalTimeZone )
	
	local time = string.format( formatString, year, month, day, hour, minute, second )
				
	-- Add the "Z" designator for UTC
	if not optUseLocalTimeZone then
		time = time .. "Z"
	end
	
	return time

end

--------------------------------------------------------------------------------

--[[!
     Retrieves the current or a specified date-time as an ISO 8601 Date string.

     	\param optTime If supplied, the date-time value to format. If nil, retrieves
			and formats the current time.
	\param optUseLocalTimeZone When true, uses the time zone of device's local calendar setting.
			If nil or false, uses GMT.
 
	\return The formatted date-time string.
]]--

function KSDate.timeInIso8601Format( optTime, optUseLocalTimeZone )

	return getTime( optTime, optUseLocalTimeZone, "%04d-%02d-%02dT%02d:%02d:%06.3f" )

end

--------------------------------------------------------------------------------

--[[!
     Retrieves the current or a specified date-time as an ISO 8601 Date string, rounding
     to the nearest whole second.

     	\param optTime If supplied, the date-time value to format. If nil, retrieves
			and formats the current time.
	\param optUseLocalTimeZone When true, uses the time zone of device's local calendar setting.
			If nil or false, uses GMT.
 
	\return The formatted date-time string.

]]--

function KSDate.timeInIso8601FormatNearestSecond( optTime, optUseLocalTimeZone )

	return getTime( optTime, optUseLocalTimeZone, "%04d-%02d-%02dT%02d:%02d:%02d" )

end

--------------------------------------------------------------------------------

--[[!
     Retrieves the current time zone of the device's local calendar setting.
  
	\return The time zone in ISO 8601 format.

]]--

function KSDate.timeZoneInIso8601Format()

	return KSDateImp.timeZoneInIso8601Format()

end

--------------------------------------------------------------------------------

--[[!
     Retrieves the current or a specified date-time as an ISO 8601 Date string,
     with a time zone component from the device's local calendar setting.

     	\param optTime If supplied, the date-time value to format. If nil, retrieves
			and formats the current time.
  
	\return The formatted date-time string.
]]--

function KSDate.timeInIso8601FormatWithTimeZone( optTime )

	return KSDate.timeInIso8601FormatNearestSecond( optTime, true )
		.. KSDate.timeZoneInIso8601Format()
	
end

--------------------------------------------------------------------------------
--[[!
   Truncates an IS08601 date-time string, returning the first 19 characters and discarding
   any subsecond time and timezone information.

   \param dateTime The IS08601 date-time string.
   
   \return An IS08601 date-time string in the format 'YYYY-MM-DDTHH:mm:ss'.
]]--

function KSDate.splitIso8601DateUptoSeconds( dateTime )

	return dateTime:sub( 1, 19 ), string.sub( dateTime, 20 )

end

--------------------------------------------------------------------------------

local politeDateCacheSize = 100

-- Caching for polite dates based on the whether the value is a timestamp
-- or the formatting style specified.
-- Dates that don't fall into the above categories are cached in the default cache

local politeDateCache  = {}

local cachedPoliteDateKeys = {
	default = true,
	medium = true,
	long = true,
}

local KSMemoryCache -- lazy loaded


--------------------------------------------------------------------------------
 
--[[!
	Creates an ISO 8601 date string from component values and caches the new
	string.

	\param y The numeric year component.
	\param m The numeric month component.
	\param d The numeric day component.
	\param optStyle	The format style, 'medium' (includes a time with no seconds component)
			or 'long'.

	\return The new string formatted in the given style.

]]--

function KSDate.politeDate( y, m, d, optStyle )

	local key = string.format( '%s-%s-%s', y, m, d )
	
	local cacheType = 'default'

	if optStyle then
		cacheType = optStyle
	end
	
	if not KSMemoryCache then
		KSMemoryCache = assert( require 'KSMemoryCache' )
	end
	
	if cachedPoliteDateKeys[ optStyle ] and not politeDateCache[ cacheType ] then
		politeDateCache[ cacheType ] = KSMemoryCache( politeDateCacheSize )
	end
	
	-- If this date is cached, return the cached polite date.

	local result = politeDateCache[ cacheType ] and politeDateCache[ cacheType ]:getData( key )
	
	if result then
		return result
	end

	-- If not, get the formatted date and write it to cache.

	result = KSDateImp_politeDate( y, m, d, nil, nil, nil, optStyle )
	
	if politeDateCache[ cacheType ] then
		politeDateCache[ cacheType ]:add( key, result )
	end
	
	return result

end

--------------------------------------------------------------------------------
 
--[[!
	Formats a timestamp value to an ISO 8601 date string.
	Does not cache the formatted values (as is done by politeDate)

	\param timestamp The numeric timestamp value.
	\param optDateStyle The format style for the date component.
	\param optTimeStyle The format style for the time component.

	\return The formatted string.

]]--

function KSDate.politeDateFromTimestamp( timestamp, optDateStyle, optTimeStyle )

	local format = "(%d+)-(%d+)-(%d+)T(%d+):(%d+):(%d+)"

	local y, m, d, hh, mm, ss = timestamp:match( format )
	return KSDateImp_politeDate( y, m, d, hh, mm, ss, optDateStyle, optTimeStyle )

end

--------------------------------------------------------------------------------

--[[!
      Adds days to a date-time.

      \param optTime The date-time value, an ISO 8601 Date string.
       \param optUseLocalTimeZone When true, normalizes the date-time to
       		the time zone of device's local calendar setting.
      \param daysToAdd The number of days to add.
      
      \return The new date-time.

]]--

function KSDate.addDaysInIso8601Format( optTime, optUseLocalTimeZone, daysToAdd )

	local year, month, day, hour, minute, second = KSDateImp.addDays( optTime, optUseLocalTimeZone, daysToAdd )

	local formatString = '%04d-%02d-%02dT%02d:%02d:%06.3f'
	local time = string.format( formatString, year, month, day, hour, minute, second )
	-- Add the "Z" designator for UTC
	if not optUseLocalTimeZone then
		time = time .. "Z"
	end

	return time
end
--------------------------------------------------------------------------------
 
--[[!
     Reports the difference between two dates in days, ignoring the time components.

       \param startDate The first date-time value, an ISO 8601 date string.
       \param endDate The second date-time value, an ISO 8601 date string.
 
       \return A number of days, negative if the first day is later than the second day.


]]--

function KSDate.differenceInDays( startDate, endDate )

	return KSDateImp.diffInDays( startDate, endDate )
	
end

--------------------------------------------------------------------------------
 
--[[!
     Reports the difference between two date-times in seconds.

       \param startDate The first date-time value, an ISO 8601 date string.
       \param endDate The second date-time value, an ISO 8601 date string.
       \param optUseLocalTimeZone When true, normalizes both date-times to
       		the time zone of device's local calendar setting.

       \return A number of seconds, negative if the first time is later than the second time.

]]--

function KSDate.differenceInSeconds( startDate, endDate, optUseLocalTimeZone )

	return KSDateImp.diffInSeconds( startDate, endDate, optUseLocalTimeZone )

end

--------------------------------------------------------------------------------

 
--[[!
	Reports the difference between two date-times, for each component
	(year, month, day, hour, minute, second). For example, if the
	two dates differ by 3 seconds, all other components have a difference
	of 0. Differences are computed relative to the device's local calendar
	setting.

       \param startDate The first date-time value, an ISO 8601 date string.
       \param endDate The second date-time value, an ISO 8601 date string.


	\return	 First return value is true if the difference is negative,
		meaning that the first date is later than the second, false otherwise.
		The next six return values are the differences in each date component
		in order: year, month, day, hour, minute, second.

]]--

function KSDate.localizedDiffInAllComponents( startDate, endDate )
	
	return KSDateImp.localizedDiffInAllComponents( startDate, endDate )
	
end

--------------------------------------------------------------------------------
 
--[[!
      Adds seconds to a date-time.

      \param date The numeric date-time value.
      \param seconds The number of seconds to add.
      
      \return The new date-time.
]]--

function KSDate.addTimeInSeconds( date, seconds )

	return KSDateImp.addTimeInSeconds( date, seconds )

end

--------------------------------------------------------------------------------

--[[!
    Formats a numeric date-time value in the local timezone into ISO8601 format.

    \param date The numeric date-time value.

    \return The formatted date-time.
]]--

function KSDate.formatDateTimeAsLocal( date )
	
	return KSDateImp.formatDateTimeAsLocal( date )
	
end

--------------------------------------------------------------------------------

--[[!
    Formats a numeric date-time value in the UTC timezone into ISO8601 format.

    \param date The numeric date-time value.

    \return The formatted date-time.
]]--

function KSDate.formatDateTimeAsUTC( date )
	
	return KSDateImp.formatDateTimeAsUTC( date )
	
end

--------------------------------------------------------------------------------

return KSDate
