--[[----------------------------------------------------------------------------

KSDirectoryEnumeratorTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author epoley 2011-04-07

------------------------------------------------------------------------------]]

local fileUtilsTest = {}

local pi = require 'WFPlatformInfo'

if ( not pi.isiOS and not pi.isAndroid ) then

	local KSPathUtils = require 'KSPathUtils'
	local KSDirectoryEnumerator = require 'KSDirectoryEnumerator'
	
	
	--------------------------------------------------------------------------------
	
	
	function fileUtilsTest.testDirectoryEnumeratorJustFiles()
	
		
		local filesFound = {}
				
		local path = KSPathUtils.getStandardFilePath( "testdata" )
		path = KSPathUtils.child( path, "ksfileutils" )
		local enum = KSDirectoryEnumerator.create( path, true )
		local curFile = KSDirectoryEnumerator.nextFile( enum )
		
		while ( curFile ) do
			table.insert( filesFound, KSPathUtils.leafName( curFile ) )
			curFile =  KSDirectoryEnumerator.nextFile( enum )
		end
		
		local expectedFiles = {
			"FileInDotDirectory.txt",
			"file.txt",
			"FooFile.txt",
			"File With Spaces.txt",
			"File_With_Very_Long_Extension.superlongfiletypeextension",
			"noextensionfile",
			"BarFile.txt",
			"BaxFile.txt",
			"file-2.txt",
			"file-3.txt"
		}
		
		assert( #expectedFiles == #filesFound, "Number of files found did not match expected value" )
	
		-- Not the most efficient algorithm, but it allows us to avoid the sticky mess of
		-- trying to extract just the file name from the path. We only check that the
		-- paths the enumerator gave us end with the expected filenames, but this is
		-- quadratic if we cannot assume the order of those it found. (Nor do I feel it
		-- necessary to write a binary search for an expectedFiles list of 7 items...)
		
		for _, item in pairs( filesFound ) do
			
			local foundMatch = false
			
			for _, expectedFile in pairs( expectedFiles ) do
			
				-- compare expectedFile with item
				if expectedFile.upper == item.upper then
				
					foundMatch = true
					break
				
				end
				
			end
			
			assert( foundMatch, "Enumerator found an unexpected file: " .. item )
			
		end
	end
	
	--------------------------------------------------------------------------------
	
	function fileUtilsTest.testRecursiveFilesIteratorFunction()
		
		local filesFound = {}
				
		local path = KSPathUtils.getStandardFilePath( "testdata" )
		path = KSPathUtils.child( path, "ksfileutils" )
		local KSRecursiveFilesIterator = require 'KSRecursiveFilesIterator'
		
		for curFile in KSRecursiveFilesIterator( path, true ) do
			table.insert( filesFound, KSPathUtils.leafName( curFile ) )
		end
		
		local expectedFiles = {
			"FileInDotDirectory.txt",
			"file.txt",
			"FooFile.txt",
			"File With Spaces.txt",
			"File_With_Very_Long_Extension.superlongfiletypeextension",
			"noextensionfile",
			"BarFile.txt",
			"BaxFile.txt",
			"file-2.txt",
			"file-3.txt"
		}
		
		assert( #expectedFiles == #filesFound, "Number of files found did not match expected value" )
	
		for _, item in pairs( filesFound ) do
			
			local foundMatch = false
			
			for _, expectedFile in pairs( expectedFiles ) do
			
				-- compare expectedFile with item
				if expectedFile.upper == item.upper then
				
					foundMatch = true
					break
				
				end
				
			end
			
			assert( foundMatch, "Enumerator found an unexpected file: " .. item )
			
		end
	end
	
	--------------------------------------------------------------------------------
	
	
	function fileUtilsTest.testDirectoryEnumeratorAllItems()
	
		local filesFound = {}
				
		local path = KSPathUtils.getStandardFilePath( "testdata" )
		path = KSPathUtils.child( path, "ksfileutils" )
		local enum = KSDirectoryEnumerator.create( path, true )
		local curItem = KSDirectoryEnumerator.next( enum )
		
		--halt()
		while ( curItem ) do
			table.insert( filesFound, KSPathUtils.leafName( curItem ) )
			curItem =  KSDirectoryEnumerator.next( enum )
		end
		
		local expectedFiles = {
			-- files
			"FileInDotDirectory.txt",
			"File.txt",
			"FooFile.txt",
			"File With Spaces.txt",
			"File_With_Very_Long_Extension.superlongfiletypeextension",
			"noextensionfile",
			"BarFile.txt",
			"BaxFile.txt",
			"file-2.txt",
			"file-3.txt",
			
			-- folders
			"foo",
			"baz",
			"bar",
			"directory.with.dots.in.name"
		}
		
		assert( #expectedFiles == #filesFound, "Number of files found did not match expected value" )
	
		-- Not the most efficient algorithm, but it allows us to avoid the sticky mess of
		-- trying to extract just the file name from the path. We only check that the
		-- paths the enumerator gave us end with the expected filenames, but this is
		-- quadratic if we cannot assume the order of those it found. (Nor do I feel it
		-- necessary to write a binary search for an expectedFiles list of 7 items...)
		
		for _, item in pairs( filesFound ) do
			
			local foundMatch = false
			
			for _, expectedFile in pairs( expectedFiles ) do
							
				if expectedFile.upper == item.upper then
				
					foundMatch = true
					break
				
				end
				
			end
			
			assert( foundMatch, "Enumerator found an unexpected file: " .. item )
			
		end
	
	end

--------------------------------------------------------------------------------


--------------------------------------------------------------------------------

-- FIXME [epoley 2011-04-07]: Add a unit test for the currently unimplemented
-- recursiveFilesAndDirectories

--------------------------------------------------------------------------------
end
return fileUtilsTest


