--[[----------------------------------------------------------------------------

KSFileUtils.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author epoley 2011-04-07

--------------------------------------------------------------------------------

WARNING [scouten 2011-08-05]: Some of the functions in this namespace are only
available on Mac. (May no longer be true ...)

FIXME [scouten 2011-08-05]: In Lightroom, we had separate utils for file manipulation
and path manipulation. I would prefer that we have that same distinction here.
Not worth that additional refactoring while in crunch mode, though.

------------------------------------------------------------------------------]]

	-- core.os
local KSPathUtils = require 'KSPathUtils'

--============================================================================--

-- @mark KSFileUtils namespace

--[[!

\class KSFileUtils

\brief This namespace provides cross-platform file-manipulation utilities.

]]--


local KSFileUtils = {}

--------------------------------------------------------------------------------


local string_match = string.match
local string_lower = string.lower
local string_find = string.find
local string_sub = string.sub

local type = type

local WFPlatformInfo = require 'WFPlatformInfo'

--------------------------------------------------------------------------------
--[[!
	Creates a unique file name for a file to be created at an existing path.

	\param path (string) The absolute path to an existing folder.

	\return (string) The file name.
]]--

function KSFileUtils.getUniqueFileName( --[[string]] path )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Copies a file from one path to another. If a file already exists at the
	destination path, chooses a new name at the destination so as not to overwrite
	the existing file.

	\param source (string) The absolute path to the file to copy.
	\param destination (string) The absolute path to the destination folder.
	
	\return (string) The complete path to the new file.
]]--

function KSFileUtils.copyToUniquelyNamedFile( --[[string]] source, --[[string]] destination )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Moves a file from one path to another. If a file already exists at the
	destination path, chooses a new name at the destination so as not to overwrite
	the existing file.
	
	\param source (string) The absolute path to the file to move.
	\param dest (string) The absolute path to the destination folder.
	
	\return (string) The complete path to the file in its new folder.
]]--

function KSFileUtils.moveToUniquelyNamedFile( --[[string]] source, --[[string]] dest )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Reports whether a folder or file exists.

	\param path (string) The path to confirm.
	
	\return (string) "directory" or "file" if the path references an
		existing folder or file, otherwise (boolean) false.
]]--

function KSFileUtils.exists( --[[string]] path )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Copies a file.

	\param src (string) The absolute path of the source file.
	\param dest (string) The absolute path of the destination.
	
	\return (boolean) true if successful, false and a message string if not.
]]--

function KSFileUtils.copy( --[[string]] src, --[[string]] dest )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Moves a file.
	
	\param src (string) The absolute path of the source file.
	\param dest (string) The absolute path of the destination.
	
	\return (boolean) true if successful, false and a message string if not.
]]--

function KSFileUtils.move( --[[string]] src, --[[string]] dest )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Immediately deletes a file, or a folder with all of its contents.
	
	\warning Be careful with this. In most cases calling 'moveToTrash' would be better.

	\param path (string) The absolute path of a file or folder.
	
	\return (boolean) true if successful, false and a message string if not.
]]--

function KSFileUtils.delete( --[[string]] path )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Creates a new folder.

	\param path (string) The absolute path of the new folder.

	\return (boolean) true if successful, false and a message string if not.
]]--

function KSFileUtils.mkdir( --[[string]] path )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Sets the "do not back up" attribute on a file. Only meaningful on iOS; no-op
	on Mac OS.

	\param file (string) The absolute path of the file.

--]]

function KSFileUtils.setDoNotBackUpFlag( --[[string]] file )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--------------------------------------------------------------------------------
--[[!
	Reports the size, creation data, and modification date of a file.

	\param path (string) The absolute path of the file.
 
 	\return (table) A table with the following entries:
		<ul><li>fileSize: the number of bytes </li>
		<li>fileCreationDate: the number of seconds since Jan 1, 2000 </li>
		<li>fileModificationDate: the number of seconds since Jan 1, 2000 </li>
		</ul>
		If the file does not exist or the OS calls to retrieve these values fails,
		returns an empty table.
	
]]--

function KSFileUtils.fileAttributesAtPath( --[[string]] path )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).

	--[[Note on dates: There are many choices of reference dates to use in a Date API
	(Unix Reference Date, Cocoa reference date, etc). I choose the Cocoa one
	because it is the only one that is normally expressed as a double precision number
	already. It is obviously possible to store the number of milliseconds since
	1970 as a floating point number, but typically Unix programs manipulate this
	quantity as an integer.	]]--

end

--------------------------------------------------------------------------------
--[[!
	Sets the file-system date attributes of a file.

	\param path (string) The absolute path of the file.
	\param attributes (table) A table that can have entries for
		<b>fileCreationDate</b> and <b>fileModificationDate</b>.
]]--

function KSFileUtils.setFileAttributesAtPath( --[[string]] path, --[[table]] attributes )
	-- This wrapper exists only for documentation purposes; real implementation
	-- comes from native code (see "Real implementation" below).
end

--==============================================================================
-- @mark Real implementation below this line
--==============================================================================

-- Replace with native implementation

local KSFileUtilsImp = require 'KSFileUtilsImp'
for k, v in pairs( KSFileUtils ) do KSFileUtils[ k ] = KSFileUtilsImp[ k ] end

--------------------------------------------------------------------------------

local stringWithAppropriateTrashFolderPath = ( not WFPlatformInfo.isiOS and not WFPlatformInfo.isAndroid )
					and assert( KSFileUtilsImp.stringWithAppropriateTrashFolderPath )

--------------------------------------------------------------------------------
--[[!
	Moves a file to the Trash or Recycle folder for its volume.
	
	\param filePath (string) The absolute path of the file.
	
	\return (boolean) true if successful, false and a message string if not.
	
	\warning In iOS, there is no such thing as a Trash folder, so this function
		performs an immediate deletion.
]]--

function KSFileUtils.moveToTrash( --[[string]] filePath )

	if WFPlatformInfo.isiOS or WFPlatformInfo.isAndroid then
		return KSFileUtils.delete( filePath )
	end
	
	if WFPlatformInfo.isWin7 then
		return KSFileUtils.delete( filePath, false )
			-- false indicates soft-delete
	end

	if not KSFileUtils.exists( filePath ) then
		return false, "file does not exist"
	end

	-- What is the correct path to the trash for the given file?
	-- (The path to the trash may differ depending on the Disk Volume, etc.)

	if not stringWithAppropriateTrashFolderPath then
		return false, "no trash folder available"
	end
	
	local trashPath = stringWithAppropriateTrashFolderPath( filePath )
	
	-- Is the file already already sitting in the trash? If so, not much point
	-- in moving it.
	
	local folder, file = KSFileUtils.splitFolderAndFile( filePath, true )

	if trashPath == folder then
		return true
	end
	
	if file == "" then
		-- this can happen if filePath is a directory path and ends in "/"
		file = KSPathUtils.leafName( folder )
	end

	local insideTrashPath = KSPathUtils.child( trashPath, file )

	KSFileUtils.moveToUniquelyNamedFile( filePath, insideTrashPath )

	return true

end

--==============================================================================

-- @mark path manipulation functions
	-- FIXME [scouten 2011-08-05]: IMHO these belong in a separate namespace
	-- like KSPathUtils.

--------------------------------------------------------------------------------

--[[!
	Reports whether a file has a given extension.
	
	\param filePath (string) The absolute path of the file.
	\param extension (string) The extension to match.
	
	\return (boolean) true if the file exists and has the extension,
		false if not.
]]--

function KSFileUtils.pathHasExtension( --[[string]] filePath, --[[string]] extension )

	if not filePath then
		return false
	end

	local pathExtension = string_match( filePath, "%.([%a%d]*)$" )
	
	pathExtension = pathExtension and string_lower( pathExtension )
	
	pathExtension = pathExtension or ""
	
	if type( extension ) == 'string' then
		return pathExtension == extension
	else
		return not not extension[ pathExtension ]
	end

end

--------------------------------------------------------------------------------

--[[!
	In Mac OS and iOS, splits a path into its file and folder components.
	Not available in Windows.
	
	\param filePath (string) The absolute path of the file.
	\param trimFolderSeparator (number) When supplied, trims the path
		to the given depth above the file.

		<em>how does this work?</em>
	
	\return The folder name and the file name.
]]--

function KSFileUtils.splitFolderAndFile( --[[string]] filePath, --[[number]] trimFolderSeparator )

	if WFPlatformInfo.isWinRT then
		error "This function does not work on windows"
	end
	
	local sepIndex = string_find( filePath, '%/[^%/]*$' )
	
	if not sepIndex then
		return "", filePath
	end
	
	local folderName, fileName
	
	if trimFolderSeparator then
		if sepIndex == 1 then
			folderName = ""
		else
			folderName = string_sub( filePath, 1, sepIndex - 1 )
		end
	else
		folderName = string_sub( filePath, 1, sepIndex )
	end
	
	if sepIndex == #filePath then
		fileName = ""
	else
		fileName = string_sub( filePath, sepIndex + 1 )
	end
	
	return folderName, fileName

end

--------------------------------------------------------------------------------

--[[!
	Reports how much space is available at a given path.
	
	\param path (string) The absolute path to report on.
 	
	\return  The number of bytes free, and the number of bytes used.
		((is this right?))
]]--

function KSFileUtils.checkFreeSpaceAtPath( --[[string]] path )

	local attrs = KSFileUtilsImp.fileSystemAttributesAtPath( path )

	local fileSystemFreeSize, fileSystemSize

	if attrs then
		fileSystemFreeSize = attrs.fileSystemFreeSize
		fileSystemSize = attrs.fileSystemSize
	end

	return fileSystemFreeSize, fileSystemSize
end

--------------------------------------------------------------------------------

if WFPlatformInfo.isWinRT then

	local move = KSFileUtils.move
	
	-- If successful, we move the file from tempFile to path. If unsuccessful,
	-- we delete. We then return the underlying value from pcall.
	
	local function finishDoForAuthorizedPath( tempFile, path, success, ... )
		if success then
			move( tempFile, path )
			return ...
		else
			KSFileUtils.delete( tempFile )
			error( ( ... ) )
		end
	end

	function KSFileUtils.doForAuthorizedPath( path, func )

		local tempDir = KSPathUtils.getStandardFilePath( "temp" )

		local leafName
		if type( path ) == 'string' then
			leafName = KSPathUtils.leafName( path )
		elseif type( path ) == 'table' then
			leafName = assert( path.filename )
		end
		
		local tempFile = KSPathUtils.child( tempDir, leafName )

		tempFile = KSFileUtils.getUniqueFileName( tempFile )
		
		return finishDoForAuthorizedPath( tempFile, path, pcall( func, tempFile ) )
		
	end
	
else

	function KSFileUtils.doForAuthorizedPath( path, func )
		if type( path ) == 'table' then
			local collisionBehavior = path.collisionBehavior
			path = KSPathUtils.child( path.folder, path.filename )
			if collisionBehavior == "GenerateUniqueName" then
				path = KSFileUtils.getUniqueFileName( path )
			elseif collisionBehavior == "FailIfExists" then
				assert( not KSFileUtils.exists( path ), "File already exists and collisionBehavior is FailIfExists" )
			end
		end
		return func( path )
	end
end

--------------------------------------------------------------------------------

return KSFileUtils
