--[[----------------------------------------------------------------------------

KSFileUtilsTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author epoley 2011-04-07

------------------------------------------------------------------------------]]

local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'
local KSTableUtils = require 'KSTableUtils'
local KSFileUtilsImp = require 'KSFileUtilsImp'
local WFPlatformInfo = require 'WFPlatformInfo'

--============================================================================--

local KSFileUtilsTest = {}

local runCount = 0
local dstPath

function KSFileUtilsTest.setUpClass()

	dstPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "carouselDocuments" ), "KSFileUtilsTestFolder" ) .. runCount
	runCount = runCount + 1
	pcall( KSFileUtils.mkdir, dstPath )

end

function KSFileUtilsTest.tearDownClass()

	pcall( KSFileUtils.delete, dstPath )
	
end


--------------------------------------------------------------------------------

local function mustFail( ... )
	if pcall( ... ) then
		assert( false, "Expected to fail: " .. debug.traceback() )
	end
end

--------------------------------------------------------------------------------


function KSFileUtilsTest.testPathHasExtension()

	local KSTestUtils = require 'KSTestUtils'
	
	local ftc = KSTestUtils.functionTestCase
	
	local kJpegExtensions = { jpg = true, jpeg = true }
	
	local testCases = {
		ftc "Simple" ( "Foo/Baz.jpg", "jpg" )( true ),
		ftc "Set" ( "Foo/Baz.jpg", kJpegExtensions )( true ),
		ftc "Uppercase" ( "Foo/Baz.JPG", "jpg" )( true ),
		ftc "Set Uppercase" ( "Foo/Baz.JPG", kJpegExtensions )( true ),
		ftc "Failure" ( "Foo/Baz.tif", "jpg" )( false ),
		ftc "Set Failure" ( "Foo/Baz.tif", kJpegExtensions )( false ),
		ftc "No extension" ( "Foo/Baz", "jpg" )( false ),
		ftc "Set no extension" ( "Foo/Baz", kJpegExtensions )( false ),
		ftc "Empty target extension" ( "Foo/Baz", "" )( true ),
		ftc "Nil path" ( nil, "jpg" )( false ),
	}
	
	KSTestUtils.runFunctionTestCases( KSFileUtils.pathHasExtension, testCases )

end

--------------------------------------------------------------------------------

-- Disable test on WinRT -- jandrew
if not WFPlatformInfo.isWinRT then

	function KSFileUtilsTest.testSplitFolderAndFile()
		
		local KSTestUtils = require 'KSTestUtils'
		
		local ftc = KSTestUtils.functionTestCase
		
		local testCases = {
			ftc "Simple Split" ( "/Foo/Baz/Bax" )( "/Foo/Baz/", "Bax" ),
			ftc "No Folder"    ( "NoFolder"     )( "", "NoFolder"    ),
			ftc "No File"      ( "NoFile/"      )( "NoFile/", ""     ),
			ftc "Root Folder"  ( "/"            )( "/", ""           ),
			ftc "Empty Path"   ( ""             )( "", ""            ),
			ftc "Trim Split"        ( "/Foo/Baz/Bax", true )( "/Foo/Baz", "Bax" ),
			ftc "Trim No Folder"    ( "NoFolder", true     )( "", "NoFolder"    ),
			ftc "Trim No File"      ( "NoFile/", true      )( "NoFile", ""     ),
			ftc "Trim Root Folder"  ( "/", true            )( "", ""           ),
			ftc "Trim Empty Path"   ( "", true             )( "", ""            )
		}
		
		KSTestUtils.runFunctionTestCases( KSFileUtils.splitFolderAndFile, testCases )

	end

end

--------------------------------------------------------------------------------

function KSFileUtilsTest.testPathHasExtension()

	local KSTestUtils = require 'KSTestUtils'
	
	local ftc = KSTestUtils.functionTestCase
	
	local testCases = {
		ftc "Just filename" ( "filename.jpg", "jpg" )( true ),
		ftc "Filename in path" ( "/Users/epoley/filename.jpg", "jpg" )( true ),
		ftc "No File" ( "", "jpg" )( false ),
		ftc "Long extension" ( "/Users/epoley/filename.longextension", "longextension" )( true ),
		ftc "Number in extension" ( "/Users/epoley/filename.cr2", "cr2" )( true ),
		ftc "Period in filename" ( "/Users/epoley/filename.old.jpg", "jpg" )( true ),
		ftc "Many periods in filename" ( "/Users/epoley/filename.2010.03.07.old.jpg", "jpg" )( true ),
		ftc "Many periods in filename, extension inside" ( "/Users/epoley/filename.2010.03.07.jpg.txt", "jpg" )( false ),
		ftc "Periods in path" ( "/Users/epoley/foo.bar/filename.jpg", "jpg" )( true ),
		ftc "Extension starts with"  ( "filename.jpgfoo", "jpg" )( false ),
	}
	 
	KSTestUtils.runFunctionTestCases( KSFileUtils.pathHasExtension, testCases )

end

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-08-06]: When out of crunch mode, we should write a real
-- unit test that could be run repeatedly. This Q&D test will at least give you
-- a framework for running this func so you can manually observe its behavior.

--function KSFileUtilsTest.testDelete()
--
--	debug.print( "delete returned ",
--		KSFileUtils.delete( "/Users/scouten/Desktop/foo.jpg" ) )
--
--end

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-08-06]: When out of crunch mode, we should write a real
-- unit test that could be run repeatedly. This Q&D test will at least give you
-- a framework for running this func so you can manually observe its behavior.

--function KSFileUtilsTest.testMoveToTrash()
--
--	debug.print( "moveToTrash returned ",
--		KSFileUtils.moveToTrash( "/Users/scouten/Desktop/foo.jpg" ) )
--
--end

--------------------------------------------------------------------------------

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testExists()
	
		local KSTestUtils = require 'KSTestUtils'
		local basePath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		local ftc = KSTestUtils.functionTestCase
		
		local testCases = {
			ftc "basic name" ( KSPathUtils.child( basePath, "file.txt" ) ) ( "file" ),
			ftc "name with spaces" ( KSPathUtils.child( basePath, "File With Spaces.txt" ) ) ( "file" ),
			ftc "name with no extension" ( KSPathUtils.child( basePath, "noextensionfile" ) ) ( "file" ),
			ftc "Long extension" ( KSPathUtils.child( basePath, "File_With_Very_Long_Extension.superlongfiletypeextension" ) ) ( "file" ),
			ftc "in folder" ( KSPathUtils.child( basePath, KSPathUtils.child( "bar", "BarFile.txt" ) ) ) ( "file" ),
		}
	
		KSTestUtils.runFunctionTestCases( KSFileUtils.exists, testCases )
	
	end

end

--------------------------------------------------------------------------------

--[[
	this test tests two basic things, first that a known set of items are populated and two,
	that the file size if correct
	
	We are not testing the acutual values of the properties for dates as I think we can't
	get those values to be fixed

--]]

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testFileAttributesAtPath()
	
		local basePath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		local expectedKeys = { fileSize = true, fileCreationDate = true, fileCreationDateTimezoneOffset = true, fileModificationDate = true, fileModificationDateTimezoneOffset = true }
		local fileAttribs = KSFileUtils.fileAttributesAtPath( KSPathUtils.child( basePath, "file.txt" ) )
		local actualKeys = {}
		
		for k, v in pairs( expectedKeys ) do
		
			if fileAttribs[ k ] ~= nil then
				actualKeys[ k ] = true
			else
				error( "fileAttributesAtPath, missing a file attribute : " .. k  )
			end
		end
		
		assert( KSTableUtils.deepEquals( actualKeys, expectedKeys ), "fileAttributesAtPath, missing key" )
		assert( fileAttribs[ "fileSize" ] == 5, "fileAttributesAtPath, fileSize is wrong" )
		
	end

end


--------------------------------------------------------------------------------

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testFileSystemAttributesAtPath()
	
		local basePath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		local expectedKeys = { fileSystemSize = true, fileSystemFreeSize = true, fileSystemNumber = true }
		local fileAttribs = KSFileUtilsImp.fileSystemAttributesAtPath( basePath )
		local actualKeys = {}
		
		
		for k, v in pairs( expectedKeys ) do
		
			if fileAttribs[ k ] ~= nil then
				actualKeys[ k ] = true
			else
				error( "FileSystemAttributesAtPath, missing a file attribute : " .. k  )
			end
		end
		
		assert( KSTableUtils.deepEquals( actualKeys, expectedKeys ), "fileAttributesAtPath, missing key" )
		
	
	end

end


--------------------------------------------------------------------------------

function KSFileUtilsTest.testFileAttributesAtPath()

	local KSDate = require "KSDate"
	
	local testFile = KSPathUtils.child( KSPathUtils.getStandardFilePath( "temp" ), "testSetFileSystemAttributesAtPath.txt" )
	testFile = KSFileUtils.getUniqueFileName( testFile )
	local file = io.open( testFile, "w" )
	file:write( "KSFileUtilsTest.testSetFileSystemAttributesAtPath\n" )
	file:close()
	
	local bDate = "2003-03-08T00:30:00"
	local wDate = "2005-05-27T13:30:00"
	
	KSFileUtils.setFileAttributesAtPath( testFile, {
		fileCreationDate = bDate,
		fileModificationDate = wDate,
	} )
	
	local fileAttribs = KSFileUtils.fileAttributesAtPath( testFile )
	
	local modDate = assert( KSDate.formatDateTimeAsUTC( fileAttribs.fileModificationDate +  fileAttribs.fileModificationDateTimezoneOffset ), "formatDateTimeAsUTC returned nil" )
	assert( modDate == wDate, modDate .. " != " .. wDate )
		
	if (not WFPlatformInfo.isAndroid) then	-- Android does not support creation date.
		local createDate = assert( KSDate.formatDateTimeAsUTC( fileAttribs.fileCreationDate + fileAttribs.fileCreationDateTimezoneOffset ), "formatDateTimeAsUTC returned nil" )
		assert( createDate == bDate, createDate .. " != " .. bDate )
	end
	
	KSFileUtils.delete( testFile )

end

--------------------------------------------------------------------------------

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testCopyToUniquelyNamedFile()
	
		-- build source and destination paths
		
		local srcPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		-- do a baic copy, filename shouldn't change since file doesn't exist
		
		local dstFile = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( dstPath, "file.txt" ) )
		assert( dstFile == KSPathUtils.child( dstPath, "file.txt" ), "CopyToUniquelyNamedFile, no conflict name failed" )
		assert( KSFileUtils.exists( KSPathUtils.child( dstPath, "file.txt" ) ), "CopyToUniquelyNamedFile, copied file doesn't exist" )
		
		-- copy file to the same place as last time, this time, filename should change since a file already existed
		
		dstFile = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( dstPath, "file.txt" ) )
		assert( dstFile ~= KSPathUtils.child( dstPath, "file.txt" ), "CopyToUniquelyNamedFile, conflict name failed" )
		assert( KSFileUtils.exists( dstFile ), "CopyToUniquelyNamedFile, copied file doesn't exist" )
		
		-- one more copy using original name, this time make sure the result doesn't match either of the first two names
		
		local dstFile2 = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( dstPath, "file.txt" ) )
		assert( dstFile2 ~= KSPathUtils.child( dstPath, "file.txt" ), "CopyToUniquelyNamedFile, copied filenames shouldn't match" )
		assert( dstFile2 ~= dstFile, "CopyToUniquelyNamedFile, copied filenames shouldn't match" )
		assert( KSFileUtils.exists( dstFile2 ), "CopyToUniquelyNamedFile, copied file doesn't exist" )
	
	end

end

--------------------------------------------------------------------------------

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testMkDir()
		
		-- create new folder should work
		
		assert( pcall( KSFileUtils.mkdir, KSPathUtils.child( dstPath, "newFolder" ) ), "testMkDir, failed to create a new folder" )
	
		-- create a file and then create a folder using the same name, the folder creation should fail
		
		local basePath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( basePath, "file.txt" ), KSPathUtils.child( dstPath, "folderFail" ) )
		mustFail( KSFileUtils.mkdir, KSPathUtils.child( dstPath, "folderFail" ) )
	
	end

end


--------------------------------------------------------------------------------

if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testMoveToUniquelyNamedFile()
	
		local srcPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		-- first I have to get a file into a writable area.
		
		local srcFile = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( dstPath, "FileToMove.txt" ) )
		
		-- now try to move it, ensuring it doesn't already exist
		
		local dstFile = KSFileUtils.moveToUniquelyNamedFile( srcFile, KSPathUtils.child( dstPath, "MovedFile.txt" ) )
		assert( not KSFileUtils.exists( srcFile ), "testMoveToUniquelyNamedFile, source file should not exist" )
		assert( dstFile == KSPathUtils.child( dstPath, "MovedFile.txt" ), "testMoveToUniquelyNamedFile, moved filename is incorrect" )
		assert( KSFileUtils.exists( dstFile ), "testMoveToUniquelyNamedFile, moved file doesn't exist" )
	
		-- try the same move again, this time it should fail, source doesn't exist
		mustFail( KSFileUtils.moveToUniquelyNamedFile, srcFile, KSPathUtils.child( dstPath, "MovedFile.txt" ) )
		
		-- move the dst file onto itself, we should end up with dst file gone and a new filename
		-- this is based on current implementation, it seems likely we might want to check src == dst and do nothing
		
		local dstFile2 = KSFileUtils.moveToUniquelyNamedFile( dstFile, KSPathUtils.child( dstPath, "MovedFile.txt" ) )
		assert( dstFile ~= dstFile2, "testMoveToUniquelyNamedFile, filename should have changed" )
		assert( KSFileUtils.exists( dstFile2 ), "testMoveToUniquelyNamedFile, file not created" )
		assert( not KSFileUtils.exists( dstFile ), "testMoveToUniquelyNamedFile, source file should not exist" )
	
	end

end


--------------------------------------------------------------------------------

function KSFileUtilsTest.testStringWithAppropriateTrashFolderPath()

	-- FIXME [nicholas: 26 June oh-12] this wont be implemented on Windows, haven't looked at testing Mac
	
end

--------------------------------------------------------------------------------


if (not WFPlatformInfo.isiOS) then

	function KSFileUtilsTest.testDelete()
	
		local srcPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( "testdata" ), "ksfileutils" )
		
		-- first I have to get a file into a writable area.
		
		local srcFile = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( dstPath, "FileToDelete.txt" ) )
		
		-- test basic delete (file)
		
		local deleted, errorString = KSFileUtils.delete( srcFile )
		assert( not KSFileUtils.exists( srcFile ), "testDelete, file wasn't deleted" )
		assert( deleted, "testDelete, deleted the wrong file" )
		assert( errorString == nil, "testDelete, shouldn't have an error" )
		
		-- try to delete again to ensure we fail
		
		deleted, errorString = KSFileUtils.delete( srcFile )
		assert( errorString ~= nil, "testDelete, we should have gotten an error" )
		
		-- now create a folder to delete
		
		local folder = KSPathUtils.child( dstPath, "FolderToDelete" )
		KSFileUtils.mkdir( folder )
		assert( KSFileUtils.exists( folder ), "testDelete, folder deletion precondition failed" )
		
		-- and delete it
		
		deleted, errorString = KSFileUtils.delete( folder )
		assert( errorString == nil, "testDelete, empty folder failed to delete" )
		assert( not KSFileUtils.exists( folder ), "testDelete,folder wasn't deleted" )
		
		-- delete it again, we should fail this time
		
		deleted, errorString = KSFileUtils.delete( folder )
		assert( errorString ~= nil, "testDelete, non-existant folder deletion failed to report error" )
		
		--[[ now create a folder and put something into it
		
		FIXME [nicholas 28 June 12]  I am sick of this.  So on windows the problem is that when I copy a file into a folder
		then try to delete it, it fails to delete.  There is some timing issue as if I manually (from Windows Explorer) copy
		files into the folder and then let the delete everything works fine.  Or if I let the create part run one time
		then run the deletes on a second run, everything is also fine.  So the code being tested works, but there is
		something weird going on and I really need to move on.
		
		KSFileUtils.mkdir( folder )
		assert( KSFileUtils.exists( folder ), "testDelete, non-empty folder deletion precondition failed" )
		srcFile = KSFileUtils.copyToUniquelyNamedFile( KSPathUtils.child( srcPath, "file.txt" ), KSPathUtils.child( folder, "FileToDelete.txt" ) )
		assert( KSFileUtils.exists( srcFile ), "testDelete, non-empty folder deletion precondition failed" )
		
		-- try to delete the non empty folder
		
		deletedFileName, errorString = KSFileUtils.delete( folder )
		assert( errorString == nil, "testDelete, non-empty folder failed to delete" )
		assert( not KSFileUtils.exists( folder ), "testDelete, non-empty folder deletion failed" )
		
		-- delete it again, we should fail this time
		
		deletedFileName, errorString = KSFileUtils.delete( folder )
		assert( errorString ~= nil, "testDelete, non-existant non-empty folder deletion failed to report error" )
		--]]
	end

end





return KSFileUtilsTest
