--[[----------------------------------------------------------------------------

KSImportJpegServiceEPInit.lua

@author erwright 2011-03-31

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

FIXME [erwright 2011-05-5]: future work on import: we need to revisit the overall
import structure for handling assets from the asset library, data straight from
the camera sensor, and filesystem files on the desktop.  Are file paths and
camera sensor data two distinct import paths?  Or are file paths just a high-level
layer on top of a core data-import component?

------------------------------------------------------------------------------]]

local require = require

local WFCurrentThread = require 'WFCurrentThread'

local KSImportJpegService = require 'KSImportJpegService'

local KSFileUtils = require 'KSFileUtils'

local KSPathUtils = require 'KSPathUtils'

local KSDate = require 'KSDate'

local WFQueueing = require 'WFQueueing'

local WFCloseSupport = require 'WFCloseSupport'

local KSClass = require 'KSClass'

local KSData = require 'KSData'

local WFURI = require 'WFURI'

local KSTableUtils = require 'KSTableUtils'

local pairs = pairs
local type = type
local string = string

local WFCloseSupport_hasBeenOrWillBeClosed = WFCloseSupport.hasBeenOrWillBeClosed

local WFTimer = require 'WFTimer'

local WFPlatformInfo = require 'WFPlatformInfo'

local KSStringUtils = require 'KSStringUtils'

local ZSTR = require( 'WFZString' ).fromZString

local KSAssetsLibraryImportService = nil
local KSImportStats = require "KSImportStats"

local KSImportStats_addImportStatEntry = KSImportStats.addImportStatEntry
local KSImportStats_updateTotalQueued = KSImportStats.updateTotalQueued
local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt )

if WFPlatformInfo.isiOS then
	KSAssetsLibraryImportService = require 'KSAssetsLibraryImportService'
end

-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

-- libs.request
local WFRequestDispatcher = require 'WFRequestDispatcher'

local metricsLog = KSLogging.log 'KSImportJpegService :: metrics '
--        :setLevel( KSLogging.DEBUG )

local KSReactive = require 'KSReactive'

local KSPromises = require 'KSPromises'

local WFArguments = require 'WFArguments'

local WFArguments_check = WFArguments.check

local KSImportJpegServiceMetrics = require 'KSImportJpegServiceMetrics'
local adjustQueueEntryCounts = KSImportJpegServiceMetrics.adjustQueueEntryCounts
local getNumTotalQueueEntries = KSImportJpegServiceMetrics.getNumTotalQueueEntries
local getNumProcessedQueueEntries = KSImportJpegServiceMetrics.getNumProcessedQueueEntries
local resetQueueEntryCounts = KSImportJpegServiceMetrics.resetQueueEntryCounts

local COLLECT_IMPORT_METRICS = true

-- If enabled, we collect metrics on all import activities, which are then provided
-- upon request to WFDiagnosticLog. The metrics include data such as: total count,
-- counts by result (e.g. success, duplicate, error, canceled), total file size, total
-- megapixels, peak memory for the import EP before and after, as well as extra information
-- for large imports (100+ photos): every 10% increment we record a "checkpoint" that includes
-- much of the data above, timings broken down by import task, as well as WFDiagnosticLog-style
-- thread statistics on memory consumption, etc.

local threadStats = nil

local recentImportStats = assert( KSImportStats.recentImportStats )
local autoImportEnumerationPromises = setmetatable( { }, { __mode = 'kv' } )
--------------------------------------------------------------------------------

local KSImportUtils = require "KSImportUtils"

local getOzAccountService = KSImportUtils.makeLateRequire 'OzAccountService'
local getOzCatalogService = KSImportUtils.makeLateRequire 'OzCatalogService'

local contentTypeForFilePath = KSImportUtils.contentTypeForFilePath
local isValidFilePath = KSImportUtils.isValidFilePath

--------------------------------------------------------------------------------

local function OzAccountService_currentUserB_valueNow()
	return getOzAccountService().currentUserB():valueNow()
end

--------------------------------------------------------------------------------

local function OzCatalogService( ... )
	return getOzCatalogService()( ... )
end

--------------------------------------------------------------------------------

WFCurrentThread.setThreadPriority( '< low >' )

--------------------------------------------------------------------------------

local manualImportRequestIdToHandler = { }
local autoImportRequestIdToHandler = { }
	-- A mapping from request id to import handler object
	
local statusCallbackHandlers = { }

local didCancelAllImports = false

local autoImportIsRunning = false

local pauseImports = false

--------------------------------------------------------------------------------

local function sendStatusMessage( ... )

	for _, handler in pairs( statusCallbackHandlers ) do
		handler:send( ... )
	end
	
end

--------------------------------------------------------------------------------

local function openCatalogP( catalogId, user )
	
	return KSPromises.resolvedPromiseP( 'data', true )
		:mapDataP( function() return 'data', OzCatalogService( catalogId, user ) end )
			-- Capture errors in the promise structure

end

--------------------------------------------------------------------------------

local function withCatalogDoP( catalogId, user, fnP )

	return openCatalogP( catalogId, user ):usingDoP( fnP )

end

--------------------------------------------------------------------------------

if COLLECT_IMPORT_METRICS then
	
	local KSSystemNotifications = require 'KSSystemNotifications'
	threadStats = {}
	
	KSSystemNotifications.notificationStreamE( 'receiveThreadStats' )
	:onDataE( function( stats )

		threadStats[ #threadStats + 1 ] = stats

	end )
	
end

local function threadStatsSortFn( a, b )
	if a.currentSize ~= b.currentSize then
		return a.currentSize > b.currentSize
	else
		return a.threadName > b.threadName
	end
end

--------------------------------------------------------------------------------

-- Sorts (in place) the threadStats table and writes it to the given table

local function writeThreadStats( startTime, theTable )
	
	if recentImportStats[ startTime ] then
		
		table.sort( threadStats, threadStatsSortFn )
		
		theTable.threadStats = threadStats

	end

end

--------------------------------------------------------------------------------

local function doCancelAutoRequests()
	local didCloseAHandler = false
	for promise in pairs( autoImportEnumerationPromises ) do
		promise:close()
		didCloseAHandler = true
		autoImportEnumerationPromises[ promise ] = nil
	end
	return didCloseAHandler
end

--------------------------------------------------------------------------------

local function cancelImportRequests( cancelManualRequests, cancelAutoRequests, optForceCancellationHandling )

--debug.print( "cancelImportRequests called", cancelManualRequests, cancelAutoRequests )

	-- Note: We must set the didCancelAllImports flag before we go about
	-- closing import handlers so that any handlers remaining to be processed
	-- will be able to send back a canceled message instead of completed.
	-- The completion logic will reset this flag to false.
	
	-- But we need to gracefully handle the case that this command is executed
	-- when the queue is empty. If we don't end up closing any import handlers,
	-- then set the flag back ourselves.
	
	didCancelAllImports = true
	
	local didCloseAHandler = false
	
	-- GRUNGE [epoley 2011-11-30]: We can't cancel the enumeration request itself,
	-- so we use this map to signal that a cancellation has occurred. Any auto
	-- import enumeration callbacks that were alive when we receive this cancellation
	-- request get "canceled." When the callback gets called back, it will check
	-- its entry in this dictionary and no-op if it has been canceled.
	
	if cancelAutoRequests then
		if doCancelAutoRequests() then
			didCloseAHandler = true
		end
	end
	
	if cancelManualRequests then
		for _, handler in pairs( manualImportRequestIdToHandler ) do
			if handler then
				if COLLECT_IMPORT_METRICS then
					KSImportStats.updateStatsAfterCancelation( handler )
				end
				handler:close()
				didCloseAHandler = true
			end
		end
	end
	
	if cancelAutoRequests then
		for _, handler in pairs( autoImportRequestIdToHandler ) do
			if handler then
				if COLLECT_IMPORT_METRICS then
					KSImportStats.updateStatsAfterCancelation( handler )
				end
--				debug.print("closing auto-import handler")
				handler:close()
				didCloseAHandler = true
			end
		end
		
		autoImportIsRunning = false
--		debug.print( "Set autoimportisrunning = false upon cancellation" )
	end
		
	if not didCloseAHandler and not optForceCancellationHandling then
		didCancelAllImports = false
	end
	
end

--------------------------------------------------------------------------------

WFRequestDispatcher.onMessage ( 'dumpImportStats', 

	function()
	
		for k, v in pairs( recentImportStats ) do
			metricsLog:debug( "Import stats at %f", k )
			metricsLog:debug( v )
		end
		
	end )
	
--------------------------------------------------------------------------------

WFRequestDispatcher.onMessage ( 'cancelAllImportRequests', 

	function()
	
		cancelImportRequests( true, true )
	
	end )

--------------------------------------------------------------------------------

WFRequestDispatcher.onMessage ( 'cancelAllAutoImportRequests' , 
	function()
	
		cancelImportRequests( false, true )
	
	end )

--------------------------------------------------------------------------------

local KSStatusCallbackHandler = KSClass 'KSStatusCallbackHandler' {

	_requestId = false,
	_responseE = false,
	
	__init = function( self )
		self._requestId = #statusCallbackHandlers + 1
		self._responseE = KSReactive.sourceE()
		statusCallbackHandlers[ self._requestId ] = self
	end,
	
	close = function( self )
		statusCallbackHandlers[ self._requestId ] = nil
		self._responseE:post( 'done' )
		self._responseE:close()
	end,
	
	send = function( self, ... )
		self._responseE:post( ... )
	end

}

--------------------------------------------------------------------------------

WFRequestDispatcher.onMessage ( 'getStatusCallback', 
	function()
	
		local statusHandler = KSStatusCallbackHandler()
		return statusHandler._responseE

	end )

--------------------------------------------------------------------------------

local function addKeyword( imageInfo, keyword )

	if not keyword then
		return
	end

	local keywords = imageInfo.keywords
	
	if not keywords then
		keywords = { }
		imageInfo.keywords = keywords
	end
	
	keywords[ keyword ] = true

end

--------------------------------------------------------------------------------

local function makePlaceholderFilename()
	
	local y, m, d = KSDate.currentDateTimeInLocalizedComponents()
	
	local dateTimeNow = string.format( "%04d-%02d-%02d", y, m, d )
		
	return ZSTR( "$$$/Import/PhotoPlaceholderFilename=Adobe_Revel_Photo_^1^2", dateTimeNow, ".jpg" )

end

--------------------------------------------------------------------------------

-- Validates that the date string is of the format "YYYY-MM-DD'T'HH:MM:SS*" 
-- (* represents any additional data (e.g. subseconds, time zone offset) that is accepted
-- but ignored by Revel). A few minor non-conformities are corrected--such as ':' or '.'
-- being used as a delimiter. The corrected (or already valid) string is returned.
-- If the string cannot conform to the format, returns nil.

local function validateDateFormat( dateString, hasTimeZoneInfo )
	
	if not dateString or type( dateString ) ~= "string" or #dateString < 19 then
		return nil
		-- FIXME [epoley 2012-06-04]: Per David F., we should support less precise strings
		-- (e.g. simply a year is valid EXIF and ISO 8601). Doing so would have big UI and
		-- backwards compatibility implications.
	end

	local validatedDate = nil
	
	-- GRUNGE [erwright 2011-08-30]: EXIF calls for YYYY:MM:DD, but we've seen files in the wild
	-- that have a YYYY.MM.DD format instead (including files created by Photoshop Express).
	-- So, we'll try handling both cases, and if we don't recognize either, the file will end up
	-- with an unknown date.
	
	local date, replaced = string.gsub( string.sub( dateString, 1, 10 ), ":", "-" )
	if replaced ~= 2 then
		date, replaced = string.gsub( string.sub( dateString, 1, 10 ), "%.", "-" )
	end
	
	local dateFormatMatches = false

	-- Verify that we are in the correct format (whether we were already or due to the above)
	if date == "0000-00-00" then
		
		-- [tgaul]: Some files have a date/time of all 0s. Rather than treating 
		-- this is a valid date (which puts things in an odd place in the UI) 
		-- return nil so that the import date is used instead.
	
	elseif string.match( date, "%d%d%d%d%-%d%d%-%d%d" ) then
		
		dateFormatMatches = true
		
	end
	
	if dateFormatMatches then			
		
		local time = string.sub( dateString, 12 )
		
		if hasTimeZoneInfo then
			-- HH:MM:SS+HH:MM	with time zone information -- [%+%-]
			if string.match( string.sub( time, 1, 14 ), "%d%d:%d%d:%d%d[%-%+]%d%d:%d%d" ) then
				validatedDate = date .. "T" .. time
			end
			
		else
			if string.match( string.sub( time, 1, 8 ), "%d%d:%d%d:%d%d" ) then
				validatedDate = date .. "T" .. time
			end
		end
		
	end
	
	return validatedDate

end

--------------------------------------------------------------------------------	

local function convertAnalysisInfo( analysisInfo, filePath, optLocationName, fileCreationDate, inAppCamera, folderPath, fileName, originalFilenameFromALOrNil )
	
	--local KSLogging = require 'KSLogging'
	--KSLogging.log("debug"):debug( imageData )
	
	--[[
		This is the asset info that OzDocStore is expecting:
					
		local imageInfo = {
		
		    width = 1500,
		    height = 1000,
		
			croppedWidth = 1500,	-- If missing, just use width and height
			croppedHeight = 1000,
		    
		    keywords = {
		        [ "Fred" ] = true,
		        [ "Barney" ] = true,
		        [ "rubble" ] = true,
		    },
		    
			importSource = {
				folder = "/Users/scouten/Pictures/",
				fileName = "IMG_0549.JPG",
				fileSize = 1534245,
				sha256 = "(hash)",
				
				-- new in 2.0
				contentType = "image/jpeg",
					-- note that content-type can have additional properties by adding a semicolon
					-- and then comma delimited key=value pairs
					-- see RFC 1341 http://www.w3.org/Protocols/rfc1341/7_2_Multipart.html
			},
		    
		    captureDate = "2011-04-05T15:11:01", -- ISO 8601 date format with/without "Z" (UTC specifier)
		}
	--]]
	
	-- Convert the metadata returned by KSImportJpegService to the format above
	
	local imageInfo = {}
	imageInfo.width = analysisInfo[ "PixelWidth" ]
	imageInfo.height = analysisInfo[ "PixelHeight" ]
	
	local importSource = {}
	imageInfo.importSource = importSource

	importSource.contentType = contentTypeForFilePath( fileName )
	
	importSource.folder = optLocationName or folderPath
	importSource.fileSize = analysisInfo.fileSize
	importSource.sha256 = analysisInfo.sha256Hash

	-- Keep precise track of where this file came from; this way if we lose the original
	-- before it's uploaded (for example, if they user quits and deletes the app, then
	-- re-installs the app) we can find it again.
	importSource.folderPath = folderPath
	importSource.filePath = filePath
	
	-- TODO: For sandboxed mac app we should create a scoped bookmark so we can access
	-- this path again, even after a relaunch
	
	if inAppCamera then
		importSource.type = "inAppCamera"
	end

	local fileCreationDate_raw
	if WFPlatformInfo.isMac and filePath then
		local fileAttrs = KSFileUtils.fileAttributesAtPath( filePath )
        if fileAttrs and fileAttrs.fileCreationDate then
        	 fileCreationDate_raw = KSDate.formatDateTimeAsLocal( fileAttrs.fileCreationDate )
        end
    else
        if fileCreationDate then
        	fileCreationDate_raw = KSDate.formatDateTimeAsLocal( fileCreationDate )
		elseif inAppCamera then
			fileCreationDate_raw = KSDate.timeInIso8601FormatNearestSecond( nil, true )
        else
			local fileAttrs = KSFileUtils.fileAttributesAtPath( filePath )
	        if fileAttrs and fileAttrs.fileCreationDate then
	        	 fileCreationDate_raw = KSDate.formatDateTimeAsLocal( fileAttrs.fileCreationDate )
			elseif fileAttrs and fileAttrs.fileModificationDate then	-- Some platforms don't support creation date...
				fileCreationDate_raw = KSDate.formatDateTimeAsLocal( fileAttrs.fileModificationDate )
	        end
		end
	end
	
	if fileCreationDate_raw then
		importSource.fileCreationDate = fileCreationDate_raw
	end
	

	-- We'll use the original filename from the assets library, if we have it, but we fall back to the
	-- GUID if it's not available (e.g., pre-iOS 5.0)
	
	if folderPath and string.find( folderPath, 'assets-library://', nil, true ) and not originalFilenameFromALOrNil then

		-- Look for 'asset.JPG?id=1000000537&ext=JPG' and pull out the id for the filename
		if string.sub( string.lower( fileName ), 1, 10 ) == 'asset.jpg?' then
			local theRest = string.sub( fileName, 11 )
			if theRest and #theRest > 0 then
				local params = KSStringUtils.split( theRest, '&' )
				
				for i = 1, #params do
					local components = KSStringUtils.split( params[ i ], '=' )
					if #components == 2 and components[ 1 ] == 'id' then
						fileName = components[ 2 ] .. ".jpg"
						break
					end
				end
			end
		end
		
	end

	importSource.fileName = originalFilenameFromALOrNil or fileName or makePlaceholderFilename()

	-- check for creation date time in IPTC since it can contain
	-- time zone information that will allow us to convert to UTC time
	-- the Tornado camera inserts IPTC date time with time zone information
	-- into the image's metadata. Use it.
	
	local convertedDateTime = nil
	local iptcDateTime = nil
	
	local iptcTable = analysisInfo[ "{IPTC}" ]
	if iptcTable then
		local dateCreated = iptcTable[ "DateCreated" ]
				
		if dateCreated then
			iptcDateTime = validateDateFormat( dateCreated, true )
			if iptcDateTime then
				convertedDateTime = iptcDateTime
			end
		end
	end
	

	if iptcDateTime == nil then

		-- Capture date: we need to convert [ "DateTimeOriginal" ] from
		--   "2010:09:18 13:07:13" to
		--   "2011-04-05T15:11:01Z"

		local exifDateTimeOriginal = nil
	
		-- FIXME [erwright 2011-08-18]: it would be good to support sub-second capture time when it is available.
		-- Per @scouten: "I remember that we had *lots* of complaints in Lightroom around this from sports shooters.
		-- This will eventually become an issue for us, too."
	
		local exifTable = analysisInfo[ "{Exif}" ];
		if exifTable then
			exifDateTimeOriginal = exifTable[ "DateTimeOriginal" ]
		end
	
		local dateTimeOriginal = exifDateTimeOriginal
	
		-- Step 1: If we found a relevant date in the metadata analysis, try to validate its format and use that.
		if dateTimeOriginal then
			convertedDateTime = validateDateFormat( dateTimeOriginal, false )
		end
	
		-- Step 2: If no metadata date was present or it wasn't valid, try using the filesystem date as a fallback.
		-- This will catch the in-app camera photos as well, whose file creation date is set to system time above.
	
		if not convertedDateTime and importSource.fileCreationDate then
		
			convertedDateTime = validateDateFormat( importSource.fileCreationDate, false )
        
			if convertedDateTime then
			    debug.print( "import: dateTimeOriginal not found in file, using creation date '" .. convertedDateTime .. "' instead: " .. ( filePath or "( no filepath )" ) )
    	    end

		end
	
		-- Step 3: Failing all else, the 'unknown date' marker (shouldn't happen in practice).
	
		if not convertedDateTime then
		
			convertedDateTime = "0000-00-00T00:00:00"
		
			debug.print( "import: dateTimeOriginal not found in data, using timestamp '" .. convertedDateTime .. "' instead" )
	
		end
		
	end

	imageInfo.captureDate = convertedDateTime
	
	local iptcTable = analysisInfo[ "{IPTC}" ]
	if iptcTable then

		local keywords = iptcTable[ "Keywords" ]
		if keywords and type( keywords ) == "table" then
		
			local convertedKeywords = { }

			for _, keyword in ipairs( keywords ) do
				convertedKeywords[ keyword ] = true			
			end
	
			imageInfo.keywords = convertedKeywords			
		end

		local rating = iptcTable[ "StarRating" ]
		if rating and type( rating ) == "number" and rating > 0 then
		
			-- FIXME [erwright 2011-04-07]: set this image to a favorite for the current user
			--imageInfo.isFavorite = true
			
		end

		-- [cmoskalo 2012.04.05]
		-- Note that the actual field name in the IPTC table
		-- for the image is called "description", however the Oz
		-- catalog calls this "caption" so that's the field name
		-- used in the image info.
		local description = iptcTable[ "Description" ]
		if description ~= nil then
			imageInfo.caption = description
		end

	end
	
	if false then
	
		-- FIXME [erwright 2011-04-25]: adding the leaf folder path as a keyword is a crutch until
		-- we have an importedFolder path or some such in OzDocs to remember where the file came from.
		
		-- [mhamburg 2012-03-08]: We do now capture the folder path elsewhere so this should almost
		-- certainly go away. That said, I am simply commenting it out for now until we are convinced
		-- we want it gone. (We might, after all, want to reference this logic elsewhere in using the
		-- folder path.) Please delete when we finally resolve this.
		
		if not optLocationName and folderPath and folderPath ~= '' and folderPath ~= '/' then
			local trimmedFolderPath = folderPath
			if string.sub( trimmedFolderPath, -1 ) == '/' then
				trimmedFolderPath = string.sub( trimmedFolderPath, 1, -2 )
			end
			local _, leafFolder = KSFileUtils.splitFolderAndFile( trimmedFolderPath )
			addKeyword( imageInfo, leafFolder )
		elseif optLocationName then
			addKeyword( imageInfo, optLocationName )
		end
		
	end
	
	local localOnly = {}
	localOnly.localUrl = filePath
	
	imageInfo._localOnly = localOnly
	
	return imageInfo
	
end

--------------------------------------------------------------------------------

local enqueue, dequeue, isEmpty = WFQueueing.makePriorityQueue()

local enqueueErr, dequeueErr = WFQueueing.makeQueue()

-- NOTE [epoley 2011-08-08]:

-- Here is our strategy for dealing with synchronously-resolvable errors
-- that should contribute to overall progress percentages: rather than
-- enqueueing them in the same queue as the async import tasks, we maintain
-- a separate error queue. (Note that the standard import queue can contain
-- entries that resolve as errors; the error queue is simply for those that
-- are able to do so synchronously.)  Right before sending the 'complete' 
-- message (when the import queue has been drained), we synchronously process 
-- all of the entries in the error queue. Enqueueing an error does not kick the
-- import queue, but we do kick the queue if a batch contains only errors (so
-- that the 'complete'/error draining logic can still be triggered).

-- This approach has eliminated problems with determining when the queue is
-- completed; however, it is only effective if photos that are meant to be
-- imported in the same batch (i.e. those multi-selected or dragged by the user)
-- are imported together using 'importFiles' or 'importAssets' instead of
-- multiple calls to their singular equivalents. If we do not do this, it mostly
-- works, except when a sync-error enters the queue before any async entries
-- have been queued. In that case, we will decide right away that we are done.

-- I have implemented 'importFiles' and 'importAssets' appropriately, and changed
-- the client code (esp. in native code) to actually use them with drag and drop,
-- the file picker, and the assets library picker.

--------------------------------------------------------------------------------

local runningImportPromiseP = false

-- The storage for the import counts
local importCountDups = 0
local importCountErrors = 0
local importCountSuccess = 0

--------------------------------------------------------------------------------

local function analysis_finish( sourceKSData, success, ... )

	-- If we have data that came through, it will have been reconstructed as
	-- a new KSData object by the analysis code.

	if sourceKSData then
		sourceKSData:close()
	end

	if success then
		return KSPromises.resolvedPromiseP( 'data', ... )
	else
		return KSPromises.resolvedPromiseP( 'error', ... )
	end

end

--------------------------------------------------------------------------------

local function makeAnalysisRequestP( filePath, optKSData )

	return analysis_finish( optKSData, pcall( KSImportJpegService.analyzeImageData, filePath, optKSData ) )

end

--------------------------------------------------------------------------------

local function closeMultiple( ... )

	for i = 1, select( '#', ... ) do
		local v = select( i, ... )
		if v then
			v:close()
		end
	end

end

--------------------------------------------------------------------------------

-- Pre-condition: COLLECT_IMPORT_METRICS is true

local function updateResultCountsForImportStats( startTime, resultName, optErrorData )

	if recentImportStats[ startTime ] then
		
		local resultCounts = recentImportStats[ startTime ].ResultCounts
		
		if resultCounts.pending > 0 then
			resultCounts.pending = resultCounts.pending - 1
		end
		
		resultCounts[ resultName ] = resultCounts[ resultName ] + 1
		
		if resultCounts.pending == 0 then
			KSImportStats.updateStatsUponTermination( startTime )
		end
		
		if optErrorData then
			resultCounts.errorDetails[ #resultCounts.errorDetails + 1 ] = optErrorData 
		end
		
    end

end

--------------------------------------------------------------------------------

local function makeImportPromiseP( entry )

	local importHandler = entry.importHandler

	local catalogService = importHandler._catalogService
	
	local fromAssetsLibrary = entry.fromAssetsLibrary
	
	local isAutoImported = entry.isAutoImported
	
	local addToAlbumId = entry.addToAlbumId
	
	local prefToSetForAssetAdd = entry.prefToSetForAssetAdd

	local randomId = entry.randomId
	
	-- GRUNGE: This maps away asset library paths in a somewhat heavy-handed detection
	-- path.
	
	local fileName = entry.fileName
	local folderPath = entry.folderPath
	
	local filePath
	
	if folderPath and WFURI.parse( folderPath ).scheme == "assets-library" then
		filePath = fileName and folderPath .. fileName
	else
		filePath = ( folderPath and fileName ) and KSPathUtils.child( folderPath, fileName ) or nil
	end
	
	local filePathForAddPhoto = filePath
	
	local startIndex = -1
	if filePath then
		startIndex = string.find( filePath, "assets-library://", nil, true )
	end
	
	-- For import metrics, we'll accumulate a dictionary of import task name -> start time
	-- so that, when this entry's import has completed, we can report a breakdown of
	-- time spent in each component of the import (e.g. KSAssetsLibraryImportService_makeAssetDataRequestP, addPhotoP, etc.)
	-- As each component finishes executing, it will stop its 'clock' and start the next one.
	
	local startTimesThisRun = COLLECT_IMPORT_METRICS and {} or nil
	
	local importRequestP
	
	if startIndex == 1 then
		filePathForAddPhoto = nil
	end
	
	local function startClock( componentName )
		if COLLECT_IMPORT_METRICS then
			startTimesThisRun[ componentName ] = KSDate.timeNowForAnimation()
		end
	end
	
	local function stopClock( componentName )
		if COLLECT_IMPORT_METRICS and startTimesThisRun[ componentName ] then
			importHandler.componentTimings[ componentName ] = ( importHandler.componentTimings[ componentName ] or 0 ) + ( KSDate.timeNowForAnimation() - startTimesThisRun[ componentName ] ) 
		end	
	end
	
	if fromAssetsLibrary then
		
		startClock( 'makeAssetDataRequestP' )

		importRequestP = KSAssetsLibraryImportService.makeAssetDataRequestP( filePath )
			:mapDataP( function( status, imgData, creationDate, originalFilename )
				
				stopClock( 'makeAssetDataRequestP' )
				startClock( 'makeAnalysisRequestP' )
				
				return makeAnalysisRequestP( filePath, imgData )
                    :mapDataP( function( status, _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy )
                        return status, _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilename
                    end )

			end )

	else

		-- We first analyze the file...

		startClock( 'makeAnalysisRequestP' )
		importRequestP = makeAnalysisRequestP( filePath, entry.ksdata )

	end
	
	-- Filter out items that we can't add to the catalog
	
	importRequestP = importRequestP
		:mapDataP( function( status, _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilenameFromALOrNil )
			
			stopClock( 'makeAnalysisRequestP' )
			startClock( 'verify file size' )
									
			if analysisInfo then
				
				local fileSize = analysisInfo.fileSize / ( 1024 * 1024 )
				local mp = ( analysisInfo[ "PixelWidth" ] * analysisInfo[ "PixelHeight" ] ) / 1000000
				
				if COLLECT_IMPORT_METRICS and recentImportStats[ importHandler.metricsStartTime ] then
					local stats = recentImportStats[ importHandler.metricsStartTime ]
					stats.TotalFileSize = stats.TotalFileSize + fileSize
					stats.TotalMegapixels = stats.TotalMegapixels + mp 
				end
				
			end
			
			-- No remap, just returns input
			
		end )
		
		
	-- Filter out duplicates. We need to general import filtering and special filtering
	-- for auto imports (which respect deletions).
		
	-- If this is an auto-import, we check the auto-import records to see if we've ever
	-- encountered this hash.  Normally, we wouldn't get this far in import because the
	-- asset URL was already compared and we'd stop earlier.  But whenever the iOS
	-- assets library URLs change (as happened between iOS 4 and iOS 5), our URL-comparison
	-- short-circuiting approach will fail.  The hash comparison is our backstop in that
	-- event, and if we get a match, we write the new assets library URL to our local
	-- database to short-circuit the check on next auto-import.
	
	importRequestP = importRequestP
		:mapDataP( function( status, _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilenameFromALOrNil )
			
			stopClock( 'verify file size' )
			
			local sha256 = analysisInfo.sha256Hash
			
			if entry.isAutoImported then
			
				startClock( 'hasAutoImportedPhotoWithSha256P' )
							
				return catalogService:hasAutoImportedPhotoWithSha256P( sha256 )
					:mapDataP( function( _, alreadyAutoImportedId, uploadStatusForView )
						stopClock( 'hasAutoImportedPhotoWithSha256P' )
						if alreadyAutoImportedId then
							catalogService:recordAutoImportUrl( filePath )
							closeMultiple( originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, maybeProxy.proxyKSData )
							if uploadStatusForView == 'originalLost' then
								-- whee!  We thought we lost this asset and suddenly the user found it for us
								return 'recovered', alreadyAutoImportedId, _analyzedFilePath_
							end
							return 'error', 'duplicate', nil
						else
							return 'data', _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilenameFromALOrNil 
						end
					end )

			else
			
				startClock( 'hasPhotoBySha256P' )
				
				return catalogService:hasPhotoBySha256P( sha256 )
					:mapDataP( function( _, existingDocId, uploadStatusForView )
						stopClock( 'hasPhotoBySha256P' )
						if existingDocId then
							closeMultiple( originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, maybeProxy.proxyKSData )
							if uploadStatusForView == 'originalLost' then
								-- whee!  We thought we lost this asset and suddenly the user found it for us
								return 'recovered', existingDocId, _analyzedFilePath_
							end
							return 'error', 'duplicate', existingDocId
						else
							return 'data', _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilenameFromALOrNil
						end
					end )
					
			end
			
		end )
				
	-- Check to see whether we should be adding it to an album even if it is a duplicate
		
	if addToAlbumId then

		importRequestP = importRequestP
			:mapP{ error = function( status, firstParam, secondParam, ... )
		
				-- Include a mapP (not mapDataP) to intercept any duplicate errors and if an albumId is present,
				-- add them to the album (since that was the user's intention anyway.)
				--
				-- Otherwise (or if an error occurs adding to the album), forward through to the next part of the promise chain.
				if status == 'error' and firstParam == 'duplicate' and type( secondParam ) == 'string' then
					
					startClock( 'addAssetsToAlbumP (for duplicates)' )				
					
					return catalogService:addAssetsToAlbumP( addToAlbumId, { secondParam } ):mapP( function( status, ... )
						
						stopClock( 'addAssetsToAlbumP (for duplicates)' )								
						
						if status == 'data' then
							return 'error', 'duplicateAddedToAlbum', ...
						else
							return status, ...
						end
						
					end )
				
				else
					return status, firstParam, secondParam, ...
				end
				
			end  }
		
	end
	
	-- Actually import the photo
	
	local clockToStop = 'addPhotoP'
	
	importRequestP = importRequestP
		:mapDataP( function( status, _analyzedFilePath_, originalKSData, previewKSData, thumbKSData, highDpiThumbKSData, analysisInfo, maybeProxy, creationDate, originalFilenameFromALOrNil  )

			startClock( 'addPhotoP' )
			return catalogService:addPhotoP {
				imageInfo = convertAnalysisInfo( analysisInfo, filePath, entry.locationName, creationDate, entry.fromInAppCamera, folderPath, fileName, originalFilenameFromALOrNil ),
				originalData = originalKSData,
				originalPath = filePathForAddPhoto,
				thumbData = thumbKSData,
				previewData = previewKSData,
				isAutoImported = isAutoImported,
				proxyData = maybeProxy.proxyKSData,
				thumbnail2xData = highDpiThumbKSData,
			}
			
		end )
		

	-- GRUNGE [tgaul]: this is a hack to communicate the photoID of a single added photo
	-- back to the UI code so we can loupe the photo after it's imported in some cases.
	
	if prefToSetForAssetAdd then
		
		importRequestP = importRequestP
			:mapDataP( function( status, docId )				
				if type( docId ) == 'string' then
					local KSPrefs = require 'KSPrefs'
					KSPrefs[ prefToSetForAssetAdd ] = docId
				end
				
			end )
	
	end


	-- Add to album if appropriate.
	
	if addToAlbumId then
		
		importRequestP = importRequestP
			:mapDataP( function( status, docId )
				
				if type( docId ) == 'string' then
				
					stopClock( 'addPhotoP' )
					clockToStop = 'addAssetsToAlbumP'
					startClock( clockToStop )
					return catalogService:addAssetsToAlbumP( addToAlbumId, { docId } )
							:mapDataP( function() return status, docId end )
				
				end
				
			end )
	
	end
	
	
	-- Then we report back status. We use a source connection for the global
	-- report while using the import handler to provide an actual connection
	-- point for the action and hence we make the lifetime dependent on the
	-- lifetime of the import handler.
	
	importRequestP = importRequestP:observeP( function( status, data, ... )
		
		stopClock( clockToStop )
		
		adjustQueueEntryCounts( 1, 0 )
		
		if status == 'data' or ( status == 'error' and data == 'TimeoutOnUpload' ) then
			
			-- In the timout case:
			-- Even though addPhotoP has resolved in error, we don't treat this
			-- as a user-facing error, since it involves upload, not import. 
			-- The photos will still be added to the catalog locally.

			importCountSuccess = importCountSuccess + 1
			
			sendStatusMessage( 'imported', filePath, { success = importCountSuccess }, randomId )
			
			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'success' )
            end

		elseif status == 'error' and data == 'duplicate' then

			-- Suppress the 'Duplicate' warning while doing auto-import

			if not isAutoImported then
            
                importCountDups = importCountDups + 1 
				sendStatusMessage( 'duplicate', filePath, { duplicate = importCountDups } )
                
			end
			
			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'duplicate' )
            end
		
		elseif status == 'error' and data == 'duplicateAddedToAlbum' then
			
			importCountSuccess = importCountSuccess + 1
			
			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'duplicate' )
            end

		elseif status == 'error' and data == 'LowDiskSpace' then

			importCountErrors = importCountErrors + 1
            sendStatusMessage( 'failedLowDiskSpace', filePath, { errors = importCountErrors }, randomId )

			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'error', { message = data, file = filePath } )
            end

			cancelImportRequests( true, true, true )
			
		elseif status == 'error' and data == 'UsageLimitExceeded' then
		
			importCountErrors = importCountErrors + 1
            sendStatusMessage( data, filePath , { errors = importCountErrors}, randomId)

			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'error', { message = data, file = filePath } )
            end

			cancelImportRequests( true, true, true )
			
		elseif status == 'error' then

			-- Not an error for the import itself, just for the file

			importCountErrors = importCountErrors + 1
            sendStatusMessage( 'failed', filePath, { errors = importCountErrors}, randomId )

			if COLLECT_IMPORT_METRICS then
				updateResultCountsForImportStats( importHandler.metricsStartTime, 'error', { message = data, file = filePath } )
            end
		
		elseif status == 'recovered' and data then
			catalogService:recoverOriginalP( data, ... )
		end
		
		sendStatusMessage( 'processedQueueEntries', nil, { processedQueueEntries = getNumProcessedQueueEntries(), totalQueueEntries = getNumTotalQueueEntries() } )
		
	end )
	
	-- We don't need to process cancel messages here because the only
	-- way we cancel is if the import handler has closed. This contrasts
	-- with the global queue case where we need to process the cancel
	-- messages to observe that we've processed more queue entries.
	
	importRequestP:whenResolvedDo {

		data = function( status, revisionId )
			importHandler:finishedOneImport( true, filePath )
		end,

		error = function( status, errorMsg )
			importHandler:finishedOneImport( false, filePath, errorMsg )
		end

	}:attachTo( importHandler )
	
	return importRequestP
		
end
		
--------------------------------------------------------------------------------

local function kickImportQueue()

	if runningImportPromiseP and not WFCloseSupport_hasBeenOrWillBeClosed( runningImportPromiseP ) then
		return
	end
	
	if pauseImports then
		return
	end
	
	local priority, entry
	
	while true do
	
		-- Fetch the next queue item
	
		priority, entry = dequeue()
		
		-- If there is no next item, reset the counters and send the complete
		-- message to any observers so they can know that the queue has been
		-- drained.
		
		if not priority then
			
			local errEntry = dequeueErr()
			
			while errEntry do
				local errFilePath = errEntry.filePath -- Error entry so it does have a file path
				local errStatus = errEntry.status
                if errStatus == 'duplicate' then

					-- If the photo was determined to be a duplicate but the user has requested
					-- that it be added to an album, do that now and tally it as a success rather than a dupe.
					
					local data = errEntry.data
					local album = data and data.album or nil
					local photoId = data and data.photoId or nil
					local catalogId = data and data.catalogId or nil
					local user = data and data.user or nil
					
					if album and photoId and catalogId and user then
					
						withCatalogDoP( catalogId, user, function( catalog )
							return catalog:addAssetsToAlbumP( album, { photoId } )
						end ):runUntilDone()
						
						importCountSuccess = importCountSuccess + 1
						
					else
						importCountDups = importCountDups + 1
					end
					
					if COLLECT_IMPORT_METRICS and errEntry.metricsStartTime then
						updateResultCountsForImportStats( errEntry.metricsStartTime, 'duplicate' )
            		end
					
                else
                    importCountErrors = importCountErrors + 1

					if COLLECT_IMPORT_METRICS and errEntry.metricsStartTime then
						updateResultCountsForImportStats( errEntry.metricsStartTime, 'error', { file = errFilePath, message = errEntry.data } )
            		end

                end
				sendStatusMessage( errStatus, errFilePath )
				errEntry = dequeueErr()
			end
            
            local importCounts = { success = importCountSuccess, error = importCountErrors, duplicate = importCountDups, total = getNumTotalQueueEntries() }

			resetQueueEntryCounts()

            importCountDups = 0
            importCountErrors = 0
            importCountSuccess = 0
			
			if didCancelAllImports then
				sendStatusMessage( 'canceled', nil, importCounts )
				
				didCancelAllImports = false
			else
				sendStatusMessage( 'complete', nil, importCounts )
				
				autoImportIsRunning = false
				debug.print( "set autoImportIsRunning = false upon completion" )
					-- NOTE: This means that if we have started an auto import, we will not
					-- permit another to start until the import queue has been drained.
					-- This could mean that if a manual import gets tacked on after the auto-import,
					-- we will not permit another auto-import crawl until all outstanding imports
					-- are done processing. This seems acceptable; it's a queue anyway.
			end
			
			return
		end
		
		-- If the import handler hasn't been closed, then we start an import.
		-- When that completes, we will clear it as the current import promise
		-- and we will kick the queue again.
		
		if not WFCloseSupport_hasBeenOrWillBeClosed( entry.importHandler ) then
		
			-- GRUNGE [mhamburg 2011-08-03]: This is a kludge to get this action to
			-- run later than the status report action by doing a no-op promise mapping
			-- so that the earlier actions get queued first.
			
			-- We also want to delay execution so that we can finish set up before
			-- the observation gets to run.
						
			local p
			
			p = makeImportPromiseP( entry )
					:waitForP( require 'WFIdle'.laterP )
					:observeP( function()
						if runningImportPromiseP == p then
							runningImportPromiseP = nil
							kickImportQueue()
						end
					end )
					:runUntilDone()
					
			runningImportPromiseP = p
			
			break

		end
		
		-- Otherwise, just treat this as having been processed and loop around.
		
		adjustQueueEntryCounts( 1, 0 )

	end

end

--------------------------------------------------------------------------------

local count = 0

--------------------------------------------------------------------------------

local kSecondsPerDay = 24 * 60 * 60

--------------------------------------------------------------------------------

local function getFilePriority( filePath )

	local attr = KSFileUtils.fileAttributesAtPath( filePath )
	
	local creationDateInSeconds = attr.fileCreationDate
	
	if creationDateInSeconds == nil then	-- this happens on iOS
		return 0
	end
	
	local creationDateInDays = math.floor( creationDateInSeconds / kSecondsPerDay )
	
	local dayCount = math.max( 0, creationDateInDays + 20 * 365 )
		-- The time stamp is based off of 2000. Let's allow for stamps
		-- all the way back to 1980 before we end up pinning. (Yes, this
		-- ignores the effects of leap year.)
		
	local remainderSeconds = creationDateInSeconds - creationDateInDays * kSecondsPerDay
	
	-- We should have 0 <= remainderSeconds < kSecondsPerDay
	
	return math.floor( dayCount * kSecondsPerDay + kSecondsPerDay - remainderSeconds )
		-- We want to make higher dates have higher priority but we want to make
		-- earlier times on those days also work out to have higher priorities.
		-- We also ensure that we are returning an integer.

end

--------------------------------------------------------------------------------

local function enqueueForImport( importHandler, uriOrNativeFilePath, optKSData, optLocationName, optIsFromAL, optIsAutoImported, optIsFromInAppCamera, optAlbumId, optPrefToSetForAssetAdd, randomId )

	-- GRUNGE uriOrNativeFilePath can be too many kinds of things
	--		assets-library://asset/asset.JPG?id=1000000001&ext=JPG
	--			(iOS import from photo library)
	--		/Users/user/Pictures/photo.jpg
	--			(mac os import from disk)
	--		C:\Users\user\Pictures\photo.jpg
	--			(windows user)
	
	-- We save space by splitting the folder from the file name. This helps
	-- since folder names are used repeatedly.

	local folderPath, fileName

	if uriOrNativeFilePath then

		local parsed = WFURI.parse( uriOrNativeFilePath )
	
		if parsed.scheme == 'assets-library' then
			-- not a file path, split along URI boundary
			fileName = parsed.path .. "?" .. parsed.query
			folderPath = parsed.scheme .. "://" .. parsed.authority
		else
			folderPath, fileName = KSPathUtils.parent( uriOrNativeFilePath ), KSPathUtils.leafName( uriOrNativeFilePath )
				-- Can happen if we had to convert the file to a temp file
				-- e.g. a UIImage on the pasteboard, or an unsupported
				-- maybe in the future some file type (e.g. GIF) was in the camera roll.
			optIsFromAL = false
		end
	end
	
	-- If we have data, we want to process it faster because it's probably
	-- chewing memory.
	
	local priority = 0
	
	if optKSData then
		priority = 20000000000
	else
		local success, filePriority = pcall( getFilePriority, uriOrNativeFilePath )
		if success then
			priority = filePriority
		end
	end
	
	if isEmpty() then
		count = 0
	end
	
	count = count + 1
	
	-- This results in later enqueues having lower priorities and hence sorting
	-- later. The multiplicative factor is designed to cause the fraction to
	-- cover most of ( 0, 1 ).
	
	local fract = 2047 / ( 1024 * ( count + 1 ) )

	enqueue( priority + fract, {
		importHandler = importHandler,
		folderPath = folderPath,
		fileName = fileName,
		ksdata = optKSData,
		fromInAppCamera = optIsFromInAppCamera,
		locationName = optLocationName,
		fromAssetsLibrary = optIsFromAL,
		isAutoImported = optIsAutoImported,
		addToAlbumId = optAlbumId,
		randomId = randomId,
		prefToSetForAssetAdd = optPrefToSetForAssetAdd,
			-- [sdaniel 2012-03-27]: Per sumner's request all 1.2+ clients should generate high dpi thumbnails
	} )
	
	adjustQueueEntryCounts( 0, 1 )
	
	sendStatusMessage( 'totalQueueEntries', nil, { processedQueueEntries = getNumProcessedQueueEntries(), totalQueueEntries = getNumTotalQueueEntries() } )
	
	kickImportQueue()

end

--------------------------------------------------------------------------------

local function enqueueError( filePath, optOtherStatusCode, optExtraData, optImportStartTime )

	enqueueErr( { filePath = filePath, status = optOtherStatusCode or 'failed', data = optExtraData, metricsStartTime = optImportStartTime } )
	
	adjustQueueEntryCounts( 0, 1 )
	
	-- Errors should not kick the import queue!

end

--------------------------------------------------------------------------------

local function collectTimedCheckpointData( importHandler )
	
	local checkpointNum = 0
	local startTimer
	
	startTimer = function()
	
		local beforeImportCountDups  = importCountDups
		local beforeImportCountErrors = importCountErrors
		local beforeImportCountSuccess =  importCountSuccess
		local beforeImportCountProcessed = importHandler._processedFiles
		
		WFTimer.afterDelayDo( 900, function()
		
			checkpointNum = checkpointNum + 1
			
			local stats = recentImportStats[ importHandler.metricsStartTime ]
			if not stats.TimedCheckpointData then
				stats.TimedCheckpointData = {}
			end 
			
			local WFAppMemoryUsage = require 'WFAppMemoryUsage'
			local currentRM, currentVM = WFAppMemoryUsage.getCurrentUsage()
			
			local function convertToMegabytes( bytes )
				return bytes / 1024 / 1024
			end
			
			local timedData = {}
			timedData = {
						CheckPointTime = KSDate.timeNowForAnimation(),
						DuplicateCount = importCountDups - beforeImportCountDups,
						ErrorCount = importCountErrors - beforeImportCountErrors,
						SuccessCount = importCountSuccess - beforeImportCountSuccess,
						ProcessedCount = importHandler._processedFiles - beforeImportCountProcessed,
						CurrentAppMemory_RM = convertToMegabytes( currentRM ),
						CurrentAppMemory_VM = convertToMegabytes( currentVM )
			}
			
			recentImportStats[ importHandler.metricsStartTime ].TimedCheckpointData[ "Checkpoint " .. checkpointNum ] = timedData
	
			if importHandler._processedFiles ~= importHandler._totalFiles then 
				startTimer()
			end
		
		end )
	end	
	
	local KSPrefs = require 'KSPrefs'
	if KSPrefs[ 'WFTroubleshooting_automationOutputEnabled' ] then 
		startTimer()
	end
	
end

--------------------------------------------------------------------------------

local KSImportHandler = KSClass 'KSImportHandler' {

	_totalFiles = 0,
		-- The number of items queued against this import handler.
		
	_processedFiles = 0,
		-- The number of items for which processing has been completed.
		
	_allQueued = false,
		-- Have we received the signal that we're done queueing
		
	_requestP = false,
		-- The promise that resolves when this import handler closes.
		
	_resolver = false,
		-- The resolver for the above promise.
		
	_catalogService = false,
		-- The service for the catalog
		
	_catalogId = false, 
		-- The catalog Id for this catalog
	
	_ozUser = false,
		-- (OzUser) user who is doing the import
	
	_requestId = false,
		-- The request id for this import
		
	_overQuotaOccurred = false,
	
	_hasAtLeastOneAutoImportedPhoto = false,
	
	_hasAtLeastOneManuallyImportedPhoto = false,
		
	metricsStartTime = false,
	componentTimings = false,
	
	_metricsCheckpointTarget = false,
		
	__init = function( self, catalogId, ozUserId, isAutoImport )
	
		if COLLECT_IMPORT_METRICS then
			self.componentTimings = {}
			KSImportStats.checkMemoryPollingForMetricsEnabledAndStartIfNeeded()
		end

		self._allQueued = false
		
		self._requestP, self._resolver = KSPromises.promiseP()
		
		-- limbo errors?? What can this promise be attached to?
		
		self._requestP:runUntilDone()
		
		self._ozUser = OzAccountService_currentUserB_valueNow()
		assert( self._ozUser and self._ozUser:id() == ozUserId )
		
		self._catalogService = OzCatalogService( catalogId, self._ozUser )
		
		self._catalogId = catalogId
		
		-- If the current catalog is no longer accessible, then we shut
		-- down this import
		 
		self._catalogService:canAccessB():onDataB( function( canAccess, reason )
		
			if type( canAccess ) == 'boolean' and not canAccess then
				debug.print( "Can't access import catalog any more; shutting down import" )
				self:close()
			end
			
		end ):addAsCloseTargetOf( self )
		
		-- If the current user changes, then we shut down this import.
		
		self._ozUser:isCurrentUserB():onDataB( function( isCurrent )
			if not isCurrent then
				self:close()
			end
		end ):addAsCloseTargetOf( self )

		if isAutoImport then
			self._requestId = #autoImportRequestIdToHandler + 1
			autoImportRequestIdToHandler[ self._requestId ] = self
		else
			self._requestId = #manualImportRequestIdToHandler + 1
			manualImportRequestIdToHandler[ self._requestId ] = self
		end

	end,
	
	timeForMetricsCheckpoint = function( self )
	
		if not self._allQueued or self._totalFiles < 100 then
			return false
		end
		
		if not self._metricsCheckpointTarget then
			
			local target = self._totalFiles / 10
			
			-- When it does not divide evenly, use standard rounding to 
			-- determine the target point for checkpoints (this will ensure
			-- the checkpoints end up as evenly split as possible, but 
			-- it means the last checkpoint may have either more or less
			-- than the others.)
			
			local flooredTarget = math.floor( target )
			if target - flooredTarget >= .5 then
				target = math.ceil( target )
			else
				target = flooredTarget
			end
			
			self._metricsCheckpointTarget = target
			
		end
		
		-- Treat the last checkpoint as special--it may have slightly more than the others.
		
		return self._processedFiles == self._totalFiles or 
			( self._metricsCheckpointTarget * 10 > self._processedFiles 
				and self._processedFiles % self._metricsCheckpointTarget == 0 )
		
	end,
	
	close = function( self )
		autoImportRequestIdToHandler[ self._requestId ] = nil
		manualImportRequestIdToHandler[ self._requestId ] = nil
		
		if self._overQuotaOccurred then
			self._resolver( 'error', 'UsageLimitExceeded' )
		else
			self._resolver( 'data', true )
		end
			
		self._catalogService:close()
		
	end,
	
	WFCloseSupport.addCloseSupportMethods,
	
	-- Arguments to enqueueForImport:
	-- ( importHandler, filePath, optKSData, optLocationName, optIsFromAL, optIsAutoImported, optIsFromInAppCamera, optAlbumId, optPrefToSetForAssetAdd )
	
	queueFilePath = function( self, filePath, optAlbumId, optPrefToSetForAssetAdd , randomId)
		enqueueForImport( self, filePath, nil, nil, nil, nil, nil, optAlbumId, optPrefToSetForAssetAdd, randomId)
		self._totalFiles = self._totalFiles + 1
	end,

	queueAssetUrl = function( self, assetUrl, isAutoImported, optAlbumId )
		
		if isAutoImported then
			self._hasAtLeastOneAutoImportedPhoto = true
		else
			self._hasAtLeastOneManuallyImportedPhoto = true
		end
		
		enqueueForImport( self, assetUrl, nil, nil, true, isAutoImported, false, optAlbumId )
		self._totalFiles = self._totalFiles + 1
		
	end,

	queueData = function( self, fromPath, ksData, optLocationName, optFromCamera, optAlbumId )
		enqueueForImport( self, fromPath, ksData, optLocationName, false, false, optFromCamera, optAlbumId )
		self._totalFiles = self._totalFiles + 1
	end,
	
	markAllQueued = function( self )
		self._allQueued = true
		if self._totalFiles <= self._processedFiles then
			self:close()
		end
	end,
	
	-- Callback from the import process when we complete an import.
	
	finishedOneImport = function( self, success, filePath, errorMsg )
		self._processedFiles = self._processedFiles + 1
		
		if not success and errorMsg == 'UsageLimitExceeded' then
			self._overQuotaOccurred = true
		end
		
		-- With each file that is processed, if we are collecting metrics, we check whether it is time
		-- for one of the 10% checkpoints. If it is, we record various data: how long this 10% took,
		-- how many files, total filesize, etc. We also request thread statistics. The thread statistics
		-- that have been reported since the previous request are recorded for that previous 10%. The last 
		-- 10% is reported after a 15 second timer.
		
		if COLLECT_IMPORT_METRICS and recentImportStats[ self.metricsStartTime ] and self:timeForMetricsCheckpoint() then
		
			local checkpointNumber 
			
			if self._processedFiles == self._totalFiles then
				checkpointNumber = 10
			else
				checkpointNumber = math.floor( self._processedFiles / self._metricsCheckpointTarget )
			end
			
			local checkpointKey = 'checkpoint-%d'
			
			local endTime = KSDate.timeNowForAnimation()
			local endCount = self._processedFiles
			local endFileSize = recentImportStats[ self.metricsStartTime ].TotalFileSize
			local endMegapixels = recentImportStats[ self.metricsStartTime ].TotalMegapixels
			
			local previousCheckpoint
			if checkpointNumber > 1 then
				previousCheckpoint = recentImportStats[ self.metricsStartTime ][ string.format( checkpointKey, checkpointNumber - 1 ) ]
			end
			
			local totalTime = previousCheckpoint and endTime - previousCheckpoint.endTime or endTime - self.metricsStartTime
			
			recentImportStats[ self.metricsStartTime ][ string.format( checkpointKey, checkpointNumber ) ] = {
				
				Timings = self.componentTimings,
				
				endTime = endTime,
				endCount = self._processedFiles,
				endFileSize = endFileSize,
				endMegapixels = endMegapixels, 
				
				TotalTime = totalTime,
				TotalCount = previousCheckpoint and endCount - previousCheckpoint.endCount or endCount,
				TotalFileSize = previousCheckpoint and endFileSize - previousCheckpoint.endFileSize or endFileSize,
				TotalMegapixels = previousCheckpoint and endMegapixels - previousCheckpoint.endMegapixels or endMegapixels,
				
				TimeComparedToFirstCheckpoint = previousCheckpoint and 
					totalTime / ( KSTableUtils_derefOpt( recentImportStats, self.metricsStartTime,  string.format( checkpointKey, 1 ), 'TotalTime' ) or 1 ),
					
			}
			
			if previousCheckpoint then
				writeThreadStats( self.metricsStartTime, previousCheckpoint )
			end
			
			threadStats = {}
			KSImportJpegService.sendThreadStatsRequest()
			
			if self._processedFiles == self._totalFiles then
				WFTimer.afterDelayDo( 15, function()
					if recentImportStats[ self.metricsStartTime ] then
						writeThreadStats( self.metricsStartTime, recentImportStats[ self.metricsStartTime ][ string.format( checkpointKey, checkpointNumber ) ] )
					end
				end )
			end
				
			self.componentTimings = {}
		end
		
		if self._allQueued and self._totalFiles <= self._processedFiles then
			self:close()
		end
	end,
	
}

--------------------------------------------------------------------------------
-- If we are using the events functionlity, we want to stop any uploading...

WFRequestDispatcher.onMessage ( 'pauseImport', 
	function()
	
		debug.print( "pauseImport was called." )
		pauseImports = true

	end )

--------------------------------------------------------------------------------
-- When done using the events functionlity, we want to start uploading...

WFRequestDispatcher.onMessage ( 'resumeImport',
	function()
	
		debug.print( "resumeImport was called." )
		pauseImports = false
		kickImportQueue()

	end )

--------------------------------------------------------------------------------

local function processAutoImportAssetListP( importHandler, assetList )

	if not importHandler then
		return
	end

	local startTime
	
	if COLLECT_IMPORT_METRICS then
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )
		importHandler.metricsStartTime = startTime		
		collectTimedCheckpointData( importHandler )	
	end
			
	local currentUser = OzAccountService_currentUserB_valueNow()
	assert( currentUser and currentUser:id() == importHandler._ozUser:id() )
        
	local p = withCatalogDoP( importHandler._catalogId, currentUser, function( catalog )
    
		local onlyErrorsFound = true
	
	    local queuedCount = 0
		local queuedErrors = 0
		
		local function checkOneAssetInfoP( assetInfo )
		
			local url = assetInfo.url
			
			local pp = catalog:hasAutoImportedPhotoAtUrlP( url )
			
			pp = pp:mapP( function( status, isDuplicate )
			
				-- If we haven't yet canceled, but we get an error, then
				-- cancel by closing the import handler.
			
				if importHandler and status ~= 'data' then
					importHandler:close()
					importHandler = nil
				end
				
				-- If we've canceled, then report that as the result
			
				if not importHandler then
					return 'error', 'canceled'
				end
				
				-- If we get to this point, the status must be 'data'
				
				if not isDuplicate then
					onlyErrorsFound = false
					importHandler:queueAssetUrl( url, true )
					queuedCount = queuedCount + 1
				else
					queuedErrors = queuedErrors + 1
				end
				
				-- Report success
				
				return 'data', true
				
			end )
			
			return pp
		
		end
		
		-- Check all of the asset infos
		
		local ppp = KSPromises.allP( checkOneAssetInfoP, assetList )
		
		-- If the check succeeded, then update the import statistics.
		
		ppp = ppp:mapDataP( function()
		
			local totalQueued = queuedCount + queuedErrors

			KSImportStats_updateTotalQueued( startTime, totalQueued )
		
			return 'data', true

		end )
		
		-- Finish up the import logic.
		
		ppp = ppp:mapP( function( ... )
		
			if importHandler then
				importHandler:markAllQueued()
			end
		
			if onlyErrorsFound then
				debug.print( "autoImport: Nothing found to import" )
				kickImportQueue()
			end
			
			return ...
			
		end )
		
		return ppp
	
	end )
	
	return p:runUntilDoneP()
	
end

--------------------------------------------------------------------------------

-- GRUNGE [epoley 2011-06-01]: Auto-import ignores duplicates, but not when multipe
-- autoImport requests are processing simultaneously. Is there something that we want
-- to do to avoid this if the autoImport message is sent a second time before
-- the first has finished?

WFRequestDispatcher.onMessage ( 'autoImport', 

	function( catalogId, ozUserId )
	
		if WFPlatformInfo.isMac then
			error( "autoImport cannot be performed on non-iOS platforms.", 2 )
		end
	
		if autoImportIsRunning then
			debug.print( "Another auto-import request is already running; ignoring this duplicate request." )
			return
		end
			
		autoImportIsRunning = true
--	    debug.print( "set autoimportisrunning = true" )
	
		local importHandler = KSImportHandler( catalogId, ozUserId, true )
		
		local getAssetsP = KSAssetsLibraryImportService.getAssetsListP( )
			:mapP( function( status, message, ... )
				if status == 'error' then
					if message ~= 'canceled' then
						debug.print( "error in autoImport:", message )
					end
					importHandler:close()
				elseif status == 'data' then
					return processAutoImportAssetListP( importHandler, message )
				end
			end )
			:runUntilDone()
		
		autoImportEnumerationPromises[ getAssetsP ] = true
				
		return importHandler._requestP

end )

--------------------------------------------------------------------------------

local recursiveFiles

if WFPlatformInfo.isMac then

	recursiveFiles = require 'KSRecursiveFilesIterator'
	
else

	recursiveFiles = function()
	
		error "recursiveFiles not implemented on this platform"
	
	end

end

--------------------------------------------------------------------------------

-- Public API: import all JPG files contained within the folder path

-- onDataE:  sends ( filePath, success, percentComplete ) for every file encountered in the folder tree
-- onDoneE:  sends done when finished

WFRequestDispatcher.onMessage ( 'importFolder',  function( catalogId, ozUserId, folderPath, optAlbumId )
	
	WFArguments_check( 'string', 'string', 'string' )	
	
	local importHandler = KSImportHandler( catalogId, ozUserId )
	
	-- Accumulation needed?
	
	local onlyErrorsFound = true
	
    local queuedCount = 0
	local queuedErrors = 0

	local startTime
	
	if COLLECT_IMPORT_METRICS then
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )
		importHandler.metricsStartTime = startTime
		collectTimedCheckpointData( importHandler )
	end 
   
	for filePath in recursiveFiles( folderPath, true ) do
	
		if isValidFilePath( filePath ) then
			onlyErrorsFound = false
			importHandler:queueFilePath( filePath, optAlbumId )
            queuedCount = queuedCount + 1
		else
			enqueueError( filePath, 'failed', 'invalid file path/extension', startTime ) 
			queuedErrors = queuedErrors + 1	 				
		end
		
	end
	
	if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
		local totalQueued = queuedCount + queuedErrors
		recentImportStats[ startTime ].TotalEnqueued = totalQueued
		recentImportStats[ startTime ].ResultCounts.pending = totalQueued
	end
	
	importHandler:markAllQueued()
	
	if onlyErrorsFound then
		kickImportQueue()
	end
		-- Note [epoley]: We musn't kick the import queue in enqueueError
		-- (in case an import starts with an error; we don't want to falsely
		-- declare done.) However, since only queueFilePath kicks the queue,
		-- we must throw in an extra kick once everything has been queued,
		-- in case only errors have been queued.
		
	return importHandler._requestP

end )

--------------------------------------------------------------------------------

-- Public API: import the single JPG file specified by filePath.

-- onDataE:  sends ( filePath, success, % complete ) for the data with a boolean status of the file import
-- onDoneE:  sends done when finished

WFRequestDispatcher.onMessage ( 'importFile' , function( catalogId, ozUserId, filePath, optAlbumId, optPrefToSetForAssetAdd, randomId )

	WFArguments_check( 'string', 'string', 'string' )
	
	local onlyErrorsFound = true
	local importHandler
	
	local startTime
	if COLLECT_IMPORT_METRICS then
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )	
	end
	
	if isValidFilePath( filePath ) then
	
		onlyErrorsFound = false
		importHandler = KSImportHandler( catalogId, ozUserId )
		
		if COLLECT_IMPORT_METRICS then
			importHandler.metricsStartTime = startTime
			collectTimedCheckpointData( importHandler )
		end
		
		importHandler:queueFilePath( filePath, optAlbumId, optPrefToSetForAssetAdd, randomId )
		
		if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
			recentImportStats[ startTime ].TotalEnqueued = 1
			recentImportStats[ startTime ].ResultCounts.pending = 1
		end
		
		importHandler:markAllQueued()
	
	else
		enqueueError( filePath, 'failed', 'invalid file path/extension', startTime ) 
	
		if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
			recentImportStats[ startTime ].TotalEnqueued = 1
			recentImportStats[ startTime ].ResultCounts.pending = 1
		end
		
	end
	
	if onlyErrorsFound then
		kickImportQueue()
	end
		-- Note [epoley]: We musn't kick the import queue in enqueueError
		-- (in case an import starts with an error; we don't want to falsely
		-- declare done.) However, since only queueFilePath kicks the queue,
		-- we must throw in an extra kick once everything has been queued,
		-- in case only errors have been queued.

	if importHandler then
		return importHandler._requestP
	else
		error( "Invalid filepath", 2 )
	end
	
end )

--------------------------------------------------------------------------------

-- Public API: import the JPG files specified by the filePaths array.

-- onDataE:  sends ( filePath, success, % complete ) for the data with a boolean status of the file import
-- onDoneE:  sends done when finished

WFRequestDispatcher.onMessage ( 'importFilesAndFolders' , function( catalogId, ozUserId, filePaths, optAlbumId )

	WFArguments_check( 'string', 'string', 'table' )
	
	local onlyErrorsFound = true
	
    local queuedCount = 0
	local queuedErrors = 0
    
	local importHandler = KSImportHandler( catalogId, ozUserId )
	
	local startTime
	
	if COLLECT_IMPORT_METRICS then
		
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )
		importHandler.metricsStartTime = startTime
		collectTimedCheckpointData( importHandler )
		
	end
	
	for i = 1, #filePaths do
	
		local filePath = filePaths[ i ]
		
		local exists = KSFileUtils.exists( filePath )
		
		if exists == 'file' then
		
			if isValidFilePath( filePath ) then

				onlyErrorsFound = false
				importHandler:queueFilePath( filePath, optAlbumId )
                queuedCount = queuedCount + 1

			else
				enqueueError( filePath, 'failed', 'invalid file path/extension', startTime ) 
				queuedErrors = queuedErrors + 1	
			end
			
		elseif exists == 'directory' then -- is folder
		
			for filePathInFolder in recursiveFiles( filePath, true ) do
	
				if isValidFilePath( filePathInFolder ) then

					onlyErrorsFound = false
					importHandler:queueFilePath( filePathInFolder, optAlbumId )
                    queuedCount = queuedCount + 1
				
				else
					enqueueError( filePathInFolder, 'failed', 'invalid file path/extension', startTime )
					queuedErrors = queuedErrors + 1	
				end
			
			end
		
		end
		
	end
	
	if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
		local totalQueued = queuedCount + queuedErrors
		recentImportStats[ startTime ].TotalEnqueued = totalQueued
		recentImportStats[ startTime ].ResultCounts.pending = totalQueued
	end
	
	importHandler:markAllQueued()
	
	if onlyErrorsFound then
		kickImportQueue()
	end
		-- Note [epoley]: We musn't kick the import queue in enqueueError
		-- (in case an import starts with an error; we don't want to falsely
		-- declare done.) However, since only queueFilePath kicks the queue,
		-- we must throw in an extra kick once everything has been queued,
		-- in case only errors have been queued.
		
	return importHandler._requestP
	
end )

--------------------------------------------------------------------------------

-- Public API: import the single JPG KSData object.

-- onDataE:  sends ( filePath, success, % complete ) for the data with a boolean status of the file import
-- onDoneE:  sends done when finished

-- fromPath is used for local duplicate detection, and is parsed for the destination folder if
-- optLocationName is not provided. optLocationName should be provided when fromPath is not a 
-- user-meaningful filepath (e.g. an assets library URL, camera). In the case of the assets library,
-- optLocationName should be the name of the group or album where the photo was found.

WFRequestDispatcher.onMessage ( 'importData', function( catalogId, ozUserId, ksData, fromPath, optLocationName, optFromInAppCamera, optAlbumId )

	assert( type( catalogId ) == 'string'  )
	assert( type( ozUserId ) == 'string' )
	assert( KSData.isData( ksData ) )
	local importHandler = KSImportHandler( catalogId, ozUserId )
	
	local startTime
	
	if COLLECT_IMPORT_METRICS then
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )
		importHandler.metricsStartTime = startTime	
		collectTimedCheckpointData( importHandler )
	end
	
	importHandler:queueData( fromPath, ksData, optLocationName, optFromInAppCamera, optAlbumId )
	
	if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
		recentImportStats[ startTime ].TotalEnqueued = 1
		recentImportStats[ startTime ].ResultCounts.pending = 1
	end
	
	importHandler:markAllQueued()

	return importHandler._requestP

end )

--------------------------------------------------------------------------------

local function importAssets( catalogId, ozUserId, assetUrls, optLocationName, optAlbumId )

	WFArguments_check( 'string', 'string', 'table', '?string', '?string' )
	
	local currentUser = OzAccountService_currentUserB_valueNow()
	assert( currentUser and currentUser:id() == ozUserId )

	local numExpected = #assetUrls

	if numExpected == 0 then
		return
	end
	
	local importHandler = KSImportHandler( catalogId, ozUserId )
	
	local onlyErrorsFound = true
	
    local queuedCount = 0
	local queuedErrors = 0
	local startTime
	
	if COLLECT_IMPORT_METRICS then
		startTime = KSDate.timeNowForAnimation()
		KSImportStats_addImportStatEntry( startTime )
		importHandler.metricsStartTime = startTime
		collectTimedCheckpointData( importHandler )
	end
	
	-- Check all of the assets against the catalog for duplicates.
	
	local checkAssetsP = withCatalogDoP( catalogId, currentUser, function( catalog )
	
		local function checkOneP( assetUrl )
		
			local hasPhotoByLocalUrlP = catalog:hasPhotoByLocalUrlP( assetUrl )
			
			return hasPhotoByLocalUrlP:mapP( function( status, existingDocId )
				if status == 'data' and not existingDocId then
					onlyErrorsFound = false
					importHandler:queueAssetUrl( assetUrl, false, optAlbumId )
                    queuedCount = queuedCount + 1
				else
					enqueueError(
						assetUrl,
						'duplicate',
						optAlbumId and { album = optAlbumId, photoId = existingDocId, catalogId = catalogId, user = currentUser } or nil,
						startTime )
					queuedErrors = queuedErrors + 1
				end
				return 'data', true
			end )
			
		end
	
		return KSPromises.allP( checkOneP, assetUrls )

	end )
	
	-- Once we've finished doing the check, kick the import queue into gear.
	
	local startImportP = checkAssetsP:mapP( function()
	
		if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
			local totalQueued = queuedCount + queuedErrors
			recentImportStats[ startTime ].TotalEnqueued = totalQueued
			recentImportStats[ startTime ].ResultCounts.pending = totalQueued
		end
		
		importHandler:markAllQueued()

		if onlyErrorsFound then
			kickImportQueue()
		end

	end )
	
	startImportP:runUntilDone()
	
	return importHandler._requestP
	
end
--------------------------------------------------------------------------------


-- Public API: import a single asset from the Assets Library

-- onDataE:  sends ( filePath, success, % complete ) for the data with a boolean status of the file import
-- onDoneE:  sends done when finished

-- assetUrl is used for local duplicate detection.  
-- optLocationName should be the name of the group or album where the photo was found.

-- Note [epoley 2011-08-22]: At the time of this writing, there are no clients of 'importAsset'
-- (singular). We may wish to deprecate it. Note that unlike its plural sibling API, it does not
-- exercise care in minimizing the image data sitting in memory at any time. In other words, the
-- following snippet of pseduocode may cause low-memory warning crashes:
--
--  for asset in assetsList do
-- 		importAsset( asset )
-- 	end
-- 
-- Whereas the seemingly equivalent snippet below will not:
--
-- importAssets( assetsList )


-- NOTE [epoley 2011-09-29]: Got tired of maintaining this duplicated, unused code. Rather than deprecating it,
-- we forward it to importAssets. Note that the above caution about memory usage is still true.

WFRequestDispatcher.onMessage ( 'importAsset' , function( catalogId, ozUserId, assetUrl, optLocationName, optAlbumId )

	WFArguments_check( 'string', 'string', 'string', '?string' )
			
	return importAssets( catalogId, ozUserId, { assetUrl }, optLocationName, optAlbumId )
	
end )

--------------------------------------------------------------------------------

-- Public API: imports a list of assets from the Assets Library

-- onDataE:  sends ( filePath, success, % complete ) for the data with a boolean status of the file import
-- onDoneE:  sends done when finished

-- assetUrls are used for local duplicate detection and fetching the data.
-- optLocationName should be the name of the group or album where the photo was found.

-- NOTE [epoley 2011-08-22]:

-- Requesting the data from the AL here proved to be problematic
-- (we were fetching them faster than we could analyze them and
-- add them to the catalog, so we ended up with too many data 
-- objects sitting in memory.) So instead, we enqueue a URL
-- and fetch its data when we get around to processing this request.
-- That means that we chain promises as follows: 
--
-- fetchALData > analyzeData > checkSha256Hash > addPhoto
--
-- To avoid having more than one asset's data in memory at a time,
-- we do not move on to the next asset until that entire chain
-- has resolved for the current asset. This memory-conservative
-- approach comes at a performance cost, but we can make this fancier
-- and more aggressive when OzCatalogService:addPhoto accepts
-- a data stream as input.

WFRequestDispatcher.onMessage ( 'importAssets', importAssets )

--------------------------------------------------------------------------------
