--[[----------------------------------------------------------------------------

KSImportStats.lua

@author arahn 2012-10-24

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

FIXME [erwright 2011-05-5]: future work on import: we need to revisit the overall
import structure for handling assets from the asset library, data straight from
the camera sensor, and filesystem files on the desktop.  Are file paths and
camera sensor data two distinct import paths?  Or are file paths just a high-level
layer on top of a core data-import component?

------------------------------------------------------------------------------]]

local WFPlatformInfo = require "WFPlatformInfo"

local KSPrefs = require 'KSPrefs'
local KSDate = require "KSDate"
local WFCurrentThread = require "WFCurrentThread"

local KSImportUtils = require "KSImportUtils"

--==============================================================================

local getWFAppMemoryUsage = KSImportUtils.makeLateRequire 'WFAppMemoryUsage'

local COLLECT_IMPORT_METRICS = true

local MAX_IMPORT_STATS = 5

local recentImportStatsCount = 0

local recentImportStats = {}

--------------------------------------------------------------------------------

local previousMemoryPollingEnabledValue

local function checkMemoryPollingForMetricsEnabledAndStartIfNeeded()

	local enabled = COLLECT_IMPORT_METRICS 
		and not _G.DISTRIBUTION
		and WFPlatformInfo.isCocoa
		and KSPrefs[ 'WFTroubleshooting_automationOutputEnabled' ]
		
	if enabled and not previousMemoryPollingEnabledValue then
		getWFAppMemoryUsage().startPolling()
	end
	
	previousMemoryPollingEnabledValue = enabled
	
	return enabled
	
end

--------------------------------------------------------------------------------

if COLLECT_IMPORT_METRICS then
	
	checkMemoryPollingForMetricsEnabledAndStartIfNeeded()
		
end

--------------------------------------------------------------------------------

local function convertToMegabytes( bytes )
	return bytes / 1024 / 1024
end

--------------------------------------------------------------------------------
		
local function updateStatsUponTermination( startTime )

	if not recentImportStats[ startTime ] then return end
	
	recentImportStats[ startTime ].TotalTime = KSDate.timeNowForAnimation() - startTime
	recentImportStats[ startTime ].InProgress = false
	recentImportStats[ startTime ].ImportThreadPeakMemoryAfter = WFCurrentThread.DEBUG_getPeakMemory() / 1024
	
	if checkMemoryPollingForMetricsEnabledAndStartIfNeeded() then

		local peakVM, peakRM = getWFAppMemoryUsage().getPeakUsage()
		local currentVM, currentRM = getWFAppMemoryUsage().getCurrentUsage()
		
		recentImportStats[ startTime ].PeakAppMemoryAfter_VM = convertToMegabytes( peakVM )
		recentImportStats[ startTime ].PeakAppMemoryAfter_RM = convertToMegabytes( peakRM )
		recentImportStats[ startTime ].CurrentAppMemoryAfter_VM = convertToMegabytes( currentVM )
		recentImportStats[ startTime ].CurrentAppMemoryAfter_RM = convertToMegabytes( currentRM )
		
	end
	
	-- Accumulate totals for all checkpoint component timings
	
	recentImportStats[ startTime ].TotalTimings = {}
	
	for k, v in pairs( recentImportStats[ startTime ] ) do
		
		if string.match( k, 'checkpoint' ) and v.Timings then
			
			for kk, vv in pairs( v.Timings ) do
				
				if not recentImportStats[ startTime ].TotalTimings[ kk ] then
					recentImportStats[ startTime ].TotalTimings[ kk ] = 0
				end
				
				recentImportStats[ startTime ].TotalTimings[ kk ] = recentImportStats[ startTime ].TotalTimings[ kk ] + vv
				
			end
			
		end
		
	end

end

--------------------------------------------------------------------------------

local function updateStatsAfterCancelation( handler )

	if handler.metricsStartTime and recentImportStats[ handler.metricsStartTime ] then
		
		local resultCounts = recentImportStats[ handler.metricsStartTime ].ResultCounts
		resultCounts.canceled = resultCounts.pending
		resultCounts.pending = 0
		
		updateStatsUponTermination( handler.metricsStartTime )

	end

end

--------------------------------------------------------------------------------

if COLLECT_IMPORT_METRICS then

	require 'WFDiagnosticLog'.addLogSectionFunc( function()
	
		if not recentImportStats then return end
		
		local modifiedImportStats = {}
		
		local function removeEndFields( t )
			
			local newT = {}
			
			for k, v in pairs( t ) do
				if not string.match( k, "^end" ) then
					newT[ k ] = v
				end
			end
			
			return newT
			
		end
		
		for time, stat in pairs( recentImportStats ) do
		
			local formattedTime = "Import started at mach_absolute_time: " .. time
			
			modifiedImportStats[ formattedTime ] = {}
		
			for k, v in pairs( stat ) do
				
				local checkpointNum = string.match( k, "checkpoint%-(%d+)" )
				
				if checkpointNum then
				
					if not modifiedImportStats[ formattedTime ].tenthsCheckpointData then
						modifiedImportStats[ formattedTime ].tenthsCheckpointData = {}
					end
					
					modifiedImportStats[ formattedTime ].tenthsCheckpointData[ "Checkpoint " .. checkpointNum - 1 ] = removeEndFields( v )
				else
					modifiedImportStats[ formattedTime ][ k ] = v
				end
				
			end
	
		end
		
		return {
			name = "Import Metrics",
			id = 'ImportMetrics',
			weight = 1600,
			outputForAutomation = true,
			body = modifiedImportStats,
		}
	
	end )

end
--------------------------------------------------------------------------------

local function addImportStatEntry( timestamp, optEntry )

	assert( timestamp )
	
	local entry = optEntry or	
		{ 
			InProgress = true,
			ResultCounts = {
				canceled = 0,
				success = 0,
				duplicate = 0,
				error = 0,
				pending = 0,
				errorDetails = {},
			},
			TotalFileSize = 0,
			TotalMegapixels = 0,
			ImportThreadPeakMemoryBefore = WFCurrentThread.DEBUG_getPeakMemory() / 1024,
		}
	
	if checkMemoryPollingForMetricsEnabledAndStartIfNeeded() then

		local peakVM, peakRM = getWFAppMemoryUsage().getPeakUsage()
		local currentVM, currentRM = getWFAppMemoryUsage().getCurrentUsage()
		
		entry.PeakAppMemoryBefore_VM = convertToMegabytes( peakVM )
		entry.PeakAppMemoryBefore_RM = convertToMegabytes( peakRM )
		entry.CurrentAppMemoryBefore_VM = convertToMegabytes( currentVM )
		entry.CurrentAppMemoryBefore_RM = convertToMegabytes( currentRM )
		
	end
	
	if recentImportStatsCount == MAX_IMPORT_STATS then
		
		-- Clear out the oldest timestamp'ed entry.
		
		local oldestKey
		for k, _ in pairs( recentImportStats ) do
			if not oldestKey or k < oldestKey then
				oldestKey = k
			end
		end
		
		recentImportStats[ oldestKey ] = nil
	
	else
		recentImportStatsCount = recentImportStatsCount + 1
	end
	
	recentImportStats[ timestamp ] = entry

end

--------------------------------------------------------------------------------

local function updateTotalQueued( startTime, totalQueued )

	if COLLECT_IMPORT_METRICS and recentImportStats[ startTime ] then
		recentImportStats[ startTime ].TotalEnqueued = totalQueued
		recentImportStats[ startTime ].ResultCounts.pending = totalQueued
	end

end

--==============================================================================

return {
	COLLECT_IMPORT_METRICS = COLLECT_IMPORT_METRICS,

	addImportStatEntry = addImportStatEntry,
	recentImportStats = recentImportStats,
	updateStatsAfterCancelation = updateStatsAfterCancelation,
	updateStatsUponTermination = updateStatsUponTermination,
	checkMemoryPollingForMetricsEnabledAndStartIfNeeded = checkMemoryPollingForMetricsEnabledAndStartIfNeeded,
	updateTotalQueued = updateTotalQueued,
}