--[[----------------------------------------------------------------------------

KSJsonDecoder.lua

@author scouten 2011-02-22
@reviewed mhamburg 2011-03-11

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

Q&D JSON decoder for interacting with prototype ps.com server.

------------------------------------------------------------------------------]]

	-- core.lua
local KSData = require 'KSData'

	-- third_party.yajl
local yajl = require 'yajl'


--============================================================================--

-- @mark KSJsonDecoder namespace

--[[!
	\class KSJsonDecoder

	\brief This namespace provides translation utilities between JSON-encoded
		REST request contents and the Lua structures used by the LUA API calls.
		These functions are typically used as the translation functions for
		\c mapE or \c mapB calls.
]]--

local KSJsonDecoder = {}

--------------------------------------------------------------------------------

local type = type
local tostring = tostring

local yajl_null = yajl.null

--------------------------------------------------------------------------------
--[[!
	\brief Translates a JSON string into the equivalent Lua data structure.
	
	\param json  The JSON string.
	\param ignoreEmpty (boolean) When true, ignores empty string. Otherwise,
		prints a debugging warning when it encounters an empty string in the
		source JSON string.

     \return The Lua table.
]]--

function KSJsonDecoder.decodeWebSafeJson( --[[string]] json,
										  --[[boolean]] ignoreEmpty )

	-- Convert KSData to string if needed.
	
	if KSData.isData( json ) then
		json = tostring( json )
	end
	
	-- Still not a string? Just pass through unaffected.

	if type( json ) ~= 'string' then
		return json
	end
	
	if #json == 0 then
		if not ignoreEmpty then
			debug.print( "attempted to parse empty JSON string at ", debug.traceback() )
		end
		return {}
	end
	
	-- Strip the while(1) silliness if it's present.
	
	json = json:gsub( "^%s*while%s*%(%s*1%s*%)%s*%{%}%s*", "", 1 )
	
	-- Convert to Lua.
	
	return yajl.to_value( json )

end

--------------------------------------------------------------------------------
--[[
	Removes NULL values from the Lua data structure created by translating a
	JSON string with decodeWebSafeJson. (Not a public API function.)
]]--

local function stripJsonNullFromTable( --[[table]] t )

	for k, v in pairs( t ) do
	
		if v == yajl_null then
			t[ k ] = nil
		elseif type( v ) == 'table' then
			stripJsonNullFromTable( v )
		end
	
	end
	
	return t

end

KSJsonDecoder.stripJsonNullFromTable = stripJsonNullFromTable

--------------------------------------------------------------------------------
--[[!
	\brief Translates a JSON string into the equivalent Lua data structure,
		converting any JSON NULL values to Lua nil.
	
	\param json (string) The JSON string.
	
	\return The Lua table.
]]--

function KSJsonDecoder.decodeWebSafeJson_stripNull( --[[string]] json )

	local result = KSJsonDecoder.decodeWebSafeJson( json )
	
	if type( result ) == 'table' then
		stripJsonNullFromTable( result )
	end
	
	return result
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Translates a JSON string into the equivalent Lua data structure,
		converting any JSON NULL values to Lua nil, but ignoring empty body text.
	
	Empty body text can happen when an HTTP request encounters errors.
	
	\param json The JSON string.
	
	\return The Lua table.
]]--

function KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( --[[string]] json )

	local result = KSJsonDecoder.decodeWebSafeJson( json, true )
	
	if type( result ) == 'table' then
		stripJsonNullFromTable( result )
	end
	
	return result
	
end

--------------------------------------------------------------------------------
--[[!
	\brief This mapping function can be used as the only argument to
		\c :mapP on an \c WFHttpClient.getP  or \c WFHttpClient.putP result.
		It transforms the HTTP body from a JSON string into the equivalent Lua
		data structure, converting any JSON NULL values to Lua nil.
	
	\param status A status value that identifies the type of the promise result,
		"error" or "data". If the status is "error", it is passed through along
		with its parameters.
	\param json The JSON string.
	\param ... Addtional parameters of an error result to pass through.
	
	\return The status value, the Lua table, and additional pass-through parameters
		for an error.
]]--

function KSJsonDecoder.decodeWebSafeJson_stripNullP( --[[string]] status,
													 --[[string]] json,
													 ... )

	if status == 'data' then

		local result = KSJsonDecoder.decodeWebSafeJson( json )
		
		if type( result ) == 'table' then
			stripJsonNullFromTable( result )
		end
		
		return 'data', result, ...

	else
		
		return status, json, ...
	
	end
	
end

--------------------------------------------------------------------------------
--[[!
	\brief This mapping function can be used as the only argument to \c :mapP
		on an \c WFHttpClient.getP or \c WFHttpClient.putP result. It transforms
		the HTTP body from a JSON string into the equivalent Lua data structure,
		converting any JSON NULL values to Lua nil, but ignoring empty body text.
	
	Empty body text can happen when an HTTP request encounters errors.
	
	\param status A status value that identifies the type of the promise result,
		"error" or "data". If the status is "error", it is passed through along
		with its parameters.
	\param json The JSON string.
	\param ... Addtional parameters of an error result to pass through.
	
	\return The status value, the Lua table, and additional pass-through
		parameters for an error.
]]--

function KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP( --[[string]] status,
																 --[[string]] json,
																 ... )

	if status == 'data' then

		local result = KSJsonDecoder.decodeWebSafeJson( json, true )
		
		if type( result ) == 'table' then
			stripJsonNullFromTable( result )
		end
		
		return 'data', result, ...
	
	else
	
		return status, json, ...
	
	end
	
end

--------------------------------------------------------------------------------

return KSJsonDecoder
