--[[----------------------------------------------------------------------------

KSJsonEncoderTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-03-08

------------------------------------------------------------------------------]]

	-- libs.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- libs.table
local KSTableArchiving = require 'KSTableArchiving'


--============================================================================--

local KSJsonEncoderTest = {}

--------------------------------------------------------------------------------

-- GRUNGE [scouten 2011-02-22]: Some of this code is duplicated from
-- KSBlondeTest. Opportunity for refactoring code.

--------------------------------------------------------------------------------

local function convertTableToStringForDebugIfNeeded( value )

	if type( value ) == 'table' then
		return KSTableArchiving.simpleTableToChunkString( value )
	else
		return value
	end

end

--------------------------------------------------------------------------------

local compareValues, deepCompareTables

--------------------------------------------------------------------------------

-- Performs a deep comparison of two tables, nested up to 20 levels deep. Returns
-- true if every key-value pair in t1 also appears in t2 and the number of
-- key-value pairs in t1 and t2 are equal. Returns false otherwise.

function deepCompareTables( t1, t2, depth )

    if depth == 0 then
        error( "Tables are too deeply nested to be compared" )
    end
    
    local t1count = 0
    local t2count = 0
    
    for k, v in pairs( t1 ) do
        t1count = t1count + 1
        local equal = compareValues( v, t2[ k ], depth )
        if not equal then return false end
    end
    
    for k, v in pairs( t2 ) do
        t2count = t2count + 1
    end
    
    return t1count == t2count

end

--------------------------------------------------------------------------------

function compareValues( v1, v2, level )

    if type( v2 ) == "table" and type( v1 ) == "table" then
        if not level then level = 20 end
        return deepCompareTables( v1, v2, level - 1 )
    else
        return ( v2 == v1 )
    end
        
end

--------------------------------------------------------------------------------

local function testWebSafeEncodeThenDecode( luaValue )

	local actualJson = KSJsonEncoder.encodeWebSafeJson( luaValue )

	assert( type( actualJson ) == 'string',
		"KSJsonEncode.encodeWebSafeJson: did not return a string value" )

	assert( actualJson:sub( 1, 13 ) == 'while (1) {}\n',
		"KSJsonEncode.encodeWebSafeJson: did not include exec guard" )

	-- [scouten 2011-03-11]: Ideally, one would want to have an expected
	-- string and simply compare against that. But ... the table keys
	-- are encoded in random order, so we can't do that. Instead we
	-- re-decode the JSON and make sure it matches the original Lua value.
	-- (Note that the JSON decode does work as a standalone test.)
	
	local reconstructedLuaValue = KSJsonDecoder.decodeWebSafeJson( actualJson )
	
	if not compareValues( luaValue, reconstructedLuaValue ) then
		local message = string.format(
					"KSJsonEncoder.encodeWebSafeJson: value did not decode correctly\n\nexpected:\n%s\n\nactual:\n%s\n",
					convertTableToStringForDebugIfNeeded( luaValue ),
					convertTableToStringForDebugIfNeeded( reconstructedLuaValue ) )
		print( message )
		error( message )
	end

end

--------------------------------------------------------------------------------

function KSJsonEncoderTest.testBasicString()

	testWebSafeEncodeThenDecode {
		status = 'ok',
		last_event_received = '2011-02-22T21:21:40.711Z',
		events = {
			{
				id = 'bc63f8cecc09469fbc40fe544119251a',
				href = '/v1.0/accounts/7a6b9ad57de3404cbc54305878afbc35/partitions/9a24b25d114c4a008ba051beb54d0ab2/events/bc63f8cecc09469fbc40fe544119251a',
				partition = {
					id = '9a24b25d114c4a008ba051beb54d0ab2',
					href = '/v1.0/accounts/7a6b9ad57de3404cbc54305878afbc35/partitions/9a24b25d114c4a008ba051beb54d0ab2'
				},
				date_created = '2011-02-22T21:21:40.711Z',
				date_received = '2011-02-22T21:21:40.711Z',
				class = 'favorite',
				value = 'true'
			},
			{
				id = '5667d946f3eb4733a0986a72f1d24108',
				href = '/v1.0/accounts/7a6b9ad57de3404cbc54305878afbc35/partitions/d4e259b3dad84a629d2a32e0c074ffb4/events/5667d946f3eb4733a0986a72f1d24108',
				partition = {
					id = 'd4e259b3dad84a629d2a32e0c074ffb4',
					href = '/v1.0/accounts/7a6b9ad57de3404cbc54305878afbc35/partitions/d4e259b3dad84a629d2a32e0c074ffb4'
				},
				date_created = '2011-02-22T21:21:40.711Z',
				date_received = '2011-02-22T21:21:40.711Z',
				class = 'favorite',
				value = 'false'
			},
		},
	}

end

--------------------------------------------------------------------------------

return KSJsonEncoderTest
