--[[----------------------------------------------------------------------------

KSJsonRestUtils.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-09-19

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local KSLuaUtils = require 'KSLuaUtils'
local WFArguments = require 'WFArguments'

	-- libs.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- wichitafoundation.web.http_client
local WFHttpClient = require 'WFHttpClient'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local dpcall = assert( KSLuaUtils.dpcall )

--============================================================================--

-- @mark KSJsonRestUtils namespace
--[[!

	\class KSJsonRestUtils

	\brief This namespace provides utilities for packing and unpacking
		REST request contents.

]]--
local KSJsonRestUtils = {}

--------------------------------------------------------------------------------

local function mapStatusJsonAndHeaders( status, rawJson, ... )

	local json

	if rawJson then
		dpcall( 'mapStatusJsonAndHeaders', function()
			json = KSJsonDecoder.decodeWebSafeJson( rawJson )
				-- Add config option for converting JSON null to Lua nil?
		end )
	end
	
	return json and status or 'error', json or rawJson, ...

end

--------------------------------------------------------------------------------

--[[!
	\brief Sends an HTTP GET request using WFHttpClient, and transforms the response body
		 from JSON encoded data into a Lua data structure

	\param url (string) The URL for the request.
	\param optArgs (table) Additional parameters for the request.

	\sa WFHttpClient.getP

	\return (KSPromise) A promise that resolves to an error if the request
	has an error result, otherwise to the decoded result and response headers.

]]--

function KSJsonRestUtils.jsonGetP( --[[string]] url, --[[table]] optArgs )

	check( 'string', '?table' )

	optArgs = optArgs or {}
	optArgs.requestHeaders = optArgs.requestHeaders or {}
	optArgs.requestHeaders[ #optArgs.requestHeaders + 1 ]
		= { header = 'Accept', value = 'application/json' }
	
	return WFHttpClient.getP( url, optArgs )
		:mapP( mapStatusJsonAndHeaders )
		
end

--------------------------------------------------------------------------------

--[[!
	\brief Sends an HTTP PUT request using WFHttpClient, and transforms the response body
		 from JSON encoded data into a Lua data structure

	\param url (string) The URL for the request.
	\param data (string) The data to PUT.
	\param optArgs (table) Additional parameters for the request.

	\sa WFHttpClient.putP

	\return (KSPromise) A promise that resolves to an error if the request
	has an error result, otherwise to the decoded result and response headers.

]]--

function KSJsonRestUtils.jsonPutP( --[[string]] url, --[[string]] data, --[[table]] optArgs )

	check( 'string', '_', '?table' )
		-- FIXME [scouten 2011-09-17]: Is there a way to check for string or KSData?

	optArgs = optArgs or {}
	optArgs.requestHeaders = optArgs.requestHeaders or {}

	optArgs.requestHeaders[ #optArgs.requestHeaders + 1 ]
		= { header = 'Content-Type', value = 'application/json' }
	optArgs.requestHeaders[ #optArgs.requestHeaders + 1 ]
		= { header = 'Accept', value = 'application/json' }

	local jsonData = data and KSJsonEncoder.encodeOzUploadJson( data )
	
	return WFHttpClient.putP( url, jsonData, optArgs )
		:mapP( mapStatusJsonAndHeaders )
		
end

--------------------------------------------------------------------------------

return KSJsonRestUtils
