--[[----------------------------------------------------------------------------

KSLogging.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author epoley 2010-09-15

@reviewed epoley 2010-11-02
@reviewed scouten 2010-11-02
@reviewed mhamburg 2010-11-02
@reviewed erwright 2010-11-02

--------------------------------------------------------------------------------

[epoley 9/15/10]

KSLogging is a simple lua library providing support for configurable logging in
Lua. Logs are always named, and there can be arbitrarily many of them.

Example usage:

	require "KSLogging"
	KSLogging.log( "foo" ):warn( "Impending doom..." )

KSLogging is built on top of LuaLogging, adding support for named logs. I have
also tweaked the logging module to support formatted messages and to fix a bug.
This means that KSLogging supports LuaLogging's logging methods, because
KSLogging.log(name) returns a LuaLogger object.

Search for 'epoley' in logging.lua to see what I changed.

The self-explanatory log methods are:

:log( level, message )
:debug( message )
:info( message )
:warn( message )
:error( message )
:fatal( message )

There are currently five ways to make a log:

1) Calling KSLogging.log( name ) when a log with name does not exist will create a
"default" log (See below).

2) KSLogging.makeDefaultLog( name ) makes a log that uses debug.print (which routes
to NSLog) to print messages in the following format:

> KSLog name: LEVEL message

3) KSLogging.makeLog( name, logFunction ) makes a custom log that logs by calling
the given function. It can take self, level, and message as parameters, and should
return true or false.

4) KSLogging.makeDefaultLogWithFile( name, filepath ) makes a log that uses the
standard lua function to output to a file at the given path.

5) KSLogging.makeDatedLogWithFile( name, filepath, datePattern): as above, except
that datePattern is used in composing the filename, in order to create daily or
monthly log files. From LuaLogging documentation:

(http://www.keplerproject.org/lualogging/file.html)

datePattern:
This is an optional parameter that can be used to specify a date pattern that will be passed to the
os.date function to compose the filename. This is useful to create daily or monthly log files.
If the user wants to create one log file per day he specifies a "%Y-%m-%d" pattern and a filename
like "temp%s.log".
	

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.table
local KSTableArchiving -- = require 'KSTableArchiving' (lazy loaded)
	-- FIXME [scouten 2011-05-04]: Reference from core to libs.


local KSLogging = { logs = {} }

--=============================================================================

--			LuaLogging

-------------------------------------------------------------------------------

-- This section of code originates from LuaLogging, but has been liberally tweaked
-- to serve our needs. Rather than a module, it is now mostly a set of local
-- functions used by KSLogging. Changes include demodulization and string.format.
-- However, it remains the case that this level has no concept of named logs. [epoley]

-------------------------------------------------------------------------------

-- @author Ellie Poley (epoley@adobe.com)

-- @author Danilo Tuler (tuler@ideais.com.br)
-- @author Andre Carregal (info@keplerproject.org)
-- @author Thiago Costa Ponte (thiago@ideais.com.br)
--
-- @copyright 2004-2007 Kepler Project
-- @release $Id$

-------------------------------------------------------------------------------

-- The DEBUG Level designates fine-grained informational events that are most
-- useful to debug an application
KSLogging.DEBUG = "DEBUG"

-- The INFO level designates informational messages that highlight the
-- progress of the application at coarse-grained level
KSLogging.INFO = "INFO"

-- The WARN level designates potentially harmful situations
KSLogging.WARN = "WARN"

-- The ERROR level designates error events that might still allow the
-- application to continue running
KSLogging.ERROR = "ERROR"

-- The FATAL level designates very severe error events that will presumably
-- lead the application to abort
KSLogging.FATAL = "FATAL"

local LEVEL = {
	[ KSLogging.DEBUG ] = 1,
	[ KSLogging.INFO ]  = 2,
	[ KSLogging.WARN ]  = 3,
	[ KSLogging.ERROR ] = 4,
	[ KSLogging.FATAL ] = 5,
}

-------------------------------------------------------------------------------

-- Prepares the log message

--[[ GRUNGE [scouten 2010-10-29]: Halide says this function isn't being used...
local function prepareLogMsg( pattern, dt, level, message )

    local logMsg = pattern or "%date %level %message\n"
    message = string.gsub(message, "%%", "%%%%")
    logMsg = string.gsub(logMsg, "%%date", dt)
    logMsg = string.gsub(logMsg, "%%level", level)
    logMsg = string.gsub(logMsg, "%%message", message)
    return logMsg
    
end
]]--

-------------------------------------------------------------------------------

local threadName

-- Converts a Lua value to a string
-- Converts Table fields in alphabetical order

local function logger_tostring( value )

	-- If it's not a table, make it a string.

	if value == ':traceback:' then
		
		return debug.traceback( '', 4 )
	
	elseif value == ':thread:' then
		
		if not threadName then
			threadName = require 'WFCurrentThread'.DEBUG_getThreadName()
		end
		
		return threadName

    elseif type( value ) ~= 'table' then

        if type( value ) == 'string' then
            return string.format( "%q", value )
        else
            return tostring( value )
        end

	else
	
		if not KSTableArchiving then
			KSTableArchiving = assert( require 'KSTableArchiving' )
		end
	
		local success, result = pcall(
			KSTableArchiving.simpleTableToChunkString, value )

		if not ( success and result ) then
			success, result = pcall( tostring, value )
		end
		
		return result
			-- NOTE [scouten 2011-05-04]: If tostring fails,
			-- we'll just use the error message as a proxy for
			-- the table value.
		
	end
	
	--[[ -- Old code for listing tables. Doesn't format nested tables as nicely.
		 -- FIXME [scouten 2011-05-04]: Remove this?
	
	else
	
		

	-- If it's a table, recursively build a string representation.
	-- List keys in alphabetical order.

    else
    
        local auxTable = {} -- String based keys
		local numTable = {} -- Numeric keys
		local trueKeyTable = { } -- Mapping from encoded keys to true keys
    
    	for k, v in pairs( value ) do
        	if type( k ) == 'number' then
				numTable[ #numTable + 1 ] = k
            else
				local sk = logger_tostring( k )
                auxTable[ #auxTable + 1 ] = sk
				trueKeyTable[ sk ] = k
            end
        end
        
        table.sort( auxTable )
        table.sort( numTable )

		local entries = { }
		
		for _, fieldName in ipairs( auxTable ) do
			local key = trueKeyTable[ fieldName ]
			local entry = "[ " .. fieldName .. " ] = " .. logger_tostring( value[ key ] )
			entries[ #entries + 1 ] = entry
		end
		
		for _, k in ipairs( numTable ) do
			local entry = "[ " .. tostring( k ) .. " ] = " .. logger_tostring( value[ k ] )
			entries[ #entries + 1 ] = entry
		end
			
		return "{ " .. table.concat( entries, ", " ) .. " }"

    end

	--]]
    
end

-------------------------------------------------------------------------------
-- Creates a new logger object
-- @param append Function used by the logger to append a message with a
--	log-level to the log stream.
-- @return Table representing the new logger object.
-------------------------------------------------------------------------------

local function new( append )

	if type( append ) ~= "function" then
		return nil, "Appender must be a function."
	end

	local logger = {}
	logger.level = KSLogging.WARN
	logger.append = append

	logger.setLevel = function ( self, level )
		assert( LEVEL[ level ], string.format( "undefined level `%s'", logger_tostring( level ) ) )
		self.level = level
		return self
	end

-- START changes by [epoley 9/15/10]

--[[ In order to support string formatting, the following chunk has been modified to support varargs
	 and to call string.format when message is a string, rather than doing nothing ]]

	logger.log = function ( self, level, message, ... )

		assert( LEVEL[ level ], string.format( "undefined level `%s'", logger_tostring( level ) ) )
		
		if LEVEL[ level ] < LEVEL[ self.level ] then
			return
		end
		
		-- Handle non-strings.

		if type( message ) ~= "string" then

		 	local success, result = pcall( logger_tostring, message )
		  	
		  	if success and type( result ) == 'string' then
		  		message = result
		 	else
		    	message = "tostring failed"
		  	end
		  
		-- Handle format strings.

		elseif select( "#", ... ) > 0 then

			local args = {}
			for i = 1, select( "#", ... ) do
				
				args[ i ] = logger_tostring( select( i, ... ) )
				
			end

		  	message = string.format( message, unpack( args, 1, select( "#", ... ) ) )

		end
		
		return logger:append( level, message )
	
	end

	-- Set up some shortcuts.
	
	logger.debug = function ( logger, ... ) return logger:log( KSLogging.DEBUG, ... ) end
	logger.info  = function ( logger, ... ) return logger:log( KSLogging.INFO,  ... ) end
	logger.warn  = function ( logger, ... ) return logger:log( KSLogging.WARN,  ... ) end
	logger.error = function ( logger, ... ) return logger:log( KSLogging.ERROR, ... ) end
	logger.fatal = function ( logger, ... ) return logger:log( KSLogging.FATAL, ... ) end
	
	-- END changes by [epoley 9/15/10]
	
	return logger
	
end

--=============================================================================

--			KSLogging

-------------------------------------------------------------------------------

-- This section of code builds on top of the LuaLogging core code above, adding
-- support for named (and cached) logs, routing through NSLog via debug.print
-- (by default), and other convenience methods for creating (and destroying) logs.

-------------------------------------------------------------------------------

-- This function only affects the log objects themselves. Obviously any output
-- and/or files that already exist will not be affected. This is useful for
-- tearing down between unit tests.

function KSLogging.removeAllLogs()

	KSLogging.logs = {}
	
end

------------------------------------------------------------------------------

function KSLogging.log( name )

	if type( name ) ~= 'string' then
		error( string.format( "KSLogging.log: name should be a string, was %s",
							type( name ) ), 2 )
	end

	local log = KSLogging.logs[ name ]
	
	if log then
		return log
	end
	
	-- No log, so go create one.
	
	KSLogging.makeDefaultLog( name )
		
	return KSLogging.logs[ name ]
	
end

------------------------------------------------------------------------------

function KSLogging.makeLog( name, logFunction )

	KSLogging.logs[ name ] = new( logFunction )

end

------------------------------------------------------------------------------

function KSLogging.makeDefaultLog( name )

	local heading = string.format( "> KSLog %s: ", name )
	local logprint = function( self, level, message )
						debug.print( heading, level, message )
						return true
					end
	
	KSLogging.makeLog( name, logprint )

end

------------------------------------------------------------------------------

function KSLogging.makeDefaultLogWithFile( name, filepath )

	local logging_file = require "logging.extensions.file" -- FIXME [scouten 2011-12-12]: Probably doesn't work. Do we care?
	KSLogging.logs[ name ] = logging_file( filepath )

end

------------------------------------------------------------------------------

function KSLogging.makeDatedLogWithFile( name, filepath, datePattern )

	local logging_file = require "core.logging.extensions.file" -- FIXME [scouten 2011-12-12]: Probably doesn't work. Do we care?
	KSLogging.logs[ name ] = logging_file( filepath, datePattern )

end

------------------------------------------------------------------------------

return KSLogging
