--[[----------------------------------------------------------------------------

KSLuaYajlTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

Adapted from test.lua in the lua-yajl distribution. Raw test of Lua YAJL in our
environment.

--------------------------------------------------------------------------------

@reviewed mhamburg 2011-03-11

------------------------------------------------------------------------------]]

local yajl = require 'yajl'
local WFPlatformInfo = require 'WFPlatformInfo'


--============================================================================--

local KSLuaYajlTest = {}

--------------------------------------------------------------------------------

local function ok( condition, message )

	if not condition then
		error( message, 2 )
	end

end

--------------------------------------------------------------------------------

local function toValue( string )

	local result

	local stack = {
		function( val )
			result = val
		end
	}

	local obj_key

	local events = {

		value = function( _, val )
			stack[ #stack ]( val )
		end,

		open_array = function()

			local arr = {}

			local idx = 1
			stack[ #stack ]( arr )

			table.insert( stack, function( val )
				arr[ idx ] = val
				idx = idx + 1
			end )
		end,

		open_object = function()

			local obj = {}

			stack[ #stack ]( obj )
			table.insert( stack, function( val )
				obj[ obj_key ] = val
			end )

		end,

		object_key = function( _, val )
			obj_key = val
		end,

		close = function()
			stack[ #stack ] = nil
		end,

	}
	
	yajl.parser( { events = events } )( string )

	return result

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testToValue()

	local json = '[["float",1.5,"integer",5,"string","hello","empty",[],"false",false,"custom","custom json serializer","ostr_key",{"key":"value"},"obool_key",{"true":true},"onull_key",{"null":null}],10,10.3,10.3,"a string",null,false,true]' --]]
	
	local val = yajl.to_value( json )
	local got = yajl.to_string( val )
	
	ok( got == json,
		"yajl.to_value(" .. json .. ") -> " .. got )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testGenerator()
	
	local strings = {}
	local generator = yajl.generator {
		printer = function( string )
			table.insert( strings, string )
		end
	}
	
	local custom = {
		__gen_json = function( self, gen )
			gen:string( "custom json serializer" )
		end
	}
	setmetatable( custom, custom )
	
	generator:open_array()

	generator:value( {
			"float", 1.5,
			"integer", 5,
			"string", "hello",
			"empty", {}, -- defaults to an empty array.
			"false", false,
			"custom", custom,
			"ostr_key", { key = "value" },
			"obool_key", { [ true ] = true },
			"onull_key", { [ yajl.null ] = yajl.null },
		} )

	generator:integer( 10.3 )
	generator:double( 10.3 )
	generator:number( 10.3 )
	generator:string( "a string" )
	generator:null()
	generator:boolean( false )
	generator:boolean( true )
	generator:open_object()
	generator:close()
	generator:close()
	
	local expect = '[["float",1.5,"integer",5,"string","hello","empty",[],"false",false,"custom","custom json serializer","ostr_key",{"key":"value"},"obool_key",{"true":true},"onull_key",{"null":null}],10,10.3,10.3,"a string",null,false,true,{}]' --]]

	local got = table.concat( strings )
	ok( expect == got, expect .. " == " .. got )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testSimple()

	local expect =
		'[' ..
			'"float",1.5,' ..
			'"integer",5,' ..
			'"true",true,' ..
			'"false",false,' ..
			'"null",null,' ..
			'"string","hello",' ..
			'"array",[1,2],' ..
			'"object",{"key":"value"}' ..
		']'
	
	-- Input to toValue matches the output of to_string:

	local got = yajl.to_string( toValue( expect ) )
	ok( expect == got, expect .. " == " .. tostring( got ) )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testNullAtEndOfArray()
	
	local expect = '["something",null]'
	local got = yajl.to_string( toValue( expect ) )
	
	ok( expect == got, expect .. " == " .. tostring( got ) )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testNullObjectValue()
	
	local expect = '{"something":null}'
	local got = yajl.to_string( toValue( expect ) )
	
	ok( expect == got, expect .. " == " .. tostring( got ) )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testWeirdNumbers()
	
	-- [scouten 2011-02-22]: Comment from original lua-yajl test file:
	-- See note in README about how we are broken when it comes to big numbers.

	local expect = '[1e+666,-1e+666,-0]'

	-- [nicholas 2012-12-17]: Android printf turns -0 into "0" and thus the
	-- -0 part of the test fails.  I am sure this won't get fixed anytime soon
	if ( WFPlatformInfo.isAndroid ) then
		expect = '[1e+666,-1e+666,0]'
	end
		
	local got = yajl.to_string( toValue( expect ) )
	ok( expect == got, expect .. " == " .. tostring( got ) )
	
	local nan = 0/0
	got = yajl.to_string { nan };
	ok( "[-0]" == got, "NaN converts to -0 (got: " .. got .. ")" )

end

--------------------------------------------------------------------------------

function KSLuaYajlTest.testNumberInString()
	
	-- Thanks to agentzh for this bug report!

	local expect = '{"2":"two"}'
	local got = yajl.to_string { [ "2" ] = "two" }

	ok( expect == got, expect .. " == " .. tostring( got ) )

end

--------------------------------------------------------------------------------

return KSLuaYajlTest
