--[[----------------------------------------------------------------------------

KSOAuth2SessionImp.lua

@author scouten 2011-11-21

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This is the backing event processor for KSOAuth2Session. This API is considered
private and should only be accessed via the KSOAuth2Session class.

One event processor is created per session ID.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'

	-- wichitafoundation.core.lua.threading
local WFRequestDispatcher = require 'WFRequestDispatcher'
local WFTimer = require 'WFTimer'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.string
local WFZString = require 'WFZString'

	-- wichitafoundation.core.utils
local KSTableUtils = require 'KSTableUtils'

	-- libs.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- libs.prefs
local KSPrefs = require 'KSPrefs'

	-- os.localization
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- wichitafoundation.web.http_client
local WFHttpClient = require 'WFHttpClient'

	-- services.keychain_store
local KSKeychainService = require 'KSKeychainService'


--------------------------------------------------------------------------------

local ZSTR = assert( WFZString.fromZString )


--==============================================================================

local IMS_TOKEN_SERVICE = "/token"
local IMS_LOGOUT_SERVICE = "/logout"
local IMS_TOKEN_EXCHANGE = "/exchange/access_token"

--------------------------------------------------------------------------------

local oAuth2ImpLog = KSLogging.log 'OAuth2SessionImp'
--	:setLevel( KSLogging.DEBUG )

WFRequestDispatcher.enableDiagnosticLogForRequestStats()

--------------------------------------------------------------------------------

local userInfoB = KSReactive.variableB( {} )

local hasValidLoginB = KSReactive.variableB( true )

local languageCode = WFLocalizationUtils.getLanguageCode()

local KSData_isData = assert( KSData.isData )

--------------------------------------------------------------------------------
-- Shared code to process new user info (whether read from keychain or
-- internal source).

local function updateUserInfoB( userInfoJson )

	local strippedCopy = KSTableUtils.deepCopy( userInfoJson )
	strippedCopy.access_token = nil
	strippedCopy.refresh_token = nil
	
	local oldUserInfo = userInfoB:valueNow()
	
	if not oldUserInfo or not KSTableUtils.deepEquals( oldUserInfo, strippedCopy ) then
		userInfoB:set( strippedCopy )
	end

end

--------------------------------------------------------------------------------
-- On startup, read the previous state (if any) from the keychain.
-- Populate the initial token value from this info.

local initMessageProcessor = function( sessionId )
	
	assert( type( sessionId ) == 'string', "sessionId must be a string" )
	assert( #sessionId == 32, "sessionId must be a compact UUID (32 characters" )
	
	local accessAndRefreshTokensP
	local accessAndRefreshTokensP_original
	
	accessAndRefreshTokensP = KSKeychainService.readP( sessionId )
		:mapP {
	
			data = function( status, rawJson )
	
				if rawJson then
				
					local decodedJson = KSJsonDecoder.decodeWebSafeJson_stripNull( rawJson )
					
					if decodedJson and decodedJson.access_token and decodedJson.refresh_token then
						
						updateUserInfoB( decodedJson )
	
						return 'data', decodedJson.access_token, decodedJson.refresh_token
					
					elseif decodedJson and decodedJson.access and decodedJson.refresh then
					
						-- [scouten 2011-11-22]: Prior to OAuth2 refactoring, we used
						-- these (non-standard) terms. This is here so we don't cause
						-- previous clients to be logged out.
	
						decodedJson.access_token = decodedJson.access
						decodedJson.access = nil
						
						decodedJson.refresh_token = decodedJson.refresh
						decodedJson.refresh = nil
						
						updateUserInfoB( decodedJson )
	
						return 'data', decodedJson.access_token, decodedJson.refresh_token
					
					end
				
				end
				
				-- No data or incomplete data in keychain. Make this into an error.
	
				oAuth2ImpLog:error( "no refresh/access token found in keychain result: %s", rawJson )
							
				-- FIXME [scouten 2011-11-29]: We're threading a very tricky balance between
				-- two bugs (#3057613 and #3056308). In the former case we just turn the
				-- connection light red when the tokens are missing and leave it that
				-- way when we should force a logout, but in the latter case we are
				-- forcing a logout immediately after logging in (when we should be
				-- expecting a storeTokensP message in just a moment). Relying on a timer
				-- to "fix" this now, but we should probably look for a more solid way
				-- to pass in the initial tokens in the first-login case in a future
				-- version.
				
				WFTimer.afterDelayDo( 10, function()
				
					-- This if clause answers the question: Did someone else provide newer
					-- tokens since we were here originally? If so, don't force a logout.
					-- If not, that probably means we'll never get new tokens.
				
					if accessAndRefreshTokensP_original == accessAndRefreshTokensP then
						oAuth2ImpLog:error( "forcing logout because no tokens were provided" )
						hasValidLoginB:set( false )
					end
				
				end )
				
				return 'error', "No refresh/access token stored in keychain"
				
			end,
			
			error = function( status, message )
			
				oAuth2ImpLog:error( "forcing logout because error occurred reading keychain" )
				hasValidLoginB:set( false )
	
				return status, message
			
			end,
		
		}
	
	accessAndRefreshTokensP_original = accessAndRefreshTokensP
	
	accessAndRefreshTokensP:runUntilDone()
	
	--------------------------------------------------------------------------------
	-- Internal code to store new user info.
	
	local function storeNewUserInfoP( userInfo )
	
		if not userInfo or not userInfo.access_token or not userInfo.refresh_token then
			oAuth2ImpLog:error( "storeNewUserInfoP: userInfo = %s", userInfo )
		else
			oAuth2ImpLog:debug( "storeNewUserInfoP: userInfo = %s", userInfo )
		end
	
		assert( userInfo.access_token, "storeNewUserInfoP: missing access_token" )
		assert( userInfo.refresh_token, "storeNewUserInfoP: missing refresh_token" )
		
		hasValidLoginB:set( true )
		
		return KSKeychainService.writeP( sessionId,
										 KSJsonEncoder.encodeOzUploadJson( userInfo ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'storeTokensP',
	function( refreshToken, accessToken )
	
		-- Hack up a small user info packet for now.
	
		if not refreshToken then
			error "KSOAuth2SessionImp:storeTokensP: missing refreshToken"
		end
	
		if not accessToken then
			error "KSOAuth2SessionImp:storeTokensP: missing accessToken"
		end
	
		local userInfo = {
			access_token = accessToken,
			refresh_token = refreshToken
		}
		
		oAuth2ImpLog:info( "received new refresh and access tokens" )
		
		accessAndRefreshTokensP = KSPromises.resolvedPromiseP( 'data',
															   userInfo.access_token,
															   userInfo.refresh_token )
	
		local resultP = storeNewUserInfoP( userInfo )
		
		resultP:runUntilDone()
		
		return resultP
	
	end )
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'userInfoB', function()
		
		return userInfoB:relayB()
	
	end )
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'hasValidLoginB', function()
	
		return hasValidLoginB:relayB()
	
	end )
	
	--------------------------------------------------------------------------------
	
	local function deleteAuthKeys()
	
		KSKeychainService.deleteP( sessionId )
		
		userInfoB:set( {} )
		
		accessAndRefreshTokensP = KSPromises.resolvedPromiseP( 'error', "Logged out" )
	
	end
	
	--------------------------------------------------------------------------------
	
	
	local function getRefreshTokenUrlAndPostBodyP( imsBaseUrl, refreshToken )
	
		--	On entry: ( imsBaseUrl refreshToken )
		--	On exit:  ( url )
				
		local url = string.format( "%s%s",
								 imsBaseUrl,
								 IMS_TOKEN_SERVICE )
				
	
		local Bablu = require "Bablu"
		
		return Bablu:getImsClientIdAndSecretP():mapDataP( function( _, clientId, clientSecret )
			assert( clientId ~= nil, "clientId is nil in getRefreshTokenUrlAndPostBodyP" )
			assert( clientSecret ~= nil, "clientSecret is nil in getRefreshTokenUrlAndPostBodyP" )
			
			local postBody = string.format( "grant_type=refresh_token&refresh_token=%s&client_id=%s&client_secret=%s",
									refreshToken,
									clientId,
									clientSecret )
			
			return 'data', url, postBody
		
		end )
	
	end
	
	--------------------------------------------------------------------------------
	
	local function refreshRequestP( refreshToken, accessToken, authHeader )
	
		-- FIXME [scouten 2011-11-28]: Don't like the hard-coded reference to
		-- Oz's Www-Authenticate header, but not worth fixing right now.
	
		if not authHeader then
			error( "refreshRequestP: missing authHeader parameter", 2 )
		end
	
		local authUrl = authHeader:match( [[auth_url="([^"]+)"]] )
									--" (fix up Halide's syntax coloring)
		
		if not authUrl then
			error( "refreshRequestP: authHeader did not contain a valid URL", 2 )
		end
		
		authUrl = authUrl:match( [[(.*)/authorize]] )
	
		local userDataP = getRefreshTokenUrlAndPostBodyP( authUrl, refreshToken ):mapDataP( function( status, refreshUrl, postBody )
			
			oAuth2ImpLog:debug( "refreshing using url = %s\n\npostBody = %s",
								refreshUrl, postBody )
		
			oAuth2ImpLog:debug( "Access Token = %s", accessToken )
			oAuth2ImpLog:debug( "Refresh Token = %s", refreshToken )
			
			local refreshPostArgs = {
				requestMethod = 'POST',
				requestHeaders = {
					{ header = 'Content-Type', value = 'application/x-www-form-urlencoded' },
				},
			}
			
			return WFHttpClient.putP( refreshUrl, postBody, refreshPostArgs )
				:mapP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
		end )
		
		userDataP:whenResolvedDo( function( status, userInfo )
			
			oAuth2ImpLog:debug( "refreshing yielded status = %s data = %s",
								status, userInfo )
	
			if status == 'data' then
	
				if not userInfo.refresh_token then
					userInfo.refresh_token = refreshToken
				end
	
				storeNewUserInfoP( userInfo ):runUntilDone()
	
			end
			
		end ):runUntilDone()
	
		local resultP = userDataP:mapP {
		
			data = function( _, userInfo )
			
				oAuth2ImpLog:debug( "refreshRequestP returned %s", userInfo )
				
				if not userInfo.access_token then
					oAuth2ImpLog:error( "refreshRequestP missing access_token :: returned %s", userInfo )
					return 'error', "missing access_token"
				end
	
				return 'data', userInfo.access_token, userInfo.refresh_token or refreshToken
			
			end,
	
			error = function( _, data, headers )
			
				local statusCode = headers and headers.status_code
				
				oAuth2ImpLog:error( "refreshToken call failed with\n%s\n%s", headers, data )
				oAuth2ImpLog:error( "Refresh Failed. Previous Access Token = %s", accessToken )
				oAuth2ImpLog:error( "Refresh Failed. Previous Refresh Token = %s", refreshToken )
				
				if not statusCode then
				
					oAuth2ImpLog:error( "refreshToken call failed with transient error:\n\ndata = %s\n\nheaders = %s",
										data, headers )
				
				elseif statusCode >= '400' and statusCode <= '499' then
				
					oAuth2ImpLog:error( "refreshToken call failed with permanent error: forcing new login\n%s\n%s", headers, data )
					
					oAuth2ImpLog:error( "forcing logout because refreshToken failed" )
					hasValidLoginB:set( false )
	
					deleteAuthKeys()
	
				end
			
				return 'error', data, headers
			
			end,
		
		}
	
		resultP:runUntilDone()
		
		return resultP
	
	end
	
	--------------------------------------------------------------------------------
	
	local function getPutImpP( url, args, getPutInnerFunc )
	
		oAuth2ImpLog:info( "getPutImp for URL %s", url )
	
		if not accessAndRefreshTokensP then
			error "Not logged in"
		end
		
		local accessAndRefreshTokensP_atStart = accessAndRefreshTokensP
		
		return accessAndRefreshTokensP:mapDataP( function( _, accessToken, refreshToken )
	
			-- [scouten 2011-06-28]: If this happens, please let me know. If possible,
			-- please capture a stack trace and Xcode debug log. Thanks!
			-- [scouten 2011-07-02]: I think this is fixed now, but leaving this
			-- in place for now.
	
			if not accessToken then
				debug.print( "KSOAuth2SessionImp: missing access token" )
				return 'error', "missing access token"
			end
	
			if not refreshToken then
				debug.print( "KSOAuth2SessionImp: missing refresh token" )
				return 'error', "missing refresh token"
			end
	
			-- If we have to get a new access token ("refresh"), this is where it happens.
		
			local function reauthAndRetryP( authorizeHeader )
			
				-- [scouten 2011-11-21]: Here's the fun and tricky new part: If the
				-- access+tokens promise object that we used when we started this request
				-- is still in place, then it's time to trigger a token refresh and
				-- suspend all future requests until we have the new token.
				
				-- If the promise object has already been changed, then we simply
				-- reissue this request with the new token set once we have it.
	
				if accessAndRefreshTokensP_atStart == accessAndRefreshTokensP then
					accessAndRefreshTokensP = refreshRequestP( refreshToken, accessToken, authorizeHeader )
				end
				
				return getPutImpP( url, args, getPutInnerFunc )
			
			end
			
			-- Configure progress rerouting if needed.
	
			-- FIXME [scouten 2011-11-18]: The wantDownloadProgress feature is currently
			-- broken. We need to fix this post-1.1. (Nobody is using it currently...)
			
			local currentProgressBB
	
	--		if args.wantDownloadProgress then
	--
	--			currentProgressBB = KSReactive.variableB( nil )
	--			local resultProgressB = currentProgressBB:mapSwitchB( function( progressB )
	--
	--				if progressB then
	--					return progressB
	--				else
	--					return KSReactive.variableB( 0, 1 )
	--				end
	--
	--			end )
	--
	--		end
		
			-- In order to implement the 401 trapping, we need to suppress the
			-- conversion of HTTP errors to reactive errors.
	
			args = args or {}
		
			local originalReturnHttpErrorsAsData = args.returnHttpErrorsAsData
			args.returnHttpErrorsAsData = true
			args.retainDataAfterRequest = true
		
			-- Strip any previous Authorization and x-token-ttl headers.
			
			local headers = args.requestHeaders or {}
			args.requestHeaders = headers
			
			for i = #headers, 1, -1 do
			
				if headers[ i ].header == 'Authorization'
					or headers[ i ].header == 'x-token-ttl'
				then
					table.remove( headers, i )
				end
			
			end
	
			-- Create the request based on current access token.
	
			headers[ #headers + 1 ] =
				{ header = 'Authorization', value = 'Bearer ' .. accessToken }
	
--			headers[ #headers + 1 ] =
--				{ header = 'x-token-ttl', value = '30' }
					-- Should normally be commented out.
					
			-- Add an Accept-Language header.
	
			headers[ #headers + 1 ] =
				{ header = 'Accept-Language', value = languageCode }
	
			oAuth2ImpLog:debug( "sending request to getPutInnerFunc with args = %s", args )
	
			local rawResponseP, rawProgressB = getPutInnerFunc( args )
			
			if currentProgressBB then
				currentProgressBB:set( rawProgressB )
			end
	
			return rawResponseP:mapDataP( function( _, data, headers )
				
				-- Close data after processing this request.
				
				if KSData_isData( data ) then
					WFRequestDispatcher.deferredClose( data )
				end
				
				-- Debug output.
				
				if headers and headers.status_code == '401' then
					oAuth2ImpLog:info( "OAuth2 token invalid; refreshing" )
				elseif headers and headers.status_code >= '400' then
					oAuth2ImpLog:error(
							"\error on URL %s\n\ndata = %s\n\nheaders = %s",
							url, data, headers )
				else
					oAuth2ImpLog:debug(
							"\nnormal response on URL %s\n\nheaders = %s",
							url, headers )
				end
				
				-- 401 errors mean we have an invalid access token. This means we should
				-- use the refresh token and try again.
				
				if headers and headers.status_code == '401' then
					
					-- FIXME [scouten 2011-11-28]: Don't like the hard-coded reference to
					-- Oz's Www-Authenticate header, but not worth fixing right now.
	
					oAuth2ImpLog:info( "Detected invalid access token %s", headers )
					local authHeader = headers[ 'Www-Authenticate' ] or headers[ 'WWW-Authenticate' ]
					return reauthAndRetryP( authHeader )
				
				end
				
				-- 403 errors (forbidden) may mean that Oz is forcing a client upgrade
				-- or that the catalog has been unshared from this user.
				
				-- FIXME [scouten 2011-11-23]: Consider generalizing this beyond 403
				-- errors in case some code looks for other error conditions.
	
				if headers and headers.status_code == '403' and args.compositeErrorData then
			
					-- Try parsing the JSON data response.
	
					local result, eData = pcall( KSJsonDecoder.decodeWebSafeJson, data, true )
	
					-- If the error code returned by Oz is 4999 we need to force an update
					-- and need to capture the iTMS link that is returned with it to direct the
					-- user to the download location for the upgrade.
	
					local link = result
								and eData
								and eData.code == 4999
								and eData.links
								and eData.links[ '/rels/client_app' ]
								and eData.links[ '/rels/client_app' ].href
								or ''
	
					if eData and eData.code then
						return 'error', string.format( "%s,%s,%s", headers.status_code, eData.code, link ), headers
					end
					
					-- Otherwise, just let it pass through as normal error handling.
					
				end
	
				-- If client requested that we return HTTP errors as errors, we restore
				-- that behavior here.
		
				if headers and headers.status_code >= '400'
					and not originalReturnHttpErrorsAsData
				then
				
					if args.returnHttpErrorsAsStatusCode then
						return 'error', headers.status_code
					else
						return 'error', data, headers
					end
	
				-- Happy path.
		
				else
	
					return 'data', data, headers
		
				end
			
			end )
			
		end )
	
	end
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'getP', function( url, optArgs, optSequencing )
		
		return getPutImpP( url, optArgs, function( newArgs )
		
			return WFHttpClient.getP( url, newArgs, optSequencing )
		
		end )
	
	end )
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'putP', function( url, data, optArgs )
		
		local resultP = getPutImpP( url, optArgs, function( newArgs )
	
			return WFHttpClient.putP( url, data, newArgs )
	
		end )
		
		if KSData_isData( data ) then
			resultP = resultP:observeP( function() data:close() end )
		end
		
		return resultP
	
	end )
	
	--------------------------------------------------------------------------------
	--[[
		Should only be used in debug builds! Simulate a failed token refresh.
	]]--
	
	WFRequestDispatcher.onMessageNoCancel( 'simulateForcedLogoutP', function()
		
		oAuth2ImpLog:error( "simulate forced logout" )
		hasValidLoginB:set( false )
	
		deleteAuthKeys()
	
	end )
	
	--------------------------------------------------------------------------------
	
	
	local function getImsLogoutUrl( imsBaseUrl, accessToken, clientId, clientSecret )
	
		return string.format( "%s%s?client_id=%s&client_secret=%s&access_token=%s",
					   imsBaseUrl,
					   IMS_LOGOUT_SERVICE,
					   clientId,
					   clientSecret,
					   accessToken )
	
	end
	
	--------------------------------------------------------------------------------
	
	local function logoutP( imsBaseUrl, accessToken )
		--	From https://zerowing.corp.adobe.com/display/renga/Logout :
		--
		--	Request:
		--	  GET /ims/logout/v1?client_id=YOUR_CLIENT_ID&client_secret=YOUR_CLIENT_SECRET&access_token=ACCESS_TOKEN HTTP/1.1
		--	  Host: ims-na1-qa1.adobelogin.com
		--
		--	Response:
		--	  HTTP/1.1 200
		--	  Cache-Control: no-store
	
		local Bablu = require "Bablu"
		
		return Bablu:getImsClientIdAndSecretP():mapDataP( function( _, clientId, clientSecret )
			
			local logoutUrl = getImsLogoutUrl( imsBaseUrl, accessToken, clientId, clientSecret )
	
			-- debug.print( "Requesting: %@", url.absoluteString )
		
			return WFHttpClient.putP( logoutUrl, nil, {
				requestMethod = "GET",
			} ):mapDataP( function( _, data, headers )
				
				local responseStatusCode = headers.status_code
	
				-- SPAM [arahn 2012-05-31]:  Why is the status_code a string type instead of number?
				-- going with the flow here, trying not to rock the boat.
				if responseStatusCode ~= "200" then
					
					local responseString = tostring( data )
					
					local message = ZSTR( "$$$/IdentityManagement/LogoutError=Received unexpected response code: ^1\nResponse:^2",
										  responseStatusCode,
										  responseString )
	
					-- debug.print( message )
					
					return 'error', message
					
				else
				
					-- Done? really?
					return 'data', 'done'
					
				end
					
				
			end )
	
		end )
			
	end
	
	--------------------------------------------------------------------------------
	
	local function getImsExchangeAccessTokenUrl( imsBaseUrl, accessToken, clientId, clientSecret )
	
		return string.format( "%s%s?client_id=%s&client_secret=%s&access_token=%s",
					   imsBaseUrl,
					   IMS_TOKEN_EXCHANGE,
					   clientId,
					   clientSecret,
					   accessToken )
	
	end
	
	--------------------------------------------------------------------------------
	
	local function exchangeAccessTokenP( imsBaseUrl, accessToken )
	
		local Bablu = require "Bablu"
		
		return Bablu:getImsClientIdAndSecretP():mapDataP( function( _, clientId, clientSecret )
			
			local exchangeAccessTokenUrl = getImsExchangeAccessTokenUrl( imsBaseUrl, accessToken, clientId, clientSecret )
	
			-- debug.print( "Requesting: %@", url.absoluteString )
		
			return WFHttpClient.putP( exchangeAccessTokenUrl,
				KSJsonEncoder.encodeWebSafeJson {
					client_id = clientId,
					client_secret = clientSecret,
					access_token = accessToken,
				},
				{
				    requestHeaders = {
				        { header = 'Content-Type', value = 'application/json' },
				        { header = 'Accept', value = 'application/json' },
				    },
					requestMethod = "POST",
				} ):mapDataP( function( _, data, headers )
					
					local responseStatusCode = headers.status_code
		
					if responseStatusCode ~= "200" then
						
						return 'error', data
						
					else
						
						local decodedJson = KSJsonDecoder.decodeWebSafeJson_stripNull( data )
						
						if decodedJson.facebook_token then
							return 'data', 'facebook', decodedJson.facebook_token, decodedJson.expires_at, decodedJson.scope
						elseif decodedJson.google_token then
							return 'data', 'google', decodedJson.google_token, decodedJson.expires_at, decodedJson.scope
						else
							return 'error', decodedJson.error
						end
						
					end
				
				end )
	
		end )
			
	end
	
	--------------------------------------------------------------------------------
	--[[
		Log the user out. Deletes all store tokens and pretend this session never
		existed.
	
		The promise may safely be ignored.
	]]--
	
	WFRequestDispatcher.onMessageNoCancel( 'logoutP', function()
	
		-- Setting hasValidLoginB to true seems odd, but it's an essential part
		-- of getting things spun up correctly when we try to log this user in
		-- again. If we don't, there's a race condition (which usually blows up)
		-- in which the client UI reads hasValidLoginB as false and forces another
		-- logout before we can propagate the news that we have a new valid
		-- login again. [scouten 2011-11-30, PR# 3050626]
	
		hasValidLoginB:set( true )
	
		if not accessAndRefreshTokensP then
			return 'data', true		-- we're already set!
		end
		
		return accessAndRefreshTokensP:mapP( function( status, accessToken )
		
			deleteAuthKeys()
	
			if status == 'data' then
			
				-- FIXME [scouten 2011-11-21]: This assumes we're using IMS. Won't hold
				-- up if we ever need to do the OAuth2 dance with anyone else.
	
				return logoutP( KSPrefs.imsBaseUrl, accessToken )
	
			else
	
				return 'error', accessToken
	
			end
			
		end )
	
	end )
	
	--------------------------------------------------------------------------------
	--[[
		Exchange the IMS token for the original token.
	]]--
	
	WFRequestDispatcher.onMessageNoCancel( 'exchangeAccessTokenP', function()
	
		if not accessAndRefreshTokensP then
			return 'error', "not logged in"
		end
		
		return accessAndRefreshTokensP:mapP( function( status, accessToken )
	
			if status == 'data' then
			
				return exchangeAccessTokenP( KSPrefs.imsBaseUrl, accessToken )
	
			else
	
				return 'error', accessToken
	
			end
			
		end )
	
	end )
	
	--------------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'DEBUG_dumpCoverageP', function( path )
	
		require 'SetupCoverage'.dumpCoverage( path )
	
		return require 'KSPromises'.resolvedPromiseP( 'data', true )
	
	end )

end

--==============================================================================

return {
	initMessageProcessor = initMessageProcessor,
}
