--[[----------------------------------------------------------------------------

KSReactive.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2011-01-11

@reviewed scouten 2011-03-11

------------------------------------------------------------------------------]]

--[[

	KSReactive: Reactive Programming Support
	
	We support two fundamental types: event streams and behaviors. All behaviors
	are also event streams that fire when the behavior changes value. The
	key differences between the two is that behaviors have a current value and
	only fire when that value actually changes whereas event streams have no
	persistent value and can fire the same value repeatedly.
	
	function KSReactive.sourceE() -> source event node
	
	function KSReactive.variableB() -> variable behavior node
	
	function KSReactive.constantB( value... ) -> constant behavior node with the
		given value.
		
	function KSReactive.constantErrorB( message... ) -> constant behavior node with the
		given value as an error.
		
	function KSReactive.anyE( ... ) -> merged event stream
		The vararg portion is a list of event streams. This combines multiple event streams.
		If any of the streams fire in a given cycle, this fires true.
		
	function KSReactive.zeroE()
		Returns an event stream that starts out (and remains) done.
		
	function KSReactive.callB( fn, ... ) -> behavior node for fn( ... )
		It will re-evaluate when the function or parameters change. If the function
		or the arguments are not behaviors, they will be run through constantB
		to get converted to behaviors.
		
	function KSReactive.argTupleB( ... ) -> behavior for ... with error detection
		If none of ... is an error, then the value is the standard argument format extension
		of ... in which the last argument can supply arbitrarily many values. If any of the
		values is an error, then the result is one of those errors.
		
	function KSReactive.dataCallB( fn, ... ) -> behavior for fn( ... )
		It will re-evaluate when the function or parameters change. If the function
		or the arguments are not behaviors, they will be run through constantB
		to get converted to behaviors. The function will not be called for errors and
		instead errors will pass through unchanged.
		
	function KSReactive.liftB( fn ) -> a function that accepts behaviors as inputs
		and returns a behavior as a result. Effectively:
			 function( ... ) return dataCallB( fn, ... ) end
			
	function KSReactive.lift( t ) -> a table in which the entries for any field with
		a name ending in a capital B has been lifted via KSReactive.liftB.
				
	function KSReactive.laterB()
		Returns a behavior that transitions from false to true when the event processor
		gets a chance to process later.
		
	function KSReactive.timerE( delayTime ) -> timer event stream
		Returns an event stream that will fire the value true after the given delay time.
		Delay times are rounded and grouped so that we can avoid creating too many
		separate timers. If one passes 'later' for delayTime, the event stream fires
		essentially at the same time laterB would fire.
		
	function KSReactive.timedSeqB( value0, delay, value1 )
		Returns a behavior that starts at value0 and transitions to value1 after the indicated
		delay.
	
	function KSReactive.falseThenTrueB( delay )
		Returns a behavior that starts at false and transitions to true after the indicated
		delay.
	
	function KSReactive.trueThenFalseB( delay )
		Returns a behavior that starts at true and transitions to false after the indicated
		delay.
	
	function KSReactive.isEventNode( v ) -> v_is_event_node, v_is_behavior
		Test whether v is a reactive event stream node and furthermore whether
		it is a behavior.
	
	function KSReactive.isBehavior( v ) -> v_is_behavior
		Test whether v is a behavior. (Short form for the second parameter from
		isEventNode.)
	
	-- FIXME [mhamburg 2011-03-14] : [scouten 2011-03-11]: Looks like this is out of date with the
	-- current set of reactive nodes.
	
	function KSReactive.makeWeakEventTable( opt_initial_value_for_behaviors )
		A class for constructing a table to manage a weak mapping to event
		nodes -- either events or behaviors. These can be managed as follows:
		
			et =  KSReactive.makeWeakEventTable( < initial value for behaviors > )
			
			local E = et:eventE( 'foo' )
			local B = et:behaviorB( 'baz' )
			
			-- All functions do nothing if the event doesn't currently exist. These
			-- return true if the event exists and false if it doesn't.
			
			et:postToEvent( 'foo', 3 )
			et:postErrorToEvent( 'foo', "Disk full" )
			et:closeEvent( 'baz' )
			
			-- But we might also want to test for event existence
			
			et:eventExists( 'bar' )
			
			-- We can get a list of the events currently in existence.
			
			et:eventKeys()
			
	Event stream modifiers:
	
		e:holdB( value... )
			Returns a behavior that takes value... as its initial value and
			updates whenever e fires.
			
		e:closingHoldB( value... )
			Returns a behavior that takes value... as its initial value and
			updates whenever e fires. Takes close responsibility for e.
			
		e:relayE( autoClose )
			Returns an event stream that supplies the exact same values as
			e. If autoClose is true and this node is closed or simply GC'd,
			we will check to see whether there are any other active relays
			for e and if not we will close e. Relays can be used to protect
			underlying event streams from external modification, closure, etc.
			
		b:relayB( autoClose )
			Relay logic for behaviors. See relayE.
	
		e:mapE( fn )
			Takes an event stream and runs each of the events through fn and
			returns the mapped value. Errors are passed through unchanged.
			
		e:closingMapE( fn )
			Takes an event stream and runs each of the events through fn and
			returns the mapped value. Errors are passed through unchanged.
			Sets the source as a close target.
			
		b:mapB( fn )
			Takes a behavior and runs its value through fn to generate a new
			mapped value. Errors are passed through unchanged.
			
		b:closingMapB( fn )
			Does a mapB on the behavior and sets the behavior as a close target
			of the result.
			
		e:rawMapE( fn )
			fn takes values in the form 'data', < data > and 'error', < error  message >
			and gets to map them to similar format values.
			
		b:rawMapB( fn )
			fn takes values in the form 'data', < data > and 'error', < error  message >
			and gets to map them to similar format values.
			
		e:filterE( fn )
			Takes an event stream and tests each value against fn and passes
			through only those values for which fn returns a true value.
			Error values pass through unchanged. (FIXME: If fn generates an
			error, it isn't clear what should happen.)
			
		e:constantE( value... )
			Turns every value from e into value... Passes errors through
			unchanged.
			
		-- OBSOLETE: e:switchE()
			e should be an event stream valued event stream. Each time it fires,
			we change which event stream we are listening to to produce the
			output. This passes errors through unchanged. It is done when both
			e is done and the last event stream if any supplied by e is done.
			
		e:mapSwitchE( fn )
			fn should be a function that transforms the values provided by e into
			event streams. We then switch over these streams. As such, it is
			equivalent to e:mapE( fn ):switchE(). The distinction between this
			version built out of the primitive modifiers and mapSwitchE is
			that mapSwitchE takes ownership of the mapped values and will
			close them when switching.
			
		e:closingMapSwitchE( fn )
			Runs e through fn as a mapSwitchE and sets e as a close target for
			the result.
			
		b:mapSwitchB( fn )
			fn should map the values from b to behaviors. As with mapSwitchE,
			this is essentially equivalent to b:mapB( fn ):switchB() with
			the added property that it takes ownership of the values produced
			by the map and closes them when switching. (BUG: Documentation in
			terms of the now non-existent switchB is a bit of a problem.)
			
		b:closingMapSwitchB( fn )
			Does a mapSwitchB on the behavior and sets the behavior as a close
			target for the result.
			
		b:zipMapSwitchB( fn )
			Like mapSwitchB but prepends the value of b (only one position if
			b is a multiple value) to the values from the behavior produced
			by fn. This is useful if you need to keep track of where a value
			came from.
			
		[ NOT USED: DISABLED ] e:skipFirstE()
			Returns an event stream that drops the first event to fire from
			e (either data or error). If e is already firing this cycle, it
			will drop that value.
			
		e:skipErrorsE()
			Returns an event stream that drops any error values from e and
			passes everything else through unchanged.
			
		e:endAfterErrorE()
			Returns an event stream that sets itself to done after the first
			error value.
			
		e:takeE( n )
			Take n events and then terminate. If the source delivers fewer
			than n events, terminate early. n should be a number greater than
			zero. (We effectively take the ceiling when counting down.)
			
		e:accumulateE( fn, accum )
			For each event, we compute fn( accum, ... ). The first result is
			taken as the new accumulator value and all of the results are
			returned. Errors are passed through without affecting the accumulator.
			If you want errors to be sticky, follow or precede this with
			endAfterErrorsE.
			
		e:accumulateB( fn, accum, ... )
			A behavior that starts out at accum, ... and then builds up values
			in the same manner as accumulateE.
			
		[ NOT USED: DISABLED ] e:takeWhileE( predicate )
			As long as the values on e pass predicate, this passes them through.
			At the first value that fails or at the first error, this terminates.
			
		[ NOT USED: DISABLED ] e:takeUntilE( predicate )
			Forward values from e until either e terminates or we receive a value
			that meets the predictate.
			
		[ NOT USED: DISABLED ] e:takeThroughE( predicate )
			Forward values from e until either e terminates or we receive a value
			that meets the predictate. If we do encounter a value that meets the
			predicate, we pass it on and then terminate the event stream.
			
		[ NOT USED: DISABLED ] e:whenDoneE()
			Fires true when the source reports that it is done (and then terminates
			iteself).
			
		[ NOT USED: DISABLED ] e:delayE( when )
			Delays propagation of the values from e. There are three options
			for when: 'next', 'stable', and 'idle'. 'next' will fire in the
			next event processing cycle which will run automatically after
			this one. 'stable' will fire when we cease to have input values
			to process or next delays to deliver. 'idle' will run when the
			event processor gets told to run idle events. (Right now that
			simply corresponds to a trip back to the main event loop. It
			should evolve to really mean idle.)
			
		e:mergeE( otherStreamE )
			Merge the event stream e with the event stream otherStreamE. If
			both streams fire on a given cycle, then only the receiver of this
			method's value is propagated.
			
		e:calmE( interval )
			We are either active or calm. When active, an event flows through
			and sets the state to calm for the time period specified by interval.
			While calm, events just accumulate. When we exit the calm state, we
			will transmit the last value to come through -- if any -- and if there
			is a value re-enter the calm state. Errors propagate immediately and
			also take the node out of the calm state.
			
		b:calmB( interval )
			Like calmE but working with behaviors. Changes to the behavior get
			calmed.
			
		b:mcallB( methodName, optDefault )
			For data values (as opposed to errors) call a method on the values
			to get a new value. Basically a shorthand for mapB. The additional
			arguments will be passed as arguments to the method. (FUTURE: Allow
			the arguments to themselves be behaviors.) If the current value is
			nil or false, we will use the default if supplied or the current
			value if not supplied.
			
		b:mcallSwitchB( methodName, optDefault )
			Like mcallB but this works with methods that return behaviors by
			then doing a switch to drop back down to values.
			
		b:propertyB( propertyName, optDefault )
			Extracts the field [ propertyName ] from the incoming value.
			Uses optDefault if incoming value is nil or false. If optDefault is not
			supplied, then it returns the incoming value in these cases.
			
		b:toBooleanB()
			= b:mapB( function( v ) return not not v end )
			
		b:filterB( fn, ... )
			Filters the values of b so that the value stream only contains values
			meeting the test of the predicate fn. If the initial values don't meet
			the predicate, then we use ... as a starting point. Errors pass through
			unchanged.
			
		b:skipErrorsB( ... )
			Remove errors from the stream of values for the behavior. If the initial
			value is an error, replace it with ....
			
		b:valueNow()
			The current value of the behavior. Use with caution since the behavior
			may not have updated if called during the current event cycle. Also
			beware that behaviors also often represent values to be delivered
			eventually and as such may not have the expected value at this particular
			moment. Errors come through as packed values (KSRPackedValues). You
			probably want to use the next routine to avoid needing to deal with
			KSRPackedValues.
			
		b:isErrorNow()
			Returns true/false and the error message if any if true. This reports
			whether the behavior value is currently an error.
			
	The escape hatches. These are used to connect the network up to the outside
	world. As such they stay alive as long as their sources chains are alive even
	if they aren't explicitly referenced anywhere else. Hence, if you don't want
	this to happen, you need to be good about explicitly closing them.
	
		e:onDataE( fn )
			Runs fn when e fires with normal data passing it the data.
			
		e:onErrorE( fn )
			Runs fn when e fires an error.
			
		e:onDoneE( fn )
			Runs fn when e announces that it is done.
			
		b:onDataB( fn )
			Runs fn on the current value of b (as of this cycle) and whenever
			b changes provided that the value of b is not an error value.
			This is also a behavior returning the same values as b (including
			errors).
			
	We can also cause a node to live until it is done.
		
		e:runUntilDone()
			
	We can monitor a node to see what values it sends:
	
		n:setDebugHook( [ optHookFn = debug.print ], [ optHookName ] )
			
	Close v Done. Closing a source node starts propagating a done signal through
	the network. This generally results in nodes closing as they lose all of their
	sources. Behaviors, however, need not close when done (though they will notify
	their successors that they done). We can adjust this for behaviors by
	calling:
	
		b:setCloseWhenDone( state )
	
]]

--============================================================================--

--[[

Requests:

	b:callWhenE( value, fn, ... )
		When the b:valueNow() == value, call fn( ... ) and return the result
		(or the error) via an event stream constructed by this call.
		See undoRedoImp in OzDocStore.lua.

]]

--============================================================================--

-- local assertf = require 'KSLuaUtils'.assertf
-- local string_format = string.format
-- local pairs = pairs
local assert = assert
local type = type
local select = select
local rawset = rawset
local unpack = unpack
local pcall = pcall
local tostring = tostring
local pairs = pairs
local setmetatable = setmetatable

local WFArguments = require 'WFArguments'
local WFArguments_check = assert( WFArguments.check )

local string_format = assert( string.format )

local math_max = math.max

local KSLeaks = require 'KSLeaks'

local KSLeaks_registerObject = assert( KSLeaks.registerObject )

local WFCloseSupport = require 'WFCloseSupport'

local WFCloseSupport_hasBeenOrWillBeClosed = assert( WFCloseSupport.hasBeenOrWillBeClosed )

local WFQueueing = require 'WFQueueing'

local WFQueueing_makeQueue = assert( WFQueueing.makeQueue )

local WFVarArgs = require 'WFVarArgs'
local WFVarArgs_map = assert( WFVarArgs.map )
local WFVarArgs_findFirst = assert( WFVarArgs.findFirst )

local KSRPackedValues = require 'KSRPackedValues'

local KSRPackedValues_pack = KSRPackedValues.pack
local KSRPackedValues_packOne = KSRPackedValues.packOne
local KSRPackedValues_unpack = KSRPackedValues.unpack
local KSRPackedValues_error = KSRPackedValues.error
local KSRPackedValues_isError = KSRPackedValues.isError
local KSRPackedValues_first = KSRPackedValues.first
local KSRPackedValues_testEquals = KSRPackedValues.testEquals
local KSRPackedValues_packPCall = KSRPackedValues.packPCall

local KSRPackedValues_sameValue = KSRPackedValues.same()
local KSRPackedValues_noneValue = { }

local KSPrivateKeys = require 'KSPrivateKeys'
local KSPrivateKeys_makePrivateKey = KSPrivateKeys.makePrivateKey
local WFDebugTraceback_traceback = require 'WFDebugTraceback'.traceback

local traceback = function()
	return WFDebugTraceback_traceback { "KSReactive.lua" }
end

-- Done is an internal signal value.

local kDoneValue = { }

-- Schedule request is an internal signal value

local kScheduleRequestValue = { }

-- Should we close the node when done?

local keyCloseWhenDone = KSPrivateKeys_makePrivateKey( 'KSReactive.closeWhenDone' )

-- The debug hook for event nodes

local keyDebugHook = KSPrivateKeys_makePrivateKey( 'KSReactive.debugHook' )

local DEBUG_NODE_BURSTS = false  -- should normally be false

local ALLOW_EMPTY_TUPLES_FROM_MAP = false
	-- Can mapE and mapB return empty tuples and have those propagate properly?
	-- We ought to support this but historically we have used that as an indicator
	-- to just leave the value unchanged. Try flipping the flag and see what
	-- hapepens...
	
local AUTO_CLOSE = false
	-- When true we will roll up the behavior chain closing items when we close.
	-- We will not do so from non-auto-close relay nodes.

--============================================================================--

-- FIXME: Move this elsewhere or make it a standard utility

local function lazyFunction( moduleName, functionName )

	local fn = nil

	return function( ... )
	
		if fn == nil then
			fn = require( moduleName )[ functionName ]
		end
		
		return fn( ... )
	
	end

end


--============================================================================--

-- Bypass grabbing all of KSLuaUtils for non-debug builds.

local KSLuaUtils, dpcall, assertft

if _G.DEBUG then

	KSLuaUtils = require 'KSLuaUtils'
	dpcall = assert( KSLuaUtils.dpcall )
	assertft = assert( KSLuaUtils.assertft )

else

	-- Minimally compatible version of above functions with no debug output.

	dpcall = function( ident, ... )
		return pcall( ... )
	end
	
	assertft = function( value )
		return assert( value )
	end

end

--------------------------------------------------------------------------------

-- Bypass grabbing all of KSTableUtils just to be able to make fully weak tables.

local fullyWeakMT = { __mode = 'kv' }

local function makeFullyWeakTable()
	return setmetatable( {}, fullyWeakMT )
end


--============================================================================--

--[[

Implementation Notes:

Posting to an event stream or setting the value on a reactive variable triggers
a reactive cycle. At the base level, this is handled via two mechanisms: the
delivery "queue" and the schedule queue.

THE DELIVERY QUEUE
------------------
The delivery queue is a sequence of node/value pairs for which we want to deliver
the given value to the given node. Delivery will invoke the appropriate method on the
node based on the value type:

	_receiveData( packedValue ): Receives a normal packed value as data
	_receiveError( errorValue ): Receives an error value
	_receiveDone( allSourcesGone ): This ties to the magic done value,
		but instead of receiving that value, it receives a flag indicating
		whether all of the sources for the node are done.
	
Each can return a value to be handed on to the node's successors or can
return kScheduleRequestValue which will queue the node to run at the appropriate
time in the topological sort at which point its _run method will be invoked
and this can in turn return a packed value to pass on up the chain.

node_deliverPackedValue( node, packedValue ) is the dispatch function that
runs this logic by deciding which method to invoke and how to handle the result.

node_deliverPackedValue is fed from the delivery queue which contains pairs of
nodes and packed values to deliver.

node_sendPackedValue is the routine called with a new value to be fed into
the system. Generally, it enqueues the value onto the delivery queue for all
of the successors for the node. If the value is the done value, it will also
initiate a tear down of the node including running any close targets and other
pieces of the close notification system. This will occur before the next node
gets a chance to run.

If node_deliverPackedValue fails, we will invoke node_deliveryFailed which will
______. FIXME

The optional table at valuesSentThisCycle records the values delivered by
node_sendPackedValue. Any given node should only have node_sendPackedValue
called once per cycle.

THE SCHEDULE QUEUE
------------------
The schedule queue implements topological sorting on the reactive graph. If
while processing a delivered message the node returns kScheduleRequestValue,
then the node will be placed on the schedule queue. The schedule queue is
a priority queue sorted by node depth at the time of insertion. When we remove
nodes from the queue, we double-check their depth since it is possible that the
depth will have increased while they were queued. (Depth never decreases.) If
it has increased, we requeue and try again. If the depth matches, then we invoke
the _run method on the node which like the various delivery methods above can
return values to pass to the node's successors via node_sendPackedValue which
in turn puts values onto the delivery queue.

THE CORE REACTIVE CYCLE
-----------------------
We pull from the delivery queue until it is empty and then pull from and process
nodes on the schedule queue until the delivery queue is again non-empty. The
core reactive cycle is done when both of these queues are empty. When both queues
are empty, we discard our knowledge of values sent during this reactive cycle
(stored in valuesSentThisCycle). This logic is encoded in runScheduledQueue.

OUTER CYCLES
------------
We have a series of levels of outer cycles where we will accumulate node/value
pairs for later delivery. These will only accumulate one (packed) value per
node so multiple accumulations for the same node will only keep the last value.
Nodes using these levels are known as "delayed" nodes and are implemented through
the delayE constructor with appropriate parameter values to select the delay
level. These nodes break cycles in the topological sorting by always having a depth
of zero since they don't fire until at least one cycle after they receive a value.

These levels are:

next tick: Used by nodes constructed by delayE. These values will be sent
immediately after we complete a core reactive cycle and will trigger the next
reactive cycle.

stable: These values will be sent when the input queue (see below) is empty.

idle: These values will be sent when KSReactive.runIdle is invoked by outer
code detecting that we're idle. [ NOT USED: DISABLED ]

THE INPUT QUEUE
---------------
Sitting between the next tick level and the stable level is the input queue.
This is a sequence of node/value pairs (possibly with multiple entries per
node). This is where outside events feed in. As we drain the input queue
(runInputQueue), we run core reactive cycle and next tick cycles as appropriate.

]]

--============================================================================--

--[[

KSRDirectedAcyclicGraph:

	This module provides a routine for creating a table full of functions for
	manipulating a DAG built out of tables with private keys. The DAG recognizes
	both strong and weak nodes where weak nodes are not kept alive by their
	predecessors but nodes always keep their predecessors alive.
	
	When we create the table of functions, we supply it with a table of
	customization options including:
	
		nodeName: A function to take a node table and return a string for debugging
			-- GRUNGE [mhamburg 2011-01-12]: This could just use __tostring.
			
		gainedFirstSuccessor: A function taking a node. This will be called
			when a node gains its first successor.
			
		lostSuccessor: A function taking a node to be called if the node loses
			a successor.
			
	The functions created in the table include:
	
		function allocNode( optStrongSuccessor )
			This creates a table with various fields filled in. The strong flag
			indicates whether this will be a strong node. The node is not
			actually ready for use until you also call:
			
		function initNode( node, nodeKind, isBase, ...predecessors... )
			This finishes the node initialization and wires it up appropriately.
			The varargs section lists its predecessors. This returns the node.
			
			So the allocation pattern is:
			
				node = allocNode( optStrongSuccessor )
				... set values before initialization completes ...
				initNode( node, nodeKind, isBase, ... predecessors... )
				
			A short version with no extra initialization is:
			
				return initNode( allocNode( strong ), nodeKind, isBase, ...predecessors... )
			
		function addLink( predecessor, successor )
			Add a link from predecessor to successor.
			
		function removeLink( predecessor, successor )
			Remove the link if any from predecessor to successor.
			
		function iterSuccessors( node )
			Returns an iterator for the successors of a node. It is an error
			(undetected at present) to add successors while iterating successors.
			
		function getNodeDepth( node )
			Get the current depth for node.
			
		function isNode( v )
			Returns true iff v is a node. (Note done nodes are still nodes.)
			
	It also includes the following debugging functions. This is where the node
	name actually matters.
	
		function assertNode( node [ , optExtraMessage ] )
			Check that node is a valid node for this DAG. Throw if it is not.
			
		function checkDepths( rootNode )
			Makes sure that the depth values are all properly increasing as
			we move from predecessors to successors.
			
		function checkSuccessors( node, ...successors... )
			Check that the successors to node are exactly the listed nodes.

]]

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-01-12]: We could optimize the predecessors case for
-- single predecessors I haven't done that yet, because it isn't critical
-- in order to get this working.

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-01-12]: Move to KSRIterators

local function iterNone()
	return nil
end

--------------------------------------------------------------------------------

-- Add element to a set at t[ k ] creating the set if need be. Returns true
-- if the set didn't exist before (which is akin to it being empty though
-- not quite the same).

local function addSetFieldElement( t, k, element )
	local s = t[ k ]
	local result = false
	if not s then
		s = { }
		t[ k ] = s
		result = true
	end
	s[ element ] = element
	return result
end

--------------------------------------------------------------------------------

-- Remove an element from an optional set at t[ k ]. Removes the set if it is
-- now empty and returns true.

local function removeSetFieldElement( t, k, element )
	local s = t[ k ]
	if s and s[ element ] then
		s[ element ] = nil
		if next( s ) == nil then
			t[ k ] = false
			return true
		end
	end
	return false
end

--------------------------------------------------------------------------------

-- Iterate a set field if not empty.

local function iterSetField( t, k )
	local s = t[ k ]
	if not s then
		return iterNone
	else
		return pairs( s )
	end
end

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-03-14]
-- This code makes a private namespace of functions so that it could be
-- used by something other than the reactive code. This is design overkill
-- and should probably go away.

local keyDepth = KSPrivateKeys_makePrivateKey( 'KSRDirectedAcyclicGraph.depth' )
local keyPredecessors = KSPrivateKeys_makePrivateKey( 'KSRDirectedAcyclicGraph.predecessors' )
local keyAllSuccessors = KSPrivateKeys_makePrivateKey( 'KSRDirectedAcyclicGraph.allSuccessors' )
local keyStrongSuccessors = KSPrivateKeys_makePrivateKey( 'KSRDirectedAcyclicGraph.strongSuccessors' )
local keyNodeIsStrongSuccessor = KSPrivateKeys_makePrivateKey( 'KSRDirectedAcyclicGraph.nodeIsStrongSuccessor' )

--------------------------------------------------------------------------------

-- We need to track whether or not something is a node for debugging purposes.
-- We store the node kind as the value.

local allNodes = makeFullyWeakTable()
local previousActiveNodeCount = 0
local nodeEpoch = 0
local WFCurrentThread_DEBUG_ONLY

if DEBUG_NODE_BURSTS then

	local WFTimer = require 'WFTimer'
	WFCurrentThread_DEBUG_ONLY = require 'WFCurrentThread'
	
	local nodeEpochDuration = 5 -- seconds
	
	local function kickTimer()
	
		WFTimer.afterDelayDo( nodeEpochDuration, function()

			local activeNodeCount = 0
			local tracebacksForThisEpoch = {}

			for node in pairs( allNodes ) do
				
				activeNodeCount = activeNodeCount + 1
				
				if node.epoch == nodeEpoch then
				
					local traceback = node.allocTraceback
					tracebacksForThisEpoch[ traceback ] = ( tracebacksForThisEpoch[ traceback ] or 0 ) + 1
				
				end
			
			end

			local activeNodeDelta = activeNodeCount - previousActiveNodeCount
			
			if activeNodeDelta >= 100 and nodeEpoch > 20 then

				local sortedTracebacks = {}
				local stbCount = 0

				for tb, count in pairs( tracebacksForThisEpoch ) do
					stbCount = stbCount + 1
					sortedTracebacks[ stbCount ] = { tb = tb, count = count }
				end
				
				table.sort( sortedTracebacks, function( a, b )
					if a.count ~= b.count then
						return a.count > b.count
					else
						return a.tb < b.tb
					end
				end )
				
				local sortedTbStrings = {}
				for i = 1, #sortedTracebacks do
					local tbItem = sortedTracebacks[ i ]
					sortedTbStrings[ i ] = string_format( "%d new reactive nodes :: %s", tbItem.count, tbItem.tb )
				end

				debug.print( string_format( "%s: Created net %d new reactive nodes in past %d seconds\n\n\n%s\n\n----------------------------------\n",
											WFCurrentThread_DEBUG_ONLY.DEBUG_getThreadName(),
											activeNodeDelta,
											nodeEpochDuration,
											table.concat( sortedTbStrings, '\n\n' ) ) )
				
			end

			previousActiveNodeCount = activeNodeCount
			nodeEpoch = nodeEpoch + 1

			kickTimer()
		
		end )
	
	end
	
	kickTimer()

end

--------------------------------------------------------------------------------

local function isNode( node )
	return not not allNodes[ node ]
end

--------------------------------------------------------------------------------

local function assertNode( node, msg )
	if isNode( node ) then
		return node
	end
	local tback = traceback()
	local errMsg
	if msg then
		errMsg = string_format( "Expected to be passed a node (%s): %s", tostring( msg ), tback )
	else
		errMsg = string_format( "Expected to be passed a node: %s", tback )
	end
	error( errMsg, 0 )
end

--------------------------------------------------------------------------------

-- We iterate successors using the optional set field logic. Note, however,
-- that because we apply different rules to successors, we do not use those
-- routines for adding and removing successors.

-- FIXME [mhamburg 2011-03-14] : [scouten 2011-03-11]: Scratching my head over this comment. Perhaps
-- you could explain what is different about successors and how that impacts
-- the add/remove logic.

--------------------------------------------------------------------------------

local function node_iterSuccessors( node )
	return iterSetField( node, keyAllSuccessors )
end

--------------------------------------------------------------------------------

local function node_iterPredecessors( node )
	return iterSetField( node, keyPredecessors )
end

--------------------------------------------------------------------------------

-- Notification and update for a node achieving strong status by gaining its first
-- strong successor.
	
local function node_becameStrong( node )
	for pred in node_iterPredecessors( node ) do
		if addSetFieldElement( pred, keyStrongSuccessors, node ) then
			if not pred[ keyNodeIsStrongSuccessor ] then
				node_becameStrong( pred )
			end
		end
	end
end

--------------------------------------------------------------------------------

-- Notification and update for a node losing strong status by losing all of its
-- strong successors.
	
local function node_becameWeak( node )
	for pred in node_iterPredecessors( node ) do
		if removeSetFieldElement( pred, keyStrongSuccessors, node ) then
			if not pred[ keyNodeIsStrongSuccessor ] then
				node_becameWeak( pred )
			end
		end
	end
end

--------------------------------------------------------------------------------

-- Does the node have a predecessor? This relies on the clearing logic dropping
-- us down to false instead of a table.

local function node_hasPredecessor( node )
	return not not node[ keyPredecessors ]
end

--------------------------------------------------------------------------------

-- Does the node have a successor?

local function node_hasSuccessor( node )

	local allSuccessors = node[ keyAllSuccessors ]
	
	if not allSuccessors then
		return false
	end
	
	if next( allSuccessors ) ~= nil then
		return true
	end

	node[ keyAllSuccessors ] = false	
	
	return false

end

--------------------------------------------------------------------------------

-- We use the depth field to mark nodes as done by setting it to -1. Since we won't
-- allow done nodes in links, this shouldn't matter.

local function node_isDone( node )
	return node[ keyDepth ] == -1
end

--------------------------------------------------------------------------------

-- We can force a node to be a base node. When this is the case, it has depth zero.

-- FIXME [mhamburg 2011-03-14] : [scouten 2011-03-11]: What does it mean to be a base node? Is this an init-time
-- operation, or can you mark a node as base at any time? If init-time, perhaps this
-- condition should be tested/verified?

local function node_markAsBase( node )
	assert( not node_isDone( node ), "Cannot change base status for a done node." )
	node[ keyDepth ] = 0
end

--------------------------------------------------------------------------------

local function node_isBase( node )
	return node[ keyDepth ] == 0
end

--------------------------------------------------------------------------------

-- Recursively walk the successor links from a node making sure that the depth
-- is at least the supplied depth and increases by at least one each time we
-- follow a link. This will run forever if the graph has cycles, but we use it
-- in addLink to prevent that.

local function node_raiseDepth( node, newDepth )

	if node_isBase( node ) then
		return
	end

	if newDepth <= node[ keyDepth ] then
		return
	end
	
	node[ keyDepth ] = newDepth
	
	local nextDepth = newDepth + 1
	
	for successor in node_iterSuccessors( node ) do
		node_raiseDepth( successor, nextDepth )
	end
	
end

--------------------------------------------------------------------------------

-- Add a link from pred to succ.

local function nodes_addLink( pred, succ )

	assert( not WFCloseSupport_hasBeenOrWillBeClosed( pred ),
			"addLink: predecessor should not have been closed" )

	assert( not WFCloseSupport_hasBeenOrWillBeClosed( succ ),
			"addLink: successor should not have been closed" )

	-- The successor should not be done.
	
	assert( not node_isDone( succ ), "addLink: successor should not be done" )
	
	-- If the predecessor is done, we simply fail to construct the link.
	
	if node_isDone( pred ) then
		return
	end
	
	-- Make sure the depth is right (this also detects cycles though in doing so it leaves the depths incremented)

	local predDepth = pred[ keyDepth ]
	
	node_raiseDepth( succ, predDepth + 1 )

	assert( pred[ keyDepth ] == predDepth, "Cannot add cyclic link" )
	
	-- Check the strong status for each node. We may be changing the status for the
	-- predecessor node, so we only check whether the local flag would block
	-- propagation.

	local succIsStrong = succ[ keyNodeIsStrongSuccessor ] or succ[ keyStrongSuccessors ]
	local predNodeIsStrong = pred[ keyNodeIsStrongSuccessor ]

	local allSuccessors = pred[ keyAllSuccessors ]
	
	if not allSuccessors then
		allSuccessors = makeFullyWeakTable()
		pred[ keyAllSuccessors ] = allSuccessors
		-- FIXME? gainedFirstSuccessor( pred )
	end
	
	allSuccessors[ succ ] = succ
	
	addSetFieldElement( succ, keyPredecessors, pred )

	-- If strong, then we need to add it to the set of strong successors and
	-- possibly run up the tree adding strong successors.
	
	if succIsStrong then
		if addSetFieldElement( pred, keyStrongSuccessors, succ ) then
			if not predNodeIsStrong then
				node_becameStrong( pred )
			end
		end
	end
	
end

--------------------------------------------------------------------------------

-- Remove the link if any from pred to succ.

local function nodes_removeLink( pred, succ )

	-- We exit early if there isn't a link from pred to succ

	local allSuccessors = pred[ keyAllSuccessors ]
	
	if not allSuccessors then
		return
	end
	
	if not allSuccessors[ succ ] then
		return
	end
	
	-- Deal with strong status first.
	
	if removeSetFieldElement( pred, keyStrongSuccessors, succ ) then
		if not pred[ keyNodeIsStrongSuccessor ] then
			node_becameWeak( pred )
		end
	end
	
	-- Now break the weak link.
	
	allSuccessors[ succ ] = nil
	
	-- Remove the predecessor link.
	
	removeSetFieldElement( succ, keyPredecessors, pred )
	
	-- And do any necessary notification.

	-- FIXME?
	-- if lostSuccessor then
	--		lostSuccessor( succ )
	-- end

end

--------------------------------------------------------------------------------

-- Get the depth for a node.

local function getNodeDepth( node )
	assertNode( node, "getNodeDepth" )
	return node[ keyDepth ]
end

--------------------------------------------------------------------------------

-- Get the kind for a node

local function getNodeKind( node )
	assertNode( node, "getNodeKind" )
	return allNodes[ node ]
end

--------------------------------------------------------------------------------

--[[

--------------------------------------------------------------------------------

-- Set up a default node name function for debugging.

local nodeName = function() return "< >" end
	-- FIXME [mhamburg 2011-03-14]: [scouten 2011-03-11]: Use the constant function?

--------------------------------------------------------------------------------

-- Check that a link exists both ways between pred and succ.

local function checkLink( pred, succ )

	local successors = pred[ keyAllSuccessors ]
	assertf( successors and successors[ succ ], "Successor not linked (%s -> %s)", nodeName( pred ), nodeName( succ ) )

	local predecessors = succ[ keyPredecessors ]
	assert( predecessors and predecessors[ pred ], "Predecessor not linked (%s <- %s)", nodeName( pred ), nodeName( succ ) )

end

--------------------------------------------------------------------------------

-- Check that the depth values are properly increasing as we follow successor chains.
-- helper which handles recursion and API routine that hides extra parameters for recursion.

local function checkDepths_helper( node, minDepth )
	local nodeDepth = node[ keyDepth ]
	assert( node_isBase( node ) or ( minDepth <= nodeDepth ), "node depth is too small" )
	local nextDepth = nodeDepth + 1
	for successor in node_iterSuccessors( node ) do
		checkDepths_helper( successor, nextDepth )
	end
end

--------------------------------------------------------------------------------

-- Check that the depth values are properly increasing as we follow successor chains.
-- Exported function.

local function checkDepths( node )

	assertNode( node, "checkDepths" )

	checkDepths_helper( node, 0 )

end

--------------------------------------------------------------------------------

-- Check that node has exactly the expected successors.

local function checkSuccessors( node, ... )

	assertNode( node, "checkSuccessors: node" )

	local successors = { }

	for i = 1, select( '#', ... ) do
		local succ = select( i, ... )
		assertNode( succ, "checkSuccessors: successor" )
		successors[ succ ] = false
		checkLink( node, succ )
	end
	
	for succ in node_iterSuccessors( node ) do
		local status = successors[ succ ]
		assert( status ~= nil, "Only expected successors in list" )
		assert( status == false, "Only expected to visit successors once" )
		successors[ succ ] = true
	end
	
	for succ, value in pairs( successors ) do
		assertf( value, "Failed to find successor for %s", nodeName( succ ) )
	end

end

--------------------------------------------------------------------------------

]]

--------------------------------------------------------------------------------

-- A utility function for making a pre-initialized node table. This needs to be
-- handed to initNode to finish the initialization.

local function nodeDAG_allocNode( optStrongSuccessor, optWeakPredecessor )

	return {
		[ keyDepth ] = 1,
		[ keyPredecessors ] = false,
		[ keyAllSuccessors ] = false,
		[ keyStrongSuccessors ] = false,
		[ keyNodeIsStrongSuccessor ] = not not optStrongSuccessor,
	}

end

--------------------------------------------------------------------------------

-- Finish node construction potentially wiring the node into the graph based on the
-- list of predecessors.

local function nodeDAG_initNode( node, nodeKind, isBase, ... )

	assert( node[ keyDepth ], "Table does not appear to have come from allocNode." )
	assert( not allNodes[ node ], "Node table appears to have already been used." )

	-- The varargs section provides us with a list of predecessors. We use those
	-- to determine the initial depth and after construction we wire them up as
	-- predecessors.
	
	local predDepth = 0
	
	local predCount = select( '#', ... )
	
	for i = 1, predCount do
		local pred = select( i, ... )
		if pred then
			assertNode( pred, "initNode: predecessor" )
			if not node_isDone( pred ) then
				predDepth = math_max( predDepth, pred[ keyDepth ] )
			end
		end
	end
	
	-- We don't actually raise the depth for base nodes. We leave in the computation
	-- because it also does some validation.
	
	if isBase then
		node_markAsBase( node )
	else
		node[ keyDepth ] = predDepth + 1
	end
	
	-- Mark it as a node

	allNodes[ node ] = nodeKind

	-- Where did this come from?
	
	if DEBUG_NODE_BURSTS then

		node.epoch = nodeEpoch
		node.allocTraceback = traceback( WFCurrentThread_DEBUG_ONLY.DEBUG_getThreadName(), 4 )
	
	end
	
	-- Handle counting
	
	KSLeaks_registerObject( nodeKind, node )
	
	-- Wire up the predecessors
	
	for i = 1, predCount do
		local pred = select( i, ... )
		if pred and not node_isDone( pred ) then
			-- assertNode( pred, "nodeDAG_initNode:addLink: predecessor" )
			-- assertNode( node, "nodeDAG_initNode:addLink: successor" )
			nodes_addLink( pred, node )
		end
	end
	
	-- Return it for chaining
	
	return node

end

--------------------------------------------------------------------------------

local keyBlockPredecessorAutoClose =
	KSPrivateKeys_makePrivateKey( 'KSReactive.keyBlockPredecessorAutoClose' )

local keyBlockSelfAutoClose =
	KSPrivateKeys_makePrivateKey( 'KSReactive.keyBlockSelfAutoClose' )

--------------------------------------------------------------------------------

local possibleAutoCloseNodes = makeFullyWeakTable()

local checkingAutoCloseNodes = false

local haveAutoCloseNodes = false

--------------------------------------------------------------------------------

local function checkAutoCloseNodes_imp()

	-- debug.print( "Checking auto close nodes" )

	repeat
	
		local foundNode = false
		
		-- We need to explicitly manage the iteration because we are adding nodes
		-- even as we are removing them. This messes with the way next ordinarily
		-- works.
		
		local node = next( possibleAutoCloseNodes )
		
		while node ~= nil do
		
			local nextNode = next( possibleAutoCloseNodes, node )
				-- Advance before we remove it from the table. While removal
				-- is safe, removal followed by addition makes things unhappy.
		
			possibleAutoCloseNodes[ node ] = nil
			
			if not node_hasSuccessor( node ) then
				foundNode = true
				-- debug.print( "Auto closing node for lack of successors" )
				node:close()
			end
			
			node = nextNode

		end
		
	until not foundNode

end

--------------------------------------------------------------------------------

local function checkAutoCloseNodes()

	if not haveAutoCloseNodes then
		return
	end
	
	if checkingAutoCloseNodes then
		return
	end
	
	checkingAutoCloseNodes = true
		-- Re-entrancy protection

	-- pcall checkAutoCloseNodes_imp until it succeeds

	repeat
	until pcall( checkAutoCloseNodes_imp )
	
	checkingAutoCloseNodes = false

	haveAutoCloseNodes = false
		-- We have cleared the set of nodes to look at
	
end

--------------------------------------------------------------------------------

-- Mark the node as done. This disconnects it in both directions and adjusts
-- the depth.
	
local function disconnectNodeAndMarkAsDone( node )

	assertNode( node, "disconnectNodeAndMarkAsDone" )
	
	if node_isDone( node ) then
		return
	end
	
	-- Remove successor links
	
	for succ in node_iterSuccessors( node ) do
		nodes_removeLink( node, succ )
	end
	
	-- Remove predecessor links
	
	local autoClose = AUTO_CLOSE and not rawget( node, keyBlockPredecessorAutoClose )
	
	for pred in node_iterPredecessors( node ) do
		nodes_removeLink( pred, node )
		if autoClose and not rawget( pred, keyBlockSelfAutoClose ) then
			possibleAutoCloseNodes[ pred ] = true
			haveAutoCloseNodes = true
		end
	end
	
	-- Set the depth to mark the node as done

	node[ keyDepth ] = -1
	
end

--------------------------------------------------------------------------------

local function wrapAssertNode( msg, fn )
	return function( node )
		assertNode( node, msg )
		return fn( node )
	end
end

--------------------------------------------------------------------------------

local nodeDAG_nodeHasPredecessor = wrapAssertNode( "nodeHasPredecessor", node_hasPredecessor )
local nodeDAG_nodeIsBase = wrapAssertNode( "nodeIsBase", node_isBase )
local nodeDAG_iterSuccessors = wrapAssertNode( "iterSuccessors", node_iterSuccessors )
local nodeDAG_nodeIsDone = wrapAssertNode( "nodeIsDone", node_isDone )

--============================================================================--

-- FIXME [mhamburg 2011-01-18]: Move to generic utilities.

local function makeMetatable_populateFrom( source )
	
	return {
		__index = function( t, k )
			local v = source[ k ]
			rawset( t, k, v )
			return v
		end,
		__metatable = "< metatable: populate from >"
	}

end

--------------------------------------------------------------------------------

local function makeNodePackedValueQueue()

	local enqueue1, dequeue1 = WFQueueing_makeQueue()
	
	local function enqueue( node, packedValue )
		enqueue1( node )
		enqueue1( packedValue )
	end
	
	local function dequeue()
		return dequeue1(), dequeue1()
	end
	
	return enqueue, dequeue

end

--------------------------------------------------------------------------------

-- The delivery queue contains the messages to be delivered asap this cycle
-- without regard to further scheduling constraints. These are messages sent
-- by sources and via this queue arriving at destinations.

local deliveryQueue_pushNodePackedValue, deliveryQueue_popNodePackedValue =
			makeNodePackedValueQueue()

--------------------------------------------------------------------------------

-- The input queue represents inputs from the public API. Post operations
-- and close operations from outside get encoded here. delay nodes run
-- before the input queue. stable nodes run after the input queue. The values
-- here represent values to be provided to the node send routine.

local inputQueue_pushNodePackedValue, inputQueue_popNodePackedValue =
			makeNodePackedValueQueue()
			
--------------------------------------------------------------------------------

-- We run the scheduler based on negative node depths so that higher node
-- depths are processed later. We manage this through a priority queue.

local scheduleQueue_enqueue, scheduleQueue_dequeue = WFQueueing.makePriorityQueue()

--------------------------------------------------------------------------------

-- We avoid putting a node into the schedule queue multiple times by tracking
-- whether it is in the schedule queue. We only set this for nodes that actually
-- end up in the queue.

local keyIsInScheduleQueue = KSPrivateKeys_makePrivateKey( 'KSReactive.isInScheduleQueue' )

--------------------------------------------------------------------------------

local function node_addToScheduleQueue( node )

	if node[ keyIsInScheduleQueue ] then
		return
	end

	assert( not nodeDAG_nodeIsBase( node ),
		"Base nodes should not need to be scheduled" )

	local depth = getNodeDepth( node )
	
	scheduleQueue_enqueue( -depth, node )
	
	node[ keyIsInScheduleQueue ] = true

end

--------------------------------------------------------------------------------

local function nextScheduledNode()

	while true do
	
		local priority, node = scheduleQueue_dequeue()
		
		-- If the queue comes back empty, we're done.

		if not priority then
			return nil
		end

		-- It's no longer in the scheduled queue.

		node[ keyIsInScheduleQueue ] = false
		
		-- We can only return it, however, if the depth is right.
		
		local depth = getNodeDepth( node )
		
		if depth <= -priority then
			return node
		end
		
		-- Otherwise, we need to reschedule.
		
		scheduleQueue_enqueue( -depth, node )
		
		node[ keyIsInScheduleQueue ] = true
		
	end

end

--------------------------------------------------------------------------------

local valuesSentThisCycle = nil
	-- A table mapping nodes to values for everything sent this cycle. Cleared
	-- after each cycle. Used when linking to nodes that have already fired.
	-- A cycle is defined as the end of running the propagation for one event
	-- in the network.

--------------------------------------------------------------------------------

local function node_enqueuePackedValueForDelivery( node, packedValue )

	if valuesSentThisCycle == nil then
		valuesSentThisCycle = { }
	end
	
	assert( valuesSentThisCycle[ node ] == nil, "We can only send one value per node per cycle" )

	valuesSentThisCycle[ node ] = packedValue
	
	for successor in nodeDAG_iterSuccessors( node ) do
		deliveryQueue_pushNodePackedValue( successor, packedValue )
	end
	
end

--------------------------------------------------------------------------------

local function node_sendPackedValue( node, packedValue )

	if nodeDAG_nodeIsDone( node ) then
		return
	end

	node_enqueuePackedValueForDelivery( node, packedValue )
		-- All of the sinks will receive the packed value via node_deliverPackedValue.
		
	local debugHook = node[ keyDebugHook ]
	
	if debugHook then
		debugHook( node, packedValue )
	end
		
	-- If the value is kDoneValue, then this node is done and we need to
	-- disconnect it from the network and mark it as done. Note that we don't
	-- check ahead of time to see whether this node has already been marked done.
	-- If it were marked done, then the above enqueue should be a no-op and the
	-- disconnect, etc shouldn't have any work to do.
	
	if packedValue == kDoneValue then
		disconnectNodeAndMarkAsDone( node )
		if node[ keyCloseWhenDone ] then
			node:close()
		end
	end

end

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-05-05]: Failure here should probably be treated as fatal
-- since it signals a problem putting things onto queues.

local function node_finishDeliveryProcessing( node, value )

	if value == nil then
		return
	end
	
	if value == kScheduleRequestValue then
		node_addToScheduleQueue( node )
		return
	end
	
	assert( not node._delayed, "Delayed nodes should not return values from their receive functions" )
	
	node_sendPackedValue( node, value )

end

--------------------------------------------------------------------------------

local function node_deliverPackedValue( node, packedValue )

	-- Don't deliver to done nodes.

	if nodeDAG_nodeIsDone( node ) then
		return
	end
	
	-- We don't pcall here because we expect the delivery queue to handle the
	-- case where delivery fails.

	local resultValue
	
	-- Dispatch based on the type of value to deliver.
	
	if packedValue == kDoneValue then
		resultValue = node:_receiveDone( not nodeDAG_nodeHasPredecessor( node ) )
	elseif KSRPackedValues_isError( packedValue ) then
		resultValue = node:_receiveError( packedValue )
	else
		resultValue = node:_receiveData( packedValue )
	end
	
	-- Process the results of the delivery.
	
	node_finishDeliveryProcessing( node, resultValue )

end

--------------------------------------------------------------------------------

local function resendPackedValueForThisCycleFromTo( source, sink )

	if nodeDAG_nodeIsDone( source ) then
		node_deliverPackedValue( sink, kDoneValue)
		return
	end
	
	if not valuesSentThisCycle then
		return
	end
	
	local sentValue = valuesSentThisCycle[ source ]
	
	if not sentValue then
		return
	end
	
	-- FIXME [mhamburg 2011-03-11]: Could we avoid error handling logic issues
	-- here by instead just putting the sent value on the delivery queue?
	
	node_deliverPackedValue( sink, sentValue )
	
end

--------------------------------------------------------------------------------

local function node_deliveryFailed( node, packedValue, errorMessage )

	debug.print( "Delivery failed for node: ", node, packedValue, errorMessage,
				 "\nin event processor " .. require 'WFCurrentThread'.DEBUG_getThreadName(),
				 "\n\nvia " .. traceback() .. "\n\n" )
	
	-- FIXME [mhamburg 2011-05-05]: What if the node is a delayed node? Can
	-- we say those just don't fail?
	
	-- FIXME [mhamburg 2011-05-05]: Can we somehow delay this until it's time
	-- for the node to run so that it doesn't get a chance to fire again this
	-- cycle if it has multiple inputs?
	
	pcall( node_sendPackedValue, node, KSRPackedValues_error( errorMessage ) )
		-- The pcall just represents paranoia around not bringing down the
		-- system when we're in error handling code.

end

--------------------------------------------------------------------------------

local prevDeliveryNode = nil
local prevDeliveryValue = nil
	-- The last node for which we attempted delivery. Used in error handling.

--------------------------------------------------------------------------------

-- Run the delivery queue as far as we can until we encounter an error or
-- finish the queue. Record the node being currently processed as we go so
-- that we can clean up if need be. Note that this logic is not re-entrant.

local function runDeliveryQueue_core()

	for node, packedValue in deliveryQueue_popNodePackedValue do
		prevDeliveryNode = node
		prevDeliveryValue = packedValue
		node_deliverPackedValue( node, packedValue )
	end
	
	prevDeliveryNode = nil
	prevDeliveryValue = nil

end

--------------------------------------------------------------------------------

-- Call runDeliveryQueue_core repeatedly cleaning up from errors each time.
-- When it returns without error, the queue must be done.

local function runDeliveryQueue()

	while true do
		local success, message = pcall( runDeliveryQueue_core )
		if success then
			return
		end
		if prevDeliveryNode then
			node_deliveryFailed( prevDeliveryNode, prevDeliveryValue, message )
		end
		prevDeliveryNode = nil
		prevDeliveryValue = nil
	end

end

--------------------------------------------------------------------------------

local function node_runFailed( node, errorMessage )

	debug.print( "Run failed for node: ", node, errorMessage )
	
	-- Send the error message on as the value.

	pcall( node_sendPackedValue, node, KSRPackedValues_error( errorMessage ) )

end

--------------------------------------------------------------------------------

local function runScheduledQueue()

	while true do
		runDeliveryQueue()
		local node = nextScheduledNode()
		if not node then
			break
		end
		local success, value = pcall( node._run, node )
		if success then
			if value then
				node_sendPackedValue( node, value )
			end
		else
			node_runFailed( node, value )
		end
	end
	
	valuesSentThisCycle = nil
		-- The cycle ends when we finish running all of the scheduled nodes

end

--------------------------------------------------------------------------------

local nextTickNodePackedValuePairs = { }
-- NOT USED: DISABLED: local stableNodePackedValuePairs = { }
-- NOT USED: DISABLED: local idleNodePackedValuePairs = { }

--------------------------------------------------------------------------------

local sendingNode = nil
local sendingValue = nil
	-- Used for error handling
	
local didSend = false

--------------------------------------------------------------------------------

local function sendDelayedNodePackedValuePairs_core( nodePackedValuePairs )

	for node, packedValue in pairs( nodePackedValuePairs ) do
		nodePackedValuePairs[ node ] = nil
		sendingNode = node
		sendingValue = packedValue
		if packedValue then
			didSend = true
			node_sendPackedValue( node, packedValue )
		end
	end
	
	sendingNode = nil
	sendingValue = nil

end

--------------------------------------------------------------------------------

local function node_delayedSendFailed( node, packedValue, errorMessage )

	debug.print( "Delayed send failed for node: ", errorMessage )

	-- FIXME [mhamburg 2011-02-15]

end

--------------------------------------------------------------------------------

local function sendDelayedNodePackedValuePairs( nodePackedValuePairs )

	-- GRUNGE [scouten 2011-03-11]: Seems like this batching-pcall loop metaphor
	-- appears frequently. Opportunity for consolidation of code?
	--
	-- [mhamburg 2011-03-11]: Maybe, but it gets tricky to do so without generating
	-- closure creation overhead.
	
	didSend = false

	while true do
	
		local success, message =
			pcall( sendDelayedNodePackedValuePairs_core, nodePackedValuePairs )

		if success then
			break
		end

		-- assert( false, "Send delayed queue experienced failure: " .. tostring( message ) )
		
		node_delayedSendFailed( sendingNode, sendingValue, message )

		sendingNode = nil
		sendingValue = nil

	end
	
	return didSend

end

--------------------------------------------------------------------------------

local function node_sendDelayedPackedValue( accumulator, node, packedValue )
	accumulator[ node ] = packedValue
end

--------------------------------------------------------------------------------

local function node_sendDoneOnNextTick( node )
	node_sendDelayedPackedValue( nextTickNodePackedValuePairs, node, kDoneValue )
end

--------------------------------------------------------------------------------

local runningScheduledAndNextTick = false

--------------------------------------------------------------------------------

local function runScheduledAndNextTick()

	if runningScheduledAndNextTick then
		return
	end
	
	runningScheduledAndNextTick = true

	repeat
		runScheduledQueue()
		local didSend = sendDelayedNodePackedValuePairs( nextTickNodePackedValuePairs )
	until not didSend
	
	runningScheduledAndNextTick = false

end

--------------------------------------------------------------------------------

-- We avoid reentrancy on the code to run the input queue since once it's
-- running, it will run until drained.

local runningInputQueue = false

--------------------------------------------------------------------------------

-- Flush values out of the input queue running the scheduled and next tick
-- actions between them. Once the input queue goes empty, send the stable
-- values if any. We are done when everything except the idle values are done.

local function runInputQueue()

	if runningInputQueue then
		return
	end

	runningInputQueue = true

	while true do

		runScheduledAndNextTick() -- Generally does nothing at first.

		local node, packedValue = inputQueue_popNodePackedValue()
		
		-- If we don't have a node, then we've exhausted the input. Send out
		-- any stable values.
		
		if node then
		
			node_sendPackedValue( node, packedValue )
				-- FIXME [mhamburg 2011-02-02]: We should ensure that this doesn't result in errors
		
		else
		
			-- The input queue is empty, so send the stable values. If those send
			-- anything, it's possible that we will have more work to do and hence
			-- we cycle around again.
			
			break
			
			--[[ NOT USED: DISABLED
		
			if not sendDelayedNodePackedValuePairs( stableNodePackedValuePairs ) then
				break
			end
			
			]]
			
		end
		
	end
	
	runningInputQueue = false

end

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

local runningIdle = false

--------------------------------------------------------------------------------

local function runIdle()

	if runningIdle then
		return
	end
	
	runningIdle = true

	runInputQueue()
		-- This should not do anything if we've executed properly, but flush
		-- it just to be safe.
		
	local didSend = sendDelayedNodePackedValuePairs( idleNodePackedValuePairs )
	
	runInputQueue()
	
	runningIdle = false

	return didSend
		-- FIXME [mhamburg 2011-02-02]: Actually we want to report whether there
		-- are any more idle tasks to run.

end

--------------------------------------------------------------------------------

local idleQueued = false

--------------------------------------------------------------------------------

local function processQueuedIdle()
	idleQueued = false
	runIdle()
end

--------------------------------------------------------------------------------

local function queueIdle()
	if not idleQueued then
		if not WFIdle_lazy then
			WFIdle_lazy = require "WFIdle"
		end
		WFIdle_lazy.onIdleDo( processQueuedIdle )
		idleQueued = true
	end
end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

local function isRunning()

	return runningInputQueue or runningIdle

end

]] -- NOT USED: DISABLED

--============================================================================--

-- @mark Node allocation and initialization

--------------------------------------------------------------------------------

local function allocNode( strongSuccessor )

	local node = nodeDAG_allocNode( strongSuccessor )
	
	return node

end

--------------------------------------------------------------------------------

local function initNodeWithMetatable( node, nodeKind, mt, ... )

	local sourceCount = select( '#', ... )
	
	for i = 1, sourceCount do
		local source = select( i, ... )
		if source then
			assertNode( source )
		end
	end
	
	setmetatable( node, mt )
	
	local isBase = node._delayed
	
	if node._delayed then
		isBase = true
	end

	nodeDAG_initNode( node, nodeKind, isBase, ... )
	
	node[ keyDebugHook ] = false
	
	-- If we've already sent values this cycle, then we need to send to
	-- the newly created node.
	
	-- FIXME [mhamburg 2011-03-11]: If any of these resends fail, we should
	-- potentially tear the node back down. Or we need to handle this with
	-- appropriate pcall logic around the delivery.
	
	for i = 1, sourceCount do
		local source = select( i, ... )
		if source then
			resendPackedValueForThisCycleFromTo( source, node )
		end
	end
	
	-- Either we aren't in mid-cycle in which case no values will have been
	-- delivered or we are in which case we will pick up the need to process
	-- this node as part of the cycle logic.
	
	return node

end

--============================================================================--

-- @mark Link adjustment

--------------------------------------------------------------------------------

local function removeLink( pred, succ )

	assertNode( pred, "removeLink: predecessor" )
	assertNode( succ, "removeLink: successor" )
	
	return nodes_removeLink( pred, succ )
	
end

--------------------------------------------------------------------------------

local function addLink( source, sink )

	-- Check that we've got nodes
	
	assertNode( source, "addLink: source" )
	assertNode( sink, "addLink: sink" )
	
	nodes_addLink( source, sink )

	-- FIXME [mhamburg 2011-03-11]: If this fails, we should potentially tear the
	-- link back down. Or we need to handle this with appropriate pcall logic around the delivery.

	resendPackedValueForThisCycleFromTo( source, sink )

end


--============================================================================--

-- @mark Debugging

--------------------------------------------------------------------------------

local function eventNodeToString( node )

	return "KSReactive: " .. getNodeKind( node )

end


--============================================================================--

-- @mark Event nodes

--------------------------------------------------------------------------------

local EventNodeMethods = {

	[ keyCloseWhenDone ] = true

}

--------------------------------------------------------------------------------

local populateEventNodeMethodsMT = makeMetatable_populateFrom( EventNodeMethods )

--------------------------------------------------------------------------------

local function makeEventNodeMetatableForMethods( methods )

	setmetatable( methods, populateEventNodeMethodsMT )
	
	local mt = {
		__index = methods,
		__tostring = eventNodeToString,
		__metatable = "< reactive event node metatable >"
	}
	
	return mt

end

--------------------------------------------------------------------------------

local function makeInitEventNodeForMethods( nodeKind )

	WFArguments_check( 'string' )

	return function ( methods )
	
		WFArguments_check( 'table' )
	
		local mt = makeEventNodeMetatableForMethods( methods )
		
		return function( node, ... )
			return initNodeWithMetatable( node, nodeKind, mt, ... )
		end
		
	end

end

--------------------------------------------------------------------------------

EventNodeMethods.isDone = nodeDAG_nodeIsDone

--------------------------------------------------------------------------------

function EventNodeMethods:close()

	-- If we're already marked as done, then we should have severed links and
	-- sent the done value, so there is no other work to do.

	if not nodeDAG_nodeIsDone( self ) then

		inputQueue_pushNodePackedValue( self, kDoneValue )

		runInputQueue()

		-- Check to see whether this has triggered any work through the auto close
		-- mechanism.

		if AUTO_CLOSE then
			checkAutoCloseNodes()
		end

	end

end

--------------------------------------------------------------------------------

WFCloseSupport.addCloseSupportMethods( EventNodeMethods )
	-- This will wrap the preceding close logic.

--------------------------------------------------------------------------------

-- Block recursion for close target accumulation

local keyBlockCloseTargetRecursion =
	KSPrivateKeys_makePrivateKey( 'KSReactive.keyBlockCloseTargetRecursion' )

--------------------------------------------------------------------------------

function EventNodeMethods:collectCloseTargets()

	-- FIXME_MCH: This doesn't seem to be helping

	if true then
		return
	end

	if AUTO_CLOSE then
		return
	end
	
	local stack = { self }
	local stackTop = 1
	local collected = { }
	
	while 1 <= stackTop do
		local node = stack[ stackTop ]
		stackTop = stackTop - 1
		if collected[ node ] then
			-- do nothing
		else
			collected[ node ] = true
			if node ~= self then
				self:addCloseTarget( node, true ) -- we won't be removing
			end
			if not node[ keyBlockCloseTargetRecursion ] then
				for pred in node_iterPredecessors( node ) do
					stackTop = stackTop + 1
					stack[ stackTop ] = pred
				end
			end
		end
	end
	
end

--------------------------------------------------------------------------------

function EventNodeMethods:blockAutoClose()
	if AUTO_CLOSE then
		rawset( self, keyBlockSelfAutoClose, true )
	end
	return self
end

--------------------------------------------------------------------------------

EventNodeMethods._delayed = false

--------------------------------------------------------------------------------

function EventNodeMethods:_receiveData( packedValue )

	return packedValue

end

--------------------------------------------------------------------------------

function EventNodeMethods:_receiveError( errorValue )

	return errorValue

end

--------------------------------------------------------------------------------

function EventNodeMethods:_receiveDone( allSourcesGone )

	if allSourcesGone then
		return kDoneValue
	end

end

--============================================================================--

-- @mark Behavior Nodes

--------------------------------------------------------------------------------

local behaviors = makeFullyWeakTable()
	-- A set to track all of the behaviors for type testing

--------------------------------------------------------------------------------

local BehaviorNodeMethods = setmetatable( { }, { __index = EventNodeMethods } )

--------------------------------------------------------------------------------

BehaviorNodeMethods[ keyCloseWhenDone ] = false
	-- Behavior nodes do not close by default when done.
	-- FIXME [mhamburg 2011-04-26]: The behavior prior to today was that
	-- behaviors actually ate the done signal which seems overly extreme.
	-- So, now we transmit done but eat close. Is this what we want to do?

--------------------------------------------------------------------------------

function BehaviorNodeMethods:setCloseWhenDone( closeWhenDone )

	rawset( self, keyCloseWhenDone, not not closeWhenDone )

end

--------------------------------------------------------------------------------

local populateBehaviorNodeMethodsMT = makeMetatable_populateFrom( BehaviorNodeMethods )

--------------------------------------------------------------------------------

local keyBehavior_currentValue = KSPrivateKeys_makePrivateKey( 'KSReactive.behaviorCurrentValue' )

--------------------------------------------------------------------------------

local function makeBehaviorNodeMetatableForMethods( methods )

	setmetatable( methods, populateBehaviorNodeMethodsMT )
	
	local mt = {
		__index = methods,
		__tostring = eventNodeToString,
		__metatable = "< reactive behavior node metatable >"
	}
	
	return mt

end

--------------------------------------------------------------------------------

local function makeInitBehaviorNodeForMethods( nodeKind )

	WFArguments_check( 'string' )

	return function( methods )
	
		WFArguments_check( 'table' )
	
		local mt = makeBehaviorNodeMetatableForMethods( methods )
		
		return function( node, initialPackedValue, ... )
			behaviors[ node ] = node
			node[ keyBehavior_currentValue ] = initialPackedValue
			return initNodeWithMetatable( node, nodeKind, mt, ... )
		end
		
	end

end

--------------------------------------------------------------------------------

local function BehaviorNode_packedValueNow( self )

	return self[ keyBehavior_currentValue ]

end

--------------------------------------------------------------------------------

local function BehaviorNode_valueNow( self )

	return KSRPackedValues_unpack( BehaviorNode_packedValueNow( self ) )

end

--------------------------------------------------------------------------------

BehaviorNodeMethods.valueNow = BehaviorNode_valueNow

--------------------------------------------------------------------------------

local function BehaviorNode_isErrorNow( self )

	return KSRPackedValues_isError( BehaviorNode_packedValueNow( self ) )

end

--------------------------------------------------------------------------------

BehaviorNodeMethods.isErrorNow = BehaviorNode_isErrorNow

--------------------------------------------------------------------------------

local function BehaviorNode_setPackedValue( self, newPackedValue )

	local oldPackedValue = self[ keyBehavior_currentValue ]
	
	if KSRPackedValues_testEquals( oldPackedValue, newPackedValue ) then
		return
	end
	
	self[ keyBehavior_currentValue ] = newPackedValue
	
	return newPackedValue

end

--------------------------------------------------------------------------------

function BehaviorNodeMethods:_receiveDone( allSourcesGone )
	-- We still have a value, but we're done changing so let the rest of
	-- the world know.
	if allSourcesGone then
		return kDoneValue
	end
end

--============================================================================--

local function isEventNode( v )

	-- Note that nodeDAG_isNode will report true even if v is closed. This is
	-- important because what we often care about is whether it has torn down
	-- the event graph for that node not whether it is externally closed.
	
	-- FIXME [mhamburg 2011-04-26]: This used to report false if the node
	-- was done but it's still an event node in that case. Does this change
	-- anything?
	
	local vIsEventNode = isNode( v )
		and not WFCloseSupport_hasBeenOrWillBeClosed( v )
	
	local vIsBehavior = vIsEventNode and behaviors[ v ]
	
	return vIsEventNode, vIsBehavior

end

--------------------------------------------------------------------------------

local function isBehavior( v )

	local isEventNode, isBehavior = isEventNode( v )
	
	return isEventNode and isBehavior

end

--============================================================================--

-- @mark Debug hooks

--------------------------------------------------------------------------------

local function formatValues( ... )

	local n = select( '#', ... )
	
	if n == 0 then
		return '< none >'
	end
	
	if n == 1 then
		return tostring( ( ... ) )
	end
	
	if n == 2 then
		local x, y = ...
		return string_format( "%s, %s", tostring( x ), tostring( y ) )
	end

	if n == 3 then
		local x, y, z = ...
		return string_format( "%s, %s, %s", tostring( x ), tostring( y ), tostring( z ) )
	end

	local x, y, z = ...
	return string_format( "%s, %s, %s,...", tostring( x ), tostring( y ), tostring( z ) )

end

--------------------------------------------------------------------------------

function EventNodeMethods:setDebugHook( hookFn, hookName )

	if not hookFn then
		self[ keyDebugHook ] = false
		return self
	end
	
	if type( hookFn ) == 'string' then
		assert( hookName == nil, "When we get a string as the first parameter to setDebugHook, there should be no second parameter." )
		hookName = hookFn
		hookFn = debug.print
	end

	self[ keyDebugHook ] = function( sender, packedValue )
	
		local nameStr
		
		if hookName then
			assert( type( hookName ) == 'string', "Hook name should be a string" )
			local nameFmtStr = behaviors[ sender ] and "B< %s : %s >-> " or "E< %s : %s >-> "
			nameStr = string_format( nameFmtStr, hookName, tostring( sender ) )
		else
			local nameFmtStr = behaviors[ sender ] and "B< %s >-> " or "E< %s >-> "
			nameStr = string_format( nameFmtStr, tostring( sender ) )
		end

		if packedValue == kDoneValue then
			hookFn( string_format( "%s DONE", nameStr ) )
			return
		end
		
		local isError, message = KSRPackedValues_isError( packedValue )
					
		if isError  then
			hookFn( string_format( "%s ERROR: %s", nameStr, tostring( message ) ) )
			return
		end
		
		hookFn( string_format( "%s DATA: %s", nameStr, formatValues( KSRPackedValues.unpack( packedValue ) ) ) )
		
	end
	
	return self

end


--============================================================================--

-- @mark constantB, constantErrorB

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-05-04]: I was preventing constant nodes from closing
-- or being close targets but this interferes with optimizing things down
-- to constant nodes. The result of removing this is that we can't memoize
-- them but that also fought with using them as optimization points. We could
-- perhaps memoize just the public entry point but even that gets risky.

--------------------------------------------------------------------------------

local constantB_initNode = makeInitBehaviorNodeForMethods 'constantB' {

	_receiveData = function()
		error( "Constant nodes should not receive values" )
	end,
	
	_receiveError = function()
		error( "Constant nodes should not receive error values" )
	end,
	
}

--------------------------------------------------------------------------------

local constantB_packedValue = function( packedValue )

	local node = constantB_initNode( allocNode( false ), packedValue )
	
	disconnectNodeAndMarkAsDone( node )
		-- A constant node should not generate any permanent successors
		-- since it won't ever change its value.
	
	return node

end

--------------------------------------------------------------------------------

local function constantB( ... )

	return constantB_packedValue( KSRPackedValues_pack( ... ) )

end

--------------------------------------------------------------------------------

local function constantErrorB( ... )

	return constantB_packedValue( KSRPackedValues_error( ... ) )

end


--============================================================================--

-- @mark holdB

--------------------------------------------------------------------------------

local holdB_initNode = makeInitBehaviorNodeForMethods 'holdB' {

	_receiveData = BehaviorNode_setPackedValue,
	
	_receiveError = BehaviorNode_setPackedValue,
	
}

--------------------------------------------------------------------------------

local function EventNode_holdB_packed( self, packedInitialValue, closesSelf )

	if nodeDAG_nodeIsDone( self ) then
		return constantB_packedValue( packedInitialValue )
	end

	local holdB = holdB_initNode( allocNode( false ), packedInitialValue, self )
	
	if closesSelf then
		holdB:addCloseTarget( self )
	end
	
	return holdB

end

--------------------------------------------------------------------------------

function EventNodeMethods:holdB( ... )
	
	return EventNode_holdB_packed( self, KSRPackedValues_pack( ... ), false )

end
	
--------------------------------------------------------------------------------

function EventNodeMethods:closingHoldB( ... )
	
	return EventNode_holdB_packed( self, KSRPackedValues_pack( ... ), true )

end
	

--============================================================================--

-- @mark zeroE

--------------------------------------------------------------------------------

-- PERF [mhamburg 2011-04-27]? We could optimize a variety of cases to
-- return zeroE if they are already done when constructed.

--------------------------------------------------------------------------------

local zeroE -- forward declaration

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local function zeroE_noInputs()
		debug.print( "zeroE should never receive data; eating value" )
	end
	
	--------------------------------------------------------------------------------
	
	local zeroE_initNode = makeInitEventNodeForMethods 'zeroE' {
		_receiveData = zeroE_noInputs,
		_receiveError = zeroE_noInputs,
		_receiveDone = zeroE_noInputs,
	}
	
	--------------------------------------------------------------------------------
	
	zeroE = function()
	
		local node = zeroE_initNode( allocNode( false ) )
		
		disconnectNodeAndMarkAsDone( node )
		
		assert( nodeDAG_nodeIsDone( node ), "Zero node is not done" )
		
		return node
	
	end
	
	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark filterE

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyPredicate = KSPrivateKeys_makePrivateKey( 'KSReactive.filterE_predicate' )
	
	--------------------------------------------------------------------------------
	
	local filterE_initNode = makeInitEventNodeForMethods 'filterE' {
	
		_receiveData = function( self, packedValue )
			local predicate = self[ keyPredicate ]
			if predicate( KSRPackedValues_unpack( packedValue ) ) then
				return packedValue
			end
		end,
		
		--[[
		
		-- Inherit the behavior of passing errors through unchanged
		
		_receivedErrorValue = function( self, errorValue )
			return errorValue -- pass errors through. FIXME [mhamburg 2011-02-15]: Desired behavior?
		end
		
		]]
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:filterE( predicate )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
		
		local node = allocNode( false )
		
		node[ keyPredicate ] = predicate
		
		return filterE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark skipErrorsE

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local skipErrorsE_initNode = makeInitEventNodeForMethods 'skipErrorE' {
	
		_receiveError = function() end
			-- Just eat error values
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:skipErrorsE()
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
	
		return skipErrorsE_initNode( allocNode( false ), self )
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark mapE, mapB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyMapFunction = KSPrivateKeys_makePrivateKey( 'KSReactive.mapFunction' )
	
	--------------------------------------------------------------------------------
	
	local function mapE_nodeReceivePackedValue_finish( oldPackedValue, ... )
	
		local n = select( '#', ... )
		
		if ( not ALLOW_EMPTY_TUPLES_FROM_MAP ) and ( n == 0 ) then
			return
		end
		
		if n == 1 then
			local v = ( ... )
			if KSRPackedValues_sameValue == v then
				return oldPackedValue
			elseif KSRPackedValues_noneValue == v then
				return nil
			else
				return KSRPackedValues_packOne( v )
			end
		end
		
		return KSRPackedValues_pack( ... )
	
	end
	
	--------------------------------------------------------------------------------
	
	local mapE_initNode = makeInitEventNodeForMethods 'mapE' {
	
		_receiveData = function( self, packedValue )
		
			local fn = self[ keyMapFunction ]
			
			return mapE_nodeReceivePackedValue_finish(
						packedValue,
						fn( KSRPackedValues_unpack( packedValue ) ) )
		
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:mapE( fn )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
	
		local node = allocNode( false )
		
		node[ keyMapFunction ] = fn
		
		return mapE_initNode( node, self )
	
	end
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:closingMapE( fn )
	
		return self:mapE( fn ):addCloseTarget( self )
	
	end
	
	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:mapB( fn )
	
		WFArguments_check( '_', 'callable' )
		
		-- FIXME [mhamburg 2011-05-04]: Assert that we aren't inside a close
		-- component or escape out of the component if we are.
	
		local initialPackedValue = mapE_nodeReceivePackedValue_finish(
					BehaviorNode_packedValueNow( self ),
					fn( BehaviorNode_valueNow( self ) ) )
					
		if nodeDAG_nodeIsDone( self ) then
			return constantB_packedValue( initialPackedValue )
		end
		
		return EventNode_holdB_packed( self:mapE( fn ), initialPackedValue, true )
	
	end
	
	--------------------------------------------------------------------------------

	function BehaviorNodeMethods:closingMapB( fn )
	
		local mappedB = self:mapB( fn )
		
		mappedB:addCloseTarget( self )
		
		return mappedB
	
	end
	
	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark rawMapE, rawMapB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyRawMapFunction = KSPrivateKeys_makePrivateKey( 'KSReactive.rawMapFunction' )
	
	--------------------------------------------------------------------------------
	
	local function rawMapE_nodeReceivePackedValue_finish( oldPackedValue, ... )
	
		local n = select( '#', ... )
		
		-- In the raw case, we can't process the nothing case so we exit early.
		
		if n == 0 then
			return
		end
		
		local status, firstValue = ...
		
		if status == KSRPackedValues_sameValue or status == 'same' then
			assert( n == 1, "Can only use same when returning no other values" )
			return oldPackedValue
		end
		
		if status == KSRPackedValues_noneValue then
			return
		end
		
		if status == 'data' then
			if n == 2 then
				return KSRPackedValues_packOne( firstValue ) -- common case
			else
				return KSRPackedValues_pack( select( 2, ... ) )
			end
		elseif status == 'error' then
			return KSRPackedValues_error( firstValue )
		else
			debug.print( "Bad status for rawMap: ", status )
		end
	
	end
	
	--------------------------------------------------------------------------------
	
	local function callRawMap( fn, packedValue )
	
		local isError, errorMessage = KSRPackedValues_isError( packedValue )
		
		if isError then
			return fn( 'error', errorMessage )
		else
			return fn( 'data', KSRPackedValues_unpack( packedValue ) )
		end
	
	end
	
	--------------------------------------------------------------------------------
	
	local function rawMapE_receiveNonDone( self, packedValue )
	
		return rawMapE_nodeReceivePackedValue_finish(
					packedValue,
					callRawMap( self[ keyRawMapFunction ], packedValue ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	local rawMapE_initNode = makeInitEventNodeForMethods 'rawMapE' {
	
		_receiveData = rawMapE_receiveNonDone,
		_receiveError = rawMapE_receiveNonDone,
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:rawMapE( fn )
	
		WFArguments_check( '_', 'callable' )
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
	
		local node = allocNode( false )
		
		node[ keyRawMapFunction ] = fn
		
		return rawMapE_initNode( node, self )
	
	end
	
	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:rawMapB( fn )
	
		WFArguments_check( '_', 'callable' )
		
		local packedValue = BehaviorNode_packedValueNow( self )
		
		local initialPackedValue = rawMapE_nodeReceivePackedValue_finish(
					packedValue,
					callRawMap( fn, packedValue ) )
					
		if nodeDAG_nodeIsDone( self ) then
			return constantB_packedValue( initialPackedValue )
		end
		
		return EventNode_holdB_packed( self:rawMapE( fn ), initialPackedValue, true )
	
	end
	
	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark mcallB, mcallSwitchB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local defaultToMethodFunctionCacheWantValues = { }
	local defaultToMethodFunctionCacheWantBehaviors = { }
	
	local kNil = { }
	local kSame = { }

	--------------------------------------------------------------------------------

	local function methodFunction( wantBehaviors, methodName, ... )
	
		WFArguments_check( 'boolean', 'string' )
		
		local cacheCache = wantBehaviors
				and defaultToMethodFunctionCacheWantBehaviors
				 or defaultToMethodFunctionCacheWantValues
		
		local default = ( ... )
		local defaultKey = default
		
		if default == nil then
			if select( '#', ... ) == 0 then
				default = kSame
				defaultKey = kSame
			else
				defaultKey = kNil
			end
		end
		
		local cache = cacheCache[ defaultKey ]
		
		if not cache then
			cache = makeFullyWeakTable()
			cacheCache[ defaultKey ] = cache
		end
		
		local cached = cache[ methodName ]
		
		if cached then
			return cached
		end
		
		local result
		
		if wantBehaviors then
			if default == kSame then
				result = function( obj )
					if not obj then return constantB( obj ) end
					return obj[ methodName ]( obj )
				end
			else
				result = function( obj )
					if not obj then return constantB( default ) end
					return obj[ methodName ]( obj )
				end
			end
		else
			if default == kSame then
				result = function( obj )
					if not obj then return obj end
					return obj[ methodName ]( obj )
				end
			else
				result = function( obj )
					if not obj then return default end
					return obj[ methodName ]( obj )
				end
			end
		end
		
		cache[ methodName ] = result
		
		return result

	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:mcallB( methodName, ... )
		return self:mapB( methodFunction( false, methodName, ... ) )
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:mcallSwitchB( methodName, ... )
		return self:mapSwitchB( methodFunction( true, methodName, ... ) )
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark propertyB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local defaultToPropertyFunctionCache = { }
	
	local kNil = { }
	local kSame = { }

	--------------------------------------------------------------------------------
	
	local function propertyFunction( propertyName, ... )
	
		local default = ( ... )
		local defaultKey = default
		
		if default == nil then
			if select( '#', ... ) == 0 then
				default = kSame
				defaultKey = kSame
			else
				defaultKey = kNil
			end
		end
		
		local cache = defaultToPropertyFunctionCache[ defaultKey ]
		
		if not cache then
			cache = makeFullyWeakTable()
			defaultToPropertyFunctionCache[ defaultKey ] = cache
		end
		
		local cached = cache[ propertyName ]
		
		if cached then
			return cached
		end
		
		local result
		
		if default == kSame then
			result = function( obj )
				if not obj then return obj end
				return obj[ propertyName ]
			end
		else
			result = function( obj )
				if not obj then return default end
				return obj[ propertyName ]
			end
		end
		
		cache[ propertyName ] = result
		
		return result
	
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:propertyB( propertyName, ... )
		WFArguments_check( '_', 'string' )
		return self:mapB( propertyFunction( propertyName, ... ) )
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark toBooleanB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local function toBoolean( v )
		return not not v
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:toBooleanB()
		return self:mapB( toBoolean )
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark filterB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:filterB( fn, ... )
		local filteredE = self:filterE( fn )
		if self:isErrorNow() or fn( self:valueNow() ) then
			return filteredE:holdB( self:valueNow() )
		else
			return filteredE:holdB( ... )
		end
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark skipErrorsB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:skipErrorsB( ... )
		local skippedE = self:skipErrorsE()
		if not self:isErrorNow() then
			return skippedE:holdB( self:valueNow() )
		else
			return skippedE:holdB( ... )
		end
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark relayE, relayB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	-- We provide support for hanging a set of auto close relays off of a behavior
	-- or event such that when the last relay closes we will close the behavior.
	-- This will probably get replaced with a check for whether we have a strong
	-- dependent on a node.
	
	local keyRelaySet = KSPrivateKeys_makePrivateKey( 'KSReactive.relaySet' )
	
	local function setupAutoCloseRelay( wantAutoClose, base, relay )
	
		-- Note that we don't do anything if we are running the auto close logic
		-- since this will already do the same thing.
	
		if wantAutoClose and not AUTO_CLOSE then
	
			-- Add the result to the relay set for the behavior
			
			local relaySet = rawget( base, keyRelaySet )
			
			if not relaySet then
				relaySet = makeFullyWeakTable()
				rawset( base, keyRelaySet, relaySet )
			end
			
			relaySet[ relay ] = relay
			
			-- When we close the relay, we clear it from the relay set and if it was
			-- the last item in the set, we close the source behavior.
			
			relay:addCloseAction( function()
				relaySet[ relay ] = nil
				if next( relaySet ) == nil then
					base:close()
				end
			end )
			
			-- NOTE: We do not use mustClose on these because that will result on
			-- the finalizer running at an arbitrary time which could close the
			-- underlying relayed object just as we were reviving it from a weak
			-- cache. [mhamburg 2011-08-24]
			
		end
		
		-- Relays block close target recursion
		
		if AUTO_CLOSE then
			if not wantAutoClose then
				rawset( relay, keyBlockPredecessorAutoClose, true )
			end
		else
			rawset( relay, keyBlockCloseTargetRecursion, true )
		end
		
		return relay
	
	end
	
	--------------------------------------------------------------------------------
	
	local relayE_initNode = makeInitEventNodeForMethods 'relayE' { }
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:relayE( wantAutoClose )
	
		-- FIXME [mhamburg 2011-05-04]: ZeroE?
	
		return setupAutoCloseRelay( wantAutoClose,
					self,
					relayE_initNode( allocNode( false ), self ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:relayB( wantAutoClose )
	
		return setupAutoCloseRelay( wantAutoClose,
					self,
					self:holdB( self:valueNow() ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:noCloseRelayE()
	
		return self:relayE():bounceClosure()
	
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:noCloseRelayB()
	
		return self:relayB():bounceClosure()
	
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:relayForAutoCloseB()
	
		if AUTO_CLOSE then
			return self:relayB()
		else
			return self
		end
		
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark onDataE, onErrorE, onDoneE, onDataB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyAction = KSPrivateKeys_makePrivateKey( 'KSReactive.action' )
	
	--------------------------------------------------------------------------------
	
	local onDataE_initNode = makeInitEventNodeForMethods 'onDataE' {
	
		_receiveData = function( self, packedValue )
			local actionFn = self[ keyAction ]
			dpcall( 'onDataE', actionFn, KSRPackedValues_unpack( packedValue ) )
			return packedValue
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:onDataE( fn )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
	
		local node = allocNode( true )
		
		node[ keyAction ] = fn
		
		onDataE_initNode( node, self )
		
		return node
	
	end
	
	--------------------------------------------------------------------------------
	
	local onErrorE_initNode = makeInitEventNodeForMethods 'onErrorE' {
	
		_receiveError = function( self, errorValue )
			local actionFn = self[ keyAction ]
			local isError, errorMessage = KSRPackedValues_isError( errorValue )
			assert( isError, "Expected to be called with an error value" )
			dpcall( 'onErrorE', actionFn, errorMessage )
			return errorValue
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:onErrorE( fn )
	
		WFArguments_check( '_', 'callable' )
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end
	
		local node = allocNode( true )
		
		node[ keyAction ] = fn
		
		onErrorE_initNode( node, self )
	
		return node
	
	end
	
	--------------------------------------------------------------------------------
	
	local onDoneE_initNode = makeInitEventNodeForMethods 'onDoneE' {
	
		_receiveDone = function( self, allSourcesGone )
			if allSourcesGone then
				local actionFn = self[ keyAction ]
				dpcall( 'onDoneE', actionFn )
				return kDoneValue
			end
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:onDoneE( fn )
	
		WFArguments_check( '_', 'callable' )
		
		-- Here we care about the done status, so we don't simply switch to
		-- a zeroE node.
	
		local node = allocNode( true )
		
		node[ keyAction ] = fn
		
		onDoneE_initNode( node, self )
	
		return node
	
	end
	
	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:onDataB( fn )
	
		local pv = BehaviorNode_packedValueNow( self )
		
		if not KSRPackedValues_isError( pv ) then
			fn( KSRPackedValues_unpack( pv ) )
		end
		
		return EventNode_holdB_packed( self:onDataE( fn ), pv, true )
	
	end
	
	--------------------------------------------------------------------------------
	
	local runUntilDoneSet = { }
	
	function EventNodeMethods:runUntilDone()
	
		local doneE
		
		local function doneFn()
			runUntilDoneSet[ doneE ] = nil
			self:close() -- We actually close because we may want to close behaviors
		end
		
		doneE = self:onDoneE( doneFn )
		
		runUntilDoneSet[ doneE ] = true
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark skipFirstE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keySkipped = KSPrivateKeys_makePrivateKey( 'KSReactive.skipped' )
	
	--------------------------------------------------------------------------------
	
	local skipFirstE_receiveNonDone = function( self, packedValue )
		if self[ keySkipped ] then
			return packedValue
		else
			self[ keySkipped ] = true
		end
	end
	
	--------------------------------------------------------------------------------
	
	local skipFirstE_initNode = makeInitEventNodeForMethods 'skipFirstE' {
	
		_receiveData = skipFirstE_receiveNonDone,
		_receiveError = skipFirstE_receiveNonDone
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:skipFirstE()
	
		-- GRUNGE [mhamburg 2011-01-12]: If we could rewire after a node had
		-- activated, we could actually just have this node remove itself from
		-- the graph.
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keySkipped ] = false
		
		return skipFirstE_initNode( node, self )
	
	end
	
	--------------------------------------------------------------------------------

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED


--============================================================================--

-- @mark switchE, mapSwitchE, mapSwitchB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keySwitchE_output_pendingValue = KSPrivateKeys_makePrivateKey( 'KSReactive.switchE_output_pendingValue' )
	local keySwitchE_output_linkStream = KSPrivateKeys_makePrivateKey( 'KSReactive.switchE_output_linkStream' )
	local keySwitchE_input_previousEventStream = KSPrivateKeys_makePrivateKey( 'KSReactive.switchE_input_previousEventStream' )
	local keySwitchE_input_outputNode = KSPrivateKeys_makePrivateKey( 'KSReactive.switchE_input_outputNode' )
	
	--------------------------------------------------------------------------------
	
	local function switchE_output_receiveNonDone( self, packedValue )
		self[ keySwitchE_output_pendingValue ] = packedValue
		return kScheduleRequestValue
	end
	
	--------------------------------------------------------------------------------
	
	local switchE_output_initNode = makeInitEventNodeForMethods 'switchE_output' {
	
		_receiveData = switchE_output_receiveNonDone,
		_receiveError = switchE_output_receiveNonDone,
		
		_receiveDone = function( self, allSourcesGone )
		
			-- Check to see whether the link stream just terminated
			-- and if so close it now (probably superfluous) if appropriate
			-- and unlink.
			
			local linkStream = self[ keySwitchE_output_linkStream ]
			
			if linkStream and nodeDAG_nodeIsDone( linkStream ) then
				self[ keySwitchE_output_linkStream ] = false
				linkStream:close()
			end

			-- If this is the end, then terminate.
		
			if allSourcesGone then
				return kDoneValue
			end
			
		end,
		
		_run = function( self )
			local value = self[ keySwitchE_output_pendingValue ]
			self[ keySwitchE_output_pendingValue ] = false
			if value then
				return value
			end
		end,
	
	}
	
	--------------------------------------------------------------------------------
	
	local function switchE_output_updateLinkStream( outputNode, newStream )
	
		local oldStream = outputNode[ keySwitchE_output_linkStream ]
		
		if oldStream == newStream then
			return
		end
		
		if oldStream then
			outputNode[ keySwitchE_output_linkStream ] = false
			removeLink( oldStream, outputNode )
			outputNode:removeCloseTarget( oldStream )
			oldStream:close()
		end
		
		if newStream then
			outputNode:addCloseTarget( newStream )
			addLink( newStream, outputNode )
			outputNode[ keySwitchE_output_linkStream ] = newStream
		end
		
	end
	
	--------------------------------------------------------------------------------
	
	local switchE_input_initNode = makeInitEventNodeForMethods 'switchE_input' {
	
		_receiveData = function( self, packedValue )
		
			local outputNode = self[ keySwitchE_input_outputNode ]
			
			local newStream = KSRPackedValues_first( packedValue ) or false
				-- Note that we force nil to false.
			
			-- Don't switch for duplicate values.
			
			-- GRUNGE [mhamburg 2011-05-05]: Does this conflict with the logic
			-- that sends values from behaviors?
			
			if self[ keySwitchE_input_previousEventStream ] == newStream then
				return
			end
			
			self[ keySwitchE_input_previousEventStream ] = newStream
			
			-- If we don't have a new stream, then we tear things down.
			
			if not newStream then
				outputNode[ keySwitchE_output_pendingValue ] = false
				switchE_output_updateLinkStream( outputNode, false )
				return
			end
			
			-- We need to set the value from the switch. If this is a behavior, we treat the current
			-- value as having fired (subject to getting updated later in this event cycle). If not,
			-- we specify no value and look to either the link operation or a later firing to set
			-- any value. In the former case, we need to schedule the output node to run since it
			-- may need to produce a value.
			
			local nsIsEventStream, nsIsBehavior = isEventNode( newStream )
			
			if not nsIsEventStream then
				debug.print( "switchE expects to receive event nodes; got ", tostring( newStream ) )
				return self:_receiveError( KSRPackedValues_error( "Expected to receive event node" ) )
			end
			
			if nsIsBehavior then
				outputNode[ keySwitchE_output_pendingValue ] = BehaviorNode_packedValueNow( newStream )
				node_addToScheduleQueue( outputNode )
			else
				outputNode[ keySwitchE_output_pendingValue ] = false
			end
			
			-- Link to receive events from the stream if it isn't done. This
			-- also handles close ownership.
			
			local linkStream = newStream
			
			if nodeDAG_nodeIsDone( linkStream ) then
				linkStream = false
			end
			
			switchE_output_updateLinkStream( outputNode, linkStream )
			
		end,
		
		_receivedError = function( self, errorValue )
		
			local outputNode = self[ keySwitchE_input_outputNode ]
			
			self[ keySwitchE_input_previousEventStream ] = false
			
			switchE_output_updateLinkStream( outputNode, false )
	
			outputNode[ keySwitchE_output_pendingValue ] = errorValue
			
			node_addToScheduleQueue( outputNode )
		
		end,
	
	}
	
	--------------------------------------------------------------------------------
	
	-- Construct a switch. We can optionally have the switch close the streams
	-- it receives. We can also specify an initial stream that it holds ahead
	-- of receiving any events. We will not look to this stream for values the
	-- way we would if it were a behavior since we may not be in the midst of
	-- processing event logic. Any use of initial values would be in the context
	-- of behaviors and would be handled via a holdB as a later stage.
	
	local function EventNode_constructSwitchE( source, initialOutputStream )
	
		-- Initial output streams are only interesting if they haven't finished
		-- firing.
	
		if initialOutputStream and nodeDAG_nodeIsDone( initialOutputStream ) then
			initialOutputStream = false
		end
	
		initialOutputStream = initialOutputStream or false
		
		--  If the source is done and there is no initial output stream, then
		-- we will generate nothing.
	
		if nodeDAG_nodeIsDone( source ) and not initialOutputStream then
			return zeroE()
		end
		
		local inputNode = allocNode( false )
		local outputNode = allocNode( false )
		
		inputNode[ keySwitchE_input_previousEventStream ] = initialOutputStream
			-- Debounce the input.
			
		inputNode[ keySwitchE_input_outputNode ] = outputNode
		
		outputNode[ keySwitchE_output_pendingValue ] = false
		outputNode[ keySwitchE_output_linkStream ] = false
		
		-- Start hooking things up. We wait to hook up the source because we
		--  want the internal wiring in place before we allow events to
		-- come in from the source.
		
		switchE_input_initNode( inputNode )
		switchE_output_initNode( outputNode, inputNode )
		
		outputNode:addCloseTarget( inputNode )
		
		-- If we have an initial stream to drive the output, hook it up now.
		-- Note that the addLink here may result in the output node receiving
		-- a value.
		
		if initialOutputStream then
			switchE_output_updateLinkStream( outputNode, initialOutputStream )
		end
		
		-- Now we can hook up the source.
		
		addLink( source, inputNode )

		return outputNode
	
	end
	
	--------------------------------------------------------------------------------
	
	--[[ switchE is no more...
	
	function EventNodeMethods:switchE()
	
		return EventNode_constructSwitchE( self, false, false )
			The middle false is no longer supported
	
	end
	
	]]
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:mapSwitchE( fnE )
	
		local mappedE = self:mapE( fnE )
		
		local switchedE = EventNode_constructSwitchE( mappedE, false )
		
		switchedE:addCloseTarget( mappedE )
		
		return switchedE
	
	end
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:closingMapSwitchE( fnE )
	
		return self:mapSwitchE( fnE ):addCloseTarget( self )
	
	end
	
	--------------------------------------------------------------------------------
	
	local function BehaviorNode_computeInitialStreamAndValue( self, requireBehavior )
	
		-- The initial stream is the initial stream from which we get events
		-- for the output. The initial packed value is the initial output
		-- value. If nothing else were changing, our naive answer would be
		-- initialStream:holdB( initialPackedValue ).
		
		local initialStream, initialPackedValue
		
		local selfPackedValue = BehaviorNode_packedValueNow( self )
		
		if KSRPackedValues_isError( selfPackedValue ) then
			initialStream = false
			initialPackedValue = selfPackedValue
		elseif not selfPackedValue then -- FIXME [mhamburg 2011-05-05]: Allow nil?
			initialStream = false
			initialPackedValue = nil
		else
			local selfValue = KSRPackedValues_first( selfPackedValue )
			local selfValueIsEventNode, selfValueIsBehavior = isEventNode( selfValue )
			assertft( selfValueIsEventNode, "Value should have been an event node." )
			assertft( selfValueIsBehavior or not requireBehavior, "Value should have been a behavior." )
			initialStream = selfValue
			if selfValueIsBehavior then
				initialPackedValue = BehaviorNode_packedValueNow( selfValue )
			end
		end
		
		-- If the initial stream won't be firing, we don't want to assume ownership
		-- for it and we will be able to just use a constant behavior if it never
		-- fires.
		
		if initialStream and nodeDAG_nodeIsDone( initialStream ) then
			initialStream = false
		end
		
		return initialStream, initialPackedValue

	end
	
	--------------------------------------------------------------------------------
	
	local function BehaviorNode_constructSwitchB( self )
	
		-- The initial stream is the initial stream from which we get events
		-- for the output. The initial packed value is the initial output
		-- value. If nothing else were changing, our naive answer would be
		-- initialStream:holdB( initialPackedValue ).
		
		local initialStream, initialPackedValue =
			BehaviorNode_computeInitialStreamAndValue( self, true )
		
		-- The standard case. self isn't done, so we want to switch on the
		-- streams coming in from self. We utilize the behavior that switchE
		-- treats switching to a behavior as firing the current value of the
		-- behavior if the behavior doesn't fire in some other way during the
		-- cycle. We initialize the switch as listening to the initial stream
		-- if any as the source for its output so that we have a value even
		-- before self fires.
		
		if not nodeDAG_nodeIsDone( self ) then
			local switchE = EventNode_constructSwitchE( self, initialStream )
			return EventNode_holdB_packed( switchE, initialPackedValue, true )
		end
		
		-- If self is done but has a value that isn't finished, then we can
		-- listen to that value and optionally close it if we close our input.
		
		if initialStream then
			return EventNode_holdB_packed( initialStream, initialPackedValue, true )
		end
		
		-- Otherwise, we just have a constant node.
		
		return constantB_packedValue( initialPackedValue )

	end
	
	--------------------------------------------------------------------------------
	
	-- mapSwitchB maps a value to a behavior valued behavior and in turn to the value
	-- of the current behavior. It owns the intermediate behaviors and will close
	-- them when switching.
	
	function BehaviorNodeMethods:mapSwitchB( fnB )
	
		local mappedB = self:mapB( fnB )
		
		local switchedB = BehaviorNode_constructSwitchB( mappedB )
		
		switchedB:addCloseTarget( mappedB )
		
		return switchedB
	
	end

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:closingMapSwitchB( fnB )
	
		local mappedB = self:mapSwitchB( fnB )
		
		mappedB:addCloseTarget( self )
		
		return mappedB
	
	end
	
	--------------------------------------------------------------------------------
	
	--[[ UNUSED
	
	local function BehaviorNode_constructSwitchE( self, closesStreams )
	
		-- The initial stream is the initial stream from which we get events
		-- for the output. The initial packed value is the initial output
		-- value. If nothing else were changing, our naive answer would be
		-- initialStream:holdB( initialPackedValue ).
		
		local initialStream, _ = BehaviorNode_computeInitialStreamAndValue( self, false )
		
		-- The standard case. self isn't done, so we want to switch on the
		-- streams coming in from self. We utilize the behavior that switchE
		-- treats switching to a behavior as firing the current value of the
		-- behavior if the behavior doesn't fire in some other way during the
		-- cycle. We initialize the switch as listening to the initial stream
		-- if any as the source for its output so that we have a value even
		-- before self fires.
		
		if not nodeDAG_nodeIsDone( self ) then
			local switchE = EventNode_constructSwitchE( self, closesStreams, initialStream )
			return switchE
		end
		
		-- If self is done but has a value that isn't finished, then we can
		-- listen to that value and optionally close it if we close our input.
		
		if initialStream and not nodeDAG_nodeIsDone( initialStream ) then
			return initialStream
		end
		
		-- Otherwise, we just have a constant node.
		
		return zeroE()

	end
	
	]]
	
	--------------------------------------------------------------------------------
	
	--[[ switchE is no more...
	
	-- switchE for behaviors needs to support an initial event stream unlike
	-- switchE for event nodes
	
	function BehaviorNodeMethods:switchE()
	
		return BehaviorNode_constructSwitchE( self, false )
	
	end
	
	]]
	
	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark delayE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyDelayE_delayTable = KSPrivateKeys_makePrivateKey( 'KSReactive.delayE_delayTable' )
	local keyDelayE_onQueueDo = KSPrivateKeys_makePrivateKey( 'KSReactive.delayE_onQueueDo' )
	
	--------------------------------------------------------------------------------
	
	local function delayE_receive( self, packedValue )
		node_sendDelayedPackedValue( self[ keyDelayE_delayTable ], self, packedValue )
		self[ keyDelayE_onQueueDo ]()
	end
	
	--------------------------------------------------------------------------------
	
	local delayE_initNode = makeInitEventNodeForMethods 'delayE' {
	
		_delayed = true,
		
		_receiveData = delayE_receive,
		_receiveError = delayE_receive,
		_receiveDone = function( self, allSourcesGone )
			if allSourcesGone then
				delayE_receive( self, kDoneValue )
			end
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	local function nop()
		-- Do nothing
		-- We don't use KSFunctionUtils.nop because we don't want to pull in that
		-- namespace just to get this one function since the import is probably
		-- almost as expensive as the declaration.
	end

	--------------------------------------------------------------------------------
	
	function EventNodeMethods:delayE( when )
	
		local delayTable = nil
		local action = nop
		
		if when == nil or when == 'next' then
			delayTable = nextTickNodePackedValuePairs
		elseif when == 'stable' then
			delayTable = stableNodePackedValuePairs
		elseif when == 'idle' then
			delayTable = idleNodePackedValuePairs
			action = queueIdle
		else
			assert( false, "Invalid parameter to delayE: " .. tostring( when ) )
		end
		
		local node = allocNode( false )
		
		node[ keyDelayE_delayTable ] = delayTable
		node[ keyDelayE_onQueueDo ] = action
	
		return delayE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED


--============================================================================--

-- @mark constantE

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyConstantE_value = KSPrivateKeys_makePrivateKey( 'KSReactive.constantE_value' )
	
	--------------------------------------------------------------------------------
	
	local constantE_initNode = makeInitEventNodeForMethods 'constantE' {
	
		_receiveData = function( self )
		
			return self[ keyConstantE_value ]
		
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:constantE( ... )
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyConstantE_value ] = KSRPackedValues_pack( ... )
		
		return constantE_initNode( node, self )
	
	end

--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark takeE

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyTakeE_counter = KSPrivateKeys_makePrivateKey( 'KSReactive.takeE_counter' )
	
	--------------------------------------------------------------------------------
	
	local function takeE_receive( self, packedValue )
		local counter = self[ keyTakeE_counter ]
		counter = counter - 1
		if counter <= 0 then
			node_sendDoneOnNextTick( self )
		else
			self[ keyTakeE_counter ] = counter
		end
		return packedValue
	end
	
	--------------------------------------------------------------------------------
	
	local takeE_initNode = makeInitEventNodeForMethods 'takeE' {
		_receiveData = takeE_receive,
		_receiveError = takeE_receive
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:takeE( n )
	
		WFArguments_check( '_', 'number' )
	
		assert( 0 < n, "Expected positive numeric parameter for takeE." )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyTakeE_counter ] = n
		
		return takeE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark endAfterErrorE

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyEndAfterErrorE_blocked =
		KSPrivateKeys_makePrivateKey( 'KSReactive.keyEndAfterErrorE_blocked' )
	
	--------------------------------------------------------------------------------
	
	local endAfterErrrorE_initNode = makeInitEventNodeForMethods 'endAfterErrorE' {
		_receiveError = function( self, packedValue )
			if self[ keyEndAfterErrorE_blocked ] then
				return
			end
			node_sendDoneOnNextTick( self )
			self[ keyEndAfterErrorE_blocked ]  = true
			return packedValue
		end
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:endAfterErrorE( n )
	
		WFArguments_check( '_' )
	
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyEndAfterErrorE_blocked ] = false
		
		return endAfterErrrorE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark accumulateE, accumulateB

--------------------------------------------------------------------------------

function EventNodeMethods:accumulateE( fn, accumulator )

	WFArguments_check( '_', 'callable' )
	
	local function finish( ... )
		accumulator = ( ... )
		return ...
	end
	
	return self:mapE( function( ... )
		return finish( fn( accumulator, ... ) )
	end )

end

--------------------------------------------------------------------------------

function EventNodeMethods:accumulateB( fn, accumulator, ... )

	local e = self:accumulateE( fn, accumulator )
		-- Generate an event stream through accumulation
	
	local r = e:holdB( accumulator, ... )
		-- Start with the initial accumulator
	
	r:addCloseTarget( e )
		-- Closing the result closes the underlying event stream
	
	return r

end


--============================================================================--

-- @mark whenDoneE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local kPackedTrue = KSRPackedValues.pack( true )

	--------------------------------------------------------------------------------
	
	local function whenDoneE_receiveNonDone( self )
		-- Just eat the value
	end
	
	--------------------------------------------------------------------------------
	
	local whenDoneE_initNode = makeInitEventNodeForMethods 'whenDoneE' {
	
		_receiveData = whenDoneE_receiveNonDone,
		_receiveError = whenDoneE_receiveNonDone,
		
		_receiveDone = function( self, allSourcesGone )
			if allSourcesGone then
				node_sendDoneOnNextTick( self )
				return kPackedTrue
			end
		end
		
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:whenDoneE()
	
		return whenDoneE_initNode( allocNode( false ), self )
	
	end

	--------------------------------------------------------------------------------

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED

--============================================================================--

-- @mark takeWhileE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyTakeWhileE_predicate = KSPrivateKeys_makePrivateKey( 'KSReactive.takeWhileE_predicate' )
	
	--------------------------------------------------------------------------------
	
	local takeWhileE_initNode = makeInitEventNodeForMethods 'takeWhileE' {
	
		_receiveData = function( self, packedValue )
			local predicate = self[ keyTakeWhileE_predicate ]
			if predicate( KSRPackedValues_unpack( packedValue ) ) then
				return packedValue
			else
				return kDoneValue
			end
		end,
		
		_receiveError = function( self, packedValue )
			return kDoneValue
		end
		
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:takeWhileE( predicate )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyTakeWhileE_predicate ] = predicate
		
		return takeWhileE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED


--============================================================================--

-- @mark takeUntilE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

do

	--------------------------------------------------------------------------------
	
	local keyTakeUntilE_predicate = KSPrivateKeys_makePrivateKey( 'KSReactive.takeUntilE_predicate' )
	
	--------------------------------------------------------------------------------
	
	local takeUntilE_initNode = makeInitEventNodeForMethods 'takeUntilE' {
	
		_receiveData = function( self, packedValue )
			local predicate = self[ keyTakeUntilE_predicate ]
			if predicate( KSRPackedValues_unpack( packedValue ) ) then
				return kDoneValue
			else
				return packedValue
			end
		end,
		
		-- Just pass errors through
		
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:takeUntilE( predicate )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyTakeUntilE_predicate ] = predicate
		
		return takeUntilE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end

]] -- NOT USED: DISABLED


--============================================================================--

-- @mark takeThroughE

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyTakeThroughE_predicate = KSPrivateKeys_makePrivateKey( 'KSReactive.takeThroughE_predicate' )
	
	--------------------------------------------------------------------------------
	
	local takeThroughE_initNode = makeInitEventNodeForMethods 'takeThroughE' {
	
		_receiveData = function( self, packedValue )
			local predicate = self[ keyTakeThroughE_predicate ]
			if predicate( KSRPackedValues_unpack( packedValue ) ) then
				node_sendDoneOnNextTick( self )
			end
			return packedValue
		end,
		
		-- Just pass errors through
		
	}
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:takeThroughE( predicate )
	
		WFArguments_check( '_', 'callable' )
		
		if nodeDAG_nodeIsDone( self ) then
			return zeroE()
		end

		local node = allocNode( false )
		
		node[ keyTakeThroughE_predicate ] = predicate
		
		return takeThroughE_initNode( node, self )
	
	end

	--------------------------------------------------------------------------------

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED


--============================================================================--

-- @mark sourceE

--------------------------------------------------------------------------------

local function sourceE_finishPost( self, packedValue )
	inputQueue_pushNodePackedValue( self, packedValue )
	runInputQueue()
	return self
end

--------------------------------------------------------------------------------

local sourceE_initNode = makeInitEventNodeForMethods 'sourceE' {

	_receiveData = function()
		error( "Source nodes should not receive values" )
	end,
	
	post = function( self, ... )
		return sourceE_finishPost( self, KSRPackedValues_pack( ... ) )
	end,
	
	postError = function( self, message )
		return sourceE_finishPost( self, KSRPackedValues_error( message ) )
	end,
	
}

--------------------------------------------------------------------------------

local function sourceE()

	return sourceE_initNode( allocNode( false ) )

end


--============================================================================--

-- @mark variableB

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

local keyBindSource = KSPrivateKeys_makePrivateKey( 'KSReactive.variableB_bindSource' )

]]

--------------------------------------------------------------------------------

local function variableB_setPacked( self, packedValue )
	local setResult = BehaviorNode_setPackedValue( self, packedValue )
	if setResult ~= nil then
		inputQueue_pushNodePackedValue( self, setResult )
		runInputQueue()
	end
	return self
end

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

local function variableB_setWhileBoundError()

	error( "Cannot use :set on a bound variable" )

end

]]

--------------------------------------------------------------------------------

local variableB_initNode = makeInitBehaviorNodeForMethods 'variableB' {

	_receiveData = function()
		error( "Variable nodes should not receive values" )
	end,
	
	_receiveError = function()
		error( "Variable nodes should not receive error values" )
	end,
	
	set = function( self, ... )
		return variableB_setPacked( self, KSRPackedValues_pack( ... ) )
	end,
	
	setError = function( self, message )
		return variableB_setPacked( self, KSRPackedValues_error( message ) )
	end,
	
	--[[ UNUSED
	
	bind = function( self, v )
	
		assert( not isRunning(), "We cannot rebind a variable while the network is running." )
	
		local oldBindSource = self[ keyBindSource ]
		
		if oldBindSource == v then
			return
		end

		-- We also allow for unbinding
	
		if v == nil then
			if oldBindSource then
				removeLink( oldBindSource, self )
				self[ keyBindSource ] = nil
				-- Go back to the standard methods
				self._receiveData = nil
				self.set = nil
			end
			return
		end
		
		-- Otherwise, this needs to be an event node
	
		local vIsEventNode, vIsBehavior = isEventNode( v )
		
		assert( vIsEventNode, "Expected a behavior to be passed to variableB:bind" )
		
		-- Rebind. Note that we don't unbind until we've successfully bound.
		
		-- Check that we've got nodes
		
		-- assertNode( v, "variableB:bind:addLink: predecessor" )
		-- assertNode( self, "variableB:bind:addLink: successor" )
	
		nodes_addLink( v, self )
		
		if oldBindSource then
			removeLink( oldBindSource, self )
		end
		
		self[ keyBindSource ] = v
		
		-- Pick up receive packed from the node methods and block set locally.
		
		self._receiveData = BehaviorNode_setPackedValue
		self._receiveError = BehaviorNode_setPackedValue
		self.set = variableB_setWhileBoundError
		
		-- Handle values from behaviors and event nodes that have already fired
		
		if vIsBehavior then
			variableB_setPacked( BehaviorNode_packedValueNow( v ) )
		end
		
		return self
	
	end
	
	]]

}

--------------------------------------------------------------------------------

local function variableB( ... )

	local node = variableB_initNode( allocNode( false ), KSRPackedValues_pack( ... ) )
	
	return node

end


--============================================================================--

-- @mark mergeE, anyE

--------------------------------------------------------------------------------

local anyE -- forward declarations

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	-- We build a structure in which the output node directly handles the left side
	-- of the merge while the input node handles the right side of the merge by
	-- setting the right side of the output node in response to any events and
	-- queueing the output node to run. We need two nodes because we have two
	-- different actions.
	
	--------------------------------------------------------------------------------
	
	local keyMerge2E_output_leftValue = KSPrivateKeys_makePrivateKey( 'KSReactive.merge2E_output_leftValue' )
	local keyMerge2E_output_rightValue = KSPrivateKeys_makePrivateKey( 'KSReactive.merge2E_output_rightValue' )
	local keyMerge2E_input_outputNode = KSPrivateKeys_makePrivateKey( 'KSReactive.merge2E_input_outputNode' )
	
	--------------------------------------------------------------------------------
	
	local function merge2E_output_receiveNonDone( self, packedValue )
		self[ keyMerge2E_output_leftValue ] = packedValue
		return kScheduleRequestValue
	end
	
	--------------------------------------------------------------------------------
	
	local merge2E_output_initNode = makeInitEventNodeForMethods 'merge2E_output' {
	
		_receiveData = merge2E_output_receiveNonDone,
		_receiveError = merge2E_output_receiveNonDone,
	
		_run = function( self )
			local leftValue = self[ keyMerge2E_output_leftValue ]
			local rightValue = self[ keyMerge2E_output_rightValue ]
			self[ keyMerge2E_output_leftValue ] = false
			self[ keyMerge2E_output_rightValue ] = false
			return leftValue or rightValue
		end
	
	}
	
	--------------------------------------------------------------------------------
	
	local function merge2E_input_receiveNonDone( self, packedValue )
		local outputNode = self[ keyMerge2E_input_outputNode ]
		outputNode[ keyMerge2E_output_rightValue ] = packedValue
		node_addToScheduleQueue( outputNode )
			-- No values ever flow out directly
	end
	
	--------------------------------------------------------------------------------
	
	local merge2E_input_initNode = makeInitEventNodeForMethods 'merge2E_input' {
		
		_receiveData = merge2E_input_receiveNonDone,
		_receiveError = merge2E_input_receiveNonDone,
	
	}
	
	--------------------------------------------------------------------------------
	
	-- Construct a merge for two event streams. Handles nil for the streams and
	-- returns nil if both streams are nil or done.
	
	local function merge2E_imp( leftE, rightE )
	
		-- Nodes that are done have nothing to contribute
	
		if leftE and nodeDAG_nodeIsDone( leftE ) then
			leftE = nil
		end
	
		if rightE and nodeDAG_nodeIsDone( rightE ) then
			rightE = nil
		end
		
		-- If we are missing a stream, then return the other stream
	
		if not leftE then
			return rightE
		end
		
		if not rightE then
			return leftE
		end
		
		-- Construct an appropriate node
	
		local inputNode = allocNode( false )
		local outputNode = allocNode( false )
		
		inputNode[ keyMerge2E_input_outputNode ] = outputNode
		
		outputNode[ keyMerge2E_output_leftValue ] = false
		outputNode[ keyMerge2E_output_rightValue ] = false
		
		merge2E_input_initNode( inputNode, rightE )
		
		merge2E_output_initNode( outputNode, leftE, inputNode )
		
		outputNode:addCloseTarget( inputNode )
		
		return outputNode
	
	end
	
	--------------------------------------------------------------------------------
	
	-- Concatenate a sequence of streams using merge2E_imp in right associative
	-- fashion.
		
	local function mergeE_imp( ... )
	
		local n = select( '#', ... )
		
		if n == 0 then
			return nil
		end
		
		if n == 1 then
			return (...)
		end
		
		return merge2E_imp( (...), mergeE_imp( select( 2, ... ) ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	anyE = function( ... )
	
		local mergedE = mergeE_imp( ... ) or zeroE()
		
		local resultE = mergedE:rawMapE( function() return 'data', true end )
			-- We want to map errors to true as well.
			
		-- Check to see whether mergedE is amongst the input nodes. If not, we
		-- want to close it when the result closes.
		
		local mergedFromInputE = WFVarArgs_findFirst(
				function( x ) return x == mergedE end,
				... )
				
		if not mergedFromInputE then
			resultE:addCloseTarget( mergedE )
		end
		
		return resultE
	
	end
	
	--------------------------------------------------------------------------------
	
	function EventNodeMethods:mergeE( otherStreamE )
	
		return merge2E_imp( self, otherStreamE ) or zeroE()
	
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

local function liftIfNeededB( v )

	if behaviors[ v ] then
		return v
	end
	
	return constantB( v )

end


--============================================================================--

-- @mark callB

--------------------------------------------------------------------------------

local callB -- forward

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyCallB_unpackLimit = KSPrivateKeys_makePrivateKey( 'KSReactive.callB_unpackLimit' )
	
	--------------------------------------------------------------------------------
	
	local function callB_constructFunctionAndArguments( node, ... )
	
		local n = select( '#', ... )
		
		node[ keyCallB_unpackLimit ] = n
		
		for i = 1, n do
			node[ i ] = liftIfNeededB( select( i, ... ) )
		end
	
	end
	
	--------------------------------------------------------------------------------
	
	local function callB_receiveNonDone()
		return kScheduleRequestValue
	end
	
	--------------------------------------------------------------------------------
	
	local function callB_evalToPacked( self )
		return KSRPackedValues_packPCall(
					WFVarArgs_map(
						BehaviorNode_valueNow,
						unpack( self, 1, self[ keyCallB_unpackLimit ] ) ) )
	end
	
	--------------------------------------------------------------------------------
	
	local callB_initNode = makeInitBehaviorNodeForMethods 'callB' {
	
		_receiveData = callB_receiveNonDone,
		_receiveError = callB_receiveNonDone,
		
		_run = function( self )
			return BehaviorNode_setPackedValue( self, callB_evalToPacked( self ) )
		end,
	
	}
	
	--------------------------------------------------------------------------------
	
	callB = function( fn, ... )
	
		local node = allocNode( false )
		
		callB_constructFunctionAndArguments( node, fn, ... )
		
		local initialValue = callB_evalToPacked( node )
		
		return callB_initNode( node, initialValue,
			unpack( node, 1, node[ keyCallB_unpackLimit ] ) )
	
	end
	
	--------------------------------------------------------------------------------
	
end


--============================================================================--

-- @mark argTupleB

--------------------------------------------------------------------------------

local argTupleB_imp -- forward

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local keyArgTupleB_inputs = KSPrivateKeys_makePrivateKey( 'KSReactive.argTupleB_inputs' )
	
	--------------------------------------------------------------------------------
	
	local function argTupleB_constructInputs( ... )
	
		local n = select( '#', ... )
		
		local inputs = { n = n, ... }
		
		for i = 1, n do
			inputs[ i ] = liftIfNeededB( inputs[ i ] )
		end
		
		return inputs
	
	end
	
	--------------------------------------------------------------------------------
	
	local function argTupleB_receiveNonDone()
		return kScheduleRequestValue
	end
	
	--------------------------------------------------------------------------------
	
	local function argTupleB_evalToPacked_expanded( ... )
	
		local hasErrorB = WFVarArgs_findFirst( BehaviorNode_isErrorNow, ... )
		
		if hasErrorB then
			return BehaviorNode_packedValueNow( hasErrorB )
		end
		
		return KSRPackedValues_pack( WFVarArgs_map( BehaviorNode_valueNow, ... ) )
	
	end
	
	--------------------------------------------------------------------------------
	
	local function argTupleB_evalToPacked( self )
	
		local inputs = self[ keyArgTupleB_inputs ]
	
		return argTupleB_evalToPacked_expanded( unpack( inputs, 1, inputs.n ) )
		
	end
	
	--------------------------------------------------------------------------------
	
	local argTupleB_initNode = makeInitBehaviorNodeForMethods 'callB' {
	
		_receiveData = argTupleB_receiveNonDone,
		_receiveError = argTupleB_receiveNonDone,
		
		_run = function( self )
			return BehaviorNode_setPackedValue( self, argTupleB_evalToPacked( self ) )
		end,
	
	}
	
	--------------------------------------------------------------------------------
	
	-- Find a non-constant behavior
	
	local function isNonConstantB( x )
		return behaviors[ x ] and not nodeDAG_nodeIsDone( x )
	end
	
	--------------------------------------------------------------------------------
	
	local function isConstantErrorB( x )
		return behaviors[ x ] and nodeDAG_nodeIsDone( x ) and BehaviorNode_isErrorNow( x )
	end
	
	--------------------------------------------------------------------------------
	
	local function optBehavior_valueNow( x )
	
		if behaviors[ x ] then
			return BehaviorNode_valueNow( x )
		else
			return x
		end
	
	end
	
	--------------------------------------------------------------------------------
	
	local function optBehavior_close( x )
	
		if x and behaviors[ x ] then
			x:close()
		end
	
	end
	
	--------------------------------------------------------------------------------
	
	argTupleB_imp = function( closeInputs, ... )
	
		local n = select( '#', ... )
		
		if n == 0 then
			return constantB()
		end
		
		if n == 1 then
			local x = ( ... )
			if behaviors[ x ] then
				return x:relayB( closeInputs )
			else
				return constantB( x )
			end
		end
		
		-- Check for a constant error.
		
		local constantErrorB = WFVarArgs_findFirst( isConstantErrorB, ... )
		
		if constantErrorB then
			return constantErrorB
		end
	
		-- Check to find something non-constant.
	
		local nonConstantB = WFVarArgs_findFirst( isNonConstantB, ... )
		
		if nonConstantB then
		
			local node = allocNode( false )
			
			local inputs = argTupleB_constructInputs( ... )
			
			node[ keyArgTupleB_inputs ] = inputs
			
			local initialValue = argTupleB_evalToPacked( node )
			
			local resultB = argTupleB_initNode( node, initialValue, unpack( inputs, 1, inputs.n ) )
			
			if closeInputs then
				for i = 1, inputs.n do
					resultB:addCloseTarget( inputs[ i ] )
				end
			end
			
			return resultB
		
		end
		
		-- Everything is constant and not an error. Just return a constantB node.
		
		local resultB = constantB( WFVarArgs_map( optBehavior_valueNow, ... ) )
		
		if closeInputs then
			WFVarArgs_map( optBehavior_close, ... )
		end
		
		return resultB
	
	end
	
	--------------------------------------------------------------------------------
	
end

--------------------------------------------------------------------------------

local function argTupleB( ... )
	return argTupleB_imp( false, ... )
end


--============================================================================--

-- @mark dataCallB

--------------------------------------------------------------------------------

local function dataCallB( fn, ... )

	assert( fn, "No function provided." )
	
	if behaviors[ fn ] then
		local argsB = argTupleB( fn, ... )
		return argsB:closingMapSwitchB( function( f, ... )
			return argTupleB_imp( true, f( ... ) )
		end )
	else
		WFArguments_check( 'callable' )
		local argsB = argTupleB( ... )
		return argsB:closingMapSwitchB( function( ... )
			return argTupleB_imp( true, fn( ... ) )
		end )
	end

end


--============================================================================--

-- @mark liftB

--------------------------------------------------------------------------------

local liftB, lift

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	local liftBCache = makeFullyWeakTable()
		-- functions to their lifted forms. This makes it safer to lift repeatedly.
	
	--------------------------------------------------------------------------------
	
	liftB = function( fn )
	
		local cached = liftBCache[ fn ]
		
		if cached then
			return cached
		end
		
		local result = function( ... )
			return dataCallB( fn, ... )
		end
		
		liftBCache[ fn ] = result
		liftBCache[ result ] = result -- No sense in lifting twice
		
		return result
	
	end
	
	--------------------------------------------------------------------------------
	
	local string_sub = string.sub
	
	--------------------------------------------------------------------------------
	
	local function getLifter( k, lifters )
	
		if type( k ) ~= 'string' or k == '' then
			return nil
		end
		
		return lifters[ string_sub( k, -1 ) ]
	
	end
	
	--------------------------------------------------------------------------------
	
	local function liftTable( t, lifters )
	
		local result = { }
		
		for k, v in pairs( t ) do
		
			local lifter = getLifter( k, lifters )
			
			if lifter then
				v = lifter( v )
			end
			
			result[ k ] = v
		
		end
		
		return result
	
	end
	
	--------------------------------------------------------------------------------
	
	local functionLifters = {
		[ 'B' ] = liftB
	}
	
	--------------------------------------------------------------------------------
	
	lift = function( t )
	
		return liftTable( t, functionLifters )
	
	end
	
	--------------------------------------------------------------------------------
	
end


--============================================================================--

-- @mark tableB

--------------------------------------------------------------------------------

local tableB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	-- We use special designators for inputs to the template. This is a collection
	-- we build as need be.
	
	local inputObjects = { }
	
	-- We maintain a mapping from input objects to the index they represent
	
	local inputObjectToIndex = { }
	
	--------------------------------------------------------------------------------
	
	local function getInputObject( idx )
	
		local obj = inputObjects[ idx ]
		
		if not obj then
		
			obj = { }
			
			inputObjects[ idx ] = obj
			
			inputObjectToIndex[ obj ] = idx
		
		end
		
		return obj
	
	end
	
	--------------------------------------------------------------------------------
	
	local isSource = isBehavior
	
	--------------------------------------------------------------------------------
	
	local KSPromises_isPromise_lazy = nil
	
	--------------------------------------------------------------------------------
	
	local function checkValidTableForTemplate( v )
	
		local is_event_node, is_behavior = isEventNode( v )
	
		if is_event_node and not is_behavior then
			error "We do not support event nodes in templates for tableB. holdB?"
		end
		
		if KSPromises_isPromise_lazy( v ) then
			error "We do not support promises in templates for tableB. untilResolvedB?"
		end
		
	end
	
	--------------------------------------------------------------------------------
	
	local function convertToTemplate_imp( t, inputList, inputMap )
	
		local tt = type( t )
		
		if tt ~= 'table' then
			return t
		end
		
		checkValidTableForTemplate( t )
		
		if isSource( t ) then
			local idx = inputMap[ t ]
			if not idx then
				idx = #inputList + 1
				inputList[ idx ] = t
				inputMap[ t ] = idx
			end
			return getInputObject( idx )
		end
		
		local result = { }
		
		for k, v in pairs( t ) do
			result[ k ] = convertToTemplate_imp( v, inputList, inputMap )
		end
		
		return result
	
	end

	--------------------------------------------------------------------------------
	
	local function convertToTemplate( t )
	
		if KSPromises_isPromise_lazy == nil then
			KSPromises_isPromise_lazy = require 'KSPromises' . isPromise
		end
		
		local inputList = { }
		local inputMap = { }
		
		local template = convertToTemplate_imp( t, inputList, inputMap )
		
		return template, inputList
	
	end
	
	--------------------------------------------------------------------------------
	
	local function instantiateTemplate( t, inputs )
	
		local tt = type( t )
		
		if tt ~= 'table' then
			return t
		end
		
		local idx = inputObjectToIndex[ t ]
		
		if idx then
			return inputs[ idx ]
		end
		
		local result = { }
		
		for k, v in pairs( t ) do
			result[ k ] = instantiateTemplate( v, inputs )
		end
		
		return result
	
	end
	
	--------------------------------------------------------------------------------
	
	tableB = function( t )
	
		local template, inputList = convertToTemplate( t )
		
		-- OPTIMIZATION: We could special case specific lengths for the source list
		
		local function eval( ... )
		
			return instantiateTemplate( template, { ... } )
		
		end

		local resultB = dataCallB( eval, unpack( inputList ) )
		
		for _, inputB in pairs( inputList ) do
			resultB:addCloseTarget( inputB )
		end
		
		return resultB
	
	end
	
	--------------------------------------------------------------------------------

end


--============================================================================--

-- @mark zipMapSwitchB

--------------------------------------------------------------------------------

do

	--------------------------------------------------------------------------------
	
	function BehaviorNodeMethods:zipMapSwitchB( fn )
		return argTupleB( self, self:mapSwitchB( fn ) )
	end

	--------------------------------------------------------------------------------

end


--============================================================================--

local KSReactiveTimer_laterB = lazyFunction( 'KSReactiveTimer', 'laterB' )
local KSReactiveTimer_timerE = lazyFunction( 'KSReactiveTimer', 'timerE' )
local KSReactiveTimer_timedSeqB = lazyFunction( 'KSReactiveTimer', 'timedSeqB' )
local KSReactiveTimer_falseThenTrueB = lazyFunction( 'KSReactiveTimer', 'falseThenTrueB' )
local KSReactiveTimer_trueThenFalseB = lazyFunction( 'KSReactiveTimer', 'trueThenFalseB' )
local KSReactiveTimer_calmE = lazyFunction( 'KSReactiveTimer', 'calmE' )
local KSReactiveTimer_calmB = lazyFunction( 'KSReactiveTimer', 'calmB' )


--============================================================================--

-- @mark calmE, calmB

--------------------------------------------------------------------------------

function EventNodeMethods:calmE( interval, startTimer )
	return KSReactiveTimer_calmE( self, interval, startTimer )
end

--------------------------------------------------------------------------------

function BehaviorNodeMethods:calmB( interval, startTimer )
	return KSReactiveTimer_calmB( self, interval, startTimer )
end


--============================================================================--

return {
	sourceE = sourceE,
	variableB = variableB,
	constantB = constantB,
	constantErrorB = constantErrorB,
	zeroE = zeroE,
	anyE = anyE,
	callB = callB,
	argTupleB = argTupleB,
	dataCallB = dataCallB,
	tableB = tableB,
	liftB = liftB,
	lift = lift,
	-- NOT USED: DISABLED: runIdle = runIdle,
	isEventNode = isEventNode,
	isBehavior = isBehavior,
	timerE = KSReactiveTimer_timerE,
	timedSeqB = KSReactiveTimer_timedSeqB,
	falseThenTrueB = KSReactiveTimer_falseThenTrueB,
	trueThenFalseB = KSReactiveTimer_trueThenFalseB,
	laterB = KSReactiveTimer_laterB,
	makeWeakEventTable = lazyFunction( 'KSWeakEventTable', 'makeWeakEventTable' )
}
