--[[----------------------------------------------------------------------------

KSReactiveBinding.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2011-02-15

@reviewed erwright 2011-03-10

------------------------------------------------------------------------------]]

--[[

	KSReactiveBinding: Binding Support for Reactive Programming
	
	function KSReactiveBinding.wrapSetFieldForReactiveBinding( setField )
		Return a function with signature function( t, k, v ) which will
		recognize reactive values v as needing to be handled with a binding
		rather than a simple assignment. Often used to wrap __newindex
		functions. The resulting function will call through to setField.
		This will only bind sets to string keys.
			
	function KSReactiveBinding.addReactiveBindingSupportMethods( methodTable )
		Adjust the __newindex entry to do reactive binding. Beware that
		for this to work, the underlying table should do something to
		encode the keys or submerge the set values so that the proxy logic
		always gets a chance to fire.
	
]]

--============================================================================--

--@mark KSReactiveBinding namespace

--[[!
	\class 	KSReactiveBinding

	\brief Provides binding support for reactive programming.

]]--

-- local assertf = require 'KSLuaUtils'.assertf

local type = type
local rawset = rawset
local rawget = rawget
local assert = assert

local debug_print = debug.print

local KSPrivateKeys = require 'KSPrivateKeys'
local WFCloseSupport = require 'WFCloseSupport'

local KSReactive = require 'KSReactive'

local KSReactive_isEventNode = assert( KSReactive.isEventNode )

local KSPromises = require 'KSPromises'

local KSPromises_isPromise = assert( KSPromises.isPromise )

local WFCloseSupport_hasBeenOrWillBeClosed = assert( WFCloseSupport.hasBeenOrWillBeClosed )

local WFIdle = require 'WFIdle'

local WFCurrentThread = require 'WFCurrentThread'

--============================================================================--

local keyBindings = KSPrivateKeys.makePrivateKey( 'KSReactiveBinding.bindings' )

--------------------------------------------------------------------------------

local function keySupportsReactiveBinding( k )

	if type( k ) ~= 'string' then
		return false
	end
	
	return true

end

--------------------------------------------------------------------------------

local gcRequestQueued = false

--------------------------------------------------------------------------------

local function requestGC()

	gcRequestQueued = false

	WFCurrentThread.requestIdleGC()

end

--------------------------------------------------------------------------------

local function didUnbindValue( value, depth )

	if gcRequestQueued then
		return
	end

	-- nil and false are common enough as values we handle them without even
	-- doing the type test

	if not value then
		return
	end

	local tvalue = type( value )
	
	-- If we unbind userdata, we let it know

	if tvalue == 'userdata' then
	
		WFIdle.onIdleDo( requestGC )
		
		gcRequestQueued = true
		
	-- For tables, we recurse down the values side
		
	elseif tvalue == 'table' then
	
		depth = depth or 1
		
		if depth < 5 then
			local nextDepth = depth + 1
			for k, v in pairs( value ) do
				didUnbindValue( v, nextDepth )
			end
		end
	
	end

end

--------------------------------------------------------------------------------

--[[!
	Creates a function with signature function( t, k, v ) that will
	recognize reactive values v as needing to be handled with a binding
	rather than a simple assignment. This will only bind sets to string keys.
	
	Often used to wrap \c __newindex functions.
	
	\param setField The created function calls through to this.
	\param testNewIndexHook (internal: controls whether we look for rebinds)

	\return The new function.
		
]]--

local function wrapSetFieldForReactiveBinding( setField, testNewIndexHook )

	-- Let's not allow setField to get called with closed values.

	local function safeSetField( self, k, v )

		-- Note that hasBeenOrWillBeClosed also reports true for the value nil,
		-- but this is a perfectly reasonable value to use when setting a field.

		if v and WFCloseSupport_hasBeenOrWillBeClosed( v ) then
			debug_print( "Attempt to set field to a closed value: " .. debug.traceback() )
			return
		end

		setField( self, k, v )

	end

	return function( self, k, v )
	
		if not keySupportsReactiveBinding( k ) then
			safeSetField( self, k, v )
			return
		end
	
		local bindings = rawget( self, keyBindings )
		
		-- No matter what happens, we don't want an old binding.
		
		local oldBinding = bindings and bindings[ k ]
		
		if oldBinding then
			oldBinding:close()
				-- Old binding entry cleared by an action on the binding itself.
		end
		
		local vIsClosed = v and WFCloseSupport_hasBeenOrWillBeClosed( v )
			-- Note that hasBeenOrWillBeClosed also reports true for the value nil,
			-- but this is a perfectly reasonable value to use when setting a field.
		
		-- If we bind to a closed value, we probably have an error and hence we log it.
		-- Even so, we will attempt to handle the binding as far as we are able.
		
		if vIsClosed then
			debug_print( "Attempt to bind or set to a closed value: " .. debug.traceback() )
		end
		
		if KSPromises_isPromise( v ) then

			assert( ( not testNewIndexHook ) or ( rawget( self, k ) == nil ),
				"We won't be able to catch re-bindings if the setField call set the public field" )

			if not bindings then
				bindings = { }
				rawset( self, keyBindings, bindings )
			end
			
			-- Note that we are able to use the logic in attachTo to manage the entry in
			-- the bindings table.

			v:whenResolvedDo( function( status, resolvedValue )
				if status == 'data' then
					safeSetField( self, k, resolvedValue )
				else
					debug_print( "Binding received promise result with status: %s. No action taken.",
						tostring( status ) )
				end
			end ):attachTo( self, bindings, k )

			return

		end
		
		local isEventNode, isBehavior = KSReactive_isEventNode( v )
		
		if isEventNode then
		
			local prevValue = nil
		
			if isBehavior then
				prevValue = v:valueNow()
				safeSetField( self, k, v:valueNow() )
			end
			
			-- Even closed behaviors have values, so we wanted to do the set. But having done
			-- so, we want to avoid actually establishing a binding. Note that we already
			-- logged that this was closed above.
			
			if vIsClosed then
				return
			end
		
			assert( ( not testNewIndexHook ) or ( rawget( self, k ) == nil ),
				"We won't be able to catch re-bindings if the setField call set the public field" )
		
			if not bindings then
				bindings = { }
				rawset( self, keyBindings, bindings )
			end
			
			-- Events cause us to update the property. We will also notify the old value that
			-- it is being unbound.
			
			-- GRUNGE: The following is a horrendous abuse of mapE but we don't want the
			-- strong successor logic from onDataE since it will result in keeping the binding
			-- target alive unnecessarily. This would be a problem if having strong successors
			-- started having meaning in the system but that's only likely to happen when
			-- we shift over to something more like the actions system for promises.
		
			local binding = v:mapE( function( vv )
				safeSetField( self, k, vv )
				didUnbindValue( prevValue )
				prevValue = vv
				return vv
			end )
			
			binding:addCloseTarget( v )
				-- FIXME [mhamburg 2011-02-24]: This is a total hack and is
				-- unsafe if we bind multiple items to the same event stream!
				
			-- Closing self closes the binding
			
			self:addCloseTarget( binding )
			
			-- Track the binding so we can find it again
			
			bindings[ k ] = binding
			
			-- Closing the binding stops the tracking.
			
			binding:addCloseAction( function()
				if bindings[ k ] == binding then
					bindings[ k ] = false
				end
				didUnbindValue( prevValue )
			end )

			return
			
		end
		
		-- Refuse to set to closed values. This recapitulates the logic
		-- from safeSetField, but we have already dumped the log once.
				
		if not vIsClosed then
			setField( self, k, v )
		end
		
	end

end

--============================================================================--

--[[! Adjusts the __newindex entry of a method table to do reactive binding.

	\param methodTable  The method table.

	\warning For this to work, the underlying table should do something to
		encode the keys or submerge the set values so that the proxy logic
		always gets a chance to fire.
]]--

local function addReactiveBindingSupportMethods( methodTable )

	local oldNewIndex = methodTable.__newindex
	
	assert( oldNewIndex,
		"We need to have a __newindex metamethod to reliably intercept subsequent sets." )
	
	methodTable.__newindex = wrapSetFieldForReactiveBinding( oldNewIndex, true )
	
	-- FIXME [mhamburg 2011-02-15]: It might be good to handle __init as well but
	-- I think we will often get the right support just from having init end
	-- up going through __newindex.

end

--============================================================================--

return {
	wrapSetFieldForReactiveBinding = wrapSetFieldForReactiveBinding,
	addReactiveBindingSupportMethods = addReactiveBindingSupportMethods,
}
