--[[----------------------------------------------------------------------------

KSReactiveTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2011-01-11

------------------------------------------------------------------------------]]

local KSReactiveTest = { }

--------------------------------------------------------------------------------

local assert = assert

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-05-04]: Test that relayE and relayB do not unnaturally
-- prolong the lifespan of their results when passed the value true. There is
-- some evidence that they do.

--------------------------------------------------------------------------------

-- Confirm that t is an array containing the provided sequence of values

local function testSequence( t, ... )

	local count = select( '#', ... )
	
	if #t ~= count then
		return false, "Table length should match sequence length: " .. debug.traceback()
	end
	
	for i = 1, count do
		if t[ i ] ~= select( i, ... ) then
			return false, "Elements mismatch"
		end
	end
	
	return true

end

--------------------------------------------------------------------------------

function KSReactiveTest.testSourceE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	
	node:onDataE( function( v ) t[ #t + 1 ] = v end )
	
	node:post( 1 )
	node:post( 2 )
	node:post( 3 )
	
	assert( testSequence( t, 1, 2, 3 ), "Expected to accumulate 1, 2, 3" )
		
	node = KSReactive_sourceE()
	
	t = { }
	
	node:mapE( function( x ) return x * x end ):filterE( function( x ) return 10 < x end ):onDataE( function( v ) t[ #t + 1 ] = v end )
	
	node:post( 1 )
	node:post( 2 )
	node:post( 3 )
	node:post( 4 )
	node:post( 5 )
	
	assert( testSequence( t, 16, 25 ), "Expected to accumulate 16, 25" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testVariableB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_variableB = KSReactive.variableB

	local node = KSReactive_variableB( 0 )
	
	assert( node:valueNow() == 0, "Initial value should be zero" )
	
	local t = { }
	
	node:onDataE( function( v ) t[ #t + 1 ] = v end )
	
	node:set( 1 )
	node:set( 2 )
	node:set( 2 ) -- Duplicate should be ignored
	node:set( 3 )
	
	assert( testSequence( t, 1, 2, 3 ), "Expected to accumulate 1, 2, 3" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testMergeE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	local node2 = KSReactive_sourceE()
	
	local node3 = node1:mergeE( node2 )
	
	local t = { }
	
	node3:onDataE( function( v ) t[ #t + 1 ] = v end )
	
	node1:post( 1 )
	node1:post( 2 )
	node2:post( 3 )
	node1:post( 4 )
	
	assert( testSequence( t, 1, 2, 3, 4 ), "Expected to accumulate 1, 2, 3, 4" )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testFilterB()

	local KSReactive = require 'KSReactive'

	local KSReactive_variableB = KSReactive.variableB

	local nodeB = KSReactive_variableB( 0 )
	
	local oddsOnlyB = nodeB:filterB( function( x ) return x % 2 == 1 end, "Foo" )
	
	local oddsOnlyB2 = nodeB:filterB( function( x ) return x % 2 == 1 end )
	
	local evensOnlyB = nodeB:filterB( function( x ) return x % 2 == 0 end, "Foo" )
	
	assert( oddsOnlyB:valueNow() == "Foo", "Should have filtered out initial 0." )
	
	assert( oddsOnlyB2:valueNow() == nil, "Should have filtered out initial 0 and returned nil." )
	
	assert( evensOnlyB:valueNow() == 0, "Should not have filtered out initial 0." )
	
	nodeB:set( 1 )
	
	assert( oddsOnlyB:valueNow() == 1, "We should have updated for the value 1." )
	assert( oddsOnlyB2:valueNow() == 1, "We should also have updated for the value 1." )
	assert( evensOnlyB:valueNow() == 0, "We should not have updated for the value 1." )

	nodeB:set( 2 )
	
	assert( oddsOnlyB:valueNow() == 1, "We should not have updated for the value 2." )
	assert( oddsOnlyB2:valueNow() == 1, "We should also not have updated for the value 2." )
	assert( evensOnlyB:valueNow() == 2, "We should have updated for the value 2." )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testFilterBPassesErrors()

	local KSReactive = require 'KSReactive'

	local KSReactive_variableB = KSReactive.variableB

	local nodeB = KSReactive_variableB( 0 )
	
	nodeB:setError( "This is an error" )
	
	local oddsOnlyB = nodeB:filterB( function( x ) return x % 2 == 1 end, "Foo" )
	
	assert( oddsOnlyB:isErrorNow(), "filterB should have passed the error through" )
	
	nodeB:set( 0 )
	
	assert( oddsOnlyB:isErrorNow(), "filterB should not have updated for 0" )
	
	nodeB:set( 1 )
	
	assert( oddsOnlyB:valueNow() == 1, "filterB should have updated for 1" )
	
	nodeB:setError( "Back to error" )
	
	assert( oddsOnlyB:isErrorNow(), "filterB should be back to being an error" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testSkipErrorsB()

	local KSReactive = require 'KSReactive'

	local KSReactive_variableB = KSReactive.variableB

	local nodeB = KSReactive_variableB( 0 )
	
	nodeB:setError( "This is an error" )
	
	local skippedB = nodeB:skipErrorsB( "Started with error" )
	
	assert( skippedB:valueNow() == "Started with error", "Should have transformed initial error." )
	
	nodeB:set( 0 )
	
	assert( skippedB:valueNow() == 0, "Should have updated for 0." )
	
	nodeB:set( 1 )
	
	assert( skippedB:valueNow() == 1, "Should have updated for 1." )
	
	nodeB:setError( "Another error" )
	
	assert( skippedB:valueNow() == 1, "Should not have updated for error." )
	
	nodeB:set( 2 )
	
	assert( skippedB:valueNow() == 2, "Should not have updated for 2." )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testAnyE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	local node2 = KSReactive_sourceE()
	
	local node3 = KSReactive.anyE( node1, node2 )
	
	local t = { }
	
	node3:onDataE( function( v ) t[ #t + 1 ] = v end )
	
	node1:post( 1 )
	node1:post( 2 )
	node2:post( 3 )
	node1:post( 4 )
	
	assert( testSequence( t, true, true, true, true ), "Expected to accumulate 1, 2, 3, 4" )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testMergeEMultipath()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	local node2 = KSReactive_sourceE()
	
	local node3 = node1:mapE( function( x ) return x * x end )
	
	local nodeM1 = node3:mergeE( node1 ):mergeE( node2 )
	
	local t1 = { }
	
	nodeM1:onDataE( function( v ) t1[ #t1 + 1 ] = v end )
	
	local nodeM2 = node1:mergeE( node2 ):mergeE( node3 )
	
	local t2 = { }
	
	nodeM2:onDataE( function( v ) t2[ #t2 + 1 ] = v end )
	
	local isDone = false
	
	nodeM2:onDoneE( function() isDone = true end )
	
	node1:post( 1 )
	node1:post( 2 )
	node2:post( 3 )
	node1:post( 4 )
	
	assert( testSequence( t1, 1, 4, 3, 16 ), "Expected to accumulate 1, 4, 3, 16" )

	assert( testSequence( t2, 1, 2, 3, 4 ), "Expected to accumulate 1, 2, 3, 4" )
	
	node1:close()
	node2:close()
	
	assert( isDone, "We should have closed the merged stream." )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testAnyEMultipath()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	local node2 = KSReactive_sourceE()
	
	local node3 = node1:mapE( function( x ) return x * x end )
	
	local nodeM1 = KSReactive.anyE( node3, node1, node2 )
	
	local t1 = { }
	
	nodeM1:onDataE( function( v ) t1[ #t1 + 1 ] = v end )
	
	local nodeM2 = KSReactive.anyE( node1, node2, node3 )
	
	local t2 = { }
	
	nodeM2:onDataE( function( v ) t2[ #t2 + 1 ] = v end )
	
	local isDone = false
	
	nodeM2:onDoneE( function() isDone = true end )
	
	node1:post( 1 )
	node1:post( 2 )
	node2:post( 3 )
	node1:post( 4 )
	
	assert( testSequence( t1, true, true, true, true ), "Expected to accumulate 1, 4, 3, 16" )

	assert( testSequence( t2, true, true, true, true ), "Expected to accumulate 1, 2, 3, 4" )
	
	node1:close()
	node2:close()
	
	assert( isDone, "We should have closed the merged stream." )

end

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeWhileE_1()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeWhileE = node:takeWhileE( function( x ) return x % 2 == 1 end )
	
	takeWhileE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:post( 4 )
	
	assert( testSequence( t, 1, 3 ), "Sequence should have been 1, 3" )
	assert( isDone, "Stream should now be closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3 ), "Sequence should still have been 1, 3" )
	
	isDone = false
	
	node:close()
	
	assert( not isDone, "onDoneE should only fire once" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeWhileE_2()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeWhileE = node:takeWhileE( function( x ) return x % 2 == 1 end )
	
	takeWhileE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:postError( 'foo' )
	
	assert( testSequence( t, 1, 3 ), "Sequence should have been 1, 3" )
	assert( isDone, "Stream should now be closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3 ), "Sequence should still have been 1, 3" )
	
	isDone = false
	
	node:close()
	
	assert( not isDone, "onDoneE should only fire once" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeWhileE_3()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeWhileE = node:takeWhileE( function( x ) return x % 2 == 1 end )
	
	takeWhileE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeWhileE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:close()
	
	assert( testSequence( t, 1, 3 ), "Sequence should have been 1, 3" )
	assert( isDone, "Stream should now be closed" )
	
end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeUntilE_1()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeUntilE = node:takeUntilE( function( x ) return x % 2 == 0 end )
	
	takeUntilE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeUntilE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeUntilE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:post( 4 )
	
	assert( testSequence( t, 1, 3 ), "Sequence should have been 1, 3" )
	assert( isDone, "Stream should now be closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3 ), "Sequence should still have been 1, 3" )
	
	isDone = false
	
	node:close()
	
	assert( not isDone, "onDoneE should only fire once" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeUntilE_2()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeUntilE = node:takeUntilE( function( x ) return x % 2 == 0 end )
	
	takeUntilE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeUntilE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeUntilE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:postError( 'foo' )
	
	assert( testSequence( t, 1, 3, 'foo' ), "Sequence should have been 1, 3, 'foo'" )
	assert( not isDone, "Stream should not have closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3, 'foo', 7 ), "Sequence should now be 1, 3, 'foo', 7" )
	
	node:close()
	
	assert( isDone, "Stream should now be closed" )

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeThroughE_1()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeThroughE = node:takeThroughE( function( x ) return x % 2 == 0 end )
	
	takeThroughE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeThroughE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeThroughE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:post( 4 )
	
	assert( testSequence( t, 1, 3, 4 ), "Sequence should have been 1, 3, 4" )
	assert( isDone, "Stream should now be closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3, 4 ), "Sequence should still have been 1, 3, 4" )
	
	isDone = false
	
	node:close()
	
	assert( not isDone, "onDoneE should only fire once" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeThroughE_2()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	local isDone = false
	
	local takeThroughE = node:takeThroughE( function( x ) return x % 2 == 0 end )
	
	takeThroughE:onDataE( function( v ) t[ #t + 1 ] = v end )
	takeThroughE:onErrorE( function( v ) t[ #t + 1 ] = v end )
	takeThroughE:onDoneE( function() isDone = true end )
	
	node:post( 1 )
	node:post( 3 )
	node:postError( 'foo' )
	
	assert( testSequence( t, 1, 3, 'foo' ), "Sequence should have been 1, 3, 'foo'" )
	assert( not isDone, "Stream should not have closed" )
	
	node:post( 7 )
	
	assert( testSequence( t, 1, 3, 'foo', 7 ), "Sequence should now be 1, 3, 'foo', 7" )
	
	node:close()
	
	assert( isDone, "Stream should now be closed" )

end

--------------------------------------------------------------------------------

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

function KSReactiveTest.testDebugHook()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	local node2 = KSReactive_sourceE()
	
	local node3 = node1:mapE( function( x ) return x * x end )
	
	local node4 = KSReactive_sourceE()
	
	local nodeM1 = KSReactive.anyE( node3, node1, node2, node4:mapE( function( x ) return x + x end ) )
	
	local t = { }
	
	nodeM1:setDebugHook( function( str ) t[ #t + 1 ] = str end, "Unit Test Merge" )
	
	node1:post( 1 )
	node1:post( 2 )
	node2:post( 3 )
	node1:post( 4 )
	
	assert( #t == 4, "We should have four message" )
	
	node4:postError( 'foo' )
	
	assert( #t == 5, "We should have five entries" )
	
	node1:close()
	node2:close()
	node4:close()

	-- table.foreachi( t, debug.print )
	
	assert( #t == 6, "We should have six entries" )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testZeroE()

	local KSReactive = require 'KSReactive'
	
	local zeroNode = KSReactive.zeroE()
	
	local isDone = false
	
	zeroNode:onDoneE( function() isDone = true end )
	
	assert( isDone, "We should have triggered the done action for the new zero node." )

end

---------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

function KSReactiveTest.testWhenDoneE()

	local KSReactive = require 'KSReactive'
	
	local sourceE = KSReactive.sourceE()
	
	local whenDoneE = sourceE:whenDoneE()
	
	local whenDoneGotData = false
	local whenDoneIsDone = false
	
	whenDoneE:onDataE( function( x ) whenDoneGotData = x end )
	whenDoneE:onDoneE( function() whenDoneIsDone = true end )
	
	sourceE:post( 4 )
	sourceE:post( 'foo' )
	sourceE:post()
	
	sourceE:postError( 'baz' )
	
	assert( whenDoneGotData == false, "Should have no data yet" )
	assert( whenDoneIsDone == false, "Should not be done yet" )
	
	sourceE:close()
	
	assert( whenDoneGotData == true, "Should have received true" )
	assert( whenDoneIsDone == true, "Should now be done" )

end

]] -- NOT USED: DISABLED

---------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

function KSReactiveTest.testSkipFirstE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local t = { }
	
	node:skipFirstE():onDataE( function( v ) t[ #t + 1 ] = v end  )
	
	node:post( 1 )
	node:post( 2 )
	node:post( 3 )
	node:post( 4 )
	
	assert( testSequence( t, 2, 3, 4 ), "Expected to accumulate 2, 3, 4" )

end

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

-- Test that hooking up a node after another node has sent still gets the
-- event that was sent.

function KSReactiveTest.testIntracycleResend()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local result = 0
	
	node:onDataE( function()
		node:mapE( function( y ) return y * y end )
				:onDataE( function( v )
					result = result + v
				end )
	end )
	
	node:post( 4 )
	
	assert( result == 16, "We should still receive the 4 and turn it into 16. Got " .. tostring( result ) )
	
	local result2 = nil
	
	node:onDataE( function( v ) result2 = v end )
	
	assert( result2 == nil, "We should not have propagated the previous value out of cycle." )
	
	node:post( 5 )
	
	assert( result2 == 5, "We should now have captured the 5." )
	
	assert( result == 66, "Expected 4 * 4 + 5 * 5 + 5 * 5 = 66, got " .. tostring( result ) )

end

--------------------------------------------------------------------------------

--[[ SwitchE is no more. See mapSwitchE.

function KSReactiveTest.testSwitchE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local nodeES = KSReactive_sourceE()
	
	local node1 = KSReactive_sourceE()
	
	local node2 = KSReactive_sourceE()
	
	local t = { }
	
	nodeES:switchE():onDataE( function( x ) t[ #t + 1 ] = x end )
	
	assert( testSequence( t ), "t should read as empty" )
	
	nodeES:post( node1 )
	
	assert( testSequence( t ), "t should read as empty" )
	
	node1:post( 'A' )
	
	assert( testSequence( t, 'A' ), "t should contain A" )
	
	node2:post( 'AA' )
	
	assert( testSequence( t, 'A' ), "t should still contain A" )
	
	node1:post( 'B' )
	
	assert( testSequence( t, 'A', 'B' ), "t should contain A, B" )
	
	nodeES:post( node2 )
	
	assert( testSequence( t, 'A', 'B' ), "t should still contain A, B" )
	
	node1:post( 'C' )
	
	assert( testSequence( t, 'A', 'B' ), "t should even still contain A, B" )
	
	node2:post( 'CC' )
	
	assert( testSequence( t, 'A', 'B', 'CC' ), "t should even still contain A, B, CC" )
	
	node2:post( 'DD' )
	
	assert( testSequence( t, 'A', 'B', 'CC', 'DD' ), "t should even still contain A, B, CC, DD" )

end

]]

--------------------------------------------------------------------------------

--[[ NOT USED: DISABLED

function KSReactiveTest.testDelayE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	
	local node2 = node1:mapE( function( x ) return x * x end ):delayE()
	
	local node3 = node2:delayE()
	
	local node4 = node1:mergeE( node2 ):mergeE( node3 )
	
	local t = { }
	
	node4:onDataE( function( x ) t[ #t + 1 ] = x end )
	
	node1:post( 4 )
	node1:post( 5 )
	
	assert( testSequence( t, 4, 16, 16, 5, 25, 25 ) )

end

]] -- NOT USED: DISABLED

--------------------------------------------------------------------------------

function KSReactiveTest.testCloseCallsTargets()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node1 = KSReactive_sourceE()
	
	local node2 = node1:mapE( function( x ) return x * x end )
	
	local node1Closed = false
	local node2Closed = false
	
	node1:addCloseAction( function() node1Closed = true end )
	node2:addCloseAction( function() node2Closed = true end )
	
	node1:close()
	
	assert( node1Closed, "Node 1 did not close" )
	assert( node2Closed, "Node 2 did not close" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testHoldB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	local node = KSReactive_sourceE()
	
	local behavior = node:holdB( "foo" )
	
	local count = 0
	
	behavior:onDataE( function() count = count + 1 end )
	
	assert( behavior:valueNow() == "foo", "Initial value should be foo" )
	
	node:post( "baz" )
	
	assert( behavior:valueNow() == "baz", "Behavior should have updated" )
	
	node:post( 1, 2, 3 )
	
	assert( testSequence( { behavior:valueNow() }, 1, 2, 3 ), "Multi-value behaviors should work" )
	
	node:post( "fraggle" )
	node:post( "fraggle" )
	node:post( "frizz" )
	node:post( "fraggle" )
	
	assert( count == 5, "Behavior should have fired five times." )
	
	node:close()
	
	assert( behavior:valueNow() == "fraggle", "valueNow should survive a close" )
	
	assert( count == 5, "count should not be affected by close" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testRawMapE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local node = KSReactive_sourceE()
	
	local rawMapE1 = node:rawMapE( function( status, x )
		if status == 'error' then
			if x == 'foo' then
				return 'same'
			else
				return 'error', 'baz'
			end
		else
			return 'data', x + 1
		end
	end )
	
	local t1 = { }
	local isDone1 = false
	
	rawMapE1:onDataE( function( v ) t1[ #t1 + 1 ] = v end )
	rawMapE1:onErrorE( function( v ) t1[ #t1 + 1 ] = v end )
	rawMapE1:onDoneE( function() isDone1 = true end )
	
	local rawMapE2 = node:rawMapE( function( ... ) return 'data', ( ... ) end )
	
	local t2 = { }
	local isDone2 = false

	rawMapE2:onDataE( function( v ) t2[ #t2 + 1 ] = v end )
	rawMapE2:onErrorE( function( v ) t2[ #t2 + 1 ] = v end )
	rawMapE2:onDoneE( function() isDone2 = true end )
	
	node:post( 1 )
	node:post( 2 )
	node:postError( 'skronk' )
	node:post( 4 )
	node:postError( 'foo' )
	node:postError( 'baz' )
	node:post( 10 )
	node:close()
	
	assert( testSequence( t1, 2, 3, 'baz', 5, 'foo', 'baz', 11 ), "Bad results for sequence 1" )
	assert( isDone1, "First sequence should be done" )
		
	assert( testSequence( t2, 'data', 'data', 'error', 'data', 'error', 'error', 'data' ),
			"Bad results for sequence 2" )
	assert( isDone2, "Second sequence should be done" )
	
end

--------------------------------------------------------------------------------

-- FIXME: Test rawMapB.

--------------------------------------------------------------------------------

--[[

FIXME: We need non-terminating error support

function KSReactiveTest.testHoldB_error()

	local KSReactive = require 'KSReactive'
	local KSRPackedValues = require 'KSRPackedValues'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	local node = KSReactive_sourceE()
	
	local behavior = node:holdB( "foo" )
	
	assert( behavior:valueNow() == "foo", "Initial value should be foo" )
	
	node:post( "baz" )
	
	assert( behavior:valueNow() == "baz", "Behavior should have updated" )
		
	node:close( "phlegm" )
	
	FIXME We no longer pass errors to close
	
	local currentValue = behaior:valueNow()
	
	local isError, errorMessage = KSRPackedValues.isNonTerminatingError( currentValue )
	
	assert( isError, "current value should be a non-terminating error" )
	assert( errorMessage == "phlegm", "Wrong error message" )

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testTimerE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testMapB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testMapSwitchB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testBindVariable()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testIsEventNode()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testSkipErrorsE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testRelayE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testRelayB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

--[[ switchE is no more. See mapSwitchE.

FIXME: We probably need to test this case.

function KSReactiveTest.testBehaviorSwitchE()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE

	local e1 = KSReactive_sourceE()
	local e2 = KSReactive_sourceE()
	local e3 = KSReactive_sourceE()
	local e4 = KSReactive_sourceE()
	
	local b = KSReactive.variableB( e1 )
	
	local se = b:switchE()
	
	local t = { }
	
	se:onDataE( function( x ) t[ #t + 1 ] = x end )
	
	e1:post( 1 )
	e2:post( "missed" )
	e3:post( "missed 2" )
	e1:post( 2 )
	
	b:set( e2 )
	
	e1:post( "not this" )
	
	e2:post( 3 )
	
	b:set( e3 )
	b:set( e4 )
	
	e3:post( "hidden" )
	
	e4:post( 4 )
	
	assert( #t == 4, "We should have captured four values" )
	
	for i = 1, 4 do
		assert( t[ i ] == i, "t[ " .. i .. " ] should equal " .. i )
	end

end

]]

--------------------------------------------------------------------------------

function KSReactiveTest.testMapSwitchB()

	local KSReactive = require 'KSReactive'
	
	local b = KSReactive.variableB( 0 )
	
	local bb
	
	local smb = b:mapSwitchB( function() bb = KSReactive.variableB( false ); return bb end )
	
	assert( smb:valueNow() == false, "value should come through as false" )
	
	bb:set( true )
	
	assert( smb:valueNow() == true, "should see change from underlying behavior" )
	
	b:set( 1 )
	
	assert( smb:valueNow() == false, "value should be back to false" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testMapSwitchE()

	local KSReactive = require 'KSReactive'
	
	local r = { }
	local re = { }
	local rdone = false
	
	local e = KSReactive.sourceE()
	
	local ee
	
	local sme = e:mapSwitchE( function() ee = KSReactive.sourceE(); return ee end )
	
	sme:onDataE( function( x ) r[ #r + 1 ] = x end )
	sme:onErrorE( function( x ) re[ #re + 1 ] = x end )
	sme:onDoneE( function() rdone = true end )
	
	assert( ee == nil, "We should not have an event stream yet" )
	
	e:post( true )
	
	assert( ee ~= nil, "We should now have an event stream" )
	
	local closed_ee = false
	
	ee:addCloseAction( function() closed_ee = true end )
	
	ee:post( 'foo' )
	
	assert( r[ #r ] == 'foo', "We should have received the event foo" )
	
	ee:post( 'baz' )
	
	assert( r[ #r ] == 'baz', "We should have received the event baz" )
	
	ee:postError( "ee_error" )
	
	assert( re[ #re ] == 'ee_error', "We should have received the error ee_error" )
	
	e:post( 3.14 )
	
	assert( closed_ee, "We should have closed the original event stream" )
	
	ee:post( 'bar' )
	
	assert( r[ #r ] == 'bar', "We should have received the event bar" )
	
	e:post( { } )
	
	ee:close()
	
	assert( not rdone, "Closing the event stream shouldn't have closed the switch" )
	
	e:postError( 'e_error' )
	
	assert( re[ #re ] == 'e_error', "We should have received the error e_error" )
	
	e:post( 'a' )
	
	e:close()
	
	ee:post( 'after_close' )
	
	assert( r[ #r ] == 'after_close', "We should have received the event bar" )
	
	ee:close()
	
	assert( rdone, "We should now have closed the mapSwitchE" )
	
	assert( #r == 4, "We should have received 4 values" )
	
	assert( #re == 2, "We should have received 2 error" )

end

--------------------------------------------------------------------------------

--[[

FIXME

function KSReactiveTest.testConstantB()

	local KSReactive = require 'KSReactive'
	
	local KSReactive_sourceE = KSReactive.sourceE
	
	return KSReactive_sourceE()
	
	-- FIXME [mhamburg 2011-??-??]: Implement this

end

]]

--------------------------------------------------------------------------------

function KSReactiveTest.testZipMapSwitchB()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local b = KSReactive.variableB( 0 )
	
	local bb
	
	local smb = b:zipMapSwitchB( function() bb = KSReactive.variableB( false ); return bb end )
	
	KSTestUtils.assertEquals "initial state" ( 0, false )( smb:valueNow() )
	
	bb:set( true )
	
	KSTestUtils.assertEquals "first sub-behavior change" ( 0, true )( smb:valueNow() )
	
	b:set( 1 )
	
	KSTestUtils.assertEquals "first root behavior change" ( 1, false )( smb:valueNow() )
	
	b:set( 2 )
	
	KSTestUtils.assertEquals "second root behavior change" ( 2, false )( smb:valueNow() )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testCallB()

	local KSReactive = require 'KSReactive'
	
	local fnB = KSReactive.variableB( function( a, b ) return a + b end )

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local cB = KSReactive.callB( fnB, aB, bB )
	
	local dB = KSReactive.callB( function( a, b ) return a + b end, cB, bB )
	
	assert( dB:valueNow() == 0, "Expected initial value to be 0." )
	
	bB:set( 4 )
	
	assert( dB:valueNow() == 8, "Expected value after b = 4 to be 8." )
	
	aB:set( 5 )
	
	assert( dB:valueNow() == 13, "Expected value after a = 5 to be 13." )
	
	fnB:set( function( a, b ) return a - b end )
	
	assert( dB:valueNow() == 5, "Expected value after changing function to be 5." )
	
	fnB:set( function( a, b ) return a or b end )
	
	assert( dB:valueNow() == 9, "Expected value after changing function to or to be 9." )
	
	aB:set( nil )
	
	assert( dB:valueNow() == 8, "Setting a to nil should give us b + b = 8." )
	
if false then
	-- Re-enable this block when 3332265 has been fixed.
	
	-- MarkH: I disagree that there is a bug here. Just as we don't look at deep equality,
	-- we assume that differences in the number of arguments might matter somewhere.

	aB:set( true )
	
	local numValues = 0
	dB = KSReactive.callB( function( a )
		if a then
			return true
		else
			return true, nil
		end
	end, aB ):mapB( function( x )
		numValues = numValues + 1
		return x
	end )
	
	assert( dB:valueNow() == true, "true" )
	assert( numValues == 1, "Only one result so far" )
	aB:set( false )
	assert( dB:valueNow() == true, "true" )
	assert( numValues == 1, "An extra nil added to the result tuple should not count as a difference" )
	
end

	-- FIXME: Confirm that we don't over eval

end

--------------------------------------------------------------------------------

function KSReactiveTest.testArgTupleB()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local zB = KSReactive.constantB( "foo" )
	
	local cB = KSReactive.argTupleB( aB, 7, zB, bB )
	
	KSTestUtils.assertEquals "initial" ( 0, 7, "foo", 0 )( cB:valueNow() )
	
	bB:set( 4 )
	
	KSTestUtils.assertEquals "after bB = 4" ( 0, 7, "foo", 4 )( cB:valueNow() )

	aB:set( 5 )
	
	KSTestUtils.assertEquals "after aB = 5" ( 5, 7, "foo", 4 )( cB:valueNow() )

	aB:set( nil )
	
	KSTestUtils.assertEquals "after aB = nil" ( nil, 7, "foo", 4 )( cB:valueNow() )

	bB:set( nil )
	
	KSTestUtils.assertEquals "after bB = nil" ( nil, 7, "foo", nil )( cB:valueNow() )
	
	bB:set()
	
	KSTestUtils.assertEquals "after bB = nothing" ( nil, 7, "foo" )( cB:valueNow() )
	
	aB:setError( "Test error propagation" )
	
	KSTestUtils.assertEquals "error propagation" ( true, "Test error propagation" )( cB:isErrorNow() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testArgTupleB_done()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local zB = KSReactive.constantB( "foo" )
	
	local cB = KSReactive.argTupleB( aB, 7, zB, bB )
	
	KSTestUtils.assertEquals "initial" ( false )( cB:isDone() )
	
	aB:close()
	
	KSTestUtils.assertEquals "aB closed" ( false )( cB:isDone() )
	
	bB:close()
	
	KSTestUtils.assertEquals "bB closed" ( true )( bB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testArgTupleB_zero()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local tB = KSReactive.argTupleB()
	
	KSTestUtils.assertEquals "empty" ( )( tB:valueNow() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testArgTupleB_const()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.constantB( 3 )
	
	local bB = KSReactive.constantB( 4 )
	
	local tB = KSReactive.argTupleB( aB, bB )
	
	KSTestUtils.assertEquals "empty" ( true )( tB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testArgTupleB_constError()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.constantB( 3 )
	
	local bB = KSReactive.variableB( 4 )
	
	bB:setError( "Test error propagation" )
	
	bB:close() -- done changing
	
	local tB = KSReactive.argTupleB( aB, bB )
	
	KSTestUtils.assertEquals "isDone" ( true )( tB:isDone() )
	
	KSTestUtils.assertEquals "isErrorNow" ( true, "Test error propagation" )( tB:isErrorNow() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testDataCallB()

	local KSReactive = require 'KSReactive'
	
	local fnB = KSReactive.variableB( function( a, b ) return a + b end )

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local cB = KSReactive.dataCallB( fnB, aB, bB )
	
	local dB = KSReactive.dataCallB( function( a, b ) return a + b end, cB, bB )
	
	assert( dB:valueNow() == 0, "Expected initial value to be 0." )
	
	bB:set( 4 )
	
	assert( dB:valueNow() == 8, "Expected value after b = 4 to be 8." )
	
	aB:set( 5 )
	
	assert( dB:valueNow() == 13, "Expected value after a = 5 to be 13." )
	
	fnB:set( function( a, b ) return a - b end )
	
	assert( dB:valueNow() == 5, "Expected value after changing function to be 5." )
	
	fnB:set( function( a, b ) return a or b end )
	
	assert( dB:valueNow() == 9, "Expected value after changing function to or to be 9." )
	
	aB:set( nil )
	
	assert( dB:valueNow() == 8, "Setting a to nil should give us b + b = 8." )
	
	aB:setError( "Test error propagation" )
	
	local dB_isError, dB_msg = dB:isErrorNow()
	
	assert( dB_isError, "Setting to an error should propagate as the error." )
	
	assert( dB_msg == "Test error propagation", "The error message should propagate." )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testDataCallB_autoSwitch()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.variableB( false )
	
	local bB = KSReactive.variableB( 4 )
	
	local function fn( v )
		return v and bB or "foo"
	end
	
	local cB = KSReactive.dataCallB( fn, aB )
	
	KSTestUtils.assertEquals "initial" ( "foo" )( cB:valueNow() )
	
	bB:setError( "Hidden error" )
	
	KSTestUtils.assertEquals "hidden error" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "error revealed" ( true, "Hidden error" )( cB:isErrorNow() )
	
	bB:set( "hello" )
	
	KSTestUtils.assertEquals "error cleared" ( "hello" )( cB:valueNow() )
	
	aB:set( false )
	
	KSTestUtils.assertEquals "switch away" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "switch back" ( "hello" )( cB:valueNow() )
	
	bB:close()
	
	KSTestUtils.assertEquals "bB closed" ( false )( cB:isDone() )
	
	aB:close()
	
	KSTestUtils.assertEquals "aB closed" ( true )( cB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testDataCallB_autoSwitch_close2()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.variableB( false )
	
	local bB = KSReactive.variableB( "hello" )
	
	local function fn( v )
		return v and bB or "foo"
	end
	
	local cB = KSReactive.dataCallB( fn, aB )
	
	KSTestUtils.assertEquals "initial" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "switch" ( "hello" )( cB:valueNow() )
	
	aB:close()
	
	KSTestUtils.assertEquals "aB closed" ( false )( cB:isDone() )
	
	bB:set( "greetings" )
	
	KSTestUtils.assertEquals "update" ( "greetings" )( cB:valueNow() )
	
	bB:close()
	
	KSTestUtils.assertEquals "bB closed" ( true )( cB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testWeakEventNodesTable()

	local KSReactive = require 'KSReactive'
	
	local WFCloseSupport = require 'WFCloseSupport'
	
	local et = KSReactive.makeWeakEventTable( "< initial >" )
	
	local b = et:behaviorB( 'foo' )
	local e = et:eventE( 'baz' )

	assert( b:valueNow() == "< initial >", "Initial value for behavior is wrong." )
	
	et:postToEvent( 'foo', "Hello" )
	
	assert( b:valueNow() == "Hello", "Value failed to propagate" )
	
	local errorProcessed = false
	
	e:onErrorE( function() errorProcessed = true end )
	
	et:postErrorToEvent( 'baz', "BOOM!" )
	
	assert( errorProcessed, "The error failed to propagate" )
	
	et:postToEvent( 'blah', 4 ) -- Check that we can post to non-existent keys
	
	local ek = et:eventKeys()
	
	assert( #ek == 2, "We should have two active event keys." )
	assert( ( ek[ 1 ] == 'foo' and ek[ 2 ] == 'baz' ) or ( ek[ 1 ] == 'baz' and ek[ 2 ] == 'foo' ),
			"The wrong keys seem to be active." )
	
	et:closeEvent( 'baz' )
	
	assert( WFCloseSupport.hasBeenOrWillBeClosed( e ), "Close did not propagate" )
	
	assert( et:eventExists( 'foo' ), "Foo should exist" )
	assert( not et:eventExists( 'baz' ), "Baz should no longer exist" )
	assert( not et:eventExists( 'blah' ), "Blah should never have existed" )
	
	local ek2 = et:eventKeys()

	assert( #ek2 == 1, "We should only have one active event key now." )
	assert( ek2[ 1 ] == 'foo', "'foo' should be the active event key." )
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( et:eventExists( 'foo' ), "Foo should still exist after garbage collection" )
	
	local beffect = b:onDataB( function() end )
		-- Anchor it
		
	local fn = function() end
	fn( beffect ) -- silence the lint about beffect going unused
		
	local weak = setmetatable( { }, { __mode = 'kv' } )
	
	weak[ 1 ] = b
	
	b = nil
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ], "beffect = onDataB should have kept b alive" )
	assert( et:eventExists( 'foo' ), "b being alive should have kept the event alive" )
	
	et:postToEvent( 'foo', "Still here!" )
	
	assert( weak[ 1 ]:valueNow() == "Still here!", "Event wiring should still work for foo" )
	
	et:close()

	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( not weak[ 1 ], "Closing the event table should have allowed b to be collected" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testLiftB()

	local KSReactive = require 'KSReactive'
	
	local fB = KSReactive.liftB( function( a, b ) return a + b end )

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local cB = fB( aB, bB )
	
	local dB = fB( cB, bB )
	
	assert( dB:valueNow() == 0, "Expected initial value to be 0." )
	
	bB:set( 4 )
	
	assert( dB:valueNow() == 8, "Expected value after b = 4 to be 8." )
	
	aB:set( 5 )
	
	assert( dB:valueNow() == 13, "Expected value after a = 5 to be 13." )
	
	aB:setError( "Test error propagation" )
	
	local dB_isError, dB_msg = dB:isErrorNow()
	
	assert( dB_isError, "Setting to an error should propagate as the error." )
	
	assert( dB_msg == "Test error propagation", "The error message should propagate." )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testLiftB_autoSwitch()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.variableB( false )
	
	local bB = KSReactive.variableB( 4 )
	
	local fnB = KSReactive.liftB( function( v )
		return v and bB or "foo"
	end )
	
	local cB = fnB( aB )
	
	KSTestUtils.assertEquals "initial" ( "foo" )( cB:valueNow() )
	
	bB:setError( "Hidden error" )
	
	KSTestUtils.assertEquals "hidden error" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "error revealed" ( true, "Hidden error" )( cB:isErrorNow() )
	
	bB:set( "hello" )
	
	KSTestUtils.assertEquals "error cleared" ( "hello" )( cB:valueNow() )
	
	aB:set( false )
	
	KSTestUtils.assertEquals "switch away" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "switch back" ( "hello" )( cB:valueNow() )
	
	bB:close()
	
	KSTestUtils.assertEquals "bB closed" ( false )( cB:isDone() )
	
	aB:close()
	
	KSTestUtils.assertEquals "aB closed" ( true )( cB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testLiftB_autoSwitch_close2()

	local KSReactive = require 'KSReactive'
	
	local KSTestUtils = require 'KSTestUtils'
	
	local aB = KSReactive.variableB( false )
	
	local bB = KSReactive.variableB( "hello" )
	
	local fnB = KSReactive.liftB( function( v )
		return v and bB or "foo"
	end )
	
	local cB = fnB( aB )
	
	KSTestUtils.assertEquals "initial" ( "foo" )( cB:valueNow() )
	
	aB:set( true )
	
	KSTestUtils.assertEquals "switch" ( "hello" )( cB:valueNow() )
	
	aB:close()
	
	KSTestUtils.assertEquals "aB closed" ( false )( cB:isDone() )
	
	bB:set( "greetings" )
	
	KSTestUtils.assertEquals "update" ( "greetings" )( cB:valueNow() )
	
	bB:close()
	
	KSTestUtils.assertEquals "bB closed" ( true )( cB:isDone() )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testLift()

	local KSReactive = require 'KSReactive'
	
	local t = KSReactive.lift {
		fnB = function( a, b ) return a + b end
	}
	
	local fnB = t.fnB

	local aB = KSReactive.variableB( 0 )
	
	local bB = KSReactive.variableB( 0 )
	
	local cB = fnB( aB, bB )
	
	local dB = fnB( cB, bB )
	
	assert( dB:valueNow() == 0, "Expected initial value to be 0." )
	
	bB:set( 4 )
	
	assert( dB:valueNow() == 8, "Expected value after b = 4 to be 8." )
	
	aB:set( 5 )
	
	assert( dB:valueNow() == 13, "Expected value after a = 5 to be 13." )
	
	aB:setError( "Test error propagation" )
	
	local dB_isError, dB_msg = dB:isErrorNow()
	
	assert( dB_isError, "Setting to an error should propagate as the error." )
	
	assert( dB_msg == "Test error propagation", "The error message should propagate." )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testWeakEventNodesTable()

	local KSReactive = require 'KSReactive'
	
	local WFCloseSupport = require 'WFCloseSupport'
	
	local et = KSReactive.makeWeakEventTable( "< initial >" )
	
	local b = et:behaviorB( 'foo' )
	local e = et:eventE( 'baz' )

	assert( b:valueNow() == "< initial >", "Initial value for behavior is wrong." )
	
	et:postToEvent( 'foo', "Hello" )
	
	assert( b:valueNow() == "Hello", "Value failed to propagate" )
	
	local errorProcessed = false
	
	e:onErrorE( function() errorProcessed = true end )
	
	et:postErrorToEvent( 'baz', "BOOM!" )
	
	assert( errorProcessed, "The error failed to propagate" )
	
	et:postToEvent( 'blah', 4 ) -- Check that we can post to non-existent keys
	
	local ek = et:eventKeys()
	
	assert( #ek == 2, "We should have two active event keys." )
	assert( ( ek[ 1 ] == 'foo' and ek[ 2 ] == 'baz' ) or ( ek[ 1 ] == 'baz' and ek[ 2 ] == 'foo' ),
			"The wrong keys seem to be active." )
	
	et:closeEvent( 'baz' )
	
	assert( WFCloseSupport.hasBeenOrWillBeClosed( e ), "Close did not propagate" )
	
	assert( et:eventExists( 'foo' ), "Foo should exist" )
	assert( not et:eventExists( 'baz' ), "Baz should no longer exist" )
	assert( not et:eventExists( 'blah' ), "Blah should never have existed" )
	
	local ek2 = et:eventKeys()

	assert( #ek2 == 1, "We should only have one active event key now." )
	assert( ek2[ 1 ] == 'foo', "'foo' should be the active event key." )
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( et:eventExists( 'foo' ), "Foo should still exist after garbage collection" )
	
	local beffect = b:onDataB( function() end )
		-- Anchor it
		
	local fn = function() end
	fn( beffect ) -- silence the lint about beffect going unused
		
	local weak = setmetatable( { }, { __mode = 'kv' } )
	
	weak[ 1 ] = b
	
	b = nil
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ], "beffect = onDataB should have kept b alive" )
	assert( et:eventExists( 'foo' ), "b being alive should have kept the event alive" )
	
	et:postToEvent( 'foo', "Still here!" )
	
	assert( weak[ 1 ]:valueNow() == "Still here!", "Event wiring should still work for foo" )
	
	et:close()

	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( not weak[ 1 ], "Closing the event table should have allowed b to be collected" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testConstantE()

	local KSReactive = require 'KSReactive'
	
	local source = KSReactive.sourceE()
	
	local resultFoo = source:constantE( 'foo' )
	
	local fooGotValue = false
	local fooResult = nil
	
	local function checkFooGotValue( vtype, result )
		assert( fooGotValue == vtype, "Expected to get value of type " .. vtype )
		assert( fooResult == result, "Expected value " .. result )
		fooGotValue = false
		fooResult = nil
	end
	
	resultFoo:onDataE( function( ... )
		assert( select( '#', ... ) == 1, "Expected 1 value" )
		fooGotValue = 'data'
		fooResult = ( ... )
	end )
	
	resultFoo:onErrorE( function( message )
		fooGotValue = 'error'
		fooResult = message
	end )
	
	source:post( 1 )
	checkFooGotValue( 'data', 'foo' )
	
	source:post( 2 )
	checkFooGotValue( 'data', 'foo' )
	
	source:postError( 'baz' )
	checkFooGotValue( 'error', 'baz' )

	source:post( 3 )
	checkFooGotValue( 'data', 'foo' )
	
end

--------------------------------------------------------------------------------

function KSReactiveTest.testTakeE()

	local KSReactive = require 'KSReactive'
	
	local source = KSReactive.sourceE()
	
	local taker = source:takeE( 3 )
	
	local results = { }
	local didFinish = false
	
	taker:onDataE( function( x ) results[ #results + 1 ] = x end )
	taker:onDoneE( function() didFinish = true end )
	
	local didFinish10 = false
	
	source:takeE( 10 ):onDoneE( function() didFinish10 = true end )

	local didFinish11 = false
	
	source:takeE( 11 ):onDoneE( function() didFinish11 = true end )

	local didFinish12 = false
	
	source:takeE( 12 ):onDoneE( function() didFinish12 = true end )

	for i = 1, 10 do
		source:post( i )
	end
	
	assert( didFinish, "taker should have finished by now" )
	assert( #results == 3, "We should have three results" )
	
	for i = 1, #results do
		assert( results[ i ] == i, "Wrong results at position: " .. i )
	end
	
	assert( didFinish10, "We should have signaled done for the exact match on count" )
	
	assert( not didFinish11, "We should not have finished 11." )
	assert( not didFinish12, "We should not have finished 12." )
	
	source:close()
	
	assert( didFinish11, "But closing should mark us as done (11)" )
	assert( didFinish12, "But closing should mark us as done (12)" )
	
end
	
--------------------------------------------------------------------------------

function KSReactiveTest.testTableB()

	local KSReactive = require 'KSReactive'
	local KSTableUtils = require 'KSTableUtils'
	
	local B1 = KSReactive.variableB( "foo" )
	local B2 = KSReactive.variableB( 3 )
	local B3 = B2:mapB( function( x ) return x * x end )
	local B4 = KSReactive.constantB( "baz" )
	
	local TB = KSReactive.tableB {
		a = B1,
		b = B2,
		c = { d = B3, e = B4 }
	}
	
	local match = { a = "foo", b = 3, c = { d = 9, e = "baz" } }
	
	local case = 'initial'
	
	local success = true
	local failedCase = 'none'
	
	TB:onDataB( function( data )
		if not KSTableUtils.deepEquals( match, data ) then
			success = false
			failedCase = case
		end
	end )
	
	match = { a = "foo", b = 4, c = { d = 16, e = "baz" } }
	case = '3 to 4'
	
	B2:set( 4 )
	
	match = { a = "flok", b = 4, c = { d = 16, e = "baz" } }
	case = 'foo to flok'
	
	B1:set( "flok" )
	
	match = { a = { glom = 7 }, b = 4, c = { d = 16, e = "baz" } }
	case = 'flok to glok'
	
	B1:set( { glom = 7 } )
	
	assert( success, "Match failed for case: " .. failedCase )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testTableBRejectsReactiveObjects()

	local KSReactive = require 'KSReactive'
	local KSPromises = require 'KSPromises'
	
	local p = KSPromises.resolvedPromiseP( 'data', 3 )
	
	assert( not pcall( KSReactive.tableB, { foo = { baz = p } } ),
			"We should have rejected a table containing a promise" )

	local e = KSReactive.sourceE()
	
	assert( not pcall( KSReactive.tableB, { foo = { baz = e } } ),
			"We should have rejected a table containing an event stream" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testMapSwitchBIsConstantFriendly()

	local KSReactive = require 'KSReactive'
	
	local constantBehaviors = {
		KSReactive.constantB( 1 ),
		KSReactive.constantB( 2 )
	}

	local closed = 0

	local function mapFn( i )

		local b = constantBehaviors[ i % 4 ]
		
		if not b then
			b = KSReactive.variableB( 'foo' ):addCloseAction( function() closed = closed + 1 end )
		end
		
		return b

	end
	
	local varB = KSReactive.variableB( 0 )

	local finished = false
	
	varB:mapSwitchB( mapFn ):onDoneE( function() finished = true end )
	
	for i = 1, 17 do -- Designed to run through to a constant behavior
		varB:set( i )
	end
	
	assert( not finished, "Finished too soon" )
	
	varB:close()
	
	assert( finished, "Finished too late" )
	
	assert( closed == 9, "We should have closed 9 behaviors, did close " .. tostring( closed ) )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testRunUntilDone()

	local KSReactive = require 'KSReactive'
	
	local sourceE = KSReactive.sourceE()
	
	local weak = setmetatable( {}, { __mode = 'kv' } )
	
	weak[ 1 ] = sourceE
	
	sourceE:runUntilDone()
	
	sourceE = nil
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ] ~= nil, "source was prematurely collected" )
	
	weak[ 1 ]:post( 'foo' )
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ] ~= nil, "source was prematurely collected (2)" )
	
	weak[ 1 ]:postError( 'baz' )
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ] ~= nil, "source was prematurely collected (2)" )
	
	weak[ 1 ]:close()
	
	collectgarbage()
	collectgarbage()
	collectgarbage()
	
	assert( weak[ 1 ] == nil, "source should have been collected" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testEndAfterErrorE()

	local KSReactive = require 'KSReactive'
	
	local sourceE = KSReactive.sourceE()
	
	local outE = sourceE:endAfterErrorE()
	
	local data = nil
	local done = false
	local emsg = nil
	
	outE:onDataE( function( x )
		data = x
	end )
	
	outE:onDoneE( function()
		done = true
	end )
	
	outE:onErrorE( function( msg )
		emsg = msg
	end )
	
	sourceE:post( 'baz' )
	sourceE:post( 'foo' )
	sourceE:postError( "First error" )
	sourceE:postError( "Second error" )
	sourceE:post( 'bleen' )
	sourceE:postError( "Third error" )
	
	assert( data == 'foo', "Should have ended up at data = foo" )
	assert( emsg == "First error", "We should only have passed the first error" )
	assert( done, "We should be done by now." )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testAccumulateE_counter()

	local KSReactive = require 'KSReactive'
	
	local sourceE = KSReactive.sourceE()
	
	local outE = sourceE:accumulateE( function( accum ) return accum + 1 end, 0 )
	
	local data = nil
	local emsg = nil

	outE:onDataE( function( x )
		data = x
	end )
	
	outE:onErrorE( function( msg )
		emsg = msg
	end )
	
	sourceE:post( 'baz' )
	sourceE:post( 'foo' )
	sourceE:postError( "First error" )
	sourceE:postError( "Second error" )
	sourceE:post( 'bleen' )
	sourceE:postError( "Third error" )
	
	assert( data == 3, "Should have ended up at data = foo" )
	assert( emsg == "Third error", "We should only have passed the first error" )

end

--------------------------------------------------------------------------------

function KSReactiveTest.testAccumulateB_counter()

	local KSReactive = require 'KSReactive'
	
	local sourceE = KSReactive.sourceE()
	
	local outB = sourceE:accumulateB( function( accum ) return accum + 1 end, 0 )
	
	local data = nil
	local emsg = nil

	outB:onDataB( function( x )
		data = x
	end )
	
	assert( data == 0, "We should start out at zero" )
	
	outB:onErrorE( function( msg )
		emsg = msg
	end )
	
	sourceE:post( 'baz' )
	sourceE:post( 'foo' )
	sourceE:postError( "First error" )
	sourceE:postError( "Second error" )
	sourceE:post( 'bleen' )
	sourceE:postError( "Third error" )
	
	assert( data == 3, "Should have ended up at data = foo" )
	assert( emsg == "Third error", "We should only have passed the first error" )

end

--------------------------------------------------------------------------------

return KSReactiveTest
