--[[----------------------------------------------------------------------------

KSReactiveTimer.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2011-01-11

@reviewed scouten 2011-03-11

------------------------------------------------------------------------------]]

--[[

	KSReactiveTimer: Timed event support
	
	function KSReactiveTimer.laterB()
		Returns a behavior that transitions from false to true when the event processor
		gets a chance to process later.
		
	function KSReactiveTimer.timerE( delayTime ) -> timer event stream
		Returns an event stream that will fire the value true after the given delay time.
		Delay times are rounded and grouped so that we can avoid creating too many
		separate timers. If one passes 'later' for delayTime, the event stream fires
		essentially at the same time laterB would fire.
		
	function KSReactiveTimer.timedSeqB( value0, delay, value1 )
		Returns a behavior that starts at value0 and transitions to value1 after the indicated
		delay.
	
	function KSReactiveTimer.falseThenTrueB( delay )
		Returns a behavior that starts at false and transitions to true after the indicated
		delay.
	
	function KSReactiveTimer.trueThenFalseB( delay )
		Returns a behavior that starts at true and transitions to false after the indicated
		delay.
		
	function KSReactiveTimer.calmE( e, interval, startTimer )
	
	function KSReactiveTimer.calmB( b, interval, startTimer )
		
]]

--@mark KSReactiveTimer namespace

--[[!
      \class KSReactiveTimer

      \brief This namespace provides support for timed events.
--]]

--============================================================================--

local KSReactive = require 'KSReactive'

local KSReactive_variableB = assert( KSReactive.variableB )
local KSReactive_sourceE = assert( KSReactive.sourceE )
local KSReactive_constantB = assert( KSReactive.constantB )

local WFIdle = require 'WFIdle'

local WFIdle_laterP = assert( WFIdle.laterP )

local WFTimer = require 'WFTimer'

local WFTimer_afterDelayP = assert( WFTimer.afterDelayP )


--============================================================================--

local cacheLaterB = nil

--------------------------------------------------------------------------------

--[[!
	Creates a behavior that transitions from false to true when the worker thread
	gets a chance to process it.

	\return The new behavior object.

]]--

local function laterB()

	if not cacheLaterB then

		local laterVarB = KSReactive_variableB( false )
		
		local trigger = function()
			cacheLaterB = nil -- We will also need a new behavior
			laterVarB:set( true )
			laterVarB:close()
		end
		
		WFIdle_laterP():whenResolvedDo( trigger ):runUntilDone()
		
		cacheLaterB = laterVarB:noCloseRelayB()

	end
	
	return cacheLaterB

end


--============================================================================--

--[[!
       Creats an event stream that fires the value true after a given delay time.
       	Delay times are rounded and grouped so that we can avoid creating too many
       	separate timers.

	\param delay The number of seconds to delay, or 'later' to fire true when
		the worker thread gets a chance to process it.

	\return The new timer event-stream object.
]]--

local function timerE( delay )

	if delay == 'later' then
		return laterB()
	end

	-- FIXME [scouten 2012-03-28]: Not sure about the 0 => 0.1 mapping.

	if delay == 0 then
		delay = .01
	end

	if delay <= 0 then
		error( "KSReactive.timerE: delay must be a positive number", 2 )
	end

	-- NOTE: The quantization that used to occur in this function is now
	-- handled in WFTimer.

	local resultTimerE = KSReactive_sourceE()
	
	WFTimer_afterDelayP( delay )
		:whenResolvedDo( function()
			resultTimerE:post( true )
			resultTimerE:close()
		end )
		:attachTo( resultTimerE )
	
	return resultTimerE:relayE()

end


--============================================================================--

-- @mark timedSeqB, falseThenTrueB

--------------------------------------------------------------------------------
--[[!
       Creates a behavior whose initial value transitions to another value
	after a delay.

	\param value0 The initial value.
	\param delay The number of seconds to delay.
	\param value1 The post-delay value.

	\return The new behavior object.
]]--

local function timedSeqB( value0, delay, value1 )

	return timerE( delay ):constantE( value1 ):holdB( value0 )

end

--------------------------------------------------------------------------------
--[[!
       Creates a behavior whose initial value is false, which transitions to true
	after a delay.

 	\param delay The number of seconds to delay.
 
	\return The new behavior object.
]]--

local function falseThenTrueB( delay )

	return timedSeqB( false, delay, true )

end

--------------------------------------------------------------------------------
--[[!
       Creates a behavior whose initial value is true, which transitions to false
	after a delay.

 	\param delay The number of seconds to delay.
 
	\return The new behavior object.

]]--

local function trueThenFalseB( delay )

	return timedSeqB( true, delay, false )

end


--============================================================================--

-- @mark calmE, calmB

--------------------------------------------------------------------------------
--[[!
       (what does this do?)(how is it different from calmB?)

       \param e	 A timer event-stream object (??)
       \param interval	A number of seconds.
       \param startTimer (boolean) (??)

	\return An event-stream object.
]]--

local function calmE( e, interval, startTimer )

	-- We have a current timer if it hasn't actually fired.
	
	local currentTimerE = nil
	
	local function startCurrentTimer()

		if not currentTimerE then

			-- Create a timer if necessary. Appended with logic to clear the timer
			-- reference once it fires. Note that since we reveal the map function,
			-- then the reference will get cleared before anything else can see the
			-- event.

			local thisTimerE

			thisTimerE = timerE( interval ):mapE( function( ... )
				if currentTimerE == thisTimerE then
					currentTimerE = nil
				end
				return ...
			end )

			currentTimerE = thisTimerE

		end

	end
	
	-- Errors clear out the current timer so that a new value won't get
	-- delayed.
	
	local clearForErrorsE = e:rawMapE( function( status, ... )
		if status ~= 'data' then
			currentTimerE = nil
		end
		return 'same'
	end )
	
	-- If we have a timer, then we wait to fire. If not, we pass it through
	-- immediately.
	
	local timerFiredE = clearForErrorsE:mapSwitchE( function( ... )
		if currentTimerE then
			return currentTimerE:constantE( ... )
		else
			return KSReactive_constantB( ... )
		end
	end )
	
	-- Whenever we fire a data value, we start the timer.
	
	timerFiredE = timerFiredE:mapE( function( ... )
		startCurrentTimer()
		return ...
	end )
	
	if startTimer then
		startCurrentTimer()
	end
	
	return timerFiredE

end

--------------------------------------------------------------------------------
--[[!
       (what does this do?)(how is it different from calmB?)

       \param b	 A timer behavior object (??)
       \param interval	A number of seconds.
       \param startTimer (boolean) (??)

	\return (??)

]]--

local function calmB( b, interval, startTimer )

	return calmE( b, interval, startTimer ):holdB( b:valueNow() )

end


--============================================================================--

return {
	timerE = timerE,
	timedSeqB = timedSeqB,
	falseThenTrueB = falseThenTrueB,
	trueThenFalseB = trueThenFalseB,
	laterB = laterB,
	calmE = calmE,
	calmB = calmB
}
