--[[----------------------------------------------------------------------------

KSStringUtils.lua

@author erwright 2011-08-31

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.string
local KSStringUtilsImp = require 'KSStringUtilsImp'

	-- wf.core.lua
local KSData = require 'KSData'

local string_find = assert( string.find )
local string_sub = assert( string.sub )


-- @mark KSStringUtils namespace

--[[!
	\class KSStringUtils
	\brief This namespace provides a cross-platform string-manipulation utility.

--]]

--==============================================================================


--============================================================================--

local KSStringUtils = { }

--------------------------------------------------------------------------------

-- GRUNGE [mhamburg 2011-03-02]: This could probably use string.gmatch and
-- simplify the logic a bit but that would mess with the split pattern.

--[[!
	 Separates a string into table entries using a given separator character.

	 \param str The string.
	 \param inSplitPattern	A string containing the separator character(s).
	 \param outResults An optional table in which to return the result.

	 \return The output table with the new entries appended,
	 	or a new table containing the substrings.
--]]

function KSStringUtils.split( --[[string]] str, --[[string]] inSplitPattern, --[[table]] outResults )
	if not outResults then
		outResults = { }
	end
	local theStart = 1
	local theSplitStart, theSplitEnd = string_find( str, inSplitPattern, theStart )
	while theSplitStart do
		outResults[ #outResults + 1 ] = string_sub( str, theStart, theSplitStart - 1 )
		theStart = theSplitEnd + 1
		theSplitStart, theSplitEnd = string_find( str, inSplitPattern, theStart )
	end
	outResults[ #outResults + 1 ] = string_sub( str, theStart )
	return outResults
end

--------------------------------------------------------------------------------
-- This function does an Unicode string compare, not a localized string compare.
--[[!
	 Compares two strings (using a Unicode, not localized, comparison).
	 
		When doing string comparison in Lua, numbers in the string are
	 	detected and compared numerically, so that n2 comes before n10,
	 	which is not the case with a strict character comparison.
	 	For example, this order is observed:
	 
	 	<pre>
	 	A
	 	N
	 	n
	 	N1
	 	n1
	 	N2
	 	n2
	 	N10
	 	n10
	 	&Ntilde;
	 	&Ntilde;1
	 	&Ntilde;2
	 	&Ntilde;10
	 	Z
	 	</pre>

	 \param string1 The first string.
	 \param string2 The second string.
	 \return  0 if the strings are equal,
	 	-1 if string1 < string2,
	 	1 if string1 > string2.
--]]

function KSStringUtils.localizedStringCompare( --[[string]] string1, --[[string]] string2 )
	return KSStringUtilsImp.i18n_collate( string1, string2 )
end

--------------------------------------------------------------------------------
--[[!
	 Compares two strings using a combination of localized, case insensitive,
	 and numeric sort.

	 \param string1 The first string.
	 \param string2 The second string.
	 \return  0 if the strings are equal,
	 	-1 if string1 < string2,
	 	1 if string1 > string2.
--]]

function KSStringUtils.localizedStandardStringCompare( --[[string]] string1, --[[string]] string2 )
	return KSStringUtilsImp.i18n_localized_standard_compare( string1, string2 )
end

--------------------------------------------------------------------------------

--[[!
	 Converts a string to all lowercase characters.

	 \param string  The string.
 	 \return A new string.
--]]

function KSStringUtils.lower( --[[string]] string )
	return KSStringUtilsImp.lower( string )
end

--------------------------------------------------------------------------------

--[[!
	 Converts a string to all uppercase characters.

	 \param string  The string.
 	 \return A new string.
--]]
function KSStringUtils.upper( --[[string]] string )
	return KSStringUtilsImp.upper( string )
end

--------------------------------------------------------------------------------

--[[!
	 Removes whitespace (space and tab characters) from
	 the front and back ends of a string.

	 \param string The string.
 	 \return A new string.
--]]

function KSStringUtils.trim( --[[string]] string )
	return KSStringUtilsImp.trim( string )
end

--------------------------------------------------------------------------------

--[[!
	Encode a KSData as Base64.
	
	\param data input KSData
	\return A new string.
--]]

function KSStringUtils.encodeBase64( --[[KSData]] data )
	return KSStringUtilsImp.encodeBase64( data )
end

--------------------------------------------------------------------------------

--[[!
	Decode a string containing Base64 data into KSData.

	\param string input
	\return A new KSData
--]]

function KSStringUtils.decodeBase64( --[[string]] string )
	return KSStringUtilsImp.decodeBase64( string, 'blob' )
end

--------------------------------------------------------------------------------

--[[!
	Decode a string containing Base64 data into another string.

	\param string input
	\return A new string
--]]

function KSStringUtils.decodeBase64AsString( --[[string]] string )
	return KSStringUtilsImp.decodeBase64( string )
end

--------------------------------------------------------------------------------

--[[!
	 Converts a string that contains HTML code to one that
	 contains escaped characters for the "<", ">", and '&' character.

	 \param s The string.
 	 \return A new string.
--]]

function KSStringUtils.escapeHtml( --[[string]] s )

	-- FIXME [scouten 2012-05-30]: Probably not entirely correct, but good
	-- enough to get us started for diagnostic log ...

	s = tostring( s ) -- just in case we got passed a number or something else

	local result = s:gsub( '&', '&amp;' )
	result = result:gsub( '<', '&lt;' )
	result = result:gsub( '>', '&gt;' )

	return result

end

--------------------------------------------------------------------------------

--[[!
	Converts a string or KSData into hexadecimal.
	
	\param input( string or KSData ) string or KSData to be converted to hex.
	\return str( string ) converted to hexadecimal.

--]]

function KSStringUtils.convertToHex( --[[ string|KSData ]] input )

	assert( KSData.isData( input ) or type( input ) == 'string', "KSStringUtils:convertToHex input should be a KSData or string value" )

	if KSData.isData( input ) then
		input = tostring( input )
	end

	return string.gsub( input,"(.)", 
			function( ch )
				return string.format("%02x",string.byte( ch ) )
			end )

end

--------------------------------------------------------------------------------

return KSStringUtils
