--[[----------------------------------------------------------------------------

KSTableUtils.lua

@author mhamburg 2010-09-24

@reviewed scouten 2010-11-02
@reviewed aksherry 2011-03-14
@reviewed mhamburg 2011-03-14

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

--[[

KSTableUtils:

	function KSTableUtils.makeWithMode( mode )
		-- mode is a weak table mode string (nil, "", "k", "v", or "kv")
		-- result is a table with the appropriate metatable (or no metatable
		-- if mode is nil or the empty string)
		
	function KSTableUtils.makeFullyWeak()
		-- Optimizes for KSTableUtils.makeWithMode( "kv" )
		
	function KSTableUtils.make()
		-- Makes an empty table
		
	function KSTableUtils.makeWithMetatable( mt )
		-- Makes a table with the given metatable
		
	function KSTableUtils.makeFactoryWithMetatable( mt )
		-- Makes a function which makes tables with the given metatable
		
	function KSTableUtils.makeWrapWithMetatable( mt )
		-- Makes a function that wraps its parameter with the given metatable
		
	function KSTableUtils.makeReadOnly( t )
		-- Returns a table that provides a read-only view of t if t is a table.
		-- Otherwise, this just return t.
	
	function KSTableUtils.readOnlyPairs( t )
		-- A version of pairs that is savvy about read-only tables.
		
	function KSTableUtils.mapArrayToSet( [ fn, ] array )
		-- Runs the array through fn (if provided) and builds a set of the
		-- results. Optimized for use with field extractors (see KSFunctionUtils).
		
	function KSTableUtils.merge( base, a, b [ , getMergeAlgorithm ] )
		-- Perform a 3-way recursive table merge from b going to a and b.
		-- Returns the merged result if it succeeds or throws an error if
		-- it fails.
		
		-- The naive merge algorithm is based on a dictionary variation
		-- on set merges where the result of a set merge is calculated
		-- as ( a - base ) + ( b - base ) + ( a * b ) -- i.e., those
		-- elements added in either a or b unioned with the elements that
		-- were in both. In the dictionary case, where we deal with more
		-- than just membership, we treat the keys as the set elements,
		-- but we need to recursively process the values. We treat a
		-- value of nil -- i.e., missing entries -- as the empty table
		-- when merging with the additional behavior that we will map
		-- this promotion away if the resulting merged table is nil.
		-- Furthermore, we allow any base value to be promoted to a pair
		-- of tables and allow for further recursion. We also allow
		-- a base value and a matching successor on one side to lead to
		-- any atomic value on the other side.
		
		-- As a further wrinkle, we also allow the specification of
		-- alternative merge algorithms that may handle the merge
		-- process in other ways. For example, an atomic merge will
		-- insist on equality rather than recursing. We do this via the
		-- optional fourth parameter which can specify a function to determine
		-- the merge algorithm to use based on the path (as an array) and
		-- the target tables. (Signature: path, base, a, b)
		-- It should return one of the following values:
		--		nil, '', or 'dictionary': A standard dictionary merge
		--		'atomic': Treat values atomically using deep equals
		--		function( forPath, base, a, b, mergeForPath ): A custom merge function
				
		-- Finally, this routine guarantees that the result that it
		-- produces will not share any mutable values with its source
		-- material -- i.e., it will perform a full clone.
	
	function KSTableUtils.derefOpt( t, ... )
		-- Dereference a potentially long chain of tables in which any
		-- member might be nil. Use this to avoid the pattern of

			-- local foo = t and t.x and t.x.y and t.x.y.z (etc.)
	
		-- The above could now be written as:
			
			-- local foo = KSTableUtils.derefOpt( t, 'x', 'y', 'z' )

]]

--------------------------------------------------------------------------------

local assert = assert
local string = string
local string_format = string.format
local tostring = tostring
local setmetatable = setmetatable
local pairs = pairs
local ipairs = ipairs
local type = type

local KSFunctionUtils -- = require 'KSFunctionUtils' (lazy loaded)

local WFArguments = require 'WFArguments'
local WFArguments_check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local KSTableUtils = { }

--------------------------------------------------------------------------------

local weakMetas = { }

for _, mode in ipairs( { 'k', 'v', 'kv' } ) do
	weakMetas[ mode ] = {
		__mode = mode,
		__metatable = string_format( "< Weak Metatable: %s >", mode )
	}
end

--------------------------------------------------------------------------------

function KSTableUtils.makeWithMode( mode )

	local result = { }

	if mode == nil or mode == "" then
		return result
	end
	
	local mt = weakMetas[ mode ]
	if not mt then
		error( string_format( "Invalid mode for weak metas: %s", mode ), 2 )
	end
	
	return setmetatable( result, mt )

end

--------------------------------------------------------------------------------

local function makeFullyWeak()

	return setmetatable( { }, weakMetas[ 'kv' ] )

end

--------------------------------------------------------------------------------

KSTableUtils.makeFullyWeak = makeFullyWeak

--------------------------------------------------------------------------------

function KSTableUtils.make()

	return { }

end

--------------------------------------------------------------------------------

function KSTableUtils.makeWithMetatable( mt )

	local result = { }
	
	if mt ~= nil then
		assert( type( mt ) == 'table', "metatable should be a table" )
		setmetatable( result, mt )
	end

	return result

end

--------------------------------------------------------------------------------

function KSTableUtils.makeFactoryWithMetatable( mt )

	if mt == nil then
		return KSTableUtils.make
	end

	assert( type( mt ) == 'table', "metatable should be a table" )

	return function()
		return setmetatable( { }, mt )
	end

end

--------------------------------------------------------------------------------

function KSTableUtils.makeWrapWithMetatable( mt )

	assert( type( mt ) == 'table', "metatable should be a table" )

	return function( t )
		return setmetatable( t, mt )
	end

end

--------------------------------------------------------------------------------

-- Signal an error on attempts to write to a read-only field

local function readOnlyError( t, k, v )
	local err = string_format( "Attempt to write to read-only field '%s'", tostring( k ) )
	error( err )
end

--------------------------------------------------------------------------------

-- A map from the read-only wrapper back to the original. Needed for iteration. (Ugh.)

local readOnlyBase = KSTableUtils.makeFullyWeak()

-- A map from tables to their pre-existing read-only counterparts (to avoid repeated
-- wrapping). Wraps new values appropriately as they come in.

local readOnlyVersion = setmetatable( { }, {
	__mode = 'kv',
	__index = function( t, k )
		if type( k ) == 'table' then
			local v = setmetatable( { }, { __index = k, __newindex = readOnlyError, __metatable = "< read-only table >" } )
			readOnlyBase[ v ] = k
			t[ k ] = v
			return v
		else
			return k
		end
	end
} )

--------------------------------------------------------------------------------

-- Return a table that provides a read-only view of the supplied table.

function KSTableUtils.makeReadOnly( t )

	return t and readOnlyVersion[ t ]

end

--------------------------------------------------------------------------------

function KSTableUtils.readOnlyPairs( t )

	return pairs( ( t and readOnlyBase[ t ] ) or t )
		-- If t isn't nil or false, look for a read-only base table
		-- If we don't have or look for a read-only base table, just use t.

end

--------------------------------------------------------------------------------

-- Test two Lua values to see whether they are equal. Recurses through tables
-- to a depth limit of 10 at which point it decides that there is a problem
-- and reports false (which could be erroneous). It does not look at metatables.
-- Keys in tables are expected to match exactly.

--- deepEquals_helper does the real work... It assumes that
-- a ~= b. This assumption allows us to focus on the recursive case while
-- leaving the common case for higher up in the processing thereby avoiding
-- a function call.

local function deepEquals_helper( a, b, depth, depthLimit, printDebugOutput )

	if type( a ) == 'table' and type( b ) == 'table' then
	
		depth = depth + 1
		
		if depthLimit < depth then
			if printDebugOutput then
				debug.print( "exceeded depth limit at key", printDebugOutput )
			end
			return false
		end
	
		-- All key/value pairs in a must match with b
	
		for k, av in pairs( a ) do
			local bv = b[ k ]
			if av ~= bv and not deepEquals_helper( av, bv, depth, depthLimit,
									printDebugOutput and printDebugOutput .. ":" .. tostring( k ) )
			then
				if printDebugOutput then
					debug.print( "tables mismatched at key ", printDebugOutput .. ":" .. tostring( k ),
						"  left = ", tostring( a[ k ] ),
						"  right = ", tostring( b[ k ] ) )
				end
				return false
			end
		end
		
		-- b must contain no other pairs
		
		for k in pairs( b ) do
			if a[ k ] == nil then
				if printDebugOutput then
					debug.print( "tables mismatched at key ", printDebugOutput .. ":" .. tostring( k ),
						"  left = ", tostring( a[ k ] ),
						"  right = ", tostring( b[ k ] ) )
				end
				return false
			end
		end
		
		return true
	
	end
	
	if printDebugOutput then
		debug.print( "tables mismatched at key ", printDebugOutput,
			"  left = ", tostring( a ),
			"  right = ", tostring( b ) )
	end
	
	return false

end

--------------------------------------------------------------------------------

local function deepEquals( a, b, depthLimit, printDebugOutput )

	if a == b then
		return true
	end
	
	depthLimit = depthLimit or 10
	
	if printDebugOutput then
		printDebugOutput = ""
	end
	
	return deepEquals_helper( a, b, 1, depthLimit, printDebugOutput )

end

--------------------------------------------------------------------------------

KSTableUtils.deepEquals = deepEquals

--============================================================================--

local kMaxLevels = 20

local function deepCopyImp( v, depth, copied )

	-- Copied from Lightroom's AgTableUtils.deepCopy and deepCopyImp.

	if v == nil then
		return v
	end

	if kMaxLevels < depth then
		error( "Depth limit exceeded for deepCopy." )
	end

	local tv = type( v )
	
	-- Userdata is not supported.
	
	if tv == 'userdata' then
		error( "deepCopy does not support userdata: " .. tostring( v ) )
	end
	
--	if isNilSentinel( v ) then
--		return v	-- This is immutable (and has a mt), so handle it specially.
--	end
	
	-- Copy tables an entry at a time
	
	if tv == 'table' then
	
		if getmetatable( v ) ~= nil then
			error( "deepCopy does not support metatables" )
		end
	
		local vcopy = copied[ v ]
		
		if vcopy ~= nil then
			return vcopy
		end
		
		vcopy = {}
		
		copied[ v ] = vcopy
			-- Generate the link back now
		
		for k, v2 in pairs( v ) do
			local ck = deepCopyImp( k, depth + 1, copied )
			local cv = deepCopyImp( v2, depth + 1, copied )
			vcopy[ ck ] = cv
		end
		
		return vcopy
	
	end
	
	-- All other cases are handled by just returning the value
	
	return v

end

--------------------------------------------------------------------------------

local function deepCopy( v )

	local tv = type( v )
	
	if tv == 'table' then
		return deepCopyImp( v, 0, { } )
	end
	
	if tv == 'userdata' then
		error( "deepCopy does not support userdata: " .. tostring( v ) )
	end

	return v

end

--------------------------------------------------------------------------------

KSTableUtils.deepCopy = deepCopy

--============================================================================--

local function mapArrayToSet( mapFn, array )

	if array == nil and type( mapFn ) == 'table' then
		return mapArrayToSet( nil, mapFn )
	end

	local result = { }

	if not KSFunctionUtils then
		KSFunctionUtils = assert( require 'KSFunctionUtils' )
	end
	
	-- Handle the case where we have no map to apply.

	if mapFn == nil or mapFn == KSFunctionUtils.identity then
	
		for i = 1, #array do
			local v = array[ i ]
			result[ v ] = v or true
		end
		
		return result
		
	end
	
	local processed = { }
		-- Track processed items so that we only process once
		
	-- Special case field extraction to avoid the function call overhead.

	local isFieldExtractor, fieldSelector = KSFunctionUtils.isFieldExtractorFn( mapFn )
	
	if isFieldExtractor then
		
		for _, v in ipairs( array ) do
			if not processed[ v ] then
				processed[ v ] = true
				local mappedV = v[ fieldSelector ]
				if mappedV ~= nil then
					result[ mappedV ] = mappedV or true
				end
			end
		end
		
		return result
		
	end
	
	-- Handle everything else
	
	for _, v in ipairs( array ) do
		if not processed[ v ] then
			processed[ v ] = true
			local mappedV = mapFn( v )
			if mappedV ~= nil then
				result[ mappedV ] = mappedV or true
			end
		end
	end
	
	return result

end

KSTableUtils.mapArrayToSet = mapArrayToSet

--============================================================================--

-- FIXME [mhamburg 2011-08-02]: The merge algorithm could really use a depth
-- limit. We could impose this in the dictMergeForPath code since we have the
-- path length there.

--------------------------------------------------------------------------------

-- Convert a path to a string for use in error messages. Just returns the
-- trailing element for now.

local function pathToString( path )

	if type( path ) == 'table' then
		return pathToString( path[ #path ] )
	end
	
	if path == nil then
		return '< none >'
	end
	
	return tostring( path )

end

--------------------------------------------------------------------------------

local function atomicMergeForPath( forPath, base, a, b, mergeForPath )

	WFArguments_check( 'table' )
	
	if deepEquals( a, b ) then
		return deepCopy( a )
	end
	
	if deepEquals( base, a ) then
		return deepCopy( b )
	end
	
	if deepEquals( base, b ) then
		return deepCopy( a )
	end
	
	error( "Atomic merge failed for key: " .. pathToString( forPath ) )

end

--------------------------------------------------------------------------------

local function dictMergeForPath( forPath, base, a, b, mergeForPath )

	WFArguments_check( 'table', 'table', 'table', 'table', 'function' )
	
	local r = { }
	
	-- Process all of the keys in a.
	
	local pathLength = #forPath + 1
	
	for k, v in pairs( a ) do
		forPath[ pathLength ] = k
		r[ k ] = mergeForPath( forPath, base[ k ], v, b[ k ] )
		forPath[ pathLength ] = nil
	end
	
	-- Process the keys in b that weren't in a
	
	for k, v in pairs( b ) do
		if a[ k ] == nil then
			forPath[ pathLength ] = k
			r[ k ] = mergeForPath( forPath, base[ k ], nil, v )
			forPath[ pathLength ] = nil
		end
	end
	
	-- That's everything
	
	return r

end

--------------------------------------------------------------------------------

local gMergeAlgorithms = {
	dictionary = dictMergeForPath,
	atomic = atomicMergeForPath
}

--------------------------------------------------------------------------------

-- Map empty tables to nil. Return other tables. Expects the parameter to be
-- a table.

local function mapEmptyToNil( t )

	if next( t ) == nil then
		return nil
	end
	
	return t

end

--------------------------------------------------------------------------------

local mergeFunctionsCache = makeFullyWeak()

--------------------------------------------------------------------------------

local function makeMergeFunction( getMergeAlgorithm )

	local cacheKey = getMergeAlgorithm or "< none >"
	
	local fn = mergeFunctionsCache[ cacheKey ]
	
	if fn then
		return fn
	end
	
	----------------------------------------------------------------------------
	
	local mergeForPath -- forward declaration
	
	----------------------------------------------------------------------------
	
	-- tableMergeForPath figures out which merge algorithm to use and then
	-- invokes it passing in mergeForPath to support appropriate recursion.
	
	local tableMergeForPath
	
	if getMergeAlgorithm then
	
		tableMergeForPath = function( forPath, base, a, b )
	
			WFArguments_check( 'table', 'table', 'table', 'table' )
			
			-- Figure out the merge algorithm to use.
			
			local algorithm = getMergeAlgorithm( forPath, base, a, b )
				
			if algorithm == nil or algorithm == '' then
			
				algorithm = dictMergeForPath
				
			else
			
				if type( algorithm ) == 'string' then
				
					algorithm = gMergeAlgorithms[ algorithm ]
					
					if not algorithm then
						error( "Unknown merge algorithm: " .. algorithm )
					end

				end
				
				assert( type( algorithm ) == 'function', "Merge algorithm needs to be a function" )
			
			end
			
			return algorithm( forPath, base, a, b, mergeForPath )
			
		end
		
	else
	
		tableMergeForPath = function( forPath, base, a, b )
		
			return dictMergeForPath( forPath, base, a, b, mergeForPath )
		
		end
	
	end

	----------------------------------------------------------------------------
	
	mergeForPath = function( forPath, base, a, b )
	
		-- Handle the easy equality cases that should behave the
		-- same regardless of algorithm.
		
		if a == b then
			return deepCopy( a )
		end
		
		if base == a then
			return deepCopy( b )
		end
		
		if base == b then
			return deepCopy( a )
		end
		
		-- Check the value types to see if it's a case we can handle via
		-- recursion.
		
		local type_base = type( base )
		local type_a = type( a )
		local type_b = type( b )
		
		-- Tables all across.
		
		if type_base == 'table' and type_a == 'table' and type_b == 'table' then
			return tableMergeForPath( forPath, base, a, b )
		end
		
		-- Any non-table value type promoted to two tables results in a recursion
		-- with an empty base table.
		
		if type_a == 'table' and type_b == 'table' then
			-- Pretend we had an empty base entry and merge from there. We won't
			-- be going back to nil, because both of the values are non-nil.
			return tableMergeForPath( forPath, { }, a, b )
		end
		
		-- Moving from a table to a table on one branch and to nil on the other
		-- is handled by treating the nil as an empty table and then checking
		-- to see whether the merge result is empty.
		
		if type_base == 'table' and type_a == 'nil' and type_b == 'table' then
			return mapEmptyToNil( tableMergeForPath( forPath, base, { }, b ) )
		end
		
		if type_base == 'table' and type_a == 'table' and type_b == 'nil' then
			return mapEmptyToNil( tableMergeForPath( forPath, base, a, { } ) )
		end
		
		-- We can also handle the case where base is a table and one of the targets
		-- is an identical table by checking for deep equality.
		
		if type_base == 'table' and ( type_a == 'table' or type_b == 'table' ) then
			return atomicMergeForPath( forPath, base, a, b )
		end
		
		-- Otherwise, we're in a case we can't handle via recursion and equality
		-- tests failed above, so error out.
		
		local pathName = pathToString( forPath )
		
		error( string_format(
				"Types do not support recursive merge %s -> %s and %s for path: %s",
				type_base, type_a, type_b, pathName ) )
				
	end

	----------------------------------------------------------------------------
	
	mergeFunctionsCache[ cacheKey ] = mergeForPath
	
	return mergeForPath
	
end

--------------------------------------------------------------------------------

function KSTableUtils.merge( base, a, b, getMergeAlgorithm )

	local mergeForPath = makeMergeFunction( getMergeAlgorithm )
	
	return mergeForPath( { }, base, a, b )

end

--------------------------------------------------------------------------------

function KSTableUtils.derefOpt( t, ... )

	local path = { ... }
	local nPathSegments = select( '#', ... )
	
	local result = t
	
	for i = 1, nPathSegments do
	
		if type( result ) ~= 'table' then
			return nil
		end
		
		if path[ i ] == nil then
			return nil
		end
		
		result = result[ path[ i ] ]
	
	end
	
	return result

end

--------------------------------------------------------------------------------

return KSTableUtils
