--[[----------------------------------------------------------------------------

KSTableUtilsTest.lua

@author mhamburg 2010-09-24

@reviewed scouten 2010-10-28
@reviewed epoley 2011-03-08
@reviewed aksherry 2011-03-14
@reviewed mhamburg 2011-03-14

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2010 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local T = { }

--------------------------------------------------------------------------------

local assertf = require 'KSLuaUtils'.assertf

local WFArguments = require 'WFArguments'

--------------------------------------------------------------------------------

local function checkMetaForMode( t, mode )

	assert( type( t ) == 'table', "Expected table" )

	local mt = debug.getmetatable( t )
	
	assert( mt, "No metatable" )
	
	assertf( mt.__mode == mode,
		    "Wrong mode in metatable. Expected %s got %s.",
			mode, mt.__mode )

end

--------------------------------------------------------------------------------

function T.testMakeWithMode()

	local TU = require 'KSTableUtils'
	
	for _, mode in ipairs( { 'k', 'v', 'kv' } ) do
		checkMetaForMode( TU.makeWithMode( mode ), mode )
	end
	
	assert( debug.getmetatable( TU.makeWithMode() ) == nil,
		"Unexpected metatable in no parameter case" )
		
	assert( debug.getmetatable( TU.makeWithMode( '' ) ) == nil,
		"Unexpected metatable in no empty string case" )
		
	assert( pcall( TU.makeMetaWithMode, 'foo' ) == false,
		"Expected failure for bad mode string" )
		
	assert( pcall( TU.makeMetaWithMode, { } ) == false,
		"Expected failure for non-string mode" )
		
end

--------------------------------------------------------------------------------

function T.testMakeFullyWeak()

	local TU = require 'KSTableUtils'
	
	checkMetaForMode( TU.makeFullyWeak(), 'kv' )
	
	assert( TU.makeFullyWeak() ~= TU.makeFullyWeak(), "Unexpected table recycling" )
		
end

--------------------------------------------------------------------------------

function T.testMake()

	local TU = require 'KSTableUtils'
	
	local t = TU.make()
	
	assert( type( t ) == 'table', "make should produce a table" )
	
	assert( debug.getmetatable( t ) == nil, "That table should not have a metatable" )
	
	assert( TU.make() ~= TU.make(), "Unexpected table recycling" )
		
end

--------------------------------------------------------------------------------

function T.testMakeWithMetatable()

	local TU = require 'KSTableUtils'
	
	local mt = { }
	
	local t = TU.makeWithMetatable( mt )
	
	assert( type( t ) == 'table', "makeWithMetatable should produce a table" )
	
	assert( debug.getmetatable( t ) == mt, "That table should have the correct metatable" )
	
	assert( TU.makeWithMetatable( mt ) ~= TU.makeWithMetatable( mt ), "Unexpected table recycling" )
	
	assert( pcall( TU.makeWithMetatable, 'foo' ) == false,
		"makeWithMetatable should reject non tables as metatables" )
		
end

--------------------------------------------------------------------------------

function T.testMakeFactoryWithMetatable()

	local TU = require 'KSTableUtils'
	
	local mt = { }
	
	local factory = TU.makeFactoryWithMetatable( mt )
	
	-- We don't test that factory is a function because what we really care
	-- about is that it is callable which we test by calling it.
	
	local t = factory()
	
	assert( type( t ) == 'table', "makeWithMetatable should produce a table" )
	
	assert( debug.getmetatable( t ) == mt, "That table should have the correct metatable" )
	
	assert( factory() ~= factory(), "Unexpected table recycling" )
	
	assert( pcall( TU.makeFactoryWithMetatable, 'foo' ) == false,
		"makeFactoryWithMetatable should reject non tables as metatables" )
		
end

--------------------------------------------------------------------------------

function T.testMakeWrapWithMetatable()

	local TU = require 'KSTableUtils'
	
	local mt = { }
	
	local wrapper = TU.makeWrapWithMetatable( mt )
	
	-- We don't test that wrapper is a function because what we really care
	-- about is that it is callable which we test by calling it.
	
	local t = { }
	
	local tt = wrapper( t )
	
	assert( t == tt, "wrapper should not change table identity" )
	
	assert( debug.getmetatable( t ) == mt, "That table should have the correct metatable" )
	
	local mt2 = { }
	
	local wrapper2 = TU.makeWrapWithMetatable( mt2 )
	
	local t2 = { }
	
	local tt2 = wrapper2( t2 )
	
	assert( t2 == tt2, "wrapper should not change table identity (2)" )
	
	assert( debug.getmetatable( t2 ) == mt2, "That table should have the correct metatable (2)" )
	
	local t3 = { }
	
	local tt3 = wrapper( t3 )
	
	assert( t3 == tt3, "wrapper should not change table identity (3)" )
	
	assert( debug.getmetatable( t3 ) == mt, "That table should have the correct metatable (3)" )
	
	assert( pcall( TU.makeWrapWithMetatable, 'foo' ) == false,
		"makeWrapWithMetatable should reject non tables as metatables" )
		
	assert( pcall( TU.makeWrapWithMetatable, nil ) == false,
		"makeWrapWithMetatable should reject nil metatables" )
		
end

--------------------------------------------------------------------------------

function T.testMakeReadOnly()

	local TU = require 'KSTableUtils'

	local makeReadOnly = assert( TU.makeReadOnly )
	
	for _, v in ipairs{ 1, "a", true, false } do
		assert( makeReadOnly( v ) == v, "makeReadOnly should preserve non-table values" )
	end
	
	assert( makeReadOnly( nil ) == nil, "makeReadOnly should preserve nil" )
	
	local t = { foo = 42, baz = 'hello' }
	
	local tt = makeReadOnly( t )
	
	for k, v in pairs( t ) do
		assert( tt[ k ] == v, "read-only table should match source values" )
	end
	
	for k in TU.readOnlyPairs( tt ) do
		assert( t[ k ] ~= nil, "read-only table should have only the same values as the source" )
	end
	
	assert( pcall( function() tt.x = 9 end ) == false, "read-only table should reject writes" )
	
	assert( t.x == nil, "Write should not have changed base table." )
	
	assert( tt.x == nil, "Write should not have changed read-only table." )
	
	assert( pcall( function() tt.foo = 54 end ) == false, "read-only table should reject writes to existing fields" )
	
	assert( t.foo == 42, "Write should not have changed base table." )
	
	assert( tt.foo == 42, "Write should not have changed read-only table." )
	
	assert( pcall( function() t.foo = 54 end ) == true, "Write to base table should succeed" )
	
	assert( tt.foo == 54, "Write to base table should be visible in read-only table" )
	
	assert( tt == makeReadOnly( t ), "Wrappers should be shared" )

end

--------------------------------------------------------------------------------

function T.testMapArrayToSetIdentityCase()

	local TU = require 'KSTableUtils'
	
	assert( TU.deepEquals(
				TU.mapArrayToSet{ foo = true, 'a', 'b', false, 5 },
				{ a = 'a', b = 'b', [ false ] = true, [ 5 ] = 5 } ),
			"Incorrect set computed" )

end

--------------------------------------------------------------------------------

function T.testMapArrayToSetFieldExtractorCase()

	local TU = require 'KSTableUtils'
	local FU = require 'KSFunctionUtils'
	
	local mapped = TU.mapArrayToSet(
		FU.fieldExtractorFn( 'id' ),
		{ { id = 'a' }, { id = 'b' }, { id = false }, { no_id = true }, { id = 5 } } )
	
	assert( TU.deepEquals(
				mapped,
				{ a = 'a', b = 'b', [ false ] = true, [ 5 ] = 5 } ),
			"Incorrect set computed" )

end

--------------------------------------------------------------------------------

function T.testMapArrayToSetFunctionCase()

	local TU = require 'KSTableUtils'
	
	assert( TU.deepEquals(
				TU.mapArrayToSet(
					function( x ) return #x end,
					{ "cat", "dog", "cat", "bird", "gorilla" } ),
				{ [ 3 ] = 3, [ 4 ] = 4, [ 7 ] = 7 } ),
			"Incorrect set computed" )

end

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2010-11-16]: We need unit tests for deepEquals
-- FIXME [mhamburg 2011-03-25]: We need unit tests for deepCopy

--------------------------------------------------------------------------------

local function successfulMergeTest( name )
	return function( p )
		T[ name ] = function()
			local TU = require 'KSTableUtils'
			local dc = function( key ) return TU.deepCopy( p[ key ] ) end
			local getMergeAlgorithm = p.getMergeAlgorithm
			assert( TU.deepEquals( TU.merge( dc 'base', dc 'a', dc 'b' , getMergeAlgorithm ),
								dc 'merged', nil, true ),
				"Incorrect merge results (1)" )
			if not p.assymetric then
				assert( TU.deepEquals( TU.merge( dc 'base', dc 'b', dc 'a', getMergeAlgorithm ),
								dc 'merged', nil, true ),
					"Incorrect merge results (2)" )
				-- Also test a recursive version
				assert( TU.deepEquals( TU.merge(
							{ rec = dc 'base', rec2 = dc 'base' },
							{ rec = dc 'a', rec2 = dc 'b' },
							{ rec = dc 'b', rec2 = dc 'a' },
							p.getMergeAlgorithm ),
					{ rec = dc 'merged', rec2 = dc 'merged' }, nil, true ),
					"Recursive merge results were wrong" )
			end
		end
	end
end

--------------------------------------------------------------------------------

local function failureMergeTest( name )
	return function( p )
		T[ name ] = function()
			local TU = require 'KSTableUtils'
			assert( false == pcall( TU.merge, p.base, p.a, p.b, p.getMergeAlgorithm ),
				"Expected merge to fail (1)" )
			assert( false == pcall( TU.merge, p.base, p.b, p.a, p.getMergeAlgorithm ),
				"Expected merge to fail (2)" )
		end
	end
end

--------------------------------------------------------------------------------

successfulMergeTest "testSimpleMergeAB" {
	base = "foo", a = "baz", b = "baz", merged = "baz"
}

--------------------------------------------------------------------------------

successfulMergeTest "testSimpleMergeBaseA" {
	base = "foo", a = "foo", b = "baz", merged = "baz"
}

--------------------------------------------------------------------------------

successfulMergeTest "testTableTableAtomicMerge" {
	base = { }, a = { }, b = "foo", merged = "foo"
}

--------------------------------------------------------------------------------

-- A real use would memoize the key match, but we don't bother with that
-- given that this is for unit tests.

local function getMergeAlgorithm_atomic( forPath, base, a, b )

	WFArguments.check( 'table', 'table', 'table', 'table' )
	
	local lastKey = forPath[ #forPath ]
	
	if type( lastKey ) == 'string' and string.find( lastKey, '_atomic$' ) then
		return 'atomic'
	end
	
	-- Otherwise just fall through

end

--------------------------------------------------------------------------------

successfulMergeTest "testTableTableTableMerge" {
	base = {
		core = 0,
		foo = 1,
		baz = 2,
		frex = 7,
		dimple = "dent"
	},
	a = {
		core = 0,
		foo = 1,
		baz = "Hello",
		frex = 11,
		bleen = "Hmmm",
		brax = "Huh?"
	},
	b = {
		core = 0,
		foo = "Green",
		baz = 2,
		frex = 11,
		dimple = "dent",
		bleen = "Hmmm"
	},
	merged = {
		core = 0, 	-- Same in all
		foo = "Green", -- b
		baz = "Hello", -- a
		frex = 11, -- Changed in both
		bleen = "Hmmm", -- both (but new)
		-- dimple deleted in a
		brax = "Huh?", -- new in a
	}
}

--------------------------------------------------------------------------------

successfulMergeTest "testMergeWithAtomic" {
	base = {
		core = 0,
		foo = 1,
		baz = 2,
		frex = 7,
		dimple = "dent",
		smash = { so = "cool" },
		spatter_atomic = { so = "cool" }
	},
	a = {
		core = 0,
		foo = 1,
		baz = "Hello",
		frex = 11,
		bleen = "Hmmm",
		brax = "Huh?",
		smash = { so = "cool", yet = "not" },
		flump_atomic = { "big", "bigger" },
		spatter_atomic = { so = "cool" }
	},
	b = {
		core = 0,
		foo = "Green",
		baz = 2,
		frex = 11,
		dimple = "dent",
		bleen = "Hmmm",
		flump_atomic = { "big", "bigger" }
	},
	merged = {
		core = 0, 	-- Same in all
		foo = "Green", -- b
		baz = "Hello", -- a
		frex = 11, -- Changed in both
		bleen = "Hmmm", -- both (but new)
		-- dimple deleted in a
		brax = "Huh?", -- new in a
		smash = { yet = "not" }, -- one key deleted in b but not both
		flump_atomic = { "big", "bigger" } -- new in both and atomic match
		-- spatter_atomic deleted in b
	},
	getMergeAlgorithm = getMergeAlgorithm_atomic
}

--------------------------------------------------------------------------------

failureMergeTest "testSimpleMergeFailure" {
	base = "foo", a = "baz", b = 3
}

--------------------------------------------------------------------------------

failureMergeTest "testTableTableAtomicMergeFailure" {
	base = { }, a = { entry = 'Hello' }, b = 3
}

--------------------------------------------------------------------------------

failureMergeTest "testTableTableAtomicMergeFailure2" {
	base = { entry = 'Hello' }, a = { entry = 'Howdy' }, b = 3
}

--------------------------------------------------------------------------------

failureMergeTest "testMergeFailureWithAtomic" {
	base = {
		core = 0,
		foo = 1,
		baz = 2,
		frex = 7,
		dimple = "dent",
		smash = { so = "cool" },
		spatter_atomic = { so = "cool" }
	},
	a = {
		core = 0,
		foo = 1,
		baz = "Hello",
		frex = 11,
		bleen = "Hmmm",
		brax = "Huh?",
		smash = { so = "cool", yet = "not" },
		flump_atomic = { "big", "bigger" },
		spatter_atomic = { so = "cool", yet = "not" } -- atomic mismatch
	},
	b = {
		core = 0,
		foo = "Green",
		baz = 2,
		frex = 11,
		dimple = "dent",
		bleen = "Hmmm",
		flump_atomic = { "big", "bigger" }
	},
	getMergeAlgorithm = getMergeAlgorithm_atomic
}

--------------------------------------------------------------------------------

local function derefOptTest( name )

	return function( p )

		T[ name ] = function()

			local TU = require 'KSTableUtils'
			local src = TU.deepCopy( p.src )
			local actualResult = TU.derefOpt( src, unpack( p.path, 1, #p.path ) )
			local expectedResult = p.result

			if type( actualResult ) == 'table' and type( expectedResult ) == 'table' then
				assert( TU.deepEquals( actualResult, expectedResult ),
						"Incorrect results" )
			else
				assert( actualResult == expectedResult,
						string.format( "Incorrect result: was %s, expected %s",
									   tostring( actualResult ),
									   tostring( expectedResult ) ) )
			end

		end

	end

end

--------------------------------------------------------------------------------

derefOptTest "testDerefOpt_emptyTable" {
	
	src = {},
	path = { 'a', 'b', 'c' },
	result = nil,
	
}

--------------------------------------------------------------------------------

derefOptTest "testDerefOpt_threeDeep" {
	
	src = {
		a = {
			b = {
				c = 42,
			},
		},
	},
	path = { 'a', 'b', 'c' },
	result = 42,
	
}

--------------------------------------------------------------------------------

derefOptTest "testDerefOpt_partialHierarchy" {
	
	src = {
		a = {},
	},
	path = { 'a', 'b', 'c' },
	result = nil,
	
}

--------------------------------------------------------------------------------

derefOptTest "testDerefOpt_nonTableInMiddle" {
	
	src = {
		a = {
			b = 'foo',
		},
	},
	path = { 'a', 'b', 'c' },
	result = nil,
	
}

--------------------------------------------------------------------------------

derefOptTest "testDerefOpt_notATable" {
	
	src = 56,
	path = { 'a', 'b', 'c' },
	result = nil,
	
}

--------------------------------------------------------------------------------

function T.testDerefOpt_nilInPath()

	local TU = require 'KSTableUtils'

	local src = {
		a = {
			b = {
				c = 42,
			},
		},
	}

	local result = TU.derefOpt( src, 'a', nil, 'c' )

	assert( result == nil )

end

--------------------------------------------------------------------------------

function T.testDerefOpt_nilPath()

	local TU = require 'KSTableUtils'

	local src = {
		a = {
			b = {
				c = 42,
			},
		},
	}

	local result = TU.derefOpt( src, nil )

	assert( result == nil )

end

--------------------------------------------------------------------------------

return T
