--[[----------------------------------------------------------------------------

KSUndoManager.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author erwright 2011-02-05

@reviewed mhamburg 2011-03-11

--------------------------------------------------------------------------------

	KSUndoManager is a simple helper for tracking undoable changes.
	We could consider reusing Lightroom's AgUndo.lua, but we would need to
	disentangle it from its Substrate dependencies.

------------------------------------------------------------------------------]]

local KSClass = require 'KSClass'

local KSReactive = require 'KSReactive'

local assert = assert
local type = type


local KSUndoManager = { }

--------------------------------------------------------------------------------

-- the current KSTransaction to add undo/redo functions to
local activeTransaction = nil

-- the lists of KSTransactions that can be undone/redone
local undoList = { }
local redoList = { }

-- the reactive variables used by canUndoB and canRedoB
local mutableCanUndoB = KSReactive.variableB( false )
local mutableCanRedoB = KSReactive.variableB( false )

--------------------------------------------------------------------------------

local transactionClass = KSClass 'TransactionClass' {

	__init = function( self, name )
		assert( name )
		assert( type( name ) == "string" )
		self.name = name
		self._undoFunctionList = { }
	end,

	name = "missing name",
	
	_undoFunctionList = false,
	
	add = function( self, undoFunction, redoFunction )

		assert( undoFunction )
		assert( redoFunction )
		assert( type( undoFunction ) == "function" )
		assert( type( redoFunction ) == "function" )

		self._undoFunctionList[ #self._undoFunctionList + 1 ] = { _undo = undoFunction, _redo = redoFunction }
		
	end,
	
	undo = function( self )
		for i = #self._undoFunctionList, 1, -1 do
			self._undoFunctionList[ i ]._undo()
		end
	end,
	
	redo = function( self )
		for i = 1, #self._undoFunctionList, 1 do
			self._undoFunctionList[ i ]._redo()
		end
	end,
}

--------------------------------------------------------------------------------

local function updateUndoRedoVariables()

	-- FIXME [erwright 2011-02-22]: how do we make these variables more reactive
	-- so that we don't have to explicitly update their values like this?
	
	-- FIXME [mhamburg 2011-03-11]: Active transaction could probably be handled
	-- via an existing reactive variable and an appropriate combiner function.
	-- We need reactive lists or the like to handle the other case since otherwise
	-- object identity is likely to nail us. Or we need to address the object
	-- identity problem...

	local canUndo = not activeTransaction and #undoList > 0
	local canRedo = not activeTransaction and #redoList > 0
	
	mutableCanUndoB:set( canUndo )
	mutableCanRedoB:set( canRedo )

end

--------------------------------------------------------------------------------

-- only one transaction may be active at a time
-- consider nested transactions later

function KSUndoManager:beginTransaction( name )

	assert( not activeTransaction )
	activeTransaction = transactionClass( name )
	
	updateUndoRedoVariables()
	
end

--------------------------------------------------------------------------------

-- commit() only succeeds if there is an active transaction, otherwise no-op

function KSUndoManager:commit()

	if activeTransaction then
	
		undoList[ #undoList + 1 ] = activeTransaction
		activeTransaction = nil
		
		updateUndoRedoVariables()
		
	end
	
end

--------------------------------------------------------------------------------

-- rollback() only succeeds if there is an active transaction, otherwise no-op

-- FIXME [erwright 2011-03-15]: scouten pointed out that this is behaving more like
-- "abandon" than rollback.  This function should invoke any undo units already
-- registered in the current transaction before releasing the transaction.

function KSUndoManager:rollback()

	activeTransaction = nil

	updateUndoRedoVariables()

end


--------------------------------------------------------------------------------

function KSUndoManager:isTransactionActive()

	return activeTransaction ~= nil
	
end

--------------------------------------------------------------------------------

-- Adds an undo and redo function to the current transaction's list of undoable actions.
-- It's up to the caller to execute the redoFunction, either before or after adding it.
-- Fails if there is no activeTransaction.

function KSUndoManager:addUndoRedoFunctions( undoFunction, redoFunction )

	assert( activeTransaction )
	activeTransaction:add( undoFunction, redoFunction )

	updateUndoRedoVariables()

end

--------------------------------------------------------------------------------

function KSUndoManager:canUndoB()

	-- We get burned by the binding code being aggressive about closing the
	-- binding. This needs be fixed, but the simple fix would hurt us on
	-- other cancelations, so for now we just tear off a new external view.
	
	-- FIXME [mhamburg 2011-02-28]

	return mutableCanUndoB:relayB()

end

--------------------------------------------------------------------------------

function KSUndoManager:canRedoB()

	-- We get burned by the binding code being aggressive about closing the
	-- binding. This needs be fixed, but the simple fix would hurt us on
	-- other cancelations, so for now we just tear off a new external view.
	
	-- FIXME [mhamburg 2011-02-28]

	return mutableCanRedoB:relayB()

end

--------------------------------------------------------------------------------

-- fails if a transaction is active, no-op if there is nothing to undo

function KSUndoManager:undo()

	assert( KSUndoManager:canUndoB():valueNow() )
	
	local transaction = undoList[ #undoList ]
	undoList[ #undoList ] = nil
	transaction:undo()
	redoList[ #redoList + 1 ] = transaction
	
	updateUndoRedoVariables()

end

--------------------------------------------------------------------------------

-- fails if a transaction is active, no-op if there is nothing to redo

function KSUndoManager:redo()

	assert( KSUndoManager:canRedoB():valueNow() )
	
	local transaction = redoList[ #redoList ]
	redoList[ #redoList ] = nil
	transaction:redo()
	undoList[ #undoList + 1 ] = transaction
	
	updateUndoRedoVariables()

end

--------------------------------------------------------------------------------

-- Removes all undo and redo units from the stack.  Cannot be called if we're in the
-- midst of a transaction.

function KSUndoManager:clearStack()

	assert( not activeTransaction )

	undoList = { }
	redoList = { }
	updateUndoRedoVariables()
	
end

--------------------------------------------------------------------------------

return KSUndoManager
