--[[----------------------------------------------------------------------------

OzAccountService.lua

@author scouten 2011-06-09

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.string
local WFZString = require 'WFZString'

	-- libs.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- libs.prefs
local KSPrefs = require 'KSPrefs'

	-- os.localization
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- oz.account
local OzUser = require 'OzUser'

	-- oz.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- oz.doc_store
local OzDocStore = require 'OzDocStore'
	
	-- wichitafoundation.oz_client
local Bablu = require "Bablu"

	-- wichitafoundation.web.http_client
local WFHttpClient = require 'WFHttpClient'

	-- services.oauth2_client
local KSOAuth2Session = require 'KSOAuth2Session'


--------------------------------------------------------------------------------

-- This doc store is used as a local cache for user profile information.

local userInfoCache
if KSPrefs.currentUserId and type( KSPrefs.currentUserId ) == 'string' and #KSPrefs.currentUserId == 32 then
	userInfoCache = OzDocStore( KSPrefs.currentUserId )
end

local accountLog = KSLogging.log 'OzAccountService'
--	:setLevel( KSLogging.DEBUG )

local ZSTR = assert( WFZString.fromZString )

local notLoggedInStr = ZSTR "$$$/OzAccountService/User/NotLoggedIn=(not logged in)"


--============================================================================--

-- @mark OzAccountService namespace

--[[!
	\class OzAccountService

	\brief This namespace manages the information about what Oz/IMS user accounts (logins)
	are known to an application.
	
	You can call these functions from any thread.
]]--

local OzAccountService = {}

--------------------------------------------------------------------------------

local currentUserB = KSPrefs.prefB( 'currentUserId' )
	:mapB( function( ozUserId )

		accountLog:info( "new currentUserId = %s", ozUserId )

		if ozUserId then
			return OzUser.make( ozUserId )
		else
			return nil
		end
	
	end )

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the current user object. If any user is
		currently logged in, the user object is available immediately upon return.
		Note that behaviors within the user object are not populated synchronously.
	
	\return <code>B< OzUser|nil ></code> A Behavior whose value is the
		current user object, or nil if no user is logged in.
]]--

function OzAccountService.currentUserB()

	return currentUserB:relayB()

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Promise) a snapshot of information about a user who
		is currently logged in. It is an error to call this function when no user
		is logged in.

	\param userId (optional, string) The user's unique ID. Default is the next
		user who is logged in.
	
	\return <code>P< table ></code> A Promise that resolves to the information packet.
		The looks like this (with some additional fields that are not guaranteed over time):
	
	<code>
	{
		country = "US",
		email = "scouten@adobe.com",
		expires_in = 86399991,
		first_name = "Eric",
		last_name = "Scouten",
		mrktPerm = "PHONE_RENT:false,EMAIL_RENT:false,MAIL:false,FAX:false,MAIL_RENT:false,PHONE:false,EMAIL:false",
		full_name = " Eric  Scouten ",
		token_type = "bearer",
		userId = "0824042C4DFBA2710A490D16@AdobeID",
		utcOffset = "null",
	}
	</code>
]]--

function OzAccountService.userInfoP( --[[string]] userId )

	if not userId then
		error( "OzAccountService.userInfoP: no userId specified", 2 )
	end

	return userInfoCache:docValueOnceP( 'user', userId )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Behavior) the current user's full name.

	\return <code>B< string ></code> A Behavior whose value is the
		the name, or if no user is logged in, a localized version
		of the string "(not logged in)".
]]--

function OzAccountService.currentUserFullNameB()

	return OzAccountService.currentUserB():mcallSwitchB( 'fullNameB', notLoggedInStr )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Behavior) the current user's e-mail address.

	\return <code>B< string ></code> A Behavior whose value is the
		the e-mail address, or if no user is logged in, an empty string.
 ]]--

function OzAccountService.currentUserEmailB()

	return OzAccountService.currentUserB():mcallSwitchB( 'emailB', "" )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Behavior) the current user's account status.

	\return <code>B< string ></code> A Behavior whose value is the
		the status, or if no user is logged in, an empty string.
 ]]--

function OzAccountService.currentUserAccountStatusB()

	return OzAccountService.currentUserB():mcallSwitchB( 'accountStatusB', "" )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Behavior) the current user's catalog list.

	\return <code>B< nil|list< catalogIds > ></code> A Behavior whose value is the
		list of catalog IDs, or  nil if no user is logged in.
 ]]--

function OzAccountService.currentUserCatalogsB()

	return OzAccountService.currentUserB():mcallSwitchB( 'catalogsB', nil )

end

--------------------------------------------------------------------------------
--[[!
	\brief Makes an Oz user the currently active user, or disables all users.
 
 	\param ozUserId (string) The user's unique ID, or nil to disable all users.
		You must already have provided authorization credentials for this user.
	
	\sa :storeNewUserCredentials
 ]]--

function OzAccountService.setCurrentUser( --[[string]] ozUserId )

	if ozUserId and ( type( ozUserId ) ~= 'string' or #ozUserId ~= 32 ) then
		error( "OzAccountService.setCurrentUser: ozUserId is invalid (must be UUID or nil)", 2 )
	end
	
	KSPrefs.currentUserId = ozUserId

end

--------------------------------------------------------------------------------

local HACK_keepOAuth2SessionAlive
	-- FIXME [scouten 2011-11-22]: Remove this when #3045522 is fixed, which is
	-- not likely for 1.1. See comment inside storeNewImsCredentialsP for
	-- details.

--------------------------------------------------------------------------------
--[[!
	\brief Updates IMS credentials for an Oz user, and makes that user the current user.

	Call when a user has successfully authenticated against the IMS server.
	Triggers a call to the Oz server to obtain further user information
	(such as user name, subscription status, and so on).
 
	\param accessToken (string) The new access token.
	\param refreshToken (string) The new refresh token.
	
 	 \return A Promise that signals completion. If the request is successful,
 	 	the change is reflected in \c OzAccountService.currentUserB.
               	If unsuccessful, \c OzAccountService.currentUserB is unchanged and the
		promise resolves as an error.
]]--

function OzAccountService.storeNewImsCredentialsP( --[[string]] accessToken,
												   --[[string]] refreshToken )

	if type( accessToken ) ~= 'string'
		or type( refreshToken ) ~= 'string'
	then
		error( "OzAccountService.storeNewImsCredentials: both accessToken and refreshToken must be string values", 2 )
	end
	
	accountLog:info( "storeNewImsCredentials( accessToken = %s, refreshToken = %s )",
						accessToken, refreshToken )
	
	local authHeader = {
		requestHeaders = {
			{ header = 'Authorization', value = 'Bearer ' .. accessToken },
--			{ header = 'x-token-ttl', value = '30' },
				-- [scouten 2011-07-01]: Normally this should be commented out.
				-- Useful when testing the refresh logic. This specifies the lifetime
				-- of the access token in seconds.

		}
	}
	
	local ozAccountFetchP = WFHttpClient.getP(
		OzCatalogUrls.baseServerUrl
			.. '/v1.0/accounts/00000000000000000000000000000000', authHeader )
		:mapP( KSJsonDecoder.decodeWebSafeJson_stripNullP )

	local userId, userInfo
	
	local writeUserInfoP = ozAccountFetchP:mapDataP( function( _, data )

		userInfo = data
		userId = data.id and data.id:sub( -32 )

		if userId then
			accountLog:info( "token corresponds to user %s", data )
		else
			return 'error', "info packet from Oz was missing user ID"
		end
		
		-- FIXME [scouten 2011-11-22]: Should be a local that we close right
		-- away, but there's a race condition involving opening an event processor
		-- right away after closing the same one previously (#3045522, not
		-- likely to be fixed for 1.1), so we use a file-scope global instead.
		
		HACK_keepOAuth2SessionAlive = KSOAuth2Session( userId )
		HACK_keepOAuth2SessionAlive:storeTokensP( refreshToken, accessToken )
			:runUntilDone()

		if userInfoCache then
			userInfoCache:close()
		end

		userInfoCache = OzDocStore( userId )

		return userInfoCache:setServerValueP( 'user', userId, userInfo )
		
	end )
	
	writeUserInfoP:whenResolvedDo( function( status, data )

		if status == 'data' then
			KSPrefs.currentUserId = userId
		else
			KSPrefs.currentUserId = nil
		end

	end ):runUntilDone()
	
	return writeUserInfoP

end

--------------------------------------------------------------------------------
--[[
	E< table > = OzAccountService.listAllUsersOnceE()
	
	Returns (via event stream) a list of all users who have successfully
	authenticated against Oz, along with summary information about them.
]]--

--[[ UNUSED
function OzAccountService.listAllUsersOnceE()

	local resultE = KSReactive.sourceE()
	
	-- FIXME [scouten 2011-06-21]: When we rewrite this to query the account
	-- store, be sure to only return those that have valid login credentials.
	-- We may also store info about other users we know about (i.e. share
	-- participants) and those should be exluded from this list.
	
	WFTimer.afterDelayDo( .5, function()
	
		resultE:post {
	
			{
				countryCode = nil,
				expires_in = 60996,
				first_name = nil,
				organization = nil,
				phoneNumber = nil,
				address = nil,
				email = "VADER.ABUSE.AGENT@GMAIL.COM",
				userId = "3A7629BA4A5B410B9920FAAF@adobeId",
				name = "darth vader",
				last_name = nil,
				utcOffset = 0,
				token_type = "bearer",
				mrktPer = nil,
				refresh_token = "92faf6e0-26bf-4889-968a-9e0ba678161a",
				displayName = "VADER.ABUSE.AGENT@GMAIL.COM",
				access_token = "9b74cd96-ae7a-4db7-ba15-69bbedaa37ac",
			}
		
		}
	
	end )
	
	return resultE:relayE()

end
--]]

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the current user's unique ID.
	
	\return (string) The ID.
]]--

function OzAccountService.getCurrentUserId()

	return KSPrefs.currentUserId
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the base URL for IMS.
	
	\return (string) The URL.
]]--

function OzAccountService.getImsBaseUrl()
	
	return KSPrefs.imsBaseUrl

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the authorization URL for IMS.
	
	\return <code>P < string ></code> A Promise that resolves to the URL.
]]--

function OzAccountService.getImsAuthUrlP()

	accountLog:info( "OzAccountService.getImsAuthUrlP called via %s", ':traceback:' )
		-- Bug 3499868 would suggest that this is happening too often.

	local opts = {
		returnHttpErrorsAsData = true,
		requestHeaders = {
			{ header = 'Authorization', value = 'Bearer' },
				-- intentionally empty bearer token; this prompts Oz to tell us how to auth
		}
	}
	
	local ozAccountFetchP = WFHttpClient.getP(
		OzCatalogUrls.baseServerUrl
			.. '/v1.0/accounts/00000000000000000000000000000000', opts )
		:mapP {
		
			data = function( _, data, headers )
			
				if headers.status_code ~= '401' then
					return 'error', "Unexpected HTTP response status: " .. headers.status_code
				end
				
				local authHeader = headers[ 'Www-Authenticate' ] or headers[ 'WWW-Authenticate' ]
					-- This is really WWW-Authenticate, but somewhere in the client
					-- libs it seems to be switching the case on us. :-(
					-- Update (rseeler): Android returns "WWW-Authenticate", so I'm checking both.

				local authUrl = authHeader and authHeader:match( [[auth_url="([^"]+)"]] )
									--" (fix up Halide's syntax coloring)
				
				if authUrl then

					local clientIdAndSecretP = Bablu:getImsClientIdAndSecretP()
					
					local authUrlAndImsBaseUrlP = clientIdAndSecretP:mapDataP( function( status, imsClientId, imsSecret )
					
						local authUrl = authUrl:gsub( 'CarouselTest1', imsClientId )
							
						authUrl = authUrl .. '&locale=' .. WFLocalizationUtils.getLanguageCode() .. '_' .. WFLocalizationUtils.getRegionCode()
					
						local imsBaseUrl = authUrl:match( [[(.*)/authorize]] )
					
						KSPrefs.imsBaseUrl = imsBaseUrl
							-- GRUNGE [scouten 2011-08-04]: Hacky way to get base URL.
							-- Should talk to Alex about improving on this.
						
						return 'data', authUrl, imsBaseUrl
						
					end )
					
					return authUrlAndImsBaseUrlP

				else
					return 'error', "WWW-Authenticate header missing or auth_url field missing"
				end
			
			end,
		
		}

	return ozAccountFetchP

end

--------------------------------------------------------------------------------

return OzAccountService
