--[[----------------------------------------------------------------------------

OzBackupFlagAgent.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the content of a single catalog's data cache and ensures that
the "do not back up" flag is properly set for those originals files where it's
relevant.

Suggest we cycle back and remove this code in summer or fall of 2012, on the
assumption that most people will have upgraded through one or more versions that
is running it by then.

------------------------------------------------------------------------------]]

	-- core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFTimer = require 'WFTimer'

	-- core.os
local KSFileUtils = require 'KSFileUtils'

	-- core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

	-- oz.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'


--------------------------------------------------------------------------------

local backupFlagPollingInterval = 40	-- seconds; feel free to tune

local backupFlagLog = KSLogging.log 'OzBackupFlagAgent'
--	:setLevel( KSLogging.DEBUG )

local timerLog = KSLogging.log 'OzBackupFlagAgent :: timer'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark OzBackupFlagAgent

local backupFlagAgentMethods = {

	_catalogId = false,
		-- (string) catalog ID as known to the Oz system (UUID)
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)
	
	_flagBatchP = false,
		-- (P< true >) currently running backup flag setter operation, if any
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function backupFlagAgentMethods:__init( catalogId )

	self._catalogId = assert( catalogId )
	
	self._docStore = OzDocStore( catalogId )
	self:addCloseTarget( self._docStore )

	self._indexStore = WFIndexStore( self._docStore, 'OzCatalogIndexDefinitions.lua' )
	self:addCloseTarget( self._indexStore )

	self:scheduleBackupFlagBatch( 20 )
		-- Start setting do not back up flags, but not right away. Give priority
		-- to the rest of the UI starting up. 20 seconds is an arbitrary choice.
		-- Tune as we get used to it.
	
end

--------------------------------------------------------------------------------

function backupFlagAgentMethods:scheduleBackupFlagBatch( afterDelay )

	-- GRUNGE [scouten 2011-04-26]: There's almost certainly a proper reactive
	-- way to code this up, but I've spent too much time trying to think through
	-- what it is. This will work, even if it's messy.

	-- If we're already downloading or have already scheduled the timer,
	-- this is a no-op.

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		return
	end

	if self._flagBatchP then
		return
	end
	
	local delaySeconds = afterDelay or backupFlagPollingInterval

	timerLog:debug( "delay timer for doBackupFlagBatchP launched :: waiting %s seconds",
						delaySeconds )

	self._flagBatchP = WFTimer.afterDelayP( delaySeconds )
		:mapDataP( function()
		
			timerLog:debug( "delay timer for doBackupFlagBatchP fired" )

			return self:doBackupFlagBatchP()
		
		end )

	self._flagBatchP:whenResolvedDo( function( status, message )
			
		if status ~= 'data' then
			timerLog:error( "doBackupFlagBatchP failed with status %s and message %s",
							status, message )
		end
		
		self._flagBatchP = false
		
		self:scheduleBackupFlagBatch()
	
	end ):attachTo( self )
	
end

--------------------------------------------------------------------------------

function backupFlagAgentMethods:doBackupFlagBatchP()

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) or WFCloseSupport.hasBeenOrWillBeClosed( self._indexStore ) then
		return
	end
 
	local photosThatNeedDoNotBackupFlagP = self._indexStore
		:queryP( [[SELECT docId, path FROM needsDoNotBackUpFlag LIMIT 100;]] )

	return self:updatePhotosInListP( photosThatNeedDoNotBackupFlagP, 1 )
	
end

--------------------------------------------------------------------------------

function backupFlagAgentMethods:updatePhotosInListP( photosThatNeedDoNotBackupFlagP,
					startFromIndex )

	return photosThatNeedDoNotBackupFlagP:mapDataP( function( _, photosToUpdate )
		
		backupFlagLog:info( "updatePhotosInListP\n\nphotosToUpdate = %s\n\nstartFromIndex = %s",
							photosToUpdate, startFromIndex )
	
		local itemAtIndex = photosToUpdate and photosToUpdate[ startFromIndex ]
		
		if not itemAtIndex then
			return 'data', true -- we've exhausted the victim pool
		end
		
		local originalFilePath = itemAtIndex[ 2 ]
		
		local overlay

		if KSFileUtils.exists( originalFilePath ) then

			KSFileUtils.setDoNotBackUpFlag( originalFilePath )
			
			overlay = {
				_localOnly = {
					original = {
						doNotBackUpFlagSet = true,
					},
				},
			}

		else
		
			-- No original any more? Delete the reference.
		
			overlay = {
				_localOnly = {
					original = { _delete = true },
					originalFilePath = { _delete = true },
				},
			}

		end
		
		local updateDocStatusP = self._docStore:applyOverlayToDocumentP {
			docType = 'asset',
			docId = itemAtIndex[ 1 ],
			overlay = overlay,
		}
		
		-- Now chain forward and remove the next victim if need be.
		
		return updateDocStatusP:mapDataP( function()
		
			startFromIndex = startFromIndex + 1
			
			return self:updatePhotosInListP( photosThatNeedDoNotBackupFlagP,
								startFromIndex )
		
		end )
		
		
	end )
	
end

--------------------------------------------------------------------------------

return KSClass 'OzBackupFlagAgent'( backupFlagAgentMethods )
