--[[----------------------------------------------------------------------------

OzCatActivity.lua

@author arahn 2013-04-01

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require "WFArguments"
local WFUnknown = require "WFUnknown"
local KSReactive = require "KSReactive"
local WFTimer = require 'WFTimer'

local OzCatComments = require "OzCatComments"
local OzCatActivityImp = require 'OzCatActivityImp'
local KSTableUtils = require 'KSTableUtils'
local WFCloseSupport = require 'WFCloseSupport'

local activityMP = require "WFWorkerThread" "OzCatActivityImp.lua"

local isUnknown = WFUnknown.isUnknown

--============================================================================--

-- @mark OzCatalogService class

local Methods = {
}

--------------------------------------------------------------------------------

local function anyUnknown( ... )
	for i = 1, select( '#', ... ) do
		if isUnknown( select( i, ... ) ) then
			return true
		end
	end
end

--------------------------------------------------------------------------------

--[[!
 	\brief Fetch an array of activity items. Activity consists of notes about
	things that the catalog user did NOT do (ie. stuff done by others).
	The resulting list will contain at most \c limit raw items
	up to an internal maximum (TBD, probably around 300).  NOTE: Only the first
	call gets to specify the limit; afterwards we will return a cached
	activity behavior that reflects the previously asked-for limit.
	
	The raw items are processed internally by the specified \c optRollup
	parameter (which if omitted a default rollup is used).  The \optRollup if
	present is an array of rollup functions which are used to condense the raw
	list of activity into a summarized list.  For example, if one photo has
	been favorited by multiple people, a rollup function could emit a single
	'new favorite' activity event that lists all those users.  Similarly, if
	one user has favorited multiple photos, another rollup could emit a single
	'new favorite' activity that lists all those photos with a single user.
	Each rollup function provides a weight, and the activity consolidator
	competes them one against another to determine the optimal rollup.
	
	The array of items is sorted in reverse-chronological order by date of
	the newest part of the roll up.  For example, if a photo was favorited
	by three different people on Jan 1, Jan 2 and Jan 3, the rolled up activity
	entry for 'new favorite' would have the date of Jan 3, and it would also
	be sorted by that date.
	
	\see WFActivity for some built-in rollup functions, and the default
	rollup array.
	
	Each activity item is a table with the following fields:
		album = {
			-- only present for 'new asset in album' type
			name = 'name',
			id = '< user uuid >',
		},
		andMaybeMore = < boolean >,
			-- will be true if the number of photos involved in this
			-- activity is equal to or exceeds the limit.  When generating
			-- user-visible messages, don't rely on the #photos to be accurate
			-- if activity.andMaybeMore == true.
		bookmark = '< bookmark data >',
			-- a non-human readable string value that points in to the
			-- point in time represented by this activity.
		comment = {
			-- only present for 'new comment' actions
			body = 'comment body',
			id = 'comment id',
		},
		date = '2013-03-02T12:34:56Z', --
		id = '< uuid >'
			-- a uuid that identifies this activity; it isn't guaranteed
			-- to persist beyond a single session of the app (but it might).
		photos = {
			-- array of one or more photo ids
			'< photo uuid >',
		},
		type = 'new comment|new asset|new favorite|new asset in album',
		users = {
			-- array of one or more user info in same format as comment
			-- user
			{
				fullName = 'name',
				lastName = 'last name',
				firstName = 'first name',
				id = '< user uuid >',
			},
		},
		yourPhotos = true | false,
			-- on 'new comment' and 'new favorite' activity types,
			-- we will additionally specify whether the activity
			-- was on your photo or on some other user's photo,
			-- and when rolling up activity by a user, we only
			-- roll up if the yourPhotos value is the same.
			-- (E.g., "User commented on 5 photos" AND "User commented
			-- on 3 of your photos").
	},
	There will never be both more than one user and more than one photo in a single item.
	
	\param limit the maximum number of raw elements to fetch.
	\param rollup an array of rollup functions, or nil for default behavior.

	\return <code>B< array_of_activity_items | WFUnknown.value ></code> The activity array.  Initially unknown, the
	transitions to a stream of values as activity updates
]]--

function Methods:activityB( optLimit, optRollup )
	
	WFArguments.check( "OzCatalogServiceImp", "?number", "?table" )
	
	optLimit = optLimit or 300
	
	if not self._activityInfo then
		self._activityInfo = {
			cache = KSTableUtils.makeFullyWeak()
		}
	end
	
	local cached = WFCloseSupport.fetchCached( self._activityInfo.cache, "cachedB" )
	if cached then
		return cached:relayB( true )
	end
	
	local commonPredicate = [[WHERE userId IS NOT ? AND userId IS NOT NULL AND activityDate IS NOT NULL ORDER BY activityDate DESC LIMIT ]] .. tostring( optLimit )
	
	local favoriteB = self._indexStore:queryB {
		sql = [[SELECT docId, userId, activityDate, importedBy FROM favorite_activity ]] .. commonPredicate,
		self._ozUserId,
		calm = 5,
	}
	
	local newAssetB = self._indexStore:queryB {
		sql = [[SELECT docId, userId, activityDate FROM new_asset_activity ]] .. commonPredicate,
			self._ozUserId,
		calm = 5,
	}
	
	local newAlbumAssetB = self._indexStore:queryB {
		sql = [[SELECT assetId, userId, album_asset.updateTimeStamp AS activityDate, albumId, name FROM album_asset INNER JOIN albums ON albums.id = album_asset.albumId ]] .. commonPredicate,
			self._ozUserId,
		calm = 5,
	}
	
	local newCommentsB = self._indexStore:queryB {
		sql = "SELECT "
			.. table.concat( OzCatActivityImp.kCommentFields, ", " )
			.. ", "
			.. OzCatComments.kCommentFields
			.. [[ FROM comment INNER JOIN coreInfo ON coreInfo.docId == comment.assetId ]]
			.. commonPredicate,
		self._ozUserId
	}
	
	local retryB = KSReactive.variableB( 1 )
	
	local yourUserId = self._ozUserId
	
	local resultB = KSReactive.dataCallB( function( favorites, newAssets, newComments, newAlbumAssets, retry )
		if anyUnknown( favorites, newAssets, newComments, newAlbumAssets ) then
			return WFUnknown.value()
		end
		return favorites, newAssets, newComments, newAlbumAssets, retry
	end, favoriteB, newAssetB, newCommentsB, newAlbumAssetB, retryB )
	:addCloseTarget( favoriteB )
	:addCloseTarget( newAssetB )
	:addCloseTarget( newCommentsB )
	:addCloseTarget( newAlbumAssetB )
	:addCloseTarget( retryB )
	:closingMapSwitchB( function( favorites, newAssets, newComments, newAlbumAssets )
		if WFUnknown.isUnknown( favorites ) then
			return KSReactive.constantB( favorites )
		elseif self._activityInfo.processingP then
			self._activityInfo.needsRetry = self._activityInfo.needsRetry or 'needsRetry'
			return self._activityInfo.processingP
				:untilResolvedB( 'data', WFUnknown.value() )
		else
			self._activityInfo.processingP = activityMP:requestP( 'processActivity',
					optLimit, yourUserId, favorites, newAssets, newComments, newAlbumAssets )
			
			return self._activityInfo.processingP
				:observeP( function()
					self._activityInfo.processingP = nil
				end )
				:untilResolvedB( 'data', WFUnknown.value() )
		end
	end )
	:closingMapSwitchB( function( orderedActivities, needUsers )
		local resultB = KSReactive.variableB()
		if needUsers then
			-- The only way to find out about users is if they're collaborators
			self:catalogUsersP()
				:mapDataP( function( status, usersInfo )
					local userInfoForId = {}
					for _, userInfo in ipairs( usersInfo ) do
						local userId = userInfo._id
						if userId then
							userInfoForId[ userId ] = userInfo
						end
					end
					for _, activity in ipairs( orderedActivities ) do
						for _, entry in ipairs( activity.users or {} ) do
							local userInfo = userInfoForId[ entry.id ]
							if userInfo then
								entry.firstName = userInfo[ 'first_name' ]
								entry.lastName = userInfo[ 'last_name' ]
								entry.fullName = userInfo[ 'full_name' ]
							end
						end
					end

					resultB:set( orderedActivities )
				end )
			:whenResolvedDo( function( status )
				if status == 'data' then
					resultB:set( orderedActivities )
				end
			end )
			:attachTo( resultB, resultB, "nameRemapping" )
		else
			resultB:set( orderedActivities )
		end
		
		if self._activityInfo.needsRetry == 'needsRetry' then
			self._activityInfo.needsRetry = 'retryDelaying'
			WFTimer.afterDelayDo( 5, function()
				self._activityInfo.needsRetry = false
				retryB:set( retryB:valueNow() + 1 )
			end )
		end
		
		return resultB
	end )
	
	self._activityInfo.cache.cachedB = resultB
	
	resultB:addCloseAction( function()
		self._activityInfo = false
	end )
	
	return resultB:relayB( true )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Updates the catalog user's bookmark for how much activity 
	has been "seen". This value is synced via the user document to all
	other clients, so that read/unread status can be displayed.
	
	It is expected that for the duration of the app session, the client
	code will use the temporary activity.id value to track what specific
	activity items have been viewed.  However upon next launch, or when
	viewing activity on another device, any activity newer than bookmark
	should be treated as 'unread', and all activity older than or equal
	to bookmark will be	'read'.
	
	\param bookmark a bookmark from one of the activity values returned
	by activityB, or nil to mark all activity as unread.
]]--

function Methods:setActivityBookmarkP( bookmark )
		
	WFArguments.check( 'OzCatalogServiceImp', 'string' )

	return self._ozUser:setActivityBookmarkP( self._catalogId, bookmark )
	
end

--------------------------------------------------------------------------------
--[[!
 	\brief Retreive the user's activity bookmark on this catalog.
	
	\param <code>B< bookmark ></code> the bookmark.
]]--

function Methods:activityBookmarkB( )
			
	WFArguments.check( 'OzCatalogServiceImp' )

	return self._ozUser:getActivityBookmarkB( self._catalogId )

end

--============================================================================--

return {
	ActivityMethods = Methods,
}