--[[----------------------------------------------------------------------------

OzCatAlbums.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSLogging = require "KSLogging"

local WFArguments = require "WFArguments"
local KSStringUtils = require "KSStringUtils"
local KSIdGenerator = require "KSIdGenerator"
local KSDate = require "KSDate"
local KSReactive = require "KSReactive"
local KSPromises = require "KSPromises"
local KSReactivePromiseBridge = require "KSReactivePromiseBridge"
local WFZString = require 'WFZString'
local KSTableUtils = require 'KSTableUtils'
local WFPlatformInfo = require 'WFPlatformInfo'

--==============================================================================

local ZSTR = assert( WFZString.fromZString )

-- @mark OzCatalogService class

local AlbumMethods = {
}

local OzCatalogServiceArg = 'OzCatalogServiceImp'

--------------------------------------------------------------------------------

local albumsLog = KSLogging.log 'OzCatalogService :: albums '
--  :setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------
--[[!
	\brief Creates an album in this catalog.
	
	\param name (string) The name of the new album.
	\param bypassUniquenessCheck (boolean) When \c true, do not check for a unique album name.
	
	\return <code>P< string ></code> A Promise object that resolves to the ID of the
		newly-created album when the operation is complete.
]]--

function AlbumMethods:createAlbumP( --[[string]] name,
				--[[boolean]] bypassUniquenessCheck )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:createAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local resultP = self._indexStore:queryP( [[SELECT DISTINCT nameLC FROM albums WHERE nameLC == ?]], KSStringUtils.lower( name ) )
		:mapP( function( status, rows )
		
			local nameIsUnique = true
			if status == 'data' and rows[ 1 ] and rows[ 1 ][ 1 ] then
				nameIsUnique = false
			end
			
			if nameIsUnique or bypassUniquenessCheck then
			
				local albumId = KSIdGenerator.uuidInLowercaseHex()
				local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						name = name,
						_localOnly = {
							updateTimeStamp = updateTimeStamp,
						},
					},
				}
				
				return innerResultP:mapDataP( function( status,_ )
					return status, albumId
				end )
				
			else
			
				return 'error', 'NameExists'
				
			end
						
		end )

	
	resultP:runUntilDone()
	
	return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes an album from this catalog.
	
	\param albumId (string) The unique ID of the album to delete.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:deleteAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:deleteAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	-- PR #3312056: Unshare the album first, then delete the album, too.
	-- Regardless of success/failure of unsharing the album, we're still going
	-- to delete it, thus the use of :mapP instead of :mapDataP.
	
	return self:unshareAlbumP( albumId )
		:mapP( function()
	
			local resultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album',
				docId = albumId,
				overlay = {
						_deleted = true,
					},
				}
								
			return resultP:mapDataP( function ()
				return 'data', true
			end )
		
		end )
		:runUntilDoneP()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Renames an album in this catalog.
	
	It is an error to rename an album to match the name of an existing album,
	regardless of case.
	
	\param albumId (string) The unique ID of the album to rename.
	\param name (string) The new name for the album.
	\param bypassUniquenessCheck (boolean)  When \c true, do not check for a unique album name.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:renameAlbumP( --[[string]] albumId,
											 --[[string]] name,
											 --[[boolean]] bypassUniquenessCheck )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string', '?boolean' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:renameAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:renameAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local resultP = self._indexStore:queryP( [[SELECT DISTINCT nameLC FROM albums WHERE id != ? AND nameLC == ? ]], albumId, KSStringUtils.lower( name ) )
		:mapP( function( status, rows )
		
			local nameIsUnique = true
			if status == 'data' and rows[ 1 ] and rows[ 1 ][ 1 ] then
				nameIsUnique = false
			end
			
			if nameIsUnique or bypassUniquenessCheck then
			
				local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						name = name,
						_localOnly = {
							updateTimeStamp = updateTimeStamp,
						},
					},
				}
				
				return innerResultP
				
			else
				return 'error', 'NameExists'
			end
						
		end )
						
	resultP:runUntilDone()

	return resultP:mapDataP( function ()
		return 'data', true
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the timestamp for an album in this catalog.
	
	\param albumId (string) The unique ID of the album to touch.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:touchAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:touchAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )

	local resultP = self._docStore:applyOverlayToDocumentP {
		docType = 'album',
		docId = albumId,
		overlay = {
			_localOnly = {
				updateTimeStamp = updateTimeStamp,
			},
		},
	}
						
	resultP:runUntilDone()
	
	return resultP:mapDataP( function ()
		return 'data', true
	end )
	
end

--------------------------------------------------------------------------------

local function nameForSmartAlbumId( self, id )
	
	local albumsNow = self:smartAlbumsB():valueNow()
		-- use public API to force lazy creation
	
	for _, album in pairs( albumsNow ) do
		if album.id == id then
			return album.name
		end
	end

	return nil -- ID not found
	
end

--------------------------------------------------------------------------------

-- FIXME [scouten 2012-10-16]: What is the difference in intent between
-- albumNameForIdB and getAlbumNameB? I can't tell. Also, I think we need
-- an albumNameForIdP based on a conversation with Jon earlier today.

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the name of an album.

	\param albumId (string) The unique ID of the album.

	\return <code>B< string | nil ></code> A Behavior object whose value is the name
		or nil if the album is not found.
]]--

function AlbumMethods:albumNameForIdB( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	return self._indexStore:queryB(
		[[SELECT id, name FROM albums WHERE id==?]], albumId )
			
		:closingMapB( function( data )
			
			if not data[ 1 ] then return nil end
					
			return data[ 1 ][ 2 ]
				
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the name of an album.
	
	\param albumId (string) The unique ID of the album.

	\return <code>P< string | nil ></code> A Promise object that resolves to the name
		when the operation is complete, or nil if the album is not found.
		
]]--

function AlbumMethods:albumNameForIdP( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	return self._indexStore:queryP(
		[[SELECT name FROM albums WHERE id==?]], albumId )
			
		:mapDataP( function( status, data )
			
			if not data[ 1 ] then return 'error', 'noSuchAlbum' end
					
			return status, data[ 1 ][ 1 ]
				
		end )

end

--------------------------------------------------------------------------------
--[[!
	\deprecated Do not use. Instead, get current album from data source and call \c getAlbumNameB()
]]--

function AlbumMethods:currentAlbumNameB()
error "Deprecated; get current albumb from data source and call getAlbumNameB"
end

--[[!
	\brief Retrieves the name of an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is the name
		or nil if the album is not found.
]]--

function AlbumMethods:getAlbumNameB( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	if not albumId then return KSReactive.constantB( nil ) end
	
	local smartAlbumName = nameForSmartAlbumId( self, albumId )
	
	if smartAlbumName then
		return KSReactive.constantB( smartAlbumName )
	end
	
	return self._indexStore:queryB(
		[[SELECT name FROM albums WHERE id==?]], albumId )
	
	:closingMapB( function( data )

		if not data[ 1 ] then return nil end
		
		return data[ 1 ][ 1 ]
	
	end )
			
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a list of albums for this catalog, sorted by last update date
		(latest first).
	
	\return <code>B< table ></code> A Behavior object whose value is an array of album
		descriptions where each entry in the array is a table containing
		 \c id and \c name and \c updateTimeStamp members.
]]--

function AlbumMethods:albumsDateSortedB()

	if not self._albumsDateSortedB then
		self._albumsDateSortedB = self._indexStore:queryB(
			[[SELECT DISTINCT a.id as id, a.name as name, MAX(
				CASE WHEN a.updateTimeStamp IS NULL THEN 0 ELSE a.updateTimeStamp END,
				CASE WHEN a.localTimeStamp IS NULL THEN 0 ELSE a.localTimeStamp END,
				CASE WHEN b.maxUpdateTimeStamp IS NULL THEN 0 ELSE b.maxUpdateTimeStamp END,
				CASE WHEN c.maxUpdateTimeStamp IS NULL THEN 0 ELSE c.maxUpdateTimeStamp END) as timeStamp, (CASE WHEN b.assetCount IS NULL THEN 0 else b.assetCount END) as albumAssetCount FROM
					(SELECT id, name, updateTimeStamp, localTimeStamp FROM albums) a
					LEFT JOIN
					(SELECT albumId, MAX(updateTimeStamp) as maxUpdateTimeStamp, count(*) as assetCount FROM (album_asset LEFT JOIN coreInfo on album_asset.assetId == coreInfo.docId) WHERE (coreInfo.contentTypeNumber == 1001 OR coreInfo.contentTypeNumber == 1002) GROUP BY albumId) b
						ON a.id == b.albumId
					LEFT JOIN
					(SELECT albumId, MAX(updateTimeStamp) as maxUpdateTimeStamp FROM space_album GROUP BY albumId) c
						ON a.id == c.albumId
							ORDER BY timeStamp DESC
			]] )
			:closingMapB( function( data )
				albumsLog:debug( "Albums date sorted: %s", data )
				local albums = {}
				for i = 1, #data do
					albums[ i ] = {}
					albums[ i ].id = data[ i ][ 1 ]
					albums[ i ].name = data[ i ][ 2 ]
					albums[ i ].updateTimeStamp = data[ i ][ 3 ]
					albums[ i ].assetCount = data[ i ][ 4 ]
				end
				return albums
			end )
			
		self:addCloseTarget( self._albumsDateSortedB )
	end
	
	return self._albumsDateSortedB:relayB( true )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a list of albums for this catalog, sorted by the most recent captureDate
	of any assets in each album. if there are no assets in an album, the updateTimestamp is used.
	
	\return <code>B< table ></code> A Behavior object whose value is an array of album
		descriptions where each entry in the array is a table containing
		 \c id and \c name and \c updateTimeStamp \c captureDate members.
		The MAX captureDate will be nil if there are no assets in an album.
]]--

function AlbumMethods:albumsCaptureDateSortedB()

	if not self._albumsCaptureDateSortedB then
		self._albumsCaptureDateSortedB = self._indexStore:queryB( [[ 
			SELECT albums.id as id, albums.name as name, albums.updateTimestamp as timestamp, a.captureDate FROM albums 
				LEFT OUTER JOIN 
					( SELECT album_asset.albumId as albumId, MAX(coreInfo.captureDate) as captureDate FROM coreInfo 
						INNER JOIN album_asset ON coreInfo.docId == album_asset.assetId 
							GROUP BY album_asset.albumId ) as a 
				ON a.albumId == albums.id 
					ORDER BY CASE WHEN a.captureDate IS NULL THEN albums.updateTimestamp ELSE a.captureDate END DESC
				]] )
			:mapB( function( data )
				albumsLog:debug( "Albums sorted by max captureDate: %s", data )
				local albums = {}
				for i = 1, #data do
					albums[ i ] = {}
					albums[ i ].id = data[ i ][ 1 ]
					albums[ i ].name = data[ i ][ 2 ]
					albums[ i ].updateTimeStamp = data[ i ][ 3 ]
					albums[ i ].captureDate = data[ i ][ 4 ]
				end
				return albums
			end )
			
		self:addCloseTarget( self._albumsCaptureDateSortedB )

	end
	
	return self._albumsCaptureDateSortedB:relayB()
end

function AlbumMethods:albumCoverImageB( --[[string]] albumId )

	debug.print(" Trying to get cover asset for albumm is : " .. albumId)
	local coverImageB = self._indexStore:queryB( [[
			SELECT coeInfo.docId FROM coreInfo INNER JOIN album_asset
			ON alsbum_asset.assetId == coreInfo.docId
			WHERE albumId == ?
			ORDER BY CASE coreInfo.favoritedCount WHEN coreInfo.favoritedCount = '0' or coreInfo.favoritedCount IS NULL THEN 0 ELSE 1 END DESC,
			captureDate ASC
			LIMIT 1	]], albumId
		)
	:mapB( function( data )
			result = {}
			result['albumId'] = albumId
			result['coverAssetId'] = data[1][1]

			return result
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves an unsorted list of albums for this catalog.
	
	\return <code>P< table ></code> A Promise object that resolves to a table of album
		descriptions where each entry in the table is a table containing keys
		 \c id and \c name.
]]--

function AlbumMethods:albumsP()

	return self._indexStore:queryP( [[ SELECT id, name FROM albums ]] )
		:mapDataP( function( status, data )
		
			albumsLog:debug( "AlbumsP: %s", data )
			local albums = {}
			for i = 1, #data do
				albums[ i ] = {}
				albums[ i ].id = data[ i ][ 1 ]
				albums[ i ].name = data[ i ][ 2 ]
			end
			return 'data', albums
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a list smart albums for this catalog.
	
	\return <code>B< table ></code> A Behavior object whose value is a table of album
		descriptions where each entry in the table is a table containing keys
		 \c id and \c name.
		 
 ]]--

function AlbumMethods:smartAlbumsB()

	if not self._smartAlbumsB then
		local albums = {}
		albums[ 1 ] = {}
		albums[ 1 ].id = 'favorites' .. self._ozUserId
		albums[ 1 ].name = ZSTR( "$$$/SourceList/SmartAlbums/Favorites=My Favorites" )
		self._smartAlbumsB = KSReactive.constantB( albums )
		self:addCloseTarget( self._smartAlbumsB )
	end
	
	return self._smartAlbumsB:relayB()
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds assets to an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to add.
	
	TODO: The return results from this function are weird.  It should consistently
	return a boolean, or a list of ids, not a mix of the two as it does currently.
	
	\return <code>P< true | array ></code>
		A Promise object that resolves to \c true if the assets were already in the album,
		or else to the list of asset IDs that were actually added to the album.
		
]]--

function AlbumMethods:addAssetsToAlbumP( --[[string]] albumId,
					--[[string]] assetIds )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )

	local undoRedoAssetIds = {}
	
	local assetsPresent = {}
	
	local function checkForExistingAssetsP()
		
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return self._indexStore:queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				assetsPresent[ idRows[ i ][ 1 ] ] = true
			end
			return 'data', assetsPresent
		end )

	end
	
	local function createAlbumAssetDocument( assetId )
		
		if assetsPresent[ assetId ] then
			return 'data', true
		else
			local albumAssetId = KSIdGenerator.uuidInLowercaseHex()
			local innerResultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = albumAssetId,
				overlay = {
					album = {
						_id = albumId,
					},
					asset = {
						_id = assetId,
					},
				},
			}

			table.insert( undoRedoAssetIds, assetId )
		
			return innerResultP
			
		end
					
	end
	
	local resultP = self._indexStore:queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return checkForExistingAssetsP()
				:mapDataP( function ()
					return KSPromises.allP( createAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
			
		end )

	resultP:mapDataP( function()
		self:touchAlbumP( albumId ):runUntilDone()
	end ):runUntilDone()

	return resultP:mapDataP( function( _, assetIds )
		return 'data', assetIds, undoRedoAssetIds
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes assets from an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to delete.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:removeAssetsFromAlbumP( --[[string]] albumId,
													   --[[string]] assetIds )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )
		
	local albumAssetsPresent = {}

	local function getAlbumAssetIdsP()
		
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT docId, assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return self._indexStore:queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				albumAssetsPresent[ idRows[ i ][ 2 ] ] = idRows[ i ][ 1 ]
			end
			return 'data', albumAssetsPresent
		end )

	end
		
	local function deleteAlbumAssetDocument( assetId )
	
		local albumAssetId = albumAssetsPresent[ assetId ]
		
		if albumAssetId then
		
			local localOnly = {
				album = {
					_id = albumId,
				},
				asset = {
					_id = assetId,
				},
			}
					
			local innerResultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = albumAssetId,
				overlay = {
					_deleted = true,
					_localOnly = localOnly,
				},
			}
						
			return innerResultP
			
		else
			return 'data', true
		end

	end
	
	local resultP
	
	if albumId == ( 'favorites' .. self._ozUserId ) then
		resultP = self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = assetIds,
						overlay = {
							favorites = {
								[ self._ozUserId ] = { _delete = true },
							},
						},
					}
	else
		resultP = self._indexStore:queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return getAlbumAssetIdsP()
				:mapDataP( function ()
					return KSPromises.allP( deleteAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
		
		end )
				
		resultP:mapDataP( function()
			self:touchAlbumP( albumId ):runUntilDone()
		end ):runUntilDone()
	end
	
	return resultP
	
end

--------------------------------------------------------------------------------

--[[!
	\deprecated Do not use. Instead, get current album from data source.
]]--

function AlbumMethods:setCurrentAlbumP( --[[string]] albumId )
	error "Moved to data source"
end

--------------------------------------------------------------------------------
--[[!
	\brief Is an albumId valid for this catalog
	
	\param albumId (string) The unique ID of the album to modify.
	
	\return <code>P< boolean ></code> A Promise object that resolves to a boolean
		if the album id exists and has not been deleted for this catalog.
]]--

function AlbumMethods:isValidAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )
	
	return self._indexStore:queryP( [[ SELECT 1 FROM albums WHERE id = ? LIMIT 1 ]], albumId )
		:mapDataP( function( status, albumList )
			
			return 'data', not not albumList[ 1 ]

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Is an albumId valid for this catalog
	
	\param albumId (string) The unique ID of the album to modify.
	
	\return <code>B< boolean ></code> A Behavior object that resolves to a boolean
		if the album id exists and has not been deleted for this catalog.
]]--

function AlbumMethods:isValidAlbumB( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )
	
	-- use a count so we can differentiate the initial empty results from
	-- the none-exists result.
	return self._indexStore:queryB( [[ SELECT count( id ) FROM albums WHERE id = ? LIMIT 1 ]], albumId )
		:closingMapB( function( albumList )
			
			return ( albumList[ 1 ] and albumList[ 1 ][ 1 ] ) ~= 0

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\deprecated Do not use. Instead, get current album from data source.
]]--

function AlbumMethods:getCurrentAlbumB()
	
	error "Moved to data source"
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Shares an album in this catalog, creating an associated space document.
	
	\param albumId (string) The unique ID of the album.
	\param spaceSubtype (string) The subtype for the album's space; usually nil.
	
	\return <code>P< string ></code> A Promise object that resolves to the ID of
		the new space if the request succeeded.
]]--

function AlbumMethods:shareAlbumP( --[[string]] albumId, --[[string]] spaceSubtype )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	-- The current sharing model for albums calls for a 1-1 mapping from an 'album' to
	-- a 'space'. In order to share an album, a 'space' needs to be created and the 'album'
	-- added to it.

	-- This function creates a space and adds the provides album to it by creating a
	-- 'space_album' document. If a space already exists for this album, it returns the ID
	-- of that 'space'.
	
	local resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function( _, idRows )
		
			-- If there is a space for this album, return the ID for that space.
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
			if spaceId then
				return 'data', spaceId
			end
			
			spaceId = KSIdGenerator.uuidInLowercaseHex()
			
			local space_album_id = KSIdGenerator.uuidInLowercaseHex()
			-- Save the album information in a localOnly section in the space itself. Oz does not have a separate
			-- Add Album to Space API for 2.0. We need to send a single request to create a space and add the album to it.
			-- When this API is present, we should sync the space_album document separately as is done for album_asset documents.
			
			local spaceResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						_localOnly = {
							album = {
								albumId = albumId,
								space_album_id = space_album_id,
							},
						},
						download = true,
						private = false,
						subtype = spaceSubtype,
					},
				}
		
			return spaceResultP:mapDataP( function()
			
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space_album',
					docIds = space_album_id,
					overlay = {
						space = {
							_id = spaceId,
						},
						album = {
							_id = albumId,
						},
					},
				}
				
				return innerResultP:mapDataP( function()
					return 'data', spaceId
				end )
			
			end )
			
		end )
		
	return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Unshares an album in this catalog, deleting the associated space
		and space_album documents.

	\param albumId (string) The unique ID of the album.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:unshareAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultP = self._indexStore:queryP( [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function( _, idRows )
		
			local space_album_id = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 2 ]
			
			if not idRows[ 1 ] then
				return 'error', "Album is not shared"
			end
			
			local spaceResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						_deleted = true,
					},
				}
		
			return spaceResultP:mapDataP( function ()
			
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space_album',
					docIds = space_album_id,
					overlay = {
						_deleted = true,
					},
				}
				
				return innerResultP:mapDataP( function ()
					return 'data', true
				end )
			
			end )
			
		end )
		
	return resultP
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Checks whether an album in this catalog is shared.
	
	\param albumId (string) The unique ID of the album.
		
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if the album is shared, \c false if not
]]--

function AlbumMethods:isAlbumSharedB( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryB( [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]], albumId )
		:closingMapB( function ( idRow )
			if idRow and idRow[ 1 ] and idRow[ 1 ][ 1 ] and idRow[ 1 ][ 2 ] then
				return true
			else
				return false
			end
		end )
		
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether an album in this catalog is shared.
	
	\param albumId (string) The unique ID of the album.
		
	\return <code>P< boolean ></code> A Promise object that resolves to \c true
		if the album is shared, \c false if not
]]--

function AlbumMethods:isAlbumSharedP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultP = self._indexStore:queryP( [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, idRow )
		
			if idRow[ 1 ] and idRow[ 1 ][ 1 ] and idRow[ 1 ][ 2 ] then
				return 'data', true
			else
				return 'data', false
			end
		end )
		
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

local function sharedAlbumUrlB_imp( self, spaceId )

	return self._docStore:docValueB( 'space', spaceId )
		:closingMapB( function ( docValue )
		
			return KSTableUtils.derefOpt( docValue, 'links', '/rels/web_space', 'href' )
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the public URL for an album in this catalog, if the album is shared.
	
	\param albumId (string) The unique ID of the album.
	\param optSpaceId (optional string) The ID of the sharing space record for
		this album. If provided, can help speed the query.
	
	\return <code>B< string ></code> A Behavior object whose value is the public URL
		for the shared album.
]]--

function AlbumMethods:sharedAlbumUrlB( --[[string]] albumId,
									   --[[optional string]] optSpaceId )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string' )
	
	if optSpaceId then
		return sharedAlbumUrlB_imp( self, optSpaceId )
	end
	
	local resultB = self._indexStore:queryB( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:closingMapSwitchB( function( idRow )
		
			local spaceId = idRow and idRow[ 1 ] and idRow[ 1 ][ 1 ]
			if spaceId then
				return sharedAlbumUrlB_imp( self, spaceId )
			else
				return KSReactive.constantB( nil )
			end

		end )
		
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the public URL for an album in this catalog, if the album is shared.
 	
	\param albumId (string) The unique ID of the album.
	
	\param nonNilResult (boolean) When \c true, the returned promise does not resolve
		until there is a non-nil value; when \c false, the promise resolves to
		whatever value is present in the local database.
	
	\return <code>P< string ></code> A Promise object that resolves to the public URL
		for the shared album, if any.
 ]]--

function AlbumMethods:sharedAlbumUrlP( --[[string]] albumId,
												--[[boolean]] nonNilResult )

	WFArguments.check( OzCatalogServiceArg, 'string', '?boolean' )
	local resultP
	
	if nonNilResult then

		local sharedAlbumUrlB = self:sharedAlbumUrlB( albumId )
		
		-- FIXME_AUTO_CLOSE: Do I need the relay here?
		
		resultP = KSReactivePromiseBridge.behaviorToPromise( sharedAlbumUrlB:relayForAutoCloseB(),
			function ( data )
				return data ~= nil
			end )

	else

		resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, idRow )
		
			local spaceId = idRow and idRow[ 1 ] and idRow[ 1 ][ 1 ]
			if spaceId then
				return self._docStore:docValueOnceP( 'space', spaceId )
					:mapDataP( function( _, docValue )
				
					local albumUrl = docValue and docValue.links and docValue.links["/rels/web_space"] and docValue.links["/rels/web_space"].href
					return 'data', albumUrl
				end )

			else
				return 'data', nil
			end
			
		end )

	end

	return resultP

end

--------------------------------------------------------------------------------
--[[!
	\brief Invites a list of users to share an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param userEmails (table) The list of user email IDs to  invite.
	\param spaceSubtype (optional string) The subtype for the album's space.

	\warning Currently performs the operation only locally, because
		the corresponding API on Oz is not yet implemented.
		The Revel client does yet use this function.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:inviteUsersToAlbumP( --[[string]] albumId,
					--[[table]] userEmails,
					--[[string]] spaceSubtype )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table', '?string' )

	local resultP = self:shareAlbumP( albumId, spaceSubtype )
	
	return resultP:mapDataP( function ( status, spaceId )
	
		-- Get the spaceID for this album and for each user being added to the album
		-- create a 'space_user' document.
		
		local function addUserToSpace( userEmail )
		
			local innerResultP = self._indexStore:queryP( [[SELECT docId FROM space_user WHERE spaceId == ? AND userEmail == ?]], spaceId, KSStringUtils.lower( userEmail ) )
				:mapP( function( _, idRows )

				if idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ] then
					-- This user is already added to this space.
					return 'data', true
				end
				
				local addUserP = self._docStore:applyOverlayToDocumentP {
					docType = 'space_user',
					docIds = KSIdGenerator.uuidInLowercaseHex(),
					overlay = {
						space = {
							_id = spaceId,
						},
						user = {
							email = KSStringUtils.lower( userEmail ),
						},
					},
				}
								
				return addUserP
			
			end )
			
			return innerResultP
		
		end
		
		return KSPromises.allP( addUserToSpace, userEmails )
			:mapDataP( function ()
				return 'data', true
			end )
	
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Removes a list of sharing users from an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param userEmails (table) The list of user email IDs to remove.

	\warning Currently performs the operation only locally, because
		the corresponding API on Oz is not yet implemented.
		The Revel client does yet use this function.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:removeUsersFromAlbumP( --[[string]] albumId,
													  --[[table]] userEmails )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )

	local resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, rows )
		
			return 'data', rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
		end )
	
	return resultP:mapDataP( function ( status, spaceId )
	
		-- Get the spaceID for this album and delete the space_user documents for each user.
		local function removeUserFromSpace( userEmail )
		
			local innerResultP = self._indexStore:queryP( [[SELECT docId FROM space_user WHERE spaceId == ? AND userEmail == ?]], spaceId, KSStringUtils.lower( userEmail ) )
				:mapP( function( _, idRows )
				
				local docId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
				if not docId then
					return 'data', true
				end
				
				local removeUserP = self._docStore:applyOverlayToDocumentP {
					docType = 'space_user',
					docIds = docId,
					overlay = {
						_deleted = true,
					},
				}
								
				return removeUserP
			
			end )
			
			return innerResultP
		
		end
		
		return KSPromises.allP( removeUserFromSpace, userEmails )
			:mapDataP( function ()
				return 'data', true
			end )
	
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Modifies the 'Require Login' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param requireLogin (boolean) The new value for the setting.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:setRequireLoginForAlbumP( --[[string]] albumId,
						--[[boolean]] requireLogin )

	WFArguments.check( OzCatalogServiceArg, 'string', 'boolean' )
	
	local resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						private = requireLogin,
					},
				}
								
				return innerResultP
					:mapDataP( function ()
						return 'data', true
					end )
			else
				return 'error', 'AlbumNotShared'
			end
			
		end )
			
	return resultP
 
end

--------------------------------------------------------------------------------
--[[!
	\brief Modifies the 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param allowDownloads (boolean) The new value for the setting.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:setAllowDownloadsForAlbumP( --[[string]] albumId,
						--[[boolean]] allowDownloads )

	WFArguments.check( OzCatalogServiceArg, 'string', 'boolean' )
	
	local resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						download = allowDownloads,
					},
				}
								
				return innerResultP
					:mapDataP( function ()
						return 'data', true
					end )
			else
				return 'error', 'AlbumNotShared'
			end
			
		end )
			
	return resultP
 
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the current 'Require Login' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> A Behavior object whose value is the current value of the setting.
]]--

function AlbumMethods:getRequireLoginForAlbumB( --[[string]] albumId )
	
	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryB( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
	
		:closingMapSwitchB( function ( rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
			
				return self._docStore:fieldValueB( 'space', spaceId,
							'private' )
				:mapB( function ( data )
					return data
				end )
			else
				return KSReactive.constantB( nil )
			end
			
		end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the current 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> A Behavior object whose value is the current value of the setting.

 ]]--

function AlbumMethods:getAllowDownloadsForAlbumB( --[[string]] albumId )
	
	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryB( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
	
		:closingMapSwitchB( function ( rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
			
				return self._docStore:fieldValueB( 'space', spaceId,
							'download' )
				:mapB( function ( data )
					return data
				end )
			else
				return KSReactive.constantB( nil )
			end
			
		end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
--[[!
	Reports the current 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> A Behavior object whose value is the current value of the setting.

 ]]--

function AlbumMethods:getAllowDownloadsForAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
	
		:mapDataP( function ( status, rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
			
				return self._docStore:docValueOnceP( 'space', spaceId )
					:mapDataP( function( _, docValue )
					
					local downloadsForAlbum = docValue and docValue.download
					return 'data', downloadsForAlbum
				end )
			
			else
				return 'data', nil
			end
			
		end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the assetId for the cover image for an album in this catalog.
	Rules governing the selection of the photo:
	Select first favorite in chronological order as an album cover. 
	If there is no favorite, select the first photo in the album in chronological order.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< string ></code> A Behavior object whose value is the assetId
	of the current cover photo for this album.

 ]]--

function AlbumMethods:albumCoverImageB( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryB( [[SELECT coreInfo.docId FROM coreInfo
						INNER JOIN album_asset ON album_asset.assetId == coreInfo.docId 
						WHERE albumId == ? AND (coreInfo.contentTypeNumber == 1001 OR coreInfo.contentTypeNumber == 1002)
						ORDER BY CASE coreInfo.favoritedCount WHEN coreInfo.favoritedCount = '0' or coreInfo.favoritedCount IS NULL THEN 0 ELSE 1 END DESC, 
							captureDate ASC 
							LIMIT 1]], albumId )
	
		:closingMapB( function ( rows )
					
			local coverAssetId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			return coverAssetId
			
		end )
		
	return resultB
		
end

--==============================================================================

return {
	AlbumMethods = AlbumMethods,
}
