--[[----------------------------------------------------------------------------

OzCatAlbumsTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFOzSyncTestUtils = require 'WFOzSyncTestUtils'
local WFAsyncTestCase = require 'WFAsyncTestCase'
local WFPlatformInfo = require 'WFPlatformInfo'
local OzCatalogTestUtils = require 'OzCatalogTestUtils'

local TEST_WAIT_TIME = 2

if WFPlatformInfo.isAndroid then
	TEST_WAIT_TIME = 6		-- PERFORMANCE_ANDROID
end

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local OzCatalogService, WFIndexStoreSchema

local T = {}
local docStore, ozUser, catalog

--------------------------------------------------------------------------------

local kAllison = string.rep( 'a', 32 )
local kBarney = string.rep( 'b', 32 )

local kPhoto1 = string.rep( '1', 32 )
local kPhoto2 = string.rep( '2', 32 )
local kPhoto3 = string.rep( '3', 32 )
local kPhoto4 = string.rep( '10', 16 )

local kAlbumAsset1 = string.rep( '4', 32 )
local kAlbumAsset2 = string.rep( '5', 32 )
local kAlbumAsset3 = string.rep( '6', 32 )
local kAlbumAsset4 = string.rep( '12', 16 )

local kAlbum1 = string.rep( '7', 32 )
local kAlbum2 = string.rep( '8', 32 )
local kAlbum3 = string.rep( '9', 32 )

local kAlbum1Name = "Album1"
local kAlbum2Name = "Album2"
local kAlbum3Name = "Album3"

local kCatalogId = string.rep( '0', 32 )

--------------------------------------------------------------------------------

local kUsers = {
	[ kAllison ] = {
		first_name = 'Alli',
		last_name = 'Ansi',
		full_name = 'Allison Ansi',
	},
	[ kBarney ] = {
		first_name = 'Barney',
		last_name = 'Barnes',
		full_name = 'Barney B Barnes',
	},
}

--------------------------------------------------------------------------------

local curTime = 0
local function newDate()
	curTime = curTime + 1
	local KSDate = require 'KSDate'
	return KSDate.timeInIso8601Format( curTime * 3600 )
end

--------------------------------------------------------------------------------

local function newAssetDoc( overlay )
	local result = {
		width = 1,
		height =1,
		_serverProvidedMetadata = {
			created = newDate(),
			links = {
				['/rels/rendition_type/144h'] = {
					href = 'somewhere',
				}
			}
		}
	}
	for k,v in pairs( overlay ) do
		result [ k ] = v
	end
	return result
end

--------------------------------------------------------------------------------

local function makeOzUser( id )
	local src = kUsers[ id ]
	local oz = {
		_id = id,
		first_name = src.first_name,
		last_name = src.last_name,
		full_name = src.full_name,
	}
	return oz
end

--------------------------------------------------------------------------------

local function addAlbumAsset( albumAssetId, albumId, assetId, updatedDate )

	docStore:setServerValueP(
		'album_asset',
		albumAssetId,
		{
			album = {
				_id = albumId,
			},
			asset = {
				_id = assetId,
			},
			updated = updatedDate,
		}
	)
	
end

--------------------------------------------------------------------------------

local function addAlbum( albumId, name, updatedTimestamp )

	docStore:setServerValueP(
		'album',
		albumId,
		{
			name = name,
			updated = updatedTimestamp,
		}
	)
	
end

--==============================================================================

function TestSuite.setUp()
	
	curTime = 0
	
		-- Utterly fake oz user.  Vends canned responses to GETs and stores
		-- of what was PUT for later analysis.
	ozUser = WFOzSyncTestUtils.makeFakeOzUser( T, makeOzUser( kAllison ) )
	
	docStore, catalog = OzCatalogTestUtils.buildMocksForTestingCatalogService( T, ozUser, kCatalogId )

end

--------------------------------------------------------------------------------

function TestSuite.tearDown()
	catalog:close()
	docStore, ozUser, catalog, OzCatalogService, WFIndexStoreSchema = nil
	T = {}
end

--==============================================================================

function TestSuite.test_albumCoverImageB_favorite_chronological( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},	
			captureDate = "2013-01-30T18:10:29",
		}
	)
	
	-- This should be the cover image since it is the first favorite in chronological order.
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kAllison = newDate(),
			},
			captureDate = "2013-01-30T18:07:29",

		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kBarney = newDate(),
				kAllison = newDate(),
			},
			captureDate = "2013-01-30T18:10:29",
		}
	)
		
	addAlbumAsset( kAlbumAsset1, kAlbum1, kPhoto1, "2013-04-09" )
	addAlbumAsset( kAlbumAsset2, kAlbum1, kPhoto2, "2013-04-09" )
	addAlbumAsset( kAlbumAsset3, kAlbum1, kPhoto3, "2013-04-09" )
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:albumCoverImageB( kAlbum1 ) )
		:continueWhenValueMatches( kPhoto2 )
		

end

--------------------------------------------------------------------------------

function TestSuite.test_albumCoverImageB_nofavorite_chronological( test )
	
	-- kPhoto1 should be the cover image since there are no favorites
	-- and this is the first asset in chronological order.
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},	
			captureDate = "2012-12-30T18:10:29",
		}
	)
		
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
			captureDate = "2013-01-30T18:07:29",

		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
			captureDate = "2013-02-30T18:10:29",
		}
	)
	
	addAlbumAsset( kAlbumAsset1, kAlbum1, kPhoto1, "2013-04-09" )
	addAlbumAsset( kAlbumAsset2, kAlbum1, kPhoto2, "2013-04-09" )
	addAlbumAsset( kAlbumAsset3, kAlbum1, kPhoto3, "2013-04-09" )
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:albumCoverImageB( kAlbum1 ) )
		:continueWhenValueMatches( kPhoto1 )
		

end


--------------------------------------------------------------------------------

function TestSuite.test_albumsCaptureDateSortedB( test )
	
	-- Create 3 albums
	
	addAlbum( kAlbum1, kAlbum1Name, "2012-01-30T01:10:29" )
	addAlbum( kAlbum2, kAlbum2Name, "2012-03-30T01:10:29" )
	addAlbum( kAlbum3, kAlbum3Name, "2012-02-30T01:10:29" )
	
	-- Create 4 assets
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},	
			captureDate = "2012-12-30T18:10:29",
		}
	)
		
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = "2013-01-30T18:07:29",

		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = "2013-02-30T18:10:29",
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto4,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = "2011-01-30T18:10:29",
		}
	)
	
	addAlbumAsset( kAlbumAsset1, kAlbum1, kPhoto1, "2013-04-09" )
	addAlbumAsset( kAlbumAsset2, kAlbum2, kPhoto2, "2013-04-09" )
	
	-- Add 2 assets to album3
	
	addAlbumAsset( kAlbumAsset3, kAlbum3, kPhoto3, "2013-04-09" )
	addAlbumAsset( kAlbumAsset4, kAlbum3, kPhoto4, "2013-04-09" )

	-- All albums have assets, they should be returned in order
	-- of the most recent captureDate (of any of their assets)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:albumsCaptureDateSortedB() )
		:continueWhenValueMatches( {
			{
				captureDate = "2013-02-30T18:10:29",
				id = "99999999999999999999999999999999",
				name = "Album3",
				updateTimeStamp = "2012-02-30T01:10:29",
			},
			{
				captureDate = "2013-01-30T18:07:29",
				id = "88888888888888888888888888888888",
				name = "Album2",
				updateTimeStamp = "2012-03-30T01:10:29",
			},
			{
				captureDate = "2012-12-30T18:10:29",
				id = "77777777777777777777777777777777",
				name = "Album1",
				updateTimeStamp = "2012-01-30T01:10:29",
			},
		}	
	)	

end

--------------------------------------------------------------------------------

function TestSuite.test_albumsCaptureDateSortedB_noAssets( test )
	
	-- create 3 albums
	addAlbum( kAlbum1, kAlbum1Name, "2012-01-30T01:10:29" )
	addAlbum( kAlbum2, kAlbum2Name, "2012-03-30T01:10:29" )
	addAlbum( kAlbum3, kAlbum3Name, "2013-04-30T01:10:29" )
	
	-- create 3 assets
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},	
			captureDate = "2012-12-30T18:10:29",
		}
	)
		
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = "2013-01-30T18:07:29",

		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = "2013-02-30T18:10:29",
		}
	)

	addAlbumAsset( kAlbumAsset1, kAlbum1, kPhoto1, "2013-04-09" )
	addAlbumAsset( kAlbumAsset2, kAlbum2, kPhoto2, "2013-04-09" )
		
	-- No assets in kAlbum3 but the updateTimestamp is most recent so that should sort first.
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:albumsCaptureDateSortedB() )
		:continueWhenValueMatches( {
			{
				id = "99999999999999999999999999999999",
				name = "Album3",
				updateTimeStamp = "2013-04-30T01:10:29",
			},
			{
				captureDate = "2013-01-30T18:07:29",
				id = "88888888888888888888888888888888",
				name = "Album2",
				updateTimeStamp = "2012-03-30T01:10:29",
			},
			{
				captureDate = "2012-12-30T18:10:29",
				id = "77777777777777777777777777777777",
				name = "Album1",
				updateTimeStamp = "2012-01-30T01:10:29",
			},
		}
	)	

end

--==============================================================================

return TestSuite