--[[----------------------------------------------------------------------------

OzCatCaptions.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require "WFArguments"

--==============================================================================

local OzCatalogServiceArg = 'OzCatalogServiceImp'

-- @mark OzCatalogService class

local CaptionMethods = {
}

--------------------------------------------------------------------------------

-- @mark captions

--------------------------------------------------------------------------------
--[[!
	\brief Sets the caption for an asset in this catalog.
	
	\param assetId (string) The unique ID of the asset.
	\param caption (string) The new caption.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function CaptionMethods:setAssetCaptionP( --[[string]] assetId,
												 --[[string]] caption )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string' )
	
    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = assetId,
        overlay = {
			caption = caption,
        },
    }

    return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Removes the caption for an asset in this catalog, if any.
	
	\param assetId (string) The unique ID of the asset.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function CaptionMethods:removeAssetCaptionP( --[[string]] assetId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = assetId,
        overlay = {
			caption = {
				_delete = true,
			},
        },
    }

    return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the caption for an asset in this catalog.
	
	\param assetId (string) The unique ID of the asset.
	
	\return <code>P < string | nil ></code> A Promise object that resolves to
		the caption, or to nil if there is no caption.
]]--

function CaptionMethods:getAssetCaptionP( --[[string]] assetId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
    local resultP = self._docStore:docValueOnceP( 'asset', assetId )
		:mapDataP( function( status, docValue )
		
			return 'data', docValue and docValue.caption
			
		end )
		
    return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the caption for an asset in this catalog.
	
	\param assetId (string) The unique ID of the asset.
	
	\return <code>B < string | nil ></code> A Behavior object whose value is
		the caption, or to nil if there is no caption.
]]--

function CaptionMethods:getAssetCaptionB( --[[string]] assetId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
  	return self._docStore:fieldValueB( 'asset', assetId,
							'caption' )
	
end

--==============================================================================

return {
	CaptionMethods = CaptionMethods,
}
