--[[----------------------------------------------------------------------------

OzCatCatalog.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSFileUtils = require "KSFileUtils"
local KSFunctionUtils = require "KSFunctionUtils"
local KSLogging = require "KSLogging"
local KSPathUtils = require "KSPathUtils"
local KSPrefs = require "KSPrefs"
local KSPromises = require "KSPromises"
local KSReactive = require "KSReactive"
local WFWorkerThread = require "WFWorkerThread"

local OzCatHelperFns = require "OzCatHelperFns"

--==============================================================================

local BabluEP = WFWorkerThread( "BabluImp.lua" )

local OzCatHelperFns_hasDownloadedSuccessfullyB = OzCatHelperFns.hasDownloadedSuccessfullyB

local canAccessLog = KSLogging.log 'OzCatalogService :: canAccessB'
	--	:setLevel( KSLogging.DEBUG )

-- @mark OzCatalogService class

local CatalogMethods = {
}

--------------------------------------------------------------------------------

-- @mark catalog access

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-04-22]: Simplify this memoizeFn pattern?

--------------------------------------------------------------------------------

local function accessErrorB( catalogId )

	return BabluEP:requestB {
		message = 'accessErrorB',
		arguments = catalogId,
		startingValue = false,
	}

end

--------------------------------------------------------------------------------

local existsOnOzB = KSFunctionUtils.memoizeFn( function( catalogId )

	return BabluEP:requestB {
		message = 'existsOnOzB',
		arguments = catalogId,
		startingValue = true,
	}

end )

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether the current user is allowed to access this catalog, and
	if not, reports why not.
 
	\return <code>B< nil | boolean >, string</code> A Behavior whose value is \c true
		if the current user is allowed to access this catalog. This value
		can change while the user is currently using the
		catalog.  The default value is nil while loading data.
			
		If the user does not have access, the Behavior value is \c false,
		and the function also returns a string describing the reason, one of:

		<ul><li>'doesNotExist' : The catalog does not exist.</li>
		   <li>'noAccess' : If the user is the catalog owner, but the user's account has
			   	completely expired or the current device is set to "cached mode" for images.</li>
		   <li>'noAccess' : This is a catalog shared with the user,
		           but the user is not currently on the collaborators list, or the owner's
		           account has expired, or the current device is set to "cached mode" for images.</li>
		   <li>'forceUpdate' : 	In this case, call \c OzCatalogService:itmsLinkP() to retrieve
			   the ITMS link via s promise. </li> </ul>

]]--

function CatalogMethods:canAccessB()

	if not self._canAccessB then

		-- Is the server reachable?
		
        local reachableB = self._ozHostIsReachableB

		-- Has this catalog's info been downloaded succesfully?
		
		local hasDownloadedB = OzCatHelperFns_hasDownloadedSuccessfullyB( self._catalogId )

		local isSyncEnabledB = self:syncEnabledB()
		
		-- Does this catalog exist on Oz?

		local existsB = existsOnOzB( self._catalogId )
			
		-- did we get a 403 of some kind?
			
		local accessErrorB = accessErrorB( self._catalogId )

		-- rules of access

		local function rules( accessError, localAccessError )

			if accessError == 4301 or localAccessError == 4301 then
					
				canAccessLog:info( "Owner's state is expired :: Checking for download originals" )

				local downloadOriginals = KSPrefs[ "DownloadOriginals" ]

				if downloadOriginals then
					canAccessLog:info( "rules == true :: download originals is set to true" )
					return true
				else
					canAccessLog:info( "rules == false :: download originals is set to false" )
					return false, 'noAccess'
				end
			else
				canAccessLog:info( "rules == true :: owner state is valid" )
				return true
			end
		end

		local function mergeResultsB( reachable, syncEnabled, hasDownloaded, exists, accessError, hasBeenDeleted, localAccessError )

			canAccessLog:info( "mergeResultsB: reachable = %s, syncEnabled = %s, hasDownloaded = %s, exists = %s, accessError = %s, hasBeenDeleted = %s localAccessError = %s",
							   reachable, syncEnabled, hasDownloaded, exists, accessError, hasBeenDeleted, localAccessError )

			-- if we just deleted the catalog, return 'false', 'doesNotExist'
			
			accessError = tonumber( accessError )

			if hasBeenDeleted then
				return false, 'doesNotExist'
			end

			if reachable ~= 'unreachable' then
			
				if localAccessError == 4999 then
					canAccessLog:info( "access denied :: local access error: got 403 with 4999 on catalog from Oz" )
					return false, 'forceUpdate'
				end
	
				if localAccessError == 4300 then
					canAccessLog:info( "access denied :: local access error: got 403 with 4300 on catalog from Oz" )
					return false, 'noAccess'
				end

				-- if the servers are reachable but this catalog has not been synced,
				-- assume the catalog is valid and return true

				if not syncEnabled then
					canAccessLog:info( "access granted :: sync not enabled" )
					return true
				end

				-- if the catalog doesn't exist on Oz, return false
				if not exists then
					canAccessLog:info( "access denied :: got 404 on catalog from Oz" )
					return false, 'doesNotExist'
				end
			
				if accessError == 4999 then
					canAccessLog:info( "access denied :: got 403 with 4999 on catalog from Oz" )
					return false, 'forceUpdate'
				end

				if accessError == 4300 then
					canAccessLog:info( "access denied :: got 403 with 4300 on catalog from Oz" )
					return false, 'noAccess'
				end
				
				local packedRulesResult = { rules( accessError, localAccessError ) }
				local rulesResult = packedRulesResult[ 1 ]
				
				-- if the catalog info has downloaded correctly, then access as per rules

				if hasDownloaded then
					if rulesResult then
						canAccessLog:info( "access granted based on rules calculation" )
					else
						canAccessLog:info( "access denied based on rules calculation" )
					end
					
					return unpack( packedRulesResult )
				else

				-- if catalog info has not downloaded correctly AND we got a 403, access as per rules

					if accessError then
						canAccessLog:info( "Got a 403 :: access according to rules" )
						if rulesResult then
							canAccessLog:info( "access granted based on rules calculation" )
						else
							canAccessLog:info( "access denied based on rules calculation" )
						end
					return unpack( packedRulesResult )
				
					else
						canAccessLog:info( "access granted :: haven't synced yet" )
						return true
					end
				end
			else
				return true

			end

		end

        self._canAccessB = KSReactive.dataCallB( mergeResultsB, reachableB, isSyncEnabledB,
							hasDownloadedB, existsB, accessErrorB, self._hasBeenDeletedB, self._accessErrorB )
		
		self:addCloseTarget( accessErrorB )
		self:addCloseTarget( self._canAccessB )

	end
	
	return self._canAccessB:relayB()

end


--------------------------------------------------------------------------------

-- @mark catalog deletion

--------------------------------------------------------------------------------
--[[!
	\brief Removes all traces of this catalog from local storage.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true when
		the operation is complete.
	
]]--

function CatalogMethods:deleteLocalCatalogAndAssetsP()

	-- Debounce requests. We tend to get duplicate requests for this.
	-- I think this is because we delete the catalog from Oz and then
	-- process the 404 catalog doesn't exist error separately from the
	-- internal call to this method.

	if self._isDeletingCatalog then
		debug.print( "Received duplicate call to deleteLocalCatalogAndAssetsP from ",
					debug.traceback() )
		return KSPromises.canceledPromiseP()
	end
	
	self._isDeletingCatalog = true

	-- hack to fire a last false access value
	self._hasBeenDeletedB:set( true )
	
	local docStore = self._docStore
	local dataCache = docStore:dataCache()

	-- First, stop syncing and wait for confirmation.

	return BabluEP:requestP( 'stopSync', self._catalogId )
		
	-- Stop the download originals agent.
	
	:mapDataP( function( )
		return BabluEP:requestP( 'stopOriginalsAgent', self._catalogId )
	end )

	-- Shut down and delete the index store and wait for confirmation.

	:mapDataP( function( )
			return self._indexStore:haltAndDeleteP()
	end )

	-- Stop local data cache.

	:mapDataP( function( status, path )

		self._indexStore:close()
		local resultP = dataCache:haltAndDeleteP()
		
		dataCache = nil
	
		return resultP
	
	end )
	
	-- Shut down the doc store and wait for confirmation.
	-- We don't delete it because it's part of the overall package
	-- that we'll be placing in the Trash.

	:mapDataP( function( )

		return docStore:halt()

	end )
	
	-- Now close the docStore and
	-- Delete local catalog and cache.

	:mapDataP( function()
	
		docStore:close()

		-- Delete (move to trash, if possible) local catalog DB folder.
		
		-- FIXME [scouten 2011-08-07]: Fragile path manipulation.
			
		local pattern = KSPathUtils.child( '', 'Originals$' )
		local localFolderPath = self._originalsFolderDocuments:gsub( pattern, '' )
		if localFolderPath then
			KSFileUtils.moveToTrash( localFolderPath )
		end
		-- Delete the caches tree- since now we may have originals stored there as well
			
		local cachesFolderPath = self._originalsFolderCaches:gsub( pattern, '' )
			
		KSFileUtils.moveToTrash( cachesFolderPath )
		
		-- Older catalogs may have originals under the old 'caches' path. Delete that too.
		
 		local oldCachesFolderPath = KSPathUtils.child( KSPathUtils.getStandardFilePath 'caches', self._catalogId )
						
		KSFileUtils.moveToTrash( oldCachesFolderPath )
		
		return 'data', true
		
	end )

end

--==============================================================================

return {
	CatalogMethods = CatalogMethods,
}
