--[[----------------------------------------------------------------------------

OzCatCollectionQueries.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'
local WFDiagnosticLog = require "WFDiagnosticLog"

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'

	-- wichitafoundation.core.utils
local WFUnknown = require 'WFUnknown'

	-- wichitafoundation.core.reactive
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.string
local KSStringUtils = require 'KSStringUtils'
local WFZString = require 'WFZString'

	-- wichitafoundation.os.prefs
local KSPrefs = require 'KSPrefs'

	-- wichitafoundation.oz_client.catalog
local OzCatEvents = require 'OzCatEvents'

local WFOzContentTypes = require 'WFOzContentTypes'


--------------------------------------------------------------------------------

local OzCatEvents_kEventCoalesceThreshold = assert( OzCatEvents.kEventCoalesceThreshold )
local OzCatEvents_decodeId = assert( OzCatEvents.decodeId )
local ZSTR = assert( WFZString.fromZString )

local unknownUserStr = ZSTR "$$$/OzCatalogService/UnknownUser=(unknown user)"

local tostring = tostring

--------------------------------------------------------------------------------

local importMetricsLog = KSLogging.log 'OzCatalogService :: metrics '
--	:setLevel( KSLogging.DEBUG )

local eventsLog = KSLogging.log 'OzCatalogService :: events '
--	:setLevel( KSLogging.DEBUG )

--==============================================================================

local kEventDateSelectPart = [[
	SELECT
		coreInfo.displayDate, coreInfo.eventName, coreInfo.eventNameLC, MIN(coreInfo.captureDate) as minCaptureDate
	FROM
		coreInfo
	]]

local kJoinToAlbumPart = [[
	INNER JOIN
		album_asset
	ON
		coreInfo.docId == album_asset.assetId
	]]

local kWhereClauseForAlbumPart = [[
	album_asset.albumId == ?
	]]

local kCollectionInfoSelectPart = [[
	SELECT
		coreInfo.docId, coreInfo.aspectRatio, coreInfo.thumbnailMD5, coreInfo.favoritedCount, coreInfo.captureDate, coreInfo.displayDate, coreInfo.timestamp, coreInfo.copyTimestamp
	FROM
		coreInfo
	]]
	
--------------------------------------------------------------------------------

local function makeJoinForAlbumId( albumId )
	if albumId then
		return kJoinToAlbumPart
	else
		return " "
	end
end
	
--------------------------------------------------------------------------------

local function makeWhereExpressionForAlbumId( albumId )
	if albumId then
		return kWhereClauseForAlbumPart
	end
end
	
--------------------------------------------------------------------------------

local function addQueryParamsForAlbumId( albumId )
	if albumId then
		return albumId
	end
end

--------------------------------------------------------------------------------

local function whereClauseForExpressions( ... )
	local n = select( "#", ... )
	local exprs = {}
	-- create table of non-nil entries
	for i = 1, n do
		local x = select( i, ... )
		if x then
			exprs[ #exprs + 1 ] = x
		end
	end
	if #exprs == 0 then
		return ""
	end
	return [[ WHERE ]] .. table.concat( exprs, " AND " ) .. " "
end

--==============================================================================

local uploadStatusCollectionNames = {

	uploaded = ZSTR "$$$/OzCatalogService/Uploaded=Uploaded",

	[ 'notUploadedStage2-original' ]
		= ZSTR "$$$/OzCatalogService/Original=Uploading Originals",
	
	[ '_conflict' ] = ZSTR "$$$/OzCatalogService/ConflictReel=Conflicts",
	
	notUploaded = ZSTR "$$$/OzCatalogService/NotUploaded=Not Uploaded",
	notUploaded_metadataOnly = ZSTR "$$$/OzCatalogService/NotUploaded=Not Uploaded",

	uploadingNow = ZSTR "$$$/OzCatalogService/UploadingNow=Uploading Now", -- not sure if this status will exist

	notUploadedNew = ZSTR "$$$/OzCatalogService/NotUploadedNew=Uploading New Photos",

	notUploadedChanged = ZSTR "$$$/OzCatalogService/NotUploadedChanged=Uploading Changes",

	uploadingOtherDevice = "$$$/OzCatalogService/uploadingOtherDevice=Uploading Originals from ^1",

	downloadingOriginal = ZSTR "$$$/OzCatalogService/uploadingFromOtherDevice=Downloading Originals",

	recoveryEffort_original = ZSTR "$$$/OzCatalogService/recoveryEffort_original=Recovering Originals",
	
	originalLost = ZSTR "$$$/OzCatalogService/originalLost=Missing",
}


--------------------------------------------------------------------------------

local function makeCollectionInfo( name, originalPhotoRows, displayDate, eventName, startDate, endDate, eventNameFirst )

	if not originalPhotoRows or #originalPhotoRows == 0 then
		return { name = name, displayDate = displayDate, photos = { }, eventName = eventName, startDate = startDate, endDate = endDate, favoritedAssets = 0, assetCount = 0, }
	end
				
	local nPhotos = #originalPhotoRows
	
	local collectionInfo = {
		name = name,
		photos = {},
		displayDate = displayDate,
		eventName = eventName,
		startDate = startDate,
		endDate = endDate,
		eventNameFirst = eventNameFirst,
        favoritedAssets = 0,
		assetCount = nPhotos,
	}
	
	local photos = collectionInfo.photos
	
	for i = 1, nPhotos do
		
		local originalPhotoRow = originalPhotoRows[ i ]
		
		photos[ i ] = {
			-- FIXME [rseeler 2012-01-13]: Since this data is collected
			-- from many different SQL queries (below), the data in the different
			-- table rows can be different, making it quite difficult
			-- to know which value is which.  It might be better for
			-- the key to be the type, not an index, IMO.
			-- [mhamburg 2012-03-02]: That, however, is a lot more table construction
			-- in and out of SQLite and this is a heavily traveled path. I agree,
			-- however, that this is excessively fragile and we should find a way to
			-- make it less so.
			
			id = originalPhotoRow[ 1 ], -- coreInfo.docId
			aspectRatio = originalPhotoRow[ 2 ], -- aspectRatio
			thumbnailMD5 = originalPhotoRow[ 3 ], -- thumbnailMD5
            favoritedCount = originalPhotoRow[ 4 ], -- favoritedCount
			captureDate = originalPhotoRow[ 5 ], --captureDate
		}

		if photos[ i ].favoritedCount and photos[ i ].favoritedCount > 0 then
            collectionInfo.favoritedAssets = collectionInfo.favoritedAssets + 1
        end
	
	end

	return collectionInfo

end

--------------------------------------------------------------------------------

-- Make an event row from a database query row.

local function makeEventRow( row )

	local displayDate = row[ 1 ] or ''
	local eventName = row[ 2 ] or ''
	local eventNameLC = row[ 3 ] or ''
	
	if displayDate == '' then
	
		-- if this row has no display date, skip it
		eventsLog:debug( " Skipping row without a displayDate " )
		
	elseif eventName == '' or eventNameLC == '' then
	
		-- if this row has no event name, skip it
		eventsLog:debug( " Skipping row without an event name " )
		
	else
	
		eventsLog:debug( " Event Name %s ", eventName )

		return {
			eventName = eventName,
			startDate = displayDate,
			endDate = displayDate,
			eventNameLC = eventNameLC
		}
								
	end

end

--------------------------------------------------------------------------------

-- Coalesce two event rows if possible. Returns true if they are coalesced
-- and false otherwise.

local function coalesceEventRows( prevRow, nextRow )

	if prevRow == nil or nextRow == nil then
		return false
	end
	
	if prevRow.eventNameLC ~= nextRow.eventNameLC then
		return false
	end
	
	if prevRow.eventName:sub( 1, 1 ) == '~' then
		return false
	end
	
	if OzCatEvents_kEventCoalesceThreshold < KSDate.differenceInDays( nextRow.endDate, prevRow.startDate ) then
		return false
	end
	
	-- Coalesce nextRow into prevRow by moving startDate back
	
	prevRow.startDate = nextRow.startDate
	prevRow.eventName = nextRow.eventName
	
	return true

end
														
--------------------------------------------------------------------------------

local function eventDateCollectionContentsQueryComponents(
						self, activeAlbum, eventNameLC, startDate, endDate )

	-- If this is an event collection
	eventsLog:debug(" Event Name LC %s", eventNameLC )
	eventsLog:debug(" Start Date %s", startDate )
	eventsLog:debug(" End Date %s", endDate)
	
	local orderBy = [[ ORDER BY captureDate, timestamp ASC, case when copyTimestamp is null then 0 else 1 end, copyTimestamp DESC, coreInfo.docId]]
	
	local whereClauseExpressions = {
		[[coreInfo.eventNameLC == ?]],
	}
	local sqlParams = {
		eventNameLC,
	}
	
	if startDate == endDate then
		
		whereClauseExpressions[ #whereClauseExpressions + 1 ] = [[coreInfo.displayDate == ?]]
		sqlParams[ #sqlParams + 1 ] = startDate
	
	else

		whereClauseExpressions[ #whereClauseExpressions + 1 ] = [[coreInfo.displayDate >= ?]]
		whereClauseExpressions[ #whereClauseExpressions + 1 ] = [[coreInfo.displayDate <= ?]]
		sqlParams[ #sqlParams + 1 ] = startDate
		sqlParams[ #sqlParams + 1 ] = endDate
	
	end

	whereClauseExpressions[ #whereClauseExpressions + 1 ] = makeWhereExpressionForAlbumId( activeAlbum )
	sqlParams[ #sqlParams + 1 ] = addQueryParamsForAlbumId( activeAlbum )
	
	local nArgs = #sqlParams
	
	local cacheKey = string.format( "%s:%s:%s:%s",
		tostring( activeAlbum ),
		tostring( eventNameLC ),
		tostring( startDate ),
		tostring( endDate ) )
	
	return {
		cacheKey = cacheKey,
		sql = kCollectionInfoSelectPart
			.. makeJoinForAlbumId( activeAlbum )
			.. whereClauseForExpressions( unpack( whereClauseExpressions ) )
			.. orderBy,
		nArgs = nArgs,
		unpack( sqlParams, 1, nArgs )
	}

end

--------------------------------------------------------------------------------

-- Build out the query logic for date queries. We can optionally filter the
-- results based on fields in coreInfo (though using items other than event name
-- or display date could have complex consequences). We use the same collection
-- identifiers in each case because we expect the collections to be shared across
-- different date views.

local function makeDateQueryBuilder( whereClauseOrig )

	return {
	
		queryRootCollectionInfoB = function( self, activeAlbum )
		
			-- Note: We do the join and the min capture date calculation based on the event name rather than the lowercase event name because
			-- we want the enhanced granularity. Selecting distinct displayDate, eventName, eventNameLC is the same as selectin distinct
			-- displayDate, eventName since eventNameLC is determined by eventName.
			
			local groupByOrderByPredicate = [[
				GROUP BY
					displayDate, eventName
				ORDER BY
					displayDate DESC, case when eventName != '' then 1 else '9999-01-01' end, minCaptureDate DESC ]]
									
			local dateQueryB

			dateQueryB = self._indexStore:queryB {
				sql = kEventDateSelectPart
					.. makeJoinForAlbumId( activeAlbum )
					.. whereClauseForExpressions( whereClauseOrig, makeWhereExpressionForAlbumId( activeAlbum ) )
					.. groupByOrderByPredicate,
				addQueryParamsForAlbumId( activeAlbum ),
				calm = 1.5
			}

			-- Coalesce tracks with the same event that are close enough. This closeness is defined in terms of days.
			-- Look for 'OzCatEvents_kEventCoalesceThreshold' defined in the beginning of this file for the current value.
			-- Example: If the value of OzCatEvents_kEventCoalesceThreshold is 3 and:
			-- If 12/9/2011 and 12/12/2011 have the same event name, they will be coalesced into one row.
			-- If 12/9/2011 and 12/13/2011 have the same event name, they will be NOT be coalesced.
			
			return dateQueryB:closingMapB( function( originalDisplayDateRows )

				-- Get import metrics
				local startTime = KSDate.timeNowForAnimation()

				-- We go through the data set returned one by one and keep adding to a "event-date" table
				-- Each event row has the following fields - eventName, startDate, endDate
				-- Each date row has the fields - displayDate
				local newDateEventRows = {}
				
				-- We also maintain a table of the unique event names we encounter in order to determine if
				-- a row can be combined with a previously found event of the same name.
				-- Each row in this table is indexed by the event name and the value is its row# in newDateEventRows
				-- The latest event of that name replaces the previous event with that name in this table
				-- For example: if there is an event of name "Foo" on 2011-12-16. And another event of the name "Foo"
				-- on 2011-12-12. This event will replace the earlier occurance in the this table once encountered.
				
				local latestEvents = {}
				
				-- This function moves rows up by 1 within newDateEventRows from the given index
				-- to the end of the table
				
				local function moveRowsUp( index )
				
					if #newDateEventRows - index > 10 then
						debug.print( "OzCatalogService: collectionQueryBuilders Moved more than 10 rows for multi-day event sort." )
					end
				
					for i = index, #newDateEventRows - 1 do
						local moveRow = newDateEventRows[ i + 1 ]
						newDateEventRows[ i ] = moveRow
						-- Update the latest event rows with the new row number for this event
						if moveRow.eventName then
							latestEvents[ moveRow.eventNameLC ] = i
						end
					end
					
					-- Remove the last row, since we just moved it up
					
					newDateEventRows[ #newDateEventRows ] = nil
					
				end

				for i = 1, #originalDisplayDateRows do
				
					local row = originalDisplayDateRows[ i ]
				
					local eventName = row[ 2 ]
					local displayDate = row[ 1 ]
					
					if not displayDate then
					
						-- if this row has no display date, skip it
						eventsLog:debug( " Skipping row without a displayDate " )
						
					elseif eventName == '' then
					
						-- if this row has no associated event, add a track for the date
							local newDate = {}
							newDate.displayDate = displayDate
							newDateEventRows[ #newDateEventRows + 1 ] = newDate
					
					else
					
						eventsLog:debug( " Event Name %s ", eventName )
				
						-- We check if this row can squished with a previously encountered event
						-- If the eventName is the same as the previous event that is within the OzCatEvents_kEventCoalesceThreshold,
						-- squish this row with the previous one. NOTE: the rows are in DESC order
						
						local eventNameLC = row[ 3 ]

						local existingEvent = latestEvents[ eventNameLC ]
						
						local existingEventRow = existingEvent and newDateEventRows[ existingEvent ]
																				
						if existingEvent and KSDate.differenceInDays( displayDate, existingEventRow.startDate ) <= OzCatEvents_kEventCoalesceThreshold then
							-- if this previous event was within the threshold
							existingEventRow.startDate = displayDate
							existingEventRow.eventName = eventName
							existingEventRow.eventNameLC = eventNameLC
							-- Copy over the row for this event to its new location within the list
							newDateEventRows[ #newDateEventRows + 1 ] = existingEventRow
							moveRowsUp( existingEvent )
							latestEvents[ eventNameLC ] = #newDateEventRows
						else
							-- insert a new event for this displayDate, with the start and end dates the same
							local newEvent = {}
							newEvent.eventName = eventName
							newEvent.startDate = displayDate
							newEvent.endDate = displayDate
							newEvent.eventNameLC = eventNameLC
							newDateEventRows[ #newDateEventRows + 1 ] = newEvent
							-- Don't allow coalescing for names beginning with tilde
							if eventName:sub( 1, 1 ) ~= '~' then
								latestEvents[ eventNameLC ] = #newDateEventRows
							end
						end
							
					end

				end

				-- WINRT_FIXME [bparadie 2012-05-29]: remove this if clause when we are
				-- able to require this namespace in Win.

				-- Get import metrics
				if WFDiagnosticLog ~= nil then

					local timeConsumed = KSDate.timeNowForAnimation() - startTime

					local resultP = WFDiagnosticLog.makeEnhancedRequestP( 'shouldPrintImportMetricsP', 'makeDateQueryBuilder', timeConsumed )
				
					resultP:whenResolvedDo( function( status, data )
				   
					   if status == 'data' and data == true then
							importMetricsLog:debug("  makeDateQueryBuilder: Time spent = %.3f seconds", timeConsumed )
						end

					end ):runUntilDone()
				end
				return newDateEventRows
				
			end )

		end,
				
		mapRootCollectionInfoFn = function( self )
		
			return function( originalDisplayDateRows )
			
				if not originalDisplayDateRows then return nil end

				local newResult = {}
				for i = 1, #originalDisplayDateRows do

					local originalDisplayDateRow = originalDisplayDateRows[ i ]
					
					local id = originalDisplayDateRow.displayDate or ( originalDisplayDateRow.eventName .. '\t' ..
							   originalDisplayDateRow.startDate .. '\t' ..
							   originalDisplayDateRow.endDate .. '\t' ..
							   originalDisplayDateRow.eventNameLC )

					newResult[ i ] = {
						id = 'date\t' .. id,
						lcname = originalDisplayDateRow.displayDate or originalDisplayDateRow.startDate,
						name = originalDisplayDateRow.displayDate or originalDisplayDateRow.startDate,
						eventNameFirst = false
					}

					-- Add the event name, start date and end date if this is an event track

					if originalDisplayDateRow.eventName then
						newResult[ i ].eventName = originalDisplayDateRow.eventName
						newResult[ i ].startDate = originalDisplayDateRow.startDate
						newResult[ i ].endDate = originalDisplayDateRow.endDate
					else
						newResult[ i ].displayDate = originalDisplayDateRow.displayDate
					end

				end
				
				return { collections = newResult }
			
			end

		end,
		
		queryCollectionInfoQueryComponents = function( self, activeAlbum, collectionId )

			eventsLog:debug(" Collection Id is %s", collectionId )
			local collectionIdDecoded = OzCatEvents_decodeId( collectionId )
			
			-- If this is a date collection
			if not collectionIdDecoded[ 2 ] then
				return eventDateCollectionContentsQueryComponents(
							self, activeAlbum, '', collectionIdDecoded[ 1 ], collectionIdDecoded[ 1 ] )
			else
				return eventDateCollectionContentsQueryComponents(
							self, activeAlbum, collectionIdDecoded[ 4 ], collectionIdDecoded[ 2 ], collectionIdDecoded[ 3 ] )
			end

		end,
		
		mapCollectionInfoFn = function( self, collectionId )
        
			return function( originalPhotoRows )
			
				local collectionIdDecoded = OzCatEvents_decodeId( collectionId )
				
				-- The collection id is either just the display date or it is a sequence of
				-- event name, start date, end date, event name LC
				
				local displayDate = collectionIdDecoded[ 2 ] or collectionIdDecoded[ 1 ]
					-- start date if present otherwise we're in the display date case
				
				local eventName = collectionIdDecoded[ 2 ] and collectionIdDecoded[ 1 ]
					-- we can only have an event name if we have a start date
				
				return makeCollectionInfo(
							displayDate,
							originalPhotoRows,
							displayDate,
							eventName,
							collectionIdDecoded[ 2 ],
							collectionIdDecoded[ 3 ],
							false )
			
			end
			
		end,
	
	}

end

--------------------------------------------------------------------------------

local function makeMapUploadStatusCollectionInfoFn( self, collectionId )

	return function( originalPhotoRows )
	
		local components = OzCatEvents_decodeId( collectionId )
	
		return makeCollectionInfo(
					uploadStatusCollectionNames[ components[ 1 ] ],
					originalPhotoRows )
		
	end

end

--------------------------------------------------------------------------------

local function sortAddedByList( a, b )

	if a.userGroupSeq ~= b.userGroupSeq then
		return a.userGroupSeq < b.userGroupSeq
	end

	if a.lcname ~= b.lcname then
		return a.lcname < b.lcname -- localized comparison?
	end
	
	return a.id < b.id

end

--------------------------------------------------------------------------------

local sCollectionQueryBuilders = {
	
	--------------------------------------------------------------------------------
	-- @mark Oz Upload Status
	--------------------------------------------------------------------------------

	ozUploadStatus = {
	
		queryRootCollectionInfoB = function( self, activeAlbum )
		
			return self._indexStore:queryB {
				sql = [[SELECT DISTINCT uploadStatusWithDeviceId FROM ozUploadStatusForView AS coreInfo ]]
							-- This is a bit of a hack: rename table to "coreInfo" so this inner join will work
					.. makeJoinForAlbumId( activeAlbum )
					.. whereClauseForExpressions( makeWhereExpressionForAlbumId( activeAlbum ) )
					.. [[ ORDER BY uploadStatusWithDeviceId]],
				addQueryParamsForAlbumId( activeAlbum ),
				calm = 1.5
			}
			
		end,
				
		mapRootCollectionInfoFn = function( self )
			return function( originalUploadStatusRows )
			
				if not originalUploadStatusRows then return nil end
		
				local originalUploadStatusRow
				local newResult = {}

				-- Need to keep a separate count because we're squelching rows.

				local count = 1
				for i = 1, #originalUploadStatusRows do

					originalUploadStatusRow = originalUploadStatusRows[ i ]
					
					-- If we're not downloading originals on this machine, don't show the 'Downloading Originals' track.
					-- Lua having 'continue' support would be nice here.

					if originalUploadStatusRow[ 1 ] == 'downloadingOriginal' and not KSPrefs[ 'DownloadOriginals' ] then

						-- do nothing
						
					else

						-- create the string to be displayed for this row and add it to the collection names array
					
						local nameStr
						if originalUploadStatusRow[ 1 ]:match( '^uploadingOtherDevice' ) then
							nameStr = ZSTR( uploadStatusCollectionNames.uploadingOtherDevice, originalUploadStatusRow[ 1 ]:sub( 21 ) )
								-- 21 = first char position after 'uploadingOtherDevice'
						else
							nameStr = uploadStatusCollectionNames[ originalUploadStatusRow[ 1 ] ]
						end
	
						uploadStatusCollectionNames[ originalUploadStatusRow[ 1 ] ] = nameStr

						newResult[ count ] = {
							id = 'status\t' .. originalUploadStatusRow[ 1 ], -- ozUploadStatus
							lcname = originalUploadStatusRow[ 1 ],  -- ozUploadStatus (again)
							name = nameStr,
						}

						count = count + 1

					end

				end

				return { collections = newResult }
			
			end
			
		end,
		
		queryCollectionInfoQueryComponents = function( self, activeAlbum, collectionId )
		
			local components = OzCatEvents_decodeId( collectionId )

			return {
				sql = [[
					SELECT
						coreInfo.docId, aspectRatio, thumbnailMD5, favoritedCount, captureDate]] ..
												-- This is a bit of a hack: rename table to "coreInfo" so this inner join will work
					[[ FROM
						ozUploadStatusForView AS coreInfo
					]] .. makeJoinForAlbumId( activeAlbum )
					.. whereClauseForExpressions(
						[[uploadStatusWithDeviceId == ?]],
						makeWhereExpressionForAlbumId( activeAlbum )
					) .. [[
					ORDER BY captureDate DESC, coreInfo.docId]],
				components[ 1 ], addQueryParamsForAlbumId( activeAlbum ),
				calm = .5,
			}
				
		end,
		
		mapCollectionInfoFn = makeMapUploadStatusCollectionInfoFn,
	
	},
	
	--------------------------------------------------------------------------------
	-- @mark Keywords
	--------------------------------------------------------------------------------

	keywords = {
	
		queryRootCollectionInfoB = function( self, activeAlbum )
		
			return self._indexStore:queryB(
					[[SELECT DISTINCT lcKeyword, keyword FROM keywords AS coreInfo
						]] .. makeJoinForAlbumId( activeAlbum ) .. [[
						]] .. whereClauseForExpressions(
							makeWhereExpressionForAlbumId( activeAlbum ) )
						.. [[ ORDER BY lcKeyword]], addQueryParamsForAlbumId( activeAlbum ) )
			
		end,
				
		mapRootCollectionInfoFn = function( self )
			return function( originalKeywordRows )
			
				if not originalKeywordRows then return nil end
		
				local newResult = {}
				for i = 1, #originalKeywordRows do
					local originalKeywordRow = originalKeywordRows[ i ]
					newResult[ i ] = {
						id = 'keyword\t' .. originalKeywordRow[ 1 ], -- lcKeyword
						lcname = originalKeywordRow[ 1 ],  -- lcKeyword (again)
						name = originalKeywordRow[ 2 ], -- keyword
					}
				end
				
				return { collections = newResult }
			
			end
			
		end,
		
		queryCollectionInfoQueryComponents = function( self, activeAlbum, collectionId )
		
			local components = OzCatEvents_decodeId( collectionId )
			
			return [[SELECT coreInfo.docId, aspectRatio, thumbnailMD5, favoritedCount, keyword, captureDate from coreInfo
					INNER JOIN keywords
						]] .. makeJoinForAlbumId( activeAlbum )
						.. whereClauseForExpressions( 
							[[coreInfo.docId == keywords.docId]],
							[[lcKeyword == ?]],
							makeWhereExpressionForAlbumId( activeAlbum )
						)
						.. [[ ORDER BY captureDate, coreInfo.docId]],
											components[ 1 ], addQueryParamsForAlbumId( activeAlbum )
			
		end,
		
		mapCollectionInfoFn = function( self, collectionId )
		
			return function( originalPhotoRows )
			
				local components = OzCatEvents_decodeId( collectionId )
				-- FIXME [scouten 2011-04-07]: This returns the same keyword
				-- for every photo, but we only read it from the first photo.
				-- That's kind of wasteful but I don't (yet) have a way to
				-- pass results from multiple queries back.
				return makeCollectionInfo(
						components[ 1 ], -- keyword from first photo
							originalPhotoRows )

			end
		
		end,
	
	},
	
	--------------------------------------------------------------------------------
	-- @mark Date
	--------------------------------------------------------------------------------

	date = makeDateQueryBuilder(),
	
	--------------------------------------------------------------------------------
	-- @mark Date Events Only
	--------------------------------------------------------------------------------

	dateEventsOnly = makeDateQueryBuilder( "eventName != ''" ),

	--------------------------------------------------------------------------------
	-- @mark Event
	--------------------------------------------------------------------------------

	events = {
	
		-- static function, returns an array of expressions to limit query across all coreInfo
		totalCountExpressions = function( )
			return { [[coreInfo.eventName != '']] }
		end,
		
		queryRootCollectionInfoB = function( self, activeAlbum )
		
			-- Note: We do the join and the min capture date calculation based on the event name rather than the lowercase event name because
			-- we want the enhanced granularity. Selecting distinct displayDate, eventName, eventNameLC is the same as selectin distinct
			-- displayDate, eventName since eventNameLC is determined by eventName.
				
			local eventQueryB
			
				eventQueryB = self._indexStore:queryB(
				kEventDateSelectPart .. makeJoinForAlbumId( activeAlbum )
					.. whereClauseForExpressions(
						[[eventName != '']],
						makeWhereExpressionForAlbumId( activeAlbum )
					) .. [[
						GROUP BY displayDate, eventName
						ORDER BY eventNameLC ASC, displayDate DESC, minCaptureDate DESC ]], addQueryParamsForAlbumId( activeAlbum ) )
					
			-- Coalesce tracks with the same event that are close enough. This closeness is defined in terms of days.
			-- Look for 'OzCatEvents_kEventCoalesceThreshold' defined in the beginning of this file for the current value.
			-- Example: If the value of OzCatEvents_kEventCoalesceThreshold is 3 and:
			-- If 12/9/2011 and 12/12/2011 have the same event name, they will be coalesced into one row.
			-- If 12/9/2011 and 12/13/2011 have the same event name, they will be NOT be coalesced.
			
			return eventQueryB:closingMapB( function( originalDisplayEventRows )
			
				-- We go through the data set returned one by one and keep adding to a "event-date" table
				-- Each event row has the following fields - eventName, startDate, endDate
				-- Each date row has the fields - displayDate
				
				local newEventDateRows = { }
				
				for i = 1, #originalDisplayEventRows do
				
					local row = makeEventRow( originalDisplayEventRows[ i ] )
					
					if not coalesceEventRows( newEventDateRows[ #newEventDateRows ], row ) then
						newEventDateRows[ #newEventDateRows + 1 ] = row
					end
				
				end
				
				table.sort( newEventDateRows, function( a, b )
					return( KSStringUtils.localizedStandardStringCompare( a.eventName, b.eventName ) < 0 )
				end )
				
				return newEventDateRows
				
			end )

		end,
				
		mapRootCollectionInfoFn = function( self )
		
			return function( originalDisplayEventRows )
			
				if not originalDisplayEventRows then return nil end

				local newResult = {}
				for i = 1, #originalDisplayEventRows do

					local originalDisplayEventRow = originalDisplayEventRows[ i ]
					
					local id = originalDisplayEventRow.eventName .. '\t' ..
							   originalDisplayEventRow.startDate .. '\t' ..
							   originalDisplayEventRow.endDate .. '\t' ..
							   originalDisplayEventRow.eventNameLC
							   
					newResult[ i ] = {
						id = 'event\t' .. id,
						name = originalDisplayEventRow.eventName,
						lcname = originalDisplayEventRow.eventNameLC,
						eventName = originalDisplayEventRow.eventName,
						startDate = originalDisplayEventRow.startDate,
						endDate = originalDisplayEventRow.endDate,
						eventNameFirst = true
					}
					
					-- debug.print( "Collection id: ", id )

				end
				
				return { collections = newResult }
			
			end

		end,
		
		queryCollectionInfoQueryComponents = function( self, activeAlbum, collectionId )

			eventsLog:debug(" Collection Id is %s", collectionId )

			local collectionIdDecoded = OzCatEvents_decodeId( collectionId )
			
			return eventDateCollectionContentsQueryComponents(
						self, activeAlbum, collectionIdDecoded[ 4 ], collectionIdDecoded[ 2 ], collectionIdDecoded[ 3 ] )

		end,
		
		mapCollectionInfoFn = function( self, collectionId )
        
			return function( originalPhotoRows )
			
				local collectionIdDecoded = OzCatEvents_decodeId( collectionId )
				
				return makeCollectionInfo(
							collectionIdDecoded[ 1 ], -- event name
							originalPhotoRows,
							collectionIdDecoded[ 2 ], -- start date as display date
							collectionIdDecoded[ 1 ], -- event name
							collectionIdDecoded[ 2 ], -- start date
							collectionIdDecoded[ 3 ], -- end date
							true )
					
			end
			
		end,
	
	},
	
	
	--------------------------------------------------------------------------------
	-- @mark Added By
	--------------------------------------------------------------------------------

	addedBy = {
	
		queryRootCollectionInfoB = function( self, activeAlbum )
		
			local tracksQuery

			tracksQuery = self._indexStore:queryB(
					[[SELECT DISTINCT importedBy FROM coreInfo
						]] .. makeJoinForAlbumId( activeAlbum )
						.. whereClauseForExpressions(
							makeWhereExpressionForAlbumId( activeAlbum )
						) .. [[
							ORDER BY importedBy]], addQueryParamsForAlbumId( activeAlbum ) )
			
			-- Now, the fun part: We get to mix in the tracks query with
			-- a query to retrieve all users of this catalog.
			
			local usersB = self:catalogUsersB()
			local currentUserB = KSPrefs.prefB( 'currentUserId' )
			
			return KSReactive.dataCallB(
				function( tracks, users, currentUser )
					if WFUnknown.isKnown( tracks ) and users then
						local t = {}
						for k, v in pairs( tracks ) do
							t[ k ] = v
						end
						t.users = users
						t.currentUser = currentUser
						return t
					end
				end,
				tracksQuery, usersB, currentUserB )
				:addCloseTarget( tracksQuery )
			
		end,
				
		mapRootCollectionInfoFn = function( self )

			return function( originalAddedByRows )
			
				if not originalAddedByRows then return nil end
				
				local users = originalAddedByRows.users
				local currentUser = originalAddedByRows.currentUser

				local usersById = {}
				for i = 1, #users do
					local user = users[ i ]
					if user._id then
						usersById[ user._id ] = user
					end
				end
				
				local newResult = {}
				for i = 1, #originalAddedByRows do

					local originalAddedByRow = originalAddedByRows[ i ]
					local userId = originalAddedByRow[ 1 ]
					local userInfo = usersById[ userId ] or { full_name = unknownUserStr }
					local name = userInfo.full_name

					local displayName = userInfo.full_name
					
					if userId == currentUser then
						displayName = ZSTR( "$$$/OzCatalogService/Me=^1 ^U+2014 Me", displayName )
					elseif userInfo.role == 'owner' then
						displayName = ZSTR( "$$$/OzCatalogService/CatalogOwner=^1 ^U+2014 Library Owner", displayName )
					end

					newResult[ i ] = {
						id = string.format( 'addedBy\t%s\t%s', userId, displayName ),
						lcname = KSStringUtils.lower( name ),
						name = name,
						userGroupSeq = userInfo.role == 'owner' and 1
										or ( userId == currentUser and 2 )
										or ( not usersById[ userId ] and 4 )
										or 3,
					}
					
				end
				
				table.sort( newResult, sortAddedByList )
				
				return { collections = newResult }
			
			end
			
		end,
		
		queryCollectionInfoQueryComponents = function( self, activeAlbum, collectionId )
		
			local components = OzCatEvents_decodeId( collectionId )
			
			return kCollectionInfoSelectPart
						.. makeJoinForAlbumId( activeAlbum )
						.. whereClauseForExpressions(
							[[importedBy == ?]],
							makeWhereExpressionForAlbumId( activeAlbum )
						)
						.. [[ ORDER BY captureDate, coreInfo.docId]],
					components[ 1 ], addQueryParamsForAlbumId( activeAlbum )
			
		end,
		
		mapCollectionInfoFn = function( self, collectionId )
		
			return function( originalPhotoRows )
			
				local components = OzCatEvents_decodeId( collectionId )

				return makeCollectionInfo( components[ 2 ], originalPhotoRows )

			end
		
		end,
	
	},
	

}

--------------------------------------------------------------------------------

local function getBuilder( whichView )
	
	assert( type( whichView ) == 'string', "Cannot construct builder for view: invalid whichView" )
	return assert( sCollectionQueryBuilders[ whichView ], "Cannot construct builder for view : builder for view " .. whichView )

end

--------------------------------------------------------------------------------
-- synthesize queryCollectionInfoB and queryCollectionInfoP from queries
-- if necessary

for k, v in pairs( sCollectionQueryBuilders ) do
	local queryComponentsFn = v.queryCollectionInfoQueryComponents
	if queryComponentsFn then
		if not v.queryCollectionInfoP then
			local function maybeConvertTableBForm( self, maybeParams, ... )
				if type( maybeParams ) == 'table' then
					return self._indexStore:queryP( maybeParams.sql, unpack( maybeParams ) )
				else
					-- no, not a table of params like for queryB
					-- maybeParams is actually our sql stmt
					return self._indexStore:queryP( maybeParams, ... )
				end
			end
			v.queryCollectionInfoP = function( self, ... )
				return maybeConvertTableBForm( self, queryComponentsFn( self, ... ) )
			end
		end
		if not v.queryCollectionInfoB then
			v.queryCollectionInfoB = function( self, ... )
				return self._indexStore:queryB( queryComponentsFn( self, ... ) )
			end
		end
	end
	
	-- Add the queryCollectionCountB and queryCollectionCountP from totalCountExpressions
	local totalCountExpressions = {}
	if v.totalCountExpressions then
		totalCountExpressions = v.totalCountExpressions()
	end
	local function queryComponents( self, activeAlbum )
		local curExprs = { unpack( totalCountExpressions ) }
		curExprs[ #curExprs + 1 ] = makeWhereExpressionForAlbumId( activeAlbum )

		return [[
			SELECT
				count( coreInfo.rowid )
			FROM
				coreInfo
			]] .. makeJoinForAlbumId( activeAlbum ) .. 
				whereClauseForExpressions( unpack( curExprs ) )
	end
	local function cleanupCountQuery( rows )
		if rows and WFUnknown.isKnown( rows ) then
			return rows[ 1 ] [ 1 ]
		end
		return rows
	end
	v.queryCollectionCountB = function( self, activeAlbum )
		return self._indexStore:queryB( queryComponents( self, activeAlbum ) )
			:closingMapB( cleanupCountQuery )
	end
	v.queryCollectionCountP = function( self, activeAlbum )
		return self._indexStore:queryP( queryComponents( self, activeAlbum ) )
			:mapDataP( function( status, rows )
				return 'data', cleanupCountQuery( rows )
			end )
	end
end

--------------------------------------------------------------------------------

-- Optimization; copy the keys from OzCatCollections.lua into a new table that maps
-- to only true.  This lets some universes, e.g. KSLVC use only this function and
-- the big block of code above will be GC-able.
local sValidViewModes = {}

for k, v in pairs( sCollectionQueryBuilders ) do
	sValidViewModes[ k ] = true
end

local function isValidViewGrouping( whichView )
	
	if type( whichView ) == 'string' then
		-- not not converts true/nil to true/false
		return not not sValidViewModes[ whichView ]
	end

end

--------------------------------------------------------------------------------

local function tornado_makeCollectionInfo( originalPhotoRows )

	if not originalPhotoRows or #originalPhotoRows == 0 then
		return { photos = { }, favoritedAssets = 0, assetCount = 0, contentTypeNumber = 0 }
	end
				
	local nPhotos = #originalPhotoRows
	
	local collectionInfo = {
		photos = {},
        favoritedAssets = 0,
		assetCount = nPhotos,
		contentTypeNumber = 0
	}
	
	local photos = collectionInfo.photos

	for i = 1, nPhotos do
		
		local originalPhotoRow = originalPhotoRows[ i ]
		
		photos[ i ] = {
			-- FIXME [rseeler 2012-01-13]: Since this data is collected
			-- from many different SQL queries (below), the data in the different
			-- table rows can be different, making it quite difficult
			-- to know which value is which.  It might be better for
			-- the key to be the type, not an index, IMO.
			-- [mhamburg 2012-03-02]: That, however, is a lot more table construction
			-- in and out of SQLite and this is a heavily traveled path. I agree,
			-- however, that this is excessively fragile and we should find a way to
			-- make it less so.
			
			id = originalPhotoRow[ 1 ], -- coreInfo.docId
			aspectRatio = originalPhotoRow[ 2 ], -- aspectRatio
            favoritedCount = originalPhotoRow[ 3 ], -- favoritedCount
			captureDate = originalPhotoRow[ 4 ], --captureDate
			importedBy = originalPhotoRow[ 5 ], --importedBy
			favoritedUsers = originalPhotoRow[ 6 ], -- favoritedUsers
			contentType = WFOzContentTypes.contentTypeForContentTypeNumber( originalPhotoRow[ 7 ], false ) -- contentType
		}


		if photos[ i ].favoritedCount and photos[ i ].favoritedCount > 0 then
            collectionInfo.favoritedAssets = collectionInfo.favoritedAssets + 1
        end

	end

--debug.print("collectionInfo.assetCount = ", collectionInfo.assetCount)
	return collectionInfo

end

--==============================================================================

return {
	getBuilder = getBuilder,
	makeCollectionInfo = makeCollectionInfo,
	isValidViewGrouping = isValidViewGrouping,
	tornado_makeCollectionInfo = tornado_makeCollectionInfo,
}
