--[[----------------------------------------------------------------------------

OzCatCollectionQueriesTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSReactive = require "KSReactive"
local WFOzSyncTestUtils = require "WFOzSyncTestUtils"
local WFAsyncTestCase = require "WFAsyncTestCase"
local WFPlatformInfo = require 'WFPlatformInfo'
local OzCatCollectionQueriesTestData = require 'OzCatCollectionQueriesTestData'

local TEST_WAIT_TIME = 2

if WFPlatformInfo.isAndroid then
	TEST_WAIT_TIME = 6		-- PERFORMANCE_ANDROID
end

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local OzCatCollectionQueries

local T = {}
local docStore, indexStore, ozCatalogService

--------------------------------------------------------------------------------

local function makeTrimP( min, max )
	return function( status, result )
		-- It is not good form to mutate a table handed to you in a map fn.
		local cpy = {}
		for i, row in ipairs( result or {} ) do
			-- New rows beyond 5 added in my my refactoring, but
			-- I don't believe they hurt anything. -- arahn
			cpy[ i ] = { unpack( row, min, max ) }
		end
		return status, cpy
	end
end

--------------------------------------------------------------------------------

local function makeTrimB( min, max )
	return function( result )
		-- It is not good form to mutate a table handed to you in a map fn.
		local cpy = {}
		for i, row in ipairs( result or {} ) do
			-- New rows beyond 5 added in my my refactoring, but
			-- I don't believe they hurt anything. -- arahn
			cpy[ i ] = { unpack( row, min, max) }
		end
		return cpy
	end
end

--------------------------------------------------------------------------------

local function buildMocks( docStore )
	
	return {
		KSFileUtils = {
		
		},
		
		OzCatalogIndexDefinitions = false,
		
		-- Fake implementation which uses OzCatalogIndexDefinitions to build out the proper
		-- schema and mapping functions for our in-memory indexStore
		WFIndexStoreSchema = docStore:raw_test_buildMockWFIndexStoreSchema(),
		
		-- Fake impl, does nothing
		WFIndexStoreImp = {
			start = function( schema )
			
			end,
		},
	}
	
end

--------------------------------------------------------------------------------

local function catalogUsersB() 

	return KSReactive.constantB( { '1023df59ef5441dbac5f0310984925ae', } )
	
end
--------------------------------------------------------------------------------

function TestSuite.setUpClass()
	
	-- Build a fake docStore, sufficient for testing this namespace
	
	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	OzCatCollectionQueries = WFOzSyncTestUtils.importWithMocks( 'OzCatCollectionQueries', buildMocks( docStore ) )

	-- Build a fake indexStore.  It's powered by a real sql db in memory
	-- that uses the real schema and mapping fns.
	indexStore = docStore:raw_test_buildMockIndexStore( )
	
	-- fake OzCatalogService that only has an indexStore reference
	ozCatalogService = { 
		_indexStore = indexStore, 
		catalogUsersB = catalogUsersB,
	}
	
	-- Build the fake index store with our fake data
	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, OzCatCollectionQueriesTestData.docs )

	
end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_date( test )

	-- 'date' view mode
	
	local builder = OzCatCollectionQueries.getBuilder( 'date' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'date	2013-01-30' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'date	2013-01-30' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = false,
			id = "date	Lala	2013-02-06	2013-02-06	lala",
			lcname = "2013-02-06",
			name = "2013-02-06",
			startDate = "2013-02-06",
		},
		{
			displayDate = "2013-01-30",
			eventNameFirst = false,
			id = "date	2013-01-30",
			lcname = "2013-01-30",
			name = "2013-01-30",
		},
		{
			displayDate = "2012-12-12",
			eventNameFirst = false,
			id = "date	2012-12-12",
			lcname = "2012-12-12",
			name = "2012-12-12",
		},
		{
			endDate = "2011-11-25",
			eventName = "Soup",
			eventNameFirst = false,
			id = "date	Soup	2011-11-25	2011-11-25	soup",
			lcname = "2011-11-25",
			name = "2011-11-25",
			startDate = "2011-11-25",
		},
	},	} )
	-- Step 2: Match queryCollectionInfoB
	
	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.340",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.924",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP
	
	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.340",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.924",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_date_album( test )

	-- 'date' view mode with an active album

	local builder = OzCatCollectionQueries.getBuilder( 'date' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'date	2013-01-30' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'date	2013-01-30' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = false,
			id = "date	Lala	2013-02-06	2013-02-06	lala",
			lcname = "2013-02-06",
			name = "2013-02-06",
			startDate = "2013-02-06",
		},
		{
			displayDate = "2013-01-30",
			eventNameFirst = false,
			id = "date	2013-01-30",
			lcname = "2013-01-30",
			name = "2013-01-30",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.924",
		},
	} )
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
			[6] = "2013-01-30",
			[7] = "2013-04-05T12:10:31.924",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_dateEventsOnly( test )

	-- view mode 'dateEventsOnly'
	
	local builder = OzCatCollectionQueries.getBuilder( 'dateEventsOnly' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'date	Lala	2013-02-06	2013-02-06	lala' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'date	Lala	2013-02-06	2013-02-06	lala' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end

	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = false,
			id = "date	Lala	2013-02-06	2013-02-06	lala",
			lcname = "2013-02-06",
			name = "2013-02-06",
			startDate = "2013-02-06",
		},
		{
			endDate = "2011-11-25",
			eventName = "Soup",
			eventNameFirst = false,
			id = "date	Soup	2011-11-25	2011-11-25	soup",
			lcname = "2011-11-25",
			name = "2011-11-25",
			startDate = "2011-11-25",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )

	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_dateEventsOnly_album( test )

	-- collection 'dateEventsOnly' with an active album
	
	local builder = OzCatCollectionQueries.getBuilder( 'dateEventsOnly' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'date	Lala	2013-02-06	2013-02-06	lala' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'date	Lala	2013-02-06	2013-02-06	lala' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = false,
			id = "date	Lala	2013-02-06	2013-02-06	lala",
			lcname = "2013-02-06",
			name = "2013-02-06",
			startDate = "2013-02-06",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_events( test )

	-- view mode 'events'

	local builder = OzCatCollectionQueries.getBuilder( 'events' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'event	Lala	2013-02-06	2013-02-06	lala' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'event	Lala	2013-02-06	2013-02-06	lala' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = true,
			id = "event	Lala	2013-02-06	2013-02-06	lala",
			lcname = "lala",
			name = "Lala",
			startDate = "2013-02-06",
		},
		{
			endDate = "2011-11-25",
			eventName = "Soup",
			eventNameFirst = true,
			id = "event	Soup	2011-11-25	2011-11-25	soup",
			lcname = "soup",
			name = "Soup",
			startDate = "2011-11-25",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )

	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_events_album( test )

	-- view mode 'events' with an active album

	local builder = OzCatCollectionQueries.getBuilder( 'events' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'event	Lala	2013-02-06	2013-02-06	lala' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'event	Lala	2013-02-06	2013-02-06	lala' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			endDate = "2013-02-06",
			eventName = "Lala",
			eventNameFirst = true,
			id = "event	Lala	2013-02-06	2013-02-06	lala",
			lcname = "lala",
			name = "Lala",
			startDate = "2013-02-06",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data', {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
			[6] = "2013-02-06",
			[7] = "2013-04-05T12:10:33.046",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_ozUploadStatus( test )

	-- view mode 'ozUploadStatus'
	
	local builder = OzCatCollectionQueries.getBuilder( 'ozUploadStatus' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'status	notUploadedChanged' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'status	notUploadedChanged' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "status	notUploadedChanged",
			lcname = "notUploadedChanged",
			name = "Uploading Changes",
		},
	},	} )

	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = "2012-12-12T14:29:43",
		},
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = "2011-11-25T11:57:59",
		},
	} )

	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data',	{
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = "2012-12-12T14:29:43",
		},
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = "2011-11-25T11:57:59",
		},
	} )


end

--------------------------------------------------------------------------------


function TestSuite.testOzCatCollectionQueries_ozUploadStatus_album( test )

	-- view mode 'ozUploadStatus' with an active album

	local builder = OzCatCollectionQueries.getBuilder( 'ozUploadStatus' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'status	notUploadedChanged' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'status	notUploadedChanged' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "status	notUploadedChanged",
			lcname = "notUploadedChanged",
			name = "Uploading Changes",
		},
	},	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB )
		:continueWhenValueMatches( {
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP )
		:continueWhenValueMatches( 'data',{
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_keywords( test )

	-- view mode 'keywords' 

	local builder = OzCatCollectionQueries.getBuilder( 'keywords' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'keyword	1' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'keyword	1' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "keyword	(no keywords)",
			lcname = "(no keywords)",
			name = "(no keywords)",
		},
		{
			id = "keyword	1",
			lcname = "1",
			name = 1,
		},
	}, 
	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB
		:mapB( makeTrimB( 1, 6 ) ) )
		:continueWhenValueMatches( {
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = 1,
			[6] = "2011-11-25T11:57:59",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = 1,
			[6] = "2012-12-12T14:29:43",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP
		:mapDataP( makeTrimP( 1, 6 ) ) )
		:continueWhenValueMatches( 'data',	{
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = 1,
			[6] = "2011-11-25T11:57:59",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = 1,
			[6] = "2012-12-12T14:29:43",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_keywords_album( test )

	-- view mode 'keywords' with an active album

	local builder = OzCatCollectionQueries.getBuilder( 'keywords' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'keyword	(no keywords)' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'keyword	(no keywords)' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "keyword	(no keywords)",
			lcname = "(no keywords)",
			name = "(no keywords)",
		},
	}, 
	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB
		:mapB( makeTrimB( 1, 6 ) ) )
		:continueWhenValueMatches( {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "(no keywords)",
			[6] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "(no keywords)",
			[6] = "2013-02-06T18:55:46",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP
		:mapDataP( makeTrimP( 1, 6 ) ) )
		:continueWhenValueMatches( 'data', {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "(no keywords)",
			[6] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "(no keywords)",
			[6] = "2013-02-06T18:55:46",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_addedBy( test )

	-- view mode 'addedBy'

	local builder = OzCatCollectionQueries.getBuilder( 'addedBy' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , nil )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, nil, 'addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, nil, 'addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)",
			lcname = "(unknown user)",
			name = "(unknown user)",
			userGroupSeq = 4,
		},
	}, 
	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB
		:mapB( makeTrimB( 1, 5 ) ) )
		:continueWhenValueMatches( {
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = "2011-11-25T11:57:59",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = "2012-12-12T14:29:43",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP
		:mapDataP( makeTrimP( 1, 5 ) ) )
		:continueWhenValueMatches( 'data',	{
		{
			"cdc7da3f51d0431b80dd78dacf862268",
			0.75,
			[4] = 0,
			[5] = "2011-11-25T11:57:59",
		},
		{
			"90b49fc134434fddb32d7b193e72476c",
			0.75,
			[4] = 0,
			[5] = "2012-12-12T14:29:43",
		},
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"6a1fc6dd5796449abc349a93164335b6",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
	} )


end

--------------------------------------------------------------------------------

function TestSuite.testOzCatCollectionQueries_addedBy_album( test )

	-- view mode 'addedBy' with an active album

	local builder = OzCatCollectionQueries.getBuilder( 'addedBy' )

	local queryRootB = builder.queryRootCollectionInfoB( ozCatalogService , 'ee69c819bf5f49c88d51449254f0d0b5' )
	
	local queryCollectionInfoB = builder.queryCollectionInfoB( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)' )
	
	local queryCollectionInfoP = builder.queryCollectionInfoP( ozCatalogService, 'ee69c819bf5f49c88d51449254f0d0b5', 'addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)' )
	
	local mapFn = builder.mapRootCollectionInfoFn( ozCatalogService )
	
	local mappedQueryRootB
	
	if mapFn then
		mappedQueryRootB = queryRootB:mapB( mapFn )
	else 
		mappedQueryRootB = queryRootB
	end
	
	-- Step 1: Match queryRootCollectionInfoB

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( mappedQueryRootB  )
		:continueWhenValueMatches( { collections = {
		{
			id = "addedBy	1023df59ef5441dbac5f0310984925ae	(unknown user)",
			lcname = "(unknown user)",
			name = "(unknown user)",
			userGroupSeq = 4,
		},
	}, 
	} )
	
	-- Step 2: Match queryCollectionInfoB

	:waitFor( queryCollectionInfoB
		:mapB( makeTrimB( 1, 5 ) ) )
		:continueWhenValueMatches( {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
	} )
	
	-- Step 3: Match queryCollectionInfoP

	:waitFor( queryCollectionInfoP
		:mapDataP( makeTrimP( 1, 5 ) ) )
		:continueWhenValueMatches( 'data', {
		{
			"089422de03444fd29753e96b2e47f88b",
			0.75,
			[4] = 0,
			[5] = "2013-01-30T18:10:29",
		},
		{
			"7c3ca81508d841b28f57ed83893c6ade",
			0.75,
			[4] = 0,
			[5] = "2013-02-06T18:55:46",
		},
	} )

end

--------------------------------------------------------------------------------

return TestSuite
