--[[----------------------------------------------------------------------------

OzCatCollections.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require "KSLogging"

	-- wichitafoundation.core.reactive
local KSPromises = require "KSPromises"
local KSReactive = require "KSReactive"

	-- wichitafoundation.core.utils
local WFArguments = require "WFArguments"

	-- wichitafoundation.oz_client.catalog
local OzCatCollectionQueries = require "OzCatCollectionQueries"


--============================================================================--

local unpack = unpack

local OzCatalogServiceArg = 'OzCatalogServiceImp'
local kCollectionInfoFields = "coreInfo.docId, aspectRatio, favoritedCount, captureDate, displayDate, timestamp, copyTimestamp"

local OzCatCollectionQueries_getBuilder = assert( OzCatCollectionQueries.getBuilder )
local OzCatCollectionQueries_makeCollectionInfo = assert( OzCatCollectionQueries.makeCollectionInfo )
local OzCatCollectionQueries_tornado_makeCollectionInfo = assert( OzCatCollectionQueries.tornado_makeCollectionInfo )


--============================================================================--

-- @mark OzCatalogService class

local OzCatCollectionMethods = {}

--------------------------------------------------------------------------------

local catalogLog = KSLogging.log 'OzCatalogService'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local function makeAlbumPredicate( activeAlbum )
	
   return
        [[INNER JOIN album_asset
                    ON coreInfo.docId == album_asset.assetId
                        WHERE album_asset.albumId == ? AND (coreInfo.contentTypeNumber == 1001 OR coreInfo.contentTypeNumber == 1002)]],
		{ activeAlbum }

end

--------------------------------------------------------------------------------

local function builderForView( whichView )

	return OzCatCollectionQueries_getBuilder( whichView )

end

--------------------------------------------------------------------------------

local function filterByFavorites( photoInfo )

	return photoInfo.favoritedCount and photoInfo.favoritedCount > 0

end

--------------------------------------------------------------------------------

local function getFilterFn( filter )

	if filter == 'favorites' then
		return filterByFavorites
	end
	
	return filter

end

--------------------------------------------------------------------------------

local function filterCollectionInfo( rawInfo, filter )

	local filterFn = getFilterFn( filter )
	
	if not filterFn then return rawInfo end
	
	local result = {}

	for k, v in pairs( rawInfo ) do

		if k == 'photos' then
		
			local nResultPhotos = 0
			local photos = {}
			
			for i = 1, #v do
				local photoInfo = v[ i ]
				if filterFn( photoInfo ) then
					nResultPhotos = nResultPhotos + 1
					photos[ nResultPhotos ] = photoInfo
				end
			end

			result[ k ] = photos
		
		else
			result[ k ] = v
		end
	
	end
	
	return result

end

--------------------------------------------------------------------------------

local function makeAdHocCollectionQueryAndParams( activeAlbum, ascending )

    local order = 'DESC'
    
    if ascending then
        order = 'ASC'
    end

	local predicate = ""
	local queryParams = {}

	if activeAlbum then
        predicate, queryParams = makeAlbumPredicate( activeAlbum )
	end
	
	return
			[[SELECT ]] .. kCollectionInfoFields .. 
			[[ FROM coreInfo INNER JOIN TEMP_ids ON coreInfo.docId == TEMP_ids.docId ]]			
			.. predicate ..
			[[ ORDER BY captureDate ]] .. order .. [[, coreInfo.docId]], queryParams
			
end

--------------------------------------------------------------------------------

function OzCatCollectionMethods:setViewGrouping( viewMode )
	error "View grouping has moved to the catalog data source"
end

--------------------------------------------------------------------------------

function OzCatCollectionMethods:getViewModeB()
	error "View grouping has moved to the catalog data source"
end

--------------------------------------------------------------------------------
--[[!
	\brief Monitors the total number of assets in this catalog.
	
	\warning Use this Behavior sparingly; it reduces performance in
		large catalogs. When possible, use \c :catalogIsEmptyB() instead.
	
	\return <code>B< number ></code> A Behavior object whose value is the total
		asset count for this catalog.
--]]

function OzCatCollectionMethods:getAssetCountB()

	local resultB = self._indexStore:queryB {
		sql = [[SELECT count( rowid ) FROM coreInfo]],
		calm = 2
	}

	return resultB
		:closingMapB( function( data )
			return data and data[ 1 ] and data[ 1 ][ 1 ]
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Monitors whether this catalog has any assets in it.
	
	Generally preferred to \c :getAssetCountB(), as it
	places a much lighter load on the index store thread.
	
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if there are no assets in the catalog
--]]

function OzCatCollectionMethods:catalogIsEmptyB()

	if not self._catalogIsEmptyB then

		-- SQLite + index store trickiness: SELECT 1 will return an empty result
		-- set if there are no rows and a single 1 = 1 row for every row.
		-- Combine that with LIMIT 1 and you have a behavior that doesn't change
		-- value except on the transition between zero and non-zero number of assets.
		-- Also: It's very, very fast for SQLite to evaluate this query.
	
		self._catalogIsEmptyB = self._indexStore:queryB {
			sql = [[ SELECT count( a ) FROM ( SELECT 1 AS a FROM coreInfo LIMIT 1 ) ]],
			calm = 1,
		}
			:mapB( function( data )

				local result = false
				if data and #data > 0 then
					result = data [ 1 ] and data[ 1 ][ 1 ] == 0
				end

				catalogLog:debug( "catalogIsEmptyB got %s, returning %s", data, result )
				return result

			end )
	
	end
	
	return self._catalogIsEmptyB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns count of assets for a particular viewGrouping and (optional)
	activeAlbum.

	\param viewGrouping (string) Which view is active.
	\param activeAlbum (optional, string) If present, limit the tracks to those
		present within this album.
 
	\return <code>B< number ></code> A Behavior object whose value is the count of
		photos present in all the tracks for this view grouping (and, if provided
		album). It may be WFUnknown if the value is not yet known.
]]--

function OzCatCollectionMethods:rootCollectionCountB( --[[string]] viewGrouping,
													 --[[string]] activeAlbum )
	local builder = builderForView( viewGrouping )
	
	return builder.queryCollectionCountB( self, activeAlbum )
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the list of tracks that should be visible for this catalog.

	\todo Describe the permissible values for viewGrouping.

	\param viewGrouping (string) Which view is active.
	\param activeAlbum (optional, string) If present, limit the tracks to those
		present within this album.
 
	\return <code>B< array ></code> A Behavior object whose value is the list of
		track IDs in the overall catalog or album.
]]--

function OzCatCollectionMethods:rootCollectionInfoB( --[[string]] viewGrouping,
													 --[[string]] activeAlbum )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string|boolean' )

	local builder = builderForView( viewGrouping )
	
	if not builder then
		return KSReactive.constantB( { collections = { } } )
	end
	
	local queryB = builder.queryRootCollectionInfoB( self, activeAlbum )
	
	local mapFn = builder.mapRootCollectionInfoFn( self )
	
	if not mapFn then
		return queryB
	end
	
	local mappedB = queryB:mapB( mapFn )
	
	mappedB:addCloseTarget( queryB )
	
	return mappedB

end

--------------------------------------------------------------------------------
--[[!
	\brief Are there any tracks in the specified view?

	\todo Describe the permissible values for viewGrouping.

	\param viewGrouping (string) Which view is active.
	\param activeAlbum (optional, string) If present, limit the tracks to those
		present within this album.
 
	\return <code>B< array ></code> A Behavior object whose value is is \c true
		if there are no tracks in the album or catalog.
]]--

function OzCatCollectionMethods:rootCollectionInfoIsEmptyB( --[[string]] viewGrouping,
															--[[string]] activeAlbum )

	-- FIXME [mhamburg 2012-03-09]: We should actually just watch for whether we
	-- have a query result back rather than using a delay. We might want a delay
	-- for showing a compute state message.
	
	WFArguments.check( OzCatalogServiceArg, 'string', '?string|boolean' )

	local delay = 0.5
	
	local rootCollectionInfoB = self:rootCollectionInfoB( viewGrouping )
	
	local isEmptyB = rootCollectionInfoB:mapB( function( info )
		return ( info == nil )
			or ( info.collections == nil )
			or ( info.collections[ 1 ] == nil )
	end )
	
	local isEmptyDelayedB = isEmptyB:mapSwitchB( function( isEmpty )
		if isEmpty then
			return KSReactive.falseThenTrueB( delay )
		else
			return KSReactive.constantB( false )
		end
	end ):addCloseTarget( isEmptyB )
	
	return isEmptyDelayedB

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves information about favorite assets in this catalog.
	
	\todo Need better documentation for this function. It's sort of ambiguous now.
	
	\param sortOrder (optional, string) The sort order for the assets list,
		'FAVS' (favorites at top). ((makes no sense - only getting favorites?))
		If not supplied, order is by date. ((is this right?))

	\return <code>B< table ></code> A Behavior object whose value is ((what?)).
]]--

function OzCatCollectionMethods:favoritesCollectionInfoB( --[[optional string]] sortOrder )

    local order = 'ALL'
    debug.print("favoritesCollectionInfoB called = ", sortOrder )
    if sortOrder then
        order = sortOrder
    end

    local sortChanged = false

    if self._favoriteSortOrder ~= order then
        sortChanged = true
        self._favoriteSortOrder = order

    end

    if not self._favoriteSortOrderCollectionInfoB or sortChanged then
       
        if self._favoriteSortOrderCollectionInfoB then
            self._favoriteSortOrderCollectionInfoB:close()
        end

        self._favoriteSortOrderCollectionInfoB = self._activeAlbumB:mapSwitchB( function( activeAlbum )
        local resultB

        if order == 'FAVS' then
            resultB = self._indexStore:queryB(
                [[SELECT ]] .. kCollectionInfoFields .. [[ FROM coreInfo
                        ORDER BY favoritedCount DESC,  displayDate ASC, coreInfo.docId]] )
        else
                resultB = self._indexStore:queryB(
                    [[SELECT ]] .. kCollectionInfoFields .. [[ FROM coreInfo
                        ORDER BY captureDate ASC, coreInfo.docId]] )
        end

        return resultB:mapB( function( originalPhotoRows )

            return OzCatCollectionQueries_makeCollectionInfo(
                'events', -- Collection id for favorite sort, not used by UI.
                originalPhotoRows,
                true )

            end )

        end )

        self:addCloseTarget( self._favoriteSortOrderCollectionInfoB )

    end

    return self._favoriteSortOrderCollectionInfoB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the assets in this catalog or an album in this catalog
		for display in the global grid mode.
	
	\param activeAlbum (optional, string) If present, show only the assets present
		within this album. If not, show assets throughout the entire catalog. 
		activeAlbum can refer to a server album and have a 32 char UUID or refer 
		to 'virtual' albums that are not present on the server (see examples below).
		Virtual albums supported currently are favorites (set of assets favorited 
		by any user) and per-user favorites (set of assets favorited by a particular user).
		
		Examples:
		activeAlbum	= '6a1fc6dd5796449abc349a93164335b6' : Real album on Oz. 
						The collection will be filtered to return only the assets 
						in this album.
		activeAlbum = 'favorites1023df59ef5441dbac5f0310984925ae': Virtual album.
						The collection will be filtered to return only the assets 
						favorited by this user. The string is 'favorites' 
						concatenated with the user id.
		activeAlbum = 'favorites': Virtual album.
						The collection will be filtered to return assets that have
						been favorited by any user.
				
	\param ascending (optional, boolean) If \c true, date order is ascending, otherwise
		order is descending.

	\return <code>B< table ></code> A Behavior whose value is the list of asset IDs.
]]--

function OzCatCollectionMethods:globalGridCollectionInfoB( --[[optional string]] activeAlbum,
														   --[[optional boolean]] ascending )

	WFArguments.check( OzCatalogServiceArg, '?string|boolean', 'boolean' )

    local order = 'DESC'
    
    if ascending then
        order = 'ASC'
    end

	local resultB
	local predicate = ""
	local queryParams = {}

	if activeAlbum then
        predicate, queryParams = makeAlbumPredicate( activeAlbum )
	end
	
    resultB = self._indexStore:queryB(
			[[SELECT ]] .. kCollectionInfoFields .. [[ FROM coreInfo ]]
			.. predicate ..
    		[[ ORDER BY captureDate ]] .. order .. [[, coreInfo.docId]], unpack( queryParams ) )

	return resultB:closingMapB( function( originalPhotoRows )
        
		return OzCatCollectionQueries_tornado_makeCollectionInfo(
				originalPhotoRows,
				true )

	end )
							
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the assets in this catalog or an album in this catalog
		for display in the global grid mode.
		
	\param activeAlbum (optional, string) If present, show only the assets present
		within this album. If not, show assets throughout the entire catalog.
		activeAlbum can refer to a server album and have a 32 char UUID or refer 
		to 'virtual' albums that are not present on the server (see examples below).
		Virtual albums supported currently are favorites (set of assets favorited 
		by any user) and per-user favorites (set of assets favorited by a particular user).
		
		Examples:
		activeAlbum	= '6a1fc6dd5796449abc349a93164335b6' : Real album on Oz. 
						The collection will be filtered to return only the assets 
						in this album.
		activeAlbum = 'favorites1023df59ef5441dbac5f0310984925ae': Virtual album.
						The collection will be filtered to return only the assets 
						favorited by this user.The string is 'favorites' 
						concatenated with the user id.
		activeAlbum = 'favorites': Virtual album.
						The collection will be filtered to return assets that have
						been favorited by any user.

	\param ascending (optional, boolean) If \c true, date order is ascending, otherwise
		order is descending.
	\param useCurrentSortOrder (boolean) When \c true, sort ascending or descending
		according to current preference. ((Default behavior??))

	\todo Remove useCurrentSortOrder parameter. This should move up to the data source.

	\return <code>B< table ></code> A Behavior whose value is the list of asset IDs.
]]--

function OzCatCollectionMethods:globalGridCollectionInfoP( --[[optional string]] activeAlbum,
														   --[[optional boolean]] ascending,
														   --[[optional boolean]] useCurrentSortOrder )

	WFArguments.check( OzCatalogServiceArg, '?string|boolean', '?boolean', '?boolean' )

	local order
	if useCurrentSortOrder then
		order = self._globalGridSortOrder
	else
		order = 'DESC'
    
		if ascending then
       		order = 'ASC'
		end
		
	end

	local resultP
	
	-- Ensure value is safe to use in SQL.

	if order ~= 'DESC' then
		order = 'ASC'
	end

	local predicate = ""
	local queryParams = {}

	if activeAlbum then
        predicate, queryParams = makeAlbumPredicate( activeAlbum )
	end

	resultP = self._indexStore:queryP(
			[[SELECT ]] .. kCollectionInfoFields .. [[ FROM coreInfo ]]
			.. predicate ..
    		[[ ORDER BY captureDate ]] .. order .. [[, coreInfo.docId]], unpack( queryParams ) )

	return resultP:mapDataP( function( status, originalPhotoRows )
    	            
		return status, OzCatCollectionQueries_makeCollectionInfo(
				'grid', -- Collection id for grid mode, not used by UI.
				originalPhotoRows,
				true )
				
	end )
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the assets in this catalog or an album in this catalog
		for display in the global grid mode.
		
	\todo Describe the permissible values for viewGrouping and filter.

	\param viewGrouping (string) Which view is active.
	\param activeAlbum (optional, string) If present, show only the assets present
		within this album. If not, show assets throughout the entire catalog.
	\param collectionId The unique ID of a collection in the album. This ID should
		have been obtains from the results of a previous call to \c :rootCollectionInfoB.
	\param filter ((??)) Collect only members that match this
		((what? metadata val? how supplied? Optional?))

	\return <code>B< table ></code> A Behavior whose value is the list of asset IDs.
]]--

function OzCatCollectionMethods:collectionInfoB( --[[string]] viewGrouping,
												 --[[optional string]] activeAlbum,
												 --[[string]] collectionId,
												 --[[optional string]] filter )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string|boolean', 'string', '?string|boolean' )

	if not collectionId then
		return KSReactive.constantB( nil )
	end
	
	local builder = builderForView( viewGrouping )
	
	if not builder then
		return KSReactive.constantB( nil )
	end

	local queryB = builder.queryCollectionInfoB( self, activeAlbum, collectionId )
	
	local mapFn = builder.mapCollectionInfoFn( self, collectionId )
	
	local rawInfoB
	
	if mapFn then
		rawInfoB = queryB:mapB( mapFn )
	else
		rawInfoB = queryB
	end
	
	local resultB = rawInfoB:mapB( function( rawInfo )
		return filterCollectionInfo( rawInfo, filter )
	end )
	
	resultB:addCloseTarget( queryB )
	
	return resultB
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the assets in this catalog or an album in this catalog
		for display in the global grid mode.
		
	\todo Describe the permissible values for viewGrouping and filter.

	\param viewGrouping (string) Which view is active.
	\param activeAlbum (optional, string) If present, show only the assets present
		within this album. If not, show assets throughout the entire catalog.
	\param collectionId The unique ID of a collection in the album.
		((is this right?))
	\param optFilter ((??)) Collect only members that match this
		((what? metadata val? how supplied? default behavior?))

	\return <code>P< table ></code> A Promise object that resolves to
		the list of asset IDs when the operation is complete.
]]--


function OzCatCollectionMethods:collectionInfoP( --[[string]] viewGrouping,
												 --[[optional string]] activeAlbum,
												 --[[string]] collectionId,
												 --[[optional string]] optFilter )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string|boolean', 'string', '?string|boolean' )
	
	local builder = builderForView( viewGrouping )
	
	if not builder then
		return KSPromises.resolvedPromiseP( 'data', nil )
	end

	local queryP = builder.queryCollectionInfoP( self, activeAlbum, collectionId )
	
	local mapFn = builder.mapCollectionInfoFn( self, collectionId )
	
	return queryP:mapDataP( function ( status, data )
		 
		local rawInfo = mapFn( data )
		if optFilter then
			return status, filterCollectionInfo( rawInfo, optFilter )
		else
			return status, rawInfo
		end

	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the collection info for the asset IDs specified.
	
	\param assetIds (table) The assetIds for which the info is requested.		
	\param activeAlbum (optional, string) If present, information will
		only be returned for the requested assets in this particular album.
		activeAlbum can refer to a server album and have a 32 char UUID or refer 
		to 'virtual' albums that are not present on the server (see examples below).
		Virtual albums supported currently are favorites (set of assets favorited 
		by any user) and per-user favorites (set of assets favorited by a particular user).
		
		Examples:
		activeAlbum	= '6a1fc6dd5796449abc349a93164335b6' : Real album on Oz. 
						The collection will be filtered to return only the assets 
						in this album.
		activeAlbum = 'favorites1023df59ef5441dbac5f0310984925ae': Virtual album.
						The collection will be filtered to return only the assets 
						favorited by this user.The string is 'favorites' 
						concatenated with the user id.
		activeAlbum = 'favorites': Virtual album.
						The collection will be filtered to return assets that have
						been favorited by any user.

	\param ascending (optional, boolean) If \c true, date order is ascending, otherwise
		order is descending.

	\return <code>B< table ></code> Behavior object that resolves to
		a table with the collection info for the provided asset IDs.
]]--

function OzCatCollectionMethods:adHocCollectionInfoB( --[[table]] assetIds,
													  --[[optional string]] activeAlbum,
													  --[[optional boolean]] ascending )

	WFArguments.check( OzCatalogServiceArg, 'table', '?string|boolean', '?boolean' )
	
	local sql, queryParams = makeAdHocCollectionQueryAndParams( activeAlbum, ascending )
	
	local resultB = self._indexStore:queryB( { sql = sql, tempIds = assetIds, unpack( queryParams ) } )

	return resultB:closingMapB( function( originalPhotoRows )
        
		return OzCatCollectionQueries_makeCollectionInfo(
				'adHoc', -- collection id for adHoc collections
				originalPhotoRows,
				true )

	end )
							
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the collection info for the asset IDs specified.
	
	\param assetIds (table) The assetIds for which the info is requested.		
	\param activeAlbum (optional, string) If present, information will
		only be returned for the requested assets in this particular album.
		activeAlbum can refer to a server album and have a 32 char UUID or refer 
		to 'virtual' albums that are not present on the server (see examples below).
		Virtual albums supported currently are favorites (set of assets favorited 
		by any user) and per-user favorites (set of assets favorited by a particular user).
		
		Examples:
		activeAlbum	= '6a1fc6dd5796449abc349a93164335b6' : Real album on Oz. 
						The collection will be filtered to return only the assets 
						in this album.
		activeAlbum = 'favorites1023df59ef5441dbac5f0310984925ae': Virtual album.
						The collection will be filtered to return only the assets 
						favorited by this user.The string is 'favorites' 
						concatenated with the user id.
		activeAlbum = 'favorites': Virtual album.
						The collection will be filtered to return assets that have
						been favorited by any user.

	\param ascending (optional, boolean) If \c true, date order is ascending, otherwise
		order is descending.

	\return <code>P< table ></code> Promise object that resolves to
		a table with the collection info for the provided asset IDs.
]]--
function OzCatCollectionMethods:adHocCollectionInfoP( --[[ table ]] assetIds,
													  --[[optional string]] activeAlbum,
													  --[[optional boolean]] ascending )

	WFArguments.check( OzCatalogServiceArg, 'table', '?string|boolean', '?boolean' )

	local sql, queryParams = makeAdHocCollectionQueryAndParams( activeAlbum, ascending )

    local resultP = self._indexStore:queryP( { sql = sql, tempIds = assetIds, unpack( queryParams ) } )

	return resultP:mapDataP( function( status, originalPhotoRows )
        
		return status, OzCatCollectionQueries_makeCollectionInfo(
				'adHoc', -- collection id for adHoc collections
				originalPhotoRows,
				true )

	end )
							
end


--------------------------------------------------------------------------------
--[[!
	 Retrieves information about favorite assets in this catalog. 
	
	\param sortOrder (optional, string) The sort order for the assets list, 
		'FAVS' (favorites at top). ((makes no sense - only getting favorites?))
		If not supplied, order is by date. ((is this right?))

	\return <code>B< table ></code> A Behavior object whose value is ((what?)).
]]--

function OzCatCollectionMethods:tornadoCollectionInfoB( sortOrder, userId )

    local order = 'ALL'

    if sortOrder then
        order = sortOrder
    end

    local sortChanged = false
    local toggleForUserID = false


    if self._favoriteSortOrder ~= order then
        sortChanged = true
        self._favoriteSortOrder = order
    end

    if self._photoFilterUserID ~= userId then
        toggleForUserID = true
        if userId ~= nil then
            self._photoFilterUserID = userId
        else
            self._photoFilterUserID = false
        end
    end

    if sortChanged or toggleForUserID then    

        local resultB

        if order == 'FAVS' then
            if userId ~= nil  then
                resultB = self._indexStore:queryB(
                    [[SELECT coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber
                        FROM coreInfo
                            WHERE   importedBy == ? AND favoritedCount > 0
                                ORDER BY favoritedCount DESC,  captureDate ASC, coreInfo.docId]], userId )
            else
                resultB = self._indexStore:queryB(
                    [[SELECT coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber			
						FROM coreInfo
							WHERE favoritedCount > 0
								ORDER BY favoritedCount DESC,  captureDate ASC, coreInfo.docId]] )
            end
        elseif order == 'COMMENTS' then
            if userId ~= nil  then
                resultB = self._indexStore:queryB(
                    [[SELECT coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber
                        FROM coreInfo
                        	INNER JOIN comment
								ON coreInfo.docId == comment.assetId
                             		WHERE  importedBy == ?
                                        GROUP BY coreInfo.docId, comment.assetId
                                            ORDER BY COUNT(coreInfo.docId == comment.assetId) DESC, coreInfo.docId]], userId )
            else
                 resultB = self._indexStore:queryB(
                    [[SELECT  coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber
      					FROM coreInfo
                            INNER JOIN comment
                                    ON coreInfo.docId == comment.assetId
                                        GROUP BY coreInfo.docId, comment.assetId
                                            ORDER BY COUNT(coreInfo.docId == comment.assetId) DESC,  coreInfo.docId]] )
            end

        else
                if userId ~= nil  then
                    resultB = self._indexStore:queryB(
                        [[SELECT coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber
                            FROM coreInfo
                                WHERE   importedBy == ?
                                    ORDER BY captureDate DESC, coreInfo.docId]], userId )
                else
                    resultB = self._indexStore:queryB(
                        [[SELECT coreInfo.docId, aspectRatio, favoritedCount, captureDate, importedBy, favoritedUsers, contentTypeNumber
							FROM coreInfo
								ORDER BY captureDate DESC, coreInfo.docId]] )
                end

 				
        end

        return resultB:mapB( function( originalPhotoRows ) 
            
            return OzCatCollectionQueries_tornado_makeCollectionInfo( originalPhotoRows )

            end )

        end 

end

--============================================================================--

return {
	CollectionMethods = OzCatCollectionMethods,
	filterCollectionInfo = filterCollectionInfo,
}
