--[[----------------------------------------------------------------------------

OzCatCollectionsTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFOzSyncTestUtils = require 'WFOzSyncTestUtils'
local WFAsyncTestCase = require 'WFAsyncTestCase'
local WFPlatformInfo = require 'WFPlatformInfo'
local OzCatalogTestUtils = require 'OzCatalogTestUtils'

local TEST_WAIT_TIME = 2

if WFPlatformInfo.isAndroid then
	TEST_WAIT_TIME = 6		-- PERFORMANCE_ANDROID
end

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local OzCatalogService, WFIndexStoreSchema

local T = {}
local docStore, ozUser, catalog

--------------------------------------------------------------------------------

local kAllison = string.rep( 'a', 32 )
local kBarney = string.rep( 'b', 32 )

local kPhoto1 = string.rep( '1', 32 )
local kPhoto2 = string.rep( '2', 32 )
local kPhoto3 = string.rep( '3', 32 )

local kCatalogId = string.rep( '0', 32 )

--------------------------------------------------------------------------------

local kUsers = {
	[ kAllison ] = {
		first_name = 'Alli',
		last_name = 'Ansi',
		full_name = 'Allison Ansi',
	},
	[ kBarney ] = {
		first_name = 'Barney',
		last_name = 'Barnes',
		full_name = 'Barney B Barnes',
	},
}

--------------------------------------------------------------------------------

local curTime = 0
local function newDate()
	curTime = curTime + 1
	local KSDate = require 'KSDate'
	return KSDate.timeInIso8601Format( curTime * 3600 )
end

--------------------------------------------------------------------------------

local function newAssetDoc( overlay )
	local result = {
		width = 1,
		height =1,
		_serverProvidedMetadata = {
			created = newDate(),
			links = {
				['/rels/rendition_type/144h'] = {
					href = 'somewhere',
				}
			}
		}
	}
	for k,v in pairs( overlay ) do
		result [ k ] = v
	end
	return result
end

--------------------------------------------------------------------------------

local function makeOzUser( id )
	local src = kUsers[ id ]
	local oz = {
		_id = id,
		first_name = src.first_name,
		last_name = src.last_name,
		full_name = src.full_name,
	}
	return oz
end

--==============================================================================

function TestSuite.setUp()
	
	curTime = 0
	
		-- Utterly fake oz user.  Vends canned responses to GETs and stores
		-- of what was PUT for later analysis.
	ozUser = WFOzSyncTestUtils.makeFakeOzUser( T, makeOzUser( kAllison ) )
	
	docStore, catalog = OzCatalogTestUtils.buildMocksForTestingCatalogService( T, ozUser, kCatalogId )

end

--------------------------------------------------------------------------------

function TestSuite.tearDown()
	catalog:close()
	docStore, ozUser, catalog, OzCatalogService, WFIndexStoreSchema = nil
	T = {}
end

--==============================================================================

function TestSuite.test_adHocCollectionInfoB( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoB( { kPhoto1, kPhoto2 } ) )
		:continueWhenValueMatches( {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 1,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					favoritedCount = 0,
					id = "11111111111111111111111111111111",
				},
				{
					aspectRatio = 1,
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
			},
		}
	)
		

end

--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoB_emptyTable( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoB( { } ) )
		:continueWhenValueMatches( {
			assetCount = 0,
			displayDate = true,
			favoritedAssets = 0,
			name = "adHoc",
			photos = {
			},
		}
	)
		

end

--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoB_favoritesOnly_defaultSort( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T03:15:00.000Z',
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T04:15:00.000Z',
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-05-15T04:15:00.000Z',
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoB( { kPhoto1, kPhoto2, kPhoto3 }, 'favorites' ) )
		:continueWhenValueMatches( {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 2,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					captureDate = "2013-05-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "33333333333333333333333333333333",
				},
				{
					aspectRatio = 1,
					captureDate = "2013-04-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
			},
		}
	)

end


--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoB_favoritesOnly_ascendingSort( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T03:15:00.000Z',
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T04:15:00.000Z',
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-05-15T04:15:00.000Z',
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoB( { kPhoto1, kPhoto2, kPhoto3 }, 'favorites', true ) )
		:continueWhenValueMatches( {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 2,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					captureDate = "2013-04-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
				{
					aspectRatio = 1,
					captureDate = "2013-05-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "33333333333333333333333333333333",
				},
			},
		}	
	)	

end

--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoP( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoP( { kPhoto1, kPhoto2 } ) )
		:continueWhenValueMatches( 'data', {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 1,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					favoritedCount = 0,
					id = "11111111111111111111111111111111",
				},
				{
					aspectRatio = 1,
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
			},
		}
	)
		

end

--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoP_emptyTable( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kAllison = newDate(),
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
				kBarney = newDate(),
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoP( { } ) )
		:continueWhenValueMatches( 'data', {
			assetCount = 0,
			displayDate = true,
			favoritedAssets = 0,
			name = "adHoc",
			photos = {
			},
		}
	)
		

end

--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoP_favoritesOnly_defaultSort( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T03:15:00.000Z',
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T04:15:00.000Z',
			favorites = {
				kAllison = true,
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-05-15T04:15:00.000Z',
			favorites = {
				kBarney = true,
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoP( { kPhoto1, kPhoto2, kPhoto3 }, 'favorites' ) )
		:continueWhenValueMatches( 'data', {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 2,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					captureDate = "2013-05-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "33333333333333333333333333333333",
				},
				{
					aspectRatio = 1,
					captureDate = "2013-04-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
			},
		}
	)

end


--------------------------------------------------------------------------------

function TestSuite.test_adHocCollectionInfoP_favoritesOnly_ascendingSort( test )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T03:15:00.000Z',
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto2,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-04-15T04:15:00.000Z',
			favorites = {
				kAllison = true,
			},
		}
	)
	
	docStore:setServerValueP(
		'asset',
		kPhoto3,
		newAssetDoc {
			importSource = {
				importedBy = kAllison,
			},
			captureDate = '2013-05-15T04:15:00.000Z',
			favorites = {
				kBarney = true,
			},
		}
	)
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( catalog:adHocCollectionInfoP( { kPhoto1, kPhoto2, kPhoto3 }, 'favorites', true ) )
		:continueWhenValueMatches( 'data', {
			assetCount = 2,
			displayDate = true,
			favoritedAssets = 2,
			name = "adHoc",
			photos = {
				{
					aspectRatio = 1,
					captureDate = "2013-04-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "22222222222222222222222222222222",
				},
				{
					aspectRatio = 1,
					captureDate = "2013-05-15T04:15:00.000Z",
					favoritedCount = 1,
					id = "33333333333333333333333333333333",
				},
			},
		}	
	)	

end

--==============================================================================

return TestSuite