--[[----------------------------------------------------------------------------

OzCatComments.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSLogging = require "KSLogging"

local WFArguments = require "WFArguments"
local KSIdGenerator = require "KSIdGenerator"
local KSDate = require "KSDate"
local OzAccountService = require "OzAccountService"
local KSPromises = require "KSPromises"
local KSReactive = require "KSReactive"
local KSStringUtils = require "KSStringUtils"


--============================================================================--

-- @mark OzCatalogService class

local CommentMethods = {
}

local kIsCatalogComment = {}

local OzCatalogServiceArg = 'OzCatalogServiceImp'

--------------------------------------------------------------------------------

local commentsLog = KSLogging.log 'OzCatalogService :: comments '
--  :setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local function getSpaceForAlbumP( self, albumId )
	
	return self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function( _, idRows )
		
			-- If there is a space for this album, return the ID for that space.
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
			if spaceId then
				return 'data', spaceId
			end
			
			return 'error', "Album is not shared"
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a comment to an asset in an album in this catalog.
	
	\param albumId The unique ID of the shared album.
	\param assetId The unique ID of the asset.
	\param comment (string) The new comment.
	
	\return <code>P< string ></code> A Promise object that resolves to
		the unique ID of the newly-created comment.
]]--

function CommentMethods:addCommentP( albumId, assetId, comment )

	WFArguments.check( OzCatalogServiceArg, 'string | nil', 'string', 'string' )
	
	if not ( #comment > 0 ) then
		error( "OzCatalogService:addCommentP: The comment body must be a non-empty string", 2 )
	end
	
	commentsLog:debug( "Adding new comment for album %s asset %s with body : %s", albumId, assetId, comment )
	
	local resultP
	if albumId then
		resultP = getSpaceForAlbumP( self, albumId )
	else
		resultP = KSPromises.resolvedPromiseP( "data", kIsCatalogComment )
	end
	
	local commentId
	
	return resultP:mapDataP( function( _, spaceId )
		assert( spaceId, "No sharing space" )
		
		if spaceId == kIsCatalogComment then
			spaceId = nil
		end
		
		commentId = KSIdGenerator.uuidInLowercaseHex()
		local created = KSDate.timeInIso8601Format( nil, false )
		
		-- fill in the other values that we expect the server will provide
		local currentUser = assert( OzAccountService.currentUserB():valueNow() )

		local author = {
			first_name = currentUser:firstNameB():valueNow(),
			last_name = currentUser:lastNameB():valueNow(),
			full_name = currentUser:fullNameB():valueNow(),
			_id = currentUser:id(),
		}

		local resultP = self._docStore:applyOverlayToDocumentP {
			docType = 'comment',
			docId = commentId,
			overlay = {
				body = comment,
				_localOnly = {
					created = created,
					spaceId = spaceId,
					assetId = assetId,
					author = author,
				},
			},
		}
		return resultP
		
	end )
	:mapDataP( function( status,_ )
		return status, commentId
	end )
	:runUntilDoneP()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a comment.
	
	\param commentId The unique ID of the comment to be deleted.
	
	\return <code>P< boolean ></code> A Promise object that resolves to
		\c true if the request succeeded.
]]--

function CommentMethods:deleteCommentP( commentId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	commentsLog:debug( "Deleting comment %s", commentId )
	
	return self._docStore:applyOverlayToDocumentP {
		docType = 'comment',
		docIds = commentId,
		overlay = {
			_deleted = true,
		},
	}
	
end

--------------------------------------------------------------------------------

local kCommentFields = "body, created, author_first_name, author_last_name, author_full_name, author_id"
local nextIndex
do
	local index = 0
	nextIndex = function() index = index + 1 return index end
end

local kDocIdIndex				= nextIndex()
local kSourceIndex				= nextIndex()
local kAlbumIdIndex				= nextIndex()
local kAlbumNameIndex			= nextIndex()
local kUpdateTimeStampIndex		= nextIndex()
local kBodyIndex				= nextIndex()
local kCreatedIndex				= nextIndex()
local kAuthorFirstNameIndex		= nextIndex()
local kAuthorLastNameIndex		= nextIndex()
local kAuthorFullNameIndex		= nextIndex()
local kAuthorIdIndex			= nextIndex()

--------------------------------------------------------------------------------

local function formatComment( idRow, optIndexOfFirstCommentField )
	
	local comment = {}
	
	local offset = 0
	
	if optIndexOfFirstCommentField then
		offset = optIndexOfFirstCommentField - kBodyIndex
	end
	
	-- doc id is always first, regardless of offset
	comment.id = idRow[ kDocIdIndex ]
	
	comment.body = idRow[ kBodyIndex + offset ]
	comment.created = idRow[ kCreatedIndex + offset ]
	
	comment.author = {
		firstName = idRow[ kAuthorFirstNameIndex + offset ],
		fullName = idRow[ kAuthorFullNameIndex + offset ],
		lastName = idRow[ kAuthorLastNameIndex + offset ],
		id = idRow[ kAuthorIdIndex + offset ],
	}
	
	return comment

end

--------------------------------------------------------------------------------

local function getCommentCountForQuery( self, query, ... )
	
	return self._indexStore:queryB( query, ... )
		:closingMapB( function( idRows )
			
			local result = idRows and idRows[ 1 ]
			if result then
				return unpack( result, 1, 2 )
			end
			
		end )
	
end

--------------------------------------------------------------------------------

local kCommentOnSpaceByAlbumClause = [[
		FROM
			comment
		WHERE
			comment.assetId == ? AND spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ?
		)
		ORDER BY
			created]]

local kAllCommentClause = [[
			FROM
				comment
					LEFT JOIN
						space_album
					ON
						space_album.spaceId == comment.spaceId
					LEFT JOIN
						albums
					ON
						space_album.albumId == albums.id
					LEFT JOIN
						album_asset
					ON
						space_album.albumId == album_asset.albumId
					AND
						comment.assetId == album_asset.assetId
			WHERE
				comment.assetId == ?
			AND
				( %s )
			ORDER BY
				comment.spaceId, created]]


local function whereClauseForPlaces( assetId, places )

	local whereClause = { }
	local addlParams = { assetId }
	
	for _, place in ipairs( places or {} ) do
		
		if place == 'all' then
			whereClause[ #whereClause + 1 ] = "1"
		elseif place == 'archived' then
			whereClause[ #whereClause + 1 ] =
				[[comment.spaceId IS NULL
					OR album_asset.albumId IS NULL
					OR album_asset.assetId IS NULL]]
		elseif place == 'catalog' then
			whereClause[ #whereClause + 1 ] = 'comment.spaceId = 0'
		else
			whereClause[ #whereClause + 1 ] = 'space_album.albumId = ?'
			addlParams[ #addlParams + 1 ] = place
		end
		
	end
	
	if whereClause[ 1 ] then
		whereClause = string.format( kAllCommentClause, table.concat( whereClause, " OR " ) )
	else
		whereClause = "1"
	end
	
	return whereClause, addlParams
end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves the number of comments associated with an asset in this catalog.
	If an album ID is specified, then only the comments associated with this
	shared album are counted.
	
	\param assetId (string) The unique ID of the asset.
	\param places (optional array) A list of places to look for comments:
		- 'all' -- in all spaces, 'archived' and 'catalog'
		- 'archived' -- comments made on shared albums where the album is no longer
			shared (or has been re-shared) or the album has been deleted
		- 'catalog' -- comments made on the catalog
		- '< albumId >' -- comments made on a particular shared album.
	
	\return <code>B< number, number ></code> A Behavior object whose value is two numbers:
		the first is the total number of comments on the asset; the second is the number
		of comments for this user.

]]--

function CommentMethods:getNumCommentsOnAssetB( assetId, places )

	WFArguments.check( OzCatalogServiceArg, 'string', '?table' )

	local whereClause, addlParams = whereClauseForPlaces( assetId, places )

	return getCommentCountForQuery( self, [[
		SELECT
			COUNT( comment.docId ), COUNT( CASE WHEN author_id == ? THEN 1 END )
		]] .. whereClause, self._ozUserId, unpack( addlParams ) )

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves comments associated with an asset in this catalog.
	
	\param assetId (string) The unique ID of the asset.
	\param places (optional array) A list of places to look for comments:
		- 'all' -- in all spaces, 'archived' and 'catalog'
		- 'archived' -- comments made on shared albums where the album is no longer
			shared (or has been re-shared) or the album has been deleted
		- 'catalog' -- comments made on the catalog
		- '< albumId >' -- comments made on a particular shared album.
	
	\return <code>B< table ></code> A Behavior object whose value is an array of
		comment entries, sorted chronologically.  Each entry is a table
		contains keys:
		<ul><li>id (string)</li>
		<li>body (string)</li>
		<li>author (table)</li>
		<li>created (string)</li>
		</ul>
		The author subtable contains keys:
		<ul><li>firstName (string)</li>
		<li>fullName (string)</li>
		<li>lastName (string)</li><li>id (string)</li></ul>

]]--

function CommentMethods:getCommentsOnAssetB( assetId, places )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )
	
	local whereClause, addlParams = whereClauseForPlaces( assetId, places )
	
	local idRowsB = self._indexStore:queryB( [[
		SELECT
			comment.docId,
			CASE
				WHEN
					space_album.albumId IS NOT NULL
						AND
					album_asset.albumId IS NOT NULL
				THEN
					"space"
				WHEN
					comment.spaceId == 0
				THEN
					"catalog"
				ELSE
					"archived"
			END AS source,
			space_album.albumId, albums.name, albums.updateTimeStamp, ]] .. kCommentFields .. [[
		]] .. whereClause, unpack( addlParams ) )
	
	local sharedAlbumsB = self._indexStore:queryB( [[
		SELECT
			album_asset.albumId, albums.name, albums.updateTimeStamp
		FROM
			album_asset, space_album, albums
		WHERE
			album_asset.assetId == ?
		AND
			space_album.albumId == album_asset.albumId
		AND
			albums.id == album_asset.albumId
		]], assetId )
	local resultB = KSReactive.dataCallB( function( idRows, sharedAlbums )

			local allSpaces = {}
			
			for _, row in ipairs( sharedAlbums ) do
				allSpaces[ row[ 1 ] ] = {
					albumId = row[ 1 ],
					albumName = row[ 2 ],
					updateTimeStamp = row[ 3 ],
				}
			end
		
			local result = {}
			local nextSpace
			
			for i, idRow in ipairs( idRows ) do
				local needsNewSpace = false
				
				local source = idRow[ kSourceIndex ]
				
				local albumId = idRow[ kAlbumIdIndex ]
				if source == 'space' and albumId then
					allSpaces[ albumId ] = nil
				end

				if nextSpace and nextSpace.source ~= source then
					needsNewSpace = true
				elseif nextSpace and nextSpace.source == 'space' and nextSpace.albumId ~= albumId then
					needsNewSpace = true
				end
				
				if nextSpace == nil or needsNewSpace then
					local myNextSpace = {
						comments = {}
					}
					nextSpace = myNextSpace
					result[ #result + 1 ] = myNextSpace
					
					myNextSpace.source = source
					if source == 'space' then
						myNextSpace.albumId = albumId
					end
					myNextSpace.albumName = idRow[ kAlbumNameIndex ]
					myNextSpace.updateTimeStamp = idRow[ kUpdateTimeStampIndex ]
				end

				nextSpace.comments[ #nextSpace.comments + 1 ] = formatComment( idRow )
			end
			
			for albumId, info in pairs( allSpaces ) do
				result[ #result + 1 ] = {
					comments = {},
					source = 'space',
					albumId = info.albumId,
					albumName = info.albumName,
					updateTimeStamp = info.updateTimeStamp,
				}
			end
			
			table.sort( result, function( a, b )
				if not ( a and a.albumName ) then
					return false
				end
				if not ( b and b.albumName ) then
					return true
				end
				return -1 == KSStringUtils.localizedStringCompare( a.albumName, b.albumName )
			end )
			
			return result
			
		end,
		idRowsB,
		sharedAlbumsB
	)
	
	resultB:addCloseTarget( idRowsB )
	resultB:addCloseTarget( sharedAlbumsB )
	
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves comments associated with an asset in an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param assetId (string) The unique ID of the asset.
	
	\return <code>B< table ></code> A Behavior object whose value is an array of
		comment entries, sorted chronologically.  Each entry is a table
		contains keys:
		<ul><li>id (string)</li>
		<li>body (string)</li>
		<li>author (table)</li>
		<li>created (string)</li>
		</ul>
		The author subtable contains keys:
		<ul><li>firstName (string)</li>
		<li>fullName (string)</li>
		<li>lastName (string)</li><li>id (string)</li></ul>
]]--

function CommentMethods:getCommentsOnAlbumAssetB( albumId, assetId )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string' )

	return self._indexStore:queryB( [[
		SELECT
			comment.docId, "space", 0, "catalog", 0, ]] .. kCommentFields .. [[
		]] .. kCommentOnSpaceByAlbumClause, assetId, albumId )
	:closingMapB( function( idRows )
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatComment( idRow )
		end
		return result
	end )

end

--==============================================================================

return {
	CommentMethods = CommentMethods,
	
	kCommentFields = kCommentFields,
	formatComment = formatComment,
}
