--[[----------------------------------------------------------------------------

OzCatCommentsTest.lua

@author arahn 2012-10-10

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]



local WFOzSyncTestUtils = require 'WFOzSyncTestUtils'
local WFAsyncTestCase = require 'WFAsyncTestCase'
local KSJsonEncoder = require 'KSJsonEncoder'
local OzCatalogTestUtils = require 'OzCatalogTestUtils'

local TEST_WAIT_TIME = 2

--==============================================================================

local function makeFunnyUniqueAlbumName()
	
	local adjectives = { "Crazy", "Disasterous", "Uncensored", "Frightening", "Lazy",
		"Freezing", "Hot", "Catastrophic", "Mysterious", }
	local events = { "Reunion", "Wedding", "Adoption", "Space Launch", "Fishing Trip",
		"Photo Safari", "Walkabout", "Ramble", "Hike", "Camping Adventure" }
	local places = { "in Transylvania", "in Abu Dahbi", "in Pukwana", "in Peru", "in Outer Mongolia",
		"in Siberia", "on the North Shore" }
	
	return string.format( "My %s %s %s",
		adjectives[ math.random( #adjectives ) ],
		events[ math.random( #events ) ],
		places[ math.random( #places ) ] )
		
end

math.randomseed( os.time() )

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local OzCatalogService, WFIndexStoreSchema

local T = {}
local docStore, ozUser, catalog

--------------------------------------------------------------------------------

local kAllison = string.rep( 'a', 32 )
local kBarney = string.rep( 'b', 32 )
local kChloe = string.rep( 'c', 32 )
local kDick = string.rep( 'd', 32 )

local kPhoto1 = string.rep( '1', 32 )
local kCatalogId = string.rep( '0', 32 )

local kCommentId = string.rep( 'c0', 16 )
local kCommentBody = string.rep( "hello ", 12 )
local kAlbumId = string.rep( 'a1bb', 8 )

local kSpaceAlbumId = string.rep( '54acea1b', 4 )
local kSpaceId = string.rep( '5pac', 8 )

--------------------------------------------------------------------------------

local kUsers = {
	[ kAllison ] = {
		first_name = 'Alli',
		last_name = 'Ansi',
		full_name = 'Allison Ansi',
	},
	[ kBarney ] = {
		first_name = 'Barney',
		last_name = 'Barnes',
		full_name = 'Barney B Barnes',
	},
	[ kChloe ] = {
		first_name = 'Chloe',
		last_name = 'Chism',
		full_name = 'Chloe Chism-Chisholm',
	},
	[ kDick ] = {
		first_name = 'Dick',
		last_name = 'Phillip',
		full_name = 'Richard K Phillip',
	},
}

--------------------------------------------------------------------------------

local function makeOzUser( id )
	local src = kUsers[ id ]
	local oz = {
		_id = id,
		first_name = src.first_name,
		last_name = src.last_name,
		full_name = src.full_name,
	}
	return oz
end

--------------------------------------------------------------------------------

local function makeWFUser( id )
	local oz = kUsers[ id ]
	local wf = {
		id = id,
		firstName = oz.first_name,
		lastName = oz.last_name,
		fullName = oz.full_name,
	}
	return wf
end

--==============================================================================

function TestSuite.setUp()
	
		-- Utterly fake oz user.  Vends canned responses to GETs and stores
		-- of what was PUT for later analysis.
	ozUser = WFOzSyncTestUtils.makeFakeOzUser( T, makeOzUser( kAllison ) )
	
	docStore, catalog = OzCatalogTestUtils.buildMocksForTestingCatalogService( T, ozUser, kCatalogId )
	
	T.userGetPutResponse[ 'https://api.oznext.com/v1.0/catalogs/' .. kCatalogId .. '/users' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			makeOzUser( kAllison ),
			makeOzUser( kBarney ),
			makeOzUser( kDick ),
			makeOzUser( kChloe ),
		},
	}

end

--------------------------------------------------------------------------------

function TestSuite.tearDown()
	catalog:close()
	docStore, ozUser, catalog, OzCatalogService, WFIndexStoreSchema = nil
	T = {}
end

--------------------------------------------------------------------------------

local function addTestCommentDocs( numAlbumComments, numCatalogComments )
	
	docStore:setServerValueP(
		'asset',
		kPhoto1,
		{
			_serverProvidedMetadata = {
				created = '2013-01-03T12:23:45Z',
			},
			importSource = {
				importedBy = kAllison,
			},
			favorites = {
			},
		}
	)
	
	docStore:setServerValueP(
		'space_album',
		kSpaceAlbumId,
		{
			album = {
				_id = kAlbumId,
			},
			space = {
				_id = kSpaceId
			},
		}
	)
	
	local n = 0
	for i = 1, ( numAlbumComments or 1 ) do
		local authors = { kChloe, kDick, kAllison, kBarney }
		local author = authors[ ( ( i - 1 ) % 4 ) + 1 ]
		docStore:setServerValueP(
			'comment',
			kCommentId:sub( 1, 30 ) .. string.format( "%02d", n ),
			{
				created = string.format('2%03d-02-02T12:23:45Z',n),
				asset = {
					_id = kPhoto1,
				},
				body = kCommentBody,
				author = makeOzUser( author ),
				space = {
					_id = kSpaceId,
				},
			}
		)
		n = n + 1
	end

	
	for i = 1, ( numCatalogComments or 1 ) do
		local authors = { kChloe, kDick, kAllison, kBarney }
		local author = authors[ ( ( i - 1 ) % 4 ) + 1 ]
		docStore:setServerValueP(
			'comment',
			kCommentId:sub( 1, 30 ) .. string.format( "%02d", n ),
			{
				created = string.format('2%03d-02-02T12:23:45Z',n),
				asset = {
					_id = kPhoto1,
				},
				body = kCommentBody,
				author = makeOzUser( author ),
				space = {
					_id = 0,
				},
			}
		)
		n = n + 1
	end
end

--==============================================================================

function TestSuite.test_getCommentsOnAlbumAssetB( test )
	
	addTestCommentDocs( 1, 0 )
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalog:getCommentsOnAlbumAssetB( kAlbumId, kPhoto1 )
	
		end )
		:continueWhenValueMatches {
			{
				id = kCommentId:sub(1, 30) .. '00',
				body = kCommentBody,
				created = '2000-02-02T12:23:45Z',
				author = makeWFUser( kChloe )
			}
		}

end

--------------------------------------------------------------------------------

function TestSuite.test_addDelete( test )
	
	addTestCommentDocs( 0, 0 )

	local newComments = {}

	local function expectedCommentList()
		local t = {}
		for idx, comment in ipairs( newComments ) do
			t[ idx * 3 - 2 ] = comment.id
			t[ idx * 3 - 1 ] = comment.body
			t[ idx * 3 - 0 ] = makeWFUser( kAllison )
		end
		if t[1] then
			return unpack( t )
		else
			-- if KSReactive had ALLOW_EMPTY_TUPLES_FROM_MAP we wouldn't need
			-- to ensure we return nil here
			return nil
		end
	end
	
	local function addAnotherCommentP()
		local comment = {
			 body = string.format( "That sounds as much fun as %s", makeFunnyUniqueAlbumName() )
		}
		return catalog:addCommentP( kAlbumId, kPhoto1, comment.body )
			:mapDataP( function( _, newCommentId )
				comment.id = newCommentId
				table.insert( newComments, comment )
			end )
	end
	
	local function deleteFirstCommentP()
		local deleted = table.remove( newComments, 1 )
		return catalog:deleteCommentP( deleted.id )
	end
	
	local function getCommentsB( )
		return catalog:getCommentsOnAlbumAssetB( kAlbumId, kPhoto1 )
			:mapB( function( commentList )
				local t = {}
				for idx, comment in ipairs( commentList or {} ) do
					t[ idx * 3 - 2 ] = comment.id
					t[ idx * 3 - 1 ] = comment.body
					t[ idx * 3 - 0 ] = comment.author
				end
				return unpack( t )
			end )
	end
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( addAnotherCommentP )
		:waitFor( getCommentsB )
		:continueWhenValueMatchesFunction( expectedCommentList )
		:waitFor( addAnotherCommentP )
		:waitFor( getCommentsB )
		:continueWhenValueMatchesFunction( expectedCommentList )
		:waitFor( deleteFirstCommentP )
		:waitFor( getCommentsB )
		:continueWhenValueMatchesFunction( expectedCommentList )
		:waitFor( deleteFirstCommentP )
		:waitFor( getCommentsB )
		:continueWhenValueMatchesFunction( expectedCommentList )

end

--------------------------------------------------------------------------------
-- @mark Get Comments
--------------------------------------------------------------------------------

function TestSuite.test_getNumCommentsOnAssetB( test )
	
	addTestCommentDocs( 8, 12 )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalog:getNumCommentsOnAssetB( kPhoto1, { kAlbumId } )
	
		end )
		:continueWhenValueMatches( 8, 2 )
		
		:waitFor( function()
		
			return catalog:getNumCommentsOnAssetB( kPhoto1, { 'all' } )
	
		end )
		:continueWhenValueMatches( 20, 5 )

end

--------------------------------------------------------------------------------

function TestSuite.test_getCommentsOnAssetB( test )

	addTestCommentDocs( 1, 2 )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()
		
			return catalog:getCommentsOnAssetB( kPhoto1, { 'all' } )
	
		end )
		:continueWhenValueMatches {
			{
				comments = {
					{
						author = makeWFUser( kChloe ),
						body = kCommentBody,
						created = "2001-02-02T12:23:45Z",
						id = kCommentId:sub(1,30) .. "01",
					},
					{
						author = makeWFUser( kDick ),
						body = kCommentBody,
						created = "2002-02-02T12:23:45Z",
						id = kCommentId:sub(1,30) .. "02",
					},
				},
				source = "catalog",
			},
			{
				comments = {
					{
						author = makeWFUser( kChloe ),
						body = kCommentBody,
						created = "2000-02-02T12:23:45Z",
						id = kCommentId:sub(1,30) .. "00",
					},
				},
				source = "archived",
			},
		}

end

--==============================================================================

return TestSuite
