--[[----------------------------------------------------------------------------

OzCatConflicts.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- wichitafoundation.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

--============================================================================--

local OzCatalogServiceArg = 'OzCatalogServiceImp'
local DocumentIdArg = 'string'
local ResolveWithArg = 'OzCatalogService.ResolveWith'
	-- Type name for checker
	
WFArguments.registerChecker( ResolveWithArg, {
	[ 'local' ] = true,
	server = true,
	fork = true,
} )

local conflictLog = KSLogging.log 'OzCatalogService :: conflict'
	--	:setLevel( KSLogging.DEBUG )

-- @mark OzCatalogService class

local ConflictMethods = {
}

--------------------------------------------------------------------------------

-- @mark conflicts

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a photo asset in this catalog has a conflict that
		prevents upload.
	
	\param photoId (string) The unique ID of the photo asset.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if the photo has a conflict.
]]--

function ConflictMethods:hasConflictB( --[[ string ]] photoId )

	return self._docStore:fieldValueB( 'asset', photoId, "_conflict" )
		:mapB( function( fieldValue, isExisting )
		
			return not not fieldValue, isExisting
		
		end )

end

--------------------------------------------------------------------------------

local function photoConflictCountB( self, activeAlbum, limit )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )

	local resultB
	
	if activeAlbum then
		resultB = self._indexStore:queryB(
			[[SELECT COUNT( ozUploadStatus ) FROM ozUploadStatus
				INNER JOIN album_asset
					ON ozUploadStatus.docId == album_asset.assetId
			WHERE ozUploadStatus == '_conflict'
							AND album_asset.albumId == ?
			]] .. (limit and "LIMIT 1" or "" ), activeAlbum )
	else
		resultB = self._indexStore:queryB(
			[[SELECT COUNT( ozUploadStatus ) FROM ozUploadStatus
				WHERE ozUploadStatus == '_conflict'
			]] .. (limit and "LIMIT 1" or "" ) )
				
	end
	
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether any photo asset in an album in this catalog
	has a conflict that prevents upload.
	
	\param activeAlbum (string) The unique ID of the album.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if any photo in the album has a conflict.
]]--

function ConflictMethods:anyPhotoHasConflictB( activeAlbum )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )

	return photoConflictCountB( self, activeAlbum, true )
		:closingMapB( function( results )
		
			return not not ( results and results[ 1 ] )
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports how many photo assets in an album in this catalog
	have conflicts that prevents upload. If no album is specified it counts
	them in the entire catalog
	
	\param activeAlbum (string) The unique ID of the album. (Optional)

	\return <code>B< boolean ></code> A Behavior object whose value is the count, or
		nil if the value is not yet known.
]]--

function ConflictMethods:countOfPhotosWithConflictsB( activeAlbum )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )

	return photoConflictCountB( self, activeAlbum, true )
		:closingMapB( function( results )
		
			return KSTableUtils.derefOpt( results, 1, 1 )
		
		end )

end

--------------------------------------------------------------------------------
--[[
 	
	Resolve a conflict on a photo that prevents if from being uploaded.
	
	\param photoId (string) The unique ID of the photo asset.
	\param resolveWith (string) The conflict resolution method, one of:
		<ul><li>'local' : Accept the local version of the photo and discard the
				  version currently on Oz.</li>
				
		<li>'server' : Accept the server version of the photo and discard the
				version created locally.</li>
		<li>'fork' : Leave the server version as is and create a new version
			     of the photo on Oz from the local version.</li></ul>

	\return <code>P< true ></code> A Promise that resolves to \c true
		if the request succeeded.
]]--

function ConflictMethods:resolveConflictP( photoId, resolveWith )

	WFArguments.check( OzCatalogServiceArg, DocumentIdArg, ResolveWithArg )

	conflictLog:info( "resolveConflictP for photo %s resolveWith = %s",
					  photoId, resolveWith )

	local docStore = self._docStore

	local docValueOnceP = docStore:docValueOnceP( 'asset', photoId )
	
	local resolveConflictP = docValueOnceP:mapDataP( function( _, docValue )
		
		conflictLog:debug( "docValue before resolving conflict was %s", docValue )
		
		if not docValue._conflict then
			return 'error', "There was no conflict to resolve."
		end

		if resolveWith == 'fork' then
		
			-- If we're resolving by forking, we make a copy of the local
			-- version first, then "resolve" the conflict on the original photo
			-- by deleting local settings (resolve in favor of server).
			
			local resultP = self:copyPhotoP( photoId, false, true )
				:mapP {
					data = function()
						return self:resolveConflictP( photoId, 'server' )
					end,
				}
			
			return resultP
		
		end

		return docStore:resolveConflictP( 'asset', photoId,
						docValue._changeCounter, resolveWith )
	
	end )
	
	if conflictLog.level == KSLogging.DEBUG then
	
		resolveConflictP:whenResolvedDo( function()
		
			docStore:docValueOnceP( 'asset', photoId )
				:whenResolvedDo( function( _, docValue )
				
					conflictLog:debug( "docValue AFTER resolving conflict was %s", docValue )
				
				end ):runUntilDone()
		
		end ):runUntilDone()
	
	end
	
	-- FIXME [scouten 2011-08-10]: This conditional promise chaining is a bit ugly ...
	
	local finalResultP = ( resolveWith == 'fork' or resolveWith == 'local' )
		and resolveConflictP
		or resolveConflictP:mapDataP( function()
			
				-- Force client to re-download renditions from Oz.
			
				local resultP = docStore:applyOverlayToDocumentP {
					docType = 'asset',
					docId = photoId,
					overlay = {
						_localOnly = {
							devSettingsLastSetAtRev = 'reverted',
							preview = { _delete = true },
							thumbnail = { _delete = true },
							thumbnail2x = { _delete = true },
						},
					},
				}
				
				return resultP -- can't inline since applyOverlayToDocumentP has multiple return values
		
			end )

	finalResultP:runUntilDone()

	return finalResultP

end

--------------------------------------------------------------------------------

local function getNameForUserP( self, userId )
	
	return self._ozUser:getP( string.format( '%s/v1.0/accounts/%s', OzCatalogUrls.baseServerUrl, userId ) )
		:mapDataP( function( status, data )

			local userInfo = KSJsonDecoder.decodeWebSafeJson( data )
			if userInfo and userInfo.full_name then
				return 'data', userInfo.full_name
			else
				return 'error', 'getNameForUserP: could not fetch the name for this user'
			end

	end )

end

--------------------------------------------------------------------------------
--[[
	Reports the details of changes made to an asset that result in conflict.

	\param photoId (string) The unique ID of the photo asset.

	\return <code>P< table ></code> A Promise that resolves to a table with these keys:
		<ul><li>'changedBy_local' : The name of the user who made the change locally,
			or if both changes were made by the same user, the name of the device.</li>
		  <li>'changedAtTime_local': The time of the change made locally.</li>
		<li>'changedBy_remote' : The name of the user who made the change on the server,
			or if both changes were made by the same user, the name of the device.</li>
		  <li>'changedAtTime_remote': The time of the change made on the server</li>
		</ul>
]]--

function ConflictMethods:getChangeDetailsP( photoId )

	if type( photoId ) ~= 'string' or #photoId == 0 then
		error( "OzCatalogService:getChangeDetailsP photoId must be a non-empty string", 2 )
	end

	return self._docStore:docValueOnceP( 'asset', photoId )
		:mapDataP( function( status, docValue )
	
			local serverMetadata = docValue._serverProvidedMetadata
			if not serverMetadata then
				return 'error', "ImportMethods:getChangeDetailsP: Could not fetch server metadata for this asset."
			end

			local links = serverMetadata.links
			if not links then
				return 'error', "ImportMethods:getChangeDetailsP: Could not fetch links for this asset."
			end

			local conflict = docValue._conflict
			if not conflict then
				return 'error', "ImportMethods:getChangeDetailsP: This photo does not have a conflict."
			end

			-- Who made the change that's on server?
			
			local remoteChangeUserId = links[ '/rels/latest_revision' ].created_by
			remoteChangeUserId = remoteChangeUserId and remoteChangeUserId:sub( -32 )

			-- Start populating the change structure.

			local changeDetails = {
				
				-- local details
				
				changedAtTime_local = docValue.changedAtTime,
				--changedBy_local = docValue.changedOnDevice,
					-- or KSDeviceUtils.deviceName?
					-- or (localized) "This Device" or "This Computer"?
				
				-- remote details
				
				changedAtTime_remote = conflict.changedAtTime,
				--changedBy_remote = conflict.changedOnDevice,
					
			}

			-- Were changes made by same user on both devices?
			
			if remoteChangeUserId == self._ozUserId then
			
				-- Yes: Show device names instead of user names as point of change.
			
				changeDetails.changedBy_local = docValue.changedOnDevice
						-- or KSDeviceUtils.deviceName?
						-- or (localized) "This Device" or "This Computer"?

				changeDetails.changedBy_remote = conflict.changedOnDevice
				
				return 'data', changeDetails
					
			end

			-- No: Changes were made by different users. Get user name for local and remote user.
			
			changeDetails.changedBy_local = self._ozUser:fullNameB():valueNow()
					
			-- try to get the name for the remote user from the doc store, if that fails query Oz

			return self:catalogUsersP()
			:mapP( function( status, list )

				if status == 'data' then
					for i = 1, #list do

						local id = list[ i ].id:gsub("urn:carousel:account:","")
						if id == remoteChangeUserId then
							changeDetails.changedBy_remote = list[ i ].full_name
							break
						end
		
					end
					if changeDetails.changedBy_remote then
						debug.print("Got local name")
						return 'data', changeDetails
					else
						return getNameForUserP( self, remoteChangeUserId )
							:mapDataP( function ( status, name )
								changeDetails.changedBy_remote = name
								return 'data', changeDetails
						end )
					end
				else
					return getNameForUserP( self, remoteChangeUserId )
						:mapDataP( function ( status, name )
							changeDetails.changedBy_remote = name
							return 'data', changeDetails
					end )
				end

			end )
			

		end )

end

--------------------------------------------------------------------------------

return {
	ConflictMethods = ConflictMethods,
}
