--[[----------------------------------------------------------------------------

OzCatCopying.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSIdGenerator = require "KSIdGenerator"
local OzCatalogUploadStatus = require "OzCatalogUploadStatus"
local KSLogging = require "KSLogging"
local KSDate = require "KSDate"
local KSPromises = require "KSPromises"
local WFArguments = require "WFArguments"

--==============================================================================

local OzCatalogServiceArg = 'OzCatalogServiceImp'

-- @mark OzCatalogService class

local CopyingMethods = {
}

local copyPhotoLog = KSLogging.log 'OzCatalogService :: copyPhoto'
	--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

-- @mark copying

--------------------------------------------------------------------------------
--[[!
	\brief Makes a duplicate of a photo in this catalog, referencing the original master file.
		
	\param photoId The unique ID of the asset to be duplicated.
	\param activeAlbum The unique ID of the current album, or false if no album is currently active.
		The new asset is added to this album, if any.
	\param duplicateAllAlbumMembership (boolean) When \c true, the new asset is added
		to all albums to which the original asset belongs.
	
	\return <code>P< string ></code> A Promise object that resolves to the
		ID of the newly-created asset.
]]--

function CopyingMethods:copyPhotoP( photoId, activeAlbum, duplicateAllAlbumMembership )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string|boolean', 'boolean|nil' )

	local docValueP = self._docStore:docValueOnceP( 'asset', photoId )

	local newDocId = KSIdGenerator.uuidInLowercaseHex()
	
	local createNewAssetP = docValueP:mapDataP( function( _, docValue )

		-- Copying is actually pretty stinkin' easy.

		copyPhotoLog:debug( "copyPhotoP\n\noriginal photo = %s",
							docValue )

		-- Check if the original is present locally or on Oz, if not return an error

		local localPath = docValue
			and docValue._localOnly
			and docValue._localOnly.originalFilePath
			
		local copyMasterFrom = docValue
			and docValue._localOnly
			and docValue._localOnly.copyMasterFrom

		local ozUrl = docValue
			and docValue._serverProvidedMetadata
			and docValue._serverProvidedMetadata.links
			and docValue._serverProvidedMetadata.links[ "/rels/master" ]
			and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
			
		local assetOriginalIsValid = OzCatalogUploadStatus.assetOriginalIsValid( docValue )
		
		local assetLatestOriginalIsValid = OzCatalogUploadStatus.assetLatestOriginalIsValid( docValue )

		-- As a fix for WATSON #3133097 - Duplication Error: Failed to duplicate the duplicated photo.
		-- If you duplicate the photo fast enough and we don't have the local copy and Oz hasn't updated the master Url yet,
		-- we end up returning an error. Check the copyMasterFrom field as well if the local copy and Oz links
		-- aren't available yet.

		if not ( localPath or ozUrl or copyMasterFrom ) then
			return 'error', 'originalNotFound'
		end

		docValue._changeCounter = nil
		docValue._conflict = nil
		docValue._docId = nil
		docValue._docType = nil
		docValue._ozMetadataUploadStatus = nil

		local links = docValue._serverProvidedMetadata
			and docValue._serverProvidedMetadata.links

		docValue.copied_from_href = links
			and links[ '/rels/latest_revision' ]
			and links[ '/rels/latest_revision' ].href

		docValue._serverProvidedMetadata = nil
		
		-- Eric S's superfunky fix for the ordering of duplicates of duplicates

		if docValue.importSource.copyTimestamp then
			docValue.importSource.copyTimestamp = docValue.
				importSource.copyTimestamp:gsub( ':X$',
								':' .. KSDate.timeInIso8601Format() .. ':X' )
		else
			docValue.importSource.copyTimestamp = string.format( "%s:%s:X",
								docValue.importSource.importTimestamp,
								KSDate.timeInIso8601Format() )
		end
		
		docValue.favorites = nil

		local localOnly = docValue._localOnly
		if not localOnly then
			localOnly = {}
			docValue._localOnly = localOnly
		end
		
		-- Since we are duplicating the asset, we need to make sure
		-- the new asset's localOnly.originalFilePath contains the local path
		-- if any for the latest original
		
		if assetOriginalIsValid then

			if localOnly.originalFilePath then

				-- Temporary check to correct the local file path and replace
				-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
				-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
				-- that updates the database with the correct file paths but this is not
				-- feasible in the current time frame.

				docValue._localOnly.originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localOnly.originalFilePath )

			end

		else

			if assetLatestOriginalIsValid and localOnly.latestOriginal and localOnly.latestOriginal.filePath then
				docValue._localOnly.originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localOnly.latestOriginal.filePath )
			end

		end
		
		if localOnly.original then
			docValue._localOnly.original.ozUrl = { _delete = true }
		end
		
		-- We also delete the latest original section for this new asset since
		-- it only knows of one original at present

		if localOnly.latestOriginal then
			docValue._localOnly.latestOriginal = { _delete = true }
		end
		
		if localOnly.preview then
			localOnly.preview.revId = 'pending'
		end
		
		if localOnly.thumbnail then
			localOnly.thumbnail.revId = 'pending'
		end

		if localOnly.thumbnail2x then
			localOnly.thumbnail2x.revId = 'pending'
		end
		
		localOnly.devSettingsLastSetAtRev = 'pending'
		localOnly.localUrl = nil

		localOnly.copyMasterFrom = links
							   and links[ '/rels/master' ]
							   and links[ '/rels/master' ].href or copyMasterFrom

		copyPhotoLog:debug( "copyPhotoP\n\ndupe photo = %s",
							docValue )

		local applyOverlayP = self._docStore:applyOverlayToDocumentP {
							docType = 'asset',
							docId = newDocId,
							overlay = docValue,
							addRevTimestamp = true,
						}

		-- Note that :applyOverlayToDocumentP has a multi-valued return
		-- so we can't simply return its results. Promise system wants
		-- exactly one result when you're promise chaining.

		return applyOverlayP
			:mapDataP( function ()
				
				return self._indexStore:queryP( [[ SELECT albumId from album_asset WHERE assetId == ?]], photoId )
					:mapDataP( function ( status, rows )

						local function addAssetToAlbum( row )

							local albumId = row and row[ 1 ]
							
							-- Don't try to add this asset to the favorites smart album, that doesn't go so well with Oz!

							if albumId == ( 'favorites' .. self._ozUserId ) then
								return 'data', true
							end

							-- Add this new asset to the album that was currently selected when the duplication happened.
							-- If the library was selected or duplicateAllAlbumMembership is true, add it to all albums.

							if not activeAlbum or ( albumId == activeAlbum ) or duplicateAllAlbumMembership then

								local albumAssetId = KSIdGenerator.uuidInLowercaseHex()

								local innerResultP = self._docStore:applyOverlayToDocumentP {
									docType = 'album_asset',
									docIds = albumAssetId,
									overlay = {
										album = {
											_id = albumId,
											},
										asset = {
											_id = newDocId,
										},
									},
								}
		
								return innerResultP

							else
								return 'data', true
							end

						end

						return KSPromises.allP( addAssetToAlbum, rows )

					end )
			end )

	end )
	
	local returnNewAssetIdP = createNewAssetP:mapDataP( function()

		return 'data', newDocId
		-- SPAM [scouten 2011-08-10]: This gets tricker when we
		-- have to also return the undo overlay object.

	end )
	
	returnNewAssetIdP:runUntilDone()

	return returnNewAssetIdP

end

--==============================================================================

return {
	CopyingMethods = CopyingMethods,
}
