--[[----------------------------------------------------------------------------

OzCatDateCorrection.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSPromises = require "KSPromises"
local KSDate = require "KSDate"

--==============================================================================

-- @mark OzCatalogService class

local DateCorrectionMethods = {
}

--------------------------------------------------------------------------------

-- @mark date correction

--------------------------------------------------------------------------------
--[[!
	\brief Sets the capture date for a set of assets in this catalog.
	
	The exact same date-time value is applied to all photos, regardless
	of any existing date-time value.
	
	\todo [scouten 2012-04-23]: IMHO this function should be renamed to
		\c correctCaptureDateAbsoluteP.
	
	\param photoIds (array of strings) The asset IDs.
	\param newCaptureDate (string) The new capture date in ISO 8601 format.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function DateCorrectionMethods:correctCaptureDateP( --[[array of strings]] photoIds,
													--[[string]] newCaptureDate )

    if not( type( newCaptureDate ) == 'string' and #newCaptureDate > 0 ) then
        error( "OzCatalogService:correctCaptureDateP: captureDate must be non-empty string", 2 )
    end

	local function correctDateP( assetId )
					
		return self._docStore:docValueOnceP( 'asset', assetId )
			:mapDataP( function ( status, photoInfo )
						
				local overlay = {}
						
				if not photoInfo.originalCaptureDate then
					overlay.originalCaptureDate = photoInfo.captureDate
				end
												
				overlay.captureDate = newCaptureDate
								
				local promiseP = self._docStore:applyOverlayToDocumentP {
       				docType = 'asset',
					docId = assetId,
				   	overlay = overlay,
				}
					
				return promiseP
	
			end )
		
	end

	local resultP = KSPromises.allP( correctDateP, photoIds )
		
	resultP:runUntilDone()
		
	return resultP

end

--------------------------------------------------------------------------------
--[[!
	\brief Corrects the capture date of a set of assets in this catalog.
	The date correction is applied to all the assets in the set by applying
	the delta between the dates given for the first asset.
	
	\todo [scouten 2012-04-23]: IMHO this function should be renamed to
		\c correctCaptureDateWithDeltaP.
	
	\param assetIds (array of strings) The asset IDs to modify
	\param oldCaptureDate (string) The old capture date for the first asset, in ISO 8601 format
	\param newCaptureDate (string) The new capture date for the first asset, in ISO 8601 format
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function DateCorrectionMethods:correctCaptureDateForTrackP( --[[array of strings]] assetIds,
							--[[string]] oldCaptureDate,
							--[[string]] newCaptureDate )

	if not( type( oldCaptureDate ) == 'string' and #oldCaptureDate > 0 ) then
        error( "OzCatalogService:correctCaptureDateForTrackP: old captureDate must be non-empty string", 2 )
    end

	if not( type( newCaptureDate ) == 'string' and #newCaptureDate > 0 ) then
        error( "OzCatalogService:correctCaptureDateForTrackP: new captureDate must be non-empty string", 2 )
    end

    if not( type( assetIds  ) == 'table' ) then
        error( "OzCatalogService:correctCaptureDateForTrackP: The photoIds must be a table", 2 )
    end

	local delta = KSDate.differenceInSeconds( oldCaptureDate, newCaptureDate, true )
	
	-- Update the captureDate for each asset by the same delta as selected
	-- by the user for the first asset in the track.
		
	local function correctDateP( assetId )
					
		return self._docStore:docValueOnceP( 'asset', assetId )
			:mapDataP( function ( status, photoInfo )
					
				if not photoInfo.captureDate then
					return 'data', true
				end
						
				local overlay = {}
						
				if not photoInfo.originalCaptureDate then
					overlay.originalCaptureDate = photoInfo.captureDate
				end
				
				-- Remove the TimeZone prefix before performing date math and append it once done.

				local captureDateUptoSeconds, timeZoneSuffix = KSDate.splitIso8601DateUptoSeconds( photoInfo.captureDate )

				local newCaptureDate = KSDate.addTimeInSeconds( captureDateUptoSeconds, delta )
							
				overlay.captureDate = newCaptureDate .. timeZoneSuffix
								
				local promiseP = self._docStore:applyOverlayToDocumentP {
       				docType = 'asset',
					docId = assetId,
				   	overlay = overlay,
				}
					
				return promiseP
	
			end )
		
		end

	-- Special case for managing correction of dates for Unknown date tracks.

	local resultP
	if oldCaptureDate == '0000-00-00T00:00:00' then
		resultP = self:correctCaptureDateP( assetIds, newCaptureDate )
	else
		resultP = KSPromises.allP( correctDateP, assetIds )
	end
		
	resultP:runUntilDone()
		
	return resultP

end

--==============================================================================

return {
	DateCorrectionMethods = DateCorrectionMethods,
}
