--[[----------------------------------------------------------------------------

OzCatDeletion.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.os
local KSFileUtils = require 'KSFileUtils'

local WFArguments = require "WFArguments"
local KSLogging = require "KSLogging"
local KSPromises = require "KSPromises"
local KSTableUtils = require "KSTableUtils"
local KSReactivePromiseBridge = require "KSReactivePromiseBridge"
local KSReactive = require "KSReactive"
local WFUnknown = require 'WFUnknown'

local OzCatalogUploadStatus = require "OzCatalogUploadStatus"

local OzCatHelperFns = require "OzCatHelperFns"

--==============================================================================

local OzCatalogServiceArg = 'OzCatalogServiceImp'

local OzCatHelperFns_findAndCorrectOriginalPath = OzCatHelperFns.findAndCorrectOriginalPath

local deletePhotoLog = KSLogging.log 'OzCatalogService :: deletePhoto'
	--	:setLevel( KSLogging.DEBUG )

-- @mark OzCatalogService class

local DeletionMethods = {
}

--==============================================================================
-- @mark utility fns
--==============================================================================

local function deleteOnePhotoP( docStore, indexStore, assetId, currentUserId, catalogOwnerId )

	-- Step 1: Fetch the existing document value.
		
	return docStore:docValueOnceP( 'asset', assetId )

	-- Step 2: Use that to delete the local copy of our files.

	:mapDataP( function( status, docValue )
		
		local photoOwnerId = KSTableUtils.derefOpt( docValue, "importSource", "importedBy" )
		
		if photoOwnerId then
			
			if currentUserId ~= photoOwnerId and catalogOwnerId ~= currentUserId then
				
				-- This user is not permitted to delete this asset
				return 'error', "permissionDenied"
				
			end
			
		end
		
		if not( docValue and docValue._changeCounter ) then
			return 'error', "No such photo"
		end
		
		deletePhotoLog:debug( "Deleting original for %s", assetId )
	
		local localOnly = docValue._localOnly
		local originalFilePath = OzCatHelperFns_findAndCorrectOriginalPath( docValue )
	
		local latestOriginalFilePath = localOnly
						and localOnly.latestOriginal
						and localOnly.latestOriginal.filePath
	
		latestOriginalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( latestOriginalFilePath )
	
		if originalFilePath then
			KSFileUtils.moveToTrash( originalFilePath )
		end
	
		if latestOriginalFilePath then
			KSFileUtils.moveToTrash( latestOriginalFilePath )
		end
	
		local proxyFilePath = localOnly
					and localOnly.proxy
					and localOnly.proxy.filePath
	
		proxyFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( proxyFilePath )
	
		if proxyFilePath then
			KSFileUtils.delete( proxyFilePath )
		end
	
		return 'data', true
	
	end )
	
	-- Step 3: Now delete the document and references to those files.

	:mapDataP( function( )

		-- Remove references to these files so we don't circle back and delete
		-- them again later when we get the deletion confirmation from Oz about
		-- this photo. The file names might get re-used if we re-import the same
		-- photo again. [scouten 2012-03-06, PR# 3127099]
	
		local resultP =  docStore:applyOverlayToDocumentP {
			docType = 'asset',
			docId = assetId,
			overlay = {
				_deleted = true,
				_localOnly = {
					originalFilePath = { _delete = true },
					latestOriginal = {
						filePath = { _delete = true },
					},
					proxy = {
						filePath = { _delete = true },
					},
				},
			},
		}
		
		return resultP
			-- Don't fold this together since :applyOverlay
			-- is a multi-valued return.
	end )
	
	-- Step 4 : Fetch the list of comments
	
	:mapDataP( function( )
	
		return indexStore:queryP(
			[[SELECT docId FROM comment WHERE
					assetId == ?]], assetId )
	
	end )
	
	-- Step 5 : Delete any comments
	
	:mapDataP( function( status, data )
		
		local assetIds = {}
		
		if not data[ 1 ] then
			return 'data', true
		end
		
		for index, row in ipairs( data ) do
			assetIds[ index ] = row[ 1 ]
		end
		
		return KSPromises.allP(
			function( docId )
				return docStore:setServerValueP( 'comment', docId, { _deleted = true } )
			end,
			assetIds )

	end )
	
	:mapDataP( function()
		return 'data', true
	end )
	
end
--------------------------------------------------------------------------------

local makeDeletePhotoP = function( docStore, indexStore, currentUserId, catalogOwnerId, failedDeletes )

	return function( assetId )
	
		-- FIXME : Why is limbo check failing without this?
		return KSPromises.suspendLimbo( function()
			return deleteOnePhotoP( docStore, indexStore, assetId, currentUserId, catalogOwnerId )

			:mapP( function( status, reason, ... )
				-- We could eventually support other reasons but only have one so far...
				if status == 'error' and reason == 'permissionDenied' then
				
					failedDeletes[ assetId ] = reason
					return 'data', true
					
				end

			end )

		end )
		
	end

end

--------------------------------------------------------------------------------

-- @mark delete photo

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a photo from this catalog.
	
	Signals completion via promise, but this promise may safely be ignored
	by the caller.
]]--

function DeletionMethods:deletePhotoP( photoId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	-- Now that we have cache purging working, we don't need to clean up
	-- content in the data cache. That will happen later. Just need to make
	-- sure the doc DB entry and the original file (if present) are killed.

	-- Delete the doc DB entry.
	
	local deleteDocP = self:catalogOwnerIdP()
		:mapDataP( function( _, catalogOwnerId )
			return deleteOnePhotoP( self._docStore, self._indexStore, photoId, self._ozUserId, catalogOwnerId )
		end )
	
	-- Make sure this runs to completion, regardless of whether client
	-- is listening.
	
	deleteDocP:runUntilDone()
	
	return deleteDocP

end

--------------------------------------------------------------------------------

local function canUserDeleteAsset_forMapDataP( currentUserId )
	return function( status, docValue )
		local photoOwnerId = KSTableUtils.derefOpt( docValue, "importSource", "importedBy" )
	
		if photoOwnerId then
			return 'data', currentUserId == photoOwnerId
		else
			-- This asset doesn't indicate what user imported it; so we can't know
			-- whether to allow or not.  Thus we have to choose...
			return 'data', true
		end
	end
end

--------------------------------------------------------------------------------

local function tryToFindIfUserCanDelete( self, assetIds, limit )
	local i = 1
	local n = #assetIds
	local currentUserId = self._ozUserId

	local function tryNext()
		return self._docStore:docValueOnceP( 'asset', assetIds[ i ] )
		:mapDataP( canUserDeleteAsset_forMapDataP( currentUserId ) )
		:mapDataP( function( _, result )
			if result then
				-- yes we found one, can stop iterating now
				return 'data', result
			end
			if i == limit then
				-- stop iterating, but we don't know for sure
				return 'data', "maybe"
			end
			if i == n then
				-- stop iterating, we're done.  none were found
				return 'data', false
			end
			-- keep trying
			i = i + 1
			return tryNext()
		end )
	end
	return tryNext
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns whether the photoIds can be deleted by this user on this catalog.
	
	\param limit (optional number) the maximum number of asset records to scan in
		order to make a decision.  If omitted, this API will scan at most 10 documents.
		
	\return <code>P< boolean|string ></code> A Promise object that to resolves
		to a true if the current user has permission to delete any of the
		selected the assets and the selection is less than 100 photos; if the
		selection exceeds limit photos this API will resolve to the string "maybe".
]]--

function DeletionMethods:canDeleteSelectedPhotosB( limit )
	
	limit = limit or 10
	local currentUserId = self._ozUserId
	
	return self:catalogOwnerIdB()
	:mapSwitchB( function( catalogOwnerId )
		if catalogOwnerId == nil then
			return KSReactive.constantB( false )
		end
		if currentUserId == catalogOwnerId then
			return KSReactive.constantB( true )
		else
			return self._indexStore:queryB(
					[[SELECT selectedPhotos.docId FROM selectedPhotos LIMIT ?]], limit )
			:mapSwitchB( function ( assetIds )
				local n = #assetIds
				if n == 0 then
					return KSReactive.constantB( false )
				end
				local compactedArray = {}
				for i, v in ipairs( assetIds ) do
					compactedArray[ i ] = assetIds[ i ][ 1 ]
				end
				local tryNext = tryToFindIfUserCanDelete( self, compactedArray, limit )
				return KSReactivePromiseBridge.promiseToBehavior( tryNext(), WFUnknown.value() )
			end )
		end
	end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns whether the photoIds can be deleted by this user on this catalog.
	
	\param assetId (string or table) The unique asset ID, or an array of ids
	\param optLimit (optional number) the maximum number of documents to scan when looking for
	the answer
	
	\return <code>P< boolean ></code> A Promise object that to resolves
		to a true if the current user has permission to delete the asset, 
		'maybe' if we couldn't tell, or false if not.
]]--

function DeletionMethods:canDeletePhotosP( assetIds, optLimit )
	
	local limit = optLimit or 50
	
	return self:catalogOwnerIdP()
		:mapDataP( function( _, catalogOwnerId )
			local currentUserId = self._ozUserId
			
			if currentUserId == catalogOwnerId then
			
				return 'data', true
			
			end
			
			if #assetIds == 0 then

				return 'data', false

			end
			local tryNext = tryToFindIfUserCanDelete( self, assetIds, limit )

			return tryNext()

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns whether the photoIds can be deleted by this user on this catalog.
	
	\param assetId (string) The unique asset ID
	the answer
	
	\return <code>P< boolean ></code> A Promise object that to resolves
		to a true if the current user has permission to delete the asset
]]--

function DeletionMethods:canDeletePhotoP( assetId )
	return self:canDeletePhotosP { assetId }
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a set of photos from this catalog, along with any
		local copies of the files.
	
	\param assetIds (table) The unique asset IDs.
	
	\return <code>P< true ></code> A Promise object that to resolves \c true
		when all of the deletions have been completed; can be safely be ignored.
]]--

function DeletionMethods:deletePhotosP( --[[table]] assetIds )

	WFArguments.check( OzCatalogServiceArg, 'string|table' )

    if not( type( assetIds ) == 'table' ) then
        error( "deletePhotosP: A set of assetIds must be provided", 2 )
    end
	
	local currentUserId = self._ozUserId
	local failedDeletes = {}
	
	local resultP = self:catalogOwnerIdP()
		:mapDataP( function( _, catalogOwnerId )
			local deleteOriginalP = makeDeletePhotoP( self._docStore, self._indexStore, currentUserId, catalogOwnerId, failedDeletes )
		
			return KSPromises.allP( deleteOriginalP, assetIds )
		end )
		:mapDataP( function( )
			if next( failedDeletes ) then
				return 'error', failedDeletes
			else
				return 'data', true
			end
		end )
		
	-- Make sure this runs to completion, regardless of whether client
	-- is listening.
	
	resultP:runUntilDone()
		
	return resultP
	
end

--==============================================================================

return {
	DeletionMethods = DeletionMethods,
}
