--[[----------------------------------------------------------------------------

OzCatDeletionTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSClass = require "KSClass"
local KSPromises = require "KSPromises"
local KSReactive = require "KSReactive"
local KSTableUtils = require "KSTableUtils"
local KSIdGenerator = require "KSIdGenerator"

local WFAsyncTestCase = require "WFAsyncTestCase"
local OzCatDeletion = require "OzCatDeletion"
local WFPlatformInfo = require 'WFPlatformInfo'

local TEST_WAIT_TIME = WFPlatformInfo.isAndroid and 4 or 3		-- FIXME_ANDROID

--==============================================================================

local function makeFakeObjects( catalogOwnerId )

	local docStore = KSClass "OzDocStore" {
		
		_docValues = false,
		_docOverlays = false,
		
		__init = function( self )
			self._docValues = {}
			self._docOverlays = {}
		end,
		
		testing_PrepopulateDocValue = function( self, docType, docId, docValue )
			self._docValues[ docType ] = self._docValues[ docType ] or {}
			self._docValues[ docType ][ docId ] = docValue
		end,
		
		docValueOnceP = function( self, docType, docId )
			local value = KSTableUtils.derefOpt( self._docValues, docType, docId )
			if value then
				return KSPromises.resolvedPromiseP( 'data', value )
			else
				return KSPromises.resolvedPromiseP( 'error', "invalid docType + docId combination" )
			end
		end,
		
		applyOverlayToDocumentP = function( self, params )
			self._docOverlays[ params.docType ] = self._docOverlays[ params.docType ] or {}
			self._docOverlays[ params.docType ][ params.docId ] = params.overlay
			return KSPromises.resolvedPromiseP( 'data', true )
		end,
		
		setServerValueP = function( self, docType, docId, docValue )
			self:testing_PrepopulateDocValue( docType, docId, docValue )
			return KSPromises.resolvedPromiseP( 'data', true )
		end,
	}

	local indexStore = KSClass "WFIndexStore" {
		_cannedQueryResults = false,
		
		__init = function( self )
			
			self._cannedQueryResults = {}
			
		end,
		
		testing_PrepopulateCannedQueryResults = function( self, queryString, result )
			self._cannedQueryResults[ queryString ] = result
		end,
		
		queryP = function( self )
			-- FIXME : Test returning some comments, too
			return KSPromises.resolvedPromiseP( 'data', {} )
		end,
		queryB = function( self, queryString )
			local result = self._cannedQueryResults[ queryString ]
			return KSReactive.constantB( result or {} )
		end
	}
	
	local catalogService = KSClass "OzCatalogServiceImp" {
		_indexStore = false,
		_docStore = false,
		_ozUserId = false,
		
		OzCatDeletion.DeletionMethods,
		
		__init = function( self, docStore, indexStore )
			self._docStore = docStore
			self._indexStore = indexStore
		end,
		
		catalogOwnerIdP = function( self )
			return KSPromises.resolvedPromiseP( 'data', catalogOwnerId )
		end,
		
		catalogOwnerIdB = function( self )
			return KSReactive.constantB( catalogOwnerId )
		end,
	}
	
	return docStore, indexStore, catalogService

end

--------------------------------------------------------------------------------

local DeletionTests = {
}

--==============================================================================

function DeletionTests.testCatalogOwnerCanDeletePhotoP( test )
	
	local catalogOwnerId = KSIdGenerator.uuid()
	local otherUserId = KSIdGenerator.uuid()
	local currentUserId = catalogOwnerId
	local assetId = KSIdGenerator.uuid()

	local makeDocStore, makeIndexStore, makeCatalogService = makeFakeObjects( catalogOwnerId )
	
	local docStore = makeDocStore( )
	
	docStore:testing_PrepopulateDocValue( 'asset', assetId, {
		importSource = {
			importedBy = otherUserId,
		},
		_changeCounter = 1,
	} )
	
	local indexStore = makeIndexStore()
	local catalogService = makeCatalogService( docStore, indexStore )
	catalogService._ozUserId = currentUserId
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( catalogService:canDeletePhotoP( assetId ) )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return catalogService:deletePhotosP( { assetId } )
		end )
		:continueWhenValueMatches( 'data', true )
	
end

--------------------------------------------------------------------------------

function DeletionTests.testAssetCreatorCanDeletePhotoP( test )
	
	local catalogOwnerId = KSIdGenerator.uuid()
	local otherUserId = KSIdGenerator.uuid()
	local currentUserId = otherUserId
	local assetId = KSIdGenerator.uuid()

	local makeDocStore, makeIndexStore, makeCatalogService = makeFakeObjects( catalogOwnerId )
	
	local docStore = makeDocStore( )
	
	docStore:testing_PrepopulateDocValue( 'asset', assetId, {
		importSource = {
			importedBy = otherUserId,
		},
		_changeCounter = 1,
	} )
	
	local indexStore = makeIndexStore()
	local catalogService = makeCatalogService( docStore, indexStore )
	catalogService._ozUserId = currentUserId
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( catalogService:canDeletePhotoP( assetId ) )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return catalogService:deletePhotosP( { assetId } )
		end )
		:continueWhenValueMatches( 'data', true )
	
end

--------------------------------------------------------------------------------

function DeletionTests.testCollaboratorCannotDeleteOwnersPhotoP( test )
	
	local catalogOwnerId = KSIdGenerator.uuid()
	local currentUserId = KSIdGenerator.uuid()
	local assetId = KSIdGenerator.uuid()

	local makeDocStore, makeIndexStore, makeCatalogService = makeFakeObjects( catalogOwnerId )
	
	local docStore = makeDocStore( )
	
	docStore:testing_PrepopulateDocValue( 'asset', assetId, {
		importSource = {
			importedBy = catalogOwnerId,
		},
		_changeCounter = 1,
	} )
	
	local indexStore = makeIndexStore()
	local catalogService = makeCatalogService( docStore, indexStore )
	catalogService._ozUserId = currentUserId
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( function()
			return catalogService:canDeletePhotoP( assetId )
		end )
		:continueWhenValueMatches( 'data', false )
		:waitFor( function()
			return catalogService:deletePhotosP( { assetId } )
		end )
		:continueWhenValueMatches( 'error', {
			[ assetId ] = "permissionDenied",
		} )
	
end

--------------------------------------------------------------------------------

function DeletionTests.testSmallishSelectionOfOthersPhotosAreNotDeletable( test )
	
	local catalogOwnerId = KSIdGenerator.uuid()
	local currentUserId = KSIdGenerator.uuid()
	local assetIds = {}
	local selectedPhotosQueryResults = {}
	local deletionFailureExpected = {}

	local makeDocStore, makeIndexStore, makeCatalogService = makeFakeObjects( catalogOwnerId )
	
	local docStore = makeDocStore( )
	
	for i = 1, 49 do
		local assetId = KSIdGenerator.uuid()
		assetIds[ i ] = assetId
		selectedPhotosQueryResults[ i ] = { assetId }
		deletionFailureExpected[ assetId ] = 'permissionDenied'
		
		docStore:testing_PrepopulateDocValue( 'asset', assetId, {
			importSource = {
				importedBy = catalogOwnerId,
			},
			_changeCounter = i,
		} )
	end
	
	local indexStore = makeIndexStore()
	
	indexStore:testing_PrepopulateCannedQueryResults( [[SELECT selectedPhotos.docId FROM selectedPhotos LIMIT ?]], selectedPhotosQueryResults )
	
	local catalogService = makeCatalogService( docStore, indexStore )
	catalogService._ozUserId = currentUserId
	
	-- This test takes a bit longer on WinRT and Android and iOS
	WFAsyncTestCase( test, TEST_WAIT_TIME * ( not WFPlatformInfo.isMac and 2 or 1 ) )
		:waitFor( function()
			return catalogService:canDeleteSelectedPhotosB( 50 )
		end )
		:continueWhenValueMatches( false )
		:waitFor( function()
			return catalogService:canDeleteSelectedPhotosB( 49 )
		end )
		:continueWhenValueMatches( "maybe" )
		:waitFor( function()
			return catalogService:deletePhotosP( assetIds )
		end )
		:continueWhenValueMatches( 'error', deletionFailureExpected )
	
end

--------------------------------------------------------------------------------



--==============================================================================

return DeletionTests
