--[[----------------------------------------------------------------------------

File: OzCatErrors.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSFunctionUtils = require "KSFunctionUtils"
local KSReactive = require "KSReactive"
local KSPromises = require "KSPromises"
local WFWorkerThread = require "WFWorkerThread"
local WFCloseSupport = require 'WFCloseSupport'

--==============================================================================

local BabluEP = WFWorkerThread( "BabluImp.lua" )

-- @mark OzCatalogService class

local ErrorMethods = {
}

--------------------------------------------------------------------------------

-- @mark error reporting

--------------------------------------------------------------------------------

local raw_uploadErrorB = WFCloseSupport.memoizeFn( function( self )

	return BabluEP:requestB {
		message = 'uploadErrorB',
		arguments = self._catalogId,
		startingValue = WFWorkerThread.kNilValue,
	}
	:addAsCloseTargetOf( self )

end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports all upload errors for this catalog. Use for debugging.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is
		the error message or error code when an upload error occurs, nil otherwise.
 ]]--

function ErrorMethods:raw_uploadErrorB()

	return raw_uploadErrorB( self ):relayB( true )

end

--------------------------------------------------------------------------------

local uploadErrorB = WFCloseSupport.memoizeFn( function( self )

	local rawUploadErrorB = raw_uploadErrorB( self ):relayB( true )

	return rawUploadErrorB:closingMapSwitchE( function( error )
		if error then
			return KSReactive.timerE( 15 ):constantE( error )
		else
			return KSReactive.timerE( 1 ):constantE( nil )
		end
	end ):closingHoldB( rawUploadErrorB:valueNow() )
	:addAsCloseTargetOf( self )

end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports upload errors for this catalog, resetting error conditions to
	 nil after 15 seconds.	Use for UI updates.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is
		the error message or error code when an upload error occurs, nil otherwise.
	 
]]--

function ErrorMethods:uploadErrorB()

	return uploadErrorB( self ):relayB( true )

end

--------------------------------------------------------------------------------

local raw_downloadErrorB = WFCloseSupport.memoizeFn( function( self )

	return BabluEP:requestB {
		message = 'downloadErrorB',
		arguments = self._catalogId,
		startingValue = WFWorkerThread.kNilValue,
	}
	:addAsCloseTargetOf( self )

end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports all download errors for this catalog. Use for debugging.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is
		the error message or error code when an upload error occurs, nil otherwise.
 ]]--

function ErrorMethods:raw_downloadErrorB()

	return raw_downloadErrorB( self ):relayB( true )

end

--------------------------------------------------------------------------------

local downloadErrorB = WFCloseSupport.memoizeFn( function( self )

	local rawDownloadErrorB = raw_downloadErrorB( self ):relayB( true )

	return rawDownloadErrorB:closingMapSwitchE( function( error )
		if error then
			return KSReactive.timerE( 15 ):constantE( error )
		else
			return KSReactive.timerE( 1 ):constantE( nil )
		end
	end ):closingHoldB( rawDownloadErrorB:valueNow() )
	:addAsCloseTargetOf( self )
	
end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports download errors for this catalog, resetting error conditions to
	 nil after 15 seconds. Use for UI updates.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is
		the error message or error code when an upload error occurs, nil otherwise.
	 
]]--

function ErrorMethods:downloadErrorB()

	return downloadErrorB( self ):relayB( true )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the ITMS link returned by Oz for a forced upgrade.

	\return A Promise object that resolves to the ITMS link.
]]--

function ErrorMethods:itmsLinkP()

	local itmsLinkP, resolver = KSPromises.promiseP()

	local itmsLinkBridge = self._itmsLinkB
	itmsLinkBridge
		:onDataB( function ( data )
			if data == nil then return end
			resolver( 'data', data )
			itmsLinkBridge:close()
	end )

	return itmsLinkP

end

--------------------------------------------------------------------------------
--[[!
	\brief Enables or disables automatic synchronization for this catalog.

	\param enabled (boolean) \c true to enable sync, \c false to disable.
]]--

function ErrorMethods:setSyncEnabled( --[[boolean]] enabled )

	assert( type( enabled ) == 'boolean' )

--	debug.print( string.format( "User Id: %s switching to Catalog Id: %s", self._ozUserId, self._catalogId ) )

	--debug.print( "Setting syncEnabled for catalog ", self._catalogId, " to ", enabled )

	if enabled then
		self._syncDisabledBy = false
	else
		self._syncDisabledBy = debug.traceback()
	end

	BabluEP:requestP( enabled and 'startSync' or 'stopSync', self._catalogId )
		:runUntilDone()

end

--------------------------------------------------------------------------------

local getSyncEnabledB = KSFunctionUtils.memoizeFn( function( catalogId )

	return BabluEP:requestB {
		message = 'syncEnabledB',
		arguments = catalogId,
		startingValue = false
	}

end )

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether automatic synchronization is currently enabled for this catalog.
	
	/return <code>B< boolean ></code> A Behavior object whose value is
		\c true if sync is enabled, \c false otherwise.
	 
 ]]--
function ErrorMethods:syncEnabledB()

	return getSyncEnabledB( self._catalogId ):relayB()

end

--==============================================================================

return {
	ErrorMethods = ErrorMethods,
}
