--[[----------------------------------------------------------------------------

OzCatEvents.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSStringUtils = require "KSStringUtils"
local KSDate = require "KSDate"
local WFArguments = require "WFArguments"

--==============================================================================

local OzCatalogServiceArg = 'OzCatalogServiceImp'

-- Constants for events

-- This is the threshold for determining if two events with the same name
-- are 'close enough' to be considered the same event and combined into
-- one track.
local kEventCoalesceThreshold = 6

local secondsInADay = 24 * 60 * 60

-- @mark OzCatalogService class

local EventMethods = {
}

--------------------------------------------------------------------------------

-- @mark events

--------------------------------------------------------------------------------
--[[!
    \brief Attaches an event to a set of photos in this catalog, replacing any
    event previously associated with them.
    
	\param photoIds (array of strings) The list of one or more unique asset IDs.
	\param eventName (string) The event name.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function EventMethods:attachEventToPhotosP( --[[array of strings]] photoIds, --[[string]] eventName )

    if not( type( eventName ) == 'string' and #eventName > 0 ) then
        error( "OzCatalogService:attachEventToPhotosP: eventName must be non-empty string", 2 )
    end

    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = photoIds,
        overlay = {
            eventName = KSStringUtils.trim( eventName ),
			eventTimestamp = KSDate.timeInIso8601Format( nil, false ),
        },
    }

    resultP:runUntilDone()

    return resultP
end

--------------------------------------------------------------------------------
--[[!
    \brief Removes any assigned event from a set of photos in this catalog.  Removing an event from
    a photo that currently does not have an assigned event is valid.

	\param photoIds (array of strings) The list of one or more unique asset IDs.
 
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.

]]--

function EventMethods:removeEventFromPhotosP( --[[array of strings]] photoIds )
	
	WFArguments.check( OzCatalogServiceArg, "table" )
	
    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = photoIds,
        overlay = {
            eventName = {
                _delete = true,
            },
			eventTimestamp = KSDate.timeInIso8601Format( nil, false ),
        },
    }

    resultP:runUntilDone()

    return resultP
end

--------------------------------------------------------------------------------
--[[!
	\brief Renames the event attached to a set of photos in this catalog.

	\param newEventName (string) The new name.
	\param photoIds (array of strings) The list of one or more unique asset IDs.
 
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.

]]--

function EventMethods:renameEventP( --[[string]] newEventName, --[[array of strings]] photoIds )

    if not( type( newEventName ) == 'string' and #newEventName > 0 ) then
        error( "OzCatalogService:renameEventP: The new event name must be non-empty string", 2 )
    end

    if not( type( photoIds  ) == 'table' ) then
        error( "OzCatalogService:renameEventP: The photoIds must be a table", 2 )
    end

	local resultP = self:attachEventToPhotosP( photoIds, newEventName )

  	resultP:runUntilDone()

	return resultP

end

--------------------------------------------------------------------------------

-- We use tab delimited id's for collection id's. The first value is the type
-- of collection.

local function decodeId( collectionId )
	local components = { }
	local kind = nil
	for s in string.gmatch( collectionId or '', '[^\t]+' ) do
		if not kind then
			kind = s
		else
			components[ #components + 1 ] = s
		end
	end
	return components, ( kind or '' )
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports on event labels that are near (in the sort order) to a given
	track in this catalog.
 	
	\param collectionId (string) The unique ID of the track.
	\param activeAlbum (string) The unique ID of the currently active album in this catalog.
	\param n (optional, number) The number of results to return. Default is 4.

	\return <code>P< list ></code> A Promise object that resolves to the list of nearby events.
	
]]--

function EventMethods:nearbyEventsP( --[[string]] collectionId, --[[string]] activeAlbum, --[[number]] n )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string|boolean', 'number|nil' )

    if not( type( collectionId  ) == 'string' and #collectionId  > 0 ) then
        error( "OzCatalogService:nearbyEventsP: The collectionId must be non-empty string", 2 )
    end
	
	local collectionIdDecoded = decodeId( collectionId )
	
	local resultP
	
	if collectionIdDecoded[ 2 ] then
	
		-- This is an event track with a possible date range
		local eventNameLC = collectionIdDecoded[ 4 ]
		local startDate = collectionIdDecoded[ 2 ]
		local endDate = collectionIdDecoded[ 3 ]
		
		local range = KSDate.differenceInDays( startDate, endDate )
				
		-- This is the middle of the date range for this event.
		-- We will look for events that are closest to this date.
		
		local midOfRange
		
		-- Limit the dates we're looking at to those that fall between startDate - kEventCoalesceThreshold
		-- and endDate + kEventCoalesceThreshold
		
		local startLimit = KSDate.addTimeInSeconds( startDate.. 'T00:00:00', ( kEventCoalesceThreshold * -1 ) * secondsInADay ):sub( 1, 10 )

		local endLimit = KSDate.addTimeInSeconds( endDate.. 'T00:00:00', ( kEventCoalesceThreshold ) * secondsInADay ):sub( 1, 10 )
		
		if ( range > 1 ) then
			-- Calculate the date in the middle of this range by adding range/2 number of days to the startDate
			-- Append a time component to prevent the dateformatter from rejecting this date
			midOfRange = KSDate.addTimeInSeconds( startDate.. 'T00:00:00', ( range / 2 ) * secondsInADay ):sub( 1, 10 )
		else
			-- if the range is just 1 day or 0 then we assume the start date as the mid range
			midOfRange = startDate
		end
									
		if activeAlbum then
		
		resultP = self._indexStore:queryP(
				[[SELECT DISTINCT coreInfo.eventName, coreInfo.displayDate, abs(julianday(coreInfo.displayDate) - julianday(?)) as delta FROM coreInfo
					INNER JOIN album_asset
						ON coreInfo.docId == album_asset.assetId
							WHERE coreInfo.eventNameLC != '' and coreInfo.displayDate between ? and ? and coreInfo.eventName != ? AND album_asset.albumId == ? ORDER BY delta ASC, coreInfo.displayDate DESC, coreInfo.eventName ASC ]],
								midOfRange, startLimit, endLimit, eventNameLC, activeAlbum )
		else
		
			resultP = self._indexStore:queryP(
				[[SELECT DISTINCT coreInfo.eventName, coreInfo.displayDate, abs(julianday(coreInfo.displayDate) - julianday(?)) as delta FROM coreInfo
						WHERE coreInfo.eventNameLC != '' and coreInfo.displayDate between ? and ? and coreInfo.eventName != ? ORDER BY delta ASC, coreInfo.displayDate DESC, coreInfo.eventName ASC ]],
							midOfRange, startLimit, endLimit, eventNameLC )
							
		end
			
	else
		-- This is a date track
		if activeAlbum then
		
		resultP = self._indexStore:queryP(
				[[SELECT DISTINCT coreInfo.eventName, coreInfo.displayDate, abs( julianday(coreInfo.displayDate) - julianday(?)) as delta FROM coreInfo
					INNER JOIN album_asset
						ON coreInfo.docId == album_asset.assetId
							WHERE coreInfo.eventName != '' and delta <= ? AND album_asset.albumId == ? ORDER BY delta ASC, coreInfo.displayDate DESC, coreInfo.eventName ASC ]],
								collectionIdDecoded[ 1 ], kEventCoalesceThreshold, activeAlbum )
		else
			resultP = self._indexStore:queryP(
				[[SELECT DISTINCT coreInfo.eventName, coreInfo.displayDate, abs( julianday(coreInfo.displayDate) - julianday(?)) as delta FROM coreInfo
					WHERE coreInfo.eventName != '' and delta <= ? ORDER BY delta ASC, coreInfo.displayDate DESC, coreInfo.eventName ASC ]],
						collectionIdDecoded[ 1 ], kEventCoalesceThreshold )
	end
	end
	
	return resultP:mapDataP( function( status, eventNames )
		
		local eventRows = {}

		-- Checks if the eventName has already been seen
		-- FIXME [ngautam 2012-01-18] If the value of n becomes larger than 3 or 4
		-- We need to cache this eventName lookup instead of doing it O(n) each time
		
		local function eventNameExists( eventName )
		
			for i = 1, #eventRows do
				if eventName == eventRows[ i ].eventName then
					return i
				end
			end
			return false
			
		end
		
		local i = 1
		
		-- Select the first n distinct event names from the returned set

		while ( i <= #eventNames and #eventRows < ( n or 4 ) ) do
		
			local eventNameExistsAt = eventNameExists( eventNames[ i ][ 1 ] )
			if not eventNameExistsAt then
				eventRows[ #eventRows + 1 ] = {}
				eventRows[ #eventRows ].eventName = eventNames[ i ][ 1 ]
				eventRows[ #eventRows ].displayDate = eventNames[ i ][ 2 ]
			else
				eventRows[ eventNameExistsAt ].displayDate = eventNames[ i ][ 2 ]
			end
			i = i + 1
			
		end
		-- FIXME [ngautam 2012-01-18]  Need to refactor this out so it can used elsewhere
		
		local function insertionSort()
		
			local row, j, done
		
			for i = 2, #eventRows do
		
		       	row  = eventRows[ i ]
    		    j = i - 1
        		done = false

		        repeat

					local result = KSDate.differenceInDays( row.displayDate, eventRows[ j ].displayDate )
	
    	    	    if result < 0 then
        	    	    eventRows[ j + 1 ] = eventRows[ j ]
            	    	j = j - 1
	            	    if j < 1 then
    	            	    done = true
						end
    	        	else
        	        	done = true
					end
			
	    	    until done

	        	eventRows[ j + 1 ] = row

   			end

		end
		
		-- Sort the coalesced event rows now, by the displayDate in descending order
		insertionSort()
		
		-- Extract the event names and be done
		
		local extractedEventNames = {}
		for k = 1, #eventRows do
			extractedEventNames[ k ] = eventRows[ k ].eventName
		end
		return 'data', extractedEventNames
						
	end )
end

--==============================================================================

return {
	kEventCoalesceThreshold = kEventCoalesceThreshold,
	decodeId = decodeId,
	
	EventMethods = EventMethods,
}
