--[[----------------------------------------------------------------------------

OzCatExport.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require "WFArguments"

local KSDate = require "KSDate"
local KSIdGenerator = require "KSIdGenerator"

--==============================================================================


local OzCatalogServiceArg = 'OzCatalogServiceImp'

-- @mark OzCatalogService class

local ExportMethods = {
}

--------------------------------------------------------------------------------

-- @mark exporting photos

--------------------------------------------------------------------------------
--[[!
	\brief Call this function to remember when and to where a photo has been exported.
	
	\param args A table with these keys (shown with sample values):
	       <pre>
		args = {
	
			photoId = "199ca8f035694cd698e02cb964f95dcd",
				-- photo ID
	
			serviceId = "com.facebook",
				-- use reverse DNS syntax for the service
			
			serviceName = "Facebook",
				-- user-visible name of service
			
			remoteId = "10150303662780942",
				-- (optional) ID assigned by the service for this photo
			
			url = "https://www.facebook.com/photo.php?fbid=10150303662780942&set=a.10150302349660942.342548.137189215941&type=1&theater",
				-- (optional) public URL for the photo on this service, if applicable
			
			exportedAtTime = "2011-09-27T22:45:09-08:00",
				-- (optional) time in ISO 8601 format when the photo was exported;
				-- if omitted, this API will fill in current timestamp
			
			developSettingsDigest = "199ca8f035694cd698e02cb964f95dcd",
				-- (optional) develop settings hash at time of export;
				-- if omitted, this API will fill in properly
			
			sha256 = "b7c90f6eff095e98f2c76c29ffae3c414eec53af2a625e7eb30bb828edd545dc",
				-- (optional) SHA-256 of the file at time of export;
				-- if omitted, this API will fill in current hash

			userOrientation = 1,
				-- (optional) rotation of the photo at time of export;
				-- if omitted, this API will fill in current orientation
			
			[ "com.facebook" ] = { ... },
				-- (optional) free-form additional storage for the export service;
				-- if present, must be a table with only scalar values (may contain
				-- nested tables that conform to that rule); key must match serviceId
	
		}
		</pre>
	
	See http://wichita.corp.adobe.com/2011/09/23/spec-for-asset-data-for-carousel-version-1-x/#export
	for the definitive spec on what should be in this table.
	
	\return <code>P < string></code> A Promise object that resolves to the UUID
			that was used to track the export. This promise can safely be ignored.
]]--

function ExportMethods:recordExportP( args )

	-- Validate arguments.

	WFArguments.check( OzCatalogServiceArg, 'table' )

	if not args.serviceId then
		error( "OzCatalogService:recordExportP: serviceId missing", 2 )
	end
	
	if not args.serviceName then
		error( "OzCatalogService:recordExportP: serviceName missing", 2 )
	end
	
	if not args.photoId then
		error( "OzCatalogService:recordExportP: photoId missing", 2 )
	end
	
	for k, v in pairs( args ) do
	
		local tv = type( v )
		
		if k == 'photoId' or k == 'serviceId' or k == 'serviceName'
			or k == 'remoteId' or k == 'url' or k == 'userId'
			or k == 'sha256' or k == 'exportedAtTime'
		then
			if tv ~= 'string' then
				error( string.format( "OzCatalogService:recordExportP: %s should be a string", k ), 2 )
			end
		
		elseif k == 'userOrientation' then
			if tv ~= 'number' then
				error( string.format( "OzCatalogService:recordExportP: %s should be a string", k ), 2 )
			end

		elseif k == 'developSettingsDigest' then
			if tv ~= 'table' then
				error( string.format( "OzCatalogService:recordExportP: %s should be a table", k ), 2 )
			end

		else
			error( string.format( "OzCatalogService:recordExportP: unacceptable key %s", k ), 2 )
		end
	
	end

	-- OK, data packet is valid. Store it. Fill in missing info.
	
	local photoId = args.photoId
	args.photoId = nil
	
	if not args.userId then
		args.userId = self._ozUser:id()
	end
	
	if not args.exportedAtTime then
		args.exportedAtTime = KSDate.timeInIso8601Format( nil, false )
	end
	
	local docValueP = self._docStore:docValueOnceP( 'asset', photoId )
	
	local exportId = KSIdGenerator.uuidInLowercaseHex()
	
	-- Update doc value.
	
	local applyOverlayP = docValueP:mapDataP( function( _, docValue )
		
		if not args.developSettingsDigest then
			args.developSettingsDigest = docValue
						and docValue.develop
		end
		
		if not args.sha256 then
			args.sha256 = docValue and docValue.importSource
						and docValue.importSource.sha256
		end
		
		if not args.userOrientation then
			args.userOrientation = docValue
								and docValue.develop
								and docValue.develop.userOrientation
		end
	
		local resultP = self._docStore:applyOverlayToDocumentP {
			docType = 'asset',
			docId = photoId,
			overlay = {
				exports = {
					[ exportId ] = args,
				},
			},
		}
		
		return resultP
		
	end )
	
	-- Return export ID in case anyone cares.
	
	local exportIdP = applyOverlayP:mapDataP( function()
	
		return 'data', exportId
	
	end )
	
	exportIdP:runUntilDone()
	
	return exportIdP

end

--==============================================================================

return {
	ExportMethods = ExportMethods,
}
