--[[----------------------------------------------------------------------------

OzCatFavorites.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.string
local WFZString = require 'WFZString'

local KSUndoManager = require "KSUndoManager"
local KSLogging = require "KSLogging"
local KSDate = require "KSDate"

--------------------------------------------------------------------------------

local ZSTR = assert( WFZString.fromZString )


--==============================================================================


local favoritesLog = KSLogging.log 'OzCatalogService :: favorites'
	--	:setLevel( KSLogging.DEBUG )

-- @mark OzCatalogService class

local FavoritesMethods = {
}

--------------------------------------------------------------------------------

-- @mark favorites

--------------------------------------------------------------------------------

--[[!
	\brief Reports how many users have marked a photo in this catalog as a favorite.

	\param photoId (string) The unique ID of the asset.
	\return <code>B < number ></code> A Behavior object whose value is the
		number of users.
]]--

function FavoritesMethods:countFavoritesB( --[[string]] photoId )

	return self._docStore:fieldValueB( 'asset', photoId, "favorites" )
		:mapB( function( fieldValue, isExisting )

			local count = 0
			if fieldValue then
				for k, isFavorite in pairs( fieldValue ) do
					if k ~= '_delete' then
						count = count + ( isFavorite and 1 or 0 )
					end
				end
			end

			return count, isExisting

		end )
		
end

--------------------------------------------------------------------------------

-- Event stream that describes who has favorited this photo.

function FavoritesMethods:favoritedByB( photoId )

	return self._docStore:fieldValueB( 'asset', photoId, "favorites" )
		:mapB( function( fieldValue, isExisting )

			-- FIXME [erwright 2011-04-28]: this just returns a list of userId
			-- strings.  Do we need to return a list of { userId, userFullName } tuples?
			-- [rseeler 2013-05-31]: Yes, we do.

			local users = { }
			if fieldValue then
				for userName, isFavorite in pairs( fieldValue ) do
					if isFavorite then users[ #users + 1 ] = userName end
				end
			end
			
			return users, isExisting
			
		end )

end

--------------------------------------------------------------------------------

-- Behavior for determining if a photo in this catalog has been marked as a favorite.
--[[!
	\brief Reports whether a photo has been marked as a favorite.

	\param photoId (string) The unique ID of the asset.
	\return <code>B < boolean ></code> A Behavior object whose value is
		\c true if the photo has been marked as a favorite.
]]--

function FavoritesMethods:isMyFavoriteB( --[[string]] photoId )

	return self._docStore:fieldValueB( 'asset', photoId, self._myFavoriteKey )
		:mapB( function( fieldValue, isExisting )
				
			-- View code gets annoyed by fieldValue == nil, thus the
			-- not not below.
			
			return not not fieldValue, isExisting
			
		end )
		
end

--------------------------------------------------------------------------------

-- Asynchronous API for toggling the favorite setting for a photo.
--[[!
	\brief Toggles the favorite setting for a photo in this catalog.

	\param photoId (string) The unique ID of the asset.
]]--

function FavoritesMethods:toggleMyFavorite( --[[string]] photoId )

	local favoriteB = self._docStore:fieldValueB( 'asset', photoId, self._myFavoriteKey )
		-- [scouten 2011-03-24]: Note that we're basically assuming that there
		-- is an existing favorite behavior for this field (IOW, that we're
		-- toggling favorite status for a photo that's currently on screen).
		-- If that isn't the case, we're going to take an immediate snapshot
		-- of the value which won't have populated yet, and our definition of
		-- "toggle" will fall back to "set favorite."

		-- [erwright 2011-03-24 via code review]: I'd be OK with changing this to
		-- setFavorite( photoId, isFavorite ). Not needed for this changeset, just
		-- something to consider.
		
		-- [rseeler 2012-03-12]:  If this code is ever changed, be sure to make
		-- the appropriate change to toggleSelectedPhotoP as well.
	
	KSUndoManager:beginTransaction( ZSTR "$$$/Commands/ToggleMyFavorite=Toggle Favorite" )

	if favoritesLog.level == KSLogging.DEBUG then
	
		-- Wrapping in the if statement because we don't want to call
		-- debug.traceback() if we're tossing away the results.
	
		favoritesLog:debug( "toggleMyFavorite for photo %s called by %s",
							photoId, debug.traceback() )

	end

	self._docStore:applyOverlayToDocumentP {
						--undoManager = KSUndoManager,
							-- TODO [scouten 2013-04-10]: Undo no longer supported by
							-- OzDocStore.
						docType = 'asset',
						docId = photoId,
						overlay = {
							favorites = {
								[ self._ozUserId ] = favoriteB:valueNow()
									and { _delete = true }
									or KSDate.timeInIso8601Format(),
							},
						},
					}

	favoriteB:close() -- maybe not necessary; it'll get GC'd soon enough

	KSUndoManager:commit()

end

--------------------------------------------------------------------------------
-- Asynchronous API for the favorite setting for a photo.
--[[! 
	Sets the favorite setting for a photo in this catalog.

	\param photoId (string) The unique ID of the asset.
	
	\param isMyFavorite (boolean) true if this photo is a favorite, false if not.

]]--

function FavoritesMethods:setMyFavoriteP( --[[string]] photoId, --[[boolean]] isMyFavorite )

	if favoritesLog.level == KSLogging.DEBUG then
	
		-- Wrapping in the if statement because we don't want to call
		-- debug.traceback() if we're tossing away the results.
	
		favoritesLog:debug( "setMyFavoriteP for photo %s called by %s",
							photoId, debug.traceback() )
				
	end

	return self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = photoId,
						overlay = {
							favorites = {
								[ self._ozUserId ] = isMyFavorite and true or { _delete = true }

							},
						},
					}


end

--==============================================================================

return {
	FavoritesMethods = FavoritesMethods,
}
