--[[----------------------------------------------------------------------------

OzCatImport.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- wichitafoundation.core.utils
local KSDeviceUtils = require 'KSDeviceUtils'
local KSIdGenerator = require 'KSIdGenerator'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'
local OzCatHelperFns = require 'OzCatHelperFns'
local OzEpoch = require 'OzEpoch'

	-- wichitafoundation.oz_client.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'

	-- wichitafoundation.oz_client.utils
local OzFilePaths = require 'OzFilePaths'

local BabluEP = WFWorkerThread( "BabluImp.lua" )

--------------------------------------------------------------------------------

local OzEpoch_cacheEpochB = OzEpoch.cacheEpochB

local importLog = KSLogging.log( 'OzCatImport' )
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark OzCatalogService class

local ImportMethods = {}

--------------------------------------------------------------------------------

-- @mark auto-import duplicate avoidance

--------------------------------------------------------------------------------
--[[!
	\brief Records a local device URL for a photo that has been auto-imported.
	
	Use this in concert with \c hasAutoImportedPhotoByUrl() to prevent
	duplicate attempts to auto-import the same photo, even if the user
	deleted the photo.

	\param localUrl (string) The local device URL of the asset.
	
	\return A Promise object that can safely be ignored.
]]--

function ImportMethods:recordAutoImportUrl( --[[string]] localUrl )

	-- debug.print( ":recordAutoImportUrl", localUrl )

	if not( type( localUrl ) == 'string' and #localUrl > 0 ) then
		error( "OzCatalogService:recordAutoImportUrl: localUrl must be non-empty string", 2 )
	end

	local resultP = self._docStore:applyOverlayToDocumentP {
		docType = 'autoImport',
		docId = KSIdGenerator.uuidInLowercaseHex(),
		overlay = {
			_localOnly = {
				localUrl = localUrl,
			},
		},
	}
						
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a photo at a given local device URL
	has already been auto-imported into this catalog,
	 regardless of whether it is still in the catalog.
	
	\param localUrl (string) The local device URL of the asset.
	\return <code>P< boolean ></code> A Promise object that resolves to \c true
		if the photo has been auto-imported.
]]--

function ImportMethods:hasAutoImportedPhotoAtUrlP( --[[string]] localUrl )

	if not( type( localUrl ) == 'string' and #localUrl > 0 ) then
		error( "OzCatalogService:hasAutoImportedPhotoAtUrlP: localUrl must be non-empty string", 2 )
	end

	return self._indexStore:queryP( [[SELECT docId FROM autoImportLog WHERE localUrl == ? LIMIT 1]], localUrl )
		:mapDataP( function( _, idRows )
		
			--debug.print( ":hasAutoImportedPhotoAtUrlP",
			-- not not ( idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ] ) )

			return 'data',
				   not not ( idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ] )
		
		end )
		:mapDataP( function( _, hasAutoImported )

			if not hasAutoImported then
				return self:hasPhotoByLocalUrlP( localUrl )
			else
				return 'data', true
			end
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Records an SHA-256 hash for a photo that has been auto-imported.
	
	Use in concert with \c hasAutoImportedPhotoBySha256P() to prevent
	 duplicate attempts to auto-import the same photo, even if the user deleted the photo.
	
	\param sha256 (string) The hash	value.
	\return A Promise object that can safely be ignored.
]]--

function ImportMethods:recordAutoImportSha256( --[[string]] sha256 )

	if not( type( sha256 ) == 'string' and #sha256 == 64 ) then
		error( "OzCatalogService:recordAutoImportSha256: sha256 must be a 64-byte hex string", 2 )
	end

	local resultP = self._docStore:applyOverlayToDocumentP {
		docType = 'autoImport',
		docId = KSIdGenerator.uuidInLowercaseHex(),
		overlay = {
			_localOnly = {
				sha256 = sha256,
			},
		},
	}
						
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

local function makeDocIdQuery( tableName )

	return string.format(
		[[SELECT
			SOURCE.docId, STATUS.ozUploadStatus
		FROM
			%s as SOURCE
		LEFT JOIN
			ozUploadStatusForView as STATUS
		ON
			STATUS.docId = SOURCE.docId
			AND STATUS.ozUploadStatus = "originalLost"
		WHERE
			SOURCE.sha256=?
		LIMIT
			1
		]],
			tableName )
	
end

--------------------------------------------------------------------------------

local function analyzeHashResults( _, idRows )

	local row1 = idRows[ 1 ]
	local id, status
	if row1 then
		id, status = row1[ 1 ], row1[ 2 ]
	end
	
	return 'data', false

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a photo with a given hash value
	has already been auto-imported into this catalog,
	 regardless of whether it is still in the catalog.
	
	\param sha256 (string) The hash	value.
	\return <code>P< boolean ></code> A Promise object that resolves to \c true
		if the photo has been auto-imported.
]]--

function ImportMethods:hasAutoImportedPhotoWithSha256P( --[[string]] sha256 )

	if not( type( sha256 ) == 'string' and #sha256 == 64 ) then
		error( "OzCatalogService:hasAutoImportedPhotoWithSha256P: sha256 must be a 64-byte hex string", 2 )
	end

	return self._indexStore:queryP( makeDocIdQuery( "autoImportLog" ), sha256 )
		:mapDataP( analyzeHashResults )
		:mapDataP( function( _, hasAutoImported, ... )

			if not hasAutoImported then
				return self:hasPhotoBySha256P( sha256 )
			else
				return 'data', hasAutoImported, ...
			end
		
		end )

end

--------------------------------------------------------------------------------

local function assertArgType( args, argName, expectedType, level )

	level = level or 3
	
	local result = args[ argName ]
	
	if expectedType == 'KSData' then
		if not KSData.isData( result ) then
			error( string.format( "%s is not a KSData", argName ), level )
		end
	elseif type( result ) ~= expectedType then
		error( string.format( "%s is not %s", argName, expectedType ), level )
	end
	
	return result

end

--------------------------------------------------------------------------------

local function getDataCacheP( docStore )
	
	return docStore:dataCache():pathP()
	
end

--------------------------------------------------------------------------------

local function checkDiskNotTooFullP( status, dataCachePath )

	local fileSystemFreeSpace, _ -- fileSystemSize
		= KSFileUtils.checkFreeSpaceAtPath( dataCachePath )

	local spaceNeeded = WFOzPurgeSettings.minimumFreeSpace() - fileSystemFreeSpace

	if spaceNeeded > 0 then
		return 'error', 'LowDiskSpace'
	end
	
	return 'data', false
	
end

--------------------------------------------------------------------------------

local function addToRoomyDisk( self, args )
	-- FIXME [scouten 2012-06-11]: Remove this when 3210911 is fixed (large import is failing).

	local imageInfo = assertArgType( args, 'imageInfo', 'table' )
			
	local fileSize = ( imageInfo.importSource and imageInfo.importSource.fileSize ) / ( 1024 * 1024 )
			
	if not( fileSize < 200 ) then
		return 'error', 'incompatibleFileSize'
	end
	
	local thumbData = assertArgType( args, 'thumbData', 'KSData' )
	local previewData = assertArgType( args, 'previewData', 'KSData' )
	local proxyData
	
	if args.proxyData then
		proxyData = assertArgType( args, 'proxyData', 'KSData' )
	end

	local thumbnail2xData
	if args.thumbnail2xData then
		thumbnail2xData = assertArgType( args, 'thumbnail2xData', 'KSData' )
	end
	
	local originalData, originalPath, originalDataSize
	
	if args.originalData then
	
		originalData = assertArgType( args, 'originalData', 'KSData' )
		originalDataSize = KSData.length( originalData )
		
	elseif args.originalPath then
	
		originalPath = assertArgType( args, 'originalPath', 'string' )
		local cacheFileAttrs = KSFileUtils.fileAttributesAtPath( originalPath )
		originalDataSize = cacheFileAttrs.fileSize
		
	else
		error( "OzCatalogService:addPhoto: either originalData or originalPath must be provided", 2 )
	end

	if not imageInfo.captureDate
		or not imageInfo.captureDate:match( "^%d%d%d%d%-%d%d%-%d%d" )
	then
		error( "OzCatalogService:addPhoto: imageInfo.captureDate must contain a date in ISO 8601 format yyyy-mm-dd", 2 )
	end
	
	imageInfo.originalCaptureDate = imageInfo.captureDate

	if not imageInfo.importSource then
		imageInfo.importSource = {}

	end
	imageInfo.importSource[ 'importedOnDevice' ] = KSDeviceUtils.deviceName()
	imageInfo.importSource[ 'importedBy' ] = self._ozUserId
	imageInfo.importSource[ 'importTimestamp' ] = KSDate.timeInIso8601Format( nil, true )

	local sha256 = imageInfo
			   and imageInfo.importSource
			   and imageInfo.importSource.sha256
	
	-- FIXME [scouten 2011-05-24]: We should probably just have a canonical
	-- validate SHA-256 function somewhere.
	
	if not( type( sha256 ) == 'string' and #sha256 == 64 ) then
		error( "OzCatalogService:addPhoto: imageInfo.imageSource.sha256 missing or invalid", 2 )
	end
	
	local newDocId = KSIdGenerator.uuidInLowercaseHex()

	if args.isAutoImported then

		-- Record this import info which will prevent auto-import from picking up this
		-- photo again.

		self:recordAutoImportSha256( sha256 )
		
		local url = imageInfo
		        and imageInfo._localOnly
		        and imageInfo._localOnly.localUrl
		
		if url then
			self:recordAutoImportUrl( url )
		end
	end
	
	-- Introduce a schema version that we can use to migrate data later if needed.
	
	imageInfo.version = 1
	
	local docStore = assert( self._docStore )
	local overlay
	
	local function createDocAndRenditionsP( destPath, proxyDestPath )

		if not imageInfo._localOnly then
			imageInfo._localOnly = {}
		end
		
		imageInfo._localOnly.originalFilePath = destPath
			-- FIXME [scouten 2011-06-01]: Store as path
			-- relative to catalog or originals folder?
			-- Would be more robust against renaming.

		local thumbMD5 = thumbData:digest( 'MD5' )
		local previewMD5 = previewData:digest( 'MD5' )
		local thumb2xMD5
		if thumbnail2xData then
			thumb2xMD5 = thumbnail2xData:digest( 'MD5' )
		end
		
		local epochNow = OzEpoch_cacheEpochB( self ):valueNow()

		imageInfo._localOnly.thumbnail = {
			md5 = thumbMD5,
			revId = 'pending',
			epoch = epochNow,
			size = KSData.length( thumbData ),
		}

		if thumbnail2xData then
			imageInfo._localOnly.thumbnail2x = {
				md5 = thumb2xMD5,
				revId = 'pending',
				epoch = epochNow,
				size = KSData.length( thumbnail2xData ),
			}
		end
		
		imageInfo._localOnly.preview = {
			md5 = previewMD5,
			revId = 'pending',
			epoch = epochNow,
			size = KSData.length( previewData ),
		}
		
		imageInfo._localOnly.original = {
			epoch = epochNow,
			size = originalDataSize,
			isInCacheFolder = false,
		}
		
		if proxyData then
			imageInfo._localOnly.proxy = {
				epoch = epochNow,
				size = KSData.length( proxyData ),
				filePath = proxyDestPath,
			}
		end
		
		if args.isMetered then
			imageInfo._localOnly.freemiumUploadEmbargoUntil = KSDate.timeNow() + 60
		end

		local dataCache = docStore:dataCache()
		
		-- FIXME [ngautam 2011-10-25] This is a fix for the 'gray thumbnails'
		-- issue which, I think, is being caused by the data cache write not completing
		-- in time for the read. By chaining the write to the doc store at the very end,
		-- we try to ensure that the preview and thumbnail data is already present in the
		-- cache when we look for it, thus avoiding the cache miss.

		-- The previous event stream version of this was using KSCompositeRequest
		-- to combine these requests and as per out understanding ( EricW and Neha )
		-- this meant these requests executed in parallel giving the cache writes enough
		-- time to complete.

		return dataCache:putP( previewMD5, previewData )
			:mapDataP( function()
				return dataCache:putP( thumbMD5, thumbData )
					:mapDataP( function( status, data )
						if thumbnail2xData and thumb2xMD5 then
							return dataCache:putP( thumb2xMD5, thumbnail2xData )
						else
							return status, data
						end
					end )
			end )
			:mapDataP( function()
				local resultP
				resultP, overlay = docStore:applyOverlayToDocumentP {
					docType = 'asset',
					docId = newDocId,
					overlay = imageInfo,
					addRevTimestamp = true,
				}
				return resultP:mapDataP( function ()
					return 'data', newDocId
				end )

			end )
		
	end
	
	-- FIXME [scouten 2012-01-13]: Don't like passing proxyDestPath through saveOriginalFileP.
	-- Rewrite this using promise coordination when that's available.
	
	local function saveOriginalFileP( proxyDestPath )
	
		-- FIXME [scouten 2011-05-31]: Straight copying doesn't seem to be
		-- too bad of a performance hit so far (given relatively small files).
		-- If we still think doing hard link with a follow-on sweep to copy
		-- the file is the right thing to do, this would be the place to do it.

		-- Determine new location for file and copy file there.
		-- The original for this photo hasn't been uploaded yet so we keep it
		-- in the documents folder until the upload completes
		
		local folder = OzFilePaths.originalsFolderForDate( self._originalsFolderDocuments,
														   imageInfo.captureDate )
		
		KSFileUtils.mkdir( folder )
		
		local leafName = OzCatHelperFns.importFileNameFromPhotoInfo( imageInfo )
		
		debug.print( folder, leafName )
		
		local destPath = KSPathUtils.child( folder, leafName )
		
		if not originalData then
		
			-- FIXME [scouten 2011-05-31]: Consider implementing an async
			-- version of this function. For 3 MB JPEG files (typical
			-- size for iPhone 4), this is taking only ~10 ms per file on my
			-- MacBook Pro, so this is perhaps only an issue when we attempt
			-- to scale up.

			debug.print( "Starting file copy, copying from ", originalPath )
			destPath = KSFileUtils.copyToUniquelyNamedFile( originalPath, destPath )
			debug.print( "Done with file copy" )

		else
		
			local saveDataFn = function()
				destPath = KSData.saveToUniquelyNamedFile( originalData, destPath )
			end
			
			-- Wrap this in a pcall so that we always close originalData, even if
			-- an error occurred in saving the data.
			
			local success, result = pcall( saveDataFn )
			
			originalData:close()
						
			if not success then
				error( result )
			end
		
		end

		-- When done, we must call through to:
		
		return createDocAndRenditionsP( destPath, proxyDestPath )
	
	end
	
	local function saveProxyFileP()

		local folder = OzFilePaths.originalsFolderForDate( self._originalsFolderDocuments,
														   imageInfo.captureDate )
		
		KSFileUtils.mkdir( folder )

		local leafName = OzCatHelperFns.importFileNameFromPhotoInfo( imageInfo, '_proxy.dng' )
		
		debug.print( folder, leafName )
		
		local destPath = KSPathUtils.child( folder, leafName )
		
		local saveDataFn = function()
			destPath = KSData.saveToUniquelyNamedFile( proxyData, destPath )
		end
			
		-- Wrap this in a pcall so that we always close originalData, even if
		-- an error occurred in saving the data.
			
		local success, result = pcall( saveDataFn )
			
		proxyData:close()
						
		if not success then
			error( result )
		end
		
		return saveOriginalFileP( destPath )
	end
	
	-- Make sure we clean up at the end.
	
	local resultP = self:hasPhotoBySha256P( sha256 )
		:mapDataP( function( _, docId )
		
			if docId then
				return 'data', docId
			else
				if proxyData then
					return saveProxyFileP()
				else
					return saveOriginalFileP()
				end
			end
		
		end )
	
	resultP:whenResolvedDo( function( status )
	
		self._nPhotosAdded = self._nPhotosAdded + 1
	
		previewData:close()
		thumbData:close()
		if thumbnail2xData then
			thumbnail2xData:close()
		end
		if status ~= 'data' and overlay then
			overlay:undoP()
		end
	
	end ):runUntilDone()

	return resultP
	
end

--------------------------------------------------------------------------------

local function importAndCheckDiskSpaceP( self, args )

	return getDataCacheP( self._docStore )
		:mapP( checkDiskNotTooFullP )
		:mapDataP( function()
		
			return addToRoomyDisk( self, args )
			
		end )

end

--------------------------------------------------------------------------------

local function importWithMeteredLimitsP( self, ownerUser, args )

	return ownerUser:refreshP()
		:mapDataP( function()
			
			return ownerUser:uploadLimitAndUsageP()
		
		end )
		:mapDataP( function( _, limit, usage )
		
			importLog:info( "upload limit = %s, usage = %s", limit, usage )
			
			if type( limit ) == 'number'
				and type( usage ) == 'number'
				and usage >= limit
			then
				return 'error', 'UsageLimitExceeded'
			else
				return importAndCheckDiskSpaceP( self, args )
			end

		end )
		:mapDataP( function( _, docId )
		
			-- FIXME [scouten 2012-11-08]: This the fully formed catalog URL
			-- should probably just be an attribute of the catalog object.
		
			local catalogUrl = string.format( '%s/%s',
											  OzCatalogUrls.catalogsUrl,
											  self._catalogId )

			return BabluEP:requestP( 'uploadP', {
				catalogId = self._catalogId,
				docType = 'asset',
				docId = docId,
				uploadStatus = 'notUploaded_metadataOnly',
				ozUserId = self._ozUserId,
				catalogUrl = catalogUrl,
				-- TODO [arahn 2013-04-04]: USE_LINK_TEMPLATES server should provide this URL
				docUrl = string.format( '%s/%ss/%s', args.catalogUrl, args.docType,
										args.docId ),
				errorCaptureForDoc = {},
			} )
			:mapP( function( status, value )

				importLog:info( "uploadP completed with status %s, value = %s",
								status, value )

				if status == 'error' then
					return 'error', 'TimeoutOnUpload'
				else
					return ownerUser:adjustUploadUsageP( 1 )
						:mapDataP( function( )
							return 'data', docId
						end )
				end

			end )

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a photo asset to this catalog.

	This function also checks for free disk space and returns an error with
	'LowDiskSpace' as the message if the free space falls below the minimum as
	defined in  \c oz/purge_agent/WFOzPurgeSettings.lua.

	\param args (table) The image data. See example.

	\return <code>P< docId ></code> A Promise object that resolves to the unique
		ID of the new asset.
	
	For example:
	<pre>
	P< docId > = catalog:addPhotoP {
	
		imageInfo = {
			-- sample image document data
			captureDate = "2003-04-12T13:54:37Z",
			height = 2032,
			importSource = {
				fileName = "es-0601-006.jpg",
				fileSize = 1116049,
				folder = "/Users/scouten/Pictures/",
				sha256 = "3cddf5178eba5b6025aeeb13f95bcc554e0857a9cdd4cf7e1f5c05d058c199b2",
			},
			keywords = {
				friends = true,
				people = true,
			},
			width = 1352,
		},
		
		thumbData = (KSData),
		previewData = (KSData),
		thumbnail2xData = (KSData),	-- optional

		originalData = (KSData),	-- original file data
		originalPath = (string),	-- path to original file
					
		-- One or the other of originalData or originalPath is required.
		
		isAutoImported = (boolean),
					-- optional, records extra info for auto-import de-duping
		}
	}
	</pre>
]]--

function ImportMethods:addPhotoP( --[[table]] args )

	return self:catalogOwnerUserP()
		:mapDataP( function( _, ownerUser )
		
			return ownerUser:isMeteredFreemiumUserP()
				:mapDataP( function( _, isMetered )
				
					args.isMetered = isMetered
					self._ozUser:addToPendingUploadCountP( self._catalogId, 1 )
					
					if isMetered then
						return importWithMeteredLimitsP( self, ownerUser, args )
					else
						return importAndCheckDiskSpaceP( self, args )
					end
		
				end )
		
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a photo with a matching SHA256 hash already exists in this
	catalog.

	If multiple photos match, only one doc ID is sent. Which one is undefined, but if any
	match has \c ozUploadStatusForView.ozUploadStatus of \c originalLost, that ID
	is returned before any other.
	
	\param sha256 (string) The hash	value.
	
	\return <code>P< id, status | false ></code> A Promise object that resolves to
		the document ID of a matching photo, if any, AND the \c ozUploadStatus for that photo.
	 	  
	 	If no matching photo is found, resolves to \c false.
]]--

function ImportMethods:hasPhotoBySha256P( --[[string]] sha256 )

	if not( type( sha256 ) == 'string' and #sha256 == 64 ) then
		error( "OzCatalogService:hasPhotoBySha256P: sha256 must be a 64-byte hex string", 2 )
	end

	return self._indexStore:queryP( makeDocIdQuery( "importSource" ), sha256 )
		:mapDataP( analyzeHashResults )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the document ID of the photo at a local URL, if any.
	If multiple photos match, only one doc ID is sent; which one is undefined.

	\param localUrl (string) The local device URL of the asset.
	
	\return <code>P< docId or false ></code> A Promise object that resolves to
		the unique ID of the asset, or \c false if no asset is found.
 ]]--

function ImportMethods:hasPhotoByLocalUrlP( --[[string]] localUrl )

	if not( type( localUrl ) == 'string' and #localUrl > 0 ) then
		error( "OzCatalogService:hasPhotoByLocalUrlP: localUrl must be non-empty string", 2 )
	end

	return self._indexStore:queryP( [[SELECT docId FROM importSource WHERE localUrl == ? LIMIT 1]], localUrl )
		:mapDataP( function( _, idRows )
		
			return 'data',
				   ( idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ] ) or false
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the name of the device where an asset was imported.

	\param photoId (string) The unique asset ID.

	\return <code>P< string ></code> A Promise object that resolves to
		the device name, or an error message is none is found.

]]--

function ImportMethods:getImportDeviceForPhotoP( --[[string]] photoId )

	return self._docStore:docValueOnceP( 'asset', photoId )
	:mapP( function ( status, photoInfo )

		if status  == 'data' then

			if photoInfo.importSource and photoInfo.importSource.importedOnDevice then

				return 'data', photoInfo.importSource.importedOnDevice
			else
				return 'error', 'ImportMethods:getImportDeviceForPhotoP no import device data for this asset'
			end

		else
			return 'error', 'ImportMethods:getImportDeviceForPhotoP no import device data for this asset'
		end

	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the name of the user who imported an asset into this catalog.

	\param photoId (string) The unique asset ID.

	\return <code>P< string ></code> A Promise object that resolves to
		the user name, or an error message is none is found.

Returns via promise the user id for the user who imported this asset.
]]--

function ImportMethods:getImportedByForPhotoP( --[[string]] photoId )

	return self._docStore:docValueOnceP( 'asset', photoId )
	:mapP( function ( status, photoInfo )

		if status  == 'data' then

			if photoInfo.importSource and photoInfo.importSource.importedBy then

				return 'data', photoInfo.importSource.importedBy
			else
				return 'error', 'ImportMethods:getImportedByForPhotoP no import data for this asset'
			end

		else
			return 'error', 'ImportMethods:getImportedByForPhotoP no import data for this asset'
		end

	end )

end

--==============================================================================

return {
	ImportMethods = ImportMethods,
}
