--[[----------------------------------------------------------------------------

OzCatInit.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local KSFileUtils = require 'KSFileUtils'

	-- wichitafoundation.core.reactive
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local _ = require 'WFArguments_uuid'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'
local OzFilePaths = require "OzFilePaths"

	-- wichitafoundation.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- wichitafoundation.web.http_client
local WFReachability = require 'WFReachability'

	-- wichitafoundation.core.os
local WFPlatformInfo = require 'WFPlatformInfo'


--------------------------------------------------------------------------------

local BabluEP = WFWorkerThread( 'BabluImp.lua' )

local catalogLog = KSLogging.log 'OzCatalogService'
	--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark OzCatalogService class

local InitMethods = {}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a catalog instance for an Oz user.
	
	\param catalogId (string) The UUID for the catalog
	\param ozUser (OzUser) The user who is accessing the catalog.
]]--

function InitMethods:__init( catalogId, ozUser )

	WFArguments.check( 'OzCatalogServiceImp', 'uuid', 'OzUser' )

	self._catalogId = catalogId
	self._refsToThisCatalog = KSTableUtils.makeFullyWeak()
	
	self._ozUser = ozUser
	self._ozUserId = ozUser:id()

	-- Make sure that we are now the current user and that we close
	-- this catalog instance if/when that changes.

	local userIsCurrentB = ozUser:isCurrentUserB()
		:addAsCloseTargetOf( self )

	if not userIsCurrentB:valueNow() then

		catalogLog:error( "Creating OzCatalogService with invalid OzUser object at %s",
						  debug.traceback() )

		error( "OzCatalogService: ozUser is not current user", 4 )

	end
	
	userIsCurrentB:onDataE( function( isCurrent )
		if not isCurrent then
			self:close()
		end
	end )

	self._myFavoriteKey = string.format( "favorites/%s", self._ozUserId )

	-- Spin up doc store and index store.

	self._docStore = OzDocStore( catalogId )
	self:addCloseTarget( self._docStore )
	
	self._indexStore = WFIndexStore( self._docStore, 'OzCatalogIndexDefinitions.lua' )
	self:addCloseTarget( self._indexStore )
	
	-- GRUNGE [scouten 2011-05-31]: Should really have a centralized location
	-- where we calculate and return catalog paths. This code is approximately
	-- duplicated from OzDocStoreImp.

	self._originalsFolderDocuments = OzFilePaths.getStandardOriginalFolderDocumentsPath( catalogId )
										
	self._originalsFolderCaches = OzFilePaths.getStandardOriginalFolderCachesPath( catalogId )
	
	KSFileUtils.mkdir( self._originalsFolderDocuments )
	KSFileUtils.mkdir( self._originalsFolderCaches )
	
	-- GRUNGE [ ngautam 08-25-2011 ] This code for getting the preference setting and
	-- enabling downloading of originals needs a cleanup (post V1).

	self._downloadOriginalsB = BabluEP:requestB {
		message = 'getOzDownloadOriginalsEnabled',
		startingValue = WFWorkerThread.kNilValue,
	}
	self:addCloseTarget( self._downloadOriginalsB )

	self._downloadOriginalsB:onDataB( function( shouldDownload )

		if shouldDownload ~= nil then
			self:setDownloadOriginalsEnabled( shouldDownload )
		end

	end )

	-- FIXME [ngautam 2011-08-31] Need to make this promise-based.

	self._itmsLinkB = BabluEP:requestB {
		message = 'itmsLinkB',
		arguments = self._catalogId,
		startingValue = WFWorkerThread.kNilValue,
	}
	self:addCloseTarget( self._itmsLinkB )

	self._hasBeenDeletedB = KSReactive.variableB( false )

	self._accessErrorB = KSReactive.variableB( false )
	self:addCloseTarget( self._accessErrorB )
	
	-- Make sure we get status updates for catalog owner, not just current user.
	
	-- FIXME [rseeler 2013-05-03]: We're getting extra hacky with the thread name comparison, but since it wasn't fixed 
	-- for the standard case (KSLVCEventProcessorScript.lua), we're not going to attempt to fix it just for Tornado.
	if WFCurrentThread.DEBUG_getThreadName() == 'KSLVCEventProcessorScript.lua' or WFCurrentThread.DEBUG_getThreadName() == 'REEventProcessorScript.lua' then

		-- FIXME [scouten 2012-11-01]: Need to prevent this logic from running in
		-- more than one thread. Hacky way to get there.

		self:watchCatalogOwnerUserStatus()

	end

	-- Track reachability status notifications.
	
	self._ozHostIsReachableB = WFReachability.hostIsReachableB( OzCatalogUrls.baseServerUrl )
	self:addCloseTarget( self._ozHostIsReachableB )
	
	self._ozHostReachabilityLog = {}
	
	self._ozHostIsReachableB:onDataB( function( newStatus )
	
		local logSize = #( self._ozHostReachabilityLog )
		
		self._ozHostReachabilityLog[ logSize + 1 ] = {
			status = newStatus or 'nil',
			time = KSDate.timeInIso8601FormatWithTimeZone(),
		}
		
		if logSize >= 10 then
			table.remove( self._ozHostReachabilityLog, 1 )
		end
	
	end )
	
end

--------------------------------------------------------------------------------


--[[!
	\brief Retrieves the Oz user who owns this catalog.
	
 	\return (OzUser) The Oz user instance.
]]--

function InitMethods:ozUser()

	return self._ozUser

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the unique ID of this catalog.
	
 	\return (string) The UUID.
]]--

function InitMethods:getCatalogId()

	return self._catalogId

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the name of this catalog.
	
 	\return <code>B< string ></code> A Behavior object whose value is the catalog name.
]]--

function InitMethods:nameB()

	if not self._nameB then

		self._nameB = self._docStore:fieldValueB( 'catalog', self._catalogId,
							'name' )

		self:addCloseTarget( self._nameB )
	
	end
	
	return self._nameB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the document store instance that backs this catalog.
	
	\warning Do not close this document store; doing so causes
		any other clients of this catalog instance to fail.
	
	\return (OzDocStore) The document store instance.
--]]

function InitMethods:docStore()

	return self._docStore

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the index store instance that backs this catalog.
	
	\warning Do not close this index store; doing so causes
		any other clients of this catalog instance to fail.
	
	\return (WFIndexStore) The index store instance.
--]]

function InitMethods:indexStore()

	return self._indexStore

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the number of documents left to index in this catalog.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return <code>B< number ></code>A Behavior object whose value is the number
		of documents left to index.
--]]

function InitMethods:sizeOfIndexQueueB()

	return self._indexStore:requestE( 'sizeOfIndexQueueB' ):closingHoldB( nil )

end

--------------------------------------------------------------------------------

return {
	InitMethods = InitMethods,
}
