--[[----------------------------------------------------------------------------

OzCatProxies.lua

@author arahn 2012-08-30

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSData = require "KSData"
local KSFileUtils = require "KSFileUtils"
local KSIdGenerator = require "KSIdGenerator"
local KSLogging = require "KSLogging"
local KSPathUtils = require "KSPathUtils"
local KSReactive = require "KSReactive"
local OzCatalogUploadStatus = require "OzCatalogUploadStatus"
local OzCatalogUrls = require "OzCatalogUrls"

local OzCatHelperFns = require "OzCatHelperFns"
local OzEpoch = require "OzEpoch"

local WFOzRenditionServerUtils = require 'WFOzRenditionServerUtils'


--============================================================================--

local OzCatHelperFns_nilP = assert( OzCatHelperFns.nilP )
local OzCatHelperFns_progressIncompleteB = assert( OzCatHelperFns.progressIncompleteB )
local OzCatHelperFns_DISABLE_MOST_FETCHES = OzCatHelperFns.DISABLE_MOST_FETCHES
local OzCatHelperFns_processHttpRequestResult = assert( OzCatHelperFns.processHttpRequestResult )
local WFOzRenditionServerUtils_touchRendition = assert( WFOzRenditionServerUtils.touchRendition )
local OzEpoch_cacheEpochB = OzEpoch.cacheEpochB

local catalogLog = KSLogging.log 'OzCatalogService'
	--	:setLevel( KSLogging.DEBUG )

-- @mark OzCatalogService class

local ProxyMethods = {
}

--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
--[[!
	\brief Checks whether a valid proxy for a photo in this catalog
	is available either locally or on the server.
	A valid proxy is one that corresponds to latest original.
	
	\param photoId (string) The unique ID of the asset.
	\param checkLocally (optional, boolean) When \c true, check only for
		local availability of the proxy.  Otherwise, also check
		on the server.
	
	\return <code>P< boolean ></code> A Promise object that resolves to
		\c true if the proxy is available, \c false if not
]]--

function ProxyMethods:proxyAvailableP( --[[string]] photoId,
					--[[optional boolean]] checkLocally )

	local docValueP = self._docStore:docValueOnceP( 'asset', photoId )

	return docValueP:mapDataP( function( status, docValue )
	
		local localPath = docValue
			and docValue._localOnly
			and docValue._localOnly.proxy
			and docValue._localOnly.proxy.filePath

		localPath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath )

		local ozUrl = docValue
			and docValue._serverProvidedMetadata
			and docValue._serverProvidedMetadata.links
			and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ]
			and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ].href
		
		if OzCatalogUploadStatus.assetProxyIsValid( docValue ) then
			if checkLocally then
				if localPath then
					return 'data', true
				else
					return 'data', false
				end
			elseif not ( localPath or ozUrl ) then
				return 'data', false
			else
				return 'data', true
			end
		else
			return 'data', false
		end

	end )

end

--------------------------------------------------------------------------------

local function downloadAndCacheProxyPathP( self, proxyUrl, photoId, imageInfo )
	
	local optArgs = {}
	optArgs[ 'wantDownloadProgress' ] = true
	
	local resultP, progressB = self._ozUser:getP( proxyUrl, optArgs )
	
	resultP = resultP:observeP( function( status, data, headers )
	
		OzCatHelperFns_processHttpRequestResult( self, status, data, headers )
		
	end )
	
	local resultPP = resultP:mapDataP( function( _, proxyData )
		
			local captureDate = imageInfo.captureDate:sub( 1, 10 )
			
			-- FIXME [scouten 2012-02-07]: Should move instances of this folder path logic into one place
			local subPath = KSPathUtils.child( captureDate:sub( 1, 4 ), captureDate:sub( 1, 7 ) ) -- yyyy, yyyy-mm
			KSPathUtils.child( subPath, captureDate ) -- yyyy-mm-dd
						
			local folder
			local overlay = {
							_localOnly = {
								proxy = {
									},
								},
							}
							
			folder = KSPathUtils.child( self._originalsFolderCaches, subPath )
							
			KSFileUtils.mkdir( folder )
			
			local leafName = imageInfo.importSource
								and imageInfo.importSource.fileName
		
			if not leafName then
				leafName = imageInfo.captureDate:sub( 12 ):gsub( '[:Z]', '' ) .. "_proxy.dng"
			else
				leafName = leafName:gsub( '\.[^.]*$' , '_proxy.dng' )
			end
			
			--debug.print( folder, leafName )
			
			local destPath = KSPathUtils.child( folder, leafName )

			destPath = KSData.saveToUniquelyNamedFile( proxyData, destPath )
			
			local dataSize = KSData.length( proxyData )
			
			overlay._localOnly.proxy.filePath = destPath
			overlay._localOnly.proxy.ozUrl = imageInfo
											and imageInfo._serverProvidedMetadata
											and imageInfo._serverProvidedMetadata.links
											and imageInfo._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ]
											and imageInfo._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ].href
			overlay._localOnly.proxy.size = dataSize
			overlay._localOnly.proxy.epoch = OzEpoch_cacheEpochB( self ):valueNow()
			
			proxyData:close()
			
			self._docStore:applyOverlayToDocumentP {
				docType = 'asset',
				docId = photoId,
				overlay = overlay,
			}
			
			return 'data', destPath
			
		end )
	
	return resultPP, progressB
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves the local path of a proxy photo document,
	and optionally downloads the asset itself.
 
	\param photoId (string) The asset's unique ID.
	\param doNotFetch (optional, boolean) When true, do not retrieve the asset.
	
	\return <code>P < string ></code>, <code>B < number ></code>
		A Promise object that resolves on completion to the
		path of the downloaded asset, or to nil if the asset has not been
		previously downloaded and has not been fetched by this call.
	    	Also returns a Behavior whose value reports the
	    	progress of the asynchronous operation.
]]--

function ProxyMethods:proxyDataPathP( --[[string]] photoId, --[[boolean]] doNotFetch )

	if OzCatHelperFns_DISABLE_MOST_FETCHES or not photoId then
		return OzCatHelperFns_nilP
	end
	
	local docValueOnceP = self._docStore:docValueOnceP( 'asset', photoId )

	local currentProgressB = KSReactive.variableB( nil )
	
	local progressB = currentProgressB:mapSwitchB( function( progressB )
		if progressB then
			return progressB
		else
			return OzCatHelperFns_progressIncompleteB
		end
	end )

	local resultP = docValueOnceP:mapDataP( function( status, docValue )
	
		if not docValue then
			return 'data', nil -- or is this an error state, too?
		end
		
		if not OzCatalogUploadStatus.assetProxyIsValid( docValue ) then
			-- FIXME [sdaniel 2012-02-07]: Request Oz to generate a proxy
		end
		
		-- First preference: Use proxy that's already stored locally.
		
		local localPath = docValue
						and docValue._localOnly
						and docValue._localOnly.proxy
						and docValue._localOnly.proxy.filePath

		localPath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath )
		
		if localPath and KSFileUtils.exists( localPath ) then
			WFOzRenditionServerUtils_touchRendition( self._docStore, OzEpoch_cacheEpochB( self ),
													 photoId, 'proxy', docValue )
			return 'data', localPath
		end
		
		if doNotFetch then
			return 'data', nil
		end
		
		-- Second preference: Download from Oz.
		
		local ozUrl = docValue
						and docValue._serverProvidedMetadata
						and docValue._serverProvidedMetadata.links
						and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ]
						and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ].href
		
		if ozUrl then

			ozUrl = OzCatHelperFns.checkForRelativeUrlAndAppendBase( ozUrl, OzCatalogUrls.baseServerUrl )

			local resultP, rawProgressB = downloadAndCacheProxyPathP( self, ozUrl, photoId, docValue )
			
			currentProgressB:set( rawProgressB )
			return resultP

		end
		
		return 'error', "Proxy can not be found"
			-- FIXME [scouten 2011-07-25]: Does this need to be localized?
	
	end )
	
	return resultP, progressB

end

--------------------------------------------------------------------------------

--[[!
     \brief Sends a request to the Oz server to generate a proxy version of
       a photo in this catalog.

       \param photoId (string) The asset's unique ID.

       \return <code>P < ?? ></code> A Promise object that resolves to ((what?))
]]--
function ProxyMethods:requestOzGenerateProxyP( --[[string]] photoId )
	
	if not photoId then
		return OzCatHelperFns_nilP
	end
	
	local docValueOnceP = self._docStore:docValueOnceP( 'asset', photoId )
	
	local resultP = docValueOnceP:mapDataP( function( status, docValue )
			
		local proxyPostUrl = docValue._serverProvidedMetadata
						 and docValue._serverProvidedMetadata.link_templates
						 and docValue._serverProvidedMetadata.link_templates[ "/rels/proxy_generate" ]
						 and docValue._serverProvidedMetadata.link_templates[ "/rels/proxy_generate" ].href

		if not proxyPostUrl then
			return 'error', "No URL template available. Can't ask Oz to generate proxy."
		end

		proxyPostUrl = proxyPostUrl:gsub( '{proxy_id}', KSIdGenerator.uuidInLowercaseHex() )
		
		-- If the url from the server is partial, only then append the base. ASSET documents don't 
		-- contain the base server url, it would be nice to have that.

		if proxyPostUrl:match( "^/" ) then
			proxyPostUrl = string.format( "%s%s", OzCatalogUrls.baseServerUrl, proxyPostUrl )
		end

		catalogLog:debug( "Attempting to POST %s", proxyPostUrl )
	
		local proxyPostArgs = {
			requestHeaders = {
				{
					header = 'Content-Type',
					value = 'image/dng'
				},
				{
					header = 'X-Proxy-Type',
					value = '2048'
				},
				{
					header = 'Accept',
					value = 'application/json'
				},
			},
			requestMethod = 'POST'
		}
		
		local proxyPostP = self._ozUser:putP( proxyPostUrl, nil, proxyPostArgs )
			
		local proxyPostStatusP = proxyPostP:mapDataP( function( _, data, headers )
		
			if not headers or headers.status_code ~= '201' then
				catalogLog:error(
				"Unexpected result code %s on POST to %s\n\ndata = %s",
									headers.status_code,
									proxyPostUrl,
									data )

				return 'error', data, headers
			end
			
			-- Oz generated the proxy, now retrieve it.
			
			local proxyDataPathP = downloadAndCacheProxyPathP( self, proxyPostUrl, photoId, docValue )
	
			return proxyDataPathP
	
		end )
		
		return proxyPostStatusP
		
	end )
	
	return resultP
	
end


--==============================================================================

return {
	ProxyMethods = ProxyMethods,
}
