--[[----------------------------------------------------------------------------

OzCatSelection.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require "WFArguments"
local KSPromises = require "KSPromises"

--==============================================================================

local OzCatalogServiceArg = 'OzCatalogServiceImp'

-- @mark OzCatalogService class

local SelectionMethods = {
}

--------------------------------------------------------------------------------

-- @mark selection

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the currently selected assets in this catalog.
	
	\return <code>P< table ></code> A Promise object that resolves
		to the list of selected asset IDs.
]]--

function SelectionMethods:getSelectedPhotosP()

	return self._indexStore:queryP(
					[[SELECT docId FROM selectedPhotos]] )
		:mapDataP( function( _, rows )
	
			local extractedRows = {}
			for i = 1, #rows do
				extractedRows[ i ] = rows[ i ][ 1 ]
			end
			return 'data', extractedRows
		
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves (as a Promise) the number of currently selected assets in this catalog.
	
	\return <code>P< number ></code> A Promise object that resolves
		to the number of selected assets.
]]--

function SelectionMethods:selectedPhotosCountP()

	return self._indexStore:queryP(
					[[SELECT COUNT( rowid ) FROM selectedPhotos]] )
		:mapDataP( function( _, data )
	
			return 'data', data and data[ 1 ] and data[ 1 ][ 1 ] or 0
		
		end )
		:runUntilDoneP()
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the number of currently selected assets in this catalog.
	
	\return <code>B< number ></code> A Behavior object whose value
		is the number of selected assets.
	
	\warning If you are only interested in whether \b any photos are selected,
		 use <code>:hasAnySelectedPhotosB()</code>, which is much
		more efficient.
	
 ]]--

function SelectionMethods:selectedPhotosCountB()

	if not self._selectedPhotosCountB then

		self._selectedPhotosCountB =  self._indexStore:queryB(
				[[SELECT COUNT( rowid ) FROM selectedPhotos]] )
			:mapB( function ( data )
				return data and data[ 1 ] and data[ 1 ][ 1 ] or 0
			end )

	end

	return self._selectedPhotosCountB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether any photos in this catalog are currently selected.
	
	\return <code>B< boolean ></code> A Behavior object whose value
		is \c true if any photos are currently selected.
--]]

function SelectionMethods:hasAnySelectedPhotosB()

	if not self._hasAnySelectedPhotosB then

		local resultB =
				-- SQLite + index store trickiness: SELECT 1 will return an empty result
				-- set if there are no rows and a single 1 = 1 row for every row.
				-- Combine that with LIMIT 1 and you have a behavior that doesn't change
				-- value except on the transition between zero and non-zero number of assets.
				-- Also: It's very, very fast for SQLite to evaluate this query.
			
				self._indexStore:queryB( [[SELECT count( a )
					FROM ( SELECT 1 AS a FROM selectedPhotos LIMIT 1 ) ]] )

		self._hasAnySelectedPhotosB = resultB
			:mapB( function ( data )

				local result = false

				if data and #data > 0 then
					result = data[ 1 ][ 1 ] == 1
				end

				return result

			end )

	end

	return self._hasAnySelectedPhotosB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether exactly one photo in this catalog is currently selected.
	
	\return <code>B< boolean ></code> A Behavior object whose value
		is \c true if exactly one photo is currently selected.
		The value is \c false if no photos, or two or more photos, are selected.
--]]

function SelectionMethods:hasExactlyOneSelectedPhotoB()

	if not self._hasExactlyOneSelectedPhotoB then

		local resultB =
				-- SQLite + index store trickiness: SELECT 1 will return an empty result
				-- set if there are no rows and a single 1 = 1 row for every row.
				-- Combine that with LIMIT 2 and you have a behavior that doesn't change
				-- value except on the transition between zero, one, and more-than-one assets.
				-- Also: It's very, very fast for SQLite to evaluate this query.
			
				self._indexStore:queryB( [[SELECT count( a )
					FROM ( SELECT 1 AS a FROM selectedPhotos LIMIT 2 ) ]] )

		self._hasExactlyOneSelectedPhotoB = resultB
			:mapB( function ( data )

				local result = nil

				if data and #data > 0 then
					result = data[ 1 ][ 1 ] == 1
				end

				return result

			end )

	end

	return self._hasExactlyOneSelectedPhotoB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the selected assets for this catalog.
	
	\return <code>B< table ></code> A Behavior object whose value is the list of
		asset IDs.
]]--

function SelectionMethods:selectedPhotosB()
	
	return self._indexStore:queryB(
			[[SELECT selectedPhotos.docId FROM selectedPhotos]] )
		:mapB( function ( data )
			local extractedRows = {}
			for i = 1, #data do
				extractedRows[ i ] = data[ i ][ 1 ]
			end
			return extractedRows
		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deselects one or more photos in this catalog.

	\param photoIds (array of strings) The unique asset IDs of the photos.

	\return <code>P< boolean ></code> A Promise object that resolves to
		\c true if the request succeeded.
]]--

function SelectionMethods:clearSelectedPhotosP( --[[array]] photoIds )

    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = photoIds,
        overlay = {
			_localOnly = {
				selected = {
					_delete = true,
				},
			},
        },
    }

    return resultP
end

--------------------------------------------------------------------------------
--[[!
	\brief Trims the selection to include only photos that are currently selected
	and also belong to a given album in this catalog.
	((that's not what it looks like - what does it do?))
	
	\param viewGrouping (string) The current view grouping (such as 'date')((valid values?))
	\param activeAlbum (string or false) The unique album ID, or \c false to indicate that no album is selected
	\param collectionId (string) The unique ID of the current track.	((??))
	\param filter (string or nil) The name of a filter (such as 'favorites'), or nil ((valid values?))
	\return <code>P< table ></code> list of assetIds
]]--

function SelectionMethods:intersectSelectionWithSourceP( --[[string]] viewGrouping, --[[string]] activeAlbum,
							--[[string]] collectionId, --[[string]] filter )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string|boolean', '?string|boolean', '?string|boolean' )
	
	local whereClause = "1"
	local queryParams = {}
	
	if ( collectionId and collectionId ~= 'grid' ) or ( filter and #filter > 0 ) or viewGrouping == 'ozUploadStatus' or viewGrouping == 'dateEventsOnly' or viewGrouping == 'events' then
		-- TODO [arahn 2012-09-11]: Some day it'd be nice to not drop the selection in these cases
		-- However, that is best done after the query's in OzCatCollectionQueries.lua are refactored
		-- such that the membership of a particular viewGrouping, activeAlbum, collectionId, filter
		-- can be shared with other kinds of queries (in this case a query for selected assets).
	else

		-- As a proof of concept, let's try not dropping selection when switching albums
		if activeAlbum then
			whereClause = "docId NOT IN ( SELECT assetId FROM album_asset WHERE albumId = ? )"
			queryParams = { activeAlbum }
		else
			-- No album selected, nor a collectionId nor any filter --
			-- AND we're not viewing ozUploadStatus
			-- therefore all the photos are visible, no need to winnow
			-- the selection at all.
			return KSPromises.resolvedPromiseP( 'data', nil )
		end
	
	end
	
	
	return self._indexStore:queryP(
		[[SELECT selectedPhotos.docId FROM selectedPhotos WHERE ]].. whereClause, unpack( queryParams ) )
		
		:mapDataP( function( status, photosToDeselect )
			
			for idx, v in ipairs( photosToDeselect ) do
				photosToDeselect[ idx ] = photosToDeselect[ idx ][ 1 ]
			end
			return self:clearSelectedPhotosP( photosToDeselect )
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Selects one or more photos in this catalog.

	\param photoIds (array of strings) The unique asset IDs of the photos.

	\return <code>P< boolean ></code> A Promise object that resolves to
		\c true if the request succeeded.
]]--

function SelectionMethods:setSelectedPhotosP( --[[array of strings]] photoIds )

    local resultP = self._docStore:applyOverlayToDocumentP {
        docType = 'asset',
        docIds = photoIds,
        overlay = {
			_localOnly = {
				selected = true,
			},
        },
    }

    return resultP
end

--------------------------------------------------------------------------------
--[[!
	\brief Toggles the selection state of a photo in this catalog.

	\param photoId (string) The unique asset ID of the photo.

	\return <code>P< boolean ></code> A Promise object that resolves to
		\c true if the request succeeded.
]]--

function SelectionMethods:toggleSelectedPhotoP( --[[string]] photoId )

	-- [rseeler 2012-03-12]: See comments in toggleMyFavorite() as they apply here as well.

	local selectedB = self._docStore:fieldValueB( 'asset', photoId, "_localOnly/selected" )

	local resultP = self._docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = photoId,
		overlay = {
			_localOnly = {
				selected = selectedB:valueNow()
					and { _delete = true }
					or true,
			},
		},
	}
	
	selectedB:close()
	return resultP
end


--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a photo in this catalog is currently selected.

	\param photoId (string) The unique asset ID of the photo.
	
	\return <code>B< boolean ></code> A Behavior object whose value is
		\c true if the photo is selected.
]]--

function SelectionMethods:selectedPhotoB( --[[string]] photoId )

	return self._docStore:fieldValueB( 'asset', photoId, "_localOnly/selected" )
		:mapB( function( fieldValue, isExisting )

			return not not fieldValue, isExisting
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the list of all photos in this catalog.
	
	\return <code>P< table ></code> A Promise object that resolves to
		the list of asset IDs.
]]--

function SelectionMethods:getAllPhotosP()

	return self._indexStore:queryP(
					[[SELECT coreInfo.docId FROM coreInfo ]] )
		:mapDataP( function( _, rows )
	
			local extractedRows = {}
			for i = 1, #rows do
				extractedRows[ i ] = rows[ i ][ 1 ]
			end
			return 'data', extractedRows
		
		end )
	
end

--==============================================================================

return {
	SelectionMethods = SelectionMethods,
}
