--[[----------------------------------------------------------------------------

OzCatSyncStatus.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSReactive = require "KSReactive"
local KSJsonEncoder = require "KSJsonEncoder"
local WFWorkerThread = require "WFWorkerThread"
local KSTableUtils = require 'KSTableUtils'

local KSTableUtils_derefOpt = KSTableUtils.derefOpt
local WFCloseSupport = require 'WFCloseSupport'

--==============================================================================

local BabluEP = WFWorkerThread( "BabluImp.lua" )

-- @mark OzCatalogService class

local SyncStatusMethods = {
}

--==============================================================================

local serviceAvailableB = WFCloseSupport.memoizeFn( function( catalogId )

		return BabluEP:requestB {
			message = 'serviceAvailableB',
			arguments = catalogId,
			startingValue = true
		}
		
end )

--------------------------------------------------------------------------------
--[[!
     \brief Reports whether Oz service is available.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
	Oz service is available. If Oz is unavailable, also returns a 503 error
	with the following response to signal unavailability:
	<code>
	while (1) {}	{"code":9999,"description":"Service is temporarily unavailable"}
	</code>
	
	((is this described correctly?))
]]--

function SyncStatusMethods:serviceAvailableB()

	return serviceAvailableB( self._catalogId ):relayB( true )

end

--------------------------------------------------------------------------------

local raw_isUploadingToOzB = WFCloseSupport.memoizeFn( function( self )

	return BabluEP:requestB {
		message = 'isUploadingToOzB',
		arguments = self._catalogId,
		startingValue = false,
	}
	:addAsCloseTargetOf( self )

end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether the sync engine is currently uploading changes to Oz.
	
	This version returns the real, immediate status, which is probably too twitchy
	for most circumstances; compare \c isUploadingToOzB()..

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		uploading is in progress.

]]--

function SyncStatusMethods:raw_isUploadingToOzB()

	return raw_isUploadingToOzB( self ):relayB( true )

end

--------------------------------------------------------------------------------

local isUploadingToOzB = WFCloseSupport.memoizeFn( function( self )

	local rawUploadingB = raw_isUploadingToOzB( self ):relayB( true )

	return rawUploadingB:closingMapSwitchE( function( status )
		if status then
			return KSReactive.timerE( 1 ):constantE( true )
		else
			return KSReactive.timerE( 3 ):constantE( status )
		end
	end ):closingHoldB( rawUploadingB:valueNow() )
	:addAsCloseTargetOf( self )
	
end )

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether the sync engine is currently uploading changes to Oz.
	
	This version has a delay that prevents over-reporting;
	compare \c raw_isUploadingToOzB().

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		uploading is in progress.

]]--

function SyncStatusMethods:isUploadingToOzB()

	return isUploadingToOzB( self ):relayB( true )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether the sync engine is currently pulling changes from Oz.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		downloading is in progress.
]]--

function SyncStatusMethods:isDownloadingFromOzB()

	local BabluEP = WFWorkerThread( "BabluImp.lua" )

	return BabluEP:requestB {
		message = 'isDownloadingFromOzB',
		arguments = self._catalogId,
		startingValue = false
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Provide an estimate of the download progress.
	
	This is calculated based on breadcrumbs left in the catalog during a previous
	sync by WF by this user on this catalog.  If no breadcrumbs are found, then progress
	will be indeterminate.
	
	\return <code>B< progress, scope ></code> Either a progress tuple,
		or false for the indeterminate state.  A progress tuple is a pair of results:
		The first is the progress as a number between 0 and 1, inclusive.  Second is
		the scope of the work, a number that reflects roughly of how many documents are involved.
		For example, after importing one photo, progress might be 0.33 because the asset's
		metadata has been uploaded but not its preview or its original; and the scope would
		be 3 (metadata, preview, original).
]]--

function SyncStatusMethods:downloadingFromOzProgressB()

	local BabluEP = WFWorkerThread( "BabluImp.lua" )

	return BabluEP:requestB {
		message = 'downloadingFromOzProgressB',
		arguments = self._catalogId,
		startingValue = false
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Provide an estimate of the upload progress.
	
	This is calculated based on upload status in the index store.
	
	\return <code>B< progress, scope ></code> Either a progress tuple,
		or false for the indeterminate state.  See downloadingFromOzProgressB for a description
		of a progress tuple return value.
]]--

function SyncStatusMethods:uploadingToOzProgressB()

	local numNeedingUploadB = self._indexStore:queryB(
		[[SELECT count( rowid ) FROM ozUploadStatus WHERE ozUploadStatus != '_conflict']] )
	local totalNumB = self._ozUser:getPendingUploadCountB( self._catalogId )
		
	return KSReactive.dataCallB( function( numNeedingUpload, totalNum )
		numNeedingUpload = KSTableUtils_derefOpt( numNeedingUpload, 1, 1 )
		
		if not numNeedingUpload or not totalNum then
			return false
		end
		if numNeedingUpload == 0 then
			self._ozUser:resetPendingUploadCountP( self._catalogId )
			return 1, 0
		end
		if totalNum == 0 then
			return false
		end
		if numNeedingUpload > totalNum then
			numNeedingUpload = totalNum
		end
		return ( totalNum - numNeedingUpload ) / totalNum, totalNum
	end, numNeedingUploadB, totalNumB )
	:addCloseTarget( numNeedingUploadB )
	:addCloseTarget( totalNumB )

end

--------------------------------------------------------------------------------
--[[
	P< string > = catalog:DEBUG_generatePhotoStatusReportP( photoId )
	
	This is intended for use in debug builds only. Generates a textual representation
	of recent status of this photo. The intended use case is that this text gets
	e-mailed by the app to scouten@adobe.com so I can debug any remaining
	"stuck upload" bugs.
]]--

function SyncStatusMethods:DEBUG_generatePhotoStatusReportP( photoId )

	return self._docStore:docValueOnceP( 'asset', photoId )
		:mapDataP( function( _, docValue )
	
			return BabluEP:requestP( 'DEBUG_statusReport',
									  self._catalogId,
									  photoId )

				:mapDataP( function( _, uploadStatus )
						
					return 'data',
						string.format( "Debug information for photo ID %s\n\ndocValue:\n\n%s\n\nlast upload status report:\n\n%s",
							photoId,
							KSJsonEncoder.encodeOzUploadJson( docValue ),
							uploadStatus )
				
				end )
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether this catalog has ever fully synced with Oz. Note that this
	does \b not signal that the catalog is \b currently synced with Oz,
	but whether it has fully completed the \b first sync with Oz.
		
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		once the catalog has fully synced with Oz, \c false before then.

	\warning Returns a false positive during the brief interval
		before it has had a chance to communicate with the database thread.

--]]

function SyncStatusMethods:hasCompletedInitialSyncB()
	
	return self._docStore:localStateB( 'isDoingInitialSync' )
		:mapB( function( value )
			return not value
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Triggers an immediate sync cycle with Oz.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		when the sync cycle is done
--]]

function SyncStatusMethods:syncNowP()

	return BabluEP:requestP( 'syncNowP', self._catalogId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the stage of upload currently being performed and how many photos
	are left at this stage.
	
	\return <code>B< string, number ></code> A Behavior object with two values,
		the upload stage (currently one of "idle", "photos", "originals", or "conflicts")
		and the number of photos that remain to be uploaded at this stage.
--]]

function SyncStatusMethods:uploadStatusB()

	local groupedQueryB = self._indexStore:queryB(
		[[SELECT uploadStatusWithDeviceId, count( rowid ) FROM ozUploadStatusForView
			GROUP BY ozUploadStatus]] )
	
	local resultB = groupedQueryB:mapB( function( queryResults )
	
		if not queryResults then return 'idle', 0 end
		
		local groups = {}
		for i = 1, #queryResults do
			local row = queryResults[ i ]
			local rowCount = row[ 2 ]
			if type( rowCount ) == 'number' and rowCount > 0 then
				groups[ row[ 1 ] ] = rowCount
			end
		end
		
		-- Choose the "most important" group to return and ignore all others.
		
		if groups[ 'notUploadedNew' ] then
			return 'photos', groups[ 'notUploadedNew' ]
		--elseif groups[ 'notUploadedChanged' ] then
			--return 'photosChanged', groups[ 'notUploadedChanged' ]
		elseif groups[ 'notUploadedStage2-original' ] then
			return 'originals', groups[ 'notUploadedStage2-original' ]
		elseif groups[ '_conflict' ] then
			return 'conflicts', groups[ '_conflict' ]
		else
			return 'idle', 0
		end
	
	end )
	
	resultB:addCloseTarget( groupedQueryB )
	self:addCloseTarget( resultB )

	return resultB

end

--==============================================================================

return {
	SyncStatusMethods = SyncStatusMethods,
}