--[[----------------------------------------------------------------------------

OzCatalogIndexDefinitions.lua

@author scouten 2011-04-07

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.string
local KSStringUtils = require 'KSStringUtils'

	-- wichitafoundation.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- wichitafoundation.core.utils
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'
local WFOzSyncConstants = require 'WFOzSyncConstants'

	-- wichitafoundation.oz_client.index_store
local WFIndexStoreImp = require 'WFIndexStoreImp'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'

local KSDate = require 'KSDate'

local WFOzContentTypes = require 'WFOzContentTypes'

--------------------------------------------------------------------------------

local uploadStatusLog = KSLogging.log 'OzCatalogIndexDefinitions::uploadStatus'
--	:setLevel( KSLogging.DEBUG )

local missingOriginalsLog = KSLogging.log 'OzCatalogIndexDefinitions::missingOriginals'
--	:setLevel( KSLogging.DEBUG )

local cacheReferencesLog = KSLogging.log 'OzCatalogIndexDefinitions::cacheReferences'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )


--============================================================================--

local indexStoreSchema = WFIndexStoreSchema()

local uploadOrderForDocType = WFOzSyncConstants.uploadOrderForDocType

--------------------------------------------------------------------------------

local function assetHasThumbnail( docValue )

	-- Per discussion on bug #2942278, this filter should be applied to all
	-- indexes that drive user-visible content. (Note, for instance, that we
	-- exclude it from the debug-only upload status filter.)

	-- Exclude documents other than assets.

	if docValue._docType ~= 'asset' then return false end
	
	-- Include if we have a local thumbnail.
	
	if derefOpt( docValue, '_localOnly', 'thumbnail', 'md5' ) then
		return true
	end
	
	-- Include if we can get a thumbnail from Oz.
	
	if derefOpt( docValue, '_serverProvidedMetadata', 'links',
				 '/rels/rendition_type/144h', 'href' )
	then
		return true
	end
	
	-- No thumbnail? Pretend it doesn't exist yet.
	
	return false

end

--------------------------------------------------------------------------------

local function coreMD5FromDocValue( docValue )
	
	local renditions = { 'thumbnail', 'thumbnail2x', 'preview', 'smallPreview', 'master' }
	local results = {}
	local localOnly = docValue._localOnly
	
	for index, rendition in ipairs( renditions ) do
		local thumbnail = localOnly and localOnly[ rendition ]
		local md5 = thumbnail and thumbnail.md5
		results[ index ] = md5
	end
	
	return unpack ( results, 1, 5 )

end

--------------------------------------------------------------------------------

local function coreInfoFromDocValue( docValue )

	local width = docValue.width
	local height = docValue.height
	
	-- Cropped width and height win.
	
	local develop = docValue.develop
	
	if develop and develop.croppedWidth and develop.croppedHeight then
		width = develop.croppedWidth
		height = develop.croppedHeight
	end
	
	-- Local only thumbnail wins.
	-- No, because this causes a lot of extra computation as large
	-- grid view results are re-sent when one thumbnail is loaded
	-- arahn 05/25/2013
	local localOnly = docValue._localOnly
	local thumbnail = localOnly and localOnly.thumbnail
	
	if width == nil and height == nil and thumbnail and thumbnail.referenceWidth and thumbnail.referenceHeight then
		width = thumbnail.referenceWidth
		height = thumbnail.referenceHeight
	end

	local aspectRatio

	if docValue._docType == 'asset'
		and not docValue._deleted
		and type( width ) == 'number' and width > 0
		and type( height ) == 'number' and height > 0
	then
		aspectRatio = width / height
	end
	
	local thumbnailMD5 = thumbnail and thumbnail.md5
	
	-- FIXME [ngautam 2011-12-13]: Need to replace string.lower by
	-- localization-friendly string conversion.

	local favoritedCount = 0
	local favorites = docValue and docValue.favorites

	if favorites then
		for k, v in pairs( favorites ) do
			if v then
				favoritedCount = favoritedCount + 1
			end
		end
	end

	return aspectRatio, thumbnailMD5, favoritedCount

end

--------------------------------------------------------------------------------

local function assetHasOriginal( docValue )

	if docValue
		and docValue._serverProvidedMetadata
		and docValue._serverProvidedMetadata.links
		and docValue._serverProvidedMetadata.links[ "/rels/master" ]
		and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
		then
			return true
		end
	
	return false
end

--------------------------------------------------------------------------------

local function Tornado_coreInfoFromDocValue( docValue )

	local width = docValue.width
	local height = docValue.height
	-- Cropped width and height win
	
	local develop = docValue.develop
	
	if develop and develop.croppedWidth and develop.croppedHeight then
		width = develop.croppedWidth
		height = develop.croppedHeight
	end
	
	-- Local only thumbnail wins
	
	local localOnly = docValue._localOnly
	local thumbnail = localOnly and localOnly.thumbnail
	
	if thumbnail and thumbnail.referenceWidth and thumbnail.referenceHeight then
		width = thumbnail.referenceWidth
		height = thumbnail.referenceHeight
	end

	local aspectRatio

	if docValue._docType == 'asset'
		and not docValue._deleted
		and type( width ) == 'number' and width > 0
		and type( height ) == 'number' and height > 0
	then
		aspectRatio = width / height
	end
		
	local thumbnailMD5 = thumbnail and thumbnail.md5

	-- FIXME [ngautam 2011-12-13] Need to replace string.lower by localization-friendly string conversion

    local favoritedCount = 0
    local favoritedUsers = ""
    if docValue and docValue.favorites then
        for k in pairs(docValue.favorites) do
            favoritedCount = favoritedCount + 1
			if favoritedUsers ~= "" then
				favoritedUsers = favoritedUsers .. ','
			end
			favoritedUsers = favoritedUsers .. k
        end
    end
    
	local importedBy = ""
    if docValue and docValue.importSource then
        importedBy = docValue.importSource.importedBy
    end

	local hasOriginal = assetHasOriginal( docValue )

	return aspectRatio, thumbnailMD5, favoritedCount, importedBy, favoritedUsers, hasOriginal

end

--------------------------------------------------------------------------------

local function convertToUTCTime( dateTime )
	
	-- three cases to handle here
	-- 1. time has time zone info - convert time zone time to UTC time
	-- YYYY-MM-DDTHH:MM:SS-HH:MM
	
	-- 2. time is in UTC time, no conversion necessary
	-- YYYY-MM-DDTHH:MM:SSZ
	
	-- 3. time has no time zone and not in UTC time, use current time zone to
	-- convert to UTC time
	-- YYYY-MM-DDTHH:MM:SS
	
	local year = string.sub( dateTime, 1, 4 )
	local month = string.sub( dateTime, 6, 7 )
	local day = string.sub( dateTime, 9, 10 )
	
	local hours = string.sub( dateTime, 12, 13 )
	
	local minutes = string.sub( dateTime, 15, 16 )
	local seconds = string.sub( dateTime, 18, 19 )
	
	local timeInSeconds = os.time{ year=year, month=month, day=day, hour=hours, min=minutes, sec=seconds }

	if timeInSeconds == nil then
		--[[ 
			mscarafo 3/13/13
			This means the input capture date was 00-00-00-00... All zero's
			From the oz api's:
			Capture datetime, if known. Unknown capture datetime is represented as "0000-00-00T00:00:00"
		--]]
		return dateTime
	end
	
	-- get the time portion of the string	
	local time = string.sub( dateTime, 12 )

	if string.match( string.sub( time, 1, 14 ), "%d%d:%d%d:%d%d[%-%+]%d%d:%d%d" ) then
		local offset = string.sub( time, 9, 9 )
		local tzHours = tonumber( string.sub( time, 10, 11 ) )
		local tzMinutes = tonumber( string.sub( time, 13, 14 ) )
		
		-- convert hours to seconds
		tzHours = tzHours * 3600
		
		-- convert minutes to seconds
		tzMinutes = tzMinutes * 60
		
		-- convert to UTC time in seconds
		if offset == '+' then
			timeInSeconds = timeInSeconds - tzHours - tzMinutes
		else
			timeInSeconds = timeInSeconds + tzHours + tzMinutes
		end
		
    elseif #dateTime >= 19 and string.sub( dateTime, 20 ) ~= 'Z' then
		
		-- get the local time zone
		-- +/-HHMM
		local timeZoneString = KSDate.timeZoneInIso8601Format( )
		
		local offset = string.sub( timeZoneString, 1, 1 )
		local tzHours = tonumber( string.sub( timeZoneString, 2, 3 ) )
		local tzMinutes = tonumber( string.sub( timeZoneString, 4, 5 ) )
		
		-- convert hours to seconds
		tzHours = tzHours * 3600
		
		-- convert minutes to seconds
		tzMinutes = tzMinutes * 60
		
		-- convert to UTC time in seconds
		if offset == '+' then
			timeInSeconds = timeInSeconds - tzHours - tzMinutes
		else
			timeInSeconds = timeInSeconds + tzHours + tzMinutes
		end

	end
	
	return timeInSeconds
end

--------------------------------------------------------------------------------
-- @mark coreInfo
--------------------------------------------------------------------------------

-- FIXME [rseeler 2013-05-03]:  There needs to be a way to create application specific tables...

if WFPlatformInfo.isTornado then

indexStoreSchema:addTable {


	sql = [[CREATE TABLE coreInfo( aspectRatio, captureDate, displayDate, timestamp, copyTimestamp, eventName, eventNameLC, favoritedCount, importedBy, favoritedUsers, hasOriginal, contentTypeNumber )]],

	--docType = 'asset',
		-- FIXME [scouten 2011-03-29]: Currently ignored by update system.

	--version = 1,
		-- FIXME [scouten 2011-03-29]: Currently ignored by update system.
		-- Will use this to cause the table to be rebuilt when functions
		-- called by map() are changed in a materially significant fashion.

	map = function( docValue, emit )

		if not assetHasThumbnail( docValue ) then return end

		local aspectRatio, thumbnailMD5, favoritedCount, importedBy, favoritedUsers, hasOriginal = Tornado_coreInfoFromDocValue( docValue )

		local displayDate

		-- Capture date may have a "Z" appended to the end, signifying that the timestamp is in UTC (20 chars).
		-- Otherwise, the timestamp is treated as local time (19 chars).
		-- Either way, we just care about getting the YYYY-MM-DD from the beginning of the string.
		
		if docValue.captureDate and string.len( docValue.captureDate ) > 10 then
			displayDate = string.gsub( string.sub( docValue.captureDate, 1, 10 ), ":", "-" )
		end

		local timestamp = docValue
			and docValue.importSource
			and docValue.importSource.importTimestamp

		local copyTimestamp = docValue
			and docValue.importSource
			and docValue.importSource.copyTimestamp
			
		local eventName = docValue
			and docValue.eventName or ""
			
		local contentTypeNumber = 0
				
		local contentType = derefOpt( docValue, 'importSource', 'contentType' )
			or derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'content_type' )
		if contentType then
			contentTypeNumber = WFOzContentTypes.getContentTypeNumberForContentType( contentType )
		else
			local subtype = derefOpt( docValue, '_serverProvidedMetadata', 'subtype' )
			contentTypeNumber = WFOzContentTypes.getContentTypeNumberRangeForSubType( subtype )
		end
		
		if aspectRatio then
			emit( aspectRatio, docValue.captureDate, displayDate, timestamp, copyTimestamp, eventName, KSStringUtils.lower( eventName ), favoritedCount, importedBy, favoritedUsers, hasOriginal, contentTypeNumber )
		end
			
	end,
	
}

else -- !isTornado

indexStoreSchema:addTable {
	
	-- Instead of specifying CREATE TABLE SQL, in order to use
	-- niceToHave, you must specify the field names and let the index
	-- store generate the SQL
	tableName = "coreInfo",
	
	fields = { 
		"aspectRatio", "captureDate", "displayDate", "timestamp",
		"copyTimestamp", "thumbnailMD5", "eventName", "eventNameLC",
		"favoritedCount", "importedBy" },
	
	-- queryB won't send updates for changes that only affect niceToHave
	-- columns
	niceToHave = {
		"thumbnailMD5",
	},

	map = function( docValue, emit )

		if not assetHasThumbnail( docValue ) then return end

		local aspectRatio, thumbnailMD5, favoritedCount = coreInfoFromDocValue( docValue )

		local displayDate
		
		-- Capture date may have a "Z" appended to the end, signifying that the
		-- timestamp is in UTC (20 chars). Otherwise, the timestamp is treated as
		-- local time (19 chars). Either way, we just care about getting the
		-- YYYY-MM-DD from the beginning of the string.

		if docValue.captureDate and string.len( docValue.captureDate ) > 10 then
			displayDate = string.gsub( string.sub( docValue.captureDate, 1, 10 ), ":", "-" )
		end

		local timestamp = derefOpt( docValue, 'importSource', 'importTimestamp' )

		local copyTimestamp = derefOpt( docValue, 'importSource', 'copyTimestamp' )
			
		local eventName = docValue.eventName or ''
		
		local importedBy = derefOpt( docValue, 'importSource', 'importedBy' ) or ''
			
		if aspectRatio then
			emit( aspectRatio, docValue.captureDate, displayDate, timestamp,
				  copyTimestamp, thumbnailMD5, eventName, KSStringUtils.lower( eventName ),
				  favoritedCount, importedBy )
		end
			
	end,

}

end -- !isTornado

-- We do many extractions based on events plus a date range. This index covers
-- those cases.

indexStoreSchema:addIndex
	[[CREATE INDEX coreInfo_eventDateTrackList ON coreInfo( eventNameLC, displayDate )]]
	
-- We seed the date and event views based on extracting display date, event name,
-- and capture date. This index covers all of those elements. We shouldn't actually
-- need the lowercase event name since it is derivable but we include it here so
-- that we can ensure consistency.

indexStoreSchema:addIndex
	[[CREATE INDEX coreInfo_eventDateTrackContents ON coreInfo( displayDate, eventName,
																eventNameLC, captureDate )]]

-- New index for imported by index.

indexStoreSchema:addIndex
	[[CREATE INDEX coreInfo_importedBy ON coreInfo( importedBy, captureDate, docId )]]

--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE coreMD5( thumbnail, thumbnail2x, preview, smallPreview, master )]],

	map = function( docValue, emit )

		if assetHasThumbnail( docValue ) then
	
			local thumbnail, thumbnail2x, preview, smallPreview, master = coreMD5FromDocValue( docValue )
			
			if thumbnail or thumbnail2x or preview or smallPreview or master then
				emit( thumbnail, thumbnail2x, preview, smallPreview, master )
			end
			
		end
			
	end,

}

--------------------------------------------------------------------------------
-- @mark selectedPhotos
--------------------------------------------------------------------------------
-- @mark selectedPhotos
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE selectedPhotos( isSelected )]],

	map = function( docValue, emit )

		local isSelected = docValue
			and docValue._localOnly
			and docValue._localOnly.selected
			and not docValue._deleted

		if isSelected then
			emit( true )
		end
			
	end,

}

--------------------------------------------------------------------------------
-- @mark keywords
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE keywords( keyword, lcKeyword )]],

	map = function( docValue, emit )

		if not assetHasThumbnail( docValue ) then return end

		if docValue._docType == 'asset' and not docValue._deleted then

			local keywords = docValue.keywords
			
			if keywords and next( keywords ) then
			
				for keywordName in pairs( keywords ) do
					emit( keywordName, string.lower( keywordName ) )
				end
			
			else
				
				emit( "(no keywords)", "(no keywords)" )
			
			end
		
		end
	
	end,

}

-- We need an index to find the photos for a keyword. This index may also help
-- with extracting unique keywords since it is sorted conveniently.

indexStoreSchema:addIndex
	[[CREATE INDEX keywords_index ON keywords( lcKeyword, keyword )]]

--------------------------------------------------------------------------------
-- @mark ozUploadStatus
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	-- Warning [ arahn 11/19/2012 ]: this table definition must match the one defined by OzUserInfoCacheIndexDefinitions
	-- because some queries operate on both indexes, for example processUploadQueueP()'s query
	-- in WFOzSyncCycle.
	-- If you add fields to this table, be sure to update setUpIndex() fn in the WFOzSyncCycleTest unit test
	sql = [[CREATE TABLE ozUploadStatus( ozUploadStatus, docType, captureDate, albumId, assetId, uploadOrder NOT NULL, changeCounter, freemiumUploadEmbargoUntil, uploadWave )]],

	map = function( docValue, emit )

		-- [scouten 2011-08-14]: If this feature becomes user-visible
		-- (disregarding the current debug-only feature), then we should create
		-- a second index with the assetHasThumbnail filter applied. This filter
		-- must remain as-is because its primary purpose is to drive the upload
		-- process.

		uploadStatusLog:debug( 'ozUploadStatus reindexing %s', docValue )

		-- TODO [arahn 2012-10-11]: This list should be stored someplace so that
		-- we can handle new types of "generic" documents without needing
		-- to make changes in so many places.

		if docValue._docType == 'asset'
			or docValue._docType == 'receipt'
				-- FIXME [ arahn 2013-01-29 ] I believe that receipt documents do not get added to this
				-- doc store. Receipts are stored AFAIK in the "OzUserInfoCacheIndexDefinitions".
				-- This "receipt" case should be removed from this.
			or docValue._docType == 'album'
			or docValue._docType == 'album_asset'
			or docValue._docType == 'space'
			or docValue._docType == 'comment'
		then
			
			local result = OzCatalogUploadStatus.calculateUploadStatus( docValue )

			if result == 'uploaded' then return end
			
			uploadStatusLog:debug( "docValue = %s\n\nuploadStatus = %s\n\n",
								   docValue, result )
			
			local albumId = ( docValue.album and docValue.album._id )
				or docValue._localOnly and docValue._localOnly.album and docValue._localOnly.album._id
				or nil

			local assetId = ( docValue.asset and docValue.asset._id )
				or docValue._localOnly and docValue._localOnly.asset and docValue._localOnly.asset._id
				or nil
			
			local changeCounter = docValue._changeCounter
			
			local uploadWave = ( docValue and docValue._localOnly and docValue._localOnly.uploadWave ) or 1
			
			local uploadOrder = uploadOrderForDocType[ docValue._docType ]
			
			if docValue._docType == 'asset' then
				if result == 'notUploaded' then
					uploadOrder = uploadOrderForDocType.asset_preview_pixels
				elseif result == 'notUploadedStage2-original' then
					uploadOrder = uploadOrderForDocType.asset_full_pixels
				end
			end
			
			emit( result, docValue._docType, docValue.captureDate, albumId,
				  assetId, uploadOrder or uploadOrderForDocType.default_order, changeCounter or nil,
				  docValue._localOnly and docValue._localOnly.freemiumUploadEmbargoUntil, uploadWave )
		
		end

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX ozUploadStatus_status ON ozUploadStatus(
					ozUploadStatus, captureDate, validRev, uploadOrder, changeCounter ASC, freemiumUploadEmbargoUntil )]]

indexStoreSchema:addIndex
	[[CREATE INDEX ozUploadStatus_docType_status ON ozUploadStatus(
					docType, ozUploadStatus, captureDate, validRev, uploadOrder, changeCounter ASC, freemiumUploadEmbargoUntil )]]
					
indexStoreSchema:addIndex
	[[CREATE INDEX ozUploadStatus_ozUploadStatus_docType_freemi_uploadOrder ON ozUploadStatus( uploadOrder, freemiumUploadEmbargoUntil, docType, ozUploadStatus )]]

--------------------------------------------------------------------------------
-- @mark ozUploadStatusForView
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE ozUploadStatusForView( ozUploadStatus, docType, captureDate, deviceName, uploadStatusWithDeviceId, aspectRatio, thumbnailMD5, favoritedCount )]],

	map = function( docValue, emit )

		uploadStatusLog:debug( 'ozUploadStatusForView reindexing %s' )

		if docValue._docType == 'asset' and assetHasThumbnail( docValue ) and not docValue._deleted then
			
			local aspectRatio, thumbnailMD5, favoritedCount = coreInfoFromDocValue( docValue )

			local result = OzCatalogUploadStatus.calculateUploadStatus( docValue )
			
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
				
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder

			local ozOriginalUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href

			local ozThumbnailUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/144h" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/144h" ].href
				
			local ozThumbnail2xUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/thumbnail2x" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/thumbnail2x" ].href

			local ozPreviewUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/1024" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/1024" ].href
			
			-- Is this a new photo not uploaded or a changed photo?

			if WFOzSyncConstants.isNotUploaded( result ) then

				-- If we don't have any renditions uploaded, it is a new photo.

				if ozOriginalUrl and ozThumbnailUrl and ozThumbnail2xUrl and ozPreviewUrl then
					result = 'notUploadedChanged'
				else
					result = 'notUploadedNew'
				end

			end

			local deviceName = ''

			-- If the status is uploaded, check originals.

			if result == 'uploaded' then
	
				if not ozOriginalUrl then
					
					if docValue.importSource and docValue.importSource.originalStatus == 'lost' then
						result = 'originalLost'
					else
						result = 'uploadingOtherDevice'
						deviceName = ( docValue and docValue.importSource and docValue.importSource.importedOnDevice ) or 'unknown device'
					end

				else

					-- We need to check for originals (base and/or latest) that are either not present locally
					-- or present in caches so that the user can see them in the 'Downloading Originals' status
					-- view and know that these are not in robust storage yet.

					if OzCatalogUploadStatus.assetOriginalIsValid( docValue ) then
						if not localPath or ( localPath and isInCaches ) then
							result = 'downloadingOriginal'
						end
					else
						if ( not localPath_latestOriginal or ( localPath_latestOriginal and isInCaches_latestOriginal ) ) or
						( not localPath or ( localPath and isInCaches ) ) then
							result = 'downloadingOriginal'
						end
					end

				end

			end

			uploadStatusLog:debug( "docValue = %s\n\nuploadStatus = %s\n\n",
								   docValue, result )

			-- Create a unique id for this row, status + device name.

			local uploadStatusWithDeviceId = result..deviceName
			
			if result ~= 'uploaded' and result ~= 'deleted' then
				emit( result, docValue._docType, docValue.captureDate, deviceName, uploadStatusWithDeviceId, aspectRatio, thumbnailMD5, favoritedCount )
			end
		
		end

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX ozUploadStatusForView_trackContents ON ozUploadStatusForView(
					uploadStatusWithDeviceId, captureDate DESC, docId )]]

--------------------------------------------------------------------------------
-- @mark importSource
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE importSource( sha256, localUrl )]],

	map = function( docValue, emit )

		if docValue._docType == 'asset' and not docValue._deleted then
		
			local sha256 = docValue.importSource
								and docValue.importSource.sha256
			
			local localUrl = docValue._localOnly
								and docValue._localOnly.localUrl
			
			if sha256 or localUrl then
				emit( sha256, localUrl )
			end
		
		end

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX importSource_sha256 ON importSource( sha256 )]]

indexStoreSchema:addIndex
	[[CREATE INDEX importSource_localUrl ON importSource( localUrl )]]

--------------------------------------------------------------------------------
-- @mark autoImportLog
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE autoImportLog( sha256, localUrl )]],

	map = function( docValue, emit )

		if docValue._docType == 'autoImport' and not docValue._deleted then
		
			local sha256 = docValue._localOnly
					   and docValue._localOnly.sha256
			
			local localUrl = docValue._localOnly
					     and docValue._localOnly.localUrl
			
			if sha256 or localUrl then
				emit( sha256, localUrl )
			end
		
		end

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX autoImportLog_sha256 ON autoImportLog( sha256 )]]

indexStoreSchema:addIndex
	[[CREATE INDEX autoImportLog_localUrl ON autoImportLog( localUrl )]]

--------------------------------------------------------------------------------
--[[
	These constants are only used when existing photos (imported on 2011-08-12
	and before are in the system and we didn't capture the rendition size.
]]--

local defaultSizeForRendition = {
	original = 20000000,
	preview = 150000,
	thumbnail = 12000,
	thumbnail2x = 48000,
}

--------------------------------------------------------------------------------
-- @mark cacheReferences
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE cacheReferences( md5, renditionName, size, epoch, hasUploaded )]],

	map = function( docValue, emit )

		local localOnly = docValue._localOnly
		
		local serverLinks =
			  ( docValue._serverProvidedMetadata
			and docValue._serverProvidedMetadata.links ) or {}

		local function emitForRenditionName( renditionName, ozNameForRendition )
		
			local renditionInfo = localOnly and localOnly[ renditionName ]
			if not renditionInfo then return end
			
			-- we aren't making destructive edits to originals yet, so we do not upload
			-- originals other than the base original
			
			local renditionKey
			
			if renditionName == 'latestOriginal' then
				renditionKey = OzCatalogUploadStatus.checkAndCorrectLocalPath(
								localOnly.latestOriginal
									and localOnly.latestOriginal.filePath )
			elseif renditionName == 'original' then
				-- Temporary check to correct the local file path and replace
				-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
				-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
				-- that updates the database with the correct file paths but this is not
				-- feasible in the current time frame.
				renditionKey = OzCatalogUploadStatus.checkAndCorrectLocalPath( localOnly.originalFilePath )
			else
				renditionKey = renditionInfo.md5
			end

			if renditionInfo and renditionKey then
			
				local serverLinksForRendition = serverLinks[ ozNameForRendition ]
				local shouldUpload
				
				if renditionName ~= 'latestOriginal' then
					shouldUpload = ( not serverLinksForRendition )
						or serverLinksForRendition.invalid
						or renditionInfo.revId == 'pending'
				else
					-- We don't upload the latest original
					shouldUpload = false
				end
					
			
				emit( renditionKey,
					  renditionName,
					  renditionInfo.size or defaultSizeForRendition[ renditionName ] or 1000000,
					  renditionInfo.epoch or 1,
					  shouldUpload and 0 or 1 )
			
				cacheReferencesLog:debug( "photoId = %s | renditionName = %s | size = %s | epoch  %s | hasUploaded = %s",
					docValue._docId, renditionName, renditionInfo.size, renditionInfo.epoch or 1,
					not shouldUpload )
			
			end
		
		end

		emitForRenditionName( 'thumbnail', '/rels/rendition_type/144h' )
		emitForRenditionName( 'thumbnail2x', '/rels/rendition_type/thumbnail2x' )
		emitForRenditionName( 'preview', '/rels/rendition_type/1024' )
		emitForRenditionName( 'original', '/rels/master' )
		emitForRenditionName( 'latestOriginal', '' )

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX cacheReferences_epoch ON cacheReferences( epoch DESC, size )]]

indexStoreSchema:addIndex
	[[CREATE INDEX cacheReferences_md5 ON cacheReferences( md5 ASC, docId )]]

indexStoreSchema:addIndex
	[[CREATE INDEX cacheReferences_potentialVictims
		ON cacheReferences( epoch ASC, size DESC, docId )]]

--------------------------------------------------------------------------------
-- @mark nonOriginalsCacheReferences
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE nonOriginalsCacheReferences( md5 )]],

	map = function( docValue, emit )
		
		if docValue._deleted or docValue._docType ~= 'asset' then
			return
		end
		
		local localOnly = docValue._localOnly
	
		local function emitForRenditionName( renditionName )
		
			local renditionInfo = localOnly and localOnly[ renditionName ]
			if not renditionInfo then return end
			
			local md5 = renditionInfo.md5

			if md5 then
				emit( md5 )
			end
		
		end

		emitForRenditionName( 'thumbnail' )
		emitForRenditionName( 'thumbnail2x' )
		emitForRenditionName( 'preview' )

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX nonOriginalsCacheReferences_md5 ON nonOriginalsCacheReferences( md5 )]]

--------------------------------------------------------------------------------
-- @mark missingOriginals
--------------------------------------------------------------------------------

indexStoreSchema:addTable {
	
	sql = [[CREATE TABLE missingOriginals( photoId, captureDate, missingInOz )]],
	
	map = function( docValue, emit )
	
		if docValue._docType == 'asset' and not docValue._deleted then
	
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
								
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local ozUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
				
			local ozUrl_firstMaster = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ].href
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder
				
			-- If we've tried to fetch this original and the request returns with a 404, mark this asset as missing its original in Oz
			-- and place it lower in the queue.
			
			local missingInOz_original = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.missingInOz and 1 or 0
				
			local missingInOz_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.missingInOz and 1 or 0
				
			-- In older documents, 'isInCacheFolder' can be missing.
			-- In that case if there is a local path, the original must be in the documents folder.
			
			-- If there are server urls for the original and (possibly) the latest original
			-- and no local paths or if these are in caches, then these are "missing originals."

			if localPath then
				missingOriginalsLog:debug( "scanning %s", docValue )
			end
			
			if OzCatalogUploadStatus.assetOriginalIsValid( docValue ) then
			
				-- If there have been no destructive edits and the original has not changed
				-- check if we have the original.
				
				if ozUrl and ( not localPath or isInCaches ) then
					missingOriginalsLog:debug( "flagged %s for download because original missing", docValue._docId )
					emit( docValue._docId, docValue.captureDate, missingInOz_original )
				end
				
			else
			
				-- If destructive edits have been made, check if we need either the base original
				-- or the latest original.
				
				if ( ozUrl and ( not localPath_latestOriginal or isInCaches_latestOriginal ) )
					or ( ozUrl_firstMaster and ( not localPath or isInCaches ) )
				then
					missingOriginalsLog:debug( "flagged %s for download because original out of date", docValue._docId )
					emit( docValue._docId, docValue.captureDate, missingInOz_latestOriginal )
				end
				
			end
	
		end
	
	end,
	
}

indexStoreSchema:addIndex
	[[CREATE INDEX missingOriginals_index ON missingOriginals( photoId, missingInOz )]]

--------------------------------------------------------------------------------
-- @mark currentOriginalsInDocuments
--------------------------------------------------------------------------------

-- This index rolls up assets for which we have a current ( for some definition of current )
-- base original or latest original already downloaded and present in the documents folder.
-- This is used by WFOzPurgeAgent to move these originals from Documents to Caches if
-- 'Store Originals' is off

indexStoreSchema:addTable {
	
	sql = [[CREATE TABLE currentOriginalsInDocuments( photoId, originalFilePath, latestOriginalFilePath )]],
	
	map = function( docValue, emit )
	
		if docValue._docType == 'asset' and not docValue._deleted then
	
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
				
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder
				
			local ozUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
	
			-- Add a check to see if this original is valid ( current )
			-- If isInCaches is nil, that means this is an older document and
			-- we assume that the original is present in Documents
			
			local originalFilePath = nil
			local latestOriginalFilePath = nil
			
			if ozUrl and localPath_latestOriginal and OzCatalogUploadStatus.assetLatestOriginalIsValid( docValue ) and not isInCaches_latestOriginal then
				latestOriginalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath_latestOriginal )
			end
			-- We check for the master url and the first revision url here to ensure that the upload
			-- of this original has completed before moving it to Caches
			
			if ozUrl and localPath and not isInCaches then
				-- Temporary check to correct the local file path and replace
				-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
				-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
				-- that updates the database with the correct file paths but this is not
				-- feasible in the current time frame.
				originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath )
			end
			if latestOriginalFilePath or originalFilePath then
				emit( docValue._docId, originalFilePath, latestOriginalFilePath )
			end
		end
	
	end,
	
}

indexStoreSchema:addIndex
	[[CREATE INDEX currentOriginalsInDocuments_index ON currentOriginalsInDocuments( photoId )]]

--------------------------------------------------------------------------------
-- @mark needsDoNotBackUpFlag
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE needsDoNotBackUpFlag( path )]],
	
	map = function( docValue, emit )
	
		-- This code should only need to run against first originals since we've
		-- covered all of the cases for latest original. (@scouten)

		-- Short-circuit some easy cases.
	
		if ( not WFPlatformInfo.isiOS and not WFPlatformInfo.isAndroid )
			or docValue._docType ~= 'asset'
			or not docValue._localOnly
			or not docValue._serverProvidedMetadata
			or not docValue._serverProvidedMetadata.links
		then
			return
		end
		
		-- Do we need to cancel back up for first original?

		if
				docValue._serverProvidedMetadata.links[ '/rels/master' ]
			and docValue._serverProvidedMetadata.links[ '/rels/master' ].href
			and docValue._localOnly.originalFilePath
			and not ( docValue._localOnly.original
				  and docValue._localOnly.original.doNotBackUpFlagSet )
		then
			-- Temporary check to correct the local file path and replace
			-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
			-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
			-- that updates the database with the correct file paths but this is not
			-- feasible in the current time frame.
			emit( OzCatalogUploadStatus.checkAndCorrectLocalPath( docValue._localOnly.originalFilePath ) )
		end
		
	end,

}

--------------------------------------------------------------------------------
-- @mark albums
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE albums( id, name, nameLC, updateTimeStamp, localTimeStamp )]],

	map = function( docValue, emit )

		if docValue and docValue._docType == 'album' and not docValue._deleted then
		
			local name = docValue.name
			
			local id = docValue._docId
			local updateTimeStamp = docValue.updated
			local localTimeStamp = docValue._localOnly and docValue._localOnly.updateTimeStamp
			
			if id and name then
				emit( id, name, KSStringUtils.lower( name ), updateTimeStamp, localTimeStamp )
			end
		
		end

	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX albums_name ON albums( nameLC )]]

--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE album_asset( assetId, userId, albumId, updateTimeStamp )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted then
		
			if docValue._docType == 'album_asset' then
				local albumId = docValue.album and docValue.album._id
				local assetId = docValue.asset and docValue.asset._id
				local updateTimeStamp = docValue.updated
				local userId = docValue.added_by
				if userId and string.len( userId ) > 32 then
					userId = userId:sub( -32 )
				end
				if albumId and assetId then
				
					emit( assetId, userId, albumId, updateTimeStamp )
				
				end
				
			end
			
			-- Support for virtual albums of favorited assets.

			if docValue._docType == 'asset' and not docValue._deleted 
				and docValue.favorites and next( docValue.favorites ) ~= nil then
				
					emit( docValue._docId, nil, 'favorites', nil )
					
					for k in pairs( docValue.favorites ) do
						emit( docValue._docId, nil, 'favorites' .. k, nil )
					end
					
			end
		end
	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX album_asset_index ON album_asset( albumId, assetId )]]

--------------------------------------------------------------------------------
-- @mark Sharing
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE space_album( spaceId, albumId, updateTimeStamp )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted and docValue._docType == 'space_album' then
			
			local albumId = docValue.album and docValue.album._id
			local spaceId = docValue.space and docValue.space._id
			local updateTimeStamp = docValue.updated
			
			if spaceId and albumId then
				
				emit( spaceId, albumId, updateTimeStamp )
				
			end
				
		end
					
	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX space_album_index ON space_album( spaceId, albumId )]]
	
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE space_user( spaceId, userId, userEmail )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted and docValue._docType == 'space_user' then
			
			-- All users will have an associated email, users in the Revel system
			-- will have an ID as well.
				
			local spaceId = docValue.space and docValue.space._id
			local userId = docValue.user and docValue.user._id
			local userEmail = docValue.user and docValue.user.email
			
			if spaceId and userEmail then
				
				emit( spaceId, userId, userEmail )
			
			end
				
		end
					
	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX space_user_index ON space_user( spaceId, userId, userEmail )]]

--------------------------------------------------------------------------------
-- @mark Comments
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE comment( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted and docValue._docType == 'comment' then
			
			local localOnly = docValue._localOnly or {}
			local spaceId = ( docValue.space and docValue.space._id ) or ( localOnly.spaceId )
			local assetId = ( docValue.asset and docValue.asset._id ) or ( localOnly.assetId )
			local created = docValue.created or ( localOnly.created )
				-- check _localOnly.created, to allow indexing of comments that we've created locally
				-- but haven't yet gotten synced back from Oz

			local body = docValue.body
			local authorInfo = docValue.author or ( localOnly.author )
			local author_first_name, author_last_name, author_full_name, author_id
			if authorInfo then
				author_first_name, author_last_name, author_full_name, author_id =
					authorInfo.first_name,
					authorInfo.last_name,
					authorInfo.full_name,
					authorInfo._id
			
			end
			
			if assetId and created then
				
				if not spaceId then
					
					spaceId = WFOzSyncConstants.kGlobalSpace
				
				end
				
				emit( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id )
				
			end
				
		end
					
	end,

}

indexStoreSchema:addIndex
	[[CREATE INDEX comment_asset_space ON comment( assetId, spaceId )]]

--------------------------------------------------------------------------------
-- @mark Activity
--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE favorite_activity( userId, activityDate, importedBy )]],

	map = function( docValue, emit )
		if docValue and not docValue._deleted and docValue._docType == 'asset' and assetHasThumbnail( docValue ) then
			local importedBy = derefOpt( docValue, 'importSource', 'importedBy' )
			for userId, favoriteDate in pairs( docValue.favorites or {} ) do
				if type( favoriteDate ) == 'string' then
					emit( userId, favoriteDate, importedBy )
				end
			end
		end
	end
}

--------------------------------------------------------------------------------

indexStoreSchema:addTable {

	sql = [[CREATE TABLE new_asset_activity( assetId, userId, activityDate )]],

	map = function( docValue, emit )
		if docValue and not docValue._deleted and docValue._docType == 'asset' and assetHasThumbnail( docValue ) then
			local importedBy = derefOpt( docValue, 'importSource', 'importedBy' )
			local importTimestamp = derefOpt( docValue, '_serverProvidedMetadata', 'created' )
			if importedBy and importTimestamp then
				emit( docValue._id, importedBy, importTimestamp )
			end
		end
	end
}

--------------------------------------------------------------------------------

indexStoreSchema:addIndex
	[[CREATE INDEX favorite_activity_date_userId ON favorite_activity( activityDate DESC, userId )]]

indexStoreSchema:addIndex
	[[CREATE INDEX new_asset_activity_date_userId ON new_asset_activity( activityDate DESC, userId )]]

indexStoreSchema:addIndex
	[[CREATE INDEX new_comment_activity_date_userId ON comment( created DESC, author_id )]]
--	[[CREATE INDEX comment_asset_space ON comment( assetId, spaceId )]]

--------------------------------------------------------------------------------

local initMessageProcessor = function( ... )
	WFIndexStoreImp.start( indexStoreSchema, ... )
end

--==============================================================================

return {
	test_indexStoreSchema = indexStoreSchema,
	initMessageProcessor = initMessageProcessor,
}

