--[[----------------------------------------------------------------------------

OzCatalogService.lua

@author scouten 2011-02-07

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.oz_client.catalog
local OzCatAbuse = require 'OzCatAbuse'
local OzCatActivity = require 'OzCatActivity'
local OzCatAlbums = require 'OzCatAlbums'
local OzCatCache = require 'OzCatCache'
local OzCatCaptions = require 'OzCatCaptions'
local OzCatCatalog = require 'OzCatCatalog'
local OzCatCollections = require 'OzCatCollections'
local OzCatComments = require 'OzCatComments'
local OzCatConflicts = require 'OzCatConflicts'
local OzCatCopying = require 'OzCatCopying'
local OzCatDeletion = require 'OzCatDeletion'
local OzCatDiagnosticLog = require 'OzCatDiagnosticLog'
local OzCatErrors = require 'OzCatErrors'
local OzCatEvents = require 'OzCatEvents'
local OzCatExport = require 'OzCatExport'
local OzCatFavorites = require 'OzCatFavorites'
local OzCatFields = require 'OzCatFields'
local OzCatImport = require 'OzCatImport'
local OzCatInit = require 'OzCatInit'
local OzCatOriginals = require 'OzCatOriginals'
local OzCatPhotoInfo = require 'OzCatPhotoInfo'
local OzCatProxies = require 'OzCatProxies'
local OzCatSelection = require 'OzCatSelection'
local OzCatSyncStatus = require 'OzCatSyncStatus'
local OzCatUsers = require 'OzCatUsers'
local OzCatDateCorrection = require 'OzCatDateCorrection'


--------------------------------------------------------------------------------

local assert = assert
local WFCloseSupport_fetchCached = assert( WFCloseSupport.fetchCached )


--============================================================================--

--------------------------------------------------------------------------------
-- @mark class synthesis Utility
--------------------------------------------------------------------------------

-- Use end sentinel so that we can assert on missing method adders
local kEnd = function()end

local function verifyClass( subclasses )
	local i = 1
	while true do
		if subclasses[ i ] == kEnd then break end
		local subclass = subclasses[ i ]
		assert( subclass, "Subclass " .. tostring( i ) .. " not found" )
		i = i + 1
	end
end

--------------------------------------------------------------------------------
-- @mark OzCatalogService class

--[[!
	\class OzCatalogService
	\brief  This namespace provides access to methods that allow you to
		manage and manipulate Oz catalogs and their contents.

		The methods available through this class are defined in
		a number of different files. To instantiate this class, use:
		<code>self_catalog = OzCatalogService( catalogId, ozUser )</code>

		\sa OzUser
	 
]]--

local catalogServiceImpMethods = {

	OzCatAbuse.AbuseMethods,
	OzCatActivity.ActivityMethods,
	OzCatAlbums.AlbumMethods,
	OzCatCache.CacheMethods,
	OzCatCaptions.CaptionMethods,
	OzCatCatalog.CatalogMethods,
	OzCatCollections.CollectionMethods,
	OzCatComments.CommentMethods,
	OzCatConflicts.ConflictMethods,
	OzCatCopying.CopyingMethods,
	OzCatDateCorrection.DateCorrectionMethods,
	OzCatDeletion.DeletionMethods,
	OzCatDiagnosticLog.DiagnosticLogMethods,
	OzCatErrors.ErrorMethods,
	OzCatEvents.EventMethods,
	OzCatExport.ExportMethods,
	OzCatFavorites.FavoritesMethods,
	OzCatImport.ImportMethods,
	OzCatInit.InitMethods,
	OzCatOriginals.OriginalsMethods,
	OzCatPhotoInfo.PhotoInfoMethods,
	OzCatProxies.ProxyMethods,
	OzCatSelection.SelectionMethods,
	OzCatSyncStatus.SyncStatusMethods,
	OzCatUsers.UserMethods,
	WFCloseSupport.addCloseSupportMethods,

	-- This must be the last include.
	kEnd,
}

--------------------------------------------------------------------------------

-- Assert in case any of the above requires is nil, or if any duplicate methods
-- are found.
verifyClass( catalogServiceImpMethods )
local methods = {}

for _, methodList in pairs( assert( catalogServiceImpMethods ) ) do
	if type( methodList ) == 'table' then
		for k,v in pairs( methodList ) do
			assert( methods[ k ] == nil, "Method already implemented" )
			methods[ k ] = true
		end
	end
end

for k,v in pairs( assert( OzCatFields ) ) do
	assert( catalogServiceImpMethods[ k ] == nil, "Field already present" )
	assert( methods[ k ] == nil, "Method and field with same name" )
	catalogServiceImpMethods[ k ] = v
	methods[ k ] = true
end

--------------------------------------------------------------------------------
-- @mark OzCatalogService Definition
--------------------------------------------------------------------------------
 
local OzCatalogServiceImp = KSClass 'OzCatalogServiceImp'( catalogServiceImpMethods )

local catalogsById = KSTableUtils.makeFullyWeak()

local function makeOzCatalogServiceImp( catalogId, ozUser )

	local result = WFCloseSupport_fetchCached( catalogsById, catalogId )
	
	if result and result._ozUser ~= ozUser then
		debug.print( "Closing existing reference to catalog: changing user ID" )
		catalogsById[ catalogId ] = nil
		result:close()
		result = nil
	end

	if not result then
		result = OzCatalogServiceImp( catalogId, ozUser )
		catalogsById[ catalogId ] = result
	end
	
	return result

end

-- @end


--============================================================================--

local catalogRefMethods = {

	_catalogService = false,
		-- (OzCatalogServiceImp)
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function catalogRefMethods:__init( catalogId, ozUser )

	self._catalogService = makeOzCatalogServiceImp( catalogId, ozUser )
	self._catalogService._refsToThisCatalog[ self ] = true

end

--------------------------------------------------------------------------------

function catalogRefMethods:close()

	if self._catalogService then

		self._catalogService._refsToThisCatalog[ self ] = nil

		if not next( self._catalogService._refsToThisCatalog ) then
			self._catalogService:close()
		end
	
	end

end

--------------------------------------------------------------------------------

function catalogRefMethods:__index( k )

	local result = self._catalogService[ k ]

	if result and type( result ) == 'function' then
	
		local catalogImpMethod = result
		
		result = function( _, ... )
			return catalogImpMethod( self._catalogService, ... )
		end
		
		rawset( self, k, result )
	
	end
	
	return result

end

--------------------------------------------------------------------------------

return KSClass 'OzCatalogService'( catalogRefMethods )
