--[[----------------------------------------------------------------------------

OzCatalogTestUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local KSTableUtils = require "KSTableUtils"
local KSStringUtils = require "KSStringUtils"
local KSPromises = require "KSPromises"
local KSData = require "KSData"
local WFWorkerThread = require 'WFWorkerThread'

--==============================================================================

local OzCatalogTestUtils = {}

--------------------------------------------------------------------------------

-- Notice the one %s in this template, which is in the middle of the
-- "software" field in the Exif data for the tiny jpeg.  We can insert
-- any base64 valid char we want there and it will still decode into
-- a valid JPEG (but with a new, unique SHA256).
local kTinyJpegBase64Tmplt =
[[/9j/4QAoRXhpZgAASUkqAAgAAAABAJiCAgADAAAANDIAAAAAAAA0MgAAAAD/7AARRHVja3]] ..
[[kAAQAEAAAAAAAA/+0ALFAAAAAAA%sAAAAAAAAAAOEJJTQQlAAAAAAAQAAAAAAAAAAAAAAAA]] ..
[[AAAAAP/uAA5BZG9iZQBkwAAAAAH/2wCEABsaGikdKUEmJkFCLy8vQkc/Pj4/R0dHR0dHR0]] ..
[[dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0cBHSkpNCY0PygoP0c/NT9H]] ..
[[R0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR//AAB]] ..
[[EIAEAAQAMBIgACEQEDEQH/xABOAAEBAQAAAAAAAAAAAAAAAAAABQMBAQEBAAAAAAAAAAAA]] ..
[[AAAAAAADBRABAAAAAAAAAAAAAAAAAAAAABEBAAAAAAAAAAAAAAAAAAAAAP/aAAwDAQACEQ]] ..
[[MRAD8A1ARaYAAAAACGAszFwBFpgAIYCzMAAXAEWmAAhgLMwABcARaYACGAszH/2Q==]]

local kFakeImageInfo = {
	_localOnly = {
		localUrl = "/fake/path/to/fake/file/%s",
	},
	captureDate = "2012-05-02T20:53:22.00",
	height = 16,
	importSource = {
		fileCreationDate = "2004-11-02T08:37:24",
		fileName = "%s",
		filePath = "/fake/path/to/fake/file/%",
		fileSize = 13493,
		folder = "/fake/path/to/fake",
		folderPath = "/fake/path/to/fake",
	},
	keywords = {
		hawaii = true,
	},
	width = 16,
}

--------------------------------------------------------------------------------

local function createImageInfo( originalData, fakeFileName, captureDate )
	
	local result = KSTableUtils.deepCopy( kFakeImageInfo )
	result._localOnly.localUrl = string.format( result._localOnly.localUrl, fakeFileName )
	result.importSource.fileName = string.format( result.importSource.fileName, fakeFileName )
	result.captureDate = captureDate
	result.importSource.sha256 = originalData:digest( "SHA256" )
	return result
	
end

--------------------------------------------------------------------------------

local function createUniqJpg( base64Char )
	assert( #base64Char == 1 )
	local uniq = string.format( kTinyJpegBase64Tmplt, base64Char )
	return KSStringUtils.decodeBase64( uniq )
end

--------------------------------------------------------------------------------

function OzCatalogTestUtils.importPhotosP( catalogService, inputs )
	
	local colors = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
	local fakeFileInfos = {}
	if type( inputs ) == 'number' then
		for i = 1, inputs do
			fakeFileInfos[ i ] = {
				fileName = string.format( "fakeFile%02d.jpg", i ),
				randomBase64Byte = string.sub(colors,i,i),
			}
		end
	else
		for i, template in ipairs( inputs ) do
			fakeFileInfos[ i ] = {
				fileName = template.fileName or string.format( "fakeFile%02d.jpg", i ),
				randomBase64Byte = string.sub(colors,i,i),
				captureDate = template.captureDate,
			}
		end
	end
	
	return KSPromises.allP( function( fakeFileInfo )
		
		local data = createUniqJpg( fakeFileInfo.randomBase64Byte )
		local fakeFileName = fakeFileInfo.fileName
		local captureDate = fakeFileInfo.captureDate or "1973-03-08T00:37:24"
		local imageInfo = createImageInfo( data, fakeFileName, captureDate )
		
		return catalogService:addPhotoP {
			imageInfo = imageInfo,
			originalData = data,
			originalPath = imageInfo.importSource.filePath,
			thumbData = KSData.make( data ),
			previewData = KSData.make( data ),
			isAutoImported = false,
			proxyData = nil, -- it seems like this is optional
			thumbnail2xData = KSData.make( data ),
		}
		
	end, fakeFileInfos )
	:mapDataP( function( status )
		return status, true
	end )
	:runUntilDoneP()
	
end

--------------------------------------------------------------------------------

function OzCatalogTestUtils.makeWorkerThreadMock( optAllowedWorkers )
	return setmetatable( {
		kNilValue = { n = 1 },
	}, {
		__call = function( self, name, ... )
			if optAllowedWorkers and optAllowedWorkers[ name ] then
				return WFWorkerThread( name, ... )
			end
			return {
				requestP = function()
					local KSPromises = require "KSPromises"
					return KSPromises.resolvedPromiseP( 'data', true )
				end,
				requestE = function()
					local KSReactive = require "KSReactive"
					return KSReactive.sourceE()
				end,
				requestB = function()
					local KSReactive = require "KSReactive"
					return KSReactive.constantB()
				end,
			}
		end,
	})
end

--------------------------------------------------------------------------------

function OzCatalogTestUtils.buildMocksForTestingCatalogService( T, ozUser, catalogId, optAllowedWorkerThreads )
	
	local WFOzSyncTestUtils = require 'WFOzSyncTestUtils'
	local KSReactive = require 'KSReactive'
	local WFIndexStoreSchema
	local docStore, indexStore
	
	local mocks = {
		KSFileUtils = {
			mkdir = function()
			end,
		},
		
		Bablu = {
		
		},
		
		OzCatInit = false,
		OzCatalogService = false,
		
		OzAccountService = {
			currentUserB = function()
				return KSReactive.constantB( ozUser )
			end,
		},
		
		OzCentralUploadTask = WFOzSyncTestUtils.makeFakeOzCentralUploadTask( T ),
		
		WFWorkerThread = OzCatalogTestUtils.makeWorkerThreadMock( optAllowedWorkerThreads ),
		
		OzCatalogIndexDefinitions = false,
		
		-- Fake implementation which uses OzCatalogIndexDefinitions to build out the proper
		-- schema and mapping functions for our in-memory indexStore
		WFIndexStoreSchema = function( ... )
			if not WFIndexStoreSchema then
				WFIndexStoreSchema = docStore:raw_test_buildMockWFIndexStoreSchema()( ... )
			end
			return WFIndexStoreSchema
		end,
		
		-- Fake impl, does nothing
		WFIndexStoreImp = {
			start = function( schema )
			
			end,
		},
		
		OzDocStore = function( catalogId )
			docStore = WFOzSyncTestUtils.makeFakeDocStore()
			local WFIndexStoreSchema = require 'WFIndexStoreSchema'
			return WFIndexStoreSchema and docStore
		end,
		
		WFIndexStore = function( docStore, fname )
			-- Build a fake indexStore.  It's powered by a real sql db in memory
			-- that uses the real schema and mapping fns.
			indexStore = docStore:raw_test_buildMockIndexStore( )
			return indexStore
		end,
	}

	local OzCatalogService = WFOzSyncTestUtils.importWithMocks( 'OzCatalogService', mocks )
	
	local catalog = OzCatalogService( catalogId, ozUser )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )

	return docStore, catalog

end

--==============================================================================

return OzCatalogTestUtils