--[[----------------------------------------------------------------------------

OzCatalogUploadStatus.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-06-01

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

local KSDeviceUtils = require "KSDeviceUtils"
local KSDeviceUtils_deviceName = KSDeviceUtils.deviceName

local WFOzSyncConstants = require "WFOzSyncConstants"

--------------------------------------------------------------------------------

local uploadStatusLog = KSLogging.log 'OzCatalogUploadStatus'
--	:setLevel( KSLogging.DEBUG )

local uploadStatusDetailedLog = KSLogging.log 'OzCatalogUploadStatusDetails'
--	:setLevel( KSLogging.DEBUG )

local conflictLog = KSLogging.log 'OzCatalogUploadStatus :: conflict'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

local OzCatalogUploadStatus = {}

local WFPlatformInfo = require 'WFPlatformInfo'

--------------------------------------------------------------------------------

local kRevisionStartIndex = 0 - ( 32 + # "/renditions/" + 32 )
local kRevisionEndIndex = 0 - ( # "/renditions/" + 32 + 1 )

assert( kRevisionStartIndex == -76, "Revision String Substitution Index May be incorrect" )
assert( kRevisionEndIndex == -45, "Revision String Substitution Ending Index May be incorrect" )

function OzCatalogUploadStatus.shouldUploadRendition( devSettingsLastChanged, localDescription,
					serverLinksForRendition )

	-- Don't upload if we're not current.

	-- FIXME [scouten 2011-07-29]: We need Oz changes to avoid a possible race
	-- condition. If the server state advances quickly enough that we don't tag the
	-- develop change as coming from our rev, then we will never think we're
	-- current in that sense, and never upload the renditions. Obscure, but possible,
	-- race condition. (MattA and I discussed since this and agreed that leaving
	-- thumbnails out of non-current nodes in the revision history isn't a fatal
	-- flaw.)

	local localRenditionIsForRev = localDescription and localDescription.revId
	
	if localRenditionIsForRev == 'pending' and not serverLinksForRendition then
		return true
	end
	
	if not localRenditionIsForRev
		or devSettingsLastChanged ~= localRenditionIsForRev
		or not localDescription
		or localDescription.md5 == 'missing'
		or localDescription.md5 == nil
	then
		return false
	end

	-- We're current. Does Oz need this rendition?
	
	if not serverLinksForRendition then return true end
	
	if serverLinksForRendition.invalid then return true end
	
	if type( serverLinksForRendition.href ) == 'string' and
			string.sub( serverLinksForRendition.href, kRevisionStartIndex,kRevisionEndIndex ) ~= localDescription.revId then
		return true
	end
	
	return false

end

--------------------------------------------------------------------------------

local shouldUploadRendition = OzCatalogUploadStatus.shouldUploadRendition

function OzCatalogUploadStatus.calculateUploadStatus( docValue )
	
	local firstPassWithAsset = false
	local docType = docValue._docType
	local uploadedRollup = docValue._ozMetadataUploadStatus
	local localOnly = docValue._localOnly or {}
	
	if localOnly._uploadedValues then
		uploadedRollup = 'uploaded'
	end

	if not uploadedRollup then
		if docType == 'asset' then
			firstPassWithAsset = true
		end
		uploadedRollup = 'notUploaded'
	end
	
	local serverLinks =
		  ( docValue._serverProvidedMetadata
		and docValue._serverProvidedMetadata.links ) or {}
	
	local importedOnDevice = docValue.importSource
		and docValue.importSource.importedOnDevice
		
	local originalUploadedRollup = uploadedRollup

	if (
			-- TODO [arahn 2012-10-11]: We should consolidate the list of
			-- supported, generic document types, so that this if statement
			-- doesn't need to be updated everytime we start handling
			-- a new kind of document.
			docType == 'asset' or
			docType == 'album' or
			docType == 'album_asset' or
			docType == 'space' or
			docType == 'comment'
		
		) and docValue._deleted and WFOzSyncConstants.isNotUploaded( originalUploadedRollup ) then
	
		-- This is a document we need to delete
		uploadedRollup = 'deleted'

	elseif docValue._deleted then
		
		-- original uploaded rollup stands; no rendition uploads permitted

	elseif docValue._conflict then

		conflictLog:info( "detected conflict on %s %s :: _conflict = %s",
							docType, docValue._docId, docValue._conflict )

		uploadedRollup = '_conflict'

	elseif docType == 'asset'
				-- For any image asset
			and ( WFOzSyncConstants.isNotUploaded( uploadedRollup ) or uploadedRollup == 'uploaded' )
				-- that we know is not uploaded, OR even if it's marked "uploaded" because this
				-- apparently can happen *even when it's not true* :(
			and ( localOnly.recoveryEffort == 'original'
				-- And we haven't previously tried recovery and failed
					or (
						localOnly.recoveryEffort == nil
						and importedOnDevice ~= nil
						and KSDeviceUtils_deviceName() == importedOnDevice
						and localOnly.localUrl == nil
						and localOnly.originalFilePath == nil ) )
						-- or it's clear that this document has no original and it was synced from the cloud
						-- so we are not currently mid-way through importing it
			and not ( serverLinks[ "/rels/master" ] and serverLinks[ "/rels/master" ].href )  then
				-- And there is no master
				
		uploadedRollup = 'recoveryEffort_original'

	elseif docType == 'asset' and uploadedRollup ~= 'notUploaded' and uploadedRollup ~= 'notUploaded_metadataOnly' then

		local devSettingsLastChanged = localOnly.devSettingsLastSetAtRev

		if shouldUploadRendition( devSettingsLastChanged, localOnly.thumbnail,
					serverLinks[ "/rels/rendition_type/144h" ] )
		then
			
			uploadStatusLog:info( "setting back to notUploaded for photo %s because thumbnail isn't uploaded",
								  docValue._docId )

			uploadedRollup = 'notUploaded'
			firstPassWithAsset = false
			
		-- If we have the 2x thumbnail locally and it needs uploading, set the status of this asset to
		-- 'notUploaded'
		elseif localOnly.thumbnail2x and shouldUploadRendition( devSettingsLastChanged, localOnly.thumbnail2x,
					serverLinks[ "/rels/rendition_type/thumbnail2x" ] )
		then
			
			uploadStatusLog:info( "setting back to notUploaded for photo %s because 2x thumbnail isn't uploaded",
								  docValue._docId )

			uploadedRollup = 'notUploaded'
			firstPassWithAsset = false

		elseif shouldUploadRendition( devSettingsLastChanged, localOnly.preview,
					serverLinks[ "/rels/rendition_type/1024" ] )
		then
		
			uploadStatusLog:info( "setting back to notUploaded for photo %s because preview isn't uploaded",
								  docValue._docId )

			uploadedRollup = 'notUploaded'
			firstPassWithAsset = false

		-- FIXME [scouten 2011-07-29]: Someday when master images can change
		-- we will need to change this. NiV1
		
		elseif ( ( localOnly.originalFilePath and localOnly.originalFilePath ~= 'missing' )
				or localOnly.copyMasterFrom )
			and not ( serverLinks[ "/rels/master" ]
				  and serverLinks[ "/rels/master" ].href )
		then
	
			if not docValue._ozCentralRevisionId then
				uploadStatusLog:warn( "setting photo ID %s to upload stage 2 (original), but localOnly._ozCentralRevisionId is empty", docValue._docId )
			end
		
			uploadedRollup = 'notUploadedStage2-original'
		
		elseif localOnly.originalFilePath == 'missing' and (
				localOnly.recoveryEffort == 'original' or
				localOnly.recoveryEffort == nil ) then
				
			uploadedRollup = 'recoveryEffort_original'

		end
	
	end

	if uploadedRollup == 'notUploaded' and firstPassWithAsset then
		uploadedRollup = 'notUploaded_metadataOnly'
	end

	uploadStatusLog:debug( "ozUploadStatus rollup:\n\n  original rollup: %s\n\nuploaded: %s\n",
		originalUploadedRollup, uploadedRollup )
	uploadStatusDetailedLog:debug( "\n  docValue: %s\n\n", docValue )

	return uploadedRollup

end

--------------------------------------------------------------------------------

function OzCatalogUploadStatus.assetOriginalIsValid( docValue )

	local ozUrl = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.ozUrl
				
	local spm_master_href = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
				
	local spm_first_master = docValue
			    and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ].href
		
	
	-- if we don't have server metadata this original hasn't been uploaded yet and
	-- we return true
	
	if not spm_master_href then
		return true
	end
	-- If there is a value for the URL this original was downloaded from, we compare its value to
	-- the value from the server for the master
	if ozUrl then
		return ozUrl == spm_master_href
	elseif spm_first_master then
		return spm_master_href == spm_first_master
	else
		return true
	end
				
end

--------------------------------------------------------------------------------

function OzCatalogUploadStatus.assetLatestOriginalIsValid( docValue )

	local ozUrl_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.ozUrl
				
	local spm_master_href = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href

	-- if the URL we downloaded the latest original from matches the master
	-- it is the latest
	return ozUrl_latestOriginal == spm_master_href
		
end

--------------------------------------------------------------------------------

-- FIXME [sdaniel 2012-02-07]: Needs to check whether proxy corresponds to latest original (check by revision?)

function OzCatalogUploadStatus.assetProxyIsValid( docValue )

	local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.proxy
				and docValue._localOnly.proxy.filePath

	local ozUrl = docValue
				and docValue._localOnly
				and docValue._localOnly.proxy
				and docValue._localOnly.proxy.ozUrl
				
	local spm_proxy_href = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ]
				and docValue._serverProvidedMetadata.links[ "/rels/proxy_type/2048" ].href
		
	
	-- if we don't have server metadata this proxy hasn't been uploaded or generated on Oz yet and
	-- we return true if it is available locally
	
	if not spm_proxy_href and localPath then
		return true
	end
	-- If there is a value for the URL this proxy was downloaded from, we compare its value to
	-- the value from the server for the master
	if ozUrl then
		return ozUrl == spm_proxy_href
	else
		return true
	end
	
	--[[
		 Even if we (or Oz) has a proxy if someone else uploads an original with destructive edits
		then our proxy is no longer valid
	--]]
				
end

--------------------------------------------------------------------------------

function OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath )

	-- if the path passed in was nil, return nil
	if not localPath then return localPath end
	
	local renamedPath = localPath
	
	-- On Mac and only if the library/caches has been moved in 2.0, return the corrected path

	if WFPlatformInfo.isMac then
		if renamedPath:match( "Adobe Carousel Library.cr" ) then
			renamedPath = renamedPath:gsub( "Adobe Carousel Library.cr", "Adobe Revel/Adobe Revel Library.rvl")
			
		elseif renamedPath:match( "Adobe Revel Library.cr" ) then
			renamedPath = renamedPath:gsub( "Adobe Revel Library.cr", "Adobe Revel/Adobe Revel Library.rvl")
		-- For caches, substitute 'Adobe Carousel' with 'Adobe Revel'
		
		elseif renamedPath:match( "Adobe Carousel" ) then
			renamedPath = renamedPath:gsub( "Adobe Carousel", "Adobe Revel")
		end
	end
	
	return renamedPath
	
end

--------------------------------------------------------------------------------

return OzCatalogUploadStatus
