--[[----------------------------------------------------------------------------

OzCentralSyncAgent.lua

@author scouten 2011-02-22

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class represents an active sync conversation between one local Oz catalog
and its corresponding representation on the Oz central service.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFTimer = require 'WFTimer'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- wichitafoundation.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- wichitafoundation.oz_client.sync_agent
local WFOzSyncCycle = require 'WFOzSyncCycle'

	-- wichitafoundation.web.http_client
local WFReachability = require 'WFReachability'


--------------------------------------------------------------------------------

local hostIsReachableB = WFReachability.hostIsReachableB( OzCatalogUrls.baseServerUrl )

--------------------------------------------------------------------------------

local minDownloadPollingInterval = 2 -- seconds

-- Maximum download polling interval based on reachability (seconds)

local maxDownloadPollingIntervals = {
	unreachable = 90,
	wifi = 30,
	wwan = 60,
	serviceUnavailable = 60,
	default = 60 -- backup value
}

local downloadPollingIntervalIncreaseFactor = 1.5
	-- Increase the time by 50% for each failure to get data

local downloadIdleLog = KSLogging.log 'OzCentralSyncAgent::Download (idle)'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

-- We attempt to know progress by collaborating with ourselves on other devices
-- Whenever a device reaches the end of the feed, they write in to a shared space
-- the "external" count of the length of the feed.  Every time a device consumes
-- some of the feed, they update an internal track of how far along they are.
-- The ratio of the two is progress.
local function updateInternalProgress( syncAgent, done )
	local cumulativeInternalProgress = ( syncAgent._oldInternalProgressCount or 0 ) + syncAgent._internalProgressCount
	local denom = syncAgent._externalProgressCount
	if done then
		syncAgent._downloadingFromOzProgressB:set( 1, denom or 1 )
		if syncAgent._oldInternalProgressCount then
			syncAgent._ozUser:setDownloadBookmarkP( syncAgent._catalogId, cumulativeInternalProgress, done )
		end
	else
		if syncAgent._oldInternalProgressCount then
			if not denom or denom == 0
				or denom < cumulativeInternalProgress then
				
				-- we don't know
				syncAgent._downloadingFromOzProgressB:set( false )
			else
				syncAgent._downloadingFromOzProgressB:set( cumulativeInternalProgress / denom, denom )
			end
			syncAgent._ozUser:setDownloadBookmarkP( syncAgent._catalogId, cumulativeInternalProgress, done )
		else
				-- we don't know yet
			syncAgent._downloadingFromOzProgressB:set( false )
		end
	end

end

--============================================================================--

-- @mark OzCentralSyncAgent class

--[[!
	\class 	OzCentralSyncAgent
	\brief This namespace provides control over synchronization options
		between a Wichita Foundation client and the Oz server.

]]--

local syncAgentMethods = {

	_ozUser = false,
		-- (OzUser) user who is doing the syncing
	
	_catalogId = false,
		-- (string) catalog ID as known to the Oz system (UUID)
	
	_catalogUrl = false,
		-- (string) Oz URL for this catalog
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)
	
	_isInSyncCycleB = false,
		-- (B< boolean >) are we currently running a sync cycle with Oz?
	
	_isDownloadingB = false,
		-- (B< boolean >) a calmer version of isInSyncCycle, useful for UI on sync status?
	
	_downloadingFromOzProgressB = false,
		-- (B< number >) progress of the download from Oz
	
	_internalProgressCount = false,
	_oldInternalProgressCount = false,
	_externalProgressCount = false,
		-- Progress counters; optimistically the external count (which comes from
		-- another device's experience with syncing this catalog) will be higher
		-- than our internal progress, and gives us a sense of where we are in the sync.
	
	_hasDownloadedSuccessfullyB = false,
		-- (B< boolean >) has this catalog downloaded successfully?

	_rawErrorB = false,
		--	(B< string >) reports upload or download errors

	_existsOnOzB = false,
		-- (B< boolean >) does this catalog exist on Oz?

	_accessErrorB = false,
		-- (B< boolean >) have we received a 403 from Oz?
		
	_serviceAvailableB = false,
		-- (B< boolean >) is Oz service is available?
	
	_hasAnythingToUploadB = false,
		-- (B< boolean >) true if there is anything to upload

	_statusReport = false,
		-- (table) status report for failed uploads
	
	_syncCycleTimerAction = false,
		-- (KSPromiseAction) current upload timer trigger
	
	_downloadPollingInterval = false,
		-- (number) how many seconds to wait before syncing again (adaptive)
	
	-- FIXME [ngautam 2011-08-31] Last minute hack to get the ITMS link for a
	-- force upgrade error. All the 403/404 error handling possibly belongs in the HttpClient
	-- code.

	_itmsLinkB = false,
		-- (B< string >) the ITMS link for this client

	_recentErrors = false,
		-- (table) array of info 10? most recent sync errors

	_syncNowP = false,
		-- (P< true >) true when we are fully synced (only if needed)

	_syncNowResolver = false,
		-- (KSPromiseResolver) resolver for _syncNowP
	
	_nCyclesWithError = 0,
		-- (number) how many cycles failed?
	
	_nCyclesWithDidAnything = 0,
		-- (number) how many cycles completed saying they "did anything"?
	
	_nCyclesCompleted = 0,
		-- (number) how many sync cycles have completed?

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	Creates a new sync-agent object for a catalog and Oz user.

	\param catalogId (string) The catalog's unique ID.
	\param ozUser (ozUser) The Oz-user object.

]]--

function syncAgentMethods:__init( --[[string]] catalogId, --[[ozUser]] ozUser )

	self._ozUser = assert( ozUser )
	local ozUserId = assert( ozUser:id() )
	local isUserCatalog = ozUserId == catalogId
		-- Sync agent is used for user "catalog" (notably Apple Store receipt uploads).

	self._catalogId = assert( catalogId )
	
	self._isInSyncCycleB = KSReactive.variableB( false )
	self:addCloseTarget( self._isInSyncCycleB )
	
	self._isDownloadingB = KSReactive.variableB( false )
	self:addCloseTarget( self._isDownloadingB )
	
	self._downloadingFromOzProgressB = KSReactive.variableB( false )
	self:addCloseTarget( self._downloadingFromOzProgressB )
	
	self._hasDownloadedSuccessfullyB = KSReactive.variableB( false )
	self:addCloseTarget( self._hasDownloadedSuccessfullyB )

	self._rawErrorB = KSReactive.variableB( nil )
	self:addCloseTarget( self._rawErrorB )
	
	self._existsOnOzB = KSReactive.variableB( true )
		-- This must start off with true, or we run the risk of deleting the catalog!
	self:addCloseTarget( self._existsOnOzB )

	self._accessErrorB = KSReactive.variableB( false )
	self:addCloseTarget( self._accessErrorB )
	
	self._serviceAvailableB = KSReactive.variableB( true )
	self:addCloseTarget( self._serviceAvailableB )

	self._itmsLinkB = KSReactive.variableB( nil )
	self:addCloseTarget( self._itmsLinkB )

	self._recentErrors = {}

	self._statusReport = {}
		-- FIXME [scouten 2011-09-15]: Make this an MRU cache?

	if not isUserCatalog then
		self._catalogUrl = string.format( '%s/%s', OzCatalogUrls.catalogsUrl, catalogId )
			-- FIXME [scouten 2011-04-22]: This URL should probably be passed down
			-- from OzCatalogList rather than computed. NNTD
	end

	self._docStore = OzDocStore( catalogId )
	self:addCloseTarget( self._docStore )
	
	self._downloadPollingInterval = minDownloadPollingInterval

	-- FIXME [scouten 2011-07-13]: Ugh. Don't like hard-coding the index
	-- definition based on catalog ID.

	self._indexStore = WFIndexStore( self._docStore,
		isUserCatalog
			and 'OzUserInfoCacheIndexDefinitions.lua'
			or 'OzCatalogIndexDefinitions.lua' )
	self:addCloseTarget( self._indexStore )

	-- Do we have anything to upload? If so, trigger a new upload cycle.
	-- We compute the actual upload manifest later.
	
	self._hasAnythingToUploadB = self._indexStore:queryB(
		[[SELECT docId FROM ozUploadStatus WHERE ozUploadStatus != '_conflict' LIMIT 1]] )
		:mapB( function( rawUploadQueue )
			return ( rawUploadQueue and #rawUploadQueue > 0 ) or false
		end )

	self:addCloseTarget( self._hasAnythingToUploadB )
	
	self._hasAnythingToUploadB:onDataB( function( hasAnything )
		if hasAnything then
			self:scheduleSyncCycle()
		end
	end )

	if not isUserCatalog then
		self:scheduleSyncCycle()

		self._internalProgressCount = 0
		self._ozUser:getDownloadBookmarkP( self._catalogId )
			:whenResolvedDo( function( status, external, internal )
				self._oldInternalProgressCount = internal
				self._externalProgressCount = external
			end )
			:runUntilDone()
	end

	-- If the reachability transitions to a better connection, try a new session now.
	
	local oldReachable = 'default'
	
	local function mergeB( newReachable, serviceAvailable )
	
		if serviceAvailable then
			return newReachable
		else
			return 'serviceUnavailable'
		end
	
	end
	
	KSReactive.dataCallB( mergeB, hostIsReachableB, self._serviceAvailableB )
		:onDataB( function( newReachable )
	
			local oldMaxTime = maxDownloadPollingIntervals[ oldReachable ]
				or maxDownloadPollingIntervals.default
	
			local newMaxTime = maxDownloadPollingIntervals[ newReachable ]
				or maxDownloadPollingIntervals.default
	
			oldReachable = newReachable
	
			if newMaxTime < oldMaxTime and not isUserCatalog then
				self:scheduleSyncCycle()
			end
	
		end ):addAsCloseTargetOf( self )
	
	WFDiagnosticLog.addLogProvider( self )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether an upload operation is in progress.

	\return	<code>B < boolean ></code> A Behavior whose value is true if upload is
		is in progress.

]]--
function syncAgentMethods:isUploadingToOzB()

	return self._hasAnythingToUploadB:relayB()

end

--------------------------------------------------------------------------------

--[[!
	\brief Reports (as a Behavior) whether a download operation is in progress.

	\return	<code>B < boolean ></code> A Behavior whose value is true if download is
		is in progress.

]]--

function syncAgentMethods:isDownloadingFromOzB()

	return self._isDownloadingB:relayB()
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Reports (as a Behavior) the download progress.

	\return	<code>B < number ></code> A Behavior whose value is between zero and
		one for the download progress.  Or false if it's indeterminate.

]]--

function syncAgentMethods:downloadingFromOzProgressB()

	return self._downloadingFromOzProgressB:relayB()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether a download operation has completed.

	\return	<code>B < boolean ></code> A Behavior whose value is true if the
		most recent download has completed successfully.

]]--

function syncAgentMethods:hasDownloadedSuccessfullyB()

	return self._hasDownloadedSuccessfullyB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) any access error for the current upload
		or download operation. 	((is this right??)

	\return <code>B < string ></code> A Behavior whose value is the error message.
		((is this right??)

]]--

function syncAgentMethods:accessErrorB()

	return self._accessErrorB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether synchronization service is currently
		available.

	\return	<code>B < boolean ></code> A Behavior whose value is true if
		the service is available.

]]--

function syncAgentMethods:serviceAvailableB()

	return self._serviceAvailableB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether ((what??) currently exists on the Oz server.

	\return	<code>B < boolean ></code> A Behavior whose value is true if ((what??) exists.

]]--

function syncAgentMethods:existsOnOzB()

	return self._existsOnOzB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the current server status of a photo.

	\param photoId (string) The photo's unique ID.

	\return	(string) The status report or an error message if no report is found.

]]--

function syncAgentMethods:DEBUG_statusReport( --[[string]] photoId )

	return self._statusReport[ photoId ] or "(no status report for this photo)"

end

--------------------------------------------------------------------------------
--[[!
	\brief ((what?))

	\return

]]--

function syncAgentMethods:itmsLinkB()

	return self._itmsLinkB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) any access error for the current upload
		operation. ((is this right??)

	\return <code>B < string ></code> A Behavior whose value is the error message.
		((??)

]]--

function syncAgentMethods:uploadErrorB()

	-- FIXME [scouten 2012-04-16]: Uploading == downloading == talking now.
	-- Should simplify API surface, but taking the easy way out for the moment.

	return self._rawErrorB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) any access error for the current download
		operation. ((is this right??)

	\return <code>B < string ></code> A Behavior whose value is the error message.
		((??)

]]--

function syncAgentMethods:downloadErrorB()

	-- FIXME [scouten 2012-04-16]: Uploading == downloading == talking now.
	-- Should simplify API surface, but taking the easy way out for the moment.

	return self._rawErrorB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Increases the interval for download polling by half a second.

		The interval is limited according to host reachability.

]]--

function syncAgentMethods:increaseDownloadInterval()

	local newInterval = self._downloadPollingInterval * downloadPollingIntervalIncreaseFactor
		
	-- Keep it to half-second increments.
		
	newInterval = 0.5 * math.floor( 2 * newInterval + 0.5 )

	-- Don't go past the limit determined based on host reachability

	local limit = maxDownloadPollingIntervals[ hostIsReachableB:valueNow() or 'default' ]
				or maxDownloadPollingIntervals.default

	if limit < newInterval then
		newInterval = limit
	end
		
	self._downloadPollingInterval = newInterval

end

--------------------------------------------------------------------------------
--[[!
	\brief Decreases the interval for download polling by half a second.

]]--

function syncAgentMethods:decreaseDownloadInterval()

	-- Go back to polling frequently.
	
	self._downloadPollingInterval = minDownloadPollingInterval

end

--------------------------------------------------------------------------------

local syncCycleCounter = 0

--------------------------------------------------------------------------------
--[[!
	\brief Initiates a synchronization operation.

	\return	A Promise that signals completion of the request.

]]--

function syncAgentMethods:syncNowP()

	if not self._syncNowP then
	
		self._syncNowP, self._syncNowResolver = KSPromises.promiseP()
		self._syncNowP:whenResolvedDo( function() end ):attachTo( self )

		self:scheduleSyncCycle()
	
	end

	return self._syncNowP

end

--------------------------------------------------------------------------------
--[[!
	\brief Schedules a synchronization operation for the next available cycle. ((??))

]]--

function syncAgentMethods:scheduleSyncCycle()

	-- Sync agent closed? We're done.

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		return
	end

	-- If we're already doing a sync cycle, don't start another one.
	
	if self._isInSyncCycleB:valueNow() then
		return
	end

	-- Bump sync cycle counter (used only for debugging).
	
	syncCycleCounter = syncCycleCounter + 1
	downloadIdleLog:debug( "Starting sync cycle %s for catalog %s", syncCycleCounter, self._catalogUrl )
	
	-- If we have an existing timer for scheduling another sync cycle,
	-- cancel it and do it now.
	
	if self._syncCycleTimerAction then
		self._syncCycleTimerAction:close()
		self._syncCycleTimerAction = false
	end
	
	-- Start a sync cycle.
	
	local errorCapture = { cycleStartTime = KSDate.timeInIso8601FormatWithTimeZone() }
	
	local syncCycleP, didSomeWorkE = WFOzSyncCycle.runOneCycleP {
		docStore = self._docStore,
		indexStore = self._indexStore,
		ozUser = self._ozUser,
		catalogId = self._catalogId,
		catalogUrl = self._catalogUrl,
		statusReport = self._statusReport,
		itmsLinkB = self._itmsLinkB,
		existsOnOzB = self._existsOnOzB,
		accessErrorB = self._accessErrorB,
		serviceAvailableB = self._serviceAvailableB,
		hasDownloadedSuccessfullyB = self._hasDownloadedSuccessfullyB,
		isUserCatalog = self._catalogId == self._ozUser:id(),
		errorCapture = errorCapture,
	}
	
	didSomeWorkE:onDataE( function()
		self._isDownloadingB:set( true )
	end )
	
	self._isInSyncCycleB:set( true )

	syncCycleP:whenResolvedDo( function( status, didAnything, numItemsDownloaded )
		
		if status == 'data' and numItemsDownloaded then
			self._internalProgressCount = self._internalProgressCount + numItemsDownloaded
		end
		
		didSomeWorkE:close()
		self._isInSyncCycleB:set( false )
		
		self._rawErrorB:set( status ~= 'data' and didAnything or nil )

		self._nCyclesCompleted = self._nCyclesCompleted + 1
		
		if status ~= 'data' then

			self._nCyclesWithError = self._nCyclesWithError + 1
			
			errorCapture.cycleEndTime = KSDate.timeInIso8601FormatWithTimeZone()
			errorCapture.message = didAnything
			
			didAnything = false

			self._recentErrors[ #self._recentErrors + 1 ] = errorCapture
			if #self._recentErrors > 10 then
				table.remove( self._recentErrors, 1 )
			end

		end
		
		-- If we had any activity, reschedule for sync right now.
		-- Otherwise, wait a little while.

		if didAnything then

			self._isDownloadingB:set( true )
			updateInternalProgress( self, false )
			
			self._nCyclesWithDidAnything = self._nCyclesWithDidAnything + 1

			downloadIdleLog:debug( "had sync activity in previous session, scheduling new sync cycle immediately" )

			self:scheduleSyncCycle()
			self:decreaseDownloadInterval()

		else

			self._isDownloadingB:set( false )
			updateInternalProgress( self, true )

			-- If we have a syncNow promise outstanding, resolve it now.
			
			if self._syncNowP then
				self._syncNowResolver( 'data', true )
				self._syncNowResolver = false
				self._syncNowP = false
			end

			self:increaseDownloadInterval()
			
			downloadIdleLog:debug( "no sync activity in previous session, delaying for %s seconds",
								   self._downloadPollingInterval)

			self._syncCycleTimerAction = WFTimer.afterDelayP( self._downloadPollingInterval )
				:whenResolvedDo( function()

					downloadIdleLog:debug( "sync timer fired for catalog %s", self._catalogUrl )
					self:scheduleSyncCycle()
				
				end ):attachTo( self )
			
		end
	
	end ):attachTo( self )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on synchronization to the log.

]]--

function syncAgentMethods:sendDiagnosticLogData()

	local isUserCatalog = self._catalogId == self._ozUser:id()

	return {

		name = isUserCatalog and "Sync Agent (User Info)" or "Sync Agent (Catalog)",
		id = 'WFOzSyncAgent-' .. ( isUserCatalog and 'user-' or '' ) .. self._catalogId,
		weight = 500,

		body = {
		
			catalogId = self._catalogId,
			catalogUrl = self._catalogUrl,
		
			hostIsReachable = hostIsReachableB,
			
			isInSyncCycle = self._isInSyncCycleB,
			isDownloadingB = self._isDownloadingB,
			
			hasDownloadedSuccessfully = self._hasDownloadedSuccessfullyB,

			rawError = self._rawErrorB,
			existsOnOz = self._existsOnOzB,
			accessErrorB = self._accessErrorB,
			serviceAvailable = self._serviceAvailableB,
			downloadPollingInterval = self._downloadPollingInterval,

			recentErrors = self._recentErrors,
			
			syncCycles = {
				nCycles = self._nCyclesCompleted,
				nCyclesWithDidAnything = self._nCyclesWithDidAnything,
				pctCyclesWithDidAnything = self._nCyclesCompleted > 0 and string.format( "%.2d",
					self._nCyclesWithDidAnything / self._nCyclesCompleted * 100 ) or "0",
				nCyclesWithError = self._nCyclesWithError,
				pctCyclesWithError = self._nCyclesCompleted > 0 and string.format( "%.2d",
					self._nCyclesWithError / self._nCyclesCompleted * 100 ) or "0",
			},
			
			nAlbums = ( not isUserCatalog )
						and self._indexStore:queryP 'SELECT count() FROM albums'
							:runUntilDoneP(),
			nSpaces = ( not isUserCatalog )
						and self._indexStore:queryP 'SELECT count() FROM space_album'
							:runUntilDoneP(),
			nComments = ( not isUserCatalog )
						and self._indexStore:queryP 'SELECT count() FROM comment'
							:runUntilDoneP(),

			refreshOnceAgain_albums = self._docStore:localStateP 'refreshOnce-1.7+/rels/albums'
							:runUntilDoneP(),
            refreshOnceAgain_spaces = self._docStore:localStateP 'refreshOnce-1.7+/rels/catalog_spaces'
							:runUntilDoneP(),
			
			refreshOnce_albums = self._docStore:localStateP 'refreshOnce+/rels/albums'
							:runUntilDoneP(),
            refreshOnce_spaces = self._docStore:localStateP 'refreshOnce+/rels/catalog_spaces'
							:runUntilDoneP(),
		
		}
	
	}

end

--------------------------------------------------------------------------------

return KSClass 'OzCentralSyncAgent'( syncAgentMethods )
