--[[----------------------------------------------------------------------------

OzCentralUploadTask.lua

@author scouten 2011-08-23

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This namespace is used exclusively by OzCentralSyncAgent. It provides code to
upload missing information for a single asset or other document up to Oz.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.diagnostic_log
local WFAppMemoryUsage -- = require 'WFAppMemoryUsage' -- lazy loaded
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'

	-- wichitafoundation.core.lua.threading
local WFIdle = require 'WFIdle'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'

	-- wichitafoundation.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.os.prefs
local KSPrefs = require 'KSPrefs'

	-- wichitafoundation.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'
local OzCatalogUrls = require 'OzCatalogUrls'
local OzOriginalsRecovery = require 'OzOriginalsRecovery'
local OzCatHelperFns = require 'OzCatHelperFns'

	-- wichitafoundation.oz_client.sync_agent
local WFOzDownloadTask = require 'WFOzDownloadTask'

	-- third_party.yajl
local yajl = require 'yajl'

--------------------------------------------------------------------------------

local uploadLog = KSLogging.log 'OzCentralUploadTask'
--	:setLevel( KSLogging.DEBUG )

local yajl_null = yajl.null

local COLLECT_UPLOAD_METRICS = true

local ENABLE_MEMORY_POLLING_FOR_METRICS = COLLECT_UPLOAD_METRICS
	and not _G.DISTRIBUTION
	and WFPlatformInfo.isCocoa

local timeStampAtFirstUpload
local uploadStatusReportBins = {}
local KSTableUtils_deepCopy = KSTableUtils.deepCopy
local KSTableUtils_derefOpt = KSTableUtils.derefOpt


--============================================================================--

-- @mark OzCentralUploadTask namespace

--[[!
	\class OzCentralUploadTask
	\brief This namespace provides the ability to upload documents to the Oz server.

]]--

local OzCentralUploadTask = {}

--------------------------------------------------------------------------------

local onIdleP = WFIdle.onIdleP
	-- FIXME [scouten 2011-08-23]: Got into a special promise hell when I used a
	-- a pre-resolved promise in mapDataP functions. No time to investigate that
	-- tonight but should talk to Mark about why this didn't work. Goal would be
	-- to remove this function and use nopP instead.
	
local uploadRate = 768 * 1024

--------------------------------------------------------------------------------

local function applyOverlayWithUpdatedValues( context, uploadedValues )
	return function()
		local docStatusOverlay = {
			_localOnly = {
				_uploadedValues = uploadedValues,
			}
		}
		
		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline because :applyOverlay has multiple returns

	end

end

--------------------------------------------------------------------------------

local function deleteAssetP( context, docValue, localOverlay )

	uploadLog:debug( "DELETE %s\n%s", context.docUrl )

	local httpOptions = {
		requestMethod = 'DELETE',
		returnHttpErrorsAsData = true,
	}
	
	context.nop = false

	-- Special case for an asset that was deleted as part of
	-- a batch delete.
	if docValue._localOnly.deletedAtOz then
	
		return context.docStore:setServerValueP( 'asset',
											context.docId,
											{ _deleted = true } )
	end

	local deleteP = context.ozUser:getP( context.docUrl, httpOptions )
		
	return deleteP:mapDataP( function( _, data, headers )

		if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then

			return context.docStore:setServerValueP( 'asset',
													 context.docId,
													 { _deleted = true, _ozMetadataUploadStatus = 'uploaded', } )
			
		else
			
			uploadLog:error(
				"Unexpected result code %s on DELETE to %s\n\ndata = %s",
									headers.status_code,
									context.docUrl,
									data )

			return 'error', data, headers

		end
	
	end )

end

--------------------------------------------------------------------------------

local function deleteAlbumP( context, docValue, localOverlay )

	uploadLog:debug( "DELETE %s", context.docUrl )

	local httpOptions = {
		requestMethod = 'DELETE',
		returnHttpErrorsAsData = true,
	}
	
	context.nop = false

	local deleteP = context.ozUser:getP( context.docUrl, httpOptions )
		
	return deleteP:mapDataP( function( _, data, headers )

		if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then

			return context.docStore:setServerValueP( 'album',
													 context.docId,
													 { _deleted = true } )
			
		else
			
			uploadLog:error(
				"Unexpected result code %s on DELETE to %s\n\ndata = %s",
									headers.status_code,
									context.docUrl,
									data )

			return 'error', data, headers

		end
	
	end )

end

--------------------------------------------------------------------------------

local function deleteSpaceP( context, docValue, localOverlay )

	uploadLog:debug( "DELETE %s", docValue )

	local httpOptions = {
		requestMethod = 'DELETE',
		returnHttpErrorsAsData = true,
	}
	
	context.nop = false

	local deleteP = context.ozUser:getP( context.docUrl, httpOptions )
		
	return deleteP:mapDataP( function( _, data, headers )

		if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then

			return context.docStore:setServerValueP( 'space',
													 context.docId,
													 { _deleted = true } )
			
		else
			
			uploadLog:error(
				"Unexpected result code %s on DELETE to %s\n\ndata = %s",
									headers.status_code,
									context.docUrl,
									data )

			return 'error', data, headers

		end
	
	end )

end

--------------------------------------------------------------------------------

local function replaceDeleteWithJsonNull( t )

	for k, v in pairs( t ) do
	
		if type( v ) == 'table' then
			
			if v._delete == true then
				t[ k ] = yajl_null
			else
				replaceDeleteWithJsonNull( v )
			end

		end
	
	end

end

--------------------------------------------------------------------------------

local function removeAlreadyUploadedValues( deltaForUpload, alreadyUploaded )
	
	for k, v in pairs( alreadyUploaded ) do
		
		if type( v ) == 'table' then
			local deltaSubValue = deltaForUpload[ k ]
			if type( deltaSubValue ) == 'table' then
				removeAlreadyUploadedValues( deltaSubValue, v )
				
				if next( deltaSubValue ) == nil then
					deltaForUpload[ k ] = nil
				end
			end
		elseif v == deltaForUpload[ k ] then

			deltaForUpload[ k ] = nil

		end
		
	end
	
end

--------------------------------------------------------------------------------

local function getDeltaForLocalOverlay( localOverlay )
	
	local deltaForUpload = {}
	
	for k, v in pairs( localOverlay ) do
		if not k:match( '^_' ) then
			deltaForUpload[ k ] = v
		end
	end
	
	local alreadyUploaded = localOverlay._localOnly and localOverlay._localOnly._uploadedValues
	
	if type( alreadyUploaded ) == 'table' then
		removeAlreadyUploadedValues( deltaForUpload, alreadyUploaded )
	end

	return deltaForUpload, next( deltaForUpload ) ~= nil
end

--------------------------------------------------------------------------------

local function freeKsData( data )
	return function( ... )
		data:close()
		return ...
	end
end

--------------------------------------------------------------------------------

local function filter201( context, key, uploadUrl, uploadBody, optBoolNilDataNotAnError )

	return function( status, data, headers, ... )

		-- TODO [ arahn 10/05/2012 ] Maybe we can always test for nil data?
		
		if optBoolNilDataNotAnError and data == nil then
			return 'data', nil
		end

		-- If not a 201, something's borked.
	
		if not headers or headers.status_code ~= '201' then

			context.errorCaptureForDoc[ key .. "-putErrorData" ] = tostring( data )
			context.errorCaptureForDoc[ key .. "-putErrorHeaders" ] = headers
			
			local logFmtStr =  "Got status %s back from Oz\n\n"
			local logParams = { headers.status_code }
			if uploadUrl then
				logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
				logParams[ #logParams + 1 ] = uploadUrl
			end
			if uploadBody then
				logFmtStr = logFmtStr .. "what we uploaded in body was:\n\n%s\n\n"
				logParams[ #logParams + 1 ] = uploadBody
			end
			logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = data
			
			uploadLog:error( logFmtStr, unpack( logParams ) )

			return 'error', data, headers

		end

		return status, data, headers, ...

	end
	
end

--------------------------------------------------------------------------------

local function handleNothingMeaningful( context, docValue )

	uploadLog:debug( "nothing to upload for %s ID %s", docValue._docType, docValue._docId )

	if docValue._ozMetadataUploadStatus == 'uploaded' then
	
		uploadLog:debug( "returning on idleP for doc ID %s", docValue._docId )
	
		return onIdleP()
	
	else

		-- FIXME [scouten 2011-06-02]: Setting _ozMetadataUploadStatus
		-- from here is evil. It should go away.
		
		uploadLog:debug( "setting metadata status for doc ID %s", docValue._docId )

		context.nop = false

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = {
				_ozMetadataUploadStatus = 'uploaded',
			},
		}
		
		return resultP -- can't inline because :applyOverlay has multiple returns
	
	end

end

--------------------------------------------------------------------------------

local function getJsonPutArgs( optRequestMethod )

	return {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
		requestMethod = optRequestMethod,
		returnHttpErrorsAsData = true,
	}

end

--------------------------------------------------------------------------------

local function uploadAssetMetadataP( context, docValue, localOverlay )

	-- [PR #2942322]: Don't attempt another upload if we haven't received
	-- confirmation of the previous one.

	uploadLog:debug( "uploadAssetMetadataP %s", docValue._docId )

	--Dont want to send the sha
	if docValue and docValue.importSource then
		docValue.importSource.sha256 = nil 
	end

	-- Generate new rev ID for this upload.

	local newRevId = KSIdGenerator.uuidInLowercaseHex()
	
	local childRevisionTemplate = KSTableUtils_derefOpt(
			docValue,
			'_serverProvidedMetadata',
			'link_templates',
			'/rels/child_revision',
			'href' )
	local putUrl
	
	if childRevisionTemplate then
		putUrl = string.gsub( childRevisionTemplate, "{revision_id}", newRevId )
		putUrl = context.baseUrl .. putUrl
	else
		putUrl = string.format( '%s/revisions/%s', context.docUrl, newRevId )
	end
	
	local needToBumpDevSettingsRevId

	-- Keep track of when we last pushed dev settings (or original upload).

	if not docValue._serverProvidedMetadata
		or( localOverlay and localOverlay.develop )
	then
		needToBumpDevSettingsRevId = true
	end

	uploadLog:debug( "link template for photo ID %s = %s", docValue._docId, childRevisionTemplate )

	-- Strip out Wichita's chatter. What remains in the local
	-- overlay is what we need to upload.

	local deltaForUpload, hasMeaningfulUpload = getDeltaForLocalOverlay( localOverlay )

	-- If nothing left, just mark it uploaded and be done.
	
	if not hasMeaningfulUpload then
		
		return handleNothingMeaningful( context, docValue )

	end
	
	context.nop = false
	context.newRevId = newRevId
	
	-- Remember what values we sent
	local uploadedValues = KSTableUtils_deepCopy( deltaForUpload )

	-- Send the local overlay.

	replaceDeleteWithJsonNull( deltaForUpload )

	local uploadContent = {
		user_id = context.ozUser:id(),
		payload = deltaForUpload,
	}
	
	-- Upload metadata for this revision.
	
	local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )

	context.errorCaptureForDoc.metadataPutUrl = putUrl
	context.errorCaptureForDoc.metadataPutContent = jsonData

	uploadLog:debug( "PUT %s\n%s", putUrl, jsonData )

	local putArgs = getJsonPutArgs()
	
	local uploadP = context.ozUser:putP( putUrl, jsonData, putArgs )
	
	return uploadP
		:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP )
		:mapDataP( function( _, data, headers )
			if headers.status_code == '403' and data.code == 1002 then
				
				debug.print( "We attempted to upload a duplicate asset: ", localOverlay and uploadContent and jsonData )
				
			end
		end )
		:mapDataP( filter201( context, "metadata", putUrl, jsonData ) )
		:mapDataP( function( _, data, headers )
		
		-- [PR #2942322]: Record the revision ID that we're pushing
		-- so we can handle that differently when we get the expected
		-- :setServerValue call later.
		
		-- Calculate new doc status overlay to reflect
		-- what we've just uploaded.
		
		local docStatusOverlay = {
			_ozCentralRevisionId = newRevId,
		}
		local localOnly = {
			_uploadedValues = uploadedValues,
			freemiumUploadEmbargoUntil = { _delete = true },
		}

		docStatusOverlay._localOnly = localOnly

		if needToBumpDevSettingsRevId then

			localOnly.devSettingsLastSetAtRev = newRevId

			local function tweakRevIdForLocalRendition( renditionName )
			
				local incomingLocalOnly = docValue._localOnly

				local renditionStruct = incomingLocalOnly
									and incomingLocalOnly[ renditionName ]

				if renditionStruct and renditionStruct.revId == 'pending' then
				
					localOnly[ renditionName ] = localOnly[ renditionName ] or {}
					localOnly[ renditionName ].revId = newRevId
				
				end
			
			end

			tweakRevIdForLocalRendition( 'thumbnail' )
			tweakRevIdForLocalRendition( 'thumbnail2x' )
			tweakRevIdForLocalRendition( 'preview' )

			uploadLog:debug(
				"bumped dev settings for rev ID: local overlay = %s",
					newRevId )
		
		else
		
			uploadLog:debug( "don't need to bump dev settings" )

		end
		
		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline because :applyOverlay has multiple returns

	end )

end

--------------------------------------------------------------------------------

local function uploadAssetRenditionP( context, docValue, localOverlay,
								      localRenditionName, ozRenditionName )

	-- Do we need to actually upload this rendition?

	uploadLog:debug( "uploadAssetRenditionP photoId = %s renditionName = %s",
					 docValue._docId, localRenditionName )

	local localOnly = docValue._localOnly or {}

	local whichRevision = localOnly.devSettingsLastSetAtRev
	local wasPending = whichRevision == 'pending'
			
	if whichRevision == 'pending' or not whichRevision then
		whichRevision = context.newRevId
	end

	if whichRevision == 'pending' or not whichRevision then
		whichRevision = docValue._ozCentralRevisionId
		uploadLog:warn( "Unable to find latest rev ID to attach rendition to, falling back to current rev ID" )
	end
	
	if whichRevision == 'pending' or not whichRevision then
	
		local currentRevLink = docValue._serverProvidedMetadata
						   and docValue._serverProvidedMetadata.links
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ]
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ].href

		whichRevision = currentRevLink and currentRevLink:sub( -32 )
		
	end

	if whichRevision == 'pending' or not whichRevision then

		uploadLog:error( "attempting to upload photo %s, but whichRevision is 'pending': docValue = %s",
						 context.docId, docValue )

		error( string.format( "attempting to upload photo %s, but whichRevision is 'pending'",
							  context.docId ) )
			-- FIXME [scouten 2011-08-01]: If you see this, please turn up the
			-- debug log level in OzCatalogUploadStatus.lua (uncomment the
			-- setLevel command at line 37) and send the log to me. Thanks!

	end
	
	local serverLinks = ( docValue._serverProvidedMetadata
					  and docValue._serverProvidedMetadata.links ) or {}

	if not wasPending
		and not OzCatalogUploadStatus.shouldUploadRendition( whichRevision,
									   localOnly[ localRenditionName ],
									   serverLinks[ "/rels/rendition_type/" .. ozRenditionName ] )
	then
	
		-- Rendition appears to be already up on Oz. Don't upload again.
	
		uploadLog:debug( "uploadAssetRenditionP photoId = %s renditionName = %s -- nothing to upload\n\ndocValue = %s",
						 docValue._docId, localRenditionName, docValue )

		return onIdleP()

	end

	-- Where will we put this rendition?
	
	local renditionId = KSIdGenerator.uuidInLowercaseHex()

	local renditionPutUrl = string.format( '%s/revisions/%s/renditions/%s',
										   context.docUrl,
										   whichRevision,
										   renditionId )

	uploadLog:debug( "asking for rendition for %s\n  :: size = %s\n\n\n\n",
					 renditionPutUrl,
					 localRenditionName )

	-- Fetch the data for the rendition.

	local renditionMD5 = docValue._localOnly
					 and docValue._localOnly[ localRenditionName ]
					 and docValue._localOnly[ localRenditionName ].md5

	if not renditionMD5 then
		return onIdleP()
	end
	
	local dataCache = context.docStore:dataCache()
	
	local renditionDataP = dataCache:getP( renditionMD5 )
		:mapP( function( status, data )
		
			if status == 'data' and data and KSData.length( data )then

				return 'data', data

			else

				uploadLog:warn( "No data in cache for %s for photo ID %s. Huh.",
								localRenditionName, context.docId )

				-- Remove the now-bogus link from doc DB.

				local docStatusOverlay = {
					_localOnly = {
						[ localRenditionName ] = { md5 = 'missing' },
					},
				}
		
				local resultP = context.docStore:applyOverlayToDocumentP {
					docType = context.docType,
					docId = context.docId,
					overlay = docStatusOverlay,
				}
				resultP:runUntilDone()
				
				-- Abort out of this part of the upload process.
				
				return 'data', nil
				
			end
		
		end )
	
	-- Got the data, now upload it.
	
	local uploadDataP = renditionDataP:mapDataP( function( _, data )

		if data == nil then
			return 'data', nil
		end

		uploadLog:debug( "Attempting to PUT %s", renditionPutUrl )
		
		context.nop = false

		local renditionPutArgs = {
			requestHeaders = {
				{
					header = 'Content-Type',
					value = 'image/jpeg'
				},
				{
					header = 'Accept',
					value = 'application/json'
				},
				{
					header = 'X-Rendition-Type',
					value = ozRenditionName,
				},
			},
		}
		
		context.errorCaptureForDoc[ localRenditionName .. '-putUrl' ] = renditionPutUrl

		local resultP = context.ozUser:putP( renditionPutUrl, data, renditionPutArgs )
		return resultP -- can't inline this because :putP can return two results
	
	end )
	
	-- Assuming data uploaded, record our success.
	
	local updateDocStatusP = uploadDataP
		:mapDataP( filter201( context, localRenditionName, renditionPutUrl, nil, true ) )
		:mapDataP( function( _, data, headers )

		if data == nil then
			return 'data', nil
		end

		local rels_imageSize = string.format( "/rels/rendition_type/%s", ozRenditionName )

		local partialRenditionUrl = renditionPutUrl:gsub( "^.*//([^/]+", "" )
			-- Fielding has progressed from killing kittens to
			-- mountain lions.
		
		local docStatusOverlay = {
			_localOnly = {
				[ localRenditionName ] = {
					renditionId = renditionId,
				},
			},
			_serverProvidedMetadata = {
				links = {
					[ rels_imageSize ] = {
						href = partialRenditionUrl,
						invalid = { _delete = true },
					},
				},
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline since there are multiple return values

	end )
	
	-- did we updload a sub-res rendition? If so, ask Oz to generate a full res rendition on our behalf
	local resultP = updateDocStatusP
	local ozUpdateRendition = docValue
						  and docValue._localOnly
						  and docValue._localOnly[ localRenditionName ]
						  and docValue._localOnly[ localRenditionName ].ozUpdateRendition
	
	if ozUpdateRendition then
	
		local updateRenditionPostUrl = string.format( '%s/revisions/%s/renditions/%s',
														context.docUrl,
														whichRevision,
														KSIdGenerator.uuidInLowercaseHex() )

		context.errorCaptureForDoc[ localRenditionName .. '-updateRenditionPostUrl' ] = updateRenditionPostUrl

		local ozUpdateRenditionP = updateDocStatusP:mapDataP( function( status )
		
			local updateRenditionPostArgs = {
				requestHeaders = {
					{
						header = 'Accept',
						value = 'application/json'
					},
					{
						header = 'X-Rendition-Type',
						value = ozRenditionName
					},
					{
						header = 'X-Rendition-Width',
						value = ozUpdateRendition.width
					},
					{
						header = 'X-Rendition-Height',
						value = ozUpdateRendition.height
					},
				},
				requestMethod = 'POST',
				returnHttpErrorsAsData = true,
			}
		
			uploadLog:debug( "Attempting to POST %s", updateRenditionPostUrl )
			
			local updateRenditionPostP = context.ozUser:putP( updateRenditionPostUrl, nil, updateRenditionPostArgs )
			return updateRenditionPostP
			
		end )
		
		-- Assuming post succeeded, update overlay
		
		resultP = ozUpdateRenditionP
		:mapDataP( filter201( context, localRenditionName, updateRenditionPostUrl ) )
		:mapDataP( function( _, data, headers)

			local rels_imageSize = string.format( "/rels/rendition_type/%s", ozRenditionName )

			local partialRenditionUrl = updateRenditionPostUrl:gsub( "^.*//([^/]+", "" )
			-- Fielding has progressed from killing kittens to
			-- mountain lions.
		
			local docStatusOverlay = {
				_localOnly = {
					[ localRenditionName ] = {
						ozUpdateRendition = { _delete = true },
					},
				},
				_serverProvidedMetadata = {
					links = {
						[ rels_imageSize ] = { href = partialRenditionUrl },
					},
				},
			}

			local applyOverlayP = context.docStore:applyOverlayToDocumentP {
				docType = context.docType,
				docId = context.docId,
				overlay = docStatusOverlay,
			}
			
			return applyOverlayP -- can't inline since there are multiple return values
		
		end )
		
	end

	return resultP

end

--------------------------------------------------------------------------------

local function shouldUploadOriginal( docValue )

	-- FIXME [scouten 2011-08-23]: When originals can change (after V1),
	-- we will need to revisit this algorithm to ensure that we upload
	-- newer originals.

	-- If we don't have original stored locally, then nothing to upload.
	
	local localOnly = docValue._localOnly or {}
	
	if not( localOnly.originalFilePath or localOnly.copyMasterFrom ) then
		return false
	end
	
	-- Does Oz think it has the original?
	
	local serverLinks = docValue._serverProvidedMetadata
					and docValue._serverProvidedMetadata.links
					 or {}
	
	return not serverLinks[ '/rels/master' ]

end

--------------------------------------------------------------------------------

local function uploadAssetOriginalViaCopyP( context, docValue, localOverlay )

	local localOnly = docValue._localOnly or {}

	local postBodyJson = KSJsonEncoder.encodeOzUploadJson {
		file_href = docValue._localOnly.copyMasterFrom,
	}
	
	local ksData = KSData.make( postBodyJson )

	-- Calculate revision to attach the master to.
		-- FIXME [scouten 2011-08-23]: Should attach to initial revision, but we
		-- don't always have a way to know what that is.

	local whichRevision = localOnly.devSettingsLastSetAtRev
			
	if whichRevision == 'pending' or not whichRevision then
		whichRevision = context.newRevId
	end

	if whichRevision == 'pending' or not whichRevision then
		whichRevision = docValue._ozCentralRevisionId
		uploadLog:warn( "Unable to find first rev ID to attach master to, falling back to current rev ID" )
	end

	if whichRevision == 'pending' or not whichRevision then

		error( "attempting to upload photo %s, but whichRevision is 'pending'", context.docId )
			-- FIXME [scouten 2011-08-01]: If you see this, please turn up the
			-- debug log level in OzCatalogUploadStatus.lua (uncomment the
			-- setLevel command at line 37) and send the log to me. Thanks!

	end

	-- TODO [arahn 2013-04-04]: USE_LINK_TEMPLATES server should provide this URL
	local masterPutUrl = string.format( '%s/%ss/%s/revisions/%s/master',
										context.catalogUrl,
										context.docType,
										context.docId,
										whichRevision )

	-- Got the data, now upload it.

	uploadLog:debug( "Attempting to PUT %s as a copy\n%s",
						masterPutUrl, postBodyJson )

	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = 'application/json'
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
		},
		returnHttpErrorsAsData = true,
	}
	
	local uploadDataP = context.ozUser:putP( masterPutUrl, ksData, masterPutArgs )
	
	-- Assuming data uploaded, record our success.

	context.nop = false
	
	local updateDocStatusP = uploadDataP
	:mapP( freeKsData( ksData ) )
	:mapDataP( filter201( context, "assetOriginalViaCopy", masterPutUrl ) )
	:mapDataP( function( _, data, headers )

		local partialMasterUrl = masterPutUrl:gsub( "^.*//([^/]+", "" )
		
		local docStatusOverlay = {
			_localOnly = {
				copyMasterFrom = { _delete = true },
				original = { ozUrl = partialMasterUrl },
			},
			_serverProvidedMetadata = {
				links = {
					[ '/rels/master' ] = { href = partialMasterUrl },
				},
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline since there are multiple return values

	end )

	return updateDocStatusP

end

--------------------------------------------------------------------------------

local function shouldUploadProxy( docValue )

	-- FIXME [sdaniel 2012-02-1]: When originals can change (after V1),
	-- we will need to revisit this algorithm to ensure that we upload
	-- newer proxies. See shouldUploadOriginal().

	-- If we don't have proxy stored locally, then nothing to upload.
	
	local localOnly = docValue._localOnly or {}
	
	if not ( localOnly.proxy and localOnly.proxy.filePath ) then
		return false
	end
	
	-- Does Oz think it has the original?
	
	local serverLinks = docValue._serverProvidedMetadata
					and docValue._serverProvidedMetadata.links
					 or {}
	
	return not serverLinks[ '/rels/proxy_type/2048' ]

end

--------------------------------------------------------------------------------

local function uploadAssetProxyP( context, docValue, localOverlay )
	
	if not shouldUploadProxy( docValue ) then
		return onIdleP()
	end

	uploadLog:debug( "uploadAssetProxyP photoId = %s", docValue._docId )
	
	local localOnly = docValue._localOnly
	
	local proxyFilePath = assert( localOnly.proxy and localOnly.proxy.filePath )
	
	proxyFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( proxyFilePath )
	
	if not KSFileUtils.exists( proxyFilePath ) then

		-- Remove the now-bogus link from doc DB.

		local docStatusOverlay = {
			_localOnly = {
				proxy = {
					filePath = 'missing',
				},
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		resultP:runUntilDone()
		
		return onIdleP()
	
	end
	
	local ksData = KSData.memoryMappedContentsOfFile( proxyFilePath )
	
	local whichRevision = localOnly.devSettingsLastSetAtRev
			
	if whichRevision == 'pending' or not whichRevision then
		whichRevision = context.newRevId
	end

	if whichRevision == 'pending' or not whichRevision then
		whichRevision = docValue._ozCentralRevisionId
		uploadLog:warn( "Unable to find first rev ID to attach master to, falling back to current rev ID" )
	end

	if whichRevision == 'pending' or not whichRevision then
		
		local currentRevLink = docValue._serverProvidedMetadata
						   and docValue._serverProvidedMetadata.links
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ]
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ].href

		whichRevision = currentRevLink and currentRevLink:sub( -32 )

		uploadLog:warn( "Unable to find first rev ID or even local current rev ID to attach master to, falling back to Oz-provided current rev ID" )

	end

	if whichRevision == 'pending' or not whichRevision then

		error(
			string.format(
				"attempting to upload photo %s, but whichRevision is %s", context.docId ) )

			-- FIXME [scouten 2011-08-01]: If you see this, please turn up the
			-- debug log level in OzCatalogUploadStatus.lua (uncomment the
			-- setLevel command at line 37) and send the log to me. Thanks!

	end

	-- TODO [arahn 2013-04-04]: USE_LINK_TEMPLATES server should provide this URL
	local proxyPutUrl = string.format( '%s/%ss/%s/revisions/%s/proxies/%s',
										context.catalogUrl,
										context.docType,
										context.docId,
										whichRevision,
										KSIdGenerator.uuidInLowercaseHex() )
										
	-- Got the data, now upload it.

	context.errorCaptureForDoc[ 'proxy-putUrl' ] = proxyPutUrl

	uploadLog:debug( "Attempting to PUT %s", proxyPutUrl )
	
	local proxyPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = 'image/dng'
			},
			{
				header = 'X-Proxy-Type',
				value = '2048'
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
		},
		returnHttpErrorsAsData = true,
	}
	
	local uploadDataP = context.ozUser:putP( proxyPutUrl, ksData, proxyPutArgs )
	
	-- Assuming data uploaded, record our success.
	context.nop = false
	
	local updateDocStatusP = uploadDataP
	:mapP( freeKsData( ksData ) )
	:mapDataP( filter201( context, "proxy", proxyPutUrl ) )
	:mapDataP( function( _, data, headers )

		-- Move the proxy from documents to caches.
		
		local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
		local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'
						
		local proxyFilePath_dirSub = proxyFilePath:sub( 1, #documentsDir )
		local newFilePath = proxyFilePath
			
		if proxyFilePath_dirSub == documentsDir then
			newFilePath = string.format( "%s%s", cachesDir, proxyFilePath:sub( #documentsDir + 1 ) )
		end
						
		-- WARNING [scouten 2012-05-15]: Need to add WFOzFileLock logic here if we move sync
		-- logic out of the BABLU imp thread. For now, it's OK as is.

		KSFileUtils.mkdir( KSPathUtils.parent( newFilePath ) )
			
		newFilePath = KSFileUtils.moveToUniquelyNamedFile( proxyFilePath, newFilePath )

		-- Update the file path and the location of the file.
			
		local filePath_proxy = newFilePath

		local partialProxyUrl = proxyPutUrl:gsub( "^.*//([^/]+", "" )
		
		local docStatusOverlay = {
			_localOnly = {
				proxy = {
					ozUrl = partialProxyUrl,
					filePath = filePath_proxy,
				 },
			},
			_serverProvidedMetadata = {
				links = {
					[ '/rels/proxy_type/2048' ] = { href = partialProxyUrl },
				},
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline since there are multiple return values
	end )

	return updateDocStatusP

end

--------------------------------------------------------------------------------

local function recoverOriginalAssetP( context, docValue, localOverlay )
	return OzOriginalsRecovery.recoveryEffortOriginalP( context.docStore, docValue )
end

--------------------------------------------------------------------------------

local function uploadAssetOriginalP( context, docValue, localOverlay )

	uploadLog:debug( "uploadAssetOriginalP photoId = %s", docValue._docId )

	-- Do we need to actually upload this original?
		-- FIXME [scouten 2011-08-23]: When originals can change (after V1),
		-- we will need to revisit this algorithm to ensure that we upload
		-- newer originals.

	if not shouldUploadOriginal( docValue ) then
		return onIdleP()
	end
	
	-- Are we copying from another pre-existing original?
	
	local localOnly = docValue._localOnly or {}
	
	if localOnly.copyMasterFrom then
		return uploadAssetOriginalViaCopyP( context, docValue, localOverlay )
	end
	
	-- Nope, upload the file then.
	
	local originalFilePath = assert( localOnly.originalFilePath )
	
	-- Temporary check to correct the local file path and replace
	-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
	-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
	-- that updates the database with the correct file paths but this is not
	-- feasible in the current time frame.
	
	originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
	
	if not KSFileUtils.exists( originalFilePath ) then

		-- Remove the now-bogus link from doc DB.

		local docStatusOverlay = {
			_localOnly = {
				originalFilePath = 'missing',
				recoveryEffort = 'original',
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		resultP:runUntilDone()
		
		return onIdleP()
	
	end

	local ksData = KSData.memoryMappedContentsOfFile( originalFilePath )
		-- FIXME [scouten 2011-06-02]: Verify SHA-256 hash before uploading?
		-- GRUNGE [ arahn 2012-08-22]: Using MMAP means we can't relocate this
		-- file on Windows so long as the KSData is alive.  We must remember
		-- to close this before trying to move the file!
		
	-- Calculate revision to attach the master to.
		-- FIXME [scouten 2011-08-23]: Should attach to initial revision, but we
		-- don't always have a way to know what that is. (Actually, now we do...
		-- courtesy of the first_revision link.)

	-- FIXME [scouten 2011-09-01]: This which revision logic should be aiming for
	-- the *first* revision, not current. Only will be relevant when we start
	-- having changing masters, so not changing today.

	local whichRevision = localOnly.devSettingsLastSetAtRev
			
	if whichRevision == 'pending' or not whichRevision then
		whichRevision = context.newRevId
	end

	if whichRevision == 'pending' or not whichRevision then
		whichRevision = docValue._ozCentralRevisionId
		uploadLog:warn( "Unable to find first rev ID to attach master to, falling back to current rev ID" )
	end

	if whichRevision == 'pending' or not whichRevision then
		
		local currentRevLink = docValue._serverProvidedMetadata
						   and docValue._serverProvidedMetadata.links
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ]
						   and docValue._serverProvidedMetadata.links[ "/rels/latest_revision" ].href

		whichRevision = currentRevLink and currentRevLink:sub( -32 )

		uploadLog:warn( "Unable to find first rev ID or even local current rev ID to attach master to, falling back to Oz-provided current rev ID" )

	end

	if whichRevision == 'pending' or not whichRevision then

		error(
			string.format(
				"attempting to upload photo %s, but whichRevision is %s", context.docId ) )

			-- FIXME [scouten 2011-08-01]: If you see this, please turn up the
			-- debug log level in OzCatalogUploadStatus.lua (uncomment the
			-- setLevel command at line 37) and send the log to me. Thanks!

	end

	-- TODO [arahn 2013-04-04]: USE_LINK_TEMPLATES server should provide this URL
	local masterPutUrl = string.format( '%s/%ss/%s/revisions/%s/master',
										context.catalogUrl,
										context.docType,
										context.docId,
										whichRevision )

	-- Got the data, now upload it.

	context.errorCaptureForDoc[ 'original-putUrl' ] = masterPutUrl

	uploadLog:debug( "Attempting to PUT %s", masterPutUrl )
	local masterContentType = 'image/jpeg'
	if docValue.importSource and docValue.importSource.contentType then
		masterContentType = docValue.importSource.contentType
	end
	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = masterContentType
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
		},
		returnHttpErrorsAsData = true,
	}
	
	local fileSize = docValue
					 and docValue.importSource
					 and docValue.importSource.fileSize
	if type( fileSize ) == 'number' and fileSize > 0 then

		masterPutArgs.timeout = math.max( 60, ( fileSize * 8 ) / uploadRate )
		
		uploadLog:debug( "uploadAssetOriginalP calculated timeout value of %.2f for file size %d",
								  masterPutArgs.timeout, fileSize )

	end
	
	local uploadDataP = context.ozUser:putP( masterPutUrl, ksData, masterPutArgs )
	
	-- Assuming data uploaded, record our success.
	
	context.nop = false
	
	local updateDocStatusP = uploadDataP
	:mapP( freeKsData( ksData ) )
	:mapDataP( filter201( context, "original", masterPutUrl ) )
	:mapDataP( function( _, data, headers )

		-- If Store Originals is not on, move the uploaded original from
		-- the Documents directory to the Caches directory.
		
		local downloadOriginals = KSPrefs[ 'DownloadOriginals' ]
		
		local filePath_original = originalFilePath
		local originalIsInCacheFolder
		
		if not downloadOriginals then

			-- Move the original from documents to caches.
		
			local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
			local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'
						
			local originalFilePath_dirSub = originalFilePath:sub( 1, #documentsDir )
			local newFilePath = originalFilePath
			
			if originalFilePath_dirSub == documentsDir then
				newFilePath = string.format( "%s%s", cachesDir, originalFilePath:sub( #documentsDir + 1 ) )
			end
						
			-- WARNING [scouten 2012-05-15]: Need to add WFOzFileLock logic here if we move sync
			-- logic out of the BABLU imp thread. For now, it's OK as is.

			KSFileUtils.mkdir( KSPathUtils.parent( newFilePath ) )
			
			newFilePath = KSFileUtils.moveToUniquelyNamedFile( originalFilePath, newFilePath )

			-- Update the file path and the location of the file.
			
			filePath_original = newFilePath
			originalIsInCacheFolder = true

		else
			
			-- This may not be relevant any more, but if we're leaving the file in the
			-- Documents folder, then make sure it is not eligible for backup.

			KSFileUtils.setDoNotBackUpFlag( originalFilePath )
		
		end
		
		local partialMasterUrl = masterPutUrl:gsub( "^.*//([^/]+", "" )
		
		local docStatusOverlay = {
			_localOnly = {
				original = {
					ozUrl = partialMasterUrl,
					isInCacheFolder = originalIsInCacheFolder,
					doNotBackUpFlagSet = true,
				 },
				originalFilePath = filePath_original,
			},
			_serverProvidedMetadata = {
				links = {
					[ '/rels/master' ] = { href = partialMasterUrl },
				},
			},
		}

		local resultP = context.docStore:applyOverlayToDocumentP {
			docType = context.docType,
			docId = context.docId,
			overlay = docStatusOverlay,
		}
		
		return resultP -- can't inline since there are multiple return values

	end )

	return updateDocStatusP

end

--------------------------------------------------------------------------------

-- FIXME [ngautam 2012-05-14] Per Eric, need to consolidate the upload implementations
-- for Asset and Album documents.

local function uploadAlbumP( context, docValue, localOverlay )

	uploadLog:debug( "uploadAlbumP albumId = %s %s", docValue._docId, docValue)

	local putUrl = context.docUrl

	local deltaForUpload, hasMeaningfulUpload = getDeltaForLocalOverlay( localOverlay )

	-- If nothing left, just mark it uploaded and be done.
	
	if not hasMeaningfulUpload then

		return handleNothingMeaningful( context, docValue )

	end
	
	uploadLog:debug( "have things to upload for album ID %s %s", docValue._docId, deltaForUpload )

	context.nop = false
	-- Send the local overlay.

	replaceDeleteWithJsonNull( deltaForUpload )

	local uploadContent = deltaForUpload
		
	local requestMethod = 'PUT'
	-- If the album has a created field, it already exists on Oz and this is a rename request.
	if docValue.created then
		requestMethod = 'POST'
	end
	
	local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )

	uploadLog:debug( "%s %s\n%s", requestMethod, putUrl, jsonData )

	local putArgs = getJsonPutArgs( requestMethod )
	
	local uploadP = context.ozUser:putP( putUrl, jsonData, putArgs )
	
	return uploadP:mapDataP( function( _, data, headers )

		-- Fix @WATSON #3196038 -  Conflict resolve:  rename an album while delete the SAME album on a different device.
		-- Need to watch out for a 404 here. If the album has been deleted and we happen to try and update it at the same time,
		-- the sync cycle can get stuck waiting for the upload to complete (never will since the album just got deleted)
		-- we never download the tombstone because we're stuck trying to upload first.
		
		if headers and headers.status_code == '404' then
		
			return context.docStore:setServerValueP( 'album', docValue._docId, { _deleted = true } )

		end

		-- Expect a 201 for an album create, 204 for album update.
			
		if not headers or not ( headers.status_code == '201' or headers.status_code == '204' ) then
		
			uploadLog:error( "Got status %s back from Oz\n\nwhat we uploaded in body was:\n\n%s\n\nHTML response body was:\n\n%s\n\n",
							 headers.status_code, jsonData, data )

			return 'error', data, headers

		end
		
		-- [scouten 2012-08-17, PR# 3317838]: Force immediate reload of the document
		-- so we get album state from Oz.
		
		return WFOzDownloadTask.reloadDocumentP( context,
												 context.docType,
												 context.docId,
												 context.docUrl )

	end )

end

--------------------------------------------------------------------------------

local function commentUrl( catalogUrl, localOnly, docId )

	-- TODO [ arahn 10/08/2012 ] This Oz api is not rooted by a catalog path, and the context doesn't
	-- have the base url anywhere, so it is a total hack that I split off the catalog part of this URL
	-- catalogUrl is something like https://api.oznext.com/v1.0/catalogs/597c1a08b9c24082987af24a201b6fef

	local _, _, baseUrl = string.find( catalogUrl, "(.+)/catalogs/[0-9a-f]+" )
	
	if not baseUrl then
		return "error", "Can't parse catalogUrl"
	end

	local spaceId = localOnly.spaceId
	if spaceId then
		return string.format( '%s/spaces/%s/assets/%s/comments/%s', baseUrl, spaceId, localOnly.assetId, docId )
	else
		return string.format( '%s/assets/%s/comments/%s', catalogUrl, localOnly.assetId, docId )
	end

end

--------------------------------------------------------------------------------

local function uploadCommentP( context, docValue, localOverlay )

	uploadLog:debug( "uploadCommentP %s id = %s %s", docValue._docType, docValue._docId, docValue)

	local putUrl = commentUrl( context.catalogUrl, localOverlay._localOnly, context.docId )
	context.docUrl = putUrl

	local deltaForUpload, hasMeaningfulUpload = getDeltaForLocalOverlay( localOverlay )

	-- If nothing left, just mark it uploaded and be done.
	
	if not hasMeaningfulUpload then
		
		return handleNothingMeaningful( context, docValue )
		
	end
	
	uploadLog:debug( "have things to upload for %s ID %s %s", docValue._docType, docValue._docId, deltaForUpload )
	
	context.nop = false
	-- Send the local overlay.
	
	local uploadedValues = KSTableUtils_deepCopy( deltaForUpload )
	replaceDeleteWithJsonNull( deltaForUpload )

	local uploadContent = deltaForUpload
		
	local requestMethod = 'PUT'

	-- If the album has a created field, it already exists on Oz
	-- Since comments are not editable, we are done.
	if docValue.created then

		-- If we've already uploaded this, since comments are immutable, just stamp in the
		-- the value to prevent an infinite loop (which is what happens if we error here).
		docValue._ozMetadataUploadStatus = 'uploaded'
		return context.docStore:setServerValueP( context.docType,
												 context.docId,
												 docValue )
	end
	
	local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )
	
	uploadLog:debug( "%s %s\n%s", requestMethod, putUrl, jsonData )

	local putArgs = getJsonPutArgs( requestMethod )
	
	local uploadP = context.ozUser:putP( putUrl, jsonData, putArgs )
	
	return uploadP
	:mapDataP( function( status, data, headers )

		if headers and headers.status_code == '404' then
			
			local result = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )

			if result.code == 1000 then
				
				-- If Oz cannot accept this comment then we have no choice but to delete it
				return context.docStore:setServerValueP(
					context.docType,
					docValue._docId,
					{ _deleted = true }
				)
				:mapP( function()
					return status, data, headers
				end )
				
			end

		end
		
	end )
	:mapDataP( filter201( context, "comment", putUrl, jsonData ) )
	:mapDataP( applyOverlayWithUpdatedValues( context, uploadedValues ) )

end

--------------------------------------------------------------------------------

local function deleteCommentP( context, docValue, localOverlay )

	local putPath = docValue and docValue.links and docValue.links.self and docValue.links.self.href
	local _, _, baseUrl = string.find( context.catalogUrl, "(https://[^/]+)/" )
	
	if not putPath or not baseUrl then
		return context.docStore:setServerValueP( context.docType,
												 context.docId,
												 { _deleted = true } )
	end
	
	local putUrl = OzCatHelperFns.checkForRelativeUrlAndAppendBase( putPath, OzCatalogUrls.baseServerUrl )

	uploadLog:debug( "DELETE %s", putUrl )

	local httpOptions = {
		requestMethod = 'DELETE',
		returnHttpErrorsAsData = true,
	}
	
	context.nop = false

	local deleteP = context.ozUser:getP( putUrl, httpOptions )
		
	return deleteP:mapDataP( function( _, data, headers )

		if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then

			return context.docStore:setServerValueP( context.docType,
													 context.docId,
													 { _deleted = true } )
			
		else
			
			uploadLog:error(
				"Unexpected result code %s on DELETE to %s\n\ndata = %s",
									headers.status_code,
									putUrl,
									data )

			return 'error', data, headers

		end
	
	end )

end

--------------------------------------------------------------------------------

local function uploadSpaceP( context, docValue, localOverlay )

	uploadLog:debug( "uploadSpaceP spaceId = %s %s", docValue._docId, docValue )

	local putUrl = context.docUrl

	local deltaForUpload, hasMeaningfulUpload = getDeltaForLocalOverlay( localOverlay )

	-- If nothing left, just mark it uploaded and be done.
	
	if not hasMeaningfulUpload then

		uploadLog:debug( "nothing to upload for space ID %s", docValue._docId )

		if docValue._ozMetadataUploadStatus == 'uploaded' then
		
			uploadLog:debug( "returning on idleP space ID %s", docValue._docId )
		
			return onIdleP()
		
		else

			-- FIXME [scouten 2011-06-02]: Setting _ozMetadataUploadStatus
			-- from here is evil. It should go away.
			
			uploadLog:debug( "setting metadata status for space ID %s", docValue._docId )

			context.nop = false
	
			-- FIXME [scouten 2012-08-04, PR #3227713]: Calling setServerValueP
			-- is even more evil, but it hacks around this particular problem.
			-- Relying on reading back the value from the server is the root problem.
			-- Gotta fix that SOON!
	
			local simulatedServerValue = {}
			
			for k, v in pairs( docValue ) do
			
				if not k:match( '^_' ) then
					simulatedServerValue[ k ] = v
				end
			
			end
	
			return context.docStore:setServerValueP( context.docType,
													 context.docId,
													 simulatedServerValue )
	
		end

	end
	
	uploadLog:debug( "have things to upload for space ID %s %s", docValue._docId, deltaForUpload )

	context.nop = false

	-- Send the local overlay.
		
	local requestMethod

	-- If the space has a created field, it already exists on Oz and this is a update request.

	if docValue.created then

		requestMethod = 'POST'

	else

		requestMethod = 'PUT'

		-- Add the albums section to the overlay.

		deltaForUpload.albums = {
			{
				_id = docValue._localOnly.album.albumId,
				space_album_id = docValue._localOnly.album.space_album_id,
			},
		}

	end
	
	replaceDeleteWithJsonNull( deltaForUpload )

	local uploadContent = deltaForUpload
	local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )

	uploadLog:debug( "%s %s\n%s", requestMethod, putUrl, jsonData )

	local putArgs = getJsonPutArgs( requestMethod )
	
	if requestMethod == 'PUT' then
		putArgs.requestHeaders[ #putArgs.requestHeaders + 1 ] = {
			header = 'If-None-Match', value = docValue._localOnly.album.albumId
		}
	end
	
	local uploadP = context.ozUser:putP( putUrl, jsonData, putArgs )
	
	return uploadP:mapDataP( function( status, data, headers )
			
		if headers and headers.status_code == '412' then
			
			-- We will get a 412 with an error code of 1028 if the album is already part of another space.

			local result = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )

			if result and result.code == 1028 then
				local resultP = context.docStore:setServerValueP(
								'space',
								docValue._docId,
								{ _deleted = true, _ozMetadataUploadStatus = 'uploaded', } )
		
				return resultP
					:mapDataP( function ()
						
						return context.docStore:setServerValueP(
							'space_album',
							docValue._localOnly.album.space_album_id,
							{ _deleted = true, _ozMetadataUploadStatus = 'uploaded', } )
										
					end )
					
			end
			
		end

		if not headers or not ( headers.status_code == '201' or headers.status_code == '204' ) then
		
			-- Expect a 201 for a space create, 204 for space update

			uploadLog:error( "Got status %s back from Oz\n\nwhat we uploaded in body was:\n\n%s\n\nHTML response body was:\n\n%s\n\n",
							 headers.status_code, jsonData, data )
							
			return 'error', data, headers

		end
		
		-- [scouten 2012-08-17, PR# 3317838]: Force immediate reload of the document
		-- so we get the space URL from Oz.
		
		return WFOzDownloadTask.reloadDocumentP( context,
												 context.docType,
												 context.docId,
												 context.docUrl )
		
	end )

end

--------------------------------------------------------------------------------

local function uploadAssetStage1P( context, docValue, localOverlay )

	-- FIXME [scouten 2011-06-01]: Not really RESTful.

	uploadLog:debug( "uploadAssetStage1P photoId = %s", context.docId )
	
	-- Upload metadata if necessary.

	local uploadMetadataP = uploadAssetMetadataP( context, docValue, localOverlay )
	
	-- Upload thumbnail if necessary.
	
	local uploadThumbnailP
	
	-- If there is a 2x thumbnail to upload, upload both the 1x and 2x
	-- else upload only the 1x.
	
	if docValue
		and docValue._localOnly
		and docValue._localOnly.thumbnail2x then
		
			uploadThumbnailP = uploadMetadataP:mapDataP( function()

				uploadLog:debug( "uploadAssetStage1P -> uploadThumbnailP photoId = %s", context.docId )
	
				return uploadAssetRenditionP( context, docValue, localOverlay, 'thumbnail', '144h' )

			end )
			:mapDataP( function()
			
				uploadLog:debug( "uploadAssetStage1P -> uploadThumbnail2xP photoId = %s", context.docId )
				
				return uploadAssetRenditionP( context, docValue, localOverlay, 'thumbnail2x', 'thumbnail2x' )

			end )
	else
		
		uploadThumbnailP = uploadMetadataP:mapDataP( function()

				uploadLog:debug( "uploadAssetStage1P -> uploadThumbnailP photoId = %s", context.docId )
	
				return uploadAssetRenditionP( context, docValue, localOverlay, 'thumbnail', '144h' )

		end )
	end
	
	-- Upload preview if necessary.
	
	local uploadPreviewP = uploadThumbnailP:mapDataP( function()

		uploadLog:debug( "uploadAssetStage1P -> uploadPreviewP photoId = %s", context.docId )

		return uploadAssetRenditionP( context, docValue, localOverlay, 'preview', '1024' )

	end )
	
	local previousUploadP
	
	if docValue
		and docValue._localOnly
		and docValue._localOnly.proxy
	then
		previousUploadP = uploadPreviewP:mapDataP( function()
			
			uploadLog:debug( "uploadAssetStage1P -> uploadProxyP photoId = %s", context.docId )
			
			return uploadAssetProxyP( context, docValue, localOverlay )
			
			end )
	else
		previousUploadP = uploadPreviewP
	end
	
	return previousUploadP

end

--------------------------------------------------------------------------------

local function uploadReceiptP( context, docValue, localOverlay )

	assert( docValue.userId == context.ozUser:id() )

	-- Only thing we need to send is the receipt data.

	local uploadContent = {
		receipt_data = assert( docValue.transactionReceipt ),
		store = 'ios',
	}
	
	-- Upload metadata for this revision.
	
	local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )

	-- FIXME [scouten 2011-12-06]: Bump warning level to error for this particular
	-- log so we can tell if we are actually getting the store receipts. Return
	-- this to lower debug level after we confirm bug is fixed (presumably after
	-- 1.1.) This instrumentation is to help us understand bug 3061577.
	-- PR #3062887: Reminder to restore this to :debug.

	uploadLog:error( "::3061577:: POST %s\n%s", OzCatalogUrls.receiptsUrl, jsonData )
--	uploadLog:debug( "POST %s\n%s", OzCatalogUrls.receiptsUrl, jsonData )

	local putArgs = getJsonPutArgs( 'POST' )

	local uploadP = context.ozUser:putP( OzCatalogUrls.receiptsUrl, jsonData, putArgs )

	return uploadP:mapDataP( function( _, data, headers )

		context.nop = false

		local newServerValue = docValue
		newServerValue._changeCounter = nil
		newServerValue._docType = nil
		newServerValue._docId = nil

		if headers.status_code == '201'
			or headers.status_code == '200'
		then
		
			-- 200 or 201: Receipt accepted.

			newServerValue._ozMetadataUploadStatus = 'uploaded'
			newServerValue.receiptStatus = 'accepted'
			
		elseif headers.status_code < '500' then
		
			-- 4xx error (probably): Receipt is invalid.
			
			newServerValue._ozMetadataUploadStatus = 'uploaded'
			newServerValue.receiptStatus = 'invalid'
			newServerValue.statusCode = headers.status_code

			local success, decodedData = pcall( KSJsonDecoder.decodeWebSafeJson, data )
			
			if success and decodedData then

				newServerValue.ozErrorCode = decodedData.code

				newServerValue.externalStoreErrorCode =
					decodedData.upstream_error_code
			
			end

		else
		
			-- 5xx error: Server is wacked, so we'll just let
			-- this remain in the upload queue and try again
			-- later.
			
			uploadLog:error(
				"Unexpected result code %s on PUT to %s\n\ndata = %s",
									headers.status_code,
									OzCatalogUrls.receiptsUrl,
									data )

			return 'error', data, headers

		end
		
		-- Record new values and we're done.
		
		return context.docStore:setServerValueP( context.docType,
												 context.docId,
												 newServerValue )
	
	end )

end

--------------------------------------------------------------------------------

local uploadTaskForDocTypeAndStatus = {

	asset = {
		notUploaded_metadataOnly = uploadAssetMetadataP,
		notUploaded = uploadAssetStage1P,
		[ 'notUploadedStage2-original' ] = uploadAssetOriginalP,
		recoveryEffort_original = recoverOriginalAssetP,
		deleted = deleteAssetP,
	},

	receipt = {
		notUploaded = uploadReceiptP,
	},
	
	album = {
		notUploaded = uploadAlbumP,
		deleted = deleteAlbumP,
	},

	comment = {
		notUploaded = uploadCommentP,
		deleted = deleteCommentP,
	},
	
	space = {
		notUploaded = uploadSpaceP,
		deleted = deleteSpaceP,
	}

}

--------------------------------------------------------------------------------
--[[!
	Initiates upload of a document to a catalog on the Oz server.
		
	\param args (table) The upload task specification. A table that can have these keys:
	       <ul><li> \c docStore (required, OzDocStore):
	       		The document store that contains the document to upload.
	       		((is this right?)) </li>
		
		<li> \c docType (required, string): The document type.</li>
		<li> \c docId (required, string): The document's unique ID. </li>
		<li> \c uploadStatus (required, string): The upload status returned by
			OzCatalogUploadStatus.calculateUploadStatus; valid upload status
			codes vary by docType. </li>
		 <li> \c ozUser (required, OzUser): The user requesting the upload.</li>
		 <li> \c catalogUrl (required, string): The root URL of the catalog on the Oz server.</li>
		 </ul>

	\return <code>P < true(??) ></code> A Promise that resolves when the upload task is completed.
		((resolves to what?))
]]--

function OzCentralUploadTask.uploadP( args )
	
	local context = {
		docStore = tostring( assert( args.docStore ) ),
		docType = assert( args.docType ),
		docId = assert( args.docId ),
		uploadStatus = assert( args.uploadStatus ),
		ozUser = tostring( assert( args.ozUser ) ),
		baseUrl = OzCatalogUrls.baseServerUrl,
		catalogUrl = args.catalogUrl,
			-- [scouten 2012-05-21, PR#3190759]: There is no catalogUrl for user "catalog".
			-- TODO [arahn 2013-04-04]: USE_LINK_TEMPLATES server should provide this URL
		docUrl = string.format( '%s/%ss/%s', args.catalogUrl, args.docType, args.docId ),
		nop = true,
		errorCaptureForDoc = args.errorCaptureForDoc,
	}

	local uploadTaskFunc = uploadTaskForDocTypeAndStatus[ args.docType ]
		and uploadTaskForDocTypeAndStatus[ args.docType ][ args.uploadStatus ]
	
	if not uploadTaskFunc then
		return KSPromises.resolvedPromiseP( 'error',
			string.format(
				"OzCentralUploadTask.uploadP: unknown docType %s and/or uploadStatus %s",
				args.docType,
				args.uploadStatus ) )
	end

	uploadLog:info( "starting upload task for %s", context )

	-- Track time of first/last upload at each stage.
	
	if not timeStampAtFirstUpload then
		timeStampAtFirstUpload = KSDate.timeNowForAnimation()
	end

	local uploadStatus = args.uploadStatus
	local uploadStatusReportBin = uploadStatusReportBins[ uploadStatus ]
	
	if not uploadStatusReportBin then

		uploadStatusReportBin = {
			peakRealMemory = 0,
			peakVirtualMemory = 0,
			nUploads = 0,
			nFails = 0,
		}

		uploadStatusReportBins[ uploadStatus ] = uploadStatusReportBin
	
	end
	
	uploadStatusReportBin.nUploads = uploadStatusReportBin.nUploads + 1
	
	-- Optional: Track memory consumption.
	
	if ENABLE_MEMORY_POLLING_FOR_METRICS then

		if not WFAppMemoryUsage then
			WFAppMemoryUsage = require 'WFAppMemoryUsage'
		end
	
		local peakVM, peakRM = WFAppMemoryUsage.getPeakUsage()
		
		uploadStatusReportBin.peakRealMemory
			= math.max( uploadStatusReportBin.peakRealMemory, peakRM )
		uploadStatusReportBin.peakVirtualMemory
			= math.max( uploadStatusReportBin.peakVirtualMemory, peakVM )

	end

	-- We have to patch in these after the log statement; otherwise, we get
	-- "table not simple" nastygrams.

	context.docStore = args.docStore
	context.ozUser = args.ozUser
	
	-- All upload tasks need the doc value.

	local docValueP = context.docStore:docValueOnceP( context.docType, context.docId )

	-- After that, they diverge ...
	
	local uploadTaskP = docValueP:mapDataP( function( _, docValue, localOverlay )
	
		args.errorCaptureForDoc.docValue = docValue
		args.errorCaptureForDoc.localOverlay = localOverlay
	
		return uploadTaskFunc( context, docValue, localOverlay )
		
	end )
	
	-- Finally, make sure we did *something* in this upload pass. If not,
	-- trigger a reindex.
	
	return uploadTaskP:mapP( function( status, data, headers )

		uploadStatusReportBin.elapsedTime = KSDate.timeNowForAnimation() - timeStampAtFirstUpload

		if status ~= 'data' then

			uploadStatusReportBin.nFails = uploadStatusReportBin.nFails + 1

			return status, data, headers, context.docUrl

		end

		if context.nop then

			-- Use the meaningless value _bump to cause a reindex.
		
			local resultP = context.docStore:applyOverlayToDocumentP {
				docType = context.docType,
				docId = context.docId,
				overlay = {
					_bump = KSIdGenerator.uuidInLowercaseHex(),
				},
			}
			
			return resultP
		
		else
		
			return 'data', true
		
		end
	
	end )
	
end

--------------------------------------------------------------------------------

WFDiagnosticLog.addLogSectionFunc( function()

	return {

		name = "Upload Status",
		id = 'WFOzUploadTask',
		weight = 600,
		outputForAutomation = true,

		body = {
		
			uploadStatusByStage = uploadStatusReportBins,
		
		},
	
	}

end )

--------------------------------------------------------------------------------

return OzCentralUploadTask
