--[[----------------------------------------------------------------------------

OzDevelopSession.lua

@author scouten 2011-07-22

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.oz_client.catalog
local OzEpoch = require 'OzEpoch'

local OzReferenceDimension = require "OzReferenceDimension"

--------------------------------------------------------------------------------

local developSessionLog = KSLogging.log 'OzDevelopSession'
	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local OzReferenceDimension_getReferenceDimensions = assert( OzReferenceDimension.getReferenceDimensions )

local check = WFArguments.check

local OzDevelopSession = 'OzDevelopSession'
local OzCatalogService = 'OzCatalogService'
local PhotoId = 'string'

local DEBUG_SAVE_RENDITIONS_TO_DESKTOP = false -- will work on desktop and simulator only
local DEBUG_CONFIRM_THUMB_PREVIEW_WRITE_TO_CACHE = false

--------------------------------------------------------------------------------

local recentDevelopSessions = {}

--------------------------------------------------------------------------------

WFDiagnosticLog.addLogSectionFunc( function()

	if #recentDevelopSessions > 0 then

		return {
		
			name = "Recent Develop Sessions",
			id = 'OzDevelopSession',
			weight = 1200,
			
			body = {
				recentSessions = recentDevelopSessions,
			},
		
		}
	
	end

end )


--============================================================================--

-- @mark OzDevelopSession

local developSessionMethods = {

	_catalogService = false,
		-- (OzCatalogService)

	_docStore = false,
		-- (OzDocStore)
	
	_photoId = false,
		-- (string) photo's asset ID
	
	_clearTempXMP = false,
		-- (boolean) if set to true, then we will forcibly overwrite the
		-- TEMP_XMP field on exit.
	
	_originalDocValueP = false,
		-- (P< table >) private: complete document value when we started this session
	
	_originalDevelopSettingsP = false,
		-- (P< table >) develop settings value when we started this session
	
	_cacheEpochB = false,
		-- (B< number >) epoch number behavior swiped from OzCatalogService
	
	_tsLogEntry = false,
		-- (table) info in recentDevelopSessions log about this entry
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function developSessionMethods:__init( ozCatalogService, photoId )

	check( OzDevelopSession, OzCatalogService, PhotoId )

	self._catalogService = ozCatalogService
	self._docStore = ozCatalogService._docStore
	self._cacheEpochB = OzEpoch.cacheEpochB( ozCatalogService._catalogService )
		-- SPAM [arahn 2012-10-05]: Groping internals of ozCatalogService
	
	self._photoId = photoId
	
	self._tsLogEntry = {
		catalogId = ozCatalogService:getCatalogId(),
		photoId = photoId,
		timeAtEntry = KSDate.timeInIso8601Format(),
	}
	
	recentDevelopSessions[ #recentDevelopSessions + 1 ] = self._tsLogEntry
	if #recentDevelopSessions > 10 then
		table.remove( recentDevelopSessions, 1 )
	end
	
	developSessionLog:info( "acquiring original develop settings for photo %s", photoId )
	
	self._originalDocValueP = self._docStore:docValueOnceP( 'asset', photoId )
	
	self._originalDevelopSettingsP = self._originalDocValueP:mapP( function( status, docValue )
	
		developSessionLog:debug( "mapping originalDocValue through to develop settings" )

		self._tsLogEntry.docValueStatus = status

		if status == 'data' then
		
			developSessionLog:info( "acquired initial develop settings for photo %s\n\ndocValue = %s", photoId, docValue )
	
			self._tsLogEntry.docValueAtEntry = docValue
	
			local develop = docValue.develop or {}
		
			self._clearTempXMP = develop.TEMP_XMP ~= nil
			
			return 'data', {
				xmpSingleBlock = develop.TEMP_XMP or "", -- Legacy field
				xmpBase = develop.xmpBase or "",
				xmpLook = develop.xmpLook or "",
				userOrientation = develop.userOrientation
			}
				-- Replace "" with whatever we should use as default develop settings.
				-- When this becomes a table, we need to construct it each time through
				-- because the caller is free to modify it.
				
		else

			self._tsLogEntry.docValueError = docValue

			developSessionLog:error( "FAILED to acquire develop settings for photo %s %s",
								photoId, docValue )
	
			return status, docValue

		end
	
	end )
	self._originalDevelopSettingsP:runUntilDone()

end

--------------------------------------------------------------------------------

function developSessionMethods:developSettingsP()

	return self._originalDevelopSettingsP

end

--------------------------------------------------------------------------------

-- function developSessionMethods:hasPendingChangeFromOzB()
	-- is this of interest?

--------------------------------------------------------------------------------

local resolvedPromiseP = KSPromises.resolvedPromiseP( 'data', true )

--------------------------------------------------------------------------------

function developSessionMethods:cancelP()

	self._tsLogEntry.exitStatus = 'cancel'
	self._tsLogEntry.timeAtExit = KSDate.timeInIso8601Format()

	return resolvedPromiseP
		-- FIXME [scouten 2011-09-04]: Perhaps we no longer need cancel API
		-- at all?

end

--------------------------------------------------------------------------------

function developSessionMethods:saveAndCloseP( changes )

	self._tsLogEntry.exitStatus = 'saveAndClose'
	self._tsLogEntry.timeAtExit = KSDate.timeInIso8601Format()
	
	self._tsLogEntry.changes = {
		newBaseSettings = changes.xmpBase or '(nil)',
		newLookSettings = changes.xmpLook or '(nil)',
		newUserOrientation = changes.userOrientation or '(nil)',
		newPreview = KSData.length( changes.preview ) .. ' bytes',
		newThumbnail = KSData.length( changes.thumbnail ) .. ' bytes',
		newHighDpiThumbnail = KSData.length( changes.highDpiThumbnail ) .. ' bytes',
		croppedWidth = changes.croppedWidth or '(nil)',
		croppedHeight = changes.croppedHeight or '(nil)',
		negativeIsInadequate = changes.negativeIsInadequate or '(nil)',
	}

	local newBaseSettings = changes.xmpBase
	local newLookSettings = changes.xmpLook
	local newUserOrientation = changes.userOrientation
	local newPreview = changes.preview
	local newThumbnail = changes.thumbnail
	local newHighDpiThumbnail = changes.highDpiThumbnail
	local croppedWidth = changes.croppedWidth
	local croppedHeight = changes.croppedHeight

	assert( type( newBaseSettings or "" ) == 'string',
		"OzDevelopSession:saveAndCloseP: develop base settings should be a string" )
		
	assert( type( newLookSettings or "" ) == 'string',
		"OzDevelopSession:saveAndCloseP: develop look settings should be a string" )
		
	assert( type( newUserOrientation ) == 'number',
		"New orientation should be a number" )

	assert( KSData.isData( newPreview ),
		"OzDevelopSession:saveAndCloseP: newPreview missing" )
	
	assert( KSData.isData( newThumbnail ),
		"OzDevelopSession:saveAndCloseP: newThumbnail missing" )
		
	assert( KSData.isData( newHighDpiThumbnail ),
		"OzDevelopSession:saveAndCloseP: newHighDpiThumbnail missing" )
	
	debug.print( "New base settings: ", newBaseSettings )
	debug.print( "New look settings: ", newLookSettings )
	
	-- We will use a _replace overlay so clear the fields that are now irrelevant.
	
	if false then
	
		-- FIXME [mhamburg 2011-08-30]: We've seen a series of bugs that suggest that
		-- _replace may be failing to clear out fields. Let's go with empty strings
		-- for now just in case.
		
		if newLookSettings == "" then
			newLookSettings = nil
		end
		
	else
	
		newBaseSettings = newBaseSettings or ""
		newLookSettings = newLookSettings or ""
	
	end
	
	-- On the other hand, we really want to have some settings.
	
	newBaseSettings = newBaseSettings or ""
	
	local previewMD5 = newPreview:digest( 'MD5' )
	local thumbnailMD5 = newThumbnail:digest( 'MD5' )
	local thumbnail2xMD5 = newHighDpiThumbnail:digest( 'MD5' )
	
	local dataCache = assert( self._docStore:dataCache() )
	
	-- Write preview and thumbnail to data cache.
	
		-- GRUNGE [scouten 2011-07-28]: There's probably a pattern to be made
		-- here to support promise chaining that looks something like
		-- WFAsyncTestCase.
	
		-- FIXME [scouten 2011-08-11]: This code should probably become a
		-- method on OzCatalogService since it contains internal references
		-- to such.

	self._tsLogEntry.saveStatus = {}

	developSessionLog:debug( "writing preview (MD5 = %s)", previewMD5 )

	local writePreviewP = dataCache:putP( previewMD5, newPreview )

	writePreviewP:whenResolvedDo( function( status, message )
		self._tsLogEntry.saveStatus.preview = string.format( "%s (%s) at %s",
															 status,
															 KSData.isData( message ) and ( KSData.length( message ) .. ' bytes' ) or message,
															 KSDate.timeInIso8601Format() )
	end ):runUntilDone()

	local writeThumbnailP = writePreviewP:mapP {
		data = function( status, data )
			developSessionLog:debug( "writing thumbnail (MD5 = %s)", thumbnailMD5 )
			return dataCache:putP( thumbnailMD5, newThumbnail )
		end,
	}
	writeThumbnailP:whenResolvedDo( function( status, message )
		self._tsLogEntry.saveStatus.thumbnail = string.format( "%s (%s) at %s",
															   status,
															   KSData.isData( message ) and ( KSData.length( message ) .. ' bytes' ) or message,
															   KSDate.timeInIso8601Format() )
	end ):runUntilDone()

	local writeThumbnail2xP = writeThumbnailP:mapP {
		data = function( status, data )
			developSessionLog:debug( "writing thumbnail2x (MD5 = %s)", thumbnail2xMD5 )
			return dataCache:putP( thumbnail2xMD5, newHighDpiThumbnail )
		end,
	}
	writeThumbnail2xP:whenResolvedDo( function( status, message )
		self._tsLogEntry.saveStatus.thumbnail2x = string.format( "%s (%s) at %s",
															     status,
															     KSData.isData( message ) and ( KSData.length( message ) .. ' bytes' ) or message,
															     KSDate.timeInIso8601Format() )
	end ):runUntilDone()
	
	-- Once those are done, then write new info to document.
	
	local epochNow = self._cacheEpochB:valueNow()
	
	local thumbnailSize = KSData.length( newThumbnail )
	local previewSize = KSData.length( newPreview )
	local thumbnail2xSize = KSData.length( newHighDpiThumbnail )
	
	if not newThumbnail or thumbnailSize < 1 then
		developSessionLog:warn( "Develop handed us a thumbnail that is empty" )
	end
	
	if not newPreview or previewSize < 1 then
		developSessionLog:warn( "Develop handed us a preview that is empty" )
	end
	
	if not newHighDpiThumbnail or thumbnail2xSize < 1 then
		developSessionLog:warn( "Develop handed us a thumbnail2x that is empty" )
	end

	-- DEBUG ONLY: Read back the thumbnail and preview and make sure it's good.
	
	if DEBUG_CONFIRM_THUMB_PREVIEW_WRITE_TO_CACHE then
	
		writePreviewP:whenResolvedDo( function( status )
		
			dataCache:getP( previewMD5 )
				:whenResolvedDo( function( status, data )
				
					if status ~= 'data' or not data or KSData.length( data ) ~= previewSize then
						developSessionLog:error( "Failed to read back the preview we just wrote out from develop !!!" )
					else
						developSessionLog:info( "Confirmed preview was written back successfully" )
					end
				
				end )
				:runUntilDone()
		
		end )
		:runUntilDone()
	
		writeThumbnailP:whenResolvedDo( function( status )
		
			dataCache:getP( thumbnailMD5 )
				:whenResolvedDo( function( status, data )
				
					if status ~= 'data' or not data or KSData.length( data ) ~= thumbnailSize then
						developSessionLog:error( "Failed to read back the thumbnail we just wrote out from develop !!!" )
					else
						developSessionLog:info( "Confirmed thumbnail was written back successfully" )
					end
				
				end )
				:runUntilDone()
		
		end )
		:runUntilDone()
		
		writeThumbnail2xP:whenResolvedDo( function( status )
		
			dataCache:getP( thumbnail2xMD5 )
				:whenResolvedDo( function( status, data )
				
					if status ~= 'data' or not data or KSData.length( data ) ~= thumbnail2xSize then
						developSessionLog:error( "Failed to read back the thumbnail2x we just wrote out from develop !!!" )
					else
						developSessionLog:info( "Confirmed thumbnail2x was written back successfully" )
					end
				
				end )
				:runUntilDone()
		
		end )
		:runUntilDone()
	
	end
	
	local tempXMP
	
	if self._clearTempXMP then
		tempXMP = ""
	end
	
	-- The reference width and height tells us how big the actual image is.
		
	local referenceWidth, referenceHeight = croppedWidth, croppedHeight
	
	if not referenceWidth or not referenceHeight then
		referenceWidth, referenceHeight = OzReferenceDimension_getReferenceDimensions( newPreview )
	end
	
	local overlay = {
	
		develop = {
		
			--_replace = true,
				-- FIXME [scouten 2011-07-27]: Make sure _replace = true
				-- remains in place when we start having tables.
				
				-- [scouten 2011-09-04]: FYI: we add this in later now.
				
			xmpBase = newBaseSettings,
			xmpLook = newLookSettings,
			
			userOrientation = newUserOrientation,
			
				-- cropped width and height are largely functions of develop
			croppedWidth = croppedWidth,
			croppedHeight = croppedHeight,
			
			TEMP_XMP = tempXMP
			
		},

		_localOnly = {
		
			devSettingsLastSetAtRev = 'pending',
			thumbnail = {
				md5 = thumbnailMD5,
				revId = 'pending',
				epoch = epochNow,
				size = KSData.length( newThumbnail ),
				referenceWidth = referenceWidth,
				referenceHeight = referenceHeight
			},
			preview = {
				md5 = previewMD5,
				revId = 'pending',
				epoch = epochNow,
				size = KSData.length( newPreview ),
				referenceWidth = referenceWidth,
				referenceHeight = referenceHeight
			},
			thumbnail2x = {
				md5 = thumbnail2xMD5,
				revId = 'pending',
				epoch = epochNow,
				size = KSData.length( newHighDpiThumbnail ),
				referenceWidth = referenceWidth,
				referenceHeight = referenceHeight
			},
		
		},

	}
	
	if changes.negativeIsInadequate then
	
		overlay._localOnly.thumbnail.ozUpdateRendition = {
			width = changes.desiredThumbnailWidth,
			height = changes.desiredThumbnailHeight,
		}
		overlay._localOnly.preview.ozUpdateRendition = {
			width = changes.desiredPreviewWidth,
			height = changes.desiredPreviewHeight,
		}
		
	end
	
	self._tsLogEntry.overlay = overlay
	
	-- Get old develop settings and new doc value and make sure we didn't
	-- have any surprises.

	local oldDevelopSettingsP = writeThumbnailP:mapDataP( function()
	
		return self._originalDevelopSettingsP
	
	end )
	
	local developSettings
	
	local currentDocValueP = oldDevelopSettingsP:mapDataP( function( _, _developSettings )
	
		developSettings = _developSettings
	
		return self._docStore:docValueOnceP( 'asset', self._photoId )
	
	end )

	local saveSettingsP = currentDocValueP:mapP {

		data = function( status, currentDocValue )

			self._tsLogEntry.docValueAtSave = currentDocValue
			
			-- Tweak overlay if develop settings changed out from under us.
			
			-- FIXME [scouten 2011-09-04]: This would be better located down in OzDocStoreImp
			-- (it could then be atomic), but this seems like a less intrusive change to make
			-- right now.
			
			local originalDevelop = {
				xmpBase = developSettings and developSettings.xmpBase or "",
				xmpLook = developSettings and developSettings.xmpLook or "",
				xmpSingleBlock = developSettings and developSettings.xmpSingleBlock or "",
				userOrientation = developSettings and developSettings.userOrientation or 1,
			}
			
			local newDevelop = {
				xmpBase = currentDocValue.develop and currentDocValue.develop.xmpBase or "",
				xmpLook = currentDocValue.develop and currentDocValue.develop.xmpLook or "",
				xmpSingleBlock = currentDocValue.develop
					and currentDocValue.develop.xmpSingleBlock or "",
				userOrientation = currentDocValue.develop
					and currentDocValue.develop.userOrientation or 1,
			}
			
			if KSTableUtils.deepEquals( originalDevelop, newDevelop ) then
			
				developSessionLog:debug( "NOT flagging: \n\noriginal = %s\n\ncurrent = %s",
										 developSettings, currentDocValue.develop )
				
			else
			
				self._tsLogEntry.generatingConflict = true

				developSessionLog:error( "flagging conflict: \n\noriginal = %s\n\ncurrent = %s",
										 developSettings, currentDocValue.develop )
				
				overlay._conflict = {
					changedAtTime = currentDocValue.changedAtTime,
					changedOnDevice = currentDocValue.changedOnDevice,
					develop = currentDocValue.develop,
				}
			
			end
			
			developSessionLog:debug( "writing new overlay %s", overlay )

			local promise, _ = self._docStore:applyOverlayToDocumentP {
	
				_replace = true,
	
				-- undoManager = ???, -- FIXME [scouten 2011-07-23]: Where is this reference?
					-- check with EricW or Akiko
					-- or, never mind ... Undo got NiV1'd ...
		
				docType = 'asset',
				docId = self._photoId,
				overlay = overlay,
				addRevTimestamp = true,
			
			}
			
			return promise
				-- FIXME [scouten 2011-07-29]: Promise map code complains about
				-- multi-valued returns.
		
		end,

	}

	saveSettingsP:whenResolvedDo( function( status, message )
		self._tsLogEntry.saveStatus.docValue = string.format( "%s (%s) at %s",
															  status,
															  message,
															  KSDate.timeInIso8601Format() )
	end ):runUntilDone()

	-- TEMPORARY [scouten 2011-07-30]: Debug code to see what's getting generated by ImageCore.

	if DEBUG_SAVE_RENDITIONS_TO_DESKTOP then
		KSData.saveToUniquelyNamedFile( newThumbnail, "~/Desktop/thumbnail.jpg" )
		KSData.saveToUniquelyNamedFile( newPreview, "~/Desktop/preview.jpg" )
		KSData.saveToUniquelyNamedFile( newHighDpiThumbnail, "~/Desktop/thumbnail2x.jpg" )
	end
	
	return saveSettingsP

end

--------------------------------------------------------------------------------

function developSessionMethods:wantsProxyP( photoId )

	if not photoId then
		return KSPromises.resolvedPromiseP( 'data', nil )
	end
	
	local docValueOnceP = self._docStore:docValueOnceP( 'asset', photoId )
	
	local resultP = docValueOnceP:mapDataP( function( status, docValue )
	
		local width = ( docValue and docValue.develop ) and docValue.develop.croppedWidth
						 or ( docValue and docValue.width or 0 )
						
		local height = ( docValue and docValue.develop ) and docValue.develop.croppedHeight
						 or ( docValue and docValue.height or 0 )
		
		local proxySize = 2048
		local wantsProxy = width >= proxySize or height >= proxySize

		return 'data', wantsProxy
	
	end )
	
	return resultP
	
end

--------------------------------------------------------------------------------

return KSClass 'OzDevelopSession'( developSessionMethods )
