--[[----------------------------------------------------------------------------

OzDocStore.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This code is the public API for Oz document access. It can be used in any event
processor.

"Documents" are effectively glorified Lua tables placed under revision control.
Documents can contain only scalar values and may use only strings for keys.
Strings with leading underscores are reserved as keys for use by the doc store
system (see below); general purpose use of such keys is prohibited. While the
slash character is permitted in keys, it will make them unreachable by
fieldValueB, so it is not recommended. Values may be tables and those tables
must observe the same rules.

Documents are identified by a document type (string) and a document ID
(a 32-digit lower-case hex UUID).

Special field names used by the doc store system:

	_localOnly: If present, this is typically a table containing information that
		is not synced to Oz Central or any other device. Use this for device-specific
		information about the document.
		
	_serverProvidedMetadata: If present, this data contains auxiliary information
		about the photo provided by Oz Central. It is not writeable by clients.
		For example, in photo asset descriptions, this is used to convey the URLs
		to the uploaded thumbnails and previews.
	
	_ozCentralRevisionId: If present, the ID of the current revision from Oz.
		DEPRECATED (2011-08-13): Now using the links field.
	
	_conflict: If present, this contains the server's version of metadata that
		conflicts with local changes. The keys within this table match the outer
		key for the document.
	
	_docType (read-only, string): The document type (as specified at creation).
	
	_docId (read-only, string): The document ID (as specified at creation).
	
	_changeCounter (read-only, number): A monotonically increasing change counter
		across the entire doc store database.
	
--------------------------------------------------------------------------------

OzDocStore API:

	OzDocStore = require 'OzDocStore'

	docStore = OzDocStore( catalogId )
		-- Construct a document store from an existing microkernel service
		-- object. Normally the catalog ID is a UUID for the catalog but you
		-- can also create a catalog with the special ID ":memory:" which
		-- creates a transient catalog in memory (primarily for unit testing).
		-- You can also use the special ID ":temp:" which creates a transient
		-- catalog on disk (again, primarily for unit testing) that will be
		-- deleted when the DB is no longer needed.
	
	docStore:close()
		-- Close the doc store when no longer needed.
	
	P< value > = docStore:localStateP( name )
		-- Return the named local state value. If none has been set yet,
		-- return nil.
	
	B< value > = docStore:localStateB( name )
		-- Tracks the named local state value. If none has been set yet,
		-- the value will be nil.
	
	P< true > = docStore:setLocalStateP( name, value )
		-- Set the named local state value. Value may be any transitable
		-- Lua data type (nil, boolean, number, string, or table). Functions,
		-- userdata, and threads are not allowed anywhere (including nested
		-- in tables). Tables must not have metatables.
		
		-- As the name would imply, local state is not propagated to other
		-- Oz databases (i.e. currently not to Oz central) when synced.
	
	P< table, table > = docStore:docValueOnceP( docType, docId, opts )
		-- Returns a promise that will resolve with the current value
		-- for the document. The second value is the "local overlay" of
		-- the document (i.e. the portion that is not synchronized to Oz).
		
		-- If present, opts is a table with flags that can modify the behavior
		-- of this method. Accepted values are:
		
			-- separateValues: (boolean) if true, the first return value
			-- changes to the current value as read back from Oz.
		
	B< table, boolean > = docStore:docValueB( docType, docId )
		-- Returns a behavior that will track the current value of the
		-- document. The behavior returns a second value that is set to true
		-- when the document is first updating itself and nil for all subsequent
		-- updates.
		
		-- Performance note: It's not wise to have many hundreds or more of
		-- these active at any time. Watching values across many documents is
		-- a sign that you should be using WFIndexStore.
		
		-- Note that there is no currently  provision for configuring the
		-- behavior of docValueB along the lines of what is available for
		-- docValueOnceP. Let me know (@scouten) if you need one.

	B< value, boolean > = docStore:fieldValueB( docType, docId, fieldPath )
		-- Returns a behavior that will track the current value of a single
		-- field in a document. The behavior returns a second value that is
		-- set to true when the document is first updating itself and nil for
		-- all subsequent updates.
		
		-- Field paths are specified as slash-delimited strings, much like
		-- Unix paths, except that the root is implied. So a key "bar" within
		-- a table at key "foo" is found by using fieldValue( _, _, "foo/bar" ).
		-- If no such table "foo" exists, then the value for all of its
		-- hypothetical children is nil. Path segments that can be understood
		-- as an integer value will be converted. (In other words, the field
		-- path "foo/1" would look for the first member of the array at key
		-- "foo".)
		
		-- Performance note: Field value behaviors are basically derived from
		-- the corresponding docValueB, so it's unwise to watch a single field
		-- across many documents. There is a relatively small incremental cost
		-- to watch additional fields within the same document.
	
	P< requestStatus > = docStore:setServerValueP( docType, docId, serverValue, options )
		-- This API is intended only for use by the sync engine, and will
		-- only be viable so long as we have a centralized model. This API
		-- is called when new server values become available. It updates
		-- the combined value to the extent possible. It may introduce conflicts
		-- if there have been local edits since the previous server value.
		
		-- If options is present, it must be a table. The only option recognized
		-- at this time is { clearLocalOverlay = true }. This has the effect of
		-- stripping any local changes except for those stored in the _localOnly
		-- section.

	P< requestStatus >, overlay = docStore:applyOverlayToDocumentP {
						docType = docType,			-- string
						docId = docId,				-- string
						docIds = docIds,			-- array of strings
						overlay = overlay,			-- table
					}
		-- Modify the specified document by applying overlay to it.
		-- An overlay is a table in the same format as a document and overlays
		-- any existing values. Within subtables, existing values are left as-is
		-- unless individually replaced by the overlay. (Exception: if a subtable
		-- has a key named _replace with a true value, the entire subtable
		-- is released.) Existing values can be deleted by replacing them with
		-- a table containing _delete = true. (For example: favorite = { _delete = true }
		-- will remove any existing favorite value.)

		-- You must provide docId or docIds but not both.
		
		-- Promise will resolve with a new revision ID (only in the single-
		-- document case) or an error when the request is done.
		
	B< number > = docStore:docChangeCounterB()
		-- Returns a counter behavior that increments for every change to any
		-- document. Be aware that this value may jump, especially if a large
		-- number of documents are changed at once.
	
	P< number > = docStore:docChangeCounterP()
		-- Returns a value for docChangeCounter that includes every operation
		-- that has completed to this point.
	
	P< table > = docStore:changedDocsP {
		fromChangeCounter = (number),
			-- value obtained from :docChangeCounterB
		limit = (optional, number),
			-- max # of documents to return (limit of 50 or 1000 is always enforced)
		docs = (optional, array of doc IDs),
		idsOnly = (optional, boolean),
	}
		-- Returns a promise that will resolve with the set of documents that
		-- have changed since a given previous marker.
		-- The promise data will contain an array of document values. The
		-- array/table will also contain a top-level element named continueFrom
		-- that will give you the change counter number that you should use
		-- in the next :changedDocs call. That number may be less than the
		-- current value of :docChangeCounterB, for instance, if the number
		-- of documents was limited.
		
		-- If docs is provided, only changed documents matching that list of
		-- IDs will be matched. All others will be skipped and will not count
		-- against the limit.
		
		-- If idsOnly is true, then the limit is raised to a higher number
		-- (currently 1000).
		
		-- API CHANGE [scouten 2012-03-23]: This now returns a promise instead
		-- of an event stream.
	
	(KSDataCache) = docStore:dataCache()
		-- Returns the data cache associated with this doc store. The data cache
		-- is to be used for binary attachments. Managing content in the data
		-- cache is not the responsibility of OzDocStore.
	
	PLANNED (not yet implemented):

		-- fieldValueOnceE
		-- hasPendingUpdateB
		-- hasConflictB
		
		-- * valueB and fieldValueB return a second value in the tuple which
		-- describes the source of the value (pre-existing, local, or external)

--------------------------------------------------------------------------------

ROADMAP: Significant work that is planned in this file (in addition to the FIXMEs):

	- describe how to delete documents
	
--------------------------------------------------------------------------------

@author scouten 2011-03-02

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.os
local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- wichitafoundation.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- libs.reactive
local KSReactive = require 'KSReactive'

	-- services.data_cache
local KSDataCache_lazy -- lazy loaded



--============================================================================--

local docStoreLog = KSLogging.log 'OzDocStore'
--	:setLevel( KSLogging.DEBUG )

local userDocLog = KSLogging.log 'OzDocStore :: user'
--	:setLevel( KSLogging.DEBUG )

local docStoreInitLog = KSLogging.log 'OzDocStore :: init'
--	:setLevel( KSLogging.DEBUG )

local DEBUG = _G.DEBUG -- make lint happy

local string_format = string.format

local type = type

--------------------------------------------------------------------------------

-- types for argument checking

local OzDocStore = 'OzDocStore'
local DocumentType = 'string'
local DocumentId = 'string'


--============================================================================--

-- @mark OzDocStore class

--[[!
	\class OzDocStore

	\brief Provides access to documents on the Oz server.
	
	Documents are represented as Lua tables, placed under revision control.
	Documents are identified by a document type (string) and
	a document ID (a 32-digit lower-case hex UUID).
	
	Documents can contain only scalar values, and keys must be strings.
	<ul><li>Do not use the slash character in key strings.</li>
	<li>Strings with leading underscores are reserved as keys for use by the doc store
	system; general purpose use of such keys is prohibited.</li>
	<li>Values can be tables and those tables must observe the same rules.</li></ul>

	These special field names are used by the doc store system:
	<table>
	<tr>
	<td>_localOnly</td><td>If present, this is typically a table containing information that
		is not synced to Oz Central or any other device. Use this for device-specific
		information about the document.</td>
	</tr>
	<tr>
	<td>_serverProvidedMetadata</td><td>If present, this data contains auxiliary information
		about the photo provided by Oz Central. It is not writeable by clients.
		For example, in photo asset descriptions, this is used to convey the URLs
		to the uploaded thumbnails and previews.</td>
	</tr>
	<tr>
	<td>_ozCentralRevisionId</td><td>If present, the ID of the current revision from Oz.
		DEPRECATED (2011-08-13): Now using the links field.</td>
	</tr>
	<tr>
	<td>_conflict</td><td>If present, this contains the server's version of metadata that
		conflicts with local changes. The keys within this table match the outer
		key for the document.</td>
	</tr>
	<tr>
	<td>_docType (read-only, string)</td><td>The document type (as specified at creation).</td>
	</tr>
	<tr>
	<td>_docId (read-only, string)</td><td>The document ID (as specified at creation).</td>
	</tr>
	<tr>
	<td>_changeCounter (read-only, number)</td><td>A monotonically increasing change counter
		across the entire doc store database.</td>
	</tr>
	</table>
]]--

local docStoreMethods = {

	_catalogId = false,
		-- (string) catalog ID

	_docStoreEP = false,
		-- (WFWorkerThread) reference to the event processor
	
	_dataCache = false,
		-- (KSDataCache) data cache for this doc store
		-- FIXME [scouten 2011-04-05]: This usage is probably temporary.

	_docChangeCounterB = false,
		-- (behavior) most recent doc change counter value. Initially nil.
		-- Lazily instantiated.

	_currentDocValueBehaviors = false,
		-- (fully weak table) document value behaviors that are currently active
			-- key: (string) doc ID (will probably change to branch rev ID)
			-- value: (B< table >) behavior for that doc value

	_currentFieldValueBehaviors = false,
		-- (fully weak table) field value behaviors that are currently active
			-- key: (string) doc ID + field path
			-- value: (B< table >) behavior for that field value

	_currentDocChangeCounter = 0,
		-- (number) all behaviors in _currentDocValueBehaviors are up to date
		-- as of this change counter value. Note that field value behaviors are
		-- mapB extractions from the doc value behaviors, so we don't need to
		-- track those separately.

	_isUpdatingDocValues = false,
		-- (boolean) true if a request for changedDocValues is in flight;
		-- used to prevent overloading of the request system

	-- FIXME [scouten 2011-11-07]: Remove these after bug #2992540 is fixed.
	-- What follows is debug instrumentation for that bug.
	
	_nApplyOverlayRequestsInInterval = 0,
	_applyOverlayIntervalStarted = false,

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local updateDocValues

--------------------------------------------------------------------------------

--[[!
	\brief Constructs a new document object from a catalog.
	
	\param catalogId (string) The unique ID of an existing catalog.
 		For testing, can use the special ID ":memory:" which
		creates a transient catalog in memory, or ":temp:"
		which creates a transient catalog on disk that is deleted
		when the DB is no longer needed.
 ]]--
function docStoreMethods:__init( --[[string]] catalogId )

	docStoreInitLog:info( "Creating new doc store ID %s in thread %s\n\nvia %s\n\n",
						  catalogId, ':thread:', ':traceback:' )

	-- Validate catalog ID: Must be either ":memory:", ":temp:", or 32-digit
	-- lower-case hex UUID.

	if catalogId == ':memory:' or catalogId == ':temp:' then

		-- Make sure :memory: and :temp: DBs don't wind up conflicting.

		catalogId = catalogId .. tostring( KSIdGenerator.localCounter() )

	else

		if type( catalogId ) ~= 'string' then
			error( "OzDocStore: catalogId must be a string", 4 )
		end

		if catalogId:match( "^:memory:%d+$" ) or catalogId:match( "^:temp:%d+$" ) then

			-- this is OK (used by WFIndexStore, for example)

		elseif not catalogId:match( "^[0-9a-f]+$" ) or #catalogId ~= 32 then

			local errorMsg = string_format(
				"OzDocStore: catalogId '%s' must be ':memory:', ':temp:', or a 32-digit lowercase hex ID",
				catalogId )
			error( errorMsg , 4 )

		end
		
	end

	WFCloseSupport.mustClose( self )

	self._catalogId = catalogId

	self._docStoreEP = WFWorkerThread( "OzDocStoreEventProcessorImp.lua",
											catalogId )
	self:addCloseTarget( self._docStoreEP )
	
	-- Keep track of all of the doc values that are outstanding.

	self._currentDocValueBehaviors = KSTableUtils.makeFullyWeak()
		-- FIXME [scouten 2011-03-28]: Revise this to use Mark's new code
		-- in KSReactive.makeWeakEventTable.

	self._currentFieldValueBehaviors = KSTableUtils.makeFullyWeak()

end

--------------------------------------------------------------------------------
--[[!
     \brief Closes this document object when no longer needed.
]]--
function docStoreMethods:close()

	-- Close out any doc value behaviors that are still active.
	-- We don't make these close targets of the doc store because that
	-- would essentially track exactly what we're already tracking in
	-- this table.
	
	-- REVIEW [scouten 2011-03-22]: Is it still true that addCloseTarget
	-- causes a strong reference? If not, then we could switch over to
	-- using KSFunctionUtils.memoizeFn.
	
	-- FIXME [mhamburg 2011-04-11]: This isn't true any more so we
	-- could use addCloseTarget if that would help. Need to make sure
	-- that memoizeFn is close savvy.
	
	for _, docValueB in pairs( self._currentDocValueBehaviors ) do
		docValueB:close()
	end
	self._currentDocValueBehaviors = nil

	for _, fieldValueB in pairs( self._currentFieldValueBehaviors ) do
		fieldValueB:close()
	end
	self._currentFieldValueBehaviors = nil

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the catalog used to create this document.

	\return (string) The catalog's unique ID.
]]--

function docStoreMethods:catalogId()

	return self._catalogId

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this document as a string.

	\return (string) The string representation.
]]--


function docStoreMethods:__tostring()

	return string.format( "OzDocStore( %s )", self._catalogId )

end

--------------------------------------------------------------------------------

-- @mark local documents (not synced with Oz)

--------------------------------------------------------------------------------
 
--[[!
	\brief Retrieves (as a Promise) a current local state value for this document.
	
	A local value is one that has not been synced with the Oz server.

	\param name (string) The name of the state value.

	\return <code>P< string|nil ></code> A Promise that resolves to the named
		local state value, or nil if no value has been set.
]]--

function docStoreMethods:localStateP( --[[string]] name )

	return self._docStoreEP:requestP( 'localStateP', name )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) a local state value for this document.
	
	A local value is one that has not been synced with the Oz server.

	\param name (string) The name of the state value.

	\return <code>B< string|nil ></code> A Behavior whose value is the named
		local state value, or nil if no value has been set.
]]--


function docStoreMethods:localStateB( --[[string]] name )

	return self._docStoreEP:requestB {
		message = 'localStateB',
		arguments = name,
		startingValue = WFWorkerThread.kNilValue,
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets a local state value for this document.
	
	Local state is not propagated to other Oz databases (currently not to Oz central)
	when synced.

	\param name (string) The name of the state value.
	\param value (nil, boolean, number, string, or table) The new value.
		Functions, userdata, and threads are not allowed anywhere (including nested
		in tables). Tables must not have metatables.

	\return <code>P< result ></code> A Promise that resolves to
		\c true if the request succeeded. ((is this right?))
]]--

function docStoreMethods:setLocalStateP( --[[string]] name, --[[any]] value )

	local resultP = self._docStoreEP:requestP( 'setLocalState', name, value )
	
	resultP:runUntilDone()
	
	return resultP
	
end

--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local function validateDocTypeAndId( docType, docId, funcName, depth )

	if type( docType ) ~= 'string' or docType == '' then
		error(
			string_format( "%s: docType must be a non-empty string",
							funcName ), ( depth or 2 ) + 1 )
	end
	
	if type( docId ) ~= 'string'
		or not docId:match( "^[0-9a-f]+$" )
		or #docId ~= 32
	then
		error(
			string_format( "%s: docID must be a 32-digit lowercase hex ID",
							funcName ), ( depth or 2 ) + 1 )
	end
end

-- [scouten 2011-03-12]: Disable above validations for release builds.

if not DEBUG then
	validateDocTypeAndId = function() end
end

--------------------------------------------------------------------------------

-- @mark :docValue

--------------------------------------------------------------------------------
 		
--[[!
       \brief Retrieves this document.

       \param docType (string) The document type. ((valid values??))
       \param docId (string) The document's unique ID.
       \param opts (optional, table) Flags that can modify the operation. Currently
       		there is just one key, "separateValues" (boolean):when \c true,
       		two values are returned, the first being the current document,
       		as read from the Oz server, and the second the "local overlay" of
		the document; that is, the portion that is not synchronized to Oz.

       \return <code>P< table, table ></code> A Promise that resolves to one or
       		two tables representing the document store.
]]--

function docStoreMethods:docValueOnceP( --[[string]] docType, --[[string]] docId, --[[table]] opts )

	validateDocTypeAndId( docType, docId, 'OzDocStore:docValueOnceP', 2 )
	
	return self._docStoreEP:requestP( 'docValue', docType, docId, opts )

end

--------------------------------------------------------------------------------

local function docValueIsEmpty( docValue )

	if not docValue then return true end
	
	for k in pairs( docValue ) do
		if k ~= '_docId' and k ~= '_docType' then return false end
	end
	
	return true

end

--------------------------------------------------------------------------------

updateDocValues = function( self, docChangeCounter )

	-- Don't allow multiple requests to be in flight at once.
	
	if self._isUpdatingDocValues then return end

	-- Gather up the list of doc IDs that we care about.

	local docIds = {}
	local nDocIds = 0

	for docId in pairs( self._currentDocValueBehaviors ) do
		nDocIds = nDocIds + 1
		docIds[ nDocIds ] = docId
	end

	-- Short-circuit if there are no behaviors active.
	
	if nDocIds == 0 then
		self._currentDocChangeCounter = docChangeCounter
		return
	end
	
	-- Go request a bunch of doc values.
	
	self._isUpdatingDocValues = true
	
	self:changedDocsP { fromChangeCounter = self._currentDocChangeCounter, docs = docIds }
		:whenResolvedDo( function( status, changedDocValues )
		
			if status == 'data' then
		
				for i = 1, #changedDocValues do
				
					local changedDocValue = changedDocValues[ i ]
				
					local docValueB =
						WFCloseSupport.fetchCached( self._currentDocValueBehaviors,
													changedDocValue._docId )
						-- Note that the docValueB might have been GCd while we were
						-- waiting for this request to complete. Nil is an expected response here.
					
					if docValueB then
	
						local oldDocValue = docValueB:valueNow()
	
						-- It's possible that the "old" doc value is actually newer,
						-- for example if the doc value behavior was recently created and
						-- received its initial update more recently than this request.
						-- Don't replace it with an older value.
						
						if not ( oldDocValue and oldDocValue._changeCounter )
							or oldDocValue._changeCounter < changedDocValue._changeCounter
						then

							if docValueIsEmpty( oldDocValue ) then
								docValueB:set( changedDocValue, true )
							else
								docValueB:set( changedDocValue )
							end

						end
					
					end
	
				end
				
				self._currentDocChangeCounter = changedDocValues.continueFrom or 0
					-- FIXME [scouten 2011-03-22]: Warn if we didn't get a continueFrom value?
					-- Should also go in unit tests.
			
			else
			
				docStoreLog:error( "Failed to update doc values. :changedDocsP failed with error %s",
								   changedDocValues )
			
				-- Not much to do but mark all active doc value behaviors
				-- as in an error state and pretend we're caught up.
	
				self._currentDocChangeCounter = self:docChangeCounterB():valueNow()
			
				for _, docValueB in pairs( self._currentDocValueBehaviors ) do
					docValueB:setError( changedDocValues )
				end
			
			end
			
			-- Clear semaphore so we can grab more updates.

			self._isUpdatingDocValues = false
			
			-- If we're not caught up yet, come back for more documents.
			
			if self._currentDocChangeCounter < self:docChangeCounterB():valueNow() then
				updateDocValues( self, docChangeCounter )
			end
		
		end ):attachTo( self )

end

--------------------------------------------------------------------------------

local function docValueB_imp( self, docType, docId )

	local resultB = WFCloseSupport.fetchCached( self._currentDocValueBehaviors, docId )
	
	if not resultB then

		-- Make sure doc change counter observations are in place.
		-- (We now do this lazily.)
	
		if not self._docChangeCounterB then
			self:docChangeCounterB()
		end

		resultB = KSReactive.variableB( nil )
		self._currentDocValueBehaviors[ docId ] = resultB
		
		self:docValueOnceP( docType, docId )
			:whenResolvedDo( function( status, newValue )
			
				if status == 'data' then

					if resultB:valueNow() == nil then
						if docValueIsEmpty( newValue ) then
							resultB:set( newValue )
						else
							resultB:set( newValue, true )
						end
					end

				else
		
					if resultB:valueNow() == nil then
						resultB:setError( newValue ) -- in this case, the error message
					end
		
				end
			
			end )
			:attachTo( self )

	end

	return resultB

end

--------------------------------------------------------------------------------
--[[!
       \brief Reports (as a Behavior) the current value of this document.

       \param docType (string) The document type. ((valid values??))
       \param docId (string) The document's unique ID.

	\return <code>B< table, boolean ></code> A Behavior whose value is
		the current value of the document. The behavior returns a second value
		that is \c true when the document is first updating itself and
		nil for all subsequent updates.
       		
       \warning It's not wise to have many hundreds or more of these active at any time.
       		 Watching values across many documents is a sign that you should be using
       		 WFIndexStore.
]]--

function docStoreMethods:docValueB( --[[string]] docType, --[[string]] docId, --[[string]] branchRevId )

	return docValueB_imp( self, docType, docId, branchRevId ):relayB( true )

end

--------------------------------------------------------------------------------
--[[!
       \brief Reports (as a Behavior) the current value of a field in this document.

       \param docType (string) The document type. ((valid values??))
       \param docId (string) The document's unique ID.
       \param fieldPath (string) The field path, a slash-delimited string like
		a Unix paths, except that the root is implied. Use the path to
		identify values within tables or nested tables.	You can use number
		values to get the nth member of a table; for example, "tabA/1" to get
		the first member of an array in "tabA".

	\return <code>B< value, boolean ></code> A Behavior whose value is
		the current value of the field. The behavior returns a second value
		that is \c true when the document is first updating itself and
		nil for all subsequent updates.


 	\warning: It is inefficient to watch a single field across many documents.
		 There is a relatively small incremental cost to watch additional
		 fields within the same document.
]]--

function docStoreMethods:fieldValueB( --[[string]] docType, --[[string]] docId, --[[string]] fieldPath )

	if type( fieldPath ) ~= 'string' then
		error( "OzDocStore:fieldValueB: fieldPath is not a string", 2 )
	end
	
	local docIdPlusFieldPath = string_format( "%s:%s", docId, fieldPath )
	
	local resultB = WFCloseSupport.fetchCached( self._currentFieldValueBehaviors, docIdPlusFieldPath )
	
	if not resultB then

		local fieldPathSegments = {}
		local nSegments = 0
		
		fieldPath:gsub( "([^/]+)",
			function( segment )
				nSegments = nSegments + 1
				if segment:match( "^-?%d+$" ) then
					fieldPathSegments[ nSegments ] = tonumber( segment )
				else
					fieldPathSegments[ nSegments ] = segment
				end
			end )
		
		if nSegments == 0 then
			error( "OzDocStore:fieldValueB: fieldPath didn't contain a path", 2 )
		end
		
		local firstTime = true
		
		resultB = docValueB_imp( self, docType, docId ) --, branchRevId )
			:mapB( function( docValue, isNew )
	
				if docValue == nil or docValueIsEmpty( docValue ) then
					return nil
				end
	
				local wasFirstTime = firstTime
				firstTime = nil
				
				local t = docValue
				
				-- All parent nodes must be tables. If they are anything else,
				-- we just pretend they don't exist since we can't index
				-- into them. (Note that we don't support userdata values
				-- in documents since they don't persist or translate cleanly
				-- to JSON.)
				
				for i = 1, nSegments - 1 do
					t = t and t[ fieldPathSegments[ i ] ]
					if type( t ) ~= 'table' then
						t = nil
					end
				end
	
				local result = t and t[ fieldPathSegments[ nSegments ] ]
				return result, not not ( wasFirstTime or isNew )
			
			end )

		self._currentFieldValueBehaviors[ docIdPlusFieldPath ] = resultB
		
	end
	
	return resultB:relayB()

end

--------------------------------------------------------------------------------

-- @mark :setServerValue

--------------------------------------------------------------------------------
--[[!
       \brief Updates the value of a field in this document. For internal use by
		the sync engine.

	\param docType (string) The document type.
	\param docId  (string) The document ID.
	\param serverValue (string) The value.
	\param options (table) Options that identify the field.
]]--

function docStoreMethods:setServerValueP( --[[string]] docType, --[[string]] docId, --[[string]] serverValue, --[[table]] options )
	
	docStoreLog:debug( "OzDocStore:setServerValueP %s %s %s %s",
					   docType, docId, serverValue, options )

	if docType == 'user' then
		userDocLog:debug( "OzDocStore:setServerValueP %s %s %s %s\n\nvia %s\n\nin thread %s",
						  docType, docId, serverValue, options, ':traceback:', ':thread:' )
	end

	validateDocTypeAndId( docType, docId, 'OzDocStore:setServerValueP', 2 )

	local resultP = self._docStoreEP:requestP( 'setServerValue', docType, docId,
											   serverValue, options )
		
	resultP:runUntilDone() -- In case the result is otherwise ignored
	
	return resultP

end

--------------------------------------------------------------------------------

-- @mark :resolveConflictP

--------------------------------------------------------------------------------

local ResolveWith = 'OzDocStore.resolveWith'

WFArguments.registerChecker( ResolveWith, {
	[ 'local' ] = true,
	server = true,
	fork = true,
} )

--------------------------------------------------------------------------------
--[[!
       \brief Resolves a conflict between two versions of this document.

       \param docType (string) The document type.
       \param docId (string) The document's unique ID.
       \param changeCounter (number) The current revision number.
       \param resolveWith (string) The resolution method to use, one of: ((valid values?))

       \return <code>P < ?? ></code> A Promise that resolves to ((what?))
]]--

function docStoreMethods:resolveConflictP( docType, docId, changeCounter, resolveWith )

	WFArguments.check( OzDocStore, DocumentType, DocumentId, 'number', ResolveWith )

	docStoreLog:debug( "OzDocStore:resolveConflictP %s %s %d %s", docType, docId,
						changeCounter, resolveWith )

	return self._docStoreEP:requestP( 'resolveConflict', docType, docId,
									  changeCounter, resolveWith )
		-- FIXME [scouten 2011-08-02]: Return value from imp's resolveConflict method
		-- is an overlay ID, so theoretically resolveConflict could be undoable.
		-- We're not doing undo for V1, so we'll let that value fall on the floor for now.

end

--------------------------------------------------------------------------------

-- @mark :applyOverlay

--------------------------------------------------------------------------------

local underscoreByte = string.byte( '_' )

--------------------------------------------------------------------------------

local acceptableSpecialKeys = {

	_deleted = true,
	_replace = true,
	
	_ozCentralRevisionId = true,
	_ozMetadataUploadStatus = true,
		-- FIXME [scouten 2011-06-02]: Should be removed from this list.
	
	_localOnly = true,
		-- free-form data that is not synced to Oz

	_serverProvidedMetadata = true,
		-- data provided by Oz server that should not be modified by Wichita

	_halt = true,
	
	_bump = true,
		-- special key used to trigger reindexing

	_conflict = true,
		-- FIXME [scouten 2011-09-04]: Would be better expressed as a
		-- _valueImExpectingToReplace member for develop, but this is deemed
		-- a less-intrusive solution.

}

--------------------------------------------------------------------------------

--[[!
       \brief Modifies this document or a set of documents by applying an overlay.
		An overlay is a table in the same format as a document that overlays
		any existing values.


	\param args (table) The arguments, a table with these keys:
		<ul><li>docType (string)</li>
			<li>docId (string) OR docIds(array of strings)</li>
			<li>overlay (table)</li></ul>
					 
 		You must provide docId or docIds but not both.
		
		Within subtables, existing values are left as-is
		 unless individually replaced by the overlay, unless  a subtable
		 has a key named _replace with a true value; in this case, the entire
		 subtable is released.
		
		To delete existing values, replace them with
		a table containing _delete = true. For example:
		<code>favorite = { _delete = true }</code>
		 removes any existing favorite value.

		
	\return <code>P< requestStatus ></code> A Promise that signals completion; it
		resolves to a new revision ID (in the single-document case) or to an error.
]]--

function docStoreMethods:applyOverlayToDocumentP( --[[table]] args )
	
	-- FIXME [scouten 2011-03-15]: Parameter validation...

	docStoreLog:debug( "OzDocStore:applyOverlayToDocumentP %s", args )

	if args.docType == 'user' then
		userDocLog:debug( "OzDocStore:applyOverlayToDocumentP %s\n\nvia %s\n\nin thread %s",
						  args, ':traceback:', ':thread:' )
	end

	local overlay = assert( args.overlay )

	for k in pairs( overlay ) do
	
		if k:byte( 1 ) == underscoreByte and not acceptableSpecialKeys[ k ] then
			error( string_format(
				"OzDocStore:applyOverlayToDocument: illegal key %s", k ), 3 )
		end
	
		-- FIXME [scouten 2011-04-06]: In debug builds, at least, we should also
		-- verify that any nested tables also don't have keys with the reserved
		-- _ prefix.
	
	end
	
	local docIds = assert( args.docIds or args.docId,
		"OzDocStore:applyOverlayToDocument: missing docId or docIds argument" )
	
	if type( docIds ) == 'string' then
		docIds = { docIds }
	end
	
	local requestP = self._docStoreEP:requestP( 'applyOverlay',
							{
								docType = assert( args.docType ),
								docIds = docIds,
								overlay = overlay,
								addRevTimestamp = args.addRevTimestamp,
							} )

	requestP:runUntilDone()

	return requestP

end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------
--[[!
       \brief Tracks changes in this document by incrementing a counter for each change.
		Be aware that this value can jump, especially if
		many documents are changed at once.

	\return <code>B< number ></code> A Behavior whose value is
		the current change count.

]]--

function docStoreMethods:docChangeCounterB()

	-- Watch for doc change counter updates. This is how we know when
	-- to update the doc/field value behaviors.

	if not self._docChangeCounterB then

		local docChangeCounterIdRequestB = self._docStoreEP:requestB {
			message = 'addDocChangeCounterRequest',
			startingValue = WFWorkerThread.kNilValue,
		}
		
		self._docChangeCounterB = docChangeCounterIdRequestB:calmB( .25 )
							
		self._docChangeCounterB:onDataE( function( ... )
			return updateDocValues( self, ... )
		end )
		
		self:addCloseTarget( docChangeCounterIdRequestB )
		self:addCloseTarget( self._docChangeCounterB )
		
	end

	return self._docChangeCounterB:relayB()

end

--------------------------------------------------------------------------------
--[[!
       \brief Reports (as a Promise) the current change count for this document.
		This count includes every operation that has completed to this point.
 
 	\return <code>P< number ></code> A Promise that resolves to
		the current change count.

]]--

function docStoreMethods:docChangeCounterP()

	return self._docStoreEP:requestP( 'docChangeCounter' )

end

--------------------------------------------------------------------------------
--[[!
       \brief Retrieves the set of documents that changed since a specified point.

	\param args (table) A table with these keys:
		<ul><li>fromChangeCounter (number) A change-counter value.</li>
		<li>limit (optional, number) The maximum number of documents to return
			(limit of 50 or 1000 is always enforced)</li>
		<li>docs (optional, array of doc IDs) </li>
		<li>idsOnly (optional, boolean) When \c true, the default limit is raised
			to a higher number (currently 1000).</li></ul>
	
	
	\return <code>P< table ></code> A Promise that resolves to the changed-document
		values. The array/table also contains a top-level element named continueFrom
		whose value is the change-counter number to use	in the next :changedDocs call.
		That number can be less than the current value of :docChangeCounterB,
		for instance, if the number of documents was limited.
 ]]--

function docStoreMethods:changedDocsP( --[[table]] args )

	return self._docStoreEP:requestP( 'changedDocs', args )

end

--------------------------------------------------------------------------------

-- @mark :dataCache

--------------------------------------------------------------------------------
--[[!
       \brief Retrieves the data cache associated with this doc store,
       		to be used for binary attachments. Managing content in the data
		cache is not the responsibility of OzDocStore.

	\return (KSDataCache)  The data cache.
]]--

function docStoreMethods:dataCache()

	local result = self._dataCache
	
	if not result then

		-- Spin up the data cache used for binary data.

		local cachePath

		if self._catalogId:sub( 1, 8 ) == ':memory:' then
	
			cachePath = ':memory:'
		
		elseif self._catalogId:sub( 1, 6 ) == ':temp:' then
	
			cachePath = ':temp:'
		
		else
	
			local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
			
			local catalogDir = KSPathUtils.child( documentsDir, self._catalogId )
			catalogDir = KSPathUtils.child( catalogDir, "Catalog.ozcat" )
	
			docStoreLog:info( 'catalogDir = %s', catalogDir )
			
			KSFileUtils.mkdir( catalogDir )
	
			cachePath = KSPathUtils.child( catalogDir, "previewCache" )
			
		end

		KSDataCache_lazy = KSDataCache_lazy or require 'KSDataCache'

		result = KSDataCache_lazy( cachePath )

		self._dataCache = result
		self:addCloseTarget( result )
	
	end
	
	return result

end

--------------------------------------------------------------------------------

-- @mark database cleanup

--------------------------------------------------------------------------------
--[[!
       \brief ((what??))

]]--

function docStoreMethods:halt()

	if self._docStoreEP and not WFCloseSupport.hasBeenOrWillBeClosed( self._docStoreEP ) then
		return self._docStoreEP:requestP( 'halt' )
	end

end

--------------------------------------------------------------------------------

-- @mark performance instrumentation

--------------------------------------------------------------------------------
--[[!
       \brief Begins collecting performance statistics.	((is this right?))
       \return A Promise that resolves when the request has succeeded. ((is this right?))

]]--

function docStoreMethods:PERF_collectQueryPerformanceStats()

	-- Should be used only by OzDocStoreTest.lua.

	local resultP = self._docStoreEP:requestP( 'PERF_collectQueryPerformanceStats' )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
       \brief Writes collected performance statistics to standard output.((is this right?))
       \return A Promise that resolves when the request has succeeded. ((is this right?))

]]--

function docStoreMethods:PERF_dumpQueryPerformanceStats()

	-- Should be used only by OzDocStoreTest.lua.

	local resultP = self._docStoreEP:requestP( 'PERF_dumpQueryPerformanceStats' )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
       \brief Writes collected coverage statistics to standard output.((is this right?))
       \return A Promise that resolves when the request has succeeded. ((is this right?))

]]--

function docStoreMethods:DEBUG_dumpCoverageP( path )

	-- Should be used only by OzDocStoreTest.lua.

	local resultP = self._docStoreEP:requestP( 'DEBUG_dumpCoverageP', path )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

return KSClass 'OzDocStore'( docStoreMethods )
