--[[----------------------------------------------------------------------------

OzDocStoreImp.lua

@author scouten 2011-03-16

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This file implements the microkernel's portion of OzDocs. It lives in the Oz
microkernel event processor and manages the SQL tables where fully-composed
document data is stored and converted into partition events eligible for
synchronization.

The public API corresponding to this file is OzDocStore.lua.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.blonde
local blonde = require 'blonde'

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFRequestDispatcher = require 'WFRequestDispatcher'

	-- wichitafoundation.core.os
local KSDate -- = require 'KSDate' (lazy loaded)
local KSPathUtils = require 'KSPathUtils'

	-- wichitafoundation.core.reactive
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSDeviceUtils = require 'KSDeviceUtils'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.sqlite.database
local KSDatabase = require 'KSDatabase'
local KSDatabaseSchema = require 'KSDatabaseSchema'
local WFSqliteUtils = require 'WFSqliteUtils'

	-- wichitafoundation.oz_client.doc_store
local WFOzDocStorePath = require 'WFOzDocStorePath'


--------------------------------------------------------------------------------

local docStoreImpLog = KSLogging.log 'OzDocStoreImp'
--	:setLevel( KSLogging.DEBUG )

local docStoreAssetLog = KSLogging.log 'OzDocStoreImp :: asset'
--	:setLevel( KSLogging.DEBUG ) -- dupe of above, but only for docType == 'asset'

local docStoreSpaceLog = KSLogging.log 'OzDocStoreImp :: space'
--	:setLevel( KSLogging.DEBUG ) -- dupe of above, but only for docType == 'space'

local applyOverlayLog = KSLogging.log 'OzDocStoreImp :: applyOverlay'
--	:setLevel( KSLogging.DEBUG )

local conflictLog = KSLogging.log 'OzDocStoreImp :: conflict'
--	:setLevel( KSLogging.DEBUG )

local favoritesLog = KSLogging.log 'OzDocStoreImp :: favorites'
--	:setLevel( KSLogging.DEBUG )

local localStateLog = KSLogging.log 'OzDocStoreImp :: localState'
--	:setLevel( KSLogging.DEBUG )

local changedDocsLog = KSLogging.log 'OzDocStoreImp :: changedDocs'
--	:setLevel( KSLogging.DEBUG )

local blonde_encode = blonde.encode
local blonde_decode = blonde.decode

local DEBUG = _G.DEBUG -- make lint happy

WFRequestDispatcher.enableDiagnosticLogForRequestStats()

--------------------------------------------------------------------------------

-- types for argument checking

local OzDocStoreImp = 'OzDocStoreImp'
local DocumentType = 'string'
local DocumentId = 'string'


--============================================================================--

-- @mark database schema

local schema = KSDatabaseSchema( '1.0' ):useWAL()

--------------------------------------------------------------------------------
--[[
	This table provides a home for miscellaneous local state information.
	This information is not propagated to Oz central or other databases.
]]--

schema:addTable
	[[CREATE TABLE localState(
						name UNIQUE NOT NULL,
						dataType,
						value )]]
	:onConflict 'replace'
		-- FIXME [scouten 2011-04-26]: 'replace' is a totally inappropriate
		-- migration strategy!

schema:addIndex
	[[CREATE INDEX localState_name ON localState( name )]]

--------------------------------------------------------------------------------
--[[
	This table contains the summary view of all documents in this catalog.
	
	-- FIXME [scouten 2011-04-19]: Write more about it later ...
	
	-- PERF [scouten 2011-03-21]: Intern the UUIDs and convert them to
	-- database-local integers for search performance?
]]--

schema:addTable
	[[CREATE TABLE documents(
						docId UNIQUE NOT NULL,
						docType,
						changeCounter,
						lockCount NOT NULL DEFAULT 0,
						embargoedServerValue,
						serverBaseValue,
						localOverlay )]]
	:onConflict 'replace'
		-- FIXME [scouten 2011-04-19]: 'replace' is a totally inappropriate
		-- migration strategy!

		-- FIXME [scouten 2011-10-04]: lockCount and embargoedServerValue
		-- are no longer used, but we don't really have support for schema
		-- migration, so leave them here for now.

schema:addIndex
	[[CREATE INDEX documents_changeCounter ON documents( changeCounter )]]

--------------------------------------------------------------------------------
--[[
	This table is used in the implementation of the specific docs
	version of changedDocs. It lets us bind to a variable number of items.
]]--

schema:addTable
	[[CREATE TEMPORARY TABLE TEMP_docIds( docId )]]


--============================================================================--

-- @mark OzDocStoreImp

local docStoreImpMethods = {

	_database = false,
		-- (KSDatabase) microkernel database

	_path = false,
		-- (string) path to DB file

	_currentChangeCounterB = false,
		-- (B< number >) most recent change counter in documents

	_localStateBehaviors = false,
		-- (table) behaviors (if needed) for local state variables

	_conflictDetectedLog = false,
		-- (table) most recent setServerValue events that led to conflict
		-- detection

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function docStoreImpMethods:__init( catalogId )

	local dbPath

	if catalogId:match( '^:memory:' ) then

		dbPath = ':memory:'
	
	elseif catalogId:match( '^:temp:' ) then

		dbPath = ':temp:'
	
	else

		local catalogDir = WFOzDocStorePath.dirPathForCatalogId( catalogId )

		dbPath = KSPathUtils.child( catalogDir, 'database' )
		
		self._path = dbPath
		
	end

	docStoreImpLog:info( 'dbPath = %s', dbPath )

	local db = KSDatabase( dbPath, schema )
	self._database = db

	self:addCloseTarget( db )

	WFCloseSupport.mustCloseAtThreadTermination( self )
	
	-- Set up change notification.

	local changeCounterSeedRow = db:execOnce( "SELECT max( changeCounter ) FROM documents" )

	local currentChangeCounter = ( changeCounterSeedRow
		and changeCounterSeedRow[ "max( changeCounter )" ] or 0 ) + 1
	
	db:setChangeCounter( currentChangeCounter )

	self._currentChangeCounterB = KSReactive.variableB( currentChangeCounter )
	self:addCloseTarget( self._currentChangeCounterB )
	
	self._localStateBehaviors = KSTableUtils.makeFullyWeak()
	
	self._conflictDetectedLog = {}
	
	WFRequestDispatcher.enableDiagnosticLogForRequestStats()

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

-- @mark local documents (not synced with Oz)

--------------------------------------------------------------------------------

function docStoreImpMethods:localStateP( name )

	local db = self._database

	local localStateRow = db:exec( [[SELECT dataType, value FROM localState
					WHERE name = ?]], name )
	
	if localStateRow then
	
		return WFSqliteUtils.sqliteValueAndTypeToLuaValue( localStateRow.value,
														   localStateRow.dataType )
	
	else
	
		return nil
			-- Use an explicit nil return here so that # parameters is > 0.
			-- That will cause an onDataE callback in the client thread.
			-- We want that to happen.
	
	end

end

--------------------------------------------------------------------------------

function docStoreImpMethods:localStateB( name )

	local resultB = WFCloseSupport.fetchCached( self._localStateBehaviors, name )
	
	if not resultB then
	
		local initialValue = self:localStateP( name )
		
		localStateLog:debug( "creating localStateB for %s with initial value %s", name, initialValue )
	
		resultB = KSReactive.variableB( initialValue )
		
		self._localStateBehaviors[ name ] = resultB
	
	end
	
	return resultB:relayB()

end

--------------------------------------------------------------------------------

function docStoreImpMethods:setLocalState( name, value )

	local db = self._database

	local sqliteValue, sqliteDataType = WFSqliteUtils.luaValueToSqliteValueAndType( value )

	db:exec( [[INSERT OR REPLACE INTO localState( name, dataType, value )
					VALUES( ?, ?, ? )]], name, sqliteDataType, sqliteValue )

	local resultB = WFCloseSupport.fetchCached( self._localStateBehaviors, name )
	
	localStateLog:debug( "setLocalState for %s :: has resultB? %s", name, not not resultB )

	if resultB then
		resultB:set( value )
	end

end

--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local function updateChangeCounter( self, db )

	local newChangeCounter = db:getChangeCounter()

	changedDocsLog:info( "newChangeCounter = %s", newChangeCounter )
	self._currentChangeCounterB:set( newChangeCounter )
	
	return newChangeCounter
	
end

--------------------------------------------------------------------------------

local function validateDocTypeAndId( docType, docId, funcName, depth )

	if type( docType ) ~= 'string' or docType == '' then
		error(
			string.format( "%s: docType must be a non-empty string",
							funcName ), ( depth or 2 ) + 1 )
	end
	
	if type( docId ) ~= 'string'
		or not docId:match( "^[0-9a-f]+$" )
		or #docId ~= 32
	then
		error(
			string.format( "%s: docID must be a 32-digit lowercase hex ID",
							funcName ), ( depth or 2 ) + 1 )
	end

end

-- Disable above validations for release builds.

if not DEBUG then
	validateDocTypeAndId = function() end
end

--------------------------------------------------------------------------------

local function applyOverlayToTable( tableToModify, overlay, retainDeleteAndReplaceEvents )

	tableToModify = tableToModify or {}

	-- Short-circuit. If this is a top-level delete, then just trash
	-- the entire old document, but leave _localState.

	if overlay._deleted == true then
		return {
			links = tableToModify.links,
				-- Save links, as links.self is needed to send this delete to Oz
			_deleted = true,
			_localOnly = overlay._localOnly
				and applyOverlayToTable( tableToModify._localOnly,
								overlay._localOnly, retainDeleteAndReplaceEvents )
				or tableToModify._localOnly,
		}
	end
	
	-- Replacing the whole table? Press the Easy Button!
	
	if overlay._replace then
		overlay._replace = nil
		return applyOverlayToTable( { _replace = true }, overlay, false )
	end

	-- Updates a docValue record in place with an overlay.

	for k, v in pairs( overlay ) do

		if k ~= '_replace' then

			tableToModify._delete = nil

			if type( v ) == 'table' then
			
				if v._delete == true then
				
					if retainDeleteAndReplaceEvents and k ~= '_deleted' then
						tableToModify[ k ] = { _delete = true }
					else
						tableToModify[ k ] = nil
					end
					
				else

					local subTable
					
					if v._replace then
						subTable = {}
					else

						subTable = tableToModify[ k ]
						if subTable then
							if type( subTable ) ~= 'table' then
								docStoreImpLog:warn( "Replacing key %s with a table", k )
								subTable = {}
							end
						else
							subTable = {}
						end
					
					end
					
					tableToModify[ k ] = applyOverlayToTable( subTable, v,
												retainDeleteAndReplaceEvents )
					
				end
			
			else
		
				tableToModify[ k ] = v
			
			end
		
		end
	
	end
	
	return tableToModify
	
end

--------------------------------------------------------------------------------

-- @mark :docValue

--------------------------------------------------------------------------------

local function rawFetchServerAndLocalValue( self, db, db_exec, docType, docId )

	-- Does this doc exist?

	local docRow = db_exec( db,
		[[SELECT docType, changeCounter, serverBaseValue, localOverlay
			FROM documents WHERE docId == ?]],
		docId )

	if not docRow then
		return
	end

	if docType and docRow.docType ~= docType then
		error( string.format( "docType mismatch: expected %s got %s (id = %s)",
								docRow.docType, docType, docId ) )
	end

	local rawServerBaseValue = docRow.serverBaseValue
	local serverBaseValue = rawServerBaseValue
				and blonde_decode( rawServerBaseValue )
	
	local rawLocalOverlay = docRow.localOverlay
	local localOverlay = rawLocalOverlay
				and blonde_decode( rawLocalOverlay )
	
	return serverBaseValue, localOverlay, docRow.changeCounter
	
end

--------------------------------------------------------------------------------

local function mergeServerAndLocalDocValue( serverBaseValue, localOverlay )

	local result

	docStoreImpLog:debug( "mergeServerAndLocalDocValue\n\nserverBaseValue = %s\n\nlocalOverlay = %s",
							serverBaseValue, localOverlay )

	if serverBaseValue and localOverlay then
		result = applyOverlayToTable( serverBaseValue, localOverlay )
	elseif serverBaseValue then
		result = serverBaseValue
	elseif localOverlay then
		result = applyOverlayToTable( {}, localOverlay )
			-- Ugly, but we have to do this to handle { _delete = true } values
			-- properly.
	else
		result = nil
	end

	docStoreImpLog:debug( "result = %s", result )

	return result

end

--------------------------------------------------------------------------------

local function combinedDocValueImp( self, db, db_exec, docType, docId )

	local serverBaseValue, localOverlay, changeCounter
		= rawFetchServerAndLocalValue( self, db, db_exec, docType, docId )

	local result = mergeServerAndLocalDocValue( serverBaseValue, localOverlay )
	
	return result, changeCounter, localOverlay

end

--------------------------------------------------------------------------------

function docStoreImpMethods:docValue( docType, docId, opts )

	validateDocTypeAndId( docType, docId, 'OzDocStoreImp:docValue', 2 )

	-- Identify the current head revision for this branch.
	
	local db = self._database
	local db_exec = db.exec

	if opts and opts.separateValues then

		local serverBaseValue, localOverlay, changeCounter
			= rawFetchServerAndLocalValue( self, db, db_exec, docType, docId )

		if serverBaseValue then
			serverBaseValue._docId = docId
			serverBaseValue._docType = docType
			serverBaseValue._changeCounter = changeCounter
		end

		return serverBaseValue, localOverlay

	else

		local rawDocValue, changeCounter, localOverlay =
			combinedDocValueImp( self, db, db_exec, docType, docId )
		
		if rawDocValue then
		
			-- PERF [scouten 2011-04-13]: How many clients can we move over
			-- to using rawDocValue instead?
		
			local docValue = rawDocValue
			
			docValue._docId = docId
			docValue._docType = docType
			docValue._changeCounter = changeCounter
			
			if docValue._conflict then
				docValue._ozMetadataUploadStatus = 'conflict'
			end
			
			return docValue, localOverlay
			
		else
	
			-- Synthesize a new empty document.
	
			return {
				_docId = docId,
				_docType = docType,
			}, {}
	
		end
	
	end

end

--------------------------------------------------------------------------------

-- @mark :setServerValue

--------------------------------------------------------------------------------

local function acceptableGpsSlop( k, newServerValue_k, oldLocalOverlay_k )

	local delta = math.abs( newServerValue_k - oldLocalOverlay_k )
	local maxDelta = k == 'altitude' and 0.125 or 0.00000125
			
	return delta < maxDelta

end

--------------------------------------------------------------------------------

local function isFullyDeletedHierarchy( t, optType )
	optType = optType or type( t )
	if optType ~= 'table' then
		return false
	end
	if t._delete == true then
		return true
	end
	-- if all values are tables and all are fully deleted
	-- then we are also fully deleted
	for k, v in pairs( t ) do
		if not isFullyDeletedHierarchy( v ) then
			return false
		end
	end
	return true
end

--------------------------------------------------------------------------------

local function valueMappingDeletes( t, k )
	local ret = t and t[ k ]
	if type( ret ) == 'table' and ret._deleted then
		return nil
	end
	return ret
end

--------------------------------------------------------------------------------

local function calculateNewOverlay( oldServerValue,
									oldLocalOverlay,
									newServerValue,
									conflictJail,
									uploadedValues )

	-- If (for instance, for photodocks), we want to replace the entire document,
	-- we either match the whole thing or we don't. Don't try to calculate the usual
	-- partial overlay.

	if oldLocalOverlay._replace then

		-- newServerValue wouldn't normally have a _replace member, but it could
		-- in the special case we've created to handle PR# 2942322.

		newServerValue._replace = nil

		oldLocalOverlay._conflict = nil
		oldLocalOverlay._replace = nil

		if KSTableUtils.deepEquals( newServerValue, oldLocalOverlay ) then

			return nil, nil

		else
		
			oldLocalOverlay._replace = true
			return oldLocalOverlay, newServerValue
		
		end
	
	end

	-- Not doing full-on replacement: Calculate incremental deltas.

	local newOverlay, newConflictJail
	
	local allKeys = {}
	if oldServerValue then
		for k in pairs( oldServerValue ) do
			allKeys[ k ] = true
		end
	end
	
	if oldLocalOverlay then
		for k in pairs( oldLocalOverlay ) do
			allKeys[ k ] = true
		end
	end
	
	if newServerValue then
		for k in pairs( newServerValue ) do
			allKeys[ k ] = true
		end
	end

	if conflictJail then
		for k in pairs( conflictJail ) do
			allKeys[ k ] = true
		end
	end

	allKeys._conflict = nil

	for k in pairs( allKeys ) do

		local oldServerValue_k = valueMappingDeletes( oldServerValue, k )
		local oldLocalOverlay_k = valueMappingDeletes( oldLocalOverlay, k )
		local newServerValue_k = valueMappingDeletes( newServerValue, k )
		local conflictJail_k = valueMappingDeletes( conflictJail, k )
		local uploadedValues_k = valueMappingDeletes( uploadedValues, k )
		if uploadedValues_k ~= nil then
			oldServerValue_k = uploadedValues_k
		end

		local oldServerValue_kType = type( oldServerValue_k )
		local oldLocalOverlay_kType = type( oldLocalOverlay_k )
		local newServerValue_kType = type( newServerValue_k )

		if ( oldServerValue_k == oldLocalOverlay_k or oldLocalOverlay_k == nil ) then
		
			-- No local change. Let whatever is in the new server value stand.
		
		elseif oldLocalOverlay_k == newServerValue_k then
		
			-- Current local value has co-evolved to match the server's value.
			-- Remove from local overlay.

		elseif oldLocalOverlay_kType == 'table'
			and oldLocalOverlay_k._delete == true
			and newServerValue_k == nil
		then
		
			-- Special case of above. We attempted to delete the value from server
			-- and server has done so. Remove the _delete placeholder from localOverlay.
			-- [scouten 2011-09-01, PR #2958408]
		
		elseif ( k == 'latitude' or k == 'longitude' or k == 'altitude' )
			and oldLocalOverlay_kType == 'number'
			and newServerValue_kType == 'number'
			and acceptableGpsSlop( k, newServerValue_k, oldLocalOverlay_k )
		then
			
			-- Server rounded off the GPS coordinates. If it's within
			-- known tolerance, we'll quietly accept new value.
			-- Remove from local overlay. [scouten 2012-10-10, PR #3335547]
			
		elseif oldServerValue_k == newServerValue_k and conflictJail_k == nil then
		
			-- No remote change. Existing overlay change (if any) stands.
			
			newOverlay = newOverlay or {}
			newOverlay[ k ] = oldLocalOverlay[ k ]

		elseif k == '_serverProvidedMetadata' and newServerValue_k then
		
			-- Server provided metadata always overwrites any local overlay.

		elseif newServerValue_k == nil and isFullyDeletedHierarchy( oldLocalOverlay_k, oldLocalOverlay_kType )
		then
		
			-- Server has deleted a key that we asked it to.
			-- Remove from local overlay.

		elseif ( oldServerValue_kType == 'table'
					and newServerValue_kType == 'table'
					and ( oldLocalOverlay_k == nil or oldLocalOverlay_kType == 'table' ) )
			or ( oldServerValue_k == nil
					and oldLocalOverlay_kType == 'table' and newServerValue_kType == 'table' )
		then
			
			-- Both old and new server values are tables, dig deeper.
			
			local newSubOverlay, subConflictJail
					= calculateNewOverlay( oldServerValue_k or {},
										   oldLocalOverlay_k,
										   newServerValue_k,
										   conflictJail_k,
										   uploadedValues_k or {} )
			
			if newSubOverlay and next( newSubOverlay ) then
				newOverlay = newOverlay or {}
				newOverlay[ k ] = newSubOverlay
			end
			
			if subConflictJail and next( subConflictJail ) then
				newConflictJail = newConflictJail or {}
				newConflictJail[ k ] = subConflictJail
			end
			
		else

			-- Old server value doesn't match either new server value
			-- or current local overlay. This means we have a conflict.
			
			-- OR we might be here because we were already in conflict
			-- and the new server values doesn't match the local, so
			-- we are still in conflict.
	
			newConflictJail = newConflictJail or {}
			if newServerValue_k == nil then
				-- if we are in conflict and the new server value is nil
				-- we must remember that
				newConflictJail[ k ] = { _deleted = true }
			else
				newConflictJail[ k ] = newServerValue_k
			end
			
			newOverlay = newOverlay or {}
			newOverlay[ k ] = oldLocalOverlay[ k ]
		
		end

	end
	
	return newOverlay, newConflictJail

end

--------------------------------------------------------------------------------

local function updateUploadStatusInLocalOverlay( localOverlay, forceUploadStatusToUploaded )

	if not localOverlay then return end

	localOverlay._ozMetadataUploadStatus = 'uploaded'
	
	if forceUploadStatusToUploaded then
		return
	end
	
	for k in pairs( localOverlay ) do

		if type( k ) ~= 'string' or not k:match( '^_' ) then
		
			localOverlay._ozMetadataUploadStatus = nil
			break

		end

	end

end

--------------------------------------------------------------------------------

local function lazyLoad_KSDate()

	if not KSDate then
		KSDate = assert( require 'KSDate' )
	end

end

--------------------------------------------------------------------------------

local function setServerValue_inTransaction( self, db, db_exec,
											 docType, docId, newServerValue,
											 options )
	
	options = options or {}
	
	-- Does this doc already exist?

	local docRow = db_exec( db,
		[[SELECT docType, changeCounter, serverBaseValue, localOverlay
			FROM documents WHERE docId == ?]],
		docId )

	if not docRow then
	
		-- No existing doc: This is very easy.
		
		db_exec( db,
			[[INSERT INTO documents( docId, docType, changeCounter,
									serverBaseValue, localOverlay )
							VALUES( ?, ?, incrementChangeCounter(), ?, ? )]],
						docId, docType, blonde_encode( newServerValue ),
						blonde_encode { _ozMetadataUploadStatus = 'uploaded' } )
		
		-- FIXME [scouten 2011-05-03]: Want to get rid of this knowledge
		-- of sync status down at this level.
	
	else
		
		local forceUploadStatusToUploaded
		if newServerValue._deleted then
			forceUploadStatusToUploaded = true
		end
		
		local oldServerValue = docRow.serverBaseValue
								and blonde_decode( docRow.serverBaseValue )
								or {}

		-- Logging for false-favorite behavior.
		
		local osv_hasFavorites = not not oldServerValue.favorites
		local nsv_hasFavorites = not not newServerValue.favorites
		
		if osv_hasFavorites ~= nsv_hasFavorites
			or ( nsv_hasFavorites
				and not KSTableUtils.deepEquals( oldServerValue.favorites,
												 newServerValue.favorites ) )
		then
			favoritesLog:debug( "setServerValue: changing favorites for doc %s\n\nold = %s\n\nnew = %s",
								docId, oldServerValue.favorites, newServerValue.favorites )
		end

		-- No-op if old and new values are same.

		if KSTableUtils.deepEquals( oldServerValue, newServerValue ) then
			return
		end

		-- Calculate new local overlay and conflict jail.
		
		local oldLocalOverlay = docRow.localOverlay
								and blonde_decode( docRow.localOverlay )
								or {}

		-- Clear local overlay if so requested.
		
		if options.clearLocalOverlay then
			oldLocalOverlay = { _localOnly = oldLocalOverlay._localOnly }
		end
		
		-- Pull out whatever values we've already told the server about
		local uploadedValues = oldLocalOverlay._localOnly and oldLocalOverlay._localOnly._uploadedValues
		if uploadedValues then
			oldLocalOverlay._localOnly._uploadedValues = nil
		end
		
		-- Did we have conflicts before?
		
		local conflictJail = oldLocalOverlay._conflict
		local hadPriorConflict = not not conflictJail

		docStoreImpLog:debug( "oldServerValue = %s\noldLocalOverlay = %s\nnewServerValue = %s",
								oldServerValue, oldLocalOverlay, newServerValue )

		if docType == 'asset' then
			docStoreAssetLog:debug( "oldServerValue = %s\noldLocalOverlay = %s\nnewServerValue = %s",
									oldServerValue, oldLocalOverlay, newServerValue )
		end

		if docType == 'space' then
			docStoreSpaceLog:debug( "oldServerValue = %s\noldLocalOverlay = %s\nnewServerValue = %s",
									oldServerValue, oldLocalOverlay, newServerValue )
		end

		local newLocalOverlay, newConflictJail
			= calculateNewOverlay(
							oldServerValue,
							oldLocalOverlay,
							newServerValue,
							conflictJail,
							uploadedValues or {} )

		-- FIXME [scouten 2011-05-03]: Want to get rid of this knowledge
		-- of sync status down at this level.
	
		if not newLocalOverlay then
			newLocalOverlay = {}
		end
		
		-- FIXME [scouten 2011-08-12]: Hacky way to fix up the spurious conflict
		-- after going into develop. The problem, fundamentally, is a race condition.
		-- If we cycle through changes with timestamp revisions fast enough that
		-- we're one or two behind the server when we make a new one, then it looks
		-- like a three-way merge conflict, even though it really isn't. So, for now,
		-- we just enforce the rule that changedAtTime by itself won't generate a
		-- conflict. (Fixes Watson bugs 2941131, 2942071, 2941777, 2941837.)
		
		if newConflictJail then
		
			if next( newConflictJail ) == 'changedAtTime'
				and next( newConflictJail, 'changedAtTime' ) == nil
			then
				newConflictJail.changedAtTime = nil
			end
		
		end
		
		-- FIXME [scouten 2011-11-17]: Repeat above hack for favorites. Note that
		-- the act of favoriting doesn't bump changedAtTime, so the two conditions
		-- shouldn't happen simultaneously. (Fixes Watson bug 3021617.)
		
		if newConflictJail then
		
			if next( newConflictJail ) == 'favorites'
				and next( newConflictJail, 'favorites' ) == nil
			then
				newConflictJail.favorites = nil
			end
		
		end
		
		-- Special intercept for eventName conflicts.
		-- We handle these by keeping the latest change i.e the
		-- eventName with the latest eventTimestamp value.

		if newConflictJail then
		
			-- If both of these values are not present in the conflict jail
			-- we retain the local value.
			
			local serverEventName = newConflictJail.eventName
			local localEventName = newLocalOverlay.eventName
				
			local serverTimestamp = newConflictJail.eventTimestamp
			local localTimestamp = newLocalOverlay.eventTimestamp
			
			-- If we don't have these four values, we can't make a decision
			-- so do nothing. Noticed a console log where we were erroring
			-- out due to no localTimestamp being present. Adding this check
			-- to prevent that from happening.
			
			if serverEventName and localEventName and serverTimestamp and localTimestamp then
				
				lazyLoad_KSDate()

				local diff = KSDate.differenceInSeconds( serverTimestamp, localTimestamp, false )
								
				-- Keep the most recent change.
				
				if diff >= 0 then
					newLocalOverlay.eventName = localEventName
					newLocalOverlay.eventTimestamp = localTimestamp
				else
					newLocalOverlay.eventName = serverEventName
					newLocalOverlay.eventTimestamp = serverTimestamp
				end
					
			end
			
			-- Delete from conflict jail to prevent the conflict from reaching the user.

			newConflictJail.eventName = nil
			newConflictJail.eventTimestamp = nil

		end
		
		-- Ignore any conflicts generated for album names. The last name to the server wins.

		if newConflictJail and next( newConflictJail ) == 'name' and docType == 'album' then
			newConflictJail = nil
		end
		
		updateUploadStatusInLocalOverlay( newLocalOverlay, forceUploadStatusToUploaded )

		if newConflictJail and next( newConflictJail ) then

			newLocalOverlay._conflict = newConflictJail
			
			conflictLog:debug( "flagged new conflict on %s %s\n\nnew localOverlay is %s",
							   docType, docId, newLocalOverlay )

			-- [scouten 2012-06-09, PR #2957019]: Make sure we don't lose the changedOnDevice
			-- from the conflict. If oldServerValue.changedOnDevice and oldLocalOverlay.changedOnDevice
			-- happened to match, the normal three-way merge logic fails us here.
			
			if oldServerValue.changedOnDevice == oldLocalOverlay.changedOnDevice
				and oldServerValue.changedOnDevice ~= newServerValue.changedOnDevice
			then
				newLocalOverlay.changedOnDevice = oldLocalOverlay.changedOnDevice
				newConflictJail.changedOnDevice = newServerValue.changedOnDevice
			end
			
			local conflictDetectedLog = self._conflictDetectedLog
			
			if conflictDetectedLog[ 21 ] then
				table.remove( conflictDetectedLog, 1 )
			end
			
			lazyLoad_KSDate()
			
			conflictDetectedLog[ #conflictDetectedLog + 1 ] = {
			
				docId = docId,
				docType = docType,
				time = KSDate.timeInIso8601FormatWithTimeZone(),
				
				-- Note that we recompute oldServerValue, etc., because
				-- calculateNewOverlay can mutate the underlying tables.
				-- Copying those tables in the default case is too expensive.
				-- So, we assume that conflict detection (and logging) is the
				-- abnormal case, and pay the cost of redoing the BLONDE
				-- decode here.
				
				oldServerValue = docRow.serverBaseValue
									and blonde_decode( docRow.serverBaseValue )
									or "(missing)",

				oldLocalOverlay = docRow.localOverlay
									and blonde_decode( docRow.localOverlay )
									or "(missing)",
				
				newServerValue = newServerValue,
				newLocalOverlay = newLocalOverlay,
			
			}

		else
		
			if hadPriorConflict then
			
				local conflictDetectedLog = self._conflictDetectedLog
				
				if conflictDetectedLog[ 21 ] then
					table.remove( conflictDetectedLog, 1 )
				end
				
				lazyLoad_KSDate()
				
				conflictDetectedLog[ #conflictDetectedLog + 1 ] = {
				
					docId = docId,
					docType = docType,
					time = KSDate.timeInIso8601FormatWithTimeZone(),
					
					conflictResolved = true,
					
					-- Note that we recompute oldServerValue, etc., because
					-- calculateNewOverlay can mutate the underlying tables.
					-- Copying those tables in the default case is too expensive.
					-- So, we assume that conflict detection (and logging) is the
					-- abnormal case, and pay the cost of redoing the BLONDE
					-- decode here.
					
					oldServerValue = docRow.serverBaseValue
										and blonde_decode( docRow.serverBaseValue )
										or "(missing)",
	
					oldLocalOverlay = docRow.localOverlay
										and blonde_decode( docRow.localOverlay )
										or "(missing)",
					
					newServerValue = newServerValue,
					newLocalOverlay = newLocalOverlay,
				
				}
	
			end
		
			-- GRUNGE [scouten 2011-07-28]: Don't like pushing knowledge of specific
			-- content types down to this level, but I've gotta do it.
			-- Keep track of the rev where develop settings changed.
			
			local oldDevelopSettings = oldServerValue.develop or {}
			local newDevelopSettings = newServerValue.develop or {}
			
			if not KSTableUtils.deepEquals( oldDevelopSettings, newDevelopSettings ) then
			
				local revId = newServerValue._serverProvidedMetadata
					and newServerValue._serverProvidedMetadata.links
					and newServerValue._serverProvidedMetadata.links[ '/rels/latest_revision' ]
					and newServerValue._serverProvidedMetadata.links[ '/rels/latest_revision' ].href
				
				if revId then
					revId = revId:sub( -32 )
				end
				
				newLocalOverlay._localOnly = newLocalOverlay._localOnly or {}
				
				newLocalOverlay._localOnly.devSettingsLastSetAtRev = revId
				
			end
		
		end
		
		-- Stash it.

		docStoreImpLog:debug( "newLocalOverlay = %s", newLocalOverlay )

		if docType == 'asset' then
			docStoreAssetLog:debug( "newLocalOverlay = %s", newLocalOverlay )
		end
		
		if docType == 'space' then
			docStoreSpaceLog:debug( "newLocalOverlay = %s", newLocalOverlay )
		end
		
		db_exec( db,
			[[UPDATE documents
				SET changeCounter = incrementChangeCounter(),
					serverBaseValue = ?,
					localOverlay = ?
				WHERE docId = ?]],
					blonde_encode( newServerValue ),
					( newLocalOverlay and next( newLocalOverlay ) )
						and blonde_encode( newLocalOverlay )
						or nil,
					docId )
	
	end
	
	updateChangeCounter( self, db )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:setServerValue( docType, docId, serverValue, options )

	docStoreImpLog:debug( "setServerValue %s %s %s %s", docType, docId, serverValue, options )

	if docType == 'asset' then
		docStoreAssetLog:debug( "setServerValue %s %s %s %s", docType, docId, serverValue, options )
	end

	if docType == 'space' then
		docStoreSpaceLog:debug( "setServerValue %s %s %s %s", docType, docId, serverValue, options )
	end

	validateDocTypeAndId( docType, docId, 'OzDocStoreImp:setServerValue', 2 )

	assert( type( serverValue ) == 'table',
		"OzDocStoreImp:setServerValue: serverValue must be a table" )

	local db = self._database
	local db_exec = db.exec

	return db:withTransactionDo( function()

		return setServerValue_inTransaction( self, db, db_exec, docType, docId,
					serverValue, options ) or true
	
	end )

end

--------------------------------------------------------------------------------

-- @mark :resolveConflictP

--------------------------------------------------------------------------------

local ResolveWith = 'OzDocStore.resolveWith'

WFArguments.registerChecker( ResolveWith, {
	[ 'local' ] = true,
	server = true,
	fork = true,
} )

local allowSpecialKeys = {} -- special internal option for :applyOverlay
local clearLocalOverlay = {} -- special internal option for :applyOverlay

--------------------------------------------------------------------------------

function docStoreImpMethods:resolveConflict( docType, docId, changeCounter, resolveWith )

	WFArguments.check( OzDocStoreImp, DocumentType, DocumentId, 'number', ResolveWith )

	docStoreImpLog:debug( "OzDocStoreImp:resolveConflict %s %s %d %s", docType, docId,
						changeCounter, resolveWith )

	local db = self._database
	local db_exec = db.exec

	local _, localOverlay, actualChangeCounter
		= rawFetchServerAndLocalValue( self, db, db_exec, docType, docId )
	
	if actualChangeCounter ~= changeCounter then
		error "document has evolved since expected change counter"
	end

	if not localOverlay._conflict then
		error "there is no conflict to resolve"
	end
	
	if resolveWith == 'local' then
	
		-- Resolve with local is really easy. We just kill the _conflict jail
		-- and make this document eligible for upload with the local values.

		return self:applyOverlay {
			docType = docType,
			docIds = { docId },
			overlay = { _conflict = { _delete = true } },
			allowSpecialKeys = allowSpecialKeys,
		}
	
	elseif resolveWith == 'server' then
	
		-- Resolve with server is a little more tricky. We have to go hack on
		-- the local overlay; deleting the _conflict jail and all local values.
		-- That would be easy, except for the part where we have to preserve
		-- most of the local _-prefixed state variables.
		
		return self:applyOverlay {
			docType = docType,
			docIds = { docId },
			overlay = {},
			clearLocalOverlay = clearLocalOverlay,
			allowSpecialKeys = allowSpecialKeys,
		}
	
	else
	
		error( string.format( "unknown value for resolveWith value: %s",
							tostring( resolveWith ) ) )
	
	end

end

--------------------------------------------------------------------------------

-- @mark :applyOverlay

--------------------------------------------------------------------------------

local underscoreByte = string.byte( '_' )

--------------------------------------------------------------------------------

local acceptableSpecialKeys = {

	_deleted = true,
	_replace = true,

	_ozCentralRevisionId = true,
	_ozMetadataUploadStatus = true,
		-- FIXME [scouten 2011-06-02]: Should be removed from this list.

	_localOnly = true,
		-- free-form data that is not synced to Oz

	_serverProvidedMetadata = true,
		-- data provided by Oz server that should not be modified by Wichita

	_halt = true,

	_bump = true,
		-- special key used to trigger reindexing

	_conflict = true,
		-- FIXME [scouten 2011-09-04]: Would be better expressed as a
		-- _valueImExpectingToReplace member for develop, but this is deemed
		-- a less-intrusive solution.

}

--------------------------------------------------------------------------------

local function applyOverlayToOneDoc_inTransaction( self, db, db_exec,
					docType, docId, overlay, clearLocalOverlay, clearDelete )

	if overlay.favorites then
		favoritesLog:debug( "applyOverlay: changing favorites for doc %s\n\noverlay = %s",
							docId, overlay )
	end

	local serverBaseValue, localOverlay = rawFetchServerAndLocalValue( self,
												db, db_exec, docType, docId )

	if serverBaseValue or localOverlay then
	
		local overlay_replace = overlay._replace
		if overlay_replace == nil then
			overlay_replace = localOverlay._replace
		end
	
		docStoreImpLog:debug( "applyOverlay started with\n\nserverBaseValue = %s\n\nlocaloverlay = %s\nnewOverlay = %s\n\n", serverBaseValue, localOverlay, overlay )

		if docType == 'asset' then
			docStoreAssetLog:debug( "applyOverlay started with\n\nserverBaseValue = %s\n\nlocaloverlay = %s\nnewOverlay = %s\n\n", serverBaseValue, localOverlay, overlay )
		end

		if docType == 'space' then
			docStoreSpaceLog:debug( "applyOverlay started with\n\nserverBaseValue = %s\n\nlocaloverlay = %s\nnewOverlay = %s\n\n", serverBaseValue, localOverlay, overlay )
		end

		-- Special case used by resolveConflict: Toss everything in local overlay
		-- except for the local state stuff.

		if clearLocalOverlay then
		
			for k in pairs( localOverlay ) do
				if k:byte( 1 ) ~= underscoreByte then
					localOverlay[ k ] = nil
				end
			end
			localOverlay._conflict = nil
			overlay_replace = nil
		
		end
		
		local newLocalValue = applyOverlayToTable( localOverlay, overlay, true )
		
		if overlay_replace == true then
			newLocalValue._replace = true
		end

		-- This case can occur when we create a document, undo it, and then redo the creation.
		-- clearDelete is set to true when we're coming through undo and redo.
		-- Otherwise, we generally want to leave deletion in place.
		
		if clearDelete and not overlay._deleted then
			localOverlay._deleted = nil
		end
		
		updateUploadStatusInLocalOverlay( localOverlay )

		docStoreImpLog:debug( "applyOverlay generated\n\nnewLocalValue = %s\n\n", newLocalValue )

		if docType == 'asset' then
			docStoreAssetLog:debug( "applyOverlay generated\n\nnewLocalValue = %s\n\n", newLocalValue )
		end
		
		if docType == 'space' then
			docStoreSpaceLog:debug( "applyOverlay generated\n\nnewLocalValue = %s\n\n", newLocalValue )
		end
		
		-- FIXME [scouten 2011-04-28]: Remove values from overlay
		-- if they match server. Also clear conflict markers.
		
		db_exec( db, [[UPDATE documents SET
							changeCounter = incrementChangeCounter(),
							localOverlay = ?
							WHERE docId == ?]],
						blonde_encode( newLocalValue ),
						docId )

	else

		if not docType then
			error "OzDocStoreImp:applyOverlay: docType must be provided when creating new documents"
		end

		-- FIXME [scouten 2011-08-13]: Hackiness. I need to clean up the handling of _delete.

		local newOverlay = {}
		for k, v in pairs( overlay ) do
			newOverlay[ k ] = v
		end
		
		if newOverlay._deleted ~= true then
			newOverlay._deleted = nil
		end

		db_exec( db, [[INSERT INTO documents( docId, docType,
							changeCounter, localOverlay )
							VALUES( ?, ?,
								incrementChangeCounter(), ? )]],
						docId, docType, blonde_encode( newOverlay ) )
	
	end

end

--------------------------------------------------------------------------------

function docStoreImpMethods:applyOverlay( args )

	assert( type( args ) == 'table' )
	
	local docType = args.docType
	local docIds = assert( args.docIds )
	local overlay = args.overlay

	applyOverlayLog:debug( "OzDocStore:applyOverlay %s", args )

	if type( overlay ) ~= 'table' then
		error( "OzDocStoreImp:applyOverlay: overlay must be a table" )
	end
	
	if args.clearLocalOverlay then
		assert( args.clearLocalOverlay == clearLocalOverlay,
			"OzDocStoreImp:applyOverlay: clearLocalOverlay is an internal option only" )
	end

	if args.allowSpecialKeys ~= allowSpecialKeys then

		for k in pairs( overlay ) do
		
			if k:byte( 1 ) == underscoreByte and not acceptableSpecialKeys[ k ] then
				error( string.format(
					"OzDocStoreImp:applyOverlay: illegal key %s",
						k ), 2 )
			end
		
		end
	
	end

	-- Insert data and device name for change.

	if args.addRevTimestamp then

		lazyLoad_KSDate()

		overlay.changedAtTime = KSDate.timeInIso8601Format( nil, true )
		overlay.changedOnDevice = KSDeviceUtils.deviceName()

		applyOverlayLog:debug( "applied timestamp %s to %s %s", overlay.changedAtTime,
						docType, docIds[ 1 ] )

	end

	local db = self._database
	local db_exec = db.exec

	return db:withTransactionDo( function()

		if overlay._halt and DEBUG then
			overlay._halt = nil
			halt()
		end

		local nDocs = #docIds
		
		for i = 1, nDocs do

			applyOverlayToOneDoc_inTransaction( self, db, db_exec,
					docType, docIds[ i ], overlay,
					args.clearLocalOverlay, args.clearDelete )

		end

		updateChangeCounter( self, db )
		
	end )
			
end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------

function docStoreImpMethods:addDocChangeCounterRequest()

	return self._currentChangeCounterB:relayB()

end

--------------------------------------------------------------------------------

function docStoreImpMethods:docChangeCounter()

	return self._currentChangeCounterB:valueNow()

end

--------------------------------------------------------------------------------

local changedDocs_maxLimit = 50
local changedDocs_maxLimitForIds = 1000
	-- Both of these are arbitrary; tune based on observed behavior.

--------------------------------------------------------------------------------

function docStoreImpMethods:changedDocs( args )

	changedDocsLog:debug( "changedDocs is called with args %s", args )

	local db = self._database

	local from = args.fromChangeCounter
	if not from then
		error( "OzDocStoreImp:changedDocs: fromChangeCounter argument missing", 2 )
	end

	local maxLimit = args.idsOnly
						and changedDocs_maxLimitForIds
						or changedDocs_maxLimit
	
	local limit = args.limit or maxLimit
	if type( limit ) ~= 'number' then
		error( "OzDocStoreImp:changedDocs: limit must be a number or nil for default", 2 )
	end
	
	if limit > maxLimit then
		limit = maxLimit
	end

	local docs = args.docs
	if docs and ( type( docs ) ~= 'table' or #docs == 0 ) then
		error( "OzDocStoreImp:changedDocs: docs must be a non-empty array of doc IDs or nil for all docs", 2 )
	end

	local idsOnly = args.idsOnly
	
	local rawDocRows = db:withTransactionDo( function()

		local rawDocRows

		if docs then
	
			if #docs > 100 then
				changedDocsLog:warn( "WARNING: doc inclusion list is awfully large %s", #docs )
			end
	
			db:exec [[DELETE FROM TEMP_docIds]]

			for i = 1, #docs do
				db:exec( [[INSERT INTO TEMP_docIds VALUES( ? )]], docs[ i ] )
			end

			rawDocRows = db:gatherArrayOfArray(
				[[SELECT documents.docId, docType, changeCounter
								FROM documents
									INNER JOIN TEMP_docIds
									WHERE changeCounter >= ?
										AND documents.docId
											== TEMP_docIds.docId
									ORDER BY changeCounter
									LIMIT ? ]],
								from,
								limit + 1 )

			-- PERF [scouten 2011-04-28]: With new DB architecture, could
			-- also grab serverBaseValue and localOverlay and thus bypass
			-- the DB lookups in rawDocValueImp below.

		else
		
			rawDocRows = db:gatherArrayOfArray(
				[[SELECT docId, docType, changeCounter
								FROM documents
									WHERE changeCounter >= ?
									ORDER BY changeCounter
									LIMIT ? ]],
								from,
								limit + 1 )
	
		end
		
		return rawDocRows
	
	end )
	
	local result = {}
	
	if idsOnly then

		for i = 1, math.min( #rawDocRows, limit ) do
		
			result[ i ] = rawDocRows[ i ][ 1 ]
			
		end
	
	else
	
		local db_exec = db.exec

		for i = 1, math.min( #rawDocRows, limit ) do
		
			local rawDocRow = rawDocRows[ i ]
			local docType = rawDocRow[ 2 ]
			local docId = rawDocRow[ 1 ]
		
			local thisDoc = combinedDocValueImp( self, db, db_exec,
								docType, docId )
		
			thisDoc._docType = docType
			thisDoc._docId = docId
			thisDoc._changeCounter = rawDocRow[ 3 ]
			
			result[ i ] = thisDoc
		
		end
	
	end

	if rawDocRows[ limit + 1 ] then
		result.continueFrom = rawDocRows[ limit + 1 ][ 3 ]
	else
		result.continueFrom = self._currentChangeCounterB:valueNow()
	end

	changedDocsLog:debug( "changedDocs is returning %s", result )

	return result

end

--------------------------------------------------------------------------------

-- @mark performance instrumentation

--------------------------------------------------------------------------------

function docStoreImpMethods:halt()

	local path = self._path
	
	self:close() -- should cause DB object to close

	return path or false

end

--------------------------------------------------------------------------------

-- @mark performance instrumentation

--------------------------------------------------------------------------------

function docStoreImpMethods:PERF_collectQueryPerformanceStats()

	self._database:PERF_collectQueryPerformanceStats()

end

--------------------------------------------------------------------------------

function docStoreImpMethods:PERF_dumpQueryPerformanceStats()

	self._database:PERF_dumpQueryPerformanceStats()

end

--------------------------------------------------------------------------------

function docStoreImpMethods:DEBUG_dumpCoverageP( path )

	require 'SetupCoverage'.dumpCoverage( path )

	return require 'KSPromises'.resolvedPromiseP( 'data', true )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:sendDiagnosticLogData()

	-- For now, all we're interested in from doc store imp is the conflict
	-- detection log. Disable this log section unless we actually have conflicts
	-- so as not to cause unnecessary alarm to those without conflicts.

	if self._conflictDetectedLog and self._conflictDetectedLog[ 1 ] then
	
		return {
		
			name = "Conflicts Detected :: " .. WFCurrentThread.DEBUG_getThreadName(),
			id = 'OzDocStoreImp-' .. WFCurrentThread.DEBUG_getThreadName():gsub( ' :: ', '+' ),
			weight = 1500,
		
			body = self._conflictDetectedLog,
		
		}
	
	end

end

--------------------------------------------------------------------------------

-- @mark event processor infrastructure

--------------------------------------------------------------------------------

function docStoreImpMethods:addEventProcessorMessageHandlers()

	WFRequestDispatcher.onMessageDo {
		message = 'localStateP',
		target = self,
		onlyIfCallbackIsStillValid = true,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'localStateB',
		target = self,
		onlyIfCallbackIsStillValid = true,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'setLocalState',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'docValue',
		target = self,
		onlyIfCallbackIsStillValid = true,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'setServerValue',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'resolveConflict',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'applyOverlay',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'changedDocs',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'addDocChangeCounterRequest',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'docChangeCounter',
		target = self,
		onlyIfCallbackIsStillValid = true,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'halt',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'PERF_collectQueryPerformanceStats',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'PERF_dumpQueryPerformanceStats',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'DEBUG_dumpCoverageP',
		target = self,
	}
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:__tostring()

	return string.format( "OzDocStoreImp( %s )", self._database:path() )

end

--------------------------------------------------------------------------------

return KSClass 'OzDocStoreImp'( docStoreImpMethods )
