--[[----------------------------------------------------------------------------

OzDocStoreTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-03-01

FIXME [scouten 2011-04-28]: With new DB architecture, needs deeper testing of
nested table cases.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- core.utils
local KSTableUtils = require 'KSTableUtils'
local KSTestUtils = require 'KSTestUtils'
local KSIdGenerator = require 'KSIdGenerator'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- oz.doc_store
local OzDocStore = require 'OzDocStore'


--============================================================================--

local OzDocStoreTest = {}

local docStore, didProfileSqlQueries

local TEST_WAIT_TIME = WFPlatformInfo.isAndroid and 9 or 3 		-- PERFORMANCE_ANDROID

local randomDocId = '9b572d0e423c4654b078893ddd41c230'

--------------------------------------------------------------------------------

function OzDocStoreTest.setUp()

	docStore = OzDocStore( ':memory:' )

	didProfileSqlQueries = false

end

--------------------------------------------------------------------------------

function OzDocStoreTest.tearDown( coveragePath )

	if docStore then
		if coveragePath then
			local oldDocStore = docStore
			docStore:DEBUG_dumpCoverageP( coveragePath )
				:whenResolvedDo( function()
					oldDocStore:close()
				end )
				:runUntilDone()
		else
			docStore:close()
		end
		docStore = nil
	end

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalState_boolean( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', false )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', false )
		
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', true )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', true )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalState_nilByDefault( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:localStateP( nil )
		
		end )
		:continueWhenValueMatches( 'data', nil )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalState_noCrossPollution( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', 'bar' )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', 'bar' )
		
		:waitFor( function()
			
			return docStore:setLocalStateP( 'bar', 'foo' )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'bar' )
		
		end )
		:continueWhenValueMatches( 'data', 'foo' )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', 'bar' )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalState_stringThenNil( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', 'bar' )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', 'bar' )
		
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', nil )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( nil )
		
		end )
		:continueWhenValueMatches( 'data', nil )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalStateB( test )

	local fooB = docStore:localStateB( 'foo' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Initial state should be nil.

		:waitFor( fooB )		-- step 1
		:continueWhenValueMatches( nil )
	
		-- Set state and wait for it to read back.
	
		:waitFor( function()	-- step 2
			
			return docStore:setLocalStateP( 'foo', 'bar' )
			
		end )
		
		:waitFor( fooB )		-- step 3
		:continueWhenValueMatches( 'bar' )
		
		:waitFor( function()	-- step 4
			
			return docStore:setLocalStateP( 'foo', nil )
			
		end )
		
		:waitFor( fooB )		-- step 5
		:continueWhenValueMatches( nil )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testLocalState_table( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:setLocalStateP( 'foo', {} )
			
		end )
		
		:waitFor( function()

			return docStore:localStateP( 'foo' )
		
		end )
		:continueWhenValueMatches( 'data', {} )
		
end

--------------------------------------------------------------------------------

-- @mark :setServerValueP

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_conflict( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new title as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = "Local New Title",
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				title = "Local New Title",
			},
			{
				title = "Local New Title",
			}
		)

		-- Meanwhile, the title changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					title = "Server's New Title",
				},
				favorite = true,
				title = "Local New Title",
			},
			{
				_conflict = {
					title = "Server's New Title",
				},
				title = "Local New Title",
			}
		)

		-- lets ensure we propagate conflicts forward from one
		-- server change to another...

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						baz = "Unrelated Field Changed",
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 4,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					title = "Server's New Title",
				},
				favorite = true,
				baz = "Unrelated Field Changed",
				title = "Local New Title",
			},
			{
				_conflict = {
					title = "Server's New Title",
				},
				title = "Local New Title",
			}
		)

		-- lets ensure we propagate conflicts forward
		-- even if it involves deletion

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						baz = "Another Unrelated Field Changed",
						-- notice title has been deleted here.
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 5,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					title = { _deleted = true },
				},
				favorite = true,
				baz = "Another Unrelated Field Changed",
				title = "Local New Title",
			},
			{
				_conflict = {
					title = { _deleted = true },
				},
				title = "Local New Title",
			}
		)

		-- Now lets see if we autoresolve when the local
		-- version also goes to deleted

		:waitFor( function()
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = { _deleted = true },
						},
					}
			
		end )
		
		
		-- (Resolution only happens if we set a new server value.)
		
		:waitFor( function()
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						baz = "Yet another Unrelated Field Changed",
						-- notice title has been deleted here.
					} )
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 7,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				baz = "Yet another Unrelated Field Changed",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_noConflict( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set initial value as though it came from server.

		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Title changed on server again.

		:waitFor( function()		-- step 3
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "Server's New Title",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new title as though it came from local user.
		-- That should override server value without conflict.

		:waitFor( function()		-- step 5
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = "Local New Title",
						},
					}
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				favorite = true,
				title = "Local New Title",
			},
			{
				title = "Local New Title",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_coevolvedValues( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
					TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Family and Friends/",
					TEMP_rootRevId = "925abca260194ec2b86a44985829f512",
					_ozCentralRevisionId = "437a9b4ecdcf490eb1adc49a237e053f",
					captureDate = "2009-08-17T18:12:22Z",
					favorites = {
						fredx235 = false,
					},
					height = 3263,
					keywords = {
						Ayla = true,
						["Family and Friends"] = true,
						["my family"] = true,
					},
					originalFileName = "es-1512-044.jpg",
					width = 2175,
				} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Family and Friends/",
				TEMP_rootRevId = "925abca260194ec2b86a44985829f512",
				_ozCentralRevisionId = "437a9b4ecdcf490eb1adc49a237e053f",
				_ozMetadataUploadStatus = "uploaded",
				captureDate = "2009-08-17T18:12:22Z",
				favorites = {
					fredx235 = false,
				},
				height = 3263,
				keywords = {
					Ayla = true,
					["Family and Friends"] = true,
					["my family"] = true,
				},
				originalFileName = "es-1512-044.jpg",
				width = 2175,
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new values as though they came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							_ozCentralRevisionId = "4cf4af113d5a43119862344fcc77634f",
							_ozMetadataUploadStatus = "uploaded",
							favorites = {
								fredx235 = true,
							},
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Family and Friends/",
				TEMP_rootRevId = "925abca260194ec2b86a44985829f512",
				_ozCentralRevisionId = "4cf4af113d5a43119862344fcc77634f",
				captureDate = "2009-08-17T18:12:22Z",
				favorites = {
					fredx235 = true,
				},
				height = 3263,
				keywords = {
					Ayla = true,
					["Family and Friends"] = true,
					["my family"] = true,
				},
				originalFileName = "es-1512-044.jpg",
				width = 2175,
			},
			{
				_ozCentralRevisionId = "4cf4af113d5a43119862344fcc77634f",
				favorites = {
					fredx235 = true,
				},
			}
		)

		-- Now read these values back from server. They should not
		-- be a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Family and Friends/",
						TEMP_rootRevId = "925abca260194ec2b86a44985829f512",
						_ozCentralRevisionId = "4cf4af113d5a43119862344fcc77634f",
						captureDate = "2009-08-17T18:12:22Z",
						favorites = {
							fredx235 = true,
						},
						height = 3263,
						keywords = {
							Ayla = true,
							["Family and Friends"] = true,
							["my family"] = true,
						},
						originalFileName = "es-1512-044.jpg",
						width = 2175,
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Family and Friends/",
				TEMP_rootRevId = "925abca260194ec2b86a44985829f512",
				_ozCentralRevisionId = "4cf4af113d5a43119862344fcc77634f",
				_ozMetadataUploadStatus = "uploaded",
				captureDate = "2009-08-17T18:12:22Z",
				favorites = {
					fredx235 = true,
				},
				height = 3263,
				keywords = {
					Ayla = true,
					["Family and Friends"] = true,
					["my family"] = true,
				},
				originalFileName = "es-1512-044.jpg",
				width = 2175,
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_coevolvedValuesFromEmpty( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Public/",
					captureDate = "2009-05-30T09:02:06Z",
					height = 2449,
					keywords = {
						America = true,
						Hawaii = true,
						Maui = true,
						[ "North America" ] = true,
						Public = true,
						[ "R&R" ] = true,
						Summer = true,
						[ "U.S." ] = true,
						US = true,
						USA = true,
						[ "United States" ] = true,
						[ "United States of America" ] = true,
						architectural = true,
						architecture = true,
						building = true,
						[ "commercial building" ] = true,
						concept = true,
						concepts = true,
						conceptual = true,
						edifice = true,
						edifices = true,
						[ "getting away from it all" ] = true,
						paradise = true,
						[ "rest and relaxation" ] = true,
						seasons = true,
						shop = true,
						store = true,
						structures = true,
						summertime = true,
						tropical = true,
					},
					originalFileName = "es-1437-022.jpg",
					width = 3673,
				},
			}
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Public/",
				captureDate = "2009-05-30T09:02:06Z",
				height = 2449,
				keywords = {
					America = true,
					Hawaii = true,
					Maui = true,
					[ "North America" ] = true,
					Public = true,
					[ "R&R" ] = true,
					Summer = true,
					[ "U.S." ] = true,
					US = true,
					USA = true,
					[ "United States" ] = true,
					[ "United States of America" ] = true,
					architectural = true,
					architecture = true,
					building = true,
					[ "commercial building" ] = true,
					concept = true,
					concepts = true,
					conceptual = true,
					edifice = true,
					edifices = true,
					[ "getting away from it all" ] = true,
					paradise = true,
					[ "rest and relaxation" ] = true,
					seasons = true,
					shop = true,
					store = true,
					structures = true,
					summertime = true,
					tropical = true,
				},
				originalFileName = "es-1437-022.jpg",
				width = 3673,
			},
			{
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Public/",
				captureDate = "2009-05-30T09:02:06Z",
				height = 2449,
				keywords = {
					America = true,
					Hawaii = true,
					Maui = true,
					[ "North America" ] = true,
					Public = true,
					[ "R&R" ] = true,
					Summer = true,
					[ "U.S." ] = true,
					US = true,
					USA = true,
					[ "United States" ] = true,
					[ "United States of America" ] = true,
					architectural = true,
					architecture = true,
					building = true,
					[ "commercial building" ] = true,
					concept = true,
					concepts = true,
					conceptual = true,
					edifice = true,
					edifices = true,
					[ "getting away from it all" ] = true,
					paradise = true,
					[ "rest and relaxation" ] = true,
					seasons = true,
					shop = true,
					store = true,
					structures = true,
					summertime = true,
					tropical = true,
				},
				originalFileName = "es-1437-022.jpg",
				width = 3673,
			}
		)

		-- Set same value as though it came from server. Should not cause a conflict.

		:waitFor( function()		-- step 3
			
			return docStore:setServerValueP( 'sample', randomDocId, {
					TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Public/",
					TEMP_rootRevId = "facfea24c62d4fbfa6ea210a7a75eeb0",
					_ozCentralRevisionId = "facfea24c62d4fbfa6ea210a7a75eeb0",
					captureDate = "2009-05-30T09:02:06Z",
					height = 2449,
					keywords = {
						America = true,
						Hawaii = true,
						Maui = true,
						[ "North America" ] = true,
						Public = true,
						[ "R&R" ] = true,
						Summer = true,
						[ "U.S." ] = true,
						US = true,
						USA = true,
						[ "United States" ] = true,
						[ "United States of America" ] = true,
						architectural = true,
						architecture = true,
						building = true,
						[ "commercial building" ] = true,
						concept = true,
						concepts = true,
						conceptual = true,
						edifice = true,
						edifices = true,
						[ "getting away from it all" ] = true,
						paradise = true,
						[ "rest and relaxation" ] = true,
						seasons = true,
						shop = true,
						store = true,
						structures = true,
						summertime = true,
						tropical = true,
					},
					originalFileName = "es-1437-022.jpg",
					width = 3673,
				} )
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				TEMP_importedFromPath = "/Users/scouten/Pictures/Desktop Photos/Source/Public/",
				TEMP_rootRevId = "facfea24c62d4fbfa6ea210a7a75eeb0",
				_ozCentralRevisionId = "facfea24c62d4fbfa6ea210a7a75eeb0",
				_ozMetadataUploadStatus = "uploaded",
				captureDate = "2009-05-30T09:02:06Z",
				height = 2449,
				keywords = {
					America = true,
					Hawaii = true,
					Maui = true,
					[ "North America" ] = true,
					Public = true,
					[ "R&R" ] = true,
					Summer = true,
					[ "U.S." ] = true,
					US = true,
					USA = true,
					[ "United States" ] = true,
					[ "United States of America" ] = true,
					architectural = true,
					architecture = true,
					building = true,
					[ "commercial building" ] = true,
					concept = true,
					concepts = true,
					conceptual = true,
					edifice = true,
					edifices = true,
					[ "getting away from it all" ] = true,
					paradise = true,
					[ "rest and relaxation" ] = true,
					seasons = true,
					shop = true,
					store = true,
					structures = true,
					summertime = true,
					tropical = true,
				},
				originalFileName = "es-1437-022.jpg",
				width = 3673,
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_serverMutatedToOriginal( test )
	
	local myRandomId = KSIdGenerator.uuidInLowercaseHex()
	local count = 1
	local setServerValue = function()
			count = count + 1
			return docStore:setServerValueP( 'frabble', myRandomId, {
				captureDate = "0000-00-00T00:00:00Z",
				updated = string.format("2012-10-24T23:32:1%d.926084Z", count ),
				height = 2449,
				keywords = {
				},
				originalFileName = "mutatedBackToOriginal.jpg",
				width = 3673,
			} )
			
		end
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( setServerValue )
		
		:waitFor( function()		-- step 2
		
			return docStore:applyOverlayToDocumentP {
				docType = 'frabble',
				docId = myRandomId,
				overlay = {
					captureDate = "1997-05-01T12:00:00.0+1:00",
						-- this is an illegal date, timezone must be two digits for hour
					_localOnly = {
						_uploadedValues = {
							captureDate = "1997-05-01T12:00:00.0+1:00",
						},
					}
				},
			}
		
		end )
	
		:waitFor( setServerValue )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'frabble', myRandomId )
		
		end )

		:continueWhenValueMatches(
			'data',
			{
				_changeCounter = 3,
				_docId = myRandomId,
				_docType = "frabble",
				_localOnly = {
				},
				_ozMetadataUploadStatus = "uploaded",
				captureDate = "0000-00-00T00:00:00Z",
				height = 2449,
				keywords = {
				},
				originalFileName = "mutatedBackToOriginal.jpg",
				updated = "2012-10-24T23:32:13.926084Z",
				width = 3673,
			},
			{
				_localOnly = {
				},
				_ozMetadataUploadStatus = "uploaded",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValue_serverMutatedToNovel( test )
	
	local myRandomId = KSIdGenerator.uuidInLowercaseHex()
	local count = 1
	local setServerValue = function( captureDate )
		return function()
			count = count + 1
			return docStore:setServerValueP( 'novel', myRandomId, {
				captureDate = captureDate,
				updated = string.format("2012-10-24T23:32:1%d.926084Z", count ),
				height = 2449,
				keywords = {
				},
				originalFileName = "mutatedBackToOriginal.jpg",
				width = 3673,
			} )
			
		end
	end
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( setServerValue "1973-03-08T00:30:00.0-07:00" )
		
		:waitFor( function()		-- step 2
		
			return docStore:applyOverlayToDocumentP {
				docType = 'novel',
				docId = myRandomId,
				overlay = {
					captureDate = "1997-05-01T12:00:00.0+1:00",
						-- this is an illegal date, timezone must be two digits for hour
					_localOnly = {
						_uploadedValues = {
							captureDate = "1997-05-01T12:00:00.0+1:00",
						},
					}
				},
			}
		
		end )
	
		:waitFor( setServerValue "0000-00-00T00:00:00Z" )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'novel', myRandomId )
		
		end )

		:continueWhenValueMatches(
			'data',
			{
				_changeCounter = 3,
				_docId = myRandomId,
				_docType = "novel",
				_localOnly = {
				},
				_ozMetadataUploadStatus = "uploaded",
				captureDate = "0000-00-00T00:00:00Z",
				height = 2449,
				keywords = {
				},
				originalFileName = "mutatedBackToOriginal.jpg",
				updated = "2012-10-24T23:32:13.926084Z",
				width = 3673,
			},
			{
				_localOnly = {
				},
				_ozMetadataUploadStatus = "uploaded",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testSetServerValueAndApplyOverlay_replace( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 1,
							b = 2,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				develop = {
					a = 1,
					b = 2,
				},
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new develop settings as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							develop = {
								_replace = true,
								c = 3,
							},
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				develop = {
					c = 3,
				},
			}
		)

		-- Meanwhile, the develop settings changed on server, too.
		-- Since we replaced the entire develop settings table, the
		-- new server values should not appear.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 4,
							b = 5,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				develop = {
					c = 3,
				},
			}
		)

end

--------------------------------------------------------------------------------

-- @mark :resolveConflict

--------------------------------------------------------------------------------

function OzDocStoreTest.testResolveConflict_local( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new title as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = "Local New Title",
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				title = "Local New Title",
			},
			{
				title = "Local New Title",
			}
		)

		-- Meanwhile, the title changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data', {
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					title = "Server's New Title",
				},
				favorite = true,
				title = "Local New Title",
			},
			{
				_conflict = {
					title = "Server's New Title",
				},
				title = "Local New Title",
			}
		)
				
		-- Now instruct Oz to override server changes with local.
		
		:waitFor( function()		-- step 7
		
			return docStore:resolveConflictP( 'sample', randomDocId, 3, 'local' )
		
		end )

		:waitFor( function()		-- step 8
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 4,
				-- _conflict is now removed, allowing this to be uploaded
				favorite = true,
				title = "Local New Title",
			},
			{
				_conflict = {
					_delete = true,
				},
				title = "Local New Title",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testResolveConflict_localDevelop( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 1,
							b = 2,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				develop = {
					a = 1,
					b = 2,
				},
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new develop settings as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							develop = {
								_replace = true,
								c = 3,
							},
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				develop = {
					c = 3,
				},
			}
		)

		-- Meanwhile, the develop settings changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 4,
							b = 5,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				develop = {
					c = 3,
				},
			}
		)
		
		-- Now instruct Oz to override server changes with local.
		
		:waitFor( function()		-- step 7
		
			return docStore:resolveConflictP( 'sample', randomDocId, 3, 'local' )
		
		end )

		:waitFor( function()		-- step 8
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 4,
				-- _conflict is now removed, allowing this to be uploaded
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				_conflict = {
					_delete = true,
				},
				develop = {
					c = 3,
				},
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testResolveConflict_server( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new title as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = "Local New Title",
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				title = "Local New Title",
			},
			{
				title = "Local New Title",
			}
		)

		-- Meanwhile, the title changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					title = "Server's New Title",
				},
				favorite = true,
				title = "Local New Title",
			},
			{
				_conflict = {
					title = "Server's New Title",
				},
				title = "Local New Title",
			}
		)
				
		-- Now instruct Oz to override local changes with server.
		
		:waitFor( function()		-- step 7
		
			return docStore:resolveConflictP( 'sample', randomDocId, 3, 'server' )
		
		end )

		:waitFor( function()		-- step 8
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 4,
				_ozMetadataUploadStatus = 'uploaded',
				favorite = true,
				title = "Server's New Title",
			},
			{
				_ozMetadataUploadStatus = 'uploaded',
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testResolveConflict_serverDevelop( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( function()		-- step 1
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 1,
							b = 2,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				develop = {
					a = 1,
					b = 2,
				},
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

		-- Set new develop settings as though it came from local user.
		-- That should override server value.

		:waitFor( function()		-- step 3
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							develop = {
								_replace = true,
								c = 3,
							},
						},
					}
			
		end )
		
		:waitFor( function()		-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				develop = {
					c = 3,
				},
			}
		)

		-- Meanwhile, the develop settings changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()		-- step 5
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						develop = {
							a = 4,
							b = 5,
						},
					} )
			
		end )
		
		:waitFor( function()		-- step 6
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				_ozMetadataUploadStatus = "conflict",
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				favorite = true,
				develop = {
					_replace = true,
					c = 3,
				},
			},
			{
				_conflict = {
					develop = {
						a = 4,
						b = 5,
					},
				},
				develop = {
					c = 3,
				},
			}
		)
		
		-- Now instruct Oz to override server changes with local.
		
		:waitFor( function()		-- step 7
		
			return docStore:resolveConflictP( 'sample', randomDocId, 3, 'server' )
		
		end )

		:waitFor( function()		-- step 8
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 4,
				_ozMetadataUploadStatus = 'uploaded',
				favorite = true,
				develop = {
					a = 4,
					b = 5,
				},
			},
			{
				_ozMetadataUploadStatus = "uploaded",
			}
		)

end

--------------------------------------------------------------------------------

-- also test server deletes an existing key and local changes it

--------------------------------------------------------------------------------

-- @mark :docValueOnceP

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueOnceP_empty( test )

	local resultFromFirstDocValueCall

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()	-- step 1
		
			return docStore:docValueOnceP( 'sample', randomDocId )
				:mapDataP( function( _, doc )
				
					resultFromFirstDocValueCall = doc
						-- save this so we can hack on it in a moment
				
				end )
			
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
			},
			{}
		)
		
		:waitFor( .1 )			-- step 2
		
		:waitFor( function()	-- step 3
		
			-- Make some changes to doc. Make sure doc store doesn't
			-- pick them up for next caller.
			
			resultFromFirstDocValueCall.favorite = true
			resultFromFirstDocValueCall.title = "My Nifty Photo"
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
			},
			{}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueOnceP_separateValues( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.
	
		:waitFor( function()
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId,
											{ separateValues = true } )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				favorite = true,
				title = "My Nifty Photo",
			}, {
				_ozMetadataUploadStatus = "uploaded",
			} )

		-- Set new title as though it came from local user.
		-- That should override server value.

		:waitFor( function()
			
			return docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docId = randomDocId,
						overlay = {
							title = "Local New Title",
						},
					}
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId,
											{ separateValues = true } )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				title = "Local New Title",
			} )

		-- Meanwhile, the title changed on server, too.
		-- That should show up as a conflict.

		:waitFor( function()
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "Server's New Title",
					} )
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId,
											{ separateValues = true } )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 3,
				favorite = true,
				title = "Server's New Title",
			},
			{
				_conflict = {
					title = "Server's New Title",
				},
				title = "Local New Title",
			} )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueOnceP_nopChangeDoesntBumpCounter( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Set initial value as though it came from server.

		:waitFor( function()
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId,
											{ separateValues = true } )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				favorite = true,
				title = "My Nifty Photo",
			}, {
				_ozMetadataUploadStatus = "uploaded",
			} )

		-- Set same value from server again. This should not
		-- bump the change counter (i.e. it should be effectively
		-- ignored).

		:waitFor( function()
			
			return docStore:setServerValueP( 'sample', randomDocId, {
						favorite = true,
						title = "My Nifty Photo",
					} )
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId,
											{ separateValues = true } )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				favorite = true,
				title = "My Nifty Photo",
			}, {
				_ozMetadataUploadStatus = "uploaded",
			} )

end

--------------------------------------------------------------------------------

-- @mark :applyOverlayToDocumentP

--------------------------------------------------------------------------------

function OzDocStoreTest.testApplyOverlayToDocumentP( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()	-- step 1
			
			return docStore:applyOverlayToDocumentP {
								docType = 'sample',
								docId = randomDocId,
								overlay = {
									favorite = true,
									title = "My Nifty Photo",
								},
							}
			
		end )
		
		:waitFor( function()	-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			 {
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				favorite = true,
				title = "My Nifty Photo",
			}, {
				favorite = true,
				title = "My Nifty Photo",
			}
		)
	
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueAfterTwoOverlays( test )

	local nDocIds = 5

	local docIds = {}
	for i = 1, nDocIds do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local victimDocId = docIds[ 4 ] -- just choose one at random
	local nonVictimDocId = docIds[ 5 ] -- any other doc

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					},
				}
			
		end )
		
		:waitFor( function()

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = victimDocId,
					overlay = {
						bar = 94,
					},
				}
		
		end )
		
		-- Spot check: Make sure we didn't affect any other documents.
		-- (Been there, done that!)
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', nonVictimDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = nonVictimDocId,
				_changeCounter = 5,
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				favorite = true,
				title = "My Nifty Photo",
			}
		)

		-- And, of course, make sure we did affect the desired doc.
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', victimDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = victimDocId,
				_changeCounter = nDocIds + 1,
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueAfterReplace( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
			
			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					},
				}
			
		end )
		
		:waitFor( function()

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = randomDocId,
					overlay = {
						_replace = true,
						bar = 94,
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				_replace = true,
				bar = 94,
			},
			{
				--_replace = true, -- FIXME [scouten 2011-08-22]: Shouldn't this be here?
				bar = 94,
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocTypeInvariant_read( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'samplex', randomDocId )
		
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )
	
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocTypeInvariant_write( test )

	-- GRUNGE [scouten 2011-03-15]: There might be a race condition where two
	-- transactions attempt to write to the same new document ID with different
	-- docTypes at the same time. I'm not sure what failure modes might exist
	-- in this case. OTOH, that seems like an unlikely programmer error so it
	-- may not be worth the time to make that iron tight.

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'samplex',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
		
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )
			
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocTypeNotAllowedWhenWriting( test )

	KSTestUtils.testFails( function()
	
		docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					_docType = 'foo',
					favorite = true,
					title = "My Nifty Photo"
				},
			}
	
	end )
	
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDeletingDocLeavesLocalState( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()	-- step 1
			
			return docStore:applyOverlayToDocumentP {
								docType = 'sample',
								docId = randomDocId,
								overlay = {
									_localOnly = {
										foo = 'bar',
									},
									favorite = true,
									title = "My Nifty Photo",
								},
							}
			
		end )
		
		:waitFor( function()	-- step 2
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				_localOnly = {
					foo = 'bar',
				},
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				_localOnly = {
					foo = 'bar',
				},
				favorite = true,
				title = "My Nifty Photo",
			}
		)
	
		:waitFor( function()	-- step 3
			
			return docStore:applyOverlayToDocumentP {
								docType = 'sample',
								docId = randomDocId,
								overlay = {
									_deleted = true,
								},
							}
			
		end )
		
		:waitFor( function()	-- step 4
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				_localOnly = {
					foo = 'bar',
				},
				_deleted = true,
			},
			{
				_localOnly = {
					foo = 'bar',
				},
				_deleted = true,
			}
		)
	
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testEmbeddedNullSafety( test )

	-- BLONDE encoding uses a zero byte to encode the number 0. This test
	-- validates that SQLite bridge code doesn't mishandle the Lua string data
	-- containing embedded nulls. (It did at one point ...)

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						rating = 0,
						title = "My Nifty Photo",
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				title = "My Nifty Photo",
				rating = 0,
			}, {
				rating = 0,
				title = "My Nifty Photo",
			} )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDeleteValueFromTable( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					}
				}
		
		end )
		
		:waitFor( function()
	
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = { _delete = true },
						rating = 0,
					}
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				title = "My Nifty Photo",
				rating = 0,
			},
			{
				favorite = {
					_delete = true,
				},
				rating = 0,
				title = "My Nifty Photo",
			} )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testOverlaySubtable( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						stuff = {
							foo = 42,
						},
						title = "My Nifty Photo",
					},
				}
		end )
			
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						stuff = { bar = 94 },
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				stuff = {
					foo = 42,
					bar = 94,
				},
				title = "My Nifty Photo",
			}, {
				stuff = {
					bar = 94,
					foo = 42,
				},
				title = "My Nifty Photo",
			} )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testReplaceSubtable( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()		-- step 1

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						stuff = {
							foo = 42,
						},
						title = "My Nifty Photo",
					}
				}
			
		end )
		
		:waitFor( function()		-- step 2
			
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_changeCounter = 1,
				_docType = 'sample',
				_docId = randomDocId,
				stuff = {
					foo = 42,
				},
				title = "My Nifty Photo",
			},
			{
				stuff = {
					foo = 42,
				},
				title = "My Nifty Photo",
			}
		)
		
		:waitFor( function()		-- step 3

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						stuff = { _replace = true, bar = 94 },
					},
				}
			
		end )
		
		:waitFor( function()		-- step 4
			
			return docStore:docValueOnceP( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_changeCounter = 2,
				_docType = 'sample',
				_docId = randomDocId,
				stuff = {
					_replace = true,
					bar = 94,
				},
				title = "My Nifty Photo",
			},
			{
				stuff = {
					bar = 94,
				},
				title = "My Nifty Photo",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testNopChangeDoesntCreateNewRev( test )

	-- unimplemented

	-- Call applyOverlay a second time with all values same as existing.
	-- Make sure we don't create a new revision.

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testBulkInsert( test )

	local docIds = {}
	for i = 1, 1000 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local victimDocId = docIds[ 4 ] -- just choose one at random

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					}
				}
			
		end )
		
		:waitFor( function()

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = victimDocId,
					overlay = {
						bar = 94,
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', victimDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = victimDocId,
				_changeCounter = #docIds + 1,
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testBulkInsertAndUpdate( test )

	local docIds = {}
	
	for i = 1, 1000 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local victimDocId = docIds[ 4 ] -- just choose one at random

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docIds = docIds,
					overlay = {
						bar = 94,
					},
				}
			
		end )

		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', victimDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = victimDocId,
				_changeCounter = 1004,
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			},
			{
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			} )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocChangeCounterB( test )

	local docChangeCounterB = docStore:docChangeCounterB()

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 1 )
	
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}

		end )

		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 2 )

		:waitFor( function()

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						rating = 0,
					},
				}
			
		end )

		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 3 )
	
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testChangedDocsP( test )

	WFAsyncTestCase( test )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = { randomDocId, '0f5bb24855b24ed0ade8ad3006fc47ec' },
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:changedDocsP { fromChangeCounter = 1 }
		
		end )
		
		:continueWhenValueMatches(
			'data',
			{
				{
					_docType = 'sample',
					_docId = randomDocId,
					_changeCounter = 1,
					favorite = true,
					title = "My Nifty Photo",
				},
				{
					_docType = 'sample',
					_docId = '0f5bb24855b24ed0ade8ad3006fc47ec',
					_changeCounter = 2,
					favorite = true,
					title = "My Nifty Photo",
				},
				
				continueFrom = 3,
			
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testChangedDocsPObservesLimit( test )

	local docIds = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:changedDocsP { fromChangeCounter = 1 }
		
		end )
		
		:whenResolvedDo( function( status, changedDocs )
		
			if status ~= 'data' then
				error "Promise resolved with non-data status"
			elseif #changedDocs ~= 50 then
				error "did not return 50 changed documents"
			elseif changedDocs.continueFrom ~= 51 then
				error "result did not include expected continueFrom value"
			end
		
		end )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testChangedDocsPObservesInclusionList( test )

	local docIds = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	WFAsyncTestCase( test )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
		
			return docStore:changedDocsP {
					fromChangeCounter = 1,
					docs = { docIds[ 4 ], docIds[ 17 ] },
				}
		
		end )
		
		:continueWhenValueMatches(
			'data',
			{
			
				{
					_docType = 'sample',
					_docId = docIds[ 4 ],
					_changeCounter = 4,
					favorite = true,
					title = "My Nifty Photo",
				},
				{
					_docType = 'sample',
					_docId = docIds[ 17 ],
					_changeCounter = 17,
					favorite = true,
					title = "My Nifty Photo",
				},
				
				continueFrom = 101,
			
			}
		)

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueIdsOnly( test )

	-- FIXME [scouten 2011-03-29]: Write this test. New option on changedDocsP.

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueB( test )

	local expectedDocValue = {
		_docType = 'sample',
		_docId = randomDocId,
	}
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:docValueB( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches( expectedDocValue )
		
		:waitFor( function()
		
			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueB( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 1,
				favorite = true,
				title = "My Nifty Photo",
			}, true )
		
		:waitFor( function()

			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						rating = 0,
					},
				}

		end )
		
		:waitFor( function()
		
			return docStore:docValueB( 'sample', randomDocId )
		
		end )
		:continueWhenValueMatches(
			{
				_docType = 'sample',
				_docId = randomDocId,
				_changeCounter = 2,
				favorite = true,
				title = "My Nifty Photo",
				rating = 0,
			}, ':ignore:' ) -- GC means this result might be a first time
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testDocValueBehaviorsAreWeaklyHeld( test )

	local WFCloseSupport = require 'WFCloseSupport'

	local docValueBehaviors = KSTableUtils.makeFullyWeak()
	local docValueB = docStore:docValueB( 'sample', randomDocId )
	
	docValueBehaviors[ 1 ] = docValueB

	local docChangeCounterB = docStore:docChangeCounterB()

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 1 )
	
		:waitFor( function()

			return docStore:applyOverlayToDocumentP
				{
					docType = 'sample',
					docId = randomDocId,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()
			
			assert( docValueBehaviors[ 1 ] == docValueB,
					"docValueB disappeared already ?!?" )
			
			assert( not WFCloseSupport.hasBeenOrWillBeClosed( docValueB ),
					"docValueB has already been closed ?!?" )

			return true
			
		end )
		
		:waitFor( function()

			docValueB = nil
			collectgarbage()
			collectgarbage()
			collectgarbage()
			
			return true
		
		end )
		
		:waitFor( function()
		
			return tostring( docValueBehaviors[ 1 ] )
		
		end )
		:continueWhenValueMatches( 'data', 'nil' )

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testFieldValueB( test )

	local fieldValueB = docStore:fieldValueB( 'sample', randomDocId, 'rating' )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( fieldValueB )		-- step 1
		:continueWhenValueMatches( nil )
		
		:waitFor( function()		-- step 2
		
			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					rating = 2,
					title = "My Nifty Photo",
				},
			}

		end )
	
		:waitFor( fieldValueB )		-- step 3
		:continueWhenValueMatches( 2, true )
		
		:waitFor( function()		-- step 4
		
			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					rating = 0,
				},
			}

		end )
	
		:waitFor( fieldValueB )		-- step 5
		:continueWhenValueMatches( 0, false )
		
end

--------------------------------------------------------------------------------

function OzDocStoreTest.testFieldValueBSilentWhenDifferentFieldChanges( test )

	-- Make sure fieldValueB doesn't send out a spurious value change
	-- when some other value in the document changes.
	
	-- This gets interesting when you consider the second value ("is existing value")
	-- in the tuple.

end

--------------------------------------------------------------------------------

function OzDocStoreTest.testNestedFieldValueB( test )

	local fieldValueB = docStore:fieldValueB( 'sample', randomDocId, 'foo/1/rating' )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( fieldValueB )		-- step 1
		:continueWhenValueMatches( nil )
		
		:waitFor( function()		-- step 2
		
			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					foo = {
						{
							rating = 2
						}
					},
					title = "My Nifty Photo",
				},
			}
		
		end )
		
		:waitFor( fieldValueB )		-- step 3
		:continueWhenValueMatches( 2, true )
		
		:waitFor( function()		-- step 4
		
			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					foo = { _delete = true },
				},
			}
			
		end )
		
		:waitFor( fieldValueB )		-- step 5
		:continueWhenValueMatches( nil, false )
		
		:waitFor( function()		-- step 6

			return docStore:applyOverlayToDocumentP {
				docType = 'sample',
				docId = randomDocId,
				overlay = {
					foo = {
						{
							rating = 4
						}
					}
				},
			}
		
		end )

		:waitFor( fieldValueB )		-- step 7
		:continueWhenValueMatches( 4, false )
		
end

--------------------------------------------------------------------------------

return OzDocStoreTest
