--[[----------------------------------------------------------------------------

OzOriginalsAgent.lua

@uathor ngautam 2011-08-25

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the assets in a single catalog and scheduled downloads of originals
for the assets that are missing them.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafouncation.core.lua.threading
local WFTimer = require 'WFTimer'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- wichitafoundation.core.reactive
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

	-- oz.account
local OzUser = require 'OzUser'

	-- oz.catalog
local OzCatalogService = require 'OzCatalogService'
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'

	-- oz.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- wichitafoundation.oz_client.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'


--------------------------------------------------------------------------------

local downloadOriginalsLog = KSLogging.log 'OzOriginalsAgent'
--	:setLevel( KSLogging.INFO )


--============================================================================--

-- @mark OzOriginalsAgent

local originalsAgentMethods = {

	_catalog = false,
		-- (OzCatalogService) catalog

	_catalogId = false,
		-- (string) catalog ID as known to the Oz system (UUID)
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)
	
	_indexStoreQueueIsLargeB = false,
		-- (B< boolean >) true if we have an index store backlog
	
	_checkAndDownloadP = false,

	_dataCache = false,
		-- (KSDataCache) data cache for this doc store

	_dataCachePathP = false,
		-- (P< string >) path to cache

	_photoToDownloadB = false,
		-- (B< table of string >) which photo should we be downloading now?

	_recentErrors = false,
		-- (table) array of info 10? most recent sync errors

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function originalsAgentMethods:__init( catalogId, userId )

	downloadOriginalsLog:info( "Starting originals agent for catalog %s", catalogId )

	self._catalogId = assert( catalogId )
	
	self._catalog =  OzCatalogService( self._catalogId, OzUser.make( userId ) )
	self:addCloseTarget( self._catalog )

	self._docStore = OzDocStore( catalogId )
	self:addCloseTarget( self._docStore )

	self._dataCache = assert( self._docStore:dataCache() )

	self._dataCachePathP = self._dataCache:pathP()
	self._dataCachePathP:runUntilDone()
	
	self._indexStore = WFIndexStore( self._docStore, 'OzCatalogIndexDefinitions.lua' )
	self:addCloseTarget( self._indexStore )
	
	self._indexStoreQueueIsLargeB = self._indexStore:sizeOfIndexQueueB()
		:mapB( function( indexQueueSize )
			return ( not indexQueueSize ) or ( indexQueueSize > 50 )
		end )
		:mapSwitchB( function( isLarge )
			if isLarge then
				return KSReactive.constantB( true )
			else
				return KSReactive.trueThenFalseB( 10 )
			end
		end )
	self:addCloseTarget( self._indexStoreQueueIsLargeB )

	self._photoToDownloadB = self._indexStoreQueueIsLargeB
		:mapSwitchB( function( isLarge )
			
			if isLarge then
				return KSReactive.constantB( {} )
			else
				return self._indexStore:queryB {
					sql = [[SELECT photoId FROM missingOriginals
								ORDER BY missingInOz, captureDate, docId LIMIT 1]],
					calm = 2,
				}
			end
		
		end )

	self:addCloseTarget( self._photoToDownloadB )

	self._photoToDownloadB:onDataB( function( photoId )
		downloadOriginalsLog:info( "photoToDownloadB updated to %s", photoId )
	end )
	
	self._photoToDownloadB:addCloseAction( function()
		downloadOriginalsLog:info( "photoToDownloadB closed" )
	end )
	
	WFCloseSupport.mustClose( self._photoToDownloadB )

	self._recentErrors = {}

	-- Kick off a new download.
	
	-- FIXME [scouten 2012-03-08]: This might better be expressed using Mark's proposed
	-- promise state machine system. Consider rewriting when/if that becomes available.

	local function startNewDownload( photoId )
	
		if not photoId then return end

		if self._checkAndDownloadP then
			downloadOriginalsLog:info( "skipping startNewDownload because there's a download in progress" )
			return
		end
		
		self._checkAndDownloadP = self:downloadOriginalP( photoId )
		
		self._checkAndDownloadP:whenResolvedDo( function( status, message )
		
			self._checkAndDownloadP = false
			
			if status == 'data' then

				downloadOriginalsLog:info( "downloadOriginalP finished with message %s",
										   message )

				-- Fix a possible race condition [scouten 2012-09-21, PR# 3331492]:
				-- Possible self._photoToDownloadB will have changed before the promise
				-- resolves. If so, we wouldn't have a way to kick off the next download.
				
				local newPhotoIdData = self._photoToDownloadB:valueNow()
				local newPhotoId = newPhotoIdData and newPhotoIdData[ 1 ] and newPhotoIdData[ 1 ][ 1 ]
				
				if newPhotoId and newPhotoId ~= photoId then
					startNewDownload( newPhotoId )
				end
				
			else

				downloadOriginalsLog:error( "downloadOriginalP failed with status %s and message %s",
											status, message )

				local errorCapture = {
					error = message,
					time = KSDate.timeInIso8601Format(),
					photoId = photoId,
				}
				
				-- Fill in the doc value when we can. We might miss one once in a while, but that's probably OK.
				
				self._docStore:docValueOnceP( 'asset', photoId )
					:whenResolvedDo( function( status, docValue )
						errorCapture.docValue = docValue
					end )
					:runUntilDone()

				self._recentErrors[ #self._recentErrors + 1 ] = errorCapture
				if #self._recentErrors > 10 then
					table.remove( self._recentErrors, 1 )
				end
	
				-- If we fail (for instance, due to network failure), kick the
				-- process again after a little while.
				
				WFTimer.afterDelayDo( 30, function()
				
					downloadOriginalsLog:info( "downloadOriginalP retrying download of photo ID %s after delay",
											   photoId )

					startNewDownload( photoId )
				
				end )
				
			end
			
		end ):attachTo( self )
		
	end

	-- Start checking for missing originals and downloading. Wait a little while after startup, though.

	WFTimer.afterDelayDo( 12, function()

		if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
			return
		end
		
		downloadOriginalsLog:debug( "attaching action to query behavior" )
		
		self._photoToDownloadB:onDataB( function( data )
		
			local photoId = data and data[ 1 ] and data[ 1 ][ 1 ]

			downloadOriginalsLog:debug( "queryB result was %s", data )

			startNewDownload( photoId )

		end )
	
	end )

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

local function moveOriginalToDocuments( originalFilePath )

	local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
	local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'
	local oldCachesDir = KSPathUtils.getStandardFilePath 'caches'
	
	local originalFilePath_dirSub = originalFilePath:sub( 1, #cachesDir )
	local originalFilePath_dirSubOld = originalFilePath:sub( 1, #oldCachesDir )
	
	local newFilePath = originalFilePath
			
	if originalFilePath_dirSub == cachesDir then
		newFilePath = string.format( "%s%s", documentsDir, originalFilePath:sub( #cachesDir + 1 ) )
	elseif originalFilePath_dirSubOld == oldCachesDir then
		newFilePath = string.format( "%s%s", documentsDir, originalFilePath:sub( #oldCachesDir + 1 ) )
	end

	KSFileUtils.mkdir( KSPathUtils.parent( newFilePath ) )
	
	newFilePath = KSFileUtils.moveToUniquelyNamedFile( originalFilePath, newFilePath )

	return newFilePath
	
end

--------------------------------------------------------------------------------

function originalsAgentMethods:downloadOriginalP( photoId )

	downloadOriginalsLog:debug( "downloadOriginalP %s", photoId )

	local spaceNeededP = self._dataCachePathP:mapDataP( function( _, path )
	
		local fileSystemFreeSpace, _ -- fileSystemSize
			= KSFileUtils.checkFreeSpaceAtPath( path )

		local amountUnderMinimumFreeSpace = WFOzPurgeSettings.minimumFreeSpace() - fileSystemFreeSpace

		local spaceNeeded = amountUnderMinimumFreeSpace
		
		return 'data', spaceNeeded
	
	end )
	
	return spaceNeededP:mapDataP( function( _, spaceNeeded )
	
		downloadOriginalsLog:debug( "downloadOriginalP %s spaceNeeded result was %s", photoId, spaceNeeded )

		-- If we're out of space, do not download.

		if spaceNeeded > 0 then
			downloadOriginalsLog:warn( "Not downloading anything: Low on disk space!" )
			return 'error', 'OzOriginalsAgent: Out of space'
		end

		downloadOriginalsLog:info( "Will download original for photo ID %s", photoId )

		-- Check if we have the base original or latest original in the caches directory, if so
		-- move it, update the doc store and be done.
		
		return self._catalog:photoInfoP( photoId )
			:mapDataP( function( status, docValue )

				downloadOriginalsLog:debug( "Doc value is %s", docValue )

				local isInCaches = docValue
							   and docValue._localOnly
							   and docValue._localOnly.original
							   and docValue._localOnly.original.isInCacheFolder
					
				local isInCaches_latestOriginal = docValue
											  and docValue._localOnly
											  and docValue._localOnly.latestOriginal
											  and docValue._localOnly.latestOriginal.isInCacheFolder
				
				local originalFilePath = docValue
									 and docValue._localOnly
									 and docValue._localOnly.originalFilePath
				
				local resultP
													
				if isInCaches then
				
					-- Temporary check to correct the local file path and replace
					-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'

					-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
					-- that updates the database with the correct file paths but this is not
					-- feasible in the current time frame.

					originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
				
					local originalStatusOverlay = {
						_localOnly = {
							original = {},
						}
					}
				
					if KSFileUtils.exists( originalFilePath ) then
					
						downloadOriginalsLog:debug( "Moving base original for asset %s to Documents", photoId )
						
						-- Move this asset to Documents, write the new value for the path and
						-- mark isInCacheFolder = false.
						
						local newFilePath = moveOriginalToDocuments( originalFilePath )
					
						-- Update the file path and the location of the file.

						originalStatusOverlay._localOnly.originalFilePath = newFilePath
						originalStatusOverlay._localOnly.original.isInCacheFolder = false
						
					else
					
						-- The file is missing, delete the section for the original and originalFilePath, so
						-- that the next time, we'll know we need to download this.

						originalStatusOverlay._localOnly.originalFilePath = { _delete = true }
						originalStatusOverlay._localOnly.original = { _delete = true }
						
					end
					
					resultP = self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = photoId,
						overlay = originalStatusOverlay,
					}
					
				else
				
					-- If not, trigger the download of the base original.

					downloadOriginalsLog:debug( "Downloading base original for asset %s", photoId )
					
					resultP = self._catalog:baseOriginalDataPathP( photoId )
						
				end
				
				local latestOriginalFilePath = docValue
										   and docValue._localOnly
										   and docValue._localOnly.latestOriginal
										   and docValue._localOnly.latestOriginal.filePath
										
				latestOriginalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( latestOriginalFilePath )

				if isInCaches_latestOriginal and OzCatalogUploadStatus.assetLatestOriginalIsValid( docValue ) then

					local latestOriginalStatusOverlay = {
						_localOnly = {
							latestOriginal = {},
						}
					}
				
					if KSFileUtils.exists( latestOriginalFilePath ) then
					
						downloadOriginalsLog:debug( "Moving latest original for asset %s to Documents", photoId )

						-- Move this asset to Documents, write the new value for the path and
						-- mark isInCacheFolder = false.
			
						local newFilePath = moveOriginalToDocuments( latestOriginalFilePath )

						-- Update the file path and the location of the file.
					
						latestOriginalStatusOverlay._localOnly.latestOriginal.filePath = newFilePath
						latestOriginalStatusOverlay._localOnly.latestOriginal.isInCacheFolder = false
					
					else
					
						-- The file is missing, delete the section for latestOriginal, so
						-- that the next time, we'll know we need to download this.

						latestOriginalStatusOverlay._localOnly.latestOriginal = { _delete = true }
						
					end
							
					return resultP:mapDataP( function( status, data )
						
						downloadOriginalsLog:debug( "Applying overlay to photo %s = %s",
												    photoId,
												    latestOriginalStatusOverlay )
						
						return self._docStore:applyOverlayToDocumentP {
							docType = 'asset',
							docId = photoId,
							overlay = latestOriginalStatusOverlay,
						}
							
					end )
					
				elseif not OzCatalogUploadStatus.assetOriginalIsValid( docValue ) then
				
					-- If we are tracking a latest version of the original, trigger the
					-- download of the original.

					downloadOriginalsLog:debug( "Downloading latest original for asset %s", photoId )

					return resultP:mapDataP( function( _, data )

						return self._catalog:originalDataPathP( photoId )

					end )

				end
				
				return resultP
				
			end )
		
	end )

end

--------------------------------------------------------------------------------

function originalsAgentMethods:sendDiagnosticLogData()

	WFDiagnosticLog.addSectionToLog {

		name = "Download Originals Agent",
		id = 'OzOriginalsAgent-' .. self._catalogId,
		weight = 650,

		body = {
		
			catalogId = self._catalogId,

			dataCachePath = self._dataCachePathP:runUntilDoneP(),
			photoToDownload = self._photoToDownloadB,

			recentErrors = self._recentErrors,
			
			sizeOfDownloadQueue = self._indexStore:queryP [[SELECT count( photoId ) FROM missingOriginals]]:runUntilDoneP(),

		}
	
	}
			
end

--------------------------------------------------------------------------------

return KSClass 'OzOriginalsAgent'( originalsAgentMethods )
