--[[----------------------------------------------------------------------------

@file OzOriginalsRecovery.lua

@author arahn
@date 10/25/2012

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSDeviceUtils = require "KSDeviceUtils"
local OzFilePaths = require "OzFilePaths"
local WFPlatformInfo = require "WFPlatformInfo"
local KSPathUtils = require "KSPathUtils"
local KSData = require "KSData"
local KSFileUtils = require "KSFileUtils"
local KSPromises = require "KSPromises"
local KSDate = require "KSDate"
local WFTimer = require "WFTimer"

local WFCloseSupport = require "WFCloseSupport"

local KSDeviceUtils_deviceName = KSDeviceUtils.deviceName
local KSDate_formatDateTimeAsLocal = KSDate.formatDateTimeAsLocal

--------------------------------------------------------------------------------

local function recoverySuccessfulP( docStore, docValue, originalPath )

	local docStatusOverlay = {
		_localOnly = {
			originalFilePath = originalPath,
			recoveryEffort = 'mission accomplished',
		},
	}

	local resultP = docStore:applyOverlayToDocumentP {
		docType = docValue._docType,
		docId = docValue._docId,
		overlay = docStatusOverlay,
	}
	resultP:runUntilDone()
				
	return resultP
end

--------------------------------------------------------------------------------

local function recoveryFailedP( docStore, docValue )
	-- Alas, it is not found
	local docStatusOverlay = {
		_localOnly = {
			recoveryEffort = 'situation hopeless',
		},
		importSource = {
			originalStatus = 'lost',
		}
	}

	local resultP = docStore:applyOverlayToDocumentP {
		docType = docValue._docType,
		docId = docValue._docId,
		overlay = docStatusOverlay,
	}
	resultP:runUntilDone()
	
	return resultP:mapDataP( function()
		return 'data', true
	end )

end

--------------------------------------------------------------------------------

-- cache our mapping between assetUrl and creationDate, so long as we have
-- any recovery operations in flight
local numInFlight = 0

local _assetDateMapP

--------------------------------------------------------------------------------

local function getAssetLookupMapsP()
	
	if _assetDateMapP and not WFCloseSupport.hasBeenOrWillBeClosed( _assetDateMapP ) then
		return _assetDateMapP
	end

	local KSAssetsLibraryImportService = require "KSAssetsLibraryImportService"

	_assetDateMapP = KSAssetsLibraryImportService.getAssetsListP()
	
	:mapDataP( function( status, assetList )
		local dateMap = {}
		local filenameMap = {}
		local assetLibraryIdMap = {}
		for _, assetInfo in ipairs( assetList ) do
			local creationDate = assetInfo.creationDate
			local urlArray

			local dateStr = KSDate_formatDateTimeAsLocal( creationDate )
			urlArray = dateMap[ dateStr ]
			if not urlArray then
				urlArray = {}
				dateMap[ dateStr ] = urlArray
			end
			urlArray[ #urlArray + 1 ] = assetInfo
			
			local filename = assetInfo.filename
			urlArray = filenameMap[ filename ]
			if not urlArray then
				urlArray = {}
				filenameMap[ filename ] = urlArray
			end
			urlArray[ #urlArray + 1 ] = assetInfo

			local assetLibraryId = assetInfo.url
			local _
			_, _, assetLibraryId = string.find( assetLibraryId, "id=([-0-9A-F]+)" )
			if assetLibraryId then
				urlArray = assetLibraryIdMap[ assetLibraryId ]
				if not urlArray then
					urlArray = {}
					assetLibraryIdMap[ assetLibraryId ] = urlArray
				end
				urlArray[ #urlArray + 1 ] = assetInfo
			end

		end
		
		return 'data', dateMap, filenameMap, assetLibraryIdMap
	end )
	
	return _assetDateMapP
		
end

--------------------------------------------------------------------------------

local function clearCaches()

	_assetDateMapP = nil
	
end

--------------------------------------------------------------------------------


local function getAssetIdsForFilenameAndDateP( filename, dateStr )
	
	return getAssetLookupMapsP()
	
	:mapDataP( function( status, dateMap, filenameMap, assetLibraryIdMap )
		
		local maybeAssetId = filename and string.match( filename, "^([^.]+)%." )
			-- I've seen revel looking for a file with no master where the filename was the assetId
		local combined = {}
		local usedUrls = {}
		for _, t in pairs{ dateMap[ dateStr ] or {}, filenameMap[ filename ] or {}, assetLibraryIdMap[ maybeAssetId or filename ] or {} } do
			for _, assetInfo in ipairs( t ) do
				if not usedUrls[ assetInfo.url ] then
					combined[ #combined + 1 ] = assetInfo
					usedUrls[ assetInfo.url ] = 1
				end
			end
		end
		
		return 'data', combined
				
	end )
	
end

--------------------------------------------------------------------------------

local function attemptRecoveryByScanningAssetLibraryP( docStore, docValue, originalPathFn )
	numInFlight = numInFlight + 1
	local KSAssetsLibraryImportService = require "KSAssetsLibraryImportService"
	
	local wantCaptureDate = docValue.originalCaptureDate or docValue.captureDate
	local wantFilename = docValue.importSource and docValue.importSource.fileName
		-- Not all clients provided the fileName. :(
		
	local newOriginalPath, matchingAssetUrl

	return getAssetIdsForFilenameAndDateP( wantFilename, wantCaptureDate )
	
	:mapDataP( function( status, assetInfos )
		local wantSha256 = docValue.importSource.sha256
		
		return KSPromises.allP( function( assetInfo )
			local assetUrl = assetInfo.url
			-- for efficiency, skip last param until we find one with matching date
			return KSAssetsLibraryImportService.makeAssetDataRequestP( assetUrl )
				:mapDataP( function( status, imgData, creationDate, originalFilename )
					if imgData:digest( "SHA256" ) == wantSha256 then
						
						local originalPath = originalPathFn( originalFilename )
						KSFileUtils.mkdir( KSPathUtils.parent( originalPath ) )

						newOriginalPath = KSData.saveToUniquelyNamedFile( imgData, originalPath )
						matchingAssetUrl = assetUrl
					end
					imgData:close()
					return 'data', true
				end )
		end, assetInfos )
		
	end )
	:mapDataP( function( status )
		if newOriginalPath then
		
			return recoverySuccessfulP( docStore, docValue, newOriginalPath, matchingAssetUrl )
		else

			return recoveryFailedP( docStore, docValue )
		end
	end )
	:mapP( function( status )
		WFTimer.afterDelayDo( 5, function()
			numInFlight = numInFlight - 1
			if numInFlight == 0 then
				clearCaches()
			end
		end )
		
		if status ~= 'data' then

			return recoveryFailedP( docStore, docValue )
			
		end
		
	end )

end

--------------------------------------------------------------------------------

local function recoverFromAssetLibraryUrl( docStore, docValue, originalPathFn, whereWeFoundFile )

	local KSAssetsLibraryImportService = require "KSAssetsLibraryImportService"
	
	return KSAssetsLibraryImportService.makeAssetDataRequestP( whereWeFoundFile )

	:mapDataP( function( status, imgData, creationDate, originalFilename )
		
		local originalPath = originalPathFn( originalFilename )
		KSFileUtils.mkdir( KSPathUtils.parent( originalPath ) )

		originalPath = KSData.saveToUniquelyNamedFile( imgData, originalPath )
		
		return recoverySuccessfulP( docStore, docValue, originalPath )
	
	end )
	
	:mapP( function( status )
		
		if status ~= 'data' then
			return recoveryFailedP( docStore, docValue )
		end
		
	end )
	
end

--------------------------------------------------------------------------------

local function recoveryEffortOriginalP( docStore, docValue )

	-- see if we can find that sucker anywhere

	local importSource = docValue.importSource
	local catalogId = docStore:catalogId()
	
	local captureDate = docValue.originalCaptureDate or docValue.captureDate
	
	if not ( captureDate and importSource ) then
		return recoveryFailedP( docStore, docValue )
	end
	
	if catalogId and KSDeviceUtils_deviceName() == importSource.importedOnDevice then
		
		-- Maybe there is a chance the original file is still around
		-- As of 10/25/2012, We remember that original import path.  Try that first.
		local whereWeFoundFile = importSource.filePath
		
		if not whereWeFoundFile then
			-- If the user hasn't deleted the catalog they originally imported from, we might
			-- have this value tucked away in the localOnly section.
			local localUrl = docValue._localOnly
					     and docValue._localOnly.localUrl
			
			if localUrl then
				whereWeFoundFile = localUrl
			end
		
		end
		
		local originaFolderDocuments = OzFilePaths.getStandardOriginalFolderDocumentsPath( catalogId )
		local originalFolder = OzFilePaths.originalsFolderForDate( originaFolderDocuments, captureDate )
		local originalPathFn =
			function( newOriginalFilePathOrFileName )
				local fileName = newOriginalFilePathOrFileName and KSPathUtils.leafName( newOriginalFilePathOrFileName )
				fileName = fileName or importSource.fileName or "missing_file"
				local result = KSPathUtils.child( originalFolder, fileName )
				KSFileUtils.mkdir( originalFolder )
				return result
			end
		
		if not whereWeFoundFile then
			local folder = importSource.folder
			
			-- prior to 10/25/2012 we didn't store the actual asset import source in a form
			-- that's readily re-usable.  But we can reconstruct it..
			-- What we want is something like "assets-library://asset/asset.JPG?id=B13DE6E1...."
			-- Unfortunately that last part is not available.
			-- But we *do* have the sha256, and the fileName
			-- So we can scan the asset library and search by fileName, then get the blob and compare
			-- the sha256
			if folder and string.find( folder, "^assets%-library:" ) and WFPlatformInfo.isiOS then
				return attemptRecoveryByScanningAssetLibraryP( docStore, docValue, originalPathFn )
			end
		end

		if whereWeFoundFile and string.find( whereWeFoundFile, "^assets%-library:" ) and WFPlatformInfo.isiOS then

			return recoverFromAssetLibraryUrl( docStore, docValue, originalPathFn, whereWeFoundFile )

		end
		
		if whereWeFoundFile then
		
			if KSFileUtils.exists( whereWeFoundFile ) then
	
				local success, originalPath = pcall( KSFileUtils.copyToUniquelyNamedFile, whereWeFoundFile, originalPathFn( whereWeFoundFile ) )
				
				-- If not success : If we're on a mac, we could use powerbox to get permission to access
				-- this file.
				
				if success then
	
					return recoverySuccessfulP( docStore, docValue, originalPath )
	
				end
				
			end

		end
	end
	
	return recoveryFailedP( docStore, docValue )

end

--------------------------------------------------------------------------------

local function recoverLostOriginalP( docStore, docId, filePath )
	
	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docId = docId,
		overlay = {
			importSource = {
				filePath = filePath,
				importedOnDevice = KSDeviceUtils_deviceName(),
			},
			_localOnly = {
				recoveryEffort = "original",
			}
		},
	}
	:runUntilDoneP()
	
end

--==============================================================================

return {
	recoveryEffortOriginalP = recoveryEffortOriginalP,
	recoverLostOriginalP = recoverLostOriginalP,
}
