local WFAsyncTestCase = require "WFAsyncTestCase"
local KSPromises = require "KSPromises"
local KSData = require "KSData"
local KSDeviceUtils = require "KSDeviceUtils"

--==============================================================================

local fakeFileSystemItems = {}
local fakeAssets = {
	{
		url = "assets-library:asset.jpg?id=1",
		filename = "testRecoveryFromAssetLibraryByFilename.JPG",
		creationDate = 110000000,
	},
	{
		url = "assets-library:asset.jpg?id=2",
		filename = "testRecoveryFromAssetLibraryByDate.JPG",
		creationDate = 100000000,
	},
	{
		url = "assets-library:asset.jpg?id=3",
		filename = "testRecoveryFromAssetLibraryByLocalUrl.JPG",
		creationDate = 120000000,
	},
	{
		url = "assets-library:asset.jpg?id=58BEC4F1-FCB6-42E2-B8E2-6668DFCB0B76",
		filename = "testRecoveryFromAssetLibraryByIdInference.JPG",
		creationDate = 140000000,
	},
}
	


local function buildMocks()
	
	local getAssetsListP = function()
		return KSPromises.resolvedPromiseP( 'data', fakeAssets )
	end
			
	return {
		KSAssetsLibraryImportService = {
			getAssetsListP = getAssetsListP,
			
			makeAssetDataRequestP = function( assetId )
				
				local asset
				for k, v in pairs( fakeAssets ) do
					if v.url == assetId then
						asset = v
						break
					end
				end

				if asset then
				
					return KSPromises.resolvedPromiseP( 'data', KSData.make( assetId ), asset.creationDate, asset.filename )
				
				end
				
				return KSPromises.resolvedPromiseP( 'error', 'no such asset' )
				
			end,
		},
		
		KSDate = {
			formatDateTimeAsLocal = function( date )
				-- This doesn't even come close to formatting
				-- that value as an accurate date; but that doesn't
				-- matter, I just want to make sure we return a consistent
				-- value regardless of what timezone/daylight savings values
				-- are in effect on the machine running this test
				local fakeYear = math.floor( date / 60 / 60 / 24 / 365 )
				date = date - fakeYear * 60 * 60 * 24 * 365
				local fakeMonth = math.floor( date / 60 / 60 / 24 / 30 )
				date = date - fakeMonth * 60 * 60 * 24 * 30
				local fakeDay = math.floor( date / 60 / 60 / 24 )
				date = date - fakeDay * 60 * 60 * 24
				local fakeHour = math.floor( date / 60 / 60 )
				date = date - fakeHour * 60 * 60
				local fakeMin = math.floor( date / 60 )
				date = date - fakeMin * 60
				
				return string.format( "%04d-%02d-%02dT%02d:%02d:%02dZ",
					fakeYear + 2000, fakeMonth, fakeDay,
					fakeHour, fakeMin, date )
			end
		},
		
		KSData = {
			saveToUniquelyNamedFile = function( data, path )
				return path
			end,
		},
		
		KSFileUtils = {
			mkdir = function( path )
				-- we could probably track these calls and test that they were made correctly.
			end,
			exists = function( path )
				return fakeFileSystemItems[ path ]
			end,
			copyToUniquelyNamedFile = function( src, dest )
				return dest
			end,
		},
		
		WFPlatformInfo = {
			isiOS = true,
				-- allows us to unit test ios code even on mac
		},
		
		KSPathUtils = {
			-- These are mocked so we can have the same file path separators and
			-- standard file paths regardless of platform and user id
			child = function( a, b )
				return string.format( "%s/%s", a, b )
			end,
			getStandardFilePath = function( key )
				return string.format( "standard file path < %s >", key )
			end,
			parent = function( a )
				return string.gsub( a, "/[^/]+", "" )
			end,
			leafName = function( a )
				return string.gsub( a, ".+/", "" )
			end,
		},
	}
end

local OzOriginalsRecovery, KSDate, docStore


--==============================================================================

local TestSuite = {}

--==============================================================================
-- @mark setup
--==============================================================================

function TestSuite.setUpClass()

	local LOADED = debug.getregistry()[ "_LOADED" ]

	for k, namespace in pairs( buildMocks() ) do
		LOADED[ k ] = namespace
	end
	
	LOADED.OzOriginalsRecovery = nil
	OzOriginalsRecovery = require 'OzOriginalsRecovery'
	KSDate = require 'KSDate'
	
	docStore = {
		catalogId = function() return "catalog_id_14159265" end,
		applyOverlayToDocumentP = function( _, t )
			return KSPromises.resolvedPromiseP( 'data', 'overlay', t )
		end
	}

end

--------------------------------------------------------------------------------

function TestSuite.tearDownClass()

	local LOADED = debug.getregistry()[ "_LOADED" ]

	for k, namespace in pairs( buildMocks() ) do
		LOADED[ k ] = nil
	end
	
	LOADED.OzOriginalsRecovery = nil
	
end

--==============================================================================

function TestSuite.testZero( test )

-- Uncomment to increase coverage
--	WFAsyncTestCase( test, 10 )
--		:waitFor( require 'WFTimer'.afterDelayP( 6 ):runUntilDoneP() )

end

--------------------------------------------------------------------------------

function TestSuite.testRecoveryFromAssetLibraryByLocalUrl( test )
	
	local recoveryP = OzOriginalsRecovery.recoveryEffortOriginalP( docStore, {
		captureDate = "2003-09-23T21:20:00Z",
		importSource = {
			folder = "assets-library:",
			sha256 = "75cd0d84d626b4881c3191836fcd05aa21d2da9028e204fc899207bbb40d99eb",
			importedOnDevice = KSDeviceUtils.deviceName(),
		},
		_localOnly = {
			localUrl = "assets-library:asset.jpg?id=3",
		},
	} )
		:runUntilDoneP()
	
	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		:continueWhenValueMatches( 'data', 'overlay', {
			overlay = {
				_localOnly = {
					originalFilePath = "standard file path < carouselDocuments >/catalog_id_14159265/Originals/2003/2003-09/2003-09-23/testRecoveryFromAssetLibraryByLocalUrl.JPG",
					recoveryEffort = "mission accomplished",
				},
			},
		} )
end

--------------------------------------------------------------------------------

function TestSuite.testRecoveryFromFileSystemViaLocalUrl( test )
	
	local fakeOriginalPath = "/path/to/testRecoveryFromFileSystemViaLocalUrl.jpg"
	fakeFileSystemItems[ fakeOriginalPath ] = true
	
	local recoveryP = OzOriginalsRecovery.recoveryEffortOriginalP( docStore, {
		captureDate = "2004-03-21T21:20:00Z",
		importSource = {
			folder = "/path/to",
			sha256 = "not used",
			importedOnDevice = KSDeviceUtils.deviceName(),
		},
		_localOnly = {
			localUrl = fakeOriginalPath,
		},
	} )
		:runUntilDoneP()
	
	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		:continueWhenValueMatches( 'data', 'overlay', {
			overlay = {
				_localOnly = {
					originalFilePath = "standard file path < carouselDocuments >/catalog_id_14159265/Originals/2004/2004-03/2004-03-21/testRecoveryFromFileSystemViaLocalUrl.jpg",
					recoveryEffort = "mission accomplished",
				},
			},
		} )
end

--------------------------------------------------------------------------------

function TestSuite.testRecoveryFromAssetLibraryByFilename( test )
	
	local recoveryP = OzOriginalsRecovery.recoveryEffortOriginalP( docStore, {
		captureDate = "0000-00-00T00:00:00Z",
				-- deliberately does not match, forcing fileName lookup
		importSource = {
			fileName = "testRecoveryFromAssetLibraryByFilename.JPG",
			folder = "assets-library:",
			sha256 = "75cd0d84d626b4881c3191836fcd05aa21d2da9028e204fc899207bbb40d99eb",
			importedOnDevice = KSDeviceUtils.deviceName(),
		},
	} )
		:runUntilDoneP()
	
	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		:continueWhenValueMatches( 'data', 'overlay', {
			overlay = {
				_localOnly = {
					originalFilePath = "standard file path < carouselDocuments >/catalog_id_14159265/Originals/0000/0000-00/0000-00-00/testRecoveryFromAssetLibraryByFilename.JPG",
					recoveryEffort = "mission accomplished",
				},
			},
		} )
end

--------------------------------------------------------------------------------

function TestSuite.testRecoveryFromAssetLibraryByDate( test )
	
	local recoveryP = OzOriginalsRecovery.recoveryEffortOriginalP( docStore, {
		captureDate = "2003-02-02T09:46:40Z",
		importSource = {
--			fileName = "testRecoveryFromAssetLibraryByFilename.JPG",
				-- No filename for this test.
			folder = "assets-library:",
			sha256 = "4e1b9edd53aa2e3f9e1eaf6f4a315ffbcedbc60d06c6b4e4f8609813b3b76702",
			importedOnDevice = KSDeviceUtils.deviceName(),
		},
	} )
		:runUntilDoneP()
	
	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		:continueWhenValueMatches( 'data', 'overlay', {
			overlay = {
				_localOnly = {
					originalFilePath = "standard file path < carouselDocuments >/catalog_id_14159265/Originals/2003/2003-02/2003-02-02/testRecoveryFromAssetLibraryByDate.JPG",
					recoveryEffort = "mission accomplished",
				},
			},
		} )

end

--------------------------------------------------------------------------------

function TestSuite.testRecoveryFromAssetLibraryByIdInference( test )
	
	local recoveryP = OzOriginalsRecovery.recoveryEffortOriginalP( docStore, {
		captureDate = "0000-00-00T00:00:00Z",
			-- deliberately cause date matching to miss
		importSource = {
			fileName = "58BEC4F1-FCB6-42E2-B8E2-6668DFCB0B76.JPG",
				-- No filename for this test.
			folder = "assets-library:",
			sha256 = "ae6c2ab23377a3bc7f311f30ad7c0a244fcd91b12cbcdc961a8e363040136c43",
			importedOnDevice = KSDeviceUtils.deviceName(),
		},
	} )
		:runUntilDoneP()
	
	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		:continueWhenValueMatches( 'data', 'overlay', {
			overlay = {
				_localOnly = {
					originalFilePath = "standard file path < carouselDocuments >/catalog_id_14159265/Originals/0000/0000-00/0000-00-00/testRecoveryFromAssetLibraryByIdInference.JPG",
					recoveryEffort = "mission accomplished",
				},
			},
		} )

end

--------------------------------------------------------------------------------

function TestSuite.testRecoverLostOriginalP( test )

	local recoveryP = OzOriginalsRecovery.recoverLostOriginalP( docStore, "some id", "/path" )
		:runUntilDoneP()

	WFAsyncTestCase( test, 999999 )
		:waitFor( recoveryP )
		
		-- Deliberately omitting any validation, because the implementation of recoverLostOriginalP
		-- is so incredibly simple that the validation would only be a copy/paste of recoverLostOriginalP()

end

--==============================================================================

return TestSuite
