--[[----------------------------------------------------------------------------

OzSyncManager.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This namespace coordinates the sync conversations with the central Oz server
run by Adobe across all catalogs.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.lua.threading
local WFRequestDispatcher = require 'WFRequestDispatcher'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- libs.reactive
local KSReactive = require 'KSReactive'

	-- oz.account
local OzAccountService = require 'OzAccountService'

	-- oz.sync_agent
local OzCentralSyncAgent = require 'OzCentralSyncAgent'


--============================================================================--

-- @mark OzSyncManager

local OzSyncManager = {}

--------------------------------------------------------------------------------

local syncManagerLog = KSLogging.log 'OzSyncManager'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local activeSyncAgentBehaviors = {}
	-- key: (string) catalogId
	-- value: (B< nil or OzCentralSyncAgent >) sync agent, but only if
		-- sync is currently enabled for this catalog

	-- NOTE: This is intentionally not a weak map. Once referenced, the
	-- behaviors stay present indefinitely so we can do mapB and mapSwitchB
	-- on them. We do clear out the agent itself.

local previousUser

--------------------------------------------------------------------------------

local function syncAgentBehaviorForCatalogId( catalogId )

	local resultB = activeSyncAgentBehaviors[ assert( catalogId ) ]
	
	if not resultB then

		syncManagerLog:debug( "Creating new syncAgentB for catalog ID %s", catalogId )
	
		resultB = KSReactive.variableB( nil )
		activeSyncAgentBehaviors[ catalogId ] = resultB
	
	end
	
	return resultB

end

--------------------------------------------------------------------------------

local currentUserB = OzAccountService.currentUserB()

local userSyncAgentAction = OzAccountService.currentUserB():onDataB( function( ozUser )

	local userId = nil

	-- Start sync for the current user.

	if ozUser ~= nil then

		-- If we got a new user without a nil in between, remove previous sync agents.
		
		if ozUser:id() ~= previousUser then

			for catalogId, syncAgentB in pairs( activeSyncAgentBehaviors ) do

			local oldSyncAgent = syncAgentB:valueNow()

				if oldSyncAgent then
					oldSyncAgent:close()
					syncAgentB:set( nil )
				end

			end

			activeSyncAgentBehaviors = {}

		end
	
		userId = ozUser:id()

		local userSyncAgentB = syncAgentBehaviorForCatalogId( userId )
	
		if not userSyncAgentB:valueNow() then

			syncManagerLog:debug( "Creating new syncAgentB for user ID %s due to user ID change", userId )
	
			userSyncAgentB:set( OzCentralSyncAgent( userId, ozUser ) )

		end

		previousUser = ozUser:id()

	else

		-- Close all the existing sync agents and empty the map.

		for catalogId, syncAgentB in pairs( activeSyncAgentBehaviors ) do
	
			syncManagerLog:debug( "Closing all existing sync agents due to user ID change" )

			local oldSyncAgent = syncAgentB:valueNow()
		
			if oldSyncAgent then
				oldSyncAgent:close()
				syncAgentB:set( nil )
			end
	
		end

		activeSyncAgentBehaviors = {}

	end
	
end )

rawset( _G, {}, userSyncAgentAction )
	-- [scouten 2011-11-05]: Ensure that this monitoring of the user
	-- ID doesn't get GC'd. Hacky, but it will work.


--============================================================================--

-- @mark sync agent management

--------------------------------------------------------------------------------

local function startSyncForCatalogId( catalogId )
			
	assert( type( catalogId ) == 'string',
		"startSync: catalogId was not a string" )
	
	local syncAgentB = syncAgentBehaviorForCatalogId( catalogId )
	
	if not syncAgentB:valueNow() then

		syncManagerLog:debug( "Creating OzCentralSyncAgent for catalog %s",
							  catalogId )

		syncAgentB:set( OzCentralSyncAgent( catalogId, currentUserB:valueNow() ) )

	end
	
end

--------------------------------------------------------------------------------

function OzSyncManager.installMessageHandlers()

	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'startSync', startSyncForCatalogId )
	
	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'stopSync',
		function( catalogId )
		
			assert( type( catalogId ) == 'string',
				"stopSync: catalogId was not a string" )
			
			local syncAgentB = syncAgentBehaviorForCatalogId( catalogId )
			
			local syncAgent = syncAgentB:valueNow()
			
			if syncAgent then

				syncManagerLog:debug( "Closing OzCentralSyncAgent for catalog %s",
									   catalogId )

				syncAgentB:set( nil )
					-- Do this first, so behaviors that are mapped from
					-- syncAgentB don't also get closed.

				syncAgent:close()

			end
			
		end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'syncEnabledB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId ):toBooleanB()
		
	end )

	----------------------------------------------------------------------------

	-- FIXME [scouten 2011-04-22]: Refactor this mapSwitchB pattern?

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'isUploadingToOzB', function( catalogId )
	
		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'isUploadingToOzB', false )
		
	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'isDownloadingFromOzB', function( catalogId )
	
		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'isDownloadingFromOzB', false )
		
	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'downloadingFromOzProgressB', function( catalogId )
	
		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'downloadingFromOzProgressB', false )
		
	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'hasDownloadedSuccessfullyB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'hasDownloadedSuccessfullyB', false )

	end )
		
	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'uploadErrorB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'uploadErrorB', nil )

	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'downloadErrorB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'downloadErrorB', nil )

	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'existsOnOzB', function( catalogId )
	
		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'existsOnOzB', true )
	
	end )
	
	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'DEBUG_statusReport', function( catalogId, photoId )
	
		local syncAgentB = syncAgentBehaviorForCatalogId( catalogId )
		
		local syncAgent = syncAgentB:valueNow()
		
		return syncAgent and syncAgent:DEBUG_statusReport( photoId )
			or "(not currently syncing)"
	
	end )
	
	----------------------------------------------------------------------------
		
	WFRequestDispatcher.onMessage( 'itmsLinkB', function( catalogId )
		
		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'itmsLinkB', nil )
		
	end )
	
	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'accessErrorB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'accessErrorB', false )

	end )
	
	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessage( 'serviceAvailableB', function( catalogId )

		return syncAgentBehaviorForCatalogId( catalogId )
			:mcallSwitchB( 'serviceAvailableB', true )

	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'syncNowP',
		function( catalogId )
	
			assert( type( catalogId ) == 'string',
				"stopSync: catalogId was not a string" )
			
			local syncAgentB = syncAgentBehaviorForCatalogId( catalogId )
			
			local syncAgent = syncAgentB:valueNow()
			
			if syncAgent then

				return syncAgent:syncNowP()

			else
			
				-- Manufacture an error promise to avoid halting here in the debugger.
				-- Due to the way we use KSPrefs to manage the current user,
				-- it is all to common to hit this error when logging out.
				local KSPromises = require "KSPromises"
				KSPromises.resolvedPromiseP( "error", "sync not enabled for this catalog" )

			end
			
	end )

end

--------------------------------------------------------------------------------

return OzSyncManager
