--[[----------------------------------------------------------------------------

OzUser.lua

@author scouten 2011-07-11

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class manages the information about a single user in the Oz/IMS ecosystem.

This code may be used from any thread/event processor.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'
local WFTimer = require "WFTimer"

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local WFPlatformInfo = require 'WFPlatformInfo'


	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- wichitafoundation.core.string
local WFZString = require 'WFZString'
local KSStringUtils = require 'KSStringUtils'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.os.prefs
local KSPrefs = require 'KSPrefs'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- wichitafoundation.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- wichitafoundation.web.http_client
local WFHttpClient = require 'WFHttpClient'

	-- wichitafoundation.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- wichitafoundation.web.oauth2_client
local KSOAuth2Session = require 'KSOAuth2Session'

	-- third_party.yajl
local yajl = require 'yajl'


--------------------------------------------------------------------------------

local userLog = KSLogging.log 'OzUser'
--	:setLevel( KSLogging.DEBUG )

local catalogListLog = KSLogging.log 'OzCatalogList'
--	:setLevel( KSLogging.DEBUG )

local ZSTR = assert( WFZString.fromZString )

local unknownUserStr = ZSTR "$$$/OzAccountService/User/Unknown=(unknown)"

local table_sort = table.sort

local string_lower = string.lower

local string_sub = string.sub

local catalogsBforUserId = KSTableUtils.makeFullyWeak()

local sharedCatalogsBforUserId = KSTableUtils.makeFullyWeak()

local currentUserB -- lazy instantiated (cyclic dependency, ugh!)

local forceUpgradeBforUserId = KSTableUtils.makeFullyWeak()


--============================================================================--

-- @mark OzUser class

--[[!
       \class OzUser

      \brief This class manages the information about a single user in the Oz/IMS ecosystem.
       
       You can call the functions from any thread. Note that for operations on catalogs,
       you might want to access information about the catalog's owner through that
       catalog object.

       \sa OzCatUsers
]]--

local userMethods = {

	_ozUserId = false,
		-- (string) Adobe user ID

	_userInfoCache = false,
		-- (OzDocStore) this doc store retains info about this user

	_userInfoIndex = false,
		-- (WFIndexStore) index store for user info doc store

	_oAuth2Session = false,
		-- (KSOAuth2Session) OAuth2 session for talking to Oz via IMS

	_userInfoB = false,
		-- (B< table >) doc value behavior for user info cache

	_rawReloadCatalogListNowE = false,
		-- (E< boolean >) internal trigger that catalog list should be
		-- reloaded immediately

	_rawReloadSharedCatalogListNowE = false,
		-- (E< boolean >) internal trigger that the shared catalog list should be
		-- reloaded immediately

	_calmedReloadCatalogListNowE = false,
		-- (E< boolean >) internal trigger that catalog list should be
		-- reloaded immediately (with calming filter)

	_calmedReloadSharedCatalogListNowE = false,
		-- (E< boolean >) internal trigger that the shared catalog list should be
		-- reloaded immediately (with calming filter)

	_receiptNotificationQueueB = false,
		-- (B< list< receiptDocId, receiptStatus, ozErrorCode, externalStoreErrorCode > >)
			-- app store receipts that have been processed but not acknowledged
			-- by the UI

	_catalogQuotaB = false,
		-- (B< number >) the catalog quota for this user

	_catalogUpdateTimerE = false,
		-- (E< true >) event stream used to trigger a deferred update for catalog status

	_previousRefreshTime = false,
		-- (number) time when previous refreshP was obtained
	
	_refreshP = false,
		-- (promise) cached refreshP that is used to prevent multiple calls to the oz service
		-- within a brief (currently 5 second) window
	
	_wantsRefresh = false,
		-- (boolean) if we ever vend the cached refreshP, we set this flag so that we will
		-- refresh after our timer expires.
	
	_omnitureOptInB = false,
		-- (boolean) value for the user's omniture setting
		
	_omnitureOptInLocalB = false,
		-- (boolean or nil) our local value after use changes the setting
	
	_lastInternalProgressCount = false,
	_lastExternalProgressCount = false,
		-- (number) values for tracking download progress
	
	_pendingUploadP = false,
		-- (promise) the last request to PendingUpload; used to chain
		-- multiple calls together to ensure we don't skip over values
	
	WFCloseSupport.addCloseSupportMethods,

-- TORNADO HACK_1.1
	_serviceAvailableB = false,
-- END TORNADO HACK

	_catalogUpateTime = 120,
		-- amount of time between catalog updates

}

--------------------------------------------------------------------------------
--[[!
      \brief Constructs an OzUser from the database user ID.

      \param ozUserId (string) The unique user ID for the Oz database.
]]--

function userMethods:__init( ozUserId )

	if type( ozUserId ) ~= 'string' then
		error( "OzUser: must initialize with a valid user UUID", 4 )
			-- FIXME [scouten 2011-08-05]: This should really enforce that it's a UUID,
			-- not just a string.
	end

	if WFPlatformInfo.isTornado then
		self._catalogUpateTime = 120
		self._serviceAvailableB = KSReactive.variableB( true )
	end

	self._userInfoCache = OzDocStore( ozUserId )
	self:addCloseTarget( self._userInfoCache )

	self._userInfoIndex = WFIndexStore( self._userInfoCache,
										'OzUserInfoCacheIndexDefinitions.lua' )
	self:addCloseTarget( self._userInfoIndex )

	self._ozUserId = ozUserId

	self._oAuth2Session = KSOAuth2Session( ozUserId )

	self._userInfoB = self._userInfoCache:docValueB( 'user', self._ozUserId )
		:addAsCloseTargetOf( self )

	self._rawReloadCatalogListNowE = KSReactive.sourceE()
		:addAsCloseTargetOf( self )

	self._calmedReloadCatalogListNowE = self._rawReloadCatalogListNowE:calmE( .5 )
		:addAsCloseTargetOf( self )
	self._calmedReloadCatalogListNowE:onDataE( function()
		catalogListLog:debug( "Requesting catalog list via _calmedReloadCatalogListNowE" )
		self:catalogsP():runUntilDone()
	end )
		:addAsCloseTargetOf( self )

	self._rawReloadSharedCatalogListNowE = KSReactive.sourceE()
		:addAsCloseTargetOf( self )

	self._calmedReloadSharedCatalogListNowE = self._rawReloadSharedCatalogListNowE:calmE( .5 )
		:addAsCloseTargetOf( self )
	self._calmedReloadSharedCatalogListNowE:onDataE( function()
		self:sharedCatalogsP():runUntilDone()
	end )

	self._receiptNotificationQueueB = self._userInfoIndex:queryB(
			[[SELECT docId, receiptStatus, ozErrorCode, externalStoreErrorCode
				FROM pendingReceiptNotifications
				WHERE userId == ?
				ORDER BY receiptStatus, docId]],
					self._ozUserId )
		:addAsCloseTargetOf( self )
		
	self:refreshP()

	self._receiptNotificationQueueB:onDataB( function( pendingReceiptNotifications )

		for i = 1, #pendingReceiptNotifications do

			local pendingReceipt = pendingReceiptNotifications[ i ]
			local receiptStatus = pendingReceipt[ 2 ]
			if receiptStatus == 'accepted' then
				self:refreshP()
				break
			end

		end

	end )
	
	self._omnitureOptInLocalB = KSReactive.variableB( nil )
		:addAsCloseTargetOf( self )

end

--------------------------------------------------------------------------------

-- @mark basic info about user

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the unique ID of this Oz user.

	\return (string) The ID.
]]--

function userMethods:id()

	return self._ozUserId

end

--------------------------------------------------------------------------------

if WFPlatformInfo.isTornado then

--[[!
	Reports (as a Behavior) whether this Oz service is avaliable.
	\return <code>B < boolean ></code> A Behavior object whose value is
		\c true if service is available, \c false otherwise.
]]--

function userMethods:isServiceAvailableB()
	return self._serviceAvailableB:relayB()
end

end
--------------------------------------------------------------------------------

--[[!
	\brief Reports (as a Behavior) whether this Oz user is current.

	\return <code>B < boolean ></code> A Behavior whose value is
		\c true if this user is current, \c false otherwise.
]]--

function userMethods:isCurrentUserB()

	return KSPrefs.prefB( 'currentUserId' )
		:closingMapB( function( userId )

			return userId == self._ozUserId

		end )

	-- FIXME [scouten 2011-06-09]: Once OzCatalogService is wired up to OzUser,
	-- then it should be watching this behavior and shutting itself down if
	-- this behavior goes to false.

end

--------------------------------------------------------------------------------

local function isCurrentUserNow( self )
	return KSPrefs.currentUserId == self._ozUserId
end

--------------------------------------------------------------------------------

local function mapNilToUnknown( originalValue )

	return originalValue or unknownUserStr

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the first name of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the first name.
]]--

function userMethods:firstNameB()

	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, 'first_name' )
		:mapB( mapNilToUnknown )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the last name of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the last name.
]]--

function userMethods:lastNameB()

	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, 'last_name' )
		:mapB( mapNilToUnknown )

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves (as a Behavior) the full name of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the full name.
]]--

function userMethods:fullNameB()

	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, 'full_name' )
		:mapB( mapNilToUnknown )

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves (as a Behavior) the email address of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the email address.
]]--

function userMethods:emailB()

	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, 'email' )
		:mapB( function( rawEmailValue )

			return rawEmailValue or ""

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) information about this Oz user.

	\return <code>P < table ></code> A Promise that resolves to
		the user information.
]]--


function userMethods:userInfoP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )

end

--------------------------------------------------------------------------------

-- @mark subscription/trial status

--------------------------------------------------------------------------------
--[[!

	\brief Retrieves (as a Behavior) the current account status of this Oz user.
 
	\return <code>B < nil | string ></code> A Behavior whose value is
		initially nil, then the account status, one of:
		<ul><li>created</li>
		<li>trial</li>
		<li>trial_grace_period</li>
		<li>trial_expired</li>
		<li>subscriber</li>
		<li>subscription_grace_period</li>
		<li>subscription_expired</li>
		</ul>
]]--

function userMethods:accountStatusB()

	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, 'status' )
		:mapB( function( status )

			userLog:debug( "accountStatusB updated to %s", status or '(nil)' )

			return status or nil

		end )

end

--------------------------------------------------------------------------------

 -- IMPORTANT NOTE on Freemium user APIS (from epoley, 10/19/2012):
 -- There are two different Freemium user APIs to choose from.
 -- catalog:isCatalogOwnerFreemiumOrTrialUserP/B() apply to
 -- most cases--they tell you whether freemium restrictions should
 -- be enforced for operations on the current catalog. This takes into account
 -- the freemium status of the catalog's owner, not of the current user.
 --
 -- In most cases related to catalog data, the below APIs on OzUser are not
 -- what you want. But they may be needed in some cases, such as enforcing
 -- restrictions on library creation, or perhaps driving certain local notifications.

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether this Oz user is subject to Freemium limits.

	\return <code>B < boolean | nil ></code> A Behavior whose value is
		initially nil, then \c true if this user is subject to limits,
		\c false otherwise.

	\warning Applies to this user, not the current catalog owner;
		for most operations on catalogs, use the catalog-owner version,
		<code>catalog:isCatalogOwnerFreemiumOrTrialUserB()</code>.
		You can use this in some cases, such as enforcing restrictions
		on library creation, or driving certain local notifications.
]]--

function userMethods:isFreemiumOrTrialUserB()

	return self._userInfoCache:docValueB( 'user', self._ozUserId )
		:mapB( function( docValue )

			userLog:debug( "isFreemiumOrTrialUserB evaluating docValue %s", docValue )

			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local uploadLimit = KSTableUtils.derefOpt( docValue, 'limits', 'uploads' )

			return status == 'trial' or type( uploadLimit ) == 'number'

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) whether this Oz user is subject to Freemium limits.

	\return <code>P < boolean ></code> A Promise that resolves to
		 \c true if this user is subject to limits, \c false otherwise.

	\warning Applies to this user, not the current catalog owner;
		for most operations on catalogs, use the catalog-owner version,
		<code>catalog:isCatalogOwnerFreemiumOrTrialUserP()</code>.
		You can use this in some cases, such as enforcing restrictions
		on library creation, or driving certain local notifications.
]]--

function userMethods:isFreemiumOrTrialUserP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local uploadLimit = KSTableUtils.derefOpt( docValue, 'limits', 'uploads' )

			return 'data', ( status == 'trial' or type( uploadLimit ) == 'number' )

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether this Oz user is subject to metered Freemium limits.

	\return <code>B < boolean | nil ></code> A Behavior whose value is
		initially nil, then \c true if this user is subject to limits,
		\c false otherwise.

	\warning Applies to this user, not the current catalog owner;
		for most operations on catalogs, use the catalog-owner version,
		<code>catalog:isCatalogOwnerMeteredFreemiumUserB()</code>.
		You can use this in some cases, such as enforcing restrictions
		on library creation, or driving certain local notifications.
]]--

function userMethods:isMeteredFreemiumUserB()

	return self:uploadLimitB()
		:mapB( function( uploadLimit )

			if type( uploadLimit ) == 'number' then
				return true
			else
				return uploadLimit
					-- could be false (not freemium user)
					-- or nil (don't know yet)
			end
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) whether this Oz user is subject to metered Freemium limits.

	\return <code>P < boolean ></code> A Promise that resolves to
		 \c true if this user is subject to limits, \c false otherwise.

	\warning Applies to this user, not the current catalog owner;
		for most operations on catalogs, use the catalog-owner version,
		<code>catalog:isCatalogOwnerMeteredFreemiumUserP()</code>.
		You can use this in some cases, such as enforcing restrictions
		on library creation, or driving certain local notifications.
]]--

function userMethods:isMeteredFreemiumUserP()

	return self:uploadLimitP()
		:mapDataP( function( _, uploadLimit )

			return 'data', type( uploadLimit ) == 'number'

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the trial start date if this user is in trial mode.
	
	\return <code>B< string | nil ></code> A Behavior whose value is
		the user's trial start date (in ISO 8601 time) or nil if this user is not
		in trial mode.
]]--

function userMethods:trialStartDateB()

	return self._userInfoCache:docValueB( 'user', self._ozUserId )
		:mapB( function( docValue )
		
			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local trialStart = KSTableUtils.derefOpt( docValue, 'trial', 'start' )
			
			if status == 'trial' then
				return trialStart
			else
				return nil
			end
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the trial start date if this user is in trial mode.
	
	\return <code>P< string | nil ></code> A Promise that resolves to
		the user's trial start date (in ISO 8601 time) or nil if this user is not
		in trial mode.
]]--

function userMethods:trialStartDateP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
		
			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local trialStart = KSTableUtils.derefOpt( docValue, 'trial', 'start' )
			
			if status == 'trial' then
				return 'data', trialStart
			else
				return 'data', nil
			end
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the trial end date if this user is in trial mode.
	
	\return <code>B< string | nil ></code> A Behavior whose value is
		the user's trial end date (in ISO 8601 time) or nil if this user is not
		in trial mode.
]]--

function userMethods:trialEndDateB()

	return self._userInfoCache:docValueB( 'user', self._ozUserId )
		:mapB( function( docValue )

			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local trialEnd = KSTableUtils.derefOpt( docValue, 'trial', 'end' )

			if status == 'trial' then
				return trialEnd
			else
				return nil
			end

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the trial end date if this user is in trial mode.
	
	\return <code>P< string | nil ></code> A Promise that resolves to
		the user's trial end date (in ISO 8601 time) or nil if this user is not
		in trial mode.
]]--

function userMethods:trialEndDateP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local status = KSTableUtils.derefOpt( docValue, 'status' )
			local trialEnd = KSTableUtils.derefOpt( docValue, 'trial', 'end' )

			if status == 'trial' then
				return 'data', trialEnd
			else
				return 'data', nil
			end

		end )

end

--------------------------------------------------------------------------------

local function daysRemainingInTrialForTrialEndDate( trialEndDate )

	if type( trialEndDate ) == 'string' then

	    local timeNow = KSDate.timeInIso8601Format( nil, false )

		local secondsRemainingInTrial = KSDate.differenceInSeconds( timeNow,
											trialEndDate, false )

		return secondsRemainingInTrial / ( 24 * 60 * 60 )

	else
		return false
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the number of days remaining in the trial
	if this user is in trial mode. The number can be fractional.
	
	\return <code>P< number | false ></code> A Promise that resolves to
		the number of days, or \c false if this user is not
		in trial mode.
]]--

function userMethods:daysRemainingInTrialP()

	return self:trialEndDateP()
		:mapDataP( function( _, trialEndDate )
			
			return 'data', daysRemainingInTrialForTrialEndDate( trialEndDate )

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the maximum number of photos this user is
	allowed to upload in the current period.
	
	\return <code>B< number | false | nil ></code> A Behavior whose value is
		initially nil, then the maximum number of photos, or \c false if
		uploads are unlimited.
]]--

function userMethods:uploadLimitB()

	return self._userInfoCache:docValueB( 'user', self._ozUserId )
		:mapB( function( docValue )

			local uploadLimit = KSTableUtils.derefOpt( docValue, 'limits', 'uploads' )

			if docValue then
				return uploadLimit or false
			else
				return nil
			end

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the maximum number of photos this user
	is allowed to upload in the current period.
	
	\return <code>P< number | false ></code> A Promise that resolves to
	 the maximum allowed number of uploads, or \c false if uploads are unlimited.
]]--

function userMethods:uploadLimitP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
		
			local uploadLimit = KSTableUtils.derefOpt( docValue, 'limits', 'uploads' )
			
			return 'data', uploadLimit or false

		end )
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the number of photos this user
	has uploaded so far in the current period.
	
	\return <code>B< number | false | nil ></code> A Behavior whose value is
	 the current number of uploads, or nil if the number of uploads is not yet known.
]]--

function userMethods:uploadUsageB()

	return self._userInfoCache:docValueB( 'user', self._ozUserId )
		:mapB( function( docValue )

			local uploadUsage = KSTableUtils.derefOpt( docValue, 'usage', 'uploads' )

			if docValue then
				return uploadUsage or 0
			else
				return nil
			end

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adjusts the local copy of the upload usage. This change persists
	until the next \c refreshP call.  If the account is not limited,
	does nothing but still returns a promise with true data.
	
	\param amount (number) The amount by which to adjust usage.

	\return <code>P< boolean ></code> A Promise that resolves to \c true
		when the request has succeeded.

]]--

function userMethods:adjustUploadUsageP( --[[number]] amount )
	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
			
			local currentUsage = KSTableUtils.derefOpt( docValue, 'usage', 'uploads' )
			if type( currentUsage ) == 'number' then
				
				return self._userInfoCache:applyOverlayToDocumentP {
					docType = 'user',
					docId = self._ozUserId,
					overlay = {
						usage = {
							uploads = currentUsage + amount,
						}
					}
				}
				
			end
			
			-- we are not metered; don't record
			return 'data', true
		end )
end

--------------------------------------------------------------------------------
--[[!
 	\brief Reports (as a Promise) whether this user is subject to upload limits,
 	and if so, the difference between the limit and the current number of uploads.

	\return <code>P< number | false ></code> A Promise that resolves to
		 the greater of <code>limit - usage</code> or 0 if uploads are unlimited.
]]--

function userMethods:uploadUsageP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
		
			return 'data', KSTableUtils.derefOpt( docValue, 'usage', 'uploads' ) or 0

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) both the upload limit for this user and the
	current number of photos uploaded in this period.
	
	\return <code>P< number | false, number | false ></code> A Promise that resolves to
	two values. The first is the maximum, or \c false if uploads are unlimited.
	The second value is the number of photos the user has uploaded so far in this
	period.
]]--

function userMethods:uploadLimitAndUsageP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
		
			return 'data',
				   KSTableUtils.derefOpt( docValue, 'limits', 'uploads' ) or false,
				   KSTableUtils.derefOpt( docValue, 'usage', 'uploads' ) or 0

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the number of uploads remaining if this user
	is subject to upload limits.
	
	\return <code>P< number | false ></code> A Promise that resolves to
		the the greater of limit - usage and 0, or \c false if usage is unlimited.
 ]]--

function userMethods:availableUploadsP()
	
	return self:uploadLimitAndUsageP()
		:mapDataP( function( _, uploadLimit, uploadUsage )
		
			if type( uploadLimit ) == 'number' and type( uploadUsage ) == 'number' then
				return 'data', math.max( 0, uploadLimit - uploadUsage )
			else
				return 'data', uploadLimit
			end
		
		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise)  the date and time when the current metered upload period ends.
	
	\return <code>P< string | false ></code> A Promise that resolves to
		the ISO 8601 date string for the end date, or \c false if
		usage is unlimited for this user.
--]]

function userMethods:uploadPeriodEndDateP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )
		
			userLog:info( "uploadPeriodEndDateP processing docValue = %s", docValue )
		
			return 'data', KSTableUtils.derefOpt( docValue, 'limits', 'uploads_period_end' )
						   or false

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the time remaining until the current metered upload period ends.
	
	\return <code>P< number | false ></code> A Promise that resolves to
		a number of days (which can be fractional), or \c false if
		usage is unlimited for this user.
 --]]

function userMethods:daysUntilUploadPeriodEndsP()

	return self:uploadPeriodEndDateP()
		:mapDataP( function( _, trialEndDate )

			if type( trialEndDate ) == 'string' then

			    local timeNow = KSDate.timeInIso8601Format( nil, false )

				local secondsRemainingInTrial = KSDate.differenceInSeconds( timeNow,
													trialEndDate, false )

				return 'data', secondsRemainingInTrial / ( 24 * 60 * 60 )

			else
				return 'data', false
			end

		end )

end

--------------------------------------------------------------------------------

-- @mark catalog limits

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) the number of catalogs left in this user's quota.

	\return <code>B< nil | number ></code> A Behavior whose value is the number of catalogs,
	   (updated when a catalog is added to or deleted from this user's account), or nil
	   if usage is unlimited.
]]--

function userMethods:catalogQuotaB()

	if not self._catalogQuotaB then

		self._catalogQuotaB = self._userInfoCache:fieldValueB( 'user', self._ozUserId,
															   'resources/catalogs' )

		self:addCloseTarget( self._catalogQuotaB )

	end

	return self._catalogQuotaB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the number of catalogs this user is allowed.
		
	\return <code>P< nil | number ></code> A Promise that resolves to the catalog limit,
		or nil if usage is unlimited.
]]--

function userMethods:catalogLimitP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local result = KSTableUtils.derefOpt( docValue, 'limits', 'catalogs' )

			if not result then

				-- What to do if user has no limit specified?
				-- FIXME [scouten 2011-08-23]: Still need to understand why we're
				-- not getting limits set for some users.

				userLog:warn( "OzUser:catalogLimitP: no catalog limit found in %s", docValue )
				result = 20

			end

			return 'data', result

		end )

end

--------------------------------------------------------------------------------
--[[ 	FIXME [ngautam 08/05/2011] : This might move to OzCatalogService if the
	limits are defined to be different per catalog.
]]--

--[[!
	\brief Retrieves (as a Promise) the number of collaborators this user can have per catalog.

	\return <code>P< nil | number ></code> A Promise that resoves to the number of collaborators,
		or nil if usage is unlimited.
]]--

function userMethods:collaboratorLimitP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local result = KSTableUtils.derefOpt( docValue, 'limits', 'catalog_users' )

			if not result then

				-- What to do if user has no limit specified?
				-- FIXME [scouten 2011-08-23]: Still need to understand why we're
				-- not getting limits set for some users.

				userLog:warn( "OzUser:catalogLimitP: no collaborator limit found in %s",
							  docValue )

				result = 50

			end

			return 'data', result

		end )

end

--------------------------------------------------------------------------------

-- @mark notification settings

--------------------------------------------------------------------------------

local function daysToNotifyBeforeUnlimitedImportEndsForValue( docValue )

	return KSTableUtils.derefOpt( docValue, 'config',
											'notify_daily_in_app_unlimited_import_end' )
		or 15 -- provide a default if Oz doesn't give us one

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) the number of days left before unlimited import
		 ends when we should notify this user.

	\return <code>P< number ></code> A Promise that resolves to the number of days.
]]--

function userMethods:daysToNotifyBeforeUnlimitedImportEndsP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			return 'data', daysToNotifyBeforeUnlimitedImportEndsForValue( docValue )

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) the number of days left before unlimited import
		 ends, when we should issue a local notification.

	\return <code>P< number ></code> A Promise that resolves to the number of days.
]]--

function userMethods:daysToLocalNotifyBeforeUnlimitedImportEndsP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			return 'data', KSTableUtils.derefOpt( docValue, 'config',
												  'notify_local_unlimited_import_end' )
						   or 3 -- default if Oz doesn't give us one

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) the number of days left before the import quota
		 resets when we should notify this user.

	\return <code>P< number ></code> A Promise that resolves to the number of days.
]]--

function userMethods:daysToNotifyBeforeQuotaResetsP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			return 'data', KSTableUtils.derefOpt( docValue, 'config',
												  'notify_local_quota_reset' )
						   or 3 -- default if Oz doesn't give us one

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) the number of days left before before the import quota
		 resets, when we should issue a local notification.

	\return <code>P< number ></code> A Promise that resolves to the number of days.
]]--


function userMethods:daysToNotifyLocallyWithoutAppLaunchP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			return 'data', KSTableUtils.derefOpt( docValue, 'config',
												  'notify_local_without_app_launch' )
						   or 14 -- default if Oz doesn't give us one

		end )

end

--------------------------------------------------------------------------------

-- @mark notification decisions

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) whether we are currently in the proper time period
	for notification that the upload window is closing.
	(Note that this does not remember whether a notification has previously been shown.)
 	
	\return <code>P< boolean ></code> A Promise that resolves to \c true if we are
		within the notification period.
]]--

function userMethods:shouldNotifyThatUnlimitedImportsWillEndP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local status = KSTableUtils.derefOpt( docValue, 'status' ) or 'trial'
			local daysToNotify = daysToNotifyBeforeUnlimitedImportEndsForValue( docValue )
			local uploadLimit = KSTableUtils.derefOpt( docValue, 'limits', 'uploads' )
			local trialEndDate = KSTableUtils.derefOpt( docValue, 'trial', 'end' )

			if status ~= 'trial'
				or uploadLimit
				or type( daysToNotify ) ~= 'number'
				or type( trialEndDate ) ~= 'string'
			then
				return 'data', false, 0
					-- not in unlimited trial mode, no notification
			else
				local daysRemaining = daysRemainingInTrialForTrialEndDate( trialEndDate )
				return 'data', ( type( daysRemaining ) == 'number' )
									and ( daysRemaining <= daysToNotify ), daysRemaining
			end

		end )

end

--------------------------------------------------------------------------------
-- @mark Activity Bookmark
--------------------------------------------------------------------------------

--[[!
	\brief set the activity feed for this user to \c bookmark
	\param catalogId (string) The id of the catalog for which to bookmark.
	\param bookmark the value retrieved from an activity item
	\see OzCatalogService.activityB
	\return <code> P<true></code> A promise which resolves when the book mark has been set
]]--

function userMethods:setActivityBookmarkP( catalogId, bookmark )
	
	WFArguments.check( 'OzUser', 'string', 'string' )

	userLog:debug( "setActivityBookmark bookmark = %s", bookmark )

	return self._userInfoCache:applyOverlayToDocumentP {
			docType = 'user',
			docId = self._ozUserId,
			overlay = {
				settings = {
					revel = {
						activity_bookmarks = {
							[ catalogId ] = bookmark,
						}
					}
				}
			}
		}
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief get the activity feed for this for this catalog
	\param catalogId (string) The id of the catalog for which to bookmark.
	\see OzCatalogService.activityB
	\return <code> B<bookmark></code> A promise which resolves to the current bookmark
]]--

function userMethods:getActivityBookmarkB( catalogId )
	
	WFArguments.check( 'OzUser', 'string' )

	return self._userInfoCache:fieldValueB( 'user',
											   self._ozUserId,
											   'settings/revel/activity_bookmarks/' .. catalogId )

end

--------------------------------------------------------------------------------

-- @mark Omniture permissions

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) this user's Omniture analytics opt-in setting.

 	\return <code> B< nil | boolean ></code> A Behavior whose value is the
		current setting, if any, or nil if this preference is not set.
]]--

function userMethods:omnitureOptInB()

	if not self._omnitureOptInB then

		self._omnitureOptInB = KSReactive.dataCallB( function( docValue, ourValue )
			if ourValue ~= nil then
				return ourValue
			else
				return docValue
			end
			
		end,
		self._userInfoCache:docValueB( 'user', self._ozUserId ):closingMapB( function( docValue )
			self._omnitureOptInLocalB:set( nil )
			return KSTableUtils.derefOpt( docValue, 'settings', 'revel', 'omnitureOptIn' )
		end ),
		self._omnitureOptInLocalB )

		self:addCloseTarget( self._omnitureOptInB )

	end

	return self._omnitureOptInB

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Promise) this user's Omniture analytics opt-in setting.

 	\return <code> P< nil | boolean ></code> A Promise that resolves to the
		current setting, if any, or nil if this preference is not set.
]]--

function userMethods:omnitureOptInP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			local result = KSTableUtils.derefOpt( docValue,
												  'settings',
												  'revel',
												  'omnitureOptIn' )
			return 'data', result

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the value for the user's Omniture opt-in preference.

	\param preference (optional boolean) The value, \c true if user has opted in to Omniture
		analytics, \c false if not. For debugging only, pass \c nil to unset preference.

	\return <code>P< true ></code> A Promise that resolves when the request has succeeded.
]]--

function userMethods:setOmnitureOptInP( --[[optional boolean]] preference )

	-- FIXME [scouten 2012-10-25]: Oz Next is failing with an internal server
	-- error when calling this API.

	WFArguments.check( 'OzUser', 'boolean|nil' )

	if preference == nil then
		preference = yajl.null
	end

	local settingsData = KSJsonEncoder.encodeOzUploadJson
		{
			revel = {
				omnitureOptIn = preference,
			},
		}

	local requestUrl = string.format( '%s/v1.0/accounts/%s/settings',
									  OzCatalogUrls.baseServerUrl,
									  self._ozUserId )

	self._omnitureOptInLocalB:set( preference )
	
	return self:putP( requestUrl, settingsData,
						{
							requestHeaders = {
								{
									header = 'Content-Type',
									value = 'application/json',
								},
							},
							requestMethod = 'POST',
						} )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
-- @mark Progress Count
--------------------------------------------------------------------------------

--[[!
	\brief set the progress counter for how many items were found in the feed
	\param catalogId (string) The id of the catalog for the sync info.
	\param count the count
	\param doneCounting (bool) if doneCounting is true, this value is set into the user
		and shared with other devices.
	\return <code> P<true></code> A promise which resolves when the count has been set
]]--

function userMethods:setDownloadBookmarkP( catalogId, count, doneCounting )
	
	WFArguments.check( 'OzUser', 'string', 'number', 'boolean' )

	userLog:debug( "setDownloadBookmarkP count = %d doneCounting = %s", count, doneCounting and "YES" or "NO" )
	
	local overlay = {
	}
	local preP = KSPromises.resolvedPromiseP( 'data', true )
	
	if doneCounting and ( ( not self._lastExternalProgressCount ) or count > self._lastExternalProgressCount ) then
		self._lastExternalProgressCount = count
		overlay.settings = {
			revel = {
				download_progress = {
					[ catalogId ] = count,
				}
			}
		}
		local settingsData = KSJsonEncoder.encodeOzUploadJson( overlay.settings )	
		local requestUrl = string.format( '%s/v1.0/accounts/%s/settings',
										  OzCatalogUrls.baseServerUrl,
										  self._ozUserId )
	
		preP = preP:mapDataP( function()
			return self:putP( requestUrl, settingsData,
				{
					requestHeaders = {
						{
							header = 'Content-Type',
							value = 'application/json',
						},
					},
					requestMethod = 'POST',
				} )
		end )
		
	end
	
	if self._lastInternalProgressCount == false or count > self._lastInternalProgressCount then
		self._lastInternalProgressCount = count
		overlay._localOnly = {
			download_progress = {
				[ catalogId ] = count,
			}
		}
	
		preP = preP:mapDataP( function()
			return ( self._userInfoCache:applyOverlayToDocumentP {
				docType = 'user',
				docId = self._ozUserId,
				overlay = overlay
			} )
		end )
	end
	
	return preP:runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief get the activity feed for this for this catalog
	\param catalogId (string) The id of the catalog for which to bookmark.
	\return <code> B<bookmark></code> A promise which resolves to the current bookmark
]]--

function userMethods:getDownloadBookmarkP( catalogId )
	
	WFArguments.check( 'OzUser', 'string' )
	
	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
			:mapDataP( function( status, userDoc )	
				
				self._lastExternalProgressCount = KSTableUtils.derefOpt( userDoc, 'settings', 'revel', 'download_progress', catalogId ) or false
				self._lastInternalProgressCount = KSTableUtils.derefOpt( userDoc, '_localOnly', 'download_progress', catalogId ) or 0

				return 'data', self._lastExternalProgressCount, self._lastInternalProgressCount

			end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Get the current pending upload count
	\param catalogId (string) The id of the catalog for which to bookmark.
	\return <code> B<bookmark></code> A promise which resolves to the current pending upload count
]]--

function userMethods:getPendingUploadCountB( catalogId )
	
	WFArguments.check( 'OzUser', 'string' )
	
	return self._userInfoCache:fieldValueB( 'user', self._ozUserId, '_localOnly/pending_upload_count/' .. catalogId )
	:mapB( function( count )	
		return count or 0
	end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Reset the current pending upload count.
	\param catalogId (string) The id of the catalog for which to bookmark.
	\return <code> B<bookmark></code> A promise which resolves to true when reset.
]]--

function userMethods:resetPendingUploadCountP( catalogId )
	
	WFArguments.check( 'OzUser', 'string', '?number' )
	
	userLog:debug( "addToPendingUploadCountP" )

	local preP = self._pendingUploadP or KSPromises.resolvedPromiseP( 'data', true )
	self._pendingUploadP = preP:mapP( function()
		return ( self._userInfoCache:applyOverlayToDocumentP {
			docType = 'user',
			docId = self._ozUserId,
			overlay = {
				_localOnly = {
					pending_upload_count = {
						[ catalogId ] = 0,
					}
				}
			}
		} )
	end )
	
	return self._pendingUploadP:runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief Increment our count of how many documents need to be uploaded in the
	current batch of documents.  The batch is reset to zero whenever we complete
	all documents in a batch.  This is used to give a sense of upload progress
	which can complete and then restart when the next wave of documents is received.
	\param catalogId (string) The id of the catalog for which to bookmark.
	\param numdocs (optional number) How many new documents there are.
	\return <code> B<bookmark></code> A promise which resolves to the current upload count
	once this addition has been included.
]]--

function userMethods:addToPendingUploadCountP( catalogId, numDocs )
	
	WFArguments.check( 'OzUser', 'string', '?number' )
	numDocs = numDocs or 1
	
	userLog:debug( "addToPendingUploadCountP numDocs = %d", numDocs )

	local preP = self._pendingUploadP or KSPromises.resolvedPromiseP( 'data', true )
	
	local function updateCountP( count )
		local overlay = {
			_localOnly = {
				pending_upload_count = {
					[ catalogId ] = count,
				}
			}
		}
		return ( self._userInfoCache:applyOverlayToDocumentP {
			docType = 'user',
			docId = self._ozUserId,
			overlay = overlay
		} )
		:mapDataP( function()
			return 'data', count
		end )

	end

	self._pendingUploadP = preP:mapP( function( status, countOrTrue )
		if status == 'data' and type( countOrTrue ) == 'number' then
			return updateCountP( countOrTrue + numDocs )
		else
			-- If previous was an error (or nil because we somehow
			-- got canceled, ignore that and push forward
			-- with a fresh set of updates
			return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
			:mapDataP( function( status, userDoc )	
				
				local count = KSTableUtils.derefOpt( userDoc, '_localOnly', 'pending_upload_count', catalogId ) or 0
				return updateCountP( count + numDocs )
				
			end )
		end
	end )
	
	return self._pendingUploadP:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark device tracking

--------------------------------------------------------------------------------
    -- FIXME [allany 2012-02-28]: It looks like Oz might have problems trying
    -- to write to a nested table within our root Revel settings table.
    -- I will work to resolve that with Dan Miley. For now, we write the
    -- four possible device class timestamp settings into the Revel root.
--[[!
	\brief  Retrieves (as a Promise) the most-recently-touched timestamp
	for devices used by this user, if any. There are four device
	classes, identified by keys "computer", "phone", "tablet", "web".

     	\return <code>P< nil | table ></code> A Promise that resolves to
		a table with one to four device class keys, where the value of
		each key is the timestamp; or nil if no devices are in use.
]]--

function userMethods:deviceClassesUsedP()

    return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
        :mapDataP( function( _, docValue )

            local result = KSTableUtils.derefOpt( docValue, 'settings', 'revel' ) or {}

            return 'data', {
				computer = result.timestamp_computer,
				phone = result.timestamp_phone,
				table = result.timestamp_tablet,
				web = result.timestamp_web,
			}

    end )

end

--------------------------------------------------------------------------------
--[[!
    \brief Clears this user's device-class usage setting for a single device type.
 	For DEBUGGING only, removes the device class setting
	<code>{ revel = { timestamp_DEVICECLASS = TIMESTAMP } }</code> in the
    	account so testers can test workflow from start.
	(The caller is responsible for clearing any \c KSPref associated with the device.)

    \param deviceClassName The device, one of: "computer", "phone", "tablet", "web"
    	(The function prepends "timestamp_" to the name.)

    \return A Promise to signal completion, that can be safely ignored.
]]--

function userMethods:DEBUG_clearDeviceClassUsedP( --[[string]] deviceClassName )

    WFArguments.check( 'OzUser', 'string' )

    return self:setDeviceClassUsedP( deviceClassName, yajl.null )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the most-recent-touch value for a device class, indicating that this
    	user has used this device class. (The caller is responsible for also clearing any
   	 \c KSPref associated with this device.)

   	\param deviceClassName (string) The device class, one of:
    		<ul><li>"computer"</li><li>"phone"</li><li>"tablet"</li><li>"web"</li></ul>
     		The function prepends "timestamp_" to this name.

    	\param timestamp (optional, string) The time to be associated with the usage,
		in ISO 8601 format. Default is the current time.

	\return A Promise to signal completion, that can be safely ignored.
]]--

function userMethods:setDeviceClassUsedP( --[[string]] deviceClassName,
					--[[optional string]] timestamp )

	WFArguments.check( 'OzUser', 'string', 'userdata|string|nil' )

    -- If we figure out a nested table that doen't overwrite existing values,
    -- then revive the following two lines, instead of the line above:
    -- revel.device_classes = {}
    -- revel.device_classes[ prefixedDeviceClassName ] = timestamp

	local settingsData = KSJsonEncoder.encodeOzUploadJson
		{
			revel = {
				[ 'timestamp_' .. deviceClassName ] = timestamp
											  or KSDate.timeInIso8601Format( nil, false ),
			}
		}

	local requestUrl = string.format( '%s/v1.0/accounts/%s/settings',
									  OzCatalogUrls.baseServerUrl,
									  self._ozUserId )

	return self:putP( requestUrl, settingsData,
		{
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
			},
			requestMethod = 'POST',
		} )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief Simulates logout from the current authorization session.

]]--

function userMethods:simulateForcedLogout()

	return self._oAuth2Session:simulateForcedLogoutP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether this user has a current valid connection
	to the Oz server.
	
	\return <code>B< boolean ></code> A Behavior whose value is
		\c true while there is a valid connection, and becomes \c false
		if the OAuth2 refresh token for this user is ever rejected.
]]--

function userMethods:hasValidConnectionToOzB()

	return self._oAuth2Session:hasValidLoginB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP GET request, signing the request	with the
	current access token of the current user and handles refreshing the access
	token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch.
	\param optArgs (optional, table) Options to modify the request.
	\param optSequencing (optional, string) 'LIFO' or 'FIFO'. FIFO by default.

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server

	\sa WFHttpClient.getP
]]--

function userMethods:getP( --[[string]] url, --[[table]] optArgs, --[[string]] optSequencing )

	return self._oAuth2Session:getP( url, optArgs, optSequencing )

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP PUT request, signing the request with the
	current access token of the current user and handles refreshing the access
	token if needed. Returns the response via a promise.

	\param url (string) the URL to fetch
	\param data (\c KSData or table) data to send with the request
		(use a Lua table for form data)
	\param optArgs (optional, table) options to modify the request

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server
	
	\sa HttpClient.putP
]]--

function userMethods:putP( --[[string]] url, --[[KSData or table]] data, --[[table]] optArgs )

	return self._oAuth2Session:putP( url, data, optArgs )

end

--------------------------------------------------------------------------------
-- Legacy version of above that returns an event stream. Expect this to be
-- deprecated and removed sometime after 1.1.

--[[!
	\deprecated Use \c putP()	instead.
]]--
function userMethods:putE( url, data, optArgs )

	optArgs = optArgs or {}

	optArgs.returnHttpErrorsAsStatusCode = true

	local resultP = self:putP( url, data, optArgs )
	resultP:runUntilDone()
		-- GRUNGE [scouten 2011-08-17]: Apparently the old implementation of
		-- putE always kept requests alive until done and the unit test
		-- counted on that. Use the transition to promises to enforce better
		-- clarity about lifetimes.

	return KSReactivePromiseBridge.promiseToEvent( resultP )

end

--------------------------------------------------------------------------------
--[[!
	\brief Logs this user out of the current Oz session.
]]--

function userMethods:logoutP()

	-- About the { logout = true } option here: If we lack both access and refresh
	-- tokens, then we should just quietly say we're done rather than firing an error
	-- that we don't have this information. "The item could not be deleted because
	-- it could not be found." And all that.

	userLog:info( "attempting to log out user %s", self._ozUserId )

	-- Optimistic logout: Clear out current user ID now; hope we can also convince
	-- Oz to log out as well.

	-- KSPrefs.currentUserId = nil

	-- Delete the overlay for this user.
	-- Fix for WATSON #2941964 (Settings shows other users catalog names briefly on
	-- first launch) since we are using the same doc store for storing the current
	-- logged in user's info we need to clear it on logout.

	self._userInfoCache:setServerValueP( 'user', self._ozUserId, { _deleted = true } )
		:runUntilDone()

	return self._oAuth2Session:logoutP():mapP(function(status, data)
		debug.print('Now set the currentuserid to nil')
		KSPrefs.currentUserId = nil
		end
		)

end

--------------------------------------------------------------------------------

local clearLocalOverlayOpts = {
	clearLocalOverlay = true,
}

--------------------------------------------------------------------------------

local function refreshP_imp( self, ignoreCurrentUser )
	
	if self._refreshP and not ignoreCurrentUser and not self._refreshP:isCanceled() then
		self._wantsRefresh = true
		return self._refreshP
	end
	
	if self._previousRefreshTime
		and KSDate.timeNowForAnimation() < ( self._previousRefreshTime + 1 )
	then
	
		userLog:debug( "OzUser:refreshP called %.2f seconds ago, current call via %s",
					   KSDate.timeNowForAnimation() - self._previousRefreshTime,
					   ':traceback:' )
	
	end

	local accountUrl = string.format( '%s/v1.0/accounts/%s',
									  OzCatalogUrls.baseServerUrl,
									  self._ozUserId )

	-- If we are not the current user, then we have to go find the current user
	-- and ask that user to fetch this info. Current user might not be populated
	-- right away (ain't this fun?).
	
	local currentUserP
	
	if ignoreCurrentUser or isCurrentUserNow( self ) then

		currentUserP = KSPromises.resolvedPromiseP( 'data', self )

	else

		if not currentUserB then
			currentUserB = require 'OzAccountService'.currentUserB()
		end
		
		currentUserP = KSReactivePromiseBridge.behaviorToPromise(
			currentUserB:relayForAutoCloseB(), -- FIXME_AUTO_CLOSE: Do I need the relay here?
			function( maybeUser )
				return not not maybeUser
			end )
		
	end
	
	-- We used to use applyOverlay to store what is really server value.
	-- That was a mistake. Now we get to clean that up. Ugh.

	local resultP = currentUserP
		:mapDataP( function( _, currentUser )
			return currentUser:getP( accountUrl )
		end )
		:mapP( KSJsonDecoder.decodeWebSafeJson_stripNullP )

	resultP:whenResolvedDo( function( status, userInfo )

		if status == 'data' then
			
			userLog:debug( "OzUser:refreshP received new info %s", userInfo )

			local trial = KSTableUtils.deepCopy( userInfo.trial )
			if trial then
				trial._replace = true
			end

			if WFCloseSupport.hasBeenOrWillBeClosed( self._userInfoCache ) then
				return 'error', 'user info cache closed'
			end

			self._userInfoCache:setServerValueP( 'user', self._ozUserId, userInfo,
												 clearLocalOverlayOpts )
				:runUntilDone()
			
		elseif status == 'error' then

			debug.print( "refreshP got an error:", userInfo )

		end

	end ):runUntilDone()

	self._previousRefreshTime = KSDate.timeNowForAnimation()

	resultP = resultP:mapP( function( status, userInfo, headers )
			
		if status == 'error' and headers == nil then
		
			-- Use the old results; better than causing the app to become useless
			-- when offline.
			-- We do not fall back to this case in the event Oz sends us an actual
			-- error (hence the test for headers == nil).
			return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
				:mapP( function( status, userInfo, ignoredOverlay )
					return status, userInfo
				end )

		end
		
		WFTimer.afterDelayP( 5 )
			:whenResolvedDo( function()
				self._refreshP = false
				if self._wantsRefresh then
					self._wantsRefresh = false
					self:refreshP()
				end
			end )
			:attachTo( self, nil, '_resetRefreshPAction' )
		
	end ):runUntilDoneP()
	
	self._refreshP = resultP
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
 	\brief Requests an update from Oz for this user's information.
	Use after a significant event (such as a purchase) occurs.

	\return <code>P< table ></code> A Promise that resolves to
		the new user information packet when the reload has finished.
		Callers need not attach an action to the promise.
]]--

function userMethods:refreshP()

	return refreshP_imp( self, false )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Clears any cached data, so that the next asynchronous call that requires access to
	the user info results in an HTTP call to Oz to fetch the information.

	Useful primarily for unit tests that want to verify changes immediately.

	\return This ozUser object.
]]--

function userMethods:clearCaches()

	self._refreshP = false
	local resetRefreshPAction = rawget( self, '_resetRefreshPAction' )
	if resetRefreshPAction then
		resetRefreshPAction:close()
		rawset( self, '_resetRefreshPAction', nil )
	end
	return self

end

--------------------------------------------------------------------------------
--[[!
 	\brief For testing only: Requests an update from Oz for this user's information,
 	using self to issue the HTTP request regardless of current user ID.
 
 	\return <code>P< table ></code> A Promise that resolves to
		the new user information packet when the reload has finished.
		Callers need not attach an action to the promise.
]]--

function userMethods:refreshP_ignoreCurrentUser()

	return refreshP_imp( self, true )

end

--------------------------------------------------------------------------------
--[[!
 	\brief Submits a subscription receipt from iOS or Mac Store Kit. This
	ensures that the receipt is submitted to the Oz server at the first
	opportunity, given network and Oz availability.
	(The first opportunity could be in a future app session.)

	\param receiptData (string) The receipt data.
	
	\return <code>P< boolean ></code> A Promise that resolves to
		\c true to confirms that the receipt has been successfully
		placed in the doc store for future transmission.
		Callers need not attach an action to the promise.

]]--

function userMethods:submitAppleStoreReceiptP( --[[string]] receiptData )

	userLog:debug( "submitAppleStoreReceiptP receiptData = %s", receiptData )

	local receiptDocId = KSIdGenerator.uuidInLowercaseHex()

	return self._userInfoCache:applyOverlayToDocumentP {
			docType = 'receipt',
			docId = receiptDocId,
			overlay = {
				userId = self._ozUserId,
				transactionReceipt = receiptData,
			},
		}
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a Behavior that allows a UI to provide notification of the
	eventual outcome of any store receipt submitted via :submitAppleStoreReceipt.
	Receipts remain in this queue until cleared via :acknowledgeReceiptNotification.

	\return <code>B< list< receiptDocId, receiptStatus, ozErrorCode, externalStoreErrorCode ></code>
		A Behavior whose value is zero or more tuples with the following values:
		<ol><li>receipt doc ID: use to remove this receipt with :acknowledgeReceiptNotification.</li>
		    <li>status ("accepted" or "invalid")</li>
		    <li>Oz error code if invalid</li>
		    <li>external error code (i.e. Apple Store error code) if invalid</li>
		</ol>
]]--

function userMethods:pendingReceiptNotificationsB()

	return self._receiptNotificationQueueB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Allows you to remove receipt status notifications from
	:pendingReceiptNotificationsB once the user has been notified as appropriate.

	\param receiptDocId (string) The receipt doc ID.
 ]]--

function userMethods:acknowledgeReceiptNotification( --[[string]] receiptDocId )

	assert( type( receiptDocId ) == 'string' )

	return self._userInfoCache:applyOverlayToDocumentP
		{
			docType = 'receipt',
			docId = receiptDocId,
			overlay = {
				receiptStatusAcknowledgedByUI = true,
			},
		}

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves all subscription products that are available to this user.
 
	\param storeId (string) One of 'ios' or 'mac'.

	\return <code>P< table ></code> A Promise that resolves to the list of products.
]]--

function userMethods:listAllProductsP( --[[string]] storeId )

	assert( type( storeId ) == 'string' )

	local requestUrl = string.format( '%s/v1.0/products?store=%s',
									  OzCatalogUrls.baseServerUrl,
									  storeId )

	return self:getP( requestUrl )
		:mapDataP( function( status, data )

			return 'data', KSJsonDecoder.decodeWebSafeJson( data )

		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Reports (as a Promise) the store type that this user used to purchase
	  the subscription.

	The store type is the value of the Oz account's <code>current_subs.store</code field.

	\return <code>P< string | nil ></code> A Promise that resolves to the store type, or
	nil if there is no subscription or the store type is not available from Oz.
	Current known store values include "ios", "admin", or "adobe".
	
]]--

function userMethods:currentSubscriptionStoreTypeP()

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function( _, docValue )

			return 'data', KSTableUtils.derefOpt( docValue, 'current_subs',
											'store' )

		end )

end

--------------------------------------------------------------------------------

local function writeUpdatedMinimalCatalogInfoP( catalogId, catalogInfo )

	local id = catalogId:gsub( 'urn:carousel:catalog:', '' )

	local docStore = OzDocStore( id )
 	local docValueP = docStore:docValueOnceP( 'catalog', id )

	return docValueP:mapP(
		function( status, docValue )

			local catalogOwnerId = KSTableUtils.derefOpt( docValue, 'owner', 'id' )

			if not catalogOwnerId then

				local resultP = docStore:setServerValueP( 'catalog', id, catalogInfo )

				resultP:whenResolvedDo( function()

					docStore:close()

				end ):runUntilDone()

				return resultP:runUntilDoneP()

			else
				docStore:close()
			end

		end )

end

--------------------------------------------------------------------------------

local function writeMinimalCatalogInfoP( self, catalogId, catalogInfo )

	-- [scouten 2012-07-24, PR# 3224253]: The overwhelming majority of the time
	-- (>99%) we are storing information that is the same as what we already have.
	-- So we cache the catalog information in the user doc DB. If it mismatches
	-- there, then we will write through. Otherwise, we avoid the (considerable)
	-- overhead of opening that catalog's doc DB for what is effectively a very
	-- expensive no-op.

	local id = catalogId:gsub( 'urn:carousel:catalog:', '' )

	local cachedCatalogDocP = self._userInfoCache:docValueOnceP( 'catalog', id )

	local maybeWriteCatalogDocP = cachedCatalogDocP
		:mapP( function( status, existingDocValue )

			-- Do we have a match?

			if status == 'data' then

				-- Nil out OzDocStore signal noise.

				existingDocValue._changeCounter = nil
				existingDocValue._docId = nil
				existingDocValue._docType = nil
				existingDocValue._ozMetadataUploadStatus = nil

				local didMatch = KSTableUtils.deepEquals( existingDocValue, catalogInfo )

				userLog:debug( "writeMinimalCatalogInfoP: match = %s\n\nexisting = %s\n\n"
							   .. "new = %s\n\n",
							   didMatch, existingDocValue, catalogInfo )

				if didMatch then
					return 'data', true
				end

			else

				userLog:debug( "writeMinimalCatalogInfoP: error %s", existingDocValue )

			end

			-- No match (or error reading): Write new value through to both
			-- locations.

			return writeUpdatedMinimalCatalogInfoP( catalogId, catalogInfo )
				:mapDataP( function()

					return self._userInfoCache:setServerValueP( 'catalog', id, catalogInfo )

				end )

		end )

	return maybeWriteCatalogDocP:runUntilDoneP()

end

--------------------------------------------------------------------------------

local function composeCatalogList( self, data )

	local result = {}

	local origCatalogs = data.catalogs or {}
	local nCatalogs = #origCatalogs
	
	-- process the minimal catalog list serially, otherwise for large lists of catalogs
	-- too many threads get started and kill the system.
	
	local function writeCatalogInfo( origCatalog )
		-- Check if this catalog's doc store has data, if not, write the information
		-- received now to the doc store to prevent a delay in fetching this information
		-- when it is accessed.
			
		return writeMinimalCatalogInfoP( self, origCatalog.id, origCatalog )
	
	end
		
	KSPromises.allP( writeCatalogInfo, origCatalogs, 1 ):runUntilDone()
	
	for i = 1, nCatalogs do

		local origCatalog = origCatalogs[ i ]
		local relativeUrl = KSTableUtils.derefOpt( origCatalog, 'links', 'self', 'href' )

		-- Check if this catalog's doc store has data, if not, write the information
		-- received now to the doc store to prevent a delay in fetching this information
		-- when it is accessed.

		local sharingState = origCatalog.sharing and origCatalog.sharing.state

		result[ i ] = {

			id = origCatalog.id:match( "[0-9a-f]+$" ),
			name = origCatalog.name,

			url = relativeUrl and ( data.base .. relativeUrl ),

			-- FIXME [scouten 2011-04-14]: Need proper URL handling.
			-- Matt might change this to absolute URL someday...

			created = string_sub( origCatalog.created, 1, 10 ),

			-- Shortens the creation date to YYYY-MM-DD

			sharedWith = origCatalog.user_count,
			
			ownerId = origCatalog.owner and origCatalog.owner._id,

			acceptedInvite = sharingState == 'accepted',
			
			subtype = origCatalog.subtype,
			
			payload = origCatalog.payload,

		}
	end

	table_sort( result, function( a, b )

		local a_name = string_lower( a.name )
		local b_name = string_lower( b.name )

		if a_name == b_name then
			return a.url < b.url
		else
			return a_name < b_name
		end

	end )

	return result

end

--------------------------------------------------------------------------------

local function makeRemapDataAndWriteThroughCache( self, behaviorCache, userInfoKey )

	return function( status, originalData )

		if WFPlatformInfo.isTornado then

		if status == 'error' then
			if originalData == 503 then
				self._serviceAvailableB:set(false)
			end
		end
		
		self._serviceAvailableB:set(true)
		
		end

		if not originalData or KSData.length( originalData ) == 0 then
			return 'error', nil
		end

		local data = KSJsonDecoder.decodeWebSafeJson( originalData )

		catalogListLog:debug( "%s got raw JSON %s", userInfoKey, data )

		if not data then return 'error', nil end
			-- FIXME [scouten 2011-04-14]: Is this actually necessary?

		if WFCloseSupport.hasBeenOrWillBeClosed( self._userInfoCache ) then
			return 'error', 'user info cache closed'
		end

		local result = composeCatalogList( self, data )

		catalogListLog:debug( "%s generated result table %s", userInfoKey, result )

		-- Write this new data to local cache DB (if changed).

		local catalogsB = WFCloseSupport.fetchCached( behaviorCache, self._ozUserId )

		local existingCatalogsValue = catalogsB and catalogsB:valueNow()

		if not existingCatalogsValue
			or not KSTableUtils.deepEquals( existingCatalogsValue, result )
		then

			local dataCopyWithReplace = {
				_replace = true,
			}

			for k, v in pairs( result ) do
				dataCopyWithReplace[ k ] = v
			end

			self._userInfoCache:applyOverlayToDocumentP {
				docType = 'user',
				docId = self._ozUserId,
				overlay = {
					_localOnly = {
						[ userInfoKey ] = dataCopyWithReplace,
					},
				},
			}

		end
		
		if result then
			result._replace = nil
		end

		return 'data', result

	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) all catalogs this user has created.
 
 	\return <code>B< nil | array of strings ></code> A Behavior whose
		value is initially nil, then the list of catalog IDs.
		The list is updated as catalogs are added or removed.
]]--

function userMethods:catalogsB()

	-- Force a reload from Oz (but not more than one per second or so)
	-- whenever somebody requests this behavior.

	self._rawReloadCatalogListNowE:post( true )

	-- Build the behavior if we don't have it already.

	local resultB = WFCloseSupport.fetchCached( catalogsBforUserId, self._ozUserId )

	if not resultB then

		resultB = self._userInfoCache:fieldValueB( 'user', self._ozUserId, '_localOnly/catalogInfo' )
			:mapB( function( catalogsList )
			
				if catalogsList then
					catalogsList._replace = nil
				end
				
				return catalogsList
			
			end )

		catalogsBforUserId[ self._ozUserId ] = resultB

	end

	return resultB:relayB()

end

--------------------------------------------------------------------------------
--[[
	B< boolean > = ozUser:forceUpgradeB()

	Return a boolean value indicating if the server is forcing an app version 
	upgrade.
	
]]--

function userMethods:forceUpgradeB()

	local forceUpgradeB = WFCloseSupport.fetchCached( forceUpgradeBforUserId, self._ozUserId )

	if not forceUpgradeB then
		forceUpgradeB = KSReactive.variableB( false )
		forceUpgradeBforUserId[ self._ozUserId ]  = forceUpgradeB
	end
	
	return forceUpgradeB:relayB()
	
end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrievss (as a Promise) a list of all catalogs to which this user has
	access.
	
 	\return <code>P< table ></code> A Promise that resolves to the list of
 		catalog URLs.
]]--

function userMethods:catalogsP()

	local resultP = self:getP( OzCatalogUrls.catalogsUrl )
	
	self:checkForForcedUpgrade(resultP)

	return resultP
	:mapDataP( makeRemapDataAndWriteThroughCache( self,
													  catalogsBforUserId,
													  'catalogInfo' ) )
	:runUntilDoneP()

end

--------------------------------------------------------------------------------

function userMethods:checkForForcedUpgrade( resultP )
    resultP:mapP( function( status, data, headers )
			
			if status == 'data' then
				-- No error, set forceUpgradeB to false
				local forceUpgradeB = WFCloseSupport.fetchCached( forceUpgradeBforUserId, self._ozUserId )

				if not forceUpgradeB then
					forceUpgradeB = KSReactive.variableB( false )
					forceUpgradeBforUserId[ self._ozUserId ]  = forceUpgradeB
				end

				forceUpgradeB:set( false )
			else 			
			
			if headers and type( headers ) == 'table' and headers.status_code == '403' then

				-- We have got a 403 from the server with error_code
				-- '4999' == force update

				local result, eData = pcall( KSJsonDecoder.decodeWebSafeJson, data )
	
				if result and eData then

					local error_code = tonumber( eData.code )

					if error_code == 4999 then
						local forceUpgradeB = WFCloseSupport.fetchCached( forceUpgradeBforUserId, self._ozUserId )

						if not forceUpgradeB then
							forceUpgradeB = KSReactive.variableB( false )
							forceUpgradeBforUserId[ self._ozUserId ]  = forceUpgradeB
						end
						forceUpgradeB:set( true )
					end

				end
				
			end
			
		end
	end ):runUntilDone()
end

--------------------------------------------------------------------------------
-- Request a reload of the shared catalogs. Currently called from within
-- OzCatalogService:removeUserP().

--[[!
	\brief Used internally by OzCatalogService:removeUserP().
]]--

function userMethods:requestSharedCatalogReload()

	self._rawReloadSharedCatalogListNowE:post( true )

end

--------------------------------------------------------------------------------
--[[!
	\brief Requests a reload of catalogs owned by this user.
]]--
 
function userMethods:requestCatalogReload()

	self._rawReloadCatalogListNowE:post( true )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) all catalogs that other users
	have shared with this user.
 
 	\return <code>B< nil | table ></code> A Behavior whose
		value is initially nil, then the list of catalogs.
		 Each entry in the list has the keys: 'id', 'name', 'url', 'created'
		The list is updated as catalogs are shared and unshared.

]]--

function userMethods:catalogsSharedWithUserB()

	-- Force a reload from Oz (but not more than one per second or so)
	-- whenever somebody requests this behavior.

	self._rawReloadSharedCatalogListNowE:post( true )

	-- Build the behavior if we don't have it already.

	local resultB = WFCloseSupport.fetchCached( sharedCatalogsBforUserId, self._ozUserId )

	if not resultB then

		resultB = self._userInfoCache:fieldValueB( 'user',
												   self._ozUserId,
												   '_localOnly/sharedCatalogInfo' )
			:mapB( function( catalogsList )
			
				if catalogsList then
					catalogsList._replace = nil
				end
				
				return catalogsList
			
			end )

		sharedCatalogsBforUserId[ self._ozUserId ] = resultB

	end

	return resultB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) all catalogs that other users
	have shared with this user.
 
 	\return <code>P< table ></code> A Promise that resolves to
		the list of catalogs. Each entry in the list has the keys:
		'id', 'name', 'url', 'created'
]]--

function userMethods:sharedCatalogsP()

	return self:getP( string.format( '%s?role=user', OzCatalogUrls.catalogsUrl ) )
		:mapDataP( makeRemapDataAndWriteThroughCache( self,
													  sharedCatalogsBforUserId,
													  'sharedCatalogInfo' ) )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Accepts a catalog invitation.

	\param catalogId (string) The catalog's unique ID.

	\return <code>P< true | error ></code> A Promise resolves as true if accepted;
	resolves as error otherwise.
]]--

function userMethods:acceptCatalogInviteP( --[[string]] catalogId )

	if type( catalogId ) ~= 'string' and #catalogId ~= 32 then
		error( "OzCatalogList.acceptCatalogInviteP: catalogId must be a UUID string", 2 )
	end

	local catalogUrl = string.format( '%s/%s', OzCatalogUrls.catalogsUrl, catalogId )

	local sharingStateJsonData = KSJsonEncoder.encodeOzUploadJson
		{ sharing = { state = 'accepted' } }

	local putArgs = {
		requestMethod = 'POST',
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	return self:putP( catalogUrl, sharingStateJsonData, putArgs )
		:mapDataP( function( _, data, headers )

			-- Per Mayank, we expect a 204 No Content response when the invite
			-- acceptance is recorded by Oz.

			if headers and headers.status_code == '204' then
				return 'data', true
			else
				return 'error', data, headers
			end

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new catalog on the Oz server.

	\param args (table) The description of the new catalog, a table with
		these keys:
	       <ul>
		<li>id (optional, string): Value is the UUID (in lowercase hex format) of the new
			catalog; if missing, one is created,</li>

		<li>name (required, non-empty string): Value is the user-visible name for the new
			catalog.</li>

		<li>users (optional, table): Value is the list of user emails to be invited
			to the catalog. </li>
		</ul>

	\return A Promise. On success, resolves to:
		'data', catalogUrl, headers, catalogId, didUserInvitationSucceed (boolean)
		<ul><li>If catalog creation succeeded, but only the user invitation failed, the promise returns:
		'data', catalogUrl, headers, catalogId, false</li>
		<li>If the catalog creation completely failed, the promise resolves in error.</li></ul>

	Any catalog list behaviors are reloaded upon resolution of the promise.
	\sa :catalogB
]]--

function userMethods:createCatalogP( args )

	if type( args ) ~= 'table' then
		error( "OzCatalogList.createCatalog: args must be a table", 2 )
	end

	if type( args.name ) ~= 'string' or #args.name == 0 then
		error( "OzCatalogList.createCatalog: name must be a non-empty string", 2 )
	end

	if args.id and type( args.id ) ~= 'string' and #args.id ~= 32 then
		error( "OzCatalogList.createCatalog: id, if present, must be a UUID string", 2 )
	end

	-- Process users to be invited.

	if args.users then
		for i, v in ipairs( args.users ) do
			args.users[ i ] = "+mailto:" .. args.users[ i ]
		end

	end

	local catalogId = args.id or KSIdGenerator.uuidInLowercaseHex()

	local newCatalogUrl = string.format( '%s/%s',
										 OzCatalogUrls.catalogsUrl,
										 catalogId )

	local newCatalogData
	
	local catalogData = { name = args.name }
		
	if args.subtype then
		catalogData[ 'subtype' ] = args.subtype
	end
		
	if args.payload then
		catalogData[ 'payload' ] = args.payload
	end
		
	newCatalogData = KSJsonEncoder.encodeOzUploadJson( catalogData )

	local putArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	return self:putP( newCatalogUrl, newCatalogData, putArgs )
		:mapP( function( status, data, headers )

			if status == 'error' then

				if headers
					and headers[ 'Content-Type' ] == 'application/json;charset=utf-8'
					and data
				then

					local result, eData = pcall( KSJsonDecoder.decodeWebSafeJson, data )
					return status, ( result and eData and eData.description )
								   or "unknown error"

				else
					return status, data, headers
				end

			end

			-- The only response we treat as success is 201 Created
			-- with a Location: headers. Anything else is an error.

			local location = headers
				and headers.status_code == '201'
				and headers.Location

			if location then

				self._rawReloadCatalogListNowE:post( true )

				local OzCatalogService = require 'OzCatalogService'

				local tempCatalog = OzCatalogService( catalogId, self )

				local ownerId = self._ozUserId

				-- As a fix for WATSON #2989758 (Spurious conflict reported in debug
				-- log for catalog owner ID) we check if the user id has the urn prefix,
				-- if not we attach it to avoid spurious conflicts when the sync agent
				-- for this catalog fetches the ownerId from Oz and writes it to the doc
				-- store.

				if not string.match( ownerId, 'urn:carousel:account:' ) then
					ownerId = 'urn:carousel:account:' .. ownerId
				end

				-- As a fix for #3049221, we only write the owner status to the doc
				-- store if it is not "created". This is to avoid spurious conflicts
				-- in the case when a new user is created and the default catalog is
				-- created for this new user since the server value returned by Oz
				-- is "trial." This can lead to import being disallowed for this catalog
				-- because the catalog docStore says the owner status is "created."

				-- #3176709	- Write the account status after making sure that the
				-- user info we have is refreshed from Oz. This is to make sure we
				-- don't end up with spurious conflicts if the account status has
				-- changed on Oz in the meantime.

				return self:refreshP()

					:mapDataP( function( status, userInfo )

						local ownerOverlay = {}
						ownerOverlay.id = ownerId
						local ownerStatus = userInfo.status

						if ownerStatus ~= 'created' then
							ownerOverlay.status = ownerStatus
						end

						return tempCatalog._docStore:applyOverlayToDocumentP {
							docType = 'catalog',
							docId = catalogId,
							overlay = {
								owner = ownerOverlay,
							},
						}

					end )

					:mapDataP( function( status, data )
	
						tempCatalog:close()

						-- Send a request to invite users.

						if args.users ~= nil then
							local usersData = KSJsonEncoder.encodeOzUploadJson
								{ users = args.users  }
	
							local postArgs = {
								requestHeaders = {
									{ header = 'Content-Type', value = 'application/json' },
								},
								requestMethod = 'POST',
							}
	
							return self:putP( newCatalogUrl, usersData, postArgs )
								:mapP( function ( status, data, inviteHeaders )
	
									if status == 'error' then
										return 'data', location, headers, catalogId, false
									else
										return 'data', location, headers, catalogId, true
									end
								end )
	
						else

							return 'data', location, headers, catalogId, true
	
						end
	
					end )

				:runUntilDoneP()

			else

				return 'error', false

			end

		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a catalog from the Oz server.

 	\param catalogId (string) The catalog's unique ID.
 	\param optCatalog (optional, OzCatalogService) An OzCatalogService instance
 		for the catalog to be deleted.Provide this if you need
		notification (in the form of canAccessB firing a false value) once the catalog
		deletion is complete.If provided, this is used for deleting local assets.
		If not, a temporaray instance of that catalog is created.

	\return A Promise that resolves to the result of the deletion operation.
		Any catalog list behaviors are reloaded upon resolution of the promise.
		\sa :catalogB
]]--

function userMethods:deleteCatalogP( catalogId, optCatalog )

	if type( catalogId ) ~= 'string' and #catalogId ~= 32 then
		error( "OzCatalogList.deleteCatalog: catalogId, if present, must be a UUID string",
			   2 )
	end

	if optCatalog then

		local OzCatalogService = require 'OzCatalogService'

		if not OzCatalogService:isClassOf( optCatalog )
			or optCatalog:getCatalogId() ~= catalogId
		then
			error( "OzUser.deleteCatalog: catalog instance, if present, must be an "
				   .. "instance of OzCatalogService for the same catalog", 2 )
		end

	end

	local deadCatalogUrl = string.format( '%s/%s',
										  OzCatalogUrls.catalogsUrl,
										  catalogId )

	return self:getP( deadCatalogUrl, { requestMethod = 'DELETE' } )
		:mapP( function( status, data, headers )
	
			-- The only response that is considered successful is
			-- 204 No Content. Anything else translates to an error.
	
			if headers and headers.status_code == '204' then
	
				self._rawReloadCatalogListNowE:post( true )
				self:deleteLocalCatalogAndAssetsP( catalogId, optCatalog ):runUntilDone()
	
				return 'data', true, headers
	
			else
	
				return 'error', headers and headers.status_code or data
	
			end
	
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts a trial for this account.

	\return <code>P< boolean ></code> A Promise that resolves to
		\c true if the request succeeded.
]]--

function userMethods:startTrialP()

	-- curl -X PUT -H "Authorization: Bearer $access_token"
		-- -H Content-Type:application/json --data-binary '{}'
		-- -v $API_ENDPOINT/v1.0/accounts/$account_id/trial

	local startTrialUrl = string.format( '%s/v1.0/accounts/%s/trial',
										 OzCatalogUrls.baseServerUrl,
										 self._ozUserId )

	local putArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
		requestMethod = 'PUT',
	}

	local jsonData = KSJsonEncoder.encodeOzUploadJson { empty = true }

	return self:putP( startTrialUrl, jsonData, putArgs )
		:mapDataP( function( _, data, headers )

			-- A 201 response means the trial was started.

			if headers and headers.status_code == '201' then
				return 'data', true
			else
				return 'error', data, headers
			end

		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes this account.

	\return <code>P< boolean ></code> A Promise that resolves to
		\c true if the request succeeded.
]]--


function userMethods:deleteAccountP()

	-- curl -X DELETE -H "Authorization: Bearer $access_token" -v
		-- $PRTCL://$HOST/v1.0/accounts/$account_id

	local deleteUrl = string.format( '%s/v1.0/accounts/%s',
									 OzCatalogUrls.baseAdminServerUrl,
									 self._ozUserId )

	return self:getP( deleteUrl, { requestMethod = 'DELETE' } )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

function userMethods:updateEntitlementE()

	-- FIXME [scouten 2012-10-23]: This should be a promise-based API.

	local KSDate = require 'KSDate'

	local resultE = KSReactive.sourceE()

	-- First get the user's account status and figure out if we can expire the account.
	
	local statusRequestB = self:accountStatusB()

	statusRequestB:onDataB( function( status )

		if not status then return end

		statusRequestB:close()

		local param

		if status == 'trial' then
			param = 'trial_end'
		elseif status == 'trial_grace_period' then
			param = 'trial_grace_end'
		elseif status == 'subscriber' then
			param = 'subs_expiry'
		elseif status == 'subscription_grace_period' then
			param = 'subs_grace_end'
		else
			local message = string.format(
				"Account status is '%s', can't expire this account", status )
			debug.print( message )
			return
		end

		-- Next set the expiration timestamp for the current account state

		-- curl -X POST -H Content-Type:application/json -H "Authorization: Bearer $access_token" -v https://$HOST/v1.0/accounts/$account_id?subs_expiry=1950-01-01T23:45:59.858242Z
		-- curl -X POST -H Content-Type:application/json -H "Authorization: Bearer $access_token" -v https://$HOST/v1.0/accounts/$account_id?subs_grace_end=1955-01-01T23:45:59.858242Z
		-- curl -X POST -H Content-Type:application/json -H "Authorization: Bearer $access_token" -v https://$HOST/v1.0/accounts/$account_id?trial_end=1960-01-01T23:45:59.858242Z
		-- curl -X POST -H Content-Type:application/json -H "Authorization: Bearer $access_token" -v https://$HOST/v1.0/accounts/$account_id?trial_grace_end=1965-01-01T23:45:59.858242Z

		local yesterday = KSDate.addDaysInIso8601Format( nil, false, -1 )

		local setTimeUrl = string.format( '%s/v1.0/accounts/%s?%s=%s',
										  OzCatalogUrls.baseAdminServerUrl,
										  self._ozUserId,
										  param,
										  yesterday )

		debug.print( "setTimeUrl:", setTimeUrl )

		local putArgs = {
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
				{ header = 'Accept', value = 'application/json' },
			},
			requestMethod = 'POST',
		}

		local setTimeRequestE = self:putE( setTimeUrl, '', putArgs )

		setTimeRequestE:onErrorE(
			function( ... ) resultE:postError( ... ) resultE:close() end )

		setTimeRequestE:onDataE( function()

			-- Finally force the account status to be updated.

			-- curl -X POST -H Content-Type:application/json -H "Authorization: Bearer $access_token" -v https://$HOST/v1.0/accounts/$account_id?update_entitlement

			local updateUrl = string.format( '%s/v1.0/accounts/%s?update_entitlement',
											 OzCatalogUrls.baseAdminServerUrl,
											 self._ozUserId )

			debug.print( "updateUrl:", updateUrl )

			local updateRequestE = self:putE( updateUrl, "", putArgs )
			updateRequestE:onDataE( function( ... ) resultE:post( ... ) end )
			updateRequestE:onErrorE( function( ... ) resultE:postError( ... ) end )
			updateRequestE:onDoneE( function( ... ) resultE:close() end )

		end )

	end )

	return resultE:relayE()

end

--------------------------------------------------------------------------------
--[[!
 	\brief Reports (as a Promise) whether a catalog exists on the Oz server.
 	
	\param catalogId (string) The catalog's unique ID.

	\return <code>P< boolean, string ></code> A Promise that resolves to
		\c true if the catalog does exist, or \c false if the catalog does not exist,
		with the message 'doesNotExist'. If Oz returns an error with a status code other
		than '404' and we are unable to determine existence, the message string is 'error'.
 ]]--

function userMethods:canAccessP( --[[string]] catalogId )

	local catalogUrl = string.format( '%s/%s',
									  OzCatalogUrls.catalogsUrl,
									  catalogId )

	local catalogP = self:getP( catalogUrl )

	local rawCatalogP = catalogP:mapP( function( status, data, headers )

		if status == 'data' then
			return 'data', 'noError'

		elseif headers and headers.status_code == '404' then
			return 'data', false, 'doesNotExist'

		elseif headers and headers.status_code == '403' then

			-- We have got a 403 from the server with error_code
			-- '4999' == force update
			-- '4300' == no longer collaborator
			-- '4301' == owner's status expired

			local result, eData = pcall( KSJsonDecoder.decodeWebSafeJson, data )

			if result and eData then

				local error_code = tonumber( eData.code )

				if error_code == 4999 then
					return 'data', false, 'forceUpdate'

				elseif error_code == 4300 then
					return 'data', false, 'noAccess'

				else
					return 'data', error_code
				end

			end

		else
			return 'error', "userMethods:canAccessP has failed with "
							.. tostring( headers and headers.status_code or data )
		end

	end )

	return rawCatalogP:mapP( function( status, data, msg )

		-- If the catalog does not exist or the request to the server returned in error,
		-- return error.

		if ( status == 'data' and data == false ) or status == 'error' then
			return status, data, msg
		end

		local accessError = data

		-- If we didn't get the error, all is good and we can allow access.

		if accessError == 4301 then

			local downloadOriginals = KSPrefs.DownloadOriginals

			if downloadOriginals then
				return 'data', true
			else
				return 'data', false, 'noAccess'
			end

		else
			return 'data', true
		end

	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) all users who have shared a catalog.

	\param catalogId (string) The catalog's unique ID.
 
 	\return <code>P< table ></code> A Promise that resolves to
		the list of user IDs.
]]--


function userMethods:catalogUsersP( --[[string]] catalogId )

	local catalogUsersUrl = string.format( '%s/%s/users',
					   OzCatalogUrls.catalogsUrl,
					   catalogId )

	return self:getP( catalogUsersUrl )
		:mapP( KSJsonDecoder.decodeWebSafeJson_stripNullP )

end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a catalog and its contained assets from the local cache.

 	\param catalogId (string) The catalog's unique ID.
 	\param optCatalog (optional, OzCatalogService) An OzCatalogService instance
 		for the catalog to be deleted.Provide this if you need
		notification (in the form of :canAccessB firing a false value) once the catalog
		deletion is complete.If provided, this is used for deleting local assets.
		If not, a temporaray instance of that catalog is created.

	\return A Promise that resolves to the result of the deletion operation.
 ]]--


function userMethods:deleteLocalCatalogAndAssetsP( catalogId, optCatalog )

	local shouldClose = not optCatalog
	
	if not optCatalog then
		optCatalog = require 'OzCatalogService'( catalogId, self )
	end

	local deleteP = optCatalog:deleteLocalCatalogAndAssetsP()
	
	if shouldClose then
		deleteP:whenResolvedDo( function()
			optCatalog:close()
		end ):runUntilDone()
	end
	
	return deleteP:runUntilDoneP()

end

--------------------------------------------------------------------------------
if WFPlatformInfo.isTornado then

function userMethods:setCatalogUpdateTime( updateTime )
	self._catalogUpateTime = updateTime
end

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts periodically refresh this user's catalog and user account details
	from the Oz server.
	
	Call only from one main thread (such as KSLVC or BABLU).
	Monitoring stops when this user object is closed or the user is logged out.

	\param optDelay (optional, number) The number of seconds to wait before updating (default = 60).
]]--


--function userMethods:startWatchingForCatalogUpdates( --[[number]] optDelay )

-- TORNADO HACK 
-- add cancel existing timer param
function userMethods:startWatchingForCatalogUpdates( --[[optional, number]] optDelay, --[[ boolean ]] cancelExistingTimer )

	userLog:info( "startWatchingForCatalogUpdates called for user %s with delay %.2f via %s\nin thread %s",
				  self._ozUserId, optDelay or -1, ':traceback:', ':thread:' )

	-- handle cancel existing timer
	if self._catalogUpdateTimerE and cancelExistingTimer then
		self._catalogUpdateTimerE:close()
		self._catalogUpdateTimerE = false
	else 
		if self._catalogUpdateTimerE then
			return
		end
	end

	self._catalogUpdateTimerE = KSReactive.timerE( optDelay or self._catalogUpateTime )

	self:addCloseTarget( self._catalogUpdateTimerE )

	self._catalogUpdateTimerE:onDataE( function()

		self._catalogUpdateTimerE = false

		if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
			return
		end

		-- Trigger manual reloads now.
		if isCurrentUserNow( self ) then
			self:requestSharedCatalogReload()
			self:requestCatalogReload()
		end
		self:refreshP()

		-- And do this all again a minute later.
-- TORNADO HACK 
-- TODO remove 
		debug.print( "updating catalogs" )

		self:startWatchingForCatalogUpdates()

	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief HTTP login call to IMS to get access token and refresh token for test
	cases.

	This is based on the following protocol:
	https://zerowing.corp.adobe.com/display/ims/IMS+API+-+login#IMSAPI-login-Usagefromanativeappwiththeauthorizationcodegranttypeenabled

	Intended for test cases where we need access token and refresh token
	and we want to bypass IMS login through webview/UI.

	\param imsLoginUrl (string) IMS login url
	\param postParams (string) Parameters to pass according to IMS documentation in above link, for example:
	 	client_id=RevelMacTest1&redirect_uri=air-auth://carousel&locale=en_US&scope=openid,AdobeID&
		client_secret=12d5de23-e638-41f5-815e-04a10a5e13e1&username=testUser&password=testPassword

	\return <code>P< string, string ></code> A Promise that resolves to the access token and refresh token.
]]--

function userMethods:getAccessTokenRefreshTokenFromImsP( --[[ string ]] imsLoginUrl,
							--[[ string ]] postParams )

	local postArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/x-www-form-urlencoded' },
			{ header = 'Accept', value = '*/*' },
		},
		requestMethod = 'POST',
	}

	return WFHttpClient.putP( imsLoginUrl, postParams, postArgs )
		:mapDataP( function( status, data, headers )

			if headers.status_code ~= "200" then

				local responseString = tostring( data )
				return 'error', responseString

			else

				local decodedJson = KSJsonDecoder.decodeWebSafeJson_stripNull( data )
				return 'data', decodedJson.access_token, decodedJson.refresh_token

			end

		end )

end

--------------------------------------------------------------------------------

-- Method internal to this file, for deleting any device tokens understood to be registered
-- with Oz.
-- compareTokenUrl if passed in will be compared with the device token url locally saved
-- and the deletion will not take place if the two are the same.

local function deleteDeviceTokenP( self, compareTokenUrl )

	return self._userInfoCache:docValueOnceP( 'user', self._ozUserId )
		:mapDataP( function ( status, docValue )
					
			local deleteTokenUrl = KSTableUtils.derefOpt( docValue, '_localOnly', 'device_token' )

			if deleteTokenUrl then
			
				if compareTokenUrl and compareTokenUrl == deleteTokenUrl then
					return 'data', true
				end
		
				local deleteArgs = {
					requestMethod = 'DELETE',
					returnHttpErrorsAsData = true,
					requestHeaders = {
						{ header = 'Content-Type', value = 'application/json' },
					},
				}
				
				return self:getP( deleteTokenUrl, deleteArgs )
					:mapP( function( status, data, headers )
						
						local status_code = KSTableUtils.derefOpt( headers, 'status_code' )
						
						-- If the status_code returned by Oz is '204' or '404' we interpret it as a success.
						if not ( status_code == "204" or status_code == "404" ) then
							
							if status == 'error' and data == 'No refresh/access token stored in keychain' then
								return 'error', 'no_token'
							end
							
							userLog:error( "deleteDeviceToken: status = %s, data = %s, headers = %s", 
											status or '(nil)', data or '(nil)', headers or '(nil)' ) 
							
							local responseString = 'deleteDeviceToken: error deleting token from Oz status_code: ' .. (status_code or '(nil)')
							return 'error', responseString
		
						else

							-- Delete the token location locally.

							return self._userInfoCache:applyOverlayToDocumentP {
							docType = 'user',
							docId = self._ozUserId,
							overlay = {
								_localOnly = {
										device_token = { _delete = true },
									}
								}
							}:mapDataP( function ( _, _ )
								return 'data', true
							end )
							
						end

				end )
				
			else
				-- No token location, return true
				return 'data', true
			end
			
		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief API to update the device token for this account in Oz.

	Oz API documentation:
	http://ps-sdk.corp.adobe.com/oz_public/multi_page/index.html#resources_accounts

	\param app_id (string) The app bundle id.
	\param aps_env (string) The aps-environment entitlement value. 
	It should be either "development" or "production".
	\param token (string or KSData) The device Token value as received from 
	Apple.

	\return <code>P< string|boolean ></code> Promise that resolves to the value
	'true' if the request succeeds and the error string, if it fails.
]]--

function userMethods:registerDeviceTokenP( app_id, aps_env, token )

	WFArguments.check( 'OzUser', 'string', 'string' )
	-- WFArguments doesn't like checking for KSData, returns an 'expected KSData, got userdata'
	-- error, thus checking type via the assert.
	
	assert( KSData.isData( token) or type(token) == 'string', "OzUser:registerDeviceTokenP token should be a KSData or string value" )
	
	local postArgs = {
		requestMethod = 'POST',
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	local registerTokenUrl = string.format( '%s/v1.0/accounts/%s/devices',
									  OzCatalogUrls.baseServerUrl,
									  self._ozUserId )

	-- Convert the token to hex
	
	local hexToken = KSStringUtils.convertToHex( token )
	
	local postParams = KSJsonEncoder.encodeOzUploadJson
		{
			device_token = hexToken,
			app_id = app_id,
			aps_env = aps_env,
		}
				
	return self:putP( registerTokenUrl, postParams, postArgs )
		:mapDataP( function( status, data, headers )
			
			if headers.status_code ~= "201" or not headers.Location then

				local responseString = tostring( data )
				return 'error', responseString

			else
				-- Delete any old token first, if it's NOT the same as the
				-- one we just registered. This prevents us deleting the token
				-- in the case that we try to register the same token twice.
				return deleteDeviceTokenP( self, headers.Location )
					:mapDataP( function ( _,_ )
				
						-- Then save the location of new token
						return self._userInfoCache:applyOverlayToDocumentP {
						docType = 'user',
						docId = self._ozUserId,
						overlay = {
							_localOnly = {
									device_token = headers.Location,
								}
							}
						}:mapDataP( function ( _,_ )
							return 'data', true
						end )
					
				end )

			end

		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief API to delete the device token for this account from Oz.

	Oz API documentation:
	http://ps-sdk.corp.adobe.com/oz_public/multi_page/index.html#resources_accounts

	\return <code>P< string|boolean ></code> Promise that resolves to the value
	'true' if the request succeeds and the error string, if it fails.
]]--

function userMethods:deleteDeviceTokenP()

	WFArguments.check( 'OzUser' )
	
	return deleteDeviceTokenP( self, nil )
	
end

--------------------------------------------------------------------------------

local OzUser_class = KSClass 'OzUser'( userMethods )


--============================================================================--

-- @mark OzUser namespace

local OzUser = {}

--------------------------------------------------------------------------------

local userObjectsById = KSTableUtils.makeFullyWeak()

function OzUser.make( userId )

	local result = WFCloseSupport.fetchCached( userObjectsById, userId )

	if not result then

		result = OzUser_class( userId )
		userObjectsById[ userId ] = result

	end

	return result

end

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether a UUID identifies a valid Oz user.

	\param maybeOzUser (string) The ID to check. ((is this right? or does it check if an object is of this class?))

	\return (boolean) \c true if the ID is valid.
]]--

function OzUser.isOzUser( --[[string]] maybeOzUser )

	return OzUser_class:isClassOf( maybeOzUser )

end

--------------------------------------------------------------------------------

return OzUser
