--[[----------------------------------------------------------------------------

OzUserDeviceTokenTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFAsyncTestCase = require 'WFAsyncTestCase'
local KSPromises = require 'KSPromises'
local KSData = require 'KSData'
local OzUserTestUtils = require 'OzUserTestUtils'

local OzUser
local T
local putData
local headers

local TestSuite = {}

--==============================================================================

local function buildMocks()
	
	local function deliverHttpResponse( url )
		local response = headers[ url ]
		if ( not response ) or response.status_code then
			response = { 'data', true, response }
		end
		return KSPromises.resolvedPromiseP( unpack( response, 1, response.n or 3 ) )
	end
	
	return
	{
		OzUserTestUtils.buildMocks(),
		
		KSOAuth2Session = function( userId )
			return {
				getP = function( self, url, optArgs )
					return deliverHttpResponse( url )
				end,
				putP = function( self, url, data, optArgs )
					putData[ url ] = data
					return deliverHttpResponse( url )
				end,
			}
		end
	}

end

--==============================================================================
-- @mark setup
--==============================================================================

local kUserId = string.rep( 'a', 32 )

function TestSuite.setUpClass()

	OzUserTestUtils.setUp( buildMocks() )
	
	OzUser = require 'OzUser'

end

--------------------------------------------------------------------------------

function TestSuite.tearDownClass()

	OzUserTestUtils.tearDown( buildMocks() )
	
end

--------------------------------------------------------------------------------

function TestSuite.test_registerDeviceTokenP_error( test )

	T = {}
	putData = {}
	headers = {}
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ] = { 
		status_code = "201",
		Location = 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e737472696e67',
	}
	
	-- In this test we cause Oz / http to send back an error.
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e737472696e67' ]  = {
		'data',
		'offline for maintenance',
		{ status_code = "500" },
	}
	
	local user = OzUser.make( kUserId )

	WFAsyncTestCase( test )
		:waitFor( function()
			return user:registerDeviceTokenP( "app_id", "development", "thisisatokenstring" )
		end )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return putData[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]
		end )
		:continueWhenValueMatches( 'data', 	[[{"aps_env":"development","app_id":"app_id","device_token":"74686973697361746f6b656e737472696e67"}]] )
		:waitFor( function()
			-- Clean up the token just stored
			return user:deleteDeviceTokenP()
		end )
		:continueWhenValueMatches( 'error', 'offline for maintenance' )
		
end

--------------------------------------------------------------------------------

function TestSuite.test_registerDeviceTokenP_string( test )

	T = {}
	putData = {}
	headers = {}
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e737472696e67' ] = { status_code = '204' }
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]  = { 
		status_code = "201",
		Location = 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e737472696e67',
	}
	
	local user = OzUser.make( kUserId )

	WFAsyncTestCase( test )
		:waitFor( function()
			return user:registerDeviceTokenP( "app_id", "development", "thisisatokenstring" )
		end )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return putData[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]
		end )
		:continueWhenValueMatches( 'data', 	[[{"aps_env":"development","app_id":"app_id","device_token":"74686973697361746f6b656e737472696e67"}]] )
		:waitFor( function()
			-- Clean up the token just stored
			return user:deleteDeviceTokenP()
		end )
		:continueWhenValueMatches( 'data', true )
		
end

--------------------------------------------------------------------------------

function TestSuite.test_registerDeviceTokenP_KSData( test )

	T = {}
	putData = {}
	headers = {}
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e6b7364617461' ] = { status_code = '204' }
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]  = { 
		status_code = "201" ,
		Location = 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e6b7364617461',
	}

	local user = OzUser.make( kUserId )

	WFAsyncTestCase( test )
		:waitFor( function()
			return user:registerDeviceTokenP( "app_id", "production", KSData.make( "thisisatokenksdata" ) )
		end )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return putData[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]
		end )
		:continueWhenValueMatches( 'data', [[{"aps_env":"production","app_id":"app_id","device_token":"74686973697361746f6b656e6b7364617461"}]] )
		:waitFor( function()
			-- Clean up the token just stored
			return user:deleteDeviceTokenP()
		end )
		:continueWhenValueMatches( 'data', true )
		
end

--------------------------------------------------------------------------------

function TestSuite.test_deleteDeviceTokenP_existingToken( test )

	T = {}
	putData = {}
	headers = {}
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e6b7364617461' ]  = { status_code = "204" }
	headers[ 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices' ]  = { 
		status_code = "201" ,
		Location = 'https://api.oznext.com/v1.0/accounts/' .. kUserId .. '/devices/74686973697361746f6b656e6b7364617461',
	}
	local user = OzUser.make( kUserId )

	WFAsyncTestCase( test )
		:waitFor( function()
			return user:registerDeviceTokenP( "app_id", "production", KSData.make( "thisisatokenksdata" ) )
		end )
		:continueWhenValueMatches( 'data', true )
		:waitFor( function()
			return user:deleteDeviceTokenP()
		end )
		:continueWhenValueMatches( 'data', true )
	
end

--------------------------------------------------------------------------------

function TestSuite.test_deleteDeviceTokenP_noToken( test )
	
	T = {}
	putData = {}
	headers = {}
		
	local user = OzUser.make( kUserId )

	WFAsyncTestCase( test )
		:waitFor( function()
			return user:deleteDeviceTokenP()
		end )
		:continueWhenValueMatches( 'data', true )
	
end

--------------------------------------------------------------------------------

return TestSuite