--[[----------------------------------------------------------------------------

PERF_OzDocStoreTest.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

Performance-specific tests for OzDocStore. Normally disabled.

--------------------------------------------------------------------------------

@author scouten 2011-03-01

------------------------------------------------------------------------------]]

	-- core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- wichitafoundation.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- core.utils
local KSTableUtils = require 'KSTableUtils'
local KSIdGenerator = require 'KSIdGenerator'

	-- oz.doc_store
local OzDocStore = require 'OzDocStore'

local KSDate = require 'KSDate'

local WFTestMetrics = require 'WFTestMetrics'

local KSPromises = require 'KSPromises'

--============================================================================--

local PERF_OzDocStoreTest = {}

local docStore, failBecause

local randomDocId = '9b572d0e423c4654b078893ddd41c230'

local testMetrics = nil

--------------------------------------------------------------------------------

local startTimes = {}
local executionTimes = {}

local function startClock( componentName )
	startTimes[ componentName ] = KSDate.timeNowForAnimation()
end
	
--------------------------------------------------------------------------------
	
local function stopClock( componentName )
	if startTimes[ componentName ] then
		executionTimes[ componentName ] = ( KSDate.timeNowForAnimation() - startTimes[ componentName ] )
	end
end

--------------------------------------------------------------------------------

local function validateChangeCounter( counter )

	local message
	
	if type( counter ) ~= 'number' then
		message = "change counter wasn't a number"
	elseif counter < 1 then
		message = "change counter wasn't positive"
	end
	
	if message then
		failBecause = message
		error( message )
	end
	
	return counter

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.setUp()

	docStore = OzDocStore( ':temp:' )

	failBecause = nil
	
	testMetrics = nil

	docStore:PERF_collectQueryPerformanceStats()

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.tearDown()

	if docStore then

		docStore:PERF_dumpQueryPerformanceStats()

		docStore:close()
		docStore = nil

	end

	if failBecause then
		error( failBecause )
	end

end

--------------------------------------------------------------------------------

local function failTestWithMessage( message )

	failBecause = message

end

--------------------------------------------------------------------------------

local function setServerValuePTest( test, numberOfTimes )

	local docIds = {}
	for i = 1, numberOfTimes do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local randomId = docIds[ 100 ] -- randomly choosen 
	
	local function setServerValue( docId )		-- step 1
			
			return docStore:setServerValueP( 'sample', docId, {
						favorite = true,
						title = "Perfomance Testcase",
					} )
	end 
	
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	WFAsyncTestCase( test, 200 )
	
		:waitFor( function()
		
			-- Start the clock before bulk update
			startClock( test:testName() )
			return KSPromises.allP( setServerValue, docIds )
						
		end )
		:whenResolvedDo( function()
			
			-- Stop the clock after bulk update then insert execution time in test metrics object
			stopClock( test:testName() )
			testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
			WFTestMetrics.writeTestMetrics( testMetrics )

		end )
		
		:waitFor( function()
			
			return docStore:docValueOnceP( 'sample', randomId )
		
		end )
			
		:continueWhenValueMatches(  'data', {
				_changeCounter = 100,
				_docId = randomId,
				_docType = "sample",
				_ozMetadataUploadStatus = "uploaded",
				favorite = true,
				title = "Perfomance Testcase",
			}, ':ignore:'  )

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testSetServerValueP_5000_testPerformance( test )

	setServerValuePTest( test, 5000 )
		
end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testSetServerValueP_50000_testPerformance( test )

	if not WFPlatformInfo.isAndroid then
		setServerValuePTest( test, 50000 )
	end
		
end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testChangedDocsP_limit100_testPerformance( test )

	local docIds = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	WFAsyncTestCase( test, 60 )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
			
			-- Start the clock 
			startClock( test:testName() )
			return docStore:changedDocsP { fromChangeCounter = 1,
											limit = 100,
											idsOnly = true }
		
		end )
		
		:whenResolvedDo( function( status, changedDocs )
			
			stopClock( test:testName() )
			testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
			WFTestMetrics.writeTestMetrics( testMetrics )
			
			if status ~= 'data' then
				error "Promise resolved with non-data status"
			end
			
			assert( #changedDocs == 100, "Wrong number of docIds returned, actual value is " .. #changedDocs .. ", expected result was 100" )
			assert( changedDocs.continueFrom == 101 , "Wrong value of continueFrom, actual value is " .. changedDocs.continueFrom .. ", expected result was 101" )
			
		end )

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testChangedDocsP_limit500_changeCounter_testPerformance( test )

	local docIds = {}
	for i = 1, 1000 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	WFAsyncTestCase( test, 60 )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
			
			-- Start the clock 
			startClock( test:testName() )
			return docStore:changedDocsP { fromChangeCounter = 500,
											limit = 500,
											idsOnly = true }
		
		end )
		
		:whenResolvedDo( function( status, changedDocs )
			
			stopClock( test:testName() )
			testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
			WFTestMetrics.writeTestMetrics( testMetrics )
			
			if status ~= 'data' then
				error "Promise resolved with non-data status"
			end
			
			assert( #changedDocs == 500, "Wrong number of docIds returned, actual value is " .. #changedDocs .. ", expected result was 100" )
			assert( changedDocs.continueFrom == 1000 , "Wrong value of continueFrom, actual value is " .. changedDocs.continueFrom .. ", expected result was 1000" )			
		
		end )

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testChangedDocsP_limit1000_testPerformance( test )

	local docIds = {}
	for i = 1, 1000 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	WFAsyncTestCase( test, 60 )
		
		:waitFor( function()		-- step 1
		
			return docStore:applyOverlayToDocumentP {
					docType ='sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo"
					},
				}
			
		end )
		
		:waitFor( function()		-- step 2
			
			-- Start the clock 
			startClock( test:testName() )
			return docStore:changedDocsP { fromChangeCounter = 1,
											limit = 1000,
											idsOnly = true }
		
		end )
		
		:whenResolvedDo( function( status, changedDocs )
	
			-- Stop the clock after changedDocsP and write execution time in test metrics object
			stopClock( test:testName() )
			testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
			WFTestMetrics.writeTestMetrics( testMetrics )
				
			if status ~= 'data' then
				error "Promise resolved with non-data status"
			end
			
			assert( #changedDocs == 1000, "Wrong number of docIds returned, actual value is " .. #changedDocs .. ", expected result was 100" )
			assert( changedDocs.continueFrom == 1001 , "Wrong value of continueFrom, actual value is " .. changedDocs.continueFrom .. ", expected result was 1001" )				
		end )

end

--------------------------------------------------------------------------------

local function testApplyOverlayResults( args )

	-- FIXME [scouten 2013-04-10]: This should really be removed in favor of
	-- WFAsyncTestCase.

	local docId = args.docId or randomDocId
	local test = assert( args.test )
	local resultE = assert( args.resultE )
	local doWait = args.doWait

	local expectedResult = args.expectedResult
		or {
			_docType = 'sample',
			_docId = docId,
			favorite = true,
			title = "My Nifty Photo",
		}

	resultE:whenResolvedDo( function( status, message )
	
		if status ~= 'data' then
			failTestWithMessage( message )
			return
		end

		-- Stop the clock after bulk update and insert execution time in testMetrics object
		stopClock( test:testName() )
		testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
		WFTestMetrics.writeTestMetrics( testMetrics )

		docStore:docValueOnceP( 'sample', docId )
			:whenResolvedDo( function( status, doc2 )
			
				if status ~= 'data' then
					failTestWithMessage( doc2 )
					return
				end

				if args.haltOnData then halt() end
				
				if type( doc2 ) ~= 'table' then
					failBecause = "docValueOnceP should have returned a table (2)"
					return
				end
				
				expectedResult._changeCounter = validateChangeCounter( doc2._changeCounter )
				
				if not KSTableUtils.deepEquals( doc2, expectedResult, 5, true ) then
					failBecause = "unexpected result from docValueOnceP"
					return
				end
				
				test:notifyTestDidSucceed()
			
			end )
			:runUntilDone()

	end )
	:runUntilDone()

	if doWait then
		test:waitForTestToFinish( 3600 )
	end

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testBulkInsert_testPerformance( test )

	local nDocIds = WFPlatformInfo.isMac and 500000 or 50000
		-- 500K blows out memory on iPad, guessing on what
		-- will be OK
	if WFPlatformInfo.isAndroid then		-- We want a shorter test on Android.
		nDocIds = 1000
	end

	local docIds = {}
	for i = 1, nDocIds do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local victimDocId = docIds[ 4 ] -- just choose one at random
	
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	WFAsyncTestCase( test, 300 ) --600
	
		:verbose()
	
		:waitFor( function()
		
			-- Start the clock before bulk update
			startClock( test:testName() )
			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docIds = docIds,
					overlay = {
						favorite = true,
						title = "My Nifty Photo",
					},
				}
			
		end )
		:waitFor( function()
			
			-- Stop the clock after bulk update and insert execution time in test metrics object
			stopClock( test:testName() )
			testMetrics[ "wfp_execution_time" ] = executionTimes[ test:testName() ]
			WFTestMetrics.writeTestMetrics( testMetrics )
			
			return true

		end )
		:waitFor( function()

			return docStore:applyOverlayToDocumentP {
					docType = 'sample',
					docId = victimDocId,
					overlay = {
						bar = 94,
					},
				}
		
		end )
		
		:waitFor( function()
		
			return docStore:docValueOnceP( 'sample', victimDocId )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_docType = 'sample',
				_docId = victimDocId,
				_changeCounter = nDocIds + 1,
				bar = 94,
				favorite = true,
				title = "My Nifty Photo",
			} , ':ignore:' )

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.testBulkInsertAndUpdate_testPerformance( test )

	-- TODO [scouten 2013-04-10]: This should really be replaced with
	-- WFAsyncTestCase.

	local docIds = {}
	
	local nDocIds = WFPlatformInfo.isMac and 500000 or 50000
		-- 500K blows out memory on iPad, guessing on what
		-- will be OK
	if WFPlatformInfo.isAndroid then		-- We want a shorter test on Android.
		nDocIds = 1000
	end

	for i = 1, nDocIds do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	local victimDocId = docIds[ 4 ] -- just choose one at random
	
	-- Generating test metrics, to be used by dashboard
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	
	-- Start the clock before bulk insert
	startClock( test:testName() )
	
	docStore:applyOverlayToDocumentP
		{
			docType = 'sample',
			docIds = docIds,
			overlay = {
				favorite = true,
				title = "My Nifty Photo",
			},
		}
		
		:whenResolvedDo( function( status, message )
		
			if status == 'error' then
				failTestWithMessage( message )
				return
			end

			testApplyOverlayResults {
				test = test,
				doWait = false,
				docId = victimDocId,
				expectedResult = {
					_docType = 'sample',
					_docId = victimDocId,
					bar = 94,
					favorite = true,
					title = "My Nifty Photo",
				},
				resultE = docStore:applyOverlayToDocumentP {
						docType = 'sample',
						docIds = docIds,
						overlay = {
							bar = 94,
						},
					}
			}
		
		end )
		:runUntilDone()
	
	test:waitForTestToFinish( 1800 ) --3600

end

--------------------------------------------------------------------------------

function PERF_OzDocStoreTest.DisabledtestBulkUndoRedo( test )

	--maybeProfileSqlQueries()

	-- unimplemented
	
	-- Make sure undo/redo aren't costly when we have lots of docs.

end

--------------------------------------------------------------------------------

return PERF_OzDocStoreTest
