--[[----------------------------------------------------------------------------

WFActivity.lua
arahn
04/02/2013

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSStringUtils = require "KSStringUtils"

--------------------------------------------------------------------------------

local shallowcopy = function( t )
	local result = {}
	for k,v in pairs( t ) do
		result[ k ] = v
	end
	return result
end

--==============================================================================

local WFActivity = {}

--------------------------------------------------------------------------------

-- m : Slope, how much each activity that's added to the rollup contributes
--     to it's fitness
-- b : Intercept, how important this kind of activity is relative to the others.
WFActivity.activityCoefficients = {
	[ 'new asset in album' ] = {
		m = 10,
		b = 0,
	},
	[ 'new favorite' ] = {
		m = 10,
		b = 1,
	},
	[ 'new comment' ] = {
		m = 10,
		b = 2,
	},
	[ 'new asset' ] = {
		m = 10,
		b = 3,
	},
}

--------------------------------------------------------------------------------

local function subkey( x )
	-- we roll up either by photoid, in which subkey is just
	-- x, or by user, in which case it is x.id
	if type( x ) == 'table' then
		return x.id
	end
	return x
end

--------------------------------------------------------------------------------

local function insertActivity( dest, src )
	for k,v in pairs( src ) do
		local dv = dest[ k ]
		if dv == nil then
			dest[ k ] = v
		end
		if type( dv ) == 'table' and type( v ) == 'table' then
			-- this isn't recursive because we know our data
			-- structure only has this level of depth.
			local existing = {}
			local copied = {}
			for k2, v2 in pairs( dv ) do
				existing[ subkey( v2 ) ] = true
				copied[ k2 ] = v2
			end
			dest[ k ] = copied
			for i, x in ipairs( v ) do
				-- prevent the same userid/photoid from appearing
				-- more than once in a rollup
				local sk = subkey( x )
				if not existing[ sk ] then
					existing[ sk ] = true
					copied[ #dv + 1 ] = x
				end
			end
		end
	end
end

--------------------------------------------------------------------------------

--[[
	About roll up functions: 
		The consume a list of raw activity items which are provided in reverse-
		chronological order, and return two results
			a list of rolled up activities
				each one includes the fields described in OzCatActivity
				as well as a goodness-of-fit value 'fit' (higher values mean
				the rollup was valuable, e.g. if 10 raw 'new favorite'
				activities are able to be combined into a single activity
				then that would have a higher value than if only 2 were
				combined.
			a set of indexes that were used by the rollup.
	Scan the list of activities and return a list of rolled up activites
	and a list of unused activites.
	usedIndexes is a set of values indicating which entries have already
	been included in another rollup.
--]]
local function rollUp( debugName, getRollupId, activityType, activities, usedIndices )
	
	-- scan thru the activities looking for runs of 'new favorite' activities
	-- on the same photo.
	-- emit new rollups for those runs, and give them a fitness based on how
	-- many activities were rolledup plus an offset for important 'new favorites'
	-- are in the grand scheme of things.
	local coef = assert( WFActivity.activityCoefficients[ activityType ], "Invalid activity type" )
	
	local slope = coef.m
	local intercept = coef.b
	local rollups = {}
	
	local activityKeys = {}
	local activityMap = {}
	local activityAddsMeaning = {}
	
	-- generate the list of rollup groups, in order
	
	for i, activity in ipairs( activities ) do
		if activity.type == activityType then
			local rollupId = getRollupId( activity )
			local myMap = activityMap[ rollupId ]
			if not myMap then
				myMap = {}
				activityMap[ rollupId ] = myMap
				activityKeys[ #activityKeys + 1 ] = rollupId
			end
			if not usedIndices[ i ] then
				activityAddsMeaning[ rollupId ] = true
			end
			myMap[ #myMap + 1 ] = i
		end
	end
	
	local function emitRollup( rollup, x, currentUsedIndices )
		rollup.fit = slope * x + intercept
		rollup.usedIndices = currentUsedIndices
		rollup.rollupAlgorithm = activityType .. " by " .. debugName
		table.sort( rollup.users, function( a, b )
			local fnCmp = KSStringUtils.localizedStandardStringCompare( a.first_name or "", b.first_name or "" )
			if fnCmp ~= 0 then
				return fnCmp == -1
			end
			local lnCmp = KSStringUtils.localizedStandardStringCompare( a.last_name or "", b.last_name or "" )
			if lnCmp ~= 0 then
				return lnCmp == -1
			end
			return a.id < b.id
		end )
		table.sort( rollup.photos, function( a, b )
			return a < b
		end )
		rollups[ #rollups + 1 ] = rollup
	end
	
	-- flesh out the previously determined groups
	
	for _, rollupId in ipairs( activityKeys ) do
		if activityAddsMeaning[ rollupId ] then
			local activityIndexes = activityMap[ rollupId ]
			local i = activityIndexes[ 1 ]
			local currentUsedIndices = { [i]=true }
			local x = 1
			local rollup = shallowcopy( activities[ i ] )
			
			for n = 2, #activityIndexes do
				i = activityIndexes[ n ]
				insertActivity( rollup, activities[ i ] )
				currentUsedIndices[ i ] = true
				x = x + 1
			end
			
			emitRollup( rollup, x, currentUsedIndices)
		end
	end
	
	return rollups
end

--------------------------------------------------------------------------------

local function datePart( activity )
	return activity and activity.date and string.sub( activity.date, 1, 10 ) or ""
end

--------------------------------------------------------------------------------

local function rollUpByPhotoId( ... )
	return rollUp( "byPhotoId", function( activity )
		return activity.photos[ 1 ] .. datePart( activity )
	end, ... )
end

--------------------------------------------------------------------------------

local function rollUpByUserId( ... )
	return rollUp( "byUserId", function( activity )
		return activity.users[ 1 ].id .. datePart( activity )
	end, ... )
end

--------------------------------------------------------------------------------

local function rollUpByAlbumIdAndUserId( ... )
	return rollUp( "byUserId", function( activity )
		return ( activity.album and activity.album.id or "" ) .. activity.users[ 1 ].id .. datePart( activity )
	end, ... )
end

--------------------------------------------------------------------------------

local function rollUpByAlbumIdAndUserId( ... )
	return rollUp( "byUserId", function( activity )
		return ( activity.album and activity.album.id or "" ) .. activity.users[ 1 ].id
	end, ... )
end

--------------------------------------------------------------------------------

local function rollUpByUserIdSplittingYoursAndTheirs( ... )
	return rollUp( "byUserId", function( activity )
		return ( activity.yourPhotos and "your-photo-" or "" ) .. activity.users[ 1 ].id .. datePart( activity )
	end, ... )
end

--------------------------------------------------------------------------------

--[[
	Roll up a raw activity feed by searching for multiple new favorite
	actions by the same user
	
	\param activities the array of raw activities
	\return an array of activities including rollups where possible.  Rollups
		are just like activities but include additional fields (usedIndices,
		and fit)
--]]--
function WFActivity.rollUpFavoritesByUserId( activities, usedIndices )
	return rollUpByUserIdSplittingYoursAndTheirs( 'new favorite', activities, usedIndices )
end

--------------------------------------------------------------------------------

--[[
	Roll up a raw activity feed by searching for multiple new favorite
	actions on the same photo by multiple users.
	
	\param activities the array of raw activities
	\return an array of activities including rollups where possible.
--]]--
function WFActivity.rollUpFavoritesByPhotoId( activities, usedIndices )
	return rollUpByPhotoId( 'new favorite', activities, usedIndices )
end

--------------------------------------------------------------------------------

--[[
	Roll up a raw activity feed by searching for multiple assets added to albums
	by the same user.
	
	\param activities the array of raw activities
	\return an array of activities including rollups where possible.
--]]--
function WFActivity.rollUpAlbumAssetsByUserId( activities, usedIndices )
	return rollUpByAlbumIdAndUserId( 'new asset in album', activities, usedIndices )
end

--------------------------------------------------------------------------------

--[[
	Roll up a raw activity feed by searching for multiple assets added to the library
	by the same user.
	
	\param activities the array of raw activities
	\return an array of activities including rollups where possible.
--]]--
function WFActivity.rollUpNewAssetsByUserId( activities, usedIndices )
	return rollUpByUserId( 'new asset', activities, usedIndices )
end

--------------------------------------------------------------------------------

--[[--
	\brief rollup a raw activity feed, reducing similar activity entries
	by aggregating them along common lines.
	
	It runs the provided array of rollUp functions, (or uses all of the ones
	provided by this namespace as a default).  The best fitting rollups
	are chosen and are used in place of the individual activity entries.
	
	\param activities an array of raw (not rolled up) activities ordered by time
	\param optRollUpFns an optional array of roll up functions
	
	\result an array of activities that have been simplified as best possible.
--]]--
function WFActivity.rollUpActivity( activities, limit, optRollUpFns )
	optRollUpFns = optRollUpFns or {
		WFActivity.rollUpFavoritesByUserId,
		WFActivity.rollUpFavoritesByPhotoId,
		WFActivity.rollUpAlbumAssetsByUserId,
		WFActivity.rollUpNewAssetsByUserId
	}
	local rollUps = {}
	for i, rollUpFn in ipairs( optRollUpFns ) do
		rollUps[ i ] = rollUpFn( activities, {} )
	end
	local rollupsOrderedByImportance = {}
	for _, rollUpList in ipairs( rollUps ) do
		for _, rollUp in ipairs( rollUpList ) do
			rollupsOrderedByImportance[ #rollupsOrderedByImportance + 1 ] = rollUp
		end
	end
	table.sort( rollupsOrderedByImportance, function( a,b )
		return a.fit > b.fit
	end )
	local indexesInRollups = {}
	for _, rollUp in ipairs( rollupsOrderedByImportance ) do
		for idx in pairs( rollUp.usedIndices ) do
			if not indexesInRollups[ idx ] then
				indexesInRollups[ idx ] = rollUp
			end
		end
	end
	local result = {}
	for i, activity in ipairs( activities ) do
		local rollUp = indexesInRollups[ i ]
		if rollUp then
			if rollUp.emitted then
				-- skip it, already part of result
			else
				result[ #result + 1 ] = rollUp
				rollUp.emitted = true
			end
		else
			result[ #result + 1 ] = activity
		end
	end
	
	-- clean up
	for _, activity in pairs( result ) do
		activity.emitted = nil
		activity.rollupAlgorithm = nil
		activity.usedIndices = nil
		activity.fit = nil
		activity.andMaybeMore = #(activity.photos or {}) >= limit
	end
	return result
end

--==============================================================================

return WFActivity