--[[----------------------------------------------------------------------------

WFActivityTest.lua
arahn
04/02/2013

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSDate = require "KSDate"
local WFActivity = require "WFActivity"
local WFPlatformInfo = require 'WFPlatformInfo'
local deepEquals = function( a, b )
	return require 'KSTableUtils'.deepEquals( a, b, nil, true )
end

--==============================================================================

local WFActivityTest = {}

--------------------------------------------------------------------------------

local timeStamp = 0

function WFActivityTest.setUp(  )
	timeStamp = 0
end

--------------------------------------------------------------------------------

local function makeUsers( ... )
	
	local t = {}
	for i = 1, select( '#', ... ) do
		local userId = select( i, ... )
		local _, _, fn, ln = string.find( userId, "([^ ]+) (.+)" )
		t[ i ] = {
			fullName = userId .. "'s fullname", --'
			lastName = ln or ( userId .. "'s last name" ), --'
			firstName = fn or ( userId .. "'s first name" ), --'
			id = userId,
		}
	end
	
	return unpack( t )

end

--------------------------------------------------------------------------------

local function toT( a, ... )
	if type( a ) ~= 'table' then
		return { a, ... }
	else
		return a, ...
	end
end

--------------------------------------------------------------------------------

local function makeActivity( activityType )
	return function( userIds, photoIds, optTimestamp, yourPhotos, andMaybeMore )

		if not optTimestamp then
			timeStamp = timeStamp + 1
			optTimestamp = timeStamp
		end
	
		userIds = toT( userIds )
		photoIds = toT( photoIds )
	
		return {
			type = activityType,
			bookmark = KSDate.timeInIso8601Format( ( 48 - optTimestamp ) * 3600 ),
			id = 'a uuid ' .. tostring( optTimestamp ),
			date = KSDate.timeInIso8601Format( ( 48 - optTimestamp ) * 3600 ),
			users = {
				makeUsers( unpack( userIds ) ),
			},
			photos = {
				unpack( photoIds ),
			},
			yourPhotos = not not yourPhotos,
			andMaybeMore = not not andMaybeMore,
		}
	end
end

--------------------------------------------------------------------------------

local function applyOverlay( t, v )
	for k, v2 in pairs( v ) do
		t[ k ] = v2
	end
	return t
end

--------------------------------------------------------------------------------

local function useIndices( ... )
	local result = {}
	for i = 1, select( '#', ... ), 2 do
		for o = select( i, ... ), select( i+1, ... ) do
			result[ o ] = true
		end
	end
	return result
end

--------------------------------------------------------------------------------

local function testRollUpByPhotoId( generator, activityType )
	
	local makeActivityFn = makeActivity( activityType )
	local coef = assert( WFActivity.activityCoefficients[ activityType ], "Invalid activity type" )
	
	local slope = coef.m
	local intercept = coef.b

	local input = {
		makeActivityFn( "Becka", "photo1" ),
		makeActivityFn( "Andy B", "photo1" ),
		makeActivityFn( "Carrie", "photo1" ),
		makeActivityFn( "Andy A", "photo1" ),
		makeActivityFn( "Dan", "photo2" ),
		makeActivityFn( "Ellie", "photo2" ),
		makeActivityFn( "Frank", "photo3" ),
	}
	
	local rollup = generator( input, {} )
	local expected = {
		applyOverlay( makeActivityFn( { "Andy A", "Andy B", "Becka", "Carrie" }, "photo1", 1 ), {
			fit = 4 * slope + intercept,
			usedIndices = useIndices( 1, 4 ),
			rollupAlgorithm = activityType .. " by byPhotoId",
		} ),
		applyOverlay( makeActivityFn( { "Dan", "Ellie" }, "photo2", 5 ), {
			fit = 2 * slope + intercept,
			usedIndices = useIndices( 5, 6 ),
			rollupAlgorithm = activityType .. " by byPhotoId",
		} ),
		applyOverlay( makeActivityFn( { "Frank" }, "photo3", 7 ), {
			fit = 1 * slope + intercept,
			usedIndices = useIndices( 7, 7 ),
			rollupAlgorithm = activityType .. " by byPhotoId",
		} ),
	}
	assert( deepEquals( rollup, expected ) )
	
end

--------------------------------------------------------------------------------

local function testRollUpByUserId( generator, activityType )
	
	local makeActivityFn = makeActivity( activityType )
	local coef = assert( WFActivity.activityCoefficients[ activityType ], "Invalid activity type" )
	
	local slope = coef.m
	local intercept = coef.b

	local input = {
		makeActivityFn( "Geoff", "photo1" ),
		makeActivityFn( "Hal", "photo1" ),
		makeActivityFn( "Hal", "photo4" ),
		makeActivityFn( "Hal", "photo4" ), -- *** duplicate
		makeActivityFn( "Geoff", "photo2" ),
		makeActivityFn( "Ingrid", "photo3" ),
		makeActivityFn( "Ingrid", "photo1" ),
		makeActivityFn( "Geoff", "photo4" ),
		makeActivityFn( "Ingrid", "photo5" ),
		makeActivityFn( "Ingrid", "photo6" ),
	}
	
	local rollup = generator( input, {} )
	local expected = {
		applyOverlay( makeActivityFn( "Geoff", { "photo1", "photo2", "photo4" }, 1 ), {
			fit = 3 * slope + intercept,
			usedIndices = useIndices( 1, 1, 5, 5, 8, 8 ),
			rollupAlgorithm = activityType .. " by byUserId",
		} ),
		applyOverlay( makeActivityFn( "Hal", { "photo1", "photo4" }, 2 ), {
			fit = 3 * slope + intercept,
				-- fitness includes the extra "Hal" entry that was elided because
				-- it was a duplicate (see *** above).
			usedIndices = useIndices( 2, 4 ),
			rollupAlgorithm = activityType .. " by byUserId",
		} ),
		applyOverlay( makeActivityFn( "Ingrid", { "photo1", "photo3", "photo5", "photo6" }, 6 ), {
			fit = 4 * slope + intercept,
			usedIndices = useIndices( 6, 7, 9, 10 ),
			rollupAlgorithm = activityType .. " by byUserId",
		} ),
	}
	assert( deepEquals( rollup, expected ) )
	
end

--------------------------------------------------------------------------------

function WFActivityTest.testRollUpFavoritesByPhotoId(  )
	return testRollUpByPhotoId( WFActivity.rollUpFavoritesByPhotoId, 'new favorite' )
end

--------------------------------------------------------------------------------

function WFActivityTest.testRollUpFavoritesByUserId(  )
	return testRollUpByUserId( WFActivity.rollUpFavoritesByUserId, 'new favorite' )
end

--------------------------------------------------------------------------------

function WFActivityTest.testRollUpAlbumAssetsByUserId(  )
	return testRollUpByUserId( WFActivity.rollUpAlbumAssetsByUserId, 'new asset in album' )
end

--==============================================================================

function WFActivityTest.testCombinedRollUp()
	
	local makeFavorite = makeActivity( 'new favorite' )
	local makeComment = makeActivity( 'new comment' )
	local makeNewAlbumAsset = makeActivity( 'new asset in album' )
	local makeNewAsset = makeActivity( 'new asset' )

	local input = {
		makeNewAsset( "Geoff", "photo1" ),
		makeFavorite( "Geoff", "photo1" ),
		makeFavorite( "Hal", "photo1" ),
		makeComment( "Hal", "photo1" ),
		makeNewAlbumAsset( "Hal", "photo1" ),
		makeFavorite( "Ingrid", "photo1" ),
		makeNewAsset( "Geoff", "photo2" ),
		makeFavorite( "Ingrid", "photo2" ),
		makeNewAsset( "You", "photo3" ),
		makeFavorite( "Ingrid", "photo3" ),
		makeComment( "Geoff", "photo1" ),
		makeNewAsset( "You", "photo4" ),
		makeNewAsset( "You", "photo5" ),
		makeNewAsset( "You", "photo6" ),
		makeFavorite( "Ingrid", "photo4", nil, true ),
		makeFavorite( "Ingrid", "photo5", nil, true ),
		makeFavorite( "Ingrid", "photo6", nil, true ),
		makeFavorite( "Geoff", "photo1" ), -- *** duplicate
		makeNewAlbumAsset( "Hal", "photo2" ),
		makeNewAlbumAsset( "Hal", "photo3" ),
		makeNewAlbumAsset( "Hal", "photo4" ),
	}
	
	local rollup = WFActivity.rollUpActivity( input, 4 )
	local expected = {
		makeNewAsset( "Geoff", { "photo1", "photo2" }, 1 ),
		makeFavorite( { "Geoff", "Hal", "Ingrid" }, "photo1", 2 ),
		makeComment( "Hal", "photo1", 4 ),
		makeNewAlbumAsset( "Hal", { "photo1", "photo2", "photo3", "photo4" }, 5, nil, true ),
		makeFavorite( "Ingrid", { "photo1", "photo2", "photo3" }, 6 ),
		makeNewAsset( "You", { "photo3", "photo4", "photo5", "photo6" }, 9, nil, true ),
		makeComment( "Geoff", "photo1", 11 ),
		makeFavorite( "Ingrid", { "photo4", "photo5", "photo6" }, 15, true ),
	}
	assert( deepEquals( rollup, expected ) )
	
end

--==============================================================================

return WFActivityTest