--[[----------------------------------------------------------------------------

WFAnalyticsService.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author allany 2012-01-19

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- libs.reactive
local KSPromises = require 'KSPromises'
local WFArguments = require "WFArguments"

	-- libs.prefs
-- local KSPrefs = require 'KSPrefs'

local WFAnalyticsRunUntilDonePool = KSPromises.makeRunUntilDonePool()
	-- We use our own run until done pool so that we don't mess up counts that
	-- check to see whether we have outstanding actions.

local log = require 'WFLogger' "WFAnalyticsService"

--------------------------------------------------------------------------------

local analyticsManagerEP = WFWorkerThread( "WFAnalyticsServiceImp.lua" )

--------------------------------------------------------------------------------


--============================================================================--
--[[!
    \class WFAnalyticsService
    
    \brief This namespace provides a way for client code to send analytics
    events to SiteCatalyst. The functions can be called from any thread.

    This API is a minimal wrapper over Adobe's AppMeasurement SiteCatalyst
    analytics library.
]]--

-- @mark WFAnalyticsService namespace

local WFAnalyticsService = {}

--------------------------------------------------------------------------------
--[[!
    Sends a single event to the AppMeasurement SiteCatalyst analytics
    library. Checks to see if the user has opted in to data collection.
    
    \param params  A table of parameters to send.
    	Special keys include \c appState to set and track the app state, and
	\c eventName to send an event.
    
    \return A promise to signal completion, which can be safely ignored.
]]--

function WFAnalyticsService.send( --[[table]] params )
	
	if type( params ) == 'string' then
		log:warn( "Ignoring old style API call to WFAnalyticsService.send" )
		return KSPromises.resolvedPromiseP( 'data', nil )
			:runUntilDoneP( WFAnalyticsRunUntilDonePool )
	end

	WFArguments.check( "table" )
	
	local resultP = analyticsManagerEP:requestP( 'send', params )

--    if WFPlatformInfo.isCocoa then
--
--        assert( type( pageName ) == 'string', "WFAnalyticsService.send: pageName must be a string" )
--        assert( type( channel ) == 'string', "WFAnalyticsService.send: channel must be a string" )
--        assert( ( optKeyValuePairs == nil or type( optKeyValuePairs ) == 'table' ), "WFAnalyticsService.send: optKeyValuePairs must be nil or a table" )
--
--        local allowSend = false
--        local ozAccountId = KSPrefs.currentUserId
--
--        -- Allow the event only if the user is logged in *and* has opted-in to data collection
--        if ozAccountId then
--
--            local currentPrefsOptInValue = KSPrefs[ "OmnitureOptIn" .. ozAccountId ]
--            if currentPrefsOptInValue ~= nil and currentPrefsOptInValue == true then
--                allowSend = true
--            end
--
--        end
--
--        if allowSend then
--
--			-- No need to also call analyticsManagerEP:requestP( 'enableTracking', true ) here, because
--			-- when this reaches the WFAnalyticsManager, BestPractice events and Offline Cache will be ensured,
--			-- (assuming the client has configured for BestPractice events and Offline Cache support).
--
--			resultP = analyticsManagerEP:requestP( 'send', pageName, channel, optKeyValuePairs )
--
--		else
--
--			-- Disable BestPractice events and Offline Caching
--			resultP = analyticsManagerEP:requestP( 'enableTracking', false )
--
--		end
--
--    end

    if resultP == nil then
        resultP = KSPromises.resolvedPromiseP( 'data', nil )
    end

    resultP:runUntilDone( WFAnalyticsRunUntilDonePool )

    return resultP

end

--------------------------------------------------------------------------------
--[[!
    Tells the analytics manager to enable or disable
    AppMeasurement's sending of BestPractice events and offline tracking.
    
    \param enable (boolean) \c true to enable or \c false to disable tracking events
    
    \return A promise to signal completion, which can be safely ignored.
]]--

function WFAnalyticsService.enableTracking( --[[boolean]] enable )

	local resultP = analyticsManagerEP:requestP( 'enableTracking', enable )

	resultP:runUntilDone( WFAnalyticsRunUntilDonePool )

    return resultP

end

--------------------------------------------------------------------------------
--[[!
    \brief Initializes analytics with necessary information, including the tracking
	suite and server name.
    
    \param params A table of important key/value pairs
    
    \return A promise to signal completion, which can be safely ignored.
]]--

function WFAnalyticsService.initP( --[[table]] params )

	local resultP = analyticsManagerEP:requestP( 'init', params )

	resultP:runUntilDone( WFAnalyticsRunUntilDonePool )

    return resultP

end

local appStateDataBs = {}

--------------------------------------------------------------------------------
--[[!
    Tells the analytics manager to track specific application state data values.
    
    \param key  (string) The key of the event to track.
    \param value (string) The value of the event to track.
    
    \return A promise to signal completion, which can be safely ignored.
]]--

function WFAnalyticsService.storeAppStateData( --[[string]] key, --[[string]] value )

	if appStateDataBs[ key ] then
		appStateDataBs[ key ]:close()
		appStateDataBs[ key ] = nil
	end
	
	local resultP = analyticsManagerEP:requestP( 'storeAppStateData', key, value )

	resultP:runUntilDone( WFAnalyticsRunUntilDonePool )

    return resultP

end

--------------------------------------------------------------------------------
--[[!
    Tells the analytics manager to track specific application state data values.
    
    \param key  (string) The key of the event to track.
    \param valueB (Behavior) A behavior whose value is to be tracked. ((is this right?))
    
    \return A promise to signal completion, which can be safely ignored.
]]--

function WFAnalyticsService.storeAppStateDataB( --[[string]] key, --[[Behavior]] valueB )

	local resultP = analyticsManagerEP:requestP( 'storeAppStateData', key, valueB:valueNow() )
	
	resultP:runUntilDone( WFAnalyticsRunUntilDonePool )

	if appStateDataBs[ key ] then
		appStateDataBs[ key ]:close()
		appStateDataBs[ key ] = nil
	end
	
	appStateDataBs[ key ] = valueB:relayB()
	
	appStateDataBs[ key ]:onDataB( function( value )
		analyticsManagerEP:requestP( 'storeAppStateData', key, value )
		:runUntilDone( WFAnalyticsRunUntilDonePool )
	end )

    return resultP

end

--------------------------------------------------------------------------------
--[[!
    Sends an analytics event if the promise resolves successfully.
    
    \param promise (Promise) The promise we are waiting on to resolve before sending
     the analytics data.
    \param caseTable (table) The Analytics data to report for data, error, or canceled.
    
    \return A new promise that resolves after the analytics call has been made.
]]--

function WFAnalyticsService.sendAnalyticsP( --[[Promise]] promiseP, --[[table]] caseTable )
	
	return promiseP:mapP( function( ... )
		
		local case
		
		if select( '#', ... ) == 0 then
			case = 'canceled'
		else
			case = ( ... )
		end
		
		local analytics = case and caseTable[ case ]
		
		if analytics then
			WFAnalyticsService.send( analytics )
		end
		
		return ...
		
	end )

end

--------------------------------------------------------------------------------

function WFAnalyticsService.sendFromNative( ... )

	if type( select( 1, ... ) ) ~= 'table' then
		log:warn( "Ignoring old-style call to sendAnalyticsEvent (native)" )
	else
		WFAnalyticsService.send( ... )
	end

end

--------------------------------------------------------------------------------

return WFAnalyticsService
