--[[----------------------------------------------------------------------------

WFAsyncTestCase.lua

@author scouten 2011-04-21
@reviewed scouten 2012-08-05

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.lua.threading
local WFIdle = require 'WFIdle'
local WFTimer = require 'WFTimer'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.table
local KSTableArchiving = require 'KSTableArchiving'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSLuaUtils = require 'KSLuaUtils'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local dpcall = assert( KSLuaUtils.dpcall )

local WFAsyncTestCase = 'WFAsyncTestCase'


--============================================================================--
--[[!
	\class WFAsyncTestCase

	\brief This namespace allows you to test code that makes
		asynchronous calls through event streams, behaviors, and promises.

	Create a \c WFAsyncTestCase object and attach a number of conditions
	that the test must pass in order to be marked as successful.

	Specify a condition in a <code>:waitFor( function )</code> block.
	The conditions are tested in the sequence in which you attach them;
	this roughly approximates the code you would write if you were using
	thread-blocking code.

	For example:

\code
function KSReactivePromiseBridgeTest.testPromiseToBehaviorWithRequest( test )

	WFAsyncTestCase( test )

		local requestTestHandler = WFWorkerThread( 'WFRequestTestHandler.lua' )
	
		:waitFor( function()

			local p = requestTestHandler:requestP( 'echo', "The data value" )
			
			return KSReactivePromiseBridge.promiseToBehavior( p, "Not Yet!" )
		
		end )
		:continueWhenValueMatches( 1, "The data value" )  -- condition
	
end
\endcode

	You can attach any number of \c :waitFor blocks to the test.
	Each block can return a timeout value (a hard number of seconds to wait),
	and an event stream, behavior, or promise.
	
	After each \c :waitFor block, you can specify a condition that
	must be satisfied in order for that step to pass. The typical case is
	<code>:continueWhenValueMatches</code>, which tests the reactive node's
	value against the expected values before passing the step.
	
	You can also use <code>:whenResolvedDo</code> (for promises) or
	<code>:onDataE</code> (for event streams or behaviors) if you can't
	provide an exact expected value set.
	
	If no validation steps are applied, the step passes if the event
	stream, behavior, or promise returns \b any non-error value.

	If all of the <code>:waitFor</code> steps  are completed without
	an error being raised or the timeout being reached, then the test is marked
	as successful.
--]]

-- @mark WFAsyncTestCase class

local reactiveTestCaseMethods = {

	_test = false,
		-- (KSLuaTestCase native object)
	
	_steps = false,
		-- (table) array of step descriptions
	
	_verbose = false,
		-- (boolean) true to generate extra log output
	
	_currentStepIndex = false,
		-- (number) index into _steps array; which step are we now waiting for?
	
	_failureMessage = false,
		-- (string) if present, this is the failure message for this test
	
	_currentReactiveNode = false,
		-- (event stream or behavior) the node we're currently waiting on
	
	_promiseActionHolder = false,
		-- (table) used to hold :whenResolvedDo actions used by calling code
		-- alive for the life of this test case
	
	_baselinePromiseLimboCount = false,
		-- (number) baseline result for KSPromises.DEBUG_getLimboCount().
		-- We use this to make a test fail if the count increases.
	
	_matchValueActionHolder = false,
		-- (table) used to hold :whenResolvedDo actions used by this code
		-- alive for the life of this test case
	
}

--------------------------------------------------------------------------------

local continueTest -- forward declaration
local sCurrentTest -- prevent the current async testcase from being GC'd

--------------------------------------------------------------------------------
--[[!
	\brief Constructs an async test-case object.

	\param test (WFTestCase) A generic test-case object.
	\param waitTime (optional, number) A timeout for the test in seconds.
		Default is 2 seconds.

	\return The new async test-case object.
]]--

function reactiveTestCaseMethods:__init( --[[WFTestCase]] test,
					 --[[optional number]] waitTime )
	
	check( WFAsyncTestCase, '_', '?number' )
		-- FIXME [scouten 2012-08-05]: We don't have a checker for native
		-- type WFTestCase.
	
	sCurrentTest = self

	self._test = assert( test )
	
	self._steps = {}
	self._currentStepIndex = 0
	
	self._promiseActionHolder = {}
	self._matchValueActionHolder = {}
	
	self._baselinePromiseLimboCount = KSPromises.DEBUG_getLimboCount()
	
	test:waitForTestToFinish( waitTime or 2 )
		-- Default wait time for tests is 2 seconds.
	
	WFIdle.onIdleDo( function()
	
		continueTest( self )
	
	end )

end

--------------------------------------------------------------------------------

local function debugPrintValue( test, verbose, ... )

	local values = { ... }

	for i = 1, select( '#', ... ) do

		local value = values[ i ]
		
		if type( value ) == 'table' then
			values[ i ] = KSTableArchiving.simpleTableToChunkString( value )
		else
			values[ i ] = tostring( values[ i ] )
		end
	
	end
	
	if verbose then
		debug.print( unpack( values, 1, select( '#', ... ) ) )
	end
	
	local joinedValues = table.concat( values, " " ) .. "\n"
	
	test:log( joinedValues )

	return joinedValues

end

--------------------------------------------------------------------------------

local dividerString = string.rep( '-', 80 ) .. '\n\n'

--------------------------------------------------------------------------------

continueTest = function( self )

	collectgarbage()
	collectgarbage()

	local verbose = self._verbose

	self._currentReactiveNode = false

	local test = self._test

	-- If we've logged a failure message, short-circuit out now.

	if self._failureMessage then
		test:log( self._failureMessage )
		test:notifyTestDidFail( self._failureMessage )
		sCurrentTest = nil
		return
	end
	
	-- Make sure no new promises or promise actions got left in limbo.
	-- What would be really nice is if this detected promise limbo happening
	-- on other threads, but that doesn't work yet. :-(
	
	if KSPromises.DEBUG_getLimboCount() > self._baselinePromiseLimboCount then
		self._failureMessage = "One or more promises or promise actions were left in limbo: Failing test"
		test:log( self._failureMessage )
		test:notifyTestDidFail( self._failureMessage )
		sCurrentTest = nil
		return
	end
	
	-- What's the next step?

	self._currentStepIndex = self._currentStepIndex + 1

	local currentStep = self._steps[ self._currentStepIndex ]
	debugPrintValue( test, verbose,
					 dividerString .. "TEST STEP",
					 self._currentStepIndex,
					 "(", currentStep and currentStep.whereDefined or 'END OF TEST', ")" )
	
	if not currentStep then
		debugPrintValue( test, verbose, "Test succeeded" )
		self._test:notifyTestDidSucceed()
		sCurrentTest = nil
		return
	end
	
	local success, messageOrNode
	if type( currentStep.eventOrGenFunc ) == 'function' then
		success, messageOrNode = dpcall( string.format( 'step %d (defined at %s)',
														self._currentStepIndex,
														currentStep.whereDefined ),
										 currentStep.eventOrGenFunc )
	else
		success, messageOrNode = true, currentStep.eventOrGenFunc
	end

	if success and not messageOrNode then
		success = false
		messageOrNode = ":waitFor at this step returned nil"
		debug.print( ":waitFor at this step returned nil" )
	end

	if success then
		
		-- Event gen function didn't fail out, configure the node
		-- so we can continue after this.
		
		local reactiveNode = messageOrNode
		
		if type( reactiveNode ) == 'number' then
		
			local timeout = reactiveNode
			reactiveNode = WFTimer.afterDelayP( timeout )
			
		end
		
		self._currentReactiveNode = reactiveNode or false
		
		local isReactiveNode, isBehavior = KSReactive.isEventNode( reactiveNode )
		local isPromise = ( not isReactiveNode )
						  and KSPromises.isPromise( reactiveNode )

		-- FIXME [scouten 2012-08-14]: Q&D hack: Convert things that are neither
		-- reactive nor promise to a resolved promise. Formalize and document this.

		if not isReactiveNode and not isPromise then
			isPromise = true
			reactiveNode = KSPromises.resolvedPromiseP( 'data', reactiveNode )
			self._currentReactiveNode = reactiveNode
			messageOrNode = reactiveNode
		end

		-- FIXME [scouten 2011-03-30]: Add some options to expect
		-- failure, etc., here.
	
		if currentStep.onDataE then

			assert( isReactiveNode )
			reactiveNode:onDataE( currentStep.onDataE )
		
		end

		local wrdFinishedP = reactiveNode
		local resolver
		
		-- Sometimes we have multiple functions that need to respond
		-- to the whenResolvedDo condition.  However, Promises make
		-- no guarantees about what order multiple functions attached
		-- in this way will be called.  So we will attach only a single
		-- function in this manner, and then run all of our funcs
		-- in sequence in the order in which they were attached
		
		-- What's more, we run the risk that we are waiting for a promise that
		-- has already resolved in which case the action will execute immediately
		-- and future additions to the list of functions to execute will never
		-- execute. So, we need to reset the list each time we process it.
		
		-- Worse, we really weren't focused on running the functions at this point
		-- in the process. So, if the node is already resolved then we need to actually
		-- attach to a version that fires later. We do this by always deferring the
		-- result from the reactive node.
				
		local whenResolvedDoFuncs = nil
		local function addWhenResolvedDoFn( actionFunc )
			if not whenResolvedDoFuncs then
				whenResolvedDoFuncs = { actionFunc }
				WFIdle.asapP():mapP( reactiveNode )
					:whenResolvedDo( function( ... )
						local existingFuncs = whenResolvedDoFuncs
						whenResolvedDoFuncs = nil
						for _, f in ipairs( existingFuncs ) do
							f( ... )
						end
					end )
					:attachTo( nil, self._promiseActionHolder, self._currentStepIndex )
			else
				whenResolvedDoFuncs[ #whenResolvedDoFuncs + 1 ] = actionFunc
			end
		end

		if currentStep.whenResolvedDo then
		
			assert( isPromise )
			
			local actionFunc = currentStep.whenResolvedDo
			
			-- Ensure that next step is always executed after the
			-- whenResolvedDo function from this step.
			
			if currentStep.expectedValue then
			
				reactiveNode:runUntilDone()
					-- FIXME [scouten 2011-10-14]: Ugh. Hacky way to
					-- silence the promise limbo warnings.
			
				wrdFinishedP, resolver = KSPromises.promiseP()
				
				actionFunc = function( ... )
					currentStep.whenResolvedDo( ... )
					resolver( ... )
				end
				
				wrdFinishedP:runUntilDone()
					-- FIXME [scouten 2011-10-14]: Ugh. Hacky way to
					-- silence the promise limbo warnings.
			
			end
			
			addWhenResolvedDoFn( actionFunc )
		
		end

		if currentStep.expectedValue then

			local expectedValue = currentStep.expectedValue
			
			if expectedValue.evaluate then
				local function packn( ... )
					return { n = select( '#', ... ), ... }
				end
				expectedValue = packn( expectedValue.evaluate() )
			end

			-- Guard against repeated calls to this onDataE
			-- function after we've satisfied this criteria.

			local hasMatched = false

			local function foundMismatch( message )
			
				self._test:log( message )

				-- Event streams and promises must match the expected
				-- value first time around. Behaviors can evolve to the
				-- expected value, provided they do it within the alloted
				-- time. Fail test immediately if not a behavior.

				if not isBehavior then

					self._failureMessage = message
					if isReactiveNode then
						reactiveNode:close()
					end

					self._test:notifyTestDidFail( self._failureMessage )
	
				end
			
			end
			
			local currentStepForMatch = self._currentStepIndex

			local function matchValue( ... )
			
				if currentStepForMatch ~= self._currentStepIndex then
					return
				end
			
				debugPrintValue( test, verbose,
								 "Attempting to match output for test step ",
								 self._currentStepIndex )

				local actualValuesStr = debugPrintValue( test, verbose, "actual", ... )
				local expectedValuesStr = debugPrintValue( test, verbose, "expected",
								 unpack( expectedValue, 1,
								 expectedValue.n ) )

				if hasMatched then return end

				-- Note that we expect that some behaviors may
				-- have other values before they get to the exact
				-- desired value, so we simply bail out (albeit
				-- with a log) if there's a mismatch.
				
				local count = select( '#', ... )
				
				if count ~= expectedValue.n then

					return foundMismatch( "Wrong number of results: Expected "
						.. tostring( expectedValue.n ) .. " got "
						.. tostring( count ) )

				end
				
				for i = 1, count do
			
					local paramI = select( i, ... )
					
					if expectedValue[ i ] ~= ':ignore:' then
					
						local mismatch = expectedValue[ i ] ~= paramI
						
						if mismatch
							and type( expectedValue[ i ] ) == 'table'
							and type( paramI ) == 'table'
							and KSTableUtils.deepEquals( expectedValue[ i ],
														 paramI, 20, verbose )
						then
							mismatch = false
						end
						
						if mismatch
							and type( paramI ) == 'string'
							and expectedValue[ i ] == ':*:'
						then
							mismatch = false
						end
				
						if mismatch then
						
							local mismatchInfo = "Result mismatch at position "
												.. tostring( i ) .. ": "
												.. actualValuesStr
												.. expectedValuesStr

							return foundMismatch( mismatchInfo )
						end
					
					end

				end

				-- Parameters are a complete match. Test can now proceed.
				
				hasMatched = true
				
				if isReactiveNode then
					continueTest( self )
				end
			
			end
			
			if isReactiveNode then

				-- It's possible we already have this value, in which case
				-- we just proceed right away. (Only applies to behaviors.)
				
				if isBehavior and not currentStep.expectError then
					matchValue( reactiveNode:valueNow() )
				end
				
				if not hasMatched then
					if currentStep.expectError then
						reactiveNode:onErrorE( matchValue )
					else
						reactiveNode:onDataE( matchValue )
					end
				end
			
			elseif isPromise then
			
				addWhenResolvedDoFn( matchValue )

			end
		
		end
		
		if isReactiveNode then

			reactiveNode:onErrorE( function( message )
			
				local annotatedMessage = string.format(
					"reactive node from step %d failed with message %s",
					self._currentStepIndex, message )
			
				debugPrintValue( test, verbose, annotatedMessage )
				self._failureMessage = annotatedMessage
				
				reactiveNode:close()
			
			end )
		
		end
		
		if isReactiveNode and currentStep.onDoneE == nil
			and not currentStep.expectedValue
		then

			reactiveNode:onDoneE( function()
			
				continueTest( self )
			
			end )
		
		end
		
		if isPromise then
		
			addWhenResolvedDoFn( function( status, message )

				if not currentStep.expectedValue and not currentStep.whenResolvedDo then
				
					if status ~= 'data' then
						self._failureMessage =
							string.format( "promise node failed with status = %s and message = %s",
											tostring( status ), tostring( message ) )
					end
					
				end
				
				continueTest( self )

			end )

		end
	
	else
	
		local annotatedMessage = string.format(
			"step %d failed to generate new node with message %s",
			self._currentStepIndex, messageOrNode or "(return value was nil)" )
	
		debugPrintValue( test, verbose, annotatedMessage )
		
		self._test:log( annotatedMessage )
		self._test:notifyTestDidFail( annotatedMessage )
	
	end

	-- Throw in extra GC to ensure that we don't succeed by chance if things
	-- are weakly held.

	collectgarbage()
	collectgarbage()

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a step to this test case.
	
	\param step An event stream, behavior, promise, or number; or a
		function that returns one of those types.
	
	If the ultimate result is an event stream or behavior, the default behavior
	is to wait until one non-error value is generated. If the ultimate result is
	a promise, the default behavior is to wait for that promise to resolve
	successfully. In either case, if an error result is generated, the default
	behavior is to fail the test.
	
	You can override these defaults with <code>:continueWhenValueMatches</code>
	or another method to evaluate the results of this node.
	
	If the ultimate result is a number, the behavior is to wait for that number
	of seconds. (This behavior cannot be overridden.)

	\return This object (useful for chaining).
--]]

function reactiveTestCaseMethods:waitFor( --[[reactive node or number]] step )

	check( WFAsyncTestCase, '_' )

	local steps = self._steps
	
	local whereDefined
	
	if type( step ) == 'function' then
	
		local debugInfo = debug and debug.getinfo and debug.getinfo( step )
		
		if debugInfo and debugInfo.source and debugInfo.linedefined then
			whereDefined = string.format( '%s:%s', debugInfo.source, debugInfo.linedefined )
		end
		
	else
		-- any Promise value that's passed to a waitFor step ought
		-- to just be kept alive by the test, so we'll test for that
		-- here  -- arahn.
		local isPromise = ( step )
						  and KSPromises.isPromise( step )
		if isPromise then
			step:runUntilDone()
		end
	end
	
	steps[ #steps + 1 ] = {
		eventOrGenFunc = assert( step ),
		whereDefined = whereDefined or 'source unknown',
	}
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Attaches a function to the event stream or behavior from the most
		recent <code>:waitFor</code> step via its <code>:onDataE</code> hook.
	
	\param onDataFunc (function) A function that can evaluate the event stream
		or behavior's results. It can cause the test to fail by raising an
		error.
	
	\return This object (useful for chaining).
]]--

function reactiveTestCaseMethods:onDataE( --[[function]] onDataFunc )

	check( WFAsyncTestCase, 'function' )

	local steps = self._steps
	assert( #steps > 0 )
	
	steps[ #steps ].onDataE = function( ... )
	
		local success, message = pcall( onDataFunc, ... )
		
		if not success then
			self._failureMessage = message
			debug.print( message )
			self._test:log( message )
		end
	
	end
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Attaches a function to the promise from the most recent
		<code>:waitFor</code> step via its <code>:whenResolvedDo</code> hook.
	
	\param resolvedFunc (function) A function that can evaluate the promise's
		results. It can cause the test to fail by raising an error.
	
	\return This object (useful for chaining).
]]--

function reactiveTestCaseMethods:whenResolvedDo( --[[function]] resolvedFunc )

	check( WFAsyncTestCase, 'function' )

	local steps = self._steps
	assert( #steps > 0 )
	
	steps[ #steps ].whenResolvedDo = function( ... )
	
		local success, message = pcall( resolvedFunc, ... )
		
		if not success then
			self._failureMessage = message
			debug.print( message )
			self._test:log( message )
		end
	
	end
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Provides expected values for the event stream, behavior, or promise
		from the most recent <code>:waitFor</code> step.
	
	\param ... The expected value or values.
		<ul><li>For an event stream, the very next <code>:onDataE</code>
		call \b must match this value or the test fails. </li>
		
		<li>For a behavior, other values can be permitted but the behavior
		must eventually reach the expected value or the test fails. </li>
		
		<li>For a promise, the status (\c 'data' or \c 'error') is the first parameter.
		The status and all returned values must match or the test fails.
		
		You can use the string \c ':ignore:' to skip matching any given parameter.
		For example:
<code> :continueWhenValueMatches( 'data', ':ignore:', ':ignore:', catalogId, true )</code></li>
		</ul>
	
	\return This object (useful for chaining).

--]]

function reactiveTestCaseMethods:continueWhenValueMatches( ... )

	local steps = self._steps
	assert( #steps > 0 )
	
	steps[ #steps ].expectedValue = { n = select( '#', ... ), ... }
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Provides the expected value or values for the event stream,
		behavior, or promise from the most recent <code>:waitFor</code> step.
	
	\param fn A function that returns the value or values to match,
		to be executed immediately before evaluating the match.
	
		<ul><li>For an event stream, the very next <code>:onDataE</code>
		call \b must match this value or the test fails. </li>
		
		<li>For a behavior, other values can be permitted but the behavior
		must eventually reach the expected value or the test fails. </li>
		
		<li>For a promise, the status (\c 'data' or \c 'error') is the first parameter.
		The status and all returned values must match or the test fails.
		
		You can use the string \c ':ignore:' to skip matching any given parameter.</li>
		</ul>

	\return This object (useful for chaining).

	\sa reactiveTestCaseMethods:continueWhenValueMatches

--]]

function reactiveTestCaseMethods:continueWhenValueMatchesFunction( --[[function]] fn )

	local steps = self._steps
	assert( #steps > 0 )
	
	steps[ #steps ].expectedValue = { evaluate = fn }
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Provides expected \b error values for the event stream or behavior
		from the most recent <code>:waitFor</code> step.
		
		Use to verify that the API under test returns an error
		condition when expected. The event stream or behavior
		must call <code>:onErrorE</code> with the expected value or values,
		or the test fails.
	
	\param ... The expected error values.
		<ul><li>For an event stream, the very next <code>:onDataE</code>
		call \b must match this value or the test fails. </li>
		
		<li>For a behavior, other values can be permitted but the behavior
		must eventually reach the expected value or the test fails. </li>
		
		<li>For a promise, the status (\c 'data' or \c 'error') is the first parameter.
		The status and all returned values must match or the test fails.
		
		You can use the string \c ':ignore:' to skip matching any given parameter.</li>
		</ul>
 	
		\return This object (useful for chaining).
	
--]]

function reactiveTestCaseMethods:continueWhenValueMatchesError( ... )

	local steps = self._steps
	assert( #steps > 0 )
	
	steps[ #steps ].expectedValue = { n = select( '#', ... ), ... }
	steps[ #steps ].expectError = true
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief Generates extra debug output to the console log while running the test.
	
	Use while debugging your unit test, but remove before checking in.
	
	\return This object (useful for chaining).
--]]

function reactiveTestCaseMethods:verbose()

	self._verbose = true
	
	return self

end

--------------------------------------------------------------------------------

return KSClass 'WFAsyncTestCase'( reactiveTestCaseMethods )
