--[[----------------------------------------------------------------------------

WFCloseSupport.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2011-02-02

@reviewed epoley 2011-03-08
@reviewed mhamburg 2011-03-14

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFIdle_lazy = nil

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSPrivateKeys = require 'KSPrivateKeys'
local KSUserdataProxy = require 'KSUserdataProxy' -- for the must close logic
local WFArguments = require 'WFArguments'

local KSFunctionUtils_memoize = require 'KSFunctionUtils_memoize'

--------------------------------------------------------------------------------

local WFArguments_check = assert( WFArguments.check )
local memoize = assert( KSFunctionUtils_memoize.memoizeFn )
local kForget = assert( KSFunctionUtils_memoize.kForget )

-- mscarafo 8/6/13
-- Toggle off all the statements in objectCloser_doClose
-- DO NOT MERGE THIS BACK! For Tornado only
local printDebug = false
--============================================================================--
--[[!
	\class WFCloseTarget

	\brief Represents a source or destination for data that can be explicitly closed.

	This is a mix-in class. Use the function \c WFCloseSupport.addCloseSupportMethods
	in the class declaration to add close support to that class.
	
	\todo Add an example of a class declaration.
	
	Any object that derives from \c WFCloseTarget can be explicitly closed by
	calling its <code>:close()</code> method.
	
	A \c WFCloseTarget can have zero or more <b>close targets</b>, which can form
	a directed acyclic graph of objects that depend on this object for their
	lifetime. When a close target object is closed, all objects marked as
	close targets of that object are also recursively closed.
	
	\sa WFCloseSupport
]]--

-- @mark WFCloseTarget class

local closeSupportMethods = {
	-- Note that this name and this method table are merely here for
	-- purposes of API documentation.
}

--------------------------------------------------------------------------------
--[[!
	\brief Closes this object and all of its close targets.
	
	This method can be overridden, in which case it is called once for each
	target object before any of the target objects are closed.
	If an object is already closed, this function does nothing.
]]--

function closeSupportMethods:close()
end

--------------------------------------------------------------------------------
--[[!
	\brief Converts this object so that it cannot be closed.
	
	Sets the <code>:close()</code> method to a no-op and causes
	<code>:hasBeenOrWillBeClosed()</code> to report false for this object.
]]--

function closeSupportMethods:bounceClosure()
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a target to a list of objects to close when this object closes.
		The target is weakly held.

	\param target The new target object. If not supplied, or a target that
		has already been closed, this call does nothing.
		
	\param noRemove An optional boolean indicating that we have no intention to
		remove this as a target. This will bypass the checks for double-adds.
	
	\return This object (useful for chaining).
]]--

function closeSupportMethods:addCloseTarget( --[[WFCloseTarget]] target,
											 --[[optional boolean]] noRemove )
end

--------------------------------------------------------------------------------
--[[!
	\brief Removes a target from the list of objects to close
	when this object is closed.
	
	\param target The new target object. If not supplied, or not a target
		on the list, this call does nothing.
]]--

function closeSupportMethods:removeCloseTarget( --[[WFCloseTarget]] target )
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds this object as a close target of another object.

	\param obj The object to which to add this one as a target.
 	
	\return This object (useful for chaining).
]]--

function closeSupportMethods:addAsCloseTargetOf( --[[WFCloseTarget]] obj )
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a function to be run when this object is closed.

	\param actionFn The function.
	
	\return This object (useful for chaining).
]]--
	
function closeSupportMethods:addCloseAction( --[[function]] actionFn )
end

--============================================================================--
-- @mark WFCloseSupport namespace

--[[!
	\class WFCloseSupport

	\brief This namespace helps you manage the lifetimes of objects by
	  creating classes whose objects can be closed, and which can be associated
	  with dependant objects  (targets) that are automatically closed with it, recursively.

	Typically, you use \c WFCloseSupport.addCloseSupportMethods() as a part of
	a class declaration, and then interact with the methods declared on \c WFCloseTarget.

	
	\sa WFCloseTarget
	
]]--

local WFCloseSupport = {}

--------------------------------------------------------------------------------

local function declare() end
declare( closeSupportMethods )

local assert = assert
local type = type
local pairs = pairs
local setmetatable = setmetatable
local rawset = rawset
local rawget = rawget

local DEBUG = _G.DEBUG

--------------------------------------------------------------------------------

local STOMP_CLOSED_OBJECT_METATABLES = false and DEBUG
	-- Do we want to replace the metatables for closed objects with
	-- a metatable that errors out.
	-- We don't do this for now because it makes a surprising amount of code
	-- unhappy and we need to figure out why. It could be that objects are just
	-- trying to settle into a semi-accessible state when closed and this code
	-- is too blunt an instrument to deal with that.
	
local closedObjectMT = {
	__index = setmetatable(
		{ close = function() debug.print( "Closing closed object. Ignoring. " .. debug.traceback() ) end },
		{ __index = function() error( "Attempt to access a closed object (index). " .. debug.traceback() ) end } ),
	__newindex = function() error "Attempt to access a closed object (newindex)." end,
	__call = function() error "Attempt to access a closed object (call)." end
}

local function stompClosedObjectMetatable( obj )

	if STOMP_CLOSED_OBJECT_METATABLES and type( obj ) == 'table' then
		debug.setmetatable( obj, closedObjectMT )
	end

end

--------------------------------------------------------------------------------

-- Avoid depdenency on KSTableUtils since WFCloseSupport is supposed to be
-- very primitive.

local weakKeysMT = { __mode = 'k' }
local weakSetMT = { __mode = 'kv' }

--------------------------------------------------------------------------------

-- We set the close method for objects bouncing closure to close_nop.

local function close_nop() end

--------------------------------------------------------------------------------

-- We maintain a mapping from objects with close targets to the weak set of
-- those targets. This mapping is weak-keyed in the object itself.

-- We maintain a mapping from objects that have been set as close targets to
-- the objects for which they are close targets. This is a weak mapping based
-- on a weak-keyed mapping from close targets to fully-weak sets of close sources.
-- We are safe from cyclic issues because the semi-weak table only leads to weak
-- sets which cannot lead back to the keys.

--------------------------------------------------------------------------------

local forwardCloseTargetMap = setmetatable( { }, weakKeysMT )

--------------------------------------------------------------------------------

-- We maintain a set of closed objects so that we will only close an object
-- once.

local closedObjects = setmetatable( { }, weakSetMT )

--------------------------------------------------------------------------------

-- We maintain a stack of objects to close (and functions to call) so that
-- we can avoid deep recursion when scanning the graph. This also allows us to
-- further reduce pcall overhead in most cases. We use the closedObjects set
-- to track objects that have been queued by setting their status to a magic
-- queued value.

local closeStack = { }
local closeStackTop = 0
local processingCloseStack = false

--------------------------------------------------------------------------------

local kQueued = { }
local kClosed = { }

--------------------------------------------------------------------------------

-- The following function returns true if the object is not already marked as
-- closed. It also marks the object as closed.

local function closeGuard( obj )

	if not obj then
		return false
	end

	local status = closedObjects[ obj ]
	
	if status == kClosed then
		return false
	end
	
	closedObjects[ obj ] = kClosed
	
	return true

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether this object has been closed or is scheduled for closure.

	An object can be scheduled for closure if it is in a chain of close targets
	that has not yet been fully processed.

	\param obj The object to test. If nil, the call returns true.
	
	\return \c true if and only if this object has already been closed, directly
	or indirectly, or if the object is in the queue waiting to be closed.
	This can happen, for example, if an object that marked \c obj was closed, but
	the chain of close targets hasn't been fully processed yet.)

	FIXME [mhamburg 2011-10-03]: The <code>obj == nil</code> behavior is dubious.
]]--

function WFCloseSupport.hasBeenOrWillBeClosed( --[[object]] obj )

	if obj == nil then
		return true
	end
	
	if closedObjects[ obj ] == nil then
		return false
	end
	
	if type( obj ) == 'table' and rawget( obj, 'close' ) == close_nop then
		return false
	end
	
	return true

end

local WFCloseSupport_hasBeenOrWillBeClosed = WFCloseSupport.hasBeenOrWillBeClosed

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a value from a cache, unless it is an object that has been closed.
		If the value is an object that has been closed, removes if from the table
		and returns nil.
	
	\param cache The cache table.
	\param key The key for the value to retrieve.
	
	If <code>cache[ key ]</code> points to an object that has been closed, remove
	that entry from the table and return \c nil.
	
	\return The value of the given key, or nil if that value was a closed object.
]]--

function WFCloseSupport.fetchCached( --[[table]] cache,  --[[value]] key )

	local value = cache[ key ]
	
	if value and WFCloseSupport_hasBeenOrWillBeClosed( value ) then
		cache[ key ] = nil
		value = nil
	end
	
	return value

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns a function that maintains a fully weak table mapping single
		parameter values to single result which must be an object that
		has close support.  If the result is closed then we do not return
		the cached result, but instead run fn( x ) again to generate another.
	
	\param fn the function accepting a single parameter that generates the result to be cached
	
	\see WFCloseSupport.fetchCached
	\return The value of the given key, or nil if that value was a closed object.
]]--

function WFCloseSupport.memoizeFn( --[[function]] fn )
	local memoizer = memoize( fn )
	return function( x )
		local value = memoizer( x )
		if value and WFCloseSupport_hasBeenOrWillBeClosed( value ) then
			memoizer( kForget, x )
			value = memoizer( x )
		end
		return value
	end
end

--------------------------------------------------------------------------------

local debugEnqueueHook = nil
	-- A hook for seeing when things get enqueued.

--------------------------------------------------------------------------------
--[[!
	\brief Enqueues an object for closing, but does not run the close queue.
	Currently used only by \c KSPromises implementation. Please contact MarkH
	before using this API.

	\param target The object to enqueue.
	
	\sa WFCloseSupport.runCloseQueue
]]--

function WFCloseSupport.enqueueTargetForClose( --[[WFCloseTarget]] target )

	if target and not closedObjects[ target ] then
		if debugEnqueueHook then
			debugEnqueueHook( target )
		end
		closeStackTop = closeStackTop + 1
		closeStack[ closeStackTop ] = target
		closedObjects[ target ] = kQueued
	end

end

--------------------------------------------------------------------------------

local enqueueTargetForClose = WFCloseSupport.enqueueTargetForClose

--------------------------------------------------------------------------------

local function enqueueCloseTargets( self )

	-- Note that if enqueue fails due to lack of memory we bail on the whole close
	-- process, but at that point we were in pretty bad shape anyway or we are closing
	-- a ridiculously large network of objects.
	
	local closeTargets = forwardCloseTargetMap[ self ]
	
	if closeTargets then

		forwardCloseTargetMap[ self ] = nil
		
		for target in pairs( closeTargets ) do
			enqueueTargetForClose( target )
		end
		
	end
		
end

--------------------------------------------------------------------------------

local closingObject = nil
	-- The object currently being closed. We mark it closed in the event of
	-- an error. This is basically a status value for processCloseStack to
	-- use in the event that doCloseTargetsInStack fails.

--------------------------------------------------------------------------------

local WFCloseSupport_closeMethod -- forward declaration

--------------------------------------------------------------------------------

-- Close the targets clearing them from the table as we go. This clears first
-- so that in the event of an error (not caught here) we can safely restart.

local function doProcessCloseStack()

	while 0 < closeStackTop do
	
		repeat -- continue support
	
			-- Pop it
		
			local entry = closeStack[ closeStackTop ]
			closeStack[ closeStackTop ] = nil
			closeStackTop = closeStackTop - 1
			
			-- Check to see whether we've already closed and if so don't come
			-- this way again.
			
			local status = closedObjects[ entry ]
			
			if status == kClosed then
				break -- continue
			end
			
			-- Invoke the operation
			
			closingObject = entry
		
			if type( entry ) == 'function' then
				entry()
			else
				local eclose = entry.close or WFCloseSupport_closeMethod
				eclose( entry )
			end
			
			-- We no longer need to remember that this object was in the midst of being closed.
			
			closingObject = nil
			
			-- Mark it as closed though generally this also gets handled in the
			-- close method.
			
			closedObjects[ entry ] = kClosed
			
		until true -- continue support

	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Closes all of the objects that have been queued for closure.

	Currently used only by \c KSPromises implementation. Please contact MarkH
	before using this API.

	\sa WFCloseSupport.enqueueTargetForClose()
	
]]--

function WFCloseSupport.runCloseQueue()

	if processingCloseStack then
		return
	end
	
	processingCloseStack = true
	
	-- Repeatedly pcall into doCloseTargetsInStack. Errors result in closing the
	-- object that was in the process of closing and picking up where we left
	-- off.

	repeat
		local success = pcall( doProcessCloseStack )
		if closingObject then
			closedObjects[ closingObject ] = kClosed
			closingObject = nil
		end
	until success
	
	processingCloseStack = false

end

--------------------------------------------------------------------------------

local processCloseStack = WFCloseSupport.runCloseQueue

--------------------------------------------------------------------------------

local supportsCloseTypes = {
	[ 'table' ] = true,
	[ 'userdata' ] = true,
}

local function supportsClose( target )
	return supportsCloseTypes[ type( target ) ] == true
end

--------------------------------------------------------------------------------

local function addForwardMapEntry( source, target, noRemove )

	local closeTargets = forwardCloseTargetMap[ source ]
	
	if not closeTargets then
		closeTargets = setmetatable( { }, weakSetMT )
		forwardCloseTargetMap[ source ] = closeTargets
	end
	
	-- Make sure we aren't adding the same object multiple times since that would
	-- make remove unreliable. (The other choice would be to disallow removal.)
	
	if ( not noRemove ) and closeTargets[ target ] then
		debug.print( string.format( "Can only add a close target once: %s -> %s",
			tostring( source ), tostring( target ) ) )
		debug.print( debug.traceback() )
		error( "Can only add a close target once.", 2 )
	end
	
	closeTargets[ target ] = target

end

--------------------------------------------------------------------------------

local function removeForwardMapEntry( source, target )

	local closeTargets = forwardCloseTargetMap[ source ]
	
	if not closeTargets then
		return
	end
	
	closeTargets[ target ] = nil

end

--------------------------------------------------------------------------------
--[[!
	\brief (deprecated) Adds a close target to this object.
	
	\deprecated Bypassing the \c KSClass system for creating close targets is
		no longer supported. The only remaining use of this syntax is in
		\c KSReactiveBindingTest. Once that is gone, this function will become
		non-public. Use <code>WFCloseSupport.addCloseSupportMethods</code> to
		create a full \c WFCloseTarget instance instead.
--]]

function WFCloseSupport.addCloseTargetMethod( --[[WFCloseTarget]] self,
											  --[[WFCloseTarget]] target,
											  --[[optional boolean]] noRemove )

	if target == nil then
		return self
	end
	
	if not self then
		debug.print( "addCloseTarget: self is not a viable close source:\n" .. debug.traceback() )
		error( "addCloseTarget: self is not a viable close source", 2 )
	end
	
	if closedObjects[ self ] then
		debug.print( "addCloseTarget: source for close has already been closed:\n" .. debug.traceback() )
		error( "addCloseTarget: source for close has already been closed", 2 )
	end
	
	if WFCloseSupport_hasBeenOrWillBeClosed( target ) then
		return self
	end
	
	if not supportsClose( target ) then
		debug.print( "addCloseTarget: target unsupported for close: " .. tostring( target ) .. "\n" .. debug.traceback() )
		error( "addCloseTarget: target unsupported for close", 2 )
	end
	
	addForwardMapEntry( self, target, noRemove )
	
	return self

end

local WFCloseSupport_addCloseTargetMethod = WFCloseSupport.addCloseTargetMethod

--------------------------------------------------------------------------------

local function removeCloseTargetMethod( self, target )

	if target == nil then
		return self
	end
	
	removeForwardMapEntry( self, target )
	
	return self

end

--------------------------------------------------------------------------------

local function addAsCloseTargetOfMethod( self, attachTo )

	WFCloseSupport_addCloseTargetMethod( attachTo, self )
	
	return self

end

--------------------------------------------------------------------------------

local function addCloseActionMethod( self, actionFn )

	WFArguments_check( 'table', 'callable' )

	if actionFn == nil then
		return self
	end
	
	if closedObjects[ self ] then
		debug.print( "addCloseAction: source for close has already been closed:\n" .. debug.traceback() )
		error( "addCloseAction: source for close has already been closed", 2 )
	end
	
	if closedObjects[ actionFn ] then
		debug.print( "addCloseAction: action for close has already been run or queued:\n" .. debug.traceback() )
		error( "addCloseAction: action for close has already been run or queued", 2 )
	end
	
	-- Wrap the action up inside of a function and hang it off of the root object.
	-- It clears the link when it closes so that anything retained by the
	-- action function can be collected.
	
	local actionObj
	
	actionObj = {
		close = function() actionFn(); rawset( self, actionObj, nil ) end
	}
	
	WFCloseSupport_addCloseTargetMethod( self, actionObj )
	
	rawset( self, actionObj, actionObj )
	
	return self

end

--------------------------------------------------------------------------------

local function bounceClosureMethod( self )

	rawset( self, 'close', close_nop )
	
	return self

end

--------------------------------------------------------------------------------
--[[!
	\brief (deprecated) Closes this object. If it has already been closed removes
	it as a close target and marks it as closed to avoid subsequent passes.
	
	Called at the start of the standard close methods.
	The standard usage pattern looks like:
	
\pre
function MyObj:close()

	if not WFCloseSupport.willClose( self ) then
		return
	end

	-- Do the standard close logic because we haven't already
	-- been closed.

end
\pre
	
	\param obj The object to close.

	\return \c false if the object has already been closed,
		\c true if the object will be closed as a result of this call.

	\deprecated This is currently used by the Objective-C bridge code and
		that usage will be reviewed later. Calling the <code>:close()</code>
		method on the close target directly is preferred.
--]]

function WFCloseSupport.willClose( --[[WFCloseTarget]] obj )

	if not closeGuard( obj ) then
		return false
	end
	
	-- Enqueue the close targets for this object.
	
	enqueueCloseTargets( obj )
	
	-- Process the close stack. This is known not to fail. It returns quickly if
	-- we're already processing.
	
	processCloseStack()
	
	return true

end

local WFCloseSupport_willClose = WFCloseSupport.willClose

--------------------------------------------------------------------------------
--[[!
	\brief (deprecated) Equivalent of <code>self:close()</code>, except that it can work on
	objects or values that are not \c WFCloseTarget instances. (In that case,
	it is a no-op.)
	
	\deprecated Bypassing the \c KSClass system for creating close targets is
		no longer supported. The only remaining use of this syntax is in
		\c KSReactiveBindingTest. Once that is gone, this function will become
		non-public. Use <code>WFCloseSupport.addCloseSupportMethods</code> to
		create a full \c WFCloseTarget instance instead.
]]--

function WFCloseSupport.closeMethod( --[[WFCloseTarget]] self )

	-- By default, when we close an object, we just need to spread that closure to
	-- its dependents (and remove it as a close target from anything where it
	-- currently is a close target). willClose handles all of this work for us.
	-- We wrap it in another function to hide the result of willClose.

	if WFCloseSupport.willClose( self ) then
		stompClosedObjectMetatable( self )
	end
	
end

WFCloseSupport_closeMethod = WFCloseSupport.closeMethod

--------------------------------------------------------------------------------
--[[!
	\brief Defines the close-target methods for a new class.

	\param methodTable A table to which to add the function pointers.
		This table can contain an override implementation for the
		\c WFCloseTarget:close() method, but must not contain any
		other method implementations.
	
	\sa WFCloseTarget::close()
]]--

function WFCloseSupport.addCloseSupportMethods( --[[table]] methodTable )

	assert( methodTable.addCloseTarget == nil, "Already had addCloseTarget method" )
	assert( methodTable.removeCloseTarget == nil, "Already had removeCloseTarget method" )
	assert( methodTable.addAsCloseTargetOf == nil, "Already had addAsCloseTargetOf method" )
	
	assert( methodTable.addCloseAction == nil, "Already had addCloseAction method" )
	assert( methodTable.bounceClosure == nil, "Already had methodTable.bounceClosure method" )
	
	methodTable.addCloseTarget = WFCloseSupport_addCloseTargetMethod
	methodTable.removeCloseTarget = removeCloseTargetMethod
	methodTable.addAsCloseTargetOf = addAsCloseTargetOfMethod

	methodTable.addCloseAction = addCloseActionMethod
	
	methodTable.bounceClosure = bounceClosureMethod
	
	local oldClose = methodTable.close
	
	if oldClose then

		methodTable.close = function( self )
		
			-- We need to wrap oldClose in the willClose logic so that we
			-- only close once.
					
			if WFCloseSupport_willClose( self ) then
				oldClose( self )
				stompClosedObjectMetatable( self )
			end
			
		end

	else

		methodTable.close = WFCloseSupport_closeMethod

	end
	
	return methodTable

end

--------------------------------------------------------------------------------

local function close( target )

	if not target then
		return
	end
	
	assert( supportsClose( target ), "WFCloseSupport.close: target does not support methods" )
	
	-- Queue up the object and start running the close logic.
	
	enqueueTargetForClose( target )
	
	processCloseStack()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Intended for debugging use only.
]]--

function WFCloseSupport._debug_setEnqueueHook( --[[function]] hookFn )

	debugEnqueueHook = hookFn

end


--============================================================================--

local gcQueue = nil

--------------------------------------------------------------------------------

local function processGCQueue()

	if not gcQueue then
		return false
	end
	
	local q = gcQueue
	
	gcQueue = nil
	
	for i = 1, #q do
		local obj = q[ i ]
		if not WFCloseSupport_hasBeenOrWillBeClosed( obj ) then
			close( obj )
		end
	end
	
	return true
	
end

WFCloseSupport.processGCQueue = processGCQueue

--------------------------------------------------------------------------------

local function enqueueFromGC( obj )

	if not gcQueue then
		gcQueue = { }
		if not WFIdle_lazy then
			WFIdle_lazy = require 'WFIdle'
		end
		WFIdle_lazy.onIdleDo( processGCQueue )
	end
	
	gcQueue[ #gcQueue + 1 ] = obj

end

--============================================================================--

-- We create a userdata proxy for the purpose of closing the request when
-- it gets collected or when requested to close. It stores the reference to
-- the request within the environment table.

--------------------------------------------------------------------------------

local objectCloserFactory, objectCloserTest, _, _, objectCloserMT =
	KSUserdataProxy.makeFactory( "WFCloseSupport.mustClose" )

--------------------------------------------------------------------------------

local function objectCloser_doClose( self, fromGC )

	local proxy, env = objectCloserTest( self )
	
	if proxy and env then
		local object = env._object
		if object then
			env._object = nil
			if not WFCloseSupport_hasBeenOrWillBeClosed( object ) then
				if fromGC and DEBUG then
					local label = env._label
					if label then
						if printDebug then
							debug.print( string.format( "Closing target due to GC (%s): %s",
											tostring( label ), tostring( object ) ) )
						end
					else
						if printDebug then
							debug.print( string.format( "Closing target due to GC: %s",
											tostring( object ) ) )
						end
					end
					local creationBacktrace = env._creationBacktrace
					if creationBacktrace and printDebug then
						debug.print( "Object was created at: ", creationBacktrace )
					end
				end
				if fromGC then
					enqueueFromGC( object )
				else
					close( object )
				end
			end
		end
	else
		debug.print( "Object closer failed type test" )
	end

end

--------------------------------------------------------------------------------

function objectCloserMT.__gc( self )
	objectCloser_doClose( self, true )
end

--------------------------------------------------------------------------------

-- We also want to allow explicit closure.

objectCloserMT.__index = {
	close = function( self ) objectCloser_doClose( self, false ) end
}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a user-data object that will close a given object when it gets
	garbage collected.
	
	The \c mustClose() logic is based on creating an object with a \c __gc metamethod
	that closes the target object. We don't actually have to just attach this to the
	target object and instead can create such a closer object "free-floating" such
	that we can either close it via the closer or via the closer getting collected.
	Again, be cautious about weak tables since those won't keep such closer objects
	from getting invoked. Note also that a closer represents a strong link to the
	object and hence hanging onto the closer hangs onto the object.

	\param object The object to be closed on GC.
	\param label A name for the closer object.
	\param creationBacktrack If supplied a backtrace. Otherwise one will be captured
			in debug builds.
]]--

function WFCloseSupport.makeCloser(
					--[[WFCloseTarget]] object,
					--[[string]] label,
					--[[string]] creationBacktrace )
									
	if not WFIdle_lazy then
		WFIdle_lazy = require 'WFIdle'
	end
	
	if DEBUG and not creationBacktrace then
		creationBacktrace = debug.traceback()
	end
	
	return objectCloserFactory {
				_object = object,
				_label = label,
				_creationBacktrace = creationBacktrace
			}

end

local makeObjectCloser = WFCloseSupport.makeCloser

--------------------------------------------------------------------------------

local keyMustClose = KSPrivateKeys.makePrivateKey( 'WFCloseSupport.mustClose' )

--------------------------------------------------------------------------------

local rootCloseClass = KSClass 'WFRootCloseTarget' {
	WFCloseSupport.addCloseSupportMethods,
}

local rootCloseObject = rootCloseClass()

--------------------------------------------------------------------------------
--[[!
	\brief Forces an object to be closed rather than simply GC'd.
	
	Use sparingly, but it can be helpful. Note that it can be dangerous when used
	with weak tables since the finalizer will get invoked when the object is
	only weakly referenced so it essentially undermines the weak table usage.
	Furthermore, because the object will be referenced from the closer object
	(see \c makeCloser()), it may not actually get removed from the weak table and
	you need to check it with \c hasBeenOrWillBeClosed() or simply use
	\c fetchCached().

	\param obj The \c WFCloseTarget object to be closed at thread termination.
]]--

function WFCloseSupport.mustClose( --[[WFCloseTarget]] obj )

	assert( type( obj ) == 'table' )
	
	local closer = rawget( obj, keyMustClose )
	
	if not closer then
		closer = makeObjectCloser( obj, 'mustClose' )
		rawset( obj, keyMustClose, closer )
	end
	
	WFCloseSupport.mustCloseAtThreadTermination( obj )
	
	return closer

end

--------------------------------------------------------------------------------

WFCloseSupport.makeCloser = makeObjectCloser

--------------------------------------------------------------------------------
--[[!
	\brief Forces an object to close when the thread is terminated, if it has
		not been closed or GC'd earlier.
	
	This has less overhead than \c WFCloseSupport.mustClose.
	Recommended when you need a deterministic close sequence for objects.
	
	\param obj The \c WFCloseTarget object to be closed at thread termination.
--]]

function WFCloseSupport.mustCloseAtThreadTermination( --[[WFCloseTarget]] obj )

	assert( type( obj ) == 'table' )
	
	rootCloseObject:addCloseTarget( obj )
	
end

--------------------------------------------------------------------------------

WFCurrentThread.registerTerminateHandler( function()

	rootCloseObject:close()

end )

--------------------------------------------------------------------------------

return WFCloseSupport
