--[[----------------------------------------------------------------------------

WFDiagnosticLog.lua

@author scouten 2012-05-31

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'
local KSPromises = require 'KSPromises'
local KSTableUtils = require 'KSTableUtils'
--------------------------------------------------------------------------------

local diagnosticLogImp_lazy
local installedEnhancedMessages = {}

local function diagnosticLogImp()

	if not diagnosticLogImp_lazy then
		diagnosticLogImp_lazy = ( require 'WFWorkerThread' )( 'WFDiagnosticLogImp.lua' )
	end
	
	return diagnosticLogImp_lazy

end


--============================================================================--
--[[!
	\class WFDiagnosticLog
	\brief This namespace provides a way to capture a
		log file that can be used to diagnose user-reported problems.
	
	\author Eric Scouten <scouten@adobe.com> 2012-05-30
]]--

-- @mark WFDiagnosticLog namespace

local WFDiagnosticLog = {}

--------------------------------------------------------------------------------
--[[!
	\brief Generates an HTML diagnostic log file designed to be sent to
		Revel developers.

	\param args (table) A table of arguments to pass into the report, with these keys:
		<ul>
			<li> \c appName: (optional, string) Name of the application.</li>
			<li> \c emailTarget: (optional, string) Address of e-mail recipient.</li>
		</ul>
	
	\return <code>P< string ></code> An HTML log file for diagnostic reports.
	\return <code>B< number ></code> Percentage of progress toward generation of the report.
--]]

function WFDiagnosticLog.generateHtmlReportP( --[[table]] args )

	WFArguments.check( 'table' )

	-- Load these namespaces late so we don't pay for them if user doesn't request
	-- a diagnostic log.
	
	local WFTimer = require 'WFTimer'
	local KSDate = require 'KSDate'
	local KSReactive = require 'KSReactive'
	local WFCloseSupport = require 'WFCloseSupport'
	
	-- Get ready to show progress.
	
	local progressB = KSReactive.variableB( 0 )
	local startTime = KSDate.timeNowForAnimation()
	local reportingInterval = 1 -- wrong but it will do until we know real value
	
	local function updateProgressB()
	
		if WFCloseSupport.hasBeenOrWillBeClosed( progressB ) then return end
		
		local elapsedTime = math.min( KSDate.timeNowForAnimation() - startTime, reportingInterval )
		progressB:set( elapsedTime / reportingInterval )

		WFTimer.afterDelayDo( .5, updateProgressB )
	
	end
	
	-- How long is this going to take? (Chill out, King Julian!)
	
	local reportingIntervalP = diagnosticLogImp():requestP( 'reportingInterval' )
	reportingIntervalP:whenResolvedDo( function( status, resultReportingInterval )
	
		if status == 'data' then
			reportingInterval = resultReportingInterval
			updateProgressB()
		end
	
	end ):attachTo( progressB )

	-- Now generate the actual log.

	local reportP = diagnosticLogImp():requestP( 'generateHtmlReportP', args )

	reportP:whenResolvedDo( function()
		progressB:set( 1 )
		progressB:close()
	end ):attachTo( progressB )
	
	return reportP, progressB

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates an event stream that fires when a diagnostic log is
		requested.
	
	The \c :onDataE() method for this event stream typically calls
	\c WFDiagnosticLog.addSectionToLog.

	\return The new event-stream object.
--]]

function WFDiagnosticLog.logDataRequestE()

	return ( require 'KSSystemNotifications' ).notificationStreamE 'WFDiagnosticLog_sendLogDataNow'

end

--------------------------------------------------------------------------------
-- FIXME [scouten 2012-06-04]: This should probably be in a utils file somewhere.
-- Maybe part of WFArguments?

local function assertType( v, wantType, msg )

	local tv = type( v )
	
	if tv == wantType then
		return v
	end
	
	local emsg
	
	if type( msg ) == 'string' then
		emsg = string.format( "Expected %s got %s (%s)", wantType, tv, msg )
	else
		emsg = string.format( "Expected %s got %s", wantType, tv )
	end
	
	error( emsg, 2 )

end

local kSimpleTypes = { [ 'string' ] = true, [ 'boolean' ] = true, [ 'number' ] = true }

local function isSimple( v )
	return kSimpleTypes[ type( v ) ]
end

local function stripNonTransitableValues( t, breakcycle )
	
	breakcycle = breakcycle or {}
	local dest = {}
	if breakcycle[ t ] then
		return dest
	end
	
	breakcycle[ t ] = true
	for k, v in pairs( t ) do
		
		if not isSimple( k ) then
			k = "< " .. tostring( k ) .. " >"
		end
		
		if type( v ) == 'table' then
			v = stripNonTransitableValues( v, breakcycle )
		elseif not isSimple( v ) then
			k = "< " .. tostring( v ) .. " >"
		end
		
		dest[ k ] = v
		
	end
	
	return dest
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a section to the diagnostic log, in response
		to a \c logDataRequestE event having been fired.
	
	\param args (table) A table of parameters for this section, with these keys:
		<ul>
			<li> \c name: (string) User-visible name of the section. Must be HTML escaped.</li>
			<li> \c id: (string) HTML anchor name of the section. Must be a valid HTML ID.</li>
			<li> \c weight: (number) Sort order of the section; larger numbers sink to the bottom
				of the list.</li>
			<li> \c body: (string) HTML body of the section. Must be valid HTML or a Lua table
				containing only scalar values, which are converted to HTML.</li>
			<li> \c outputForAutomation: (optional, boolean) When \c true, this section is included in
				the abridged JSON flavor of the diagnostic log that is used for automated tests.
				Enabling this property requires that the body be a Lua table rather than HTML.</li>
		</ul>
	
	\return <code>P< true ></code> A promise object that resolves when section
		has been added to log. Can be ignored.
--]]

function WFDiagnosticLog.addSectionToLog( --[[table]] args )

	if args == nil then return end

	WFArguments.check( 'table' )
	
	assertType( args.name, 'string', 'name' )
	assertType( args.id, 'string', 'id' )
	assertType( args.weight, 'number', 'weight' )
	
	local bodyType = type( args.body )
	if bodyType ~= 'string' and bodyType ~= 'table' then
		error( string.format( "Expected string or table got %s (body)", bodyType ), 2 )
	end
	
	-- If body is a table, convert behaviors and promises at root level of table
	-- to actual values. Wait 1 second to give time for that to occur.
	
	local promiseOrBehaviorPopulationDelayP = nil
	local promiseOrBehaviorPopulationDelayAction = nil
	
	local body = args.body
	
	local KSPromises = require 'KSPromises'
	local KSReactive = require 'KSReactive'
	local WFTimer = require 'WFTimer'

	if bodyType == 'table' then
	
		for k, v in pairs( body ) do
		
			local tv = type( v )
			
			if tv == 'table' then

				local isEventNode, isBehavior = KSReactive.isEventNode( v )
				if isBehavior then
					
					if not promiseOrBehaviorPopulationDelayP then
						promiseOrBehaviorPopulationDelayP = WFTimer.afterDelayP( 1 )
						promiseOrBehaviorPopulationDelayAction
							= promiseOrBehaviorPopulationDelayP:runUntilDone()
					end
					
					local myK = k
					
					v:onDataB( function( newValue )
						if newValue == nil then
							body[ myK ] = "(nil)"
						else
							body[ myK ] = newValue
						end
					end ):addAsCloseTargetOf( promiseOrBehaviorPopulationDelayAction )
					
				elseif isEventNode then
					
					body[ k ] = "(event nodes not permitted in diagnostic log)"
				
				elseif KSPromises.isPromise( v ) then
				
					local myK = k
					local myPromise = v
					
					body[ myK ] = "(promise did not resolve in time)"
				
					if not promiseOrBehaviorPopulationDelayP then
						promiseOrBehaviorPopulationDelayP = WFTimer.afterDelayP( 1 )
						promiseOrBehaviorPopulationDelayAction
							= promiseOrBehaviorPopulationDelayP:runUntilDone()
					end

					myPromise:whenResolvedDo( function( _, newValue )
						if newValue == nil then
							body[ myK ] = "(nil)"
						else
							body[ myK ] = newValue
						end
					end ):runUntilDone() -- lame
						-- :attachTo( promiseOrBehaviorPopulationDelayAction )
						-- Why doesn't this work???

				end
			
			end
		
		end
	
	end

	if promiseOrBehaviorPopulationDelayP == nil then
		promiseOrBehaviorPopulationDelayP = KSPromises.resolvedPromiseP( 'data', true )
	end
	
	local resultP = promiseOrBehaviorPopulationDelayP
		:mapP( function()
			return diagnosticLogImp():requestP( 'addSectionToLog', stripNonTransitableValues( args ) )
		end )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds a function to generate a log section. Shortcut version for
		combined calls to \c logDataRequestE() and \c addSectionToLog().
	
	\param logSectionFunc (function) A function that is called when a diagnostic
		log is requested and returns a table in the format required by
		\c addSectionToLog()

	\sa  WFDiagnosticLog.addSectionToLog
--]]

function WFDiagnosticLog.addLogSectionFunc( --[[function]] logSectionFunc )

	WFDiagnosticLog.logDataRequestE()
		:onDataE( function()
		
			WFDiagnosticLog.addSectionToLog( logSectionFunc() )
		
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds an object that generates a log section if it is still open
		when a diagnostic log is requested.
	
	\param logProvider (object) This object must have a \c sendDiagnosticLogData
		method and must also be a \c WFCloseTarget.
--]]

function WFDiagnosticLog.addLogProvider( --[[object]] logProvider )

	if logProvider == nil then return end

	local logDataRequestE = WFDiagnosticLog.logDataRequestE()
	
	logDataRequestE:onDataE( function()
		
		WFDiagnosticLog.addSectionToLog( logProvider:sendDiagnosticLogData() )
	
	end )

	logProvider:addCloseTarget( logDataRequestE )

end

--------------------------------------------------------------------------------
--[[!
	\brief Enhance the built-in diagnostic facilities with additional logic.
	
	\param namespace (string) This namespace will be required on the diagnostic
	log's dedicated message processor.
	\return A promise (boolean) for success.  The promise will runUntilDone.
--]]

function WFDiagnosticLog.registerDiagnosticUtilityP( namespace )
	WFArguments.check( "string" )
	
	return diagnosticLogImp():requestP( 'registerDiagnosticUtility', namespace )
		:mapDataP( function( status, installedMessages )
			installedEnhancedMessages = KSTableUtils.mapArrayToSet( installedMessages )
			return status, true
		end )
		:runUntilDoneP()
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Query an enhanced diagnostic utility that was installed in the diagnostic
	log's dedicated message processor via registerDiagnosticUtility.
	
	\param request name (string) The request name, as specified by WFWorkerThread:requestP.
	\param ... additional parameters to send
	\return a promise with the values returned by the enhanced utility.  It will not
	automatically run until done.
--]]

function WFDiagnosticLog.makeEnhancedRequestP( message, ... )
	WFArguments.check( "string" )
	
	if installedEnhancedMessages[ message ] then
		return diagnosticLogImp():requestP( message, ... )
	else
		return KSPromises.resolvedPromiseP( 'data', false )
	end
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Is this a simple value: a number, a boolean or a string.
	
	\return True if it is simple.  False otherwise.  Note that nil is NOT considered
	simple for the purposes of generating diagnostic logs.
--]]

function WFDiagnosticLog.isSimpleValue( v )
	return isSimple( v )
end

--------------------------------------------------------------------------------
--[[!
	\brief Copy the simple values key/value pairs.  If either the key or the value is
	a complex value (a table, userdata, function, or coroutine) then the pair is omitted.
	
	Useful for dumping a namespace into a diagnostic log, when the namespace might potentially
	grow over time to gain values that can't be transited simply.
	
	\return The copy.
--]]

function WFDiagnosticLog.shallowCopyOfSimpleValues( v )
	
	if isSimple( v ) then
		return v
	end
	
	if type( v ) == 'table' then
		local values = {}
		
		for key,value in pairs( v ) do
			if isSimple( key ) and isSimple( value ) then
				values[ key ] = value
			end
		end
		return values
		
	end
	
	return nil
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Build a diagnostic log entry that contains already formatted data
	(as HTML and JSON)
	
	\param html (string) HTML representation
	\param data (table) lua type represenation (for use when diagnostic log
		is generated as alternate representaion (e.g. JSON)
		
	\return A special table containing keys that inform the diagnostic log
	to build this entry as the HTML representation (when possible).
--]]

function WFDiagnosticLog.buildPreformattedHtmlEntry( htmlString, optLuaAlternateForm )
	return {
		[ "< preformatted html >" ] = htmlString,
		[ "< raw data >" ] = optLuaAlternateForm or "no data form available",
	}
end

--------------------------------------------------------------------------------
--[[!
	\brief Insert a sparkline into the HTML output (or just the array of numbers
		into the JSON format)
	
	\param title (string) title for the sparkline
	\param arrayOfNumbers (table) an array of numbers
		
	\return A special table containing keys that inform the diagnostic log
	to build this entry as the HTML canvas sparkline.
--]]

function WFDiagnosticLog.sparkline( title, arrayOfNumbers )
	return {
		title = title,
		[ "< sparkline >" ] = arrayOfNumbers,
	}
end

--------------------------------------------------------------------------------
--[[!
	\brief Insert a box-and-whisker graph into the HTML output (or just the array of numbers
		into the JSON format)
	
	\param title (string) title for the box-and-whisker
	\param arrayOfNumbers (table) an array of numbers
		
	\return A special table containing keys that inform the diagnostic log
	to build this entry as the HTML canvas box-and-whisker.
--]]

function WFDiagnosticLog.boxAndWhisker( title, arrayOfNumbers )
	return {
		title = title,
		[ "< box-and-whisker >" ] = arrayOfNumbers,
	}
end

--------------------------------------------------------------------------------
--[[!
	\brief Create a ring buffer for storing a series of values.  Suitable
	for feeding into a sparkline, for example.
	
	\param options (table)
		size : size of the ring buffer (default 1000)
		accumulate : add new values to previous ones
		mapFn : a function to run values thru
	\param title (string) title for the data
	\param ... titles for other fields
		
	\return WFDiagnosticRingBuffer object which contains add( num1, num2, num3 ) fn and
	visualize( 'sparkline', 'boxAndWhisker') fn.
--]]

function WFDiagnosticLog.makeRingBuffer( options )
	options = options or {}
	local size = options.size or 1000
	local accumulate = options.accumulate or false
	local mapFn = options.mapFn
	local ringBuffer = {
		head = 0,
	}
	
	-- strip down to just array elements
	local titles = { unpack( options, 1, #options ) }
	
	local function historyNodes( fieldIndex )
		local t = {}
		local idx = 1
		for _, idxs in ipairs { { ringBuffer.head + 1, #ringBuffer }, { 1, ringBuffer.head } } do
			for i = idxs[ 1 ], idxs[ 2 ] do
				local val = ringBuffer[ i ][ fieldIndex ]
				if val then
					t[ idx ] = val
					idx = idx + 1
				end
			end
		end
		return t
	end
	
	local result = {
		add = function( self, ... )
			local head = ringBuffer.head
			local lastEntry = head > 0 and ringBuffer[ head ]
			local values
			if mapFn then
				values = { mapFn( ... ) }
			else
				values = { ... }
			end
			for i = 1, #titles do
				if accumulate and lastEntry and lastEntry[ i ] then
					values[ i ] = ( values[ i ] or 0 ) + lastEntry[ i ]
				end
			end
			head = head + 1
			if head > size then
				head = 1
			end
			ringBuffer.head = head
			ringBuffer[ head ] = values
		end,
		
		visualize = function( self, kind )
			local result = {}
			local fn = kind == 'sparkline' and WFDiagnosticLog.sparkline or
				kind == 'boxAndWhisker' and WFDiagnosticLog.boxAndWhisker
			for i = 1, #titles do
				table.insert( result, fn( titles[ i ], historyNodes( i ) ) )
			end
			return unpack( result )
		end,
	}
	
	return result
	
end

--------------------------------------------------------------------------------

return WFDiagnosticLog
