--[[----------------------------------------------------------------------------

WFDocStore.lua

@author scouten 2013-04-01

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.utils
local KSClass = require 'KSClass'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFDocStore class

--------------------------------------------------------------------------------
--[[!
	\class WFDocStore

	This class implements a generic document store. It can be thought of as
	a key-value database where the key is any arbitrary (non-empty) string and
	the value is a Lua table, where all values are of the following types:
	
		* boolean
		* number
		* string
		* table
	
	Table keys must be non-empty strings. (If strings, must not contain leading
	underscores.) There is one exception: A document may contain a top-level key
	named \c _localOnly, which is intended to store document content that is
	device-specific and not synchronized with other databases. (Note, however,
	that this is a convention: \c WFDocStore intentionally has no knowledge of
	synchronization.)
	
	Tables must honor the key and value restrictions recursively and may not
	contain metatables. Exception: Nested tables may use numeric values for keys.
	
	Tables must not contain reference cycles.
	
	It's probably worth reading http://guide.couchdb.org/editions/1/en/api.html#documents
	or other primers on CouchDB or Couchbase as a basis for understanding what's
	meant by our use of the term "document database." This database isn't
	implemented using CouchDB, but it is heavily influenced by CouchDB and
	you will understand many of the concepts behind this code if you understand
	CouchDB well.
	
	\c WFDocStore differs from CouchDB in that it allows individual documents
	to have named branches which represent alternative versions of the document
	content. Think of branches in the same sense as git branches, except that
	they apply to a single document at a time, instead of the entire database.
	When unspecified, the branch name will be "local".
]]--

local docStoreMethods = {

	_path = false,
		-- (string) path to doc store DB

	_docStoreImp = false,
		-- (WFWorkerThread) reference to the worker thread
	
	_docChangeCounterB = false,
		-- (B< number >) most recent doc change counter value. Initially nil.
		-- Lazily instantiated.

	_docValueBehaviors = false,
		-- (fully weak table) document value behaviors that are currently active
			-- key: (string) doc ID + slot name
			-- value: (B< table >) behavior for that document value

	_fieldValueBehaviors = false,
		-- (fully weak table) field value behaviors that are currently active
			-- key: (string) doc ID + slot name + field path
			-- value: (B< table >) behavior for that field value

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs a new document database.

	\param path (string) path to the document database.
	
		For testing, you can use the special ID ":memory:" which
		creates a transient catalog in memory, or ":temp:"
		which creates a transient catalog on disk that is deleted
		when the DB is no longer needed. You may optionally add any
		suffix following ":memory:" or ":temp:" to create distinct
		instances of temporary document databases that can be addressed
		from multiple threads.
		
		If no database exists at this path, it will be created.
		
		The path string must be non-empty.
]]--

function docStoreMethods:__init( --[[string]] path )

	check( 'WFDocStore', '+string' )

	WFCloseSupport.mustClose( self )

	-- Make sure :memory: and :temp: DBs don't conflict with each other.

	if path == ':memory:' or path == ':temp:' then
		path = path .. tostring( KSIdGenerator.localCounter() )
	end

	self._path = path

	self._docStoreImp = WFWorkerThread( "WFDocStoreMessageProcessor.lua", path )
	self:addCloseTarget( self._docStoreImp )

	self._docValueBehaviors = KSTableUtils.makeFullyWeak()
	self:addCloseTarget( self._docValueBehaviors )

	self._fieldValueBehaviors = KSTableUtils.makeFullyWeak()
	self:addCloseTarget( self._fieldValueBehaviors )

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the value of a single document.

	\param docId (string) the document's unique ID
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>P< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):
	
		* \c _id (string): the document's unique ID
		* \c _slot (string): name of the slot that was fetched
		* \c _hash (string): hash value of the slot's contents
		* \c _empty (boolean): true if there isn't actually a value for this doc ID + slot
		* \c _transient (boolean): true if the slot is marked as transient
		* \c _changeCounter (number): a monotonically-increasing change counter
			that is increased whenever any branch of this document is revised
			(The functions \c docChangeCounterB and \c docChangeCounterP will
			return the maximum value of \c _changeCounter across the entire
			database.)
]]--

function docStoreMethods:docValueP( --[[string]] docId, --[[optional, string]] optSlotName )

	check( 'WFDocStore', '+string', '?string' )
	
	return self._docStoreImp:requestP( 'docValue', docId, optSlotName )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this value of a single document as a behavior.
	
	The behavior will be updated reasonably quickly as changes come in from
	this and other threads.
	
	\warning You should avoid watching many documents via \c docValueB or
		\c fieldValueB simultaneously for performance reasons. Console logs
		will be generated if the number of \c docValueB instances is large.
		(Note that \c fieldValueB uses \c docValueB in its implementation,
		but the performance concern is around the number of \b documents,
		not the number of fields, being watched.)

	\param docId (string) the document's unique ID
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>B< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):
	
		* \c _id (string): the document's unique ID
		* \c _slot (string): name of the slot that was fetched
		* \c _hash (string): hash value of the slot's contents
		* \c _empty (boolean): true if there isn't actually a value for this doc ID + slot
		* \c _transient (boolean): true if the slot is marked as transient
		* \c _changeCounter (number): a monotonically-increasing change counter
			that is increased whenever any branch of this document is revised
			(The functions \c docChangeCounterB and \c docChangeCounterP will
			return the maximum value of \c _changeCounter across the entire
			database.)
]]--

function docStoreMethods:docValueB( --[[string]] docId,
									--[[optional, string]] optSlotName )
	
	check( 'WFDocStore', '+string', '?string' )

	local slotName = optSlotName or 'main'

	local docValueBehaviors = self._docValueBehaviors
	local key = string.format( '%s :: %s', docId, slotName )
	
	local resultB = WFCloseSupport.fetchCached( docValueBehaviors, key )

	if not resultB then

		resultB = self._docStoreImp:requestE( 'docValueB', docId, optSlotName )
			:holdB( WFUnknown.value() )

		docValueBehaviors[ key ] = resultB
		
	end
	
	return resultB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the document value, replacing any existing value.
	
	\param value (table) new value for the entire document
	
	\c value \b must contain the following elements:
		* \c _id (string) the document's unique ID
		* \c _slot (string) the slot name to write to (often "main")
	
	\c value \b may contain the following elements:
		* \c _transient (boolean): if true, the slot is marked as transient
		* \c _hash (string): if present, the current document value must
			have the given hash value or the request will fail
		* \c _changeCounter (number): if present, the current document value
			must have the given change counter value or the request will fail
		* \c _deleted (boolean): if present, must be \c true and no other
			values may be present
		* \c _syncStatus (table): reserved for use by sync engine only

	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:setDocValueP( value )

	check( 'WFDocStore', 'table' )
	
	return self._docStoreImp:requestP( 'setDocValue', value )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a behavior) the current value of a field in this document.
	
	\param docId (string) The document's unique ID.
	
	\param fieldPath (string) The field path, a slash-delimited string like
		a Unix paths, except that the root is implied. Use the path to
		identify values within tables or nested tables.	You can use number
		values to get the nth member of a table; for example, "tabA/1" to get
		the first member of an array in "tabA".

	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	The behavior will be updated reasonably quickly as changes come in from
	this and other threads. 
	
	\warning You should avoid watching many documents via \c docValueB or
		\c fieldValueB simultaneously for performance reasons. Console logs
		will be generated if the number of \c docValueB instances is large.
		(Note that \c fieldValueB uses \c docValueB in its implementation,
		but the performance concern is around the number of \b documents,
		not the number of fields, being watched.)
	
	\return <code>B< value ></code> A behavior whose value is
		the current value of the field.
]]--

function docStoreMethods:fieldValueB( --[[string]] docId,
									  --[[string]] fieldPath,
									  --[[optional, string]] optSlotName )

	check( 'WFDocStore', '+string', '+string', '?string' )

	local slotName = optSlotName or 'main'
	
	local key = string.format( '%s:%s:%s', docId, slotName, fieldPath )
	
	local resultB = WFCloseSupport.fetchCached( self._fieldValueBehaviors, key )

	if not resultB then

		local fieldPathSegments = {}
		local nSegments = 0
		
		fieldPath:gsub( '([^/]+)',
			function( segment )
				nSegments = nSegments + 1
				if segment:match( '^-?%d+$' ) then
					fieldPathSegments[ nSegments ] = tonumber( segment )
				else
					fieldPathSegments[ nSegments ] = segment
				end
			end )
		
		if nSegments == 0 then
			error( "WFDocStore:fieldValueB: fieldPath didn't contain a path", 2 )
		end
		
		resultB = self:docValueB( docId, slotName )
			:mapB( function( docValue )
	
				if WFUnknown.isUnknown( docValue )
					or( docValue and docValue._empty )
				then
					return WFUnknown.value()
				end
	
				local t = docValue
				
				-- All parent nodes must be tables. If they are anything else,
				-- we just pretend they don't exist since we can't index
				-- into them. (Note that we don't support userdata values
				-- in documents since they don't persist or translate cleanly
				-- to JSON.)
				
				for i = 1, nSegments - 1 do
					t = t and t[ fieldPathSegments[ i ] ]
					if type( t ) ~= 'table' then
						t = nil
					end
				end
	
				return t and t[ fieldPathSegments[ nSegments ] ]
			
			end )

		self._fieldValueBehaviors[ key ] = resultB
		
	end
	
	return resultB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the document value, applying an overlay to any existing value
	
	\param overlay (table) delta to apply to the existing document value
	
	\c overlay \b must contain the following elements:
		* \c _id (string) the document's unique ID
		* \c _slot (string) the slot name to write to (often "main")
	
	\c overlay \b may contain the following elements:
		* \c _transient (boolean) if true, the slot is marked as transient
		* \c _hash (string) if present, the current document value must
			have the given hash value or the request will fail
		* \c _changeCounter (number) if present, the current document value
			must have the given change counter value or the request will fail

	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:applyOverlayP( overlay )

	check( 'WFDocStore', 'table' )
	
	return self._docStoreImp:requestP( 'applyOverlay', overlay )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark managing document slot values

--------------------------------------------------------------------------------
--[[!
	\brief Copy a value from one slot to another.

	\param docId (string) The document's unique ID.
	\param fromSlot (string) The name of the slot to copy from
	\param toSlot (string) The name of the new slot
	\param transient (optional, boolean) true if the new slot should be a transient value
	
	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:copySlotToSlotP( docId, fromSlot, toSlot, transient )

	check( 'WFDocStore', '+string', '+string', '+string', '?boolean' )

	return self._docStoreImp:requestP( 'copySlotToSlot', docId, fromSlot, toSlot, transient )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Quietly delete slot.

	\param docId (string) The document's unique ID.
	\param slotName (string) The name of the slot to delete

	\warning This function is intended mostly for use by the sync engine. It
		\b quietly deletes the slot value. By that, we mean that the value
		simply disappears from the database without any change notification.
	
	It is not permitted to delete from the slot named "main".
	
	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:quietlyDeleteSlotP( docId, slotName )

	check( 'WFDocStore', '+string', '+string' )

	return self._docStoreImp:requestP( 'quietlyDeleteSlot', docId, slotName )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------
--[[!
	\brief Tracks changes in this document by incrementing a counter for each
		change. Be aware that this value can jump, especially if many documents
		are changed at once.
	
	\return <code>B< number ></code> the current maximum value for
		\c _changeCounter across all documents in this database
]]--

function docStoreMethods:docChangeCounterB()

	check( 'WFDocStore' )

	if not self._docChangeCounterB then
		self._docChangeCounterB = self._docStoreImp:requestE( 'docChangeCounterB' )
			:holdB( WFUnknown.value() )
	end
	
	return self._docChangeCounterB:relayB()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a promise) the current change count for this document.
		This count includes every operation that has completed to this point.
 
	\return <code>P< number ></code> the current maximum value for
		\c _changeCounter across all documents in this database
]]--

function docStoreMethods:docChangeCounterP()

	check( 'WFDocStore' )

	return self._docStoreImp:requestP( 'docChangeCounter' )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the set of documents that have changed since a specified point.

	\param args (table) A table with these keys:
		* \c fromChangeCounter (number): A previous change-counter value. Documents
			whose change counter is equal to or greater than this value can match
			this query.
		* \c limit (optional, number): The maximum number of documents to return
		* \c idsOnly (optional, boolean): If true, only document IDs are returned
		* \c slot (optional, string): If present, only documents whose value has
			changed in the specified slot will match this query. If not present,
			all document value changes will match this query.
		
	Note that regardless of any value specified in \c limit, a limit of 50 documents
	is always enforced. (The limit is raised to 1000 if \c idsOnly is true.)

	\return <code>P< table ></code> An array of fully-formed document values or IDs
		(as from \c :docValueP) that satisfy the changed-documents search criteria.
	
	This array/table also contains a top-level element named \c continueFrom
	whose value is the change-counter number to use	in the next \c :changedDocs call.
	That number may be less than the current value of \c :docChangeCounterB if
	the number of documents returned in this call was limited.
]]--

function docStoreMethods:changedDocsP( args )

	check( 'WFDocStore', 'table' )
	
	return self._docStoreImp:requestP( 'changedDocs', args )

end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the path to this document store.

	\return (string) The catalog's path.
]]--

function docStoreMethods:path()

	check( 'WFDocStore' )

	return self._path

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this document store as a string.

	\return (string) The string representation.
]]--

function docStoreMethods:__tostring()

	check( 'WFDocStore' )

	return string.format( "WFDocStore( %s )", self._path )

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for debugging use only) Generates code coverage statistics
		for the worker thread that implemented this doc store.

	\warning This function should only be used by WFDocStoreTest.lua.

	\return P< true > resolves when coverage data has been generated
]]--

function docStoreMethods:DEBUG_dumpCoverageP( path )

	return self._docStoreImp:requestP( 'DEBUG_dumpCoverage', path )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

return KSClass 'WFDocStore'( docStoreMethods )
