--[[----------------------------------------------------------------------------

WFDocStoreImp.lua

@author scouten 2013-04-01

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local KSLogging = require 'KSLogging'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.table
local KSTableArchiving = require 'KSTableArchiving'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local WFDigest = require 'WFDigest'
local WFTableMerge = require 'WFTableMerge'

	-- WF.sqlite.database
local KSDatabase = require 'KSDatabase'
local KSDatabaseSchema = require 'KSDatabaseSchema'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local sha1 = assert( WFDigest.SHA1 )

local docStoreImpLog = KSLogging.log 'WFDocStoreImp'
--	:setLevel( KSLogging.DEBUG )

local changedDocsLog = KSLogging.log 'WFDocStoreImp :: changedDocs'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark database schema

local schema = KSDatabaseSchema( '1.0' ):useWAL()

--------------------------------------------------------------------------------
--[[
	Content in this document database is primarily stored via hash reference.
	In other words, the hash of the content is the primary key used to find
	the content.
--]]
	
schema:addTable
	[[CREATE TABLE content( contentHash UNIQUE NOT NULL, value NOT NULL )]]
	:onConflict 'replace'

--------------------------------------------------------------------------------
--[[
	Content hash is probably not the most user-friendly way to find content,
	so we introduce the concept of documents and value slots. The "slot" concept
	allows us to store multiple alternative values for the document (for instance,
	from external sync sources). We assert that the doc store will store identical
	content efficiently (i.e. not duplicate content) if multiple slots hold the
	same value, and also offer create copy-on-write semantics when one slot's value
	evolves.
	
	Slots are a little bit like named branches in git, but there is no assumption
	that we preserve history or support merging of values across slots, so the
	analogy is not pure.
--]]

schema:addTable
	[[CREATE TABLE documents(
						docId NOT NULL,
						slotName NOT NULL,
						changeCounter NOT NULL,
						transient,
						contentHash NOT NULL )]]
	:onConflict 'replace'

schema:addIndex
	[[CREATE UNIQUE INDEX documents_docId_slotName ON documents( docId, slotName )]]

schema:addIndex
	[[CREATE INDEX documents_transient ON documents( transient )]]

schema:addIndex
	[[CREATE INDEX documents_contentHash ON documents( contentHash )]]

schema:addIndex
	[[CREATE INDEX documents_changeCounter ON documents( changeCounter DESC, slotName )]]

--------------------------------------------------------------------------------
--[[
	This table is used in the implementation of the specific docs
	version of changedDocs. It lets us bind to a variable number of items.
]]--

schema:addTable
	[[CREATE TEMPORARY TABLE TEMP_docIds( docId )]]


--============================================================================--

-- @mark WFDocStoreImp class

--------------------------------------------------------------------------------
--[[!
	\class WFDocStoreImp

	This class is considered \b INTERNAL to the implementation of \c WFDocStore.
	Please do not access this class directly.
]]--

local docStoreImpMethods = {

	_db = false,
		-- (KSDatabase) database file

	_exec = false,
		-- (function) cached reference to db:exec

	_path = false,
		-- (string) path to doc store DB

	_docChangeCounterB = false,
		-- (B< number >) most recent doc change counter value
		-- WARNING: This value is one lower than db:get/setChangeCounter().
		-- The semantics of the underlying KSDatabase code are wrong, but it's
		-- not possible to change that in a way that doesn't break OzDocStoreImp.

	_docValueBehaviors = false,
		-- (KSReactive.makeWeakEventTable) document value behaviors that are currently active
			-- key: (string) doc ID + slot name

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function docStoreImpMethods:__init( --[[string]] path )

	check( 'WFDocStoreImp', '+string' )

	-- Map :memory:(random) and :temp:(random) paths back to SQLite pseudo-paths.

	local dbPath = path

	if dbPath:match( '^:memory:' ) then
		dbPath = ':memory:'
	elseif dbPath:match( '^:temp:' ) then
		dbPath = ':temp:'
	end

	-- Create the actual database.
	
	self._path = dbPath
	docStoreImpLog:info( 'dbPath = %s', dbPath )

	local db = KSDatabase( dbPath, schema )
	self._db = db
	
	self:addCloseTarget( db )
	WFCloseSupport.mustCloseAtThreadTermination( self )

	self._exec = assert( db.exec )

	-- Delete transient values.
	
	self:deleteTransientValues()

	-- Set up change notification.

	local changeCounterSeedRow = db:execOnce( "SELECT max( changeCounter ) FROM documents" )

	local currentChangeCounter = ( changeCounterSeedRow
		and changeCounterSeedRow[ "max( changeCounter )" ] or 0 )
	
	db:setChangeCounter( currentChangeCounter + 1 )

	self._docChangeCounterB = KSReactive.variableB( currentChangeCounter )
	self:addCloseTarget( self._docChangeCounterB )
	
	self._docValueBehaviors = KSReactive.makeWeakEventTable( nil )
	self:addCloseTarget( self._docValueBehaviors )

--	WFDiagnosticLog.addLogProvider( self ) -- add later

end

--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local function updateChangeCounter( self )

	local newChangeCounter = self._db:getChangeCounter()

	changedDocsLog:info( "newChangeCounter = %s", newChangeCounter )
	self._docChangeCounterB:set( newChangeCounter - 1 )
	
	return newChangeCounter
	
end

--------------------------------------------------------------------------------

local function maybeGCOldContentValue( self, oldContentHash )

	-- At the point when this function is called, we have a content hash
	-- that may no longer be referenced. We want to get rid of the corresponding
	-- content unless some other reference still exists.

	-- No-op if old has was nil.

	if not oldContentHash then
		return
	end

	-- Is there an existing reference?
	
	local existingRow = self._exec( self._db,
		[[SELECT 1 FROM documents WHERE contentHash == ? LIMIT 1]], oldContentHash )
	
	if existingRow then
		return
	end
	
	-- No existing reference: Delete.
	
	self._exec( self._db, [[DELETE FROM content WHERE contentHash == ?]], oldContentHash )
	
end

--------------------------------------------------------------------------------

local underscoreByte = string.byte( '_' )

local unacceptableTypes = {
	[ 'function' ] = true,
	[ 'thread' ] = true,
	[ 'userdata' ] = true,
}

--------------------------------------------------------------------------------

local approvedKeys = {
	_syncStatus = true,
}

--------------------------------------------------------------------------------

local function assertDocValueIsValid( value, isSubtable )

	if getmetatable( value ) then
		error "table contains metatable"
	end
	
	if not isSubtable then
		if value._deleted == true then
			for k in pairs( value ) do
				if k ~= '_deleted' then
					error( string.format( "key %s not allowed if _deleted == true",
										  k ) )
				end
			end
			return
		end
	end
	
	for k, v in pairs( value ) do
	
		local tk = type( k )
	
		if tk ~= 'string' and ( tk ~= 'number' or not isSubtable ) then
			error( string.format( "key %s is of type %s, must be string%s",
								  tostring( k ), tk,
								  isSubtable and " or number" or "" ) )
		end
		
		if tk == 'string' then

			if #k == 0 then
				error "value contains empty string as key"
			end
		
			if k:byte( 1 ) == underscoreByte and not approvedKeys[ k ] then
				error( string.format( "value contains key with leading underscore (%s)", k ) )
			end

		end
		
		local tv = type( v )
		
		if unacceptableTypes[ tv ] then
			error( string.format( "value at key %s is of type %s (must be scalar or table)",
								  k, tv ) )
		end
		
		if tv == 'table' then
			assertDocValueIsValid( v, true )
		end
	
	end

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------

function docStoreImpMethods:docValue( docId, optSlotName )

	check( 'WFDocStoreImp', '+string', '?string' )

	local slotName = optSlotName or 'main'

	-- Find value.

	local existingRow = self._exec( self._db,
		[[SELECT value, documents.contentHash, transient, changeCounter
			FROM documents INNER JOIN content USING( contentHash )
			WHERE docId == ? AND slotName == ?]], docId, slotName )
	
	-- If no such value, hack up an empty document.
	
	if not existingRow then
		return {
			_id = docId,
			_slot = slotName,
			_empty = true,
		}
	end

	local docValue = loadstring( "return " .. existingRow.value )()
	
	docValue._id = docId
	docValue._changeCounter = existingRow.changeCounter
	docValue._hash = existingRow.contentHash
	docValue._slot = slotName
	
	if existingRow.transient == 1 then
		docValue._transient = true
	end
	
	return docValue
	
end

--------------------------------------------------------------------------------

local function docValueBehaviorKey( docId, slotName )

	return string.format( '%s :: %s', docId, slotName )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:docValueB( docId, optSlotName )

	check( 'WFDocStoreImp', '+string', '?string' )

	local slotName = optSlotName or 'main'

	local docValueBehaviors = self._docValueBehaviors
	local key = docValueBehaviorKey( docId, slotName )
	
	if docValueBehaviors:eventExists( key ) then
	
		return docValueBehaviors:behaviorB( key ):relayB( true )
	
	else
	
		-- Create new behavior, then post current doc value to it
		-- so we know we're up to date.
	
		local resultB = docValueBehaviors:behaviorB( key ):relayB( true )
		
		docValueBehaviors:postToEvent( key, self:docValue( docId, slotName ) )
		
		return resultB
	
	end

end

--------------------------------------------------------------------------------

function docStoreImpMethods:setDocValue( value )

	check( 'WFDocStoreImp', 'table' )

	-- Extract doc ID from the value.
	
	local docId = value._id
	value._id = nil
	
	if type( docId ) ~= 'string' or #docId < 1 then
		error( "WFDocStoreImp:setDocValue: _id must be non-empty string", 2 )
	end

	-- Extract _slot from the value.
	
	local slotName = value._slot
	value._slot = nil
	
	if type( slotName ) ~= 'string' or #slotName < 1 then
		error( "WFDocStoreImp:setDocValue: _slot must be non-empty string", 2 )
	end
	
	-- We allow _transient to be populated. If present, it must be a boolean
	-- and must match the existing doc value. If not present, we default to false.
	-- Note that this implies that an existing slot value can not be transitioned
	-- from permanent to transient status. This is intentional.
	
	local transient = value._transient
	if transient and type( transient ) ~= 'boolean' then
		error( "WFDocStoreImp:setDocValue: _transient must be nil or boolean", 2 )
	end
	value._transient = nil
	
	-- We allow _hash to be populated. If present, it must match the existing
	-- doc value. If not present, we'll replace any existing doc value.
	
	local mustMatchHash = value._hash
	value._hash = nil
	
	-- We allow _changeCounter to be populated. If present, it must match the existing
	-- doc value. If not present, we'll replace any existing docValue.
	
	local mustMatchChangeCounter = value._changeCounter
	value._changeCounter = nil
	
	-- Verify that doc value is acceptable (does not contain illegal keys, etc.).

	assertDocValueIsValid( value )

	-- Look up existing content hash value. It might no longer be needed.

	local oldContentHash
	
	local existingRow = self._exec( self._db,
		[[SELECT contentHash, changeCounter, transient FROM documents
			WHERE docId == ? AND slotName == ?]],
			docId, slotName )
	
	if existingRow == nil then

		if transient == nil then
			transient = false
		end
	
	else
	
		oldContentHash = existingRow.contentHash

		if transient == nil then

			transient = existingRow.transient == 1

		else
		
			if transient ~= ( existingRow.transient == 1 ) then
				error( string.format( "WFDocStoreImp:setDocValue: _transient value %s does not match existing document transient status %s",
									  transient, existingRow.transient == 1 ), 2 )
			end
		
		end

	end
	
	if mustMatchHash and oldContentHash ~= mustMatchHash then
		error( string.format( "WFDocStoreImp:setDocValue: _hash %s does not match existing document hash %s",
							  mustMatchHash, oldContentHash ), 2 )
	end
	
	local oldChangeCounter = existingRow and existingRow.changeCounter
	
	if mustMatchChangeCounter and oldChangeCounter ~= mustMatchChangeCounter then
		error( string.format( "WFDocStoreImp:setDocValue: _changeCounter %s does not match existing document change counter %s",
							  mustMatchChangeCounter, oldChangeCounter ), 2 )
	end

	-- Convert Lua to text and hash it.
	
	local archive = KSTableArchiving.simpleTableToChunkString( value )
	local contentHash = sha1.digest( archive )
	
	-- Short-circuit if new content == existing.
	
	if contentHash == oldContentHash then
		return contentHash
	end
	
	-- Write new revision into database.
	
	return self._db:withTransactionDo( function()
	
		-- Add this new content blob into DB if not already there.
	
		self._exec( self._db,
			[[INSERT OR IGNORE INTO content VALUES( ?, ? )]], contentHash, archive )
		
		-- Update the reference in documents table.
		
		self._exec( self._db,
			[[INSERT OR REPLACE INTO documents
				VALUES( ?, ?, incrementChangeCounter(), ?, ? )]],
				docId, slotName, transient and 1 or 0, contentHash )

		-- Bump local copy of change counter.
		
		updateChangeCounter( self )

		-- Do we need to update a doc value behavior?
		
		local docValueBehaviors = self._docValueBehaviors
		local key = docValueBehaviorKey( docId, slotName )
		
		if docValueBehaviors:eventExists( key ) then
			docValueBehaviors:postToEvent( key, self:docValue( docId, slotName ) )
		end

		-- Delete old content if no longer referenced.
		
		maybeGCOldContentValue( self, oldContentHash )
	
		return contentHash
	
	end )
	
end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------

function docStoreImpMethods:applyOverlay( overlay )

	check( 'WFDocStoreImp', 'table' )

	-- Extract doc ID from the overlay.
	
	local docId = overlay._id
	overlay._id = nil
	
	if type( docId ) ~= 'string' or #docId < 1 then
		error( "WFDocStoreImp:applyOverlay: _id must be non-empty string", 2 )
	end

	-- Extract _slot from the overlay.
	
	local slotName = overlay._slot
	overlay._slot = nil
	
	if type( slotName ) ~= 'string' or #slotName < 1 then
		error( "WFDocStoreImp:applyOverlay: _slot must be non-empty string", 2 )
	end
	
	-- Get existing value, if any, apply overlay, write it back.
	
	local value = self:docValue( docId, slotName )
	value._empty = nil
	
	WFTableMerge.applyOverlay( value, overlay )
	
	return self:setDocValue( value )

end

--------------------------------------------------------------------------------

-- @mark slot management

--------------------------------------------------------------------------------

function docStoreImpMethods:copySlotToSlot( docId, fromSlot, toSlot, transient )

	check( 'WFDocStoreImp', '+string', '+string', '+string', '?boolean' )
	
	-- Look up existing content hash value. Error if it doesn't exist.

	local sourceRow = self._exec( self._db,
		[[SELECT contentHash, changeCounter FROM documents
			WHERE docId == ? AND slotName == ?]],
			docId, fromSlot )
	
	if not sourceRow then
		error( string.format( "No such document/slot: %s/%s", docId, fromSlot ), 2 )
	end
	
	-- Make sure we don't change transient status via this call.

	local existingDestRow = self._exec( self._db,
		[[SELECT transient, contentHash FROM documents WHERE docId == ? AND slotName == ?]],
			docId, toSlot )

	if existingDestRow == nil then

		if transient == nil then
			transient = false
		end
	
	else

		-- Short-circuit if dest row already has this value.

		if existingDestRow.contentHash == sourceRow.contentHash then
			return sourceRow.contentHash
		end

		-- Make sure we don't change transient status.
	
		if transient == nil then

			transient = existingDestRow.transient == 1

		else
		
			if transient ~= ( existingDestRow.transient == 1 ) then
				error( string.format( "WFDocStoreImp:copySlotToSlot: _transient value %s does not match existing slot's transient status %s",
									  transient, existingDestRow.transient == 1 ), 2 )
			end
		
		end

	end
	
	-- Plunk this in the new slot.
	
	return self._db:withTransactionDo( function()
	
		-- Update the reference in documents table.
		
		self._exec( self._db,
			[[INSERT OR REPLACE INTO documents
				VALUES( ?, ?, incrementChangeCounter(), ?, ? )]],
				docId, toSlot, transient and 1 or 0, sourceRow.contentHash )

		-- Do we need to update a doc value behavior?
		
		local docValueBehaviors = self._docValueBehaviors
		local key = docValueBehaviorKey( docId, toSlot )
		
		if docValueBehaviors:eventExists( key ) then
			docValueBehaviors:postToEvent( key, self:docValue( docId, toSlot ) )
		end

		-- Bump local copy of change counter.
		
		updateChangeCounter( self )
		
		return sourceRow.contentHash
	
	end )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:quietlyDeleteSlot( docId, slotName )

	check( 'WFDocStoreImp', '+string', '+string' )
	
	if slotName == 'main' then
		error( "WFDocStoreImp:quietlyDeleteSlot: can not delete slot 'main'", 2 )
	end
	
	-- Look up existing content hash value. It might no longer be needed.

	local existingRow = self._exec( self._db,
		[[SELECT contentHash FROM documents WHERE docId == ? AND slotName == ?]],
			docId, slotName )
	
	if existingRow == nil then
		return
	end
	
	local oldContentHash = existingRow.contentHash

	-- Write new revision into database.
	
	return self._db:withTransactionDo( function()
	
		-- Remove this reference from the documents table.
	
		self._exec( self._db,
			[[DELETE FROM documents WHERE docId == ? AND slotName == ?]],
				docId, slotName )
		
		-- Do we need to update a doc value behavior?
		
		local docValueBehaviors = self._docValueBehaviors
		local key = docValueBehaviorKey( docId, slotName )
		
		if docValueBehaviors:eventExists( key ) then
			docValueBehaviors:postToEvent( key,
				{
					_id = docId,
					_slot = slotName,
					_empty = true,
				} )
		end

		-- Delete old content if no longer referenced.
		
		maybeGCOldContentValue( self, oldContentHash )
	
	end )
	
end

--------------------------------------------------------------------------------

-- @mark transient slots

--------------------------------------------------------------------------------

function docStoreImpMethods:deleteTransientValues()

	check( 'WFDocStoreImp' )
	
	local oldContentHashRows = self._db:gatherArrayOfArray(
		[[SELECT contentHash FROM documents WHERE transient == 1]] )

	if #oldContentHashRows == 0 then
		return
	end
	
	-- Coalesce possible duplicate hash values. Only need to GC each item once.
	
	local oldContentHashes = {}
	for i = 1, #oldContentHashRows do
		oldContentHashes[ oldContentHashRows[ i ][ 1 ] ] = true
	end

	-- Pick up current change counter so we can ensure it doesn't fall backward.

	local db = self._db

	local changeCounterSeedRow = db:execOnce( "SELECT max( changeCounter ) FROM documents" )
	local currentChangeCounter = changeCounterSeedRow[ "max( changeCounter )" ]

	-- Now purge slot values and GC any associated content.
	
	return db:withTransactionDo( function()

		db:exec( [[DELETE FROM documents WHERE transient == 1]] )
		
		for hash in pairs( oldContentHashes ) do
			maybeGCOldContentValue( self, hash )
		end
		
		-- Do we need to introduce a placeholder document to ensure change
		-- counter doesn't fall backwards?
		
		local newCCRow = db:execOnce( "SELECT max( changeCounter ) FROM documents" )
		local newChangeCounter = newCCRow[ "max( changeCounter )" ]
	
		if newChangeCounter < currentChangeCounter then
		
			db:exec(
				[[INSERT OR REPLACE INTO documents
					VALUES( '(placeholder)', '(ignore me)',
							?, 1, '' )]], currentChangeCounter )
		
		end

	end )

end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------

function docStoreImpMethods:docChangeCounterB()

	check( 'WFDocStoreImp' )

	return self._docChangeCounterB:relayB()
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:docChangeCounter()

	check( 'WFDocStoreImp' )

	return self._docChangeCounterB:valueNow()

end

--------------------------------------------------------------------------------

local changedDocs_maxLimit = 50
local changedDocs_maxLimitForIds = 1000
	-- Both of these are arbitrary; tune based on observed behavior.

--------------------------------------------------------------------------------

function docStoreImpMethods:changedDocs( args )

	check( 'WFDocStoreImp', 'table' )
	
	local db = self._db

	local from = args.fromChangeCounter
	if not from then
		error( "WFDocStoreImp:changedDocs: fromChangeCounter argument missing", 2 )
	end

	local maxLimit = args.idsOnly
						and changedDocs_maxLimitForIds
						or changedDocs_maxLimit
	
	local limit = args.limit or maxLimit
	if type( limit ) ~= 'number' then
		error( "WFDocStoreImp:changedDocs: limit must be a number or nil for default", 2 )
	end
	
	if limit < 1 then
		error( "WFDocStoreImp:changedDocs: limit must be a positive number", 2 )
	end
	
	if limit > maxLimit then
		limit = maxLimit
	end

	local docs = args.docs
	if docs and ( type( docs ) ~= 'table' or #docs == 0 ) then
		error( "WFDocStoreImp:changedDocs: docs must be a non-empty array of doc IDs or nil for all docs", 2 )
	end

	local idsOnly = args.idsOnly
	
	local slot = args.slot
	if slot and ( type( slot ) ~= 'string' or #slot == 0 ) then
		error( "WFDocStoreImp:changedDocs: slot must be a non-empty string or nil for all slots", 2 )
	end
	
	local rawDocRows = db:withTransactionDo( function()

		local rawDocRows
		
		if docs then
	
			if #docs > 100 then
				changedDocsLog:warn( "WARNING: doc inclusion list is awfully large %s",
									 #docs )
			end
	
			db:exec [[DELETE FROM TEMP_docIds]]

			for i = 1, #docs do
				db:exec( [[INSERT INTO TEMP_docIds VALUES( ? )]], docs[ i ] )
			end

			if slot then

				rawDocRows = db:gatherArrayOfArray(
					[[SELECT documents.docId, slotName, changeCounter
									FROM documents
										INNER JOIN TEMP_docIds
										WHERE changeCounter >= ?
											AND documents.docId == TEMP_docIds.docId
											AND slotName == ?
										ORDER BY changeCounter
										LIMIT ? ]],
									from, slot, limit + 1 )

			else

				rawDocRows = db:gatherArrayOfArray(
					[[SELECT documents.docId, slotName, changeCounter
									FROM documents
										INNER JOIN TEMP_docIds
										WHERE changeCounter >= ?
											AND documents.docId == TEMP_docIds.docId
										ORDER BY changeCounter
										LIMIT ? ]],
									from, limit + 1 )
			
			end
	
		else
		
			if slot then
		
				rawDocRows = db:gatherArrayOfArray(
					[[SELECT docId, slotName, changeCounter
									FROM documents
										WHERE changeCounter >= ?
											AND slotName == ?
										ORDER BY changeCounter
										LIMIT ? ]],
									from, slot, limit + 1 )

			else

				rawDocRows = db:gatherArrayOfArray(
					[[SELECT docId, slotName, changeCounter
									FROM documents
										WHERE changeCounter >= ?
										ORDER BY changeCounter
										LIMIT ? ]],
									from, limit + 1 )
	
			end
	
		end
		
		return rawDocRows
	
	end )
	
	local result = {}
	
	if idsOnly then

		for i = 1, math.min( #rawDocRows, limit ) do
		
			result[ i ] = rawDocRows[ i ][ 1 ]
			
		end
	
	else
	
		for i = 1, math.min( #rawDocRows, limit ) do
		
			local rawDocRow = rawDocRows[ i ]
			result[ i ] = self:docValue( rawDocRow[ 1 ], rawDocRow[ 2 ] )
				-- PERF [scouten 2013-04-04]: Can probably scoop up results
				-- in a bulk query above using another INNER JOIN so we don't
				-- have to do *every* row via :docValue.
				
				-- But doing it this way now will get us to build the unit
				-- tests with correct expected results and we can make it fast
				-- later.
		
		end
	
	end

	if rawDocRows[ limit + 1 ] then
		result.continueFrom = rawDocRows[ limit + 1 ][ 3 ]
	else
		result.continueFrom = self._docChangeCounterB:valueNow() + 1
	end

	changedDocsLog:debug( "changedDocs is returning %s", result )

	return result

end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------

function docStoreImpMethods:path()

	check( 'WFDocStoreImp' )

	return self._path

end

--------------------------------------------------------------------------------

function docStoreImpMethods:__tostring()

	check( 'WFDocStoreImp' )

	return string.format( "WFDocStoreImp( %s )", self._path )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:DEBUG_dumpCoverage( path )

	check( 'WFDocStoreImp', '+string' )

	require 'SetupCoverage'.dumpCoverage( path )

end

--------------------------------------------------------------------------------

return KSClass 'WFDocStoreImp'( docStoreImpMethods )
