--[[----------------------------------------------------------------------------

WFDocStoreTest.lua

@author scouten 2013-04-08

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTestUtils = require 'KSTestUtils'
local WFArguments = require 'WFArguments'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

local WFPlatformInfo = require 'WFPlatformInfo'
local TEST_WAIT_TIME =  WFPlatformInfo.isAndroid and 5 or 3


--============================================================================--

local WFDocStoreTest = {}

local check = assert( WFArguments.check )

local docStore

--------------------------------------------------------------------------------

function WFDocStoreTest.setUp()

	docStore = WFDocStore( ':memory:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.tearDown( coveragePath )

	if docStore then

		if coveragePath then

			local oldDocStore = docStore
			docStore:DEBUG_dumpCoverageP( coveragePath )
				:whenResolvedDo( function()
					oldDocStore:close()
				end )
				:runUntilDone()

		else

			docStore:close()

		end

		docStore = nil

	end

end

--------------------------------------------------------------------------------

local function assertFailsAsync( test, fn )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		:waitFor( fn() )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_emptyDoc( test )

	local hash

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue._hash = _hash
			hash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testDocValueB_emptyDoc( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local docValueB = docStore:docValueB( 'testDocId' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( docValueB )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_slot = 'main',
				_empty = true,
			}
		
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( status, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )
		
		end )

		:waitFor( docValueB )
		:continueWhenValueMatches( expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testDocValueB_reuseExisting( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local docValueB = docStore:docValueB( 'testDocId' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:docValueB( 'testDocId' )
			
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_slot = 'main',
				_empty = true,
			}
		
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( status, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )
		
		end )

		:waitFor( docValueB )
		:continueWhenValueMatches( expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testDocValueB_slotCrossContamination( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local docValueB = docStore:docValueB( 'testDocId', 'other' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( docValueB )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_slot = 'other',
				_empty = true,
			}
		
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( status, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )
		
		end )

		:waitFor( docValueB )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_slot = 'other',
				_empty = true,
			}

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_replacingHash( test )

	local firstHash

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 45,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( status, hash )
			
				firstHash = hash
				return 'data', hash
			
			end )
			
		end )
		
		:waitFor( function()
			
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_hash = firstHash,
				test = 45,
			}
			:mapDataP( function( status, hash )
			
				expectedValue._hash = hash
				return 'data', hash
			
			end )
		
		end )
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_replacingChangeCounter( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 45,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		
		:waitFor( function()
			
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_changeCounter = 1,
				test = 45,
			}
			:mapDataP( function( status, hash )
			
				expectedValue._hash = hash
				return 'data', hash
			
			end )
		
		end )
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_emptyDocNestedArray( test )

	local hash

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
		array = { 'a', 'b', 'c' },
		hash = { a = 1, b = 2, c = 3 },
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				array = { 'a', 'b', 'c' },
				hash = { a = 1, b = 2, c = 3 },
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue._hash = _hash
			hash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValue_allowSyncStatusSpecialKey( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
		_syncStatus = { 'a', 'b', 'c' },
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				_syncStatus = { 'a', 'b', 'c' },
			}
			:mapDataP( function( _, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValue_allowDeletedSpecialKey( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValue_afterDelete = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_deleted = true,
		_changeCounter = 2,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( _, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

		:waitFor( function()
			
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_deleted = true,
			}
			:mapDataP( function( _, hash )
				expectedValue_afterDelete._hash = hash
				return 'data', hash
			end )
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue_afterDelete )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_emptyDocTempDB( test )

	-- Same test as above, but proving we can use ':temp:' pseudo-path as well.

	docStore:close()	
	docStore = WFDocStore( ':temp:' )

	return WFDocStoreTest.testGetSetDocValueP_emptyDoc( test )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetDocValueP_empty( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42
			}
		
		end )

		:waitFor( function()
		
			return docStore:docValueP( 'bogusDocId' )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'bogusDocId',
				_slot = 'main',
				_empty = true,
			} )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorReplacingWrongHash( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_hash = 'notTheRightHash',
				test = 45,
			}
		
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorChangingTransientStatus_trueToFalse( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = false,
				test = 45,
			}
			
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorChangingTransientStatus_falseToTrue( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = false,
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 45,
			}

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorChangingTransientStatus_nilToTrue( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 45,
			}
		
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorReplacingWrongChangeCounter( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_changeCounter = 71,
				test = 45,
			}
		
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueMissingTable()

	KSTestUtils.assertFails( function()
		return docStore:setDocValueP()
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueMissingId( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP {}
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIdNotString( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP { _id = 4 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIdEmptyString( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP { _id = '' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueMissingSlot( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP { id = 'ok' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueSlotNotString( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP { _id = 'ok', _slot = 4 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueSlotEmptyString( test )

	assertFailsAsync( test, function()
		return docStore:setDocValueP { _id = 'ok', _slot = '' }
	end )

end

--------------------------------------------------------------------------------

local function setDocValuePShouldFailAsync( test, value )

	value._id = 'testDocId'
	value._slot = 'main'

	assertFailsAsync( test, function()
		return docStore:setDocValueP( value )
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, { _transient = "true", blah = 42 } )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, { [ 42 ] = 'string keys only, please' } )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, { [ '' ] = 'non-empty string keys' } )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, { _blah = 'must not start with underscore' } )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, { func = function() end } )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test,
		{
			nested = {
				[ true ] = 'enforce non-string/number rule in nested tables',
			},
		} )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test, 
		{
			nested = {
				_stillNotOK = 'enforce leading underscore rule in nested tables',
			},
		} )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test,
		{
			nested = {
				[ '' ] = 'enforce non-empty string rule in nested tables',
			},
		} )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueIsInvalid_transientNotBoolean( test )

	setDocValuePShouldFailAsync( test,
		{
			nested = {
				nested2 = {
					[ true ] = 'enforce non-string/number rule in deeply nested tables',
				},
			},
		} )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testSetDocValueP_errorDocValueContainsDeletedAndOtherKeys( test )

	setDocValuePShouldFailAsync( test,
		{
			_deleted = true,
			nothingElseAllowed = true,
				-- If _deleted = true, then no other values are allowed in the table.
		} )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_eraseOldValue( test )

	local firstHash, secondHash

	local expectedValue1 = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValue2 = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		blah = 95,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue1._hash = _hash
			firstHash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue1 )

		-- Update doc value and test read-back again.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				blah = 95,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			assert( firstHash ~= _hash )
			
			expectedValue2._hash = _hash
			secondHash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue2 )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_differentSlots( test )

	local firstHash, secondHash

	local expectedValue1 = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValue2 = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 3,
		blah = 95,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue1._hash = _hash
			firstHash = _hash
		
		end )
	
		-- Set second slot to same doc value.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				test = 42,
			}
		
		end )
	
		-- Now change second slot value, leaving first slot at existing value.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				blah = 95,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			assert( firstHash ~= _hash )
			
			expectedValue2._hash = _hash
			secondHash = _hash
		
		end )
	
		-- Now read both values back.
	
		:waitFor( function()
			return docStore:docValueP( 'testDocId', 'main' )
		end )
		:continueWhenValueMatches( 'data', expectedValue1 )

		:waitFor( function()
			return docStore:docValueP( 'testDocId', 'other' )
		end )
		:continueWhenValueMatches( 'data', expectedValue2 )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testGetSetDocValueP_reuseExistingValue( test )

	local hash

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue._hash = _hash
			hash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

		-- Set doc value to same value and capture rev ID.
		-- Should not bump change counter.
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue._hash = _hash
			hash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------

function WFDocStoreTest.testFieldValueB( test )

	local fieldValueB = docStore:fieldValueB( 'testDocId', 'rating' )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( fieldValueB )
		:continueWhenValueMatches( WFUnknown.value() )
		
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				rating = 2,
				title = "My Nifty Photo",
			}

		end )
	
		:waitFor( fieldValueB )
		:continueWhenValueMatches( 2 )
		
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				rating = 0,
			}

		end )
	
		:waitFor( fieldValueB )
		:continueWhenValueMatches( 0 )
		
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testFieldValueB_nested( test )

	local fieldValueB = docStore:fieldValueB( 'testDocId', 'foo/1/rating' )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( fieldValueB )
		:continueWhenValueMatches( WFUnknown.value() )
		
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				foo = {
					{
						rating = 2
					}
				},
				title = "My Nifty Photo",
			}
		
		end )
		
		:waitFor( fieldValueB )
		:continueWhenValueMatches( 2 )
		
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				foo = { _delete = true },
			}
			
		end )
		
		:waitFor( fieldValueB )
		:continueWhenValueMatches( nil )
		
		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				foo = {
					{
						rating = 4
					}
				},
			}
		
		end )

		:waitFor( fieldValueB )
		:continueWhenValueMatches( 4 )
		
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testFieldValueB_errorNoPath( test )

	KSTestUtils.assertFails( function()

		return docStore:fieldValueB( 'testDocId', '/' )
	
	end )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_empty( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		-- Set doc value and capture rev ID.
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, _hash )
			
			check( 'string', 'string' )
			
			assert( status == 'data' )
			assert( #_hash == 40 )
			assert( _hash:match( '^[0-9a-f]+$' ) )
			
			expectedValue._hash = _hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_afterSetDocValueP( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueAfterOverlay = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		blah = 45,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			
		end )
		:whenResolvedDo( function( status, hash )
		
			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = 45,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueAfterOverlay._hash = hash
			
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueAfterOverlay )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_nestedTable( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		blah = {
			a = 1,
			b = 2,
			[ "foo" ] = 42,
		},
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )

			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					a = 1,
					b = 2,
					[ "foo" ] = 42,
				}
			}
			
		end )
		:whenResolvedDo( function( status, hash )

			expectedValue._hash = hash
			
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_nestedTableReplacingScalar( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		blah = {
			a = 1,
			b = 2,
			[ "foo" ] = 42,
		},
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				blah = 95,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					a = 1,
					b = 2,
					[ "foo" ] = 42,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
			
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_mergeNestedTable( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		blah = {
			a = 1,
			b = 2,
			[ "foo" ] = 42,
		},
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				blah = {
					a = 1,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					b = 2,
					[ "foo" ] = 42,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_deleteNestedTable( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				blah = {
					a = 1,
					b = 2,
					[ "foo" ] = 42,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					_delete = true,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_deleteScalar( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				blah = "zip",
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					_delete = true,
				}
			}
			
		end )
		:whenResolvedDo( function( status, hash )

			expectedValue._hash = hash
			
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_deleteNothing( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				blah = {
					_delete = true,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_replaceExisting( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		nested = {
			a = 1,
			b = 2,
			[ 42 ] = 'c',
		},
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
				nested = {
					a = 1,
					b = 2,
					[ "foo" ] = 42,
				}
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
			
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				nested = {
					_replace = true,
					a = 1,
					b = 2,
					[ 42 ] = 'c',
				},
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_replaceNothing( test )

	local expectedValue = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
		nested = {
			a = 1,
			b = 2,
			[ 42 ] = 'c',
		},
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'testDocId',
				_slot = 'main',
				nested = {
					_replace = true,
					a = 1,
					b = 2,
					[ 42 ] = 'c',
				},
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValue._hash = hash
		
		end )
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValue )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorMissingTable()

	KSTestUtils.assertFails( function()
		return docStore:applyOverlayP()
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorMissingId( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP {}
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorIdNotString( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP { _id = 4 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorIdEmptyString( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP { _id = '' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorMissingSlot( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP { id = 'ok' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorSlotNotString( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP { _id = 'ok', _slot = 4 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testApplyOverlayP_errorSlotEmptyString( test )

	assertFailsAsync( test, function()
		return docStore:applyOverlayP { _id = 'ok', _slot = '' }
	end )

end

--------------------------------------------------------------------------------

-- @mark slot management

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash

		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other' )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
		
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlot_docValueB( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}
	
	local docValueB_main = docStore:docValueB( 'testDocId', 'main' )
	local docValueB_other = docStore:docValueB( 'testDocId', 'other' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
			:mapDataP( function( status, hash )
				expectedValueMain._hash = hash
				return 'data', hash
			end )

		end )
		
		:waitFor( docValueB_main )
		:continueWhenValueMatches( expectedValueMain )

		:waitFor( docValueB_other )
		:continueWhenValueMatches {
			_id = 'testDocId',
			_slot = 'other',
			_empty = true,
		}

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other' )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
		
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( docValueB_main )
		:continueWhenValueMatches( expectedValueMain )

		:waitFor( docValueB_other )
		:continueWhenValueMatches( expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_shortCircuitExistingValueMatches( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
			-- should not bump to 3 when we do :copySlotToSlotP
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )

			expectedValueOther._hash = hash
			
		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'other', 'main' )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
			
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_errorNoSuchSource( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash

		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'bogus', 'other' )
			
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )
		
		-- Make sure we can read back expected results.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )

		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'other',
				_empty = true,
			} )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_changeTransientStatus_true( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 2,
		_transient = true,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash
			
		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other', true )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
			
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_changeTransientStatus_false( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 2,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash
			
		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other', false )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
			
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_overwriteExistingValue( test )

	local expectedValueMain = {
		_id = 'testDocId',
		_slot = 'main',
		-- _hash = (inserted later),
		_changeCounter = 1,
		test = 42,
	}

	local expectedValueOther = {
		_id = 'testDocId',
		_slot = 'other',
		-- _hash = (inserted later),
		_changeCounter = 3,
		_transient = true,
		test = 42,
	}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )
			
			expectedValueMain._hash = hash
			
		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				_transient = true,
				blah = 57,
			}
			
		end )

		:waitFor( function()

			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other' )
		
		end )
		:whenResolvedDo( function( status, hash )
			
			assert( hash == expectedValueMain._hash,
				"Huh. How did the hash value change?" )
			
			expectedValueOther._hash = hash
			
		end )
		
		-- Make sure we can read both back.
	
		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'main' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueMain )

		:waitFor( function()
		
			return docStore:docValueP( 'testDocId', 'other' )
		
		end )
		:continueWhenValueMatches( 'data', expectedValueOther )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_errorChangingTransientStatus_trueToFalse( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				_transient = true,
				test = 45,
			}
		
		end )

		:waitFor( function()
		
			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other', false )
			
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_errorChangingTransientStatus_falseToTrue( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				_transient = false,
				test = 45,
			}
		
		end )

		:waitFor( function()
		
			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other', true )
			
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testCopySlotToSlotP_errorChangingTransientStatus_nilToTrue( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				_transient = true,
				test = 42,
			}
		
		end )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'other',
				test = 45,
			}
		
		end )

		:waitFor( function()
		
			return docStore:copySlotToSlotP( 'testDocId', 'main', 'other', true )
			
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testQuietlyDeleteSlotP( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'blah',
				test = 42,
			}

		end )
		
		:waitFor( function()
			
			return docStore:docValueP( 'testDocId', 'blah' )
				:mapDataP( function( _, docValue )
					docValue._hash = nil
					return 'data', docValue
				end )

		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'blah',
				_changeCounter = 1,
				test = 42,
			} )
		
		:waitFor( function()
		
			return docStore:quietlyDeleteSlotP( 'testDocId', 'blah' )
		
		end )
		
		:waitFor( function()
			
			return docStore:docValueP( 'testDocId', 'blah' )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'blah',
				_empty = true,
			} )
		
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testQuietlyDeleteSlotP_doesntExist( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
			
			return docStore:docValueP( 'testDocId', 'blah' )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'blah',
				_empty = true,
			} )
		
		:waitFor( function()
		
			return docStore:quietlyDeleteSlotP( 'testDocId', 'blah' )
		
		end )
		
		:waitFor( function()
			
			return docStore:docValueP( 'testDocId', 'blah' )
		
		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'blah',
				_empty = true,
			} )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testQuietlyDeleteSlotP_updatesDocValueB( test )

	local expectedValue =
		{
			_id = 'testDocId',
			_slot = 'blah',
			_changeCounter = 1,
			test = 42,
		}
	
	local docValueB = docStore:docValueB( 'testDocId', 'blah' )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'blah',
				test = 42,
			}
			:mapDataP( function( _, hash )
				expectedValue._hash = hash
				return 'data', hash
			end )

		end )
		
		:waitFor( docValueB )
		:continueWhenValueMatches( expectedValue )
		
		:waitFor( function()
		
			return docStore:quietlyDeleteSlotP( 'testDocId', 'blah' )
		
		end )
		
		:waitFor( docValueB )
		:continueWhenValueMatches(
			{
				_id = 'testDocId',
				_slot = 'blah',
				_empty = true,
			} )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testQuietlyDeleteSlotP_errorCantDeleteMain( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}

		end )
		
		:waitFor( function()
			
			return docStore:docValueP( 'testDocId', 'main' )
				:mapDataP( function( _, docValue )
					docValue._hash = nil
					return 'data', docValue
				end )

		end )
		:continueWhenValueMatches( 'data',
			{
				_id = 'testDocId',
				_slot = 'main',
				_changeCounter = 1,
				test = 42,
			} )
		
		:waitFor( function()
			return docStore:quietlyDeleteSlotP( 'testDocId', 'main' )
		end )
		:continueWhenValueMatches( 'error', ':ignore:' )
		
end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------

function WFDocStoreTest.testDocChangeCounterB( test )

	local docChangeCounterB = docStore:docChangeCounterB()

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 0 )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}

		end )

		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 1 )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				rating = 3,
			}
			
		end )

		:waitFor( docChangeCounterB )
		:continueWhenValueMatches( 2 )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testDocChangeCounterP( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return docStore:docChangeCounterP()
		
		end )
		:continueWhenValueMatches( 'data', 0 )
	
		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}

		end )

		:waitFor( function()

			return docStore:docChangeCounterP()
		
		end )
		:continueWhenValueMatches( 'data', 1 )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				rating = 3,
			}
			
		end )

		:waitFor( function()

			return docStore:docChangeCounterP()
		
		end )
		:continueWhenValueMatches( 'data', 2 )
	
end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP( test )

	local expectedChangedDocsPResult =
		{

			{
				_id = 'testDocId',
				_slot = 'main',
				_changeCounter = 1,
				_hash = nil, -- filled in later
				test = 42,
			},
			{
				_id = 'anotherPhoto',
				_slot = 'main',
				_changeCounter = 2,
				_hash = nil, -- filled in later
				favorite = true,
				title = "My Nifty Photo",
			},
			
			continueFrom = 3,
		
		}

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		-- Write two documents.
		
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'testDocId',
				_slot = 'main',
				test = 42,
			}
		
		end )
		:whenResolvedDo( function( status, hash )

			expectedChangedDocsPResult[ 1 ]._hash = hash

		end )
		
		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'anotherPhoto',
				_slot = 'main',
				favorite = true,
				title = "My Nifty Photo",
			}
		
		end )
		:whenResolvedDo( function( status, hash )

			expectedChangedDocsPResult[ 2 ]._hash = hash

		end )
	
		-- Read them back via :changedDocsP.
		
		:waitFor( function()
			return docStore:changedDocsP { fromChangeCounter = 1 }
		end )
		
		:continueWhenValueMatches( 'data', expectedChangedDocsPResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_empty( test )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()
			return docStore:changedDocsP { fromChangeCounter = 1 }
		end )
		
		:continueWhenValueMatches( 'data', { continueFrom = 1 } )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorFromChangeCounterMissing( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP {}
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorLimitNotString( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, limit = 'bogus' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorLimitNotPositive( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, limit = 0 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorSlotNotString( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, slot = 0 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorSlotNameEmpty( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, slot = '' }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_observesLimit( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 51 }
	for i = 1, 50 do
		expectedResult[ i ] = {
			_id = docIds[ i ],
			_slot = 'main',
			_changeCounter = i,
			favorite = true,
		}
	end
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = 'main',
						favorite = true,
					}
					:mapDataP( function( status, hash )
						if i <= 50 then
							expectedResult[ i ]._hash = hash
						end
						return 'data', hash
					end )
				
				end, indices )

		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP { fromChangeCounter = 1, limit = 1000 }
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_observesSlotName( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 101 }
	for i = 1, 50 do
		expectedResult[ i ] = {
			_id = docIds[ i * 2 ],
			_slot = 'even',
			_changeCounter = i * 2,
			favorite = true,
		}
	end
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = i % 2 == 0 and 'even' or 'odd',
						favorite = true,
					}
					:mapDataP( function( status, hash )

						if i % 2 == 0 then
							expectedResult[ i / 2 ]._hash = hash
						end
					
					end )
					
				end, indices )

		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP { fromChangeCounter = 1, slot = 'even' }
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_observesSlotName_empty( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 101 }
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = 'main',
						favorite = true,
					}
				
				end, indices )

		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP { fromChangeCounter = 1, slot = 'nonesuch' }
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_observesDocInclusionList( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 101 }

	expectedResult[ 1 ] = {
		_id = docIds[ 17 ],
		_slot = 'main',
		_changeCounter = 17,
		favorite = true,
	}
	
	expectedResult[ 2 ] = {
		_id = docIds[ 50 ],
		_slot = 'main',
		_changeCounter = 50,
		favorite = true,
	}
	
	expectedResult[ 3 ] = {
		_id = docIds[ 51 ],
		_slot = 'main',
		_changeCounter = 51,
		favorite = true,
	}
	
	expectedResult[ 4 ] = {
		_id = docIds[ 87 ],
		_slot = 'main',
		_changeCounter = 87,
		favorite = true,
	}
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = 'main',
						favorite = true,
					}
					:mapDataP( function( status, hash )

						if i == 17 then
							expectedResult[ 1 ]._hash = hash
						elseif i == 50 then
							expectedResult[ 2 ]._hash = hash
						elseif i == 51 then
							expectedResult[ 3 ]._hash = hash
						elseif i == 87 then
							expectedResult[ 4 ]._hash = hash
						end
					
					end )
				
			end, indices )
			
		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP {
				fromChangeCounter = 1,
				docs = { docIds[ 87 ], docIds[ 50 ], docIds[ 51 ], docIds[ 17 ] },
					-- Intentionally not in order of creation.
					-- Want to ensure that changes are reported in chronological
					-- order, not the order that they appear in docs array.
			}
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_observesDocInclusionListAndSlotName( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 101 }

	expectedResult[ 1 ] = {
		_id = docIds[ 17 ],
		_slot = 'odd',
		_changeCounter = 17,
		favorite = true,
	}
	
	expectedResult[ 2 ] = {
		_id = docIds[ 51 ],
		_slot = 'odd',
		_changeCounter = 51,
		favorite = true,
	}
	
	expectedResult[ 3 ] = {
		_id = docIds[ 87 ],
		_slot = 'odd',
		_changeCounter = 87,
		favorite = true,
	}
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = i % 2 == 0 and 'even' or 'odd',
						favorite = true,
					}
					:mapDataP( function( status, hash )

						if i == 17 then
							expectedResult[ 1 ]._hash = hash
						elseif i == 51 then
							expectedResult[ 2 ]._hash = hash
						elseif i == 87 then
							expectedResult[ 3 ]._hash = hash
						end
		
					end )
				
				end, indices )
			
		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP {
				fromChangeCounter = 1,
				slot = 'odd',
				docs = { docIds[ 87 ], docIds[ 50 ], docIds[ 51 ], docIds[ 17 ] },
					-- Intentionally not in order of creation.
					-- Want to ensure that changes are reported in chronological
					-- order, not the order that they appear in docs array.
			}
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorDocInclusionListNotTable( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, docs = 42 }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_errorDocInclusionListEmptyTable( test )

	assertFailsAsync( test, function()
		return docStore:changedDocsP { fromChangeCounter = 1, docs = {} }
	end )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_warnDocInclusionListIsLarge( test )

	local docIds = {}
	local indices = {}
	for i = 1, 120 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 51 }

	for i = 1, 50 do
		expectedResult[ i ] = {
			_id = docIds[ i ],
			_slot = 'main',
			_changeCounter = i,
			favorite = true,
		}
	end

	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = 'main',
						favorite = true,
					}
					:mapDataP( function( status, hash )

						if i <= 50 then
							expectedResult[ i ]._hash = hash
						end
					
					end )
				
				end, indices )
			
		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP {
				fromChangeCounter = 1,
				docs = docIds, -- yes, that's too many, but we'll warn and continue
			}
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_idsOnlyObservesLimit( test )

	local docIds = {}
	local indices = {}
	for i = 1, 2000 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = { continueFrom = 1001 }
	for i = 1, 1000 do
		expectedResult[ i ] = docIds[ i ]
	end
	
	WFAsyncTestCase( test, TEST_WAIT_TIME * 15 )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot = 'main',
						favorite = true,
					}

				end, indices )
			
		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP {
				fromChangeCounter = 1,
				idsOnly = true,
				limit = 1787,
			}
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testChangedDocsP_idsOnlyObservesDocInclusionList( test )

	local docIds = {}
	local indices = {}
	for i = 1, 100 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
		indices[ i ] = i -- hacky
	end
	
	local expectedResult = {
		continueFrom = 101,
		docIds[ 17 ],
		docIds[ 50 ],
		docIds[ 51 ],
		docIds[ 87 ],
	}
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )
		
		:waitFor( function()

			return KSPromises.allP(
				function( i )
				
					return docStore:setDocValueP {
						_id = docIds[ i ],
						_slot ='main',
						favorite = true,
					}
				
				end, indices )
			
		end )
		
		:waitFor( function()
		
			return docStore:changedDocsP {
				fromChangeCounter = 1,
				idsOnly = true,
				docs = { docIds[ 87 ], docIds[ 50 ], docIds[ 51 ], docIds[ 17 ] },
					-- Intentionally not in order of creation.
					-- Want to ensure that changes are reported in chronological
					-- order, not the order that they appear in docs array.
			}
		
		end )
		:continueWhenValueMatches( 'data', expectedResult )

end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------

function WFDocStoreTest.testPath()

	local path = docStore:path()
	
	assert( type( path ) == 'string' and #path > 0,
		":path() should have returned a non-empty string" )

end

--------------------------------------------------------------------------------

function WFDocStoreTest.testToString()

	local docStoreAsString = tostring( docStore )
	
	assert( type( docStoreAsString ) == 'string' and #docStoreAsString > 0,
		"tostring() should have returned a non-empty string" )

end

--------------------------------------------------------------------------------

return WFDocStoreTest
