--[[----------------------------------------------------------------------------

WFHttpClient.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-02-02

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFWorkerThread = require 'WFWorkerThread'

local KSReactive = require 'KSReactive'
local KSPromises = require 'KSPromises'

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFCurrentThread = require 'WFCurrentThread'
local WFCloseSupport = require 'WFCloseSupport'
local KSClass = require 'KSClass'
local KSPromises_promiseForResolverP = require 'KSPromises'.promiseForResolverP
local KSDate = require 'KSDate'
local WFQueueing = require 'WFQueueing'

--==============================================================================

local kMaxNumSimultaneousRequests = 4

--------------------------------------------------------------------------------

local httpClientLog = KSLogging.log 'WFHttpClient'
--	:setLevel( KSLogging.DEBUG )

local check = assert( WFArguments.check )

local httpClientImp = WFWorkerThread( 'WFHttpClientImp.lua' )
local httpStats = WFDiagnosticLog.makeRingBuffer {
	title = "Number of HTTP connections",
	accumulate = true,
	"uploadsRequested", "uploadsDispatched", "downloadsRequested", "downloadsDispatched"
}

local now = KSDate.timeNowForAnimation
local kTimings = setmetatable( {}, { __mode = 'kv' } )
local httpTimingStats = WFDiagnosticLog.makeRingBuffer {
	title = "HTTP client timings",
	"GET", "PUT"
}

--==============================================================================

WFDiagnosticLog.addLogSectionFunc( function()

	return {

		name = "HTTP Client : " .. WFCurrentThread.DEBUG_getThreadName(),
		id = 'WFHttpClient-' .. WFCurrentThread.DEBUG_getThreadName():gsub( ' :: ', '+' ),
		weight = 400,
		outputForAutomation = true,

		body = {
			connections = { httpStats:visualize "boxAndWhisker" },
			timings = { httpTimingStats:visualize "boxAndWhisker" },
		}
	}
end )

--==============================================================================

local function addHttpStatsForUpload( isUpload, ... )
	if isUpload then
		httpStats:add( ... )
	else
		-- download stats are the second pair
		httpStats:add( nil, nil, ... )
	end
end

--==============================================================================

local requestPromiseResolverMethods = {

	_didResolve = false,
		-- (boolean) sanity check: have we resolved this promise already?

	_rateLimiter = false,
	
	_didStartRequest = false,
	
	_messageE = false,
	
	_progressB = false,
	
	_data = false,
	
	_options = false,
	
	_url = false,
	
	_resolutionStarted = false,
	
	WFCloseSupport.addCloseSupportMethods,
	KSPromises.addPromiseResolverMethods,
	
}

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:__init( rateLimiter, url, options, data )
	
	self._url = assert( url, "Must have url" )
	self._options = options or false
	self._data = data or false
	
	self._rateLimiter = rateLimiter
	
	self._rateLimiter:addHttpStats( 1, nil )
	
	kTimings[ self ] = now()
		
end
--------------------------------------------------------------------------------

function requestPromiseResolverMethods:_startRequest()

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		return
	end
	
	if not self:canResolvePromise() then
		self:close()
		return
	end

	self._rateLimiter:addHttpStats( nil, 1 )

	local messageE = httpClientImp:requestE( 'getOrPut', self._url, self._options, self._data )
	
	if not self._options.retainDataAfterRequest then
		local data = self._data
		if data and type( data ) ~= 'string' then
			data:close()
		end
	end

	self._options = false
	self._data = self._data and 'closed'
	
	self._messageE = messageE
	local progressB = KSReactive.variableB( 0, 0 )
	self._progressB = progressB
	
	self._rateLimiter:startedTask()
	
	local function resolverAndBreakLinks( ... )
		if self and not self._didResolve then
			self._didResolve = true
			if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				self:resolvePromise( ... )
					-- This will also trigger a close
			end
			-- FIXME: Probably irrelevant since we have either
			-- closed or we triggered a close above
			if self._messageE then
				self._messageE:close()
				self._messageE = false
				self._rateLimiter:finishedTask()
			end
			if self._progressB then
				self._progressB:close()
				self._progressB = false
			end
			progressB = false
			self = false
		end
	end
	
	messageE:onDataE( function( what, ... )
			if what == 'result' then
				resolverAndBreakLinks( ... )
			elseif what == 'progress' and progressB then
				progressB:set( ... )
			end
		end )
	messageE:onErrorE( function( ... )
			resolverAndBreakLinks( 'error', ... )
		end )
	messageE:onDoneE( function( ... )
			resolverAndBreakLinks( 'error', 'unexpected termination' )
		end )
		
	self._didStartRequest = true
	
	return true

end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:startResolution( priority )

	if self._resolutionStarted then
		return
	end
	
	self._resolutionStarted = true
	
	self._rateLimiter:enqueueTask( function()
			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				return false
			end
			return self:_startRequest()
		end )

end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:close()
	
	-- Sever links to aid GC.
	self._didResolve = 'closed'
	
	self._rateLimiter:addHttpStats( -1, self._messageE and -1 or 0 )

	local time = now() - kTimings[ self ]

	kTimings[ self ] = nil

	httpTimingStats:add( self._data and 0 or time, self._data and time or 0 )
	
	if self._messageE then
		self._messageE:close()
		self._messageE = false
		self._rateLimiter:finishedTask()
	end

	if self._progressB then
		self._progressB:close()
		self._progressB = false
	end
	
	-- FIXME: Can we close the data here in case we never started the request?
	
end

--------------------------------------------------------------------------------

local requestPromiseResolver_class
	= KSClass 'RequestPromiseResolver'( requestPromiseResolverMethods )

--==============================================================================

local function makeRateLimiter( isUpload, queueMaker )

	local enqueue, dequeue = queueMaker()

	local outstanding = 0

	local startingActions = false
	
	local function startActions()
	
		if startingActions then
			return
		end
		
		startingActions = true
		
		while outstanding < kMaxNumSimultaneousRequests do

			local fn = dequeue()

			if not fn then
				break
			end
			
			pcall( fn )
			
		end
		
		startingActions = false
		
	end
	
	return {

		addHttpStats = function( self, ... ) return addHttpStatsForUpload( isUpload, ... ) end,

		enqueueTask = function( self, fn )

			if not fn then
				return
			end

			if outstanding < kMaxNumSimultaneousRequests then
				pcall( fn )
			else
				enqueue( fn )
			end

		end,
		
		startedTask = function()
			outstanding = outstanding + 1
		end,

		finishedTask = function( self )
		
			outstanding = outstanding - 1

			startActions()

		end

	}

end

--------------------------------------------------------------------------------

local uploadRateLimiter = makeRateLimiter( true, WFQueueing.makeQueue )

local downloadFIFORateLimiter = makeRateLimiter( false, WFQueueing.makeQueue )

local downloadLIFORateLimiter = makeRateLimiter( false, WFQueueing.makeStack )

local downloadRateLimiterForSequencing = {
	[ 'FIFO' ] = downloadFIFORateLimiter,
	[ 'LIFO' ] = downloadLIFORateLimiter
}

local defaultDownloadRateLimiter = downloadFIFORateLimiter

--============================================================================--

local function makeDualLIFORateLimiters()

	local enqueueHigh, dequeueHigh = WFQueueing.makeStack()
	local enqueueLow, dequeueLow = WFQueueing.makeStack()

	local outstanding = 0

	local startingActions = false
	
	local function startActions()
	
		if startingActions then
			return
		end
		
		startingActions = true
		
		while outstanding < kMaxNumSimultaneousRequests do

			local fn = dequeueHigh() or dequeueLow()

			if not fn then
				break
			end
			
			pcall( fn )
			
		end
		
		startingActions = false
		
	end
	
	local function makeEnqueueTask( enq )
		return function( self, fn )
			if not fn then
				return
			end
	
			if outstanding < kMaxNumSimultaneousRequests then
				pcall( fn )
			else
				enq( fn )
			end
		end
	end
	
	local function startedTask()
		outstanding = outstanding + 1
	end
	
	local function finishedTask()
		outstanding = outstanding - 1
		startActions()
	end
	
	local function addHttpStats( self, ... )
		return addHttpStatsForUpload( false, ... )
	end
	
	local function makeRateLimiter( enq )
		return {
			addHttpStats = addHttpStats,
			enqueueTask = makeEnqueueTask( enq ),
			startedTask = startedTask,
			finishedTask = finishedTask
		}
	end
	
	return makeRateLimiter( enqueueHigh ), makeRateLimiter( enqueueLow )

end

--------------------------------------------------------------------------------

downloadRateLimiterForSequencing[ 'LIFO-High' ],
downloadRateLimiterForSequencing[ 'LIFO-Low' ] =
	makeDualLIFORateLimiters()

--============================================================================--
--[[!
	\class WFHttpClient
	\brief This namespace provides basic HTTP services.
]]--

-- @mark WFHttpClient namespace

local WFHttpClient = {}

--------------------------------------------------------------------------------

local function transmitP( rateLimiter, url, options, data )

	options = options or {}

	local resolver = requestPromiseResolver_class( rateLimiter, url, options, data )

	local promise = KSPromises_promiseForResolverP( resolver, 'retained | closeTarget' )
	
	return promise
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP GET request and returns the response via a promise.

	\param url (string) The URL to fetch.
	\param options (optional, table) Options to modify the request.
	\param sequencing (optional, string) A string specifying the sequencing model
			Valid values include 'LIFO' and 'FIFO'. Defaults to FIFO if it doesn't
			find a match.

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server.

	\b Usage:
\code
local options = {
    requestHeaders = {                  -- optional
        { header = 'Content-Type', value = 'application/json' },
        { header = 'Accept', value = 'application/json' },
    }
    requestMethod = 'GET',              -- optional
                -- (must be a valid HTTP request method)
    returnHttpErrorsAsData = true,      -- optional
    timeout = (# seconds),              -- optional
	manuallyManageCookies = false,		-- optional
	followRedirects = true,				-- optional
}

local urlDataP = WFHttpClient.getP( url, options, sequencing )
    :whenResolvedDo( function( status, data, headers )

        if status == 'data' then

            -- The promise resolves as "data" in the "normal" case of a
            -- valid HTTP response with a non-error (<400) HTTP status code.
            -- The promise also resolves to "data" if the server returns
            -- an HTTP error code when returnHttpErrorsAsData = true.

            -- The data in this case is a KSData with the complete response body.
	    -- and 'header' is  a Lua table of key/value pairs.
            -- A special key 'status_code' in the table contains a string
            -- with the HTTP response code (such as "200").

            -- The default HTTP request method is GET, but you can override this
            -- with options.requestMethod.

        elseif status == 'error' then

            -- The promise resolves with "error" status if the system does
            -- not receive a valid HTTP response (for example, if the connection
            -- was refused or timed out). In this case, the data argument is a string
            -- describing the error and headers are nil.

            -- The promise also resolves with "error" if a valid HTTP response
            -- is received but the HTTP status code indicates an error (>= 400).
            -- You can override this behavior with returnHttpErrorsAsData = true.
 
            -- If an HTTP error response is signaled via "error" status, then
            -- data and headers are KSData and table as in the "data" case.

        end )
\endcode

	This returns a standard promise. You are responsible for managing the lifetime
	of the promise via the action(s) you attach to it. The HTTP request is
	canceled if all of its promise actions are closed.

\code
local urlDataP, progressB = WFHttpClient.getP ...

progressB:onDataB( function( nBytesSoFar, nBytesExpected )
     -- ...
	end )
\endcode
			
	NOTE: This function does no caching. Calling this function, always
	creates network traffic.
--]]

function WFHttpClient.getP( --[[string]] url,
			--[[optional table]] options,
			--[[optional string]] sequencing )

	check( 'string', '?table' )

	httpClientLog:debug( "getP( url = %s, options = %s )", url, options )
	
	options = options or {}
	
	local rateLimiter = downloadRateLimiterForSequencing[ sequencing ]
							or defaultDownloadRateLimiter
	
	local promise = transmitP( rateLimiter, url, options )
	
	return promise
	
end


--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP PUT request and returns the response via a promise.

	\param url (string) the URL to fetch
	\param data (\c KSData or Lua Table) data to send with the request
		(use a Lua table for form data)
	\param options (optional, table) options to modify the request

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server

	For now, this only supports URL encoding. Could easily be adapted for MIME
	encoding as well, but this hasn't proven necessary as yet.

	\b Usage:
\code
local options = {
    requestHeaders = {
        { header = 'Content-Type', value = 'application/json' },
        { header = 'Accept', value = 'application/json' },
    }
    requestMethod = 'PUT',              -- optional
                -- (must be a valid HTTP request method)
    timeout = (# seconds),              -- optional
    returnHttpErrorsAsData = true,      -- optional
	manuallyManageCookies = false,		-- optional
	followRedirects = true,				-- optional
}

local urlDataP = WFHttpClient.putP( url, data, options )
    :whenResolvedDo( function( status, data, headers )

        if status == 'data' then

            -- The promise resolves as "data" in the "normal" case of a
            -- valid HTTP response with a non-error (<400) HTTP status code.
            -- The promise also resolves to "data" if the server returns
            -- an HTTP error code when returnHttpErrorsAsData = true.

            -- The data in this case is a KSData with the complete response body.
	    -- and 'header' is  a Lua table of key/value pairs.
            -- A special key 'status_code' in the table contains a string
            -- with the HTTP response code (such as "200").

            -- The default HTTP request method is PUT, but you can override this
            -- with options.requestMethod.

        elseif status == 'error' then

            -- The promise will resolve with "error" status if the system does
            -- not receive a valid HTTP response (i.e. connection refused,
            -- timeout, etc.). In this case, the data argument will be a string
            -- describing the error and headers will be nil.

            -- The promise resolves with "error" status if the system does
            -- not receive a valid HTTP response (for example, if the connection
            -- was refused or timed out). In this case, the data argument is a string
            -- describing the error and headers are nil.

            -- The promise also resolves with "error" if a valid HTTP response
            -- is received but the HTTP status code indicates an error (>= 400).
            -- You can override this behavior with returnHttpErrorsAsData = true.
 
            -- If an HTTP error response is signaled via "error" status, then
            -- data and headers are KSData and table as in the "data" case.

        end )
\endcode
			
	This returns a standard promise. You are responsible for managing the lifetime
	of the promise via the action(s) you attach to it. The HTTP request is
	canceled if all of its promise actions are closed.

\code
local urlDataP, progressB = WFHttpClient.putP ...

progressB:onDataB( function( nBytesSoFar, nBytesExpected )

    -- ...

end )
\endcode
--]]

function WFHttpClient.putP( --[[string]] url,
				--[[KSData or table]] data,
				--[[optional table]] options )

	httpStats:add( 1, nil )
--	httpClientLog:debug( "putP( url = %s, data = %s, options = %s )", url, tostring( data ), options ) -- occasionally useful, but gets REALLY chatty in Xcode
	httpClientLog:debug( "putP( url = %s, options = %s )", url, options )

	options = options or {}
	
	options.requestMethod = options.requestMethod or 'PUT'
	
	-- FIXME, we aren't reporting progress for this.
	local promise = transmitP( uploadRateLimiter, url, options, data )
	
	return promise

end

--------------------------------------------------------------------------------

return WFHttpClient
