--[[----------------------------------------------------------------------------

WFHttpClientImp.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This is the thread that manages access to the native implementation for WFHttpClient.
It should be accessed only by the public API class WFHttpClient.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFRequestDispatcher = require 'WFRequestDispatcher'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.string
local WFZString = require 'WFZString'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local KSDate = require 'KSDate'

	-- wichitafoundation.web.http_client
local WFHttpClientImp_native = require 'WFHttpClientImp_native'
local WFHttpClientHeaderParser = require 'WFHttpClientHeaderParser'

--==============================================================================

local httpCountStats = WFDiagnosticLog.makeRingBuffer {
	title = "Number of HTTP connections",
	accumulate = true,
	"GET", "PUT"
}

local now = KSDate.timeNowForAnimation
local kTimings = setmetatable( {}, { __mode = 'kv' } )
local httpTimingStats = WFDiagnosticLog.makeRingBuffer {
	title = "Native HTTP client timings",
	"GET", "PUT"
}

--------------------------------------------------------------------------------

WFDiagnosticLog.addLogSectionFunc( function()

	return {

		name = "HTTP Client Imp",
		id = 'WFHttpClientImp',
		weight = 390,
		outputForAutomation = true,

		body = {
			counts = { httpCountStats:visualize "boxAndWhisker" },
			timings = { httpTimingStats:visualize "boxAndWhisker" },
		}
	}
end )

--------------------------------------------------------------------------------

local ZSTR = assert( WFZString.fromZString )
local sUserAgent

local function getUserAgent()
	if not sUserAgent then
		-- Configure default user agent
		-- Clients other than Revel will need to add their appIdentifier
		-- to the case below if they want a custom user agent name
		
		local WFPlatformInfo = require "WFPlatformInfo"
		
		local function optFn( fnName, default )
			local fn = WFPlatformInfo[ fnName ]
			if not fn then
				return default
			end
			return fn()
		end
		
		local appNumber = optFn( 'getBuild' )
				
		if not appNumber then
			appNumber = optFn( 'getAppVersion', "0.0" ) .. ".Dev"
		end
		
		-- ensure the appNumber starts with two numbers, separated by a dot
		-- This is required for Oz to recognize us.
		if not string.find( appNumber, "^%d+%.%d+%." ) then
			appNumber = string.gsub( appNumber, "^%d+%.%d+", "%0.0" )
		end
		if not string.find( appNumber, "^%d+%.%d+%." ) then
			appNumber = string.gsub( appNumber, "^%d+", "%0.0.0" )
		end
		if not string.find( appNumber, "^%d+%.%d+%." ) then
			appNumber = "0.0." .. appNumber
		end

		local baseName = optFn( 'getUserAgentAppName' )
		if not baseName then
			baseName = optFn( 'getAppName', 'Wichita Foundation Client' )
		end
		local manufacturer = optFn( 'getDeviceManufacturer', 'Unknown' )
		local model = optFn( 'getDeviceModel', 'Unknown' )
		
		local os
		if WFPlatformInfo.isiOS then
			os = "iOS"
		elseif WFPlatformInfo.isMac then
			os = "OSX"
		elseif WFPlatformInfo.isWinRT then
			os = "WInRT"
		elseif WFPlatformInfo.isWindows then
			os = "Windows"
		elseif WFPlatformInfo.isAndroid then
			os = "Android"
		else
			os = "Unknown"
		end
		
		local osVer = optFn( 'getOSVersion', "?.?.?" )
		
		sUserAgent = string.format(
			"%s(%s;%s;%s;%s;%s)",
			baseName,
			appNumber,
			manufacturer,
			model,
			os,
			osVer )
	end
	return sUserAgent
end

WFHttpClientImp_native.setLocalizedErrorMessages(
	ZSTR "$$$/ASIHTTPRequest/RequestTimedOutError=The request timed out",
	ZSTR "$$$/ASIHTTPRequest/AuthenticationError=Authentication needed",
	ZSTR "$$$/ASIHTTPRequest/RequestCancelledError=The request was cancelled",
	ZSTR "$$$/ASIHTTPRequest/UnableToCreateRequestError=Unable to create request (bad url?)",
	ZSTR "$$$/ASIHTTPRequest/TooMuchRedirectionError=The request failed because it redirected too many times" )

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
 -- Apple concatenates the headers together with a comma and we want to split them
 -- apart.

local function FixLameAppleCookieHeader( header )
	local parsedCookies = {}
	
	-- Here is what we're up against:
	-- MY_COOKIE=BEIE%3Dchtechip; path=/; expires=Fri, 05-Oct-2007 07:41:09 GMT, M_BOKIE=%2C%29_-%3Dsas; path=/through/the/woods; expires=Fri, 05-Oct-2007 07:41:09 GMT
	
	-- Notice the catch here: The expires field has a comma in it, too!
	
	local munged = header
	
	-- Take note of all the expires fields.
	
	local allExpires = {}
	local index = 0
	for insanity = 1, 200 do
			-- bail if we encounter more than 200 cookies
		index = string.find( munged, "expires=", index+1 )
		if index == nil then
			break
		end
		table.insert( allExpires, index )
	end
	
	-- Now split on commas, but not if they're too close to the expires field.

	index = 0
	local last = 1
	for insanity = 1, 200 do
			-- bail if we encounter more than 200 fields in this cookie
			
		local index2
		index, index2 = string.find( munged, ",[ \t]*", index+1 )
		if index == nil then
			local headerValue = string.sub( munged, last )
			table.insert( parsedCookies, headerValue )
			break
		end

		local tooClose = false

		for _, x in ipairs( allExpires ) do

			if index > x and index - x < 13 then

					-- 13 > the length of 'expires=Mdy'
					-- the RFC says the field is MUCH longer, so this
					-- should be fine.  The only thing I wonder is if
					-- some server out there uses an illegal day of the
					-- week token that's more than 3 chars.

				tooClose = true

			end

		end

		if not tooClose then

			local headerValue = string.sub( munged, last, index-1 )
			last = index2+1
			table.insert( parsedCookies, headerValue )

		end

	end
			
	return parsedCookies
end

local kRawHeaderKey = "< raw : headers >"
local kCookieHeaderKey = "Set-Cookie"

local function standardizeHeaders( headers )
	-- parse headers on windows which are returned as a single string
	local rawHeaders = headers and headers[ kRawHeaderKey ]
	if rawHeaders and type( rawHeaders ) == 'string' then
		
		for k, v in pairs( WFHttpClientHeaderParser.parseHeaderString( rawHeaders ) ) do
			headers[ k ] = v
		end
		
		headers[ kRawHeaderKey ] = nil
	end
	
	-- parsing for apple headers which are a table like we want, but have multiple Set-Cookie
	-- headers sent as a single comma-separated string
	local cookieHeader = headers and headers[ kCookieHeaderKey ]
	if cookieHeader and type( cookieHeader ) == 'string' then
	
		headers[ kCookieHeaderKey ] = FixLameAppleCookieHeader( cookieHeader )
	
	end
	
	return headers
	
end

--------------------------------------------------------------------------------

local httpClientImpLog = KSLogging.log 'WFHttpClientImp'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local sherpaForRequest = {}
	-- key (native object): the native HTTP request object (only responds to :close)
	-- value (httpRequestSherpa instance)

local resolverForRequest = {}
	-- key (native object): the native HTTP request object (only responds to :close)
	-- value (uploadChunkResolver instance)

--============================================================================--

-- @mark httpRequestSherpa (internal to this file)

local httpRequestSherpaMethods = {

	_httpRequest = false,
		-- (native object) the native HTTP request object (only responds to :close)
	
	_didResolve = false,
		-- (boolean) true if we've resolved this object
	
	_url = false,
		-- (string) URL we're fetching or putting to
	
	_options = false,
		-- (table) options passed to client thread getP or putP function
	
	_optData = false,
		-- (KSData or number) data to pass with request or total size if multipart post
	
	_returnHttpErrorsAsData = false,
		-- (boolean) true to resolve HTTP status >= 400 via 'data' status;
		-- false to resolve via 'error' status
	
	_returnHttpErrorsAsStatusCode = false,
		-- (boolean) true to return HTTP errors as simply the status_code;
		-- needed for backward compatibility with the old event-stream based API
	
	_makeRequestFn = false,
		-- (function) factory for _httpRequest object
	
	_headers = false,
		-- (table) headers received (they come in before data )

	_messagesE = false,
		-- E< what, ... > an event stream for sending info back to the client. If
		-- what is "progress", then the ... = nBytesSoFar, nBytesExpected;
		-- If what is "result", then ... = data
	
	WFCloseSupport.addCloseSupportMethods,
		
}

--------------------------------------------------------------------------------





function httpRequestSherpaMethods:__init( url, options, optData )

	self._url = url
	self._returnHttpErrorsAsData = options.returnHttpErrorsAsData or false
	self._returnHttpErrorsAsStatusCode = options.returnHttpErrorsAsStatusCode or false
	
	self._options = options
	self._optData = optData or false
		-- If we don't put the "or false" clause in, then we error out with
		-- illegal read on this field later.
		
		
	self._messagesE = KSReactive.sourceE()
	
	self._messagesE:addCloseTarget( self )

	httpCountStats:add( optData and 0 or 1, optData and 1 or 0 )
	kTimings[ self ] = now()
	
	-- Make the request.
	self._httpRequest = WFHttpClientImp_native.getOrPut( self._url,
														 self._options,
														 self._optData or nil )
	if self._optData and type( self._optData ) ~= 'string'
		and type( self._optData ) ~= 'number' then
		-- TODO: MMLJ - i think we want to do this differently for optData strings ... wrap in function
		-- that sends and closes...
		
		-- release memory used by data ASAP
		self._optData:close();
		self._optData = 'closed'
	end
	
	sherpaForRequest[ self._httpRequest ] = self
	
end

--------------------------------------------------------------------------------

function httpRequestSherpaMethods:close()
	
	-- Close off the HTTP request.

	httpCountStats:add( self._optData and 0 or -1, self._optData and -1 or 0 )
	local time = now() - kTimings[ self ]
	kTimings[ self ] = nil
	httpTimingStats:add( self._optData and 0 or time, self._optData and time or 0 )
	
	local request = self._httpRequest
	
	if request then

		self._httpRequest = false
		request:close()
		
		sherpaForRequest[ request ] = nil
		
	end
	
end

--------------------------------------------------------------------------------

function httpRequestSherpaMethods:startResolution( priority )

	assert( self._httpRequest, "Requests should have been created in __init" )
	
end

--------------------------------------------------------------------------------


local HttpRequestSherpa = KSClass 'HttpRequestSherpa'( httpRequestSherpaMethods )

-- @end HttpRequestSherpa


--============================================================================--

-- @matthewj uploadChunkResolver class (internal to this file)

local uploadChunkResolverMethods = {
	
	_didResolve = false,
		-- (boolean) true if we've resolved this object
	
	_httpRequest = false,
	
	_chunkNumber = false,
		-- sequence number identifying the chunk
	
	_data = false,
		-- chunk data to pass with upload
	
	KSPromises.addPromiseResolverMethods,
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function uploadChunkResolverMethods:__init( request, chunkNum, data )

	self._httpRequest = request
	self._data = data or false
	self._chunkNumber = chunkNum
end

--------------------------------------------------------------------------------

function uploadChunkResolverMethods:close()
	
	-- Close off the upload chunk data.

	local request = self._httpRequest
	
	if ( self._data and type(self._data) ~= 'string'  ) then
		self._data:close()
	end
	
	if request then

		self._httpRequest = false
		-- note, it's not the uploadChunk's job to close the request
		if ( self == resolverForRequest[ self._chunkNumber ] ) then
			resolverForRequest[ self._chunkNumber ] = nil
		end
	end

end

--------------------------------------------------------------------------------

function uploadChunkResolverMethods:startResolution( priority )

	-- Ignore restart attempts (i.e. priority changes).
	WFHttpClientImp_native.uploadChunk( self._httpRequest, self._data or "", self._chunkNumber)
														
	resolverForRequest[ self._chunkNumber ] = self
	
end

--------------------------------------------------------------------------------

local UploadChunkResolver = KSClass 'UploadChunkResolver'( uploadChunkResolverMethods )

-- @end UploadChunkResolver

--============================================================================--

-- This function gets triggered from native code when it has data to report
-- for any of the outstanding requests.

WFHttpClientImp_native.registerCompletionHandler( function( request, status, dataType, data )

	local sherpa = sherpaForRequest[ request ]
	if not sherpa then
		httpClientImpLog:warn( "Completion handler got called for a request that we don't recognize" )
		return
	end

	if sherpa._didResolve then
		httpClientImpLog:warn( "Invoking completion handler after promise was resolved" )
		return
	end

	if status == 'error' then
		sherpa._didResolve = true
		sherpa._messagesE:post( 'result', 'error', dataType, nil,  data )
		return
	end
	
	if status == 'chunkrequest' then
		sherpa._messagesE:post( 'chunkrequest', request, data ) -- data is chunknum
		
		return
	end
	
	if status == 'chunkdone' then
		local uploadResolver = resolverForRequest[ data ]
		
		uploadResolver:resolvePromise( 'done' )

		return
	end

	if dataType == 'header' then
	
		standardizeHeaders( data )

		-- Save headers off; we'll include them as part of the final response.
		
		sherpa._headers = data

	elseif dataType == 'data' then

		if data and type( data ) ~= 'string' then
			WFRequestDispatcher.deferredClose( data )
		end

		local headers = sherpa._headers

		if not headers then
			httpClientImpLog:error( "Huh ... got data but not headers!" )
		elseif headers.status_code >= '400' then
			httpClientImpLog:info( "HTTP error on %s: \n\n%s\n\n%s", sherpa._url, headers, data )
		end

		sherpa._didResolve = true
		
		if sherpa._returnHttpErrorsAsData or ( headers and headers.status_code < '400' ) then
			sherpa._messagesE:post( 'result', 'data', data, headers )
		elseif sherpa._returnHttpErrorsAsStatusCode then
			sherpa._messagesE:post( 'result', 'error', headers and headers.status_code )
		else
			sherpa._messagesE:post( 'result', 'error', data, headers )
		end
		sherpa._messagesE:close()
		sherpa._messagesE = nil
	
	elseif status == 'progress' then

		sherpa._messagesE:post( 'progress', dataType, data )	-- really nBytesSoFar, nBytesExpected
			
	else
	
		httpClientImpLog:error( "bogus dataType from native code: %s", dataType )
	
	end
	
end )

--============================================================================--

-- return true if the headers specified by these options include a User-Agent.
-- Also return the header array, creating one in options.requestHeaders if necessary.
local function hasUserAgentHeader( options )
	local headers = options.requestHeaders
	if not headers then
		headers = {}
		options.requestHeaders = headers
	end
	local hasUserAgent = false
	for _, h in ipairs( headers ) do
		if h.header == 'User-Agent' then
			hasUserAgent = true
		end
	end
	return hasUserAgent, headers
end

--------------------------------------------------------------------------------

WFRequestDispatcher.onMessageDo {

	message = 'getOrPut',
	onlyIfCallbackIsStillValid = true,
	
	handler = function( url, options, optData )
	
		local hasUserAgent, headers = hasUserAgentHeader( options )
		if not hasUserAgent then
			headers[ #headers + 1 ] = { header = 'User-Agent', value = getUserAgent() }
		end
	
		httpClientImpLog:debug( "getOrPut( url = %s, options = %s )", url, options )

		local sherpa = HttpRequestSherpa( url, options or {}, optData )
		return sherpa._messagesE

	end,

}

WFRequestDispatcher.onMessageDo {
	message = 'uploadChunk',
	
	onlyIfCallbackIsStillValid = true,
	
	handler = function( request, chunkNum, data )
	
		httpClientImpLog:debug( "uploadChunk( request = %s )", request )
		
		local resolver = UploadChunkResolver( request, chunkNum, data )
		
		return KSPromises.promiseForResolverP( resolver, 'retained | closeTarget' )

	end,

}

--============================================================================--

WFRequestDispatcher.onMessage( 'DEBUG_dumpCoverageP', function( path )

	require 'SetupCoverage'.dumpCoverage( path )

	return require 'KSPromises'.resolvedPromiseP( 'data', true )

end )

