--[[----------------------------------------------------------------------------

WFHttpClientTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author scouten 2011-02-02

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.util
--local WFConfiguration = import "WFConfiguration"

	-- wichitafoundation.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- wichitafoundation.web.http_client
local WFHttpClient = require 'WFHttpClient'
local KSDate = require "KSDate"

--==============================================================================

-- infer how far we are from the server by examing a local time
local localTimezone = KSDate.formatDateTimeAsLocal( 0 )
local multiplier = 1

if localTimezone == "2000-12-31T18:00:00" then
	multiplier = 2
end

local function getTimeout( nearbyTimeout )
	return nearbyTimeout * multiplier
end

--============================================================================--

local WFHttpClientTest = {}

--------------------------------------------------------------------------------

function WFHttpClientTest.tearDown( coveragePath )

	if coveragePath then
		
		local WFWorkerThread = require "WFWorkerThread"
		local HttpClientEP = WFWorkerThread( "WFHttpClientImp.lua" )
		HttpClientEP:requestP( 'DEBUG_dumpCoverageP', coveragePath )
		:runUntilDone()
		
	end

end

--------------------------------------------------------------------------------

function WFHttpClientTest.testHeaderParser( test )
	
	local WFHttpClientHeaderParser = require 'WFHttpClientHeaderParser'
	assert( WFHttpClientHeaderParser, "Failed to requiredWFHttpClientHeaderParser" )
	local parseHeaderString = WFHttpClientHeaderParser.parseHeaderString
	
	local exampleInput = "Header-Name: Header value\r\nHeader2-Name:Header 2 Value\r\nHeader3-Name:	 	Header 3\r\n gots value\r\n\r\n"
	local headers = parseHeaderString( exampleInput )

	test:log( "Dumping headers:" )
	for k, v in pairs( headers ) do
		test:log( string.format( "%s = %s", k, v ) )
	end
	test:log( "Done." )
	assert( headers[ "Header-Name" ] == "Header value", "Basic space separated one-line header value" )
	assert( headers[ "Header2-Name" ] == "Header 2 Value", "Basic no-whitespace separated one-line header value" )
	assert( headers[ "Header3-Name" ] == "Header 3 gots value", "Multiple whitespaces, including tabs, with a multi-line header value" )
	
end

--------------------------------------------------------------------------------

-- needs tests for the promise version

--------------------------------------------------------------------------------

function WFHttpClientTest.testSimpleGetP( test )

	WFHttpClient.getP(
		'http://lr-httptest.pxswarm.com/http-get-test-data.txt',
		{
			requestMethod = "GET",
			timeout = 1,
		})
		:whenResolvedDo( function( status, data )
		
			if status ~= 'data' then
				
				debug.print( "unexpected promise status ", status )
				test:notifyTestDidFail()
		
			elseif tostring( data ) == "Hello, Wichita!\n" then

				test:notifyTestDidSucceed()

			else
			
				test:log( "HTTP GET returned incorrect data" )
				debug.print( "HTTP GET returned incorrect data " .. tostring( data ) )
				test:notifyTestDidFail()
			
			end
					
		end )
		:runUntilDone()

	collectgarbage()
	collectgarbage()
	collectgarbage()

	test:waitForTestToFinish( getTimeout( 10 ) )

end

--------------------------------------------------------------------------------

function WFHttpClientTest.testSimpleStringPutP( test )

	WFHttpClient.putP(
		'http://lr-httptest.pxswarm.com/AgHttpClientTest3.post.cgi',
		"Post Data",
		{
			requestMethod = "POST",
			timeout = 1,
			requestHeaders = {
				{ header = { invalid=true }, value = "invalid key, should log bug not fail", },
				{ header = "Header Has Invalid Value Will Log But Won't Fail", value = { invalid=true }, },
				{ header = "X-Unit-Test-Ignore-This", value = "ignore", },
			},
		})
		:whenResolvedDo( function( status, data )
		
			if status ~= 'data' then
				
				debug.print( "unexpected promise status ", status )
				test:notifyTestDidFail()
		
			elseif tostring( data ) == "<p>Thanks. You said: Post Data</p>" then

				test:notifyTestDidSucceed()

			else
			
				test:log( "HTTP GET returned incorrect data" )
				debug.print( "HTTP GET returned incorrect data " .. tostring( data ) )
				test:notifyTestDidFail()
			
			end
					
		end )
		:runUntilDone()

	collectgarbage()
	collectgarbage()
	collectgarbage()

	test:waitForTestToFinish( getTimeout( 10 ) )

end

--------------------------------------------------------------------------------

function WFHttpClientTest.DISABLE_testGetPWithProgress( test )

	-- FIXME [scouten 2012-03-26]: Progress doesn't seem to be used by anybody
	-- at the moment, so disabling for now.

	local resultP, progressB = WFHttpClient.getP(
		'http://lr-httptest.pxswarm.com/http-get-test-data.txt',
		{ wantDownloadProgress = true } )
	
	resultP:whenResolvedDo( function( status, data )
		
			if status ~= 'data' then
				
				debug.print( "unexpected promise status ", status )
				test:notifyTestDidFail()
		
			elseif tostring( data ) == "Hello, Wichita!\n" then

				test:notifyTestDidSucceed()

			else
			
				debug.print( "HTTP GET returned incorrect data " .. tostring( data ) )
				test:notifyTestDidFail()
			
			end
					
		end )
		:runUntilDone()

	progressB:onDataB( function( nBytesSoFar, nBytesTotal )
	
		debug.print( "progressB returned nBytesSoFar = ", nBytesSoFar,
					 "nBytesTotal = ", nBytesTotal )
	
	end )

	collectgarbage()
	collectgarbage()
	collectgarbage()

	test:waitForTestToFinish( getTimeout( 10 ) )

end

--------------------------------------------------------------------------------

function WFHttpClientTest.testGetPConverts404ToError( test )

	WFAsyncTestCase( test, getTimeout( 10 ) )

		:waitFor( function()
		
			return WFHttpClient.getP(
				'http://lr-httptest.pxswarm.com/http-get-test-data.txtx' )
			
		end )
	
		:whenResolvedDo( function( status, data, headers )
		
			if status ~= 'error' then
				error( "Unexpected status " .. tostring( status ) )
			end
			
			if not( headers and headers.status_code == '404' ) then
				error( "HTTP GET returned unexpected status" .. data .. headers )
			end
			
		end )
end

--------------------------------------------------------------------------------

function WFHttpClientTest.testGetPDisableErrorConversion( test )

	WFAsyncTestCase( test, getTimeout( 10 ) )

		:waitFor( function()
		
			return WFHttpClient.getP(
				'http://lr-httptest.pxswarm.com/http-get-test-data.txtx',
					{ returnHttpErrorsAsData = true } )
		
		end )
		
		:whenResolvedDo( function( status, data, headers )
		
			if status ~= 'data' then
				error( "Unexpected status " .. tostring( status ) )
			end
			
			if not( headers and headers.status_code == '404' ) then
				error( "HTTP GET returned unexpected status" .. data .. headers )
			end
			
		end )

end

--------------------------------------------------------------------------------

function WFHttpClientTest.testGetPWithRequestHeader( test )

	WFAsyncTestCase( test, getTimeout( 10 ) )
	
		:waitFor( function()
		
			return WFHttpClient.getP(
				'http://lr-httptest.pxswarm.com/http-get-test-data.txt',
				{
					requestHeaders = {
						{ header = 'Accept', value = 'application/json' },
					}
				} )
			
		end )
			
		:whenResolvedDo( function( status, data, headers )
		
			if status ~= 'data' then
				error( "Unexpected status " .. tostring( status ) )
			end
			
			assert( tostring( data ) == "Hello, Wichita!\n",
					"HTTP GET returned incorrect data " .. tostring( data ) )
		
		end )

end

--------------------------------------------------------------------------------

function WFHttpClientTest.testGetPWithReadResponseHeaders( test )

	WFAsyncTestCase( test, getTimeout( 10 ) )
	
		:waitFor( function()
		
			return WFHttpClient.getP(
				'http://lr-httptest.pxswarm.com/http-get-test-data.txt' )
		
		end )
		:whenResolvedDo( function( status, data, headers )
			
			if status ~= 'data' then
				error( "Unexpected status " .. tostring( status ) )
			end
			
			require 'KSLogging'.log( 'debug' ):debug( headers )
			if headers[ 'Content-Type' ] ~= "text/plain" then
				error( "Unexpected Content-Type: " .. headers[ 'Content-Type' ] )
			end
			
			if headers.status_code ~= '200' then
				error( "Unexpected status_code: " .. headers.status_code )
			end
			
			assert( tostring( data ) == "Hello, Wichita!\n",
					"HTTP GET returned incorrect data " .. tostring( data ) )
		
		end )

end
	
--------------------------------------------------------------------------------

return WFHttpClientTest
