--[[----------------------------------------------------------------------------

WFIndexStore.lua

@author scouten 2011-03-28
@reviewed scouten 2012-08-22

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.utils
local KSClass = require 'KSClass'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'
local KSBucketBrigade = require 'KSBucketBrigade'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )

local DEBUG = _G.DEBUG


--============================================================================--
--[[!
	\class WFIndexStore
	\brief Tracks an \c OzDocStore and provides SQL-style indexing of its contents.
	
	\todo Write more narrative describing how it gets used.
--]]

-- @mark WFIndexStore

local indexStoreMethods = {

	_indexStoreWorkerThread = false,
		-- (WFWorkerThread) reference to the event processor that does the indexing

	_docStoreId = false,
		-- (string) doc store ID, used for debugging
		
	_queryResultsCache = false,
		-- (fully weak table) a mapping from cache keys to query results
		
	_queryResultsBrigade = false,
		-- (KSBucketBrigade) used to avoid having GC wipe things out too fast
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Creates an index store object that tracks the contents of an Oz document store.
	
	\param docStore (\c WFDocStore or \c OzDocStore) The document store object.
	\param indexSpecScript (string) The name of an index definition script for this index
--]]

function indexStoreMethods:__init( --[[WFDocStore or OzDocStore]] docStore,
								   --[[string]] indexSpecScript )

	check( 'WFIndexStore', 'WFDocStore|OzDocStore', '+string' )

	local docStoreId
	
	if WFDocStore:isClassOf( docStore ) then
		docStoreId = docStore:path()
	else
		docStoreId = 'OzDocStore ' .. docStore:catalogId()
	end

	self._docStoreId = docStoreId
	
	self._queryResultsCache = setmetatable( { }, { __mode = 'kv' } )
	
	self._queryResultsBrigade = KSBucketBrigade.makeBrigade( 5 ) -- depth 5 (5 GC cycles)
	
	self._indexStoreWorkerThread = WFWorkerThread( indexSpecScript, docStoreId )
	self:addCloseTarget( self._indexStoreWorkerThread )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Performs a single SQL query on the index store and returns the result
		as a Promise.

		Before performing the query, the method brings all relevant index tables
		up to date. If data is added to the doc store while this request
		is being processed, the newer data might or might not appear in the query results.
		
	\param sql (string) The SQL query to be performed. OR a table:
		- \c sql: (string) SQL query to be performed
		- \c tempIds: (optional, array of doc ids) if present, the table TEMP_ids will
			be populated with these ids for use in the query, e.g. as part of
			an join such as "INNER JOIN TEMP_ids ON TEMP_ids.docId = yourTable.docId"
			or via a where in clause such as "WHERE docId IN ( SELECT docId from TEMP_ids )"
			TEMP_ids is a temporary table with only one column, docId, and no special indexes.
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	\param ... (optional) The query arguments to be substituted for ? placeholders in the query.
		ONLY if sql was provided as a string, not as the table form.
	
	\return <code>P< table ></code> A Promise that resolves to the query results,
		as an array of arrays.
--]]

function indexStoreMethods:queryP( sql, ... )

	check( 'WFIndexStore', 'string|table' )
	
	if type( sql ) == 'string' then
		sql = {
			sql = sql,
			...
		}
	end
	
	return self._indexStoreWorkerThread:requestP( 'queryP', sql )

end

--------------------------------------------------------------------------------

local function getCachedResult( self, cacheKey )

	if not cacheKey then
		return
	end
	
	local entry = self._queryResultsCache[ cacheKey ]
	
	if not entry then
		return
	end
	
	self._queryResultsBrigade( entry )
	
	return entry.data

end

--------------------------------------------------------------------------------

local function storeCachedResult( self, cacheKey, value )

	if not cacheKey or not value or not WFUnknown.isKnown( value ) then
		return
	end
	
	local entry = self._queryResultsCache[ cacheKey ]
	
	if entry then
		entry.data = value
	else
		entry = { data = value }
		self._queryResultsCache[ cacheKey ] = entry
	end
	
	self._queryResultsBrigade( entry )

end

--------------------------------------------------------------------------------

local kOutOfDate = { outOfDate = true }

local kUnknown = WFUnknown.value()
	
--------------------------------------------------------------------------------

--[[!
	\brief Reports as a Behavior  the results of a SQL query on the index store.
	
	\param args (table) arguments for query
		- \c sql: (string) SQL query to be performed
		- \c calm: (optional, number) if present, calm query result updates to no more
			than one every _n_ seconds
		- \c outOfDateNotice: (optional, boolean) if true, query results will contain
			a member "outOfDate" (boolean) which will be true if query results are
			known to be lagging behind the doc store
		- \c cacheKey: (optional, string) if supplied, this will be used as a cache
			value to look up cached query results; do not use in conjunction with
			outOfDateNotice. (outOfDateNotice will win)
		- \c nArgs: (optional, number) if present, specifies number of arguments
			(overrides #args)
		- \c tempIds: (optional, array of doc ids) if present, the table TEMP_ids will
			be populated with these ids for use in the query, e.g. as part of
			an join such as "INNER JOIN TEMP_ids ON TEMP_ids.docId = yourTable.docId"
			or via a where in clause such as "WHERE docId IN ( SELECT docId from TEMP_ids )"
			TEMP_ids is a temporary table with only one column, docId, and no special indexes.
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	
	For compatibility with an older version of this API, this function also permits
	the first argument to be a string (interpreted as the SQL query string) followed
	by any number of query parameters as ... arguments. Future code should use the
	new named-argument syntax. Note that \c calm and \c outOfDateNotice are inaccessible
	through the old syntax.
	
	You may use a behavior for \c sql and any of the query arguments. In this case
	the actual query performed will update in response to changes to the behavior(s).
	
	\return B< table > A Behavior whose value tracks query results, as an array of arrays.
	
	\b WARNING: Use this method sparingly. Creating many query behaviors will adversely
		affect the overall performance of the application.
	
	\b WARNING: This behavior may, at times, return information that is based on
		incomplete index table results. It will eventually catch up and update
		results if needed. (Out of date information is especially likely for the
		first value returned.)

--]]

function indexStoreMethods:queryB( args, ... )

	check( 'WFIndexStore', 'string|table' )

	-- Retrofit for old syntax.

	local _, isBehavior = KSReactive.isEventNode( args )

	if type( args ) == 'string' or isBehavior then
		local nArgs = select( '#', ... )
		args = { sql = args, nArgs = nArgs, ... }
	end

	-- Fill in thread and traceback info if in debug build.

	args.threadName = DEBUG and WFCurrentThread.DEBUG_getThreadName() or nil
	args.traceback = DEBUG and debug.traceback( '', 2 ) or nil
	
	-- Are there any behaviors in params?
	
	local sqlIsBehavior = KSReactive.isBehavior( args.sql )
	local argsHaveBehaviors = false
	
	local nArgs = args.nArgs or #args
	
	for i = 1, nArgs do
		if KSReactive.isBehavior( args[ i ] ) then
			argsHaveBehaviors = true
			break
		end
	end
	
	-- Reconstruct the query as needed for behavior changes.
	
	local indexStoreWorkerThread = self._indexStoreWorkerThread
	
	local function makeQueryRequestE( sql, ... )

		local nArgs = select( '#', ... )
		local savedArgs = nArgs > 0 and { ... } or nil
	
		local newArgs = {
			sql = sql,
			calm = args.calm,
			outOfDateNotice = args.outOfDateNotice,
			tempIds = args.tempIds,
			nArgs = nArgs,
			...
		}
		
		assert( not WFCloseSupport.hasBeenOrWillBeClosed( self ),
			    "queryB running after index store closed" )

		local requestE = indexStoreWorkerThread:requestE( 'queryB', newArgs )
		
		local queryE = requestE:mapE( function( results )
			
				-- Pass back the SQL and args from results so clients
				-- can see if input changes have been tracked.
			
				results.sql = sqlIsBehavior and sql or nil
				results.args = argsHaveBehaviors and savedArgs or nil

				return results
			
			end )
			
		queryE:addCloseTarget( requestE )
		
		return queryE
	
	end
	
	local function makeQueryRequestBWithInitialValue( initialValue, ... )
		local e = makeQueryRequestE( ... )
		local b = e:closingHoldB( initialValue )
		return b
	end
		
	local initialValue
	
	local cacheKey
	
	if args.outOfDateNotice then
		initialValue = kOutOfDate
	else
		cacheKey = args.cacheKey
		initialValue = getCachedResult( self, cacheKey ) or kUnknown
	end
	
	local resultB
	
	-- If we are tracking the inputs closely (something I would have thought
	-- would be the default but it isn't), then we keep dropping back to the
	-- initialValue when the inputs change. -- MarkH
	
	if args.trackInputs then

		local function makeQueryRequestB_initialValue( ... )
			return makeQueryRequestBWithInitialValue( initialValue, ... )
		end
		
		resultB = KSReactive.dataCallB( makeQueryRequestB_initialValue,
						args.sql, unpack( args, 1, nArgs ) )
										

	else
	
		local kDropThis = { }
	
		local function makeQueryRequestB_dropThis( ... )
			return makeQueryRequestBWithInitialValue( kDropThis, ... )
		end
	
		local queryRequestB = KSReactive.dataCallB( makeQueryRequestB_dropThis,
						args.sql, unpack( args, 1, nArgs ) )
						
		-- Remove drop this from the stream. Use initialValue at the start
		-- if we were starting with drop this.
						
		resultB =
			queryRequestB:filterB(
				function( x ) return x ~= kDropThis end,
				initialValue )
				
		resultB:addCloseTarget( queryRequestB )
						
	end
	
	if cacheKey then
		resultB = resultB:closingMapB( function( ... )
			storeCachedResult( self, cacheKey, ( ... ) )
			return ...
		end )
	end
	
	self:addCloseTarget( resultB )
	
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the number of documents left to index.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return (<code>B< number ></code>) A Behavior whose value is the number of documents.
--]]

function indexStoreMethods:sizeOfIndexQueueB()

	local resultB = self._indexStoreWorkerThread:requestE( 'sizeOfIndexQueueB' )
		:closingHoldB( 0 )
	
	self:addCloseTarget( resultB )
	
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Permanently shuts down this index store and deletes all related files.
--]]

function indexStoreMethods:haltAndDeleteP()

	return self._indexStoreWorkerThread:requestP( 'haltAndDelete' )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

function indexStoreMethods:DEBUG_dumpCoverageP( path )

	-- Should be used only by WFIndexStoreTest.lua.

	local resultP = self._indexStoreWorkerThread:requestP( 'DEBUG_dumpCoverageP', path )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

return KSClass 'WFIndexStore'( indexStoreMethods )
