--[[----------------------------------------------------------------------------

WFIndexStoreMapServer.lua

@author scouten 2012-08-14
@reviewed scouten 2012-08-14

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local KSLogging = require 'KSLogging'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- WF.core.utils
local KSClass = require 'KSClass'
local KSIdGenerator = require 'KSIdGenerator'
local KSLuaUtils = require 'KSLuaUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )

local mapServerLog = KSLogging.log 'WFIndexStoreMapServer'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--
--[[!
	\brief A \c WFIndexStoreMapServer is used by the index store implementation
		to keep the index store tables up to date.
	
	It should only be accessed by \c WFIndexStoreImp.
--]]

local mapServerMethods = {

	_docStore = false,
		-- (WFDocStore or OzDocStore) doc store we're getting our data from

	_docStoreIsWf = false,
		-- (boolean) true if doc store is a WFDocStore

	_db = false,
		-- (KSDatabase)
	
	_db_exec = false,
		-- (function) cached reference to self._database:exec
	
	_tableNames = false,
		-- (table) names of the index tables declared in this schema
	
	_mapFunctions = false,
		-- (table) map function for each table in this schema
			-- key: (string) name of index table
			-- value: (function) map function declared for this table
	
	_nColumnsPerTable = false,
		-- (table) how many user-defined columns are in each index table?
			-- key: (string) name of index table
			-- value: (number) how many user-defined columns?
	
	_updateTimePerTable = false,
		-- (table)
			-- key: (string) name of index table
			-- value: (table) info on statement (count, min, max, total)

	_isReadyP = false,
		-- (P< true | "out of sync" >) resolves when we have verified that
			-- index store and doc store are properly related; resolves to
			-- true in most normal circumstances or "out of sync" if index
			-- and doc store were unrelated to each other

	_docChangeCounterB = false,
		-- (B< number >) doc change counter from underlying doc store

	_needsUpdateFromBForTable = false,
		-- (table) how up to date is each index table?
			-- key: (string) name of index table
			-- value: (B< number >) change counter for first document
				-- *not* yet indexed

	_tableNeedsUpdateThru = false,
		-- (table) are there any pending update requests for each table?
			-- key: (string) name of index table
			-- value: (number) maximum change counter value requested for this table

	_existingRowQueryForTable = false,
		-- (table) query to gather existing rows
			-- key: (string) name of index table
			-- value: (string) SELECT * FROM (table name) WHERE docId == ?

	_deleteExistingRowsStmtForTable = false,
		-- (table) query to delete existing rows
			-- key: (string) name of index table
			-- value: (string) DELETE FROM (table name) WHERE docId == ?

	_insertRowStmtForTable = false,
		-- (table) sql statment to insert new rows
			-- key: (string) name of index table
			-- value: (string) INSERT INTO (table name) VALUES( ... )

	_tableUpdatedE = false,
		-- (KSReactive.makeWeakEventTable) table update event stream for each table

	_pendingChangedDocsRequestP = false,
		-- (P< table >) pending changed documents request, if any
	
	_nDocumentsToIndexB = false,
		-- (B< number >) how many documents are remaining to index?
	
	_lastTempIds = false,
	
	_niceToHaveIndices = false,
		-- (table)
			-- key (string) : table name as declared in SQL or fields
			-- value (table): a set of columns that are nice to have (often the
			-- empty set but never nil)
	
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local docStoreCrossCheckP -- function defined later

--------------------------------------------------------------------------------
--[[!
	\brief Create a map server to support an index store implementation.
	
	Note that the map server assumes ownership (and close targeting) of
	\c docStore and \c indexStoreDb.
	
	\param docStore (\c WFDocStore or \c OzDocStore) doc store providing the data
	\param indexStoreDb (\c KSDatabase) SQLite database for index store
	\param indexStoreSchema (\c WFIndexStoreSchema) index store schema description
--]]

function mapServerMethods:__init( --[[WFDocStore or OzDocStore]] docStore,
								  --[[KSDatabase]] indexStoreDb,
								  --[[WFIndexStoreSchema]] indexStoreSchema )

	check( 'WFIndexStoreMapServer', 'WFDocStore|OzDocStore', 'KSDatabase',
		   'WFIndexStoreSchema' )

	mapServerLog:info( "Spinning up map server for index store %s", tostring( docStore ) )

	self._docStore = docStore
	self:addCloseTarget( docStore )
	
	self._docStoreIsWf = WFDocStore:isClassOf( docStore )

	self._db = indexStoreDb
	self:addCloseTarget( indexStoreDb )

	self._db_exec = indexStoreDb.exec
	
	WFCloseSupport.mustCloseAtThreadTermination( self )

	self._updateTimePerTable = {}
	self._tableNeedsUpdateThru = {}

	self._tableNames = assert( indexStoreSchema:tableNames() )
	self._mapFunctions = assert( indexStoreSchema:mapFunctions() )
	self._niceToHaveIndices = assert( indexStoreSchema:niceToHaveIndices() )
	self._nColumnsPerTable = assert( indexStoreSchema:nColumnsPerTable() )

	self._existingRowQueryForTable = {}
	self._deleteExistingRowsStmtForTable = {}
	self._insertRowStmtForTable = {}

	self._tableUpdatedE = KSReactive.makeWeakEventTable()
	self:addCloseTarget( self._tableUpdatedE )
	
	self._needsUpdateFromBForTable = {}
	
	for tableName in pairs( self._tableNames ) do
		local behavior = KSReactive.variableB( 0 )
		self._needsUpdateFromBForTable[ tableName ] = behavior
		self:addCloseTarget( behavior )
	end

	self._nDocumentsToIndexB = KSReactive.variableB( 0 )

	-- Start tracking changes in doc store.
	
	self._docChangeCounterB = self._docStore:docChangeCounterB()
	self:addCloseTarget( self._docChangeCounterB )
	
 	-- Make sure doc store is the same one we thought it was before.

	self._isReadyP = docStoreCrossCheckP( self, indexStoreSchema )
	self._isReadyP:whenResolvedDo( function() end ):attachTo( self )

end

--------------------------------------------------------------------------------

local function populateNeedsUpdateFromBForTables( self )

	-- Called once by docStoreCrossCheckB. Preload change counters for all
	-- tables from database.

	local db = self._db

	local indexStatusRows = db:gatherArrayOfArray
		[[SELECT indexTableName, needsUpdateFrom FROM indexStatus]]

	for i = 1, #indexStatusRows do
	
		local row = indexStatusRows[ i ]
		local tableName = row[ 1 ]
		local needsUpdateFrom = row[ 2 ]
		
		if self._needsUpdateFromBForTable[ tableName ] then
			self._needsUpdateFromBForTable[ tableName ]:set( needsUpdateFrom )
		end
	
	end
	
end

--------------------------------------------------------------------------------

local function setDocStoreCrossCheckP( self, crossCheck )

	if self._docStoreIsWf then

		return self._docStore:setDocValueP
			{
				_id = 'indexStoreCrossCheck',
				_slot = 'wfIndexStore',
				crossCheck = crossCheck,
			}
	
	else

		return self._docStore:setLocalStateP( 'indexStoreCrossCheck',
											  crossCheck )
	
	end

end

--------------------------------------------------------------------------------

docStoreCrossCheckP = function( self, indexStoreSchema )

	local db = self._db

 	-- Make sure doc store is the same one we thought it was before.

	local existingRow = db:execOnce
		[[SELECT value FROM localState WHERE name == 'docStoreCrossCheck']]

	-- If we have no check code in the index store, then we create one.

	local expectedCheckCode = existingRow and existingRow.value
	
	local expectedCheckCodeIsNew = not expectedCheckCode

	if expectedCheckCodeIsNew then
	
		expectedCheckCode = KSIdGenerator.uuidInLowercaseHex()

		mapServerLog:debug( "writing new docStoreCrossCheck = %s",
							expectedCheckCode )

		db:execOnce(
			[[INSERT OR REPLACE INTO localState values( 'docStoreCrossCheck', ? )]],
			expectedCheckCode )

	end

	mapServerLog:debug( "expectedCheckCode is %s is new = %s",
						expectedCheckCode,
						expectedCheckCodeIsNew )

	-- Does the index store cross-check value match the doc store's cross-check?

	local docStoreCrossCheckP
	
	if self._docStoreIsWf then

		docStoreCrossCheckP = self._docStore:docValueP( 'indexStoreCrossCheck',
														'wfIndexStore' )
			:mapDataP( function( _, docValue )
				return 'data', docValue.crossCheck
			end )

	else

		docStoreCrossCheckP = self._docStore:localStateP( 'indexStoreCrossCheck' )

	end
	
	return docStoreCrossCheckP:mapDataP( function( status, docStoreCheckCode )
	
		mapServerLog:debug( "docStoreCheckCode resolved with %s, expected %s",
							docStoreCheckCode, expectedCheckCode )

		if expectedCheckCode == docStoreCheckCode then
		
			-- Nothing to do: Doc store and index store are already related.
		
			indexStoreSchema:resetTablesIfNecessary( db, false )
		
		elseif expectedCheckCodeIsNew and docStoreCheckCode == nil then

			-- Doc store and index store are not related, but it's OK because
			-- index store is new. Make them related.
			
			mapServerLog:debug( "write check code to doc store" )

			setDocStoreCrossCheckP( self, expectedCheckCode )

			indexStoreSchema:resetTablesIfNecessary( db, false )

		else

			-- Index store appears to describe some other doc store.
			-- Blow it away and start over!
		
			indexStoreSchema:resetTablesIfNecessary( db, true )

			-- Generate new check code and apply it to both doc store
			-- and index store.

			expectedCheckCode = KSIdGenerator.uuidInLowercaseHex()
		
			mapServerLog:warn(
				"index store will blow away everything :: out of sync!!!" )

			mapServerLog:debug( "docStoreCheckCode generated new check code %s",
								expectedCheckCode )

			db:execOnce(
				[[INSERT OR REPLACE INTO localState values( 'docStoreCrossCheck', ? )]],
				expectedCheckCode )
	
			populateNeedsUpdateFromBForTables( self )
	
			return setDocStoreCrossCheckP( self, expectedCheckCode )
				:mapDataP( function()
				
					return 'data', 'out of sync'
				
				end )

		end
		
		-- Signal that DB has desired schema and we can start updating
		-- the index tables.
		
		populateNeedsUpdateFromBForTables( self )

		return 'data', true
	
	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Resolve when the index store database is ready for use.
	
	This is primarily used for testing and analytics.
	
	\return <code>P< true | "out of sync" ></code> resolves when we have
		verified that index store and doc store are properly related; resolves
		to true in most normal circumstances or "out of sync" if index and
		doc store were unrelated to each other
--]]

function mapServerMethods:isReadyP()

	return self._isReadyP

end

--------------------------------------------------------------------------------

local function updateIndexForOneDoc_inTransaction( self, db, db_exec,
												   tableName, docValue, mapFn,
												   existingRowQuery,
												   deleteExistingRowsStmt,
												   insertRowStmt,
												   nColumns, niceToHaveIndices )

	local docId = docValue._id or docValue._docId

	-- Gather existing rows for this document from table.
	
	local existingRows = db:gatherArrayOfArray( existingRowQuery, docId )

	-- See what map functions says is the correct value for this doc now.

	local newRows = {}
	local nNewRows = 0
	
	local function emit( ... )
		nNewRows = nNewRows + 1
		local newRow = { _n = select( '#', ... ), ... }
		newRows[ nNewRows ] = newRow
	end
	
	mapFn( docValue, emit )

	local nExistingRows = existingRows and #existingRows or 0

	-- Short-circuit out if old and new rows are identical.
	
	if nExistingRows == 0 and nNewRows == 0 then
		return false
	end
	
	local didMatch = true
	
	if existingRows and #existingRows == nNewRows then
	
		-- Compare the two sets of rows.
	
		if nNewRows == 1 then
		
			local oldRow = existingRows[ 1 ]
			local newRow = newRows[ 1 ]
			for i = 1, nColumns do
				if oldRow[ i + 2 ] ~= newRow[ i ] then
					if niceToHaveIndices[ i ] then
						didMatch = 'niceToHave'
					else
						didMatch = false
						break
					end
				end
			end
			
			if didMatch == true then return false end
				-- Don't return early if 'niceToHave'; we must update the table
		
		else
		
			-- FIXME [scouten 2011-04-08]: If multiple rows for this doc,
			-- we should still be able to compare results. Punting on that
			-- for now.
		
		end
	
	end
	
	-- No match: Replace all old rows with new rows.
	
	if nExistingRows > 0 then
		db_exec( db, deleteExistingRowsStmt, docId )
	end
	
	if nNewRows == 0 then return true end
	
	for i = 1, nNewRows do
	
		local newRow = newRows[ i ]
		db_exec( db, insertRowStmt, docId, docValue._changeCounter,
				 unpack( newRow, 1, nColumns ) )
	
	end
	
	if didMatch == 'niceToHave' then
		return didMatch
	else
		return true
	end

end

--------------------------------------------------------------------------------

local function updateIndexTable_inTransaction( self, db, tableName, changedDocs )

	local hasNewData = false
	local hasNiceToHaveNewData = false
	local db_exec = self._db_exec

	-- Precompute some queries so we can assume they're present in the tight loop.
	
	local existingRowQuery = self._existingRowQueryForTable[ tableName ]
	if not existingRowQuery then
		existingRowQuery = string.format( [[SELECT * FROM %s WHERE docId == ?]], tableName )
		self._existingRowQueryForTable[ tableName ] = existingRowQuery
	end
	
	local deleteExistingRowsStmt = self._deleteExistingRowsStmtForTable[ tableName ]
	if not deleteExistingRowsStmt then
		deleteExistingRowsStmt = string.format( [[DELETE FROM %s WHERE docId == ?]],
												tableName )
		self._deleteExistingRowsStmtForTable[ tableName ] = deleteExistingRowsStmt
	end

	local nColumns = self._nColumnsPerTable[ tableName ]

	local insertRowStmt = self._insertRowStmtForTable[ tableName ]
	if not insertRowStmt then
		insertRowStmt = string.format( [[INSERT INTO %s VALUES( ?, ?%s )]],
									   tableName, string.rep( ", ?", nColumns ) )
		self._insertRowStmtForTable[ tableName ] = insertRowStmt
	end
	
	local mapFn = assert( self._mapFunctions[ tableName ] )
	local niceToHaveIndices = assert( self._niceToHaveIndices [ tableName ] )
	
	for i = 1, #changedDocs do

		local docValue = changedDocs[ i ]
	
		if docValue._slot ~= 'wfIndexStore' then
	
			local newData =
				updateIndexForOneDoc_inTransaction( self, db, db_exec,
												   tableName, docValue, mapFn,
												   existingRowQuery,
												   deleteExistingRowsStmt,
												   insertRowStmt,
												   nColumns, niceToHaveIndices )
			if newData == 'niceToHave' then
				hasNiceToHaveNewData = true
			elseif newData then
				hasNewData = true
			end
		
		end
	
	end
	
	db_exec( db, [[UPDATE indexStatus SET needsUpdateFrom = ? WHERE indexTableName = ?]],
			 changedDocs.continueFrom, tableName )

	return hasNewData, hasNiceToHaveNewData

end

--------------------------------------------------------------------------------

local function tableNeedsUpdate( self, tableName )

	local needsUpdateFromB = self._needsUpdateFromBForTable[ tableName ]
	
	if not needsUpdateFromB or needsUpdateFromB:isErrorNow() then
		return false
	end
	
	local needsUpdateFrom = needsUpdateFromB:valueNow()

	local needsUpdateThrough = self._tableNeedsUpdateThru[ tableName ]
	
	local needsUpdate =
				needsUpdateThrough
				and needsUpdateFrom
				and needsUpdateThrough > needsUpdateFrom
				
	if not needsUpdate then
		return false
	end
	
	return true, needsUpdateFrom, needsUpdateThrough

end

--------------------------------------------------------------------------------

local function issueNextChangedDocsRequest( self )

	-- The goal of this logic is to ensure that we issues as few changedDocsP
	-- requests as possible. We only allow one changedDocsP request to be
	-- issued at a time, so we examine the state of each table that needs
	-- updating and back up to its starting point for the next request.
	
	if self._pendingChangedDocsRequestP then return end

	local maxNeedsUpdateThrough, minNeedsUpdateFrom
	
	for tableName in pairs( self._tableNames ) do
	
		local needsUpdate, needsUpdateFrom, needsUpdateThrough
					= tableNeedsUpdate( self, tableName )
	
		if needsUpdate then

			mapServerLog:debug( "table %s needs update from %s through %s",
								tableName, needsUpdateFrom, needsUpdateThrough )
			
			if not minNeedsUpdateFrom or minNeedsUpdateFrom > needsUpdateFrom then
				minNeedsUpdateFrom = needsUpdateFrom
			end
			
			if not maxNeedsUpdateThrough or maxNeedsUpdateThrough < needsUpdateThrough then
				maxNeedsUpdateThrough = needsUpdateThrough
			end
			
		end
	
	end
	
	-- Nothing needs updating? Woot!
		
	if not minNeedsUpdateFrom or minNeedsUpdateFrom >= maxNeedsUpdateThrough then
		assert( not maxNeedsUpdateThrough ) -- We shouldn't set one and not the other.
		self._nDocumentsToIndexB:set( 0 )
		return
	end
	
	-- We're behind. Signal how far behind and then issue next request.
	
	self._nDocumentsToIndexB:set( maxNeedsUpdateThrough - minNeedsUpdateFrom )
	
	local changedDocsP = self._docStore:changedDocsP
		{ fromChangeCounter = minNeedsUpdateFrom }

	self._pendingChangedDocsRequestP = changedDocsP
	
	changedDocsP:whenResolvedDo( function( status, changedDocs )

		if WFCloseSupport.hasBeenOrWillBeClosed( self ) then return end

		self._pendingChangedDocsRequestP = false

		-- If changedDocs fails, we're basically hosed.

		if status ~= 'data' then
		
			mapServerLog:error( ":changedDocsP failed with message %s", changedDocs )

			for tableName in pairs( self._tableNames ) do
				self._needsUpdateFromBForTable[ tableName ]:setError( changedDocs )
			end

			return

		end
		
		mapServerLog:debug( "received a bunch of changed docs: now what? (n = %d)",
							#changedDocs )

		-- Apply all updates within one transaction.
		
		local db = self._db
		local hasNewDataByTable = {}
		local didReviewIndexForTable = {}

		local continueFrom = changedDocs.continueFrom

		KSLuaUtils.dpcall( 'processChangedDocs', function()

			db:withTransactionDo( function()
	
				-- Apply updates to all potentially relevant tables.
				
				for tableName in pairs( self._tableNames ) do
				
					-- Does table need updating? Skip if not.
					
					local needsUpdate, needsUpdateFrom =
							tableNeedsUpdate( self, tableName )
					
					if needsUpdate then
		
						-- Make sure the range of documents provided by this update is
						-- valid for this table's state.
						
						if minNeedsUpdateFrom <= needsUpdateFrom	-- changedDocs includes at least first doc of interest
							and continueFrom > needsUpdateFrom		-- changedDocs includes *any* docs of interest
						then
							
							-- There are relevant documents. Go index them.
							
							local startTime = KSDate.timeNowForAnimation()
					
							local success, hasNewDataOrMessage, hasNiceToHaveNewData = pcall( function()
								return updateIndexTable_inTransaction( self, db, tableName, changedDocs )
							end )
						
							-- If an error occurs in this process, set index status to error
							-- and walk away. This table will not update again during this
							-- app session.
							
							if success then
								hasNewDataByTable[ tableName ] = hasNewDataOrMessage
							else
								self._needsUpdateFromBForTable[ tableName ]:setError( hasNewDataOrMessage )
							end
							
							-- Update profiling stats for this table.

							local timeConsumed = KSDate.timeNowForAnimation() - startTime
							
							local stat = self._updateTimePerTable[ tableName ]
							
							if stat then
				
								stat.count = stat.count + 1
								stat.min = math.min( stat.min, timeConsumed )
								stat.max = math.max( stat.max, timeConsumed )
								stat.niceToHaveUpdatesIgnored = ( stat.niceToHaveUpdatesIgnored or 0 )
										+ ( ( hasNiceToHaveNewData and not hasNewDataOrMessage ) and 1 or 0 )
								stat.total = stat.total + timeConsumed
				
							else
				
								stat = {
									tableName = tableName,
									count = 1,
									min = timeConsumed,
									max = timeConsumed,
									total = timeConsumed,
								}
				
								self._updateTimePerTable[ tableName ] = stat
				
							end

							if success then
								didReviewIndexForTable[ tableName ] = true
							end
							
						end
			
					end
				
				end
			
			end )
		
		end )
		
		-- Update state cache.

		for tableName in pairs( didReviewIndexForTable ) do
--if tableName == 'missingOriginals' then
-- debug.print( "missingOriginals UPDATED to CC = ", continueFrom )
--end
			self._needsUpdateFromBForTable[ tableName ]:set( continueFrom )
		end
		
		for tableName, hasNewData in pairs( hasNewDataByTable ) do
			if hasNewData then
				self._tableUpdatedE:postToEvent( tableName, true )
			end
		end
	
		-- More updating needed?
		
		issueNextChangedDocsRequest( self )

	end ):runUntilDone()

end

--------------------------------------------------------------------------------
--[[!
	\brief Bring an index table up to date.
	
	\param tableName (string) name of index table to be brought up to date
	
	\return <code>P< true ></code> resolves when the table has been brought
		up to date with doc store as of the time this call is issued
--]]

function mapServerMethods:bringTableUpToDateP( --[[string]] tableName )

	check( 'WFIndexStoreMapServer', 'string' )
	
	if not self._tableNames[ tableName ] then
		error( "WFIndexStoreMapServer:bringTableUpToDateP: no such table " .. tableName, 2 )
	end

	local docChangeCounterP = self._docStore:docChangeCounterP()

	return docChangeCounterP
		:mapDataP( function( _, targetChangeCounter )

			if self._docStoreIsWf then
				targetChangeCounter = targetChangeCounter + 1
			end

			local tableCounterB = self._needsUpdateFromBForTable[ tableName ]
			
			-- Is table stuck?

			local isError, message = tableCounterB:isErrorNow()
			if isError then
				return 'error', message
			end

			-- Is index table already up to date? Yay!
			
			local tableCounterNow = tableCounterB:valueNow()
			if tableCounterNow and tableCounterNow >= targetChangeCounter then
				return 'data', true
			end
			
			-- If not, make it so.
			
			local existingUpdateRequest = self._tableNeedsUpdateThru[ tableName ]
			
			if not existingUpdateRequest
				or existingUpdateRequest < targetChangeCounter
			then
				self._tableNeedsUpdateThru[ tableName ] = targetChangeCounter
			end
			
			issueNextChangedDocsRequest( self )
			
			-- Resolve the promise when table is up to date with target
			-- change counter.
			
			-- FIXME_AUTO_CLOSE: Do I need the relay here?
			
			return KSReactivePromiseBridge.behaviorToPromise( tableCounterB:relayForAutoCloseB(),
				function( newChangeCounter )
					return newChangeCounter and newChangeCounter >= targetChangeCounter
				end )
	
		end )
		:mapDataP( function()
	
			-- Strip the change counter and just return true when we get
			-- to target level.
	
			return 'data', true

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Track when an index table has new data.
	
	This is not a guarantee that the table is fully up to date. It merely
	informs you when one or more document changes have been processed \b and
	have resulted in changes to the index table.
	
	Note that this method does \b not cause the table to be updated.
	Use <code>:bringTableUpToDateP()</code> to trigger an update.

	\param tableName (string) name of index table to be tracked

	\return <code>E< boolean ></code> fires whenever data is modified in the
		index table; true if table is thought to be up to date, false if not
--]]

function mapServerMethods:tableUpdatedE( --[[string]] tableName )

	check( 'WFIndexStoreMapServer', 'string' )

	if not self._tableNames[ tableName ] then
		error( "WFIndexStoreMapServer:tableUpdatedE: no such table " .. tableName, 2 )
	end

	return self._tableUpdatedE:eventE( tableName ):relayE( true )

end

--------------------------------------------------------------------------------
--[[!
	\brief Is a table up to date?
	
	Note that this method does \b not cause the table to be updated.
	Use <code>:bringTableUpToDateP()</code> to trigger an update.
	
	\param tableName (string) name of index table to be tracked

	\return <code>B< true ></code> true if the table is up to date as of
		the most recent doc change counter update from the doc store
--]]

function mapServerMethods:isTableUpToDateB( --[[string]] tableName )

	check( 'WFIndexStoreMapServer', 'string' )
	
	if not self._tableNames[ tableName ] then
		error( "WFIndexStoreMapServer:isTableUpToDateB: no such table " .. tableName, 2 )
	end

	return KSReactive.callB(
		function( tableChangeCounter, currentChangeCounter )
		
			if self._docStoreIsWf then
				currentChangeCounter = currentChangeCounter + 1
			end
		
			return type( tableChangeCounter ) == 'number'
				and type( currentChangeCounter ) == 'number'
				and tableChangeCounter >= currentChangeCounter
				or false

		end,
		self._needsUpdateFromBForTable[ tableName ],
		self._docChangeCounterB )

end

--------------------------------------------------------------------------------

--[[
	Helper function to update the TEMP_ids table (if the next query to be run
	needs it).
--]]

local BATCH_SIZE = 32

local kBatchSql = [[INSERT INTO TEMP_ids ( docId )
		SELECT ?
		]] .. string.rep( "UNION SELECT ? ", BATCH_SIZE - 1 )

local function updateTempIds( self, ids )
	local canSkip = false
	if self._lastTempIds and #self._lastTempIds == #ids then
		canSkip = true
		for i, id in ipairs( self._lastTempIds ) do
			if ids[ i ] ~= id then
				canSkip = false
				break
			end
		end
	end
	if not canSkip then
		local n = #ids
		local db = self._db
		local exec = db.exec
		self._lastTempIds = ids
		exec( db, [[DELETE FROM TEMP_ids]] )
		-- Do inserts in batches, which is more efficient.  This cuts off
		-- around 20-30% of the total time.
		for i = 1, BATCH_SIZE * math.floor( n / BATCH_SIZE ), BATCH_SIZE do
			exec( db, kBatchSql, unpack( ids, i, i + BATCH_SIZE - 1 ) )
		end
		-- Any remaining are inserted one at a time.
		for i = BATCH_SIZE * math.floor( n / BATCH_SIZE ) + 1, n do
			exec( db, [[INSERT INTO TEMP_ids VALUES ( ? )]], ids[ i ] )
		end
	end
end

--------------------------------------------------------------------------------
--[[!
	\brief Run a query against the index database.
	
	No guarantee is made by this API that the tables are up to date.
	Use <code>:bringTableUpToDateP()</code> to trigger an update.
	
	This query runs immediately in this thread.

	\param sql (string) SQL query string to run
	\param ... (optional) arguments to SQL query
	
	\return (table) array of arrays (outer arrays are result rows, inner arrays
		are columns)
--]]

function mapServerMethods:query( sql, ids, ... )

	check( 'WFIndexStoreMapServer', 'string' )
	
	if ids then
		updateTempIds( self, ids )
	end
	
	return self._db:gatherArrayOfArray( sql, ... )

end

--------------------------------------------------------------------------------
--[[!
	\brief Return the number of documents left to index.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return (<code>B< number ></code>) how many documents left to index?
--]]

function mapServerMethods:sizeOfIndexQueueB()

	return self._nDocumentsToIndexB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Return information about index update performance.
	
	Intended primarily for use in performance instrumentation (i.e. diagnostic log).
	
	\return (table) array of the following statistics, per index table that has been updated:
		- \c tableName: (string) table name as declared in SQL
		- \c max: (number) longest time it has taken to update this table
		- \c min: (number) shortest time it has taken to update this table
		- \c total: (number) total time spent for all updates to this table
		- \c count: (number) how many times has this table been updated?
	
	All times are in seconds.
--]]

function mapServerMethods:indexUpdatePerformanceStats()

	local sortedTableUpdateTimes = {}
	local nSortedTableEntries = 0
	
	for _, tableUpdatePerf in pairs( self._updateTimePerTable ) do
		nSortedTableEntries = nSortedTableEntries + 1
		sortedTableUpdateTimes[ nSortedTableEntries ] = {
			max = tableUpdatePerf.max,
			min = tableUpdatePerf.min,
			total = tableUpdatePerf.total,
			count = tableUpdatePerf.count,
			niceToHaveUpdatesIgnored = tableUpdatePerf.niceToHaveUpdatesIgnored,
			tableName = tableUpdatePerf.tableName,
			tableNeedsUpdateFrom = self._needsUpdateFromBForTable[ tableUpdatePerf.tableName ]:valueNow(),
		}
	end
	
	table.sort( sortedTableUpdateTimes, function( a, b )
		return a.total > b.total
	end )
	
	sortedTableUpdateTimes.docChangeCounterB = self._docChangeCounterB:valueNow()
	
	return sortedTableUpdateTimes
	
end

--------------------------------------------------------------------------------

return KSClass 'WFIndexStoreMapServer'( mapServerMethods )
