--[[----------------------------------------------------------------------------

WFIndexStoreMapServerMetricsTest.lua

@author jandrew 2012-10-23

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- wichitafoundation.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local WFDigest = require 'WFDigest'

	-- wichitafoundation.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'

local KSDate = require 'KSDate'
local KSPromises = require 'KSPromises'
local WFPlatformInfo = require 'WFPlatformInfo'
local WFTestMetrics = require 'WFTestMetrics'

--------------------------------------------------------------------------------

local md5 = assert( WFDigest.MD5, "Cannot find MD5 digest" )


--============================================================================--

-- @mark WFIndexStoreMapServerMetricsTest

local WFIndexStoreMapServerMetricsTest = {}

--------------------------------------------------------------------------------

local function mapForExampleSchema( docValue, emit )

	-- WARNING: This version is grossly simplified and omits error checking.
	-- See OzCatalogIndexDefinitions.lua for a better example.

	if docValue._docType == 'asset' then
		emit( docValue.width / docValue.height,
			  docValue.captureDate:sub( 1, 10 ) )
	end

end

local exampleFunction_md5 = md5.digest(
						string.dump( mapForExampleSchema ) .. '+2012-06-25' )

local sqlForExampleSchema = "CREATE TABLE coreInfo( aspectRatio, displayDate )"

local exampleSql_md5 = md5.digest
	[[CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )]]

--------------------------------------------------------------------------------

local function exampleSchemaFromApiDocs()

	-- Why not start by testing the exact schema declaration described in
	-- the API docs? :-)
	
	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {
		sql = sqlForExampleSchema,
		map = mapForExampleSchema,
	}
	
	indexStoreSchema:addIndex
		"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )"
	
	return indexStoreSchema

end

--------------------------------------------------------------------------------

local docStore, indexStoreDb, mapServer
local testMetrics

--------------------------------------------------------------------------------

local function createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {

		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",

		map = function( docValue, emit )
		
			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.
		
			if docValue._docType == 'asset' then
				emit( docValue.width / docValue.height,
					  docValue.captureDate:sub( 1, 10 ) )
			end
			
		end,

	}
	
	indexStoreSchema:addTable {

		sql = "CREATE TABLE keywords( keyword )",

		map = function( docValue, emit )
		
			if docValue.keywords then
			
				-- Intentionally unsafe access to docValue.keywords here.
				-- If keywords is non-nil, but not a table, we want the error.
				
				for k in pairs( docValue.keywords ) do
					emit( k )
				end
			
			end
		
		end,
	
	}
	
	-- Remember that indexStoreDb and mapServer are file-scoped locals.
	
	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb, indexStoreSchema )

end

--------------------------------------------------------------------------------

local startTimes = {}
local executionTimes = {}

local function startClock( componentName )
	startTimes[ componentName ] = KSDate.timeNowForAnimation()
end
	
--------------------------------------------------------------------------------
	
local function stopClock( componentName )
	if startTimes[ componentName ] then
		executionTimes[ componentName ] = ( KSDate.timeNowForAnimation() - startTimes[ componentName ] )
	end
end

--------------------------------------------------------------------------------

local function generateTestAssets( numberOfTimes )

	local docIds = {}
	
	for i = 1, numberOfTimes do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	table.sort( docIds ) -- so we have predictable results later
	
	local expectedCoreInfo = {}
	
	for i = 1, #docIds do
		expectedCoreInfo[ i ] = { docIds[ i ], i, 1, '2012-10-23' }
	end
	
	return docIds, expectedCoreInfo

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.setUp()

	docStore = OzDocStore ':memory:'
	
	indexStoreDb = nil

	mapServer = nil
	
	testMetrics = nil
	
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.tearDown()

	if docStore then
		docStore:close()
		docStore = nil
	end
	
	if indexStoreDb then
		indexStoreDb:close()
		indexStoreDb = nil
	end

	if mapServer then
		mapServer:close()
		mapServer = nil
	end
	
	testMetrics = nil

end

--------------------------------------------------------------------------------

local function bulkApplyOverlayP( docId )

	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = { docId },
		overlay = {
			width = 1,
			height = 1,
			captureDate = '2012-10-23T11:43:29',
		},
	}
	
end

--------------------------------------------------------------------------------

local function bulkApplyOverlayAndUpdateTableP( docId )

	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = { docId },
		overlay = {
			width = 1,
			height = 1,
			captureDate = '2012-10-23T11:43:29',
		},
	}
	:mapDataP( function ()
	
		return mapServer:bringTableUpToDateP( 'coreInfo' )
	
	end )
		
end

--------------------------------------------------------------------------------

local function mapServerTest_bulkApplyOverlay( test, numberOfTimes )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local TEST_WAIT_TIME = WFPlatformInfo.isAndroid and 0.03 or 0.01	-- PERFORMANCE_ANDROID
	local UPDATE_TIME = WFPlatformInfo.isAndroid and 0.003 or 0.001
	
	local waitTime = numberOfTimes * TEST_WAIT_TIME
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 3
		
			return KSPromises.allP( bulkApplyOverlayP, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
		
		-- Bring coreInfo table up to date.
		
		:waitFor( function() -- Step 4
		
			startClock( 'mapServerTest_bulkApplyOverlay: ' .. #docIds )
		
			return mapServer:bringTableUpToDateP( 'coreInfo' )
		
		end )
		:whenResolvedDo( function ()
		
			stopClock( 'mapServerTest_bulkApplyOverlay: ' .. #docIds )
		
		end )

		-- Verify data and performance is correct and acceptable
		
		:waitFor( function() -- Step 5

			local executionTime = executionTimes[ ( 'mapServerTest_bulkApplyOverlay: ' .. #docIds ) ]
			if testMetrics then
			
				testMetrics[ "wfp_execution_time" ] = executionTime
				WFTestMetrics.writeTestMetrics( testMetrics )
				
			end
			test:log( "Execution Time: " .. executionTime )
			
			assert( executionTime < #docIds * UPDATE_TIME )
		
			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]
			
		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_100_testPerformance( test )
	
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlay( test, 100 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_1000_testPerformance( test )
	
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlay( test, 1000 )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_5000_testPerformance( test )
	
	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlay( test, 5000 )
			
end

--------------------------------------------------------------------------------

local function mapServerTest_bulkApplyOverlayAndUpdateTable( test, numberOfTimes )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local TEST_WAIT_TIME = WFPlatformInfo.isAndroid and 0.05 or 0.02	-- PERFORMANCE_ANDROID
	local UPDATE_TIME = WFPlatformInfo.isAndroid and 0.05 or 0.01
	
	local waitTime = numberOfTimes * TEST_WAIT_TIME
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 3
		
			startClock( 'bulkApplyOverlayAndUpdateTableP: ' .. #docIds )
			
			return KSPromises.allP( bulkApplyOverlayAndUpdateTableP, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
		:whenResolvedDo( function ()
		
			stopClock( 'bulkApplyOverlayAndUpdateTableP: ' .. #docIds )
			
		end )
		
		-- Verify data and performance is correct and acceptable
		
		:waitFor( function() -- Step 4
			
			local executionTime = executionTimes[ ( 'bulkApplyOverlayAndUpdateTableP: ' .. #docIds ) ]
			if testMetrics then
			
				testMetrics[ "wfp_execution_time" ] = executionTime
				WFTestMetrics.writeTestMetrics( testMetrics )
				
			end
			test:log( "Execution Time: " .. executionTime )
			
			assert( executionTime < #docIds * UPDATE_TIME )
		
			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]
			
		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_100_testPerformance( test )
	
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 100 )
			
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_1000_testPerformance( test )
	
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 1000 )
			
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_5000_testPerformance( test )
	
	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 5000 )
	
end

--------------------------------------------------------------------------------

return WFIndexStoreMapServerMetricsTest
