--[[----------------------------------------------------------------------------

WFIndexStoreMapServerTest.lua

@author scouten 2012-08-14
@reviewed scouten 2012-08-14

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTestUtils = require 'KSTestUtils'
local WFDigest = require 'WFDigest'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'


--------------------------------------------------------------------------------

local md5 = assert( WFDigest.MD5, "Cannot find MD5 digest" )


--============================================================================--

-- @mark WFIndexStoreMapServerTest

local WFIndexStoreMapServerTest = {}


--------------------------------------------------------------------------------

local function mapForExampleSchema( docValue, emit )

	emit( docValue.width / docValue.height,
		  docValue.captureDate:sub( 1, 10 ) )

end

local exampleFunction_md5 = md5.digest(
						string.dump( mapForExampleSchema ) .. '+2012-06-25' )

local sqlForExampleSchema = "CREATE TABLE coreInfo( aspectRatio, displayDate )"

local exampleSql_md5 = md5.digest
	[[CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )]]

--------------------------------------------------------------------------------

local function exampleSchemaFromApiDocs()

	-- Why not start by testing the exact schema declaration described in
	-- the API docs? :-)

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {
		sql = sqlForExampleSchema,
		map = mapForExampleSchema,
	}

	indexStoreSchema:addIndex
		"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )"

	return indexStoreSchema

end

--------------------------------------------------------------------------------

local docStore, indexStoreDb, mapServer

local coreInfoTableUpdatedE, keywordTableUpdatedE
local coreInfoTableIsUpToDateB, keywordTableIsUpToDateB

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.setUp()

	docStore = WFDocStore ':memory:'

	indexStoreDb = nil

	mapServer = nil

	coreInfoTableUpdatedE = nil
	keywordTableUpdatedE = nil

	coreInfoTableIsUpToDateB = nil
	keywordTableIsUpToDateB = nil

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.tearDown()

	if docStore then
		docStore:close()
		docStore = nil
	end

	if indexStoreDb then
		indexStoreDb:close()
		indexStoreDb = nil
	end

	if mapServer then
		mapServer:close()
		mapServer = nil
	end

	if coreInfoTableUpdatedE then
		coreInfoTableUpdatedE:close()
		coreInfoTableUpdatedE = nil
	end

	if keywordTableUpdatedE then
		keywordTableUpdatedE:close()
		keywordTableUpdatedE = nil
	end

	if coreInfoTableIsUpToDateB then
		coreInfoTableIsUpToDateB:close()
		coreInfoTableIsUpToDateB = nil
	end

	if keywordTableIsUpToDateB then
		keywordTableIsUpToDateB:close()
		keywordTableIsUpToDateB = nil
	end

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testIsReadyP( test )

	local indexStoreSchema = exampleSchemaFromApiDocs()

	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb,
									   indexStoreSchema )

	WFAsyncTestCase( test )

		:waitFor( function()
			return mapServer:isReadyP()
		end )

		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testIsReadyP_existing( test )

	local indexStoreSchema = exampleSchemaFromApiDocs()

	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	local crossCheck = '5da52368defc467e82f3d03fcd1bff86'

	-- Simulate existing an existing doc + index store pair
	-- that are correctly matched.

	indexStoreDb:execOnce(
		[[INSERT OR REPLACE INTO localState values( 'docStoreCrossCheck', ? )]],
			crossCheck )

	indexStoreDb:execOnce(
		[[INSERT INTO indexStatus values( 'coreInfo', 2, ?, ? )]],
			exampleSql_md5, exampleFunction_md5 )

	-- Put some data into the index table. It should be left unchanged
	-- when we open the map server.

	indexStoreDb:execOnce
		[[INSERT INTO coreInfo values( "foo", 1, 1.5, '2010-08-17' )]]

	indexStoreDb:execOnce
		[[INSERT INTO coreInfo values( "bar", 2, 1.9, '2017-02-26' )]]

	WFAsyncTestCase( test )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'indexStoreCrossCheck',
					_slot = 'wfIndexStore',
					crossCheck = crossCheck,
				}

		end )

		:waitFor( function()

			mapServer = WFIndexStoreMapServer( docStore, indexStoreDb,
											   indexStoreSchema )

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Doc store should still have same cross-check code.

		:waitFor( function()

			return docStore:docValueP( 'indexStoreCrossCheck', 'wfIndexStore' )
				:mapDataP( function( _, docValue )
					return 'data', docValue.crossCheck
				end )

		end )
		:continueWhenValueMatches( 'data', crossCheck )

		-- Index store also should still have same cross-check code.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT value FROM localState WHERE name == 'docStoreCrossCheck']]

		end )
		:continueWhenValueMatches( 'data',
			{
				{ crossCheck }
			} )

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId DESC]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					"foo",
					1,
					1.5,
					'2010-08-17',
				},
				{
					"bar",
					2,
					1.9,
					'2017-02-26',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testIsReadyP_mismatch( test )

	local indexStoreSchema = exampleSchemaFromApiDocs()

	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	local crossCheck = '5da52368defc467e82f3d03fcd1bff86'

	-- Simulate existing an existing doc + index store pair
	-- that are correctly matched.

	indexStoreDb:execOnce(
		[[INSERT OR REPLACE INTO localState values( 'docStoreCrossCheck', ? )]],
			'bogus' )

	indexStoreDb:execOnce(
		[[INSERT INTO indexStatus values( 'coreInfo', 2, ?, ? )]],
			exampleSql_md5, exampleFunction_md5 )

	-- Put some data into the index table. It should be deleted
	-- when we open the map server.

	indexStoreDb:execOnce
		[[INSERT INTO coreInfo values( "foo", 1, 1.5, '2010-08-17' )]]

	indexStoreDb:execOnce
		[[INSERT INTO coreInfo values( "bar", 2, 1.9, '2017-02-26' )]]

	local newCrossCheck

	WFAsyncTestCase( test )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'indexStoreCrossCheck',
					_slot = 'wfIndexStore',
					crossCheck = crossCheck,
				}

		end )

		:waitFor( function()

			mapServer = WFIndexStoreMapServer( docStore, indexStoreDb,
											   indexStoreSchema )

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', "out of sync" )

		-- Find out what is the new cross-check value.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT value FROM localState WHERE name == 'docStoreCrossCheck']]

		end )
		:whenResolvedDo( function( status, newCrossCheckRows )

			assert( status == 'data' )

			newCrossCheck = newCrossCheckRows
				and newCrossCheckRows[ 1 ]
				and newCrossCheckRows[ 1 ][ 1 ]

			test:log( "Index store's cross check is now "
					  ..tostring( newCrossCheck ) )

			assert( type( newCrossCheck ) == 'string' )
			assert( newCrossCheck ~= crossCheck )

			newCrossCheck = newCrossCheck

		end )

		-- Doc store should now have the same (new) cross-check code.

		:waitFor( function()

			return docStore:docValueP( 'indexStoreCrossCheck', 'wfIndexStore' )
				:mapDataP( function( _, docValue )
					return 'data', docValue.crossCheck
				end )

		end )
		:whenResolvedDo( function( status, _newCrossCheck )

			test:log( "Doc store's cross check is " .. tostring( _newCrossCheck ) )

			assert( status == 'data' )
			assert( newCrossCheck == _newCrossCheck )

		end )

		-- Data in coreInfo table should have been deleted.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId DESC]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- indexStatus should be reset.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM indexStatus]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'coreInfo',
					1,
					exampleSql_md5,
					exampleFunction_md5
				},
			} )

end

--------------------------------------------------------------------------------

local coreInfoMapCalls = 0
local keywordMapCalls = 0

local coreInfoTableUpdateCount = 0
local keywordTableUpdateCount = 0

--------------------------------------------------------------------------------

local function createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	coreInfoMapCalls = 0
	keywordMapCalls = 0

	coreInfoTableUpdateCount = 0
	keywordTableUpdateCount = 0

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {

		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",

		map = function( docValue, emit )

			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.

			coreInfoMapCalls = coreInfoMapCalls + 1

			emit( docValue.width / docValue.height,
				  docValue.captureDate:sub( 1, 10 ) )

		end,

	}

	indexStoreSchema:addTable {

		sql = "CREATE TABLE keywords( keyword )",

		map = function( docValue, emit )

			keywordMapCalls = keywordMapCalls + 1

			if docValue.keywords then

				-- Intentionally unsafe access to docValue.keywords here.
				-- If keywords is non-nil, but not a table, we want the error.

				for k in pairs( docValue.keywords ) do
					emit( k )
				end

			end

		end,

	}

	-- Remember that indexStoreDb and mapServer are file-scoped locals.

	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb, indexStoreSchema )

	-- Track table updated events from both tables.

	coreInfoTableUpdatedE = mapServer:tableUpdatedE( 'coreInfo' )
	coreInfoTableUpdatedE:onDataE( function()
		coreInfoTableUpdateCount = coreInfoTableUpdateCount + 1
	end )

	keywordTableUpdatedE = mapServer:tableUpdatedE( 'keywords' )
	keywordTableUpdatedE:onDataE( function()
		keywordTableUpdateCount = keywordTableUpdateCount + 1
	end )

	coreInfoTableIsUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	keywordTableIsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_empty( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Bring table up to date.

		:waitFor( function()

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that coreInfo table still has no data.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors still have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_twoRows( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table now has data.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_noopChange( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table now has data.

		:waitFor( function()

			assert( coreInfoMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Make a change to doc store that shouldn't result in
		-- changes to emitted data.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 60,
				height = 40,
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				somethingUnrelated = 'blah',
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date again.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that coreInfo table has same data,
		-- but coreInfo emit function was called again.

		:waitFor( function()

			assert( coreInfoMapCalls == 4 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )


end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_singleRowChange( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table now has data.

		:waitFor( function()

			assert( coreInfoMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Make a single change to doc store that should result in
		-- changes to emitted data.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 50,
				height = 40,
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date again.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table has new data,
		-- but coreInfo emit function was called again.

		:waitFor( function()

			assert( coreInfoMapCalls == 3 )

			assert( coreInfoTableUpdateCount == 2 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					4,
					1.25,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_noRowsChanged( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table now has data.

		:waitFor( function()

			assert( coreInfoMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Bring table up to date again. (Should be a no-op.)

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that coreInfo table still has same data and no new
		-- map calls were placed.

		:waitFor( function()

			assert( coreInfoMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_empty( test )

	-- Make sure we exit out gracefully if no existing rows and no new rows.

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that keywords table is empty.

		:waitFor( function()

			assert( keywordMapCalls == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'keywords' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Verify that keywords table is now empty, but that
		-- the keywords emit function did get called twice.

		:waitFor( function()

			assert( keywordMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_empty( test )

	-- Make sure we exit out gracefully if no existing rows and no new rows.

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that keywords table is empty.

		:waitFor( function()

			assert( keywordMapCalls == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'keywords' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Verify that keywords table is now empty, but that
		-- the keywords emit function did get called twice.

		:waitFor( function()
			return tostring( keywordMapCalls )
		end )
		:continueWhenValueMatches( 'data', '2' )

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_multipleRowsChange( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				keywords = {
					bar = true,
					bop = true,
				},
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				keywords = {
					bar = true,
					foo = true,
				},
			}

		end )

		-- Verify that keywords table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'keywords' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Verify that keywords table now has data.

		:waitFor( function()

			assert( keywordMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 1 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords ORDER BY docId, keyword]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bar',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bop',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'bar',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'foo',
				},
			} )

		-- Now change both docs.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				keywords = {
					bop = { _delete = true },
					yellow = true,
				},
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				keywords = {
					bar = { _delete = true },
					blue = true,
				},
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring table up to date again.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 1 )

			return mapServer:bringTableUpToDateP( 'keywords' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Verify that keyword table has new data.

		:waitFor( function()

			assert( keywordMapCalls == 4 )

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 2 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords ORDER BY docId, keyword]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					4,
					'bar',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					4,
					'yellow',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					'blue',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					'foo',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_doesntAffectOtherTable( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test, 222 )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				keywords = {
					bar = true,
					bop = true,
				},
				captureDate = '2012-08-17T11:10:05',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				keywords = {
					bar = true,
					foo = true,
				},
				captureDate = '2012-08-15',
			}

		end )

		-- Verify that both tables are empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring keywords table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'keywords' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Verify that coreInfo table was left alone.

		:waitFor( function()

			assert( coreInfoMapCalls == 0 )

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 1 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that keywords table now has data.

		:waitFor( function()

			assert( keywordMapCalls == 2 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords ORDER BY docId, keyword]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bar',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bop',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'bar',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'foo',
				},
			} )

		-- Now change both docs.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				keywords = {
					bop = { _delete = true },
					yellow = true,
				},
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				keywords = {
					bar = { _delete = true },
					blue = true,
				},
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring coreInfo table up to date.
		-- Leave keywords table alone this time.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 1 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that keywords table was left alone, even
		-- though there is new data in doc store. (We didn't
		-- ask for *that* update!)

		:waitFor( function()

			assert( keywordMapCalls == 2 )

			assert( coreInfoTableUpdateCount == 1 )
			assert( keywordTableUpdateCount == 1 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM keywords ORDER BY docId, keyword]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bar',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					'bop',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'bar',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					'foo',
				},
			} )

		-- Verify that coreInfo table is up to date.

		:waitFor( function()

			assert( coreInfoMapCalls == 2 )

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					4,
					1.5,
					'2012-08-17',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					2/3,
					'2012-08-15',
				},
			} )


end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_validatesTableName( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return KSTestUtils.testFails( function()
				return mapServer:bringTableUpToDateP( 'bogusTable' )
			end )

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_errorInMapFunction( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				keywords = {
					bar = true,
					bop = true,
				},
				captureDate = '2012-08-17T11:10:05',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				keywords = {
					bar = true,
					foo = true,
				},
			}

		end )

		-- Bring coreInfo table up to date.

		:waitFor( function()

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Add more data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'e7a002fbcf974ef79c5b467b53c35b6b',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-17T11:10:05',
			}

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring coreInfo table up to date. Should fail again.

		:waitFor( function()

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testBringTableUpToDateP_bulk( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	-- Generate a bunch of UUIDs for doc IDs. The exact number doesn't matter
	-- but we have to have >50 so we trigger the bulk case of changedDocs.

	local docIds = {}
	for i = 1, 71 do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end

	table.sort( docIds ) -- so we have predictable results later

	local expectedCoreInfo = {}
	for i = 1, #docIds do
		expectedCoreInfo[ i ] = { docIds[ i ], i + 1, 1.5, '2012-08-17' }
	end

	WFAsyncTestCase( test, 2222 )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( true )

		-- Start adding data to doc store.

		:waitFor( function()

			return KSPromises.allP(
				function( docId )

					return docStore:applyOverlayP {
						_id = docId,
						_slot = 'main',
						width = 30,
						height = 20,
						captureDate = '2012-08-17T11:43:29',
					}

				end, docIds )

		end )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

		-- Bring coreInfo table up to date.

		:waitFor( function()

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return tostring( coreInfoMapCalls )

		end )
		:continueWhenValueMatches( 'data', tostring( #docIds ) )

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )

		-- Assert that the "up to date" behaviors have correct values.

		:waitFor( coreInfoTableIsUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordTableIsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testTableUpdatedE_validatesTableName( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return KSTestUtils.testFails( function()
				return mapServer:tableUpdatedE( 'bogusTable' )
			end )

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testIsTableUpToDateB_validatesTableName( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return KSTestUtils.testFails( function()
				return mapServer:isTableUpToDateB( 'bogusTable' )
			end )

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testQuery( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query and verify results.

		:waitFor( function()

			return mapServer:query
				[[SELECT * FROM coreInfo ORDER BY docId]]

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Make sure query args get passed through.

		:waitFor( function()

			return mapServer:query(
				[[SELECT * FROM coreInfo WHERE aspectRatio > ? ORDER BY docId]], nil, 1 )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
			} )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerTest.testIndexUpdatePerformanceStats( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local stats

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:setDocValueP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:setDocValueP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Bring table up to date.

		:waitFor( function()

			assert( coreInfoTableUpdateCount == 0 )
			assert( keywordTableUpdateCount == 0 )

			return mapServer:bringTableUpToDateP( 'coreInfo' )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Review index update stats.

		:waitFor( function()

			stats = mapServer:indexUpdatePerformanceStats()

			assert( type( stats ) == 'table' )
			assert( #stats == 1 )

			local statsRow = stats[ 1 ]

			assert( statsRow.max >= 0 )
			assert( statsRow.min >= 0 )
			assert( statsRow.max >= statsRow.min )
			assert( statsRow.total >= 0 )
			assert( statsRow.count > 0 )
			assert( statsRow.tableName == 'coreInfo' )

			return stats

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

end

--------------------------------------------------------------------------------

return WFIndexStoreMapServerTest
