--[[----------------------------------------------------------------------------

WFIndexStoreQueryServer.lua

@author scouten 2012-08-17
@reviewed scouten 2012-08-17

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.string
local KSStringUtils = require 'KSStringUtils'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSDate = require 'KSDate'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFTimer = require 'WFTimer'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )

--------------------------------------------------------------------------------

local function throttledMapE( event, timeRatio, mapFn )
	local pingB = KSReactive.variableB()
	local noSoonerB = KSReactive.variableB()
	noSoonerB:set( 0 )
	local getTimeNow = KSDate.timeNowForAnimation
	
	local count = 1
	local lastProcessed = 1
	local updateCountB = event:mapE( function( )
		count = count + 1
		return count
	end ):holdB( 'ignore' )
	local rateLimitedB = KSReactive.dataCallB( function( now, noSooner, count )
		
		if count == 'ignore' then
			return false
		end
		
		if count == lastProcessed then
			return false
		end
		
		local timeNow = getTimeNow()
		if timeNow < noSooner then
			WFTimer.afterDelayDo( noSooner - timeNow, function()
				pingB:set( getTimeNow() )
			end )
			return false
		end
		return true
	end, pingB, noSoonerB, updateCountB )
	
	local rateLimitedE = rateLimitedB:filterE( function( tooSoon )
		return tooSoon
	end )
	
	return rateLimitedE:mapE( function(...)
		local start = getTimeNow()
		local results = { mapFn(...) }
		lastProcessed = count
		local noSooner = getTimeNow()
		noSooner = noSooner + ( noSooner - start ) * timeRatio
		noSoonerB:set( noSooner )
		return unpack( results )
	end )

end

--============================================================================--
--[[!
	\brief A \c WFIndexStoreQueryServer is used by the index store implementation
		to keep the index store tables up to date.
	
	It should only be accessed by \c WFIndexStoreImp.
--]]

local queryServerMethods = {

	_mapServer = false,
		-- (WFIndexStoreMapServer) map server for this index store

	_tableNames = false,
		-- (table) names of the index tables declared in this schema
	
	_queryBRequests = false,
		-- (table) current query behavior requests
			-- key: (B< table >) event stream returned to caller
			-- value: (table) arguments to :queryB, plus
				-- relevantTables: (table) set of table names to update
				-- nUpdatesSent: (number) how many times have we broadcasted results?

	_nQueryBehaviorsCreated = 0,
		-- (number) how many query requests have ever been created?
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Create a query server to support an index store implementation.
	
	\param mapServer (\c WFIndexStoreMapServer) map server providing the index tables
	\param tableNames (table) set of table names declared in this schema
--]]

function queryServerMethods:__init( --[[WFIndexStoreMapServer]] mapServer,
									--[[table]] tableNames )

	check( 'WFIndexStoreQueryServer', 'WFIndexStoreMapServer', 'table' )

	self._mapServer = mapServer
	self._mapServer:addCloseTarget( self )
	
	self._tableNames = tableNames
	
	self._queryBRequests = setmetatable( {}, { __mode = 'k' } )

end

--------------------------------------------------------------------------------

local function identifyRelevantTables( tableNames, sql )

	-- Identify all of our index tables that might be relevant for this query.
	-- Naive implementation: Simply search for each table name in query string.

	local result = {}
	
	for tableName in pairs( tableNames ) do
	
		-- Naive search, too: Since SQL table names can't contain Lua pattern
		-- special characters, just use the table name as search pattern.
	
		if sql:find( tableName ) then
			result[ tableName ] = true
		end
	
	end
	
	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Create a behavior that tracks the results of a SQL query on the index store.
	
	\param args (table) arguments for query
		- \c sql: (string) SQL query to be performed
		- \c threadName: (optional, string) name of thread requesting the query
		- \c traceback: (optional, string) traceback of call site in requesting thread
		- \c calm: (optional, number) if present, calm query result updates to no more
			than one every _n_ seconds
		- \c outOfDateNotice: (optional, boolean) if true, query results will contain
			a member "outOfDate" (boolean) which will be true if query results are
			known to be lagging behind the doc store
		- \c nArgs: (optional, number) number of query arguments (helps with \c nil
			values in the middle of the list)
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	
	\return B< table > query results (array of arrays syntax)
	
	\b WARNING: Use this method sparingly. Creating many query behaviors will adversely
		affect the overall performance of the application.
	
	\b WARNING: This behavior may, at times, return information that is based on
		incomplete index table results. It will eventually catch up and update
		results if needed. (Out of date information is especially likely for the
		first value returned.)
--]]

function queryServerMethods:queryB( --[[table]] args )

	-- Validate args.

	check( 'WFIndexStoreQueryServer', 'table' )
	
	local sql = args.sql
	if type( sql ) ~= 'string' then
		error( "WFQueryIndexStoreQueryServer:queryB: sql arg must be a string", 2 )
	end
	
	if args.threadName ~= nil and type( args.threadName ) ~= 'string' then
		error( "WFQueryIndexStoreQueryServer:queryB: threadName arg must be a string or nil", 2 )
	end

	if args.traceback ~= nil and type( args.traceback ) ~= 'string' then
		error( "WFQueryIndexStoreQueryServer:queryB: traceback arg must be a string or nil", 2 )
	end

	if args.calm ~= nil and ( type( args.calm ) ~= 'number' or args.calm <= 0 ) then
		error( "WFQueryIndexStoreQueryServer:queryB: traceback arg must be a positive number or nil", 2 )
	end
	
	-- Identify relevant tables.
	
	local relevantTablesSet = identifyRelevantTables( self._tableNames, sql )
	
	local nTables = 0
	local relevantTables = {}
	
	for tableName in pairs( relevantTablesSet ) do
		nTables = nTables + 1
		relevantTables[ nTables ] = tableName
	end
	
	if nTables <= 0 then
		error( "WFIndexStoreQueryServer:queryB: did not find any relevant tables in query " .. sql, 2 )
	end
	
	-- Housekeeping ...
	
	local mapServer = self._mapServer
	self._nQueryBehaviorsCreated = self._nQueryBehaviorsCreated + 1
	
	args.nUpdatesSent = 0
	args.relevantTables = relevantTables

	-- Do we have new data from any of the tables?
	
	local tableUpdatedEs = {}
	
	for i = 1, nTables do
		tableUpdatedEs[ i ] = mapServer:tableUpdatedE( relevantTables[ i ] )
	end

	-- We use the laterE event to start the clock on calming even though
	-- we don't use this first event to trigger a re-query. This prevents us
	-- from re-running the query (and publishing results) any sooner than
	-- (interval) seconds after the query behavior is created.
	-- If you disable this logic, expect the test named testQueryB_outOfDate_join
	-- to fail with a "too soon" error.

	local freneticAnyTableUpdatedE = KSReactive.anyE( unpack( tableUpdatedEs, 1, nTables ) )
	
	-- But we don't want to update too often. Note that we start with the timer running so that
	-- we are calm from the get go.
	
	local anyTableUpdatedE = freneticAnyTableUpdatedE:calmE( args.calm or .125, true )
		:addCloseTarget( freneticAnyTableUpdatedE )

	-- Fetch query results.
	
	local nQueryArgs = args.nArgs or #args
	local queryResultsNow = function()
		args.nUpdatesSent = args.nUpdatesSent + 1
		return mapServer:query( sql, args.tempIds, unpack( args, 1, nQueryArgs ) )
	end
	
	-- Squelch events if the actual query results are unchanged.

	local previousResult = {}
	
	local queryResultsNowSetFirst = function()
		previousResult = queryResultsNow()
		return previousResult
	end
	
	-- Raw query event stream. Fires whenever new data is available in any of the
	-- index tables, modulo the calming factor.
	
	-- The calming factor now scales with the performance of the query;
	-- as results get big, we calm more and more.
	
	local rawQueryE = throttledMapE( anyTableUpdatedE, 4, queryResultsNow )

	local filteredQueryE = rawQueryE:filterE( function( newValue )

		if not KSTableUtils.deepEquals( previousResult, newValue ) then
			previousResult = newValue
			return true
		else
			return false
		end
	
	end )
	
	filteredQueryE:addCloseTarget( rawQueryE ) -- Mark, is this necessary?
	
	-- Convert to a behavior. Populate with whatever results we have now, even
	-- if they may be out of date.
	
	local rawQueryB = filteredQueryE:holdB( queryResultsNowSetFirst() )

	-- Now, make sure all relevant tables stay up to date.
	
	local tablesUpToDateBs = {}
	
	for i = 1, nTables do

		local tableName = relevantTables[ i ]
		
		local thisTableIsUpToDateB = mapServer:isTableUpToDateB( tableName )
		rawQueryB:addCloseTarget( thisTableIsUpToDateB )
		
		local function maybeBringTableUpToDate( upToDate )

			if not upToDate
				and not WFCloseSupport.hasBeenOrWillBeClosed( thisTableIsUpToDateB )
			then

				mapServer:bringTableUpToDateP( tableName )
					:whenResolvedDo( function( status )
					
						-- When this bring up-to-date operation completes, we check to see
						-- if we remain behind (i.e. new incoming changes). If so, trigger
						-- yet another bring up-to-date. [scouten 2012-08-17, PR# 3331492]
					
						if status == 'data' then
							maybeBringTableUpToDate( thisTableIsUpToDateB:valueNow() )
						end

					end )
					:runUntilDone()

			end
		
		end

		thisTableIsUpToDateB:onDataB( maybeBringTableUpToDate )

		tablesUpToDateBs[ i ] = thisTableIsUpToDateB

	end

	-- Make sure these actions don't get GC'd as long as result behavior survives.

	rawset( rawQueryB, '_tablesUpToDateBs', tablesUpToDateBs )
	
	-- If so requested, provide out-of-date notice in results table.
	
	local resultB
	
	if args.outOfDateNotice then
	
		local allTablesUpToDateB
		
		if nTables == 1 then
	
			allTablesUpToDateB = tablesUpToDateBs[ 1 ]
	
		else
	
			allTablesUpToDateB = KSReactive.callB( function( ... )
			
				for _, upToDate in pairs { ... } do
					if not upToDate then
						return false
					end
				end
				
				return true
			
			end, unpack( tablesUpToDateBs, 1, nTables ) )
	
		end
		
		-- Merge up to date notice with raw query behavior results.
		
		resultB = KSReactive.callB( function( queryResults, isUpToDate )

			if queryResults then
			
				local resultsClone = {}
				for k, v in pairs( queryResults ) do
					resultsClone[ k ] = v
				end
				
				resultsClone.outOfDate = not isUpToDate

				return resultsClone
			
			end
		
		end, rawQueryB, allTablesUpToDateB )
	
	else
	
		resultB = rawQueryB

	end
	
	-- Make sure we close down properly if client loses interest,
	-- or this query server gets terminated.
	
	resultB:addCloseTarget( anyTableUpdatedE )
	self:addCloseTarget( resultB )
	
	-- Remember the query.
	
	self._queryBRequests[ resultB ] = args
	
	return resultB

end 

--------------------------------------------------------------------------------
--[[!
	\brief Create a promise that performs a single SQL query on the index store.

	This promise will first ensure that the relevant index tables are brought up
	to date as of the time this method is called before performing the query.
		
	However, if data continues to be added to the doc store while this request
	is being processed, no guarantee is made as to whether the newer data will
	or will not appear in the query results.
		
	\param sql (string) SQL query to be performed
	\param ... (optional) query arguments to be substituted for ? placeholders in query
	
	\return P< table > query results (array of arrays syntax)
--]]

function queryServerMethods:queryP( params )

	check( 'WFIndexStoreQueryServer', 'table' )
	
	local sql = assert( params.sql )
	
	-- Make sure all relevant tables are up to date before doing query.

	local relevantTablesSet = identifyRelevantTables( self._tableNames, sql )
	
	local nTables = 0
	local relevantTables = {}
	
	for tableName in pairs( relevantTablesSet ) do
		nTables = nTables + 1
		relevantTables[ nTables ] = tableName
	end
	
	if nTables <= 0 then
		error( "WFIndexStoreQueryServer:queryP: did not find any relevant tables in query " .. sql, 2 )
	end
	
	local upToDateP
	local mapServer = self._mapServer
	
	if nTables == 1 then

		-- Simplify for the common case of using only one table.
		
		upToDateP = mapServer:bringTableUpToDateP( relevantTables[ 1 ] )
	
	else

		upToDateP = KSPromises.allP( function( tableName )
			return mapServer:bringTableUpToDateP( tableName )
		end, relevantTables )
	
	end
	
	-- When all tables are up to date, then run the query and return results.
	
	local nQueryArgs = params.nArgs or #params
	return upToDateP:mapDataP( function()

		return 'data', mapServer:query( params.sql, params.tempIds, unpack( params, 1, nQueryArgs ) )
	
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Return summary statistics about query behaviors that are currently
		running.
	
	\return (table) array of outstanding query behaviors. Each entry contains:
		- \c sql: (string) SQL query
		- \c args: (table, array of strings) query arguments, if any
		- \c relevantTables: (table, set of strings) which tables are used for this table?
		- \c nUpdatesSent: (number) how many times has this query sent updates?
		- \c calm: (optional, number) calming interval, if applied
		- \c threadName: (optional, string) which thread requested this behavior? (if provided)
		- \c traceback: (optional, string) traceback in requesting thread (if provided)
--]]			

function queryServerMethods:queryBStats()

	-- Send information about query requests that are outstanding.

	local filteredQueryBRequests = {}
	local nQueries = 0
	
	for resultE, queryInfo in pairs( self._queryBRequests ) do
	
		if not WFCloseSupport.hasBeenOrWillBeClosed( resultE ) then

			nQueries = nQueries + 1
			
			local traceback = queryInfo.traceback
			
			if traceback then
				traceback = KSStringUtils.split( queryInfo.traceback, '\n' )
				table.remove( traceback, 1 )
			end
	
			filteredQueryBRequests[ nQueries ] = {
				sql = queryInfo.sql,
				relevantTables = queryInfo.relevantTables,
				nUpdatesSent = queryInfo.nUpdatesSent,
				calm = queryInfo.calm,
				threadName = queryInfo.threadName,
				traceback = traceback,
			}
			
			if queryInfo[ 1 ] then

				local args = {}
				for i = 1, #queryInfo do
					args[ i ] = queryInfo[ i ]
				end
				
				filteredQueryBRequests[ nQueries ].args = args
				
			end
		
		end
		
	end
	
	table.sort( filteredQueryBRequests, function( a, b )
		return a.sql < b.sql
	end )
	
	filteredQueryBRequests.nQueryBehaviorsCreated = self._nQueryBehaviorsCreated
	
	return filteredQueryBRequests
	
end

--------------------------------------------------------------------------------

return KSClass 'WFIndexStoreQueryServer'( queryServerMethods )
