--[[----------------------------------------------------------------------------

WFIndexStoreQueryServerMetricsTest.lua

@author jandrew 2013-02-21

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'

	-- wichitafoundation.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'
local WFTestMetrics = require 'WFTestMetrics'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- wichitafoundation.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'
local WFIndexStoreQueryServer = require 'WFIndexStoreQueryServer'

local WFPlatformInfo = require 'WFPlatformInfo'

local WAIT_TIME_PER_ITEM = 0.05
local EXEC_TIME_PER_ITEM_SHORT = 0.05
local EXEC_TIME_PER_ITEM_LONG = 0.1

--============================================================================--

-- @mark WFIndexStoreQueryServerMetricsTest

local WFIndexStoreQueryServerMetricsTest = {}

--------------------------------------------------------------------------------

local docStore, indexStoreDb, mapServer, queryServer, queryB, updatedQueryCountB, testMetrics

--------------------------------------------------------------------------------

local function createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {

		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",

		map = function( docValue, emit )
		
			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.
		
			if docValue._docType == 'asset' then
				emit( docValue.width / docValue.height,
					  docValue.captureDate:sub( 1, 10 ) )
			end
			
		end,

	}
	
	indexStoreSchema:addTable {

		sql = "CREATE TABLE keywords( keyword )",

		map = function( docValue, emit )
		
			if docValue.keywords then
			
				-- Intentionally unsafe access to docValue.keywords here.
				-- If keywords is non-nil, but not a table, we want the error.
				
				for k in pairs( docValue.keywords ) do
					emit( k )
				end
			
			end
		
		end,
	
	}
	
	-- Remember that indexStoreDb and mapServer are file-scoped locals.
	
	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb, indexStoreSchema )
	
	queryServer = WFIndexStoreQueryServer( mapServer, indexStoreSchema:tableNames() )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.setUp()

	docStore = OzDocStore ':memory:'
	
	indexStoreDb = nil

	mapServer = nil
	
	queryB = nil
	
	testMetrics = nil
	
	local KSReactive_variableB = KSReactive.variableB
	
	updatedQueryCountB = KSReactive_variableB( 0 )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.tearDown()

	if docStore then
		docStore:close()
		docStore = nil
	end
	
	if indexStoreDb then
		indexStoreDb:close()
		indexStoreDb = nil
	end

	if mapServer then
		mapServer:close()
		mapServer = nil
	end
	
	if queryB then
		queryB:close()
		queryB = nil
	end
	
	if updatedQueryCountB then
		updatedQueryCountB:close()
		updatedQueryCountB = nil
	end
	
	testMetrics = nil
	
end

--------------------------------------------------------------------------------

local startTimes = {}
local executionTimes = {}

--------------------------------------------------------------------------------

local function startClock( componentName )
	startTimes[ componentName ] = KSDate.timeNowForAnimation()
end
	
--------------------------------------------------------------------------------
	
local function stopClock( componentName )
	if startTimes[ componentName ] then
		executionTimes[ componentName ] = ( KSDate.timeNowForAnimation() - startTimes[ componentName ] )
	end
end

--------------------------------------------------------------------------------

local function generateTestAssets( numberOfTimes )

	local docIds = {}
	
	for i = 1, numberOfTimes do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	table.sort( docIds ) -- so we have predictable results later
	
	local expectedCoreInfo = {}
	
	for i = 1, #docIds do
		expectedCoreInfo[ i ] = { docIds[ i ], i, 1, '2013-02-21' }
	end
	
	return docIds, expectedCoreInfo

end

--------------------------------------------------------------------------------

local function generateBulkQueryBs( numberOfTimes, expectedCoreInfo )

	local queryBs = {}
	
	for i = 1, numberOfTimes do
		queryBs[ i ] = queryServer:queryB { sql = [[SELECT * FROM coreInfo ORDER BY docId]] }
		
		queryBs[ i ]:onDataB( function( data )
				if( KSTableUtils.deepEquals( expectedCoreInfo, data, 20, false ) ) then
					updatedQueryCountB:set( updatedQueryCountB:valueNow() + 1 )
				end
				
		end )
		
	end
	
	return queryBs

end

--------------------------------------------------------------------------------

local function bulkUpdate( docId )

	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = { docId },
		overlay = {
			width = 1,
			height = 1,
			captureDate = '2013-02-21',
		},
	}
		
end


--------------------------------------------------------------------------------

-- @mark queryP Bulk Updates

--------------------------------------------------------------------------------

local function queryServerTest_queryPBulkUpdates( test, numberOfTimes )
	
	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local waitTime = numberOfTimes * WAIT_TIME_PER_ITEM
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 2
		
			startClock( 'queryPBulkUpdates: ' .. #docIds )
			
			return KSPromises.allP( bulkUpdate, docIds, #docIds )
			
		end )
		
		-- Verify that coreInfo table is empty.
		
		:waitFor( function() -- Step 3
		
			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]
			
		end )
		:continueWhenValueMatches( 'data', {} )
		
		-- Verify that map server thinks both tables are
		-- not up to date.
		
		:waitFor( coreInfoUpToDateB ) -- Step 4
		:continueWhenValueMatches( false )
		
		:waitFor( keywordsUpToDateB ) -- Step 5
		:continueWhenValueMatches( false )
		
		 -- Run a the query
		
		:waitFor( function() -- Step 6
			
			return queryServer:queryP{ sql =
				[[SELECT * FROM coreInfo ORDER BY docId]]
			}
			
		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )
		
		-- Verify performance is acceptable
		
		:waitFor( function()
			
			stopClock( 'queryPBulkUpdates: ' .. #docIds )
			
			local executionTime = executionTimes[ ( 'queryPBulkUpdates: ' .. #docIds ) ]
			if testMetrics then
			
				testMetrics[ "wfp_execution_time" ] = executionTime
				WFTestMetrics.writeTestMetrics( testMetrics )
				
			end
			test:log( "Execution Time: " .. executionTime )
			
			assert( executionTime < #docIds * EXEC_TIME_PER_ITEM_SHORT )
			
			return 0
			
		end )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryPBulkUpdates_100_testPerformance( test )

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryPBulkUpdates( test, 100 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryPBulkUpdates_1000_testPerformance( test )
	
	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryPBulkUpdates( test, 1000 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryPBulkUpdates_5000_testPerformance( test )

	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryPBulkUpdates( test, 5000 )
		
end


--------------------------------------------------------------------------------

-- @mark queryB Bulk Updates

--------------------------------------------------------------------------------

local function queryServerTest_queryBBulkUpdates( test, numberOfTimes )
	
	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local waitTime = numberOfTimes * WAIT_TIME_PER_ITEM
	
	queryB = queryServer:queryB { sql = [[SELECT * FROM coreInfo ORDER BY docId]] }
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that query result is empty.
		
		:waitFor( queryB )
		:continueWhenValueMatches( {} )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 3
		
			startClock( 'queryBBulkUpdates: ' .. #docIds )
			
			return KSPromises.allP( bulkUpdate, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
		
		 -- Wait for query behavior to update
		
		:waitFor( queryB )
		:continueWhenValueMatches( expectedCoreInfo )
		
		-- Verify performance is acceptable
		
		:waitFor( function() -- Step 4
			
			stopClock( 'queryBBulkUpdates: ' .. #docIds )
			
			local executionTime = executionTimes[ ( 'queryBBulkUpdates: ' .. #docIds ) ]
			if testMetrics then
			
				testMetrics[ "wfp_execution_time" ] = executionTime
				WFTestMetrics.writeTestMetrics( testMetrics )
				
			end
			test:log( "Execution Time: " .. executionTime )
			
			assert( executionTime < #docIds * EXEC_TIME_PER_ITEM_SHORT )
			
			return 0
			
		end )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBBulkUpdates_100_testPerformance( test )

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBBulkUpdates( test, 100 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBBulkUpdates_1000_testPerformance( test )

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBBulkUpdates( test, 1000 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBBulkUpdates_5000_testPerformance( test )

	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBBulkUpdates( test, 5000 )
		
end


--------------------------------------------------------------------------------

-- @mark queryB Many

--------------------------------------------------------------------------------

local function queryServerTest_queryBMany( test, numberOfTimes )
	
	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( 5 )
	
	local queryBs = generateBulkQueryBs( numberOfTimes, expectedCoreInfo )
	
	-- Arbitrary wait time
	local waitTime = numberOfTimes * WAIT_TIME_PER_ITEM
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 2
		
			startClock( 'queryBMany: ' .. numberOfTimes )
		
			return KSPromises.allP( bulkUpdate, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
		
		 -- Wait for all of the query behaviors to update
		
		:waitFor( updatedQueryCountB ) -- Step 3
		:continueWhenValueMatches( #queryBs )
		
		-- Verify performance is acceptable
		
		:waitFor( function() -- Step 4
			
			stopClock( 'queryBMany: ' .. numberOfTimes )
			
			local executionTime = executionTimes[ ( 'queryBMany: ' .. numberOfTimes ) ]
			if testMetrics then
			
				testMetrics[ "wfp_execution_time" ] = executionTime
				WFTestMetrics.writeTestMetrics( testMetrics )
				
			end
			test:log( "Execution Time: " .. executionTime )
			
			assert( executionTime < #queryBs * EXEC_TIME_PER_ITEM_LONG )
			
			-- Clean up the behaviors
	
			for i = 1, #queryBs do
				queryBs[ i ]:close()
				queryBs[ i ] = nil
			end
			
			return 0
			
		end )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBMany_100_testPerformance( test )

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBMany( test, 100 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBMany_1000_testPerformance( test )

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBMany( test, 1000 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServerMetricsTest.testQueryBMany_5000_testPerformance( test )

	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	testMetrics = WFTestMetrics.populateTestMetrics( test )
	queryServerTest_queryBMany( test, 5000 )
		
end

--------------------------------------------------------------------------------

return WFIndexStoreQueryServerMetricsTest
