--[[----------------------------------------------------------------------------

WFIndexStoreSchema.lua

@author scouten 2011-03-28
@reviewed scouten 2012-08-10

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFArguments = require 'WFArguments'
local WFDigest = require 'WFDigest'

	-- wichitafoundation.sqlite.database
local KSDatabase = require 'KSDatabase'
local KSDatabaseSchema = require 'KSDatabaseSchema'
local WFSqliteUtils = require 'WFSqliteUtils'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )

local md5 = assert( WFDigest.MD5, "Cannot find MD5 digest" )


--============================================================================--
--[[!
	\class WFIndexStoreSchema
	\brief Represents a database schema used by an <b>index definition file</b>
		for an index store.
	
	An <b>index definition file</b> is application-specific code that describes
	what data should be extracted (indexed) from a document store
	(\c OzDocStore) to support the queries it will need to perform.
	
	You can start a new index store at the end of the index definition file
	using <code>WFIndexStoreImp.start( schema, ... )</code>.

	Create tables and indexes to support the queries you will be performing
	against this data using the \c WFIndexStore interface.
	Use <code>:addTable</code> to add any number of tables, and
	<code>:addIndex</code> to add any number of indexes.
	Be aware of performance hits for more tables and indexes.

	A typical index definition file looks like this:

\code
local WFIndexStoreSchema = require 'WFIndexStoreSchema'
local indexStoreSchema = WFIndexStoreSchema()

indexStoreSchema:addTable {

	sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
	map = function( docValue, emit )
	
		-- WARNING: This version is grossly simplified and omits error checking.
		-- See OzCatalogIndexDefinitions.lua for a better example.
	
		if docValue._docType == 'asset' then
			emit( docValue.width / docValue.height,
				  docValue.captureDate:sub( 1, 10 ) )
		end
	end,
}

indexStoreSchema:addIndex
	"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )"

require 'WFIndexStoreImp'.start( schema, ... )
\endcode

--]]

-- @mark WFIndexStoreSchema class

local indexStoreSchemaMethods = {

	_schema = false,
		-- (KSDatabaseSchema)

	_tableInfo = false,
		-- (table)
			-- key (string): table name as declared in SQL
			-- value (table): args to :addTable method, plus several other
				-- bits of information we gather/generate along the way

	_lcTableNames = false,
		-- (table)
			-- key (string): table name as declared in SQL (normalized to LC)
			-- value: true
	
	_niceToHaveIndices = false,
		-- (table)
			-- key (string) : table name as declared in SQL or fields
			-- value (table): a set of columns that are nice to have (often the
			-- empty set but never nil)
}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new index-store schema instance.

]]--

function indexStoreSchemaMethods:__init()

	self._schema = KSDatabaseSchema( '1.0' ):useWAL()

	-- Create a table to track status of indexes.
	
	self._schema:addTable [[CREATE TABLE indexStatus(
					indexTableName UNIQUE,
					needsUpdateFrom,
					sqlSchemaHash,
					mapFunctionHash)]]
		:onConflict( 'replace' )
		
		-- indexTableName = name of table created via :addTable below
		-- needsUpdateFrom = doc change counter from OzDocStore
			-- when this table was last known to be up to date
		-- sqlSchemaHash = MD5 hash of the SQL schema string used to
			-- create this table
		-- mapFunctionHash = MD5 hash of the map function used to
			-- extract data into this table

		-- Note that there is a special entry titled "__invalidate"
		-- which is used to keep the docChangeCounter value for when
		-- invalidation noticees were sent to all other index tables.

	-- Create a table used to quickly invalidate document contents.
		-- FIXME [scouten 2012-08-10]: Not sure this will survive the
		-- refactoring, but leaving it in for now.

	self._schema:addTable [[CREATE TEMPORARY TABLE TEMP_invalidatedDocIds( docId )]]
	
	-- Create a table to store index-store local data.
	
	self._schema:addTable [[CREATE TABLE localState( name UNIQUE NOT NULL, value )]]
	
	self._schema:addTable [[CREATE TEMPORARY TABLE TEMP_ids( docId )]]

	self._tableInfo = {}
	self._lcTableNames = {}
	self._niceToHaveIndices = {}
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Adds an index table to this schema.
	
	\param args (table) description of the index table
		- \c sql: (string) a <code>CREATE TABLE</code> SQL statement
		- \c map: (function) a function taking <code>( docValue, emit )</code>
			that maps a single document value to zero or more rows in this table.
			Rows are created by calling the \c emit function with parameters in
			the sequence specified in the \c sql statement.

	The index store implementation records the values of both
	\c sql and \c map. (Actually, it stores an MD5 hash of both.) If either
	\c sql or \c map is changed, the existing contents of the index is
	discarded and the table's contents is regenerated automatically.

	These table names are reserved for use by \c WFIndexStore:
		
		- \c indexStatus
		- \c localState
		- \c TEMP_invalidatedDocIds
		- Any table name with the prefix \c WFIndexStore_
	
	Remember that SQL table names are case-insensitive, so any case variation
	of those names are also reserved.
--]]

function indexStoreSchemaMethods:addTable( --[[table]] args )

	check( 'WFIndexStoreSchema', 'table' )

	if type( args.sql ) ~= 'string' then
		if type( args.fields ) ~= 'table' or type( args.tableName ) ~= 'string' then
			error( "WFIndexStoreSchema:addTable: args.sql must be a string or ( args.fields must be an array and args.tableName must be a string )", 2 )
		end
	end
	
	if not args.fields then
		if not string.lower( args.sql ):match( "^%s*create%s+table%s+" ) then
			error( "WFIndexStoreSchema:addTable: args.sql must be a CREATE TABLE statement", 2 )
		end
	end
	
	if type( args.map ) ~= 'function' then
		error( "WFIndexStoreSchema:addTable: args.sql must be a function", 2 )
	end
	
	-- Insert standard fields.
	
	local sql
	local fields = args.fields
	local tableName = args.tableName
	local nColumns
	if not fields then
		sql = args.sql:gsub( "%(", "( docId, validRev,", 1 )
		local columns = sql:match( "%((.*)%)" )
		local _, nCommas = columns:gsub( ',', ',' )
		
		nColumns = nCommas - 1
			-- +1 because there is one more column than comma,
			-- but -2 because we don't count our two columns here
	else
		sql = string.format( "CREATE TABLE %s ( docId, validRev, %s )",
			tableName, table.concat( fields, ", " ) )
		nColumns = #fields
	end
	
	args.nColumns = nColumns
	
	if not tableName then
		tableName = WFSqliteUtils.tableNameFromCreateTableStatement( sql )
	end

	if not tableName then
		error( "WFIndexStoreSchema:addTable: could not find table name in SQL " .. sql, 2 )
	end
	
	local niceToHaveIndices = {}
	
	if args.niceToHave then
		assert( fields, "niceToHave fields must be declared via fields and tableName" )
		local columnNameIndices = {}
		for index, field in ipairs( fields ) do
			columnNameIndices[ field ] = index
		end
		for _, field in ipairs( args.niceToHave ) do
			niceToHaveIndices[ columnNameIndices[ field ] ] = true
		end
	end
	
	local lcTableName = string.lower( tableName )
	
	if lcTableName:match( "^wfindexstore_" )
		or lcTableName == 'indexstatus'
		or lcTableName == 'localstate'
		or lcTableName == 'temp_invalidateddocids'
	then
		error( string.format( "WFIndexStoreSchema:addTable: illegal table name: %s",
							  tableName ), 2 )
	end
	
	if self._lcTableNames[ lcTableName ] then
		error(
			string.format( "WFIndexStoreSchema:addTable: table named %q already exists", tableName ),
			2 )
	end

	args.sqlSchemaHash = md5.digest( sql )
	args.mapFunctionHash = md5.digest( string.dump( args.map ) .. '+2012-06-25' )
		-- [scouten 2012-06-25]: The string concat will force all index tables in
		-- all catalogs to rebuild because they might be missing data. (PR #3221655)
	
	self._tableInfo[ tableName ] = args
	self._niceToHaveIndices[ tableName ] = niceToHaveIndices
	self._lcTableNames[ lcTableName ] = true

	self._schema:addIndex( string.format(
		[[CREATE INDEX %s_docId ON %s( docId )]], tableName, tableName ) )
	
	self._schema:addIndex( string.format(
		[[CREATE INDEX %s_validRef ON %s( validRev )]], tableName, tableName ) )
	
	return self._schema:addTable( sql ):onConflict( 'replace' )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds an index to the schema.
	
	\param sql (string) A <code>CREATE INDEX</code> SQL statement.
		This index must refer to one of the tables created with
		<code>:addTable</code>.
--]]

function indexStoreSchemaMethods:addIndex( --[[string]] sql )

	check( 'WFIndexStoreSchema', 'string' )

	return self._schema:addIndex( sql )

end

--------------------------------------------------------------------------------
--[[!
	\brief Opens an existing database or creates a new one with this index
		store schema.
	
	\param path (string) The path at which to find or create the database.

	\return (\c KSDatabase) The opened database object.
--]]

function indexStoreSchemaMethods:openOrCreateDatabaseAtPath( --[[string]] path )

	check( 'WFIndexStoreSchema', 'string' )

	return KSDatabase( path, self._schema )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the set of table names declared in this schema.
	
	\return (table)
		- key (string): A table name as declared in SQL
		- value (boolean): \c true
--]]

function indexStoreSchemaMethods:tableNames()

	local result = {}
	
	for k in pairs( self._tableInfo ) do
		result[ k ] = true
	end

	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves an array of niceToHave column indices for each table
	declared in this schema.
	
	\return (table)
		- key (string): A table name as declared in SQL or via tableName
		- value (set): a table mapping updated column index to true if
		that column is niceToHave
--]]

function indexStoreSchemaMethods:niceToHaveIndices()

	return self._niceToHaveIndices

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the map function for each table declared in this schema.
	
	\return (table)
		- key (string): A table name as declared in SQL
		- value (function): The map function for this table
--]]

function indexStoreSchemaMethods:mapFunctions()

	local result = {}
	
	for k, tableInfo in pairs( self._tableInfo ) do
		result[ k ] = tableInfo.map
	end
	
	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the number of user-defined columns for each table declared
		in this schema.
	
	\return (table)
		- key (string): A table name as declared in SQL
		- value (number): The number of user-defined columns in this table
--]]

function indexStoreSchemaMethods:nColumnsPerTable()

	local result = {}
	
	for k, tableInfo in pairs( self._tableInfo ) do
	
		result[ k ] = assert( tableInfo.nColumns )
			
	end
	
	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Removes data from any tables where schema or map function have changed
		since the previous app session.
		
	\param db (\c KSDatabase) The database to be updated.
	\param forceRefresh (optional, boolean) When \c true, reset all existing tables.
--]]

function indexStoreSchemaMethods:resetTablesIfNecessary( --[[KSDatabase]] db,
														 --[[optional boolean]] forceRefresh )

	check( 'WFIndexStoreSchema', 'KSDatabase', '?boolean' )

	-- Find out the status of existing tables in the database.
	
	local indexStatusRows = db:gatherArrayOfArray
		[[SELECT indexTableName, sqlSchemaHash, mapFunctionHash, needsUpdateFrom
			FROM indexStatus]]
	
	local indexStatusByTable = {}
	
	for i = 1, #indexStatusRows do
		local row = indexStatusRows[ i ]
		indexStatusByTable[ row[ 1 ] ] = {
			sqlSchemaHash = row[ 2 ],
			mapFunctionHash = row[ 3 ],
			needsUpdateFrom = row[ 4 ],
		}
	end
	
	-- Reset any tables where we have a potential mismatch.
	
	db:withTransactionDo( function()
	
		for tableName, tableInfo in pairs( self._tableInfo ) do
		
			local indexStatus = indexStatusByTable[ tableName ]
	
			local clearTable
		
			if indexStatus then
			
				if indexStatus.sqlSchemaHash ~= tableInfo.sqlSchemaHash
					or indexStatus.mapFunctionHash ~= tableInfo.mapFunctionHash
					or forceRefresh
				then
				
					db:execOnce(
						[[DELETE FROM indexStatus WHERE indexTableName == ?]],
										tableName )
					
					clearTable = true
	
				end
			
			else
			
				clearTable = true
			
			end
	
			if clearTable then
	
				db:execOnce( [[INSERT INTO indexStatus VALUES( ?, ?, ?, ? )]],
								tableName, 1, tableInfo.sqlSchemaHash,
								tableInfo.mapFunctionHash )
			
				db:execOnce( string.format( [[DELETE FROM %s]], tableName ) )
				
				if indexStatus then
					indexStatus.needsUpdateFrom = 0
				end
			
			end
			
			tableInfo.lastUpdated = indexStatus and ( indexStatus.needsUpdateFrom - 1 ) or 0
		
		end
	
	end )

end

--------------------------------------------------------------------------------

return KSClass 'WFIndexStoreSchema'( indexStoreSchemaMethods )
