--[[----------------------------------------------------------------------------

WFIndexStoreSchemaTest.lua

@author scouten 2012-08-13
@reviewed scouten 2012-08-13

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.utils
local KSTestUtils = require 'KSTestUtils'
local WFDigest = require 'WFDigest'

	-- wichitafoundation.oz_client.index_store
local WFIndexStoreSchema = require 'WFIndexStoreSchema'

	-- wichitafoundation.sqlite.database
local KSDatabase = require 'KSDatabase'


--------------------------------------------------------------------------------

local md5 = assert( WFDigest.MD5, "Cannot find MD5 digest" )


--============================================================================--

local WFIndexStoreSchemaTest = {}

--------------------------------------------------------------------------------

local function mapForExampleSchema( docValue, emit )

	-- WARNING: This version is grossly simplified and omits error checking.
	-- See OzCatalogIndexDefinitions.lua for a better example.

	if docValue._docType == 'asset' then
		emit( docValue.width / docValue.height,
			  docValue.captureDate:sub( 1, 10 ) )
	end

end

local exampleFunction_md5 = md5.digest( string.dump( mapForExampleSchema ) .. '+2012-06-25' )

--------------------------------------------------------------------------------

local function exampleSchemaFromApiDocs()

	-- Why not start by testing the exact schema declaration described in
	-- the API docs? :-)
	
	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
		map = mapForExampleSchema,
	}
	
	indexStoreSchema:addIndex
		"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )"
	
	return indexStoreSchema

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testTableNames_simple()

	local indexStoreSchema = exampleSchemaFromApiDocs()
	
	local actualTableNames = indexStoreSchema:tableNames()
	
	KSTestUtils.assertTableMatches( actualTableNames,
		{ coreInfo = true } )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testTableNames_isACopy()

	-- Verify that the table we get back from :tableNames is a tear-off
	-- copy of the underlying data.

	local indexStoreSchema = exampleSchemaFromApiDocs()
	
	local tableNamesCopy1 = indexStoreSchema:tableNames()
	
	tableNamesCopy1.coreInfo = nil
	
	local tableNamesCopy2 = indexStoreSchema:tableNames()

	KSTestUtils.assertTableMatches( tableNamesCopy2,
		{ coreInfo = true } )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testTableNames_threeTables()

	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
		map = mapForExampleSchema,
	}
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE anotherTable( foo, bar )",
		map = mapForExampleSchema,
	}
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE blah( boop, bap )",
		map = mapForExampleSchema,
	}
	
	local actualTableNames = indexStoreSchema:tableNames()
	
	KSTestUtils.assertTableMatches( actualTableNames,
		{
			anotherTable = true,
			blah = true,
			coreInfo = true,
		} )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testOpenOrCreateDatabaseAtPath()

	-- Why not start by testing the exact schema declaration described in
	-- the API docs? :-)
	
	local indexStoreSchema = exampleSchemaFromApiDocs()
	
	-- Instead of spinning up an index store based on this, we'll apply it
	-- directly to an in-memory database and verify the result schema.

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )
	
	assert( KSDatabase:isClassOf( db ),
			":openOrCreateDatabaseAtPath did not return a KSDatabase instance" )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testOpenOrCreateDatabaseAtPath_validateArgs()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		return indexStoreSchema:openOrCreateDatabaseAtPath()
			-- No path. Should fail.

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_table()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		indexStoreSchema:addTable()
			-- No args table. Should fail.

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_sqlString()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		indexStoreSchema:addTable {
			sql = 42,
				-- FAIL: Should be a string.
			map = function() end,
		}

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_sqlCreateTable()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		indexStoreSchema:addTable {
			sql = "CREATE BOGUS TABLE foo( bar, blah )",
				-- FAIL: Should be CREATE TABLE!
			map = function() end,
		}

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_sqlCreateTableHasName()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		indexStoreSchema:addTable {
			sql = "CREATE TABLE foo WITH COLUMNS bar, blah",
				-- FAIL: Should be proper CREATE TABLE syntax!
			map = function() end,
		}

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_mapFunction()

	KSTestUtils.assertFails( function()
		
		local indexStoreSchema = WFIndexStoreSchema()

		indexStoreSchema:addTable {
			sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
			map = "This is not a function.",
				-- FAIL: Should be a function
		}

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_createTableTwice()

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {
	
		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
		
		map = function( docValue, emit )
		
			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.
		
			if docValue._docType == 'asset' then
				emit( docValue.width / docValue.height,
					  docValue.captureDate:sub( 1, 10 ) )
			end
		
		end,
	
	}
	
	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
		
			sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
			
			map = function( docValue, emit )
			
				-- WARNING: This version is grossly simplified and omits error checking.
				-- See OzCatalogIndexDefinitions.lua for a better example.
			
				if docValue._docType == 'asset' then
					emit( docValue.width / docValue.height,
						  docValue.captureDate:sub( 1, 10 ) )
				end
			
			end,
		
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_createTableTwice()

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {
	
		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
		
		map = function( docValue, emit )
		
			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.
		
			if docValue._docType == 'asset' then
				emit( docValue.width / docValue.height,
					  docValue.captureDate:sub( 1, 10 ) )
			end
		
		end,
	
	}
	
	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
		
			sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
			
			map = function( docValue, emit )
			
				-- WARNING: This version is grossly simplified and omits error checking.
				-- See OzCatalogIndexDefinitions.lua for a better example.
			
				if docValue._docType == 'asset' then
					emit( docValue.width / docValue.height,
						  docValue.captureDate:sub( 1, 10 ) )
				end
			
			end,
		
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_reservedName1()

	local indexStoreSchema = WFIndexStoreSchema()

	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
			sql = "CREATE TABLE indexstatus( aspectRatio, displayDate )",
			map = mapForExampleSchema,
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_reservedName2()

	local indexStoreSchema = WFIndexStoreSchema()

	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
			sql = "CREATE TABLE LocalState( aspectRatio, displayDate )",
			map = mapForExampleSchema,
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_reservedName3()

	local indexStoreSchema = WFIndexStoreSchema()

	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
			sql = "CREATE TABLE temp_invalidatedDocIds( aspectRatio, displayDate )",
			map = mapForExampleSchema,
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testAddTable_validateArgs_reservedName4()

	local indexStoreSchema = WFIndexStoreSchema()

	KSTestUtils.assertFails( function()
		
		indexStoreSchema:addTable {
			sql = "CREATE TABLE wfIndexStore_foo( aspectRatio, displayDate )",
			map = mapForExampleSchema,
		}
		
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_validateArgs_db()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	KSTestUtils.assertFails( function()
		indexStoreSchema:resetTablesIfNecessary()
			-- First arg should be a KSDatabase instance.
	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_validateArgs_forceRefresh()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	KSTestUtils.assertFails( function()

		local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

		indexStoreSchema:resetTablesIfNecessary( db, 42 )
			-- Second arg should be nil or boolean.

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_emptyDatabase()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	indexStoreSchema:resetTablesIfNecessary( db )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

	-- Examine the indexStatus table.
	
	local actualIndexStatus = db:gatherArrayOfArray
		[[SELECT * FROM indexStatus ORDER BY indexTableName]]

	KSTestUtils.assertTableMatches( actualIndexStatus,
		{
			{
				"coreInfo",
				1,
				"c42ffe40cc4243786d6351f0c11ddd86",
				exampleFunction_md5,
			},
		} )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_existingDatabase_unchanged()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	-- Fake some existing rows in the index status.
	
	db:execOnce( [[INSERT INTO indexStatus VALUES( 'coreInfo', 1, 'c42ffe40cc4243786d6351f0c11ddd86', ? )]],
				 exampleFunction_md5 )

	-- Fake existing content in tables.

	db:execOnce
		[[INSERT INTO coreInfo VALUES( 'docId', 'validRev', 1.5, '2009-08-11' )]]

	-- Reset the schema.

	indexStoreSchema:resetTablesIfNecessary( db )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

	-- Examine the indexStatus table.
	
	local actualIndexStatus = db:gatherArrayOfArray
		[[SELECT * FROM indexStatus ORDER BY indexTableName]]

	KSTestUtils.assertTableMatches( actualIndexStatus,
		{
			{
				"coreInfo",
				1,
				"c42ffe40cc4243786d6351f0c11ddd86",
				exampleFunction_md5,
			},
		} )

	-- Make sure the index table has been cleared.
	
	local actualCount = db:gatherArrayOfArray
		[[SELECT count() FROM coreInfo]]
	
	KSTestUtils.assertTableMatches( actualCount, { { 1 } } )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_existingDatabase_sqlChanged()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	-- Fake some existing rows in the index status.
	
	db:execOnce( [[INSERT INTO indexStatus VALUES( 'coreInfo', 1, 'bogus_sql_hash', ? )]],
				 exampleFunction_md5 )

	-- Fake existing content in tables.

	db:execOnce
		[[INSERT INTO coreInfo VALUES( 'docId', 'validRev', 1.5, '2009-08-11' )]]

	-- Reset the schema.

	indexStoreSchema:resetTablesIfNecessary( db )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

	-- Examine the indexStatus table.
	
	local actualIndexStatus = db:gatherArrayOfArray
		[[SELECT * FROM indexStatus ORDER BY indexTableName]]

	KSTestUtils.assertTableMatches( actualIndexStatus,
		{
			{
				"coreInfo",
				1,
				"c42ffe40cc4243786d6351f0c11ddd86",
				exampleFunction_md5,
			},
		} )

	-- Make sure the index table has been cleared.
	
	local actualCount = db:gatherArrayOfArray
		[[SELECT count() FROM coreInfo]]
	
	KSTestUtils.assertTableMatches( actualCount, { { 0 } } )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_existingDatabase_mapFnChanged()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	-- Fake some existing rows in the index status.
	
	db:execOnce
		[[INSERT INTO indexStatus VALUES( 'coreInfo', 1, 'c42ffe40cc4243786d6351f0c11ddd86', 'bogus_map_changed' )]]

	-- Fake existing content in tables.

	db:execOnce
		[[INSERT INTO coreInfo VALUES( 'docId', 'validRev', 1.5, '2009-08-11' )]]

	-- Reset the schema.

	indexStoreSchema:resetTablesIfNecessary( db )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

	-- Examine the indexStatus table.
	
	local actualIndexStatus = db:gatherArrayOfArray
		[[SELECT * FROM indexStatus ORDER BY indexTableName]]

	KSTestUtils.assertTableMatches( actualIndexStatus,
		{
			{
				"coreInfo",
				1,
				"c42ffe40cc4243786d6351f0c11ddd86",
				exampleFunction_md5,
			},
		} )

	-- Make sure the index table has been cleared.
	
	local actualCount = db:gatherArrayOfArray
		[[SELECT count() FROM coreInfo]]
	
	KSTestUtils.assertTableMatches( actualCount, { { 0 } } )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testResetTablesIfNecessary_existingDatabase_forceRefresh()

	local indexStoreSchema = exampleSchemaFromApiDocs()

	local db = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	-- Fake some existing rows in the index status.
	
	db:execOnce( [[INSERT INTO indexStatus VALUES( 'coreInfo', 1, 'c42ffe40cc4243786d6351f0c11ddd86', ? )]],
				 exampleFunction_md5 )

	-- Fake existing content in tables.

	db:execOnce
		[[INSERT INTO coreInfo VALUES( 'docId', 'validRev', 1.5, '2009-08-11' )]]

	-- Reset the schema.

	indexStoreSchema:resetTablesIfNecessary( db, true )

	-- Examine the schema.
	
	local actualSchema = db:gatherArrayOfArray
		[[SELECT sql FROM sqlite_master ORDER BY name, type]]
	
	KSTestUtils.assertTableMatches( actualSchema,
		{
			{
				"CREATE TABLE coreInfo( docId, validRev, aspectRatio, displayDate )",
			},
			{
				"CREATE INDEX coreInfo_displayDate ON coreInfo( displayDate )",
			},
			{
				"CREATE INDEX coreInfo_docId ON coreInfo( docId )",
			},
			{
				"CREATE INDEX coreInfo_validRef ON coreInfo( validRev )",
			},
			{
				"CREATE TABLE indexStatus(\n\t\t\t\t\tindexTableName UNIQUE,\n\t\t\t\t\tneedsUpdateFrom,\n\t\t\t\t\tsqlSchemaHash,\n\t\t\t\t\tmapFunctionHash)",
					-- Ugh, but white space has to match.
			},
			{
				"CREATE TABLE localState( name UNIQUE NOT NULL, value )",
			},
			{
				"CREATE TABLE schemaVersion( currentVersion, backwardCompatibleToVersion )",
			},
			{
			},
			{
			},
		} )

	-- Examine the indexStatus table.
	
	local actualIndexStatus = db:gatherArrayOfArray
		[[SELECT * FROM indexStatus ORDER BY indexTableName]]

	KSTestUtils.assertTableMatches( actualIndexStatus,
		{
			{
				"coreInfo",
				1,
				"c42ffe40cc4243786d6351f0c11ddd86",
				exampleFunction_md5,
			},
		} )

	-- Make sure the index table has been cleared.
	
	local actualCount = db:gatherArrayOfArray
		[[SELECT count() FROM coreInfo]]
	
	KSTestUtils.assertTableMatches( actualCount, { { 0 } } )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testMapFunctions()

	local indexStoreSchema = exampleSchemaFromApiDocs()
	
	local actualMapFunctions = indexStoreSchema:mapFunctions()

	KSTestUtils.assertTableMatches( actualMapFunctions,
									{ coreInfo = mapForExampleSchema } )

end

--------------------------------------------------------------------------------

function WFIndexStoreSchemaTest.testNColumnsPerTable()

	local indexStoreSchema = WFIndexStoreSchema()
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",
		map = mapForExampleSchema,
	}
	
	indexStoreSchema:addTable {
		sql = "CREATE TABLE blah( foo )",
		map = mapForExampleSchema,
	}
	
	local actualNColumns = indexStoreSchema:nColumnsPerTable()
	
	KSTestUtils.assertTableMatches( actualNColumns,
									{ coreInfo = 2, blah = 1 } )

end

--------------------------------------------------------------------------------

return WFIndexStoreSchemaTest
