--[[----------------------------------------------------------------------------

WFIndexStore_WFDocStore_Test.lua

@author scouten 2011-03-28
@reviewed scouten 2012-08-22

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'


--============================================================================--

local WFIndexStore_WFDocStore_Test = {}

local docStore, indexStore, failBecause

local randomDocId = '9b572d0e423c4654b078893ddd41c230'

local WFPlatformInfo = require 'WFPlatformInfo'

local testTimeout = WFPlatformInfo.isAndroid and 4 or 2

--------------------------------------------------------------------------------

-- FIXME [scouten 2012-08-22]: This file would probably benefit from some of the
-- new tests in WFIndexStoreQueryServerTest. OTOH, that gives nearly 100% coverage
-- of the underlying code, so maybe not as critical.

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.setUp()

	docStore = WFDocStore( ':memory:' )

	failBecause = nil
	
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.tearDown( coveragePath )
	
	if coveragePath then
		if indexStore then
			indexStore:DEBUG_dumpCoverageP( coveragePath )
			:runUntilDone()
		end
	end
	
	if docStore then
		docStore:close()
		docStore = nil
	end

	if failBecause then
		error( failBecause )
	end

end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testSimpleQueryB( test )

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local aspectRatiosB = indexStore:queryB(
		[[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]] )

	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches { { randomDocId, 1.5 } }

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'af21cdb6d2fa4d1c8099d08a6ba1d7c1',
				_slot = 'main',
				height = 130,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ randomDocId, 1.5 },
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
		}
		
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testReindexSingleDoc( test )

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local aspectRatiosB = indexStore:queryB(
		[[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]] )

	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:applyOverlayP {
					_id = randomDocId,
					_slot = 'main',
					height = 150,
					width = 100,
				}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches { { randomDocId, 1.5 } }

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				height = 177,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ randomDocId, 1.77 },
		}
		
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testIndexRecordsForDeletedDocsGetDeleted( test )

	-- NOTE: There's a possible gap where docs could be deleted (or more
	-- likely, their creation undone). If the index store for any given table
	-- isn't being actively updated, it's possible that deletion might linger
	-- and thus the index store data would be out of date.
	
	-- For the moment, that isn't an issue but once we implement garbage
	-- collection of primary doc store content this could crop up.

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local aspectRatiosB = indexStore:queryB(
		[[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]] )

	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}
			
		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches { { randomDocId, 1.5 } }

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = randomDocId,
				_slot = 'main',
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {}

		:waitFor( function()
		
			return docStore:setDocValueP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches { { randomDocId, 1.5 } }

end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testMultipleResultsPerDoc( test )

	indexStore = WFIndexStore( docStore, 'TestKeywords.lua' )

	local keywordsB = indexStore:queryB(
		[[SELECT docId, keyword FROM keywords
					ORDER BY docId, keyword]] )

	WFAsyncTestCase( test, testTimeout )
	
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				keywords = {
					dog = true,
					cat = true,
				},
			}

		end )
		
		:waitFor( keywordsB )
		:continueWhenValueMatches {
			{ randomDocId, "cat" },
			{ randomDocId, "dog" },
		}

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				keywords = {
					_replace = true,
					dog = true,
					bat = true,
				},
			}

		end )
		
		:waitFor( keywordsB )
		:continueWhenValueMatches {
			{ randomDocId, "bat" },
			{ randomDocId, "dog" },
		}
		
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				keywords = {
					_delete = true,
				},
			}

		end )
		
		:waitFor( keywordsB )
		:continueWhenValueMatches {}
		
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testDoesntSendRedundantUpdates( test )

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local aspectRatiosB = indexStore:queryB(
		[[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]] )

	local updateCount = 0
	local radioSilenceExpected = false

	local updateDetectorB = aspectRatiosB:mapB( function( data )
		debug.print "remapping aspectRatio data"
		updateCount = updateCount + 1
		return updateCount
	end )

	local onDataTrigger

	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches { { randomDocId, 1.5 } }

		:waitFor( .1 )
			-- Hack around a race condition: The updateDetectorB might
			-- be triggered before or after this.

		:waitFor( function()

			onDataTrigger = aspectRatiosB:onDataE( function()
				if radioSilenceExpected then
					debug.print "got unwanted onDataE callback!"
					updateCount = updateCount + 1
				end
			end )

			updateCount = 0
			radioSilenceExpected = true
			
			-- debug.print( "-------------------------------------------- should be radio silence after this point --------------------------------------------" )
		
			return docStore:applyOverlayP {
					_id = randomDocId,
					_slot = 'main',
					foo = 'bar',
						-- not part of aspecRatio calculation,
						-- should not cause index update
				}

		end )
		
		:waitFor( .75 )
		:whenResolvedDo( function()
			
			-- [scouten 2011-04-08]: I don't really expect either
			-- of these conditions to ever happen, but I'm using this
			-- to keep the two triggers from being GC'd.

			if not onDataTrigger or not updateDetectorB then
				error "huh"
			end
			
			if updateCount > 0 then
				error "Redundant update notification sent!"
			end
			
		end )
		
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testQueryB_sqlB( test )

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local sqlB = KSReactive.variableB(
		[[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]] )

	local aspectRatiosB = indexStore:queryB( sqlB )

	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}

		end )
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'af21cdb6d2fa4d1c8099d08a6ba1d7c1',
				_slot = 'main',
				height = 130,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ randomDocId, 1.5 },
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
			sql = [[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio DESC]],
		}
		
		:waitFor( function()
			sqlB:set( [[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio]] )
			return aspectRatiosB
		end )
		:continueWhenValueMatches {
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
			{ randomDocId, 1.5 },
			sql = [[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio]],
		}

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = '8793eb265f8241009129c21fc850a59d',
				_slot = 'main',
				height = 115,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ '8793eb265f8241009129c21fc850a59d', 1.15 },
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
			{ randomDocId, 1.5 },
			sql = [[SELECT docId, aspectRatio FROM aspectRatio
					ORDER BY aspectRatio]],
		}
		
end

--------------------------------------------------------------------------------

function WFIndexStore_WFDocStore_Test.testQueryB_paramB( test )

	indexStore = WFIndexStore( docStore, 'TestAspectRatio.lua' )

	local minAspectRatioB = KSReactive.variableB( 1.3 )

	local aspectRatiosB = indexStore:queryB( [[SELECT docId, aspectRatio FROM aspectRatio
					WHERE aspectRatio >= ?
					ORDER BY aspectRatio]], minAspectRatioB )


	WFAsyncTestCase( test, testTimeout )

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = randomDocId,
				_slot = 'main',
				height = 150,
				width = 100,
			}

		end )
		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = 'af21cdb6d2fa4d1c8099d08a6ba1d7c1',
				_slot = 'main',
				height = 130,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
			{ randomDocId, 1.5 },
			args = { 1.3 },
		}
		
		:waitFor( function()
			minAspectRatioB:set( 1.5 )
			return aspectRatiosB
		end )
		:continueWhenValueMatches {
			{ randomDocId, 1.5 },
			args = { 1.5 },
		}

		:waitFor( function()
		
			return docStore:applyOverlayP {
				_id = '8793eb265f8241009129c21fc850a59d',
				_slot = 'main',
				height = 115,
				width = 100,
			}

		end )
		
		:waitFor( aspectRatiosB )
		:continueWhenValueMatches {
			{ randomDocId, 1.5 },
			args = { 1.5 },
		}

		:waitFor( function()
			minAspectRatioB:set( 1 )
			return aspectRatiosB
		end )
		:continueWhenValueMatches {
			{ '8793eb265f8241009129c21fc850a59d', 1.15 },
			{ 'af21cdb6d2fa4d1c8099d08a6ba1d7c1', 1.3 },
			{ randomDocId, 1.5 },
			args = { 1 },
		}

		
end

--------------------------------------------------------------------------------

return WFIndexStore_WFDocStore_Test
