--[[----------------------------------------------------------------------------

WFLogger.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2007 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Substrate framework
local KSClass = require 'KSClass'
local WFConfiguration = require 'WFConfiguration'
local WFFileLogger = require 'WFFileLogger'
local WFLoggerImp = require 'WFLoggerImp'
local KSTableArchiving = require 'KSTableArchiving'

local string_format = string.format

--==============================================================================

local weakLoggers = setmetatable( { }, { __mode = 'kv' } )

--============================================================================--
--[[!

	\brief \c WFLogger
	This class provides a mechanism for writing debug output that can be viewed
	with an external log viewer application.
	<p>Use the imported namespace as a constructor; access the functions through
	the created objects. The constructor creates a new logger with a given name,
	or finds and returns an existing one.  If no name is provided, it returns
	an application-wide "default" <code>WFLogger</code> object.
	
	Loggers are silent until configured with <code>WFLogger:enable()</code>.</p>
	<p>Common applications for viewing debug output are:</p>
	
	<ul>
		<li>WinDBG (available for download from <a href="http://www.microsoft.com/whdc/DevTools/Debugging/default.mspx">Microsoft</a>)</li>
		<li>Microsoft Developer Studio</li>
		<li>Console (built-in application on Mac OS: look in /Applications/Utilities)</li>
		<li>Xcode</li>
	</ul>
	<p>Configure loggers using the config.lua file as follows:
	<ul>
	 <li>Create an entry for each logger of interest by specifying loggers.<i>loggerName</i> = {...}
	 <li>Arguments in the table include <code>logLevel</code> and a string "fatal", "error",
	     "warn", "info", "debug", or "trace".
	 <li>For each valid log level, you can add an entry to modify the behavior of log messages for that level.
	     For example, to log to a file, add the entry <code>trace = 'logfile',</code>
	 <li>To specify an action for all log levels, add the entry <code>action = 'logfile',</code>
	</ul>
	<p>Sample code...</p>
	<code>
	log = require 'WFLogger' 'myPlugin'
	
	log:enable 'print'		-- this can also be configured in config.lua with:  logger.myPlugin = "print"
	
	log:warn 'foo - warning'
	</code>
]]--

--------------------------------------------------------------------------------

-- @mark -
-- @mark WFLogger namespace

local logger = {

	_name = false, -- (string) name of logger (or nil for default)
	
	_actions = false,  -- (table) what to do on each action (nil or function)
								-- key: (string) name of log type (trace, warn, info, error, or fatal)
								-- value: (string) what to do for this action (console or nil)
	
	_priority = false,	-- an int indicating what log level we are at, as defined in
						-- kLogLevelToPriority below
	
	_savedConfigurations = false,	-- saved actions table and priority values for making temporary changes
									-- to the logger behavior

}

--------------------------------------------------------------------------------

function logger.__init( self, name )

	assert( type( name ) == 'string', "WFLogger:init: name, if it exists, must be a string" )
	self._name = name or '(default)'
	self._savedConfigurations = {}

	self:disable()

end

--------------------------------------------------------------------------------
--[[!
	\brief Disables all log output from this logger.
 	Equivalent to <code>logger:enable( false )</code>.
]]--

function logger:disable()

	self._actions = {}
	return self

end

--------------------------------------------------------------------------------

local function paramStrings( ... )

	local params = { ... }

	for i = 1, #params do
		local p = params[ i ]
		if type( p ) == "table" then
			local success, pp = pcall( KSTableArchiving.simpleTableToChunkString, p )
			if success then p = pp end
		end
		params[ i ] = tostring( p )
	end
	
	return params

end

local function paramMessageString( ... )
	local params = paramStrings( ... )

	return table.concat( params, ' \t ' )
end

local function paramListToString( actionType, ... )

	local message = paramMessageString( ... )
	return string_format( '%s\t%s', string.upper( actionType ), message )

end

local function loggerParamListToString( self, actionType, ... )

	local message = paramMessageString( ... )
	return string_format( '%s\t%s\t%s', self._name, string.upper( actionType ), message )

end

--------------------------------------------------------------------------------

local function makeSingleStringOutputFunc( outputFunc )

	return function( self, actionType, ... )
	
		outputFunc( loggerParamListToString( self, actionType, ... ) )
	
	end

end

--------------------------------------------------------------------------------

local function alwaysTracebackUnformatted( str )
	WFLoggerImp.alwaysLogTraceback( "%s", str )
end

--------------------------------------------------------------------------------
-- Log function lookup table.

local tracebacksDisplayedOnce = {}

local function tracebackOnce( self, actionType, ... )
	
	local tb = debug.traceback()
	
	if not tracebacksDisplayedOnce[ tb ] then
		alwaysTracebackUnformatted( loggerParamListToString( self, actionType, ... ) )
		tracebacksDisplayedOnce[ tb ] = true
	end

end

--------------------------------------------------------------------------------

local logFunctions = {

	print = makeSingleStringOutputFunc( WFLoggerImp.alwaysPrint ),
	
	traceback = makeSingleStringOutputFunc( alwaysTracebackUnformatted ),
	
	traceback_once = tracebackOnce,
	
	logfile = function( self, actionType, ... )
		WFFileLogger.writeLine( self._name, paramListToString( actionType, ... ), true )
	end,

}

--------------------------------------------------------------------------------

local function checkOptionalLogAction( action )

	if type( action ) == 'function' then
		return makeSingleStringOutputFunc( action )
	else
		return logFunctions[ action ]
	end

end

--------------------------------------------------------------------------------

local function checkLogAction( action )

	if not action then return end

	return assert( checkOptionalLogAction( action ),
			string_format( "WFLogger:enable: unknown log behavior %s", tostring( action ) ) )

end

--------------------------------------------------------------------------------
--[[!
	\brief Enables specific log output from this logger. This is usually accomplished through
	the config.lua file instead of this function.
	\param actions (string, function, or table) What action to take for each
			log action type. For each of the action types, this must be either a function
			that takes a string (the log message), or one of the following strings:
			   <ul>
				<li><code>print</code>: Print to console log</li>
				<li><code>traceback</code>: Print to console log with stack traceback</li>
				<li><code>logfile</code>: Print to a log file named for the logger.
					This log file is written to your <code>~/Documents</code> folder in Mac OS,
					or to your <code>My Documents</code> folder in Windows</li>.
			   </ul>
			<p>As a shortcut, you can pass a single string or function, which is
			applied to all of the log output from this logger.</p>
]]--

function logger:enable( --[[string/function/table]] actions )

	local actionsType = type( actions )

	if actionsType == 'string' or actionsType == 'function' then
		actions = {
			trace = actions,
			debug = actions,
			info = actions,
			warn = actions,
			error = actions,
			fatal = actions,
		}
	else
		assert( actionsType == 'table', "WFLogger:enable(): actions must be a table or string" )
	end
	
	actions.trace = checkLogAction( actions.trace )
	actions.debug = checkLogAction( actions.debug )
	actions.info = checkLogAction( actions.info )
	actions.warn = checkLogAction( actions.warn )
	actions.error = checkLogAction( actions.error )
	actions.fatal = checkLogAction( actions.fatal )
	
	self._actions = actions

	return self

end

--------------------------------------------------------------------------------

-- @mark Log Configuration

--------------------------------------------------------------------------------

local kLogLevelToPriority = {
	trace = 6,
	debug = 5,
	info = 4,
	warn = 3,
	error = 2,
	fatal = 1
}

local kPriorityToLogLevel = {
	"fatal",
	"error",
	"warn",
	"info",
	"debug",
	"trace"
}

local kDefaultAction = "print"
local kDefaultLogLevel = "error"

--------------------------------------------------------------------------------

function logger:configure( logConfiguration )
	
	local name = self._name
	local logLevel = logConfiguration.logLevel or kDefaultLogLevel
	local priority = kLogLevelToPriority[ logLevel ]
	local defaultAction = logConfiguration.action or kDefaultAction
		
	if not priority then
		debug.print( "Invalid log configuration for", name )
		return
	end
	
	local actionTable = {}
		
	for i = 1, priority do

		local levelName = kPriorityToLogLevel[ i ]
		local action = logConfiguration[ levelName ] or defaultAction
		
		if not checkOptionalLogAction( action ) then
			debug.print( "Invalid log action", action, "defined for", name )
			return
		end
		
		actionTable[ levelName ] = action

	end
	
	self._priority = priority
	self:enable( actionTable )

end

--------------------------------------------------------------------------------

function logger:saveConfiguration()

	table.insert( self._savedConfigurations, {
		priority = self._priority,
		actions = self._actions
	} )
	
end

--------------------------------------------------------------------------------

function logger:restoreConfiguration()

	local config = table.remove( self._savedConfigurations )
	
	if not config then
		assert( false, "unbalanced call to restoreConfiguration" )
	end
	
	self._priority = config.priority
	self._actions = config.actions
	
end

--------------------------------------------------------------------------------

local function pformat( ... )
	-- A protected version of string.format
	
	local success, text = pcall( string_format, ... )
	
	if success then
		return text
	else
		debug.print( "Text formatting failed:", text, ... )
	end
	
end

--------------------------------------------------------------------------------

local function addLoggerMethods( selector )

	logger[ selector ] = function( self, ... )
		local actionFunc = self._actions[ selector ]
		if actionFunc then
			pcall( actionFunc, self, selector, ... )
		end
	end
	
	logger[ selector .. 'f' ] = function( self, ... )
		local actionFunc = self._actions[ selector ]
		if actionFunc then
			local text = pformat( ... )
			if text then
				pcall( actionFunc, self, selector, text )
			end
		end
	end
	
	logger[ 'will_' .. selector ] = {
		get = function( self )
			return self._actions[ selector ] ~= nil
		end,
	}

end

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'trace' messages.
	-- @name logger:trace
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'trace' messages,
 -- using string.format to prepare the output.
	-- @name logger:tracef
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'trace' )

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'debug' messages.
	-- @name logger:debug
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'debug' messages,
 -- using string.format to prepare the output.
	-- @name logger:debugf
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'debug' )

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'info' messages.
	-- @name logger:info
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'info' messages,
 -- using string.format to prepare the output.
	-- @name logger:infof
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'info' )

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'warn' messages.
	-- @name logger:warn
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'warn' messages,
 -- using string.format to prepare the output.
	-- @name logger:warnf
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'warn' )

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'error' messages.
	-- @name logger:error
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'error' messages,
 -- using string.format to prepare the output.
	-- @name logger:errorf
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'error' )

--------------------------------------------------------------------------------
--- Feeds log output through the action function defined for 'fatal' messages.
	-- @name logger:fatal
	-- @class function
	-- @param ... Data to be printed, of any type; <code>tostring()</code> is called on all values.

--- Feeds log output through the action function defined for 'fatal' messages,
 -- using string.format to prepare the output.
	-- @name logger:fatalf
	-- @class function
	-- @param format (string) formatting instructions (see Lua's string.format for details)
	-- @param ... Arguments to string.format required by format string

addLoggerMethods( 'fatal' )

--------------------------------------------------------------------------------

local function nop() end
	-- Canonical no-op function for use in the

--------------------------------------------------------------------------------
--[[!
	\brief Creates optimized versions of specified log functions for use in tight loops.
	This version avoids the overhead of the method lookup, and is reduced
	to a no-op function when logging is disabled.
	For example:
	<pre>
		local warn, info = logger:quick( 'warn', 'info' )
		warn( 'something bad happened' )
	</pre>
	\param ... (one or more strings) The names of log functions to optimize.
	\return (one or more functions) The optimized functions.
]]--

function logger:quick( ... )

	--[[
	INTERNAL USAGE NOTE:
	The functions returned from quick bypass the check for logger configuration, so they
	should not be retained indefinitely since changes to log level and action made via the
	configure method will be ignored.
	]]

	local results = {}
	local params = { ... }

	for i, name in ipairs( params ) do
	
		local action = self._actions[ name ]
		
		if action then
			results[ i ] = function( ... )
				pcall( action, self, name, ... )
			end
		else
			results[ i ] = nop
		end
		
	end
	
	return unpack( results )

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates optimized versions of specified log functions for use in tight loops.
	This version avoids the overhead of the method lookup, and is reduced
	to a no-op function when logging is disabled. Unlike the functions returned
	by <code>logger:quick</code>, these functions take <code>string.format</code>
	instructions.
	For example:
	<pre>
		local warnf, infof = logger:quickf( 'warn', 'info' )
		warnf( 'something %s happened', 'bad' )
	</pre>
	\param ... (one or more strings) The names of log functions to optimize.
	\return (one or more functions) The optimized functions.
]]--

function logger:quickf( ... )

	local results = {}
	local params = { ... }

	for i, name in ipairs( params ) do
	
		local action = self._actions[ name ]
		
		if action then
			results[ i ] = function( ... )
				local text = pformat( ... )
				if text then
					pcall( action, self, name, text )
				end
			end
		else
			results[ i ] = nop
		end
		
	end
	
	return unpack( results )

end

--------------------------------------------------------------------------------
--	wrapMethodWithTrace & wrapMethodsWithTrace
--
--		These methods allow a method or table of method to be wrapped
--		in such a way that a "trace" will be done before entering the
--		function upon each call (but only when tracing is enabled --
--		when trace is off, these are zero-overhead).
--
--		Usage:
--			logger:wrapMethodsWithTrace( MyClassMethods )
--
--		NOTE [tgaul 2008-12-10]: We probably don't want to expose these
--		in the SDK at this point.

function logger:wrapMethodWithTrace( methodName, fn )

	local action = self._actions.trace
	
	if action then
		return function( obj, ... )
			self:trace( methodName, ... )
			return fn( obj, ... )
		end
	end
	
	return fn

end

function logger:wrapMethodsWithTrace( methodsTable )

	local action = self._actions.trace
	
	if action then
		local t = {}
		
		for k, v in pairs( methodsTable ) do
			assert( type( k ) == "string" )
			assert( type( v ) == "function" )
			
			t[ k ] = self:wrapMethodWithTrace( k, v )
		end
		
		for k, v in pairs( t ) do
			methodsTable[ k ] = v
		end
	end
	
	return methodsTable

end

--------------------------------------------------------------------------------

logger.__tostring =
	function( self )
	
	return string_format( "WFLogger( %q )", self._name )
	
end

--------------------------------------------------------------------------------

local WFLoggerClass = KSClass 'WFLogger'( logger )

--------------------------------------------------------------------------------

local function globMatch( name, pattern )
	pattern = string.gsub( pattern, "%.", "." )
	pattern = string.gsub( pattern, "%*", ".+" )
	pattern = string.gsub( pattern, "%?", ".?" )
	return string.find( name, pattern )
end

--------------------------------------------------------------------------------

local loggersByName = function( name )
	
	if weakLoggers[ name ] then
		return weakLoggers[ name ]
	end
	
	local logger = WFLoggerClass( name )
	
	if name then
		local sortedLoggers = {}
		
		for loggerName in pairs( weakLoggers ) do
			table.insert( sortedLoggers, loggerName )
		end
		
		table.sort( sortedLoggers, function( name1, name2 )
			return #name1 > #name2
		end )

		-- check for a wildcard configuration and copy its settings
		for i,loggerName in ipairs( sortedLoggers ) do
			local matchedLogger = weakLoggers[ loggerName ]

			if matchedLogger and loggerName ~= name and globMatch( name, loggerName ) then
				logger._priority = matchedLogger._priority
				logger._actions = matchedLogger._actions
				break
			end
		end
	end
	
	weakLoggers[ name ] = logger
	
	return logger
	
end

--------------------------------------------------------------------------------

-- 3.28.12 kjt - Configured loggers used to be stored in _B in Lightroom; need to figure out
-- what problems may be caused by storing them locally; can they get GC'd? Are
-- we doing too much work this way?
--	-- Use a strong table so that configured loggers don't get GC'd
--	-- also, store in _B so that it doesn't get GC'd: Can't use a local here!
--	assert( _B )
--	_B.configuredLoggers = {}

local configuredLoggers = {}

local function configureLogger( name, logConfiguration )
	
	local logger = loggersByName( name )
	configuredLoggers[ name ] = logger
	
	logger:configure( logConfiguration )

end

--------------------------------------------------------------------------------

local function setupDefaultLogger()
	if weakLoggers[ "*" ] then
		return
	end
	
	configureLogger( "*", {
		logLevel = _G.DEBUG and "info" or "error"
	} )
end

--------------------------------------------------------------------------------

local function configureLoggers( loggers )

	for name, desc in pairs( loggers ) do
		if type( desc ) == 'string' then
			desc = {
				logLevel = desc
			}
		end
		configureLogger( name, desc )
	end
	
	setupDefaultLogger()
end

if next( WFConfiguration.loggers ) then
	configureLoggers( WFConfiguration.loggers )
end

--------------------------------------------------------------------------------
 
local function WFLogger( name )
	if name == 'reportMemoryUsage' then
		local count = 0
		for k in pairs( weakLoggers ) do
			count = count + 1
		end
		WFLoggerImp.alwaysPrint( string.format( "WFLogger: %d active loggers", count ))
		return
	end
	
	local logger = weakLoggers[ name or "(default)" ]
	if not logger then
		logger = loggersByName( name or "(default)" )
	end

	return logger
end

--==============================================================================

return WFLogger


