--[[----------------------------------------------------------------------------


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFURI = require 'WFURI'
local KSPromises = require 'KSPromises'
local WFZString = require 'WFZString'
local KSPathUtils = require 'KSPathUtils'
local ZSTR = assert( WFZString.fromZString )
local KSClass = require 'KSClass'
local KSModalDialogFactories -- = require "KSModalDialogFactories"
local WFCloseSupport = require "WFCloseSupport"

-- Requires
local KSOAuth2Impl = require "KSOAuth2Impl"

	-- core.logging
local KSLogging = require 'KSLogging'

	-- libs.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- oz_client.bablu
local Bablu = require 'Bablu'

	-- os.localization
local WFLocalizationUtils = require 'WFLocalizationUtils'

local logger = KSLogging.log 'WFLoginProcess'
local tinsert = table.insert

local IMS_LOGIN_REDIRECT_URI = "air-auth://carousel";

--==============================================================================

local function readFileToString( path )

	local f = io.open( path, "rb" )
	local text
	
	if f then
		text = f:read( "*all" )
	
		f:close()
	end
	
	return text

end

--------------------------------------------------------------------------------

-- List of entities to escape
-- note if %, ], ^ or other special regex chars need to be escaped
-- in the future, the simple code below that generates sHtmlPattern
-- will need to be updated.
local sHtmlEntities = {
	[ '&' ] = "&amp;",
	[ '"' ] = "&quot;",
				-- " (This line is here to fix Halide syntax coloring)
	[ '\'' ] = "&#x27;",
	[ '>' ] = "&gt;",
	[ '<' ] = "&lt;",
}

local sHtmlPattern = {}

for k in pairs( sHtmlEntities ) do
	table.insert( sHtmlPattern, k )
end

sHtmlPattern = "[" .. table.concat( sHtmlPattern ) .. "]"

local function escapeHtmlEntities( translatedString )
	
	return string.gsub( translatedString, sHtmlPattern, function( capture )
		return sHtmlEntities[ capture ] or capture
	end )
	
end

--------------------------------------------------------------------------------

local function replaceZStrs( fileContents )

	return string.gsub( fileContents, 'ZSTR%("(%$%$%$[^"]*)"%)', function( capture )
				-- " (This line is here to fix Halide syntax coloring)
		return escapeHtmlEntities( ZSTR( capture ) )
	end )
	
end

--------------------------------------------------------------------------------

local function getResourceFilePath( resourceName )
	
	local resources = KSPathUtils.getStandardFilePath "Resources"
	return KSPathUtils.child( resources, resourceName )
	
end

--------------------------------------------------------------------------------

local function appendCustomJavascriptP( self, page, parsedUrl )
	
    local clientIdAndSecretP = Bablu:getImsClientIdAndSecretP()

	return clientIdAndSecretP:mapDataP( function( status, client_id, client_secret )
		
		local stringBuffer = { page }
		
		tinsert( stringBuffer, '<script type="text/javascript">' )
		
		local language_code = WFLocalizationUtils.getLanguageCode()
		local region_code = WFLocalizationUtils.getRegionCode()
		
		-- Set the query parameters
		local decoratedImsUrl = function( action )
			-- Base URL (scheme/host/path) with action-specific param and decorated with ims
			return string.format( "%s%s?client_id=%s%s&locale=%s_%s", self._imsLoginUrl, action, client_id, "&redirect_uri=air-auth://carousel", language_code, region_code )
		end
		
		-- AdobeId scoped URL (order of multiple scope values does not matter)
		local decoratedImsUrl_adobeIdScope = function( action )
			return string.format( "%s%s", decoratedImsUrl( action ), "&scope=openid,AdobeID" )
		end
		
		local function addCall( method, value )
			tinsert( stringBuffer, string.format( [[if( window.%s ) { %s( "%s" ); }]], method, method, value ) )
		end
		
		-- Facebook scoped URL (order of multiple scope values does not matter)
		addCall( "setFacebookLoginUrl", decoratedImsUrl( "/authorize" ) .. "&scope=openid,facebook&authentication_service=facebook" )
			
		-- Google scoped URL (order of multiple scope values does not matter)
		addCall( "setGoogleLoginUrl", decoratedImsUrl( "/authorize" ) .. "&scope=openid,Google&authentication_service=google" )
			
		-- Local Adobe login HTML page
		addCall( "setAdobeLoginUrl", "resource:///ims_workflow_adobe.html?authentication_service=adobe" )

		-- Enhanced Entry Point: Create Account
		addCall( "setAdobeCreateAccountUrl", decoratedImsUrl_adobeIdScope( "/authorize" ) ..  "&idp_flow=create_account" )

		-- Local Adobe login HTML page
		addCall( "setBackButtonUrl", "resource:///ims_workflow_adobe.html?" )
			
		-- Enhanced Entry Point: Reset Password
		addCall( "setAdobeResetPasswordFormSubmitUrl", decoratedImsUrl_adobeIdScope( "/send/v1/password_reset" ) )

		-- Pre-populate the User ID field, if a value for it is available
		if self._lastKnownUserId then
			addCall( "setUserId", self._lastKnownUserId )
		end
		
		-- Local Adobe reset password HTML page
		addCall( "setAdobeResetPasswordLinkUrl", "resource:///ims_workflow_adobe_reset.html?" )
			
		-- Enhanced Entry Point: Login
		local eepLoginUrl = decoratedImsUrl_adobeIdScope "/login/code"
		eepLoginUrl = string.format( "%s&client_secret=%s", eepLoginUrl, client_secret )
		addCall( "setAdobeLoginFormSubmitUrl", eepLoginUrl )
		
		tinsert( stringBuffer, '</script>' )
		
		return 'data', table.concat( stringBuffer, "\n" ), parsedUrl
		
	end )

end

--------------------------------------------------------------------------------

-- Take an url such as
-- file://localhost/Volumes/src/perforce/yorktown_1700/wf_win_rt/build/Debug/Adobe%20Revel.app/Contents/Resources/ims_workflow.html
-- and return a string with the ZSTR expressions replaced, and the parsed url
-- Note: the file contents must be UTF8 encoded.
local function translatePageP( self, url )
	
	local parsed = WFURI.parse( url )
	local scheme = parsed.scheme
	local path = parsed.path
	
	if scheme == 'resource' then
		-- find the path to this resource in our app package
		path = getResourceFilePath( path )
		scheme = 'file'
	end
	
	if scheme == 'file' then
		
		local contents = readFileToString( path )
		
		if contents then
			local translated = replaceZStrs( contents )
			return appendCustomJavascriptP( self, translated, parsed )
		end
	
	end
	
	return KSPromises.resolvedPromiseP( "error" )
	
	-- else return nil, unable to translate that url
	
end

--==============================================================================

local WFLoginProcess = {
	
	_workflow = "ims_workflow",
	
	_imsLoginUrl = false,
	
	_lastKnownUserId = false,
		
}

--------------------------------------------------------------------------------

function WFLoginProcess:__init( imsLoginUrl )

	self._imsLoginUrl = imsLoginUrl

end

--------------------------------------------------------------------------------

function WFLoginProcess:reloadUrl( )

	return self._workflow or "ims_workflow"

end

--------------------------------------------------------------------------------

function WFLoginProcess:translatePageP( newUrl )

	if string.find( newUrl, "^ims_" ) then
		-- Make this into a resource URL
		newUrl = string.format( "resource:///%s.html", newUrl )
	end
	
	return translatePageP( self, newUrl ):mapP( function( status, translatedHtmlContent, parsedUrl )
	
		if translatedHtmlContent then
			
			local _, _, wkflow = string.find( parsedUrl.path, "(ims_[a-z_]+)" )
			if wkflow then
				self._workflow = wkflow
			else
				self._workflow = "ims_workflow"
			end
			
			local _, _, filename = string.find( parsedUrl.path, "[/]([^/]+)" )
			return 'data', 'html', translatedHtmlContent, filename
		
		else
			
			self._workflow = "ims_workflow"
			return 'data', 'url', self._imsLoginUrl
		
		end
		
	end )

end

--------------------------------------------------------------------------------

local function decodeQuery( url )
	local parsed = WFURI.parse( url )
	
	local query = parsed.query or ""
	
	local t = {}
	
	for kvpair in string.gmatch( query, "[^&]+" ) do
		local _, _, k, v = string.find( kvpair, "^([^=]+)=(.+)$" )
		if k and v then
			if t[ k ] then
				logger:warn( "Duplicate key in url Query string: %s", k )
			end
			t[ k ] = v
		end
	end
	
	return t, parsed
	
end

--------------------------------------------------------------------------------

local function storeUserId( self, queryParams )

	-- Forward along the UserId
	local lastKnownUserId = queryParams.lastKnownUserId
	if lastKnownUserId then
		
		self._lastKnownUserId = WFURI.unescape( lastKnownUserId )
	
	end
	
end

--------------------------------------------------------------------------------
-- returns a promise that specifies what to do:
-- it may be already resolved when it returns.
-- if it is not resolved, you should disable user interaction until it does and start
-- a timer.  Stop the timer when it resolves.  Re-enable user interaction if the resolved
-- value indicates you should.
-- it resolves to tuple of these values:
-- P<
--		status -- data
--		enableUserInteraction (boolean)  -- if true, re-enable user interaction in the web view
--		action ('html', 'url', 'javascript', 'cancel', 'failure')
--			'nothing' means do nothing
--			'html' means display this HTML
--				... = the html
--			'url' means navigate to this url
--			'javascript' means run this javascript in the browser DOM
--			'cancel' means close the login process now
--			'failure' means an error occurred; display a message indicating login failed
--				...		currently unused but it'd be nice if we could indicate why action=failure with
--						more info here.
--			'success' means an error occurred; display a message indicating login failed
--				... = accessToken and refreshToken
--		...

function WFLoginProcess:webViewDidStartLoadP( url, loaded )

	-- Get query string components
	local queryParams, parsedUrl = decodeQuery( url )
	local stopSpinner = false
	local enableUserInteraction = true
	
	storeUserId( self, queryParams )
	
	local function allResults( ... )
		return queryParams.authentication_service, enableUserInteraction, stopSpinner, ...
	end
	
	if parsedUrl.scheme == 'resource' then
		return self:translatePageP( url )
			:mapDataP( function( _, ... )
				return 'data', allResults( ... )
			end )
	end
	
	-- See if it the URL starts with our special URI "air-auth://carousel"
	-- If NOT, we are done here

	if string.find( url, IMS_LOGIN_REDIRECT_URI, 1, true ) ~= 1 then
		return KSPromises.resolvedPromiseP( 'data', allResults( 'nothing' ) )
	end
	
	enableUserInteraction = false
	
	local code = queryParams.code
	local eep_login_response = queryParams.eep_login_response
	local eep_reset_password_response = queryParams.eep_reset_password_response


	if not code and not eep_login_response and not eep_reset_password_response then
	
		local errorReason = queryParams.error_reason
		
		if errorReason == "user_denied" then
			
			-- User-requested cancel, as reported by IMS in "error_reason" (currently only in third-party workflow).
			-- Silently dismiss the sign in window.
			
			return KSPromises.resolvedPromiseP( 'data', allResults 'cancel' )
			
		else
		
			logger:error( "sign in error: didn't receive initial auth code or Enhanced Entry Point response from IMS" )
			
			return KSPromises.resolvedPromiseP( 'data', allResults 'failure' )
			
		end
	
	elseif eep_login_response then

		-- IMS Enhanced Entry Point login workflow
		-- The api returns a JSON object containing, among other things, an access token and a refresh token
    
		local accessToken, refreshToken
		
		eep_login_response = WFURI.unescape( eep_login_response )

		local deserializedData = KSJsonDecoder.decodeWebSafeJson_stripNull( eep_login_response )
		if deserializedData then
	
			accessToken = deserializedData.access_token
			refreshToken = deserializedData.refresh_token
	
		end

		if accessToken and refreshToken then
        
			return KSPromises.resolvedPromiseP( 'data', allResults( 'success', accessToken, refreshToken ) )
        
		else

			-- Likely IMS has reported an error, and it might include a jump URL
			
			enableUserInteraction = true
			
			stopSpinner = true
			
			local errorCode = deserializedData.error
			if errorCode then
			
				-- See if IMS has specified a "jump URL"...if so, navigate to it
				-- This is to have the user perform additional actions for their account (like update the TOU, etc)
				
				local jumpUrl = deserializedData.jump
			
				if jumpUrl then
					
					return KSPromises.resolvedPromiseP( 'data', allResults( 'url', jumpUrl ) )
				
				else
					
					return KSPromises.resolvedPromiseP(
						'data', allResults( 'javascript',
						string.format( [[set_status( "%s" )]], errorCode ) ) )
					
				end
				
			else
				
				-- No error info sent...bail out.
				logger:error( "sign in error: didn't receive an access token and/or refresh tokn and/or error code from IMS in Enhanced Entry Point login workflow" )
				
				return KSPromises.resolvedPromiseP( 'data', allResults 'failure' )
				
			end
			
		end
	
	elseif eep_reset_password_response then
    
		-- IMS Enhanced Entry Point reset password workflow
		-- The api returns a JSON object containing a reset password status
    
		local status = nil
	
		eep_reset_password_response = WFURI.unescape( eep_reset_password_response )
	
		local deserializedData = KSJsonDecoder.decodeWebSafeJson_stripNull( eep_reset_password_response )
		
		if deserializedData then
	        
			status = deserializedData.status
	
		end
	
		enableUserInteraction = true
	
		if status then
	
			-- Responses are EMAIL_SENT and UNKNOWN
	        
			if status == "EMAIL_SENT" then
				
				-- Forward along the UserId
					local userId = queryParams.lastKnownUserId

				if userId then
				
					self._lastKnownUserId = WFURI.unescape( userId )
				
				end
				
				return self:translatePageP( "resource:///ims_workflow_adobe_recovery.html" )
				
			else
				
				stopSpinner = true
				-- Likely: UNKNOWN
				return KSPromises.resolvedPromiseP(
					'data', allResults( 'javascript',
					string.format( [[set_status( "%s" )]], status ) ) )
	
			end
	        
		else
	        
			stopSpinner = true
			logger:error( "sign in error: didn't receive a status from IMS in Enhanced Entry Point reset password workflow" )
			return KSPromises.resolvedPromiseP(
				'data', allResults( 'javascript',
				[[set_status( "" )]] ) )	-- Will cause a generic error
	        
		end
	    
	else
	
		-- FIXME [scouten 2011-08-04]: Should be obtained from the config API
		-- that Alex is going to provide, but (a) we don't have it yet, and (b)
		-- I don't have an easy way to pass this info down from Lua to ObjC.
		-- Fielding is killing lots of kittens today.
		
		-- Grab the refresh and access tokens
	
		local imsBaseUrl = self._imsLoginUrl
		
		local fetchP = KSOAuth2Impl.fetchImsTokensP( code, imsBaseUrl )
		
		return fetchP:mapP( function( status, data, headers )
			if status == 'data' and data.access_token and data.refresh_token then
				local accessToken = data.access_token
				local refreshToken = data.refresh_token
				return KSPromises.resolvedPromiseP( 'data', allResults( 'success', accessToken, refreshToken ) )
			else
				return KSPromises.resolvedPromiseP( 'data', allResults( 'failure', status == 'error' and data ) )
			end
		end )
	
	end
	
	return KSPromises.resolvedPromiseP( 'data', allResults( 'nothing' ) )

end

--------------------------------------------------------------------------------

function WFLoginProcess:webViewDidFinishLoad( ksview )

	-- After a page is loaded, we find all external links (those with target='_blank')
	-- and change the protocol to newwindow: and escape the original URL.
	-- Later, if the user clicks on one of these links, we can detect this protocol
	-- and open the link in the system web browser, rather than in-page.
	--
	-- Reference:
	-- http://www.icab.de/blog/2009/07/27/webkit-on-the-iphone-part-1/
	-- http://www.icab.de/blog/2009/08/05/webkit-on-the-iphone-part-2/
	
	local modifyLinkTargetsJS = [[
		function Carousel_ModifyLinkTargets() {
			var allLinks = document.getElementsByTagName('a')
			if (allLinks) {
				var i;
				for (i=0; i<allLinks.length; i++) {
					var link = allLinks[i];
					var target = link.getAttribute('target')
					if (target && target == '_blank') {
						link.setAttribute('target','_self')
						link.href = 'newwindow:'+escape(link.href)
					}
				}
			}
		}
		Carousel_ModifyLinkTargets()]]
	
	local nativeViewController = ksview:getNativeViewController()
	nativeViewController:runJavascriptWithWebView( modifyLinkTargetsJS )

end

--------------------------------------------------------------------------------

function WFLoginProcess:webViewDidFailLoadWithError( ksview, message, details )

	if details then
		message = string.format( "%s\n%s", message, details )
	end

	-- SPAM [arahn 2012-05-29]: We shouldn't have to compare against a localized string
	-- to understand which button was pressed.
	local hack_RetryLabel = ZSTR "$$$/Login/AlertView/RetryButton=Retry"
	
	KSModalDialogFactories.createAndShowAlertView( ksview,
	{
		title = ZSTR "$$$/Login/AlertView/LoginErrorTitle=Sign in error",
		message = message,
		cancelButtonTitle = ZSTR "$$$/Login/AlertView/CancelButton=Cancel",
		otherButtonTitles = hack_RetryLabel,
		dismissedWithButton = function( self, button )
			KSModalDialogFactories.clearAlertViewTree( self )
			
			if button ~= hack_RetryLabel then
				ksview:loginFailed()
			else
				ksview:reload()
			end
		end
	} )

end

--------------------------------------------------------------------------------

WFCloseSupport.addCloseSupportMethods( WFLoginProcess )

return KSClass "WFLoginProcess" ( WFLoginProcess )

