--[[----------------------------------------------------------------------------

WFLuaModelsImp.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@author mhamburg 2012-10-10

------------------------------------------------------------------------------]]

local assert = assert
local type = type
local tostring = tostring

local KSReactive = require 'KSReactive'
local KSReactive_isBehavior = assert( KSReactive.isBehavior )

local KSPromises = require 'KSPromises'
local KSPromises_isPromise = assert( KSPromises.isPromise )

local logger = require 'WFLogger' ( 'WFModels' )

--============================================================================--

local WFLuaModelsImp = { }

--============================================================================--

local function makePromiseBasedModel( promise, name )

	local action
		-- Note that this can get set after we've closed if the promise
		-- is already resolved, but that doesn't matter because at that
		-- point the action is minimal in what it references and this
		-- variable won't stay alive much longer.
	
	return {
	
		_bindToModelStub = function( self, stub )
		
			action = promise:whenResolvedDo( function( status, ... )
				if status == 'data' then
					stub:_postData( ... )
				elseif status == 'error' then
					stub:_postError( tostring( ( ... ) ) )
				end
				stub:_postDone()
				stub:_closeModelStub()
			end )
			
			action:attachTo( nil, self )
				-- nothing with close support to close the action but
				-- we manage its lifetime from the model
		
		end,
		
		_closeLuaModel = function( self )
			if action then
				action:close()
				action = nil
			end
		end,
		
		_receiveEvent = function( self, selector, ... )
		
			if name then
				logger:warnf( "No event support in promise-based model %s. Received %s.",
					name, tostring( selector ) )
			else
				logger:warnf( "No event support in promise-based models. Received %s.",
					tostring( selector ) )
			end
		
		end,
		
		_openModel = function( self, selector, ... )
		
			if name then
				error( string.format(
					"No submodel support for promise-based model %s: %s",
					name, tostring( selector ) ) )
			else
				error( "No submodel support for promise-based models: " .. tostring( selector ) )
			end
		
		end,
	
	}

end

--============================================================================--

local function postCurrentValueToStub( b, stub )

	local isError, message = b:isErrorNow()
	
	if isError then
		stub:_postError( tostring( message ) )
	else
		stub:_postData( b:valueNow() )
	end

end

--------------------------------------------------------------------------------

local function callCloseMethod( x )
	local method = x.close
	if method then
		method( x )
	end
end

--------------------------------------------------------------------------------

local function makeParamsBasedModel( params, name )

	local valueB = params.valueB
	assert( KSReactive_isBehavior( valueB ), "valueB should be a behavior" )
	
	local events = params.events
	assert( not events or type( events ) == 'table', "events should be a table" )
	
	local models = params.models
	assert( not models or type( models ) == 'table', "models should be a table" )
	
	valueB:collectCloseTargets()
		-- Closing valueB should close everything that feeds it

	local onValues
	local onDone
	
	return {
	
		_bindToModelStub = function( self, stub )
		
			-- What we do when done depends on whether we can receive requests
			-- for events and models 	
				
			local onDoneFn
			
			if events or models then
			
				onDoneFn = function()
					stub:_postDone()
				end
				
			else
			
				onDoneFn = function()
					stub:_postDone()
					stub:_closeModelStub()
					valueB:close()
				end
			
			end
			
			-- Send the value to the stub
	
			postCurrentValueToStub( valueB, stub )
			
			-- Wire up for any future activity
		
			if valueB:isDone() then
			
				onDoneFn()
	
			else
				
				onValues = valueB:rawMapE( function( status, ... )
					if status == 'data' then
						stub:_postData( ... )
					elseif status == 'error' then
						stub:_postError( tostring( ( ... ) ) )
					end
				end )
				
				onDone = valueB:onDoneE( onDoneFn )
				
			end
		
		end,
		
		_closeLuaModel = function( self )
			-- It would be good to close any event sources spawned from here.
			valueB:close()
			if onValues then onValues:close() end
			if onDone then onDone:close() end
			events = nil
			models = nil
			pcall( callCloseMethod, params )
		end,
		
		_receiveEvent = function( self, selector, ... )
		
			local handler = events and events[ selector ]
			
			if handler then
				handler( params, ... )
			else
				if name then
					logger:warnf(
						"No event handler for model %s for event: %s",
						name, tostring( selector ) )
				else
					logger:warnf(
						"No event handler for event: %s",
						tostring( selector ) )
				end
			end
		
		end,
		
		_openModel = function( self, selector, ... )
		
			local handler = models and models[ selector ]
			
			if not handler then
				error( "No submodel support for selector " .. tostring( selector ) )
			end
			
			return handler( params, ... )
		
		end,
	
	}

end

--------------------------------------------------------------------------------

function WFLuaModelsImp.asModel( params, name )

	if name then
		name = tostring( name )
	end

	if KSPromises_isPromise( params ) then
		return makePromiseBasedModel( params, name )
	end
	
	if KSReactive_isBehavior( params ) then
		return makeParamsBasedModel( {
			valueB = params
		}, name )
	end
	
	return makeParamsBasedModel( params, name )

end

--============================================================================--

return WFLuaModelsImp

