--[[----------------------------------------------------------------------------

WFModelsTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local WFModelsTest = {}

--------------------------------------------------------------------------------

local assert = assert

local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'
local KSReactive = require 'KSReactive'
local WFIdle = require 'WFIdle'
local WFPlatformInfo = require 'WFPlatformInfo'

--------------------------------------------------------------------------------

local ModelClient

ModelClient = KSClass 'ModelClient' {

	_modelP = false,
	_model = false,
	
	_events = false,
	
	__init = function( self, modelP )
		self._modelP = modelP
		modelP:whenResolvedDo( function( status, model )
			if status ~= 'data' then return end
			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				model:close()
				return
			end
			self._model = model
			self:addCloseTarget( model )
			local events = self._events
			if events then
				self._events = false
				for i = 1, #events do
					local event = events[ i ]
					model:postEvent( unpack( event, 1, event.n ) )
				end
			end
		end ):attachTo( self )
	end,
	
	valueB = function( self, ... )
		if self._model then
			return self._model:valueB( ... )
		end
		local args = { ... }
		local n = select( '#', ... )
		local notDone = { }
		local modelB = KSReactivePromiseBridge.promiseToBehavior( self._modelP, notDone )
		return modelB:mapSwitchB( function( x )
			if x == notDone then
				return KSReactive.constantB( unpack( args, 1, n ) )
			else
				return x:valueB( unpack( args, 1, n ) )
			end
		end )
	end,
	
	postEvent = function( self, ... )
		if self._model then
			self._model:postEvent( ... )
		else
			local events = self._events
			if not events then
				events = { }
				self._events = events
			end
			events[ #events + 1 ] = { n = select( '#', ... ), ... }
		end
	end,
	
	openModel = function( self, ... )
		local args = { ... }
		local n = select( '#', ... )
		return ModelClient( self._modelP:mapDataP( function( _, model )
			--debug.print( "About to call through to submodel creation: ", unpack( args, 1, n ) )
			return 'data', model:openModel( unpack( args, 1, n ) )
		end ) )
	end,
	
	WFCloseSupport.addCloseSupportMethods

}

--------------------------------------------------------------------------------

local function messagingProcessorP( workerThread )

	return workerThread:threadIsReadyP():waitForP( WFIdle.asapP ):mapDataP( function()
		return 'data', workerThread._messagingProcessor
	end )

end

--------------------------------------------------------------------------------

local ModelFactory = KSClass 'ModelFactory' {

	_factoryP = false,
	
	_factory = false,

	__init = function( self, workerThread )
		local WFModels = require 'WFModels'
		self._factoryP = messagingProcessorP( workerThread ):mapDataP( function( status, messagingProcessor )
				return 'data', WFModels.openFactory( messagingProcessor )
			end )
		self._factoryP:whenResolvedDo( function( status, factory )
			if status ~= 'data' then return end
			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				factory:close()
				return
			end
			self._factory = factory
			self:addCloseTarget( factory )
		 end ):attachTo( self )
		self:addCloseTarget( workerThread )
	end,
	
	openModel = function( self, ... )
		local n = select( '#', ... )
		local args = { ... }
		return ModelClient( self._factoryP:mapDataP( function( _, factory )
			return 'data', factory:openModel( unpack( args, 1, n ) )
		end ) )
	end,

	WFCloseSupport.addCloseSupportMethods

}

--------------------------------------------------------------------------------

local simpleEcho_valueB

function WFModelsTest.testSimpleEcho( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	local factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	local model = factory:openModel( "WFModelsTest_Server.simpleTest", "foo" )
	
	local valueB = model:valueB( "baz" )
	
	simpleEcho_valueB = valueB
	
	-- valueB:onDataB( function( x ) debug.print( "valueB = ", x ) end )
	
	assert( valueB:valueNow() == "baz" )
	
	local function whenB( txt ) return function( fn )
		valueB:onDataB( function( x )
			if x == txt then fn() end
		end )
	end end
	
	whenB 'foo' ( function()
		model:postEvent( 'echo', 'floon' )
	end )
	
	whenB 'floon' ( function()
		model:postEvent( 'echoUpper', 'bleen' )
	end )
	
	whenB 'BLEEN' ( function()
		valueB:close()
		model:close()
		factory:close()
		simpleEcho_valueB = nil
		test:notifyTestDidSucceed()
	end )
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

local simpleEchoWithUpper_valueB

function WFModelsTest.testSimpleEchoWithUpper( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	local factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	local model = factory:openModel( "WFModelsTest_Server.simpleTest", "foo" )
	
	local submodel = model:openModel( "toUpper" )
	
	local valueB = submodel:valueB( "baz" )
	
	simpleEchoWithUpper_valueB = valueB
	
	-- valueB:onDataB( function( x ) debug.print( "valueB = ", x ) end )
	
	assert( valueB:valueNow() == "baz" )
	
	local function whenB( txt ) return function( fn )
		valueB:onDataB( function( x )
			if x == txt then fn() end
		end )
	end end
	
	whenB 'FOO' ( function()
		model:postEvent( 'echo', 'bleen' )
	end )
	
	whenB 'BLEEN' ( function()
		valueB:close()
		model:close()
		submodel:close()
		factory:close()
		simpleEchoWithUpper_valueB = nil
		test:notifyTestDidSucceed()
	end )
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

local submodels_testSubmodel

function WFModelsTest.testSubmodel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	submodels_testSubmodel = { }
	
	local M = submodels_testSubmodel
	
	M.factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	M.model = M.factory:openModel( "WFModelsTest_Server.counter" )
	
	M.valueB = M.model:valueB( "0" )
	
	M.submodel = M.model:openModel( "sub", "foo" )
	
	M.subvalueB = M.submodel:valueB( "baz" )
	
	M.onFoo = M.subvalueB:onDataB( function( x )
		if x == "foo" then
			M.submodel:postEvent( "echo", "bleen" )
		end
	end )
	
	M.onBleen = M.subvalueB:onDataB( function( x )
		if x == "bleen" then
			if M.valueB:valueNow() ~= "1" then
					-- This is a bit of a stretch to depend on since it's a
					-- separate model, but we've gone back and forth on values
					-- on another model from the same factory and what we want
					-- to see is that the count went up and then came down.
				test:notifyTestDidFail()
			end
			M.submodel:close()
			M.onZero = M.valueB:onDataB( function( y )
				if y == "0" then
					M = nil
					submodels_testSubmodel = nil
					test:notifyTestDidSucceed()
				end
			end )
		end
	end )
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

local submodels_testResolvedPromiseSubmodel

function WFModelsTest.testResolvedPromiseSubmodel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	submodels_testResolvedPromiseSubmodel = { }
	
	local M = submodels_testResolvedPromiseSubmodel
	
	M.factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	M.model = M.factory:openModel( "WFModelsTest_Server.counter" )
	
	M.valueB = M.model:valueB( "0" )
	
	M.submodel = M.model:openModel( "resolvedPromise", "foo" )
	
	M.subvalue = M.submodel:valueB( "baz" )
	
	M.subvalue:onDoneE( function()
		submodels_testResolvedPromiseSubmodel = nil
		if M.subvalue:valueNow() == "foo" then
			test:notifyTestDidSucceed()
		else
			test:notifyTestDidFail()
		end
	end )
	
	test:waitForTestToFinish( 2 )
	
end

--------------------------------------------------------------------------------

local submodels_testPromiseSubmodel

function WFModelsTest.testPromiseSubmodel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	submodels_testPromiseSubmodel = { }
	
	local M = submodels_testPromiseSubmodel
	
	M.factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	M.model = M.factory:openModel( "WFModelsTest_Server.counter" )
	
	M.valueB = M.model:valueB( "0" )
	
	M.submodel = M.model:openModel( "promise", "foo" )
	
	M.subvalue = M.submodel:valueB( "baz" )
	
	M.subvalue:onDoneE( function()
		submodels_testPromiseSubmodel = nil
		if M.subvalue:valueNow() == "foo" then
			test:notifyTestDidSucceed()
		else
			test:notifyTestDidFail()
		end
	end )
	
	test:waitForTestToFinish( 2 )
	
end

--------------------------------------------------------------------------------

local submodels_testConstantSubmodel

function WFModelsTest.testConstantSubmodel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	submodels_testConstantSubmodel = { }
	
	local M = submodels_testConstantSubmodel
	
	M.factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	M.model = M.factory:openModel( "WFModelsTest_Server.counter" )
	
	M.valueB = M.model:valueB( "0" )
	
	M.submodel = M.model:openModel( "subConstant", "foo" )
	
	M.subvalue = M.submodel:valueB( "baz" )
	
	M.subvalue:onDoneE( function()
		submodels_testConstantSubmodel = nil
		if M.subvalue:valueNow() == "foo" then
			test:notifyTestDidSucceed()
		else
			test:notifyTestDidFail()
		end
	end )
	
	test:waitForTestToFinish( 2 )
	
end

--------------------------------------------------------------------------------

local actions

function WFModelsTest.testCloseValueClosesModel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	actions = { }
	
	local factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	local model = factory:openModel( "WFModelsTest_Server.counter" )
	
	local valueB = model:valueB( "0" )
	
	local submodel = model:openModel( "sub", "foo" )
	
	local subvalueB = submodel:valueB( "baz" )
	
	actions.onFoo = subvalueB:onDataB( function( x )
		if x == "foo" then
			submodel:postEvent( "echo", "bleen" )
		end
	end )
	
	actions.onBleen = subvalueB:onDataB( function( x )
		if x == "bleen" then
			if valueB:valueNow() ~= "1" then
					-- This is a bit of a stretch to depend on since it's a
					-- separate model, but we've gone back and forth on values
					-- on another model from the same factory and what we want
					-- to see is that the count went up and then came down.
				test:notifyTestDidFail()
			end
			subvalueB:close()
			actions = { }
			actions.onZero = valueB:onDataB( function( y )
				if y == "0" then
					actions = nil
					test:notifyTestDidSucceed()
				end
			end )
		end
	end )
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

local actions2

function WFModelsTest.testGCClosesModel( test )

	local WFWorkerThread = require 'WFWorkerThread'
	
	actions2 = { }
	
	local factory = ModelFactory( WFWorkerThread( "WFModelsTest_Server.lua" ) )
	
	local model = factory:openModel( "WFModelsTest_Server.counter" )
	
	local valueB = model:valueB( "0" )
	
	local submodel = model:openModel( "sub", "foo" )
	
	model = nil
	
	local subvalueB = submodel:valueB( "baz" )
	
	actions2.onFoo = subvalueB:onDataB( function( x )
		if x == "foo" then
			submodel:postEvent( "echo", "bleen" )
			submodel = nil
			actions2.onFoo = nil
			WFIdle.doLater( function()
				collectgarbage()
				collectgarbage()
				collectgarbage()
			end )
		end
	end )
	
	actions2.onBleen = subvalueB:onDataB( function( x )
		if x == "bleen" then
			if valueB:valueNow() ~= "1" then
					-- This is a bit of a stretch to depend on since it's a
					-- separate model, but we've gone back and forth on values
					-- on another model from the same factory and what we want
					-- to see is that the count went up and then came down.
				test:notifyTestDidFail()
			end
			subvalueB = nil
			actions2 = { }
			actions2.onZero = valueB:onDataB( function( y )
				if y == "0" then
					actions2 = nil
					test:notifyTestDidSucceed()
				end
			end )
			WFIdle.doLater( function()
				collectgarbage()
				collectgarbage()
				collectgarbage()
			end )
		end
	end )
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

if  WFPlatformInfo.isAndroid then
	return WFModelsTest and {}
else
	return WFModelsTest
end
