--[[----------------------------------------------------------------------------

WFOzDownloadTask.lua

@author scouten 2012-04-23

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This namespace is used exclusively by OzCentralSyncAgent. It provides code to
download the latest information for a catalog from Oz.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFIdle = require 'WFIdle'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafondation.core.os
local KSFileUtils = require 'KSFileUtils'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local WFTimer = require "WFTimer"
local WFURI = require "WFURI"

	-- wichitafoundation.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'
local OzCatHelperFns = require 'OzCatHelperFns'

	-- wichitafoundation.web.json
local KSJsonDecoder = require 'KSJsonDecoder'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local downloadLog = KSLogging.log 'WFOzDownloadTask'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark WFOzDownloadTask namespace

--[[!
	\class 	WFOzDownloadTask
	\brief This namespace provides the ability to manage download tasks
		for the Oz server.

]]--

local WFOzDownloadTask = {}

--------------------------------------------------------------------------------

local receivedDocumentsByType = {}
local syncHistory = WFDiagnosticLog.makeRingBuffer {
	accumulate = true,
	mapFn = function( updates )
		local deletions, assets = 0, 0
		for i, update in ipairs( updates ) do
			if update.deleted then
				deletions = deletions + 1
			elseif update.type == 'asset' then
				assets = assets + 1
			end
		end
		return #updates, deletions, assets
	end,
	"updates", "deletions", "assets"
}
--------------------------------------------------------------------------------

local updateHistoricDiagnosticData = function( updates )
	syncHistory:add( updates )
end

--------------------------------------------------------------------------------

local function buildSyncHistorySparkline()
	return {
		syncHistory:visualize( 'sparkline' ),
	}
end

--------------------------------------------------------------------------------

local function purgeAlbumAssetDocumentsP( args, whereClause, id )

	local indexStore = args.indexStore
	
	local queryBase = string.format( 'SELECT docId FROM album_asset %s', whereClause )
	
	local resultP = indexStore:queryP( queryBase, id )
		:mapDataP( function( _, rows )
				
			local function deleteAlbumAssetDocument( row )
			
				local albumAssetId = row and row[ 1 ]
							
				return args.docStore:setServerValueP( 'album_asset', albumAssetId, { _deleted = true } )
			
			end
			
			return KSPromises.allP( deleteAlbumAssetDocument, rows, 8 )
			
		end )
		
	return resultP

end

--------------------------------------------------------------------------------

local function purgeSpaceAlbumDocumentsP( args, whereClause, id )

	local indexStore = args.indexStore
	
	local queryBase = string.format( 'SELECT docId FROM space_album %s', whereClause )
	
	local resultP = indexStore:queryP( queryBase, id )
		:mapDataP( function( _, rows )
				
			local function deleteSpaceAlbumDocument( row )
			
				local spaceAlbumId = row and row[ 1 ]
							
				return args.docStore:setServerValueP( 'space_album', spaceAlbumId, { _deleted = true } )
			
			end
			
			return KSPromises.allP( deleteSpaceAlbumDocument, rows )
			
		end )
		
	return resultP

end

--------------------------------------------------------------------------------

local catalogRootInfoCache = {}

-------------------------------------------------------------------------------

local function updateCatalogRootInfoP( args )

	local catalogUrl = assert( args.catalogUrl )
	
	-- Cache catalog info for up to 5 seconds to avoid swamping first sync with repeated
	-- fetches of catalog info

	local cachedP = catalogRootInfoCache[ catalogUrl ]
	if catalogRootInfoCache[ catalogUrl ] and not cachedP:isCanceled() then
		return cachedP
	end
	
	cachedP = args.ozUser:getP( catalogUrl, { compositeErrorData = true } )
		:mapP( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP )
		:mapDataP( function( _, catalogInfo )

			if not catalogInfo or not next( catalogInfo ) then return end

			args.accessErrorB:set( false )
			args.serviceAvailableB:set( true )

			local catalogUsersUrl = string.format( '%s/users', catalogUrl )

			return args.ozUser:getP( catalogUsersUrl, { compositeErrorData = true } )
				:mapP( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP )
				:mapDataP( function( _, catalogUsers )

					if not catalogUsers or not next( catalogUsers ) then return end

					-- FIXME [scouten 2012-04-17]: Split catalog users out into a separate
					-- document?

					catalogInfo.catalogUsers = catalogUsers
					
					if not catalogUsers or #catalogUsers == 0 then
						downloadLog:warn( "attempt to write empty users block! catalogInfo = %s", catalogInfo )
						return
					end
					
					downloadLog:debug( "Setting catalog root info to %s", catalogInfo )

					return args.docStore:setServerValueP( 'catalog', args.catalogId,
														  catalogInfo  )

				end )

			:mapDataP( function()

				return 'data', catalogInfo
			
			end )
		
		end )
		:mapP( function( status )
				
			if status == 'data' then
				
				WFTimer.afterDelayP( 5 )
					:whenResolvedDo( function()
						catalogRootInfoCache[ catalogUrl ] = nil
					end )
					:runUntilDone()
			
			else
			
				catalogRootInfoCache[ catalogUrl ] = nil

			end
		
		end )
	
	catalogRootInfoCache[ catalogUrl ] = cachedP
	
	return cachedP
end

--------------------------------------------------------------------------------

local function receiveOneAssetP( docType, args, docBody )

	local docId = docBody and docBody._id

	if not docId then return end
	
	if not docBody.deleted then
	
		downloadLog:debug( "receiveOneAssetP %s", docBody )
	
		local payload = docBody.payload or {}
		
		-- Invert the relationship between payload and server metadata.
		
		payload._serverProvidedMetadata = docBody
		payload._serverProvidedMetadata.payload = nil
		payload._serverProvidedMetadata.id = nil
		payload._serverProvidedMetadata._id = nil
	
		-- Some APIs contain a base member, some do not. This mucks up the
		-- logic in setServerValue. Let's hope (pray?) that we don't have other
		-- fields that come and go like this.
	
		payload._serverProvidedMetadata.base = nil
	
		downloadLog:debug( "receiveOneAssetP :: setServerValueP %s",
						   payload )
	
		return args.docStore:setServerValueP( 'asset', docId, payload )

	else

		-- File deleted from Oz: Remove original file from cache if present.
		
		return args.docStore:docValueOnceP( 'asset', docId )
			:mapDataP( function( _, docValue )

				local originalFilePath = docValue
					and docValue._localOnly
					and docValue._localOnly.originalFilePath
			
				-- Temporary check to correct the local file path and replace
				-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
				-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
				-- that updates the database with the correct file paths but this is not
				-- feasible in the current time frame.
				
				originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
				
				if originalFilePath then
					KSFileUtils.moveToTrash( originalFilePath )
				end
				
				local latestOriginalFilePath = docValue
									and docValue._localOnly
									and docValue._localOnly.latestOriginal
									and docValue._localOnly.latestOriginal.filePath
									
				latestOriginalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( latestOriginalFilePath )
				
				if latestOriginalFilePath then
					KSFileUtils.moveToTrash( latestOriginalFilePath )
				end
				
				local proxyFilePath = docValue
								  and docValue._localOnly
								  and docValue._localOnly.proxy
								  and docValue._localOnly.proxy.filePath
								
				proxyFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( proxyFilePath )
				
				if proxyFilePath then
					KSFileUtils.delete( proxyFilePath )
				end
			
				-- Now mark the asset as deleted in local DB.
				
				return args.docStore:setServerValueP( 'asset', docId, { _deleted = true } )
					:mapDataP( function( _, _ )
						return purgeAlbumAssetDocumentsP( args, 'WHERE assetId == ?', docId )
					end )

			end )

	end

end

--------------------------------------------------------------------------------

local function receiveOneGenericDocumentP( docType, args, docBody )

	local docId = docBody and docBody._id

	if not docId then return end
	if not docType then return end
	
	if not docBody.deleted then
	
		downloadLog:debug( "receiveOneGenericDocumentP %s", docBody )
	
		local payload = docBody or {}
				
		payload._id = nil
		payload.base = nil
	
		-- Some APIs contain a base member, some do not. This mucks up the
		-- logic in setServerValue. Let's hope (pray?) that we don't have other
		-- fields that come and go like this.
		
		downloadLog:debug( "receiveOneGenericDocumentP :: setServerValueP %s",
						   payload )
	
		return args.docStore:setServerValueP( docType, docId, payload )

	else

		local resultP = args.docStore:setServerValueP( docType, docId, { _deleted = true } )

		if docType == 'album' then
			-- Album deleted from Oz: Mark the album as deleted in local DB.
			-- Also delete any album_asset documents for this album.

			resultP = resultP
				:mapDataP( function( _, _ )
					return purgeAlbumAssetDocumentsP( args, 'WHERE albumId == ?', docId )
				end )
		elseif docType == 'space' then
			-- Space deleted from Oz
			-- Also delete any space_album documents for this space.

			resultP = resultP
				:mapDataP( function( _, _ )
					return purgeSpaceAlbumDocumentsP( args, 'WHERE spaceId == ?', docId )
				end )
		end

		return resultP

	end

end


--------------------------------------------------------------------------------

local receiveFunctionsForDocTypes = {

	asset = receiveOneAssetP,

}

--------------------------------------------------------------------------------
--[[!
	\brief	Reloads a document from the Oz database.

	\param args (table) A table that contains these keys:
		<ul><li>docStore</li>
		    <li>ozUser</li>
		    <li>catalogUrl</li></ul>
	\param docType (string) The document type.
	\param docId (string) The document's unique ID.
	\param docUrl (string) The document's URL.

	\return <code>P < table ></code> A Promise that resolves to the reloaded document.

]]--
function WFOzDownloadTask.reloadDocumentP( args, docType, docId, docUrl )

	-- args contains:
		-- docStore = args.docStore,
		-- ozUser = args.ozUser,
		-- catalogUrl = args.catalogUrl,

	assert( args.ozUser )
	assert( docType )
	assert( docId )
	assert( docUrl )
	
	local resultP = args.ozUser:getP( docUrl )
		:mapP( function( status, data, headers )

			if status ~= 'data' then
				downloadLog:debug( "reloadDocumentP for %s failed\n\n%s\n\n%s",
								   docUrl, headers, data )
				return status, data, headers
			end

			downloadLog:debug( "received server metadata reload from %s\n\ndata = %s",
							   docUrl, data )
	
			local docBody = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )

			receivedDocumentsByType[ docType ] = ( receivedDocumentsByType[ docType ] or 0 ) + 1

			local receiveFunction = receiveFunctionsForDocTypes[ docType ]
	
			if not receiveFunction then
				
				receiveFunction = receiveOneGenericDocumentP
				
			end
	
			return receiveFunction( docType, args, docBody )
				:mapDataP( function()
				
					return 'data', data, headers
				
				end )
		
		end )
		
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

local function receiveNewDocumentsP( args, json, optDocType )

	-- No documents? Signal no-op to upstream code so we can back off on the
	-- download interval.

	if not json or #json == 0 then

		-- Remember that we got nothing. This is our signal that we are completely
		-- caught up with Oz.
		
		args.docStore:setLocalStateP( 'isDoingInitialSync', false )

		return KSPromises.resolvedPromiseP( 'data', false )

	end

	local function receiveOneDocumentP( docBody )
		
		local docType = docBody.type or optDocType
		
		receivedDocumentsByType[ docType ] = ( receivedDocumentsByType[ docType ] or 0 ) + 1

		local receiveFunction = receiveFunctionsForDocTypes[ docType ]

		if not receiveFunction then
			
			receiveFunction = receiveOneGenericDocumentP
			
		end

		return receiveFunction( docType, args, docBody )
	
	end

	return KSPromises.allP( receiveOneDocumentP, json )

end

--------------------------------------------------------------------------------
-- Promise should resolve to new timestamp or nil if no new documents received.

local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

local Fetch = KSClass "Fetch" {
	url = false,
	getP = false,
	__init = function( self, ozUser, url )
		self.url = url
		self.getP = ozUser:getP( url, { compositeErrorData = true } )
		self.getP:whenResolvedDo( function() end )
			:attachTo( self )
	end,
	WFCloseSupport.addCloseSupportMethods,
}

local preFetch

local function updateDocsChangedSinceP( args, previousSyncTimestamp, nextUpdatesLink )

	local updatesListUrl = string.format( '%s/updates', args.catalogUrl )
		
	if nextUpdatesLink then

		updatesListUrl = nextUpdatesLink
	
	elseif previousSyncTimestamp then
		updatesListUrl = string.format( '%s?updated_since=%s',
									  updatesListUrl, previousSyncTimestamp )
	end
	
	downloadLog:info( "HTTP GET %s", updatesListUrl )
	
	local getP
	if preFetch then
		if preFetch.url == updatesListUrl then
			getP = preFetch.getP
		else
			preFetch:close()
			preFetch = nil
		end
	end
	if not getP then
		getP = args.ozUser:getP( updatesListUrl, { compositeErrorData = true } )
	end
	return getP
		:mapDataP( function( _, data )
			
			-- FIXME [scouten 2011-10-04]: One of the effects of
			-- compositeErrorData = true is that 500 errors get reported
			-- via onDataE, not onErrorE.

			local json = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )

			if json and next( json ) then

				args.accessErrorB:set( false )
				args.serviceAvailableB:set( true )
				args.hasDownloadedSuccessfullyB:set( true )

				return receiveNewDocumentsP( args, json.updates )
					:mapDataP( function()
						local nextUpdatesLink = json.links.next.href
						
						nextUpdatesLink = OzCatHelperFns.checkForRelativeUrlAndAppendBase( nextUpdatesLink, json.base )
						
						if json.updates then
							WFIdle.doLater( function()
								updateHistoricDiagnosticData( json.updates )
							end )
						end
						if #json.updates ~= 0 and nextUpdatesLink then
							preFetch = Fetch( args.ozUser, nextUpdatesLink )
						else
							preFetch = nil
						end
						return 'data', #json.updates ~= 0, json.last_updated or false, nextUpdatesLink, #json.updates

					end )
			
			else
			
				return 'data', false
				
			end

		end )

end

--------------------------------------------------------------------------------

local function extractUpdatedSince( url )
	local _, updateDate
	if url then
		_, _, updateDate = string.find( WFURI.unescape( url ), "updated_since=([^&]+)" )
	end
	
	return updateDate
end

--------------------------------------------------------------------------------
-- This promise should always return catalogInfo so we can chain these
-- requests together.

local function refreshDocumentTypeOnceP( args, catalogInfo, linkPath, arrayName, docType, optVarName )

	-- What's the current refresh-once state for this link path?
	
	local localStateVarName = optVarName or 'refreshOnce-1.7+' .. linkPath
	local docStore = args.docStore
	local ozNextUpdatesLink
		-- it'd be nice to use the local state variable ozCentralTimestamp, but
		-- earlier versions of Revel did not set this when syncing.
	
	return args.docStore:localStateP( 'ozNextUpdatesLink' )
	
		:mapDataP( function( _, inOzNextUpdatesLink )
			ozNextUpdatesLink = inOzNextUpdatesLink
		end )
		:mapDataP( function( _, inOzCentralTimestamp )
			return docStore:localStateP( localStateVarName )
		end )
		:mapDataP( function( _, refreshState )

			-- If we're done, exit out now.

			if refreshState == 'done' then
				return 'data', catalogInfo
			end
			
			-- Calculate refresh URL from refresh state, or links section
			-- of catalog info.
			
			local refreshPartialUrl = refreshState
									  or KSTableUtils.derefOpt( catalogInfo,
																'links',
																linkPath,
																'href' )
			
			if not refreshPartialUrl then
				return 'error', "catalogInfo is incomplete"
			end
			
			local refreshUrl = OzCatHelperFns.checkForRelativeUrlAndAppendBase( refreshPartialUrl, catalogInfo.base )
			
			local refreshSinceDate = extractUpdatedSince( refreshUrl )
			local nextUpdateDate = extractUpdatedSince( ozNextUpdatesLink )
			
			if nextUpdateDate == nil or ( refreshSinceDate and refreshSinceDate > nextUpdateDate ) then
				-- we have caught up with initial sync; in the event ozNextUpdatesLink is nil, then
				-- this is an initial sync so we don't need to refresh at all
				return docStore:setLocalStateP( localStateVarName, 'done' )
					:mapDataP( function()
						return 'data', catalogInfo
					end )
			end

			return args.ozUser:getP( refreshUrl, { compositeErrorData = true } )
				:mapDataP( function( _, data )
				
					local json = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
		
					if json and next( json ) then
		
						args.accessErrorB:set( false )
						args.serviceAvailableB:set( true )
						args.hasDownloadedSuccessfullyB:set( true )
						
						local newDocs
						if type( arrayName ) == 'string' then
							newDocs = json[ arrayName ]
						elseif type( arrayName ) == 'function' then
							newDocs = arrayName( json )
						else
							error( "Invalid arrayName parameter" )
						end
						
						local finish = function()

							local nextPartialUrl = json.links
											   and json.links.next
											   and json.links.next.href
							
							if nextPartialUrl == refreshPartialUrl then -- Oz bug
								nextPartialUrl = nil
							end
							
							return docStore:setLocalStateP( localStateVarName,
													nextPartialUrl or 'done' )
								:mapDataP( function()
									return 'data', catalogInfo
								end )

						end
						
						-- Don't call into receiveNewDocumentsP with a zero-length array
						-- as that will cause our code to assume the initial sync is done.
						-- fixes 3372749
						if #newDocs == 0 then
							return finish()
						end
						
						return receiveNewDocumentsP( args, newDocs, docType )
							:mapDataP( finish )

					end
					
				end )
			
		end )

end

--------------------------------------------------------------------------------

-- FIXME [scouten 2012-04-17]: This should move out to a string utils file.

local function split( str, pat )

	local t = {}  -- NOTE: use {n = 0} in Lua-5.0
	local fpat = "(.-)" .. pat
	local last_end = 1

	local s, e, cap = str:find( fpat, 1 )

	while s do

		if s ~= 1 or cap ~= "" then
			table.insert( t, cap )
		end

		last_end = e+1
		s, e, cap = str:find( fpat, last_end )

	end

		if last_end <= #str then
			cap = str:sub( last_end )
			table.insert( t, cap )
		end

	return t

end

--------------------------------------------------------------------------------
--[[!
	\brief Downloads the latest ((what?)) from the Oz database.

	\param args (table) A table that contains these keys:
		<ul><li>docStore</li>
		    <li>catalogId</li>
		    <li>itmsLinkB</li>
		    <li>existsOnOzB</li>
		    <li>accessErrorB</li>
		    <li>serviceAvailableB</li>
		    <li>hasDownloadedSuccessfullyB</li>
		    <li>errorCapture</li></ul>

	\return <code>P < ?? ></code> A Promise that resolves to ((what?))

]]--
function WFOzDownloadTask.downloadLatestFromOzP( args )

	check( 'table' )

	assert( args.docStore )
	assert( args.catalogId )
	assert( args.itmsLinkB )
	assert( args.existsOnOzB )
	assert( args.accessErrorB )
	assert( args.serviceAvailableB )
	assert( args.hasDownloadedSuccessfullyB )
	assert( args.errorCapture )

	-- First, update catalog root info.
		-- FIXME [scouten 2012-04-17]: Requested a change from Oz to add the
		-- catalog document itself to the combined _updates feed. Can remove the
		-- distinct updateCatalogRootInfo call when that's done, which will make
		-- me happy!

	args.errorCapture.downloadLatestFromOzP = {}

	local resultP = updateCatalogRootInfoP( args )
	
		-- PR #3316853: If we might have missed any non-asset documents, let's
		-- circle back and reload them.

		:mapDataP( function( _, catalogInfo )
		
			return refreshDocumentTypeOnceP( args, catalogInfo, '/rels/albums', 'albums', 'album' )
		
		end )
		:mapDataP( function( _, catalogInfo )
			
			return refreshDocumentTypeOnceP( args, catalogInfo, '/rels/catalog_spaces', 'spaces', 'space' )
		
		end )
		:mapDataP( function( _, catalogInfo )
			
			return refreshDocumentTypeOnceP( args, catalogInfo, '/rels/updates', function( json )
				local docs = {}
				for i, jsonDoc in ipairs( json.updates ) do
					if jsonDoc.type == 'comment' then
						docs[ #docs + 1 ] = jsonDoc
					end
				end
				return docs
			end, 'comment', 'refreshOnce-1.8+comments' )
		
		end )
	
		-- Then find out when we last correctly synced with Oz.
	
		:mapDataP( function()
		
			args.errorCapture.downloadLatestFromOzP.updateCatalogRootInfoP_DONE = true

			return args.docStore:localStateP( 'ozCentralTimestamp' )
				:mapP( function( statusTimestamp, timestamp )
				
					args.errorCapture.downloadLatestFromOzP.ozCentralTimestamp = timestamp

					return args.docStore:localStateP( 'ozNextUpdatesLink' )
						:mapP( function( statusLink, link )
						
							args.errorCapture.downloadLatestFromOzP.ozNextUpdatesLink = link

							-- If we have neither timestamp (1.x sync algorithm) nor
							-- next updates link (2.0+ sync algorithm), then we're almost
							-- certainly syncing from scratch for this catalog.
							
							if not timestamp and not link then

								args.errorCapture.downloadLatestFromOzP.flaggedForInitialSync = true

								args.docStore:setLocalStateP( 'isDoingInitialSync', true )

							end
						
							local updateTimestamp = statusTimestamp == 'data' and timestamp
							local updateLink = statusLink == 'data' and link
							return 'data', updateTimestamp, updateLink
							
						end )
					
				end )
			
		end )
		
		-- Now process all documents that have been updated since then.
		
		:mapDataP( function( _, timestamp, updatesLink )
		
			return updateDocsChangedSinceP( args, timestamp, updatesLink )
		
		end )

		-- Save off new sync timestamp or the next updates link.
		
		:mapDataP( function( _, haveMoreToDownload, newTimestamp, nextUpdatesLink, numItemsDownloaded )
		
			args.errorCapture.downloadLatestFromOzP.updateTimestampOrLink = {}
		
			local function updateTimestampOrLink( info )

				if info[ 1 ] and info[ 2 ] then

					args.errorCapture.downloadLatestFromOzP.updateTimestampOrLink[ info[ 1 ] ] = info[ 2 ]
				
					return args.docStore:setLocalStateP( info[ 1 ], info[ 2 ] )

				end
			end
			
			return KSPromises.allP( updateTimestampOrLink, { { 'ozCentralTimestamp',  newTimestamp }, { 'ozNextUpdatesLink', nextUpdatesLink } } )
				:mapDataP( function()
					return 'data', haveMoreToDownload, numItemsDownloaded
				end, 2 )
		
		end )
	
		-- Make sure final result is a boolean.
		
		:mapDataP( function( _, haveMoreToDownload, numItemsDownloaded )

			-- haveMoreToDownload is true if the updates count did not match the total_count.
			-- Using this instead of waiting till the total_count comes back as 0 per Matt's request.
			return 'data', haveMoreToDownload, numItemsDownloaded

		end )

	-- Capture error info.

	resultP = resultP:observeP( function( status, data, headers )
	
		if status ~= 'data' then
		
			args.errorCapture.downloadLatestFromOzP.data = tostring( data )
			args.errorCapture.downloadLatestFromOzP.headers = headers
		
			-- Sniff the error data.
			
			local errorData = data and type( data ) == 'string' and split( data, ',' ) or {}
		
			local error_code =  errorData[ 2 ]
			local link = errorData[ 3 ]
	
			if error_code == '4999' and link and link ~= '' then
				args.itmsLinkB:set( link )
			end
				
			local message
			if type( headers ) == 'string' then
				message = headers
			elseif type( headers ) == 'table' then
				message = headers.status_code
			elseif type( data ) == 'string' then
				message = data
			else
				message = "Unknown error type: " .. type( headers ) .. ", " .. type( data )
			end
			
			downloadLog:error( "download session request failed for catalog ID %s with message %s ",
							   args.catalogId, message )
	
			-- If we get a 404 back this catalog has been deleted on Oz.
			-- If we get a 403 this catalog has been unshared.
			-- If we get a 403 with while (1) {}
			-- {"code":4999,"description":"Upgrade required"} as the payload,
			-- we need to force a client update.
			
			-- Oz will return 503 with '9999' to signify that the service is unavailable.
			-- We track this using serviceAvailableB so that the UI can inform the user.
	
			if headers and headers.status_code == '404' then
			
				args.existsOnOzB:set( false )
				
			elseif headers and headers.status_code == '403' then
	
				if error_code then
					args.accessErrorB:set( error_code )
				else
					args.accessErrorB:set( true )
				end
				
			elseif headers and headers.status_code == '503' then
			
				args.serviceAvailableB:set( false )
				
			end
			
		end
	
	end )
	
	return resultP

end

--------------------------------------------------------------------------------

WFDiagnosticLog.addLogSectionFunc( function()

	return {

		name = "Download Status",
		id = 'WFOzDownloadTask',
		weight = 650,
		outputForAutomation = true,

		body = {
		
			receivedDocumentsByType = receivedDocumentsByType,
			syncHistory = buildSyncHistorySparkline(),
		},
	
	}

end )

--------------------------------------------------------------------------------

return WFOzDownloadTask
