--[[----------------------------------------------------------------------------

WFOzDownloadTaskTest.lua

11/26/2012
arahn

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFOzSyncTestUtils = require "WFOzSyncTestUtils"
local WFAsyncTestCase = require "WFAsyncTestCase"
local KSReactive = require "KSReactive"
local KSJsonEncoder = require "KSJsonEncoder"
local WFPlatformInfo = require 'WFPlatformInfo'

local TEST_WAIT_TIME = 2

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local WFOzDownloadTask

local T = {}
local docStore, ozUser, indexStore

--------------------------------------------------------------------------------

local function buildMocks( docStore )
	
	return {
		KSFileUtils = {
		
		},
		
		OzCentralUploadTask = WFOzSyncTestUtils.makeFakeOzCentralUploadTask( T ),
		
		OzCatalogIndexDefinitions = false,
		
		-- Fake implementation which uses OzCatalogIndexDefinitions to build out the proper
		-- schema and mapping functions for our in-memory indexStore
		WFIndexStoreSchema = docStore:raw_test_buildMockWFIndexStoreSchema(),
		
		-- Fake impl, does nothing
		WFIndexStoreImp = {
			start = function( schema )
			
			end,
		},
	}
	
end

--------------------------------------------------------------------------------

function TestSuite.setUpClass()
	
	-- Build a fake docStore, sufficient for testing this namespace
	
	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	WFOzDownloadTask = WFOzSyncTestUtils.importWithMocks( 'WFOzDownloadTask', buildMocks( docStore ) )
	
	-- Utterly fake oz user.  Vends canned responses to GETs and stores
	-- of what was PUT for later analysis.
	ozUser = WFOzSyncTestUtils.makeFakeOzUser( T )

	-- Build a fake indexStore.  It's powered by a real sql db in memory
	-- that uses the real schema and mapping fns.
	indexStore = docStore:raw_test_buildMockIndexStore( )
	
end

--------------------------------------------------------------------------------

function TestSuite.testVerifyDownloadUnknownDocument( test )


	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )


end

--------------------------------------------------------------------------------

function TestSuite.test_downloadLatestFromOzP_downloadComments( test )

	-- Need to start out with a fresh docStore for each test.
	
	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )
	
	T.userGetPutResponse.LOLcatURL = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			catalogUsers = {
				"Joe User Url"
			},
			links = {
				[ '/rels/updates' ] = {
					href = "Some0Updatez",
				},
			},
			base = "TehBaseUrl/",
		},
	}
	
	T.userGetPutResponse[ 'LOLcatURL/users' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			"Joe User Info"
		},
	}
	
	T.userGetPutResponse[ 'LOLcatURL/updates?updated_since=2020-12-12T12:12:12' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			updates = {
				{
					type = 'unknown',
					_id = 'tresMysterious',
					Q = "Who knows what evil lurks in the hearts of men?",
					A = "The Shadow knows!",
				},
			},
			links = {
				next = {
					href = 'updated_NEXT_href',
				},
			},
			base = 'updATEd_base',
		},
	}
	
	
	docStore._localStateP[ 'refreshOnce-1.7+/rels/albums' ] = 'done'
	docStore._localStateP[ 'refreshOnce-1.7+/rels/catalog_spaces' ] = 'done'
	docStore._localStateP[ 'refreshOnce-1.7+/rels/updates' ] = 'done'
	docStore._localStateP[ 'ozCentralTimestamp' ] = '2020-12-12T12:12:12'
	
	local itmsLinkB = KSReactive.variableB()
	local existsOnOzB = KSReactive.variableB()
	local accessErrorB = KSReactive.variableB()
	local serviceAvailableB = KSReactive.variableB()
	local hasDownloadedSuccessfullyB = KSReactive.variableB()
	local errorCapture = {}
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( WFOzDownloadTask.downloadLatestFromOzP {
			docStore = docStore,
			catalogId = docStore:catalogId(),
			catalogUrl = "LOLcatURL",
			ozUser = ozUser,
			itmsLinkB = itmsLinkB,
			existsOnOzB = existsOnOzB,
			accessErrorB = accessErrorB,
			serviceAvailableB = serviceAvailableB,
			hasDownloadedSuccessfullyB = hasDownloadedSuccessfullyB,
			errorCapture = errorCapture,
		}:mapDataP( function( status, result )
			return status, result, docStore._docs
		end ) )
		:continueWhenValueMatches( 'data', true, {
			catalogcatalog_id_14159265 = {
				_docId = "catalog_id_14159265",
				_docType = "catalog",
				base = "TehBaseUrl/",
				catalogUsers = {
					"Joe User Info",
				},
				links = {
					["/rels/updates"] = {
						href = "Some0Updatez",
					},
				},
			},
			unknowntresMysterious = {
				A = "The Shadow knows!",
				Q = "Who knows what evil lurks in the hearts of men?",
				_docId = "tresMysterious",
				_docType = "unknown",
				type = "unknown",
			},
		} )
		

end

--------------------------------------------------------------------------------

function TestSuite.test_absoluteUrlHandling( test )

	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )
	
	T.userGetPutResponse.HOTSoup = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			catalogUsers = {
				"Noone"
			},
			links = {
				[ '/rels/updates' ] = {
					-- Pass in an absolute URL for updates
					href = "AbsoluteUrl/forUpdates",
				},
			},
			base = "HOTSoup/",
		},
	}
	
	T.userGetPutResponse[ 'HOTSoup/users' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			"Noone"
		},
	}
	
	-- If our URL handling is correct, we would get the following response
	-- when we try and fetch updates.
	T.userGetPutResponse[ 'AbsoluteUrl/forUpdates' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			updates = {
				{
					type = 'unknown',
					_id = 'someUpdate',
					Q = "Soup for me?",
					A = "No soup for you!",
				},
			},
			links = {
				next = {
					href = 'AbsoluteUpdates/Next',
				},
			},
			base = "HOTSoup/",
		},
	}
	
	-- We should get here if the previous request for updates succeeded.
	T.userGetPutResponse[ 'HOTSoup/updates?updated_since=2020-12-12T12:12:12' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			updates = {
				{
					type = 'unknown',
					_id = 'someUpdate',
					Q = "Soup for me?",
					A = "Yes, coz you're Batman!",
				},
			},
			links = {
				next = {
					-- Pass in an absolute URL for the next updates link
					href = 'HOTSoup/updates?updated_since=2400-12-12T12:12:12',
				},
			},
			base = "HOTSoup/",
		},
	}

	docStore._localStateP[ 'refreshOnce-1.7+/rels/albums' ] = 'done'
	docStore._localStateP[ 'refreshOnce-1.7+/rels/catalog_spaces' ] = 'done'
	docStore._localStateP[ 'ozNextUpdatesLink' ] = 'HOTSoup/updates?updated_since=2020-12-12T12:12:12'
	
	local itmsLinkB = KSReactive.variableB()
	local existsOnOzB = KSReactive.variableB()
	local accessErrorB = KSReactive.variableB()
	local serviceAvailableB = KSReactive.variableB()
	local hasDownloadedSuccessfullyB = KSReactive.variableB()
	local errorCapture = {}
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( WFOzDownloadTask.downloadLatestFromOzP {
			docStore = docStore,
			catalogId = docStore:catalogId(),
			catalogUrl = "HOTSoup",
			ozUser = ozUser,
			itmsLinkB = itmsLinkB,
			existsOnOzB = existsOnOzB,
			accessErrorB = accessErrorB,
			serviceAvailableB = serviceAvailableB,
			hasDownloadedSuccessfullyB = hasDownloadedSuccessfullyB,
			errorCapture = errorCapture,
		}:mapDataP( function( status, result )
			return status, result, docStore._docs, docStore._localStateP['ozNextUpdatesLink']
		end ) )
		:continueWhenValueMatches( 'data', true, {
			catalogcatalog_id_14159265 = {
				_docId = "catalog_id_14159265",
				_docType = "catalog",
				base = "HOTSoup/",
				catalogUsers = {
					"Noone",
				},
				links = {
					["/rels/updates"] = {
						href = "AbsoluteUrl/forUpdates",
					},
				},
			},
			-- Match the updates, if we failed at any point in the download sequence, 
			-- this would not match.
			unknownsomeUpdate = {
				A = "Yes, coz you're Batman!",
				Q = "Soup for me?",
				_docId = "someUpdate",
				_docType = "unknown",
				type = "unknown",
			},
			-- Match the value for ozNextUpdatesLink in the docStore, 
			-- if this matches our URL handling was correct.
		
		}, 'HOTSoup/updates?updated_since=2400-12-12T12:12:12' )


end
--------------------------------------------------------------------------------

function TestSuite.test_relativeUrlHandling( test )

	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )
	
	T.userGetPutResponse.RealHOTSoup = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			catalogUsers = {
				"Noone"
			},
			links = {
				[ '/rels/updates' ] = {
					-- Pass in a relative URL for updates
					href = "/forUpdates",
				},
			},
			base = "RealHOTSoup",
		},
	}
	
	T.userGetPutResponse[ 'RealHOTSoup/users' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			"Noone"
		},
	}
	
	-- If our URL handling is correct, we would get the following response
	-- when we try and fetch updates.
	T.userGetPutResponse[ 'RealHOTSoup/forUpdates' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			updates = {
				{
					type = 'unknown',
					_id = 'someUpdate',
					Q = "Soup for me?",
					A = "No soup for you!",
				},
			},
			links = {
				next = {
					href = '/Next',
				},
			},
			base = "RealHOTSoup",
		},
	}
	
	-- We should get here if the previous request for updates succeeded.
	T.userGetPutResponse[ 'RealHOTSoup/updates?updated_since=2020-12-12T12:12:12' ] = {
		'data',
		KSJsonEncoder.encodeWebSafeJson {
			updates = {
				{
					type = 'unknown',
					_id = 'someUpdate',
					Q = "Soup for me?",
					A = "Yes, coz you're Batman!",
				},
			},
			links = {
				next = {
					-- Pass in a relative URL for the next updates link
					href = '/updates?updated_since=2400-12-12T12:12:12',
				},
			},
			base = "RealHOTSoup",
		},
	}

	docStore._localStateP[ 'refreshOnce-1.7+/rels/albums' ] = 'done'
	docStore._localStateP[ 'refreshOnce-1.7+/rels/catalog_spaces' ] = 'done'
	docStore._localStateP[ 'ozNextUpdatesLink' ] = 'RealHOTSoup/updates?updated_since=2020-12-12T12:12:12'
	
	local itmsLinkB = KSReactive.variableB()
	local existsOnOzB = KSReactive.variableB()
	local accessErrorB = KSReactive.variableB()
	local serviceAvailableB = KSReactive.variableB()
	local hasDownloadedSuccessfullyB = KSReactive.variableB()
	local errorCapture = {}
	
	return WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( WFOzDownloadTask.downloadLatestFromOzP {
			docStore = docStore,
			catalogId = docStore:catalogId(),
			catalogUrl = "RealHOTSoup",
			ozUser = ozUser,
			itmsLinkB = itmsLinkB,
			existsOnOzB = existsOnOzB,
			accessErrorB = accessErrorB,
			serviceAvailableB = serviceAvailableB,
			hasDownloadedSuccessfullyB = hasDownloadedSuccessfullyB,
			errorCapture = errorCapture,
		}:mapDataP( function( status, result )
			return status, result, docStore._docs, docStore._localStateP['ozNextUpdatesLink']
		end ) )
		:continueWhenValueMatches( 'data', true, {
			catalogcatalog_id_14159265 = {
				_docId = "catalog_id_14159265",
				_docType = "catalog",
				base = "RealHOTSoup",
				catalogUsers = {
					"Noone",
				},
				links = {
					["/rels/updates"] = {
						href = "/forUpdates",
					},
				},
			},
			-- Match the updates, if we failed at any point in the download sequence, 
			-- this would not match.
			unknownsomeUpdate = {
				A = "Yes, coz you're Batman!",
				Q = "Soup for me?",
				_docId = "someUpdate",
				_docType = "unknown",
				type = "unknown",
			},
			-- Match the value for ozNextUpdatesLink in the docStore, 
			-- if this matches our URL handling was correct.
		
		}, 'RealHOTSoup/updates?updated_since=2400-12-12T12:12:12' )


end
--------------------------------------------------------------------------------

return TestSuite
