--[[----------------------------------------------------------------------------

WFOzFileLock.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@reviewed scouten 2012-05-15

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'


--------------------------------------------------------------------------------

local check = WFArguments.check

local WFOzFileLock = 'WFOzFileLock'


--------------------------------------------------------------------------------

local fileLockLog = KSLogging.log 'WFOzFileLock'
--	:setLevel( KSLogging.DEBUG )

local BabluEP = WFWorkerThread( 'BabluImp.lua' )


--============================================================================--
--[[!
	\class WFOzFileLock
	\brief Provides the ability to lock a file, preventing it
	from being purged or moved between documents and caches.
	
	Create an object of this type to lock the associated file, and close
	the object explicitly when you no longer need to protect the file.
	
	\author Eric Scouten <scouten@adobe.com> 2012-05-15
--]]

-- @mark WFOzFileLock class

local fileLockSessionMethods = {

	_catalogId = false,
		-- (string) catalog ID

	_filePath = false,
		-- (string) path to the file we're locking
	
	_lockP = false,
		-- (P< true >) confirmation of lock from purge agent
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs a file lock object.
	
	\param ozCatalogService (OzCatalogService) The catalog to which the file belongs.
	\param filePath (string) The full path to the file to protect.
--]]

function fileLockSessionMethods:__init( --[[OzCatalogService]] ozCatalogService,
										--[[string]] filePath )

	check( WFOzFileLock, 'OzCatalogService | OzCatalogServiceImp', 'string' )

	WFCloseSupport.mustClose( self )

	fileLockLog:debug( "Requesting lock on file at path %s", filePath )

	self._catalogId = ozCatalogService._catalogId
	self._filePath = filePath
	
	self._lockP = BabluEP:requestP( 'lockFileAtPath', self._catalogId, filePath )
		:mapP( function( status, result )
		
			fileLockLog:debug( "File lock request for path %s resolved with status %s result %s",
							   filePath, status, result )

			if status == 'data' and not result then
				return 'error', "File no longer exists"
			else
				return status, result
			end
		
		end )

	self._lockP:whenResolvedDo( function() end ):attachTo( self )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) the path to the file that this lock object
		is associated with.
	
	\return <code>P< filePath ></code> A Promise that resolves to
		the full path when the file lock is confirmed (or to an error otherwise).
--]]

function fileLockSessionMethods:lockedFilePathP()

	return self._lockP
		:mapDataP( function()
			return 'data', self._filePath
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Closes this ojbect, releasing the file lock.
--]]

function fileLockSessionMethods:close()

	fileLockLog:debug( "Releasing lock on file at path %s", self._filePath )

	BabluEP:requestP( 'unlockFileAtPath', self._catalogId, self._filePath )
		:runUntilDone()
	
end

--------------------------------------------------------------------------------

return KSClass 'WFOzFileLock'( fileLockSessionMethods )
