--[[----------------------------------------------------------------------------

WFOzPurgeAgent.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the content of a single catalog's data cache and periodically
purges it.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'

	-- wichitafoundation.core.lua.threading
local WFTimer = require 'WFTimer'

	-- wichitafoundation.core.os
local KSFileUtils = require 'KSFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.os.prefs
local KSPrefs = require 'KSPrefs'

	-- wichitafoundation.oz.doc_store
local OzDocStore = require 'OzDocStore'

	-- wichitafoundation.oz.index_store
local WFIndexStore = require 'WFIndexStore'

	-- wichitafoundation.oz.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'


--------------------------------------------------------------------------------

local cachePollingInterval = 40	-- seconds; feel free to tune

local minimumFreeSpace = WFOzPurgeSettings.minimumFreeSpace()

local purgeLog = KSLogging.log 'WFOzPurgeAgent'
--	:setLevel( KSLogging.DEBUG )

local purgeLockLog = KSLogging.log 'WFOzPurgeAgent :: filePath lock'
--	:setLevel( KSLogging.DEBUG )

local timerLog = KSLogging.log 'WFOzPurgeAgent :: timer'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--
--[[!
	\brief A \c WFOzPurgeAgent monitors the content of a single catalog's data
	cache and periodically purges it to make additional disk space available.
	It should be used only by BABLU.
	
	One \c WFOzPurgeAgent object is created per active catalog. Their lifetimes
	are managed via \c WFOzPurgeAgentManager.
--]]

-- @mark WFOzPurgeAgent

local purgeAgentMethods = {

	_catalogId = false,
		-- (string) catalog ID as known to the Oz system (UUID)
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (OzIndexStore)
	
	_dataCache = false,
		-- (KSDataCache) data cache for this doc store
	
	_dataCachePathP = false,
		-- (P< string >) path to cache
	
	_purgeCacheP = false,
		-- (P< true >) currently running purge cache operation, if any
		
	_purgeCacheAction = false,
		-- (KSPromiseAction) action attached to currently running purge cache operation, if any
		
	_purgeAllB = false,
		-- (B< boolean >) used to enable purging of all eligible renditions in response to
		-- a user request

	_checkCacheKeysFrom = false,
		-- (false or string) where do we start cross-checking cache keys?

	_warnedMissingEntries = false,
		-- (table) entries that we've warned about (so we don't keep bleating about them)

	_unreferencedFromPreviousSession = false,
		-- (table or nil) unreferenced cache entries from previous session

	_lockedFilePaths = false,
		-- (table) paths that are locked from purging
			-- key: (string) file path
			-- value: true

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Initialize a purge agent for a specific catalog.
	
	\param catalogId (string) ID for the catalog
--]]

function purgeAgentMethods:__init( --[[string]] catalogId )

	self._catalogId = assert( catalogId )
	
	self._docStore = OzDocStore( catalogId )
	self:addCloseTarget( self._docStore )

	self._indexStore = WFIndexStore( self._docStore, 'OzCatalogIndexDefinitions.lua' )
	self:addCloseTarget( self._indexStore )

	self._dataCache = assert( self._docStore:dataCache() )

	self._dataCachePathP = self._dataCache:pathP()
	self._dataCachePathP:runUntilDone()
	
	self._warnedMissingEntries = {}
	
	self._lockedFilePaths = {}
	
	self._purgeAllB = KSReactive.variableB( false )

	-- Start cache purge, but not right away. Give priority to the rest
	-- of the UI starting up. 20 seconds is an arbitrary choice. Tune as
	-- we get used to it.

	self:schedulePurgeSession( 20 )
	
end

--------------------------------------------------------------------------------

local purgeCacheForCatalogP -- forward declaration

--------------------------------------------------------------------------------
--[[!
	\brief Schedule a purge session after a delay.
	
	\param afterDelay (number) how many seconds to wait before starting the
		purge session
]]--

function purgeAgentMethods:schedulePurgeSession( --[[number]] afterDelay )

	-- GRUNGE [scouten 2011-04-26]: There's almost certainly a proper reactive
	-- way to code this up, but I've spent too much time trying to think through
	-- what it is. This will work, even if it's messy.

	-- If we're already downloading or have already scheduled the timer,
	-- this is a no-op.

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		return
	end

	if self._purgeCacheP then
		return
	end
	
	if self._purgeCacheAction and not self._purgeAllB:valueNow() then
		return
	end
	
	if self._purgeCacheAction then
		self._purgeCacheAction:close()
	end
 	
	local delaySeconds = afterDelay or cachePollingInterval

	timerLog:debug( "delay timer for purgeCacheForCatalogP launched :: waiting %s seconds",
					delaySeconds )

	self._purgeCacheP = WFTimer.afterDelayP( delaySeconds )
		:mapDataP( function()
		
			timerLog:debug( "delay timer for purgeCacheForCatalogP fired" )

			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				return 'data', true
			end

			return purgeCacheForCatalogP( self )
		
		end )
		
	self._purgeCacheAction = self._purgeCacheP:whenResolvedDo( function( status, message )
	
		if status ~= 'data' then

			timerLog:error( "purgeCacheForCatalog failed with status %s and message %s",
							status, message )

			-- If we were purging all and errored out, we need to set purgeAll back to false.
			
			if self._purgeAllB:valueNow() then
				self._purgeAllB:set( false, message )
			end

		end
		
		self._purgeCacheP = false
		self._purgeCacheAction = false
		
		-- If purge all is enabled, schedule a purge session immediately.
		
		if self._purgeAllB:valueNow() then
			self:schedulePurgeSession( 0 )
		else
			self:schedulePurgeSession()
		end
	
	end ):attachTo( self )
	
end

--------------------------------------------------------------------------------

local collectUnreferencedKeysP -- forward declaration
local freeSpaceIfNeededP -- forward declaration
local moveOriginalsFromDocumentsToCachesP -- forward declaration

--------------------------------------------------------------------------------

purgeCacheForCatalogP = function( self )

	return freeSpaceIfNeededP( self )
		:mapDataP( function( status, data )
			
			return collectUnreferencedKeysP( self )
				:mapDataP( function( status, data )
					
					return moveOriginalsFromDocumentsToCachesP( self )
						
				end )
		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Start purging for this catalog. Used by the "Purge All" command to
	make as much space as possible available.
]]--

function purgeAgentMethods:setPurgeAllEnabled()

	self._purgeAllB:set( true )

	-- Try to kick off a purge all cycle right away.

	self:schedulePurgeSession( 0 )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Disable purging for this catalog.
]]--

function purgeAgentMethods:setPurgeAllDisabled()

	self._purgeAllB:set( false )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Return a promise that will resolve when the current purge all session
	has completed.

	\return P< true > resolves when current purge operation is finished
--]]

function purgeAgentMethods:purgeFinishedP()

	local resultP, resolver = KSPromises.promiseP()
	
	self._purgeAllB
		:onDataB( function ( purgeEnabled, errorMessage )

			if not purgeEnabled then

				if resolver:canResolvePromise() then
					if not errorMessage then
						resolver( 'data', true )
					else
						resolver( 'error', errorMessage )
					end
				end

			end

		end )

	return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Request a purge lock on a file (typically an original or proxy).
	
	While this lock is in place, the file will not be purged or moved.
	
	\param filePath (string) path to lock
	
	\return true if the file exists and the lock has been granted, false otherwise
--]]

function purgeAgentMethods:lockFileAtPath( --[[string]] filePath )

	if KSFileUtils.exists( filePath ) then
	
		purgeLockLog:info( "Lock granted for file at path %s", filePath )
	
		self._lockedFilePaths[ filePath ] = true
		return true
	
	else
	
		purgeLockLog:info( "Lock denied for file at path %s (file does not exist)", filePath )

		return false
	
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Release the purge lock on a file.
	
	Releasing the lock means that the file may become eligible for purging.
	
	This function is assumed to always succeed.
	
	\param filePath (string) path to unlock
--]]

function purgeAgentMethods:unlockFileAtPath( --[[string]] filePath )

	if self._lockedFilePaths[ filePath ] then
		purgeLockLog:info( "Lock released for file at path %s", filePath )
	end

	self._lockedFilePaths[ filePath ] = nil

end

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-08-12]: We should also sniff for unreferenced originals
-- and missing originals.

--------------------------------------------------------------------------------

local function discoveredMissingEntry( self, k, photoId )

	-- FIXME [scouten 2011-08-29]: Would like to know why the enumerate API
	-- is causing false alarms just after importing photos, but since this is
	-- the only place we enumerate the data cache, it can wait IMHO.
	
	purgeLog:debug( "looks like cache entry %s for photo %s is missing ... checking",
					k, photoId )

	self._dataCache:getP( k )
		:whenResolvedDo( function( status, data )
		
			if status == 'data' and data and KSData.length( data ) > 0 then
			
				purgeLog:info( "checked cache entry %s for photo %s and it was OK",
							   k, photoId )
			
			else

				purgeLog:warn( "checked cache entry %s for photo %s and it was missing with status = %s",
							   k, photoId, status )
				
				self._warnedMissingEntries[ k ] = true

				-- FIXME [scouten 2011-08-29]: Yank the entry from the
				-- doc DB entry.

			end
		
			if data then
				data:close()
			end
		
		end ):runUntilDone()

end

--------------------------------------------------------------------------------

collectUnreferencedKeysP = function( self )

	local optStartPath = self._checkCacheKeysFrom or nil
	
	local dataCache = self._dataCache
	local indexStore = self._indexStore
	
	local keysFromCacheAsSet = {} -- filled in below
	
	local keysFromCacheP = dataCache:enumerateKeysP( optStartPath ) -- , optLimit
		:mapDataP( function( _, data )
		
			-- If we didn't get any keys from this query, then we've cycled
			-- through all keys.

			if not data or #data == 0 then
				self._checkCacheKeysFrom = false
				return 'data', false, optStartPath
			end
			
			local minKey = optStartPath
			local maxKey = data[ #data ]
			self._checkCacheKeysFrom = maxKey
			
			purgeLog:info( "\n\nminKey = %s" ..
						   "\n\nmaxKey = %s\n\n",
								minKey, maxKey )

			for i = 1, #data do
				keysFromCacheAsSet[ data[ i ] ] = true
			end
			
			return 'data', true, minKey, maxKey
			 
		end )
		
	local keysFromCatalogP = keysFromCacheP
		:mapDataP( function( _, hadKeys, minKey, maxKey )

			if hadKeys then
			
				purgeLog:info( "hadKeys == true" )
			
				return indexStore:queryP( [[SELECT md5, docId FROM nonOriginalsCacheReferences
					WHERE md5 >= ? AND md5 <= ?]], minKey or 0, maxKey )
			
			elseif optStartPath then

				purgeLog:info( "hadKeys == false; finishing range from %s", optStartPath )
			
				return indexStore:queryP( [[SELECT md5, docId FROM nonOriginalsCacheReferences
					WHERE md5 > ?]], optStartPath )
			
			else

				purgeLog:info( "looks like cache was completely empty" )

				return indexStore:queryP( [[SELECT md5, docId FROM nonOriginalsCacheReferences]] )
			
			end
			
		end ):mapDataP( function( _, data )
	
			local keysFromCatalogAsSet = {}
			for i = 1, #data do
				keysFromCatalogAsSet[ data[ i ][ 1 ] ] = data[ i ][ 2 ]
			end
			
			return 'data', keysFromCatalogAsSet
		
		end )

	-- Now that we have both key sets, cross-reference them.
	
	local unreferencedKeysFromCacheAsSet = {} -- filled in below

	local previouslyUnreferenced = self._unreferencedFromPreviousSession or {}
	self._unreferencedFromPreviousSession = previouslyUnreferenced
	
	local unfulfilledKeysFromCatalogAsSet = {} -- filled in below
	
	local crossReferenceKeysP = keysFromCatalogP
		:mapDataP( function( _, keysFromCatalogAsSet )
	
			for k in pairs( keysFromCacheAsSet ) do
				if not keysFromCatalogAsSet[ k ] then
					purgeLog:debug( "flagging cache entry %s for deletion", k )
					unreferencedKeysFromCacheAsSet[ k ] = true
				end
			end
			
			for k, photoId in pairs( keysFromCatalogAsSet ) do
				if not keysFromCacheAsSet[ k ] then
					if not self._warnedMissingEntries[ k ] then
						discoveredMissingEntry( self, k, photoId )
					end
					unfulfilledKeysFromCatalogAsSet[ k ] = photoId
				end
			end
				
			purgeLog:info( "\n\nunreferencedKeysFromCacheAsSet = %s" ..
						   "\n\nunfulfilledKeysFromCatalogAsSet = %s\n\n",
						   unreferencedKeysFromCacheAsSet,
						   unfulfilledKeysFromCatalogAsSet )

			return 'data', true
		
		end )
	
	-- Now go kill off cache entries that are unreferenced.
	-- Remove is a fire-and-forget operation, so we just queue 'em all
	-- up and hope they happen.
	
	crossReferenceKeysP:whenResolvedDo( function( status, data )
	
		for k in pairs( unreferencedKeysFromCacheAsSet ) do
		
			-- FIXME [scouten 2011-09-01, PR #2956100]: I suspect a race condition
			-- under which a newly-created rendition could get purged immediately
			-- after it's been added because the index store hasn't yet caught up
			-- to its existence.

			-- "Fix" for now by space out the purge intervals a bit more (now 40
			-- seconds) and only delete a cache entry when it appears unreferenced in
			-- two successive sweeps.
			
			-- Marking as FIXME because this is basically making the race condition
			-- (much) more difficult to repro, not decisively eliminating it. A better
			-- fix will be "interesting" because it probably entails providing some
			-- atomic commit protocol between both the data cache and the doc DB.
			-- Not something I want to try with 3 hours to GM.
		
			if previouslyUnreferenced[ k ] then
				dataCache:removeP( k )
				previouslyUnreferenced[ k ] = nil
			else
				previouslyUnreferenced[ k ] = true
			end
			
		end

	end ):runUntilDone()

	-- FIXME [scouten 2011-08-11]: Should also fix up references to
	-- missing data in cache. Be careful not to toss references to
	-- originals when doing so.
	
	return crossReferenceKeysP

end

--------------------------------------------------------------------------------

freeSpaceIfNeededP = function( self )

	local indexStore = self._indexStore

	local spaceNeededP = self._dataCachePathP
		:mapDataP( function( _, path )
		
			local fileSystemFreeSpace, _ -- fileSystemSize
				= KSFileUtils.checkFreeSpaceAtPath( path )

			local amountUnderMinimumFreeSpace = minimumFreeSpace - fileSystemFreeSpace

			local spaceNeeded = amountUnderMinimumFreeSpace
			
			-- If we need to purge all, we kick off the purging by setting space
			-- needed to whatever is the current minimum. The logic in purgeVictimsInListP
			-- ensures that we ignore this value if purgeAll is enabled and only
			-- stop the purge once all the victims have been eliminated.

			if self._purgeAllB:valueNow() then
				spaceNeeded = minimumFreeSpace
			end
						
--			purgeLog:info( "cache path = %s\n\nfree space = %s\n\nFS size = %s\n\ncacheFileSize = %s\n\ntargetCacheSize = %s\n\nspaceNeeded = %s",
--						   path, fileSystemFreeSpace, fileSystemSize, cacheFileSize,
--						   targetCacheSize, spaceNeeded )
			
			return 'data', spaceNeeded
		
		end )
	
	local potentialVictimsAndSpaceNeededP = spaceNeededP
		:mapDataP( function( _, spaceNeeded )
	
			if spaceNeeded <= 0 then
				return 'data', {}, 0
			end
			
			-- Note that we do not offer thumbnails up as potential space victims.

			local downloadOriginals = KSPrefs.DownloadOriginals
			local resultP
						
			if self._purgeAllB:valueNow() then
				
				-- We select all eligible renditions (successfully uploaded to Oz) if purge all is enabled.

				if downloadOriginals then
			
					-- Exclude originals and latest originals from the cache victim set if download originals is on
					-- FIXME [ngautam 2011-11-14] Need to add proxies to these queries once we start using those
					
					resultP = indexStore:queryP( [[SELECT docId, renditionName, md5, size FROM cacheReferences
									WHERE hasUploaded == 1
										AND renditionName == 'preview' LIMIT 100]] )

				else

					resultP = indexStore:queryP( [[SELECT docId, renditionName, md5, size FROM cacheReferences
									WHERE hasUploaded == 1
										AND renditionName != 'thumbnail'
										AND renditionName != 'thumbnail2x' LIMIT 100]] )

				end

			else
				
				-- If download originals is enabled, don't purge originals.

				if downloadOriginals then

					-- Exclude originals and latest originals from the cache victim set
					-- if download originals is on.

					resultP = indexStore:queryP( [[SELECT docId, renditionName, md5, size FROM cacheReferences
									WHERE epoch < ( SELECT max( epoch ) FROM cacheReferences )
										AND hasUploaded == 1
										AND renditionName == 'preview'
									ORDER BY epoch ASC, size DESC LIMIT 10]] )

				else
					resultP = indexStore:queryP( [[SELECT docId, renditionName, md5, size FROM cacheReferences
									WHERE epoch < ( SELECT max( epoch ) FROM cacheReferences )
										AND hasUploaded == 1
										AND renditionName != 'thumbnail'
										AND renditionName != 'thumbnail2x'
									ORDER BY epoch ASC, size DESC LIMIT 10]] )
				end
			
			end

			return resultP
				:mapDataP( function( _, victimList )
				
					if self._purgeAllB:valueNow() and #victimList == 0 then

						-- If we've purged all we could, set purgeAllB to false.

						self._purgeAllB:set( false )

					end

					return 'data', victimList, spaceNeeded

				end )
		
		end )
	
	potentialVictimsAndSpaceNeededP:whenResolvedDo( function( status, victims )
	
		purgeLog:info( "potentialVictimsP returned with status = %s victims = %s",
			status, victims )
	
	end ):attachTo( self )
	
	return self:purgeVictimsInListP( potentialVictimsAndSpaceNeededP, 1 )
	
end

--------------------------------------------------------------------------------

local emptyResolvedPromise = KSPromises.resolvedPromiseP( 'data', nil )

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeVictimsInListP( potentialVictimsAndSpaceNeededP,
												startFromIndex,
												overrideSpaceNeeded )

	return potentialVictimsAndSpaceNeededP
		:mapDataP( function( _, potentialVictims, spaceNeeded )
		
			if overrideSpaceNeeded then
				spaceNeeded = overrideSpaceNeeded
			end

			purgeLog:info( "purgeVictimsInListP\n\npotentialVictims = %s\n\nstartFromIndex = %s\n\noverrideSpaceNeeded = %s",
						   potentialVictims, startFromIndex, spaceNeeded )
		
			if spaceNeeded <= 0 and not self._purgeAllB:valueNow() then
				return 'data', true, 0 -- we're done
			end
		
			local itemAtIndex = potentialVictims and potentialVictims[ startFromIndex ]
			
			if not itemAtIndex then
				return 'data', true, spaceNeeded -- we've exhausted the victim pool
			end
			
			local filePath = ( itemAtIndex[ 2 ] == 'original' or itemAtIndex[ 2 ] == 'latestOriginal' )
				and itemAtIndex[ 3 ]
			
			local resultP = emptyResolvedPromise
			
			if filePath and self._lockedFilePaths[ filePath ] then
			
				purgeLockLog:warn( "Not purging %s for photo %s (path %s) due to file lock",
								   itemAtIndex[ 2 ], itemAtIndex[ 1 ], filePath )
				
			else

				local renditionName = itemAtIndex[ 2 ]
					-- Hard-coded indices are based on potentialVictimsAndSpaceNeededP
					-- query in :freeSpaceIfNeededP methods above.
				
				local overlay = {
					_localOnly = {
						[ renditionName ] = { _delete = true }
					}
				}
				
				-- Also remove the local path so that the original can be downloaded again
				-- if the user turns store originals on.
	
				if renditionName == 'original' then
					overlay._localOnly.originalFilePath = { _delete = true }
				end
				
				-- Delete reference in document. If, for some reason, we miss
				-- the following steps, we'll clean out the data cache entry in a
				-- subsequent unreferenced item pass.
				
				local removeReferenceP = self._docStore:applyOverlayToDocumentP {
					docType = 'asset',
					docId = itemAtIndex[ 1 ],
					overlay = overlay,
				}
				
				-- Now remove the entry from cache.
				
				resultP = removeReferenceP
					:mapDataP( function()
	
						if itemAtIndex[ 2 ] == 'original' or itemAtIndex[ 2 ] == 'latestOriginal' then
							KSFileUtils.delete( itemAtIndex[ 3 ] )
							return emptyResolvedPromise
						else
							return self._dataCache:removeP( itemAtIndex[ 3 ] )
						end
	
					end )
				
			end
			
			-- Now chain forward and remove the next victim if need be.
			
			local purgeAnotherItemP = resultP
				:mapDataP( function()
				
					startFromIndex = startFromIndex + 1
					spaceNeeded = spaceNeeded - itemAtIndex[ 4 ]
					
					return self:purgeVictimsInListP( potentialVictimsAndSpaceNeededP,
										startFromIndex, spaceNeeded )
				
				end )
			
			return purgeAnotherItemP
			
		end )

end

--------------------------------------------------------------------------------

local function moveOriginalsToCaches( originalFilePath )

	local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
	local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'

	local originalFilePath_dirSub = originalFilePath:sub( 1, #documentsDir )
	local newOriginalFilePath = originalFilePath
			
	if originalFilePath_dirSub == documentsDir then
		newOriginalFilePath = string.format( '%s%s', cachesDir, originalFilePath:sub( #documentsDir + 1 ) )
	end
	
	KSFileUtils.mkdir( KSPathUtils.parent( newOriginalFilePath ) )
	newOriginalFilePath = KSFileUtils.moveToUniquelyNamedFile( originalFilePath, newOriginalFilePath )
	
	return newOriginalFilePath
	
end

--------------------------------------------------------------------------------

moveOriginalsFromDocumentsToCachesP = function( self )
	
	local downloadOriginals = KSPrefs.DownloadOriginals
	
	if downloadOriginals then
		return emptyResolvedPromise
	end
	
	-- Find eligible originals to move from Documents to the Caches folder.
	
	local resultP = self._indexStore:queryP( [[SELECT photoId, originalFilePath, latestOriginalFilePath FROM currentOriginalsInDocuments
											 LIMIT 10]] )
									
	return resultP:mapDataP( function ( status, rows )
	
		purgeLog:debug( "Query returned %s candidates for moving to caches",
						#rows )
	
		-- For each of the selected rows, move the base and/or latest originals to
		-- caches and update the doc value to reflect the change.
			
		for i = 1, #rows do
		
			local photoId = rows[ i ][ 1 ]
			local originalFilePath = rows[ i ][ 2 ]
			local latestOriginalFilePath  = rows[ i ][ 3 ]
			
			local originalStatusOverlay = {
					_localOnly = {
						original = {
						},
					}
				}
			
			if originalFilePath then

				if KSFileUtils.exists( originalFilePath ) then

					if self._lockedFilePaths[ originalFilePath ] then
					
						purgeLockLog:info( "Not moving base original for photo %s (path %s) to caches due to file lock",
										   photoId, originalFilePath )
					
					else

						purgeLog:debug( "Moving base original for photo %s to caches",
										photoId )
				
						local newOriginalFilePath = moveOriginalsToCaches( originalFilePath )
									
						originalStatusOverlay._localOnly.originalFilePath = newOriginalFilePath
						originalStatusOverlay._localOnly.original.isInCacheFolder = true
					
					end

				else

					-- The file is missing, delete the section for the original and originalFilePath.

					originalStatusOverlay._localOnly.originalFilePath = { _delete = true }
					originalStatusOverlay._localOnly.original = { _delete = true }

				end

			end
			
			if latestOriginalFilePath then
			
				if KSFileUtils.exists( latestOriginalFilePath ) then
			
					if self._lockedFilePaths[ latestOriginalFilePath ] then
					
						purgeLockLog:info( "Not moving latest original for photo %s (path %s) to caches due to file lock",
										   photoId, latestOriginalFilePath )

					else

						originalStatusOverlay._localOnly.latestOriginal = {}
				
						purgeLog:debug( "Moving latest original for photo %s to caches",
										photoId )
				
						local newLatestOriginalFilePath = moveOriginalsToCaches( latestOriginalFilePath )
								
						originalStatusOverlay._localOnly.latestOriginal.filePath = newLatestOriginalFilePath
						originalStatusOverlay._localOnly.latestOriginal.isInCacheFolder = true
					
					end
					
				else
					
					-- The file is missing, delete the section for the latest original.

					originalStatusOverlay._localOnly.latestOriginal = { _delete = true }
					
				end

			end
										
			self._docStore:applyOverlayToDocumentP {
				docType = 'asset',
				docId = photoId,
				overlay = originalStatusOverlay,
			}:runUntilDone()

		end
		
		return 'data', true
	
	end )

end

--------------------------------------------------------------------------------

return KSClass 'WFOzPurgeAgent'( purgeAgentMethods )
