--[[----------------------------------------------------------------------------

WFOzPurgeAgentManager.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

@reviewed scouten 2012-05-14

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFRequestDispatcher = require 'WFRequestDispatcher'

	-- wichitafoundation.core.reactive
local KSReactive = require 'KSReactive'

	-- wichitafoundation.core.utils
local WFQueueing = require 'WFQueueing'

	-- wichitafoundation.os.prefs
local KSPrefs = require 'KSPrefs'

	-- wichitafoundation.oz.purge_agent
local WFOzPurgeAgent = require 'WFOzPurgeAgent'


--------------------------------------------------------------------------------

local purgeLog = KSLogging.log 'WFOzPurgeAgentLog'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--
--[[!
	\brief The \c WFOzPurgeAgentManager namespace coordinates the cache purge agents
	for all catalogs.

	\author Eric Scouten <scouten@adobe.com> 2011-08-11
]]--

-- @mark WFOzPurgeAgentManager

local WFOzPurgeAgentManager = {}

--------------------------------------------------------------------------------

local activePurgeAgentBehaviors = {}
	-- key: (string) catalogId
	-- value: (B< nil or WFOzPurgeAgent >) cache purger for this catalog

	-- NOTE: This is intentionally not a weak map. Once referenced, the
	-- behaviors stay present indefinitely so we can do mapB and mapSwitchB
	-- on them. We do clear out the agent itself.
	
local activePurgeAllBehaviors = {}
	-- key: (string) catalogId
	-- value: (B< Boolean >) tracks if there is an active (queued/executing) purge all
	-- request for a catalog

local enqueue, dequeue = WFQueueing.makeQueue()
	-- queue for maintaining the purge all requests

local currentPurgeAllRequestP = false

local currentPurgeAllCatalogId = nil

--------------------------------------------------------------------------------

local function purgeAgentBehaviorForCatalogId( catalogId )

	local resultB = activePurgeAgentBehaviors[ assert( catalogId ) ]
	
	if not resultB then
	
		resultB = KSReactive.variableB( nil )
		activePurgeAgentBehaviors[ catalogId ] = resultB
	
	end
	
	return resultB

end

--------------------------------------------------------------------------------

local function purgeAllBehaviorForCatalogId( catalogId )

	local resultB = activePurgeAllBehaviors[ assert( catalogId ) ]
	
	if not resultB then
	
		resultB = KSReactive.variableB( false )
		activePurgeAllBehaviors[ catalogId ] = resultB
	
	end
	
	return resultB

end


--============================================================================--

-- @mark purge agent management

--------------------------------------------------------------------------------

local function startPurgeAgentForCatalogId( catalogId )
			
	assert( type( catalogId ) == 'string',
		"startPurgeAgent: catalogId was not a string" )
	
	local purgeAgentB = purgeAgentBehaviorForCatalogId( catalogId )
	
	if not purgeAgentB:valueNow() then

		purgeLog:debug( "Creating WFOzPurgeAgent for catalog %s", catalogId )

		purgeAgentB:set( WFOzPurgeAgent( catalogId ) )

	end
	
end

--------------------------------------------------------------------------------

local function startNewPurgeAllRequest()
	
	local currentCatalogId = dequeue()
				
	if currentCatalogId then
				
		startPurgeAgentForCatalogId( currentCatalogId )

		local purgeAgentB = purgeAgentBehaviorForCatalogId( currentCatalogId )
			
		local purgeAgent = purgeAgentB:valueNow()
			
		if purgeAgent then

			purgeLog:debug( "Purging all eligible originals/previews for catalog %s",
							currentCatalogId )

			-- Set purgeAll to true to force the purge agent to
			-- remove all eligible renditions regardless of space requirements the next time it runs
				
			purgeAgent:setPurgeAllEnabled()
					
			currentPurgeAllRequestP = purgeAgent:purgeFinishedP()
			
			currentPurgeAllRequestP:whenResolvedDo( function( status, data )
			
				if status ~= 'data' then

					-- The purge failed for some reason: Print to console and move on.
					-- If we enqueue this catalog again, we might end up purging and failing
					-- endlessly.

					debug.print(" Purge all failed for catalog" .. currentPurgeAllCatalogId )

				end

				purgeLog:debug( "Completed a purge all for catalog %s",
								currentPurgeAllCatalogId )

				-- Update the purging all behavior for this catalog.
	
				purgeAllBehaviorForCatalogId( currentPurgeAllCatalogId ):set( false )
	
				-- Close the purge agent if this is not the active catalog.
	
				local prefsKey = 'CurrentCatalogFor' .. KSPrefs.currentUserId
	
				if KSPrefs[ prefsKey ] ~= currentPurgeAllCatalogId then
				
					local purgeAgentB = purgeAgentBehaviorForCatalogId( currentPurgeAllCatalogId )
			
					local purgeAgent = purgeAgentB:valueNow()
			
					if purgeAgent then

						purgeLog:debug( "Closing WFOzPurgeAgent for catalog %s ",
										currentPurgeAllCatalogId )

						purgeAgentB:set( nil )

						purgeAgent:close()

					end
				end

				-- Start another purge if need be.
		
				currentPurgeAllRequestP = false
	
				startNewPurgeAllRequest()
	
			end ):runUntilDone()
			
			currentPurgeAllCatalogId = currentCatalogId

		end
					
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Install message handlers for purge agent. Should only be called by
	BABLU thread.
--]]

function WFOzPurgeAgentManager.installMessageHandlers()

	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'startPurgeAgent',
										   startPurgeAgentForCatalogId )
	
	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'stopPurgeAgent',
		function( catalogId )
		
			assert( type( catalogId ) == 'string',
				"stopPurgeAgent: catalogId was not a string" )
			
			local purgeAgentB = purgeAgentBehaviorForCatalogId( catalogId )
			
			local purgeAgent = purgeAgentB:valueNow()
			
			if purgeAgent then

				purgeLog:debug( "Closing WFOzPurgeAgent for catalog %s", catalogId )

				purgeAgentB:set( nil )
					-- Do this first, so behaviors that are mapped from
					-- purgeAgentB don't also get closed.

				purgeAgent:close()

			end
			
		end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'purgeEnabledB', function( catalogId )

		return purgeAgentBehaviorForCatalogId( catalogId ):toBooleanB()
		
	end )
	
	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'isNowOrWillBePurgingB', function( catalogId )
	
		return purgeAllBehaviorForCatalogId( catalogId )
		
	end )

	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'lockFileAtPath', function( catalogId, filePath )
	
		local purgeAgentB = purgeAgentBehaviorForCatalogId( catalogId )
		
		local purgeAgent = purgeAgentB:valueNow()
		
		if purgeAgent then
		
			return purgeAgent:lockFileAtPath( filePath )
		
		else
		
			return true
		
		end

	end )
	
	----------------------------------------------------------------------------

	WFRequestDispatcher.onMessage( 'unlockFileAtPath', function( catalogId, filePath )
	
		local purgeAgentB = purgeAgentBehaviorForCatalogId( catalogId )
		
		local purgeAgent = purgeAgentB:valueNow()
		
		if purgeAgent then
			purgeAgent:unlockFileAtPath( filePath )
		end
		
		return true

	end )
	
	----------------------------------------------------------------------------
	
	WFRequestDispatcher.onMessageNoCancel( 'purgeAllEligible',
		function( catalogId )
		
			-- Add the purge all request for this catalog to the queue.
			
			purgeLog:debug( "Adding a purge all request for catalog %s", catalogId )
			
			enqueue( catalogId )
			
			purgeAllBehaviorForCatalogId( catalogId ):set( true )

			-- If a purge is not active, kick off a purge for this catalog.
				
			if not currentPurgeAllRequestP then

				-- Get a request from the queue and start the purge.

				startNewPurgeAllRequest()
				
			end
			
		end )
		
end

--------------------------------------------------------------------------------

return WFOzPurgeAgentManager
