--[[----------------------------------------------------------------------------

WFOzRenditionServer.lua

@author scouten 2012-10-04

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua
local KSData = require 'KSData'

	-- wichitafoundation.core.lua.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSPrivateKeys = require 'KSPrivateKeys'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- wichitafoundation.os.notifications
local KSSystemNotifications = require 'KSSystemNotifications'

	-- wichitafoundation.oz_client.bablu
local KSMemoryCache = require 'KSMemoryCache'

	-- wichitafoundation.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'
local OzCatHelperFns = require 'OzCatHelperFns'
local OzReferenceDimension = require 'OzReferenceDimension'

	-- wichitafoundation.oz_client.catalog.rendition_server
local WFOzRenditionServerUtils = require 'WFOzRenditionServerUtils'

	-- wichitafoundation.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

local USE_ONE_QUERY_FOR_ALL_MD5 = false

--------------------------------------------------------------------------------

local assert = assert
local check = assert( WFArguments.check )
local KSReactivePromiseBridge_promiseToBehavior = assert( KSReactivePromiseBridge.promiseToBehavior )
local KSTableUtils_makeFullyWeak = assert( KSTableUtils.makeFullyWeak )
local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt )
local OzCatHelperFns_DISABLE_MOST_FETCHES = OzCatHelperFns.DISABLE_MOST_FETCHES
local OzCatHelperFns_processHttpRequestResult = OzCatHelperFns.processHttpRequestResult
local OzReferenceDimension_getReferenceDimensions =
	assert( OzReferenceDimension.getReferenceDimensions )
local WFCloseSupport_fetchCached = assert( WFCloseSupport.fetchCached )
local WFCloseSupport_hasBeenOrWillBeClosed = assert( WFCloseSupport.hasBeenOrWillBeClosed )
local WFOzRenditionServerUtils_touchOzRendition =
	assert( WFOzRenditionServerUtils.touchRendition )

local DEBUG_DOWNLOAD_FETCH_CANCELING = false

local ozLinkNameForOzRenditionName =
	assert( WFOzRenditionServerUtils.ozLinkNameForRenditionName )
local renditionKindForOzRenditionName =
	assert( WFOzRenditionServerUtils.renditionKindForRenditionName )

local now = KSDate.timeNowForAnimation

--------------------------------------------------------------------------------

local renditionServerLog = KSLogging.log 'WFOzRenditionServer'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local jpegDecodeService = WFWorkerThread 'KSJpegDecodeServiceEPInit.lua'

--------------------------------------------------------------------------------

local DOWNLOAD_THUMBNAIL2X_IF_PREVIEW = false
	-- Should we try to download the 2x thumbnail if we have a preview
	-- even if the 2x thumbnail isn't available?
	
local OBEY_DEFER_TRIGGERS = false
	-- Should we obey defer triggers?

--------------------------------------------------------------------------------

-- Device renditions include the small preview version of the preview rendition

local deviceRenditionNameArg = 'WFOzRenditionServer.DeviceRenditionName'

WFArguments.registerChecker( deviceRenditionNameArg, {
	thumbnail = true,
	preview = true,
	master = true,
	thumbnail2x = true,
	smallPreview = true,
} )

--------------------------------------------------------------------------------

-- Oz renditions do not include the small preview

local ozRenditionNameArg = 'WFOzRenditionServer.OzRenditionName'

WFArguments.registerChecker( ozRenditionNameArg, {
	thumbnail = true,
	preview = true,
	master = true,
	thumbnail2x = true,
} )

--------------------------------------------------------------------------------

local ozRenditionNameForDeviceRenditionName = {
	thumbnail = 'thumbnail',
	thumbnail2x = 'thumbnail2x',
	preview = 'preview',
	smallPreview = 'preview',
	master = 'master',
}

--------------------------------------------------------------------------------

local decodeOptionForDeviceRenditionName = {
	smallPreview = { 720, 720 }, -- size limit
}

-- On the iPhone, we want the small previews to be 320 points wide with
-- unlimited height.

if WFPlatformInfo.isiPhone then
	if WFPlatformInfo.isHighDPIDevice then
		decodeOptionForDeviceRenditionName.smallPreview = { 640, 0 }
	else
		decodeOptionForDeviceRenditionName.smallPreview = { 320, 0 }
	end
end

--------------------------------------------------------------------------------

local cascadedBehaviorCacheKeyToDeviceRenditionNames = {
	_previewBCache2 = { 'preview', 'thumbnail2x', 'thumbnail', sequential = false },
	_smallPreviewBCache2 = { 'smallPreview', 'thumbnail2x', 'thumbnail', sequential = true },
	_thumbnailBCache2 = { 'thumbnail2x', 'thumbnail', sequential = true }
}

--------------------------------------------------------------------------------

local thumbnailCacheSize, minimalThumbnailCacheSize, previewCacheSize

if WFPlatformInfo.isMac then
	thumbnailCacheSize = 4000
	minimalThumbnailCacheSize = 4000
	previewCacheSize = 10
elseif WFPlatformInfo.isiOS then
	thumbnailCacheSize = 100
	minimalThumbnailCacheSize = 20
	previewCacheSize = 3
elseif WFPlatformInfo.isAndroid then
	thumbnailCacheSize = 100
	minimalThumbnailCacheSize = 20
	previewCacheSize = 3
else
	-- GRUNGE [arahn 2012-08-28]: What are sensible defaults for other platforms?
	thumbnailCacheSize = 100
	minimalThumbnailCacheSize = 20
	previewCacheSize = 3
end

--------------------------------------------------------------------------------

-- Cache info for renditions. We maintain a KSMemoryCache for images and a fully-weak
-- cache for behaviors.

-- imageCacheName: Name of the cache for decoded images. Indexed by MD5 hash.
-- behaviorCacheName: Name of the cache for fetch behaviors. Indexed by photo id.
-- minimalImageCacheSize: Number of entries to allow in the cache when we minimize
-- standardImageCacheSize: Number of entries to allow in the cache when not minimized

-- We map the rendition names down to a smaller set of rendition kinds so that we can employ
-- common processing where appropriate.

local cacheInfoForDeviceRenditionName = {

	thumbnail = {
		imageCacheName = '_thumbnailCache',
		behaviorCacheName = '_thumbnailBCache',
		minimalImageCacheSize = minimalThumbnailCacheSize,
		standardImageCacheSize = thumbnailCacheSize,
	},

	thumbnail2x = {
		imageCacheName = '_thumbnailCache',
			-- [mhamburg 2012-03-22]
			-- Shared thumbnail cache with the small thumbnails. This is safe
			-- because the cache is indexed by MD5 hash just like the data cache.
			-- Note that this means that we will perform some operations on this cache
			-- twice such as creating it and minimizing it but that's harmless and beats
			-- managing two cahes or having extra logic to decide which cache we are
			-- talking to.
		behaviorCacheName = '_thumbnail2xBCache',
		minimalImageCacheSize = minimalThumbnailCacheSize,
		standardImageCacheSize = thumbnailCacheSize,
	},

	preview = {
		imageCacheName = '_previewCache',
		behaviorCacheName = '_previewCacheB',
		minimalImageCacheSize = previewCacheSize,
		standardImageCacheSize = previewCacheSize,
	},

	smallPreview = {
		imageCacheName = '_previewCache',
		behaviorCacheName = '_smallPreviewCacheB',
		minimalImageCacheSize = previewCacheSize,
		standardImageCacheSize = previewCacheSize,
	},

}

--------------------------------------------------------------------------------

local kDownsampledOzRenditions = {
	preview = true,
	thumbnail = true,
	thumbnail2x = true
}

--------------------------------------------------------------------------------

--[[ FIXME

-- A class for blocking items when we want to wait until we drop below
-- a busy threshold.

local Blocker = KSClass 'Blocker' {

	_counter = 0,
	_threshold = 0,
	_promiseP = false,
	_resolver = false,

	__init = function( self, threshold )
		self._threshold = threshold or 0
	end,
	
	startBlockP = function( self, fnP, ... )

		local p = fnP( ... )
		
		self._counter = self._counter + 1

		p = p:observeP( function()
			self._counter = self._counter - 1
			if self._counter <= self._threshold and self._resolver then
				local resolver = self._resolver
				self._promiseP = false
				self._resolver = false
				resolver( 'data', true )
			end
		end )
		
		return p

	end,
	
	waitForBlockP = function( self, fnP, ... )
		
		if self._counter <= self._threshold then
			return fnP( ... )
		end
		
		local args = { n = select( '#', ... ), ... }
		
		if not self._promiseP then
			self._promiseP, self._resolver = KSPromises.promiseP()
			self._promiseP:runUntilDone() -- Silence the warnings
		end
		
		local p = self._promiseP:mapDataP( function()
			return fnP( unpack( args, 1, args.n ) )
		end )
		
		return p
		
	end

}

]]

--------------------------------------------------------------------------------

local kDiagnosticTimepoints = { "md5QueryResult", "md5", "startDownload", "finishDownload", "deliver" }

--------------------------------------------------------------------------------

local function _fetchFields( t, fields, idx, n )
	if n > idx then
		return t[ fields[ idx ] ], _fetchFields( t, fields, idx + 1, n )
	else
		return t[ fields[ idx ] ]
	end
end
local function fetchFields( t, fields )
	return _fetchFields( t, fields, 1, #fields )
end

--------------------------------------------------------------------------------

local function recordDiagnosticTiming( self, photoId, stage, optTime )
	local requestTiming = self._requestTimings[ photoId ]
	if requestTiming then
		if stage then
			if not requestTiming[ stage ] then
				requestTiming[ stage ] = ( optTime or now() ) - requestTiming.start
			end
		else
			self._requestTimings[ photoId ] = nil
			return
				fetchFields( requestTiming, kDiagnosticTimepoints )
		end
	end
end

--------------------------------------------------------------------------------

-- FIXME [mhamburg 2011-08-30]: Move this into KSReactive as a general utility?

local function fetchCachedBehavior( cache, key, checkObjectProxy )

	local resultB = WFCloseSupport_fetchCached( cache, key )

	if not resultB then
		return nil
	end
	
	local value = resultB:valueNow()
	
	if not value then
		return resultB
	end
	
	-- If the value has been closed then we probably wanted a cache miss.
	
	if WFCloseSupport_hasBeenOrWillBeClosed( value ) then
		cache[ key ] = nil
		return nil
	end
	
	-- If we are checking object proxy closure (due to GC, we can
	-- still find ourselves needing to report as closed).
	
	-- FIXME [mhamburg 2011-08-30]: This really ought to fold together with
	-- fetchCached.
	
	if checkObjectProxy and value:proxyHasBeenOrWillBeClosed() then
		cache[ key ] = nil
		return nil
	end
	
	-- Everything looks okay (at least from a quick inspection).
	
	return resultB

end

--------------------------------------------------------------------------------

local function getBehaviorCache( self, key )

	local cache = self[ key ]
	
	if not cache then
		cache = KSTableUtils_makeFullyWeak()
		self[ key ] = cache
	end
	
	return cache

end

--------------------------------------------------------------------------------

local function flushCaches( self )
	
	for _, info in pairs( cacheInfoForDeviceRenditionName ) do
		self[ info.imageCacheName ]:flush()
		self[ info.behaviorCacheName ] = false
	end
	
	for cacheName in pairs( cascadedBehaviorCacheKeyToDeviceRenditionNames ) do
		self[ cacheName ] = false
	end
	
end


--============================================================================--
--[[!
	\class WFOzRenditionServer

	\brief This class provides renditions (such as thumbnails and previews)
		to Oz clients. It manages the local cache of renditions and triggers
		downloads from the Oz server when necessary.
--]]

-- @mark WFOzRenditionServer class

local renditionServerMethods = {

	_catalogId = false,
		-- (string) catalog ID (currently used only for diagnostic log)

	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)
	
	_ozUser = false,
		-- (OzUser)
	
	_cacheEpochB = false,
		-- (B< number >) cache epoch

	_didEnterBackgroundNotificationE = false,
		-- (E<>) which fires when the app moves into the background
	
	_didReceiveMemoryWarningNotificationE = false,
		-- (E<>) which fires when the app receives a low memory warning

	_processHttpRequestResultFn = false,
		-- (function) call to process HTTP request results and update
		-- access status as appropriate
	
	_recentPreviews = false,
		-- (table) most recent preview behaviors

	_velocityBuffer = false,
		-- diagnostic ring buffer for watching how long thumbnail requests
		-- take to fulfill
	
	_requestTimings = false,
		-- weigh-points for requests as they move throught the pipe
		
	-- _activeThumbnailDataCacheFetchesBlocker = false,
		-- Blocker. Disallow other data cache fetches while we have thumbnail
		-- fetches active.
		
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

for _, info in pairs( cacheInfoForDeviceRenditionName ) do

	-- Can't assert here -- thumbnail and thumbnail2x share the same
	-- imageCacheName. (Intentional.)

	-- assert( renditionServerMethods[ info.imageCacheName ] == nil, "image cache item already defined" )
	assert( renditionServerMethods[ info.behaviorCacheName ] == nil, "behavior cache item already defined" )
	
	renditionServerMethods[ info.imageCacheName ] = false -- KSMemoryCache of images
	renditionServerMethods[ info.behaviorCacheName ] = false -- fully weak cache of behaviors

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new rendition server for a catalog.
	
	\param catalog (OzCatalogService) The catalog.
		
--]]

function renditionServerMethods:__init( --[[OzCatalogService]] catalog )

	check( 'WFOzRenditionServer', 'OzCatalogService|OzCatalogServiceImp' )
	
	self._catalogId = assert( catalog._catalogId )
	self._docStore = assert( catalog._docStore )
	self._indexStore = assert( catalog._indexStore )
	self._ozUser = assert( catalog._ozUser )
	
	self._recentPreviews = {}
	
	-- self._activeThumbnailDataCacheFetchesBlocker = Blocker()
	
	catalog:addCloseTarget( self )
	
	self._cacheEpochB = WFOzRenditionServerUtils.cacheEpochB( self._docStore,
															  self._indexStore )
	
	self:addCloseTarget( self._cacheEpochB )

	self._processHttpRequestResultFn = function( ... )
		return OzCatHelperFns_processHttpRequestResult( catalog, ... )
	end

	for _, info in pairs( cacheInfoForDeviceRenditionName ) do
		self[ info.imageCacheName ] = KSMemoryCache( info.standardImageCacheSize )
	end
	
	for cacheName in pairs( cascadedBehaviorCacheKeyToDeviceRenditionNames ) do
		rawset( self, cacheName, false )
	end

	-- Purge caches when going into background and when we receive
	-- memory warnings.

	self._didEnterBackgroundNotificationE = KSSystemNotifications.notificationStreamE( 'didEnterBackground' )
	self._didReceiveMemoryWarningNotificationE = KSSystemNotifications.notificationStreamE( 'didReceiveMemoryWarning' )
	
	self._didEnterBackgroundNotificationE:onDataE( function() flushCaches( self ) end )
	self._didReceiveMemoryWarningNotificationE:onDataE( function() flushCaches( self ) end )
	
	self:addCloseTarget( self._didEnterBackgroundNotificationE )
	self:addCloseTarget( self._didReceiveMemoryWarningNotificationE )

	WFDiagnosticLog.addLogProvider( self )

	self._velocityBuffer = WFDiagnosticLog.makeRingBuffer( kDiagnosticTimepoints )
	self._requestTimings = {}
end

--------------------------------------------------------------------------------
--[[!
	\brief Flushes all the 'preview' type caches, and drives garbage collection.
]]--

function renditionServerMethods:flushPreviewCache()

	for deviceRenditionName, info in pairs( cacheInfoForDeviceRenditionName ) do
	
		local ozRenditionName = ozRenditionNameForDeviceRenditionName[ deviceRenditionName ]

		if renditionKindForOzRenditionName[ ozRenditionName ] == 'preview' then
			self[ info.imageCacheName ]:flush()
		end

	end

	-- We've just dumped a few strong references to sizeable image data.
	-- Drive the GC.

	collectgarbage()
	collectgarbage()

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets whether to use a minimal thumbnail cache.
	
	When transitioning to a minimal thumbnail cache, the call triggers garbage collection
	in order to reclaim any image data that has been made eligible for collection.
	
	\param minimal (boolean) The new value for the preference, \c true to use minimal caching.
]]--

function renditionServerMethods:setUseMinimalThumbnailCache( --[[boolean]] minimal )

	for deviceRenditionName, info in pairs( cacheInfoForDeviceRenditionName ) do
	
		local ozRenditionName = ozRenditionNameForDeviceRenditionName[ deviceRenditionName ]

		if renditionKindForOzRenditionName[ ozRenditionName ] == 'thumbnail' then
			self[ info.imageCacheName ]:adjustMaximumEntries(
				minimal and info.minimalImageCacheSize or info.standardImageCacheSize )
		end

	end

	if minimal then

		-- We've just dumped a bunch of strong references to image data.
		-- Drive the GC.

		collectgarbage()
		collectgarbage()

	end

end
local kIndexForRendition = {
	thumbnail = 1,
	thumbnail2x = 2,
	preview = 3,
	smallPreview = 4,
	master = 5,
}

--------------------------------------------------------------------------------

local function fetchMd5( self, photoId, ozRenditionName )
	
	local resultB

	if USE_ONE_QUERY_FOR_ALL_MD5 then
		table.insert( self._assetIds, photoId )
		resultB = KSReactive.variableB()
		resultB:addCloseAction( function()
			
		end )
		if self._currentMd5sB then
			self._currentMd5sB:close()
		end
		self._currentMd5sB = self._indexStore:queryB {
			sql = [[SELECT
				thumbnail, thumbnail2x, preview, smallPreview, master, coreMD5.docId
			FROM
				coreMD5
			INNER JOIN
				TEMP_ids
			ON
				coreMD5.docId = TEMP_ids.docId]],
			tempIds = self._assetIds
		}
		self._currentMd5sB:onDataB( function( rows )
			recordDiagnosticTiming( self, photoId, 'md5QueryResult' )
			for i, row in ipairs( rows ) do
				local nextAssetId = row[ 6 ]
				local nextB = self._md5sForPhotoIdB[ nextAssetId ]
				if nextB then
					nextB:set( row )
				end
			end
		end )

	else
		local queryB = self._indexStore:queryB {
			sql = [[SELECT
				thumbnail, thumbnail2x, preview, smallPreview, master
			FROM
				coreMD5
			WHERE
				docId = ?
			]],
			photoId
		}
		resultB = queryB:mapB( function( rows )
			recordDiagnosticTiming( self, photoId, 'md5QueryResult' )
			return KSTableUtils_derefOpt( rows, 1 )
		end )
		resultB:addCloseTarget( queryB )
	end
	
	return resultB
	
end

--------------------------------------------------------------------------------

local function md5ForOzRenditionB( self, photoId, md5hint, ozRenditionName,
								   optDiagnosticLogEntry )

	-- If we have an MD5 hint, then construct a behavior that starts as the
	-- hint and then becomes the fetch.
	
	if md5hint then
	
		-- Timer
		--		> md5ForRendition behavior (initial value and subsequent values)
		--		> filter nil
		--		start with the hint value
		
		local md5E = KSReactive.timerE( 0.5 )
			:closingMapSwitchE( function()

				if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
					return nil
				end

				return md5ForOzRenditionB( self, photoId, nil, ozRenditionName )

			end )
			
		local md5B = md5E
				:filterE( function( md5 ) return md5 ~= nil end )
				:holdB( md5hint )
				:addCloseTarget( md5E )
				
		return md5B
	
	end

	return fetchMd5( self, photoId, ozRenditionName )
	:closingMapB( function( md5s )

		local md5 = md5s and md5s[ kIndexForRendition[ ozRenditionName ] ]
		
		-- Side effect: Touch this rendition.
		
		if md5 then
			if optDiagnosticLogEntry then
				optDiagnosticLogEntry.md5 = md5
				optDiagnosticLogEntry.time_md5 = KSDate.timeInIso8601Format( nil, true )
			end
		
			WFOzRenditionServerUtils_touchOzRendition( self._docStore,
													   self._cacheEpochB,
													   photoId, ozRenditionName )
	
			return md5
		end
	
	end )

end

--------------------------------------------------------------------------------

local function shouldDownloadOzRendition( docValue, photoId, ozRenditionName )

	renditionServerLog:debug( "shouldDownloadOzRendition sniffing docValue %s", docValue )

	-- If we're still waiting for docValue, punt for now.

	if not docValue then
		renditionServerLog:debug( "shouldDownloadOzRendition %s %s = false: no docValue",
								  photoId, ozRenditionName )
		return false
	end

	-- Is local version current? If so, nothing else matters.
	
	local localDescription = KSTableUtils_derefOpt( docValue, '_localOnly', ozRenditionName  )

	local localRenditionIsForRev = localDescription and localDescription.revId
	local localRenditionId = localDescription and localDescription.renditionId
	
	-- Extract the revision ID from the server link for this rendition.
	
	local ozLinkName = ozLinkNameForOzRenditionName[ ozRenditionName ]

	local serverRevisionId = KSTableUtils_derefOpt( docValue,
													'_serverProvidedMetadata',
													'links',
													ozLinkName,
													'href' )

	local serverRevisionInvalid = KSTableUtils_derefOpt( docValue,
													'_serverProvidedMetadata',
													'links',
													ozLinkName,
													'invalid' )
			
	local serverRenditionId
	
	if serverRevisionId then
		serverRevisionId, serverRenditionId = serverRevisionId:match( '([0-9a-f]+)%/renditions%/([0-9a-f]+)$' )
	end
	
	if localDescription
		and localDescription.revId
		and localDescription.md5
		and localDescription.renditionId
	then

		if docValue
			and docValue._ozMetadataUploadStatus == 'uploaded'
			and ( localRenditionIsForRev ~= serverRevisionId or localRenditionId ~= serverRenditionId )
			and ( not serverRevisionInvalid )
		then
			return true
		end

		renditionServerLog:debug( "shouldDownloadOzRendition %s %s = false: local rev is valid",
								  photoId, ozRenditionName )
		return false

	else

		renditionServerLog:debug( "shouldDownloadOzRendition %s %s rejected local rev based on"
								  .. " serverRevisionId = %s and localRenditionIsForRev = %s"
								  .. " and localRenditionId = %s",
								  photoId, ozRenditionName, serverRevisionId,
								  localRenditionIsForRev, localRenditionId )

	end

	-- Local isn't current: Download version from Oz unless Oz thinks it's invalid.
	
	local serverLinks = KSTableUtils_derefOpt( docValue, '_serverProvidedMetadata', 'links' )
	local serverDescription = KSTableUtils_derefOpt( serverLinks, ozLinkName )
	
	-- If we lack a server description for the 2x thumbnails, we will still try
	-- to download them if the server has a valid preview. Note that an invalid 2x
	-- thumbnail on the server will not result in a download.
	
	if DOWNLOAD_THUMBNAIL2X_IF_PREVIEW then

		if not serverDescription and ozRenditionName == 'thumbnail2x' then
			local previewLinkName = ozLinkNameForOzRenditionName[ 'preview' ]
			serverDescription = KSTableUtils_derefOpt( serverLinks, previewLinkName )
		end

	end

	local result = serverDescription and not serverDescription.invalid

	renditionServerLog:debug( "shouldDownloadOzRendition %s %s = %s: based on SD = %s",
							  photoId, ozRenditionName, result, serverDescription )

	return result

end

--------------------------------------------------------------------------------

local function shouldDownloadOzRenditionB( self, photoId, ozRenditionName )

	-- Returns a behavior that turns true if the rendition we have stored for
	-- this photo should be replaced by one from Oz.
	
	local docValueB = self._docStore:docValueB( 'asset', photoId )
	
	local resultB = docValueB:closingMapB( function( docValue )

		return shouldDownloadOzRendition( docValue, photoId, ozRenditionName )

	end )
	
	return resultB

end

--------------------------------------------------------------------------------

local function getCachedImage( imageCache, deviceRenditionName, key )

	if not imageCache or not key or not deviceRenditionName then
		return nil
	end
	
	key = string.format( "%s:%s", deviceRenditionName, key )

	local cachedImage = imageCache:getData( key )
	
	if not cachedImage then
		return nil
	end
	
	if WFCloseSupport_hasBeenOrWillBeClosed( cachedImage ) then
		return nil
	end
	
	if cachedImage:proxyHasBeenOrWillBeClosed() then
		return nil
	end
	
	return cachedImage

end

--------------------------------------------------------------------------------

local function storeCachedImage( imageCache, deviceRenditionName, key, value )

	if not imageCache or not key or not value or not deviceRenditionName then
		return
	end
	
	key = string.format( "%s:%s", deviceRenditionName, key )
	
	imageCache:add( key, value )

end

--------------------------------------------------------------------------------

local downloadPromiseCache = KSTableUtils_makeFullyWeak()
	-- A mapping from rendition URL to promise for its download
	-- We cache here rather than in the HTTP client because the HTTP client
	-- moves to aggressively close the data as it returns it which makes it
	-- difficult to return the same value from multiple requests.

--------------------------------------------------------------------------------

local function downloadOzRenditionForUrlP( self, ozRenditionName, renditionUrl )

	local cachedP = downloadPromiseCache[ renditionUrl ]
	
	if cachedP and not cachedP:isCanceled() then
		return cachedP
	end

	local sequencing = ( ozRenditionName == 'thumbnail' ) and 'LIFO-High' or 'LIFO-Low'

	local p = self._ozUser:getP( renditionUrl, nil, sequencing )
	
	downloadPromiseCache[ renditionUrl ] = p
	
	return p

end

--------------------------------------------------------------------------------

local function downloadOzRenditionP( self, photoId, ozRenditionName, optDiagnosticLogEntry )

	check( 'WFOzRenditionServer', 'string', ozRenditionNameArg )

	renditionServerLog:info( "downloadOzRenditionP %s %s", photoId, ozRenditionName )

	local docValueP = self._docStore:docValueOnceP( 'asset', photoId )
	
	local renditionUrl, devSettingsLastChanged
	
	local downloadP = docValueP:mapDataP( function( status, docValue )
		
		renditionServerLog:info( "downloadOzRenditionP %s %s got docValue %s",
								 photoId, ozRenditionName, docValue )
		
		local ozLinkName = ozLinkNameForOzRenditionName[ ozRenditionName ]

		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.ozLinkName = ozLinkName
			optDiagnosticLogEntry.time_ozLinkName = KSDate.timeInIso8601Format( nil, true )
		end

		local partialRenditionUrl = KSTableUtils_derefOpt( docValue,
														   '_serverProvidedMetadata',
														   'links',
														   ozLinkName,
														   'href' )
		
		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.partialRenditionUrl = partialRenditionUrl
		end

		devSettingsLastChanged = KSTableUtils_derefOpt( docValue,
													    '_localOnly',
														'devSettingsLastSetAtRev' )

		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.devSettingsLastChanged = devSettingsLastChanged
		end

		if partialRenditionUrl then

			-- If the url from the server is partial, only then append the base. ASSET documents don't 
			-- contain the base server url, it would be nice to have that.
			
			renditionUrl = OzCatHelperFns.checkForRelativeUrlAndAppendBase( partialRenditionUrl, OzCatalogUrls.baseServerUrl )
			
			if optDiagnosticLogEntry then
				optDiagnosticLogEntry.renditionUrl = renditionUrl
			end
	
			renditionServerLog:info( "renditionUrl = %s", renditionUrl )

			local resultP = downloadOzRenditionForUrlP( self, ozRenditionName, renditionUrl )
			
			resultP = resultP:observeP( function( status, data, headers )
			
				if optDiagnosticLogEntry then

					optDiagnosticLogEntry.renditionUrlStatus = status
					optDiagnosticLogEntry.renditionUrlHeaders = headers
					optDiagnosticLogEntry.time_renditionUrlFetch = KSDate.timeInIso8601Format( nil, true )
					
					if status == 'error' then
						local result, message = pcall( KSJsonDecoder.decodeWebSafeJson, data )
						if result and message then
							optDiagnosticLogEntry.renditionUrlMessage = message
						end
					end

				end
		
				self:_processHttpRequestResultFn( status, data, headers )
				
			end )

			return resultP
		
		else
		
			if docValue then
				renditionServerLog:info( "no image URL in doc data RN = %s\n\ndocValue = %s",
											ozRenditionName, docValue )
			end

			return 'data', nil
		
		end
	
	end )
	
	local writeToCacheP = downloadP:mapDataP( function( status, data )
	
		if not data then return end
	
		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.renditionReceivedSize = KSData.length( data )
		end

		renditionServerLog:debug( "got data for URL %s %d bytes :: writing to cache",
								  renditionUrl, KSData.length( data ) )

		if kDownsampledOzRenditions[ ozRenditionName ] then

			local renditionMD5 = data:digest( 'MD5' )

			local referenceWidth, referenceHeight = OzReferenceDimension_getReferenceDimensions( data )
			
			-- Note that the data cache put is treated as a runUntilDone operation
			-- since once we have the data, we really don't want to fetch it again.
				
			local resultP = self._docStore:dataCache():putP( renditionMD5, data )
				:mapP( function( status, message )
			
					if status ~= 'data' then
				
						if optDiagnosticLogEntry then
							optDiagnosticLogEntry.cacheWriteFailedWithMessage = message
						end

						renditionServerLog:error( "failed to write %s data to cache for photo %s: %s",
												  ozRenditionName,
												  photoId,
												  message )

						return status, message
					
					else
				
						local revId, renditionId
						
						if renditionUrl then
							revId, renditionId = renditionUrl:match( '([0-9a-f]+)%/renditions%/([0-9a-f]+)$' )
						end
					
						local overlay = {
							_localOnly = {
								[ ozRenditionName ] = {
									md5 = renditionMD5,
									revId = revId,
									renditionId = renditionId,
									epoch = self._cacheEpochB:valueNow(),
									size = KSData.length( data ),
									referenceWidth = referenceWidth,
									referenceHeight = referenceHeight
								}
							}
						}
				
						-- GRUNGE [scouten 2011-08-04]: Ideally we'd be getting this notion
						-- of when dev settings were last changed from Oz, but it isn't there
						-- now.
					
						if devSettingsLastChanged == 'reverted' then
							overlay._localOnly.devSettingsLastSetAtRev = revId
						end

						if optDiagnosticLogEntry then
							optDiagnosticLogEntry.cacheWriteOverlay = overlay
							optDiagnosticLogEntry.time_cacheWrite = KSDate.timeInIso8601Format( nil, true )
						end
				
						renditionServerLog:debug( "data for URL %s written to cache under MD5 key %s, "
												  .. "now applying overlay %s to photo %s",
												  renditionUrl, renditionMD5, overlay, photoId )

						local resultP = self._docStore:applyOverlayToDocumentP {
							docType = 'asset',
							docId = photoId,
							overlay = overlay,
						}
				
						return resultP:mapDataP( function()

							if optDiagnosticLogEntry then
								optDiagnosticLogEntry.time_updateDocValue = KSDate.timeInIso8601Format( nil, true )
							end

							return 'data', data

						end )
				
					end
			
				end )
			
			return resultP:runUntilDoneP()

		else
			
			return 'data', nil
		
		end

	end )
	
	return writeToCacheP

end

--------------------------------------------------------------------------------

local function suppressNilB( sourceB )

	local resultB = sourceB
		:filterE( function( v ) return v ~= nil end )
		:holdB( sourceB:valueNow() )
			
	resultB:addCloseTarget( sourceB )
	
	return resultB

end

--------------------------------------------------------------------------------

local keyDownload = KSPrivateKeys.makePrivateKey 'WFOzRenditionServer.downloadRendition'

--------------------------------------------------------------------------------

local kShouldLogRendition = {
	preview = true,
	smallPreview = true
}

--------------------------------------------------------------------------------

-- Set up diagnostic log reporting, but only for previews.

local function logRenditionFetch( self, photoId, md5hint, deviceRenditionName )

	local optDiagnosticLogEntry
		
	if kShouldLogRendition[ deviceRenditionName ] then
		
		optDiagnosticLogEntry = {
			photoId = photoId,
			md5hint = md5hint,
			renditionName = deviceRenditionName,
			time_00 = KSDate.timeInIso8601Format( nil, true ),
		}
		
		local recentPreviews = self._recentPreviews
		recentPreviews[ #recentPreviews + 1 ] = optDiagnosticLogEntry
		if #recentPreviews > 10 then
			table.remove( recentPreviews, 1 )
		end
		
	end

	return optDiagnosticLogEntry

end

--------------------------------------------------------------------------------

local function removeLocalLinkForOzRendition( docStore, photoId, ozRenditionName )

	docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docId = photoId,
		overlay = {
			_localOnly = {
				[ ozRenditionName ] = {
					_delete = true,
				},
			},
		},
	}:runUntilDone()

end

--------------------------------------------------------------------------------

local function rawDataCacheGetP( self, md5, debugRetainCount )

	return self._docStore:dataCache():getP( md5, debugRetainCount )

end

--------------------------------------------------------------------------------

-- Fetch the rendition from the data cache. If not present, then update the
-- local only data for the document to remove the reference to the Oz rendition
-- since we don't have it.

local function getCachedOzRenditionDataP(
						self, ozRenditionName, md5,
						photoId, debugRetainCount, optDiagnosticLogEntry )
						
	-- Check to see whether we already have a promise for the data. Right now,
	-- this is just working with resolved promises because those don't have
	-- issues around cancelation.
	
	-- Construct the fetch promise. We block everything else for thumbnails.
	
	--[[
						
	local p
						
	if ozRenditionName == 'thumbnail' then

		p = self._activeThumbnailDataCacheFetchesBlocker:startBlockP(
								rawDataCacheGetP, self, md5, debugRetainCount )
			
	else
	
		p = self._activeThumbnailDataCacheFetchesBlocker:waitForBlockP(
								rawDataCacheGetP, self, md5, debugRetainCount )
	
	end
	
	]]
	
	local p = rawDataCacheGetP( self, md5, debugRetainCount )
	
	p = p:mapDataP( function( status, ksData )
	
		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.imageFoundInLocalCache = not not ksData
			optDiagnosticLogEntry.time_cacheRead = KSDate.timeInIso8601Format( nil, true )
		end
		
		if not ksData then
		
			-- Cache miss: Warn about this and then delete the errant reference.

			renditionServerLog:warn( "rawGetDeviceImageB: cachedDataB for photo %s MD5 %s was nil (cache miss)",
									 photoId, md5 )
									
			removeLocalLinkForOzRendition( self._docStore, photoId, ozRenditionName )

		end
		
		return status, ksData
	
	end )
	
	return p

end

--------------------------------------------------------------------------------

local decodePromiseCache = KSTableUtils_makeFullyWeak()
	-- A mapping from MD5 hashes to promises for decoded image data

--------------------------------------------------------------------------------

local function fetchAndDecodeImageForMD5P(
						self,
						photoId,
						deviceRenditionName,
						md5,
						renditionServerLog,
						optDiagnosticLogEntry )

	renditionServerLog:info( "rawGetDeviceImageB: imageDataB fetching data for photo %s MD5 = %s",
							 photoId, md5 )
							
	local cachedDecodedImageP = decodePromiseCache[ md5 ]
	
	if cachedDecodedImageP and not cachedDecodedImageP:isCanceled() then
		return cachedDecodedImageP
	end
	
	local ozRenditionName = ozRenditionNameForDeviceRenditionName[ deviceRenditionName ]
	
	local cachedDataP = getCachedOzRenditionDataP(
				self, ozRenditionName, md5, photoId, nil, optDiagnosticLogEntry )
		-- nil == confirmed cache miss
		
	local decodedImageP = cachedDataP:mapDataP( function( _, ksData )
	
		if not ksData then
			return 'data', nil
		end
		
		-- We got an entry. Decode it.

		renditionServerLog:info( "rawGetDeviceImageB: cachedDataB for photo %s got data of size = %d",
								 photoId, KSData.length( ksData ) )
								
		local options = decodeOptionForDeviceRenditionName[ deviceRenditionName ] or { }
		
		return jpegDecodeService:requestP(
			'decode', ksData, unpack( options ) )

	end )
	
	decodePromiseCache[ md5 ] = decodedImageP
	
	return decodedImageP

end

--------------------------------------------------------------------------------

local function rawGetDeviceImageB( self, photoId, md5hint, deviceRenditionName )

	renditionServerLog:info( "rawGetDeviceImageB: photoId = %s", photoId )

	local optDiagnosticLogEntry =
		logRenditionFetch( self, photoId, md5hint, deviceRenditionName )
		
	-- Get the name of the Oz rendition
	
	local ozRenditionName = ozRenditionNameForDeviceRenditionName[ deviceRenditionName ]
	
	-- Get the MD5 key for current rendition.

	local md5B = md5ForOzRenditionB( self, photoId, md5hint, ozRenditionName,
								  	 optDiagnosticLogEntry )
								
	-- Log the MD5 info if appropriate
	
	md5B = md5B:closingMapB( function( md5 )
	
		recordDiagnosticTiming( self, photoId, 'md5' )
		renditionServerLog:info( "rawGetDeviceImageB: photoId = %s, md5 = %s", photoId, md5 )

		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.md5_resolved = md5
			optDiagnosticLogEntry.time_md5_resolved = KSDate.timeInIso8601Format( nil, true )
		end
		
		return md5

	end )
	
	-- Get the image. We relay the MD5 because we will use it elsewhere.
	
	local imageB = md5B:relayB( true ):closingMapSwitchB( function( md5 )
	
		if not md5 or md5 == '' then
			return KSReactive.constantB( nil )
		end
		
		-- Try the image cache.
		
		local cacheInfo = cacheInfoForDeviceRenditionName[ deviceRenditionName ]
	
		local imageCache = self[ cacheInfo.imageCacheName ]
		
		local cachedImage = getCachedImage( imageCache, deviceRenditionName, md5 )
		
		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.imageFoundInMemoryCache = not not cachedImage
		end
		
		if cachedImage then
			return KSReactive.constantB( cachedImage )
		end
		
		-- Fetch and decode the image.
		
		local imageP = fetchAndDecodeImageForMD5P( 
				self,
				photoId,
				deviceRenditionName,
				md5,
				renditionServerLog,
				optDiagnosticLogEntry
			)
			
		-- Stick it in the cache or log an error.
		
		imageP = imageP:waitForP( function( status, value )
			if status == 'data' and value and imageCache then
				storeCachedImage( imageCache, deviceRenditionName, md5, value )
			elseif status == 'error' then
				debug.print( "rawGetDeviceImageB: decode jpeg data failed with message %s", tostring( value ) )
			end
		end )
		
		-- Return it as a behavior so we can switch over it
			
		return imageP:untilResolvedB( 'data', nil )

	end )
	
	-- Do we think the data is out of date? We will start a new download either when
	-- shouldBeDownloadB changes or when the MD5 hash changes. The latter generally
	-- indicates a local change. By passing both values up from the call, we get a value
	-- that changes whenever either of the underlying values changes.
	
	local isInvalidB = shouldDownloadOzRenditionB( self, photoId, ozRenditionName )
	
	local downloadMD5B = md5B:relayB( true ) -- the other relay
	
	local triggerB = KSReactive.dataCallB( function( isInvalid, md5 )
										   		return isInvalid, md5
									   		end,
										    isInvalidB,
										    downloadMD5B )
						:addCloseTarget( isInvalidB )
						:addCloseTarget( downloadMD5B )
	
	local downloadB = triggerB:closingMapSwitchB( function( isInvalid )

		if optDiagnosticLogEntry then
			optDiagnosticLogEntry.shouldDownload = isInvalid
		end

		renditionServerLog:debug( "shouldDownloadOzRenditionB for %s %s is %s",
								  photoId, ozRenditionName, isInvalid )
								
		-- If we don't need to download, then we get to do nothing.
							
		if not isInvalid then
			return KSReactive.constantB( nil )
		end
		
		local downloadTaskP = downloadOzRenditionP( self, photoId, ozRenditionName, optDiagnosticLogEntry )
		recordDiagnosticTiming( self, photoId, "startDownload" )
		
		renditionServerLog:debug( "Starting download rendition task" )
		
		-- Add logging for when we finish
		
		downloadTaskP = downloadTaskP
			:waitForP( function( status )
				recordDiagnosticTiming( self, photoId, "finishDownload" )
				if optDiagnosticLogEntry then
					optDiagnosticLogEntry.downloadRenditionStatus = status
					optDiagnosticLogEntry.time_downloadRendition = KSDate.timeInIso8601Format( nil, true )
				end
				renditionServerLog:debug( "Download rendition task resolved with status %s", status )
			end )
			
		-- Then map this into a behavior we can switch over.
		
		return KSReactivePromiseBridge_promiseToBehavior( downloadTaskP )
		
	end )
			
	-- We want to keep the download logic alive as long as the image is alive.
	
	rawset( imageB, keyDownload, downloadB )
	imageB:addCloseTarget( downloadB )
	
	-- Don't let nil values while we're going to the next item slip through.
	-- We can start out nil, but we won't slip back to being nil.
	
	imageB = suppressNilB( imageB )

	if renditionServerLog.level == KSLogging.DEBUG then

		imageB:onDataB( function( newImage )
		
			renditionServerLog:debug( "imageB for %s %s returning new data checksum is probably %s",
									  photoId, deviceRenditionName,
									  newImage and downloadMD5B:valueNow() or nil )
		
		end )
	
	end
	
	return imageB
		
end

--------------------------------------------------------------------------------

-- FIXME: The image cache is no longer getting used!
-- FIXME [scouten 2012-10-04]: Who wrote this? Is it still true?

local function getDeviceImageB( self, photoId, md5hint, deferTriggerB, deviceRenditionName )
			
	check( 'WFOzRenditionServer', 'string', '?string', '_', deviceRenditionNameArg )
			
	local cacheInfo = cacheInfoForDeviceRenditionName[ deviceRenditionName ]
	
	if not cacheInfo then
		return KSReactive.constantB( nil )
	end

	local behaviorCache = getBehaviorCache( self, cacheInfo.behaviorCacheName )
	
	local cachedImageB = fetchCachedBehavior( behaviorCache, photoId, true )
	
	if cachedImageB then
		return cachedImageB:relayB( true )
	end
	
	-- If we don't already have a behavior in place for this, we might want to defer
	-- creation of it for a bit.
	
	if deferTriggerB then
	
		-- We don't cache when we're based off of a defer trigger because someone else might request
		-- it without the defer trigger.
		
		-- PERF [mhamburg 2011-03-18]: It might be nice to skip the defer if we ended up with a
		-- real value more rapidly.
		
		-- PERF [mhamburg 2011-08-14]: If someone requested the value and it actually fired, we
		-- would really prefer not to be sitting behind a defer trigger.
		
		local imageCache = self[ cacheInfo.imageCacheName ]
		
		local existingImage = getCachedImage( imageCache, deviceRenditionName, md5hint )
	
		local resultB = deferTriggerB:mapSwitchB( function( isLater )

			if isLater then

				return getDeviceImageB( self, photoId, md5hint, nil, deviceRenditionName )

			elseif existingImage then

				local existingImageB = KSReactive.constantB( existingImage )
				existingImage = nil
				return existingImageB

			else

				return KSReactive.constantB( nil )

			end

		end )

		resultB = suppressNilB( resultB )
		
		return resultB
	
	end

	-- Nothing in cache, create it.

	local imageB = rawGetDeviceImageB( self, photoId, md5hint, deviceRenditionName )
	
	if DEBUG_DOWNLOAD_FETCH_CANCELING then
	
		imageB:addCloseAction( function()

			local cacheSize = 0
			
			for _, beh in pairs( behaviorCache ) do
			
				if not WFCloseSupport_hasBeenOrWillBeClosed( beh ) then
					cacheSize = cacheSize + 1
				end
			
			end
		
			debug.print( "closing imageB for photo ID ", photoId,
						 "   remaining cache size = ", cacheSize )
			
		end )
	
	end
	
	behaviorCache[ photoId ] = imageB
	
	return imageB:relayB( true )

end

--------------------------------------------------------------------------------

local errors = setmetatable( { }, { __mode = 'kv' } )

--------------------------------------------------------------------------------

local function fnEncodeError( status, ... )

	if status == 'error' then
		local err = { n = select( '#', ... ), ... }
		errors[ err ] = true
		return 'data', err
	else
		return status, ...
	end

end

--------------------------------------------------------------------------------

-- GRUNGE [mhamburg 2012-03-22]: This probably deserves a unit test to make
-- sure it really does the right thing with dataCallB since so much of the
-- reactive system hides errors.

local function fnOrWithEncodedErrorFilter( a, b )

	-- We suppress error values for b.

	if errors[ b ] then
		b = nil
	end

	-- Error values for a are returned as a second result.
	
	if errors[ a ] then
		return b, unpack( a, 1, a.n )
	else
		return a or b
	end

end

--------------------------------------------------------------------------------

local function orB( aB, bB )

	if not aB then
		return bB
	end
	
	if not bB then
		return aB
	end
	
	local resultB = KSReactive.dataCallB(
			fnOrWithEncodedErrorFilter,
			aB:rawMapB( fnEncodeError ),
			bB:rawMapB( fnEncodeError ) )
	
	resultB:addCloseTarget( aB )
	resultB:addCloseTarget( bB )
	
	return resultB

end

--------------------------------------------------------------------------------

-- Return a behavior that filters out all errors. It's a bit of a kludge if
-- the initial value is an error plus something else.

-- FIXME [mhamburg 2012-03-22]: Move into the reactive code

local function skipErrorsB( b )

	local e = b:skipErrorsE()
	
	local r

	if b:isErrorNow() then
		r = e:holdB( nil )
	else
		r = e:holdB( b:valueNow() )
	end
	
	r:addCloseTarget( b )
	
	return r

end

--------------------------------------------------------------------------------

local function renditionImageTriggerB( renditionImageB )

	return renditionImageB:mapB( function( value ) return value ~= nil end )
				
end

--------------------------------------------------------------------------------

-- Triggers transition from false to true. They could probably become promises
-- in some future organization, but that would take more unification between
-- behaviors and promises.

local function andTriggersB( trigger1B, trigger2B )

	if trigger1B then
	
		if trigger2B then
		
			return KSReactive.dataCallB( function( a, b ) return a and b end,
				trigger1B, trigger2B )
		
		else
		
			return trigger1B
		
		end
	
	elseif trigger2B then
	
		return trigger2B
	
	else
		
		return nil 
	
	end

end

--------------------------------------------------------------------------------

-- This takes a sequence of rendition names to fetch and builds a compound
-- behavior to do the fetch preferring the first renditions in the list to
-- the later renditions. Values of false in the list result in the entry
-- being skipped. It returns a behavior for the image. The behaviors is
-- relayed (i.e., safe to close). The md5hint is for the lowest resolution version
-- being fetched.

-- GRUNGE [mhamburg 2012-03-22]: It would be nice to stop listening to the
-- lower-resolution behaviors once the higher resolution behavior has a non-nil
-- value.

local function getCascadedImageB( self, photoId, md5hint, deferTriggerB,
								  cascadedBehaviorCacheKey, deferSecondTriggerB )
								
	if not OBEY_DEFER_TRIGGERS then
		deferTriggerB = nil
		-- deferSecondTriggerB = nil
	end
								
	local deviceRenditionNames =
		cascadedBehaviorCacheKeyToDeviceRenditionNames[ cascadedBehaviorCacheKey ]
	
	assert( deviceRenditionNames, "No device renditions names for cascaded behavior" )
	
	local processSequentially = deviceRenditionNames.sequential
			
	if OzCatHelperFns_DISABLE_MOST_FETCHES then
		return KSReactive.constantB( nil )
	end

	if not photoId then
		return KSReactive.constantB( nil )
	end
	
	local behaviorCache = getBehaviorCache( self, cascadedBehaviorCacheKey )
	
	local cachedB = fetchCachedBehavior( behaviorCache, photoId, true )
		
	if cachedB then
		return cachedB:relayB( true )
	end
	
	local n = #deviceRenditionNames
	
	local imageB
	
	-- Note that the evolving trigger may refer to the individual rendition image
	-- behaviors, but those are all managed via the orB chain that gets built via
	-- imageB.
	
	local triggerB = deferTriggerB
	
	for i = n, 1, -1 do
	
		local renditionName = deviceRenditionNames[ i ]
		
		-- On Android, we do not want the 1x thumbs...
		if renditionName and not ( WFPlatformInfo.isAndroid and renditionName == 'thumbnail' ) then
		
			local renditionImageB =
				getDeviceImageB( self, photoId, md5hint, triggerB, renditionName )
				
			-- We only care about errors on the primary request.
				
			if i ~= 1 then
				renditionImageB = renditionImageB and skipErrorsB( renditionImageB )
			end
			
			if renditionImageB then
				imageB = orB( renditionImageB, imageB )
			end
			
			-- If we are processing sequentially, then we need a new trigger.

			if processSequentially and 1 < i and renditionImageB then
				triggerB = andTriggersB( triggerB, renditionImageTriggerB( renditionImageB ) )
			end
			
			if deferSecondTriggerB then
				triggerB = andTriggersB( triggerB, deferSecondTriggerB )
				deferSecondTriggerB = false -- use only once
			end

		end
		
		md5hint = nil
			-- Hint is only for first fetch.
	
	end
	
	if not imageB then
		return KSReactive.constantB( nil )
	end
	
	-- If we aren't deferred, then we can cache.
	
	if not deferTriggerB then
		behaviorCache[ photoId ] = imageB
	end
	
	return imageB:relayB( true ):closingMapB(function( image )

		if self._requestTimings[ photoId ] and image then
			-- we've delivered our first result
			recordDiagnosticTiming( self, photoId, "deliver" )
			self._velocityBuffer:add( recordDiagnosticTiming( self, photoId ) )
		end
		return image
		
	end )

			
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) a thumbnail rendition of a photo from the Oz server.

	 \param photoId (string) The photo's unique ID.
	 \param wantBigThumbnails (boolean) \c true to get big thumbnail. ((only big? small and big?))
	 \param md5hint (string) ((what?))
	 \param deferTriggerB (boolean) \c true to defer ((what?))

	 <code>B< jpegData | nil ></code> A Behavior whose value is the rendition. ((how? multiple?))

]]--

function renditionServerMethods:thumbnailB( --[[string]] photoId,
						--[[boolean]] wantBigThumbnails,
					    	--[[string]] md5hint,
					    	--[[boolean]] deferTriggerB,
							deferSecondTriggerB )

	self._requestTimings[ photoId ] = { start = now() }
	
	if OzCatHelperFns_DISABLE_MOST_FETCHES then
		return KSReactive.constantB( nil )
	end

	if not photoId then
		return KSReactive.constantB( nil )
	end
	
	if wantBigThumbnails then

		return getCascadedImageB( self, photoId, md5hint, deferTriggerB,
								  '_thumbnailBCache2', deferSecondTriggerB )

	else
	
		if not OBEY_DEFER_TRIGGERS then
			deferTriggerB = nil
			-- deferSecondTriggerB = nil
		end

		return getDeviceImageB( self, photoId, md5hint, deferTriggerB, 'thumbnail' )
			:closingMapB( function( img )
				if self._requestTimings[ photoId ] and img then
					recordDiagnosticTiming( self, photoId, "deliver" )
					self._velocityBuffer:add( recordDiagnosticTiming( self, photoId ) )
				end
				return img
			end )
	end
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) a preview rendition of a photo from the Oz server.

	 \param photoId (string) The photo's unique ID.
	 \param obsoleteAndUnused (obsolete)
	 \param skipThumbnail (boolean) \c true to fetch only preview, \c false to fetch both thumbnail and preview.
	 \param deferTriggerB (behavior) We will wait for this behavior to go true before fetching

	<code>B< image | nil ></code> A Behavior whose value is the rendition.

]]--

function renditionServerMethods:previewB( --[[string]] photoId,
					--[[boolean | 'small']] desiredSize,
					--[[boolean]] skipThumbnail,
					--[[behavior]] deferTriggerB )
					
	local cascadedBehaviorCacheKey
	
	if desiredSize == 'small' then
		cascadedBehaviorCacheKey = '_smallPreviewBCache2'
	else
		cascadedBehaviorCacheKey = '_previewBCache2'
	end

	if skipThumbnail then
	
		local deviceRenditionName =
			cascadedBehaviorCacheKeyToDeviceRenditionNames[ cascadedBehaviorCacheKey ][ 1 ]

		return getDeviceImageB( self, photoId, nil, deferTriggerB, deviceRenditionName )

	else

		return getCascadedImageB( self, photoId, nil, -- no MD5 hint
								  deferTriggerB, cascadedBehaviorCacheKey )

	end
	
end

--------------------------------------------------------------------------------

local nilP = KSPromises.resolvedPromiseP( 'data', nil )

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) a known-current encoded rendition for a photo.
	
 	\warning The fetch functions such as <code>:previewB()</code>
	 	retrieve any pixels they can, even if they are stale or incorrectly sized.
	       	This function, however, fails if it cannot find a correct rendition.
	
	\param photoId (optional, string) The photo to fetch; if not provided, the
		promise resolves to \c nil)
	\param renditionType (string) The rendition type, such as \c 'thumbnail' or
		\c 'preview'. ((are there other types?))
	\param fetchNow (optional, boolean)When \c true, waits for the rendition to become
		 available (downloading if necessary) and then returns that data.
		 When \c false or \c nil, the returned Promise resolves to \c nil data
		if the preview is not already available locally
	\param debugRetainCount (optional, boolean) When \c true, wraps the resultant data
		object in a special class that enables retain/release debugging.
	
	\return <code>P< jpegData | nil ></code> A Promise that resolves to the rendition.
]]--

function renditionServerMethods:encodedRenditionP( --[[string]] photoId,
						  --[[string]] ozRenditionName,
						--[[boolean]] fetchNow,
						--[[boolean]] debugRetainCount )

	check( 'WFOzRenditionServer', '?string', 'string', '?boolean', '?boolean' )

	if not photoId then
		return nilP
	end
	
	-- Grab data from cache, but only if we know it's current;
	-- otherwise trigger a download from Oz now.

	local docValueP = self._docStore:docValueOnceP( 'asset', photoId )
	
	-- FIXME [mhamburg 2013-05-10]: I'm not sure this routine is doing what the documentation
	-- above regarding fetchNow says it should be doing.

	local imageDataP = docValueP
		:mapDataP( function( _, docValue )

			if not docValue then
				return 'error', "encodedRenditionP: no document data available"
			end
			
			if shouldDownloadOzRendition( docValue, photoId, ozRenditionName ) then
			
				local resultP = downloadOzRenditionP( self, photoId, ozRenditionName )

				-- FIXME [ngautam 2012-02-17]: Remove this logic when downloadOzRenditionP
				-- starts returning the downloaded data.

				if fetchNow then

					return resultP:mapDataP( function( _, _ )

						local resultPromiseP = self:encodedRenditionP( photoId,
																	   ozRenditionName,
																	   nil,
																	   debugRetainCount )

						return resultPromiseP

					end )

				else
					return resultP
				end

			else
			
				local md5 = KSTableUtils_derefOpt( docValue, '_localOnly', ozRenditionName, 'md5' )

				if not md5 then
					return 'error', "encodedRendition: no rendition available either locally or on server"
				end
				
				renditionServerLog:debug( "encodedRenditionP: fetching data for MD5 = %s", md5 )
				
				return getCachedOzRenditionDataP( self, ozRenditionName, md5,
									photoId, debugRetainCount, nil )
				
			end
		
		end )

	return imageDataP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a known-current encoded preview for a photo.
	
 	\warning The fetch functions such as <code>:previewB()</code>
	 	retrieve any pixels they can, even if they are stale or incorrectly sized.
	       	This function, however, fails if it cannot find a correct rendition.
	
	\param photoId (optional, string) The photo to fetch; if not provided, the
		promise resolves to \c nil)
	\param fetchNow (optional, boolean)When \c true, waits for the rendition to become
		 available (downloading if necessary) and then returns that data.
		 When \c false or \c nil, the returned Promise resolves to \c nil data
		if the preview is not already available locally
	\param debugRetainCount (optional, boolean) When \c true, wraps the resultant data
		object in a special class that enables retain/release debugging.
	
	\return <code>P< jpegData | nil ></code> A Promise that resolves to the rendition.
]]--

function renditionServerMethods:encodedPreviewP( --[[string]] photoId,
						 --[[boolean]] fetchNow,
						 --[[boolean]] debugRetainCount )

	check( 'WFOzRenditionServer', '?string', '?boolean', '?boolean' )

	return self:encodedRenditionP( photoId, 'preview', fetchNow, debugRetainCount )

end

--------------------------------------------------------------------------------
--[[!
     \brief Enables logging of rendition operations.
]]--

function renditionServerMethods:sendDiagnosticLogData()

	WFDiagnosticLog.addSectionToLog {

		name = "Rendition Server",
		id = 'WFOzRenditionServer' .. self._catalogId,
		weight = 3000,

		body = {
		
			catalogId = self._catalogId,

			cacheEpochB = self._cacheEpochB,
			
			cacheSize = self._indexStore:queryP [[ SELECT sum( size ) FROM cacheReferences ]]
							:runUntilDoneP(),
			
			recentPreviews = self._recentPreviews,

			velocity = { self._velocityBuffer:visualize( 'boxAndWhisker' ) },
		}
	
	}
			
end

--------------------------------------------------------------------------------

local renditionServer_class = KSClass 'WFOzRenditionServer'( renditionServerMethods )

--------------------------------------------------------------------------------

local renditionServerForCatalog = KSTableUtils_makeFullyWeak()

return function( catalog )

	check( 'OzCatalogService|OzCatalogServiceImp' )
	
	local catalogImp = catalog._catalogService or catalog
	
	local result = renditionServerForCatalog[ catalogImp ]

	if not result then
	
		result = renditionServer_class( catalogImp )
		renditionServerForCatalog[ catalogImp ] = result
	
	end
	
	return result

end