--[[----------------------------------------------------------------------------

WFOzRenditionServerUtils.lua

@author scouten 2012-10-03

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- wichitafoundation.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'


--============================================================================--
--[[ Don't get too attached to this namespace. In all likelihood, this
	will disappear into the private portions of \c WFOzRenditionServer when
	that refactoring is completed.
]]--

-- @mark WFOzRenditionServerUtils

local WFOzRenditionServerUtils = {}

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local cacheEpochLog = KSLogging.log 'WFOzRenditionServerUtils :: cacheEpoch'
--	:setLevel( KSLogging.DEBUG )

--------------------------------------------------------------------------------

local cacheEpochSize

if WFPlatformInfo.isMac then
	cacheEpochSize = 50 * 1024 * 1024
elseif WFPlatformInfo.isiOS then
	cacheEpochSize = 20 * 1024 * 1024
else
	-- GRUNGE [arahn 2012-08-28]: What are sensible defaults for other platforms?
	cacheEpochSize = 20 * 1024 * 1024
end

--------------------------------------------------------------------------------

local cacheEpochBs = KSTableUtils.makeFullyWeak()
	-- key (OzDocStore)
	-- value: B< number >

--------------------------------------------------------------------------------

function WFOzRenditionServerUtils.cacheEpochB( --[[OzDocStore]] docStore,
											   --[[WFIndexStore]] indexStore )

	check( 'OzDocStore', 'WFIndexStore' )
		
	-- Do we have a cached cache epoch behavior already?
	
	local catalogId = docStore:catalogId()
	
	local resultB = WFCloseSupport.fetchCached( cacheEpochBs, catalogId )
	if resultB then
		return resultB
	end
	
	-- We don't need the cache monitoring to be precise, but we do need it
	-- to be efficient. Profile data shows that this is a major source of
	-- DB thread CPU consumption in larger catalogs, so we monitor cache epoch
	-- via a calmed behavior. The consequence of this is that we might write more
	-- data into an epoch than intended, but that's relatively harmless.
	
	cacheEpochLog:info( "created cacheEpochB for catalog ID %s", catalogId )
	
	local maxEpochB = indexStore:queryB {
		sql = 'SELECT max( epoch ) FROM cacheReferences',
		calm = 10,
	}
	:closingMapB( function( rows )
		return rows and rows[ 1 ] and rows[ 1 ][ 1 ] or 1
	end )
	
	if cacheEpochLog.level == KSLogging.DEBUG then
		maxEpochB:onDataB( function( maxEpoch )
			cacheEpochLog:debug( "maxEpochB updated to %s", maxEpoch )
		end )
	end
	
	local epochSizeB = indexStore:queryB {
		sql = 'SELECT epoch, sum( size ) FROM cacheReferences WHERE epoch = ?',
		calm = 10,
		maxEpochB,
	}
	:closingMapB( function( rows )
		local row1 = rows and rows[ 1 ]
		if row1 then
			return ( row1[ 1 ] or 1 ), ( row1[ 2 ] or 0 )
		else
			return 1, 0
		end
	end )
	:addCloseTarget( maxEpochB )
	
	if cacheEpochLog.level == KSLogging.DEBUG then
		epochSizeB:onDataB( function( epoch, size )
			cacheEpochLog:debug( "epochSizeB updated to %s, %s", epoch, size )
		end )
	end
	
	resultB = epochSizeB:closingMapB( function( epoch, size )
		
		if size >= cacheEpochSize then
			cacheEpochLog:info( "Epoch %s too large (size = %s), starting new epoch %s",
								epoch, size, epoch + 1 )
			return epoch + 1
		else
			cacheEpochLog:info( "Epoch %s size %s OK", epoch, size )
			return epoch
		end
	
	end )
	
	cacheEpochBs[ catalogId ] = resultB
	
	return resultB
	
end

--------------------------------------------------------------------------------

-- We map the rendition names down to a smaller set of rendition kinds so that we can employ
-- common processing where appropriate.

local renditionKindForRenditionName = {
	thumbnail = 'thumbnail',
	preview = 'preview',
	master = 'master',
	thumbnail2x = 'thumbnail',
}

local ozLinkNameForRenditionName = {
	thumbnail = "/rels/rendition_type/144h",
	preview = "/rels/rendition_type/1024",
	master = "/rels/master",
	thumbnail2x = "/rels/rendition_type/thumbnail2x",
}

WFOzRenditionServerUtils.renditionKindForRenditionName = renditionKindForRenditionName
WFOzRenditionServerUtils.ozLinkNameForRenditionName = ozLinkNameForRenditionName

--------------------------------------------------------------------------------
-- Touch a rendition (place it in the current epoch). This operation is
-- fire-and-forget.

function WFOzRenditionServerUtils.touchRendition( --[[OzDocStore]] docStore,
												  --[[B< number >]] cacheEpochB,
												  --[[string]] photoId,
												  --[[string]] renditionName,
												  --[[table]] optDocValue,
												  --[[string]] optSize )

	check( 'OzDocStore', '_', 'string', 'string', '?table', '?string' )

	-- Touch a rendition (place it in the current epoch). This operation is
	-- fire-and-forget.
	
	-- Don't bother touching thumbnails. We don't space manage them.
	
	if renditionKindForRenditionName[ renditionName ] == 'thumbnail' then
		return
	end

	-- Circle back if we don't have doc value. (It's better to provide optDocValue
	-- if you can!)
	
	if not optDocValue then
	
		docStore:docValueOnceP( 'asset', photoId )
			:whenResolvedDo( function( status, docValue )
			
				if status == 'data' then
					WFOzRenditionServerUtils.touchRendition( docStore, cacheEpochB, photoId,
															 renditionName, docValue,
															 optSize )
				end

			end ):runUntilDone()
		
		return
	
	end

	-- If this item is already in current epoch, ignore.

	local renditionInfo = optDocValue
					  and optDocValue._localOnly
					  and optDocValue._localOnly[ renditionName ]

	local epochNow = cacheEpochB:valueNow()

	if not epochNow then
		cacheEpochLog:warn( "cacheEpochB had no value ... not touching rendition" )
		return
	end

	if renditionInfo and renditionInfo.epoch == epochNow
		and ( renditionInfo.size == optSize or not optSize )
	then
		cacheEpochLog:debug( "not touching %s for photo %s ... already in current epoch",
							 renditionName, photoId )
		return
	end
	
	-- Update cache info.
	
	cacheEpochLog:debug( "cache touching %s for photo %s, moving to epoch %d with size %s",
						 renditionName, photoId, epochNow, optSize )

	docStore:applyOverlayToDocumentP {
		
		docType = 'asset',
		docId = photoId,
		
		overlay = {
			_localOnly = {
				[ renditionName ] = {
					epoch = epochNow,
					size = optSize,
				}
			}
		},
	
	}

end

--------------------------------------------------------------------------------

return WFOzRenditionServerUtils
