--[[----------------------------------------------------------------------------

WFOzSyncCycle.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This namespace implements a single sync "cycle" (meaning an attempt to upload
a set number of changes and download changes from Oz).

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.os
local KSDate = require 'KSDate'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require "KSReactive"

	-- wichitafoundation.core.table
local KSTableArchiving = require 'KSTableArchiving'

	-- wichitafoundation.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'

	-- wichitafoundation.oz.sync_agent
local OzCentralUploadTask = require 'OzCentralUploadTask'
local WFOzDownloadTask = require 'WFOzDownloadTask'
local WFOzSyncConstants = require 'WFOzSyncConstants'

	-- wichitafoundation.web.json
local KSJsonEncoder = require 'KSJsonEncoder'
local KSJsonDecoder = require 'KSJsonDecoder'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = KSTableUtils.derefOpt
local uploadOrderForDocType = WFOzSyncConstants.uploadOrderForDocType

local syncCycleLog = KSLogging.log 'WFOzSyncCycle'
--	:setLevel( KSLogging.DEBUG )


--============================================================================--

-- @mark WFOzSyncCycle

local WFOzSyncCycle = {}

--------------------------------------------------------------------------------

local function deleteFromDocStoreP( docStore, docType, docId )

	return docStore:setServerValueP(
								docType,
								docId,
								{ _deleted = true, _ozMetadataUploadStatus = 'uploaded', } )

end

--------------------------------------------------------------------------------

-- Function for url encoding a given string
-- FIXME [ngautam 2012-01-23] This should move to a more appropriate utils location
-- Function adopted from http://lua-users.org/wiki/StringRecipes

local function url_encode( str )
	
 	if str then
		str = string.gsub( str, "\n", "\r\n" )
		str = string.gsub( str, "([^%w ])",
        	function ( c )
				return string.format( "%%%02X", string.byte( c ) )
			end )
		str = string.gsub( str, " ", "+" )
	end
		
	return str
		
end

--------------------------------------------------------------------------------

local delayUploadUntilNextWaveP = function( docStore, indexStore, docType, docId )

	return indexStore:queryP( "SELECT max( uploadWave ) FROM ozUploadStatus" )
		:mapDataP( function( status, rows )

			local nextWave = ( rows[ 1 ][ 1 ] or 1 ) + 1
			return docStore:applyOverlayToDocumentP{
				docType = docType,
				docId = docId,
				overlay = {
					_localOnly = {
						uploadWave = nextWave,
					},
				}
			}
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

local function processDeleteAssetQueueP( args )

	return args.indexStore:queryP(
		[[SELECT docId FROM ozUploadStatus
			WHERE docType == 'asset' AND ozUploadStatus == 'deleted'
			ORDER BY captureDate DESC, validRev ASC
			LIMIT 100]] )

		:mapDataP( function( _, deleteQueue )

			if not deleteQueue or #deleteQueue == 0 then
				return 'data', false
			end
			
			syncCycleLog:debug( "processDeleteQueueP: %d items in delete queue", #deleteQueue )
		
			local currentAssets = {}
			for i = 1, #deleteQueue do
				currentAssets[ i ] = deleteQueue[ i ][ 1 ]
			end
		
			-- Delete the assets from Oz.
			
			syncCycleLog:debug( "Deleting set of assetIds %s", #currentAssets )
			
			local formattedAssetIds = url_encode( table.concat( currentAssets, ';' ) )
			
			local deleteUrl = string.format( '%s/assets?asset_ids=%s', args.catalogUrl,
											 formattedAssetIds )
						
			syncCycleLog:debug( "Sending DELETE ASSET request to %s", deleteUrl )
					
			local deleteFromOzP = args.ozUser:getP( deleteUrl, { requestMethod = 'DELETE' } )
			
			local function deleteFromLocalDatabaseP( assetId )
				
				return deleteFromDocStoreP( args.docStore, 'asset', assetId )
			
			end
				
			return deleteFromOzP:mapP( function( status, data, headers )
			
				if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then
				
					-- '204' means the assets were found and deleted.
					-- '404' means the assets were not found, so delete them locally.
								
					syncCycleLog:debug( "Oz returned  %s", headers.status_code )
		
					-- Set the server value for each asset to deleted.
			
					return KSPromises.allP( deleteFromLocalDatabaseP, currentAssets, 5 )
		
				else
		
					return 'error', data
		
				end
		
			end )
		
	end )
	
end

--------------------------------------------------------------------------------

-- if optUploadOrderFilter it must start with " AND " so it can be inserted into our sql query

local function processUploadQueueP( args, optUploadOrderFilter )

-- add something like epoch so we circle through every doc in uploadStatus once before retrying
-- mark a document as belonging to a later cycle if we get any 4xx error
-- 4xx errors should not trigger the jail (track didAnything in OzCentralSyncAgent)
	
	local predicate = [[
			WHERE ozUploadStatus != '_conflict' AND docType != 'album_asset'
				AND ( freemiumUploadEmbargoUntil ISNULL
					OR freemiumUploadEmbargoUntil < ? )
				]] .. ( optUploadOrderFilter or "" )
	
	local sql = [[SELECT docId, docType, ozUploadStatus FROM ozUploadStatus
				]] .. predicate .. [[
				AND uploadOrder IN ( SELECT min( uploadOrder ) FROM ozUploadStatus ]] .. predicate .. [[ )
			ORDER BY uploadWave ASC, uploadOrder ASC, ozUploadStatus, captureDate DESC, validRev
			LIMIT 4]]
	
	local now = KSDate.timeNow()
	
	return args.indexStore:queryP(
		sql, now, now )

		:mapDataP( function( _, uploadQueue )

			if not uploadQueue or #uploadQueue == 0 then
				return 'data', false
			end
			
			syncCycleLog:debug( "processUploadQueueP: %d items in upload queue", #uploadQueue )
		
			args.errorCapture.processOneUploadP = {}

			local function processOneUploadP( uploadQueueEntry )
			
				local docId = uploadQueueEntry[ 1 ]
				local docType = uploadQueueEntry[ 2 ]
				local uploadStatus = uploadQueueEntry[ 3 ]
			
				syncCycleLog:debug( "processUploadQueueP: kicking off upload task for %s status %s", docId, uploadStatus )

				args.errorCapture.processOneUploadP[ docId ] = {}
				args.errorCapture.processOneUploadP[ docId ].docType = docType
				args.errorCapture.processOneUploadP[ docId ].uploadStatus = uploadStatus
				
				local resultP = OzCentralUploadTask.uploadP {
					docStore = args.docStore,
					ozUser = args.ozUser,
					catalogUrl = args.catalogUrl,
					docId = docId,
					docType = docType,
					uploadStatus = uploadStatus,
					errorCaptureForDoc = args.errorCapture.processOneUploadP[ docId ],
				}
				
				resultP = resultP:mapP( function( status, data, headers, docUrl )
				
					if headers and headers.status_code == '404' then
						return deleteFromDocStoreP( args.docStore, docType, docId )
					end

					if status == 'error'
						and ( ( headers and headers.status_code ) or data == 'reload' )
					then
		
						args.errorCapture.processOneUploadP[ docId ].didAttemptReload = true
						args.errorCapture.processOneUploadP[ docId ].didReloadBecause = {
							status = status,
							data = tostring( data ),
							headers = headers,
							docUrl = docUrl,
						}

						return WFOzDownloadTask.reloadDocumentP( args, docType, docId, docUrl )
							:mapP( function( reloadStatus, reloadData, reloadHeaders )
								if reloadHeaders and reloadHeaders.status_code == '404' then
									return deleteFromDocStoreP( args.docStore, docType, docId )
								end
								if reloadData == 'reload'
									or ( reloadHeaders and
										( reloadHeaders[ 'X-Result-Code' ] == '1002'
									   or reloadHeaders[ 'X-Result-Code' ] == '1005' ) )
								then
									return 'data', true
								else
									args.errorCapture.processOneUploadP[ docId ].reloadResult = {
										status = reloadStatus,
										data = type( reloadData ) ~= 'table' and tostring( reloadData ) or reloadData,
										headers = reloadHeaders,
									}
									return status, data, headers
								end
		
							end )
		
					end
					
					if status == 'error'
						and ( headers and string.sub( headers.status_code, 1, 1 ) == '4' ) then
						
						delayUploadUntilNextWaveP( args.docStore, args.indexStore, docType, docId )
						
					end
				
				end )
				
				resultP = resultP:observeP( function( status, data, headers )
				
					if status == 'data' then
		
						syncCycleLog:debug( "upload task for %s %s %s finished with status %s %s %s\n\n%s\n\n",
											docType, docId, uploadStatus,
											status, data, headers,
											debug.traceback() )
		
						if args and args.statusReport then
							args.statusReport[ docId ] = nil
						end
		
					else
		
						local failedMessage = string.format( "upload task for %s %s %s finished with status %s %s %s\n\n%s\n\n",
											docType, docId, uploadStatus,
											status,
											tostring( data ),
											KSTableArchiving.simpleTableToChunkString( headers or {} ),
											debug.traceback() )
		
						syncCycleLog:warn( "%s", failedMessage )
		
						if args and args.statusReport then
							args.statusReport[ docId ] = failedMessage
						end
		
					end
		
				end )
				
				return resultP
			
			end
		
			return KSPromises.allP( processOneUploadP, uploadQueue, 5 )
	
	end )

end

--------------------------------------------------------------------------------

local function processCreateAlbumAssetQueueP( args, createAlbumAssetQueue )

	if not createAlbumAssetQueue or #createAlbumAssetQueue == 0 then
		return 'data', false
	end
		
	syncCycleLog:debug( "processCreateAlbumAssetQueueP: %s items in create queue", createAlbumAssetQueue )
			
	-- Go through the set of album_asset documents, collect by albumId and send one request for each albumId present.
	-- Create the list as so
			
	-- { albumId =<albumId>,
	--	 	payload = {
	--			_id = <assetId>,
	--			album_asset_id = <docId>,
	--		}
	--	}
	--
			
	local currentAlbumAssets = {}
	local albumAssetsByAssetId = {}
			
	local i = 1
	while i <= #createAlbumAssetQueue do

		local currentAlbumId = createAlbumAssetQueue[ i ][ 2 ]
		local albumRow = #currentAlbumAssets + 1

		currentAlbumAssets[ albumRow ] = {
			albumId = currentAlbumId,
			payload = {},
		}

		while i <= #createAlbumAssetQueue and createAlbumAssetQueue[ i ][ 2 ] == currentAlbumId do

			local payloadRow = #currentAlbumAssets[ albumRow ].payload + 1

			currentAlbumAssets[ albumRow ].payload[ payloadRow ] = {
				_id = createAlbumAssetQueue[ i ][ 3 ],
				album_asset_id = createAlbumAssetQueue[ i ][ 1 ],
			}

			albumAssetsByAssetId[ createAlbumAssetQueue[ i ][ 3 ] ] = createAlbumAssetQueue[ i ][ 1 ]
			i = i + 1

		end

	end
			
	local function addAssetToAlbumP( collection )
			
		local albumId = collection.albumId
		local uploadContent = {
			assets = collection.payload
		}
				
		syncCycleLog:debug( "Sending ADD ASSET TO ALBUM to %s with UPLOAD CONTENT", uploadContent )
		
		local addAssetsToAlbumUrl = string.format( '%s/albums/%s/assets', args.catalogUrl, albumId )
						
		syncCycleLog:debug( "Sending ADD ASSET TO ALBUM to %s", addAssetsToAlbumUrl )
				
		local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )
				
		syncCycleLog:debug( "Sending ADD ASSET TO ALBUM with JSON CONTENT %s", jsonData )
				
		local putArgs = {
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
				{ header = 'Accept', value = 'application/json' },
			},
			returnHttpErrorsAsData = true,
		}
					
		return args.ozUser:putP( addAssetsToAlbumUrl, jsonData, putArgs )
			:mapDataP( function( status, data, headers )
										
				syncCycleLog:debug( "Oz returned  %s", headers.status_code )
		
				-- Update the local database depending on the results.
				-- Those that succeed, mark them as uploaded.
				-- Those that fail with a duplication error, mark them as deleted.
						
				local function updateDatabaseP( record )
						
					local status = record[ 1 ]
					local docId = record[ 2 ]
					local server_album_asset = record[ 3 ]
					local assetId = record[ 4 ]
							
					if status == 'completed' then
								
						return args.docStore:applyOverlayToDocumentP{
							docType = 'album_asset',
							docId = docId,
							overlay = { _localOnly = { _uploadedValues = {} }, }
								-- adding an _uploadedValues table here signals to the
								-- uploadStatus code that we no longer need upload
						}
								
					elseif status =='duplicate' then
		
						return args.docStore:docValueOnceP( 'album_asset', docId )
								
							:mapDataP( function( _, docValue )
									
								if docId == server_album_asset._id then

									-- GRUNGE [ngautam 2012-05-18] It looks like the indexStore doesn't refresh quickly enough and we end up
									-- sending the same album_asset request twice. This can result in a '1002' asset already exists error
									-- which we handle by deleting the album_asset document we have locally.
									-- We don't want to do this for a valid document and so we check for the upload status here or the
									-- value of docValue.created and don't do anything if any of these is set.
									-- Requested Mayank to provide the album_asset_id for the document in Oz when this error is thrown so
									-- we can use that to compare our album_asset_id against.

									local overlay = server_album_asset
									overlay._ozMetadataUploadStatus = 'uploaded'
									overlay._id = nil
									
									return args.docStore:setServerValueP (
										'album_asset',
										docId,
										overlay
									)
								
								end
								
								-- Use the server_album_asset returned and write that document to the database before deleting the one that failed.

								local overlay = server_album_asset
								local newDocId = server_album_asset._id
								overlay._id = nil
								
								return args.docStore:setServerValueP(
									'album_asset',
									newDocId,
									overlay
								)
								:mapDataP( function()

									-- Have to add the uploaded status here so that the document doesn't circle
									-- back and get picked up by the queue for deleting album_asset documents.

									return deleteFromDocStoreP( args.docStore, 'album_asset', docId )

								end )
											
							end )
							:runUntilDoneP()

					elseif status == 'albumDeleted' or status == 'assetDeleted' then
						
						return deleteFromDocStoreP( args.docStore, 'album_asset', docId )

					elseif status == 'duplicateAlbumAssetId' then

						-- Oz returned an error because the album_asset_id specified by us is already in use.
						-- We will delete the album_asset document that failed and create a new one with a new id and retry.
		
						return deleteFromDocStoreP( args.docStore, 'album_asset', docId )
				
						:mapDataP( function ()
								
							return args.docStore:applyOverlayToDocumentP {
								docType = 'album_asset',
								docIds = KSIdGenerator.uuidInLowercaseHex(),
								overlay = {
									album = {
										_id = albumId,
										},
									asset = {
										_id = assetId,
										},
									},
								}
												
						end )
					
					elseif status == 'failed' then
						
						-- Push this document into the next uploadWave, to prevent perpetually stalling on this document
						
						return delayUploadUntilNextWaveP( args.docStore, args.indexStore, 'album_asset', docId )
						
					end
							
					-- We don't update the document for any other failures.
							
				end
						
				local result = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
						
				syncCycleLog:debug( "Oz returned data %s", result )
		
				-- Process the result returned from Oz. Successes contain the assetIds of ones
				-- successfully added to the album.

				local successes = result and result.assets
				local errors = result and result.errors
						
				local resultRows = {}
						
				if successes then
					for i = 1, #successes do
						resultRows[ i ] = {}
						resultRows[ i ][ 1 ] = 'completed'
						resultRows[ i ][ 2 ] = albumAssetsByAssetId[ successes[ i ]._id ]
					end
				end

				-- Errors contain the assetIds of the ones that failed.

				if errors and not errors.code then

					for i = 1, #errors do

						local count = #resultRows + 1
						resultRows[ count ] = {}

						if errors[ i ].code == 1002 then
							resultRows[ count ][ 1 ] = 'duplicate'
							resultRows[ count ][ 3 ] = errors[ i ].album_asset
						elseif errors[ i ].code == 1000 then
							resultRows[ count ][ 1 ] = 'assetDeleted'
						elseif errors[ i ].code == 1029 then
							resultRows[ count ][ 1 ] = 'duplicateAlbumAssetId'
						else
							resultRows[ count ][ 1 ] = 'failed'
						end

						resultRows[ count ][ 2 ] = albumAssetsByAssetId[ errors[ i ]._id ]
						resultRows[ count ][ 4 ] = errors[ i ]._id

					end

				end
				
				if derefOpt( errors, 'album', 1 ) == 'does not exist' then
				
					-- the album for these have been deleted.
					-- So delete all the album_asset documents
					for index, assetInfo in ipairs( collection.payload ) do
						resultRows[ index ] = {
							"albumDeleted",
							assetInfo.album_asset_id
						}
					end
				
				end
						
				return KSPromises.allP( updateDatabaseP, resultRows, 5 )
		
			end )
			
		end
				
	return KSPromises.allP( addAssetToAlbumP, currentAlbumAssets, 5 )
		
end

--------------------------------------------------------------------------------

local function processDeleteAlbumAssetQueueP( args, deleteAlbumAssetQueue )

	if not deleteAlbumAssetQueue or #deleteAlbumAssetQueue == 0 then
		return 'data', false
	end
		
	syncCycleLog:debug( "processDeleteAlbumAssetQueueP: %s items in delete queue", deleteAlbumAssetQueue )
			
	-- Go through the set of album_asset documents, collect by albumId and send one request for each albumId present.
	-- Create the list as so
			
	-- { albumId =<albumId>,
	--	 	payload = {
	--			assetId = <assetId>,
	--			album_asset_id = <docId>,
	--		}
	--	}
	--
			
	local currentAlbumAssets = {}
			
	local i = 1

	while i <= #deleteAlbumAssetQueue do

		local currentAlbumId = deleteAlbumAssetQueue[ i ][ 2 ]
		local albumRow = #currentAlbumAssets + 1

		currentAlbumAssets[ albumRow ] = {
			albumId = currentAlbumId,
			payload = {},
		}

		while i <= #deleteAlbumAssetQueue and deleteAlbumAssetQueue[ i ][ 2 ] == currentAlbumId do

			local payloadRow = #currentAlbumAssets[ albumRow ].payload + 1
			currentAlbumAssets[ albumRow ].payload[ payloadRow ] = {
				assetId = deleteAlbumAssetQueue[ i ][ 3 ],
				album_asset_id = deleteAlbumAssetQueue[ i ][ 1 ],
			}

			i = i + 1

		end

	end
			
	local function deleteAssetsFromAlbumP( collection )
			
		local albumId = collection.albumId
		local payload = collection.payload
				
		local currentAssets = {}
				
		for i = 1, #payload do
			currentAssets[ i ] = payload[ i ].assetId
		end
				
		syncCycleLog:debug( "Sending DELETE ASSET FROM ALBUM to %s with ASSET LIST", currentAssets )
		
		local formattedAssetIds = url_encode( table.concat( currentAssets, ';' ) )
			
		local deleteUrl = string.format( '%s/albums/%s/assets?asset_ids=%s', args.catalogUrl, albumId,
											 	 formattedAssetIds )
						
		syncCycleLog:debug( "Sending DELETE ASSET FROM ALBUM to %s", deleteUrl )
				
		return args.ozUser:getP( deleteUrl, { requestMethod = 'DELETE' } )
					
			:mapP( function( status, data, headers )
					
				local function deleteFromLocalDatabaseP( docId )
				
					return deleteFromDocStoreP( args.docStore, 'album_asset', docId )
			
				end
						
				if headers and ( headers.status_code == '204' or headers.status_code == '404' ) then
				
					syncCycleLog:debug( "Oz returned  %s", headers.status_code )
		
					-- '204' means the album assets were found and deleted.
					-- '404' means the album assets were not found, so delete them locally.
								
					-- Set the server value for each album asset to deleted.
			
					local docIds = {}
					for i = 1, #payload do
						docIds[ i ] = payload[ i ].album_asset_id
					end
							
					return KSPromises.allP( deleteFromLocalDatabaseP, docIds, 5 )
		
				else
		
					return 'error', data
		
				end
		
			end )
			
		end
				
	return KSPromises.allP( deleteAssetsFromAlbumP, currentAlbumAssets )
		
end

--------------------------------------------------------------------------------

local function processAlbumAssetQueueP( args, didAnyAssetUploads )

	-- FIXME [scouten 2012-11-06, PR# 3357111]: If we just did any asset uploads
	-- then that means that the set of album-asset documents we get here might
	-- reference assets that are yet to be uploaded. Oz gets grumpy about that
	-- so we'll have to wait to make those uploads.

	if didAnyAssetUploads then
		return KSPromises.resolvedPromiseP( 'data', false )
	end

	return args.indexStore:queryP(
		[[SELECT docId, albumId, assetId, ozUploadStatus FROM ozUploadStatus
			WHERE docType == 'album_asset'
			ORDER BY albumId, changeCounter, ozUploadStatus
			LIMIT 50]] )
			
		:mapDataP( function ( _, albumAssetQueue )
		
			if not albumAssetQueue or #albumAssetQueue == 0 then
				return 'data', false
			end
			
			-- We have the list of album_asset documents that either need to be created or deleted.
			-- We order them by albumId, then by the doc store change counter and then by the ozUploadStatus
			-- in the hope of atleast combining some of these as one HTTP request.
			
			local i = 1
			local albumAssetQueue_combined = {}
			while i <= #albumAssetQueue do

				local currentUploadStatus = albumAssetQueue[ i ][ 4 ]
				local currentRow_combined = #albumAssetQueue_combined + 1
				
				albumAssetQueue_combined[ currentRow_combined ] = {
					uploadStatus = currentUploadStatus,
					payload = {},
				}
				
				while i <= #albumAssetQueue and albumAssetQueue[ i ][ 4 ] == currentUploadStatus do
				
					local payloadRow = #albumAssetQueue_combined[ currentRow_combined ].payload + 1
					albumAssetQueue_combined[ currentRow_combined ].payload[ payloadRow ] = albumAssetQueue[ i ]
					i = i + 1
					
				end
		
			end
	
			-- Now take each entry in albumAssetQueue_combined and depending on the uploadStatus process each one.
	
			local function processEachItem( row )
			
				if row.uploadStatus == 'deleted' then
					return processDeleteAlbumAssetQueueP( args, row.payload )
				else
					return processCreateAlbumAssetQueueP( args, row.payload )
				end
	
			end
	
			return KSPromises.allP( processEachItem, albumAssetQueue_combined, 5 )
			
		end )
	
end

--------------------------------------------------------------------------------

local function processDownloadsP( args )

	syncCycleLog:debug( "processDownloadsP" )
	
	if args.isUserCatalog then

		return KSPromises.resolvedPromiseP( 'data', false )

	else

		return WFOzDownloadTask.downloadLatestFromOzP( args )
			:mapDataP( function( status, didAnyDownloads, numItemsDownloaded )
			
				syncCycleLog:debug( "processDownloadsP FINISHED" )
				return 'data', didAnyDownloads, numItemsDownloaded
			
			end )

	end

end

--------------------------------------------------------------------------------

local function trackStatusFireEventIfTrue( didSomeWorkE )
	local value = false
	return function( ... )
		if select( '#', ... ) == 0 then
			return value
		end
		if not value then
			value = ...
			if value and didSomeWorkE then
				didSomeWorkE:post( true )
			end
		end
	end
end

--------------------------------------------------------------------------------
--[[!
	\brief Runs a single sync cycle (attempt to upload a fixed number of changes
	and download changes from Oz).
	
	\param args (table)
		- \c docStore = (OzDocStore)
		- \c ozUser = (OzUser)
		- \c catalogUrl = (string)
		- \c uploadQueue = (table) array of upload queue entries
		- \c statusReport = (optional, table) place to collect status reports
				for up/download failure
	
	\return <code>P< boolean ></code> \c true if any data was uploaded or downloaded,
		\c false if not
	\return <code>E< boolean ></code> fires true any time we do substantial work uploading
		or downloading (may not fire at all, e.g. if the first result resolves to false).
--]]

local earlyUploadFilter = string.format( "AND uploadOrder < %d", uploadOrderForDocType.album_asset )

function WFOzSyncCycle.runOneCycleP( --[[table]] args )

	check( 'table' )
	
	syncCycleLog:debug( "runOneCycleP BEGIN" )
	
	local didSomeWorkE = KSReactive.sourceE()
	
	local didAnyDownloads = trackStatusFireEventIfTrue( didSomeWorkE )
	local didAnyAssetDeletes = trackStatusFireEventIfTrue( didSomeWorkE )
	local didAnyUploadsEarly = trackStatusFireEventIfTrue( didSomeWorkE )
	local didAnyAlbumAssetUpdates = trackStatusFireEventIfTrue( didSomeWorkE )
	local didAnyUploadsSlow = trackStatusFireEventIfTrue( didSomeWorkE )
	local numItemsDownloaded = 0
	
	-- Ugh. Using mapDataP for a side effect...

	args.errorCapture.syncCycle = { runOneCycleP = true }

	local resultP = processDownloadsP( args )
		:mapDataP( function( status, value, numItemsDd )
			didAnyDownloads( value )
			numItemsDownloaded = numItemsDd
		end )
	
		:mapDataP( function() return processDeleteAssetQueueP( args ) end )
		:mapDataP( function( status, value )
			didAnyAssetDeletes( value )
			args.errorCapture.syncCycle.processDeleteAssetQueueP_DONE = true
		end )

		:mapDataP( function() return processUploadQueueP( args, earlyUploadFilter ) end )
		:mapDataP( function( status, value )
			didAnyUploadsEarly( value )
			args.errorCapture.syncCycle.processUploadQueueP_DONE = true
		end )

		:mapDataP( function() return processAlbumAssetQueueP( args, didAnyUploadsEarly() ) end )
		:mapDataP( function( status, value )
			didAnyAlbumAssetUpdates( value )
			args.errorCapture.syncCycle.processAlbumAssetQueueP_DONE = true
		end )

		:mapDataP( function()
			if ( not didAnyUploadsEarly() ) and ( not didAnyAlbumAssetUpdates() ) then
				return processUploadQueueP( args )
			else
				return 'data', false
			end
		end )
		:mapDataP( function( status, value )
			didAnyUploadsSlow( value )
			args.errorCapture.syncCycle.processSlowUploadQueueP_DONE = true
		end )

		:mapDataP( function()
			syncCycleLog:debug( "upload queue finished, now process downloads" )
		
			return 'data', didAnyDownloads()
						or didAnyAssetDeletes()
						or didAnyUploadsEarly()
						or didAnyAlbumAssetUpdates()
						or didAnyUploadsSlow(),
					numItemsDownloaded
		end )
	
	return resultP, didSomeWorkE

end

--------------------------------------------------------------------------------

return WFOzSyncCycle
