
local KSPromises = require "KSPromises"

local KSJsonEncoder = require "KSJsonEncoder"

local WFAsyncTestCase = require "WFAsyncTestCase"
local WFOzSyncTestUtils = require "WFOzSyncTestUtils"

--==============================================================================
-- @mark setup
--==============================================================================

local TestSuite = {}

local WFOzSyncCycle

local docStore
local indexStore
local ozUser
local T = {}

--------------------------------------------------------------------------------

local function buildMocks( docStore )
	
	return {
		OzCentralUploadTask = WFOzSyncTestUtils.makeFakeOzCentralUploadTask( T ),
		
		OzCatalogIndexDefinitions = false,
		
		WFOzDownloadTask = {
			
			reloadDocumentP = function( args, docType, docId, docUrl )
				table.insert( T.ozDownloadReloadResults, {
					docId, docType, args.catalogUrl,
				} )
				local result = T.ozDownloadReloads[ docType .. docId ]
				if not result then
					return KSPromises.resolvedPromiseP( 'error', "Unexpected reload of document" )
				end
				return KSPromises.resolvedPromiseP( unpack( result ) )
				
			end,
			
			downloadLatestFromOzP = function( args )
				-- The sync cycle calls this each time thru.
				-- since we use all fake canned data there's nothing to do here...
				T.calledDownloadLatestFromOzP = true
				return KSPromises.resolvedPromiseP( 'data', false )
					:runUntilDoneP()
			end,
		},
		
		-- Fake implementation which uses OzCatalogIndexDefinitions to build out the proper
		-- schema and mapping functions for our in-memory indexStore
		WFIndexStoreSchema = docStore:raw_test_buildMockWFIndexStoreSchema(),
		
		-- Fake impl, does nothing
		WFIndexStoreImp = {
			start = function( schema )
			
			end,
		},
	}
	
end

--==============================================================================

function TestSuite.setUpClass()
	
	-- Build a fake docStore, sufficient for testing this namespace
	
	docStore = WFOzSyncTestUtils.makeFakeDocStore()

	WFOzSyncCycle = WFOzSyncTestUtils.importWithMocks( 'WFOzSyncCycle', buildMocks( docStore ) )
	
	-- Utterly fake oz user.  Vends canned responses to GETs and stores
	-- of what was PUT for later analysis.
	ozUser = WFOzSyncTestUtils.makeFakeOzUser( T )
	
	-- Build a fake indexStore.  It's powered by a real sql db in memory
	-- that uses the real schema and mapping fns.
	indexStore = docStore:raw_test_buildMockIndexStore( )
	
end

--------------------------------------------------------------------------------

local function okHttpResult( code, optData, optStatus, ... )
	return {
		optStatus or 'data',
		optData or "",
		{
			status_code = code,
		},
		...
	}
end

--------------------------------------------------------------------------------

local function runOneCycleP( args )
	
	args = args or {}
	args.errorCapture = T.errorCapture
	args.statusReport = T.statusReport
	args.indexStore = indexStore
	args.docStore = docStore
	args.ozUser = ozUser
	args.catalogUrl = "urlz"
	
	return WFOzSyncCycle.runOneCycleP( args )
	
end

--------------------------------------------------------------------------------

local kNormalFullSync = {
	processAlbumAssetQueueP_DONE = true,
	processDeleteAssetQueueP_DONE = true,
	processUploadQueueP_DONE = true,
	processSlowUploadQueueP_DONE = true,
	runOneCycleP = true,
}

--------------------------------------------------------------------------------

-- Does nothing.  But still brings us to 19% coverage in WFOzSyncCycle.lua
function TestSuite.testNoOp( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {} )
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP()
		:mapP( function( status, result )
			return status, {
				A_result = result,
				B_ozCentralUploads = T.ozCentralUploads,
				C_userGetPutRequests = T.userGetPutRequests,
				D_calledDownloadLatestFromOzP = T.calledDownloadLatestFromOzP,
				E_errorCapture = T.errorCapture,
				F_statusReport = T.statusReport,
			}
		end ))
		:continueWhenValueMatches( 'data', {
			A_result = false,
			B_ozCentralUploads = {
			},
			C_userGetPutRequests = {
			},
			D_calledDownloadLatestFromOzP = true,
			E_errorCapture = {
				syncCycle = kNormalFullSync,
			},
			F_statusReport = {}
		} )

end

--------------------------------------------------------------------------------

-- Test whether sync properly deletes one asset.  Now up to 36% coverage.
function TestSuite.testUploadDeletedAsset( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'abcdefg',
			_deleted = true,
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.userGetPutResponse[ "urlz/assets?asset_ids=abcdefg" ] = okHttpResult "204"
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_result = value,
				B_ozCentralUploads = T.ozCentralUploads,
				C_userGetPutRequests = T.userGetPutRequests,
				D_calledDownloadLatestFromOzP = T.calledDownloadLatestFromOzP,
				E_errorCapture = T.errorCapture,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_result = true,
			B_ozCentralUploads = {
			},
			C_userGetPutRequests = {
				{
					args = {
						requestMethod = "DELETE",
					},
					requestMethod = "DELETE",
					url = "urlz/assets?asset_ids=abcdefg",
				},
			},
			D_calledDownloadLatestFromOzP = true,
			E_errorCapture = {
				syncCycle = kNormalFullSync
			},
		} )

end

--------------------------------------------------------------------------------

-- Test whether sync properly deletes one asset.  Now up to 36% coverage.
function TestSuite.testOneToUpload( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'lmnop',
			stuff = 'more stuff',
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.userGetPutResponse[ "urlz/assets?asset_ids=lmnop" ] = okHttpResult "204"
	T.ozCentralUploadResults[ "assetlmnop" ] = okHttpResult "404"
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_result = value,
				B_ozCentralUploads = T.ozCentralUploads,
				C_userGetPutRequests = T.userGetPutRequests,
				E_errorCapture = T.errorCapture,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_result = true,
			B_ozCentralUploads = {
				{
					"urlz",
					"lmnop",
					"asset",
					"notUploaded",
				},
			},
			C_userGetPutRequests = {
			},
			E_errorCapture = {
				processOneUploadP = {
					lmnop = {
						docType = "asset",
						uploadStatus = "notUploaded",
					},
				},
				syncCycle = kNormalFullSync
			},
		} )

end

--------------------------------------------------------------------------------

-- Test whether sync properly reloads one asset.  Now up to 36% coverage.
function TestSuite.testUploadReload( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'pqrs',
			stuff = 'more stuff',
			_ozMetadataUploadStatus = 'notUploaded_metadataOnly',
		},
	} )
	
	T.ozCentralUploadResults[ "assetpqrs" ] = okHttpResult( "500", "reload", 'error' )
	
	T.ozDownloadReloads[ "assetpqrs" ] = okHttpResult "200"
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status,
				value,
				T.ozCentralUploads,
				T.userGetPutRequests,
				T.ozDownloadReloadResults,
				T.errorCapture
			
		end ) )
		:continueWhenValueMatches( 'error',
			'reload',
			{
				{
					"urlz",
					"pqrs",
					"asset",
					"notUploaded_metadataOnly",
				},
			},
			{ },
			{
				{
					"pqrs",
					"asset",
					"urlz",
				},
			},
			{
				processOneUploadP = {
					pqrs = {
						didAttemptReload = true,
						didReloadBecause = {
							data = "reload",
							headers = {
								status_code = "500",
							},
							status = "error",
						},
						docType = "asset",
						reloadResult = {
							data = "",
							headers = {
								status_code = "200",
							},
							status = "data",
						},
						uploadStatus = "notUploaded_metadataOnly",
					},
				},
				syncCycle = {
					processDeleteAssetQueueP_DONE = true,
					runOneCycleP = true,
				},
			}
		)

end

--------------------------------------------------------------------------------

-- Test whether sync properly reloads and then deletes one asset.
function TestSuite.testUploadReloadDelete( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'alpha',
			stuff = 'more stuff',
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.ozCentralUploadResults[ "assetalpha" ] = okHttpResult( "500", "reload", 'error' )
	
	T.ozDownloadReloads[ "assetalpha" ] = okHttpResult "404"
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_doc = docStore._docs.assetalpha,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_doc = {
				_deleted = true,
				_docId = "alpha",
				_docType = "asset",
				_ozMetadataUploadStatus = "uploaded",
			},
		} )

end

--------------------------------------------------------------------------------

-- Test whether sync properly reloads and then deletes one asset.
function TestSuite.testUploadReloadReload( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'beta',
			stuff = 'more stuff',
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.ozCentralUploadResults[ "assetbeta" ] = okHttpResult( "500", "reload", 'error' )
	
	T.ozDownloadReloads[ "assetbeta" ] = okHttpResult( "500", "reload" )
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_doc = docStore._docs.assetbeta,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_doc = {
				_docId = "beta",
				_docType = "asset",
				_ozMetadataUploadStatus = "notUploaded",
				stuff = 'more stuff',
			},
		} )

end

--------------------------------------------------------------------------------

-- Test whether sync properly deletes one asset.  Now up to 36% coverage.
function TestSuite.testUploadThenOzDeletes( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'asset',
			_docId = 'wxyz',
			stuff = 'more stuff',
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.ozCentralUploadResults[ "assetwxyz" ] = okHttpResult( "404", "reload", 'error' )
		-- 404 means that it's deleted on Oz
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_doc = docStore._docs.assetwxyz,
				E_errorCapture = T.errorCapture,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_doc = {
				_docId = "wxyz",
				_docType = "asset",
				_ozMetadataUploadStatus = "uploaded",
				_deleted = true,
			},
			E_errorCapture = {
				processOneUploadP = {
					wxyz = {
						docType = "asset",
						uploadStatus = "notUploaded",
					},
				},
				syncCycle = kNormalFullSync
			},
		} )

end

--------------------------------------------------------------------------------

-- Album asset logic is special,and extensive.  This unit test alone doubles coverage to 63%.
function TestSuite.testUploadAlbumAsset_disabled( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'album_asset',
			_docId = 'hi-ya-friend',
			album = {
				_id = 'ye-olde-album-id',
			},
			asset = {
				_id = 'yar-asset-id',
			},
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.userGetPutResponse[ "urlz/albums/ye-olde-album-id/assets" ] = okHttpResult( "204",
		KSJsonEncoder.encodeWebSafeJson {
			assets = {
				{
					_id = 'yar-asset-id',
				},
			},
			errors = {
			},
		} )
	T.ozCentralUploadResults[ "album_assethi-ya-friend" ] = okHttpResult "204"
	
	return WFAsyncTestCase( test, 600000 )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_userGetPutRequests = T.userGetPutRequests,
				E_errorCapture = T.errorCapture,
				
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_userGetPutRequests = {
				{
					args = {
						requestHeaders = {
							{
								header = "Content-Type",
								value = "application/json",
							},
							{
								header = "Accept",
								value = "application/json",
							},
						},
						returnHttpErrorsAsData = true,
					},
					body = "{\"assets\":[{\"album_asset_id\":\"hi-ya-friend\",\"_id\":\"yar-asset-id\"}]}",
					requestMethod = "GET",
					url = "urlz/albums/ye-olde-album-id/assets",
				},
			},
			E_errorCapture = {
				syncCycle = kNormalFullSync
			},
		} )

end

--------------------------------------------------------------------------------

-- Album asset logic is special,and extensive.  This unit test alone doubles coverage to 63%.
function TestSuite.testUploadAlbumAssetSendDuplicate( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'album_asset',
			_docId = 'hi-ya-friend',
			album = {
				_id = 'ye-olde-album-id',
			},
			asset = {
				_id = 'yar-asset-id',
			},
			_ozMetadataUploadStatus = 'notUploaded',
		},
	} )
	
	T.userGetPutResponse[ "urlz/albums/ye-olde-album-id/assets" ] = okHttpResult( "204",
		KSJsonEncoder.encodeWebSafeJson {
			assets = {
			},
			errors = {
				{
					album_asset = {
						_id = 'cya-later-dude',
						album = {
							_id = 'ye-olde-album-id',
						},
						asset = {
							_id = 'yar-asset-id',
						},
					},
					_id = 'yar-asset-id',
					code = 1002,
				},
			},
		} )
	T.ozCentralUploadResults[ "album_assethi-ya-friend" ] = okHttpResult "204"
	
	return WFAsyncTestCase( test, 60000 )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_userGetPutRequests = T.userGetPutRequests,
				B_docs = docStore._docs,
				E_errorCapture = T.errorCapture,
				
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_userGetPutRequests = {
				{
					args = {
						requestHeaders = {
							{
								header = "Content-Type",
								value = "application/json",
							},
							{
								header = "Accept",
								value = "application/json",
							},
						},
						returnHttpErrorsAsData = true,
					},
					body = "{\"assets\":[{\"album_asset_id\":\"hi-ya-friend\",\"_id\":\"yar-asset-id\"}]}",
					requestMethod = "GET",
					url = "urlz/albums/ye-olde-album-id/assets",
				},
			},
			B_docs = {
				["album_assetcya-later-dude"] = {
					_docId = "cya-later-dude",
					_docType = "album_asset",
					album = {
						_id = "ye-olde-album-id",
					},
					asset = {
						_id = "yar-asset-id",
					},
				},
				["album_assethi-ya-friend"] = {
					_deleted = true,
					_docId = "hi-ya-friend",
					_docType = "album_asset",
					_ozMetadataUploadStatus = "uploaded",
				},
			},
			E_errorCapture = {
				syncCycle = kNormalFullSync
			},
		} )

end

--------------------------------------------------------------------------------

-- Test process of deleting an asset from an album. Coverage @ 85%
function TestSuite.testUploadDeleteAlbumAsset( test )

	WFOzSyncTestUtils.setUpIndex( indexStore, docStore, T, {
		{
			_docType = 'album_asset',
			_docId = 'hi-ya-friend',
			_ozMetadataUploadStatus = 'deleted',
			album = {
				_id = 'ye-olde-album-id',
			},
			asset = {
				_id = 'randomassetid',
			},
		},
	} )
	
	T.userGetPutResponse[ "urlz/albums/ye-olde-album-id/assets?asset_ids=randomassetid" ] = okHttpResult "204"
	
	return WFAsyncTestCase( test )
		:waitFor( runOneCycleP {
		}:mapP( function( status, value )
			
			return status, {
				A_userGetPutRequests = T.userGetPutRequests,
				B_docs = docStore._docs,
				E_errorCapture = T.errorCapture,
			}
			
		end ) )
		:continueWhenValueMatches( 'data', {
			A_userGetPutRequests = {
				{
					args = {
						requestMethod = "DELETE",
					},
					requestMethod = "DELETE",
					url = "urlz/albums/ye-olde-album-id/assets?asset_ids=randomassetid",
				},
			},
			B_docs = {
				["album_assethi-ya-friend"] = {
					_deleted = true,
					_docId = "hi-ya-friend",
					_docType = "album_asset",
					_ozMetadataUploadStatus = "uploaded",
				},
			},
			E_errorCapture = {
				syncCycle = kNormalFullSync
			},
		} )

end

--==============================================================================

return TestSuite
