--[[----------------------------------------------------------------------------

WFOzSyncTestUtils.lua

11/26/2012
arahn

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSPromises = require "KSPromises"
local KSReactive = require "KSReactive"
local WFIdle = require 'WFIdle'

local KSDatabaseSchema = require "KSDatabaseSchema"

local WFSqliteUtils = require "WFSqliteUtils"
local KSDatabase = require "KSDatabase"

local WFUnknown = require 'WFUnknown'

local WFOzSyncTestUtils = {}

--==============================================================================

function WFOzSyncTestUtils.importWithMocks( ns, mocks )

	local LOADED = debug.getregistry()[ "_LOADED" ]

	for k, namespace in pairs( mocks ) do
		if namespace == false then
			LOADED[ k ] = nil
		else
			LOADED[ k ] = namespace
		end
	end
	
	LOADED.ns = nil
	return require( ns )

end

--------------------------------------------------------------------------------

-- Ultra simple fake version.
local function setServerValue( docStore, docType, docId, doc )

	doc._docId = docId
	doc._docType = docType
	docStore._docs[ docType .. docId ] = doc

	for _, schemaInfo in ipairs( docStore._mapFns ) do
	
		schemaInfo.map( doc, function( ... )
			return schemaInfo.emit( doc._docId, ... )
		end )
	
	end

end

--------------------------------------------------------------------------------

local function raw_test_buildMockIndexStore( docStore, fn )

	local requireFn = fn
	if requireFn then
		requireFn = string.gsub( requireFn, "%.lua", "" )
	else
		requireFn = "OzCatalogIndexDefinitions"
	end
	local OzCatalogIndexDefinitions = require ( requireFn )
		-- as a side effect of requiring that, schema is now populated
	local WFIndexStoreSchema = require 'WFIndexStoreSchema'

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema._schema:addTable( [[CREATE TEMPORARY TABLE TEMP_ids( docId )]] )

	local function prepQueryAndGetRows( self, args, ... )

		local sql
		local nQueryArgs
		local rows
		if type( args ) == 'string' then
			sql = args
			nQueryArgs = select( '#', ... )
		else
			sql = args.sql
			nQueryArgs = args.nArgs or #args
		end

		if args.tempIds then
			-- Empty the TEMP_ids table and insert new records
			self.db:exec( [[DELETE FROM TEMP_ids]] )
			for i = 1, #args.tempIds do
				self.db:exec( [[INSERT INTO TEMP_ids VALUES ( ? )]], args.tempIds[ i ] )
			end
		end
		if nQueryArgs == 0 then
			rows = self.db:gatherArrayOfArray( sql )
		else
			if type( args ) == 'string' then
				rows = self.db:gatherArrayOfArray( sql, ... )
			else
				rows = self.db:gatherArrayOfArray( sql, unpack( args, 1, nQueryArgs ) )
			end
		end

		return rows

	end

	local indexStore = {
		queryP = function( self, args, ... )
			local rows = prepQueryAndGetRows( self, args, ... )
			return KSPromises.resolvedPromiseP( 'data', rows )
		end,
		queryB = function( self, args, ... )
			local rows = prepQueryAndGetRows( self, args, ... )
			local resultB = KSReactive.variableB()
			resultB:set( WFUnknown.value() )
			WFIdle.doASAP( function()
				resultB:set( rows )
			end )
			return resultB
		end,
		db = KSDatabase( ':memory:', indexStoreSchema._schema )
	}

	indexStoreSchema._updateE
		:onDataE( function( ... )
			indexStore.db:exec( ... )
		end )

	return OzCatalogIndexDefinitions and indexStore

end

--------------------------------------------------------------------------------

local function raw_test_buildMockWFIndexStoreSchema( docStore )
	local result
	return function()
		
		if result then
			return result
		end
		
		result = {
			_schema = KSDatabaseSchema( '1.0' ),
			_updateE = KSReactive.sourceE(),
			addTable = function( self, args )
				local sql, nColumns, tableName
				if args.fields then
					tableName = args.tableName
					sql = string.format( "CREATE TABLE %s ( docId, validRev, %s )",	
						args.tableName, table.concat( args.fields, ", " ) )
					nColumns = #args.fields
				else
					sql = args.sql:gsub( "%(", "( docId, validRev,", 1 )
					local columns = sql:match( "%((.*)%)" )
					local _, nCommas = columns:gsub( ',', ',' )
					
					nColumns = nCommas - 1
					tableName = WFSqliteUtils.tableNameFromCreateTableStatement( sql )
				end
				local updateSql = string.format( "INSERT INTO %s VALUES( ?, ?%s )",
							tableName,
							string.rep( ", ?", nColumns ) )
				local deleteSql = string.format( "DELETE FROM %s WHERE docId = ?",
							tableName )

				self._schema:addTable( sql )
				local didEmitAny = false

				table.insert( docStore._mapFns, {
					map = function( doc, ... )
						didEmitAny = false

						local result = { args.map( doc, ... ) }
						if not didEmitAny then
							self._updateE:post( deleteSql, doc._docId )
						end
						return unpack( result )
						
					end,
					emit = function( docId, ... )

						if not didEmitAny then
							self._updateE:post( deleteSql, docId )
							didEmitAny = true
						end
						self._updateE:post( updateSql, docId, 1, ... )
						
					end
				} )
			end,
			addIndex = function( self, sql )
				self._schema:addIndex( sql )
			end,
		}
		
		return result
	end
end

--------------------------------------------------------------------------------

function WFOzSyncTestUtils.makeFakeDocStore( )

	return {
		_docs = {},
		_mapFns = {},
		_localStateP = {},
		catalogId = function() return "catalog_id_14159265" end,
		applyOverlayToDocumentP = function( self, t )
			local docId = t.docId or t.docIds
			local doc = self._docs[ t.docType .. docId ]
			if not doc then
				doc = {}
				self._docs[ t.docType .. docId ] = doc
			end
			if t.overlay._deleted then
				for k, v in pairs( t.overlay ) do
					doc[ k ] = nil
				end
			end
			local function deleteKeys( parent )
				for k, v in pairs( parent ) do
					
					if type( v ) == 'table' then
						if v._delete then
							parent[ k ] = nil
						else 
							deleteKeys( v )
						end
					end
					
				end
			end
			deleteKeys( t.overlay )
			
			for k, v in pairs( t.overlay ) do
				doc[ k ] = v
			end
			setServerValue( self, t.docType, docId, doc )
			return KSPromises.resolvedPromiseP( 'data', 'overlay', t )
		end,
		setServerValueP = function( self, docType, docId, value )
			setServerValue( self, docType, docId, value )
			return KSPromises.resolvedPromiseP( 'data', true )
		end,
		docValueOnceP = function( self, docType, docId )
			return KSPromises.resolvedPromiseP( 'data', self._docs[ docType .. docId ] )
		end,
		docValueB = function( self, docType, docId )
			local resultB = KSReactive.variableB( nil )
			WFIdle.doASAP( function()
				resultB:set( self._docs[ docType .. docId ] )
			end )
			return resultB
		end,
		fieldValueB = function( self, docType, docId, fieldPath )
			local resultB = KSReactive.variableB( nil )
			WFIdle.doASAP( function()
				local value = self._docs[ docType .. docId ]
				fieldPath:gsub( "([^/]+)",
					function( segment )
						if segment:match( "^-?%d+$" ) then
							segment = tonumber( segment )
						end
						value = type( value ) == 'table' and value[ segment ] or nil
					end )
				resultB:set( value )
			end )
			return resultB
		end,
		
		localStateP = function( self, localStateVarName )
			return KSPromises.resolvedPromiseP( 'data', self._localStateP[ localStateVarName ] )
		end,
		
		setLocalStateP = function( self, localStateVarName, value )
			self._localStateP[ localStateVarName ] = value
			return KSPromises.resolvedPromiseP( 'data', true )
		end,
		
		raw_test_setServerValue = setServerValue,
		
		raw_test_buildMockWFIndexStoreSchema = raw_test_buildMockWFIndexStoreSchema,
		
		raw_test_buildMockIndexStore = raw_test_buildMockIndexStore,
		
	}
end

--------------------------------------------------------------------------------

function WFOzSyncTestUtils.makeFakeOzCentralUploadTask( T )
	return {
		
		-- Dummy impl, capture what was sent so we can verify
		uploadP = function( args )
			
			table.insert( T.ozCentralUploads, {
				args.catalogUrl,
				args.docId,
				args.docType,
				args.uploadStatus,
			} )
			local uploadResults = T.ozCentralUploadResults[ args.docType .. args.docId ]
			if not uploadResults then
				error "Unit Test Needs Updating : Didn't expect oz Central to try and upload this kind of document for this test"
			end
			return KSPromises.resolvedPromiseP( unpack( uploadResults ) )
			
		end,
		
	}
end

--------------------------------------------------------------------------------

function WFOzSyncTestUtils.makeFakeOzUser( T, ozUserInfo )
	
	return setmetatable( {
		
		getOrPutP = function( self, method, url, body, args )
	
			table.insert( T.userGetPutRequests, {
				requestMethod = (args and args.requestMethod ) or method,
				url = url,
				body = body,
				args = args,
			} )
			
			local data = T.userGetPutResponse[ url ]
			if data then
				return KSPromises.resolvedPromiseP( unpack( data ) )
			else
				return KSPromises.resolvedPromiseP( 'error', 'No GET data' )
			end
		
		end,
		
		putP = function( self, url, body, args )
			
			return self:getOrPutP( (args and args.requestMethod ) or "GET", url, body, args )
			
		end,
		
		getP = function( self, url, args )
			
			return self:getOrPutP( (args and args.requestMethod ) or "GET", url, nil, args )
			
		end,
		
		id = function()
			return ozUserInfo._id
		end,
		
		isCurrentUserB = function()
			return KSReactive.constantB( true )
		end,
		
		firstNameB = function()
			return KSReactive.constantB( ozUserInfo.first_name )
		end,
		
		lastNameB = function()
			return KSReactive.constantB( ozUserInfo.last_name )
		end,
		
		fullNameB = function()
			return KSReactive.constantB( ozUserInfo.full_name )
		end,
		
		
	}, { __type = 'OzUser' } )
	
end

--------------------------------------------------------------------------------

function WFOzSyncTestUtils.setUpIndex( indexStore, docStore, testContext, docs )

	indexStore.db:exec "DELETE FROM ozUploadStatus"	
	indexStore.db:exec "DELETE FROM coreInfo"
	indexStore.db:exec "DELETE FROM album_asset"
	indexStore.db:exec "DELETE FROM keywords"

	docStore._docs = {
	}

	for _, doc in ipairs( docs ) do
		
		docStore:raw_test_setServerValue( doc._docType, doc._docId, doc )
		
	end
	
	testContext.userGetPutRequests = {}
	testContext.userGetPutResponse = {}
	testContext.ozCentralUploads = {}
	testContext.ozCentralUploadResults = {}
	testContext.calledDownloadLatestFromOzP = false
	testContext.ozDownloadReloads, testContext.ozDownloadReloadResults = {}, {}
	testContext.errorCapture, testContext.statusReport = {}, {}
	
end

--==============================================================================

return WFOzSyncTestUtils
