--[[----------------------------------------------------------------------------

WFRequestDispatcher.lua

@author scouten 2011-04-21

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

-- NOTE: This namespace is fundamental to the framework operation. Please avoid
-- introducing any other dependencies.

local assert = assert
local require = require
local type = type
local pairs = pairs

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'

	-- wichitafoundation.core.reactive (lazy load)
local KSPromises_lazy
local KSReactive_lazy

	-- wichitafoundation.core.utils
local KSPCallLoops = require 'KSPCallLoops'
local WFArguments = require 'WFArguments'


--------------------------------------------------------------------------------

local requestDispatcherLog = KSLogging.log( 'WFRequestDispatcher' )
--	:setLevel( KSLogging.DEBUG )


--============================================================================--
--[[!
	\class WFRequestDispatcher
	\brief This namespace provides the mechanism by which
	worker threads can respond to messages from other threads.
	
	This namespace wraps underlying native APIs.
--]]

local KSPromises_isPromise_lazy = function()
	return false
end

--------------------------------------------------------------------------------

local KSReactive_isEventNode_lazy = function()
	return false, false
end

--------------------------------------------------------------------------------

local function loadReactiveSystem()

	if not KSPromises_lazy then
		KSPromises_lazy = require 'KSPromises'
		KSPromises_isPromise_lazy = assert( KSPromises_lazy.isPromise )
	end
	
	if not KSReactive_lazy then
		KSReactive_lazy = require 'KSReactive'
		KSReactive_isEventNode_lazy = assert( KSReactive_lazy.isEventNode )
	end

end


--============================================================================--

-- @mark WFRequestDispatcher namespace

local WFRequestDispatcher = {}

--------------------------------------------------------------------------------

local eventNodesForRawRequests = {}
local promiseActionsForRawRequests = {}
local messageIdsForRawRequests = {}

local requestStats -- (table) request stats, used for diagnostic log if requested
	-- key: (string) request name
	-- value: (table) info on request (count, min, max, total)

local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local stuffToClose

--------------------------------------------------------------------------------
--[[!
	\brief Closes a specific object after processing current request.
	
	The object is closed when the current request processing has finished.
	 Typically, this is just after the request's data has been placed in
	transit. If the request has already been closed, it occurs when we
	detect that closure.
	
	\param object The object to close. Must be a \c WFCloseTarget or a native object
			or nil.

	\return (object) The object that was closed.
--]]

function WFRequestDispatcher.deferredClose( --[[WFCloseTarget]] object )

	check( '? userdata | table' )

	if object == nil then
		return
	end
	
	if not object.close then
		error( string.format( "WFRequestDispatcher.deferredClose: object %s (no close method)",
							  tostring( object ) ), 2 )
	end

	if not stuffToClose then
		stuffToClose = {}
	end
	
	stuffToClose[ object ] = true
	
	return object

end

--------------------------------------------------------------------------------

local function sendClose( obj )

	obj:close()

end

--------------------------------------------------------------------------------

local function processDeferredClosures()

	while stuffToClose do
	
		local currentCycle = stuffToClose
	
		stuffToClose = nil
		
		KSPCallLoops.pdoWithIterator( sendClose, nil, pairs( currentCycle ) )
		
	end
	
end

--------------------------------------------------------------------------------

WFCurrentThread.registerPriorityChangeHandler( function( rawRequest, priority )

	-- FIXME [scouten 2012-03-19]: No support for prioritization now.
	-- We can, however, cancel requests.

	if priority <= -10000 then

		local messageId = messageIdsForRawRequests[ rawRequest ]
		messageIdsForRawRequests[ rawRequest ] = nil
		
		if messageId then

			if requestStats then
				local requestStatsForMessage = requestStats[ messageId ]
				if requestStatsForMessage then
					requestStatsForMessage.cancelCount = ( requestStatsForMessage or 0 ) + 1
				end
			end

			requestDispatcherLog:debug( "Canceling request for message %s",
										messageId )

		end

		local eventNode = eventNodesForRawRequests[ rawRequest ]
		
		if eventNode then

			if rawget( eventNode, '_debugMe' ) then
				requestDispatcherLog:error( "Closing event node for message %s", rawRequest._debugMe )
			end

			eventNode:close()
			eventNodesForRawRequests[ rawRequest ] = nil

		end
	
		local promiseAction = promiseActionsForRawRequests[ rawRequest ]
		
		if promiseAction then

			if rawget( promiseAction, '_debugMe' ) then
				requestDispatcherLog:error( "Closing promise for message %s", rawRequest._debugMe )
			end

			promiseAction:close()
			promiseActionsForRawRequests[ rawRequest ] = nil

		end
	
	end

end )

--------------------------------------------------------------------------------

local function finish_sendResponse( messageId, startTime, request, success, ... )

	requestDispatcherLog:debug( "finish_sendResponse\n  messageId = %s\n  success = %s",
								messageId, success )

	local debugMe = false
					--messageId == 'addDocChangeCounterRequest'  --(example of typical use)

	local rawRequest = request:getRawRequestRef()

	if success then

		-- If first return value is an event node, then we bind the request
		-- to the node and keep it running.

		local isEventNode, isBehavior = KSReactive_isEventNode_lazy( ... )

		if isEventNode then
		
			local eventNode = ...
			
			eventNode:collectCloseTargets()
		
			eventNodesForRawRequests[ rawRequest ] = eventNode
			
			if debugMe then
				messageIdsForRawRequests[ rawRequest ] = messageId
				rawset( eventNode, '_debugMe', messageId )
			end
			
			local function sendValue( ... )

				if debugMe then
					requestDispatcherLog:error( "sending event node update\n  messageId = %s\n  values = %s, %s, %s",
												messageId, ... )
				else
					requestDispatcherLog:debug( "sending event node update\n  messageId = %s\n  values = %s, %s, %s",
												messageId, ... )
				end

				if not request:send( nil, 'data', ... ) then
				
					-- [mhamburg 2012-03-22 via scouten]: Response from
					-- request:send always seems to be false. Treating it as an
					-- unreliable signal of client thread's lack of interest.
					-- Uncomment next two lines if :send can be made to return
					-- an accurate close hint; delete if not.
					
					-- [mhamburg 2012-06-14]: I believe I've fixed this as of 2012-04-17 (CL 823884)
				
					eventNode:close()
					eventNodesForRawRequests[ rawRequest ] = nil

				end
			
				processDeferredClosures()

			end
			
			if isBehavior then
				sendValue( eventNode:valueNow() )
			end
			
			eventNode:onDataE( sendValue )

			eventNode:onErrorE( function( ... )

				if debugMe then
					requestDispatcherLog:error( "sending event node ERROR\n  messageId = %s\n  values = %s, %s, %s",
												messageId, ... )
				else
					requestDispatcherLog:debug( "sending event node ERROR\n  messageId = %s\n  values = %s, %s, %s",
												messageId, ... )
				end

				if not request:send( nil, 'error', ... ) then

					-- [mhamburg 2012-03-22 via scouten]: Response from
					-- request:send always seems to be false. Treating it as an
					-- unreliable signal of client thread's lack of interest.
					-- Uncomment next two lines if :send can be made to return
					-- an accurate close hint; delete if not.

					-- [mhamburg 2012-06-14]: I believe I've fixed this as of 2012-04-17 (CL 823884)
				
					eventNode:close()
					eventNodesForRawRequests[ rawRequest ] = nil

				end

				processDeferredClosures()

			end )
		
			eventNode:onDoneE( function()

				if debugMe then
					requestDispatcherLog:error( "sending event node CLOSED\n  messageId = %s", messageId )
				else
					requestDispatcherLog:debug( "sending event node CLOSED\n  messageId = %s", messageId )
				end

				request:send( 'done' )
				request:close()

				eventNodesForRawRequests[ rawRequest ] = nil

				processDeferredClosures()

			end )

		elseif KSPromises_isPromise_lazy( ... ) then

			local promise = ...
			
			if debugMe then
				messageIdsForRawRequests[ rawRequest ] = messageId
			end

			local function handlePromiseResult( status, ... )
				
				if debugMe then
					requestDispatcherLog:error( "sending promise result\n  messageId = %s\n  status = %s\n  values = %s, %s, %s",
												messageId, status, ... )
				else
					requestDispatcherLog:debug( "sending promise result\n  messageId = %s\n  status = %s\n  values = %s, %s, %s",
												messageId, status, ... )
				end
				
				if status == 'data' then
					request:send( 'done', 'data', ... )
				else
					request:send( 'done', 'error', ... )
				end
				
				request:close()
				promiseActionsForRawRequests[ rawRequest ] = nil

				processDeferredClosures()
			
			end
				
			local promiseAction = promise:whenResolvedDo( handlePromiseResult )
			promiseAction:runUntilDone()
			
			promiseActionsForRawRequests[ rawRequest ] = promiseAction
		
		else
			
			-- If not a reactive node or promise, then treat this as a
			-- one-off request: Send back whatever values were returned
			-- to us and close the callback.

			-- Promises on the client side work better when they have at least
			-- one value as part of their resolution, so we supply a true value
			-- as an indication of success if we got no results.
			
			if debugMe then
				requestDispatcherLog:error( "sending completed request results\n  messageId = %s\n  values = %s, %s, %s",
											messageId, ... )
			else
				requestDispatcherLog:debug( "sending completed request results\n  messageId = %s\n  values = %s, %s, %s",
											messageId, ... )
			end

			local isEmpty = select( '#', ... ) == 0
		
			if isEmpty then
				request:send( 'done', 'data', true )
			else
				request:send( 'done', 'data', ... )
			end
			
			request:close()
			
		end

	else

		if debugMe then
			requestDispatcherLog:debug( "sending completed request results WITH ERROR\n  messageId = %s\n  values = %s, %s, %s",
										messageId, ... )
		else
			requestDispatcherLog:debug( "sending completed request results WITH ERROR\n  messageId = %s\n  values = %s, %s, %s",
										messageId, ... )
		end

		request:send( 'done', 'error', ... )
		request:close()

	end
	
	if requestStats then
	
		local elapsedTime = debug.timeNow() - startTime
		
		local stat = requestStats[ messageId ]
	
		if stat then
			stat.count = stat.count + 1
			stat.min = math.min( stat.min, elapsedTime )
			stat.max = math.max( stat.max, elapsedTime )
			stat.total = stat.total + elapsedTime
			stat.errorCount = stat.errorCount + ( success and 0 or 1 )
		else
			requestStats[ messageId ] = {
				messageId = messageId,
				count = 1,
				errorCount = success and 0 or 1,
				cancelCount = 0,
				min = elapsedTime,
				max = elapsedTime,
				total = elapsedTime,
			}
		end

	end

	processDeferredClosures()

end

--------------------------------------------------------------------------------

local function sendResponseOrErrorToCallback( messageId, request, func, ... )

	local startTime = requestStats and debug.timeNow()

	finish_sendResponse( messageId, startTime, request, pcall( func, ... ) )
	
end

--------------------------------------------------------------------------------

local messageHandlers = {}
	-- key (string): message name
	-- value (function): function to be called when this message is received

--------------------------------------------------------------------------------

WFCurrentThread.registerRequestDispatchHandler( function( requestRef,
														  priority,
														  selector,
														  ... )

	requestDispatcherLog:debug( "receiving request\n  messageId = %s\n  priority = %s\n  values = %s, %s, %s",
								selector, priority, ... )

	local messageHandler = messageHandlers[ selector ]

	if messageHandler then
	
		return messageHandler( requestRef, ... )

	elseif selector == '_notifyWhenReady' then
	
		requestRef:send( 'done', nil, true )
		requestRef:close()
	
	else
	
		requestDispatcherLog:error( "No message handler defined for request\n  messageId = %s\n  priority = %s\n  values = " .. string.rep( "%s, ", select( '#', ... ) ),
									selector, priority, ... )

		requestRef:send( 'done', 'error',
						 "No message handler defined for message: " .. selector )

		requestRef:close()
	
	end

end )

--------------------------------------------------------------------------------

-- FIXME [scouten 2012-03-19]: Add support for priority change.

--------------------------------------------------------------------------------
--[[!
	\brief Registers a handler for a specific incoming message.
	
	\param args (table) A table with these keys:
		- \c message: (string)The name of the message to be handled.
		- \c handler: (function) A callback function to handle the message. You
			must provide either this handler function OR a target object
			containing a handler method, not both.
		- \c target: (object) An object to handle the message. This object must
			contain a method with the same name as the message, to be called
			if the message is received. Provide either this object OR the handler
			function, not both.
		- \c onlyIfCallbackIsStillValid: (boolean) \c true to skip the call to
			the handler if the client has lost interest
			(that is, the callback has become invalid). Default is \c false.
		
		The message handler function or method can return any number of results, which
		are interpreted as follows:
	
		- If the first result is a Behavior, it is kept alive as long as the client
			is interested, and any data posted to the node is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is a Promise, it is kept alive as long as the client
			is interested. When the promise resolves, the
			status (\c data or \c error) and any additional data is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is NOT a Reactive node, all results are passed
			back to the client via the request object, and the request is marked as done.
			All of the results passed through must be transitable (that is, scalar
			values, \c KSData, certain native object references, or tables
			containing only transitable values).
		- If the message handler fails with an error, the error status is
			passed to the client via the request object.
		- If the message handler returns no result, the request is marked as done.

]]--

function WFRequestDispatcher.onMessageDo( --[[table]] args )

	check( 'table' )

	local message = args.message
	assert( type( message ) == 'string' and #message > 0,
			"WFRequestDispatcher.onMessageDo: missing required argument 'message' (must be non-empty string)" )

	local target = args.target
	local func = assert( args.handler or args.target[ args.method or message ],
						 "WFRequestDispatcher:onMessageDo: must specify either handler or target" )
	
	if not args.simpleResponses then
		loadReactiveSystem()
	end
	
	messageHandlers[ message ] = function( request, ... )

		-- This code is called more or less directly by the "we've received a
		-- request" code so it can rely on the fact that we've filtered out
		-- canceled requests at the level above.

		if target then
			sendResponseOrErrorToCallback( args.message, request, func, target, ... )
		else
			sendResponseOrErrorToCallback( args.message, request, func, ... )
		end
	
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Registers a handler for a specific incoming message. (Simple version.)
	
	\param message (string) The name of the message to be handled.
	\param handler (function) A function to handle the message.
	
 	The message handler function can return any number of results, which
	are interpreted as follows:
	
		- If the first result is a Behavior, it is kept alive as long as the client
			is interested, and any data posted to the node is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is a Promise, it is kept alive as long as the client
			is interested. When the promise resolves, the
			status (\c data or \c error) and any additional data is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is NOT a Reactive node, all results are passed
			back to the client via the request object, and the request is marked as done.
			All of the results passed through must be transitable (that is, scalar
			values, \c KSData, certain native object references, or tables
			containing only transitable values).
		- If the message handler fails with an error, the error status is
			passed to the client via the request object.
		- If the message handler returns no result, the request is marked as done.

]]--

function WFRequestDispatcher.onMessage( --[[string]] message,
					--[[function]] handler )

	check( 'string', 'function' )

	return WFRequestDispatcher.onMessageDo {
		message = message,
		handler = handler,
		onlyIfCallbackIsStillValid = true,
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Registers a handler for a specific incoming message, but does not allow
		the request to be canceled.
	
	\param message (string) The name of the message to be handled.
	\param handler (function) A function to handle the message. The handler
		is not called if the client has lost interest
		(that is, the callback has become invalid).
	
 	The message handler function can return any number of results, which
	are interpreted as follows:
	
		- If the first result is a Behavior, it is kept alive as long as the client
			is interested, and any data posted to the node is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is a Promise, it is kept alive as long as the client
			is interested. When the promise resolves, the
			status (\c data or \c error) and any additional data is passed
			to the client via the request object. Any subsequent results
			are ignored.
		- If the first result is NOT a Reactive node, all results are passed
			back to the client via the request object, and the request is marked as done.
			All of the results passed through must be transitable (that is, scalar
			values, \c KSData, certain native object references, or tables
			containing only transitable values).
		- If the message handler fails with an error, the error status is
			passed to the client via the request object.
		- If the message handler returns no result, the request is marked as done.

]]--

function WFRequestDispatcher.onMessageNoCancel( --[[string]] message,
												--[[function]] handler )

	check( 'string', 'function' )

	return WFRequestDispatcher.onMessageDo {
		message = message,
		handler = handler,
		onlyIfCallbackIsStillValid = false,
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Enables logging for request operations.
]]--

function WFRequestDispatcher.enableDiagnosticLogForRequestStats()

	if requestStats then return end
	
	requestStats = {}
	
	-- Don't spin up WFDiagnosticLog unless thread determines it's OK.
	-- It spins up promises and reactive, which are both fairly heavyweight.
	-- If those are already in the thread, then it's no big deal. But some threads
	-- can run without, and we shouldn't rock those boats.
	
	local WFDiagnosticLog = require 'WFDiagnosticLog'
	
	WFDiagnosticLog.addLogSectionFunc( function()
	
		local sortedRequestStats = {}
		local n = 0
		
		for _, v in pairs( requestStats ) do
			n = n + 1
			sortedRequestStats[ n ] = {
				messageId = v.messageId,
				count = v.count,
				errorCount = v.errorCount,
				cancelCount = v.cancelCount,
				min = string.format( "%.3f", v.min ),
				max = string.format( "%.3f", v.max ),
				total = v.total, -- will format after sort
				avg = string.format( "%.3f", v.total / v.count ),
			}
		end
		
		table.sort( sortedRequestStats, function( a, b )
			return a.total > b.total
		end )
		
		for _, v in pairs( sortedRequestStats ) do
			v.total = string.format( "%.3f", v.total )
		end
		
		sortedRequestStats.gcSize = collectgarbage( 'count' )
		
		return {
			name = "Messaging Stats :: " .. WFCurrentThread.DEBUG_getThreadName(),
			id = 'WFRequestDispatcher-' .. WFCurrentThread.DEBUG_getThreadName():gsub( ' :: ', '+' ),
			weight = 500,
			body = sortedRequestStats,
		}
	
	end )

end


--==============================================================================

function WFRequestDispatcher.getInstalledMessages()
	local result = {}
	for k in pairs( messageHandlers ) do
		result[ #result + 1 ] = k
	end
	return result
end
--==============================================================================

-- @mark Support for memory leak tool

--------------------------------------------------------------------------------

WFRequestDispatcher.onMessageDo {

		message = 'debug_dumpUniverse',
		simpleResponses = true,
		onlyIfCallbackIsStillValid = false,

		handler = function( options )
			
			local WFLeakCheck = require 'WFLeakCheck'
			if not options.containers then
					options.containers = {}
			end
			for k, v in pairs( messageHandlers ) do
				if type( k ) == 'string' then
					k = string.gsub( k, "%..+", "" )
					options.containers[ "MH_" .. k ] = v
				end
			end
			
			WFLeakCheck.dumpThisUniverseWithOptions( options )
			
		end,
	
}

--------------------------------------------------------------------------------

return WFRequestDispatcher
