--[[----------------------------------------------------------------------------

WFSyncAgent.lua

@author scouten 2013-04-11

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local KSClass = require 'KSClass'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFSyncAgent class

--------------------------------------------------------------------------------
--[[!
	\class WFSyncAgent
	
	A sync agent brings together one doc store and one external service.
	It provides the code for processing updates from the external service
	and determining what updates to send to that service.
	
	An "external service" could be any number of things, such as:
	
		* Oz server
		* another catalog within the host app (for instance, .lrcat file)
		* another instance of the host app on another device, though that use
			case is not tested or supported today
	
	See \c WFTestSyncPeer class as an example implementation of "external
	service" that is used for unit testing this sync agent code.
]]--

local syncAgentMethods = {

	_docStore = false,
		-- (WFDocStore) this client's doc store
	
	_syncPeer = false,
		-- (object) external sync peer
	
	_name = false,
		-- (string) name of external service (used to construct doc store slot names)

	_syncStatusDocName = false,
		-- (string) name of sync status doc to use

	_inboxSlotName = false,
		-- (string) name of doc store slot we use for incoming docs

	_outboxSlotName = false,
		-- (string) name of doc store slot we use for outgoing docs

	_getChangesP = false,
		-- (KSPromise) receive request in progress (if any)

	_sendUpdatesP = false,
		-- (KSPromise) send request in progress (if any)

	_hasReconciledInbox = false,
		-- (boolean) true if we've called reconcileInboxP at least once

	_hasProcessedOutbox = false,
		-- (boolean) true if we've called processOutboxP at least once

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the sync agent.
	
	\param docStore (\c WFDocStore) local doc store to be kept in sync with external service
	\param syncPeer (\c object) wrapper that can communicate with external service
	\param name (\c string) name of external service
]]--

function syncAgentMethods:__init( --[[WFDocStore]] docStore,
								  --[[object]] syncPeer,
								  --[[string]] name )

	check( 'WFSyncAgent', 'WFDocStore', 'table', '+string' )

	self._docStore = docStore

	self._syncPeer = syncPeer
	self:addCloseTarget( syncPeer )

	self._name = name
	
	self._syncStatusDocName = string.format( 'syncPeer-%s', name )
	self._inboxSlotName = string.format( '%s-inbox', name )
	self._outboxSlotName = string.format( '%s-outbox', name )

end

--------------------------------------------------------------------------------

-- @mark :getChangesP

--------------------------------------------------------------------------------

local function reconcileOneInboxDocumentP( self, inboxDoc )

	-- TEMPORARY: Naive implementation just stomps inbox value into main branch.
	-- Will need to replace this with an actual merge algorithm soon.

	local newDocValue = {}
	
	for k, v in pairs( inboxDoc ) do
		newDocValue[ k ] = v
	end
	
	newDocValue._slot = 'main'
	newDocValue._hash = nil
	newDocValue._changeCounter = nil
	
	return self._docStore:setDocValueP( newDocValue )
		:mapDataP( function()
		
			return self._docStore:quietlyDeleteSlotP( inboxDoc._id, self._inboxSlotName )
		
		end )

end

--------------------------------------------------------------------------------

local function reconcileInboxP( self )
	
	self._hasReconciledInbox = true
	
	-- Note that we delete items out of the inbox as we process it, so the
	-- volume of data here should be relatively small.

	return self._docStore:changedDocsP
		{
			fromChangeCounter = 0,
			slot = self._inboxSlotName,
		}
		
		:mapDataP( function( _, inboxDocs )
			
			if #inboxDocs == 0 then
				return 'data', true
			end
			
			local reconcileInboxDocsP = KSPromises.allP(

				function( inboxDoc )
				
					return reconcileOneInboxDocumentP( self, inboxDoc )
				
				end,

				inboxDocs,
				5 ) -- arbitrary choice for parallelism; tune if needed
			
			-- Continue until we empty the inbox.

			return reconcileInboxDocsP:mapDataP( function()
			
				return reconcileInboxP( self )
			
			end )
			
		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Request documents from external source and try to process them through
		to "main" slot.
	
	\return P< boolean > resolves to true if any new document values were received;
		false if nothing new from external service
]]--

function syncAgentMethods:getChangesP()

	check( 'WFSyncAgent' )
	
	-- Don't do this more than once at a time.
	
	if self._getChangesP then
		return self._getChangesP
	end
	
	-- Where were we before?
	
	local resultP = self._docStore:docValueP( self._syncStatusDocName, 'wfSyncAgent' )

		-- Extract the last changed marker from previous status (if any).

		:mapDataP( function( _, docValue )
		
			return 'data', docValue.continueFrom
		
		end )

		-- Ask external service for updates since that marker.
		
		:mapDataP( function( _, since )
		
			return self._syncPeer:changedDocsP( since )
		
		end )

		-- Process those updates.
		
		:mapDataP( function( _, updates )
		
			-- Shared error function for misbehaving sync peer code.
			
			local function gripe( message, ... )
				error( string.format( "sync peer %s: changedDocsP " .. message,
									  self._name, ... ), 2 )
			end
		
			-- Make sure sync peer is behaving properly.
		
			if type( updates ) ~= 'table' then
				gripe( "did not return a table" )
			end
			
			for i = 1, #updates do
			
				local updateItem = updates[ i ]
				
				if type( updateItem ) ~= 'table' then
					gripe( "item %d in updates response was not a table", i )
				end
				
				if type( updateItem._id ) ~= 'string' or #updateItem._id == 0 then
					gripe( "item %d in updates response: _id must be non-empty string",
						   i )
				end
				
				local timestampType = type( updateItem._timestamp )
				if timestampType ~= 'string' and timestampType ~= 'number' then
					gripe( "item %d in updates response: _timestamp must be string or number",
						   i )
				end
			
			end
		
			if #updates > 0 then

				return KSPromises.allP(
					function( updateItem )
					
						local docValue = {
							_id = updateItem._id,
							_slot = self._inboxSlotName,
							_syncStatus = {
								[ self._name ] = {
									timestamp = updateItem._timestamp,
								},
							}
						}
						
						for k, v in pairs( updateItem ) do
							if not k:match( "^_" ) then
								docValue[ k ] = v
							end
						end
						
						return self._docStore:setDocValueP( docValue )
					
					end, updates )

				:mapDataP( function()
				
					return reconcileInboxP( self )
						:mapDataP( function()
							return 'data', true
						end )
				
				end )
			
			else
			
				-- We still run the reconcile inbox step at least once
				-- just in case we had received documents in a previous
				-- session that we failed to process for some reason.
			
				if self._hasReconciledInbox then

					return 'data', false

				else

					return reconcileInboxP( self )
						:mapDataP( function()
							return 'data', false
						end )
					
				end
			
			end
		
		end )
		
	-- Make sure we only have one receive request in flight at a time.
	-- Also make sure we see this request through until completion.
	
	self._getChangesP = resultP
	
	resultP:whenResolvedDo( function()
	
		self._getChangesP = false
	
	end ):runUntilDone()

	return resultP

end

--------------------------------------------------------------------------------

-- @mark :sendUpdatesP

--------------------------------------------------------------------------------

local function convertOutboxDocToUpdateP( self, outboxDoc )

	local result = {}
	
	for k, v in pairs( outboxDoc ) do
	
		if k == '_id' then

			result._id = v
		
		elseif k == '_index' then
			
			result._index = v
			
		-- FIXME [scouten 2013-05-08]: Need to pull _timestamp out from
		-- previously-recorded sync status.
		
		elseif k:sub( 1, 1 ) ~= '_' then
			
			result[ k ] = v
		
		end
	
	end

	return KSPromises.resolvedPromiseP( 'data', result )
		-- TODO [scouten 2013-05-08]: This will probably become a promise
		-- when we have to go looking for previously-reconciled doc value.

end

--------------------------------------------------------------------------------

local function peerDidAcceptChangedDocP( self, outboxDoc )

	-- FIXME [scouten 2013-05-08]: Need to record reconciled value.
	
	return self._docStore:quietlyDeleteSlotP( outboxDoc._id, self._outboxSlotName )

end

--------------------------------------------------------------------------------

local function peerDidRejectChangedDocP( self, outboxDoc, perDocStatus )

	-- FIXME [scouten 2013-05-08]: Need to record failed sync attempt.

	return KSPromises.resolvedPromiseP( 'data', true )

end

--------------------------------------------------------------------------------

local function sendOutboundChangesP( self )

	self._hasProcessedOutbox = true
	
	-- Find out how many documents we're allowed to send at once.
	
	local limit = self._syncPeer:maxUpdatesPerSend()
	if type( limit ) ~= 'number' or limit < 1 or math.floor( limit ) ~= limit then
		error( "WFSyncAgent: sync peer returned invalid response to :maxUpdatesPerSend call: "
			   .. tostring( limit ) .. " (must be a positive integer)" )
	end

	-- Note that we delete items out of the inbox as we process it, so the
	-- volume of data here should be relatively small.

	return self._docStore:changedDocsP
		{
			fromChangeCounter = 0,
			slot = self._outboxSlotName,
			limit = limit,
		}
		
		:mapDataP( function( _, outboxDocs )
			
			if #outboxDocs == 0 then
				return 'data', false
			end

			-- Call syncPeer:sendUpdatesP and process response.
			
			local updates = {}
			
			for i = 1, #outboxDocs do
				outboxDocs[ i ]._index = i
			end
			
			return KSPromises.allP( function( outboxDoc )
			
				return convertOutboxDocToUpdateP( self, outboxDoc )
					:mapDataP( function( _, outboxDoc )
					
						local index = outboxDoc._index
						outboxDoc._index = nil
						
						updates[ index ] = outboxDoc
						
						return 'data', true
					
					end )
			
			end, outboxDocs )
			:mapDataP( function()
			
				return self._syncPeer:sendUpdatesP( updates )
			
			end )
			:mapP( function( status, perDocStatus )
			
				if status == 'data' then
					
					-- If we succeeded, then remove all documents
					-- from outbox and advance the "continue from"
					-- pointer accordingly.
					
					return KSPromises.allP( function( outboxDoc )

						return peerDidAcceptChangedDocP( self, outboxDoc )

					end, outboxDocs )

					:mapDataP( function()

						return self._docStore:applyOverlayP {
							_id = self._syncStatusDocName,
							_slot = 'wfSyncAgent',
							outboundContinueFrom = outboxDocs.continueFrom,
						}
					 
					end )
				
				elseif type( perDocStatus ) == 'table' then

					-- We have a per-document status. Need to figure out
					-- what succeeded and what failed.
					
					local didHaveAnyFailures
					
					return KSPromises.allP( function( outboxDoc )
					
						if perDocStatus[ outboxDoc._id ] then
							
							didHaveAnyFailures = true
							
							peerDidRejectChangedDocP( self, outboxDoc,
													  perDocStatus[ outboxDoc._id ] )
						
						else
							
							peerDidAcceptChangedDocP( self, outboxDoc )
						
						end
					
					end, outboxDocs )
					:mapDataP( function()
					
						if not didHaveAnyFailures then
						
							return 'error',
								   "sync peer returned a table that did not have any errors"
							
						else
						
							return 'error',
								   "sync peer rejected one or more documents"

						end
					
					end )
				
				else

					-- If we didn't get a per-document status, just treat
					-- this as an overall failure.
				
					return status, perDocStatus
				
				end
			
			end )
			:mapDataP( function()
			
				-- Continue until we empty the outbox.
	
				return sendOutboundChangesP( self )
				
			end )
			
		end )

end

--------------------------------------------------------------------------------

local function putOneDocumentInOutboxP( self, possibleOutboundDoc )

	-- FIXME: Need to cross-check with previously reconciled external value.
	-- For now, just blindly put entire doc value into outbox.
	
	return self._docStore:copySlotToSlotP( possibleOutboundDoc._id,
										   'main', self._outboxSlotName )

end

--------------------------------------------------------------------------------

function syncAgentMethods:sendUpdatesP()

	check( 'WFSyncAgent' )

	-- Don't do this more than once at a time.
	
	if self._sendUpdatesP then
		return self._sendUpdatesP
	end
	
	-- What needs to be sent?
	
	local docStore = self._docStore
	
	local resultP = docStore:docValueP( self._syncStatusDocName, 'wfSyncAgent' )

		-- Extract the last checked marker from previous status (if any).

		:mapDataP( function( _, docValue )
		
			return 'data', docValue.outboundContinueFrom
		
		end )

		-- Ask local doc store for changes in 'main' slot since then.
		
		:mapDataP( function( _, outboundContinueFrom )
		
			return docStore:changedDocsP
				{
					fromChangeCounter = outboundContinueFrom
											and ( outboundContinueFrom + 1 ) or 0,
					slot = 'main',
				}
		
		end )
		
		:mapDataP( function( _, possibleOutboundDocs )
		
			if #possibleOutboundDocs > 0 then

				return KSPromises.allP(
					function( possibleOutboundDoc )
					
						return putOneDocumentInOutboxP( self, possibleOutboundDoc )

					end, possibleOutboundDocs )

				:mapDataP( function()
				
					return sendOutboundChangesP( self )
						:mapDataP( function()
							return 'data', true
						end )
				
				end )
			
			else
			
				-- We still run the process outbox step at least once
				-- just in case we had marked documents to send in a previous
				-- session that we failed to process for some reason.
			
				if self._hasProcessedOutbox then

					return 'data', false

				else

					return sendOutboundChangesP( self )
						:mapDataP( function()
							return 'data', false
						end )
					
				end
			
			end
		
		end )

	-- Make sure we only have one send request in flight at a time.
	-- Also make sure we see this request through until completion.
	
	self._sendUpdatesP = resultP
	
	resultP:whenResolvedDo( function()
	
		self._sendUpdatesP = false
	
	end ):runUntilDone()

	return resultP

end

--------------------------------------------------------------------------------

return KSClass 'WFSyncAgent'( syncAgentMethods )
