--[[----------------------------------------------------------------------------

WFTestSyncPeer.lua

@author scouten 2013-04-11

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.lua.threading
local WFTimer = require 'WFTimer'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local KSClass = require 'KSClass'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFTestSyncPeer class

--------------------------------------------------------------------------------
--[[!
	\class WFTestSyncPeer

	This class implements a sync peer that can be used for unit testing
	/c WFSyncEngine. It also serves as a reference implementation for other
	sync peer agents that talk to "real" external sources.
	
	This sync peer does not actually sync with an external source, but rather
	mocks the sync interface using local / in-memory storage.
]]--

local testSyncPeerMethods = {

	_docStore = false,
		-- (WFDocStore) in-memory doc store used to store simulated external content

	_delayNextChangedDocsP = false,
		-- (optional number) seconds to delay next :changedDocsP response

	_overrideNextChangedDocsP_status = false,
		-- (string) status for next :changedDocsP response
	
	_overrideNextChangedDocsP_value = false,
		-- (any) value to use in place of normal :changedDocsP response

	_overrideNextSendUpdatesP_status = false,
		-- (string) status for next :sendUpdatesP response
	
	_overrideNextSendUpdatesP_value = false,
		-- (any) value to use in place of normal :sendUpdatesP response

	_partiallyFailNextSendUpdatesP_value = false,
		-- (table) if present, :sendUpdatesP should fail with this table
		-- as the per-document status result

	_shouldOverrideNextMaxUpdatesPerSendResponse = false,
		-- (boolean) true to override next :maxUpdatesPerSend response
	
	_overrideNextMaxUpdatesPerSendResponse = false,
		-- (any) if _shouldOverrideNextMaxUpdatesPerSendResponse is true,
		-- use this value in response

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the test sync peer.
]]--

function testSyncPeerMethods:__init()

	self._docStore = WFDocStore( ':memory:' )
	self:addCloseTarget( self._docStore )

end

--------------------------------------------------------------------------------

local function changedDocsP_imp( self, since )

	-- Are we overriding response?
	
	if self._overrideNextChangedDocsP_status then
	
		local resultP = KSPromises.resolvedPromiseP( self._overrideNextChangedDocsP_status,
													 self._overrideNextChangedDocsP_value )
		
		self._overrideNextChangedDocsP_status = false
		rawset( self, '_overrideNextChangedDocsP_value', false )
			-- Using rawset in case the value was nil.
											
		return resultP

	else

		return self._docStore:changedDocsP
			{
				fromChangeCounter = since and ( since + 1 ) or 0,
				slot = 'main',
			}
			:mapDataP( function( _, changes )
			
				for i = 1, #changes do
				
					local change = changes[ i ]
					
					change._slot = nil
					change._hash = nil
					change._timestamp = change._changeCounter
					change._changeCounter = nil
				
				end
				
				return 'data', changes
			
			end )
		
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the set of documents that have changed since a specified point.
	
	\param since (optional, string or number) a previous change-counter value
	
	Documents that have been changed since this marker will be returned,
	up to an arbitrary limit enforced by the external peer. (The external peer
	should impose some reasonable limit so as not to swamp the client.)
	
	If \c since is \c nil, then start from the very beginning of time.
	
	\return <code>P< table ></code> An array of fully-formed document values.
	
	This array/table also contains a top-level element named \c continueFrom
	whose value is the \c since value to use in the next \c :changedDocs call.
]]--

function testSyncPeerMethods:changedDocsP( --[[optional, number or string]] since )

	check( 'WFTestSyncPeer', '?number|string' )

	-- Do we need a delay for testing?
	
	local delay = self._delayNextChangedDocsP
	
	if delay then
	
		self._delayNextChangedDocsP = false
		
		return WFTimer.afterDelayP( delay )
			:mapDataP( function()

				return changedDocsP_imp( self, since )
			
			end )
		
	else
	
		return changedDocsP_imp( self, since )
	
	end

end

--------------------------------------------------------------------------------

local function validateNestedTable( t )

	for k, v in pairs( t ) do
	
		local tk = type( k )
		local tv = type( v )
	
		if k == '_id' then
		
			assert( tv == 'string', "_id must be a string" )
			assert( #v > 0, "_id must be a non-empty string" )
			
		elseif k == '_timestamp' then
			
			assert( tv == 'number' or tv == 'string', "_timestamp must be number or string" )
		
		-- TODO [scouten 2013-05-01]: Add support for _delete, __merge, etc.
		
		elseif tk == 'string' and ( k:match( '^_' ) or #k == 0 ) then
			
			error( "Illegal key: " .. k )

		end
		
		if tv == 'table' then
			validateNestedTable( v )
		end
	
	end

end

--------------------------------------------------------------------------------

local function sendOneUpdateP( self, update )

	check( 'WFTestSyncPeer', 'table' )
	
	-- Validate the update table.
	
	for k, v in pairs( update ) do
	
		local tk = type( k )
		local tv = type( v )
	
		if tk ~= 'string' then
		
			error( string.format( "Illegal key of type %s (%s)", tk, k ) )
		
		elseif k == '_id' then
		
			assert( tv == 'string', "_id must be a string" )
			assert( #v > 0, "_id must be a non-empty string" )
			
		elseif k == '_timestamp' then
			
			assert( tv == 'number' or tv == 'string', "_timestamp must be number or string" )
		
		-- TODO [scouten 2013-05-01]: Add support for _delete, __merge, etc.
		
		elseif tk == 'string' and ( k:match( '^_' ) or #k == 0 ) then
			
			error( "Illegal key: " .. k )

		end
		
		if tv == 'table' then
			validateNestedTable( v )
		end
	
	end
	
	-- Update is valid: Apply overlay in test DB.
	
	local overlay = {}
	
	for k, v in pairs( update ) do

		if k == '_timestamp' then
			overlay._changeCounter = v
		else
			overlay[ k ] = v
		end
	
	end
	
	overlay._slot = 'main'
	
	return self._docStore:applyOverlayP( overlay )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sends updates to the remote server.
	
	Updates are sent in the form of deltas against a previously-agreed checkpoint,
	similar to the \c :applyOverlayP call on \c WFDocStore.
	
	\param updates (table) array of document updates to send
	
	Each entry in the array will contain a delta to apply to a document, in the
	form of \c :applyOverlay. The delta table will also contain the following
	entries:
	
		* \c _id - document ID
		* \c _timestamp - the revision on the peer that is being superceded
			(if this does not match, the peer should reject the change)
	
	The sync peer code may send these updates simultaneously or serially. If
	sent serially, they should be sent in the order of appearance in the updates
	aray.
	
	\return <code>P< true ></code> resolves when all updates have been sent
	
	If any updates can not be sent, will resolve with error status. The "message"
	will contain a table mapping document ID to error status.
]]-- 

function testSyncPeerMethods:sendUpdatesP( --[[table]] updates )

	-- Are we overriding response?

	if self._overrideNextSendUpdatesP_status then
	
		local resultP = KSPromises.resolvedPromiseP( self._overrideNextSendUpdatesP_status,
													 self._overrideNextSendUpdatesP_value )
		
		self._overrideNextSendUpdatesP_status = false
		rawset( self, '_overrideNextSendUpdatesP_value', false )
			-- Using rawset in case the value was nil.
											
		return resultP

	else

		check( 'WFTestSyncPeer', 'table' )
	
		local errors = self._partiallyFailNextSendUpdatesP_value or {}
		self._partiallyFailNextSendUpdatesP_value = false
		
		return KSPromises.allP(
			function( update )
	
				local id = ( type( update ) == 'table' and update._id ) or '(unknown)'
				
				-- Don't send the update if testing code requested that we
				-- fail this document. (If you're cargo-culting this code, please
				-- don't copy this part; this is just for testing.)
				
				if errors[ id ] then
					return KSPromises.resolvedPromiseP( 'data', true )
				end

				-- Send the update.

				return sendOneUpdateP( self, update )
					:mapP( function( status, message )
					
						-- If we fail on one send, we continue with the other
						-- attempts and roll up the error status.
						
						if status ~= 'data' then
	
							errors[ id ] = message
	
							return 'data', true -- continue with other updates
	
						end
						
						return status, message
					
					end )
	
			end, updates )
		
			:mapDataP( function( status, message )
			
				-- If any of the requests failed, translate back to error now.
			
				if next( errors ) then
					return 'error', errors
				else
					return status, message
				end
			
			end ):runUntilDoneP()
	
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the maximum number of updates to send to the remote peer at once.
	
	\return (number) maximum number of updates to send to the remote peer at once.
]]--

function testSyncPeerMethods:maxUpdatesPerSend()

	-- A "normal" sync peer would probably just return a constant (positive integer).
	-- The override stuff is here so we can test WFSyncAgent's handling of faulty
	-- responses.

	if self._shouldOverrideNextMaxUpdatesPerSendResponse then

		self._shouldOverrideNextMaxUpdatesPerSendResponse = false
		return rawget( self, '_overrideNextMaxUpdatesPerSendResponse' )

	else

		return 4  -- depends on remote peer
	
	end

end

--------------------------------------------------------------------------------

-- @mark testing configuration

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Provide access to the backing doc store.
	
	This is not part of the WF sync peer interface. It allows test code to
	access the underlying doc store so it can simulate changes originating
	on the external side of the connection.
	
	\return (WFDocStore) doc store
]]--

function testSyncPeerMethods:docStore()

	check( 'WFTestSyncPeer' )

	return self._docStore

end

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Introduce a delay into the next :changedDocsP response.
	
	\param (number) # of seconds to delay response
]]--

function testSyncPeerMethods:delayNextChangedDocsP( --[[number]] delay )

	check( 'WFTestSyncPeer', 'number' )

	assert( delay > 0, "ain't got no time machine here" )
	
	self._delayNextChangedDocsP = delay

end

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Set the value for the next :changedDocsP response.
	
	This is not part of the WF sync peer interface. It allows test code to
	override the next response to :changedDocsP so that the test code can
	(for example) verify correct processing of incorrect results.
	
	\param status (string) promise resolution status
	\param response (anything) value to send back as part of next :changedDocsP response
]]--

function testSyncPeerMethods:overrideNextChangedDocsPResponse( --[[string]] status,
															   --[[any]] response )

	self._overrideNextChangedDocsP_status = status
	self._overrideNextChangedDocsP_value = response

end

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Set the value for the next :sendUpdatesP response.
	
	This is not part of the WF sync peer interface. It allows test code to
	override the next response to :sendUpdatesP so that the test code can
	(for example) verify correct processing of incorrect results.
	
	\param status (string) promise resolution status
	\param response (anything) value to send back as part of next :sendUpdatesP response
]]--

function testSyncPeerMethods:overrideNextSendUpdatesPResponse( --[[string]] status,
															   --[[any]] response )

	self._overrideNextSendUpdatesP_status = status
	self._overrideNextSendUpdatesP_value = response

end

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Cause the next :sendUpdatesP response to fail with
		a given per-document status response.
	
	This is not part of the WF sync peer interface. It allows test code to
	override the next response to :sendUpdatesP so that the test code can
	(for example) verify correct processing of incorrect results.
	
	\param response (table) per-document to send back as part of next :sendUpdatesP
		error response
]]--

function testSyncPeerMethods:partiallyFailNextSendUpdatesPResponse( --[[table]] response )

	check( 'WFTestSyncPeer', 'table' )

	self._partiallyFailNextSendUpdatesP_value = response

end

--------------------------------------------------------------------------------
--[[!
	\brief (Testing only) Set the value for the next :maxUpdatesPerSend response.
	
	This is not part of the WF sync peer interface. It allows test code to
	override the next response to :maxUpdatesPerSend so that the test code can
	(for example) verify correct processing of incorrect results.
	
	\param response (anything) value to send back on next :maxUpdatesPerSend call
]]--

function testSyncPeerMethods:overrideNextMaxUpdatesPerSendResponse( --[[any]] response )

	self._shouldOverrideNextMaxUpdatesPerSendResponse = true
	rawset( self, '_overrideNextMaxUpdatesPerSendResponse', response )

end

--------------------------------------------------------------------------------

return KSClass 'WFTestSyncPeer'( testSyncPeerMethods )
