--[[----------------------------------------------------------------------------

WFTimer.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'

	-- wichitafoundation.core.os
local KSDateImp = require 'KSDateImp'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'

	-- wichitafoundation.core.utils
local WFArguments = require 'WFArguments'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local math_ceil = assert( math.ceil )

local timeNowForAnimation = assert( KSDateImp.timeNowForAnimation )

local requestPulseAfterDelay = assert( WFCurrentThread.requestPulseAfterDelay )


--============================================================================--
--[[!
	\class WFTimer
	\brief This namespace allows you to cause events to happen after
	a specified delay.
	
	These are not high-precision timers. "Nearby" timer events (currently defined
	as within the same 1/16 second interval) are coalesced together to save
	on system resources.
	
	This namespace wraps the underlying native APIs.
	
	\author Eric Scouten <scouten@adobe.com> 2012-03-20
--]]

-- @mark WFTimer namespace

local WFTimer = {}

--------------------------------------------------------------------------------

local timerPromises = {}
	-- key: (number) time when timer should fire
	-- value: (KSPromise) promise that fires at that time

local timerResolvers = {}
	-- key: (number) time when timer should fire
	-- value: (KSPromiseResolver) resolver for promise in timerPromises

local kTimerTickPrecision = 16
	-- We are precise to a 16th of a second

--------------------------------------------------------------------------------
--[[!
	\brief Returns a promise that resolves after the specified time interval
	has elapsed.
	
	This is not a high-precision timer. "Nearby" timer events (currently defined
	as within the same 1/16 second interval) are coalesced together to save
	on system resources.
	
	\param delay (number) how many seconds to wait (may be fractional)
	
	\return <code>P< true ></code> this promise will resolve to true after specified
		time interval has elapsed
]]--

function WFTimer.afterDelayP( --[[number]] delay )

	check( 'number' )

	if delay < 0 then
		error( "WFTimer.afterDelayP: delay must be a non-negative number", 2 )
	end
	
	local currentTime = timeNowForAnimation()
	
	local targetTimerTicks = math_ceil( ( currentTime + delay ) * kTimerTickPrecision )
	
	local existingTimerP = timerPromises[ targetTimerTicks ]
	
	if existingTimerP then
		return existingTimerP
	end
	
	local timerP, resolver = KSPromises.promiseP()
	
	timerPromises[ targetTimerTicks ] = timerP
	timerResolvers[ targetTimerTicks ] = resolver
	
	local timerDelay = ( targetTimerTicks / kTimerTickPrecision ) - currentTime
	
	requestPulseAfterDelay( timerDelay )
	
	return timerP

end

--------------------------------------------------------------------------------
--[[!
	\brief Call this function once after the specified time interval
	has elapsed.
	
	This is not a high-precision timer. "Nearby" timer events (currently defined
	as within the same 1/16 second interval) are coalesced together to save
	on system resources.
	
	\param delay (number) how many seconds to wait (may be fractional)
	\param func (function) function to be called when current thread becomes idle
]]--

function WFTimer.afterDelayDo( --[[number]] delay,
							   --[[function]] func )

	check( 'number' )

	if delay <= 0 then
		error( "WFTimer.afterDelayDo: delay must be a positive number", 2 )
	end

	WFTimer.afterDelayP( delay ):whenResolvedDo( func ):runUntilDone()

end

--------------------------------------------------------------------------------

WFCurrentThread.registerPulseHandler( function()

	local currentTimerTick = math_ceil( timeNowForAnimation() * kTimerTickPrecision )

	local resolversToRun = {}
	local timerTicksToRemove = {}

	for targetTime, resolver in pairs( timerResolvers ) do
	
		if currentTimerTick >= targetTime then

			resolversToRun[ resolver ] = true
			timerTicksToRemove[ targetTime ] = true
		
		end

	end
	
	for timerTick in pairs( timerTicksToRemove ) do
	
		timerPromises[ timerTick ] = nil
		timerResolvers[ timerTick ] = nil
	
	end

	for resolver in pairs( resolversToRun ) do
	
		resolver( 'data', true )
	
	end

end )

--------------------------------------------------------------------------------

return WFTimer
