--[[----------------------------------------------------------------------------

WFWorkerThread.lua

@author scouten 2011-09-21

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wichitafoundation.core.logging
local KSLogging = require 'KSLogging'

	-- wichitafoundation.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFMessagingRequest = require 'WFMessagingRequest'

	-- wichitafoundation.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- wichitafoundation.core.threading
local WFMessagingProcessor = require 'WFMessagingProcessor'

	-- wichitafoundation.core.utils
local KSClass = require 'KSClass'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'


--------------------------------------------------------------------------------

local workerThreadLog = KSLogging.log 'WFWorkerThread'
--	:setLevel( KSLogging.DEBUG )

local requestELog = KSLogging.log 'WFWorkerThread :: requestE'
--	:setLevel( KSLogging.DEBUG )

local check = assert( WFArguments.check )
local WFWorkerThread = 'WFWorkerThread'

local WFCloseSupport_fetchCached = WFCloseSupport.fetchCached
local WFCloseSupport_hasBeenOrWillBeClosed = WFCloseSupport.hasBeenOrWillBeClosed
local WFCloseSupport_makeCloser = assert( WFCloseSupport.makeCloser )

local KSTableUtils_makeFullyWeak = KSTableUtils.makeFullyWeak

local WFWorkerThread_runUntilDonePool = KSPromises.makeRunUntilDonePool()

local assert = assert
local select = select
local error = error

local KSPromises_promiseForResolverP = assert( KSPromises.promiseForResolverP )
local WFArguments_check = assert( WFArguments.check )

local debug_print = assert( debug.print )
local string_format = assert( string.format )


--============================================================================--

local CHECK_REQUEST_CLOSURE = _G.DEBUG

--------------------------------------------------------------------------------

local GCMonitor = KSClass 'WFWorkerThreads_GCMonitor' {

	_case = false,
	_worker = false,
	_message = false,

	__tostring = function( self )
			return string_format( "%s( %s : %s )",
					tostring( self._case or nil ),
					tostring( self._worker or nil ),
					tostring( self._message or nil ) )
		end,
		
	__init = function( self, case, worker, message )
		self._case = case
		self._worker = worker
		self._message = message
	end,
		
	WFCloseSupport.addCloseSupportMethods

}

--------------------------------------------------------------------------------

local function makeGCMonitor( case, worker, message, backtrace )

	if not CHECK_REQUEST_CLOSURE then
		return
	end
	
	local obj = GCMonitor( case, worker, message )
	
	return WFCloseSupport_makeCloser( obj, nil, backtrace )

end


--============================================================================--

-- @mark raw promise request processing

--------------------------------------------------------------------------------

local promiseResolverForRawRequestRef = KSTableUtils.makeWithMode 'v'
	-- key: (light userdata) raw request ref
	-- value: (RequestPromiseResolver) promise resolver for that raw request

--------------------------------------------------------------------------------

local requestPromiseResolverMethods = {

	_didResolve = false,
		-- (boolean) sanity check: have we resolved this promise already?
	
	_request = false,
		-- (WFMessagingRequest) request that will provide data for this promise

	_messagingProcessor = false,
		-- (WFMessagingProcessor) worker messaging processor

	_message = false,
		-- (string) message ID
	
	_params = false,
		-- (array) additional parameters to pass along to worker processor
		
	_monitor = false,
		-- a hook for storing a monitor object to detect when we GC rather
		-- than close a request

	WFCloseSupport.addCloseSupportMethods,
	KSPromises.addPromiseResolverMethods,
	
}

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:__init( messagingProcessor, message, ... )

	WFArguments_check( '_', '_', 'string' )

	--checkLimbo() --??

	self._messagingProcessor = messagingProcessor
	self._message = message
	self._params = { n = select( '#', ... ), ... }

end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:close()
	
	-- Sever links to aid GC.
	
	self._messagingProcessor = false
	self._request = false
	self._message = false
	self._params = false

end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:startResolution( priority )
	
	assert( self._messagingProcessor )

	if self._request then
		return
	end
	
	self._request = WFMessagingRequest.create( self._messagingProcessor,
											   self._message,
											   unpack( self._params, 1, self._params.n ) )
	
	self:addCloseTarget( self._request )
	
	promiseResolverForRawRequestRef[ self._request:getRawRequestRef() ] = self

end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:handleResponse( requestStatus, statusSelector, ... )

--	debug_print( "handleResponse: message = ", self._message,
--				 "  workerProcessor = ", self._messagingProcessor,
--				 "  requestStatus = ", requestStatus,
--				 "  statusSelector = ", statusSelector )
--
--	for i = 1, select( '#', ... ) do
--		debug_print( string_format( "Received [ %d ]: %s", i,
--									tostring( ( select( i, ... ) ) ) ) )
--	end

	-- Make sure we only resolve this promise once.

	if self._didResolve then
		debug_print "Attempting to handle second call from callback"
		return
	end
	
	self._didResolve = true
	
	-- Figure out what to do with result.
	
	local nArgs = select( '#', ... )
	
	if requestStatus == 'orphaned' then
	
		self:resolvePromise( 'error', "Request was orphaned by worker thread" )
	
	elseif statusSelector == 'error' then
	
		local message = ( ... )
		
		if message then
			self:resolvePromise( 'error', ... )
		else
			self:resolvePromise( 'error', "< Unknown Error >" )
		end
	
	elseif nArgs == 0 then -- cancel
	
		self:resolvePromise()
	
	else
	
		self:resolvePromise( 'data', ... )
	
	end
	
	-- Note: Every code path above results in a call to :resolvePromise
	-- which in turn closes this promise resolver object, so there's no
	-- further need for us to clean up here.
	
end

--------------------------------------------------------------------------------

function requestPromiseResolverMethods:setMonitor( monitor )

	if not monitor then
		return
	end
	
	self._monitor = monitor
	
	self:addCloseTarget( monitor )

end

--------------------------------------------------------------------------------

local requestPromiseResolver_class
	= KSClass 'RequestPromiseResolver'( requestPromiseResolverMethods )

--------------------------------------------------------------------------------

local function requestP( monitor, messagingProcessor, message, ... )

	assert( messagingProcessor )

	local resolver = requestPromiseResolver_class( messagingProcessor, message, ... )
	
	resolver:setMonitor( monitor )
	
	return KSPromises_promiseForResolverP( resolver, 'retained | closeTarget' )
	
end


--============================================================================--

-- @mark raw event stream request processing

--------------------------------------------------------------------------------

local internalEventStreamForRawRequestRef = KSTableUtils.makeWithMode 'v'
	-- key: (light userdata) raw request ref
	-- value: (KSEventStream) event stream for that raw request

--------------------------------------------------------------------------------

local function requestE( monitor, messagingProcessor, message, ... )

	requestELog:debug( "sending requestE( %s, %s, %s, %s )",
					   messagingProcessor, message, ... )
			
	assert( messagingProcessor )

	local internalResultE = KSReactive.sourceE()

	-- Send message off to other thread. Remember the result stream so we can
	-- post results back when we have them.

	local success, requestOrMessage = pcall( WFMessagingRequest.create, messagingProcessor, message, ... )
	
	if not success then
	
		error( string_format( "WFWorkerThread:requestE: Attempting to send a request to a closed messaging processor?"
							  .. "\n\n   messagingProcessor = %s\n\n   messageId = %s\n\n   error = %s",
							  messagingProcessor, message, requestOrMessage ), 2 )

	end

	local request = requestOrMessage

	internalEventStreamForRawRequestRef[ request:getRawRequestRef() ] = internalResultE

	-- Stash the message name in the message body. Might be helpful for debugging.

	rawset( internalResultE, '_message', message )
	
	if monitor then
		internalResultE:addCloseTarget( monitor )
		rawset( internalResultE, '_gcmonitor', monitor )
	end

	-- Tie the lifetime of the request to the lifetime of this event stream.

	rawset( internalResultE, '_request', request )
	internalResultE:addCloseTarget( request )
	
	-- Return a relay off of internalResultE, so callers can't interfere and
	-- post extra results.

	local publicResultE = internalResultE:relayE()
	publicResultE:addCloseTarget( request )
	publicResultE:addCloseTarget( internalResultE )
	
	return publicResultE

end

--------------------------------------------------------------------------------

local function handleEventStreamResponse( internalResultE, requestStatus, statusSelector, ... )

	requestELog:debug( "processing requestE response:\n  message %s\n  requestStatus = %s\n  statusSelector = %s\n  params = %s, %s, %s, %s",
					   internalResultE._message,
					   requestStatus,
					   statusSelector,
					   ... )
					
	-- If request was orphaned, signal that as an error and close event stream.
	
	if requestStatus == 'orphaned' then
		internalResultE:postError( "Request was orphaned by worker thread" )
		internalResultE:close()
		return
	end
	
	local nArgs = select( '#', ... )

	-- Post results from this response. Special case: If no extra values (... is empty)
	-- and the worker thread is saying it's done, don't post an empty result before closing.
	
	if statusSelector == 'error' then

		if nArgs > 0 then
			internalResultE:postError( ... )
		else
			internalResultE:postError( "< Unknown Error >" )
		end

	elseif requestStatus ~= 'done' or nArgs > 0 then

		internalResultE:post( ... )

	end

	-- If worker thread is done, close out the event stream.

	if requestStatus == 'done' then
		internalResultE:close()
	end
	
end


--============================================================================--

-- @mark raw message response handling

--------------------------------------------------------------------------------

local function handleResponse( rawRequestRef, requestStatus, statusSelector, ... )

	local promiseResolver = WFCloseSupport.fetchCached( promiseResolverForRawRequestRef,
														rawRequestRef )
	
	if promiseResolver then
		promiseResolver:handleResponse( requestStatus, statusSelector, ... )
		return
	end
	
	local eventStream = WFCloseSupport.fetchCached( internalEventStreamForRawRequestRef,
													rawRequestRef )
	
	if eventStream then
		handleEventStreamResponse( eventStream, requestStatus, statusSelector, ... )
	end
	
end

WFCurrentThread.registerResponseDispatchHandler( handleResponse )


--============================================================================--

-- @mark raw event processor loading

--------------------------------------------------------------------------------

local function rawLoadMessagingProcessorP( initScriptName, optArg )

	-- Should only be called if we don't have a valid MP reference in place.

	local selector = assert( initScriptName )

	if optArg then
		selector = string.format( '%s :: %s', initScriptName, optArg )
	end

	local resultMP = WFMessagingProcessor.createForId( selector )

	local resultP = requestP( nil, resultMP, '_notifyWhenReady' )
	
	-- Close the message processor if we fail
	
	resultP = resultP:mapP( function( status, ... )
		if status ~= 'data' then
			resultMP:close()
		end
	end )
	
	-- Having started initialization, we always want to finish this much
	
	resultP:runUntilDone( WFWorkerThread_runUntilDonePool )
	
	-- If we success, we return the message processor as data
	
	return resultP:mapDataP( function()
	
		return 'data', resultMP
	
	end )
	
end

--------------------------------------------------------------------------------

local cachedMessagingProcessorPromises = KSTableUtils_makeFullyWeak()
	-- key: (string) path to init script + (optional: ':' + optArg )
	-- value: (P< WFMessagingProcessor >) (only while loading is in flight)

--------------------------------------------------------------------------------

local cachedMessagingProcessors = KSTableUtils_makeFullyWeak()
	-- key: (string) path to init script + (optional: ':' + optArg )
	-- value: (WFMessagingProcessor)

--------------------------------------------------------------------------------

local function cachedLoadMessagingProcessorP( initScriptName, optArg )

	local cacheKey = optArg and string.format( '%s:%s', initScriptName, optArg )
							or initScriptName

	local messagingProcessor = WFCloseSupport_fetchCached( cachedMessagingProcessors, cacheKey )
	if messagingProcessor then
		return KSPromises.resolvedPromiseP( 'data', messagingProcessor )
	end
	
	local mpLoaderP = WFCloseSupport_fetchCached( cachedMessagingProcessorPromises, cacheKey )
	if mpLoaderP then
		return mpLoaderP
	end
	
	mpLoaderP = rawLoadMessagingProcessorP( initScriptName, optArg )
	cachedMessagingProcessorPromises[ cacheKey ] = mpLoaderP
	
	mpLoaderP:whenResolvedDo( function( status, messagingProcessorOrMessage )
	
		if status == 'data' then
			cachedMessagingProcessors[ cacheKey ] = messagingProcessorOrMessage
		end
		
		cachedMessagingProcessorPromises[ cacheKey ] = nil
	
	end ):runUntilDone( WFWorkerThread_runUntilDonePool )
	
	return mpLoaderP

end


--============================================================================--
--[[!
	\brief An object of this type acquires a worker thread instance and releases it
	when no longer needed.
	
	This class wraps the underlying native APIs and is the preferred mechanism
	for accessing worker threads (also called event processors).
--]]

-- @mark WFWorkerThread class

local workerThreadMethods = {

	_loadP = false,
		-- (P< WFMessagingProcessor >) promise that the event processor will load

	_messagingProcessor = false,
		-- (WFMessagingProcessor) filled in once MP has confirmed that is has loaded
	
	_initScriptName = false,
		-- (string) path by which this MP was initialized
	
	_optArg = false,
		-- (string) optional specialization argument for MP
	
	_creationTraceback = false,
		-- (string) in debug builds, traceback from which this object was created
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local refsForMessagingProcessor = {} -- or should this be weak keyed?
	-- key: (WFMessagingProcessor)
	-- value: (fully weak table)
		-- key: (WFWorkerThread)
		-- value: true

--------------------------------------------------------------------------------
--[[!
	\brief Constructs a worker thread for a named initialization script and
	optional specialization.

	Worker thread instances are shared and reference-counted.
	If two client threads specify a worker thread using this name and specialization pair,
	they address the same actual worker thread. When all client references
	are closed or garbage-collected, the worker thread terminates.
 
 	\param	initScriptName	The name of a Lua script.

		The reserved value <code>":self:"</code> sends this request to the current thread.
		(This usage is under review and may become deprecated in the future.
		 Please contact EricS before using this form of the API.)
		 \warning If you use the <code>":self:"</code> form, you <b>must not</b>
		 	hold a persistent reference to this \c WFWorkerThread object. If you do, you
		 	will prevent the thread from being released when other threads no longer reference it.
	\param optArg	A specialization that distinguishes this thread from any other
		that is initialized with the same script.
		Must be \c nil or a string, or a promise that resolves to \c nil or a string.

	\return A reference to the new worker thread.

]]--

function workerThreadMethods:__init( --[[string]] initScriptName,
									 --[[opt string]] optArg )

	check( WFWorkerThread, 'string', '_' )

	self._initScriptName = initScriptName

	if initScriptName == ':self:' and optArg ~= nil then
		error( "WFWorkerThread: can not specify optArg when initScriptName is ':self:'", 2 )
	end
	
	if _G.DEBUG then
		self._creationTraceback = debug.traceback( "created via", 4 )
	end
		
	if KSPromises.isPromise( optArg ) then
	
		self._loadP = optArg:mapDataP( function( _, actualArg )
		
			if actualArg and type( actualArg ) ~= 'string' then
				return 'error',
					   "WFWorkerThread: optArg promise must resolve to nil or string"
			end

			return cachedLoadMessagingProcessorP( initScriptName, actualArg )

		end )
	
	else

		if optArg and type( optArg ) ~= 'string' then
			error( "WFWorkerThread: optArg must be nil, string, or promise", 2 )
		end
		
		self._loadP = cachedLoadMessagingProcessorP( initScriptName, optArg )

	end
	
	if optArg then
		self._optArg = optArg
	end

	self._loadP:whenResolvedDo( function( status, messagingProcessor )
	
		if status == 'data' and not WFCloseSupport_hasBeenOrWillBeClosed( self ) then
		
			self._messagingProcessor = messagingProcessor
			
			if initScriptName ~= ':self:' then

				if not refsForMessagingProcessor[ messagingProcessor ] then
					refsForMessagingProcessor[ messagingProcessor ] = KSTableUtils_makeFullyWeak()
				end
				
				refsForMessagingProcessor[ messagingProcessor ][ self ] = true
			
			end
		
		end
		
	end ):attachTo( self )

end

--------------------------------------------------------------------------------
--[[!
	\brief Closes this reference to this worker thread when it is no longer needed.
	
	When all references in this thread to this worker thread are closed, this
	thread releases its reference to the worker thread.
]]--

function workerThreadMethods:close()

	-- Close our event processor if this is the last reference to it.

	local messagingProcessor = self._messagingProcessor
	
	if messagingProcessor then
	
		local rfmp = refsForMessagingProcessor[ messagingProcessor ]

		if rfmp then

			rfmp[ self ] = nil

			if not next( rfmp ) then

				workerThreadLog:debug( "closing last reference to MP( %s / %s ) from %s",
									   self._initScriptName,
									   self._optArg or "",
									   ':traceback:' )

				messagingProcessor:close()
				refsForMessagingProcessor[ messagingProcessor ] = nil

			end

		end
	
	end

end

--------------------------------------------------------------------------------

local function assertNotClosed( self, methodName )

	if WFCloseSupport_hasBeenOrWillBeClosed( self ) then
		error( string.format( "WFWorkerThread:%s: thread %s %s has been closed\n\n"
							  .. "created in thread %s\n\n%s",
							  methodName, self._initScriptName,
							  self._optArg,
							  WFCurrentThread.DEBUG_getThreadName(),
							  self._creationTraceback ), 3 )
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief This promise notifies you when the worker thread has finished
	loading and is ready to accept requests.

	You can attach actions to this if you need specialized load success/failure
	notification. By default, the class logs a warning in the event
	of a load failure.
	
	\return Resolves to \c true when the worker thread
		loads successfully. If the worker thread fails to load, resolves with
		the error message from the thread.
]]--

function workerThreadMethods:threadIsReadyP()

	assertNotClosed( self, 'threadIsReadyP' )

	return self._loadP
		:mapDataP( function()
			return 'data', true
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Initiates a message request to this worker thread and returns the result
		through a promise.
	
	If the worker thread has not finished loading, the request waits until
	loading is complete. If the worker thread fails to load, the result of
	the promise is the failure message generated by the load attempt.
	
	\param message The name of the message to send to this worker thread.
	\param ... Other arguments to pass to the message.

	\return A \c KSPromise object with the result of the message.
]]--

function workerThreadMethods:requestP( --[[string]] message, ... )

	assertNotClosed( self, 'requestP' )
	
	local backtrace
	
	if CHECK_REQUEST_CLOSURE then
		backtrace = debug.traceback()
	end

	if self._messagingProcessor then
	
		local monitor = makeGCMonitor(
				'requestP',
				self._initScriptName,
				message,
				backtrace )

		return requestP( monitor, self._messagingProcessor, message, ... )

	end
	
	local n = select( '#', ... )
	local params = { ... }
	
	local resultP = self._loadP:mapDataP( function( status, messagingProcessor )
		
		local monitor = makeGCMonitor(
				'requestP',
				self._initScriptName,
				message,
				backtrace )

		return requestP( monitor, messagingProcessor, message, unpack( params, 1, n ) )
		
	end )
	
	return resultP

end

--------------------------------------------------------------------------------
--[[!
	\brief Initiates a request to this worker thread and returns the result
	or results through an event stream.
	
	If the worker thread has not finished loading, the request waits until
	loading is complete. If the worker thread fails to load, the
	failure message generated by the load attempt is posted as an error event
	on this stream, and the stream is then closed.
	
	\warning Do <b>NOT</b> use this function for requests that have side effects.
	 The lifetime of such requests is unpredictable and can led to subtle
	silent-failure bugs. Use \c :requestP() instead.
	
	\param message The name of the message to send to this worker thread.
	\param ... Other arguments to pass to the message.
	
	\return A \c KSEventStream object with the result or results of the message.
]]--

function workerThreadMethods:requestE( --[[string]] message, ... )

	assertNotClosed( self, 'requestE' )
	
	local backtrace
	
	if CHECK_REQUEST_CLOSURE then
		backtrace = debug.traceback()
	end

	local resultE

	if self._messagingProcessor then
	
		local monitor = makeGCMonitor( 'requestE', self._initScriptName, message, backtrace )
	
		resultE = requestE( monitor, self._messagingProcessor, message, ... )
	
	else
	
		local loadEvent = KSReactivePromiseBridge.promiseToEvent( self._loadP )
		
		local n = select( '#', ... )
		local params = { ... }

		resultE = loadEvent:mapSwitchE( function( messagingProcessor )

			local monitor = makeGCMonitor( 'requestE', self._initScriptName, message, backtrace )
	
			return requestE( monitor, messagingProcessor, message, unpack( params, 1, n ) )
		
		end )
	
	end
	
	return resultE

end

--------------------------------------------------------------------------------

local function collectArgs( t, field )
	local v = t[ field ]
	if type( v ) == 'table' then
		return v, v.n or #v
	elseif v == nil then
		return {}, 0
	else
		return { v }, 1
	end
end
--------------------------------------------------------------------------------

--[[!
	\brief Like requestE, but starts with an initial value.
	
	\param params A table of parameters
		* message : The name of the message to send to this worker thread.
		* arguments : (number, string, boolean, nil, kNilValue or array) Other arguments to pass,
			packed as an array.
			If it is a sparse array, include a field 'n' to indicate the total # of arguments.
			As a convenience, if it is a value (not a table) then it will be treated as the
			single argument (and wrapped in an array for you).
			If it is nil then no arguments are sent.  To use nil (rather than
			then empty tuple) use WFWorkerThread.kNilValue.
			To send a table argument, you *must* wrap it in an array, otherwise it will get unpacked.
		* startingValue : (number, string, boolean, nil, kNilValue or array) starting value. It is also handled as conveniently just
			like arguments
	
	\return A \c Behavior with the result or results of the message.
]]--
function workerThreadMethods:requestB( params )
	check( WFWorkerThread, 'table' )
	
	local message = params.message
	local arguments, nArgs = collectArgs( params, 'arguments' )
	local E = self:requestE( message, unpack( arguments, 1, nArgs ) )
	local startingValue, nVals = collectArgs( params, 'startingValue' )
	local B = E:closingHoldB( unpack( startingValue, 1, nVals ) )
	return B
end

--------------------------------------------------------------------------------

local WFWorkerThreadClass = KSClass 'WFWorkerThread'( workerThreadMethods )

rawset( WFWorkerThreadClass, 'kNilValue', { n = 1 } )

--------------------------------------------------------------------------------

return WFWorkerThreadClass
