--[[----------------------------------------------------------------------------

WFWorkerThreadTest.lua

@author scouten 2011-02-04

@reviewed erwright 2011-03-08
@reviewed mhamburg 2011-03-13

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.utils
local KSTestUtils = require 'KSTestUtils'


--============================================================================--

local WFWorkerThreadTest = {}

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-12-16]: Convert to WFAsyncTestCase. Note that this
-- requires implementation of the planned :onErrorE handling.

--------------------------------------------------------------------------------

local requestTestHandlerFailed = nil

local requestTestHandler = WFWorkerThread( 'WFRequestTestHandler.lua' )

local currentTest = nil
local expectedTestResults = nil

--------------------------------------------------------------------------------

function WFWorkerThreadTest.setUp()

	if requestTestHandlerFailed then
		error( requestTestHandlerFailed )
	end

end


--============================================================================--

-- @mark :requestP testing

--------------------------------------------------------------------------------

local function checkTestResultsAndNotify( test, ... )

	local n = select( '#', ... )
	local expectedResults = { ... }

	return function( ... )

		if pcall( function( ... )
				KSTestUtils.assertMatchArgs( unpack( expectedResults, 1, n ) )( ... )
			end, ... ) == true
		then
			test:notifyTestDidSucceed()
		else
			test:notifyTestDidFail()
		end

	end

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestP( test )

	local p = requestTestHandler:requestP( 'echo', 1, 2, 3 )

	p:whenResolvedDo( checkTestResultsAndNotify( test, 'data', 3, 1, 2, 3 ) ):runUntilDone()
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestPTransitWithNil( test )

	local p = requestTestHandler:requestP( 'echo', 1, nil, 3 )

	p:whenResolvedDo( checkTestResultsAndNotify( test, 'data', 3, 1, nil, 3 ) ):runUntilDone()
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestPWithData( test )

	local p = requestTestHandler:requestP( 'echo', "The data vaue" )

	p:whenResolvedDo( checkTestResultsAndNotify( test, 'data', 1, "The data vaue" ) ):runUntilDone()
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestPWithError( test )

	local p = requestTestHandler:requestP( 'sendError', "Expecting this error" )

	p:whenResolvedDo( checkTestResultsAndNotify( test, 'error',
												 "Expecting this error" ) )
			-- FIXME [scouten 2012-02-09]: Don't love hard-coding the line number in,
			-- but I don't have a better solution at this time.
		:runUntilDone()
	
	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestPCancelation( test )

	local p = requestTestHandler:requestP( 'echo', "Hello!" )
	
	local canceledB = KSReactive.variableB( false )

	p = p:observeP( function( ... ) canceledB:set( select( '#', ... ) == 0 ) end )
	
	local a1 = p:whenResolvedDo( function() end ):runUntilDone()
	
	a1:close()
	
	WFAsyncTestCase( test )
	
		:waitFor( canceledB )
		:continueWhenValueMatches( true )
	
end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestPCancelationWithDelay( test )

	local echoDelayP = requestTestHandler:requestP( 'echoDelay', 2, "Hello!" )
	
	local a1 = echoDelayP:whenResolvedDo( function() end ):runUntilDone()
	
	WFAsyncTestCase( test, 4 )
	
		:waitFor( .5 )		 	-- step 1
		
		:waitFor( function()	-- step 2
		
			a1:close()
			
			return echoDelayP
		
		end )
		
		:continueWhenValueMatches() -- no results on cancel
		
end


--============================================================================--

-- @mark :requestE testing

--------------------------------------------------------------------------------

local function testCallback( ... )

	local testResults = { ... }
	
	local success = true
	
	if testResults[ 1 ] ~= expectedTestResults.n then
		debug.print( "echo didn't respond with correct # of results" )
		success = false
	end
	
	for i = 1, expectedTestResults.n do
	
		if( testResults[ i + 1 ] ~= expectedTestResults[ i ] ) then
			debug.print( "unexpected result at index ", i, "  actual: ", tostring( testResults[ i + 1 ] ),
								"   expected: ", tostring( expectedTestResults[ i ] ) )
			success = false
		end
	
	end
	
	if success then
		currentTest:notifyTestDidSucceed()
	else
		currentTest:notifyTestDidFail()
	end

end

--------------------------------------------------------------------------------

local function testRequestFinishedWithResults( test, _, ... )

	currentTest = test

	expectedTestResults = { ... }
	expectedTestResults.n = select( '#', ... )

	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestE( test )

	testRequestFinishedWithResults( test,
		requestTestHandler:requestE( 'echo', 1, 2, 3 ):onDataE( testCallback ),
		1, 2, 3 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestETransitWithNil( test )

	testRequestFinishedWithResults( test,
		requestTestHandler:requestE( 'echo', 1, nil, 3 ):onDataE( testCallback ),
		1, nil, 3 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestEWithData( test )

	testRequestFinishedWithResults( test,
		requestTestHandler:requestE( 'echo', "The data value" )
			:onDataE( testCallback )
			:onErrorE( function() test:notifyTestDidFail() end ),
		"The data value" )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestEWithError( test )

	requestTestHandler:requestE( 'sendError', "expecting this error" )
		:onDataE( function() test:notifyTestDidFail() end )
		:onErrorE( function( message )
			if message == 'expecting this error' then
				test:notifyTestDidSucceed()
			else
				test:log( "wrong message: " .. message )
				test:notifyTestDidFail()
			end
		end )

	test:waitForTestToFinish( 2 )

end

--------------------------------------------------------------------------------

function WFWorkerThreadTest.testRequestEUpdates( test )

	local counterB = requestTestHandler:requestE( 'counterE', 4, 7 ):holdB( nil )
	
	WFAsyncTestCase( test, 4 )

		:waitFor( counterB )
		:continueWhenValueMatches( 4 )
		
		:waitFor( counterB )
		:continueWhenValueMatches( 5 )
		
		:waitFor( counterB )
		:continueWhenValueMatches( 6 )
		
		:waitFor( counterB )
		:continueWhenValueMatches( 7 )
		
		-- test for counterB closing?
		
end

--------------------------------------------------------------------------------

return WFWorkerThreadTest
